/* Automatically @generated by tree-sitter v0.25.4 (726dcd1e872149d95de581589fc408fb8ea9cb0b) */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 1161
#define LARGE_STATE_COUNT 166
#define SYMBOL_COUNT 149
#define ALIAS_COUNT 4
#define TOKEN_COUNT 67
#define EXTERNAL_TOKEN_COUNT 15
#define FIELD_COUNT 0
#define MAX_ALIAS_SEQUENCE_LENGTH 8
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 6
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym__backslash_escape = 1,
  sym_entity_reference = 2,
  sym_numeric_character_reference = 3,
  anon_sym_LBRACK = 4,
  anon_sym_RBRACK = 5,
  anon_sym_LT = 6,
  anon_sym_GT = 7,
  anon_sym_BANG = 8,
  anon_sym_DQUOTE = 9,
  anon_sym_POUND = 10,
  anon_sym_DOLLAR = 11,
  anon_sym_PERCENT = 12,
  anon_sym_AMP = 13,
  anon_sym_SQUOTE = 14,
  anon_sym_STAR = 15,
  anon_sym_PLUS = 16,
  anon_sym_COMMA = 17,
  anon_sym_DASH = 18,
  anon_sym_DOT = 19,
  anon_sym_SLASH = 20,
  anon_sym_COLON = 21,
  anon_sym_SEMI = 22,
  anon_sym_EQ = 23,
  anon_sym_QMARK = 24,
  anon_sym_AT = 25,
  anon_sym_BSLASH = 26,
  anon_sym_CARET = 27,
  anon_sym__ = 28,
  anon_sym_BQUOTE = 29,
  anon_sym_LBRACE = 30,
  anon_sym_PIPE = 31,
  anon_sym_RBRACE = 32,
  anon_sym_TILDE = 33,
  anon_sym_LPAREN = 34,
  anon_sym_RPAREN = 35,
  sym__newline_token = 36,
  sym_uri_autolink = 37,
  sym_email_autolink = 38,
  sym__attribute_name = 39,
  aux_sym__attribute_value_token1 = 40,
  anon_sym_LT_BANG_DASH_DASH = 41,
  anon_sym_DASH_DASH_GT = 42,
  anon_sym_LT_QMARK = 43,
  anon_sym_QMARK_GT = 44,
  aux_sym__declaration_token1 = 45,
  anon_sym_LT_BANG_LBRACKCDATA_LBRACK = 46,
  anon_sym_RBRACK_RBRACK_GT = 47,
  sym__whitespace_ge_2 = 48,
  aux_sym__whitespace_token1 = 49,
  sym__word_no_digit = 50,
  sym__digits = 51,
  sym__error = 52,
  sym__trigger_error = 53,
  sym__code_span_start = 54,
  sym__code_span_close = 55,
  sym__emphasis_open_star = 56,
  sym__emphasis_open_underscore = 57,
  sym__emphasis_close_star = 58,
  sym__emphasis_close_underscore = 59,
  sym__last_token_whitespace = 60,
  sym__last_token_punctuation = 61,
  sym__strikethrough_open = 62,
  sym__strikethrough_close = 63,
  sym__latex_span_start = 64,
  sym__latex_span_close = 65,
  sym__unclosed_span = 66,
  sym_inline = 67,
  sym_backslash_escape = 68,
  sym_link_label = 69,
  sym_link_destination = 70,
  sym__link_destination_parenthesis = 71,
  sym__text_no_angle = 72,
  sym_link_title = 73,
  sym_code_span = 74,
  sym_latex_block = 75,
  sym__link_text = 76,
  sym__link_text_non_empty = 77,
  sym_shortcut_link = 78,
  sym_full_reference_link = 79,
  sym_collapsed_reference_link = 80,
  sym_inline_link = 81,
  sym_image = 82,
  sym__image_inline_link = 83,
  sym__image_shortcut_link = 84,
  sym__image_full_reference_link = 85,
  sym__image_collapsed_reference_link = 86,
  sym__image_description = 87,
  sym__image_description_non_empty = 88,
  sym__html_tag = 89,
  sym__open_tag = 90,
  sym__closing_tag = 91,
  sym__tag_name = 92,
  sym__attribute = 93,
  sym__attribute_value = 94,
  sym__html_comment = 95,
  sym__processing_instruction = 96,
  sym__declaration = 97,
  sym__cdata_section = 98,
  sym_hard_line_break = 99,
  sym__whitespace = 100,
  sym__word = 101,
  sym__soft_line_break = 102,
  sym__inline_base = 103,
  sym__text_base = 104,
  sym__text_inline_no_link = 105,
  sym__inline_element = 106,
  aux_sym__inline = 107,
  sym__inline_element_no_star = 108,
  aux_sym__inline_no_star = 109,
  sym__inline_element_no_underscore = 110,
  aux_sym__inline_no_underscore = 111,
  sym__inline_element_no_tilde = 112,
  aux_sym__inline_no_tilde = 113,
  sym__strikethrough = 114,
  sym__emphasis_star = 115,
  sym__strong_emphasis_star = 116,
  sym__emphasis_underscore = 117,
  sym__strong_emphasis_underscore = 118,
  sym__inline_element_no_link = 119,
  aux_sym__inline_no_link = 120,
  sym__inline_element_no_star_no_link = 121,
  aux_sym__inline_no_star_no_link = 122,
  sym__inline_element_no_underscore_no_link = 123,
  aux_sym__inline_no_underscore_no_link = 124,
  sym__inline_element_no_tilde_no_link = 125,
  aux_sym__inline_no_tilde_no_link = 126,
  sym__strikethrough_no_link = 127,
  sym__emphasis_star_no_link = 128,
  sym__strong_emphasis_star_no_link = 129,
  sym__emphasis_underscore_no_link = 130,
  sym__strong_emphasis_underscore_no_link = 131,
  aux_sym_link_label_repeat1 = 132,
  aux_sym_link_destination_repeat1 = 133,
  aux_sym_link_destination_repeat2 = 134,
  aux_sym_link_title_repeat1 = 135,
  aux_sym_link_title_repeat2 = 136,
  aux_sym_link_title_repeat3 = 137,
  aux_sym_code_span_repeat1 = 138,
  aux_sym_latex_block_repeat1 = 139,
  aux_sym_inline_link_repeat1 = 140,
  aux_sym__open_tag_repeat1 = 141,
  aux_sym__tag_name_repeat1 = 142,
  aux_sym__attribute_value_repeat1 = 143,
  aux_sym__attribute_value_repeat2 = 144,
  aux_sym__html_comment_repeat1 = 145,
  aux_sym__processing_instruction_repeat1 = 146,
  aux_sym__declaration_repeat1 = 147,
  aux_sym__inline_base_repeat1 = 148,
  alias_sym_emphasis = 149,
  alias_sym_html_tag = 150,
  alias_sym_image_description = 151,
  alias_sym_link_text = 152,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__backslash_escape] = "_backslash_escape",
  [sym_entity_reference] = "entity_reference",
  [sym_numeric_character_reference] = "numeric_character_reference",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_BANG] = "!",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_POUND] = "#",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_PERCENT] = "%",
  [anon_sym_AMP] = "&",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_STAR] = "*",
  [anon_sym_PLUS] = "+",
  [anon_sym_COMMA] = ",",
  [anon_sym_DASH] = "-",
  [anon_sym_DOT] = ".",
  [anon_sym_SLASH] = "/",
  [anon_sym_COLON] = ":",
  [anon_sym_SEMI] = ";",
  [anon_sym_EQ] = "=",
  [anon_sym_QMARK] = "\?",
  [anon_sym_AT] = "@",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_CARET] = "^",
  [anon_sym__] = "_",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_LBRACE] = "{",
  [anon_sym_PIPE] = "|",
  [anon_sym_RBRACE] = "}",
  [anon_sym_TILDE] = "~",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [sym__newline_token] = "_newline_token",
  [sym_uri_autolink] = "uri_autolink",
  [sym_email_autolink] = "email_autolink",
  [sym__attribute_name] = "_attribute_name",
  [aux_sym__attribute_value_token1] = "_attribute_value_token1",
  [anon_sym_LT_BANG_DASH_DASH] = "<!--",
  [anon_sym_DASH_DASH_GT] = "-->",
  [anon_sym_LT_QMARK] = "<\?",
  [anon_sym_QMARK_GT] = "\?>",
  [aux_sym__declaration_token1] = "_declaration_token1",
  [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = "<![CDATA[",
  [anon_sym_RBRACK_RBRACK_GT] = "]]>",
  [sym__whitespace_ge_2] = "_whitespace_ge_2",
  [aux_sym__whitespace_token1] = "_whitespace_token1",
  [sym__word_no_digit] = "_word_no_digit",
  [sym__digits] = "_digits",
  [sym__error] = "_error",
  [sym__trigger_error] = "_trigger_error",
  [sym__code_span_start] = "code_span_delimiter",
  [sym__code_span_close] = "code_span_delimiter",
  [sym__emphasis_open_star] = "_emphasis_open_star",
  [sym__emphasis_open_underscore] = "_emphasis_open_underscore",
  [sym__emphasis_close_star] = "emphasis_delimiter",
  [sym__emphasis_close_underscore] = "emphasis_delimiter",
  [sym__last_token_whitespace] = "_last_token_whitespace",
  [sym__last_token_punctuation] = "_last_token_punctuation",
  [sym__strikethrough_open] = "_strikethrough_open",
  [sym__strikethrough_close] = "emphasis_delimiter",
  [sym__latex_span_start] = "latex_span_delimiter",
  [sym__latex_span_close] = "latex_span_delimiter",
  [sym__unclosed_span] = "_unclosed_span",
  [sym_inline] = "inline",
  [sym_backslash_escape] = "backslash_escape",
  [sym_link_label] = "link_label",
  [sym_link_destination] = "link_destination",
  [sym__link_destination_parenthesis] = "_link_destination_parenthesis",
  [sym__text_no_angle] = "_text_no_angle",
  [sym_link_title] = "link_title",
  [sym_code_span] = "code_span",
  [sym_latex_block] = "latex_block",
  [sym__link_text] = "_link_text",
  [sym__link_text_non_empty] = "_link_text_non_empty",
  [sym_shortcut_link] = "shortcut_link",
  [sym_full_reference_link] = "full_reference_link",
  [sym_collapsed_reference_link] = "collapsed_reference_link",
  [sym_inline_link] = "inline_link",
  [sym_image] = "image",
  [sym__image_inline_link] = "_image_inline_link",
  [sym__image_shortcut_link] = "_image_shortcut_link",
  [sym__image_full_reference_link] = "_image_full_reference_link",
  [sym__image_collapsed_reference_link] = "_image_collapsed_reference_link",
  [sym__image_description] = "_image_description",
  [sym__image_description_non_empty] = "_image_description_non_empty",
  [sym__html_tag] = "_html_tag",
  [sym__open_tag] = "_open_tag",
  [sym__closing_tag] = "_closing_tag",
  [sym__tag_name] = "_tag_name",
  [sym__attribute] = "_attribute",
  [sym__attribute_value] = "_attribute_value",
  [sym__html_comment] = "_html_comment",
  [sym__processing_instruction] = "_processing_instruction",
  [sym__declaration] = "_declaration",
  [sym__cdata_section] = "_cdata_section",
  [sym_hard_line_break] = "hard_line_break",
  [sym__whitespace] = "_whitespace",
  [sym__word] = "_word",
  [sym__soft_line_break] = "_soft_line_break",
  [sym__inline_base] = "_inline_base",
  [sym__text_base] = "_text_base",
  [sym__text_inline_no_link] = "_text_inline_no_link",
  [sym__inline_element] = "_inline_element",
  [aux_sym__inline] = "_inline",
  [sym__inline_element_no_star] = "_inline_element_no_star",
  [aux_sym__inline_no_star] = "_inline_no_star",
  [sym__inline_element_no_underscore] = "_inline_element_no_underscore",
  [aux_sym__inline_no_underscore] = "_inline_no_underscore",
  [sym__inline_element_no_tilde] = "_inline_element_no_tilde",
  [aux_sym__inline_no_tilde] = "_inline_no_tilde",
  [sym__strikethrough] = "strikethrough",
  [sym__emphasis_star] = "_emphasis_star",
  [sym__strong_emphasis_star] = "strong_emphasis",
  [sym__emphasis_underscore] = "_emphasis_underscore",
  [sym__strong_emphasis_underscore] = "strong_emphasis",
  [sym__inline_element_no_link] = "_inline_element_no_link",
  [aux_sym__inline_no_link] = "_inline_no_link",
  [sym__inline_element_no_star_no_link] = "_inline_element_no_star_no_link",
  [aux_sym__inline_no_star_no_link] = "_inline_no_star_no_link",
  [sym__inline_element_no_underscore_no_link] = "_inline_element_no_underscore_no_link",
  [aux_sym__inline_no_underscore_no_link] = "_inline_no_underscore_no_link",
  [sym__inline_element_no_tilde_no_link] = "_inline_element_no_tilde_no_link",
  [aux_sym__inline_no_tilde_no_link] = "_inline_no_tilde_no_link",
  [sym__strikethrough_no_link] = "strikethrough",
  [sym__emphasis_star_no_link] = "_emphasis_star_no_link",
  [sym__strong_emphasis_star_no_link] = "strong_emphasis",
  [sym__emphasis_underscore_no_link] = "_emphasis_underscore_no_link",
  [sym__strong_emphasis_underscore_no_link] = "strong_emphasis",
  [aux_sym_link_label_repeat1] = "link_label_repeat1",
  [aux_sym_link_destination_repeat1] = "link_destination_repeat1",
  [aux_sym_link_destination_repeat2] = "link_destination_repeat2",
  [aux_sym_link_title_repeat1] = "link_title_repeat1",
  [aux_sym_link_title_repeat2] = "link_title_repeat2",
  [aux_sym_link_title_repeat3] = "link_title_repeat3",
  [aux_sym_code_span_repeat1] = "code_span_repeat1",
  [aux_sym_latex_block_repeat1] = "latex_block_repeat1",
  [aux_sym_inline_link_repeat1] = "inline_link_repeat1",
  [aux_sym__open_tag_repeat1] = "_open_tag_repeat1",
  [aux_sym__tag_name_repeat1] = "_tag_name_repeat1",
  [aux_sym__attribute_value_repeat1] = "_attribute_value_repeat1",
  [aux_sym__attribute_value_repeat2] = "_attribute_value_repeat2",
  [aux_sym__html_comment_repeat1] = "_html_comment_repeat1",
  [aux_sym__processing_instruction_repeat1] = "_processing_instruction_repeat1",
  [aux_sym__declaration_repeat1] = "_declaration_repeat1",
  [aux_sym__inline_base_repeat1] = "_inline_base_repeat1",
  [alias_sym_emphasis] = "emphasis",
  [alias_sym_html_tag] = "html_tag",
  [alias_sym_image_description] = "image_description",
  [alias_sym_link_text] = "link_text",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__backslash_escape] = sym__backslash_escape,
  [sym_entity_reference] = sym_entity_reference,
  [sym_numeric_character_reference] = sym_numeric_character_reference,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym__] = anon_sym__,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [sym__newline_token] = sym__newline_token,
  [sym_uri_autolink] = sym_uri_autolink,
  [sym_email_autolink] = sym_email_autolink,
  [sym__attribute_name] = sym__attribute_name,
  [aux_sym__attribute_value_token1] = aux_sym__attribute_value_token1,
  [anon_sym_LT_BANG_DASH_DASH] = anon_sym_LT_BANG_DASH_DASH,
  [anon_sym_DASH_DASH_GT] = anon_sym_DASH_DASH_GT,
  [anon_sym_LT_QMARK] = anon_sym_LT_QMARK,
  [anon_sym_QMARK_GT] = anon_sym_QMARK_GT,
  [aux_sym__declaration_token1] = aux_sym__declaration_token1,
  [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
  [anon_sym_RBRACK_RBRACK_GT] = anon_sym_RBRACK_RBRACK_GT,
  [sym__whitespace_ge_2] = sym__whitespace_ge_2,
  [aux_sym__whitespace_token1] = aux_sym__whitespace_token1,
  [sym__word_no_digit] = sym__word_no_digit,
  [sym__digits] = sym__digits,
  [sym__error] = sym__error,
  [sym__trigger_error] = sym__trigger_error,
  [sym__code_span_start] = sym__code_span_start,
  [sym__code_span_close] = sym__code_span_start,
  [sym__emphasis_open_star] = sym__emphasis_open_star,
  [sym__emphasis_open_underscore] = sym__emphasis_open_underscore,
  [sym__emphasis_close_star] = sym__emphasis_close_star,
  [sym__emphasis_close_underscore] = sym__emphasis_close_star,
  [sym__last_token_whitespace] = sym__last_token_whitespace,
  [sym__last_token_punctuation] = sym__last_token_punctuation,
  [sym__strikethrough_open] = sym__strikethrough_open,
  [sym__strikethrough_close] = sym__emphasis_close_star,
  [sym__latex_span_start] = sym__latex_span_start,
  [sym__latex_span_close] = sym__latex_span_start,
  [sym__unclosed_span] = sym__unclosed_span,
  [sym_inline] = sym_inline,
  [sym_backslash_escape] = sym_backslash_escape,
  [sym_link_label] = sym_link_label,
  [sym_link_destination] = sym_link_destination,
  [sym__link_destination_parenthesis] = sym__link_destination_parenthesis,
  [sym__text_no_angle] = sym__text_no_angle,
  [sym_link_title] = sym_link_title,
  [sym_code_span] = sym_code_span,
  [sym_latex_block] = sym_latex_block,
  [sym__link_text] = sym__link_text,
  [sym__link_text_non_empty] = sym__link_text_non_empty,
  [sym_shortcut_link] = sym_shortcut_link,
  [sym_full_reference_link] = sym_full_reference_link,
  [sym_collapsed_reference_link] = sym_collapsed_reference_link,
  [sym_inline_link] = sym_inline_link,
  [sym_image] = sym_image,
  [sym__image_inline_link] = sym__image_inline_link,
  [sym__image_shortcut_link] = sym__image_shortcut_link,
  [sym__image_full_reference_link] = sym__image_full_reference_link,
  [sym__image_collapsed_reference_link] = sym__image_collapsed_reference_link,
  [sym__image_description] = sym__image_description,
  [sym__image_description_non_empty] = sym__image_description_non_empty,
  [sym__html_tag] = sym__html_tag,
  [sym__open_tag] = sym__open_tag,
  [sym__closing_tag] = sym__closing_tag,
  [sym__tag_name] = sym__tag_name,
  [sym__attribute] = sym__attribute,
  [sym__attribute_value] = sym__attribute_value,
  [sym__html_comment] = sym__html_comment,
  [sym__processing_instruction] = sym__processing_instruction,
  [sym__declaration] = sym__declaration,
  [sym__cdata_section] = sym__cdata_section,
  [sym_hard_line_break] = sym_hard_line_break,
  [sym__whitespace] = sym__whitespace,
  [sym__word] = sym__word,
  [sym__soft_line_break] = sym__soft_line_break,
  [sym__inline_base] = sym__inline_base,
  [sym__text_base] = sym__text_base,
  [sym__text_inline_no_link] = sym__text_inline_no_link,
  [sym__inline_element] = sym__inline_element,
  [aux_sym__inline] = aux_sym__inline,
  [sym__inline_element_no_star] = sym__inline_element_no_star,
  [aux_sym__inline_no_star] = aux_sym__inline_no_star,
  [sym__inline_element_no_underscore] = sym__inline_element_no_underscore,
  [aux_sym__inline_no_underscore] = aux_sym__inline_no_underscore,
  [sym__inline_element_no_tilde] = sym__inline_element_no_tilde,
  [aux_sym__inline_no_tilde] = aux_sym__inline_no_tilde,
  [sym__strikethrough] = sym__strikethrough,
  [sym__emphasis_star] = sym__emphasis_star,
  [sym__strong_emphasis_star] = sym__strong_emphasis_star,
  [sym__emphasis_underscore] = sym__emphasis_underscore,
  [sym__strong_emphasis_underscore] = sym__strong_emphasis_star,
  [sym__inline_element_no_link] = sym__inline_element_no_link,
  [aux_sym__inline_no_link] = aux_sym__inline_no_link,
  [sym__inline_element_no_star_no_link] = sym__inline_element_no_star_no_link,
  [aux_sym__inline_no_star_no_link] = aux_sym__inline_no_star_no_link,
  [sym__inline_element_no_underscore_no_link] = sym__inline_element_no_underscore_no_link,
  [aux_sym__inline_no_underscore_no_link] = aux_sym__inline_no_underscore_no_link,
  [sym__inline_element_no_tilde_no_link] = sym__inline_element_no_tilde_no_link,
  [aux_sym__inline_no_tilde_no_link] = aux_sym__inline_no_tilde_no_link,
  [sym__strikethrough_no_link] = sym__strikethrough,
  [sym__emphasis_star_no_link] = sym__emphasis_star_no_link,
  [sym__strong_emphasis_star_no_link] = sym__strong_emphasis_star,
  [sym__emphasis_underscore_no_link] = sym__emphasis_underscore_no_link,
  [sym__strong_emphasis_underscore_no_link] = sym__strong_emphasis_star,
  [aux_sym_link_label_repeat1] = aux_sym_link_label_repeat1,
  [aux_sym_link_destination_repeat1] = aux_sym_link_destination_repeat1,
  [aux_sym_link_destination_repeat2] = aux_sym_link_destination_repeat2,
  [aux_sym_link_title_repeat1] = aux_sym_link_title_repeat1,
  [aux_sym_link_title_repeat2] = aux_sym_link_title_repeat2,
  [aux_sym_link_title_repeat3] = aux_sym_link_title_repeat3,
  [aux_sym_code_span_repeat1] = aux_sym_code_span_repeat1,
  [aux_sym_latex_block_repeat1] = aux_sym_latex_block_repeat1,
  [aux_sym_inline_link_repeat1] = aux_sym_inline_link_repeat1,
  [aux_sym__open_tag_repeat1] = aux_sym__open_tag_repeat1,
  [aux_sym__tag_name_repeat1] = aux_sym__tag_name_repeat1,
  [aux_sym__attribute_value_repeat1] = aux_sym__attribute_value_repeat1,
  [aux_sym__attribute_value_repeat2] = aux_sym__attribute_value_repeat2,
  [aux_sym__html_comment_repeat1] = aux_sym__html_comment_repeat1,
  [aux_sym__processing_instruction_repeat1] = aux_sym__processing_instruction_repeat1,
  [aux_sym__declaration_repeat1] = aux_sym__declaration_repeat1,
  [aux_sym__inline_base_repeat1] = aux_sym__inline_base_repeat1,
  [alias_sym_emphasis] = alias_sym_emphasis,
  [alias_sym_html_tag] = alias_sym_html_tag,
  [alias_sym_image_description] = alias_sym_image_description,
  [alias_sym_link_text] = alias_sym_link_text,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__backslash_escape] = {
    .visible = false,
    .named = true,
  },
  [sym_entity_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_numeric_character_reference] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [sym__newline_token] = {
    .visible = false,
    .named = true,
  },
  [sym_uri_autolink] = {
    .visible = true,
    .named = true,
  },
  [sym_email_autolink] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute_name] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__attribute_value_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_BANG_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__declaration_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK_RBRACK_GT] = {
    .visible = true,
    .named = false,
  },
  [sym__whitespace_ge_2] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__whitespace_token1] = {
    .visible = false,
    .named = false,
  },
  [sym__word_no_digit] = {
    .visible = false,
    .named = true,
  },
  [sym__digits] = {
    .visible = false,
    .named = true,
  },
  [sym__error] = {
    .visible = false,
    .named = true,
  },
  [sym__trigger_error] = {
    .visible = false,
    .named = true,
  },
  [sym__code_span_start] = {
    .visible = true,
    .named = true,
  },
  [sym__code_span_close] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_open_star] = {
    .visible = false,
    .named = true,
  },
  [sym__emphasis_open_underscore] = {
    .visible = false,
    .named = true,
  },
  [sym__emphasis_close_star] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_close_underscore] = {
    .visible = true,
    .named = true,
  },
  [sym__last_token_whitespace] = {
    .visible = false,
    .named = true,
  },
  [sym__last_token_punctuation] = {
    .visible = false,
    .named = true,
  },
  [sym__strikethrough_open] = {
    .visible = false,
    .named = true,
  },
  [sym__strikethrough_close] = {
    .visible = true,
    .named = true,
  },
  [sym__latex_span_start] = {
    .visible = true,
    .named = true,
  },
  [sym__latex_span_close] = {
    .visible = true,
    .named = true,
  },
  [sym__unclosed_span] = {
    .visible = false,
    .named = true,
  },
  [sym_inline] = {
    .visible = true,
    .named = true,
  },
  [sym_backslash_escape] = {
    .visible = true,
    .named = true,
  },
  [sym_link_label] = {
    .visible = true,
    .named = true,
  },
  [sym_link_destination] = {
    .visible = true,
    .named = true,
  },
  [sym__link_destination_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym__text_no_angle] = {
    .visible = false,
    .named = true,
  },
  [sym_link_title] = {
    .visible = true,
    .named = true,
  },
  [sym_code_span] = {
    .visible = true,
    .named = true,
  },
  [sym_latex_block] = {
    .visible = true,
    .named = true,
  },
  [sym__link_text] = {
    .visible = false,
    .named = true,
  },
  [sym__link_text_non_empty] = {
    .visible = false,
    .named = true,
  },
  [sym_shortcut_link] = {
    .visible = true,
    .named = true,
  },
  [sym_full_reference_link] = {
    .visible = true,
    .named = true,
  },
  [sym_collapsed_reference_link] = {
    .visible = true,
    .named = true,
  },
  [sym_inline_link] = {
    .visible = true,
    .named = true,
  },
  [sym_image] = {
    .visible = true,
    .named = true,
  },
  [sym__image_inline_link] = {
    .visible = false,
    .named = true,
  },
  [sym__image_shortcut_link] = {
    .visible = false,
    .named = true,
  },
  [sym__image_full_reference_link] = {
    .visible = false,
    .named = true,
  },
  [sym__image_collapsed_reference_link] = {
    .visible = false,
    .named = true,
  },
  [sym__image_description] = {
    .visible = false,
    .named = true,
  },
  [sym__image_description_non_empty] = {
    .visible = false,
    .named = true,
  },
  [sym__html_tag] = {
    .visible = false,
    .named = true,
  },
  [sym__open_tag] = {
    .visible = false,
    .named = true,
  },
  [sym__closing_tag] = {
    .visible = false,
    .named = true,
  },
  [sym__tag_name] = {
    .visible = false,
    .named = true,
  },
  [sym__attribute] = {
    .visible = false,
    .named = true,
  },
  [sym__attribute_value] = {
    .visible = false,
    .named = true,
  },
  [sym__html_comment] = {
    .visible = false,
    .named = true,
  },
  [sym__processing_instruction] = {
    .visible = false,
    .named = true,
  },
  [sym__declaration] = {
    .visible = false,
    .named = true,
  },
  [sym__cdata_section] = {
    .visible = false,
    .named = true,
  },
  [sym_hard_line_break] = {
    .visible = true,
    .named = true,
  },
  [sym__whitespace] = {
    .visible = false,
    .named = true,
  },
  [sym__word] = {
    .visible = false,
    .named = true,
  },
  [sym__soft_line_break] = {
    .visible = false,
    .named = true,
  },
  [sym__inline_base] = {
    .visible = false,
    .named = true,
  },
  [sym__text_base] = {
    .visible = false,
    .named = true,
  },
  [sym__text_inline_no_link] = {
    .visible = false,
    .named = true,
  },
  [sym__inline_element] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_star] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_star] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_underscore] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_underscore] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_tilde] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_tilde] = {
    .visible = false,
    .named = false,
  },
  [sym__strikethrough] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_star] = {
    .visible = false,
    .named = true,
  },
  [sym__strong_emphasis_star] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_underscore] = {
    .visible = false,
    .named = true,
  },
  [sym__strong_emphasis_underscore] = {
    .visible = true,
    .named = true,
  },
  [sym__inline_element_no_link] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_link] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_star_no_link] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_star_no_link] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_underscore_no_link] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_underscore_no_link] = {
    .visible = false,
    .named = false,
  },
  [sym__inline_element_no_tilde_no_link] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__inline_no_tilde_no_link] = {
    .visible = false,
    .named = false,
  },
  [sym__strikethrough_no_link] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_star_no_link] = {
    .visible = false,
    .named = true,
  },
  [sym__strong_emphasis_star_no_link] = {
    .visible = true,
    .named = true,
  },
  [sym__emphasis_underscore_no_link] = {
    .visible = false,
    .named = true,
  },
  [sym__strong_emphasis_underscore_no_link] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_link_label_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_destination_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_destination_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_code_span_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_latex_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_inline_link_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__open_tag_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__tag_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__attribute_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__attribute_value_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_comment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__processing_instruction_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__inline_base_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_emphasis] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_html_tag] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_image_description] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_link_text] = {
    .visible = true,
    .named = true,
  },
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_html_tag,
  },
  [2] = {
    [0] = alias_sym_emphasis,
  },
  [3] = {
    [0] = sym__emphasis_close_star,
  },
  [4] = {
    [1] = alias_sym_link_text,
  },
  [5] = {
    [2] = alias_sym_image_description,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__html_tag, 2,
    sym__html_tag,
    alias_sym_html_tag,
  aux_sym__inline, 2,
    aux_sym__inline,
    alias_sym_image_description,
  sym__emphasis_star, 2,
    sym__emphasis_star,
    alias_sym_emphasis,
  sym__emphasis_underscore, 2,
    sym__emphasis_underscore,
    alias_sym_emphasis,
  aux_sym__inline_no_link, 2,
    aux_sym__inline_no_link,
    alias_sym_link_text,
  sym__emphasis_star_no_link, 2,
    sym__emphasis_star_no_link,
    alias_sym_emphasis,
  sym__emphasis_underscore_no_link, 2,
    sym__emphasis_underscore_no_link,
    alias_sym_emphasis,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 3,
  [26] = 2,
  [27] = 6,
  [28] = 11,
  [29] = 11,
  [30] = 30,
  [31] = 17,
  [32] = 19,
  [33] = 21,
  [34] = 23,
  [35] = 30,
  [36] = 17,
  [37] = 19,
  [38] = 21,
  [39] = 23,
  [40] = 11,
  [41] = 30,
  [42] = 17,
  [43] = 19,
  [44] = 21,
  [45] = 23,
  [46] = 11,
  [47] = 30,
  [48] = 17,
  [49] = 19,
  [50] = 21,
  [51] = 23,
  [52] = 30,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 57,
  [59] = 53,
  [60] = 60,
  [61] = 57,
  [62] = 54,
  [63] = 18,
  [64] = 57,
  [65] = 57,
  [66] = 57,
  [67] = 57,
  [68] = 60,
  [69] = 57,
  [70] = 56,
  [71] = 54,
  [72] = 60,
  [73] = 53,
  [74] = 56,
  [75] = 54,
  [76] = 60,
  [77] = 53,
  [78] = 56,
  [79] = 54,
  [80] = 60,
  [81] = 53,
  [82] = 56,
  [83] = 54,
  [84] = 54,
  [85] = 54,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 88,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 109,
  [112] = 112,
  [113] = 101,
  [114] = 103,
  [115] = 105,
  [116] = 107,
  [117] = 112,
  [118] = 110,
  [119] = 109,
  [120] = 112,
  [121] = 101,
  [122] = 103,
  [123] = 105,
  [124] = 107,
  [125] = 110,
  [126] = 109,
  [127] = 112,
  [128] = 101,
  [129] = 103,
  [130] = 105,
  [131] = 107,
  [132] = 110,
  [133] = 133,
  [134] = 110,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 137,
  [139] = 139,
  [140] = 139,
  [141] = 139,
  [142] = 136,
  [143] = 137,
  [144] = 137,
  [145] = 136,
  [146] = 136,
  [147] = 139,
  [148] = 148,
  [149] = 149,
  [150] = 148,
  [151] = 149,
  [152] = 149,
  [153] = 149,
  [154] = 148,
  [155] = 148,
  [156] = 148,
  [157] = 149,
  [158] = 149,
  [159] = 148,
  [160] = 149,
  [161] = 149,
  [162] = 148,
  [163] = 148,
  [164] = 149,
  [165] = 148,
  [166] = 166,
  [167] = 166,
  [168] = 168,
  [169] = 166,
  [170] = 168,
  [171] = 166,
  [172] = 168,
  [173] = 166,
  [174] = 166,
  [175] = 168,
  [176] = 168,
  [177] = 168,
  [178] = 166,
  [179] = 168,
  [180] = 180,
  [181] = 168,
  [182] = 168,
  [183] = 180,
  [184] = 180,
  [185] = 180,
  [186] = 180,
  [187] = 180,
  [188] = 166,
  [189] = 180,
  [190] = 166,
  [191] = 191,
  [192] = 180,
  [193] = 193,
  [194] = 191,
  [195] = 193,
  [196] = 191,
  [197] = 191,
  [198] = 193,
  [199] = 180,
  [200] = 200,
  [201] = 191,
  [202] = 193,
  [203] = 193,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 205,
  [208] = 204,
  [209] = 204,
  [210] = 204,
  [211] = 206,
  [212] = 212,
  [213] = 205,
  [214] = 206,
  [215] = 212,
  [216] = 205,
  [217] = 212,
  [218] = 212,
  [219] = 206,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 220,
  [225] = 225,
  [226] = 226,
  [227] = 222,
  [228] = 220,
  [229] = 229,
  [230] = 226,
  [231] = 220,
  [232] = 221,
  [233] = 233,
  [234] = 234,
  [235] = 223,
  [236] = 236,
  [237] = 229,
  [238] = 222,
  [239] = 223,
  [240] = 226,
  [241] = 225,
  [242] = 220,
  [243] = 205,
  [244] = 222,
  [245] = 225,
  [246] = 226,
  [247] = 222,
  [248] = 221,
  [249] = 212,
  [250] = 225,
  [251] = 222,
  [252] = 223,
  [253] = 204,
  [254] = 229,
  [255] = 206,
  [256] = 229,
  [257] = 257,
  [258] = 221,
  [259] = 222,
  [260] = 260,
  [261] = 226,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 262,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 263,
  [283] = 264,
  [284] = 265,
  [285] = 266,
  [286] = 262,
  [287] = 268,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 269,
  [294] = 294,
  [295] = 295,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 270,
  [306] = 306,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 310,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 316,
  [317] = 317,
  [318] = 318,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 271,
  [329] = 329,
  [330] = 277,
  [331] = 278,
  [332] = 280,
  [333] = 281,
  [334] = 263,
  [335] = 264,
  [336] = 265,
  [337] = 266,
  [338] = 272,
  [339] = 268,
  [340] = 269,
  [341] = 270,
  [342] = 271,
  [343] = 272,
  [344] = 273,
  [345] = 274,
  [346] = 275,
  [347] = 273,
  [348] = 274,
  [349] = 223,
  [350] = 275,
  [351] = 314,
  [352] = 352,
  [353] = 276,
  [354] = 276,
  [355] = 315,
  [356] = 234,
  [357] = 316,
  [358] = 317,
  [359] = 292,
  [360] = 318,
  [361] = 296,
  [362] = 298,
  [363] = 299,
  [364] = 300,
  [365] = 301,
  [366] = 302,
  [367] = 288,
  [368] = 314,
  [369] = 315,
  [370] = 316,
  [371] = 317,
  [372] = 318,
  [373] = 326,
  [374] = 327,
  [375] = 222,
  [376] = 329,
  [377] = 277,
  [378] = 263,
  [379] = 264,
  [380] = 265,
  [381] = 268,
  [382] = 269,
  [383] = 223,
  [384] = 289,
  [385] = 290,
  [386] = 291,
  [387] = 319,
  [388] = 221,
  [389] = 223,
  [390] = 225,
  [391] = 294,
  [392] = 295,
  [393] = 297,
  [394] = 303,
  [395] = 229,
  [396] = 304,
  [397] = 397,
  [398] = 320,
  [399] = 223,
  [400] = 306,
  [401] = 307,
  [402] = 308,
  [403] = 309,
  [404] = 310,
  [405] = 311,
  [406] = 223,
  [407] = 312,
  [408] = 313,
  [409] = 319,
  [410] = 320,
  [411] = 321,
  [412] = 322,
  [413] = 323,
  [414] = 414,
  [415] = 325,
  [416] = 321,
  [417] = 278,
  [418] = 280,
  [419] = 281,
  [420] = 266,
  [421] = 262,
  [422] = 422,
  [423] = 422,
  [424] = 424,
  [425] = 425,
  [426] = 270,
  [427] = 271,
  [428] = 428,
  [429] = 429,
  [430] = 272,
  [431] = 273,
  [432] = 274,
  [433] = 275,
  [434] = 276,
  [435] = 322,
  [436] = 323,
  [437] = 397,
  [438] = 324,
  [439] = 325,
  [440] = 312,
  [441] = 310,
  [442] = 294,
  [443] = 443,
  [444] = 311,
  [445] = 422,
  [446] = 313,
  [447] = 424,
  [448] = 326,
  [449] = 288,
  [450] = 289,
  [451] = 428,
  [452] = 429,
  [453] = 453,
  [454] = 424,
  [455] = 290,
  [456] = 291,
  [457] = 292,
  [458] = 327,
  [459] = 294,
  [460] = 397,
  [461] = 295,
  [462] = 296,
  [463] = 297,
  [464] = 298,
  [465] = 299,
  [466] = 300,
  [467] = 301,
  [468] = 422,
  [469] = 288,
  [470] = 424,
  [471] = 289,
  [472] = 302,
  [473] = 303,
  [474] = 428,
  [475] = 429,
  [476] = 304,
  [477] = 309,
  [478] = 290,
  [479] = 306,
  [480] = 307,
  [481] = 308,
  [482] = 482,
  [483] = 397,
  [484] = 309,
  [485] = 310,
  [486] = 311,
  [487] = 312,
  [488] = 313,
  [489] = 314,
  [490] = 295,
  [491] = 422,
  [492] = 315,
  [493] = 424,
  [494] = 316,
  [495] = 317,
  [496] = 318,
  [497] = 428,
  [498] = 429,
  [499] = 319,
  [500] = 500,
  [501] = 320,
  [502] = 321,
  [503] = 397,
  [504] = 296,
  [505] = 291,
  [506] = 297,
  [507] = 298,
  [508] = 299,
  [509] = 300,
  [510] = 301,
  [511] = 322,
  [512] = 323,
  [513] = 324,
  [514] = 302,
  [515] = 428,
  [516] = 303,
  [517] = 429,
  [518] = 325,
  [519] = 326,
  [520] = 327,
  [521] = 304,
  [522] = 292,
  [523] = 329,
  [524] = 306,
  [525] = 329,
  [526] = 277,
  [527] = 278,
  [528] = 280,
  [529] = 529,
  [530] = 530,
  [531] = 307,
  [532] = 281,
  [533] = 308,
  [534] = 534,
  [535] = 324,
  [536] = 290,
  [537] = 529,
  [538] = 534,
  [539] = 500,
  [540] = 288,
  [541] = 289,
  [542] = 291,
  [543] = 294,
  [544] = 295,
  [545] = 297,
  [546] = 303,
  [547] = 304,
  [548] = 306,
  [549] = 307,
  [550] = 308,
  [551] = 309,
  [552] = 310,
  [553] = 311,
  [554] = 312,
  [555] = 313,
  [556] = 319,
  [557] = 320,
  [558] = 321,
  [559] = 322,
  [560] = 323,
  [561] = 324,
  [562] = 325,
  [563] = 278,
  [564] = 290,
  [565] = 281,
  [566] = 266,
  [567] = 262,
  [568] = 270,
  [569] = 271,
  [570] = 272,
  [571] = 273,
  [572] = 274,
  [573] = 275,
  [574] = 276,
  [575] = 290,
  [576] = 290,
  [577] = 292,
  [578] = 296,
  [579] = 298,
  [580] = 299,
  [581] = 300,
  [582] = 301,
  [583] = 302,
  [584] = 314,
  [585] = 315,
  [586] = 316,
  [587] = 317,
  [588] = 318,
  [589] = 326,
  [590] = 327,
  [591] = 329,
  [592] = 277,
  [593] = 263,
  [594] = 264,
  [595] = 265,
  [596] = 268,
  [597] = 269,
  [598] = 290,
  [599] = 280,
  [600] = 600,
  [601] = 601,
  [602] = 602,
  [603] = 603,
  [604] = 604,
  [605] = 605,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 609,
  [610] = 610,
  [611] = 604,
  [612] = 607,
  [613] = 608,
  [614] = 609,
  [615] = 607,
  [616] = 608,
  [617] = 609,
  [618] = 610,
  [619] = 604,
  [620] = 603,
  [621] = 621,
  [622] = 622,
  [623] = 601,
  [624] = 602,
  [625] = 603,
  [626] = 626,
  [627] = 621,
  [628] = 626,
  [629] = 621,
  [630] = 626,
  [631] = 601,
  [632] = 602,
  [633] = 603,
  [634] = 607,
  [635] = 608,
  [636] = 609,
  [637] = 610,
  [638] = 604,
  [639] = 621,
  [640] = 626,
  [641] = 601,
  [642] = 642,
  [643] = 602,
  [644] = 610,
  [645] = 645,
  [646] = 646,
  [647] = 647,
  [648] = 648,
  [649] = 168,
  [650] = 647,
  [651] = 645,
  [652] = 646,
  [653] = 647,
  [654] = 648,
  [655] = 645,
  [656] = 646,
  [657] = 645,
  [658] = 658,
  [659] = 647,
  [660] = 648,
  [661] = 166,
  [662] = 646,
  [663] = 646,
  [664] = 648,
  [665] = 647,
  [666] = 648,
  [667] = 645,
  [668] = 668,
  [669] = 168,
  [670] = 225,
  [671] = 229,
  [672] = 672,
  [673] = 180,
  [674] = 166,
  [675] = 221,
  [676] = 298,
  [677] = 276,
  [678] = 302,
  [679] = 180,
  [680] = 296,
  [681] = 681,
  [682] = 205,
  [683] = 204,
  [684] = 684,
  [685] = 685,
  [686] = 686,
  [687] = 687,
  [688] = 688,
  [689] = 689,
  [690] = 690,
  [691] = 691,
  [692] = 225,
  [693] = 693,
  [694] = 694,
  [695] = 229,
  [696] = 696,
  [697] = 697,
  [698] = 698,
  [699] = 699,
  [700] = 204,
  [701] = 205,
  [702] = 702,
  [703] = 703,
  [704] = 221,
  [705] = 276,
  [706] = 706,
  [707] = 263,
  [708] = 264,
  [709] = 268,
  [710] = 710,
  [711] = 269,
  [712] = 710,
  [713] = 713,
  [714] = 229,
  [715] = 706,
  [716] = 706,
  [717] = 717,
  [718] = 710,
  [719] = 719,
  [720] = 706,
  [721] = 225,
  [722] = 292,
  [723] = 221,
  [724] = 296,
  [725] = 710,
  [726] = 298,
  [727] = 299,
  [728] = 300,
  [729] = 301,
  [730] = 302,
  [731] = 314,
  [732] = 706,
  [733] = 315,
  [734] = 316,
  [735] = 317,
  [736] = 318,
  [737] = 737,
  [738] = 326,
  [739] = 327,
  [740] = 710,
  [741] = 329,
  [742] = 706,
  [743] = 277,
  [744] = 710,
  [745] = 706,
  [746] = 710,
  [747] = 265,
  [748] = 296,
  [749] = 302,
  [750] = 316,
  [751] = 751,
  [752] = 317,
  [753] = 753,
  [754] = 754,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 318,
  [759] = 751,
  [760] = 299,
  [761] = 329,
  [762] = 756,
  [763] = 763,
  [764] = 764,
  [765] = 757,
  [766] = 766,
  [767] = 763,
  [768] = 755,
  [769] = 764,
  [770] = 326,
  [771] = 327,
  [772] = 300,
  [773] = 756,
  [774] = 766,
  [775] = 753,
  [776] = 754,
  [777] = 755,
  [778] = 277,
  [779] = 292,
  [780] = 764,
  [781] = 753,
  [782] = 754,
  [783] = 755,
  [784] = 757,
  [785] = 763,
  [786] = 764,
  [787] = 757,
  [788] = 754,
  [789] = 764,
  [790] = 757,
  [791] = 791,
  [792] = 263,
  [793] = 264,
  [794] = 265,
  [795] = 756,
  [796] = 766,
  [797] = 314,
  [798] = 315,
  [799] = 268,
  [800] = 269,
  [801] = 753,
  [802] = 754,
  [803] = 755,
  [804] = 301,
  [805] = 763,
  [806] = 764,
  [807] = 757,
  [808] = 753,
  [809] = 756,
  [810] = 766,
  [811] = 756,
  [812] = 753,
  [813] = 754,
  [814] = 755,
  [815] = 766,
  [816] = 763,
  [817] = 764,
  [818] = 757,
  [819] = 756,
  [820] = 766,
  [821] = 298,
  [822] = 753,
  [823] = 754,
  [824] = 755,
  [825] = 763,
  [826] = 766,
  [827] = 763,
  [828] = 828,
  [829] = 829,
  [830] = 830,
  [831] = 831,
  [832] = 832,
  [833] = 833,
  [834] = 828,
  [835] = 833,
  [836] = 833,
  [837] = 225,
  [838] = 229,
  [839] = 828,
  [840] = 840,
  [841] = 841,
  [842] = 842,
  [843] = 833,
  [844] = 844,
  [845] = 845,
  [846] = 833,
  [847] = 828,
  [848] = 848,
  [849] = 849,
  [850] = 850,
  [851] = 851,
  [852] = 833,
  [853] = 853,
  [854] = 854,
  [855] = 828,
  [856] = 833,
  [857] = 828,
  [858] = 858,
  [859] = 859,
  [860] = 860,
  [861] = 828,
  [862] = 862,
  [863] = 863,
  [864] = 864,
  [865] = 865,
  [866] = 866,
  [867] = 298,
  [868] = 302,
  [869] = 869,
  [870] = 870,
  [871] = 871,
  [872] = 872,
  [873] = 873,
  [874] = 874,
  [875] = 875,
  [876] = 876,
  [877] = 276,
  [878] = 878,
  [879] = 879,
  [880] = 880,
  [881] = 229,
  [882] = 225,
  [883] = 229,
  [884] = 884,
  [885] = 885,
  [886] = 886,
  [887] = 885,
  [888] = 229,
  [889] = 889,
  [890] = 890,
  [891] = 225,
  [892] = 892,
  [893] = 893,
  [894] = 225,
  [895] = 895,
  [896] = 896,
  [897] = 897,
  [898] = 298,
  [899] = 302,
  [900] = 298,
  [901] = 901,
  [902] = 902,
  [903] = 302,
  [904] = 229,
  [905] = 905,
  [906] = 298,
  [907] = 302,
  [908] = 905,
  [909] = 225,
  [910] = 910,
  [911] = 302,
  [912] = 912,
  [913] = 913,
  [914] = 914,
  [915] = 298,
  [916] = 916,
  [917] = 916,
  [918] = 916,
  [919] = 916,
  [920] = 916,
  [921] = 916,
  [922] = 916,
  [923] = 717,
  [924] = 924,
  [925] = 925,
  [926] = 925,
  [927] = 927,
  [928] = 925,
  [929] = 924,
  [930] = 925,
  [931] = 931,
  [932] = 927,
  [933] = 924,
  [934] = 924,
  [935] = 927,
  [936] = 925,
  [937] = 924,
  [938] = 931,
  [939] = 927,
  [940] = 931,
  [941] = 931,
  [942] = 931,
  [943] = 927,
  [944] = 944,
  [945] = 945,
  [946] = 945,
  [947] = 944,
  [948] = 945,
  [949] = 944,
  [950] = 950,
  [951] = 951,
  [952] = 945,
  [953] = 953,
  [954] = 954,
  [955] = 955,
  [956] = 945,
  [957] = 944,
  [958] = 944,
  [959] = 945,
  [960] = 944,
  [961] = 944,
  [962] = 945,
  [963] = 963,
  [964] = 964,
  [965] = 964,
  [966] = 964,
  [967] = 967,
  [968] = 968,
  [969] = 717,
  [970] = 963,
  [971] = 964,
  [972] = 963,
  [973] = 973,
  [974] = 963,
  [975] = 964,
  [976] = 963,
  [977] = 963,
  [978] = 963,
  [979] = 964,
  [980] = 964,
  [981] = 717,
  [982] = 982,
  [983] = 983,
  [984] = 984,
  [985] = 985,
  [986] = 986,
  [987] = 987,
  [988] = 988,
  [989] = 989,
  [990] = 990,
  [991] = 984,
  [992] = 992,
  [993] = 993,
  [994] = 992,
  [995] = 995,
  [996] = 996,
  [997] = 982,
  [998] = 998,
  [999] = 999,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 1002,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 993,
  [1006] = 1002,
  [1007] = 986,
  [1008] = 995,
  [1009] = 986,
  [1010] = 983,
  [1011] = 987,
  [1012] = 988,
  [1013] = 1013,
  [1014] = 984,
  [1015] = 992,
  [1016] = 996,
  [1017] = 1017,
  [1018] = 993,
  [1019] = 987,
  [1020] = 995,
  [1021] = 996,
  [1022] = 982,
  [1023] = 998,
  [1024] = 990,
  [1025] = 983,
  [1026] = 999,
  [1027] = 985,
  [1028] = 989,
  [1029] = 1000,
  [1030] = 1004,
  [1031] = 1001,
  [1032] = 1002,
  [1033] = 1003,
  [1034] = 1004,
  [1035] = 986,
  [1036] = 987,
  [1037] = 988,
  [1038] = 998,
  [1039] = 1013,
  [1040] = 984,
  [1041] = 992,
  [1042] = 993,
  [1043] = 995,
  [1044] = 996,
  [1045] = 1045,
  [1046] = 982,
  [1047] = 998,
  [1048] = 999,
  [1049] = 1000,
  [1050] = 1001,
  [1051] = 1002,
  [1052] = 1003,
  [1053] = 1053,
  [1054] = 1054,
  [1055] = 1017,
  [1056] = 986,
  [1057] = 1017,
  [1058] = 1017,
  [1059] = 225,
  [1060] = 229,
  [1061] = 999,
  [1062] = 987,
  [1063] = 1004,
  [1064] = 1003,
  [1065] = 988,
  [1066] = 1000,
  [1067] = 1001,
  [1068] = 1002,
  [1069] = 990,
  [1070] = 1013,
  [1071] = 983,
  [1072] = 1017,
  [1073] = 984,
  [1074] = 1017,
  [1075] = 985,
  [1076] = 989,
  [1077] = 1004,
  [1078] = 1003,
  [1079] = 988,
  [1080] = 992,
  [1081] = 993,
  [1082] = 229,
  [1083] = 995,
  [1084] = 1017,
  [1085] = 996,
  [1086] = 985,
  [1087] = 982,
  [1088] = 225,
  [1089] = 1004,
  [1090] = 1013,
  [1091] = 990,
  [1092] = 983,
  [1093] = 990,
  [1094] = 985,
  [1095] = 998,
  [1096] = 999,
  [1097] = 989,
  [1098] = 1000,
  [1099] = 1001,
  [1100] = 989,
  [1101] = 1004,
  [1102] = 1013,
  [1103] = 298,
  [1104] = 302,
  [1105] = 302,
  [1106] = 298,
  [1107] = 1107,
  [1108] = 1108,
  [1109] = 1109,
  [1110] = 1110,
  [1111] = 1111,
  [1112] = 1112,
  [1113] = 1113,
  [1114] = 1114,
  [1115] = 1115,
  [1116] = 1116,
  [1117] = 1116,
  [1118] = 1118,
  [1119] = 1118,
  [1120] = 1116,
  [1121] = 1116,
  [1122] = 1118,
  [1123] = 1118,
  [1124] = 1116,
  [1125] = 1118,
  [1126] = 225,
  [1127] = 1127,
  [1128] = 1127,
  [1129] = 1127,
  [1130] = 1130,
  [1131] = 1127,
  [1132] = 1127,
  [1133] = 1127,
  [1134] = 1127,
  [1135] = 1135,
  [1136] = 1135,
  [1137] = 1137,
  [1138] = 1138,
  [1139] = 1138,
  [1140] = 1137,
  [1141] = 1141,
  [1142] = 1138,
  [1143] = 1138,
  [1144] = 1138,
  [1145] = 1145,
  [1146] = 1135,
  [1147] = 1135,
  [1148] = 1148,
  [1149] = 1137,
  [1150] = 1137,
  [1151] = 1151,
  [1152] = 1137,
  [1153] = 1138,
  [1154] = 298,
  [1155] = 1138,
  [1156] = 1135,
  [1157] = 1135,
  [1158] = 1137,
  [1159] = 1135,
  [1160] = 1137,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(2183);
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2206,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2191,
        '=', 2211,
        '>', 2193,
        '?', 2213,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2189,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0 &&
          (lookahead < '[' || '`' < lookahead)) ADVANCE(2245);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2190,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2206,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2206,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2206,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2190,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2213,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2213,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2189,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2189,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2192,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2190,
        '=', 2211,
        '>', 2193,
        '?', 2213,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2190,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2189,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2199,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2190,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2215,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 15:
      if (lookahead == '\t') ADVANCE(2242);
      if (lookahead == '\n') ADVANCE(2226);
      if (lookahead == '\r') ADVANCE(2227);
      if (lookahead == ' ') ADVANCE(2244);
      if (lookahead == '"') ADVANCE(2195);
      if (lookahead == '\'') ADVANCE(2201);
      if (lookahead != 0 &&
          (lookahead < '<' || '>' < lookahead) &&
          lookahead != '`') ADVANCE(2231);
      END_STATE();
    case 16:
      if (lookahead == '\t') ADVANCE(2242);
      if (lookahead == '\n') ADVANCE(2226);
      if (lookahead == '\r') ADVANCE(2227);
      if (lookahead == ' ') ADVANCE(2244);
      if (lookahead == '/') ADVANCE(2208);
      if (lookahead == '>') ADVANCE(2193);
      if (lookahead == ':' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(2230);
      END_STATE();
    case 17:
      if (lookahead == '-') ADVANCE(18);
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '[') ADVANCE(431);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(2238);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('.' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          lookahead == '?' ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 18:
      if (lookahead == '-') ADVANCE(2233);
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('.' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 19:
      if (lookahead == '-') ADVANCE(141);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(140);
      END_STATE();
    case 20:
      if (lookahead == '-') ADVANCE(21);
      if (lookahead == '[') ADVANCE(431);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(2239);
      END_STATE();
    case 21:
      if (lookahead == '-') ADVANCE(2232);
      END_STATE();
    case 22:
      if (lookahead == '-') ADVANCE(2178);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(143);
      END_STATE();
    case 23:
      if (lookahead == '-') ADVANCE(2178);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(143);
      END_STATE();
    case 24:
      if (lookahead == '-') ADVANCE(27);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(26);
      END_STATE();
    case 25:
      if (lookahead == '-') ADVANCE(27);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(26);
      END_STATE();
    case 26:
      if (lookahead == '-') ADVANCE(23);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(22);
      END_STATE();
    case 27:
      if (lookahead == '-') ADVANCE(23);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(22);
      END_STATE();
    case 28:
      if (lookahead == '-') ADVANCE(31);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(30);
      END_STATE();
    case 29:
      if (lookahead == '-') ADVANCE(31);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(30);
      END_STATE();
    case 30:
      if (lookahead == '-') ADVANCE(25);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(24);
      END_STATE();
    case 31:
      if (lookahead == '-') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(24);
      END_STATE();
    case 32:
      if (lookahead == '-') ADVANCE(35);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(34);
      END_STATE();
    case 33:
      if (lookahead == '-') ADVANCE(35);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(34);
      END_STATE();
    case 34:
      if (lookahead == '-') ADVANCE(29);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(28);
      END_STATE();
    case 35:
      if (lookahead == '-') ADVANCE(29);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(28);
      END_STATE();
    case 36:
      if (lookahead == '-') ADVANCE(39);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(38);
      END_STATE();
    case 37:
      if (lookahead == '-') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(38);
      END_STATE();
    case 38:
      if (lookahead == '-') ADVANCE(33);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(32);
      END_STATE();
    case 39:
      if (lookahead == '-') ADVANCE(33);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(32);
      END_STATE();
    case 40:
      if (lookahead == '-') ADVANCE(43);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(42);
      END_STATE();
    case 41:
      if (lookahead == '-') ADVANCE(43);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(42);
      END_STATE();
    case 42:
      if (lookahead == '-') ADVANCE(37);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(36);
      END_STATE();
    case 43:
      if (lookahead == '-') ADVANCE(37);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(36);
      END_STATE();
    case 44:
      if (lookahead == '-') ADVANCE(47);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(46);
      END_STATE();
    case 45:
      if (lookahead == '-') ADVANCE(47);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(46);
      END_STATE();
    case 46:
      if (lookahead == '-') ADVANCE(41);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(40);
      END_STATE();
    case 47:
      if (lookahead == '-') ADVANCE(41);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(40);
      END_STATE();
    case 48:
      if (lookahead == '-') ADVANCE(51);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(50);
      END_STATE();
    case 49:
      if (lookahead == '-') ADVANCE(51);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(50);
      END_STATE();
    case 50:
      if (lookahead == '-') ADVANCE(45);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(44);
      END_STATE();
    case 51:
      if (lookahead == '-') ADVANCE(45);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(44);
      END_STATE();
    case 52:
      if (lookahead == '-') ADVANCE(55);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(54);
      END_STATE();
    case 53:
      if (lookahead == '-') ADVANCE(55);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(54);
      END_STATE();
    case 54:
      if (lookahead == '-') ADVANCE(49);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(48);
      END_STATE();
    case 55:
      if (lookahead == '-') ADVANCE(49);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(48);
      END_STATE();
    case 56:
      if (lookahead == '-') ADVANCE(59);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(58);
      END_STATE();
    case 57:
      if (lookahead == '-') ADVANCE(59);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(58);
      END_STATE();
    case 58:
      if (lookahead == '-') ADVANCE(53);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(52);
      END_STATE();
    case 59:
      if (lookahead == '-') ADVANCE(53);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(52);
      END_STATE();
    case 60:
      if (lookahead == '-') ADVANCE(63);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(62);
      END_STATE();
    case 61:
      if (lookahead == '-') ADVANCE(63);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(62);
      END_STATE();
    case 62:
      if (lookahead == '-') ADVANCE(57);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(56);
      END_STATE();
    case 63:
      if (lookahead == '-') ADVANCE(57);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(56);
      END_STATE();
    case 64:
      if (lookahead == '-') ADVANCE(67);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(66);
      END_STATE();
    case 65:
      if (lookahead == '-') ADVANCE(67);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(66);
      END_STATE();
    case 66:
      if (lookahead == '-') ADVANCE(61);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 67:
      if (lookahead == '-') ADVANCE(61);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 68:
      if (lookahead == '-') ADVANCE(71);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(70);
      END_STATE();
    case 69:
      if (lookahead == '-') ADVANCE(71);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(70);
      END_STATE();
    case 70:
      if (lookahead == '-') ADVANCE(65);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(64);
      END_STATE();
    case 71:
      if (lookahead == '-') ADVANCE(65);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(64);
      END_STATE();
    case 72:
      if (lookahead == '-') ADVANCE(75);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(74);
      END_STATE();
    case 73:
      if (lookahead == '-') ADVANCE(75);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(74);
      END_STATE();
    case 74:
      if (lookahead == '-') ADVANCE(69);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(68);
      END_STATE();
    case 75:
      if (lookahead == '-') ADVANCE(69);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(68);
      END_STATE();
    case 76:
      if (lookahead == '-') ADVANCE(79);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(78);
      END_STATE();
    case 77:
      if (lookahead == '-') ADVANCE(79);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(78);
      END_STATE();
    case 78:
      if (lookahead == '-') ADVANCE(73);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(72);
      END_STATE();
    case 79:
      if (lookahead == '-') ADVANCE(73);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(72);
      END_STATE();
    case 80:
      if (lookahead == '-') ADVANCE(83);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(82);
      END_STATE();
    case 81:
      if (lookahead == '-') ADVANCE(83);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(82);
      END_STATE();
    case 82:
      if (lookahead == '-') ADVANCE(77);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(76);
      END_STATE();
    case 83:
      if (lookahead == '-') ADVANCE(77);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(76);
      END_STATE();
    case 84:
      if (lookahead == '-') ADVANCE(87);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(86);
      END_STATE();
    case 85:
      if (lookahead == '-') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(86);
      END_STATE();
    case 86:
      if (lookahead == '-') ADVANCE(81);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 87:
      if (lookahead == '-') ADVANCE(81);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 88:
      if (lookahead == '-') ADVANCE(91);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(90);
      END_STATE();
    case 89:
      if (lookahead == '-') ADVANCE(91);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(90);
      END_STATE();
    case 90:
      if (lookahead == '-') ADVANCE(85);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(84);
      END_STATE();
    case 91:
      if (lookahead == '-') ADVANCE(85);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(84);
      END_STATE();
    case 92:
      if (lookahead == '-') ADVANCE(95);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(94);
      END_STATE();
    case 93:
      if (lookahead == '-') ADVANCE(95);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(94);
      END_STATE();
    case 94:
      if (lookahead == '-') ADVANCE(89);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(88);
      END_STATE();
    case 95:
      if (lookahead == '-') ADVANCE(89);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(88);
      END_STATE();
    case 96:
      if (lookahead == '-') ADVANCE(99);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(98);
      END_STATE();
    case 97:
      if (lookahead == '-') ADVANCE(99);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(98);
      END_STATE();
    case 98:
      if (lookahead == '-') ADVANCE(93);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(92);
      END_STATE();
    case 99:
      if (lookahead == '-') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(92);
      END_STATE();
    case 100:
      if (lookahead == '-') ADVANCE(103);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(102);
      END_STATE();
    case 101:
      if (lookahead == '-') ADVANCE(103);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(102);
      END_STATE();
    case 102:
      if (lookahead == '-') ADVANCE(97);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(96);
      END_STATE();
    case 103:
      if (lookahead == '-') ADVANCE(97);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(96);
      END_STATE();
    case 104:
      if (lookahead == '-') ADVANCE(107);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(106);
      END_STATE();
    case 105:
      if (lookahead == '-') ADVANCE(107);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(106);
      END_STATE();
    case 106:
      if (lookahead == '-') ADVANCE(101);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(100);
      END_STATE();
    case 107:
      if (lookahead == '-') ADVANCE(101);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(100);
      END_STATE();
    case 108:
      if (lookahead == '-') ADVANCE(111);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(110);
      END_STATE();
    case 109:
      if (lookahead == '-') ADVANCE(111);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(110);
      END_STATE();
    case 110:
      if (lookahead == '-') ADVANCE(105);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(104);
      END_STATE();
    case 111:
      if (lookahead == '-') ADVANCE(105);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(104);
      END_STATE();
    case 112:
      if (lookahead == '-') ADVANCE(115);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(114);
      END_STATE();
    case 113:
      if (lookahead == '-') ADVANCE(115);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(114);
      END_STATE();
    case 114:
      if (lookahead == '-') ADVANCE(109);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(108);
      END_STATE();
    case 115:
      if (lookahead == '-') ADVANCE(109);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(108);
      END_STATE();
    case 116:
      if (lookahead == '-') ADVANCE(119);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(118);
      END_STATE();
    case 117:
      if (lookahead == '-') ADVANCE(119);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(118);
      END_STATE();
    case 118:
      if (lookahead == '-') ADVANCE(113);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(112);
      END_STATE();
    case 119:
      if (lookahead == '-') ADVANCE(113);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(112);
      END_STATE();
    case 120:
      if (lookahead == '-') ADVANCE(123);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(122);
      END_STATE();
    case 121:
      if (lookahead == '-') ADVANCE(123);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(122);
      END_STATE();
    case 122:
      if (lookahead == '-') ADVANCE(117);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(116);
      END_STATE();
    case 123:
      if (lookahead == '-') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(116);
      END_STATE();
    case 124:
      if (lookahead == '-') ADVANCE(127);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(126);
      END_STATE();
    case 125:
      if (lookahead == '-') ADVANCE(127);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(126);
      END_STATE();
    case 126:
      if (lookahead == '-') ADVANCE(121);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(120);
      END_STATE();
    case 127:
      if (lookahead == '-') ADVANCE(121);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(120);
      END_STATE();
    case 128:
      if (lookahead == '-') ADVANCE(131);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(130);
      END_STATE();
    case 129:
      if (lookahead == '-') ADVANCE(131);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(130);
      END_STATE();
    case 130:
      if (lookahead == '-') ADVANCE(125);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(124);
      END_STATE();
    case 131:
      if (lookahead == '-') ADVANCE(125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(124);
      END_STATE();
    case 132:
      if (lookahead == '-') ADVANCE(135);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(134);
      END_STATE();
    case 133:
      if (lookahead == '-') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(134);
      END_STATE();
    case 134:
      if (lookahead == '-') ADVANCE(129);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(128);
      END_STATE();
    case 135:
      if (lookahead == '-') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(128);
      END_STATE();
    case 136:
      if (lookahead == '-') ADVANCE(139);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(138);
      END_STATE();
    case 137:
      if (lookahead == '-') ADVANCE(139);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(138);
      END_STATE();
    case 138:
      if (lookahead == '-') ADVANCE(133);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(132);
      END_STATE();
    case 139:
      if (lookahead == '-') ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(132);
      END_STATE();
    case 140:
      if (lookahead == '-') ADVANCE(137);
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(136);
      END_STATE();
    case 141:
      if (lookahead == '-') ADVANCE(137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(136);
      END_STATE();
    case 142:
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      END_STATE();
    case 143:
      if (lookahead == '.') ADVANCE(2177);
      if (lookahead == '>') ADVANCE(2229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(142);
      END_STATE();
    case 144:
      if (lookahead == '1') ADVANCE(2157);
      if (lookahead == '3') ADVANCE(147);
      END_STATE();
    case 145:
      if (lookahead == '1') ADVANCE(2171);
      if (lookahead == ';') ADVANCE(2185);
      END_STATE();
    case 146:
      if (lookahead == '1') ADVANCE(373);
      if (lookahead == '2') ADVANCE(2158);
      if (lookahead == '3') ADVANCE(370);
      if (lookahead == '4') ADVANCE(149);
      if (lookahead == '5') ADVANCE(2159);
      if (lookahead == '7') ADVANCE(150);
      END_STATE();
    case 147:
      if (lookahead == '4') ADVANCE(152);
      END_STATE();
    case 148:
      if (lookahead == '4') ADVANCE(152);
      if (lookahead == 'f') ADVANCE(1668);
      END_STATE();
    case 149:
      if (lookahead == '5') ADVANCE(152);
      END_STATE();
    case 150:
      if (lookahead == '8') ADVANCE(152);
      END_STATE();
    case 151:
      if (lookahead == ':') ADVANCE(388);
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '/' ||
          lookahead == '=' ||
          lookahead == '?' ||
          ('^' <= lookahead && lookahead <= '`') ||
          ('{' <= lookahead && lookahead <= '~')) ADVANCE(390);
      if (lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(151);
      END_STATE();
    case 152:
      if (lookahead == ';') ADVANCE(2185);
      END_STATE();
    case 153:
      ADVANCE_MAP(
        ';', 2185,
        'A', 593,
        'B', 583,
        'E', 280,
        'H', 546,
        'a', 819,
        'b', 584,
        'c', 611,
        'd', 803,
        'e', 277,
        'f', 1261,
        'g', 167,
        'h', 629,
        'j', 754,
        'l', 203,
        'm', 1244,
        'n', 464,
        'o', 554,
        'p', 633,
        'r', 587,
        's', 534,
        't', 230,
        'u', 1778,
        'v', 1071,
      );
      END_STATE();
    case 154:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1762);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'q') ADVANCE(2079);
      if (lookahead == 's') ADVANCE(976);
      if (lookahead == 'x') ADVANCE(1303);
      END_STATE();
    case 155:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'A') ADVANCE(1881);
      END_STATE();
    case 156:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'V') ADVANCE(1072);
      END_STATE();
    case 157:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      END_STATE();
    case 158:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'D') ADVANCE(1652);
      END_STATE();
    case 159:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'E') ADVANCE(1740);
      END_STATE();
    case 160:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'L') ADVANCE(1074);
      END_STATE();
    case 161:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'R') ADVANCE(1325);
      END_STATE();
    case 162:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'U') ADVANCE(1696);
      END_STATE();
    case 163:
      ADVANCE_MAP(
        ';', 2185,
        'C', 1663,
        'D', 1637,
        'E', 1372,
        'G', 1897,
        'H', 2090,
        'L', 1046,
        'N', 1013,
        'P', 1858,
        'R', 1047,
        'S', 1741,
        'T', 1250,
        'V', 1086,
      );
      END_STATE();
    case 164:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'C') ADVANCE(572);
      END_STATE();
    case 165:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'D') ADVANCE(1598);
      if (lookahead == 'E') ADVANCE(1740);
      END_STATE();
    case 166:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'D') ADVANCE(445);
      END_STATE();
    case 167:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(152);
      END_STATE();
    case 168:
      ADVANCE_MAP(
        ';', 2185,
        'E', 152,
        'a', 1677,
        'c', 2068,
        'e', 229,
        'i', 1499,
        'n', 463,
        'o', 882,
        's', 1255,
        'u', 1809,
      );
      END_STATE();
    case 169:
      ADVANCE_MAP(
        ';', 2185,
        'E', 152,
        'd', 1598,
        'e', 240,
        'm', 2075,
        'n', 2160,
        'p', 1448,
        'r', 583,
        's', 1026,
      );
      END_STATE();
    case 170:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'v') ADVANCE(2173);
      END_STATE();
    case 171:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'e') ADVANCE(325);
      END_STATE();
    case 172:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'i') ADVANCE(881);
      if (lookahead == 'o') ADVANCE(1903);
      if (lookahead == 'p') ADVANCE(1829);
      END_STATE();
    case 173:
      ADVANCE_MAP(
        ';', 2185,
        'E', 297,
        'a', 821,
        'b', 1749,
        'c', 1291,
        'd', 1598,
        'e', 298,
        'f', 1750,
        'g', 280,
        'i', 1484,
        'j', 754,
        'l', 369,
        'n', 464,
        'o', 1680,
        'r', 537,
        's', 789,
        't', 231,
        'v', 1071,
      );
      END_STATE();
    case 174:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      END_STATE();
    case 175:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      if (lookahead == 'F') ADVANCE(2097);
      if (lookahead == 'G') ADVANCE(1890);
      if (lookahead == 'L') ADVANCE(972);
      if (lookahead == 'S') ADVANCE(1458);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 176:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      if (lookahead == 'F') ADVANCE(2097);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 177:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      if (lookahead == 'G') ADVANCE(1890);
      if (lookahead == 'L') ADVANCE(972);
      if (lookahead == 'S') ADVANCE(1458);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 178:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      if (lookahead == 'S') ADVANCE(1458);
      END_STATE();
    case 179:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E') ADVANCE(1740);
      if (lookahead == 'S') ADVANCE(1458);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 180:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'G') ADVANCE(152);
      END_STATE();
    case 181:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'H') ADVANCE(152);
      END_STATE();
    case 182:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'I') ADVANCE(1580);
      if (lookahead == 'S') ADVANCE(2061);
      if (lookahead == 'U') ADVANCE(1561);
      END_STATE();
    case 183:
      ADVANCE_MAP(
        ';', 2185,
        'J', 754,
        'a', 820,
        'c', 613,
        'e', 1103,
        'f', 1750,
        'l', 257,
        'm', 1243,
        'o', 1528,
        's', 786,
        'T', 152,
        't', 152,
      );
      END_STATE();
    case 184:
      ADVANCE_MAP(
        ';', 2185,
        'J', 754,
        'a', 1500,
        'b', 1749,
        'c', 980,
        'd', 1598,
        'f', 1750,
        'o', 1680,
        'r', 998,
        's', 778,
        'T', 152,
        'g', 152,
        't', 152,
      );
      END_STATE();
    case 185:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'L') ADVANCE(972);
      END_STATE();
    case 186:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'N') ADVANCE(152);
      END_STATE();
    case 187:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'P') ADVANCE(152);
      END_STATE();
    case 188:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'P') ADVANCE(1448);
      END_STATE();
    case 189:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'T') ADVANCE(152);
      END_STATE();
    case 190:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 191:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'Y') ADVANCE(152);
      END_STATE();
    case 192:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(784);
      if (lookahead == 'p') ADVANCE(1414);
      END_STATE();
    case 193:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1135);
      if (lookahead == 'o') ADVANCE(2013);
      END_STATE();
    case 194:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(2175);
      END_STATE();
    case 195:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(778);
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 196:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1750);
      END_STATE();
    case 197:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(792);
      if (lookahead == 'p') ADVANCE(152);
      END_STATE();
    case 198:
      ADVANCE_MAP(
        ';', 2185,
        'a', 1678,
        'c', 2068,
        'e', 249,
        'i', 1817,
        'n', 463,
        'p', 1673,
        's', 1255,
        'E', 152,
        'y', 152,
      );
      END_STATE();
    case 199:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(305);
      if (lookahead == 's') ADVANCE(1256);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 200:
      ADVANCE_MAP(
        ';', 2185,
        'a', 1677,
        'b', 276,
        'f', 1903,
        'h', 1339,
        'l', 1677,
        'p', 1358,
        's', 1255,
        't', 1358,
        'c', 152,
        'w', 152,
      );
      END_STATE();
    case 201:
      ADVANCE_MAP(
        ';', 2185,
        'a', 850,
        'c', 1233,
        'd', 2169,
        'm', 309,
        's', 1255,
        't', 2121,
        'b', 152,
        'e', 152,
      );
      END_STATE();
    case 202:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(850);
      if (lookahead == 'i') ADVANCE(881);
      if (lookahead == 'o') ADVANCE(1903);
      if (lookahead == 'p') ADVANCE(1823);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 203:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'c') ADVANCE(1649);
      if (lookahead == 'h') ADVANCE(648);
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 204:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'd') ADVANCE(272);
      if (lookahead == 'i') ADVANCE(1137);
      if (lookahead == 'o') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(1413);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 205:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1903);
      END_STATE();
    case 206:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1914);
      if (lookahead == 'c') ADVANCE(1233);
      if (lookahead == 'd') ADVANCE(1605);
      END_STATE();
    case 207:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1358);
      END_STATE();
    case 208:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'e') ADVANCE(921);
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 209:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1363);
      if (lookahead == 'c') ADVANCE(1958);
      if (lookahead == 'g') ADVANCE(152);
      END_STATE();
    case 210:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1503);
      if (lookahead == 'b') ADVANCE(1872);
      if (lookahead == 'c') ADVANCE(576);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 211:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1503);
      if (lookahead == 's') ADVANCE(1413);
      if (lookahead == 'd' ||
          lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 212:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(843);
      END_STATE();
    case 213:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1798);
      if (lookahead == 'c') ADVANCE(1190);
      if (lookahead == 'o') ADVANCE(1837);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 214:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1382);
      END_STATE();
    case 215:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1766);
      if (lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 216:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1712);
      if (lookahead == 'c') ADVANCE(2088);
      if (lookahead == 'e') ADVANCE(1732);
      if (lookahead == 'n') ADVANCE(670);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 217:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1263);
      if (lookahead == 'e') ADVANCE(334);
      END_STATE();
    case 218:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a') ADVANCE(1457);
      if (lookahead == 's') ADVANCE(1358);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 219:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(1598);
      if (lookahead == 'c') ADVANCE(1233);
      if (lookahead == 'f') ADVANCE(318);
      END_STATE();
    case 220:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(152);
      if (lookahead == 'd') ADVANCE(356);
      END_STATE();
    case 221:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(152);
      if (lookahead == 'h') ADVANCE(1950);
      END_STATE();
    case 222:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(196);
      END_STATE();
    case 223:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(196);
      if (lookahead == 'd') ADVANCE(152);
      END_STATE();
    case 224:
      ADVANCE_MAP(
        ';', 2185,
        'b', 276,
        'f', 1903,
        'h', 1339,
        'l', 1677,
        'p', 1358,
        's', 1255,
        't', 1358,
      );
      END_STATE();
    case 225:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(546);
      if (lookahead == 'e') ADVANCE(1732);
      END_STATE();
    case 226:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b') ADVANCE(1825);
      if (lookahead == 'c') ADVANCE(576);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'o') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 227:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(152);
      END_STATE();
    case 228:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(260);
      if (lookahead == 'f') ADVANCE(1566);
      if (lookahead == 'm') ADVANCE(1222);
      if (lookahead == 's') ADVANCE(850);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 229:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(216);
      END_STATE();
    case 230:
      ADVANCE_MAP(
        ';', 2185,
        'c', 763,
        'd', 1598,
        'h', 1799,
        'i', 1497,
        'l', 583,
        'q', 2080,
        'r', 505,
      );
      END_STATE();
    case 231:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(763);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'l') ADVANCE(504);
      if (lookahead == 'q') ADVANCE(2080);
      if (lookahead == 'r') ADVANCE(571);
      END_STATE();
    case 232:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(259);
      END_STATE();
    case 233:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(762);
      if (lookahead == 'd') ADVANCE(1642);
      if (lookahead == 'l') ADVANCE(267);
      END_STATE();
    case 234:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(762);
      if (lookahead == 'd') ADVANCE(1643);
      if (lookahead == 'g') ADVANCE(267);
      if (lookahead == 's') ADVANCE(669);
      END_STATE();
    case 235:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(600);
      if (lookahead == 'e') ADVANCE(1168);
      if (lookahead == 'l') ADVANCE(655);
      if (lookahead == 'p') ADVANCE(1810);
      END_STATE();
    case 236:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(1233);
      if (lookahead == 'w') ADVANCE(152);
      END_STATE();
    case 237:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(2068);
      if (lookahead == 'e') ADVANCE(232);
      END_STATE();
    case 238:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(2068);
      if (lookahead == 'e' ||
          lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 239:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c') ADVANCE(690);
      if (lookahead == 'f') ADVANCE(1278);
      if (lookahead == 'o') ADVANCE(876);
      if (lookahead == 't') ADVANCE(235);
      END_STATE();
    case 240:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1598);
      END_STATE();
    case 241:
      ADVANCE_MAP(
        ';', 2185,
        'd', 1598,
        'e', 325,
        'g', 167,
        'l', 167,
        'n', 926,
        'p', 1448,
        'r', 583,
      );
      END_STATE();
    case 242:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 's') ADVANCE(358);
      if (lookahead == 'E' ||
          lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 243:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(152);
      END_STATE();
    case 244:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(926);
      END_STATE();
    case 245:
      ADVANCE_MAP(
        ';', 2185,
        'd', 1599,
        'e', 240,
        'h', 1916,
        'l', 583,
        'm', 2075,
        'n', 2160,
        'p', 1448,
        's', 1026,
      );
      if (('1' <= lookahead && lookahead <= '3') ||
          lookahead == 'E') ADVANCE(152);
      END_STATE();
    case 246:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1591);
      if (lookahead == 'l') ADVANCE(986);
      if (lookahead == 'r') ADVANCE(1286);
      END_STATE();
    case 247:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1591);
      if (lookahead == 'l') ADVANCE(986);
      if (lookahead == 'u') ADVANCE(1677);
      END_STATE();
    case 248:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1591);
      if (lookahead == 'l') ADVANCE(1079);
      if (lookahead == 'q') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(1329);
      END_STATE();
    case 249:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'd') ADVANCE(1263);
      END_STATE();
    case 250:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 251:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(926);
      if (lookahead == 'm') ADVANCE(1933);
      if (lookahead == 'r') ADVANCE(1987);
      if (lookahead == 's') ADVANCE(1700);
      if (lookahead == 'z') ADVANCE(583);
      END_STATE();
    case 252:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 's') ADVANCE(1014);
      END_STATE();
    case 253:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1750);
      END_STATE();
    case 254:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1733);
      if (lookahead == 'm') ADVANCE(1315);
      if (lookahead == 'p') ADVANCE(1448);
      if (lookahead == 's') ADVANCE(1747);
      END_STATE();
    case 255:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1958);
      END_STATE();
    case 256:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1161);
      END_STATE();
    case 257:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1106);
      END_STATE();
    case 258:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(325);
      END_STATE();
    case 259:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1732);
      END_STATE();
    case 260:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1732);
      if (lookahead == 'l') ADVANCE(946);
      END_STATE();
    case 261:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(334);
      END_STATE();
    case 262:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(852);
      if (lookahead == 'i') ADVANCE(1487);
      if (lookahead == 'o') ADVANCE(905);
      END_STATE();
    case 263:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(321);
      END_STATE();
    case 264:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1737);
      END_STATE();
    case 265:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1737);
      if (lookahead == 'n') ADVANCE(991);
      END_STATE();
    case 266:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(599);
      END_STATE();
    case 267:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1903);
      END_STATE();
    case 268:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1934);
      END_STATE();
    case 269:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1558);
      if (lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 270:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1835);
      if (lookahead == 's') ADVANCE(961);
      END_STATE();
    case 271:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1791);
      if (lookahead == 's') ADVANCE(1677);
      END_STATE();
    case 272:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1801);
      if (lookahead == 'f' ||
          lookahead == 'm') ADVANCE(152);
      END_STATE();
    case 273:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e') ADVANCE(1790);
      END_STATE();
    case 274:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 275:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(1587);
      if (lookahead == 'y') ADVANCE(335);
      END_STATE();
    case 276:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(1903);
      END_STATE();
    case 277:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(1963);
      if (lookahead == 'g') ADVANCE(152);
      if (lookahead == 'q') ADVANCE(326);
      if (lookahead == 's') ADVANCE(234);
      END_STATE();
    case 278:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(1504);
      if (lookahead == 'l') ADVANCE(947);
      END_STATE();
    case 279:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f') ADVANCE(1988);
      if (lookahead == 'q') ADVANCE(326);
      if (lookahead == 's') ADVANCE(334);
      END_STATE();
    case 280:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'g') ADVANCE(152);
      END_STATE();
    case 281:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'g') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(1979);
      if (lookahead == 'm') ADVANCE(1711);
      END_STATE();
    case 282:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'g') ADVANCE(1958);
      END_STATE();
    case 283:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'g') ADVANCE(926);
      END_STATE();
    case 284:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'h') ADVANCE(152);
      END_STATE();
    case 285:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'h') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(1629);
      END_STATE();
    case 286:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'h') ADVANCE(1588);
      END_STATE();
    case 287:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1236);
      if (lookahead == 'n') ADVANCE(1118);
      if (lookahead == 'o') ADVANCE(1979);
      END_STATE();
    case 288:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(900);
      if (lookahead == 'o') ADVANCE(1520);
      END_STATE();
    case 289:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1780);
      if (lookahead == 'u') ADVANCE(1994);
      if (lookahead == 'E' ||
          lookahead == 'd' ||
          lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 290:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1780);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 291:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1538);
      if (lookahead == 'p') ADVANCE(652);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 292:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(2017);
      END_STATE();
    case 293:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(865);
      END_STATE();
    case 294:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1512);
      if (lookahead == 'n') ADVANCE(1246);
      END_STATE();
    case 295:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1440);
      END_STATE();
    case 296:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'i') ADVANCE(1577);
      if (lookahead == 'l') ADVANCE(152);
      if (lookahead == 's') ADVANCE(240);
      END_STATE();
    case 297:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 298:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(152);
      if (lookahead == 'q') ADVANCE(326);
      if (lookahead == 's') ADVANCE(233);
      END_STATE();
    case 299:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(2016);
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 300:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(1958);
      END_STATE();
    case 301:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(1629);
      END_STATE();
    case 302:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(1629);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 303:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(926);
      if (lookahead == 'd' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 304:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(2020);
      if (lookahead == 'm') ADVANCE(572);
      END_STATE();
    case 305:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'l') ADVANCE(1399);
      END_STATE();
    case 306:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'm') ADVANCE(152);
      END_STATE();
    case 307:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'm') ADVANCE(271);
      END_STATE();
    case 308:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'm') ADVANCE(606);
      END_STATE();
    case 309:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'n') ADVANCE(152);
      END_STATE();
    case 310:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'n') ADVANCE(625);
      END_STATE();
    case 311:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 312:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(330);
      END_STATE();
    case 313:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(297);
      END_STATE();
    case 314:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(1096);
      END_STATE();
    case 315:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(2127);
      END_STATE();
    case 316:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(2116);
      END_STATE();
    case 317:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(850);
      END_STATE();
    case 318:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(1804);
      END_STATE();
    case 319:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(2033);
      END_STATE();
    case 320:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(1543);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 321:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'o') ADVANCE(1568);
      END_STATE();
    case 322:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'p') ADVANCE(152);
      END_STATE();
    case 323:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(294);
      END_STATE();
    case 324:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'p') ADVANCE(1829);
      END_STATE();
    case 325:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'q') ADVANCE(152);
      END_STATE();
    case 326:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'q') ADVANCE(152);
      if (lookahead == 's') ADVANCE(1443);
      END_STATE();
    case 327:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'q') ADVANCE(326);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 328:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'q') ADVANCE(325);
      END_STATE();
    case 329:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'q') ADVANCE(2079);
      END_STATE();
    case 330:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 331:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'r') ADVANCE(564);
      if (lookahead == 's') ADVANCE(240);
      END_STATE();
    case 332:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'r') ADVANCE(1271);
      END_STATE();
    case 333:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'r') ADVANCE(1249);
      END_STATE();
    case 334:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 335:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(1750);
      END_STATE();
    case 336:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(243);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 337:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(223);
      END_STATE();
    case 338:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(2149);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 339:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(1463);
      END_STATE();
    case 340:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(1358);
      END_STATE();
    case 341:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(2021);
      END_STATE();
    case 342:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(961);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 343:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(1016);
      END_STATE();
    case 344:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(1029);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 345:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 's') ADVANCE(2083);
      END_STATE();
    case 346:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 347:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 348:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(1588);
      END_STATE();
    case 349:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(273);
      END_STATE();
    case 350:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(734);
      END_STATE();
    case 351:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(1358);
      END_STATE();
    case 352:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(832);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 353:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(679);
      END_STATE();
    case 354:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(1612);
      END_STATE();
    case 355:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 't') ADVANCE(1271);
      END_STATE();
    case 356:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 357:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'u') ADVANCE(1241);
      END_STATE();
    case 358:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 359:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'v') ADVANCE(2173);
      END_STATE();
    case 360:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'v') ADVANCE(723);
      END_STATE();
    case 361:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'v') ADVANCE(1028);
      END_STATE();
    case 362:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'a' ||
          lookahead == 'h') ADVANCE(152);
      END_STATE();
    case 363:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'b' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 364:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'c' ||
          lookahead == 'w') ADVANCE(152);
      END_STATE();
    case 365:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e' ||
          lookahead == 'g') ADVANCE(152);
      END_STATE();
    case 366:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e' ||
          lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 367:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'f' ||
          lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 368:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 369:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'E' ||
          lookahead == 'a' ||
          lookahead == 'j') ADVANCE(152);
      END_STATE();
    case 370:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == '4' ||
          lookahead == '5' ||
          lookahead == '8') ADVANCE(152);
      END_STATE();
    case 371:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'D' ||
          lookahead == 'U' ||
          lookahead == 'd' ||
          lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 372:
      if (lookahead == ';') ADVANCE(2185);
      if (lookahead == 'H' ||
          lookahead == 'L' ||
          lookahead == 'R' ||
          lookahead == 'h' ||
          lookahead == 'l' ||
          lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 373:
      if (lookahead == ';') ADVANCE(2185);
      if (('2' <= lookahead && lookahead <= '6') ||
          lookahead == '8') ADVANCE(152);
      END_STATE();
    case 374:
      if (lookahead == ';') ADVANCE(2186);
      END_STATE();
    case 375:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(376);
      END_STATE();
    case 376:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(374);
      END_STATE();
    case 377:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(375);
      END_STATE();
    case 378:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(377);
      END_STATE();
    case 379:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(378);
      END_STATE();
    case 380:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(379);
      END_STATE();
    case 381:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(374);
      END_STATE();
    case 382:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(381);
      END_STATE();
    case 383:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(382);
      END_STATE();
    case 384:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(383);
      END_STATE();
    case 385:
      if (lookahead == ';') ADVANCE(2186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(384);
      END_STATE();
    case 386:
      if (lookahead == '>') ADVANCE(2234);
      END_STATE();
    case 387:
      if (lookahead == '>') ADVANCE(2241);
      END_STATE();
    case 388:
      if (lookahead == '>') ADVANCE(2228);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != ' ' &&
          lookahead != '<') ADVANCE(388);
      END_STATE();
    case 389:
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '/' ||
          lookahead == '=' ||
          lookahead == '?' ||
          ('^' <= lookahead && lookahead <= '`') ||
          ('{' <= lookahead && lookahead <= '~')) ADVANCE(390);
      if (lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(151);
      END_STATE();
    case 390:
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 391:
      ADVANCE_MAP(
        'A', 509,
        'a', 825,
        'c', 1289,
        'f', 1750,
        'i', 1377,
        'o', 1683,
        's', 799,
        'u', 1476,
      );
      END_STATE();
    case 392:
      ADVANCE_MAP(
        'A', 593,
        'B', 583,
        'H', 546,
        'a', 772,
        'b', 584,
        'c', 611,
        'd', 802,
        'e', 209,
        'f', 1261,
        'h', 630,
        'i', 1151,
        'l', 589,
        'm', 1661,
        'n', 1477,
        'o', 555,
        'p', 637,
        'r', 583,
        's', 535,
        't', 1217,
        'u', 1403,
        'x', 152,
      );
      END_STATE();
    case 393:
      if (lookahead == 'A') ADVANCE(527);
      END_STATE();
    case 394:
      ADVANCE_MAP(
        'A', 754,
        'I', 754,
        'U', 754,
        'a', 812,
        'c', 1291,
        'f', 1750,
        'o', 1680,
        's', 778,
        'u', 1485,
      );
      END_STATE();
    case 395:
      if (lookahead == 'A') ADVANCE(448);
      END_STATE();
    case 396:
      ADVANCE_MAP(
        'A', 1575,
        'C', 1088,
        'D', 1593,
        'F', 1402,
        'R', 1324,
        'T', 1033,
        'U', 1691,
        'V', 1067,
        'a', 1881,
        'r', 1317,
      );
      END_STATE();
    case 397:
      if (lookahead == 'A') ADVANCE(518);
      END_STATE();
    case 398:
      ADVANCE_MAP(
        'A', 1769,
        'B', 640,
        'D', 579,
        'a', 1535,
        'c', 2139,
        'd', 579,
        'e', 939,
        'f', 1750,
        'l', 2016,
        'n', 1932,
        'o', 1680,
        'p', 1819,
        'r', 2016,
        's', 801,
        'z', 1264,
      );
      END_STATE();
    case 399:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'd') ADVANCE(1750);
      if (lookahead == 'e') ADVANCE(279);
      if (lookahead == 's') ADVANCE(1255);
      if (lookahead == 't') ADVANCE(333);
      END_STATE();
    case 400:
      ADVANCE_MAP(
        'A', 1769,
        'H', 546,
        'a', 1134,
        'b', 1355,
        'c', 616,
        'd', 193,
        'e', 281,
        'f', 1262,
        'h', 623,
        'i', 594,
        'j', 754,
        'l', 768,
        'o', 1406,
        'r', 730,
        's', 770,
        't', 880,
        'u', 588,
        'w', 598,
        'z', 759,
      );
      END_STATE();
    case 401:
      ADVANCE_MAP(
        'A', 1769,
        'H', 546,
        'a', 813,
        'b', 1772,
        'c', 1270,
        'd', 585,
        'f', 1262,
        'g', 1811,
        'h', 624,
        'l', 861,
        'm', 195,
        'o', 1138,
        'p', 695,
        'r', 860,
        's', 778,
        't', 877,
        'u', 590,
        'w', 598,
      );
      END_STATE();
    case 402:
      ADVANCE_MAP(
        'A', 1769,
        'a', 1277,
        'b', 546,
        'c', 1289,
        'e', 677,
        'f', 1750,
        'k', 1904,
        'o', 591,
        's', 791,
        'y', 732,
      );
      END_STATE();
    case 403:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1769);
      END_STATE();
    case 404:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'p') ADVANCE(546);
      END_STATE();
    case 405:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1762);
      if (lookahead == 'c') ADVANCE(346);
      if (lookahead == 'm') ADVANCE(1223);
      if (lookahead == 's') ADVANCE(2122);
      if (lookahead == 't') ADVANCE(1471);
      if (lookahead == 'x') ADVANCE(1958);
      END_STATE();
    case 406:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1762);
      if (lookahead == 'n') ADVANCE(2122);
      END_STATE();
    case 407:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1762);
      if (lookahead == 'n') ADVANCE(975);
      END_STATE();
    case 408:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'a') ADVANCE(1861);
      if (lookahead == 'i') ADVANCE(1154);
      if (lookahead == 't') ADVANCE(1844);
      END_STATE();
    case 409:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 't') ADVANCE(332);
      END_STATE();
    case 410:
      if (lookahead == 'A') ADVANCE(1769);
      if (lookahead == 't') ADVANCE(1839);
      END_STATE();
    case 411:
      ADVANCE_MAP(
        'A', 1884,
        'D', 1652,
        'E', 1743,
        'T', 1018,
        'a', 1881,
        'd', 1659,
        'p', 1027,
        's', 1240,
      );
      END_STATE();
    case 412:
      if (lookahead == 'A') ADVANCE(818);
      END_STATE();
    case 413:
      if (lookahead == 'A') ADVANCE(818);
      if (lookahead == 'D') ADVANCE(1596);
      if (lookahead == 'G') ADVANCE(1802);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 414:
      if (lookahead == 'A') ADVANCE(1400);
      END_STATE();
    case 415:
      ADVANCE_MAP(
        'A', 1576,
        'C', 1088,
        'D', 1593,
        'F', 1402,
        'T', 1033,
        'U', 1691,
        'V', 1067,
        'a', 1881,
      );
      END_STATE();
    case 416:
      if (lookahead == 'A') ADVANCE(1881);
      END_STATE();
    case 417:
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'D') ADVANCE(1652);
      END_STATE();
    case 418:
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'R') ADVANCE(1325);
      END_STATE();
    case 419:
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'R') ADVANCE(1325);
      if (lookahead == 'T') ADVANCE(1003);
      END_STATE();
    case 420:
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'T') ADVANCE(1003);
      END_STATE();
    case 421:
      if (lookahead == 'A') ADVANCE(1881);
      if (lookahead == 'V') ADVANCE(1072);
      END_STATE();
    case 422:
      if (lookahead == 'A') ADVANCE(1885);
      if (lookahead == 'B') ADVANCE(1749);
      if (lookahead == 'L') ADVANCE(1065);
      if (lookahead == 'R') ADVANCE(1323);
      if (lookahead == 'T') ADVANCE(1018);
      if (lookahead == 'a') ADVANCE(1881);
      END_STATE();
    case 423:
      if (lookahead == 'B') ADVANCE(549);
      if (lookahead == 'P') ADVANCE(698);
      END_STATE();
    case 424:
      ADVANCE_MAP(
        'B', 583,
        'E', 180,
        'a', 822,
        'c', 613,
        'e', 361,
        'f', 1750,
        'h', 1588,
        'i', 1132,
        'o', 1686,
        'r', 1317,
        's', 785,
        'u', 1393,
      );
      END_STATE();
    case 425:
      if (lookahead == 'B') ADVANCE(546);
      END_STATE();
    case 426:
      if (lookahead == 'B') ADVANCE(546);
      if (lookahead == 'L') ADVANCE(1272);
      if (lookahead == 'S') ADVANCE(1060);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 427:
      if (lookahead == 'B') ADVANCE(1882);
      END_STATE();
    case 428:
      if (lookahead == 'B') ADVANCE(1874);
      if (lookahead == 'n') ADVANCE(429);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(163);
      END_STATE();
    case 429:
      if (lookahead == 'B') ADVANCE(1879);
      END_STATE();
    case 430:
      if (lookahead == 'C') ADVANCE(482);
      if (lookahead == 'c') ADVANCE(2139);
      END_STATE();
    case 431:
      if (lookahead == 'C') ADVANCE(451);
      END_STATE();
    case 432:
      if (lookahead == 'C') ADVANCE(572);
      END_STATE();
    case 433:
      if (lookahead == 'C') ADVANCE(1653);
      if (lookahead == 'T') ADVANCE(1304);
      END_STATE();
    case 434:
      if (lookahead == 'C') ADVANCE(1330);
      END_STATE();
    case 435:
      if (lookahead == 'C') ADVANCE(1425);
      END_STATE();
    case 436:
      if (lookahead == 'C') ADVANCE(2087);
      END_STATE();
    case 437:
      if (lookahead == 'C') ADVANCE(1669);
      END_STATE();
    case 438:
      if (lookahead == 'C') ADVANCE(1669);
      if (lookahead == 'D') ADVANCE(1595);
      if (lookahead == 'L') ADVANCE(1077);
      if (lookahead == 'R') ADVANCE(1327);
      if (lookahead == 'U') ADVANCE(1692);
      if (lookahead == 'V') ADVANCE(1086);
      END_STATE();
    case 439:
      if (lookahead == 'D') ADVANCE(1598);
      END_STATE();
    case 440:
      if (lookahead == 'D') ADVANCE(1598);
      if (lookahead == 'M') ADVANCE(1315);
      if (lookahead == 'P') ADVANCE(1448);
      if (lookahead == 'T') ADVANCE(1304);
      END_STATE();
    case 441:
      if (lookahead == 'D') ADVANCE(1598);
      if (lookahead == 'a') ADVANCE(1769);
      END_STATE();
    case 442:
      if (lookahead == 'D') ADVANCE(1598);
      if (lookahead == 'o') ADVANCE(1958);
      END_STATE();
    case 443:
      if (lookahead == 'D') ADVANCE(1598);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 444:
      ADVANCE_MAP(
        'D', 319,
        'J', 754,
        'S', 754,
        'Z', 754,
        'a', 1136,
        'c', 616,
        'e', 1359,
        'f', 1750,
        'i', 558,
        'o', 1684,
        's', 799,
      );
      END_STATE();
    case 445:
      if (lookahead == 'D') ADVANCE(152);
      END_STATE();
    case 446:
      ADVANCE_MAP(
        'D', 442,
        'a', 816,
        'c', 614,
        'd', 1598,
        'e', 152,
        'f', 443,
        'g', 331,
        'l', 296,
        'm', 545,
        'n', 1121,
        'o', 1138,
        'p', 627,
        'q', 809,
        'r', 441,
        's', 779,
        't', 362,
        'u', 1465,
        'x', 837,
      );
      END_STATE();
    case 447:
      ADVANCE_MAP(
        'D', 439,
        'a', 764,
        'c', 1654,
        'd', 579,
        'e', 662,
        'f', 1750,
        'h', 1588,
        'i', 853,
        'l', 817,
        'n', 1708,
        'o', 917,
        'p', 152,
        's', 798,
        'u', 304,
      );
      END_STATE();
    case 448:
      if (lookahead == 'D') ADVANCE(462);
      END_STATE();
    case 449:
      ADVANCE_MAP(
        'D', 2174,
        'H', 371,
        'U', 2174,
        'V', 372,
        'b', 1603,
        'd', 2174,
        'h', 371,
        'm', 1315,
        'p', 1448,
        't', 1304,
        'u', 2174,
        'v', 372,
      );
      END_STATE();
    case 450:
      if (lookahead == 'D') ADVANCE(754);
      END_STATE();
    case 451:
      if (lookahead == 'D') ADVANCE(397);
      END_STATE();
    case 452:
      ADVANCE_MAP(
        'D', 579,
        'H', 583,
        'a', 1677,
        'd', 579,
        'g', 2166,
        'i', 1522,
        'l', 409,
        'r', 410,
        's', 1255,
      );
      END_STATE();
    case 453:
      ADVANCE_MAP(
        'D', 579,
        'b', 546,
        'c', 2139,
        'd', 582,
        'e', 929,
        'f', 1750,
        'o', 1680,
        's', 778,
        'v', 897,
      );
      END_STATE();
    case 454:
      if (lookahead == 'D') ADVANCE(579);
      if (lookahead == 'd') ADVANCE(579);
      END_STATE();
    case 455:
      if (lookahead == 'D') ADVANCE(1652);
      if (lookahead == 'L') ADVANCE(1074);
      if (lookahead == 'R') ADVANCE(1325);
      if (lookahead == 'U') ADVANCE(1696);
      END_STATE();
    case 456:
      if (lookahead == 'D') ADVANCE(1254);
      END_STATE();
    case 457:
      if (lookahead == 'D') ADVANCE(1020);
      END_STATE();
    case 458:
      if (lookahead == 'D') ADVANCE(1660);
      if (lookahead == 'E') ADVANCE(1740);
      END_STATE();
    case 459:
      if (lookahead == 'D') ADVANCE(1664);
      if (lookahead == 'T') ADVANCE(1039);
      if (lookahead == 'V') ADVANCE(1067);
      END_STATE();
    case 460:
      if (lookahead == 'D') ADVANCE(1674);
      if (lookahead == 'Q') ADVANCE(2098);
      END_STATE();
    case 461:
      ADVANCE_MAP(
        'E', 1398,
        'M', 187,
        'a', 812,
        'b', 1749,
        'c', 1270,
        'f', 1750,
        'g', 1811,
        'l', 1679,
        'm', 541,
        'n', 881,
        'o', 1138,
        'p', 1704,
        'r', 1245,
        's', 796,
        't', 1292,
        'u', 1464,
      );
      END_STATE();
    case 462:
      if (lookahead == 'E') ADVANCE(152);
      END_STATE();
    case 463:
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'a') ADVANCE(1677);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 464:
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'a') ADVANCE(1695);
      if (lookahead == 'e') ADVANCE(328);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 465:
      if (lookahead == 'E') ADVANCE(152);
      if (lookahead == 'e') ADVANCE(327);
      if (lookahead == 's') ADVANCE(1255);
      if (lookahead == 't') ADVANCE(330);
      END_STATE();
    case 466:
      if (lookahead == 'E') ADVANCE(1371);
      if (lookahead == 'U') ADVANCE(1727);
      END_STATE();
    case 467:
      ADVANCE_MAP(
        'E', 754,
        'J', 1404,
        'O', 754,
        'a', 812,
        'c', 1270,
        'd', 1598,
        'f', 1750,
        'g', 1811,
        'm', 192,
        'n', 1986,
        'o', 1139,
        's', 778,
        't', 1318,
        'u', 1347,
      );
      END_STATE();
    case 468:
      if (lookahead == 'E') ADVANCE(1370);
      END_STATE();
    case 469:
      ADVANCE_MAP(
        'E', 1404,
        'a', 812,
        'c', 1270,
        'd', 752,
        'f', 1750,
        'g', 1811,
        'm', 543,
        'o', 1680,
        'p', 990,
        'r', 152,
        's', 793,
        't', 1239,
        'u', 1464,
        'v', 1005,
      );
      END_STATE();
    case 470:
      if (lookahead == 'E') ADVANCE(1743);
      END_STATE();
    case 471:
      if (lookahead == 'E') ADVANCE(1740);
      END_STATE();
    case 472:
      if (lookahead == 'E') ADVANCE(1745);
      if (lookahead == 'F') ADVANCE(2097);
      if (lookahead == 'G') ADVANCE(1890);
      if (lookahead == 'L') ADVANCE(972);
      if (lookahead == 'S') ADVANCE(1458);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 473:
      if (lookahead == 'E') ADVANCE(1746);
      if (lookahead == 'F') ADVANCE(2097);
      if (lookahead == 'G') ADVANCE(1890);
      if (lookahead == 'L') ADVANCE(972);
      if (lookahead == 'S') ADVANCE(1458);
      if (lookahead == 'T') ADVANCE(1318);
      END_STATE();
    case 474:
      if (lookahead == 'F') ADVANCE(517);
      END_STATE();
    case 475:
      if (lookahead == 'F') ADVANCE(2091);
      END_STATE();
    case 476:
      if (lookahead == 'G') ADVANCE(152);
      END_STATE();
    case 477:
      ADVANCE_MAP(
        'G', 1122,
        'L', 1053,
        'R', 1317,
        'V', 454,
        'a', 736,
        'b', 1905,
        'c', 566,
        'd', 579,
        'e', 154,
        'f', 1750,
        'g', 465,
        'h', 404,
        'i', 336,
        'j', 754,
        'l', 399,
        'm', 1222,
        'o', 323,
        'p', 635,
        'r', 408,
        's', 771,
        't', 1142,
        'u', 307,
        'v', 452,
        'w', 407,
      );
      END_STATE();
    case 478:
      if (lookahead == 'G') ADVANCE(1890);
      END_STATE();
    case 479:
      if (lookahead == 'G') ADVANCE(1899);
      if (lookahead == 'L') ADVANCE(1055);
      END_STATE();
    case 480:
      ADVANCE_MAP(
        'H', 430,
        'O', 474,
        'a', 818,
        'c', 208,
        'f', 1750,
        'h', 1655,
        'i', 1140,
        'm', 660,
        'o', 1680,
        'q', 1776,
        's', 778,
        't', 546,
        'u', 720,
      );
      END_STATE();
    case 481:
      ADVANCE_MAP(
        'H', 502,
        'R', 395,
        'S', 486,
        'a', 2163,
        'c', 613,
        'f', 1750,
        'h', 960,
        'i', 1386,
        'o', 1680,
        'r', 1297,
        's', 799,
      );
      END_STATE();
    case 482:
      if (lookahead == 'H') ADVANCE(754);
      END_STATE();
    case 483:
      if (lookahead == 'H') ADVANCE(754);
      if (lookahead == 'J') ADVANCE(754);
      if (lookahead == 'a') ADVANCE(1701);
      if (lookahead == 'c') ADVANCE(981);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 484:
      ADVANCE_MAP(
        'H', 754,
        'O', 503,
        'a', 824,
        'c', 612,
        'd', 1598,
        'e', 914,
        'f', 1750,
        'h', 1223,
        'i', 1797,
        'l', 1602,
        'o', 1442,
        'r', 1604,
        's', 778,
        'u', 1687,
      );
      END_STATE();
    case 485:
      ADVANCE_MAP(
        'H', 754,
        'a', 818,
        'c', 616,
        'd', 1598,
        'e', 1807,
        'f', 1750,
        'o', 1680,
        's', 778,
      );
      END_STATE();
    case 486:
      if (lookahead == 'H') ADVANCE(754);
      if (lookahead == 'c') ADVANCE(2139);
      END_STATE();
    case 487:
      if (lookahead == 'H') ADVANCE(2099);
      END_STATE();
    case 488:
      if (lookahead == 'I') ADVANCE(152);
      END_STATE();
    case 489:
      if (lookahead == 'I') ADVANCE(1498);
      END_STATE();
    case 490:
      if (lookahead == 'I') ADVANCE(1565);
      END_STATE();
    case 491:
      ADVANCE_MAP(
        'J', 754,
        'a', 818,
        'c', 613,
        'e', 1143,
        'f', 1750,
        'o', 428,
        's', 778,
        't', 1292,
        'u', 152,
      );
      END_STATE();
    case 492:
      if (lookahead == 'L') ADVANCE(972);
      END_STATE();
    case 493:
      if (lookahead == 'L') ADVANCE(1272);
      END_STATE();
    case 494:
      if (lookahead == 'L') ADVANCE(1070);
      if (lookahead == 'R') ADVANCE(1325);
      END_STATE();
    case 495:
      if (lookahead == 'L') ADVANCE(1070);
      if (lookahead == 'R') ADVANCE(1325);
      if (lookahead == 'l') ADVANCE(1052);
      if (lookahead == 'r') ADVANCE(1317);
      END_STATE();
    case 496:
      if (lookahead == 'L') ADVANCE(1074);
      if (lookahead == 'R') ADVANCE(1325);
      END_STATE();
    case 497:
      if (lookahead == 'M') ADVANCE(1056);
      if (lookahead == 'T') ADVANCE(1205);
      if (lookahead == 'V') ADVANCE(1044);
      END_STATE();
    case 498:
      if (lookahead == 'M') ADVANCE(1315);
      END_STATE();
    case 499:
      ADVANCE_MAP(
        'N', 1598,
        'a', 808,
        'b', 1793,
        'c', 1619,
        'd', 1739,
        'e', 873,
        'f', 1750,
        'i', 1123,
        'k', 678,
        'l', 559,
        'n', 938,
        'o', 1685,
        'p', 1820,
        'r', 936,
        's', 780,
        'u', 1405,
      );
      END_STATE();
    case 500:
      ADVANCE_MAP(
        'N', 476,
        'T', 181,
        'a', 812,
        'c', 615,
        'd', 1598,
        'f', 1750,
        'g', 1811,
        'l', 1057,
        'm', 544,
        'o', 1138,
        'p', 1954,
        'q', 2051,
        's', 788,
        't', 532,
        'u', 1464,
        'x', 1248,
      );
      END_STATE();
    case 501:
      if (lookahead == 'O') ADVANCE(189);
      END_STATE();
    case 502:
      if (lookahead == 'O') ADVANCE(508);
      END_STATE();
    case 503:
      if (lookahead == 'P') ADVANCE(191);
      END_STATE();
    case 504:
      if (lookahead == 'P') ADVANCE(546);
      END_STATE();
    case 505:
      if (lookahead == 'P') ADVANCE(546);
      if (lookahead == 'i') ADVANCE(368);
      END_STATE();
    case 506:
      if (lookahead == 'P') ADVANCE(1448);
      END_STATE();
    case 507:
      if (lookahead == 'Q') ADVANCE(2098);
      END_STATE();
    case 508:
      if (lookahead == 'R') ADVANCE(186);
      END_STATE();
    case 509:
      if (lookahead == 'R') ADVANCE(450);
      END_STATE();
    case 510:
      if (lookahead == 'R') ADVANCE(1328);
      if (lookahead == 'T') ADVANCE(1039);
      if (lookahead == 'V') ADVANCE(1067);
      END_STATE();
    case 511:
      ADVANCE_MAP(
        'S', 152,
        'a', 815,
        'c', 1270,
        'd', 580,
        'e', 1404,
        'f', 851,
        'g', 1630,
        'h', 727,
        'i', 1516,
        'l', 586,
        'm', 542,
        'o', 1680,
        'p', 547,
        'r', 204,
        's', 794,
        't', 1287,
        'u', 1464,
        'v', 725,
      );
      END_STATE();
    case 512:
      if (lookahead == 'S') ADVANCE(1502);
      END_STATE();
    case 513:
      if (lookahead == 'S') ADVANCE(1502);
      if (lookahead == 'V') ADVANCE(1042);
      END_STATE();
    case 514:
      if (lookahead == 'S') ADVANCE(1720);
      END_STATE();
    case 515:
      if (lookahead == 'S') ADVANCE(2061);
      END_STATE();
    case 516:
      if (lookahead == 'S') ADVANCE(1747);
      END_STATE();
    case 517:
      if (lookahead == 'T') ADVANCE(754);
      END_STATE();
    case 518:
      if (lookahead == 'T') ADVANCE(393);
      END_STATE();
    case 519:
      if (lookahead == 'T') ADVANCE(1219);
      END_STATE();
    case 520:
      if (lookahead == 'T') ADVANCE(1200);
      END_STATE();
    case 521:
      if (lookahead == 'T') ADVANCE(1039);
      if (lookahead == 'V') ADVANCE(1067);
      END_STATE();
    case 522:
      if (lookahead == 'T') ADVANCE(1901);
      END_STATE();
    case 523:
      if (lookahead == 'U') ADVANCE(501);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 524:
      if (lookahead == 'V') ADVANCE(1086);
      END_STATE();
    case 525:
      if (lookahead == 'V') ADVANCE(1072);
      END_STATE();
    case 526:
      if (lookahead == 'W') ADVANCE(1274);
      END_STATE();
    case 527:
      if (lookahead == '[') ADVANCE(2240);
      END_STATE();
    case 528:
      if (lookahead == '_') ADVANCE(528);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          (lookahead < ' ' || '@' < lookahead) &&
          (lookahead < '[' || '`' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(2245);
      END_STATE();
    case 529:
      ADVANCE_MAP(
        'a', 812,
        'b', 1749,
        'c', 289,
        'e', 1398,
        'f', 330,
        'g', 1811,
        'l', 937,
        'm', 197,
        'n', 885,
        'o', 1138,
        'p', 202,
        'r', 1245,
        's', 797,
        't', 1292,
        'u', 1464,
        'w', 833,
      );
      END_STATE();
    case 530:
      ADVANCE_MAP(
        'a', 812,
        'c', 290,
        'e', 761,
        'f', 2167,
        'g', 1811,
        'i', 287,
        'j', 1404,
        'm', 538,
        'n', 239,
        'o', 758,
        'p', 1810,
        'q', 2066,
        's', 790,
        't', 295,
        'u', 1347,
      );
      END_STATE();
    case 531:
      ADVANCE_MAP(
        'a', 804,
        'c', 2139,
        'e', 839,
        'f', 1750,
        'o', 1680,
        'r', 935,
        's', 778,
        'u', 1475,
      );
      END_STATE();
    case 532:
      if (lookahead == 'a') ADVANCE(152);
      END_STATE();
    case 533:
      if (lookahead == 'a') ADVANCE(330);
      END_STATE();
    case 534:
      if (lookahead == 'a') ADVANCE(1739);
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'h') ADVANCE(152);
      if (lookahead == 'i') ADVANCE(1469);
      if (lookahead == 'q') ADVANCE(719);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 535:
      if (lookahead == 'a') ADVANCE(1739);
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'h') ADVANCE(152);
      if (lookahead == 'q') ADVANCE(719);
      END_STATE();
    case 536:
      ADVANCE_MAP(
        'a', 823,
        'c', 561,
        'd', 1598,
        'e', 907,
        'f', 1750,
        'h', 756,
        'i', 1754,
        'l', 2056,
        'o', 1452,
        'r', 592,
        's', 800,
        't', 876,
        'u', 924,
        'w', 833,
        'y', 1395,
      );
      END_STATE();
    case 537:
      if (lookahead == 'a') ADVANCE(2113);
      END_STATE();
    case 538:
      if (lookahead == 'a') ADVANCE(783);
      if (lookahead == 'o') ADVANCE(1096);
      if (lookahead == 'p') ADVANCE(925);
      END_STATE();
    case 539:
      if (lookahead == 'a') ADVANCE(1412);
      if (lookahead == 'e') ADVANCE(1725);
      if (lookahead == 'i') ADVANCE(883);
      if (lookahead == 't') ADVANCE(261);
      END_STATE();
    case 540:
      if (lookahead == 'a') ADVANCE(367);
      END_STATE();
    case 541:
      if (lookahead == 'a') ADVANCE(778);
      END_STATE();
    case 542:
      if (lookahead == 'a') ADVANCE(778);
      if (lookahead == 'e') ADVANCE(1131);
      if (lookahead == 'i') ADVANCE(845);
      END_STATE();
    case 543:
      if (lookahead == 'a') ADVANCE(778);
      if (lookahead == 'e') ADVANCE(1131);
      if (lookahead == 'i') ADVANCE(844);
      END_STATE();
    case 544:
      if (lookahead == 'a') ADVANCE(778);
      if (lookahead == 'p') ADVANCE(1985);
      END_STATE();
    case 545:
      if (lookahead == 'a') ADVANCE(778);
      if (lookahead == 'p') ADVANCE(1993);
      if (lookahead == 's') ADVANCE(1690);
      END_STATE();
    case 546:
      if (lookahead == 'a') ADVANCE(1750);
      END_STATE();
    case 547:
      if (lookahead == 'a') ADVANCE(1750);
      if (lookahead == 'e') ADVANCE(1798);
      if (lookahead == 'l') ADVANCE(2065);
      END_STATE();
    case 548:
      if (lookahead == 'a') ADVANCE(1750);
      if (lookahead == 'f') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(2065);
      END_STATE();
    case 549:
      if (lookahead == 'a') ADVANCE(1750);
      if (lookahead == 'r') ADVANCE(628);
      END_STATE();
    case 550:
      ADVANCE_MAP(
        'a', 814,
        'b', 1772,
        'c', 1270,
        'd', 752,
        'f', 1750,
        'g', 1811,
        'm', 541,
        'n', 895,
        'o', 1138,
        'p', 411,
        'r', 1260,
        's', 778,
        't', 1318,
        'u', 1464,
      );
      END_STATE();
    case 551:
      if (lookahead == 'a') ADVANCE(1516);
      END_STATE();
    case 552:
      if (lookahead == 'a') ADVANCE(358);
      END_STATE();
    case 553:
      ADVANCE_MAP(
        'a', 1875,
        'b', 1804,
        'c', 613,
        'd', 1598,
        'e', 1453,
        'f', 1750,
        'h', 967,
        'i', 1441,
        'o', 958,
        'p', 1820,
        'r', 665,
        's', 769,
        'w', 1227,
      );
      END_STATE();
    case 554:
      ADVANCE_MAP(
        'a', 1526,
        'b', 1804,
        'n', 1129,
        'o', 1726,
        'p', 548,
        't', 1304,
        'w', 596,
        'z', 269,
      );
      END_STATE();
    case 555:
      if (lookahead == 'a') ADVANCE(1526);
      if (lookahead == 'b') ADVANCE(1804);
      if (lookahead == 'p') ADVANCE(548);
      if (lookahead == 't') ADVANCE(1304);
      END_STATE();
    case 556:
      if (lookahead == 'a') ADVANCE(1958);
      END_STATE();
    case 557:
      if (lookahead == 'a') ADVANCE(1958);
      if (lookahead == 'l') ADVANCE(1252);
      if (lookahead == 't') ADVANCE(1532);
      END_STATE();
    case 558:
      if (lookahead == 'a') ADVANCE(867);
      if (lookahead == 'f') ADVANCE(1107);
      END_STATE();
    case 559:
      if (lookahead == 'a') ADVANCE(834);
      if (lookahead == 'k') ADVANCE(144);
      if (lookahead == 'o') ADVANCE(827);
      END_STATE();
    case 560:
      if (lookahead == 'a') ADVANCE(243);
      END_STATE();
    case 561:
      if (lookahead == 'a') ADVANCE(1702);
      if (lookahead == 'e') ADVANCE(906);
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'u') ADVANCE(1710);
      END_STATE();
    case 562:
      if (lookahead == 'a') ADVANCE(918);
      END_STATE();
    case 563:
      if (lookahead == 'a') ADVANCE(338);
      END_STATE();
    case 564:
      if (lookahead == 'a') ADVANCE(2111);
      END_STATE();
    case 565:
      if (lookahead == 'a') ADVANCE(774);
      if (lookahead == 'o') ADVANCE(2127);
      END_STATE();
    case 566:
      if (lookahead == 'a') ADVANCE(1678);
      if (lookahead == 'e') ADVANCE(921);
      if (lookahead == 'o') ADVANCE(1539);
      if (lookahead == 'u') ADVANCE(1677);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 567:
      if (lookahead == 'a') ADVANCE(346);
      END_STATE();
    case 568:
      ADVANCE_MAP(
        'a', 773,
        'c', 1291,
        'e', 309,
        'f', 1750,
        'i', 754,
        'o', 1680,
        's', 778,
        'u', 760,
      );
      END_STATE();
    case 569:
      ADVANCE_MAP(
        'a', 1800,
        'c', 2139,
        'f', 1750,
        'h', 1223,
        'i', 152,
        'l', 2060,
        'o', 1288,
        'r', 262,
        's', 787,
      );
      END_STATE();
    case 570:
      if (lookahead == 'a') ADVANCE(1701);
      END_STATE();
    case 571:
      if (lookahead == 'a') ADVANCE(1713);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'e') ADVANCE(1734);
      if (lookahead == 'l') ADVANCE(972);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 572:
      if (lookahead == 'a') ADVANCE(1677);
      END_STATE();
    case 573:
      ADVANCE_MAP(
        'a', 1677,
        'c', 2139,
        'e', 902,
        'f', 1750,
        'i', 1556,
        'o', 1680,
        's', 778,
        'u', 152,
      );
      END_STATE();
    case 574:
      if (lookahead == 'a') ADVANCE(1677);
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'u') ADVANCE(1677);
      END_STATE();
    case 575:
      if (lookahead == 'a') ADVANCE(1677);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 576:
      if (lookahead == 'a') ADVANCE(1677);
      if (lookahead == 'u') ADVANCE(1677);
      END_STATE();
    case 577:
      if (lookahead == 'a') ADVANCE(1120);
      END_STATE();
    case 578:
      ADVANCE_MAP(
        'a', 1460,
        'c', 2139,
        'e', 1490,
        'f', 1285,
        'i', 1404,
        'j', 1404,
        'l', 557,
        'n', 1590,
        'o', 1681,
        'p', 668,
        'r', 565,
        's', 778,
      );
      END_STATE();
    case 579:
      if (lookahead == 'a') ADVANCE(1924);
      END_STATE();
    case 580:
      if (lookahead == 'a') ADVANCE(1924);
      if (lookahead == 'b') ADVANCE(1438);
      if (lookahead == 'i') ADVANCE(2108);
      if (lookahead == 'o') ADVANCE(1958);
      if (lookahead == 's') ADVANCE(1640);
      END_STATE();
    case 581:
      if (lookahead == 'a') ADVANCE(1339);
      END_STATE();
    case 582:
      if (lookahead == 'a') ADVANCE(1927);
      END_STATE();
    case 583:
      if (lookahead == 'a') ADVANCE(1769);
      END_STATE();
    case 584:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'b') ADVANCE(1804);
      if (lookahead == 'r') ADVANCE(618);
      END_STATE();
    case 585:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'b') ADVANCE(1438);
      if (lookahead == 'h') ADVANCE(546);
      END_STATE();
    case 586:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'c') ADVANCE(1234);
      if (lookahead == 'i') ADVANCE(1538);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 587:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'c') ADVANCE(1649);
      if (lookahead == 'h') ADVANCE(649);
      if (lookahead == 'm') ADVANCE(152);
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 588:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'h') ADVANCE(546);
      END_STATE();
    case 589:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'h') ADVANCE(546);
      if (lookahead == 'm') ADVANCE(152);
      END_STATE();
    case 590:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'm') ADVANCE(297);
      END_STATE();
    case 591:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'm') ADVANCE(2000);
      if (lookahead == 'o') ADVANCE(1342);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'r') ADVANCE(725);
      END_STATE();
    case 592:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'o') ADVANCE(1929);
      END_STATE();
    case 593:
      if (lookahead == 'a') ADVANCE(1769);
      if (lookahead == 'r') ADVANCE(1750);
      if (lookahead == 't') ADVANCE(679);
      END_STATE();
    case 594:
      if (lookahead == 'a') ADVANCE(1467);
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'g') ADVANCE(661);
      if (lookahead == 's') ADVANCE(1267);
      if (lookahead == 'v') ADVANCE(288);
      END_STATE();
    case 595:
      if (lookahead == 'a') ADVANCE(2150);
      END_STATE();
    case 596:
      if (lookahead == 'a') ADVANCE(1914);
      if (lookahead == 'b') ADVANCE(546);
      END_STATE();
    case 597:
      if (lookahead == 'a') ADVANCE(1914);
      if (lookahead == 'c') ADVANCE(1289);
      if (lookahead == 'd') ADVANCE(579);
      if (lookahead == 'R' ||
          lookahead == 'S') ADVANCE(152);
      END_STATE();
    case 598:
      if (lookahead == 'a') ADVANCE(1537);
      END_STATE();
    case 599:
      if (lookahead == 'a') ADVANCE(1997);
      END_STATE();
    case 600:
      if (lookahead == 'a') ADVANCE(1358);
      END_STATE();
    case 601:
      ADVANCE_MAP(
        'a', 1756,
        'c', 2139,
        'e', 1757,
        'f', 1750,
        'h', 1237,
        'i', 352,
        'l', 617,
        'm', 152,
        'o', 1298,
        'r', 168,
        's', 787,
        'u', 1536,
      );
      END_STATE();
    case 602:
      if (lookahead == 'a') ADVANCE(1991);
      if (lookahead == 'e') ADVANCE(1938);
      if (lookahead == 'o') ADVANCE(346);
      END_STATE();
    case 603:
      if (lookahead == 'a') ADVANCE(1182);
      END_STATE();
    case 604:
      if (lookahead == 'a') ADVANCE(1356);
      END_STATE();
    case 605:
      if (lookahead == 'a') ADVANCE(1381);
      if (lookahead == 'l') ADVANCE(1272);
      if (lookahead == 's') ADVANCE(2085);
      END_STATE();
    case 606:
      if (lookahead == 'a') ADVANCE(1804);
      END_STATE();
    case 607:
      if (lookahead == 'a') ADVANCE(1917);
      END_STATE();
    case 608:
      if (lookahead == 'a') ADVANCE(1504);
      END_STATE();
    case 609:
      if (lookahead == 'a') ADVANCE(762);
      END_STATE();
    case 610:
      if (lookahead == 'a') ADVANCE(1538);
      END_STATE();
    case 611:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'e') ADVANCE(922);
      if (lookahead == 'u') ADVANCE(718);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 612:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'e') ADVANCE(906);
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'o') ADVANCE(1566);
      END_STATE();
    case 613:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'e') ADVANCE(921);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 614:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'i') ADVANCE(1780);
      if (lookahead == 'o') ADVANCE(1388);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 615:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'i') ADVANCE(1780);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 616:
      if (lookahead == 'a') ADVANCE(1803);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 617:
      if (lookahead == 'a') ADVANCE(1510);
      if (lookahead == 'u') ADVANCE(1908);
      END_STATE();
    case 618:
      if (lookahead == 'a') ADVANCE(776);
      if (lookahead == 'k') ADVANCE(930);
      END_STATE();
    case 619:
      if (lookahead == 'a') ADVANCE(1699);
      if (lookahead == 'u') ADVANCE(1699);
      END_STATE();
    case 620:
      if (lookahead == 'a') ADVANCE(887);
      END_STATE();
    case 621:
      if (lookahead == 'a') ADVANCE(1375);
      END_STATE();
    case 622:
      if (lookahead == 'a') ADVANCE(1375);
      if (lookahead == 'i') ADVANCE(1420);
      END_STATE();
    case 623:
      if (lookahead == 'a') ADVANCE(1759);
      END_STATE();
    case 624:
      if (lookahead == 'a') ADVANCE(1759);
      if (lookahead == 'b') ADVANCE(1389);
      END_STATE();
    case 625:
      if (lookahead == 'a') ADVANCE(1394);
      END_STATE();
    case 626:
      if (lookahead == 'a') ADVANCE(1394);
      if (lookahead == 'e') ADVANCE(1914);
      if (lookahead == 'i') ADVANCE(2110);
      END_STATE();
    case 627:
      if (lookahead == 'a') ADVANCE(1794);
      if (lookahead == 'l') ADVANCE(2065);
      if (lookahead == 's') ADVANCE(1230);
      END_STATE();
    case 628:
      if (lookahead == 'a') ADVANCE(777);
      END_STATE();
    case 629:
      if (lookahead == 'a') ADVANCE(1789);
      if (lookahead == 'b') ADVANCE(1389);
      END_STATE();
    case 630:
      if (lookahead == 'a') ADVANCE(1789);
      if (lookahead == 'o') ADVANCE(358);
      END_STATE();
    case 631:
      if (lookahead == 'a') ADVANCE(1721);
      END_STATE();
    case 632:
      if (lookahead == 'a') ADVANCE(862);
      END_STATE();
    case 633:
      if (lookahead == 'a') ADVANCE(1787);
      END_STATE();
    case 634:
      if (lookahead == 'a') ADVANCE(1387);
      END_STATE();
    case 635:
      if (lookahead == 'a') ADVANCE(1779);
      if (lookahead == 'o') ADVANCE(1454);
      if (lookahead == 'r') ADVANCE(237);
      END_STATE();
    case 636:
      if (lookahead == 'a') ADVANCE(1385);
      END_STATE();
    case 637:
      if (lookahead == 'a') ADVANCE(1774);
      if (lookahead == 'p') ADVANCE(1673);
      END_STATE();
    case 638:
      if (lookahead == 'a') ADVANCE(1760);
      if (lookahead == 'r') ADVANCE(709);
      END_STATE();
    case 639:
      if (lookahead == 'a') ADVANCE(1365);
      END_STATE();
    case 640:
      if (lookahead == 'a') ADVANCE(1773);
      END_STATE();
    case 641:
      if (lookahead == 'a') ADVANCE(1373);
      END_STATE();
    case 642:
      if (lookahead == 'a') ADVANCE(1360);
      END_STATE();
    case 643:
      if (lookahead == 'a') ADVANCE(1366);
      END_STATE();
    case 644:
      if (lookahead == 'a') ADVANCE(1361);
      END_STATE();
    case 645:
      if (lookahead == 'a') ADVANCE(1827);
      END_STATE();
    case 646:
      if (lookahead == 'a') ADVANCE(1379);
      END_STATE();
    case 647:
      if (lookahead == 'a') ADVANCE(1367);
      END_STATE();
    case 648:
      if (lookahead == 'a') ADVANCE(1748);
      END_STATE();
    case 649:
      if (lookahead == 'a') ADVANCE(1777);
      END_STATE();
    case 650:
      if (lookahead == 'a') ADVANCE(1753);
      END_STATE();
    case 651:
      if (lookahead == 'a') ADVANCE(1831);
      END_STATE();
    case 652:
      if (lookahead == 'a') ADVANCE(1775);
      END_STATE();
    case 653:
      if (lookahead == 'a') ADVANCE(2015);
      END_STATE();
    case 654:
      if (lookahead == 'a') ADVANCE(1842);
      END_STATE();
    case 655:
      if (lookahead == 'a') ADVANCE(1815);
      END_STATE();
    case 656:
      if (lookahead == 'a') ADVANCE(1806);
      END_STATE();
    case 657:
      if (lookahead == 'a') ADVANCE(1812);
      END_STATE();
    case 658:
      ADVANCE_MAP(
        'a', 818,
        'b', 1739,
        'c', 198,
        'd', 1627,
        'e', 405,
        'f', 1770,
        'h', 213,
        'i', 1159,
        'l', 583,
        'm', 539,
        'o', 1104,
        'p', 562,
        'q', 806,
        'r', 583,
        's', 781,
        't', 638,
        'u', 721,
        'w', 406,
        'z', 1398,
      );
      END_STATE();
    case 659:
      ADVANCE_MAP(
        'a', 818,
        'c', 616,
        'd', 1598,
        'e', 1063,
        'f', 1750,
        'h', 754,
        'i', 1153,
        'o', 1680,
        's', 778,
        'w', 1338,
      );
      END_STATE();
    case 660:
      if (lookahead == 'a') ADVANCE(1410);
      END_STATE();
    case 661:
      if (lookahead == 'a') ADVANCE(1479);
      END_STATE();
    case 662:
      if (lookahead == 'a') ADVANCE(1936);
      END_STATE();
    case 663:
      ADVANCE_MAP(
        'a', 1705,
        'c', 981,
        'f', 1750,
        'g', 1814,
        'h', 754,
        'j', 754,
        'o', 1680,
        's', 778,
      );
      END_STATE();
    case 664:
      if (lookahead == 'a') ADVANCE(856);
      END_STATE();
    case 665:
      if (lookahead == 'a') ADVANCE(903);
      if (lookahead == 'i') ADVANCE(710);
      if (lookahead == 'p') ADVANCE(965);
      END_STATE();
    case 666:
      if (lookahead == 'a') ADVANCE(1863);
      END_STATE();
    case 667:
      if (lookahead == 'a') ADVANCE(2012);
      END_STATE();
    case 668:
      if (lookahead == 'a') ADVANCE(1830);
      END_STATE();
    case 669:
      if (lookahead == 'a') ADVANCE(1712);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'e') ADVANCE(1735);
      if (lookahead == 'g') ADVANCE(1978);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 670:
      if (lookahead == 'a') ADVANCE(1712);
      if (lookahead == 'e') ADVANCE(1738);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 671:
      if (lookahead == 'a') ADVANCE(1712);
      if (lookahead == 's') ADVANCE(1255);
      END_STATE();
    case 672:
      if (lookahead == 'a') ADVANCE(842);
      END_STATE();
    case 673:
      if (lookahead == 'a') ADVANCE(1864);
      if (lookahead == 'l') ADVANCE(2141);
      if (lookahead == 'r') ADVANCE(934);
      if (lookahead == 'v') ADVANCE(1083);
      END_STATE();
    case 674:
      if (lookahead == 'a') ADVANCE(1549);
      END_STATE();
    case 675:
      if (lookahead == 'a') ADVANCE(1400);
      if (lookahead == 'k') ADVANCE(358);
      END_STATE();
    case 676:
      if (lookahead == 'a') ADVANCE(2022);
      END_STATE();
    case 677:
      if (lookahead == 'a') ADVANCE(1876);
      if (lookahead == 'l') ADVANCE(1415);
      if (lookahead == 'r') ADVANCE(826);
      END_STATE();
    case 678:
      if (lookahead == 'a') ADVANCE(1826);
      END_STATE();
    case 679:
      if (lookahead == 'a') ADVANCE(1263);
      END_STATE();
    case 680:
      if (lookahead == 'a') ADVANCE(1263);
      if (lookahead == 'i') ADVANCE(1600);
      END_STATE();
    case 681:
      if (lookahead == 'a') ADVANCE(2076);
      END_STATE();
    case 682:
      if (lookahead == 'a') ADVANCE(1422);
      END_STATE();
    case 683:
      if (lookahead == 'a') ADVANCE(1867);
      END_STATE();
    case 684:
      if (lookahead == 'a') ADVANCE(2078);
      END_STATE();
    case 685:
      if (lookahead == 'a') ADVANCE(2030);
      END_STATE();
    case 686:
      if (lookahead == 'a') ADVANCE(1550);
      END_STATE();
    case 687:
      if (lookahead == 'a') ADVANCE(1869);
      END_STATE();
    case 688:
      if (lookahead == 'a') ADVANCE(1407);
      END_STATE();
    case 689:
      if (lookahead == 'a') ADVANCE(1859);
      END_STATE();
    case 690:
      if (lookahead == 'a') ADVANCE(1818);
      END_STATE();
    case 691:
      if (lookahead == 'a') ADVANCE(1865);
      END_STATE();
    case 692:
      if (lookahead == 'a') ADVANCE(1866);
      END_STATE();
    case 693:
      if (lookahead == 'a') ADVANCE(1457);
      END_STATE();
    case 694:
      if (lookahead == 'a') ADVANCE(1881);
      END_STATE();
    case 695:
      if (lookahead == 'a') ADVANCE(1881);
      if (lookahead == 'd') ADVANCE(1659);
      if (lookahead == 'h') ADVANCE(651);
      if (lookahead == 'l') ADVANCE(2065);
      if (lookahead == 's') ADVANCE(1232);
      if (lookahead == 'u') ADVANCE(1731);
      END_STATE();
    case 696:
      if (lookahead == 'a') ADVANCE(1881);
      if (lookahead == 'd') ADVANCE(1675);
      if (lookahead == 'h') ADVANCE(651);
      END_STATE();
    case 697:
      if (lookahead == 'a') ADVANCE(1881);
      if (lookahead == 'r') ADVANCE(1317);
      END_STATE();
    case 698:
      if (lookahead == 'a') ADVANCE(1892);
      END_STATE();
    case 699:
      if (lookahead == 'a') ADVANCE(2038);
      END_STATE();
    case 700:
      if (lookahead == 'a') ADVANCE(1886);
      if (lookahead == 'h') ADVANCE(711);
      if (lookahead == 'l') ADVANCE(1075);
      if (lookahead == 'r') ADVANCE(1332);
      if (lookahead == 's') ADVANCE(1742);
      if (lookahead == 't') ADVANCE(1220);
      END_STATE();
    case 701:
      if (lookahead == 'a') ADVANCE(1886);
      if (lookahead == 'h') ADVANCE(711);
      if (lookahead == 'l') ADVANCE(1095);
      if (lookahead == 'r') ADVANCE(1326);
      if (lookahead == 't') ADVANCE(1220);
      END_STATE();
    case 702:
      if (lookahead == 'a') ADVANCE(2040);
      END_STATE();
    case 703:
      if (lookahead == 'a') ADVANCE(1887);
      END_STATE();
    case 704:
      if (lookahead == 'a') ADVANCE(1887);
      if (lookahead == 'd') ADVANCE(597);
      END_STATE();
    case 705:
      if (lookahead == 'a') ADVANCE(1888);
      END_STATE();
    case 706:
      if (lookahead == 'a') ADVANCE(1888);
      if (lookahead == 'h') ADVANCE(713);
      END_STATE();
    case 707:
      if (lookahead == 'a') ADVANCE(2042);
      END_STATE();
    case 708:
      if (lookahead == 'a') ADVANCE(1889);
      if (lookahead == 'h') ADVANCE(713);
      if (lookahead == 's') ADVANCE(1742);
      END_STATE();
    case 709:
      if (lookahead == 'a') ADVANCE(1322);
      if (lookahead == 'n') ADVANCE(1903);
      END_STATE();
    case 710:
      if (lookahead == 'a') ADVANCE(1571);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'm') ADVANCE(1315);
      if (lookahead == 'p') ADVANCE(1448);
      if (lookahead == 's') ADVANCE(718);
      if (lookahead == 't') ADVANCE(1300);
      END_STATE();
    case 711:
      if (lookahead == 'a') ADVANCE(1894);
      END_STATE();
    case 712:
      if (lookahead == 'a') ADVANCE(1578);
      END_STATE();
    case 713:
      if (lookahead == 'a') ADVANCE(1896);
      END_STATE();
    case 714:
      if (lookahead == 'a') ADVANCE(1579);
      END_STATE();
    case 715:
      if (lookahead == 'a') ADVANCE(1581);
      END_STATE();
    case 716:
      if (lookahead == 'a') ADVANCE(1582);
      END_STATE();
    case 717:
      if (lookahead == 'a') ADVANCE(1583);
      END_STATE();
    case 718:
      if (lookahead == 'b') ADVANCE(152);
      END_STATE();
    case 719:
      if (lookahead == 'b') ADVANCE(152);
      if (lookahead == 'u') ADVANCE(1589);
      END_STATE();
    case 720:
      if (lookahead == 'b') ADVANCE(343);
      if (lookahead == 'c') ADVANCE(864);
      if (lookahead == 'm') ADVANCE(152);
      if (lookahead == 'p') ADVANCE(270);
      END_STATE();
    case 721:
      if (lookahead == 'b') ADVANCE(169);
      if (lookahead == 'c') ADVANCE(775);
      if (lookahead == 'm') ADVANCE(152);
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 'p') ADVANCE(245);
      END_STATE();
    case 722:
      if (lookahead == 'b') ADVANCE(344);
      if (lookahead == 'c') ADVANCE(831);
      if (lookahead == 'p') ADVANCE(344);
      END_STATE();
    case 723:
      if (lookahead == 'b') ADVANCE(243);
      END_STATE();
    case 724:
      if (lookahead == 'b') ADVANCE(252);
      if (lookahead == 'p') ADVANCE(252);
      END_STATE();
    case 725:
      if (lookahead == 'b') ADVANCE(546);
      END_STATE();
    case 726:
      if (lookahead == 'b') ADVANCE(546);
      if (lookahead == 'g') ADVANCE(969);
      END_STATE();
    case 727:
      if (lookahead == 'b') ADVANCE(546);
      if (lookahead == 'm') ADVANCE(152);
      END_STATE();
    case 728:
      if (lookahead == 'b') ADVANCE(546);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 729:
      if (lookahead == 'b') ADVANCE(546);
      if (lookahead == 't') ADVANCE(293);
      if (lookahead == 'y') ADVANCE(519);
      END_STATE();
    case 730:
      if (lookahead == 'b') ADVANCE(1354);
      if (lookahead == 'c') ADVANCE(1648);
      END_STATE();
    case 731:
      if (lookahead == 'b') ADVANCE(250);
      if (lookahead == 'p') ADVANCE(250);
      END_STATE();
    case 732:
      if (lookahead == 'b') ADVANCE(2094);
      if (lookahead == 'p') ADVANCE(1215);
      END_STATE();
    case 733:
      if (lookahead == 'b') ADVANCE(891);
      END_STATE();
    case 734:
      if (lookahead == 'b') ADVANCE(1804);
      END_STATE();
    case 735:
      if (lookahead == 'b') ADVANCE(926);
      if (lookahead == 'p') ADVANCE(926);
      END_STATE();
    case 736:
      if (lookahead == 'b') ADVANCE(1368);
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 'p') ADVANCE(172);
      if (lookahead == 't') ADVANCE(2084);
      END_STATE();
    case 737:
      if (lookahead == 'b') ADVANCE(1909);
      END_STATE();
    case 738:
      if (lookahead == 'b') ADVANCE(533);
      END_STATE();
    case 739:
      if (lookahead == 'b') ADVANCE(1078);
      END_STATE();
    case 740:
      if (lookahead == 'b') ADVANCE(1511);
      if (lookahead == 'p') ADVANCE(1511);
      END_STATE();
    case 741:
      if (lookahead == 'b') ADVANCE(1953);
      if (lookahead == 'c') ADVANCE(863);
      if (lookahead == 'p') ADVANCE(1030);
      END_STATE();
    case 742:
      if (lookahead == 'b') ADVANCE(1953);
      if (lookahead == 'p') ADVANCE(1030);
      END_STATE();
    case 743:
      if (lookahead == 'b') ADVANCE(1424);
      END_STATE();
    case 744:
      if (lookahead == 'b') ADVANCE(1427);
      END_STATE();
    case 745:
      if (lookahead == 'b') ADVANCE(1870);
      END_STATE();
    case 746:
      if (lookahead == 'b') ADVANCE(1430);
      END_STATE();
    case 747:
      if (lookahead == 'b') ADVANCE(1431);
      END_STATE();
    case 748:
      if (lookahead == 'b') ADVANCE(1445);
      END_STATE();
    case 749:
      if (lookahead == 'b') ADVANCE(657);
      END_STATE();
    case 750:
      if (lookahead == 'b') ADVANCE(1436);
      END_STATE();
    case 751:
      if (lookahead == 'b') ADVANCE(1437);
      END_STATE();
    case 752:
      if (lookahead == 'b') ADVANCE(1438);
      END_STATE();
    case 753:
      if (lookahead == 'b') ADVANCE(1956);
      if (lookahead == 'p') ADVANCE(1956);
      END_STATE();
    case 754:
      if (lookahead == 'c') ADVANCE(2139);
      END_STATE();
    case 755:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'e') ADVANCE(2113);
      END_STATE();
    case 756:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'e') ADVANCE(836);
      if (lookahead == 'i') ADVANCE(152);
      END_STATE();
    case 757:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(1449);
      if (lookahead == 'o') ADVANCE(1682);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 758:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'g') ADVANCE(1629);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 759:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'i') ADVANCE(1153);
      END_STATE();
    case 760:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'm') ADVANCE(297);
      END_STATE();
    case 761:
      if (lookahead == 'c') ADVANCE(2139);
      if (lookahead == 'x') ADVANCE(765);
      END_STATE();
    case 762:
      if (lookahead == 'c') ADVANCE(152);
      END_STATE();
    case 763:
      if (lookahead == 'c') ADVANCE(152);
      if (lookahead == 'i') ADVANCE(1750);
      END_STATE();
    case 764:
      if (lookahead == 'c') ADVANCE(330);
      if (lookahead == 'l') ADVANCE(931);
      if (lookahead == 'p') ADVANCE(341);
      if (lookahead == 'r') ADVANCE(1351);
      END_STATE();
    case 765:
      if (lookahead == 'c') ADVANCE(297);
      END_STATE();
    case 766:
      ADVANCE_MAP(
        'c', 574,
        'd', 2016,
        'f', 1750,
        'h', 403,
        'i', 152,
        'l', 403,
        'm', 572,
        'n', 1259,
        'o', 878,
        'r', 403,
        's', 795,
        'u', 1709,
        'v', 1003,
        'w', 966,
      );
      END_STATE();
    case 767:
      if (lookahead == 'c') ADVANCE(574);
      if (lookahead == 'o') ADVANCE(879);
      if (lookahead == 's') ADVANCE(1744);
      if (lookahead == 't') ADVANCE(1900);
      if (lookahead == 'u') ADVANCE(1708);
      if (lookahead == 'v') ADVANCE(1003);
      if (lookahead == 'w') ADVANCE(966);
      END_STATE();
    case 768:
      if (lookahead == 'c') ADVANCE(1648);
      END_STATE();
    case 769:
      if (lookahead == 'c') ADVANCE(2170);
      if (lookahead == 'h') ADVANCE(754);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 770:
      if (lookahead == 'c') ADVANCE(2170);
      if (lookahead == 'o') ADVANCE(1358);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 771:
      if (lookahead == 'c') ADVANCE(238);
      if (lookahead == 'h') ADVANCE(1666);
      if (lookahead == 'i') ADVANCE(1480);
      if (lookahead == 'm') ADVANCE(1222);
      if (lookahead == 'p') ADVANCE(546);
      if (lookahead == 'q') ADVANCE(1937);
      if (lookahead == 'u') ADVANCE(722);
      END_STATE();
    case 772:
      if (lookahead == 'c') ADVANCE(932);
      if (lookahead == 'd') ADVANCE(1269);
      if (lookahead == 'e') ADVANCE(1481);
      if (lookahead == 'n') ADVANCE(1130);
      if (lookahead == 'q') ADVANCE(2054);
      if (lookahead == 'r') ADVANCE(1763);
      if (lookahead == 't') ADVANCE(680);
      END_STATE();
    case 773:
      if (lookahead == 'c') ADVANCE(2063);
      END_STATE();
    case 774:
      if (lookahead == 'c') ADVANCE(146);
      if (lookahead == 's') ADVANCE(1358);
      END_STATE();
    case 775:
      if (lookahead == 'c') ADVANCE(216);
      END_STATE();
    case 776:
      if (lookahead == 'c') ADVANCE(2165);
      END_STATE();
    case 777:
      if (lookahead == 'c') ADVANCE(927);
      END_STATE();
    case 778:
      if (lookahead == 'c') ADVANCE(1750);
      END_STATE();
    case 779:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'i') ADVANCE(1463);
      END_STATE();
    case 780:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'e') ADVANCE(1472);
      if (lookahead == 'i') ADVANCE(1473);
      if (lookahead == 'o') ADVANCE(1362);
      END_STATE();
    case 781:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'e') ADVANCE(2006);
      if (lookahead == 'm') ADVANCE(1312);
      if (lookahead == 't') ADVANCE(650);
      END_STATE();
    case 782:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'e') ADVANCE(1783);
      END_STATE();
    case 783:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'g') ADVANCE(928);
      if (lookahead == 't') ADVANCE(1183);
      END_STATE();
    case 784:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'g') ADVANCE(1282);
      END_STATE();
    case 785:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'h') ADVANCE(152);
      END_STATE();
    case 786:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'h') ADVANCE(152);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 787:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(152);
      END_STATE();
    case 788:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(1463);
      END_STATE();
    case 789:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(1468);
      END_STATE();
    case 790:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(1509);
      END_STATE();
    case 791:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'l') ADVANCE(579);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 792:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'l') ADVANCE(1120);
      END_STATE();
    case 793:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'l') ADVANCE(607);
      END_STATE();
    case 794:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'l') ADVANCE(607);
      if (lookahead == 'o') ADVANCE(1358);
      END_STATE();
    case 795:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'q') ADVANCE(847);
      END_STATE();
    case 796:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(1265);
      END_STATE();
    case 797:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 't') ADVANCE(152);
      if (lookahead == 'y') ADVANCE(1488);
      END_STATE();
    case 798:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 't') ADVANCE(1718);
      END_STATE();
    case 799:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 't') ADVANCE(1821);
      END_STATE();
    case 800:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'u') ADVANCE(731);
      END_STATE();
    case 801:
      if (lookahead == 'c') ADVANCE(1750);
      if (lookahead == 'u') ADVANCE(740);
      END_STATE();
    case 802:
      if (lookahead == 'c') ADVANCE(532);
      if (lookahead == 'l') ADVANCE(912);
      if (lookahead == 'q') ADVANCE(2074);
      if (lookahead == 's') ADVANCE(1183);
      END_STATE();
    case 803:
      if (lookahead == 'c') ADVANCE(532);
      if (lookahead == 'q') ADVANCE(2074);
      if (lookahead == 'r') ADVANCE(913);
      if (lookahead == 's') ADVANCE(1183);
      END_STATE();
    case 804:
      if (lookahead == 'c') ADVANCE(1350);
      if (lookahead == 'r') ADVANCE(2109);
      END_STATE();
    case 805:
      if (lookahead == 'c') ADVANCE(1516);
      if (lookahead == 'i') ADVANCE(1587);
      if (lookahead == 'm') ADVANCE(1263);
      if (lookahead == 'p') ADVANCE(152);
      if (lookahead == 't') ADVANCE(1019);
      END_STATE();
    case 806:
      if (lookahead == 'c') ADVANCE(619);
      if (lookahead == 's') ADVANCE(2055);
      if (lookahead == 'u') ADVANCE(215);
      END_STATE();
    case 807:
      if (lookahead == 'c') ADVANCE(1958);
      END_STATE();
    case 808:
      if (lookahead == 'c') ADVANCE(1341);
      if (lookahead == 'r') ADVANCE(2112);
      END_STATE();
    case 809:
      if (lookahead == 'c') ADVANCE(1290);
      if (lookahead == 's') ADVANCE(1257);
      if (lookahead == 'u') ADVANCE(626);
      if (lookahead == 'v') ADVANCE(1725);
      END_STATE();
    case 810:
      if (lookahead == 'c') ADVANCE(1289);
      if (lookahead == 'e') ADVANCE(886);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 'p') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(266);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 811:
      if (lookahead == 'c') ADVANCE(1289);
      if (lookahead == 'e') ADVANCE(915);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 812:
      if (lookahead == 'c') ADVANCE(2062);
      END_STATE();
    case 813:
      if (lookahead == 'c') ADVANCE(2062);
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 814:
      if (lookahead == 'c') ADVANCE(2062);
      if (lookahead == 'r') ADVANCE(1786);
      END_STATE();
    case 815:
      if (lookahead == 'c') ADVANCE(2062);
      if (lookahead == 's') ADVANCE(1958);
      END_STATE();
    case 816:
      if (lookahead == 'c') ADVANCE(2062);
      if (lookahead == 's') ADVANCE(2012);
      END_STATE();
    case 817:
      if (lookahead == 'c') ADVANCE(1677);
      if (lookahead == 'd') ADVANCE(1750);
      END_STATE();
    case 818:
      if (lookahead == 'c') ADVANCE(2095);
      END_STATE();
    case 819:
      ADVANCE_MAP(
        'c', 2095,
        'e', 1481,
        'g', 1868,
        'm', 733,
        'n', 1128,
        'p', 152,
        'q', 2054,
        'r', 1761,
        't', 217,
      );
      END_STATE();
    case 820:
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'm') ADVANCE(733);
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 'p') ADVANCE(1447);
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 821:
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'm') ADVANCE(1491);
      if (lookahead == 'p') ADVANCE(152);
      END_STATE();
    case 822:
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 'r') ADVANCE(1796);
      END_STATE();
    case 823:
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'p') ADVANCE(210);
      if (lookahead == 'r') ADVANCE(962);
      END_STATE();
    case 824:
      if (lookahead == 'c') ADVANCE(2095);
      if (lookahead == 'p') ADVANCE(292);
      if (lookahead == 'y') ADVANCE(1408);
      END_STATE();
    case 825:
      if (lookahead == 'c') ADVANCE(989);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 826:
      if (lookahead == 'c') ADVANCE(1629);
      END_STATE();
    case 827:
      if (lookahead == 'c') ADVANCE(1339);
      END_STATE();
    case 828:
      if (lookahead == 'c') ADVANCE(1348);
      END_STATE();
    case 829:
      if (lookahead == 'c') ADVANCE(1348);
      if (lookahead == 's') ADVANCE(963);
      END_STATE();
    case 830:
      if (lookahead == 'c') ADVANCE(1340);
      if (lookahead == 'n') ADVANCE(514);
      END_STATE();
    case 831:
      if (lookahead == 'c') ADVANCE(259);
      END_STATE();
    case 832:
      if (lookahead == 'c') ADVANCE(1216);
      END_STATE();
    case 833:
      if (lookahead == 'c') ADVANCE(1609);
      if (lookahead == 'i') ADVANCE(1516);
      END_STATE();
    case 834:
      if (lookahead == 'c') ADVANCE(1343);
      if (lookahead == 'n') ADVANCE(1339);
      END_STATE();
    case 835:
      if (lookahead == 'c') ADVANCE(572);
      END_STATE();
    case 836:
      if (lookahead == 'c') ADVANCE(1344);
      END_STATE();
    case 837:
      if (lookahead == 'c') ADVANCE(1358);
      if (lookahead == 'i') ADVANCE(1914);
      if (lookahead == 'p') ADVANCE(1058);
      END_STATE();
    case 838:
      if (lookahead == 'c') ADVANCE(1353);
      if (lookahead == 'n') ADVANCE(1923);
      END_STATE();
    case 839:
      if (lookahead == 'c') ADVANCE(681);
      if (lookahead == 'r') ADVANCE(1541);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 840:
      if (lookahead == 'c') ADVANCE(1345);
      if (lookahead == 'k') ADVANCE(2108);
      END_STATE();
    case 841:
      if (lookahead == 'c') ADVANCE(762);
      END_STATE();
    case 842:
      if (lookahead == 'c') ADVANCE(926);
      END_STATE();
    case 843:
      if (lookahead == 'c') ADVANCE(1197);
      END_STATE();
    case 844:
      if (lookahead == 'c') ADVANCE(1803);
      END_STATE();
    case 845:
      if (lookahead == 'c') ADVANCE(1803);
      if (lookahead == 'd') ADVANCE(152);
      if (lookahead == 'n') ADVANCE(2065);
      END_STATE();
    case 846:
      if (lookahead == 'c') ADVANCE(1923);
      END_STATE();
    case 847:
      if (lookahead == 'c') ADVANCE(2058);
      END_STATE();
    case 848:
      if (lookahead == 'c') ADVANCE(1957);
      END_STATE();
    case 849:
      if (lookahead == 'c') ADVANCE(600);
      END_STATE();
    case 850:
      if (lookahead == 'c') ADVANCE(1233);
      END_STATE();
    case 851:
      if (lookahead == 'c') ADVANCE(1233);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 852:
      if (lookahead == 'c') ADVANCE(1066);
      END_STATE();
    case 853:
      if (lookahead == 'c') ADVANCE(1792);
      if (lookahead == 'd') ADVANCE(206);
      if (lookahead == 'n') ADVANCE(2073);
      END_STATE();
    case 854:
      if (lookahead == 'c') ADVANCE(2044);
      END_STATE();
    case 855:
      if (lookahead == 'c') ADVANCE(2031);
      END_STATE();
    case 856:
      if (lookahead == 'c') ADVANCE(1062);
      END_STATE();
    case 857:
      if (lookahead == 'c') ADVANCE(2015);
      END_STATE();
    case 858:
      if (lookahead == 'c') ADVANCE(1291);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'm') ADVANCE(599);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(782);
      if (lookahead == 'u') ADVANCE(1346);
      END_STATE();
    case 859:
      if (lookahead == 'c') ADVANCE(1291);
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(782);
      if (lookahead == 'u') ADVANCE(1346);
      END_STATE();
    case 860:
      if (lookahead == 'c') ADVANCE(1650);
      if (lookahead == 'i') ADVANCE(1525);
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 861:
      if (lookahead == 'c') ADVANCE(1650);
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 862:
      if (lookahead == 'c') ADVANCE(1352);
      END_STATE();
    case 863:
      if (lookahead == 'c') ADVANCE(1032);
      END_STATE();
    case 864:
      if (lookahead == 'c') ADVANCE(1032);
      if (lookahead == 'h') ADVANCE(520);
      END_STATE();
    case 865:
      if (lookahead == 'c') ADVANCE(639);
      END_STATE();
    case 866:
      if (lookahead == 'c') ADVANCE(1423);
      END_STATE();
    case 867:
      if (lookahead == 'c') ADVANCE(1847);
      if (lookahead == 'm') ADVANCE(1635);
      END_STATE();
    case 868:
      if (lookahead == 'c') ADVANCE(643);
      END_STATE();
    case 869:
      if (lookahead == 'c') ADVANCE(2030);
      END_STATE();
    case 870:
      if (lookahead == 'c') ADVANCE(1618);
      if (lookahead == 'e') ADVANCE(1676);
      if (lookahead == 'p') ADVANCE(1820);
      if (lookahead == 's') ADVANCE(1275);
      END_STATE();
    case 871:
      if (lookahead == 'c') ADVANCE(1407);
      END_STATE();
    case 872:
      if (lookahead == 'c') ADVANCE(646);
      END_STATE();
    case 873:
      if (lookahead == 'c') ADVANCE(684);
      if (lookahead == 'm') ADVANCE(1711);
      if (lookahead == 'p') ADVANCE(1918);
      if (lookahead == 'r') ADVANCE(1564);
      if (lookahead == 't') ADVANCE(2126);
      END_STATE();
    case 874:
      if (lookahead == 'c') ADVANCE(691);
      END_STATE();
    case 875:
      if (lookahead == 'c') ADVANCE(1090);
      END_STATE();
    case 876:
      if (lookahead == 'd') ADVANCE(1598);
      END_STATE();
    case 877:
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'i') ADVANCE(1440);
      if (lookahead == 'r') ADVANCE(1229);
      END_STATE();
    case 878:
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'p') ADVANCE(1097);
      if (lookahead == 't') ADVANCE(1300);
      END_STATE();
    case 879:
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'p') ADVANCE(1448);
      if (lookahead == 't') ADVANCE(1304);
      END_STATE();
    case 880:
      if (lookahead == 'd') ADVANCE(1598);
      if (lookahead == 'r') ADVANCE(1229);
      END_STATE();
    case 881:
      if (lookahead == 'd') ADVANCE(152);
      END_STATE();
    case 882:
      if (lookahead == 'd') ADVANCE(152);
      if (lookahead == 'f') ADVANCE(605);
      if (lookahead == 'p') ADVANCE(348);
      END_STATE();
    case 883:
      if (lookahead == 'd') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(926);
      END_STATE();
    case 884:
      if (lookahead == 'd') ADVANCE(152);
      if (lookahead == 'u') ADVANCE(297);
      END_STATE();
    case 885:
      if (lookahead == 'd') ADVANCE(211);
      if (lookahead == 'g') ADVANCE(251);
      END_STATE();
    case 886:
      if (lookahead == 'd') ADVANCE(726);
      if (lookahead == 'i') ADVANCE(1021);
      END_STATE();
    case 887:
      if (lookahead == 'd') ADVANCE(1444);
      END_STATE();
    case 888:
      if (lookahead == 'd') ADVANCE(513);
      END_STATE();
    case 889:
      if (lookahead == 'd') ADVANCE(489);
      END_STATE();
    case 890:
      if (lookahead == 'd') ADVANCE(479);
      END_STATE();
    case 891:
      if (lookahead == 'd') ADVANCE(532);
      END_STATE();
    case 892:
      if (lookahead == 'd') ADVANCE(754);
      if (lookahead == 'r') ADVANCE(236);
      END_STATE();
    case 893:
      if (lookahead == 'd') ADVANCE(283);
      END_STATE();
    case 894:
      if (lookahead == 'd') ADVANCE(194);
      END_STATE();
    case 895:
      if (lookahead == 'd') ADVANCE(1005);
      if (lookahead == 'i') ADVANCE(1633);
      END_STATE();
    case 896:
      if (lookahead == 'd') ADVANCE(250);
      END_STATE();
    case 897:
      if (lookahead == 'd') ADVANCE(579);
      END_STATE();
    case 898:
      if (lookahead == 'd') ADVANCE(598);
      END_STATE();
    case 899:
      if (lookahead == 'd') ADVANCE(345);
      END_STATE();
    case 900:
      if (lookahead == 'd') ADVANCE(263);
      END_STATE();
    case 901:
      if (lookahead == 'd') ADVANCE(1266);
      END_STATE();
    case 902:
      if (lookahead == 'd') ADVANCE(1266);
      if (lookahead == 'l') ADVANCE(1456);
      END_STATE();
    case 903:
      if (lookahead == 'd') ADVANCE(926);
      END_STATE();
    case 904:
      if (lookahead == 'd') ADVANCE(2070);
      END_STATE();
    case 905:
      if (lookahead == 'd') ADVANCE(2070);
      if (lookahead == 'p') ADVANCE(1672);
      END_STATE();
    case 906:
      if (lookahead == 'd') ADVANCE(1224);
      END_STATE();
    case 907:
      if (lookahead == 'd') ADVANCE(1224);
      if (lookahead == 'm') ADVANCE(1711);
      if (lookahead == 'n') ADVANCE(349);
      END_STATE();
    case 908:
      if (lookahead == 'd') ADVANCE(1949);
      if (lookahead == 'u') ADVANCE(1193);
      END_STATE();
    case 909:
      if (lookahead == 'd') ADVANCE(940);
      END_STATE();
    case 910:
      if (lookahead == 'd') ADVANCE(1912);
      END_STATE();
    case 911:
      if (lookahead == 'd') ADVANCE(2005);
      END_STATE();
    case 912:
      if (lookahead == 'd') ADVANCE(1193);
      END_STATE();
    case 913:
      if (lookahead == 'd') ADVANCE(1193);
      if (lookahead == 'u') ADVANCE(1949);
      END_STATE();
    case 914:
      if (lookahead == 'd') ADVANCE(1299);
      if (lookahead == 'n') ADVANCE(2034);
      END_STATE();
    case 915:
      if (lookahead == 'd') ADVANCE(1145);
      END_STATE();
    case 916:
      if (lookahead == 'd') ADVANCE(1591);
      if (lookahead == 'u') ADVANCE(1677);
      END_STATE();
    case 917:
      if (lookahead == 'd') ADVANCE(1010);
      if (lookahead == 'p') ADVANCE(1096);
      END_STATE();
    case 918:
      if (lookahead == 'd') ADVANCE(1006);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 919:
      if (lookahead == 'd') ADVANCE(1023);
      END_STATE();
    case 920:
      if (lookahead == 'd') ADVANCE(1614);
      END_STATE();
    case 921:
      if (lookahead == 'd') ADVANCE(1263);
      END_STATE();
    case 922:
      if (lookahead == 'd') ADVANCE(1263);
      if (lookahead == 'i') ADVANCE(1358);
      END_STATE();
    case 923:
      if (lookahead == 'd') ADVANCE(1025);
      END_STATE();
    case 924:
      if (lookahead == 'd') ADVANCE(645);
      if (lookahead == 'e') ADVANCE(1693);
      if (lookahead == 'l') ADVANCE(666);
      if (lookahead == 'p') ADVANCE(226);
      if (lookahead == 'r') ADVANCE(673);
      if (lookahead == 'v') ADVANCE(1003);
      if (lookahead == 'w') ADVANCE(925);
      END_STATE();
    case 925:
      if (lookahead == 'e') ADVANCE(881);
      END_STATE();
    case 926:
      if (lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 927:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'k') ADVANCE(961);
      END_STATE();
    case 928:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(1272);
      if (lookahead == 'p') ADVANCE(652);
      END_STATE();
    case 929:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(729);
      END_STATE();
    case 930:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 's') ADVANCE(1364);
      END_STATE();
    case 931:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 't') ADVANCE(268);
      END_STATE();
    case 932:
      if (lookahead == 'e') ADVANCE(152);
      if (lookahead == 'u') ADVANCE(2007);
      END_STATE();
    case 933:
      if (lookahead == 'e') ADVANCE(439);
      END_STATE();
    case 934:
      if (lookahead == 'e') ADVANCE(309);
      END_STATE();
    case 935:
      if (lookahead == 'e') ADVANCE(2113);
      END_STATE();
    case 936:
      if (lookahead == 'e') ADVANCE(2113);
      if (lookahead == 'v') ADVANCE(738);
      END_STATE();
    case 937:
      if (lookahead == 'e') ADVANCE(1105);
      if (lookahead == 'p') ADVANCE(1186);
      END_STATE();
    case 938:
      if (lookahead == 'e') ADVANCE(329);
      if (lookahead == 'o') ADVANCE(1958);
      END_STATE();
    case 939:
      if (lookahead == 'e') ADVANCE(225);
      if (lookahead == 'l') ADVANCE(1415);
      if (lookahead == 'r') ADVANCE(728);
      END_STATE();
    case 940:
      if (lookahead == 'e') ADVANCE(176);
      END_STATE();
    case 941:
      if (lookahead == 'e') ADVANCE(155);
      END_STATE();
    case 942:
      if (lookahead == 'e') ADVANCE(148);
      END_STATE();
    case 943:
      if (lookahead == 'e') ADVANCE(440);
      END_STATE();
    case 944:
      if (lookahead == 'e') ADVANCE(438);
      END_STATE();
    case 945:
      if (lookahead == 'e') ADVANCE(182);
      END_STATE();
    case 946:
      if (lookahead == 'e') ADVANCE(704);
      END_STATE();
    case 947:
      if (lookahead == 'e') ADVANCE(1494);
      END_STATE();
    case 948:
      if (lookahead == 'e') ADVANCE(156);
      END_STATE();
    case 949:
      if (lookahead == 'e') ADVANCE(466);
      END_STATE();
    case 950:
      if (lookahead == 'e') ADVANCE(497);
      END_STATE();
    case 951:
      if (lookahead == 'e') ADVANCE(248);
      END_STATE();
    case 952:
      if (lookahead == 'e') ADVANCE(433);
      END_STATE();
    case 953:
      if (lookahead == 'e') ADVANCE(525);
      END_STATE();
    case 954:
      if (lookahead == 'e') ADVANCE(159);
      END_STATE();
    case 955:
      if (lookahead == 'e') ADVANCE(246);
      END_STATE();
    case 956:
      if (lookahead == 'e') ADVANCE(507);
      END_STATE();
    case 957:
      if (lookahead == 'e') ADVANCE(1750);
      END_STATE();
    case 958:
      if (lookahead == 'e') ADVANCE(532);
      if (lookahead == 'p') ADVANCE(219);
      if (lookahead == 's') ADVANCE(532);
      END_STATE();
    case 959:
      if (lookahead == 'e') ADVANCE(1516);
      END_STATE();
    case 960:
      if (lookahead == 'e') ADVANCE(1836);
      if (lookahead == 'i') ADVANCE(830);
      END_STATE();
    case 961:
      if (lookahead == 'e') ADVANCE(1958);
      END_STATE();
    case 962:
      if (lookahead == 'e') ADVANCE(1958);
      if (lookahead == 'o') ADVANCE(1504);
      END_STATE();
    case 963:
      if (lookahead == 'e') ADVANCE(436);
      END_STATE();
    case 964:
      if (lookahead == 'e') ADVANCE(427);
      END_STATE();
    case 965:
      if (lookahead == 'e') ADVANCE(2155);
      END_STATE();
    case 966:
      if (lookahead == 'e') ADVANCE(915);
      END_STATE();
    case 967:
      if (lookahead == 'e') ADVANCE(1852);
      if (lookahead == 'i') ADVANCE(838);
      if (lookahead == 'k') ADVANCE(575);
      if (lookahead == 'o') ADVANCE(1752);
      END_STATE();
    case 968:
      if (lookahead == 'e') ADVANCE(1106);
      END_STATE();
    case 969:
      if (lookahead == 'e') ADVANCE(325);
      END_STATE();
    case 970:
      if (lookahead == 'e') ADVANCE(468);
      END_STATE();
    case 971:
      if (lookahead == 'e') ADVANCE(515);
      END_STATE();
    case 972:
      if (lookahead == 'e') ADVANCE(1929);
      END_STATE();
    case 973:
      if (lookahead == 'e') ADVANCE(1732);
      END_STATE();
    case 974:
      if (lookahead == 'e') ADVANCE(916);
      END_STATE();
    case 975:
      if (lookahead == 'e') ADVANCE(546);
      END_STATE();
    case 976:
      if (lookahead == 'e') ADVANCE(546);
      if (lookahead == 'i') ADVANCE(1463);
      END_STATE();
    case 977:
      if (lookahead == 'e') ADVANCE(457);
      END_STATE();
    case 978:
      if (lookahead == 'e') ADVANCE(1110);
      END_STATE();
    case 979:
      if (lookahead == 'e') ADVANCE(334);
      END_STATE();
    case 980:
      if (lookahead == 'e') ADVANCE(921);
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 981:
      if (lookahead == 'e') ADVANCE(921);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 982:
      if (lookahead == 'e') ADVANCE(2151);
      END_STATE();
    case 983:
      if (lookahead == 'e') ADVANCE(412);
      END_STATE();
    case 984:
      if (lookahead == 'e') ADVANCE(1736);
      if (lookahead == 'v') ADVANCE(1003);
      if (lookahead == 'w') ADVANCE(966);
      END_STATE();
    case 985:
      if (lookahead == 'e') ADVANCE(437);
      END_STATE();
    case 986:
      if (lookahead == 'e') ADVANCE(1101);
      END_STATE();
    case 987:
      if (lookahead == 'e') ADVANCE(337);
      END_STATE();
    case 988:
      if (lookahead == 'e') ADVANCE(1738);
      END_STATE();
    case 989:
      if (lookahead == 'e') ADVANCE(1339);
      END_STATE();
    case 990:
      if (lookahead == 'e') ADVANCE(1517);
      END_STATE();
    case 991:
      if (lookahead == 'e') ADVANCE(1737);
      END_STATE();
    case 992:
      if (lookahead == 'e') ADVANCE(1979);
      END_STATE();
    case 993:
      if (lookahead == 'e') ADVANCE(1903);
      END_STATE();
    case 994:
      if (lookahead == 'e') ADVANCE(1914);
      END_STATE();
    case 995:
      if (lookahead == 'e') ADVANCE(1997);
      END_STATE();
    case 996:
      if (lookahead == 'e') ADVANCE(1358);
      END_STATE();
    case 997:
      if (lookahead == 'e') ADVANCE(888);
      END_STATE();
    case 998:
      if (lookahead == 'e') ADVANCE(699);
      END_STATE();
    case 999:
      if (lookahead == 'e') ADVANCE(890);
      END_STATE();
    case 1000:
      if (lookahead == 'e') ADVANCE(1504);
      END_STATE();
    case 1001:
      if (lookahead == 'e') ADVANCE(762);
      END_STATE();
    case 1002:
      if (lookahead == 'e') ADVANCE(893);
      END_STATE();
    case 1003:
      if (lookahead == 'e') ADVANCE(926);
      END_STATE();
    case 1004:
      if (lookahead == 'e') ADVANCE(910);
      END_STATE();
    case 1005:
      if (lookahead == 'e') ADVANCE(1758);
      END_STATE();
    case 1006:
      if (lookahead == 'e') ADVANCE(1909);
      END_STATE();
    case 1007:
      if (lookahead == 'e') ADVANCE(1676);
      if (lookahead == 'k') ADVANCE(570);
      if (lookahead == 'n') ADVANCE(1644);
      if (lookahead == 'p') ADVANCE(1192);
      if (lookahead == 'r') ADVANCE(286);
      if (lookahead == 's') ADVANCE(1238);
      if (lookahead == 't') ADVANCE(1209);
      END_STATE();
    case 1008:
      if (lookahead == 'e') ADVANCE(1676);
      if (lookahead == 'p') ADVANCE(1191);
      END_STATE();
    case 1009:
      if (lookahead == 'e') ADVANCE(1920);
      END_STATE();
    case 1010:
      if (lookahead == 'e') ADVANCE(1394);
      END_STATE();
    case 1011:
      if (lookahead == 'e') ADVANCE(869);
      END_STATE();
    case 1012:
      if (lookahead == 'e') ADVANCE(1558);
      END_STATE();
    case 1013:
      if (lookahead == 'e') ADVANCE(1951);
      END_STATE();
    case 1014:
      if (lookahead == 'e') ADVANCE(2003);
      END_STATE();
    case 1015:
      if (lookahead == 'e') ADVANCE(1922);
      END_STATE();
    case 1016:
      if (lookahead == 'e') ADVANCE(1981);
      END_STATE();
    case 1017:
      if (lookahead == 'e') ADVANCE(581);
      END_STATE();
    case 1018:
      if (lookahead == 'e') ADVANCE(941);
      END_STATE();
    case 1019:
      if (lookahead == 'e') ADVANCE(1530);
      END_STATE();
    case 1020:
      if (lookahead == 'e') ADVANCE(1421);
      END_STATE();
    case 1021:
      if (lookahead == 'e') ADVANCE(1798);
      END_STATE();
    case 1022:
      if (lookahead == 'e') ADVANCE(620);
      END_STATE();
    case 1023:
      if (lookahead == 'e') ADVANCE(1912);
      END_STATE();
    case 1024:
      if (lookahead == 'e') ADVANCE(604);
      END_STATE();
    case 1025:
      if (lookahead == 'e') ADVANCE(1913);
      END_STATE();
    case 1026:
      if (lookahead == 'e') ADVANCE(1968);
      if (lookahead == 'i') ADVANCE(1463);
      if (lookahead == 'u') ADVANCE(2162);
      END_STATE();
    case 1027:
      if (lookahead == 'e') ADVANCE(1764);
      END_STATE();
    case 1028:
      if (lookahead == 'e') ADVANCE(1871);
      END_STATE();
    case 1029:
      if (lookahead == 'e') ADVANCE(2004);
      END_STATE();
    case 1030:
      if (lookahead == 'e') ADVANCE(1835);
      END_STATE();
    case 1031:
      if (lookahead == 'e') ADVANCE(2011);
      END_STATE();
    case 1032:
      if (lookahead == 'e') ADVANCE(1004);
      END_STATE();
    case 1033:
      if (lookahead == 'e') ADVANCE(948);
      if (lookahead == 'r') ADVANCE(1335);
      END_STATE();
    case 1034:
      if (lookahead == 'e') ADVANCE(1751);
      END_STATE();
    case 1035:
      if (lookahead == 'e') ADVANCE(1845);
      if (lookahead == 'i') ADVANCE(1516);
      END_STATE();
    case 1036:
      if (lookahead == 'e') ADVANCE(1795);
      END_STATE();
    case 1037:
      if (lookahead == 'e') ADVANCE(1873);
      END_STATE();
    case 1038:
      if (lookahead == 'e') ADVANCE(1893);
      END_STATE();
    case 1039:
      if (lookahead == 'e') ADVANCE(953);
      END_STATE();
    case 1040:
      if (lookahead == 'e') ADVANCE(1808);
      END_STATE();
    case 1041:
      if (lookahead == 'e') ADVANCE(1080);
      END_STATE();
    case 1042:
      if (lookahead == 'e') ADVANCE(1805);
      END_STATE();
    case 1043:
      if (lookahead == 'e') ADVANCE(1768);
      END_STATE();
    case 1044:
      if (lookahead == 'e') ADVANCE(1813);
      END_STATE();
    case 1045:
      if (lookahead == 'e') ADVANCE(1771);
      END_STATE();
    case 1046:
      if (lookahead == 'e') ADVANCE(1111);
      END_STATE();
    case 1047:
      if (lookahead == 'e') ADVANCE(2115);
      if (lookahead == 'i') ADVANCE(1158);
      END_STATE();
    case 1048:
      if (lookahead == 'e') ADVANCE(1428);
      END_STATE();
    case 1049:
      if (lookahead == 'e') ADVANCE(524);
      END_STATE();
    case 1050:
      if (lookahead == 'e') ADVANCE(898);
      END_STATE();
    case 1051:
      if (lookahead == 'e') ADVANCE(678);
      if (lookahead == 'w') ADVANCE(678);
      END_STATE();
    case 1052:
      if (lookahead == 'e') ADVANCE(1108);
      END_STATE();
    case 1053:
      if (lookahead == 'e') ADVANCE(1108);
      if (lookahead == 'l') ADVANCE(152);
      if (lookahead == 't') ADVANCE(358);
      END_STATE();
    case 1054:
      if (lookahead == 'e') ADVANCE(1000);
      END_STATE();
    case 1055:
      if (lookahead == 'e') ADVANCE(1940);
      END_STATE();
    case 1056:
      if (lookahead == 'e') ADVANCE(901);
      END_STATE();
    case 1057:
      if (lookahead == 'e') ADVANCE(1493);
      END_STATE();
    case 1058:
      if (lookahead == 'e') ADVANCE(854);
      if (lookahead == 'o') ADVANCE(1574);
      END_STATE();
    case 1059:
      if (lookahead == 'e') ADVANCE(667);
      END_STATE();
    case 1060:
      if (lookahead == 'e') ADVANCE(1729);
      END_STATE();
    case 1061:
      if (lookahead == 'e') ADVANCE(1833);
      END_STATE();
    case 1062:
      if (lookahead == 'e') ADVANCE(2026);
      END_STATE();
    case 1063:
      if (lookahead == 'e') ADVANCE(2026);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 1064:
      if (lookahead == 'e') ADVANCE(749);
      END_STATE();
    case 1065:
      if (lookahead == 'e') ADVANCE(1112);
      END_STATE();
    case 1066:
      if (lookahead == 'e') ADVANCE(919);
      END_STATE();
    case 1067:
      if (lookahead == 'e') ADVANCE(855);
      END_STATE();
    case 1068:
      if (lookahead == 'e') ADVANCE(1930);
      END_STATE();
    case 1069:
      if (lookahead == 'e') ADVANCE(1716);
      END_STATE();
    case 1070:
      if (lookahead == 'e') ADVANCE(1113);
      END_STATE();
    case 1071:
      if (lookahead == 'e') ADVANCE(1832);
      if (lookahead == 'n') ADVANCE(462);
      END_STATE();
    case 1072:
      if (lookahead == 'e') ADVANCE(857);
      END_STATE();
    case 1073:
      if (lookahead == 'e') ADVANCE(1584);
      END_STATE();
    case 1074:
      if (lookahead == 'e') ADVANCE(1114);
      END_STATE();
    case 1075:
      if (lookahead == 'e') ADVANCE(1115);
      END_STATE();
    case 1076:
      if (lookahead == 'e') ADVANCE(1567);
      END_STATE();
    case 1077:
      if (lookahead == 'e') ADVANCE(1116);
      if (lookahead == 'o') ADVANCE(1544);
      END_STATE();
    case 1078:
      if (lookahead == 'e') ADVANCE(1851);
      END_STATE();
    case 1079:
      if (lookahead == 'e') ADVANCE(1109);
      END_STATE();
    case 1080:
      if (lookahead == 'e') ADVANCE(2045);
      END_STATE();
    case 1081:
      if (lookahead == 'e') ADVANCE(1898);
      END_STATE();
    case 1082:
      if (lookahead == 'e') ADVANCE(1552);
      END_STATE();
    case 1083:
      if (lookahead == 'e') ADVANCE(703);
      END_STATE();
    case 1084:
      if (lookahead == 'e') ADVANCE(1462);
      END_STATE();
    case 1085:
      if (lookahead == 'e') ADVANCE(1160);
      END_STATE();
    case 1086:
      if (lookahead == 'e') ADVANCE(1895);
      END_STATE();
    case 1087:
      if (lookahead == 'e') ADVANCE(2032);
      END_STATE();
    case 1088:
      if (lookahead == 'e') ADVANCE(1311);
      END_STATE();
    case 1089:
      if (lookahead == 'e') ADVANCE(1878);
      END_STATE();
    case 1090:
      if (lookahead == 'e') ADVANCE(923);
      END_STATE();
    case 1091:
      if (lookahead == 'e') ADVANCE(702);
      END_STATE();
    case 1092:
      if (lookahead == 'e') ADVANCE(1527);
      END_STATE();
    case 1093:
      if (lookahead == 'e') ADVANCE(707);
      END_STATE();
    case 1094:
      if (lookahead == 'e') ADVANCE(875);
      END_STATE();
    case 1095:
      if (lookahead == 'e') ADVANCE(1119);
      END_STATE();
    case 1096:
      if (lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 1097:
      if (lookahead == 'f') ADVANCE(152);
      if (lookahead == 'l') ADVANCE(2065);
      END_STATE();
    case 1098:
      if (lookahead == 'f') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(1624);
      END_STATE();
    case 1099:
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(152);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 's') ADVANCE(778);
      END_STATE();
    case 1100:
      if (lookahead == 'f') ADVANCE(1750);
      if (lookahead == 'i') ADVANCE(1516);
      if (lookahead == 'o') ADVANCE(1680);
      if (lookahead == 'p') ADVANCE(1820);
      if (lookahead == 's') ADVANCE(778);
      if (lookahead == 'u') ADVANCE(602);
      END_STATE();
    case 1101:
      if (lookahead == 'f') ADVANCE(1958);
      END_STATE();
    case 1102:
      if (lookahead == 'f') ADVANCE(1107);
      END_STATE();
    case 1103:
      if (lookahead == 'f') ADVANCE(1961);
      if (lookahead == 's') ADVANCE(1906);
      END_STATE();
    case 1104:
      if (lookahead == 'f') ADVANCE(1984);
      if (lookahead == 'l') ADVANCE(222);
      if (lookahead == 'p') ADVANCE(1096);
      END_STATE();
    case 1105:
      if (lookahead == 'f') ADVANCE(1925);
      if (lookahead == 'p') ADVANCE(1183);
      END_STATE();
    case 1106:
      if (lookahead == 'f') ADVANCE(2049);
      END_STATE();
    case 1107:
      if (lookahead == 'f') ADVANCE(1081);
      END_STATE();
    case 1108:
      if (lookahead == 'f') ADVANCE(1988);
      END_STATE();
    case 1109:
      if (lookahead == 'f') ADVANCE(2003);
      END_STATE();
    case 1110:
      if (lookahead == 'f') ADVANCE(1668);
      END_STATE();
    case 1111:
      if (lookahead == 'f') ADVANCE(1995);
      if (lookahead == 's') ADVANCE(1911);
      END_STATE();
    case 1112:
      if (lookahead == 'f') ADVANCE(1970);
      END_STATE();
    case 1113:
      if (lookahead == 'f') ADVANCE(1972);
      END_STATE();
    case 1114:
      if (lookahead == 'f') ADVANCE(1990);
      END_STATE();
    case 1115:
      if (lookahead == 'f') ADVANCE(2028);
      END_STATE();
    case 1116:
      if (lookahead == 'f') ADVANCE(1976);
      END_STATE();
    case 1117:
      if (lookahead == 'f') ADVANCE(1628);
      END_STATE();
    case 1118:
      if (lookahead == 'f') ADVANCE(1267);
      END_STATE();
    case 1119:
      if (lookahead == 'f') ADVANCE(2048);
      END_STATE();
    case 1120:
      if (lookahead == 'g') ADVANCE(152);
      END_STATE();
    case 1121:
      if (lookahead == 'g') ADVANCE(152);
      if (lookahead == 's') ADVANCE(1677);
      END_STATE();
    case 1122:
      if (lookahead == 'g') ADVANCE(152);
      if (lookahead == 't') ADVANCE(358);
      END_STATE();
    case 1123:
      if (lookahead == 'g') ADVANCE(767);
      END_STATE();
    case 1124:
      if (lookahead == 'g') ADVANCE(240);
      END_STATE();
    case 1125:
      if (lookahead == 'g') ADVANCE(240);
      if (lookahead == 'i') ADVANCE(1516);
      END_STATE();
    case 1126:
      if (lookahead == 'g') ADVANCE(495);
      END_STATE();
    case 1127:
      if (lookahead == 'g') ADVANCE(514);
      END_STATE();
    case 1128:
      if (lookahead == 'g') ADVANCE(244);
      END_STATE();
    case 1129:
      if (lookahead == 'g') ADVANCE(1461);
      END_STATE();
    case 1130:
      if (lookahead == 'g') ADVANCE(303);
      END_STATE();
    case 1131:
      if (lookahead == 'g') ADVANCE(532);
      END_STATE();
    case 1132:
      if (lookahead == 'g') ADVANCE(1199);
      END_STATE();
    case 1133:
      if (lookahead == 'g') ADVANCE(2154);
      END_STATE();
    case 1134:
      if (lookahead == 'g') ADVANCE(1146);
      if (lookahead == 'l') ADVANCE(995);
      if (lookahead == 'r') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(1187);
      END_STATE();
    case 1135:
      if (lookahead == 'g') ADVANCE(1146);
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 1136:
      if (lookahead == 'g') ADVANCE(1146);
      if (lookahead == 'r') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(1189);
      END_STATE();
    case 1137:
      if (lookahead == 'g') ADVANCE(1590);
      END_STATE();
    case 1138:
      if (lookahead == 'g') ADVANCE(1629);
      if (lookahead == 'p') ADVANCE(1096);
      END_STATE();
    case 1139:
      if (lookahead == 'g') ADVANCE(1629);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 1140:
      if (lookahead == 'g') ADVANCE(1470);
      END_STATE();
    case 1141:
      if (lookahead == 'g') ADVANCE(1188);
      END_STATE();
    case 1142:
      if (lookahead == 'g') ADVANCE(1358);
      if (lookahead == 'i') ADVANCE(1390);
      if (lookahead == 'l') ADVANCE(1120);
      if (lookahead == 'r') ADVANCE(1331);
      END_STATE();
    case 1143:
      if (lookahead == 'g') ADVANCE(676);
      if (lookahead == 's') ADVANCE(2014);
      if (lookahead == 'w') ADVANCE(493);
      END_STATE();
    case 1144:
      if (lookahead == 'g') ADVANCE(1504);
      END_STATE();
    case 1145:
      if (lookahead == 'g') ADVANCE(926);
      END_STATE();
    case 1146:
      if (lookahead == 'g') ADVANCE(957);
      END_STATE();
    case 1147:
      if (lookahead == 'g') ADVANCE(961);
      END_STATE();
    case 1148:
      if (lookahead == 'g') ADVANCE(1877);
      END_STATE();
    case 1149:
      if (lookahead == 'g') ADVANCE(1877);
      if (lookahead == 'i') ADVANCE(1516);
      if (lookahead == 't') ADVANCE(1658);
      END_STATE();
    case 1150:
      if (lookahead == 'g') ADVANCE(1775);
      END_STATE();
    case 1151:
      if (lookahead == 'g') ADVANCE(1202);
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 's') ADVANCE(1301);
      END_STATE();
    case 1152:
      if (lookahead == 'g') ADVANCE(920);
      END_STATE();
    case 1153:
      if (lookahead == 'g') ADVANCE(1784);
      END_STATE();
    case 1154:
      if (lookahead == 'g') ADVANCE(1198);
      END_STATE();
    case 1155:
      if (lookahead == 'g') ADVANCE(1978);
      END_STATE();
    case 1156:
      if (lookahead == 'g') ADVANCE(1978);
      if (lookahead == 'l') ADVANCE(972);
      END_STATE();
    case 1157:
      if (lookahead == 'g') ADVANCE(1978);
      if (lookahead == 'q') ADVANCE(1155);
      END_STATE();
    case 1158:
      if (lookahead == 'g') ADVANCE(1203);
      END_STATE();
    case 1159:
      if (lookahead == 'g') ADVANCE(1496);
      if (lookahead == 'm') ADVANCE(241);
      END_STATE();
    case 1160:
      if (lookahead == 'g') ADVANCE(1828);
      END_STATE();
    case 1161:
      if (lookahead == 'g') ADVANCE(1828);
      if (lookahead == 'r') ADVANCE(1945);
      END_STATE();
    case 1162:
      if (lookahead == 'g') ADVANCE(1204);
      END_STATE();
    case 1163:
      if (lookahead == 'g') ADVANCE(1206);
      END_STATE();
    case 1164:
      if (lookahead == 'g') ADVANCE(1207);
      END_STATE();
    case 1165:
      if (lookahead == 'g') ADVANCE(1208);
      END_STATE();
    case 1166:
      if (lookahead == 'g') ADVANCE(1407);
      END_STATE();
    case 1167:
      if (lookahead == 'g') ADVANCE(1210);
      END_STATE();
    case 1168:
      if (lookahead == 'g') ADVANCE(1040);
      if (lookahead == 'r') ADVANCE(849);
      END_STATE();
    case 1169:
      if (lookahead == 'g') ADVANCE(1211);
      END_STATE();
    case 1170:
      if (lookahead == 'g') ADVANCE(1212);
      END_STATE();
    case 1171:
      if (lookahead == 'g') ADVANCE(1429);
      END_STATE();
    case 1172:
      if (lookahead == 'g') ADVANCE(1201);
      END_STATE();
    case 1173:
      if (lookahead == 'g') ADVANCE(1431);
      END_STATE();
    case 1174:
      if (lookahead == 'g') ADVANCE(1432);
      END_STATE();
    case 1175:
      if (lookahead == 'g') ADVANCE(1433);
      END_STATE();
    case 1176:
      if (lookahead == 'g') ADVANCE(1446);
      END_STATE();
    case 1177:
      if (lookahead == 'g') ADVANCE(1434);
      END_STATE();
    case 1178:
      if (lookahead == 'g') ADVANCE(1435);
      END_STATE();
    case 1179:
      if (lookahead == 'g') ADVANCE(494);
      END_STATE();
    case 1180:
      if (lookahead == 'g') ADVANCE(694);
      END_STATE();
    case 1181:
      if (lookahead == 'g') ADVANCE(1221);
      END_STATE();
    case 1182:
      if (lookahead == 'h') ADVANCE(881);
      END_STATE();
    case 1183:
      if (lookahead == 'h') ADVANCE(152);
      END_STATE();
    case 1184:
      if (lookahead == 'h') ADVANCE(297);
      END_STATE();
    case 1185:
      if (lookahead == 'h') ADVANCE(514);
      END_STATE();
    case 1186:
      if (lookahead == 'h') ADVANCE(532);
      END_STATE();
    case 1187:
      if (lookahead == 'h') ADVANCE(358);
      END_STATE();
    case 1188:
      if (lookahead == 'h') ADVANCE(1958);
      END_STATE();
    case 1189:
      if (lookahead == 'h') ADVANCE(2108);
      END_STATE();
    case 1190:
      if (lookahead == 'h') ADVANCE(754);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 1191:
      if (lookahead == 'h') ADVANCE(1223);
      END_STATE();
    case 1192:
      if (lookahead == 'h') ADVANCE(1223);
      if (lookahead == 'i') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(1657);
      END_STATE();
    case 1193:
      if (lookahead == 'h') ADVANCE(546);
      END_STATE();
    case 1194:
      if (lookahead == 'h') ADVANCE(1677);
      END_STATE();
    case 1195:
      if (lookahead == 'h') ADVANCE(1339);
      END_STATE();
    case 1196:
      if (lookahead == 'h') ADVANCE(1339);
      if (lookahead == 'r') ADVANCE(316);
      END_STATE();
    case 1197:
      if (lookahead == 'h') ADVANCE(926);
      END_STATE();
    case 1198:
      if (lookahead == 'h') ADVANCE(2049);
      END_STATE();
    case 1199:
      if (lookahead == 'h') ADVANCE(1965);
      END_STATE();
    case 1200:
      if (lookahead == 'h') ADVANCE(556);
      END_STATE();
    case 1201:
      if (lookahead == 'h') ADVANCE(2003);
      END_STATE();
    case 1202:
      if (lookahead == 'h') ADVANCE(1967);
      END_STATE();
    case 1203:
      if (lookahead == 'h') ADVANCE(1995);
      END_STATE();
    case 1204:
      if (lookahead == 'h') ADVANCE(1980);
      END_STATE();
    case 1205:
      if (lookahead == 'h') ADVANCE(1226);
      END_STATE();
    case 1206:
      if (lookahead == 'h') ADVANCE(1971);
      END_STATE();
    case 1207:
      if (lookahead == 'h') ADVANCE(1973);
      END_STATE();
    case 1208:
      if (lookahead == 'h') ADVANCE(1990);
      END_STATE();
    case 1209:
      if (lookahead == 'h') ADVANCE(992);
      if (lookahead == 'r') ADVANCE(1334);
      END_STATE();
    case 1210:
      if (lookahead == 'h') ADVANCE(1975);
      END_STATE();
    case 1211:
      if (lookahead == 'h') ADVANCE(1977);
      END_STATE();
    case 1212:
      if (lookahead == 'h') ADVANCE(1974);
      END_STATE();
    case 1213:
      if (lookahead == 'h') ADVANCE(1068);
      END_STATE();
    case 1214:
      if (lookahead == 'h') ADVANCE(1260);
      END_STATE();
    case 1215:
      if (lookahead == 'h') ADVANCE(1000);
      END_STATE();
    case 1216:
      if (lookahead == 'h') ADVANCE(1117);
      END_STATE();
    case 1217:
      if (lookahead == 'h') ADVANCE(1799);
      if (lookahead == 'i') ADVANCE(1497);
      if (lookahead == 'r') ADVANCE(1231);
      END_STATE();
    case 1218:
      if (lookahead == 'h') ADVANCE(1022);
      END_STATE();
    case 1219:
      if (lookahead == 'h') ADVANCE(1284);
      END_STATE();
    case 1220:
      if (lookahead == 'h') ADVANCE(1880);
      END_STATE();
    case 1221:
      if (lookahead == 'h') ADVANCE(2048);
      END_STATE();
    case 1222:
      if (lookahead == 'i') ADVANCE(881);
      END_STATE();
    case 1223:
      if (lookahead == 'i') ADVANCE(152);
      END_STATE();
    case 1224:
      if (lookahead == 'i') ADVANCE(297);
      END_STATE();
    case 1225:
      if (lookahead == 'i') ADVANCE(280);
      END_STATE();
    case 1226:
      if (lookahead == 'i') ADVANCE(830);
      END_STATE();
    case 1227:
      if (lookahead == 'i') ADVANCE(2137);
      if (lookahead == 'o') ADVANCE(1218);
      END_STATE();
    case 1228:
      if (lookahead == 'i') ADVANCE(2153);
      END_STATE();
    case 1229:
      if (lookahead == 'i') ADVANCE(274);
      END_STATE();
    case 1230:
      if (lookahead == 'i') ADVANCE(302);
      END_STATE();
    case 1231:
      if (lookahead == 'i') ADVANCE(299);
      END_STATE();
    case 1232:
      if (lookahead == 'i') ADVANCE(285);
      END_STATE();
    case 1233:
      if (lookahead == 'i') ADVANCE(1750);
      END_STATE();
    case 1234:
      if (lookahead == 'i') ADVANCE(1750);
      if (lookahead == 'r') ADVANCE(1604);
      END_STATE();
    case 1235:
      if (lookahead == 'i') ADVANCE(1516);
      END_STATE();
    case 1236:
      if (lookahead == 'i') ADVANCE(1516);
      if (lookahead == 'n') ADVANCE(1958);
      END_STATE();
    case 1237:
      if (lookahead == 'i') ADVANCE(358);
      if (lookahead == 'm') ADVANCE(1495);
      if (lookahead == 'o') ADVANCE(1538);
      END_STATE();
    case 1238:
      if (lookahead == 'i') ADVANCE(1140);
      if (lookahead == 'u') ADVANCE(753);
      END_STATE();
    case 1239:
      if (lookahead == 'i') ADVANCE(1391);
      END_STATE();
    case 1240:
      if (lookahead == 'i') ADVANCE(301);
      END_STATE();
    case 1241:
      if (lookahead == 'i') ADVANCE(1958);
      END_STATE();
    case 1242:
      if (lookahead == 'i') ADVANCE(2108);
      END_STATE();
    case 1243:
      if (lookahead == 'i') ADVANCE(876);
      END_STATE();
    case 1244:
      if (lookahead == 'i') ADVANCE(876);
      if (lookahead == 'o') ADVANCE(2092);
      END_STATE();
    case 1245:
      if (lookahead == 'i') ADVANCE(1505);
      END_STATE();
    case 1246:
      if (lookahead == 'i') ADVANCE(359);
      END_STATE();
    case 1247:
      if (lookahead == 'i') ADVANCE(1952);
      END_STATE();
    case 1248:
      if (lookahead == 'i') ADVANCE(1952);
      if (lookahead == 'p') ADVANCE(1632);
      END_STATE();
    case 1249:
      if (lookahead == 'i') ADVANCE(250);
      END_STATE();
    case 1250:
      if (lookahead == 'i') ADVANCE(1386);
      END_STATE();
    case 1251:
      if (lookahead == 'i') ADVANCE(1677);
      END_STATE();
    case 1252:
      if (lookahead == 'i') ADVANCE(1120);
      END_STATE();
    case 1253:
      if (lookahead == 'i') ADVANCE(1120);
      if (lookahead == 'l') ADVANCE(1252);
      END_STATE();
    case 1254:
      if (lookahead == 'i') ADVANCE(1102);
      END_STATE();
    case 1255:
      if (lookahead == 'i') ADVANCE(1463);
      END_STATE();
    case 1256:
      if (lookahead == 'i') ADVANCE(1463);
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 1257:
      if (lookahead == 'i') ADVANCE(1463);
      if (lookahead == 'l') ADVANCE(674);
      END_STATE();
    case 1258:
      if (lookahead == 'i') ADVANCE(1629);
      END_STATE();
    case 1259:
      if (lookahead == 'i') ADVANCE(1903);
      END_STATE();
    case 1260:
      if (lookahead == 'i') ADVANCE(1525);
      END_STATE();
    case 1261:
      if (lookahead == 'i') ADVANCE(1928);
      if (lookahead == 'l') ADVANCE(1638);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 1262:
      if (lookahead == 'i') ADVANCE(1928);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 1263:
      if (lookahead == 'i') ADVANCE(1358);
      END_STATE();
    case 1264:
      if (lookahead == 'i') ADVANCE(1133);
      END_STATE();
    case 1265:
      if (lookahead == 'i') ADVANCE(1144);
      END_STATE();
    case 1266:
      if (lookahead == 'i') ADVANCE(2069);
      END_STATE();
    case 1267:
      if (lookahead == 'i') ADVANCE(1504);
      END_STATE();
    case 1268:
      if (lookahead == 'i') ADVANCE(1388);
      END_STATE();
    case 1269:
      if (lookahead == 'i') ADVANCE(762);
      END_STATE();
    case 1270:
      if (lookahead == 'i') ADVANCE(1780);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 1271:
      if (lookahead == 'i') ADVANCE(926);
      END_STATE();
    case 1272:
      if (lookahead == 'i') ADVANCE(1538);
      END_STATE();
    case 1273:
      if (lookahead == 'i') ADVANCE(688);
      END_STATE();
    case 1274:
      if (lookahead == 'i') ADVANCE(911);
      END_STATE();
    case 1275:
      if (lookahead == 'i') ADVANCE(1482);
      END_STATE();
    case 1276:
      if (lookahead == 'i') ADVANCE(1180);
      END_STATE();
    case 1277:
      if (lookahead == 'i') ADVANCE(1824);
      if (lookahead == 'l') ADVANCE(1096);
      if (lookahead == 'm') ADVANCE(1279);
      if (lookahead == 'r') ADVANCE(892);
      END_STATE();
    case 1278:
      if (lookahead == 'i') ADVANCE(1529);
      END_STATE();
    case 1279:
      if (lookahead == 'i') ADVANCE(1374);
      END_STATE();
    case 1280:
      if (lookahead == 'i') ADVANCE(993);
      END_STATE();
    case 1281:
      if (lookahead == 'i') ADVANCE(2064);
      END_STATE();
    case 1282:
      if (lookahead == 'i') ADVANCE(1569);
      END_STATE();
    case 1283:
      if (lookahead == 'i') ADVANCE(1562);
      END_STATE();
    case 1284:
      if (lookahead == 'i') ADVANCE(1507);
      END_STATE();
    case 1285:
      if (lookahead == 'i') ADVANCE(1404);
      if (lookahead == 'l') ADVANCE(1253);
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 1286:
      if (lookahead == 'i') ADVANCE(1141);
      END_STATE();
    case 1287:
      if (lookahead == 'i') ADVANCE(1392);
      END_STATE();
    case 1288:
      if (lookahead == 'i') ADVANCE(1540);
      if (lookahead == 'p') ADVANCE(1096);
      END_STATE();
    case 1289:
      if (lookahead == 'i') ADVANCE(1817);
      END_STATE();
    case 1290:
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'o') ADVANCE(1388);
      END_STATE();
    case 1291:
      if (lookahead == 'i') ADVANCE(1817);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 1292:
      if (lookahead == 'i') ADVANCE(1390);
      END_STATE();
    case 1293:
      if (lookahead == 'i') ADVANCE(745);
      END_STATE();
    case 1294:
      if (lookahead == 'i') ADVANCE(636);
      END_STATE();
    case 1295:
      if (lookahead == 'i') ADVANCE(1943);
      END_STATE();
    case 1296:
      if (lookahead == 'i') ADVANCE(2114);
      END_STATE();
    case 1297:
      if (lookahead == 'i') ADVANCE(1723);
      END_STATE();
    case 1298:
      if (lookahead == 'i') ADVANCE(1548);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'u') ADVANCE(1519);
      END_STATE();
    case 1299:
      if (lookahead == 'i') ADVANCE(1409);
      END_STATE();
    case 1300:
      if (lookahead == 'i') ADVANCE(1487);
      END_STATE();
    case 1301:
      if (lookahead == 'i') ADVANCE(1542);
      END_STATE();
    case 1302:
      if (lookahead == 'i') ADVANCE(1947);
      END_STATE();
    case 1303:
      if (lookahead == 'i') ADVANCE(1942);
      END_STATE();
    case 1304:
      if (lookahead == 'i') ADVANCE(1497);
      END_STATE();
    case 1305:
      if (lookahead == 'i') ADVANCE(1483);
      END_STATE();
    case 1306:
      if (lookahead == 'i') ADVANCE(2023);
      END_STATE();
    case 1307:
      if (lookahead == 'i') ADVANCE(1420);
      END_STATE();
    case 1308:
      if (lookahead == 'i') ADVANCE(1546);
      END_STATE();
    case 1309:
      if (lookahead == 'i') ADVANCE(1037);
      END_STATE();
    case 1310:
      if (lookahead == 'i') ADVANCE(644);
      END_STATE();
    case 1311:
      if (lookahead == 'i') ADVANCE(1439);
      END_STATE();
    case 1312:
      if (lookahead == 'i') ADVANCE(1407);
      END_STATE();
    case 1313:
      if (lookahead == 'i') ADVANCE(1646);
      END_STATE();
    case 1314:
      if (lookahead == 'i') ADVANCE(1625);
      END_STATE();
    case 1315:
      if (lookahead == 'i') ADVANCE(1563);
      END_STATE();
    case 1316:
      if (lookahead == 'i') ADVANCE(1563);
      if (lookahead == 'n') ADVANCE(152);
      END_STATE();
    case 1317:
      if (lookahead == 'i') ADVANCE(1154);
      END_STATE();
    case 1318:
      if (lookahead == 'i') ADVANCE(1440);
      END_STATE();
    case 1319:
      if (lookahead == 'i') ADVANCE(868);
      END_STATE();
    case 1320:
      if (lookahead == 'i') ADVANCE(872);
      END_STATE();
    case 1321:
      if (lookahead == 'i') ADVANCE(746);
      END_STATE();
    case 1322:
      if (lookahead == 'i') ADVANCE(1162);
      END_STATE();
    case 1323:
      if (lookahead == 'i') ADVANCE(1163);
      END_STATE();
    case 1324:
      if (lookahead == 'i') ADVANCE(1164);
      END_STATE();
    case 1325:
      if (lookahead == 'i') ADVANCE(1165);
      END_STATE();
    case 1326:
      if (lookahead == 'i') ADVANCE(1167);
      END_STATE();
    case 1327:
      if (lookahead == 'i') ADVANCE(1169);
      END_STATE();
    case 1328:
      if (lookahead == 'i') ADVANCE(1170);
      END_STATE();
    case 1329:
      if (lookahead == 'i') ADVANCE(1172);
      END_STATE();
    case 1330:
      if (lookahead == 'i') ADVANCE(1891);
      END_STATE();
    case 1331:
      if (lookahead == 'i') ADVANCE(712);
      END_STATE();
    case 1332:
      if (lookahead == 'i') ADVANCE(1181);
      END_STATE();
    case 1333:
      if (lookahead == 'i') ADVANCE(714);
      END_STATE();
    case 1334:
      if (lookahead == 'i') ADVANCE(715);
      END_STATE();
    case 1335:
      if (lookahead == 'i') ADVANCE(716);
      END_STATE();
    case 1336:
      if (lookahead == 'i') ADVANCE(717);
      END_STATE();
    case 1337:
      if (lookahead == 'j') ADVANCE(152);
      END_STATE();
    case 1338:
      if (lookahead == 'j') ADVANCE(152);
      if (lookahead == 'n') ADVANCE(1337);
      END_STATE();
    case 1339:
      if (lookahead == 'k') ADVANCE(152);
      END_STATE();
    case 1340:
      if (lookahead == 'k') ADVANCE(514);
      END_STATE();
    case 1341:
      if (lookahead == 'k') ADVANCE(870);
      END_STATE();
    case 1342:
      if (lookahead == 'k') ADVANCE(1444);
      END_STATE();
    case 1343:
      if (lookahead == 'k') ADVANCE(1417);
      END_STATE();
    case 1344:
      if (lookahead == 'k') ADVANCE(308);
      END_STATE();
    case 1345:
      if (lookahead == 'k') ADVANCE(284);
      END_STATE();
    case 1346:
      if (lookahead == 'k') ADVANCE(754);
      END_STATE();
    case 1347:
      if (lookahead == 'k') ADVANCE(754);
      if (lookahead == 'm') ADVANCE(297);
      END_STATE();
    case 1348:
      if (lookahead == 'k') ADVANCE(2131);
      END_STATE();
    case 1349:
      if (lookahead == 'k') ADVANCE(350);
      END_STATE();
    case 1350:
      if (lookahead == 'k') ADVANCE(1939);
      END_STATE();
    case 1351:
      if (lookahead == 'k') ADVANCE(957);
      END_STATE();
    case 1352:
      if (lookahead == 'k') ADVANCE(961);
      END_STATE();
    case 1353:
      if (lookahead == 'k') ADVANCE(671);
      END_STATE();
    case 1354:
      if (lookahead == 'k') ADVANCE(678);
      END_STATE();
    case 1355:
      if (lookahead == 'k') ADVANCE(678);
      if (lookahead == 'l') ADVANCE(609);
      END_STATE();
    case 1356:
      if (lookahead == 'k') ADVANCE(1308);
      END_STATE();
    case 1357:
      if (lookahead == 'l') ADVANCE(881);
      END_STATE();
    case 1358:
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 1359:
      if (lookahead == 'l') ADVANCE(347);
      END_STATE();
    case 1360:
      if (lookahead == 'l') ADVANCE(493);
      END_STATE();
    case 1361:
      if (lookahead == 'l') ADVANCE(462);
      END_STATE();
    case 1362:
      if (lookahead == 'l') ADVANCE(221);
      END_STATE();
    case 1363:
      if (lookahead == 'l') ADVANCE(291);
      END_STATE();
    case 1364:
      if (lookahead == 'l') ADVANCE(2164);
      END_STATE();
    case 1365:
      if (lookahead == 'l') ADVANCE(426);
      END_STATE();
    case 1366:
      if (lookahead == 'l') ADVANCE(413);
      END_STATE();
    case 1367:
      if (lookahead == 'l') ADVANCE(185);
      END_STATE();
    case 1368:
      if (lookahead == 'l') ADVANCE(532);
      END_STATE();
    case 1369:
      if (lookahead == 'l') ADVANCE(2143);
      END_STATE();
    case 1370:
      if (lookahead == 'l') ADVANCE(1057);
      END_STATE();
    case 1371:
      if (lookahead == 'l') ADVANCE(1057);
      if (lookahead == 'q') ADVANCE(2082);
      END_STATE();
    case 1372:
      if (lookahead == 'l') ADVANCE(1057);
      if (lookahead == 'q') ADVANCE(2081);
      if (lookahead == 'x') ADVANCE(1247);
      END_STATE();
    case 1373:
      if (lookahead == 'l') ADVANCE(478);
      END_STATE();
    case 1374:
      if (lookahead == 'l') ADVANCE(1958);
      END_STATE();
    case 1375:
      if (lookahead == 'l') ADVANCE(190);
      END_STATE();
    case 1376:
      if (lookahead == 'l') ADVANCE(471);
      END_STATE();
    case 1377:
      if (lookahead == 'l') ADVANCE(739);
      END_STATE();
    case 1378:
      if (lookahead == 'l') ADVANCE(434);
      END_STATE();
    case 1379:
      if (lookahead == 'l') ADVANCE(425);
      END_STATE();
    case 1380:
      if (lookahead == 'l') ADVANCE(516);
      END_STATE();
    case 1381:
      if (lookahead == 'l') ADVANCE(546);
      END_STATE();
    case 1382:
      if (lookahead == 'l') ADVANCE(334);
      END_STATE();
    case 1383:
      if (lookahead == 'l') ADVANCE(579);
      END_STATE();
    case 1384:
      if (lookahead == 'l') ADVANCE(255);
      END_STATE();
    case 1385:
      if (lookahead == 'l') ADVANCE(445);
      END_STATE();
    case 1386:
      if (lookahead == 'l') ADVANCE(909);
      END_STATE();
    case 1387:
      if (lookahead == 'l') ADVANCE(456);
      END_STATE();
    case 1388:
      if (lookahead == 'l') ADVANCE(1629);
      END_STATE();
    case 1389:
      if (lookahead == 'l') ADVANCE(1339);
      END_STATE();
    case 1390:
      if (lookahead == 'l') ADVANCE(896);
      END_STATE();
    case 1391:
      if (lookahead == 'l') ADVANCE(896);
      if (lookahead == 'm') ADVANCE(993);
      END_STATE();
    case 1392:
      if (lookahead == 'l') ADVANCE(896);
      if (lookahead == 'm') ADVANCE(1015);
      END_STATE();
    case 1393:
      if (lookahead == 'l') ADVANCE(977);
      END_STATE();
    case 1394:
      if (lookahead == 'l') ADVANCE(1903);
      END_STATE();
    case 1395:
      if (lookahead == 'l') ADVANCE(848);
      END_STATE();
    case 1396:
      if (lookahead == 'l') ADVANCE(1259);
      END_STATE();
    case 1397:
      if (lookahead == 'l') ADVANCE(2142);
      END_STATE();
    case 1398:
      if (lookahead == 'l') ADVANCE(1225);
      END_STATE();
    case 1399:
      if (lookahead == 'l') ADVANCE(996);
      END_STATE();
    case 1400:
      if (lookahead == 'l') ADVANCE(1358);
      END_STATE();
    case 1401:
      if (lookahead == 'l') ADVANCE(2152);
      END_STATE();
    case 1402:
      if (lookahead == 'l') ADVANCE(1638);
      END_STATE();
    case 1403:
      if (lookahead == 'l') ADVANCE(2089);
      END_STATE();
    case 1404:
      if (lookahead == 'l') ADVANCE(1252);
      END_STATE();
    case 1405:
      if (lookahead == 'l') ADVANCE(1384);
      if (lookahead == 'm') ADVANCE(1689);
      END_STATE();
    case 1406:
      if (lookahead == 'l') ADVANCE(1381);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(254);
      if (lookahead == 'u') ADVANCE(744);
      if (lookahead == 'w') ADVANCE(1508);
      END_STATE();
    case 1407:
      if (lookahead == 'l') ADVANCE(926);
      END_STATE();
    case 1408:
      if (lookahead == 'l') ADVANCE(982);
      END_STATE();
    case 1409:
      if (lookahead == 'l') ADVANCE(1368);
      END_STATE();
    case 1410:
      if (lookahead == 'l') ADVANCE(1378);
      END_STATE();
    case 1411:
      if (lookahead == 'l') ADVANCE(997);
      END_STATE();
    case 1412:
      if (lookahead == 'l') ADVANCE(1459);
      if (lookahead == 's') ADVANCE(1194);
      END_STATE();
    case 1413:
      if (lookahead == 'l') ADVANCE(1656);
      END_STATE();
    case 1414:
      if (lookahead == 'l') ADVANCE(1280);
      END_STATE();
    case 1415:
      if (lookahead == 'l') ADVANCE(1251);
      END_STATE();
    case 1416:
      if (lookahead == 'l') ADVANCE(1376);
      END_STATE();
    case 1417:
      if (lookahead == 'l') ADVANCE(1607);
      if (lookahead == 's') ADVANCE(1747);
      if (lookahead == 't') ADVANCE(1902);
      END_STATE();
    case 1418:
      if (lookahead == 'l') ADVANCE(972);
      END_STATE();
    case 1419:
      if (lookahead == 'l') ADVANCE(972);
      if (lookahead == 'q') ADVANCE(1418);
      END_STATE();
    case 1420:
      if (lookahead == 'l') ADVANCE(1293);
      END_STATE();
    case 1421:
      if (lookahead == 'l') ADVANCE(595);
      END_STATE();
    case 1422:
      if (lookahead == 'l') ADVANCE(1380);
      END_STATE();
    case 1423:
      if (lookahead == 'l') ADVANCE(943);
      END_STATE();
    case 1424:
      if (lookahead == 'l') ADVANCE(944);
      END_STATE();
    case 1425:
      if (lookahead == 'l') ADVANCE(1641);
      END_STATE();
    case 1426:
      if (lookahead == 'l') ADVANCE(933);
      END_STATE();
    case 1427:
      if (lookahead == 'l') ADVANCE(1064);
      END_STATE();
    case 1428:
      if (lookahead == 'l') ADVANCE(986);
      if (lookahead == 'r') ADVANCE(1286);
      END_STATE();
    case 1429:
      if (lookahead == 'l') ADVANCE(951);
      END_STATE();
    case 1430:
      if (lookahead == 'l') ADVANCE(952);
      END_STATE();
    case 1431:
      if (lookahead == 'l') ADVANCE(964);
      END_STATE();
    case 1432:
      if (lookahead == 'l') ADVANCE(1084);
      END_STATE();
    case 1433:
      if (lookahead == 'l') ADVANCE(974);
      END_STATE();
    case 1434:
      if (lookahead == 'l') ADVANCE(954);
      END_STATE();
    case 1435:
      if (lookahead == 'l') ADVANCE(955);
      END_STATE();
    case 1436:
      if (lookahead == 'l') ADVANCE(956);
      END_STATE();
    case 1437:
      if (lookahead == 'l') ADVANCE(983);
      END_STATE();
    case 1438:
      if (lookahead == 'l') ADVANCE(609);
      END_STATE();
    case 1439:
      if (lookahead == 'l') ADVANCE(1260);
      END_STATE();
    case 1440:
      if (lookahead == 'l') ADVANCE(903);
      END_STATE();
    case 1441:
      if (lookahead == 'l') ADVANCE(903);
      if (lookahead == 'm') ADVANCE(987);
      if (lookahead == 'n') ADVANCE(1958);
      END_STATE();
    case 1442:
      if (lookahead == 'l') ADVANCE(1631);
      if (lookahead == 'n') ADVANCE(1149);
      if (lookahead == 'p') ADVANCE(1098);
      if (lookahead == 'u') ADVANCE(1570);
      END_STATE();
    case 1443:
      if (lookahead == 'l') ADVANCE(551);
      END_STATE();
    case 1444:
      if (lookahead == 'l') ADVANCE(968);
      if (lookahead == 'r') ADVANCE(1317);
      END_STATE();
    case 1445:
      if (lookahead == 'l') ADVANCE(1049);
      END_STATE();
    case 1446:
      if (lookahead == 'l') ADVANCE(1048);
      END_STATE();
    case 1447:
      if (lookahead == 'l') ADVANCE(664);
      END_STATE();
    case 1448:
      if (lookahead == 'l') ADVANCE(2065);
      END_STATE();
    case 1449:
      if (lookahead == 'l') ADVANCE(1411);
      END_STATE();
    case 1450:
      if (lookahead == 'l') ADVANCE(610);
      END_STATE();
    case 1451:
      if (lookahead == 'l') ADVANCE(1301);
      END_STATE();
    case 1452:
      if (lookahead == 'l') ADVANCE(1636);
      if (lookahead == 'm') ADVANCE(1492);
      if (lookahead == 'n') ADVANCE(1125);
      if (lookahead == 'p') ADVANCE(275);
      END_STATE();
    case 1453:
      if (lookahead == 'l') ADVANCE(1849);
      END_STATE();
    case 1454:
      if (lookahead == 'l') ADVANCE(1235);
      END_STATE();
    case 1455:
      if (lookahead == 'l') ADVANCE(1396);
      END_STATE();
    case 1456:
      if (lookahead == 'l') ADVANCE(1283);
      END_STATE();
    case 1457:
      if (lookahead == 'l') ADVANCE(1399);
      END_STATE();
    case 1458:
      if (lookahead == 'l') ADVANCE(686);
      END_STATE();
    case 1459:
      if (lookahead == 'l') ADVANCE(1955);
      END_STATE();
    case 1460:
      if (lookahead == 'l') ADVANCE(1451);
      END_STATE();
    case 1461:
      if (lookahead == 'l') ADVANCE(1052);
      if (lookahead == 'm') ADVANCE(631);
      if (lookahead == 'r') ADVANCE(1317);
      END_STATE();
    case 1462:
      if (lookahead == 'l') ADVANCE(1079);
      if (lookahead == 'r') ADVANCE(1329);
      END_STATE();
    case 1463:
      if (lookahead == 'm') ADVANCE(152);
      END_STATE();
    case 1464:
      if (lookahead == 'm') ADVANCE(297);
      END_STATE();
    case 1465:
      if (lookahead == 'm') ADVANCE(297);
      if (lookahead == 'r') ADVANCE(1588);
      END_STATE();
    case 1466:
      if (lookahead == 'm') ADVANCE(514);
      END_STATE();
    case 1467:
      if (lookahead == 'm') ADVANCE(320);
      END_STATE();
    case 1468:
      if (lookahead == 'm') ADVANCE(366);
      END_STATE();
    case 1469:
      if (lookahead == 'm') ADVANCE(365);
      END_STATE();
    case 1470:
      if (lookahead == 'm') ADVANCE(532);
      END_STATE();
    case 1471:
      if (lookahead == 'm') ADVANCE(1316);
      END_STATE();
    case 1472:
      if (lookahead == 'm') ADVANCE(1223);
      END_STATE();
    case 1473:
      if (lookahead == 'm') ADVANCE(250);
      END_STATE();
    case 1474:
      if (lookahead == 'm') ADVANCE(1677);
      END_STATE();
    case 1475:
      if (lookahead == 'm') ADVANCE(1707);
      END_STATE();
    case 1476:
      if (lookahead == 'm') ADVANCE(1688);
      END_STATE();
    case 1477:
      if (lookahead == 'm') ADVANCE(1222);
      END_STATE();
    case 1478:
      if (lookahead == 'm') ADVANCE(1222);
      if (lookahead == 'p') ADVANCE(683);
      END_STATE();
    case 1479:
      if (lookahead == 'm') ADVANCE(1470);
      END_STATE();
    case 1480:
      if (lookahead == 'm') ADVANCE(258);
      END_STATE();
    case 1481:
      if (lookahead == 'm') ADVANCE(1711);
      END_STATE();
    case 1482:
      if (lookahead == 'm') ADVANCE(259);
      END_STATE();
    case 1483:
      if (lookahead == 'm') ADVANCE(572);
      END_STATE();
    case 1484:
      if (lookahead == 'm') ADVANCE(996);
      END_STATE();
    case 1485:
      if (lookahead == 'm') ADVANCE(1358);
      END_STATE();
    case 1486:
      if (lookahead == 'm') ADVANCE(1504);
      END_STATE();
    case 1487:
      if (lookahead == 'm') ADVANCE(926);
      END_STATE();
    case 1488:
      if (lookahead == 'm') ADVANCE(1703);
      END_STATE();
    case 1489:
      if (lookahead == 'm') ADVANCE(1698);
      END_STATE();
    case 1490:
      if (lookahead == 'm') ADVANCE(688);
      END_STATE();
    case 1491:
      if (lookahead == 'm') ADVANCE(560);
      END_STATE();
    case 1492:
      if (lookahead == 'm') ADVANCE(567);
      if (lookahead == 'p') ADVANCE(278);
      END_STATE();
    case 1493:
      if (lookahead == 'm') ADVANCE(959);
      END_STATE();
    case 1494:
      if (lookahead == 'm') ADVANCE(959);
      if (lookahead == 'x') ADVANCE(993);
      END_STATE();
    case 1495:
      if (lookahead == 'm') ADVANCE(556);
      END_STATE();
    case 1496:
      if (lookahead == 'm') ADVANCE(540);
      END_STATE();
    case 1497:
      if (lookahead == 'm') ADVANCE(993);
      END_STATE();
    case 1498:
      if (lookahead == 'm') ADVANCE(1719);
      END_STATE();
    case 1499:
      if (lookahead == 'm') ADVANCE(979);
      END_STATE();
    case 1500:
      if (lookahead == 'm') ADVANCE(1491);
      END_STATE();
    case 1501:
      if (lookahead == 'm') ADVANCE(1315);
      END_STATE();
    case 1502:
      if (lookahead == 'm') ADVANCE(682);
      END_STATE();
    case 1503:
      if (lookahead == 'n') ADVANCE(881);
      END_STATE();
    case 1504:
      if (lookahead == 'n') ADVANCE(152);
      END_STATE();
    case 1505:
      if (lookahead == 'n') ADVANCE(280);
      END_STATE();
    case 1506:
      if (lookahead == 'n') ADVANCE(422);
      END_STATE();
    case 1507:
      if (lookahead == 'n') ADVANCE(514);
      END_STATE();
    case 1508:
      if (lookahead == 'n') ADVANCE(696);
      END_STATE();
    case 1509:
      if (lookahead == 'n') ADVANCE(242);
      END_STATE();
    case 1510:
      if (lookahead == 'n') ADVANCE(840);
      END_STATE();
    case 1511:
      if (lookahead == 'n') ADVANCE(2160);
      END_STATE();
    case 1512:
      if (lookahead == 'n') ADVANCE(170);
      END_STATE();
    case 1513:
      if (lookahead == 'n') ADVANCE(521);
      END_STATE();
    case 1514:
      if (lookahead == 'n') ADVANCE(525);
      END_STATE();
    case 1515:
      if (lookahead == 'n') ADVANCE(188);
      END_STATE();
    case 1516:
      if (lookahead == 'n') ADVANCE(1958);
      END_STATE();
    case 1517:
      if (lookahead == 'n') ADVANCE(436);
      END_STATE();
    case 1518:
      if (lookahead == 'n') ADVANCE(487);
      END_STATE();
    case 1519:
      if (lookahead == 'n') ADVANCE(243);
      END_STATE();
    case 1520:
      if (lookahead == 'n') ADVANCE(2136);
      END_STATE();
    case 1521:
      if (lookahead == 'n') ADVANCE(416);
      END_STATE();
    case 1522:
      if (lookahead == 'n') ADVANCE(1118);
      END_STATE();
    case 1523:
      if (lookahead == 'n') ADVANCE(916);
      END_STATE();
    case 1524:
      if (lookahead == 'n') ADVANCE(250);
      END_STATE();
    case 1525:
      if (lookahead == 'n') ADVANCE(1120);
      END_STATE();
    case 1526:
      if (lookahead == 'n') ADVANCE(1120);
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 1527:
      if (lookahead == 'n') ADVANCE(2018);
      END_STATE();
    case 1528:
      if (lookahead == 'n') ADVANCE(1126);
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'w') ADVANCE(1027);
      END_STATE();
    case 1529:
      if (lookahead == 'n') ADVANCE(355);
      END_STATE();
    case 1530:
      if (lookahead == 'n') ADVANCE(1339);
      END_STATE();
    case 1531:
      if (lookahead == 'n') ADVANCE(258);
      END_STATE();
    case 1532:
      if (lookahead == 'n') ADVANCE(1903);
      END_STATE();
    case 1533:
      if (lookahead == 'n') ADVANCE(889);
      END_STATE();
    case 1534:
      if (lookahead == 'n') ADVANCE(253);
      END_STATE();
    case 1535:
      if (lookahead == 'n') ADVANCE(1150);
      if (lookahead == 'r') ADVANCE(1007);
      END_STATE();
    case 1536:
      if (lookahead == 'n') ADVANCE(846);
      END_STATE();
    case 1537:
      if (lookahead == 'n') ADVANCE(1166);
      END_STATE();
    case 1538:
      if (lookahead == 'n') ADVANCE(926);
      END_STATE();
    case 1539:
      if (lookahead == 'n') ADVANCE(1124);
      END_STATE();
    case 1540:
      if (lookahead == 'n') ADVANCE(874);
      END_STATE();
    case 1541:
      if (lookahead == 'n') ADVANCE(1634);
      END_STATE();
    case 1542:
      if (lookahead == 'n') ADVANCE(1152);
      END_STATE();
    case 1543:
      if (lookahead == 'n') ADVANCE(899);
      END_STATE();
    case 1544:
      if (lookahead == 'n') ADVANCE(1179);
      END_STATE();
    case 1545:
      if (lookahead == 'n') ADVANCE(957);
      END_STATE();
    case 1546:
      if (lookahead == 'n') ADVANCE(1127);
      END_STATE();
    case 1547:
      if (lookahead == 'n') ADVANCE(869);
      END_STATE();
    case 1548:
      if (lookahead == 'n') ADVANCE(2036);
      END_STATE();
    case 1549:
      if (lookahead == 'n') ADVANCE(1969);
      END_STATE();
    case 1550:
      if (lookahead == 'n') ADVANCE(1983);
      END_STATE();
    case 1551:
      if (lookahead == 'n') ADVANCE(2008);
      END_STATE();
    case 1552:
      if (lookahead == 'n') ADVANCE(2010);
      END_STATE();
    case 1553:
      if (lookahead == 'n') ADVANCE(988);
      END_STATE();
    case 1554:
      if (lookahead == 'n') ADVANCE(991);
      END_STATE();
    case 1555:
      if (lookahead == 'n') ADVANCE(1428);
      END_STATE();
    case 1556:
      if (lookahead == 'n') ADVANCE(2067);
      END_STATE();
    case 1557:
      if (lookahead == 'n') ADVANCE(207);
      END_STATE();
    case 1558:
      if (lookahead == 'n') ADVANCE(1145);
      END_STATE();
    case 1559:
      if (lookahead == 'n') ADVANCE(1148);
      END_STATE();
    case 1560:
      if (lookahead == 'n') ADVANCE(1073);
      END_STATE();
    case 1561:
      if (lookahead == 'n') ADVANCE(1258);
      END_STATE();
    case 1562:
      if (lookahead == 'n') ADVANCE(2026);
      END_STATE();
    case 1563:
      if (lookahead == 'n') ADVANCE(2065);
      END_STATE();
    case 1564:
      if (lookahead == 'n') ADVANCE(1639);
      END_STATE();
    case 1565:
      if (lookahead == 'n') ADVANCE(2027);
      END_STATE();
    case 1566:
      if (lookahead == 'n') ADVANCE(1235);
      END_STATE();
    case 1567:
      if (lookahead == 'n') ADVANCE(2025);
      END_STATE();
    case 1568:
      if (lookahead == 'n') ADVANCE(2045);
      END_STATE();
    case 1569:
      if (lookahead == 'n') ADVANCE(656);
      END_STATE();
    case 1570:
      if (lookahead == 'n') ADVANCE(2037);
      END_STATE();
    case 1571:
      if (lookahead == 'n') ADVANCE(1171);
      END_STATE();
    case 1572:
      if (lookahead == 'n') ADVANCE(2035);
      END_STATE();
    case 1573:
      if (lookahead == 'n') ADVANCE(1314);
      END_STATE();
    case 1574:
      if (lookahead == 'n') ADVANCE(1076);
      END_STATE();
    case 1575:
      if (lookahead == 'n') ADVANCE(1173);
      if (lookahead == 'r') ADVANCE(1853);
      END_STATE();
    case 1576:
      if (lookahead == 'n') ADVANCE(1173);
      if (lookahead == 'r') ADVANCE(1857);
      END_STATE();
    case 1577:
      if (lookahead == 'n') ADVANCE(2039);
      END_STATE();
    case 1578:
      if (lookahead == 'n') ADVANCE(1174);
      END_STATE();
    case 1579:
      if (lookahead == 'n') ADVANCE(1175);
      END_STATE();
    case 1580:
      if (lookahead == 'n') ADVANCE(2041);
      END_STATE();
    case 1581:
      if (lookahead == 'n') ADVANCE(1176);
      END_STATE();
    case 1582:
      if (lookahead == 'n') ADVANCE(1177);
      END_STATE();
    case 1583:
      if (lookahead == 'n') ADVANCE(1178);
      END_STATE();
    case 1584:
      if (lookahead == 'n') ADVANCE(2046);
      END_STATE();
    case 1585:
      if (lookahead == 'n') ADVANCE(705);
      END_STATE();
    case 1586:
      if (lookahead == 'n') ADVANCE(694);
      END_STATE();
    case 1587:
      if (lookahead == 'o') ADVANCE(881);
      END_STATE();
    case 1588:
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 1589:
      if (lookahead == 'o') ADVANCE(330);
      END_STATE();
    case 1590:
      if (lookahead == 'o') ADVANCE(1096);
      END_STATE();
    case 1591:
      if (lookahead == 'o') ADVANCE(2127);
      END_STATE();
    case 1592:
      if (lookahead == 'o') ADVANCE(526);
      END_STATE();
    case 1593:
      if (lookahead == 'o') ADVANCE(2106);
      END_STATE();
    case 1594:
      if (lookahead == 'o') ADVANCE(247);
      END_STATE();
    case 1595:
      if (lookahead == 'o') ADVANCE(1960);
      END_STATE();
    case 1596:
      if (lookahead == 'o') ADVANCE(1959);
      END_STATE();
    case 1597:
      if (lookahead == 'o') ADVANCE(1750);
      END_STATE();
    case 1598:
      if (lookahead == 'o') ADVANCE(1958);
      END_STATE();
    case 1599:
      if (lookahead == 'o') ADVANCE(1958);
      if (lookahead == 's') ADVANCE(2059);
      END_STATE();
    case 1600:
      if (lookahead == 'o') ADVANCE(310);
      END_STATE();
    case 1601:
      if (lookahead == 'o') ADVANCE(2138);
      END_STATE();
    case 1602:
      if (lookahead == 'o') ADVANCE(829);
      END_STATE();
    case 1603:
      if (lookahead == 'o') ADVANCE(2136);
      END_STATE();
    case 1604:
      if (lookahead == 'o') ADVANCE(1929);
      END_STATE();
    case 1605:
      if (lookahead == 'o') ADVANCE(346);
      END_STATE();
    case 1606:
      if (lookahead == 'o') ADVANCE(2116);
      END_STATE();
    case 1607:
      if (lookahead == 'o') ADVANCE(2156);
      END_STATE();
    case 1608:
      if (lookahead == 'o') ADVANCE(1677);
      END_STATE();
    case 1609:
      if (lookahead == 'o') ADVANCE(1566);
      END_STATE();
    case 1610:
      if (lookahead == 'o') ADVANCE(2117);
      END_STATE();
    case 1611:
      if (lookahead == 'o') ADVANCE(2118);
      END_STATE();
    case 1612:
      if (lookahead == 'o') ADVANCE(1463);
      END_STATE();
    case 1613:
      if (lookahead == 'o') ADVANCE(1339);
      END_STATE();
    case 1614:
      if (lookahead == 'o') ADVANCE(2013);
      END_STATE();
    case 1615:
      if (lookahead == 'o') ADVANCE(2119);
      END_STATE();
    case 1616:
      if (lookahead == 'o') ADVANCE(2132);
      END_STATE();
    case 1617:
      if (lookahead == 'o') ADVANCE(1903);
      END_STATE();
    case 1618:
      if (lookahead == 'o') ADVANCE(1525);
      END_STATE();
    case 1619:
      if (lookahead == 'o') ADVANCE(1525);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 1620:
      if (lookahead == 'o') ADVANCE(2133);
      END_STATE();
    case 1621:
      if (lookahead == 'o') ADVANCE(2125);
      END_STATE();
    case 1622:
      if (lookahead == 'o') ADVANCE(2120);
      END_STATE();
    case 1623:
      if (lookahead == 'o') ADVANCE(1358);
      if (lookahead == 'u') ADVANCE(718);
      END_STATE();
    case 1624:
      if (lookahead == 'o') ADVANCE(904);
      END_STATE();
    case 1625:
      if (lookahead == 'o') ADVANCE(1532);
      END_STATE();
    case 1626:
      if (lookahead == 'o') ADVANCE(2123);
      END_STATE();
    case 1627:
      if (lookahead == 'o') ADVANCE(1964);
      END_STATE();
    case 1628:
      if (lookahead == 'o') ADVANCE(1804);
      END_STATE();
    case 1629:
      if (lookahead == 'o') ADVANCE(1504);
      END_STATE();
    case 1630:
      if (lookahead == 'o') ADVANCE(1504);
      if (lookahead == 'r') ADVANCE(564);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 1631:
      if (lookahead == 'o') ADVANCE(1524);
      END_STATE();
    case 1632:
      if (lookahead == 'o') ADVANCE(1560);
      END_STATE();
    case 1633:
      if (lookahead == 'o') ADVANCE(1515);
      END_STATE();
    case 1634:
      if (lookahead == 'o') ADVANCE(2102);
      END_STATE();
    case 1635:
      if (lookahead == 'o') ADVANCE(1503);
      END_STATE();
    case 1636:
      if (lookahead == 'o') ADVANCE(1531);
      END_STATE();
    case 1637:
      if (lookahead == 'o') ADVANCE(2104);
      END_STATE();
    case 1638:
      if (lookahead == 'o') ADVANCE(1597);
      END_STATE();
    case 1639:
      if (lookahead == 'o') ADVANCE(2050);
      END_STATE();
    case 1640:
      if (lookahead == 'o') ADVANCE(1357);
      END_STATE();
    case 1641:
      if (lookahead == 'o') ADVANCE(828);
      END_STATE();
    case 1642:
      if (lookahead == 'o') ADVANCE(1999);
      END_STATE();
    case 1643:
      if (lookahead == 'o') ADVANCE(2001);
      END_STATE();
    case 1644:
      if (lookahead == 'o') ADVANCE(2043);
      END_STATE();
    case 1645:
      if (lookahead == 'o') ADVANCE(1555);
      END_STATE();
    case 1646:
      if (lookahead == 'o') ADVANCE(1557);
      END_STATE();
    case 1647:
      if (lookahead == 'o') ADVANCE(1523);
      END_STATE();
    case 1648:
      if (lookahead == 'o') ADVANCE(1816);
      if (lookahead == 'r') ADVANCE(1608);
      END_STATE();
    case 1649:
      if (lookahead == 'o') ADVANCE(1841);
      END_STATE();
    case 1650:
      if (lookahead == 'o') ADVANCE(1843);
      if (lookahead == 'r') ADVANCE(1608);
      END_STATE();
    case 1651:
      if (lookahead == 'o') ADVANCE(1767);
      END_STATE();
    case 1652:
      if (lookahead == 'o') ADVANCE(2128);
      END_STATE();
    case 1653:
      if (lookahead == 'o') ADVANCE(1479);
      END_STATE();
    case 1654:
      if (lookahead == 'o') ADVANCE(1479);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 1655:
      if (lookahead == 'o') ADVANCE(1822);
      END_STATE();
    case 1656:
      if (lookahead == 'o') ADVANCE(1706);
      END_STATE();
    case 1657:
      if (lookahead == 'o') ADVANCE(1717);
      END_STATE();
    case 1658:
      if (lookahead == 'o') ADVANCE(2086);
      END_STATE();
    case 1659:
      if (lookahead == 'o') ADVANCE(2134);
      END_STATE();
    case 1660:
      if (lookahead == 'o') ADVANCE(2129);
      END_STATE();
    case 1661:
      if (lookahead == 'o') ADVANCE(2092);
      END_STATE();
    case 1662:
      if (lookahead == 'o') ADVANCE(2007);
      END_STATE();
    case 1663:
      if (lookahead == 'o') ADVANCE(1559);
      if (lookahead == 'u') ADVANCE(1697);
      END_STATE();
    case 1664:
      if (lookahead == 'o') ADVANCE(2130);
      END_STATE();
    case 1665:
      if (lookahead == 'o') ADVANCE(1572);
      END_STATE();
    case 1666:
      if (lookahead == 'o') ADVANCE(1837);
      END_STATE();
    case 1667:
      if (lookahead == 'o') ADVANCE(1645);
      END_STATE();
    case 1668:
      if (lookahead == 'o') ADVANCE(1818);
      END_STATE();
    case 1669:
      if (lookahead == 'o') ADVANCE(1551);
      END_STATE();
    case 1670:
      if (lookahead == 'o') ADVANCE(1625);
      END_STATE();
    case 1671:
      if (lookahead == 'o') ADVANCE(1647);
      END_STATE();
    case 1672:
      if (lookahead == 'o') ADVANCE(1883);
      END_STATE();
    case 1673:
      if (lookahead == 'o') ADVANCE(1454);
      END_STATE();
    case 1674:
      if (lookahead == 'o') ADVANCE(2107);
      END_STATE();
    case 1675:
      if (lookahead == 'o') ADVANCE(2135);
      END_STATE();
    case 1676:
      if (lookahead == 'p') ADVANCE(1954);
      END_STATE();
    case 1677:
      if (lookahead == 'p') ADVANCE(152);
      END_STATE();
    case 1678:
      if (lookahead == 'p') ADVANCE(152);
      if (lookahead == 'r') ADVANCE(1629);
      END_STATE();
    case 1679:
      if (lookahead == 'p') ADVANCE(1186);
      END_STATE();
    case 1680:
      if (lookahead == 'p') ADVANCE(1096);
      END_STATE();
    case 1681:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'r') ADVANCE(675);
      END_STATE();
    case 1682:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'r') ADVANCE(414);
      if (lookahead == 'u') ADVANCE(1834);
      END_STATE();
    case 1683:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'r') ADVANCE(1228);
      END_STATE();
    case 1684:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(165);
      if (lookahead == 'u') ADVANCE(743);
      if (lookahead == 'w') ADVANCE(1506);
      END_STATE();
    case 1685:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 't') ADVANCE(354);
      if (lookahead == 'w') ADVANCE(2019);
      if (lookahead == 'x') ADVANCE(449);
      END_STATE();
    case 1686:
      if (lookahead == 'p') ADVANCE(1096);
      if (lookahead == 'u') ADVANCE(1533);
      END_STATE();
    case 1687:
      if (lookahead == 'p') ADVANCE(164);
      END_STATE();
    case 1688:
      if (lookahead == 'p') ADVANCE(458);
      END_STATE();
    case 1689:
      if (lookahead == 'p') ADVANCE(171);
      END_STATE();
    case 1690:
      if (lookahead == 'p') ADVANCE(145);
      END_STATE();
    case 1691:
      if (lookahead == 'p') ADVANCE(459);
      END_STATE();
    case 1692:
      if (lookahead == 'p') ADVANCE(417);
      END_STATE();
    case 1693:
      if (lookahead == 'p') ADVANCE(1750);
      if (lookahead == 's') ADVANCE(762);
      END_STATE();
    case 1694:
      if (lookahead == 'p') ADVANCE(532);
      END_STATE();
    case 1695:
      if (lookahead == 'p') ADVANCE(324);
      END_STATE();
    case 1696:
      if (lookahead == 'p') ADVANCE(416);
      END_STATE();
    case 1697:
      if (lookahead == 'p') ADVANCE(432);
      END_STATE();
    case 1698:
      if (lookahead == 'p') ADVANCE(250);
      END_STATE();
    case 1699:
      if (lookahead == 'p') ADVANCE(334);
      END_STATE();
    case 1700:
      if (lookahead == 'p') ADVANCE(1183);
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 1701:
      if (lookahead == 'p') ADVANCE(1694);
      END_STATE();
    case 1702:
      if (lookahead == 'p') ADVANCE(1903);
      if (lookahead == 'r') ADVANCE(1629);
      END_STATE();
    case 1703:
      if (lookahead == 'p') ADVANCE(259);
      END_STATE();
    case 1704:
      if (lookahead == 'p') ADVANCE(1369);
      END_STATE();
    case 1705:
      if (lookahead == 'p') ADVANCE(1714);
      END_STATE();
    case 1706:
      if (lookahead == 'p') ADVANCE(926);
      END_STATE();
    case 1707:
      if (lookahead == 'p') ADVANCE(973);
      END_STATE();
    case 1708:
      if (lookahead == 'p') ADVANCE(1448);
      END_STATE();
    case 1709:
      if (lookahead == 'p') ADVANCE(1448);
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 1710:
      if (lookahead == 'p') ADVANCE(1926);
      END_STATE();
    case 1711:
      if (lookahead == 'p') ADVANCE(1998);
      END_STATE();
    case 1712:
      if (lookahead == 'p') ADVANCE(1724);
      END_STATE();
    case 1713:
      if (lookahead == 'p') ADVANCE(1724);
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 1714:
      if (lookahead == 'p') ADVANCE(552);
      END_STATE();
    case 1715:
      if (lookahead == 'p') ADVANCE(1667);
      END_STATE();
    case 1716:
      if (lookahead == 'p') ADVANCE(1450);
      END_STATE();
    case 1717:
      if (lookahead == 'p') ADVANCE(1989);
      END_STATE();
    case 1718:
      if (lookahead == 'p') ADVANCE(1617);
      END_STATE();
    case 1719:
      if (lookahead == 'p') ADVANCE(1414);
      END_STATE();
    case 1720:
      if (lookahead == 'p') ADVANCE(672);
      END_STATE();
    case 1721:
      if (lookahead == 'p') ADVANCE(1944);
      END_STATE();
    case 1722:
      if (lookahead == 'p') ADVANCE(1849);
      if (lookahead == 's') ADVANCE(2077);
      END_STATE();
    case 1723:
      if (lookahead == 'p') ADVANCE(1426);
      END_STATE();
    case 1724:
      if (lookahead == 'p') ADVANCE(1829);
      END_STATE();
    case 1725:
      if (lookahead == 'p') ADVANCE(654);
      END_STATE();
    case 1726:
      if (lookahead == 'p') ADVANCE(703);
      END_STATE();
    case 1727:
      if (lookahead == 'p') ADVANCE(470);
      END_STATE();
    case 1728:
      if (lookahead == 'p') ADVANCE(1671);
      END_STATE();
    case 1729:
      if (lookahead == 'p') ADVANCE(687);
      END_STATE();
    case 1730:
      if (lookahead == 'p') ADVANCE(1670);
      END_STATE();
    case 1731:
      if (lookahead == 'p') ADVANCE(705);
      END_STATE();
    case 1732:
      if (lookahead == 'q') ADVANCE(152);
      END_STATE();
    case 1733:
      if (lookahead == 'q') ADVANCE(240);
      END_STATE();
    case 1734:
      if (lookahead == 'q') ADVANCE(1419);
      END_STATE();
    case 1735:
      if (lookahead == 'q') ADVANCE(1157);
      END_STATE();
    case 1736:
      if (lookahead == 'q') ADVANCE(1722);
      END_STATE();
    case 1737:
      if (lookahead == 'q') ADVANCE(325);
      END_STATE();
    case 1738:
      if (lookahead == 'q') ADVANCE(1732);
      END_STATE();
    case 1739:
      if (lookahead == 'q') ADVANCE(2053);
      END_STATE();
    case 1740:
      if (lookahead == 'q') ADVANCE(2072);
      END_STATE();
    case 1741:
      if (lookahead == 'q') ADVANCE(2105);
      if (lookahead == 'u') ADVANCE(741);
      END_STATE();
    case 1742:
      if (lookahead == 'q') ADVANCE(2093);
      END_STATE();
    case 1743:
      if (lookahead == 'q') ADVANCE(2082);
      END_STATE();
    case 1744:
      if (lookahead == 'q') ADVANCE(847);
      if (lookahead == 't') ADVANCE(546);
      END_STATE();
    case 1745:
      if (lookahead == 'q') ADVANCE(2096);
      END_STATE();
    case 1746:
      if (lookahead == 'q') ADVANCE(2100);
      END_STATE();
    case 1747:
      if (lookahead == 'q') ADVANCE(2103);
      END_STATE();
    case 1748:
      if (lookahead == 'r') ADVANCE(881);
      END_STATE();
    case 1749:
      if (lookahead == 'r') ADVANCE(935);
      END_STATE();
    case 1750:
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 1751:
      if (lookahead == 'r') ADVANCE(439);
      END_STATE();
    case 1752:
      if (lookahead == 'r') ADVANCE(309);
      END_STATE();
    case 1753:
      if (lookahead == 'r') ADVANCE(1096);
      END_STATE();
    case 1754:
      if (lookahead == 'r') ADVANCE(228);
      END_STATE();
    case 1755:
      if (lookahead == 'r') ADVANCE(322);
      END_STATE();
    case 1756:
      if (lookahead == 'r') ADVANCE(199);
      END_STATE();
    case 1757:
      if (lookahead == 'r') ADVANCE(805);
      END_STATE();
    case 1758:
      if (lookahead == 'r') ADVANCE(423);
      END_STATE();
    case 1759:
      if (lookahead == 'r') ADVANCE(2168);
      END_STATE();
    case 1760:
      if (lookahead == 'r') ADVANCE(274);
      END_STATE();
    case 1761:
      if (lookahead == 'r') ADVANCE(224);
      END_STATE();
    case 1762:
      if (lookahead == 'r') ADVANCE(1196);
      END_STATE();
    case 1763:
      if (lookahead == 'r') ADVANCE(200);
      END_STATE();
    case 1764:
      if (lookahead == 'r') ADVANCE(496);
      END_STATE();
    case 1765:
      if (lookahead == 'r') ADVANCE(364);
      END_STATE();
    case 1766:
      if (lookahead == 'r') ADVANCE(2172);
      END_STATE();
    case 1767:
      if (lookahead == 'r') ADVANCE(157);
      END_STATE();
    case 1768:
      if (lookahead == 'r') ADVANCE(175);
      END_STATE();
    case 1769:
      if (lookahead == 'r') ADVANCE(1750);
      END_STATE();
    case 1770:
      if (lookahead == 'r') ADVANCE(315);
      END_STATE();
    case 1771:
      if (lookahead == 'r') ADVANCE(478);
      END_STATE();
    case 1772:
      if (lookahead == 'r') ADVANCE(755);
      END_STATE();
    case 1773:
      if (lookahead == 'r') ADVANCE(358);
      END_STATE();
    case 1774:
      if (lookahead == 'r') ADVANCE(282);
      END_STATE();
    case 1775:
      if (lookahead == 'r') ADVANCE(1958);
      END_STATE();
    case 1776:
      if (lookahead == 'r') ADVANCE(1958);
      if (lookahead == 'u') ADVANCE(689);
      END_STATE();
    case 1777:
      if (lookahead == 'r') ADVANCE(243);
      END_STATE();
    case 1778:
      if (lookahead == 'r') ADVANCE(908);
      END_STATE();
    case 1779:
      if (lookahead == 'r') ADVANCE(218);
      END_STATE();
    case 1780:
      if (lookahead == 'r') ADVANCE(227);
      END_STATE();
    case 1781:
      if (lookahead == 'r') ADVANCE(214);
      END_STATE();
    case 1782:
      if (lookahead == 'r') ADVANCE(490);
      END_STATE();
    case 1783:
      if (lookahead == 'r') ADVANCE(754);
      END_STATE();
    case 1784:
      if (lookahead == 'r') ADVANCE(583);
      END_STATE();
    case 1785:
      if (lookahead == 'r') ADVANCE(1223);
      END_STATE();
    case 1786:
      if (lookahead == 'r') ADVANCE(317);
      END_STATE();
    case 1787:
      if (lookahead == 'r') ADVANCE(300);
      END_STATE();
    case 1788:
      if (lookahead == 'r') ADVANCE(306);
      END_STATE();
    case 1789:
      if (lookahead == 'r') ADVANCE(884);
      END_STATE();
    case 1790:
      if (lookahead == 'r') ADVANCE(876);
      END_STATE();
    case 1791:
      if (lookahead == 'r') ADVANCE(1588);
      END_STATE();
    case 1792:
      if (lookahead == 'r') ADVANCE(311);
      END_STATE();
    case 1793:
      if (lookahead == 'r') ADVANCE(1349);
      END_STATE();
    case 1794:
      if (lookahead == 'r') ADVANCE(340);
      END_STATE();
    case 1795:
      if (lookahead == 'r') ADVANCE(435);
      END_STATE();
    case 1796:
      if (lookahead == 'r') ADVANCE(351);
      END_STATE();
    case 1797:
      if (lookahead == 'r') ADVANCE(866);
      END_STATE();
    case 1798:
      if (lookahead == 'r') ADVANCE(1677);
      END_STATE();
    case 1799:
      if (lookahead == 'r') ADVANCE(1003);
      END_STATE();
    case 1800:
      if (lookahead == 'r') ADVANCE(2018);
      END_STATE();
    case 1801:
      if (lookahead == 'r') ADVANCE(314);
      END_STATE();
    case 1802:
      if (lookahead == 'r') ADVANCE(537);
      END_STATE();
    case 1803:
      if (lookahead == 'r') ADVANCE(1629);
      END_STATE();
    case 1804:
      if (lookahead == 'r') ADVANCE(1339);
      END_STATE();
    case 1805:
      if (lookahead == 'r') ADVANCE(2146);
      END_STATE();
    case 1806:
      if (lookahead == 'r') ADVANCE(2148);
      END_STATE();
    case 1807:
      if (lookahead == 'r') ADVANCE(1592);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 1808:
      if (lookahead == 'r') ADVANCE(1903);
      END_STATE();
    case 1809:
      if (lookahead == 'r') ADVANCE(996);
      END_STATE();
    case 1810:
      if (lookahead == 'r') ADVANCE(1587);
      END_STATE();
    case 1811:
      if (lookahead == 'r') ADVANCE(564);
      END_STATE();
    case 1812:
      if (lookahead == 'r') ADVANCE(2124);
      END_STATE();
    case 1813:
      if (lookahead == 'r') ADVANCE(2147);
      END_STATE();
    case 1814:
      if (lookahead == 'r') ADVANCE(1054);
      END_STATE();
    case 1815:
      if (lookahead == 'r') ADVANCE(1195);
      END_STATE();
    case 1816:
      if (lookahead == 'r') ADVANCE(1504);
      END_STATE();
    case 1817:
      if (lookahead == 'r') ADVANCE(762);
      END_STATE();
    case 1818:
      if (lookahead == 'r') ADVANCE(926);
      END_STATE();
    case 1819:
      if (lookahead == 'r') ADVANCE(1608);
      END_STATE();
    case 1820:
      if (lookahead == 'r') ADVANCE(1300);
      END_STATE();
    case 1821:
      if (lookahead == 'r') ADVANCE(1613);
      END_STATE();
    case 1822:
      if (lookahead == 'r') ADVANCE(1966);
      END_STATE();
    case 1823:
      if (lookahead == 'r') ADVANCE(1601);
      END_STATE();
    case 1824:
      if (lookahead == 'r') ADVANCE(1923);
      END_STATE();
    case 1825:
      if (lookahead == 'r') ADVANCE(835);
      END_STATE();
    case 1826:
      if (lookahead == 'r') ADVANCE(1606);
      END_STATE();
    case 1827:
      if (lookahead == 'r') ADVANCE(1759);
      END_STATE();
    case 1828:
      if (lookahead == 'r') ADVANCE(600);
      END_STATE();
    case 1829:
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 1830:
      if (lookahead == 'r') ADVANCE(2036);
      END_STATE();
    case 1831:
      if (lookahead == 'r') ADVANCE(1715);
      END_STATE();
    case 1832:
      if (lookahead == 'r') ADVANCE(2029);
      END_STATE();
    case 1833:
      if (lookahead == 'r') ADVANCE(1945);
      END_STATE();
    case 1834:
      if (lookahead == 'r') ADVANCE(1309);
      END_STATE();
    case 1835:
      if (lookahead == 'r') ADVANCE(1953);
      END_STATE();
    case 1836:
      if (lookahead == 'r') ADVANCE(978);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 1837:
      if (lookahead == 'r') ADVANCE(1982);
      END_STATE();
    case 1838:
      if (lookahead == 'r') ADVANCE(603);
      END_STATE();
    case 1839:
      if (lookahead == 'r') ADVANCE(1271);
      END_STATE();
    case 1840:
      if (lookahead == 'r') ADVANCE(1610);
      END_STATE();
    case 1841:
      if (lookahead == 'r') ADVANCE(1545);
      END_STATE();
    case 1842:
      if (lookahead == 'r') ADVANCE(1931);
      END_STATE();
    case 1843:
      if (lookahead == 'r') ADVANCE(1534);
      END_STATE();
    case 1844:
      if (lookahead == 'r') ADVANCE(1249);
      END_STATE();
    case 1845:
      if (lookahead == 'r') ADVANCE(1573);
      END_STATE();
    case 1846:
      if (lookahead == 'r') ADVANCE(1397);
      END_STATE();
    case 1847:
      if (lookahead == 'r') ADVANCE(1306);
      END_STATE();
    case 1848:
      if (lookahead == 'r') ADVANCE(1401);
      END_STATE();
    case 1849:
      if (lookahead == 'r') ADVANCE(1001);
      END_STATE();
    case 1850:
      if (lookahead == 'r') ADVANCE(1611);
      END_STATE();
    case 1851:
      if (lookahead == 'r') ADVANCE(1962);
      END_STATE();
    case 1852:
      if (lookahead == 'r') ADVANCE(942);
      if (lookahead == 't') ADVANCE(563);
      END_STATE();
    case 1853:
      if (lookahead == 'r') ADVANCE(1615);
      END_STATE();
    case 1854:
      if (lookahead == 'r') ADVANCE(1616);
      END_STATE();
    case 1855:
      if (lookahead == 'r') ADVANCE(1620);
      END_STATE();
    case 1856:
      if (lookahead == 'r') ADVANCE(1621);
      END_STATE();
    case 1857:
      if (lookahead == 'r') ADVANCE(1622);
      END_STATE();
    case 1858:
      if (lookahead == 'r') ADVANCE(1094);
      END_STATE();
    case 1859:
      if (lookahead == 'r') ADVANCE(945);
      END_STATE();
    case 1860:
      if (lookahead == 'r') ADVANCE(1626);
      END_STATE();
    case 1861:
      if (lookahead == 'r') ADVANCE(1765);
      END_STATE();
    case 1862:
      if (lookahead == 'r') ADVANCE(1050);
      END_STATE();
    case 1863:
      if (lookahead == 'r') ADVANCE(1755);
      END_STATE();
    case 1864:
      if (lookahead == 'r') ADVANCE(1788);
      END_STATE();
    case 1865:
      if (lookahead == 'r') ADVANCE(1069);
      END_STATE();
    case 1866:
      if (lookahead == 'r') ADVANCE(971);
      END_STATE();
    case 1867:
      if (lookahead == 'r') ADVANCE(693);
      END_STATE();
    case 1868:
      if (lookahead == 'r') ADVANCE(608);
      END_STATE();
    case 1869:
      if (lookahead == 'r') ADVANCE(653);
      END_STATE();
    case 1870:
      if (lookahead == 'r') ADVANCE(1281);
      END_STATE();
    case 1871:
      if (lookahead == 'r') ADVANCE(1946);
      END_STATE();
    case 1872:
      if (lookahead == 'r') ADVANCE(847);
      END_STATE();
    case 1873:
      if (lookahead == 'r') ADVANCE(2026);
      END_STATE();
    case 1874:
      if (lookahead == 'r') ADVANCE(1017);
      END_STATE();
    case 1875:
      if (lookahead == 'r') ADVANCE(1147);
      if (lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 1876:
      if (lookahead == 'r') ADVANCE(2009);
      END_STATE();
    case 1877:
      if (lookahead == 'r') ADVANCE(2071);
      END_STATE();
    case 1878:
      if (lookahead == 'r') ADVANCE(1948);
      END_STATE();
    case 1879:
      if (lookahead == 'r') ADVANCE(1024);
      END_STATE();
    case 1880:
      if (lookahead == 'r') ADVANCE(1041);
      END_STATE();
    case 1881:
      if (lookahead == 'r') ADVANCE(1826);
      END_STATE();
    case 1882:
      if (lookahead == 'r') ADVANCE(632);
      END_STATE();
    case 1883:
      if (lookahead == 'r') ADVANCE(2024);
      END_STATE();
    case 1884:
      if (lookahead == 'r') ADVANCE(1840);
      END_STATE();
    case 1885:
      if (lookahead == 'r') ADVANCE(1850);
      END_STATE();
    case 1886:
      if (lookahead == 'r') ADVANCE(1854);
      END_STATE();
    case 1887:
      if (lookahead == 'r') ADVANCE(1855);
      END_STATE();
    case 1888:
      if (lookahead == 'r') ADVANCE(1856);
      END_STATE();
    case 1889:
      if (lookahead == 'r') ADVANCE(1860);
      END_STATE();
    case 1890:
      if (lookahead == 'r') ADVANCE(1059);
      END_STATE();
    case 1891:
      if (lookahead == 'r') ADVANCE(871);
      END_STATE();
    case 1892:
      if (lookahead == 'r') ADVANCE(1082);
      END_STATE();
    case 1893:
      if (lookahead == 'r') ADVANCE(473);
      END_STATE();
    case 1894:
      if (lookahead == 'r') ADVANCE(1728);
      END_STATE();
    case 1895:
      if (lookahead == 'r') ADVANCE(2047);
      END_STATE();
    case 1896:
      if (lookahead == 'r') ADVANCE(1730);
      END_STATE();
    case 1897:
      if (lookahead == 'r') ADVANCE(1091);
      END_STATE();
    case 1898:
      if (lookahead == 'r') ADVANCE(1092);
      END_STATE();
    case 1899:
      if (lookahead == 'r') ADVANCE(1093);
      END_STATE();
    case 1900:
      if (lookahead == 'r') ADVANCE(1333);
      END_STATE();
    case 1901:
      if (lookahead == 'r') ADVANCE(1335);
      END_STATE();
    case 1902:
      if (lookahead == 'r') ADVANCE(1336);
      END_STATE();
    case 1903:
      if (lookahead == 's') ADVANCE(152);
      END_STATE();
    case 1904:
      if (lookahead == 's') ADVANCE(1051);
      END_STATE();
    case 1905:
      if (lookahead == 's') ADVANCE(322);
      if (lookahead == 'u') ADVANCE(1489);
      END_STATE();
    case 1906:
      if (lookahead == 's') ADVANCE(472);
      END_STATE();
    case 1907:
      if (lookahead == 's') ADVANCE(498);
      END_STATE();
    case 1908:
      if (lookahead == 's') ADVANCE(201);
      END_STATE();
    case 1909:
      if (lookahead == 's') ADVANCE(357);
      END_STATE();
    case 1910:
      if (lookahead == 's') ADVANCE(220);
      END_STATE();
    case 1911:
      if (lookahead == 's') ADVANCE(177);
      END_STATE();
    case 1912:
      if (lookahead == 's') ADVANCE(179);
      END_STATE();
    case 1913:
      if (lookahead == 's') ADVANCE(178);
      END_STATE();
    case 1914:
      if (lookahead == 's') ADVANCE(1958);
      END_STATE();
    case 1915:
      if (lookahead == 's') ADVANCE(492);
      END_STATE();
    case 1916:
      if (lookahead == 's') ADVANCE(1623);
      END_STATE();
    case 1917:
      if (lookahead == 's') ADVANCE(284);
      END_STATE();
    case 1918:
      if (lookahead == 's') ADVANCE(1223);
      END_STATE();
    case 1919:
      if (lookahead == 's') ADVANCE(506);
      END_STATE();
    case 1920:
      if (lookahead == 's') ADVANCE(346);
      END_STATE();
    case 1921:
      if (lookahead == 's') ADVANCE(250);
      END_STATE();
    case 1922:
      if (lookahead == 's') ADVANCE(205);
      END_STATE();
    case 1923:
      if (lookahead == 's') ADVANCE(1677);
      END_STATE();
    case 1924:
      if (lookahead == 's') ADVANCE(1183);
      END_STATE();
    case 1925:
      if (lookahead == 's') ADVANCE(2149);
      END_STATE();
    case 1926:
      if (lookahead == 's') ADVANCE(339);
      END_STATE();
    case 1927:
      if (lookahead == 's') ADVANCE(1184);
      END_STATE();
    case 1928:
      if (lookahead == 's') ADVANCE(1188);
      END_STATE();
    case 1929:
      if (lookahead == 's') ADVANCE(1903);
      END_STATE();
    case 1930:
      if (lookahead == 's') ADVANCE(1259);
      END_STATE();
    case 1931:
      if (lookahead == 's') ADVANCE(1358);
      END_STATE();
    case 1932:
      if (lookahead == 's') ADVANCE(2052);
      END_STATE();
    case 1933:
      if (lookahead == 's') ADVANCE(894);
      END_STATE();
    case 1934:
      if (lookahead == 's') ADVANCE(926);
      END_STATE();
    case 1935:
      if (lookahead == 's') ADVANCE(973);
      END_STATE();
    case 1936:
      if (lookahead == 's') ADVANCE(2101);
      END_STATE();
    case 1937:
      if (lookahead == 's') ADVANCE(2057);
      END_STATE();
    case 1938:
      if (lookahead == 's') ADVANCE(2003);
      END_STATE();
    case 1939:
      if (lookahead == 's') ADVANCE(1383);
      END_STATE();
    case 1940:
      if (lookahead == 's') ADVANCE(1915);
      END_STATE();
    case 1941:
      if (lookahead == 's') ADVANCE(1992);
      END_STATE();
    case 1942:
      if (lookahead == 's') ADVANCE(1996);
      END_STATE();
    case 1943:
      if (lookahead == 's') ADVANCE(1321);
      END_STATE();
    case 1944:
      if (lookahead == 's') ADVANCE(1989);
      END_STATE();
    case 1945:
      if (lookahead == 's') ADVANCE(1011);
      END_STATE();
    case 1946:
      if (lookahead == 's') ADVANCE(949);
      END_STATE();
    case 1947:
      if (lookahead == 's') ADVANCE(985);
      END_STATE();
    case 1948:
      if (lookahead == 's') ADVANCE(970);
      END_STATE();
    case 1949:
      if (lookahead == 's') ADVANCE(1193);
      END_STATE();
    case 1950:
      if (lookahead == 's') ADVANCE(2059);
      END_STATE();
    case 1951:
      if (lookahead == 's') ADVANCE(2014);
      END_STATE();
    case 1952:
      if (lookahead == 's') ADVANCE(2002);
      END_STATE();
    case 1953:
      if (lookahead == 's') ADVANCE(1016);
      END_STATE();
    case 1954:
      if (lookahead == 's') ADVANCE(1268);
      END_STATE();
    case 1955:
      if (lookahead == 's') ADVANCE(1031);
      END_STATE();
    case 1956:
      if (lookahead == 's') ADVANCE(1087);
      END_STATE();
    case 1957:
      if (lookahead == 't') ADVANCE(2139);
      END_STATE();
    case 1958:
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 1959:
      if (lookahead == 't') ADVANCE(152);
      if (lookahead == 'u') ADVANCE(751);
      END_STATE();
    case 1960:
      if (lookahead == 't') ADVANCE(152);
      if (lookahead == 'w') ADVANCE(1521);
      END_STATE();
    case 1961:
      if (lookahead == 't') ADVANCE(396);
      END_STATE();
    case 1962:
      if (lookahead == 't') ADVANCE(514);
      END_STATE();
    case 1963:
      if (lookahead == 't') ADVANCE(701);
      END_STATE();
    case 1964:
      if (lookahead == 't') ADVANCE(363);
      END_STATE();
    case 1965:
      if (lookahead == 't') ADVANCE(415);
      END_STATE();
    case 1966:
      if (lookahead == 't') ADVANCE(455);
      END_STATE();
    case 1967:
      if (lookahead == 't') ADVANCE(700);
      END_STATE();
    case 1968:
      if (lookahead == 't') ADVANCE(265);
      END_STATE();
    case 1969:
      if (lookahead == 't') ADVANCE(1156);
      END_STATE();
    case 1970:
      if (lookahead == 't') ADVANCE(510);
      END_STATE();
    case 1971:
      if (lookahead == 't') ADVANCE(521);
      END_STATE();
    case 1972:
      if (lookahead == 't') ADVANCE(418);
      END_STATE();
    case 1973:
      if (lookahead == 't') ADVANCE(421);
      END_STATE();
    case 1974:
      if (lookahead == 't') ADVANCE(525);
      END_STATE();
    case 1975:
      if (lookahead == 't') ADVANCE(708);
      END_STATE();
    case 1976:
      if (lookahead == 't') ADVANCE(419);
      END_STATE();
    case 1977:
      if (lookahead == 't') ADVANCE(420);
      END_STATE();
    case 1978:
      if (lookahead == 't') ADVANCE(1750);
      END_STATE();
    case 1979:
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 1980:
      if (lookahead == 't') ADVANCE(1008);
      END_STATE();
    case 1981:
      if (lookahead == 't') ADVANCE(174);
      END_STATE();
    case 1982:
      if (lookahead == 't') ADVANCE(1478);
      END_STATE();
    case 1983:
      if (lookahead == 't') ADVANCE(471);
      END_STATE();
    case 1984:
      if (lookahead == 't') ADVANCE(754);
      END_STATE();
    case 1985:
      if (lookahead == 't') ADVANCE(2140);
      END_STATE();
    case 1986:
      if (lookahead == 't') ADVANCE(256);
      if (lookahead == 'v') ADVANCE(1295);
      END_STATE();
    case 1987:
      if (lookahead == 't') ADVANCE(360);
      END_STATE();
    case 1988:
      if (lookahead == 't') ADVANCE(697);
      END_STATE();
    case 1989:
      if (lookahead == 't') ADVANCE(1588);
      END_STATE();
    case 1990:
      if (lookahead == 't') ADVANCE(416);
      END_STATE();
    case 1991:
      if (lookahead == 't') ADVANCE(1035);
      END_STATE();
    case 1992:
      if (lookahead == 't') ADVANCE(212);
      END_STATE();
    case 1993:
      if (lookahead == 't') ADVANCE(2144);
      END_STATE();
    case 1994:
      if (lookahead == 't') ADVANCE(250);
      END_STATE();
    case 1995:
      if (lookahead == 't') ADVANCE(522);
      END_STATE();
    case 1996:
      if (lookahead == 't') ADVANCE(334);
      END_STATE();
    case 1997:
      if (lookahead == 't') ADVANCE(1183);
      END_STATE();
    case 1998:
      if (lookahead == 't') ADVANCE(2145);
      END_STATE();
    case 1999:
      if (lookahead == 't') ADVANCE(313);
      END_STATE();
    case 2000:
      if (lookahead == 't') ADVANCE(1188);
      END_STATE();
    case 2001:
      if (lookahead == 't') ADVANCE(312);
      END_STATE();
    case 2002:
      if (lookahead == 't') ADVANCE(1903);
      END_STATE();
    case 2003:
      if (lookahead == 't') ADVANCE(259);
      END_STATE();
    case 2004:
      if (lookahead == 't') ADVANCE(264);
      END_STATE();
    case 2005:
      if (lookahead == 't') ADVANCE(1185);
      END_STATE();
    case 2006:
      if (lookahead == 't') ADVANCE(1486);
      END_STATE();
    case 2007:
      if (lookahead == 't') ADVANCE(926);
      END_STATE();
    case 2008:
      if (lookahead == 't') ADVANCE(1658);
      END_STATE();
    case 2009:
      if (lookahead == 't') ADVANCE(1909);
      END_STATE();
    case 2010:
      if (lookahead == 't') ADVANCE(1213);
      END_STATE();
    case 2011:
      if (lookahead == 't') ADVANCE(1501);
      END_STATE();
    case 2012:
      if (lookahead == 't') ADVANCE(957);
      END_STATE();
    case 2013:
      if (lookahead == 't') ADVANCE(1935);
      END_STATE();
    case 2014:
      if (lookahead == 't') ADVANCE(999);
      END_STATE();
    case 2015:
      if (lookahead == 't') ADVANCE(1597);
      END_STATE();
    case 2016:
      if (lookahead == 't') ADVANCE(1785);
      END_STATE();
    case 2017:
      if (lookahead == 't') ADVANCE(634);
      END_STATE();
    case 2018:
      if (lookahead == 't') ADVANCE(1294);
      END_STATE();
    case 2019:
      if (lookahead == 't') ADVANCE(1271);
      END_STATE();
    case 2020:
      if (lookahead == 't') ADVANCE(1305);
      END_STATE();
    case 2021:
      if (lookahead == 't') ADVANCE(1594);
      END_STATE();
    case 2022:
      if (lookahead == 't') ADVANCE(1296);
      END_STATE();
    case 2023:
      if (lookahead == 't') ADVANCE(1319);
      END_STATE();
    case 2024:
      if (lookahead == 't') ADVANCE(1313);
      END_STATE();
    case 2025:
      if (lookahead == 't') ADVANCE(1273);
      END_STATE();
    case 2026:
      if (lookahead == 't') ADVANCE(1753);
      END_STATE();
    case 2027:
      if (lookahead == 't') ADVANCE(1085);
      END_STATE();
    case 2028:
      if (lookahead == 't') ADVANCE(706);
      END_STATE();
    case 2029:
      if (lookahead == 't') ADVANCE(1553);
      END_STATE();
    case 2030:
      if (lookahead == 't') ADVANCE(1258);
      END_STATE();
    case 2031:
      if (lookahead == 't') ADVANCE(1651);
      END_STATE();
    case 2032:
      if (lookahead == 't') ADVANCE(1554);
      END_STATE();
    case 2033:
      if (lookahead == 't') ADVANCE(1838);
      END_STATE();
    case 2034:
      if (lookahead == 't') ADVANCE(1034);
      END_STATE();
    case 2035:
      if (lookahead == 't') ADVANCE(642);
      END_STATE();
    case 2036:
      if (lookahead == 't') ADVANCE(1235);
      END_STATE();
    case 2037:
      if (lookahead == 't') ADVANCE(1036);
      END_STATE();
    case 2038:
      if (lookahead == 't') ADVANCE(1038);
      END_STATE();
    case 2039:
      if (lookahead == 't') ADVANCE(1040);
      END_STATE();
    case 2040:
      if (lookahead == 't') ADVANCE(1043);
      END_STATE();
    case 2041:
      if (lookahead == 't') ADVANCE(1061);
      END_STATE();
    case 2042:
      if (lookahead == 't') ADVANCE(1045);
      END_STATE();
    case 2043:
      if (lookahead == 't') ADVANCE(1214);
      END_STATE();
    case 2044:
      if (lookahead == 't') ADVANCE(685);
      END_STATE();
    case 2045:
      if (lookahead == 't') ADVANCE(1304);
      END_STATE();
    case 2046:
      if (lookahead == 't') ADVANCE(1310);
      END_STATE();
    case 2047:
      if (lookahead == 't') ADVANCE(1320);
      END_STATE();
    case 2048:
      if (lookahead == 't') ADVANCE(705);
      END_STATE();
    case 2049:
      if (lookahead == 't') ADVANCE(694);
      END_STATE();
    case 2050:
      if (lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 2051:
      if (lookahead == 'u') ADVANCE(622);
      END_STATE();
    case 2052:
      if (lookahead == 'u') ADVANCE(2162);
      END_STATE();
    case 2053:
      if (lookahead == 'u') ADVANCE(1588);
      END_STATE();
    case 2054:
      if (lookahead == 'u') ADVANCE(311);
      END_STATE();
    case 2055:
      if (lookahead == 'u') ADVANCE(724);
      END_STATE();
    case 2056:
      if (lookahead == 'u') ADVANCE(737);
      END_STATE();
    case 2057:
      if (lookahead == 'u') ADVANCE(735);
      END_STATE();
    case 2058:
      if (lookahead == 'u') ADVANCE(1677);
      END_STATE();
    case 2059:
      if (lookahead == 'u') ADVANCE(718);
      END_STATE();
    case 2060:
      if (lookahead == 'u') ADVANCE(1907);
      END_STATE();
    case 2061:
      if (lookahead == 'u') ADVANCE(742);
      END_STATE();
    case 2062:
      if (lookahead == 'u') ADVANCE(1994);
      END_STATE();
    case 2063:
      if (lookahead == 'u') ADVANCE(1994);
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 2064:
      if (lookahead == 'u') ADVANCE(1463);
      END_STATE();
    case 2065:
      if (lookahead == 'u') ADVANCE(1903);
      END_STATE();
    case 2066:
      if (lookahead == 'u') ADVANCE(1009);
      END_STATE();
    case 2067:
      if (lookahead == 'u') ADVANCE(1919);
      END_STATE();
    case 2068:
      if (lookahead == 'u') ADVANCE(926);
      END_STATE();
    case 2069:
      if (lookahead == 'u') ADVANCE(1466);
      END_STATE();
    case 2070:
      if (lookahead == 'u') ADVANCE(807);
      END_STATE();
    case 2071:
      if (lookahead == 'u') ADVANCE(959);
      END_STATE();
    case 2072:
      if (lookahead == 'u') ADVANCE(600);
      END_STATE();
    case 2073:
      if (lookahead == 'u') ADVANCE(1910);
      END_STATE();
    case 2074:
      if (lookahead == 'u') ADVANCE(1589);
      END_STATE();
    case 2075:
      if (lookahead == 'u') ADVANCE(1374);
      END_STATE();
    case 2076:
      if (lookahead == 'u') ADVANCE(1934);
      END_STATE();
    case 2077:
      if (lookahead == 'u') ADVANCE(841);
      END_STATE();
    case 2078:
      if (lookahead == 'u') ADVANCE(1921);
      END_STATE();
    case 2079:
      if (lookahead == 'u') ADVANCE(1242);
      END_STATE();
    case 2080:
      if (lookahead == 'u') ADVANCE(994);
      END_STATE();
    case 2081:
      if (lookahead == 'u') ADVANCE(621);
      END_STATE();
    case 2082:
      if (lookahead == 'u') ADVANCE(1307);
      END_STATE();
    case 2083:
      if (lookahead == 'u') ADVANCE(1241);
      END_STATE();
    case 2084:
      if (lookahead == 'u') ADVANCE(1781);
      END_STATE();
    case 2085:
      if (lookahead == 'u') ADVANCE(1753);
      END_STATE();
    case 2086:
      if (lookahead == 'u') ADVANCE(1782);
      END_STATE();
    case 2087:
      if (lookahead == 'u') ADVANCE(1846);
      END_STATE();
    case 2088:
      if (lookahead == 'u') ADVANCE(1848);
      END_STATE();
    case 2089:
      if (lookahead == 'u') ADVANCE(1193);
      END_STATE();
    case 2090:
      if (lookahead == 'u') ADVANCE(1476);
      END_STATE();
    case 2091:
      if (lookahead == 'u') ADVANCE(1547);
      END_STATE();
    case 2092:
      if (lookahead == 'u') ADVANCE(1941);
      END_STATE();
    case 2093:
      if (lookahead == 'u') ADVANCE(1276);
      END_STATE();
    case 2094:
      if (lookahead == 'u') ADVANCE(1400);
      END_STATE();
    case 2095:
      if (lookahead == 'u') ADVANCE(2007);
      END_STATE();
    case 2096:
      if (lookahead == 'u') ADVANCE(641);
      END_STATE();
    case 2097:
      if (lookahead == 'u') ADVANCE(1416);
      END_STATE();
    case 2098:
      if (lookahead == 'u') ADVANCE(1662);
      END_STATE();
    case 2099:
      if (lookahead == 'u') ADVANCE(1474);
      END_STATE();
    case 2100:
      if (lookahead == 'u') ADVANCE(647);
      END_STATE();
    case 2101:
      if (lookahead == 'u') ADVANCE(1862);
      END_STATE();
    case 2102:
      if (lookahead == 'u') ADVANCE(1455);
      END_STATE();
    case 2103:
      if (lookahead == 'u') ADVANCE(690);
      END_STATE();
    case 2104:
      if (lookahead == 'u') ADVANCE(748);
      END_STATE();
    case 2105:
      if (lookahead == 'u') ADVANCE(692);
      END_STATE();
    case 2106:
      if (lookahead == 'u') ADVANCE(747);
      if (lookahead == 'w') ADVANCE(1513);
      END_STATE();
    case 2107:
      if (lookahead == 'u') ADVANCE(750);
      END_STATE();
    case 2108:
      if (lookahead == 'v') ADVANCE(152);
      END_STATE();
    case 2109:
      if (lookahead == 'v') ADVANCE(152);
      if (lookahead == 'w') ADVANCE(925);
      END_STATE();
    case 2110:
      if (lookahead == 'v') ADVANCE(166);
      END_STATE();
    case 2111:
      if (lookahead == 'v') ADVANCE(250);
      END_STATE();
    case 2112:
      if (lookahead == 'v') ADVANCE(1003);
      if (lookahead == 'w') ADVANCE(1002);
      END_STATE();
    case 2113:
      if (lookahead == 'v') ADVANCE(926);
      END_STATE();
    case 2114:
      if (lookahead == 'v') ADVANCE(950);
      END_STATE();
    case 2115:
      if (lookahead == 'v') ADVANCE(1089);
      END_STATE();
    case 2116:
      if (lookahead == 'w') ADVANCE(152);
      END_STATE();
    case 2117:
      if (lookahead == 'w') ADVANCE(158);
      END_STATE();
    case 2118:
      if (lookahead == 'w') ADVANCE(162);
      END_STATE();
    case 2119:
      if (lookahead == 'w') ADVANCE(161);
      END_STATE();
    case 2120:
      if (lookahead == 'w') ADVANCE(160);
      END_STATE();
    case 2121:
      if (lookahead == 'w') ADVANCE(1588);
      END_STATE();
    case 2122:
      if (lookahead == 'w') ADVANCE(546);
      END_STATE();
    case 2123:
      if (lookahead == 'w') ADVANCE(334);
      END_STATE();
    case 2124:
      if (lookahead == 'w') ADVANCE(966);
      END_STATE();
    case 2125:
      if (lookahead == 'w') ADVANCE(1903);
      END_STATE();
    case 2126:
      if (lookahead == 'w') ADVANCE(1054);
      if (lookahead == 'a' ||
          lookahead == 'h') ADVANCE(152);
      END_STATE();
    case 2127:
      if (lookahead == 'w') ADVANCE(1504);
      END_STATE();
    case 2128:
      if (lookahead == 'w') ADVANCE(1521);
      END_STATE();
    case 2129:
      if (lookahead == 'w') ADVANCE(1518);
      END_STATE();
    case 2130:
      if (lookahead == 'w') ADVANCE(1514);
      END_STATE();
    case 2131:
      if (lookahead == 'w') ADVANCE(1302);
      END_STATE();
    case 2132:
      if (lookahead == 'w') ADVANCE(353);
      END_STATE();
    case 2133:
      if (lookahead == 'w') ADVANCE(1428);
      END_STATE();
    case 2134:
      if (lookahead == 'w') ADVANCE(1586);
      END_STATE();
    case 2135:
      if (lookahead == 'w') ADVANCE(1585);
      END_STATE();
    case 2136:
      if (lookahead == 'x') ADVANCE(152);
      END_STATE();
    case 2137:
      if (lookahead == 'x') ADVANCE(1958);
      END_STATE();
    case 2138:
      if (lookahead == 'x') ADVANCE(259);
      END_STATE();
    case 2139:
      if (lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 2140:
      if (lookahead == 'y') ADVANCE(513);
      END_STATE();
    case 2141:
      if (lookahead == 'y') ADVANCE(984);
      END_STATE();
    case 2142:
      if (lookahead == 'y') ADVANCE(460);
      END_STATE();
    case 2143:
      if (lookahead == 'y') ADVANCE(475);
      END_STATE();
    case 2144:
      if (lookahead == 'y') ADVANCE(342);
      END_STATE();
    case 2145:
      if (lookahead == 'y') ADVANCE(2108);
      END_STATE();
    case 2146:
      if (lookahead == 'y') ADVANCE(512);
      END_STATE();
    case 2147:
      if (lookahead == 'y') ADVANCE(519);
      END_STATE();
    case 2148:
      if (lookahead == 'y') ADVANCE(488);
      END_STATE();
    case 2149:
      if (lookahead == 'y') ADVANCE(1463);
      END_STATE();
    case 2150:
      if (lookahead == 'y') ADVANCE(925);
      END_STATE();
    case 2151:
      if (lookahead == 'y') ADVANCE(1903);
      END_STATE();
    case 2152:
      if (lookahead == 'y') ADVANCE(973);
      END_STATE();
    case 2153:
      if (lookahead == 'z') ADVANCE(1665);
      END_STATE();
    case 2154:
      if (lookahead == 'z') ADVANCE(577);
      END_STATE();
    case 2155:
      if (lookahead == 'z') ADVANCE(1281);
      END_STATE();
    case 2156:
      if (lookahead == 'z') ADVANCE(1012);
      END_STATE();
    case 2157:
      if (lookahead == '2' ||
          lookahead == '4') ADVANCE(152);
      END_STATE();
    case 2158:
      if (lookahead == '3' ||
          lookahead == '5') ADVANCE(152);
      END_STATE();
    case 2159:
      if (lookahead == '6' ||
          lookahead == '8') ADVANCE(152);
      END_STATE();
    case 2160:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 2161:
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(2176);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(380);
      END_STATE();
    case 2162:
      if (lookahead == 'b' ||
          lookahead == 'p') ADVANCE(152);
      END_STATE();
    case 2163:
      if (lookahead == 'b' ||
          lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 2164:
      if (lookahead == 'd' ||
          lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 2165:
      if (lookahead == 'e' ||
          lookahead == 'k') ADVANCE(152);
      END_STATE();
    case 2166:
      if (lookahead == 'e' ||
          lookahead == 't') ADVANCE(152);
      END_STATE();
    case 2167:
      if (lookahead == 'f' ||
          lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 2168:
      if (lookahead == 'l' ||
          lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 2169:
      if (lookahead == 'o' ||
          lookahead == 'u') ADVANCE(152);
      END_STATE();
    case 2170:
      if (lookahead == 'r' ||
          lookahead == 'y') ADVANCE(152);
      END_STATE();
    case 2171:
      if (lookahead == '3' ||
          lookahead == '4') ADVANCE(152);
      END_STATE();
    case 2172:
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(152);
      END_STATE();
    case 2173:
      if (('a' <= lookahead && lookahead <= 'c')) ADVANCE(152);
      END_STATE();
    case 2174:
      if (lookahead == 'L' ||
          lookahead == 'R' ||
          lookahead == 'l' ||
          lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 2175:
      if (('a' <= lookahead && lookahead <= 'h')) ADVANCE(152);
      END_STATE();
    case 2176:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(385);
      END_STATE();
    case 2177:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(19);
      END_STATE();
    case 2178:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(142);
      END_STATE();
    case 2179:
      if (eof) ADVANCE(2183);
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2206,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2191,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 2180:
      if (eof) ADVANCE(2183);
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2191,
        '=', 2211,
        '>', 2193,
        '?', 2213,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 2181:
      if (eof) ADVANCE(2183);
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2191,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2189,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 2182:
      if (eof) ADVANCE(2183);
      ADVANCE_MAP(
        '\t', 2242,
        '\n', 2226,
        '\r', 2227,
        ' ', 2244,
        '!', 2194,
        '"', 2195,
        '#', 2196,
        '$', 2197,
        '%', 2198,
        '&', 2200,
        '\'', 2201,
        '(', 2224,
        ')', 2225,
        '*', 2202,
        '+', 2203,
        ',', 2204,
        '-', 2205,
        '.', 2207,
        '/', 2208,
        ':', 2209,
        ';', 2210,
        '<', 2191,
        '=', 2211,
        '>', 2193,
        '?', 2212,
        '@', 2214,
        '[', 2187,
        '\\', 2216,
        ']', 2188,
        '^', 2217,
        '_', 2218,
        '`', 2219,
        '{', 2220,
        '|', 2221,
        '}', 2222,
        '~', 2223,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(2246);
      if (lookahead != 0) ADVANCE(2245);
      END_STATE();
    case 2183:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 2184:
      ACCEPT_TOKEN(sym__backslash_escape);
      END_STATE();
    case 2185:
      ACCEPT_TOKEN(sym_entity_reference);
      END_STATE();
    case 2186:
      ACCEPT_TOKEN(sym_numeric_character_reference);
      END_STATE();
    case 2187:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 2188:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 2189:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      if (lookahead == ']') ADVANCE(387);
      END_STATE();
    case 2190:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 2191:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '!') ADVANCE(17);
      if (lookahead == '?') ADVANCE(2236);
      if (('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          ('^' <= lookahead && lookahead <= '`') ||
          ('{' <= lookahead && lookahead <= '~')) ADVANCE(390);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(389);
      END_STATE();
    case 2192:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '!') ADVANCE(20);
      if (lookahead == '?') ADVANCE(2235);
      END_STATE();
    case 2193:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 2194:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 2195:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 2196:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 2197:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 2198:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 2199:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 2200:
      ACCEPT_TOKEN(anon_sym_AMP);
      ADVANCE_MAP(
        '#', 2161,
        'A', 461,
        'B', 531,
        'C', 484,
        'D', 444,
        'E', 500,
        'F', 757,
        'G', 184,
        'H', 391,
        'I', 467,
        'J', 859,
        'K', 483,
        'L', 183,
        'M', 573,
        'N', 491,
        'O', 469,
        'P', 569,
        'Q', 523,
        'R', 424,
        'S', 480,
        'T', 481,
        'U', 550,
        'V', 453,
        'W', 811,
        'X', 1099,
        'Y', 394,
        'Z', 485,
        'a', 529,
        'b', 499,
        'c', 536,
        'd', 400,
        'e', 446,
        'f', 578,
        'g', 173,
        'h', 402,
        'i', 530,
        'j', 858,
        'k', 663,
        'l', 153,
        'm', 447,
        'n', 477,
        'o', 511,
        'p', 601,
        'q', 1100,
        'r', 392,
        's', 658,
        't', 553,
        'u', 401,
        'v', 398,
        'w', 810,
        'x', 766,
        'y', 568,
        'z', 659,
      );
      END_STATE();
    case 2201:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 2202:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 2203:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 2204:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 2205:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 2206:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(386);
      END_STATE();
    case 2207:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 2208:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 2209:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 2210:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 2211:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 2212:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 2213:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '>') ADVANCE(2237);
      END_STATE();
    case 2214:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 2215:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 2216:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (('!' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          ('[' <= lookahead && lookahead <= '`') ||
          ('{' <= lookahead && lookahead <= '~')) ADVANCE(2184);
      END_STATE();
    case 2217:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 2218:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2219:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 2220:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 2221:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 2222:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 2223:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 2224:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 2225:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 2226:
      ACCEPT_TOKEN(sym__newline_token);
      END_STATE();
    case 2227:
      ACCEPT_TOKEN(sym__newline_token);
      if (lookahead == '\n') ADVANCE(2226);
      END_STATE();
    case 2228:
      ACCEPT_TOKEN(sym_uri_autolink);
      END_STATE();
    case 2229:
      ACCEPT_TOKEN(sym_email_autolink);
      END_STATE();
    case 2230:
      ACCEPT_TOKEN(sym__attribute_name);
      if (lookahead == '-' ||
          lookahead == '.' ||
          ('0' <= lookahead && lookahead <= ':') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(2230);
      END_STATE();
    case 2231:
      ACCEPT_TOKEN(aux_sym__attribute_value_token1);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != ' ' &&
          lookahead != '"' &&
          lookahead != '\'' &&
          (lookahead < '<' || '>' < lookahead) &&
          lookahead != '`') ADVANCE(2231);
      END_STATE();
    case 2232:
      ACCEPT_TOKEN(anon_sym_LT_BANG_DASH_DASH);
      END_STATE();
    case 2233:
      ACCEPT_TOKEN(anon_sym_LT_BANG_DASH_DASH);
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 2234:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_GT);
      END_STATE();
    case 2235:
      ACCEPT_TOKEN(anon_sym_LT_QMARK);
      END_STATE();
    case 2236:
      ACCEPT_TOKEN(anon_sym_LT_QMARK);
      if (lookahead == '@') ADVANCE(2177);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          ('?' <= lookahead && lookahead <= 'Z') ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 2237:
      ACCEPT_TOKEN(anon_sym_QMARK_GT);
      END_STATE();
    case 2238:
      ACCEPT_TOKEN(aux_sym__declaration_token1);
      if (lookahead == '@') ADVANCE(2177);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(2238);
      if (lookahead == '!' ||
          ('#' <= lookahead && lookahead <= '\'') ||
          lookahead == '*' ||
          lookahead == '+' ||
          ('-' <= lookahead && lookahead <= '9') ||
          lookahead == '=' ||
          lookahead == '?' ||
          ('^' <= lookahead && lookahead <= '~')) ADVANCE(390);
      END_STATE();
    case 2239:
      ACCEPT_TOKEN(aux_sym__declaration_token1);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(2239);
      END_STATE();
    case 2240:
      ACCEPT_TOKEN(anon_sym_LT_BANG_LBRACKCDATA_LBRACK);
      END_STATE();
    case 2241:
      ACCEPT_TOKEN(anon_sym_RBRACK_RBRACK_GT);
      END_STATE();
    case 2242:
      ACCEPT_TOKEN(sym__whitespace_ge_2);
      END_STATE();
    case 2243:
      ACCEPT_TOKEN(sym__whitespace_ge_2);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(2243);
      END_STATE();
    case 2244:
      ACCEPT_TOKEN(aux_sym__whitespace_token1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(2243);
      END_STATE();
    case 2245:
      ACCEPT_TOKEN(sym__word_no_digit);
      if (lookahead == '_') ADVANCE(528);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          (lookahead < ' ' || '@' < lookahead) &&
          (lookahead < '[' || '`' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(2245);
      END_STATE();
    case 2246:
      ACCEPT_TOKEN(sym__digits);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(2246);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 2182, .external_lex_state = 2},
  [2] = {.lex_state = 2182, .external_lex_state = 3},
  [3] = {.lex_state = 2182, .external_lex_state = 3},
  [4] = {.lex_state = 2182, .external_lex_state = 4},
  [5] = {.lex_state = 2182, .external_lex_state = 4},
  [6] = {.lex_state = 2182, .external_lex_state = 3},
  [7] = {.lex_state = 2182, .external_lex_state = 5},
  [8] = {.lex_state = 2182, .external_lex_state = 5},
  [9] = {.lex_state = 2182, .external_lex_state = 6},
  [10] = {.lex_state = 2182, .external_lex_state = 6},
  [11] = {.lex_state = 2182, .external_lex_state = 7},
  [12] = {.lex_state = 2182, .external_lex_state = 3},
  [13] = {.lex_state = 2182, .external_lex_state = 8},
  [14] = {.lex_state = 2182, .external_lex_state = 8},
  [15] = {.lex_state = 2182, .external_lex_state = 3},
  [16] = {.lex_state = 2182, .external_lex_state = 3},
  [17] = {.lex_state = 2182, .external_lex_state = 9},
  [18] = {.lex_state = 2182, .external_lex_state = 8},
  [19] = {.lex_state = 2182, .external_lex_state = 7},
  [20] = {.lex_state = 2182, .external_lex_state = 7},
  [21] = {.lex_state = 2182, .external_lex_state = 10},
  [22] = {.lex_state = 2182, .external_lex_state = 10},
  [23] = {.lex_state = 2182, .external_lex_state = 9},
  [24] = {.lex_state = 2182, .external_lex_state = 9},
  [25] = {.lex_state = 2182, .external_lex_state = 3},
  [26] = {.lex_state = 2182, .external_lex_state = 3},
  [27] = {.lex_state = 2182, .external_lex_state = 3},
  [28] = {.lex_state = 2182, .external_lex_state = 7},
  [29] = {.lex_state = 2182, .external_lex_state = 7},
  [30] = {.lex_state = 2182, .external_lex_state = 10},
  [31] = {.lex_state = 2182, .external_lex_state = 9},
  [32] = {.lex_state = 2182, .external_lex_state = 7},
  [33] = {.lex_state = 2182, .external_lex_state = 10},
  [34] = {.lex_state = 2182, .external_lex_state = 9},
  [35] = {.lex_state = 2182, .external_lex_state = 10},
  [36] = {.lex_state = 2182, .external_lex_state = 9},
  [37] = {.lex_state = 2182, .external_lex_state = 7},
  [38] = {.lex_state = 2182, .external_lex_state = 10},
  [39] = {.lex_state = 2182, .external_lex_state = 9},
  [40] = {.lex_state = 2182, .external_lex_state = 7},
  [41] = {.lex_state = 2182, .external_lex_state = 10},
  [42] = {.lex_state = 2182, .external_lex_state = 9},
  [43] = {.lex_state = 2182, .external_lex_state = 7},
  [44] = {.lex_state = 2182, .external_lex_state = 10},
  [45] = {.lex_state = 2182, .external_lex_state = 9},
  [46] = {.lex_state = 2182, .external_lex_state = 7},
  [47] = {.lex_state = 2182, .external_lex_state = 10},
  [48] = {.lex_state = 2182, .external_lex_state = 9},
  [49] = {.lex_state = 2182, .external_lex_state = 7},
  [50] = {.lex_state = 2182, .external_lex_state = 10},
  [51] = {.lex_state = 2182, .external_lex_state = 9},
  [52] = {.lex_state = 2182, .external_lex_state = 10},
  [53] = {.lex_state = 2182, .external_lex_state = 8},
  [54] = {.lex_state = 2182, .external_lex_state = 8},
  [55] = {.lex_state = 2182, .external_lex_state = 8},
  [56] = {.lex_state = 2182, .external_lex_state = 8},
  [57] = {.lex_state = 2182, .external_lex_state = 8},
  [58] = {.lex_state = 2182, .external_lex_state = 8},
  [59] = {.lex_state = 2182, .external_lex_state = 8},
  [60] = {.lex_state = 2182, .external_lex_state = 8},
  [61] = {.lex_state = 2182, .external_lex_state = 8},
  [62] = {.lex_state = 2182, .external_lex_state = 8},
  [63] = {.lex_state = 2182, .external_lex_state = 8},
  [64] = {.lex_state = 2182, .external_lex_state = 8},
  [65] = {.lex_state = 2182, .external_lex_state = 8},
  [66] = {.lex_state = 2182, .external_lex_state = 8},
  [67] = {.lex_state = 2182, .external_lex_state = 8},
  [68] = {.lex_state = 2182, .external_lex_state = 8},
  [69] = {.lex_state = 2182, .external_lex_state = 8},
  [70] = {.lex_state = 2182, .external_lex_state = 8},
  [71] = {.lex_state = 2182, .external_lex_state = 8},
  [72] = {.lex_state = 2182, .external_lex_state = 8},
  [73] = {.lex_state = 2182, .external_lex_state = 8},
  [74] = {.lex_state = 2182, .external_lex_state = 8},
  [75] = {.lex_state = 2182, .external_lex_state = 8},
  [76] = {.lex_state = 2182, .external_lex_state = 8},
  [77] = {.lex_state = 2182, .external_lex_state = 8},
  [78] = {.lex_state = 2182, .external_lex_state = 8},
  [79] = {.lex_state = 2182, .external_lex_state = 8},
  [80] = {.lex_state = 2182, .external_lex_state = 8},
  [81] = {.lex_state = 2182, .external_lex_state = 8},
  [82] = {.lex_state = 2182, .external_lex_state = 8},
  [83] = {.lex_state = 2182, .external_lex_state = 8},
  [84] = {.lex_state = 2182, .external_lex_state = 8},
  [85] = {.lex_state = 2182, .external_lex_state = 8},
  [86] = {.lex_state = 2182, .external_lex_state = 4},
  [87] = {.lex_state = 2182, .external_lex_state = 6},
  [88] = {.lex_state = 2182, .external_lex_state = 3},
  [89] = {.lex_state = 2182, .external_lex_state = 5},
  [90] = {.lex_state = 2182, .external_lex_state = 3},
  [91] = {.lex_state = 2182, .external_lex_state = 3},
  [92] = {.lex_state = 2182, .external_lex_state = 6},
  [93] = {.lex_state = 2182, .external_lex_state = 5},
  [94] = {.lex_state = 2182, .external_lex_state = 5},
  [95] = {.lex_state = 2182, .external_lex_state = 6},
  [96] = {.lex_state = 2182, .external_lex_state = 4},
  [97] = {.lex_state = 2182, .external_lex_state = 4},
  [98] = {.lex_state = 2182, .external_lex_state = 3},
  [99] = {.lex_state = 2182, .external_lex_state = 3},
  [100] = {.lex_state = 2182, .external_lex_state = 3},
  [101] = {.lex_state = 2182, .external_lex_state = 9},
  [102] = {.lex_state = 2182, .external_lex_state = 8},
  [103] = {.lex_state = 2182, .external_lex_state = 7},
  [104] = {.lex_state = 2182, .external_lex_state = 3},
  [105] = {.lex_state = 2182, .external_lex_state = 10},
  [106] = {.lex_state = 2182, .external_lex_state = 10},
  [107] = {.lex_state = 2182, .external_lex_state = 9},
  [108] = {.lex_state = 2182, .external_lex_state = 9},
  [109] = {.lex_state = 2182, .external_lex_state = 7},
  [110] = {.lex_state = 2182, .external_lex_state = 8},
  [111] = {.lex_state = 2182, .external_lex_state = 7},
  [112] = {.lex_state = 2182, .external_lex_state = 10},
  [113] = {.lex_state = 2182, .external_lex_state = 9},
  [114] = {.lex_state = 2182, .external_lex_state = 7},
  [115] = {.lex_state = 2182, .external_lex_state = 10},
  [116] = {.lex_state = 2182, .external_lex_state = 9},
  [117] = {.lex_state = 2182, .external_lex_state = 10},
  [118] = {.lex_state = 2182, .external_lex_state = 8},
  [119] = {.lex_state = 2182, .external_lex_state = 7},
  [120] = {.lex_state = 2182, .external_lex_state = 10},
  [121] = {.lex_state = 2182, .external_lex_state = 9},
  [122] = {.lex_state = 2182, .external_lex_state = 7},
  [123] = {.lex_state = 2182, .external_lex_state = 10},
  [124] = {.lex_state = 2182, .external_lex_state = 9},
  [125] = {.lex_state = 2182, .external_lex_state = 8},
  [126] = {.lex_state = 2182, .external_lex_state = 7},
  [127] = {.lex_state = 2182, .external_lex_state = 10},
  [128] = {.lex_state = 2182, .external_lex_state = 9},
  [129] = {.lex_state = 2182, .external_lex_state = 7},
  [130] = {.lex_state = 2182, .external_lex_state = 10},
  [131] = {.lex_state = 2182, .external_lex_state = 9},
  [132] = {.lex_state = 2182, .external_lex_state = 8},
  [133] = {.lex_state = 2182, .external_lex_state = 3},
  [134] = {.lex_state = 2182, .external_lex_state = 8},
  [135] = {.lex_state = 2182, .external_lex_state = 7},
  [136] = {.lex_state = 2182, .external_lex_state = 8},
  [137] = {.lex_state = 2182, .external_lex_state = 8},
  [138] = {.lex_state = 2182, .external_lex_state = 8},
  [139] = {.lex_state = 2182, .external_lex_state = 8},
  [140] = {.lex_state = 2182, .external_lex_state = 8},
  [141] = {.lex_state = 2182, .external_lex_state = 8},
  [142] = {.lex_state = 2182, .external_lex_state = 8},
  [143] = {.lex_state = 2182, .external_lex_state = 8},
  [144] = {.lex_state = 2182, .external_lex_state = 8},
  [145] = {.lex_state = 2182, .external_lex_state = 8},
  [146] = {.lex_state = 2182, .external_lex_state = 8},
  [147] = {.lex_state = 2182, .external_lex_state = 8},
  [148] = {.lex_state = 2182, .external_lex_state = 7},
  [149] = {.lex_state = 2182, .external_lex_state = 7},
  [150] = {.lex_state = 2182, .external_lex_state = 10},
  [151] = {.lex_state = 2182, .external_lex_state = 10},
  [152] = {.lex_state = 2182, .external_lex_state = 8},
  [153] = {.lex_state = 2182, .external_lex_state = 9},
  [154] = {.lex_state = 2182, .external_lex_state = 9},
  [155] = {.lex_state = 2182, .external_lex_state = 8},
  [156] = {.lex_state = 2182, .external_lex_state = 8},
  [157] = {.lex_state = 2182, .external_lex_state = 8},
  [158] = {.lex_state = 2182, .external_lex_state = 10},
  [159] = {.lex_state = 2182, .external_lex_state = 10},
  [160] = {.lex_state = 2182, .external_lex_state = 9},
  [161] = {.lex_state = 2182, .external_lex_state = 7},
  [162] = {.lex_state = 2182, .external_lex_state = 7},
  [163] = {.lex_state = 2182, .external_lex_state = 9},
  [164] = {.lex_state = 2182, .external_lex_state = 8},
  [165] = {.lex_state = 2182, .external_lex_state = 8},
  [166] = {.lex_state = 2180, .external_lex_state = 7},
  [167] = {.lex_state = 2180, .external_lex_state = 9},
  [168] = {.lex_state = 2181, .external_lex_state = 7},
  [169] = {.lex_state = 2180, .external_lex_state = 10},
  [170] = {.lex_state = 2181, .external_lex_state = 10},
  [171] = {.lex_state = 2180, .external_lex_state = 9},
  [172] = {.lex_state = 2181, .external_lex_state = 9},
  [173] = {.lex_state = 2180, .external_lex_state = 7},
  [174] = {.lex_state = 2180, .external_lex_state = 8},
  [175] = {.lex_state = 2181, .external_lex_state = 7},
  [176] = {.lex_state = 2181, .external_lex_state = 8},
  [177] = {.lex_state = 2181, .external_lex_state = 9},
  [178] = {.lex_state = 2180, .external_lex_state = 10},
  [179] = {.lex_state = 2181, .external_lex_state = 10},
  [180] = {.lex_state = 2179, .external_lex_state = 10},
  [181] = {.lex_state = 2181, .external_lex_state = 8},
  [182] = {.lex_state = 2181, .external_lex_state = 8},
  [183] = {.lex_state = 2179, .external_lex_state = 7},
  [184] = {.lex_state = 2179, .external_lex_state = 7},
  [185] = {.lex_state = 2179, .external_lex_state = 8},
  [186] = {.lex_state = 2179, .external_lex_state = 10},
  [187] = {.lex_state = 2179, .external_lex_state = 9},
  [188] = {.lex_state = 2180, .external_lex_state = 8},
  [189] = {.lex_state = 2179, .external_lex_state = 9},
  [190] = {.lex_state = 2180, .external_lex_state = 8},
  [191] = {.lex_state = 9, .external_lex_state = 11},
  [192] = {.lex_state = 2179, .external_lex_state = 8},
  [193] = {.lex_state = 9, .external_lex_state = 11},
  [194] = {.lex_state = 9, .external_lex_state = 11},
  [195] = {.lex_state = 9, .external_lex_state = 11},
  [196] = {.lex_state = 9, .external_lex_state = 11},
  [197] = {.lex_state = 9, .external_lex_state = 11},
  [198] = {.lex_state = 9, .external_lex_state = 11},
  [199] = {.lex_state = 2179, .external_lex_state = 8},
  [200] = {.lex_state = 9, .external_lex_state = 11},
  [201] = {.lex_state = 9, .external_lex_state = 11},
  [202] = {.lex_state = 9, .external_lex_state = 11},
  [203] = {.lex_state = 9, .external_lex_state = 11},
  [204] = {.lex_state = 2182, .external_lex_state = 6},
  [205] = {.lex_state = 2182, .external_lex_state = 7},
  [206] = {.lex_state = 2182, .external_lex_state = 12},
  [207] = {.lex_state = 2182, .external_lex_state = 8},
  [208] = {.lex_state = 2182, .external_lex_state = 4},
  [209] = {.lex_state = 2182, .external_lex_state = 3},
  [210] = {.lex_state = 2182, .external_lex_state = 5},
  [211] = {.lex_state = 2182, .external_lex_state = 2},
  [212] = {.lex_state = 2182, .external_lex_state = 6},
  [213] = {.lex_state = 2182, .external_lex_state = 10},
  [214] = {.lex_state = 2182, .external_lex_state = 13},
  [215] = {.lex_state = 2182, .external_lex_state = 3},
  [216] = {.lex_state = 2182, .external_lex_state = 9},
  [217] = {.lex_state = 2182, .external_lex_state = 5},
  [218] = {.lex_state = 2182, .external_lex_state = 4},
  [219] = {.lex_state = 2182, .external_lex_state = 14},
  [220] = {.lex_state = 11, .external_lex_state = 15},
  [221] = {.lex_state = 2182, .external_lex_state = 5},
  [222] = {.lex_state = 2182, .external_lex_state = 4},
  [223] = {.lex_state = 2182, .external_lex_state = 6},
  [224] = {.lex_state = 11, .external_lex_state = 15},
  [225] = {.lex_state = 2182, .external_lex_state = 12},
  [226] = {.lex_state = 11, .external_lex_state = 15},
  [227] = {.lex_state = 2182, .external_lex_state = 6},
  [228] = {.lex_state = 11, .external_lex_state = 15},
  [229] = {.lex_state = 2182, .external_lex_state = 2},
  [230] = {.lex_state = 11, .external_lex_state = 15},
  [231] = {.lex_state = 11, .external_lex_state = 15},
  [232] = {.lex_state = 2182, .external_lex_state = 3},
  [233] = {.lex_state = 2182, .external_lex_state = 6},
  [234] = {.lex_state = 2182, .external_lex_state = 3},
  [235] = {.lex_state = 2182, .external_lex_state = 3},
  [236] = {.lex_state = 2182, .external_lex_state = 5},
  [237] = {.lex_state = 2182, .external_lex_state = 12},
  [238] = {.lex_state = 2182, .external_lex_state = 4},
  [239] = {.lex_state = 2182, .external_lex_state = 4},
  [240] = {.lex_state = 11, .external_lex_state = 15},
  [241] = {.lex_state = 2182, .external_lex_state = 14},
  [242] = {.lex_state = 11, .external_lex_state = 15},
  [243] = {.lex_state = 2182, .external_lex_state = 8},
  [244] = {.lex_state = 2182, .external_lex_state = 6},
  [245] = {.lex_state = 2182, .external_lex_state = 2},
  [246] = {.lex_state = 11, .external_lex_state = 15},
  [247] = {.lex_state = 2182, .external_lex_state = 3},
  [248] = {.lex_state = 2182, .external_lex_state = 6},
  [249] = {.lex_state = 2182, .external_lex_state = 3},
  [250] = {.lex_state = 2182, .external_lex_state = 13},
  [251] = {.lex_state = 2182, .external_lex_state = 5},
  [252] = {.lex_state = 2182, .external_lex_state = 5},
  [253] = {.lex_state = 2182, .external_lex_state = 3},
  [254] = {.lex_state = 2182, .external_lex_state = 14},
  [255] = {.lex_state = 2182, .external_lex_state = 2},
  [256] = {.lex_state = 2182, .external_lex_state = 13},
  [257] = {.lex_state = 2182, .external_lex_state = 4},
  [258] = {.lex_state = 2182, .external_lex_state = 4},
  [259] = {.lex_state = 2182, .external_lex_state = 5},
  [260] = {.lex_state = 11, .external_lex_state = 15},
  [261] = {.lex_state = 11, .external_lex_state = 15},
  [262] = {.lex_state = 2182, .external_lex_state = 9},
  [263] = {.lex_state = 2182, .external_lex_state = 7},
  [264] = {.lex_state = 2182, .external_lex_state = 7},
  [265] = {.lex_state = 2182, .external_lex_state = 7},
  [266] = {.lex_state = 2182, .external_lex_state = 10},
  [267] = {.lex_state = 2182, .external_lex_state = 10},
  [268] = {.lex_state = 2182, .external_lex_state = 7},
  [269] = {.lex_state = 2182, .external_lex_state = 7},
  [270] = {.lex_state = 2182, .external_lex_state = 10},
  [271] = {.lex_state = 2182, .external_lex_state = 10},
  [272] = {.lex_state = 2182, .external_lex_state = 10},
  [273] = {.lex_state = 2182, .external_lex_state = 10},
  [274] = {.lex_state = 2182, .external_lex_state = 10},
  [275] = {.lex_state = 2182, .external_lex_state = 10},
  [276] = {.lex_state = 2182, .external_lex_state = 9},
  [277] = {.lex_state = 2182, .external_lex_state = 8},
  [278] = {.lex_state = 2182, .external_lex_state = 8},
  [279] = {.lex_state = 1, .external_lex_state = 16},
  [280] = {.lex_state = 2182, .external_lex_state = 8},
  [281] = {.lex_state = 2182, .external_lex_state = 8},
  [282] = {.lex_state = 2182, .external_lex_state = 8},
  [283] = {.lex_state = 2182, .external_lex_state = 8},
  [284] = {.lex_state = 2182, .external_lex_state = 8},
  [285] = {.lex_state = 2182, .external_lex_state = 8},
  [286] = {.lex_state = 2182, .external_lex_state = 8},
  [287] = {.lex_state = 2182, .external_lex_state = 8},
  [288] = {.lex_state = 2182, .external_lex_state = 9},
  [289] = {.lex_state = 2182, .external_lex_state = 9},
  [290] = {.lex_state = 2182, .external_lex_state = 9},
  [291] = {.lex_state = 2182, .external_lex_state = 9},
  [292] = {.lex_state = 2182, .external_lex_state = 10},
  [293] = {.lex_state = 2182, .external_lex_state = 8},
  [294] = {.lex_state = 2182, .external_lex_state = 9},
  [295] = {.lex_state = 2182, .external_lex_state = 9},
  [296] = {.lex_state = 2182, .external_lex_state = 10},
  [297] = {.lex_state = 2182, .external_lex_state = 9},
  [298] = {.lex_state = 2182, .external_lex_state = 10},
  [299] = {.lex_state = 2182, .external_lex_state = 10},
  [300] = {.lex_state = 2182, .external_lex_state = 10},
  [301] = {.lex_state = 2182, .external_lex_state = 10},
  [302] = {.lex_state = 2182, .external_lex_state = 10},
  [303] = {.lex_state = 2182, .external_lex_state = 9},
  [304] = {.lex_state = 2182, .external_lex_state = 9},
  [305] = {.lex_state = 2182, .external_lex_state = 8},
  [306] = {.lex_state = 2182, .external_lex_state = 9},
  [307] = {.lex_state = 2182, .external_lex_state = 9},
  [308] = {.lex_state = 2182, .external_lex_state = 9},
  [309] = {.lex_state = 2182, .external_lex_state = 9},
  [310] = {.lex_state = 2182, .external_lex_state = 9},
  [311] = {.lex_state = 2182, .external_lex_state = 9},
  [312] = {.lex_state = 2182, .external_lex_state = 9},
  [313] = {.lex_state = 2182, .external_lex_state = 9},
  [314] = {.lex_state = 2182, .external_lex_state = 10},
  [315] = {.lex_state = 2182, .external_lex_state = 10},
  [316] = {.lex_state = 2182, .external_lex_state = 10},
  [317] = {.lex_state = 2182, .external_lex_state = 10},
  [318] = {.lex_state = 2182, .external_lex_state = 10},
  [319] = {.lex_state = 2182, .external_lex_state = 9},
  [320] = {.lex_state = 2182, .external_lex_state = 9},
  [321] = {.lex_state = 2182, .external_lex_state = 9},
  [322] = {.lex_state = 2182, .external_lex_state = 9},
  [323] = {.lex_state = 2182, .external_lex_state = 9},
  [324] = {.lex_state = 2182, .external_lex_state = 9},
  [325] = {.lex_state = 2182, .external_lex_state = 9},
  [326] = {.lex_state = 2182, .external_lex_state = 10},
  [327] = {.lex_state = 2182, .external_lex_state = 10},
  [328] = {.lex_state = 2182, .external_lex_state = 8},
  [329] = {.lex_state = 2182, .external_lex_state = 10},
  [330] = {.lex_state = 2182, .external_lex_state = 10},
  [331] = {.lex_state = 2182, .external_lex_state = 9},
  [332] = {.lex_state = 2182, .external_lex_state = 9},
  [333] = {.lex_state = 2182, .external_lex_state = 9},
  [334] = {.lex_state = 2182, .external_lex_state = 10},
  [335] = {.lex_state = 2182, .external_lex_state = 10},
  [336] = {.lex_state = 2182, .external_lex_state = 10},
  [337] = {.lex_state = 2182, .external_lex_state = 9},
  [338] = {.lex_state = 2182, .external_lex_state = 8},
  [339] = {.lex_state = 2182, .external_lex_state = 10},
  [340] = {.lex_state = 2182, .external_lex_state = 10},
  [341] = {.lex_state = 2182, .external_lex_state = 9},
  [342] = {.lex_state = 2182, .external_lex_state = 9},
  [343] = {.lex_state = 2182, .external_lex_state = 9},
  [344] = {.lex_state = 2182, .external_lex_state = 9},
  [345] = {.lex_state = 2182, .external_lex_state = 9},
  [346] = {.lex_state = 2182, .external_lex_state = 9},
  [347] = {.lex_state = 2182, .external_lex_state = 8},
  [348] = {.lex_state = 2182, .external_lex_state = 8},
  [349] = {.lex_state = 2182, .external_lex_state = 3},
  [350] = {.lex_state = 2182, .external_lex_state = 8},
  [351] = {.lex_state = 2182, .external_lex_state = 8},
  [352] = {.lex_state = 2182, .external_lex_state = 3},
  [353] = {.lex_state = 2182, .external_lex_state = 8},
  [354] = {.lex_state = 2182, .external_lex_state = 7},
  [355] = {.lex_state = 2182, .external_lex_state = 8},
  [356] = {.lex_state = 2182, .external_lex_state = 3},
  [357] = {.lex_state = 2182, .external_lex_state = 8},
  [358] = {.lex_state = 2182, .external_lex_state = 8},
  [359] = {.lex_state = 2182, .external_lex_state = 9},
  [360] = {.lex_state = 2182, .external_lex_state = 8},
  [361] = {.lex_state = 2182, .external_lex_state = 9},
  [362] = {.lex_state = 2182, .external_lex_state = 9},
  [363] = {.lex_state = 2182, .external_lex_state = 9},
  [364] = {.lex_state = 2182, .external_lex_state = 9},
  [365] = {.lex_state = 2182, .external_lex_state = 9},
  [366] = {.lex_state = 2182, .external_lex_state = 9},
  [367] = {.lex_state = 2182, .external_lex_state = 7},
  [368] = {.lex_state = 2182, .external_lex_state = 9},
  [369] = {.lex_state = 2182, .external_lex_state = 9},
  [370] = {.lex_state = 2182, .external_lex_state = 9},
  [371] = {.lex_state = 2182, .external_lex_state = 9},
  [372] = {.lex_state = 2182, .external_lex_state = 9},
  [373] = {.lex_state = 2182, .external_lex_state = 9},
  [374] = {.lex_state = 2182, .external_lex_state = 9},
  [375] = {.lex_state = 2182, .external_lex_state = 3},
  [376] = {.lex_state = 2182, .external_lex_state = 9},
  [377] = {.lex_state = 2182, .external_lex_state = 9},
  [378] = {.lex_state = 2182, .external_lex_state = 9},
  [379] = {.lex_state = 2182, .external_lex_state = 9},
  [380] = {.lex_state = 2182, .external_lex_state = 9},
  [381] = {.lex_state = 2182, .external_lex_state = 9},
  [382] = {.lex_state = 2182, .external_lex_state = 9},
  [383] = {.lex_state = 2182, .external_lex_state = 4},
  [384] = {.lex_state = 2182, .external_lex_state = 7},
  [385] = {.lex_state = 2182, .external_lex_state = 7},
  [386] = {.lex_state = 2182, .external_lex_state = 7},
  [387] = {.lex_state = 2182, .external_lex_state = 8},
  [388] = {.lex_state = 2182, .external_lex_state = 3},
  [389] = {.lex_state = 2182, .external_lex_state = 5},
  [390] = {.lex_state = 2182, .external_lex_state = 2},
  [391] = {.lex_state = 2182, .external_lex_state = 7},
  [392] = {.lex_state = 2182, .external_lex_state = 7},
  [393] = {.lex_state = 2182, .external_lex_state = 7},
  [394] = {.lex_state = 2182, .external_lex_state = 7},
  [395] = {.lex_state = 2182, .external_lex_state = 2},
  [396] = {.lex_state = 2182, .external_lex_state = 7},
  [397] = {.lex_state = 1, .external_lex_state = 16},
  [398] = {.lex_state = 2182, .external_lex_state = 8},
  [399] = {.lex_state = 2182, .external_lex_state = 6},
  [400] = {.lex_state = 2182, .external_lex_state = 7},
  [401] = {.lex_state = 2182, .external_lex_state = 7},
  [402] = {.lex_state = 2182, .external_lex_state = 7},
  [403] = {.lex_state = 2182, .external_lex_state = 7},
  [404] = {.lex_state = 2182, .external_lex_state = 7},
  [405] = {.lex_state = 2182, .external_lex_state = 7},
  [406] = {.lex_state = 2182, .external_lex_state = 3},
  [407] = {.lex_state = 2182, .external_lex_state = 7},
  [408] = {.lex_state = 2182, .external_lex_state = 7},
  [409] = {.lex_state = 2182, .external_lex_state = 7},
  [410] = {.lex_state = 2182, .external_lex_state = 7},
  [411] = {.lex_state = 2182, .external_lex_state = 7},
  [412] = {.lex_state = 2182, .external_lex_state = 7},
  [413] = {.lex_state = 2182, .external_lex_state = 7},
  [414] = {.lex_state = 2182, .external_lex_state = 10},
  [415] = {.lex_state = 2182, .external_lex_state = 7},
  [416] = {.lex_state = 2182, .external_lex_state = 8},
  [417] = {.lex_state = 2182, .external_lex_state = 7},
  [418] = {.lex_state = 2182, .external_lex_state = 7},
  [419] = {.lex_state = 2182, .external_lex_state = 7},
  [420] = {.lex_state = 2182, .external_lex_state = 7},
  [421] = {.lex_state = 2182, .external_lex_state = 7},
  [422] = {.lex_state = 2182, .external_lex_state = 7},
  [423] = {.lex_state = 2182, .external_lex_state = 7},
  [424] = {.lex_state = 2182, .external_lex_state = 10},
  [425] = {.lex_state = 2182, .external_lex_state = 7},
  [426] = {.lex_state = 2182, .external_lex_state = 7},
  [427] = {.lex_state = 2182, .external_lex_state = 7},
  [428] = {.lex_state = 1, .external_lex_state = 16},
  [429] = {.lex_state = 1, .external_lex_state = 16},
  [430] = {.lex_state = 2182, .external_lex_state = 7},
  [431] = {.lex_state = 2182, .external_lex_state = 7},
  [432] = {.lex_state = 2182, .external_lex_state = 7},
  [433] = {.lex_state = 2182, .external_lex_state = 7},
  [434] = {.lex_state = 2182, .external_lex_state = 10},
  [435] = {.lex_state = 2182, .external_lex_state = 8},
  [436] = {.lex_state = 2182, .external_lex_state = 8},
  [437] = {.lex_state = 1, .external_lex_state = 16},
  [438] = {.lex_state = 2182, .external_lex_state = 8},
  [439] = {.lex_state = 2182, .external_lex_state = 8},
  [440] = {.lex_state = 2182, .external_lex_state = 8},
  [441] = {.lex_state = 2182, .external_lex_state = 8},
  [442] = {.lex_state = 2182, .external_lex_state = 8},
  [443] = {.lex_state = 2182, .external_lex_state = 9},
  [444] = {.lex_state = 2182, .external_lex_state = 8},
  [445] = {.lex_state = 2182, .external_lex_state = 7},
  [446] = {.lex_state = 2182, .external_lex_state = 8},
  [447] = {.lex_state = 2182, .external_lex_state = 10},
  [448] = {.lex_state = 2182, .external_lex_state = 8},
  [449] = {.lex_state = 2182, .external_lex_state = 10},
  [450] = {.lex_state = 2182, .external_lex_state = 10},
  [451] = {.lex_state = 1, .external_lex_state = 16},
  [452] = {.lex_state = 1, .external_lex_state = 16},
  [453] = {.lex_state = 2182, .external_lex_state = 10},
  [454] = {.lex_state = 2182, .external_lex_state = 10},
  [455] = {.lex_state = 2182, .external_lex_state = 10},
  [456] = {.lex_state = 2182, .external_lex_state = 10},
  [457] = {.lex_state = 2182, .external_lex_state = 7},
  [458] = {.lex_state = 2182, .external_lex_state = 8},
  [459] = {.lex_state = 2182, .external_lex_state = 10},
  [460] = {.lex_state = 1, .external_lex_state = 16},
  [461] = {.lex_state = 2182, .external_lex_state = 10},
  [462] = {.lex_state = 2182, .external_lex_state = 7},
  [463] = {.lex_state = 2182, .external_lex_state = 10},
  [464] = {.lex_state = 2182, .external_lex_state = 7},
  [465] = {.lex_state = 2182, .external_lex_state = 7},
  [466] = {.lex_state = 2182, .external_lex_state = 7},
  [467] = {.lex_state = 2182, .external_lex_state = 7},
  [468] = {.lex_state = 2182, .external_lex_state = 7},
  [469] = {.lex_state = 2182, .external_lex_state = 8},
  [470] = {.lex_state = 2182, .external_lex_state = 10},
  [471] = {.lex_state = 2182, .external_lex_state = 8},
  [472] = {.lex_state = 2182, .external_lex_state = 7},
  [473] = {.lex_state = 2182, .external_lex_state = 10},
  [474] = {.lex_state = 1, .external_lex_state = 16},
  [475] = {.lex_state = 1, .external_lex_state = 16},
  [476] = {.lex_state = 2182, .external_lex_state = 10},
  [477] = {.lex_state = 2182, .external_lex_state = 8},
  [478] = {.lex_state = 2182, .external_lex_state = 8},
  [479] = {.lex_state = 2182, .external_lex_state = 10},
  [480] = {.lex_state = 2182, .external_lex_state = 10},
  [481] = {.lex_state = 2182, .external_lex_state = 10},
  [482] = {.lex_state = 2182, .external_lex_state = 9},
  [483] = {.lex_state = 1, .external_lex_state = 16},
  [484] = {.lex_state = 2182, .external_lex_state = 10},
  [485] = {.lex_state = 2182, .external_lex_state = 10},
  [486] = {.lex_state = 2182, .external_lex_state = 10},
  [487] = {.lex_state = 2182, .external_lex_state = 10},
  [488] = {.lex_state = 2182, .external_lex_state = 10},
  [489] = {.lex_state = 2182, .external_lex_state = 7},
  [490] = {.lex_state = 2182, .external_lex_state = 8},
  [491] = {.lex_state = 2182, .external_lex_state = 7},
  [492] = {.lex_state = 2182, .external_lex_state = 7},
  [493] = {.lex_state = 2182, .external_lex_state = 10},
  [494] = {.lex_state = 2182, .external_lex_state = 7},
  [495] = {.lex_state = 2182, .external_lex_state = 7},
  [496] = {.lex_state = 2182, .external_lex_state = 7},
  [497] = {.lex_state = 1, .external_lex_state = 16},
  [498] = {.lex_state = 1, .external_lex_state = 16},
  [499] = {.lex_state = 2182, .external_lex_state = 10},
  [500] = {.lex_state = 2182, .external_lex_state = 8},
  [501] = {.lex_state = 2182, .external_lex_state = 10},
  [502] = {.lex_state = 2182, .external_lex_state = 10},
  [503] = {.lex_state = 1, .external_lex_state = 16},
  [504] = {.lex_state = 2182, .external_lex_state = 8},
  [505] = {.lex_state = 2182, .external_lex_state = 8},
  [506] = {.lex_state = 2182, .external_lex_state = 8},
  [507] = {.lex_state = 2182, .external_lex_state = 8},
  [508] = {.lex_state = 2182, .external_lex_state = 8},
  [509] = {.lex_state = 2182, .external_lex_state = 8},
  [510] = {.lex_state = 2182, .external_lex_state = 8},
  [511] = {.lex_state = 2182, .external_lex_state = 10},
  [512] = {.lex_state = 2182, .external_lex_state = 10},
  [513] = {.lex_state = 2182, .external_lex_state = 10},
  [514] = {.lex_state = 2182, .external_lex_state = 8},
  [515] = {.lex_state = 1, .external_lex_state = 16},
  [516] = {.lex_state = 2182, .external_lex_state = 8},
  [517] = {.lex_state = 1, .external_lex_state = 16},
  [518] = {.lex_state = 2182, .external_lex_state = 10},
  [519] = {.lex_state = 2182, .external_lex_state = 7},
  [520] = {.lex_state = 2182, .external_lex_state = 7},
  [521] = {.lex_state = 2182, .external_lex_state = 8},
  [522] = {.lex_state = 2182, .external_lex_state = 8},
  [523] = {.lex_state = 2182, .external_lex_state = 8},
  [524] = {.lex_state = 2182, .external_lex_state = 8},
  [525] = {.lex_state = 2182, .external_lex_state = 7},
  [526] = {.lex_state = 2182, .external_lex_state = 7},
  [527] = {.lex_state = 2182, .external_lex_state = 10},
  [528] = {.lex_state = 2182, .external_lex_state = 10},
  [529] = {.lex_state = 2182, .external_lex_state = 8},
  [530] = {.lex_state = 2182, .external_lex_state = 7},
  [531] = {.lex_state = 2182, .external_lex_state = 8},
  [532] = {.lex_state = 2182, .external_lex_state = 10},
  [533] = {.lex_state = 2182, .external_lex_state = 8},
  [534] = {.lex_state = 2182, .external_lex_state = 8},
  [535] = {.lex_state = 2182, .external_lex_state = 7},
  [536] = {.lex_state = 2182, .external_lex_state = 8},
  [537] = {.lex_state = 2182, .external_lex_state = 8},
  [538] = {.lex_state = 2182, .external_lex_state = 8},
  [539] = {.lex_state = 2182, .external_lex_state = 8},
  [540] = {.lex_state = 2182, .external_lex_state = 8},
  [541] = {.lex_state = 2182, .external_lex_state = 8},
  [542] = {.lex_state = 2182, .external_lex_state = 8},
  [543] = {.lex_state = 2182, .external_lex_state = 8},
  [544] = {.lex_state = 2182, .external_lex_state = 8},
  [545] = {.lex_state = 2182, .external_lex_state = 8},
  [546] = {.lex_state = 2182, .external_lex_state = 8},
  [547] = {.lex_state = 2182, .external_lex_state = 8},
  [548] = {.lex_state = 2182, .external_lex_state = 8},
  [549] = {.lex_state = 2182, .external_lex_state = 8},
  [550] = {.lex_state = 2182, .external_lex_state = 8},
  [551] = {.lex_state = 2182, .external_lex_state = 8},
  [552] = {.lex_state = 2182, .external_lex_state = 8},
  [553] = {.lex_state = 2182, .external_lex_state = 8},
  [554] = {.lex_state = 2182, .external_lex_state = 8},
  [555] = {.lex_state = 2182, .external_lex_state = 8},
  [556] = {.lex_state = 2182, .external_lex_state = 8},
  [557] = {.lex_state = 2182, .external_lex_state = 8},
  [558] = {.lex_state = 2182, .external_lex_state = 8},
  [559] = {.lex_state = 2182, .external_lex_state = 8},
  [560] = {.lex_state = 2182, .external_lex_state = 8},
  [561] = {.lex_state = 2182, .external_lex_state = 8},
  [562] = {.lex_state = 2182, .external_lex_state = 8},
  [563] = {.lex_state = 2182, .external_lex_state = 8},
  [564] = {.lex_state = 2182, .external_lex_state = 8},
  [565] = {.lex_state = 2182, .external_lex_state = 8},
  [566] = {.lex_state = 2182, .external_lex_state = 8},
  [567] = {.lex_state = 2182, .external_lex_state = 8},
  [568] = {.lex_state = 2182, .external_lex_state = 8},
  [569] = {.lex_state = 2182, .external_lex_state = 8},
  [570] = {.lex_state = 2182, .external_lex_state = 8},
  [571] = {.lex_state = 2182, .external_lex_state = 8},
  [572] = {.lex_state = 2182, .external_lex_state = 8},
  [573] = {.lex_state = 2182, .external_lex_state = 8},
  [574] = {.lex_state = 2182, .external_lex_state = 8},
  [575] = {.lex_state = 2182, .external_lex_state = 7},
  [576] = {.lex_state = 2182, .external_lex_state = 10},
  [577] = {.lex_state = 2182, .external_lex_state = 8},
  [578] = {.lex_state = 2182, .external_lex_state = 8},
  [579] = {.lex_state = 2182, .external_lex_state = 8},
  [580] = {.lex_state = 2182, .external_lex_state = 8},
  [581] = {.lex_state = 2182, .external_lex_state = 8},
  [582] = {.lex_state = 2182, .external_lex_state = 8},
  [583] = {.lex_state = 2182, .external_lex_state = 8},
  [584] = {.lex_state = 2182, .external_lex_state = 8},
  [585] = {.lex_state = 2182, .external_lex_state = 8},
  [586] = {.lex_state = 2182, .external_lex_state = 8},
  [587] = {.lex_state = 2182, .external_lex_state = 8},
  [588] = {.lex_state = 2182, .external_lex_state = 8},
  [589] = {.lex_state = 2182, .external_lex_state = 8},
  [590] = {.lex_state = 2182, .external_lex_state = 8},
  [591] = {.lex_state = 2182, .external_lex_state = 8},
  [592] = {.lex_state = 2182, .external_lex_state = 8},
  [593] = {.lex_state = 2182, .external_lex_state = 8},
  [594] = {.lex_state = 2182, .external_lex_state = 8},
  [595] = {.lex_state = 2182, .external_lex_state = 8},
  [596] = {.lex_state = 2182, .external_lex_state = 8},
  [597] = {.lex_state = 2182, .external_lex_state = 8},
  [598] = {.lex_state = 2182, .external_lex_state = 9},
  [599] = {.lex_state = 2182, .external_lex_state = 8},
  [600] = {.lex_state = 2182, .external_lex_state = 7},
  [601] = {.lex_state = 2182, .external_lex_state = 9},
  [602] = {.lex_state = 2182, .external_lex_state = 9},
  [603] = {.lex_state = 2182, .external_lex_state = 9},
  [604] = {.lex_state = 2182, .external_lex_state = 8},
  [605] = {.lex_state = 2182, .external_lex_state = 9},
  [606] = {.lex_state = 2182, .external_lex_state = 8},
  [607] = {.lex_state = 2182, .external_lex_state = 7},
  [608] = {.lex_state = 2182, .external_lex_state = 7},
  [609] = {.lex_state = 2182, .external_lex_state = 7},
  [610] = {.lex_state = 2182, .external_lex_state = 7},
  [611] = {.lex_state = 2182, .external_lex_state = 7},
  [612] = {.lex_state = 2182, .external_lex_state = 8},
  [613] = {.lex_state = 2182, .external_lex_state = 8},
  [614] = {.lex_state = 2182, .external_lex_state = 8},
  [615] = {.lex_state = 2182, .external_lex_state = 9},
  [616] = {.lex_state = 2182, .external_lex_state = 9},
  [617] = {.lex_state = 2182, .external_lex_state = 9},
  [618] = {.lex_state = 2182, .external_lex_state = 9},
  [619] = {.lex_state = 2182, .external_lex_state = 9},
  [620] = {.lex_state = 2182, .external_lex_state = 10},
  [621] = {.lex_state = 2182, .external_lex_state = 7},
  [622] = {.lex_state = 2182, .external_lex_state = 10},
  [623] = {.lex_state = 2182, .external_lex_state = 7},
  [624] = {.lex_state = 2182, .external_lex_state = 7},
  [625] = {.lex_state = 2182, .external_lex_state = 7},
  [626] = {.lex_state = 2182, .external_lex_state = 10},
  [627] = {.lex_state = 2182, .external_lex_state = 7},
  [628] = {.lex_state = 2182, .external_lex_state = 10},
  [629] = {.lex_state = 2182, .external_lex_state = 7},
  [630] = {.lex_state = 2182, .external_lex_state = 10},
  [631] = {.lex_state = 2182, .external_lex_state = 8},
  [632] = {.lex_state = 2182, .external_lex_state = 8},
  [633] = {.lex_state = 2182, .external_lex_state = 8},
  [634] = {.lex_state = 2182, .external_lex_state = 10},
  [635] = {.lex_state = 2182, .external_lex_state = 10},
  [636] = {.lex_state = 2182, .external_lex_state = 10},
  [637] = {.lex_state = 2182, .external_lex_state = 10},
  [638] = {.lex_state = 2182, .external_lex_state = 10},
  [639] = {.lex_state = 2182, .external_lex_state = 7},
  [640] = {.lex_state = 2182, .external_lex_state = 10},
  [641] = {.lex_state = 2182, .external_lex_state = 10},
  [642] = {.lex_state = 2182, .external_lex_state = 8},
  [643] = {.lex_state = 2182, .external_lex_state = 10},
  [644] = {.lex_state = 2182, .external_lex_state = 8},
  [645] = {.lex_state = 2},
  [646] = {.lex_state = 2},
  [647] = {.lex_state = 2},
  [648] = {.lex_state = 2},
  [649] = {.lex_state = 8, .external_lex_state = 11},
  [650] = {.lex_state = 2},
  [651] = {.lex_state = 2},
  [652] = {.lex_state = 2},
  [653] = {.lex_state = 2},
  [654] = {.lex_state = 2},
  [655] = {.lex_state = 2},
  [656] = {.lex_state = 2},
  [657] = {.lex_state = 2},
  [658] = {.lex_state = 2, .external_lex_state = 17},
  [659] = {.lex_state = 2},
  [660] = {.lex_state = 2},
  [661] = {.lex_state = 6, .external_lex_state = 11},
  [662] = {.lex_state = 2},
  [663] = {.lex_state = 2},
  [664] = {.lex_state = 2},
  [665] = {.lex_state = 2},
  [666] = {.lex_state = 2},
  [667] = {.lex_state = 2},
  [668] = {.lex_state = 2, .external_lex_state = 17},
  [669] = {.lex_state = 10, .external_lex_state = 15},
  [670] = {.lex_state = 1, .external_lex_state = 18},
  [671] = {.lex_state = 1, .external_lex_state = 18},
  [672] = {.lex_state = 2},
  [673] = {.lex_state = 3, .external_lex_state = 11},
  [674] = {.lex_state = 7, .external_lex_state = 15},
  [675] = {.lex_state = 1, .external_lex_state = 19},
  [676] = {.lex_state = 1, .external_lex_state = 16},
  [677] = {.lex_state = 1, .external_lex_state = 16},
  [678] = {.lex_state = 1, .external_lex_state = 16},
  [679] = {.lex_state = 4, .external_lex_state = 15},
  [680] = {.lex_state = 1, .external_lex_state = 16},
  [681] = {.lex_state = 2},
  [682] = {.lex_state = 9, .external_lex_state = 11},
  [683] = {.lex_state = 9, .external_lex_state = 20},
  [684] = {.lex_state = 2, .external_lex_state = 17},
  [685] = {.lex_state = 2, .external_lex_state = 17},
  [686] = {.lex_state = 2},
  [687] = {.lex_state = 2},
  [688] = {.lex_state = 2},
  [689] = {.lex_state = 2},
  [690] = {.lex_state = 2, .external_lex_state = 17},
  [691] = {.lex_state = 2},
  [692] = {.lex_state = 9, .external_lex_state = 21},
  [693] = {.lex_state = 2},
  [694] = {.lex_state = 2},
  [695] = {.lex_state = 9, .external_lex_state = 21},
  [696] = {.lex_state = 2},
  [697] = {.lex_state = 2},
  [698] = {.lex_state = 2},
  [699] = {.lex_state = 2},
  [700] = {.lex_state = 11, .external_lex_state = 22},
  [701] = {.lex_state = 11, .external_lex_state = 15},
  [702] = {.lex_state = 2},
  [703] = {.lex_state = 2},
  [704] = {.lex_state = 9, .external_lex_state = 20},
  [705] = {.lex_state = 9, .external_lex_state = 11},
  [706] = {.lex_state = 5, .external_lex_state = 17},
  [707] = {.lex_state = 9, .external_lex_state = 11},
  [708] = {.lex_state = 9, .external_lex_state = 11},
  [709] = {.lex_state = 9, .external_lex_state = 11},
  [710] = {.lex_state = 5, .external_lex_state = 17},
  [711] = {.lex_state = 9, .external_lex_state = 11},
  [712] = {.lex_state = 5, .external_lex_state = 17},
  [713] = {.lex_state = 2182},
  [714] = {.lex_state = 11, .external_lex_state = 23},
  [715] = {.lex_state = 5, .external_lex_state = 17},
  [716] = {.lex_state = 5, .external_lex_state = 17},
  [717] = {.lex_state = 2},
  [718] = {.lex_state = 5, .external_lex_state = 17},
  [719] = {.lex_state = 2182},
  [720] = {.lex_state = 5, .external_lex_state = 17},
  [721] = {.lex_state = 11, .external_lex_state = 23},
  [722] = {.lex_state = 9, .external_lex_state = 11},
  [723] = {.lex_state = 11, .external_lex_state = 22},
  [724] = {.lex_state = 9, .external_lex_state = 11},
  [725] = {.lex_state = 5, .external_lex_state = 17},
  [726] = {.lex_state = 9, .external_lex_state = 11},
  [727] = {.lex_state = 9, .external_lex_state = 11},
  [728] = {.lex_state = 9, .external_lex_state = 11},
  [729] = {.lex_state = 9, .external_lex_state = 11},
  [730] = {.lex_state = 9, .external_lex_state = 11},
  [731] = {.lex_state = 9, .external_lex_state = 11},
  [732] = {.lex_state = 5, .external_lex_state = 17},
  [733] = {.lex_state = 9, .external_lex_state = 11},
  [734] = {.lex_state = 9, .external_lex_state = 11},
  [735] = {.lex_state = 9, .external_lex_state = 11},
  [736] = {.lex_state = 9, .external_lex_state = 11},
  [737] = {.lex_state = 2182},
  [738] = {.lex_state = 9, .external_lex_state = 11},
  [739] = {.lex_state = 9, .external_lex_state = 11},
  [740] = {.lex_state = 5, .external_lex_state = 17},
  [741] = {.lex_state = 9, .external_lex_state = 11},
  [742] = {.lex_state = 5, .external_lex_state = 17},
  [743] = {.lex_state = 9, .external_lex_state = 11},
  [744] = {.lex_state = 5, .external_lex_state = 17},
  [745] = {.lex_state = 5, .external_lex_state = 17},
  [746] = {.lex_state = 5, .external_lex_state = 17},
  [747] = {.lex_state = 9, .external_lex_state = 11},
  [748] = {.lex_state = 11, .external_lex_state = 15},
  [749] = {.lex_state = 11, .external_lex_state = 15},
  [750] = {.lex_state = 11, .external_lex_state = 15},
  [751] = {.lex_state = 13},
  [752] = {.lex_state = 11, .external_lex_state = 15},
  [753] = {.lex_state = 5},
  [754] = {.lex_state = 5},
  [755] = {.lex_state = 5},
  [756] = {.lex_state = 5},
  [757] = {.lex_state = 13},
  [758] = {.lex_state = 11, .external_lex_state = 15},
  [759] = {.lex_state = 12},
  [760] = {.lex_state = 11, .external_lex_state = 15},
  [761] = {.lex_state = 11, .external_lex_state = 15},
  [762] = {.lex_state = 5},
  [763] = {.lex_state = 5},
  [764] = {.lex_state = 12},
  [765] = {.lex_state = 13},
  [766] = {.lex_state = 5},
  [767] = {.lex_state = 5},
  [768] = {.lex_state = 5},
  [769] = {.lex_state = 12},
  [770] = {.lex_state = 11, .external_lex_state = 15},
  [771] = {.lex_state = 11, .external_lex_state = 15},
  [772] = {.lex_state = 11, .external_lex_state = 15},
  [773] = {.lex_state = 5},
  [774] = {.lex_state = 5},
  [775] = {.lex_state = 5},
  [776] = {.lex_state = 5},
  [777] = {.lex_state = 5},
  [778] = {.lex_state = 11, .external_lex_state = 15},
  [779] = {.lex_state = 11, .external_lex_state = 15},
  [780] = {.lex_state = 12},
  [781] = {.lex_state = 5},
  [782] = {.lex_state = 5},
  [783] = {.lex_state = 5},
  [784] = {.lex_state = 13},
  [785] = {.lex_state = 5},
  [786] = {.lex_state = 12},
  [787] = {.lex_state = 13},
  [788] = {.lex_state = 5},
  [789] = {.lex_state = 12},
  [790] = {.lex_state = 13},
  [791] = {.lex_state = 5},
  [792] = {.lex_state = 11, .external_lex_state = 15},
  [793] = {.lex_state = 11, .external_lex_state = 15},
  [794] = {.lex_state = 11, .external_lex_state = 15},
  [795] = {.lex_state = 5},
  [796] = {.lex_state = 5},
  [797] = {.lex_state = 11, .external_lex_state = 15},
  [798] = {.lex_state = 11, .external_lex_state = 15},
  [799] = {.lex_state = 11, .external_lex_state = 15},
  [800] = {.lex_state = 11, .external_lex_state = 15},
  [801] = {.lex_state = 5},
  [802] = {.lex_state = 5},
  [803] = {.lex_state = 5},
  [804] = {.lex_state = 11, .external_lex_state = 15},
  [805] = {.lex_state = 5},
  [806] = {.lex_state = 12},
  [807] = {.lex_state = 13},
  [808] = {.lex_state = 5},
  [809] = {.lex_state = 5},
  [810] = {.lex_state = 5},
  [811] = {.lex_state = 5},
  [812] = {.lex_state = 5},
  [813] = {.lex_state = 5},
  [814] = {.lex_state = 5},
  [815] = {.lex_state = 5},
  [816] = {.lex_state = 5},
  [817] = {.lex_state = 12},
  [818] = {.lex_state = 13},
  [819] = {.lex_state = 5},
  [820] = {.lex_state = 5},
  [821] = {.lex_state = 11, .external_lex_state = 15},
  [822] = {.lex_state = 5},
  [823] = {.lex_state = 5},
  [824] = {.lex_state = 5},
  [825] = {.lex_state = 5},
  [826] = {.lex_state = 5},
  [827] = {.lex_state = 5},
  [828] = {.lex_state = 14},
  [829] = {.lex_state = 2, .external_lex_state = 17},
  [830] = {.lex_state = 2},
  [831] = {.lex_state = 2, .external_lex_state = 17},
  [832] = {.lex_state = 2},
  [833] = {.lex_state = 14},
  [834] = {.lex_state = 14},
  [835] = {.lex_state = 14},
  [836] = {.lex_state = 14},
  [837] = {.lex_state = 2, .external_lex_state = 24},
  [838] = {.lex_state = 2, .external_lex_state = 24},
  [839] = {.lex_state = 14},
  [840] = {.lex_state = 14},
  [841] = {.lex_state = 2, .external_lex_state = 17},
  [842] = {.lex_state = 2},
  [843] = {.lex_state = 14},
  [844] = {.lex_state = 14},
  [845] = {.lex_state = 14},
  [846] = {.lex_state = 14},
  [847] = {.lex_state = 14},
  [848] = {.lex_state = 14},
  [849] = {.lex_state = 2, .external_lex_state = 17},
  [850] = {.lex_state = 2, .external_lex_state = 17},
  [851] = {.lex_state = 2, .external_lex_state = 17},
  [852] = {.lex_state = 14},
  [853] = {.lex_state = 2},
  [854] = {.lex_state = 14},
  [855] = {.lex_state = 14},
  [856] = {.lex_state = 14},
  [857] = {.lex_state = 14},
  [858] = {.lex_state = 2, .external_lex_state = 17},
  [859] = {.lex_state = 14},
  [860] = {.lex_state = 14},
  [861] = {.lex_state = 14},
  [862] = {.lex_state = 2},
  [863] = {.lex_state = 2},
  [864] = {.lex_state = 2},
  [865] = {.lex_state = 2},
  [866] = {.lex_state = 2},
  [867] = {.lex_state = 2},
  [868] = {.lex_state = 2},
  [869] = {.lex_state = 2},
  [870] = {.lex_state = 2},
  [871] = {.lex_state = 2},
  [872] = {.lex_state = 2},
  [873] = {.lex_state = 2},
  [874] = {.lex_state = 2},
  [875] = {.lex_state = 2},
  [876] = {.lex_state = 2, .external_lex_state = 17},
  [877] = {.lex_state = 2},
  [878] = {.lex_state = 2},
  [879] = {.lex_state = 2},
  [880] = {.lex_state = 2},
  [881] = {.lex_state = 12, .external_lex_state = 24},
  [882] = {.lex_state = 5, .external_lex_state = 24},
  [883] = {.lex_state = 13, .external_lex_state = 24},
  [884] = {.lex_state = 5, .external_lex_state = 17},
  [885] = {.lex_state = 13, .external_lex_state = 17},
  [886] = {.lex_state = 2182, .external_lex_state = 17},
  [887] = {.lex_state = 12, .external_lex_state = 17},
  [888] = {.lex_state = 5, .external_lex_state = 24},
  [889] = {.lex_state = 5},
  [890] = {.lex_state = 5, .external_lex_state = 17},
  [891] = {.lex_state = 12, .external_lex_state = 24},
  [892] = {.lex_state = 2},
  [893] = {.lex_state = 2},
  [894] = {.lex_state = 13, .external_lex_state = 24},
  [895] = {.lex_state = 14, .external_lex_state = 17},
  [896] = {.lex_state = 5},
  [897] = {.lex_state = 14, .external_lex_state = 17},
  [898] = {.lex_state = 13},
  [899] = {.lex_state = 13},
  [900] = {.lex_state = 12},
  [901] = {.lex_state = 2182},
  [902] = {.lex_state = 5},
  [903] = {.lex_state = 12},
  [904] = {.lex_state = 14, .external_lex_state = 24},
  [905] = {.lex_state = 13},
  [906] = {.lex_state = 5},
  [907] = {.lex_state = 5},
  [908] = {.lex_state = 12},
  [909] = {.lex_state = 14, .external_lex_state = 24},
  [910] = {.lex_state = 14, .external_lex_state = 17},
  [911] = {.lex_state = 14},
  [912] = {.lex_state = 14},
  [913] = {.lex_state = 14},
  [914] = {.lex_state = 14},
  [915] = {.lex_state = 14},
  [916] = {.lex_state = 14},
  [917] = {.lex_state = 14},
  [918] = {.lex_state = 14},
  [919] = {.lex_state = 14},
  [920] = {.lex_state = 14},
  [921] = {.lex_state = 14},
  [922] = {.lex_state = 14},
  [923] = {.lex_state = 0},
  [924] = {.lex_state = 0},
  [925] = {.lex_state = 0},
  [926] = {.lex_state = 0},
  [927] = {.lex_state = 0},
  [928] = {.lex_state = 0},
  [929] = {.lex_state = 0},
  [930] = {.lex_state = 0},
  [931] = {.lex_state = 0},
  [932] = {.lex_state = 0},
  [933] = {.lex_state = 0},
  [934] = {.lex_state = 0},
  [935] = {.lex_state = 0},
  [936] = {.lex_state = 0},
  [937] = {.lex_state = 0},
  [938] = {.lex_state = 0},
  [939] = {.lex_state = 0},
  [940] = {.lex_state = 0},
  [941] = {.lex_state = 0},
  [942] = {.lex_state = 0},
  [943] = {.lex_state = 0},
  [944] = {.lex_state = 0},
  [945] = {.lex_state = 0},
  [946] = {.lex_state = 0},
  [947] = {.lex_state = 0},
  [948] = {.lex_state = 0},
  [949] = {.lex_state = 0},
  [950] = {.lex_state = 0},
  [951] = {.lex_state = 15},
  [952] = {.lex_state = 0},
  [953] = {.lex_state = 15},
  [954] = {.lex_state = 15},
  [955] = {.lex_state = 15},
  [956] = {.lex_state = 0},
  [957] = {.lex_state = 0},
  [958] = {.lex_state = 0},
  [959] = {.lex_state = 0},
  [960] = {.lex_state = 0},
  [961] = {.lex_state = 0},
  [962] = {.lex_state = 0},
  [963] = {.lex_state = 16},
  [964] = {.lex_state = 16},
  [965] = {.lex_state = 16},
  [966] = {.lex_state = 16},
  [967] = {.lex_state = 2182},
  [968] = {.lex_state = 2182},
  [969] = {.lex_state = 15},
  [970] = {.lex_state = 16},
  [971] = {.lex_state = 16},
  [972] = {.lex_state = 16},
  [973] = {.lex_state = 2182},
  [974] = {.lex_state = 16},
  [975] = {.lex_state = 16},
  [976] = {.lex_state = 16},
  [977] = {.lex_state = 16},
  [978] = {.lex_state = 16},
  [979] = {.lex_state = 16},
  [980] = {.lex_state = 16},
  [981] = {.lex_state = 16},
  [982] = {.lex_state = 0},
  [983] = {.lex_state = 0},
  [984] = {.lex_state = 0},
  [985] = {.lex_state = 0},
  [986] = {.lex_state = 0},
  [987] = {.lex_state = 0},
  [988] = {.lex_state = 0},
  [989] = {.lex_state = 0},
  [990] = {.lex_state = 0},
  [991] = {.lex_state = 0},
  [992] = {.lex_state = 0},
  [993] = {.lex_state = 0},
  [994] = {.lex_state = 0},
  [995] = {.lex_state = 0},
  [996] = {.lex_state = 0},
  [997] = {.lex_state = 0},
  [998] = {.lex_state = 0},
  [999] = {.lex_state = 0},
  [1000] = {.lex_state = 0},
  [1001] = {.lex_state = 0},
  [1002] = {.lex_state = 0},
  [1003] = {.lex_state = 0},
  [1004] = {.lex_state = 0},
  [1005] = {.lex_state = 0},
  [1006] = {.lex_state = 0},
  [1007] = {.lex_state = 0},
  [1008] = {.lex_state = 0},
  [1009] = {.lex_state = 0},
  [1010] = {.lex_state = 0},
  [1011] = {.lex_state = 0},
  [1012] = {.lex_state = 0},
  [1013] = {.lex_state = 0},
  [1014] = {.lex_state = 0},
  [1015] = {.lex_state = 0},
  [1016] = {.lex_state = 0},
  [1017] = {.lex_state = 0},
  [1018] = {.lex_state = 0},
  [1019] = {.lex_state = 0},
  [1020] = {.lex_state = 0},
  [1021] = {.lex_state = 0},
  [1022] = {.lex_state = 0},
  [1023] = {.lex_state = 0},
  [1024] = {.lex_state = 0},
  [1025] = {.lex_state = 0},
  [1026] = {.lex_state = 0},
  [1027] = {.lex_state = 0},
  [1028] = {.lex_state = 0},
  [1029] = {.lex_state = 0},
  [1030] = {.lex_state = 0},
  [1031] = {.lex_state = 0},
  [1032] = {.lex_state = 0},
  [1033] = {.lex_state = 0},
  [1034] = {.lex_state = 0},
  [1035] = {.lex_state = 0},
  [1036] = {.lex_state = 0},
  [1037] = {.lex_state = 0},
  [1038] = {.lex_state = 0},
  [1039] = {.lex_state = 0},
  [1040] = {.lex_state = 0},
  [1041] = {.lex_state = 0},
  [1042] = {.lex_state = 0},
  [1043] = {.lex_state = 0},
  [1044] = {.lex_state = 0},
  [1045] = {.lex_state = 0},
  [1046] = {.lex_state = 0},
  [1047] = {.lex_state = 0},
  [1048] = {.lex_state = 0},
  [1049] = {.lex_state = 0},
  [1050] = {.lex_state = 0},
  [1051] = {.lex_state = 0},
  [1052] = {.lex_state = 0},
  [1053] = {.lex_state = 0},
  [1054] = {.lex_state = 16},
  [1055] = {.lex_state = 0},
  [1056] = {.lex_state = 0},
  [1057] = {.lex_state = 0},
  [1058] = {.lex_state = 0},
  [1059] = {.lex_state = 16, .external_lex_state = 24},
  [1060] = {.lex_state = 16, .external_lex_state = 24},
  [1061] = {.lex_state = 0},
  [1062] = {.lex_state = 0},
  [1063] = {.lex_state = 0},
  [1064] = {.lex_state = 0},
  [1065] = {.lex_state = 0},
  [1066] = {.lex_state = 0},
  [1067] = {.lex_state = 0},
  [1068] = {.lex_state = 0},
  [1069] = {.lex_state = 0},
  [1070] = {.lex_state = 0},
  [1071] = {.lex_state = 0},
  [1072] = {.lex_state = 0},
  [1073] = {.lex_state = 0},
  [1074] = {.lex_state = 0},
  [1075] = {.lex_state = 0},
  [1076] = {.lex_state = 0},
  [1077] = {.lex_state = 0},
  [1078] = {.lex_state = 0},
  [1079] = {.lex_state = 0},
  [1080] = {.lex_state = 0},
  [1081] = {.lex_state = 0},
  [1082] = {.lex_state = 15, .external_lex_state = 24},
  [1083] = {.lex_state = 0},
  [1084] = {.lex_state = 0},
  [1085] = {.lex_state = 0},
  [1086] = {.lex_state = 0},
  [1087] = {.lex_state = 0},
  [1088] = {.lex_state = 15, .external_lex_state = 24},
  [1089] = {.lex_state = 0},
  [1090] = {.lex_state = 0},
  [1091] = {.lex_state = 0},
  [1092] = {.lex_state = 0},
  [1093] = {.lex_state = 0},
  [1094] = {.lex_state = 0},
  [1095] = {.lex_state = 0},
  [1096] = {.lex_state = 0},
  [1097] = {.lex_state = 0},
  [1098] = {.lex_state = 0},
  [1099] = {.lex_state = 0},
  [1100] = {.lex_state = 0},
  [1101] = {.lex_state = 0},
  [1102] = {.lex_state = 0},
  [1103] = {.lex_state = 16},
  [1104] = {.lex_state = 15},
  [1105] = {.lex_state = 16},
  [1106] = {.lex_state = 15},
  [1107] = {.lex_state = 0},
  [1108] = {.lex_state = 0},
  [1109] = {.lex_state = 0},
  [1110] = {.lex_state = 0},
  [1111] = {.lex_state = 0},
  [1112] = {.lex_state = 0},
  [1113] = {.lex_state = 0},
  [1114] = {.lex_state = 0},
  [1115] = {.lex_state = 0},
  [1116] = {.lex_state = 0},
  [1117] = {.lex_state = 0},
  [1118] = {.lex_state = 0},
  [1119] = {.lex_state = 0},
  [1120] = {.lex_state = 0},
  [1121] = {.lex_state = 0},
  [1122] = {.lex_state = 0},
  [1123] = {.lex_state = 0},
  [1124] = {.lex_state = 0},
  [1125] = {.lex_state = 0},
  [1126] = {.lex_state = 0, .external_lex_state = 25},
  [1127] = {.lex_state = 0},
  [1128] = {.lex_state = 0},
  [1129] = {.lex_state = 0},
  [1130] = {.lex_state = 0},
  [1131] = {.lex_state = 0},
  [1132] = {.lex_state = 0},
  [1133] = {.lex_state = 0},
  [1134] = {.lex_state = 0},
  [1135] = {.lex_state = 0},
  [1136] = {.lex_state = 0},
  [1137] = {.lex_state = 0},
  [1138] = {.lex_state = 0},
  [1139] = {.lex_state = 0},
  [1140] = {.lex_state = 0},
  [1141] = {.lex_state = 0, .external_lex_state = 26},
  [1142] = {.lex_state = 0},
  [1143] = {.lex_state = 0},
  [1144] = {.lex_state = 0},
  [1145] = {.lex_state = 0},
  [1146] = {.lex_state = 0},
  [1147] = {.lex_state = 0},
  [1148] = {.lex_state = 0, .external_lex_state = 26},
  [1149] = {.lex_state = 0},
  [1150] = {.lex_state = 0},
  [1151] = {.lex_state = 0, .external_lex_state = 26},
  [1152] = {.lex_state = 0},
  [1153] = {.lex_state = 0},
  [1154] = {.lex_state = 0, .external_lex_state = 26},
  [1155] = {.lex_state = 0},
  [1156] = {.lex_state = 0},
  [1157] = {.lex_state = 0},
  [1158] = {.lex_state = 0},
  [1159] = {.lex_state = 0},
  [1160] = {.lex_state = 0},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__backslash_escape] = ACTIONS(1),
    [sym_entity_reference] = ACTIONS(1),
    [sym_numeric_character_reference] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [sym__newline_token] = ACTIONS(1),
    [sym_uri_autolink] = ACTIONS(1),
    [sym_email_autolink] = ACTIONS(1),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1),
    [anon_sym_DASH_DASH_GT] = ACTIONS(1),
    [anon_sym_LT_QMARK] = ACTIONS(1),
    [anon_sym_QMARK_GT] = ACTIONS(1),
    [aux_sym__declaration_token1] = ACTIONS(1),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK_RBRACK_GT] = ACTIONS(1),
    [sym__whitespace_ge_2] = ACTIONS(1),
    [aux_sym__whitespace_token1] = ACTIONS(1),
    [sym__word_no_digit] = ACTIONS(1),
    [sym__digits] = ACTIONS(1),
    [sym__error] = ACTIONS(1),
    [sym__trigger_error] = ACTIONS(1),
    [sym__code_span_start] = ACTIONS(1),
    [sym__code_span_close] = ACTIONS(1),
    [sym__emphasis_open_star] = ACTIONS(1),
    [sym__emphasis_open_underscore] = ACTIONS(1),
    [sym__emphasis_close_star] = ACTIONS(1),
    [sym__emphasis_close_underscore] = ACTIONS(1),
    [sym__last_token_whitespace] = ACTIONS(1),
    [sym__last_token_punctuation] = ACTIONS(1),
    [sym__strikethrough_open] = ACTIONS(1),
    [sym__strikethrough_close] = ACTIONS(1),
    [sym__latex_span_start] = ACTIONS(1),
    [sym__latex_span_close] = ACTIONS(1),
    [sym__unclosed_span] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_inline] = STATE(1145),
    [sym_backslash_escape] = STATE(155),
    [sym_code_span] = STATE(155),
    [sym_latex_block] = STATE(155),
    [sym__link_text] = STATE(1124),
    [sym__link_text_non_empty] = STATE(469),
    [sym_shortcut_link] = STATE(529),
    [sym_full_reference_link] = STATE(529),
    [sym_collapsed_reference_link] = STATE(529),
    [sym_inline_link] = STATE(529),
    [sym_image] = STATE(155),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(155),
    [sym__whitespace] = STATE(155),
    [sym__word] = STATE(155),
    [sym__soft_line_break] = STATE(155),
    [sym__inline_base] = STATE(529),
    [sym__text_base] = STATE(155),
    [sym__inline_element] = STATE(529),
    [aux_sym__inline] = STATE(14),
    [sym__strikethrough] = STATE(529),
    [sym__emphasis_star] = STATE(534),
    [sym__strong_emphasis_star] = STATE(529),
    [sym__emphasis_underscore] = STATE(534),
    [sym__strong_emphasis_underscore] = STATE(529),
    [aux_sym__inline_base_repeat1] = STATE(155),
    [sym__backslash_escape] = ACTIONS(3),
    [sym_entity_reference] = ACTIONS(5),
    [sym_numeric_character_reference] = ACTIONS(5),
    [anon_sym_LBRACK] = ACTIONS(7),
    [anon_sym_RBRACK] = ACTIONS(9),
    [anon_sym_LT] = ACTIONS(11),
    [anon_sym_GT] = ACTIONS(13),
    [anon_sym_BANG] = ACTIONS(15),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_POUND] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(13),
    [anon_sym_PERCENT] = ACTIONS(13),
    [anon_sym_AMP] = ACTIONS(17),
    [anon_sym_SQUOTE] = ACTIONS(13),
    [anon_sym_STAR] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(13),
    [anon_sym_COMMA] = ACTIONS(13),
    [anon_sym_DASH] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(13),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_COLON] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(13),
    [anon_sym_QMARK] = ACTIONS(13),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_BSLASH] = ACTIONS(19),
    [anon_sym_CARET] = ACTIONS(13),
    [anon_sym__] = ACTIONS(13),
    [anon_sym_BQUOTE] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_PIPE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(13),
    [anon_sym_TILDE] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(13),
    [sym__newline_token] = ACTIONS(21),
    [sym_uri_autolink] = ACTIONS(5),
    [sym_email_autolink] = ACTIONS(5),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(23),
    [anon_sym_LT_QMARK] = ACTIONS(25),
    [aux_sym__declaration_token1] = ACTIONS(27),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(29),
    [sym__whitespace_ge_2] = ACTIONS(31),
    [aux_sym__whitespace_token1] = ACTIONS(33),
    [sym__word_no_digit] = ACTIONS(5),
    [sym__digits] = ACTIONS(5),
    [sym__code_span_start] = ACTIONS(35),
    [sym__emphasis_open_star] = ACTIONS(37),
    [sym__emphasis_open_underscore] = ACTIONS(39),
    [sym__last_token_whitespace] = ACTIONS(41),
    [sym__strikethrough_open] = ACTIONS(43),
    [sym__latex_span_start] = ACTIONS(45),
    [sym__unclosed_span] = ACTIONS(5),
  },
  [STATE(2)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(52),
    [sym_full_reference_link] = STATE(52),
    [sym_collapsed_reference_link] = STATE(52),
    [sym_inline_link] = STATE(52),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(52),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(52),
    [aux_sym__inline_no_underscore] = STATE(52),
    [sym__strikethrough] = STATE(52),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(52),
    [sym__emphasis_underscore] = STATE(454),
    [sym__strong_emphasis_underscore] = STATE(52),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [ts_builtin_sym_end] = ACTIONS(47),
    [sym__backslash_escape] = ACTIONS(49),
    [sym_entity_reference] = ACTIONS(52),
    [sym_numeric_character_reference] = ACTIONS(52),
    [anon_sym_LBRACK] = ACTIONS(55),
    [anon_sym_RBRACK] = ACTIONS(58),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_GT] = ACTIONS(64),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(64),
    [anon_sym_POUND] = ACTIONS(64),
    [anon_sym_DOLLAR] = ACTIONS(64),
    [anon_sym_PERCENT] = ACTIONS(64),
    [anon_sym_AMP] = ACTIONS(70),
    [anon_sym_SQUOTE] = ACTIONS(64),
    [anon_sym_STAR] = ACTIONS(64),
    [anon_sym_PLUS] = ACTIONS(64),
    [anon_sym_COMMA] = ACTIONS(64),
    [anon_sym_DASH] = ACTIONS(64),
    [anon_sym_DOT] = ACTIONS(64),
    [anon_sym_SLASH] = ACTIONS(64),
    [anon_sym_COLON] = ACTIONS(64),
    [anon_sym_SEMI] = ACTIONS(64),
    [anon_sym_EQ] = ACTIONS(64),
    [anon_sym_QMARK] = ACTIONS(64),
    [anon_sym_AT] = ACTIONS(64),
    [anon_sym_BSLASH] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(64),
    [anon_sym__] = ACTIONS(64),
    [anon_sym_BQUOTE] = ACTIONS(64),
    [anon_sym_LBRACE] = ACTIONS(64),
    [anon_sym_PIPE] = ACTIONS(64),
    [anon_sym_RBRACE] = ACTIONS(64),
    [anon_sym_TILDE] = ACTIONS(64),
    [anon_sym_LPAREN] = ACTIONS(64),
    [anon_sym_RPAREN] = ACTIONS(64),
    [sym__newline_token] = ACTIONS(76),
    [sym_uri_autolink] = ACTIONS(52),
    [sym_email_autolink] = ACTIONS(52),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(79),
    [anon_sym_LT_QMARK] = ACTIONS(82),
    [aux_sym__declaration_token1] = ACTIONS(85),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(88),
    [sym__whitespace_ge_2] = ACTIONS(91),
    [aux_sym__whitespace_token1] = ACTIONS(94),
    [sym__word_no_digit] = ACTIONS(52),
    [sym__digits] = ACTIONS(52),
    [sym__code_span_start] = ACTIONS(97),
    [sym__emphasis_open_star] = ACTIONS(100),
    [sym__emphasis_open_underscore] = ACTIONS(103),
    [sym__last_token_punctuation] = ACTIONS(106),
    [sym__strikethrough_open] = ACTIONS(108),
    [sym__latex_span_start] = ACTIONS(111),
    [sym__unclosed_span] = ACTIONS(52),
  },
  [STATE(3)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(28),
    [sym_full_reference_link] = STATE(28),
    [sym_collapsed_reference_link] = STATE(28),
    [sym_inline_link] = STATE(28),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(28),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(28),
    [aux_sym__inline_no_star] = STATE(28),
    [sym__strikethrough] = STATE(28),
    [sym__emphasis_star] = STATE(423),
    [sym__strong_emphasis_star] = STATE(28),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(28),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [ts_builtin_sym_end] = ACTIONS(47),
    [sym__backslash_escape] = ACTIONS(114),
    [sym_entity_reference] = ACTIONS(117),
    [sym_numeric_character_reference] = ACTIONS(117),
    [anon_sym_LBRACK] = ACTIONS(120),
    [anon_sym_RBRACK] = ACTIONS(123),
    [anon_sym_LT] = ACTIONS(126),
    [anon_sym_GT] = ACTIONS(129),
    [anon_sym_BANG] = ACTIONS(132),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_POUND] = ACTIONS(129),
    [anon_sym_DOLLAR] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_AMP] = ACTIONS(135),
    [anon_sym_SQUOTE] = ACTIONS(129),
    [anon_sym_STAR] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_DOT] = ACTIONS(129),
    [anon_sym_SLASH] = ACTIONS(129),
    [anon_sym_COLON] = ACTIONS(129),
    [anon_sym_SEMI] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(129),
    [anon_sym_QMARK] = ACTIONS(129),
    [anon_sym_AT] = ACTIONS(129),
    [anon_sym_BSLASH] = ACTIONS(138),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym__] = ACTIONS(129),
    [anon_sym_BQUOTE] = ACTIONS(129),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_PIPE] = ACTIONS(129),
    [anon_sym_RBRACE] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(129),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_RPAREN] = ACTIONS(129),
    [sym__newline_token] = ACTIONS(141),
    [sym_uri_autolink] = ACTIONS(117),
    [sym_email_autolink] = ACTIONS(117),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(144),
    [anon_sym_LT_QMARK] = ACTIONS(147),
    [aux_sym__declaration_token1] = ACTIONS(150),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(153),
    [sym__whitespace_ge_2] = ACTIONS(156),
    [aux_sym__whitespace_token1] = ACTIONS(159),
    [sym__word_no_digit] = ACTIONS(117),
    [sym__digits] = ACTIONS(117),
    [sym__code_span_start] = ACTIONS(162),
    [sym__emphasis_open_star] = ACTIONS(165),
    [sym__emphasis_open_underscore] = ACTIONS(168),
    [sym__last_token_punctuation] = ACTIONS(171),
    [sym__strikethrough_open] = ACTIONS(173),
    [sym__latex_span_start] = ACTIONS(176),
    [sym__unclosed_span] = ACTIONS(117),
  },
  [STATE(4)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(30),
    [sym_full_reference_link] = STATE(30),
    [sym_collapsed_reference_link] = STATE(30),
    [sym_inline_link] = STATE(30),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(30),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(30),
    [aux_sym__inline_no_underscore] = STATE(30),
    [sym__strikethrough] = STATE(30),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(30),
    [sym__emphasis_underscore] = STATE(424),
    [sym__strong_emphasis_underscore] = STATE(30),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(179),
    [sym_entity_reference] = ACTIONS(182),
    [sym_numeric_character_reference] = ACTIONS(182),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(194),
    [anon_sym_BANG] = ACTIONS(197),
    [anon_sym_DQUOTE] = ACTIONS(194),
    [anon_sym_POUND] = ACTIONS(194),
    [anon_sym_DOLLAR] = ACTIONS(194),
    [anon_sym_PERCENT] = ACTIONS(194),
    [anon_sym_AMP] = ACTIONS(200),
    [anon_sym_SQUOTE] = ACTIONS(194),
    [anon_sym_STAR] = ACTIONS(194),
    [anon_sym_PLUS] = ACTIONS(194),
    [anon_sym_COMMA] = ACTIONS(194),
    [anon_sym_DASH] = ACTIONS(194),
    [anon_sym_DOT] = ACTIONS(194),
    [anon_sym_SLASH] = ACTIONS(194),
    [anon_sym_COLON] = ACTIONS(194),
    [anon_sym_SEMI] = ACTIONS(194),
    [anon_sym_EQ] = ACTIONS(194),
    [anon_sym_QMARK] = ACTIONS(194),
    [anon_sym_AT] = ACTIONS(194),
    [anon_sym_BSLASH] = ACTIONS(203),
    [anon_sym_CARET] = ACTIONS(194),
    [anon_sym__] = ACTIONS(194),
    [anon_sym_BQUOTE] = ACTIONS(194),
    [anon_sym_LBRACE] = ACTIONS(194),
    [anon_sym_PIPE] = ACTIONS(194),
    [anon_sym_RBRACE] = ACTIONS(194),
    [anon_sym_TILDE] = ACTIONS(194),
    [anon_sym_LPAREN] = ACTIONS(194),
    [anon_sym_RPAREN] = ACTIONS(194),
    [sym__newline_token] = ACTIONS(206),
    [sym_uri_autolink] = ACTIONS(182),
    [sym_email_autolink] = ACTIONS(182),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(209),
    [anon_sym_LT_QMARK] = ACTIONS(212),
    [aux_sym__declaration_token1] = ACTIONS(215),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(218),
    [sym__whitespace_ge_2] = ACTIONS(221),
    [aux_sym__whitespace_token1] = ACTIONS(224),
    [sym__word_no_digit] = ACTIONS(182),
    [sym__digits] = ACTIONS(182),
    [sym__code_span_start] = ACTIONS(227),
    [sym__emphasis_open_star] = ACTIONS(230),
    [sym__emphasis_open_underscore] = ACTIONS(233),
    [sym__emphasis_close_star] = ACTIONS(236),
    [sym__last_token_punctuation] = ACTIONS(238),
    [sym__strikethrough_open] = ACTIONS(240),
    [sym__latex_span_start] = ACTIONS(243),
    [sym__unclosed_span] = ACTIONS(182),
  },
  [STATE(5)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(31),
    [sym_full_reference_link] = STATE(31),
    [sym_collapsed_reference_link] = STATE(31),
    [sym_inline_link] = STATE(31),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(31),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(31),
    [aux_sym__inline_no_tilde] = STATE(31),
    [sym__strikethrough] = STATE(31),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(31),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(31),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(246),
    [sym_entity_reference] = ACTIONS(249),
    [sym_numeric_character_reference] = ACTIONS(249),
    [anon_sym_LBRACK] = ACTIONS(252),
    [anon_sym_RBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(261),
    [anon_sym_BANG] = ACTIONS(264),
    [anon_sym_DQUOTE] = ACTIONS(261),
    [anon_sym_POUND] = ACTIONS(261),
    [anon_sym_DOLLAR] = ACTIONS(261),
    [anon_sym_PERCENT] = ACTIONS(261),
    [anon_sym_AMP] = ACTIONS(267),
    [anon_sym_SQUOTE] = ACTIONS(261),
    [anon_sym_STAR] = ACTIONS(261),
    [anon_sym_PLUS] = ACTIONS(261),
    [anon_sym_COMMA] = ACTIONS(261),
    [anon_sym_DASH] = ACTIONS(261),
    [anon_sym_DOT] = ACTIONS(261),
    [anon_sym_SLASH] = ACTIONS(261),
    [anon_sym_COLON] = ACTIONS(261),
    [anon_sym_SEMI] = ACTIONS(261),
    [anon_sym_EQ] = ACTIONS(261),
    [anon_sym_QMARK] = ACTIONS(261),
    [anon_sym_AT] = ACTIONS(261),
    [anon_sym_BSLASH] = ACTIONS(270),
    [anon_sym_CARET] = ACTIONS(261),
    [anon_sym__] = ACTIONS(261),
    [anon_sym_BQUOTE] = ACTIONS(261),
    [anon_sym_LBRACE] = ACTIONS(261),
    [anon_sym_PIPE] = ACTIONS(261),
    [anon_sym_RBRACE] = ACTIONS(261),
    [anon_sym_TILDE] = ACTIONS(261),
    [anon_sym_LPAREN] = ACTIONS(261),
    [anon_sym_RPAREN] = ACTIONS(261),
    [sym__newline_token] = ACTIONS(273),
    [sym_uri_autolink] = ACTIONS(249),
    [sym_email_autolink] = ACTIONS(249),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(276),
    [anon_sym_LT_QMARK] = ACTIONS(279),
    [aux_sym__declaration_token1] = ACTIONS(282),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(285),
    [sym__whitespace_ge_2] = ACTIONS(288),
    [aux_sym__whitespace_token1] = ACTIONS(291),
    [sym__word_no_digit] = ACTIONS(249),
    [sym__digits] = ACTIONS(249),
    [sym__code_span_start] = ACTIONS(294),
    [sym__emphasis_open_star] = ACTIONS(297),
    [sym__emphasis_open_underscore] = ACTIONS(300),
    [sym__emphasis_close_star] = ACTIONS(236),
    [sym__last_token_punctuation] = ACTIONS(303),
    [sym__strikethrough_open] = ACTIONS(305),
    [sym__latex_span_start] = ACTIONS(308),
    [sym__unclosed_span] = ACTIONS(249),
  },
  [STATE(6)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(17),
    [sym_full_reference_link] = STATE(17),
    [sym_collapsed_reference_link] = STATE(17),
    [sym_inline_link] = STATE(17),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(17),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(17),
    [aux_sym__inline_no_tilde] = STATE(17),
    [sym__strikethrough] = STATE(17),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(17),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(17),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [ts_builtin_sym_end] = ACTIONS(47),
    [sym__backslash_escape] = ACTIONS(311),
    [sym_entity_reference] = ACTIONS(314),
    [sym_numeric_character_reference] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(317),
    [anon_sym_RBRACK] = ACTIONS(320),
    [anon_sym_LT] = ACTIONS(323),
    [anon_sym_GT] = ACTIONS(326),
    [anon_sym_BANG] = ACTIONS(329),
    [anon_sym_DQUOTE] = ACTIONS(326),
    [anon_sym_POUND] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(326),
    [anon_sym_PERCENT] = ACTIONS(326),
    [anon_sym_AMP] = ACTIONS(332),
    [anon_sym_SQUOTE] = ACTIONS(326),
    [anon_sym_STAR] = ACTIONS(326),
    [anon_sym_PLUS] = ACTIONS(326),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_DASH] = ACTIONS(326),
    [anon_sym_DOT] = ACTIONS(326),
    [anon_sym_SLASH] = ACTIONS(326),
    [anon_sym_COLON] = ACTIONS(326),
    [anon_sym_SEMI] = ACTIONS(326),
    [anon_sym_EQ] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(326),
    [anon_sym_AT] = ACTIONS(326),
    [anon_sym_BSLASH] = ACTIONS(335),
    [anon_sym_CARET] = ACTIONS(326),
    [anon_sym__] = ACTIONS(326),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_LBRACE] = ACTIONS(326),
    [anon_sym_PIPE] = ACTIONS(326),
    [anon_sym_RBRACE] = ACTIONS(326),
    [anon_sym_TILDE] = ACTIONS(326),
    [anon_sym_LPAREN] = ACTIONS(326),
    [anon_sym_RPAREN] = ACTIONS(326),
    [sym__newline_token] = ACTIONS(338),
    [sym_uri_autolink] = ACTIONS(314),
    [sym_email_autolink] = ACTIONS(314),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(341),
    [anon_sym_LT_QMARK] = ACTIONS(344),
    [aux_sym__declaration_token1] = ACTIONS(347),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(350),
    [sym__whitespace_ge_2] = ACTIONS(353),
    [aux_sym__whitespace_token1] = ACTIONS(356),
    [sym__word_no_digit] = ACTIONS(314),
    [sym__digits] = ACTIONS(314),
    [sym__code_span_start] = ACTIONS(359),
    [sym__emphasis_open_star] = ACTIONS(362),
    [sym__emphasis_open_underscore] = ACTIONS(365),
    [sym__last_token_punctuation] = ACTIONS(368),
    [sym__strikethrough_open] = ACTIONS(370),
    [sym__latex_span_start] = ACTIONS(373),
    [sym__unclosed_span] = ACTIONS(314),
  },
  [STATE(7)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(11),
    [sym_full_reference_link] = STATE(11),
    [sym_collapsed_reference_link] = STATE(11),
    [sym_inline_link] = STATE(11),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(11),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(11),
    [aux_sym__inline_no_star] = STATE(11),
    [sym__strikethrough] = STATE(11),
    [sym__emphasis_star] = STATE(445),
    [sym__strong_emphasis_star] = STATE(11),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(11),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(376),
    [sym_entity_reference] = ACTIONS(379),
    [sym_numeric_character_reference] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(382),
    [anon_sym_RBRACK] = ACTIONS(385),
    [anon_sym_LT] = ACTIONS(388),
    [anon_sym_GT] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(394),
    [anon_sym_DQUOTE] = ACTIONS(391),
    [anon_sym_POUND] = ACTIONS(391),
    [anon_sym_DOLLAR] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(397),
    [anon_sym_SQUOTE] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(391),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(391),
    [anon_sym_SLASH] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(391),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_EQ] = ACTIONS(391),
    [anon_sym_QMARK] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(391),
    [anon_sym_BSLASH] = ACTIONS(400),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym__] = ACTIONS(391),
    [anon_sym_BQUOTE] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_PIPE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_RPAREN] = ACTIONS(391),
    [sym__newline_token] = ACTIONS(403),
    [sym_uri_autolink] = ACTIONS(379),
    [sym_email_autolink] = ACTIONS(379),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(406),
    [anon_sym_LT_QMARK] = ACTIONS(409),
    [aux_sym__declaration_token1] = ACTIONS(412),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(415),
    [sym__whitespace_ge_2] = ACTIONS(418),
    [aux_sym__whitespace_token1] = ACTIONS(421),
    [sym__word_no_digit] = ACTIONS(379),
    [sym__digits] = ACTIONS(379),
    [sym__code_span_start] = ACTIONS(424),
    [sym__emphasis_open_star] = ACTIONS(427),
    [sym__emphasis_open_underscore] = ACTIONS(430),
    [sym__emphasis_close_underscore] = ACTIONS(433),
    [sym__last_token_punctuation] = ACTIONS(435),
    [sym__strikethrough_open] = ACTIONS(437),
    [sym__latex_span_start] = ACTIONS(440),
    [sym__unclosed_span] = ACTIONS(379),
  },
  [STATE(8)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(36),
    [sym_full_reference_link] = STATE(36),
    [sym_collapsed_reference_link] = STATE(36),
    [sym_inline_link] = STATE(36),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(36),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(36),
    [aux_sym__inline_no_tilde] = STATE(36),
    [sym__strikethrough] = STATE(36),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(36),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(36),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(443),
    [sym_entity_reference] = ACTIONS(446),
    [sym_numeric_character_reference] = ACTIONS(446),
    [anon_sym_LBRACK] = ACTIONS(449),
    [anon_sym_RBRACK] = ACTIONS(452),
    [anon_sym_LT] = ACTIONS(455),
    [anon_sym_GT] = ACTIONS(458),
    [anon_sym_BANG] = ACTIONS(461),
    [anon_sym_DQUOTE] = ACTIONS(458),
    [anon_sym_POUND] = ACTIONS(458),
    [anon_sym_DOLLAR] = ACTIONS(458),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(464),
    [anon_sym_SQUOTE] = ACTIONS(458),
    [anon_sym_STAR] = ACTIONS(458),
    [anon_sym_PLUS] = ACTIONS(458),
    [anon_sym_COMMA] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(458),
    [anon_sym_DOT] = ACTIONS(458),
    [anon_sym_SLASH] = ACTIONS(458),
    [anon_sym_COLON] = ACTIONS(458),
    [anon_sym_SEMI] = ACTIONS(458),
    [anon_sym_EQ] = ACTIONS(458),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_AT] = ACTIONS(458),
    [anon_sym_BSLASH] = ACTIONS(467),
    [anon_sym_CARET] = ACTIONS(458),
    [anon_sym__] = ACTIONS(458),
    [anon_sym_BQUOTE] = ACTIONS(458),
    [anon_sym_LBRACE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(458),
    [anon_sym_RBRACE] = ACTIONS(458),
    [anon_sym_TILDE] = ACTIONS(458),
    [anon_sym_LPAREN] = ACTIONS(458),
    [anon_sym_RPAREN] = ACTIONS(458),
    [sym__newline_token] = ACTIONS(470),
    [sym_uri_autolink] = ACTIONS(446),
    [sym_email_autolink] = ACTIONS(446),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(473),
    [anon_sym_LT_QMARK] = ACTIONS(476),
    [aux_sym__declaration_token1] = ACTIONS(479),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(482),
    [sym__whitespace_ge_2] = ACTIONS(485),
    [aux_sym__whitespace_token1] = ACTIONS(488),
    [sym__word_no_digit] = ACTIONS(446),
    [sym__digits] = ACTIONS(446),
    [sym__code_span_start] = ACTIONS(491),
    [sym__emphasis_open_star] = ACTIONS(494),
    [sym__emphasis_open_underscore] = ACTIONS(497),
    [sym__emphasis_close_underscore] = ACTIONS(433),
    [sym__last_token_punctuation] = ACTIONS(500),
    [sym__strikethrough_open] = ACTIONS(502),
    [sym__latex_span_start] = ACTIONS(505),
    [sym__unclosed_span] = ACTIONS(446),
  },
  [STATE(9)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(40),
    [sym_full_reference_link] = STATE(40),
    [sym_collapsed_reference_link] = STATE(40),
    [sym_inline_link] = STATE(40),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(40),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(40),
    [aux_sym__inline_no_star] = STATE(40),
    [sym__strikethrough] = STATE(40),
    [sym__emphasis_star] = STATE(468),
    [sym__strong_emphasis_star] = STATE(40),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(40),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(508),
    [sym_entity_reference] = ACTIONS(511),
    [sym_numeric_character_reference] = ACTIONS(511),
    [anon_sym_LBRACK] = ACTIONS(514),
    [anon_sym_RBRACK] = ACTIONS(517),
    [anon_sym_LT] = ACTIONS(520),
    [anon_sym_GT] = ACTIONS(523),
    [anon_sym_BANG] = ACTIONS(526),
    [anon_sym_DQUOTE] = ACTIONS(523),
    [anon_sym_POUND] = ACTIONS(523),
    [anon_sym_DOLLAR] = ACTIONS(523),
    [anon_sym_PERCENT] = ACTIONS(523),
    [anon_sym_AMP] = ACTIONS(529),
    [anon_sym_SQUOTE] = ACTIONS(523),
    [anon_sym_STAR] = ACTIONS(523),
    [anon_sym_PLUS] = ACTIONS(523),
    [anon_sym_COMMA] = ACTIONS(523),
    [anon_sym_DASH] = ACTIONS(523),
    [anon_sym_DOT] = ACTIONS(523),
    [anon_sym_SLASH] = ACTIONS(523),
    [anon_sym_COLON] = ACTIONS(523),
    [anon_sym_SEMI] = ACTIONS(523),
    [anon_sym_EQ] = ACTIONS(523),
    [anon_sym_QMARK] = ACTIONS(523),
    [anon_sym_AT] = ACTIONS(523),
    [anon_sym_BSLASH] = ACTIONS(532),
    [anon_sym_CARET] = ACTIONS(523),
    [anon_sym__] = ACTIONS(523),
    [anon_sym_BQUOTE] = ACTIONS(523),
    [anon_sym_LBRACE] = ACTIONS(523),
    [anon_sym_PIPE] = ACTIONS(523),
    [anon_sym_RBRACE] = ACTIONS(523),
    [anon_sym_TILDE] = ACTIONS(523),
    [anon_sym_LPAREN] = ACTIONS(523),
    [anon_sym_RPAREN] = ACTIONS(523),
    [sym__newline_token] = ACTIONS(535),
    [sym_uri_autolink] = ACTIONS(511),
    [sym_email_autolink] = ACTIONS(511),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(538),
    [anon_sym_LT_QMARK] = ACTIONS(541),
    [aux_sym__declaration_token1] = ACTIONS(544),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(547),
    [sym__whitespace_ge_2] = ACTIONS(550),
    [aux_sym__whitespace_token1] = ACTIONS(553),
    [sym__word_no_digit] = ACTIONS(511),
    [sym__digits] = ACTIONS(511),
    [sym__code_span_start] = ACTIONS(556),
    [sym__emphasis_open_star] = ACTIONS(559),
    [sym__emphasis_open_underscore] = ACTIONS(562),
    [sym__last_token_punctuation] = ACTIONS(565),
    [sym__strikethrough_open] = ACTIONS(567),
    [sym__strikethrough_close] = ACTIONS(570),
    [sym__latex_span_start] = ACTIONS(572),
    [sym__unclosed_span] = ACTIONS(511),
  },
  [STATE(10)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(41),
    [sym_full_reference_link] = STATE(41),
    [sym_collapsed_reference_link] = STATE(41),
    [sym_inline_link] = STATE(41),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(41),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(41),
    [aux_sym__inline_no_underscore] = STATE(41),
    [sym__strikethrough] = STATE(41),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(41),
    [sym__emphasis_underscore] = STATE(470),
    [sym__strong_emphasis_underscore] = STATE(41),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(575),
    [sym_entity_reference] = ACTIONS(578),
    [sym_numeric_character_reference] = ACTIONS(578),
    [anon_sym_LBRACK] = ACTIONS(581),
    [anon_sym_RBRACK] = ACTIONS(584),
    [anon_sym_LT] = ACTIONS(587),
    [anon_sym_GT] = ACTIONS(590),
    [anon_sym_BANG] = ACTIONS(593),
    [anon_sym_DQUOTE] = ACTIONS(590),
    [anon_sym_POUND] = ACTIONS(590),
    [anon_sym_DOLLAR] = ACTIONS(590),
    [anon_sym_PERCENT] = ACTIONS(590),
    [anon_sym_AMP] = ACTIONS(596),
    [anon_sym_SQUOTE] = ACTIONS(590),
    [anon_sym_STAR] = ACTIONS(590),
    [anon_sym_PLUS] = ACTIONS(590),
    [anon_sym_COMMA] = ACTIONS(590),
    [anon_sym_DASH] = ACTIONS(590),
    [anon_sym_DOT] = ACTIONS(590),
    [anon_sym_SLASH] = ACTIONS(590),
    [anon_sym_COLON] = ACTIONS(590),
    [anon_sym_SEMI] = ACTIONS(590),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_QMARK] = ACTIONS(590),
    [anon_sym_AT] = ACTIONS(590),
    [anon_sym_BSLASH] = ACTIONS(599),
    [anon_sym_CARET] = ACTIONS(590),
    [anon_sym__] = ACTIONS(590),
    [anon_sym_BQUOTE] = ACTIONS(590),
    [anon_sym_LBRACE] = ACTIONS(590),
    [anon_sym_PIPE] = ACTIONS(590),
    [anon_sym_RBRACE] = ACTIONS(590),
    [anon_sym_TILDE] = ACTIONS(590),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(590),
    [sym__newline_token] = ACTIONS(602),
    [sym_uri_autolink] = ACTIONS(578),
    [sym_email_autolink] = ACTIONS(578),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(605),
    [anon_sym_LT_QMARK] = ACTIONS(608),
    [aux_sym__declaration_token1] = ACTIONS(611),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(614),
    [sym__whitespace_ge_2] = ACTIONS(617),
    [aux_sym__whitespace_token1] = ACTIONS(620),
    [sym__word_no_digit] = ACTIONS(578),
    [sym__digits] = ACTIONS(578),
    [sym__code_span_start] = ACTIONS(623),
    [sym__emphasis_open_star] = ACTIONS(626),
    [sym__emphasis_open_underscore] = ACTIONS(629),
    [sym__last_token_punctuation] = ACTIONS(632),
    [sym__strikethrough_open] = ACTIONS(634),
    [sym__strikethrough_close] = ACTIONS(570),
    [sym__latex_span_start] = ACTIONS(637),
    [sym__unclosed_span] = ACTIONS(578),
  },
  [STATE(11)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(678),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(12)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(35),
    [sym_full_reference_link] = STATE(35),
    [sym_collapsed_reference_link] = STATE(35),
    [sym_inline_link] = STATE(35),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(35),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(35),
    [aux_sym__inline_no_underscore] = STATE(35),
    [sym__strikethrough] = STATE(35),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(35),
    [sym__emphasis_underscore] = STATE(447),
    [sym__strong_emphasis_underscore] = STATE(35),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__last_token_punctuation] = ACTIONS(722),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(13)] = {
    [sym_backslash_escape] = STATE(155),
    [sym_code_span] = STATE(155),
    [sym_latex_block] = STATE(155),
    [sym__link_text] = STATE(1124),
    [sym__link_text_non_empty] = STATE(469),
    [sym_shortcut_link] = STATE(529),
    [sym_full_reference_link] = STATE(529),
    [sym_collapsed_reference_link] = STATE(529),
    [sym_inline_link] = STATE(529),
    [sym_image] = STATE(155),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(155),
    [sym__whitespace] = STATE(155),
    [sym__word] = STATE(155),
    [sym__soft_line_break] = STATE(155),
    [sym__inline_base] = STATE(529),
    [sym__text_base] = STATE(155),
    [sym__inline_element] = STATE(529),
    [aux_sym__inline] = STATE(18),
    [sym__strikethrough] = STATE(529),
    [sym__emphasis_star] = STATE(534),
    [sym__strong_emphasis_star] = STATE(529),
    [sym__emphasis_underscore] = STATE(534),
    [sym__strong_emphasis_underscore] = STATE(529),
    [aux_sym__inline_base_repeat1] = STATE(155),
    [ts_builtin_sym_end] = ACTIONS(728),
    [sym__backslash_escape] = ACTIONS(3),
    [sym_entity_reference] = ACTIONS(5),
    [sym_numeric_character_reference] = ACTIONS(5),
    [anon_sym_LBRACK] = ACTIONS(7),
    [anon_sym_RBRACK] = ACTIONS(9),
    [anon_sym_LT] = ACTIONS(11),
    [anon_sym_GT] = ACTIONS(13),
    [anon_sym_BANG] = ACTIONS(15),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_POUND] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(13),
    [anon_sym_PERCENT] = ACTIONS(13),
    [anon_sym_AMP] = ACTIONS(17),
    [anon_sym_SQUOTE] = ACTIONS(13),
    [anon_sym_STAR] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(13),
    [anon_sym_COMMA] = ACTIONS(13),
    [anon_sym_DASH] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(13),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_COLON] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(13),
    [anon_sym_QMARK] = ACTIONS(13),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_BSLASH] = ACTIONS(19),
    [anon_sym_CARET] = ACTIONS(13),
    [anon_sym__] = ACTIONS(13),
    [anon_sym_BQUOTE] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_PIPE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(13),
    [anon_sym_TILDE] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(13),
    [sym__newline_token] = ACTIONS(21),
    [sym_uri_autolink] = ACTIONS(5),
    [sym_email_autolink] = ACTIONS(5),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(23),
    [anon_sym_LT_QMARK] = ACTIONS(25),
    [aux_sym__declaration_token1] = ACTIONS(27),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(29),
    [sym__whitespace_ge_2] = ACTIONS(31),
    [aux_sym__whitespace_token1] = ACTIONS(33),
    [sym__word_no_digit] = ACTIONS(5),
    [sym__digits] = ACTIONS(5),
    [sym__code_span_start] = ACTIONS(35),
    [sym__emphasis_open_star] = ACTIONS(37),
    [sym__emphasis_open_underscore] = ACTIONS(39),
    [sym__strikethrough_open] = ACTIONS(43),
    [sym__latex_span_start] = ACTIONS(45),
    [sym__unclosed_span] = ACTIONS(5),
  },
  [STATE(14)] = {
    [sym_backslash_escape] = STATE(155),
    [sym_code_span] = STATE(155),
    [sym_latex_block] = STATE(155),
    [sym__link_text] = STATE(1124),
    [sym__link_text_non_empty] = STATE(469),
    [sym_shortcut_link] = STATE(529),
    [sym_full_reference_link] = STATE(529),
    [sym_collapsed_reference_link] = STATE(529),
    [sym_inline_link] = STATE(529),
    [sym_image] = STATE(155),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(155),
    [sym__whitespace] = STATE(155),
    [sym__word] = STATE(155),
    [sym__soft_line_break] = STATE(155),
    [sym__inline_base] = STATE(529),
    [sym__text_base] = STATE(155),
    [sym__inline_element] = STATE(529),
    [aux_sym__inline] = STATE(18),
    [sym__strikethrough] = STATE(529),
    [sym__emphasis_star] = STATE(534),
    [sym__strong_emphasis_star] = STATE(529),
    [sym__emphasis_underscore] = STATE(534),
    [sym__strong_emphasis_underscore] = STATE(529),
    [aux_sym__inline_base_repeat1] = STATE(155),
    [ts_builtin_sym_end] = ACTIONS(730),
    [sym__backslash_escape] = ACTIONS(3),
    [sym_entity_reference] = ACTIONS(5),
    [sym_numeric_character_reference] = ACTIONS(5),
    [anon_sym_LBRACK] = ACTIONS(7),
    [anon_sym_RBRACK] = ACTIONS(9),
    [anon_sym_LT] = ACTIONS(11),
    [anon_sym_GT] = ACTIONS(13),
    [anon_sym_BANG] = ACTIONS(15),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_POUND] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(13),
    [anon_sym_PERCENT] = ACTIONS(13),
    [anon_sym_AMP] = ACTIONS(17),
    [anon_sym_SQUOTE] = ACTIONS(13),
    [anon_sym_STAR] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(13),
    [anon_sym_COMMA] = ACTIONS(13),
    [anon_sym_DASH] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(13),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_COLON] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(13),
    [anon_sym_QMARK] = ACTIONS(13),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_BSLASH] = ACTIONS(19),
    [anon_sym_CARET] = ACTIONS(13),
    [anon_sym__] = ACTIONS(13),
    [anon_sym_BQUOTE] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_PIPE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(13),
    [anon_sym_TILDE] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(13),
    [sym__newline_token] = ACTIONS(21),
    [sym_uri_autolink] = ACTIONS(5),
    [sym_email_autolink] = ACTIONS(5),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(23),
    [anon_sym_LT_QMARK] = ACTIONS(25),
    [aux_sym__declaration_token1] = ACTIONS(27),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(29),
    [sym__whitespace_ge_2] = ACTIONS(31),
    [aux_sym__whitespace_token1] = ACTIONS(33),
    [sym__word_no_digit] = ACTIONS(5),
    [sym__digits] = ACTIONS(5),
    [sym__code_span_start] = ACTIONS(35),
    [sym__emphasis_open_star] = ACTIONS(37),
    [sym__emphasis_open_underscore] = ACTIONS(39),
    [sym__strikethrough_open] = ACTIONS(43),
    [sym__latex_span_start] = ACTIONS(45),
    [sym__unclosed_span] = ACTIONS(5),
  },
  [STATE(15)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(29),
    [sym_full_reference_link] = STATE(29),
    [sym_collapsed_reference_link] = STATE(29),
    [sym_inline_link] = STATE(29),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(29),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(29),
    [aux_sym__inline_no_star] = STATE(29),
    [sym__strikethrough] = STATE(29),
    [sym__emphasis_star] = STATE(422),
    [sym__strong_emphasis_star] = STATE(29),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(29),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__last_token_punctuation] = ACTIONS(732),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(16)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(42),
    [sym_full_reference_link] = STATE(42),
    [sym_collapsed_reference_link] = STATE(42),
    [sym_inline_link] = STATE(42),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(42),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(42),
    [aux_sym__inline_no_tilde] = STATE(42),
    [sym__strikethrough] = STATE(42),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(42),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(42),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__last_token_punctuation] = ACTIONS(772),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(17)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(778),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(18)] = {
    [sym_backslash_escape] = STATE(155),
    [sym_code_span] = STATE(155),
    [sym_latex_block] = STATE(155),
    [sym__link_text] = STATE(1124),
    [sym__link_text_non_empty] = STATE(469),
    [sym_shortcut_link] = STATE(529),
    [sym_full_reference_link] = STATE(529),
    [sym_collapsed_reference_link] = STATE(529),
    [sym_inline_link] = STATE(529),
    [sym_image] = STATE(155),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(155),
    [sym__whitespace] = STATE(155),
    [sym__word] = STATE(155),
    [sym__soft_line_break] = STATE(155),
    [sym__inline_base] = STATE(529),
    [sym__text_base] = STATE(155),
    [sym__inline_element] = STATE(529),
    [aux_sym__inline] = STATE(18),
    [sym__strikethrough] = STATE(529),
    [sym__emphasis_star] = STATE(534),
    [sym__strong_emphasis_star] = STATE(529),
    [sym__emphasis_underscore] = STATE(534),
    [sym__strong_emphasis_underscore] = STATE(529),
    [aux_sym__inline_base_repeat1] = STATE(155),
    [ts_builtin_sym_end] = ACTIONS(780),
    [sym__backslash_escape] = ACTIONS(782),
    [sym_entity_reference] = ACTIONS(785),
    [sym_numeric_character_reference] = ACTIONS(785),
    [anon_sym_LBRACK] = ACTIONS(788),
    [anon_sym_RBRACK] = ACTIONS(791),
    [anon_sym_LT] = ACTIONS(794),
    [anon_sym_GT] = ACTIONS(797),
    [anon_sym_BANG] = ACTIONS(800),
    [anon_sym_DQUOTE] = ACTIONS(797),
    [anon_sym_POUND] = ACTIONS(797),
    [anon_sym_DOLLAR] = ACTIONS(797),
    [anon_sym_PERCENT] = ACTIONS(797),
    [anon_sym_AMP] = ACTIONS(803),
    [anon_sym_SQUOTE] = ACTIONS(797),
    [anon_sym_STAR] = ACTIONS(797),
    [anon_sym_PLUS] = ACTIONS(797),
    [anon_sym_COMMA] = ACTIONS(797),
    [anon_sym_DASH] = ACTIONS(797),
    [anon_sym_DOT] = ACTIONS(797),
    [anon_sym_SLASH] = ACTIONS(797),
    [anon_sym_COLON] = ACTIONS(797),
    [anon_sym_SEMI] = ACTIONS(797),
    [anon_sym_EQ] = ACTIONS(797),
    [anon_sym_QMARK] = ACTIONS(797),
    [anon_sym_AT] = ACTIONS(797),
    [anon_sym_BSLASH] = ACTIONS(806),
    [anon_sym_CARET] = ACTIONS(797),
    [anon_sym__] = ACTIONS(797),
    [anon_sym_BQUOTE] = ACTIONS(797),
    [anon_sym_LBRACE] = ACTIONS(797),
    [anon_sym_PIPE] = ACTIONS(797),
    [anon_sym_RBRACE] = ACTIONS(797),
    [anon_sym_TILDE] = ACTIONS(797),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_RPAREN] = ACTIONS(797),
    [sym__newline_token] = ACTIONS(809),
    [sym_uri_autolink] = ACTIONS(785),
    [sym_email_autolink] = ACTIONS(785),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(812),
    [anon_sym_LT_QMARK] = ACTIONS(815),
    [aux_sym__declaration_token1] = ACTIONS(818),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(821),
    [sym__whitespace_ge_2] = ACTIONS(824),
    [aux_sym__whitespace_token1] = ACTIONS(827),
    [sym__word_no_digit] = ACTIONS(785),
    [sym__digits] = ACTIONS(785),
    [sym__code_span_start] = ACTIONS(830),
    [sym__emphasis_open_star] = ACTIONS(833),
    [sym__emphasis_open_underscore] = ACTIONS(836),
    [sym__strikethrough_open] = ACTIONS(839),
    [sym__latex_span_start] = ACTIONS(842),
    [sym__unclosed_span] = ACTIONS(785),
  },
  [STATE(19)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(845),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(20)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(847),
    [sym_entity_reference] = ACTIONS(850),
    [sym_numeric_character_reference] = ACTIONS(850),
    [anon_sym_LBRACK] = ACTIONS(853),
    [anon_sym_RBRACK] = ACTIONS(856),
    [anon_sym_LT] = ACTIONS(859),
    [anon_sym_GT] = ACTIONS(862),
    [anon_sym_BANG] = ACTIONS(865),
    [anon_sym_DQUOTE] = ACTIONS(862),
    [anon_sym_POUND] = ACTIONS(862),
    [anon_sym_DOLLAR] = ACTIONS(862),
    [anon_sym_PERCENT] = ACTIONS(862),
    [anon_sym_AMP] = ACTIONS(868),
    [anon_sym_SQUOTE] = ACTIONS(862),
    [anon_sym_STAR] = ACTIONS(862),
    [anon_sym_PLUS] = ACTIONS(862),
    [anon_sym_COMMA] = ACTIONS(862),
    [anon_sym_DASH] = ACTIONS(862),
    [anon_sym_DOT] = ACTIONS(862),
    [anon_sym_SLASH] = ACTIONS(862),
    [anon_sym_COLON] = ACTIONS(862),
    [anon_sym_SEMI] = ACTIONS(862),
    [anon_sym_EQ] = ACTIONS(862),
    [anon_sym_QMARK] = ACTIONS(862),
    [anon_sym_AT] = ACTIONS(862),
    [anon_sym_BSLASH] = ACTIONS(871),
    [anon_sym_CARET] = ACTIONS(862),
    [anon_sym__] = ACTIONS(862),
    [anon_sym_BQUOTE] = ACTIONS(862),
    [anon_sym_LBRACE] = ACTIONS(862),
    [anon_sym_PIPE] = ACTIONS(862),
    [anon_sym_RBRACE] = ACTIONS(862),
    [anon_sym_TILDE] = ACTIONS(862),
    [anon_sym_LPAREN] = ACTIONS(862),
    [anon_sym_RPAREN] = ACTIONS(862),
    [sym__newline_token] = ACTIONS(874),
    [sym_uri_autolink] = ACTIONS(850),
    [sym_email_autolink] = ACTIONS(850),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(877),
    [anon_sym_LT_QMARK] = ACTIONS(880),
    [aux_sym__declaration_token1] = ACTIONS(883),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(886),
    [sym__whitespace_ge_2] = ACTIONS(889),
    [aux_sym__whitespace_token1] = ACTIONS(892),
    [sym__word_no_digit] = ACTIONS(850),
    [sym__digits] = ACTIONS(850),
    [sym__code_span_start] = ACTIONS(895),
    [sym__emphasis_open_star] = ACTIONS(898),
    [sym__emphasis_open_underscore] = ACTIONS(901),
    [sym__emphasis_close_star] = ACTIONS(904),
    [sym__strikethrough_open] = ACTIONS(906),
    [sym__latex_span_start] = ACTIONS(909),
    [sym__unclosed_span] = ACTIONS(850),
  },
  [STATE(21)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(912),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(22)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(914),
    [sym_entity_reference] = ACTIONS(917),
    [sym_numeric_character_reference] = ACTIONS(917),
    [anon_sym_LBRACK] = ACTIONS(920),
    [anon_sym_RBRACK] = ACTIONS(923),
    [anon_sym_LT] = ACTIONS(926),
    [anon_sym_GT] = ACTIONS(929),
    [anon_sym_BANG] = ACTIONS(932),
    [anon_sym_DQUOTE] = ACTIONS(929),
    [anon_sym_POUND] = ACTIONS(929),
    [anon_sym_DOLLAR] = ACTIONS(929),
    [anon_sym_PERCENT] = ACTIONS(929),
    [anon_sym_AMP] = ACTIONS(935),
    [anon_sym_SQUOTE] = ACTIONS(929),
    [anon_sym_STAR] = ACTIONS(929),
    [anon_sym_PLUS] = ACTIONS(929),
    [anon_sym_COMMA] = ACTIONS(929),
    [anon_sym_DASH] = ACTIONS(929),
    [anon_sym_DOT] = ACTIONS(929),
    [anon_sym_SLASH] = ACTIONS(929),
    [anon_sym_COLON] = ACTIONS(929),
    [anon_sym_SEMI] = ACTIONS(929),
    [anon_sym_EQ] = ACTIONS(929),
    [anon_sym_QMARK] = ACTIONS(929),
    [anon_sym_AT] = ACTIONS(929),
    [anon_sym_BSLASH] = ACTIONS(938),
    [anon_sym_CARET] = ACTIONS(929),
    [anon_sym__] = ACTIONS(929),
    [anon_sym_BQUOTE] = ACTIONS(929),
    [anon_sym_LBRACE] = ACTIONS(929),
    [anon_sym_PIPE] = ACTIONS(929),
    [anon_sym_RBRACE] = ACTIONS(929),
    [anon_sym_TILDE] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(929),
    [anon_sym_RPAREN] = ACTIONS(929),
    [sym__newline_token] = ACTIONS(941),
    [sym_uri_autolink] = ACTIONS(917),
    [sym_email_autolink] = ACTIONS(917),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(944),
    [anon_sym_LT_QMARK] = ACTIONS(947),
    [aux_sym__declaration_token1] = ACTIONS(950),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(953),
    [sym__whitespace_ge_2] = ACTIONS(956),
    [aux_sym__whitespace_token1] = ACTIONS(959),
    [sym__word_no_digit] = ACTIONS(917),
    [sym__digits] = ACTIONS(917),
    [sym__code_span_start] = ACTIONS(962),
    [sym__emphasis_open_star] = ACTIONS(965),
    [sym__emphasis_open_underscore] = ACTIONS(968),
    [sym__emphasis_close_underscore] = ACTIONS(971),
    [sym__strikethrough_open] = ACTIONS(973),
    [sym__latex_span_start] = ACTIONS(976),
    [sym__unclosed_span] = ACTIONS(917),
  },
  [STATE(23)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(979),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(24)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(981),
    [sym_entity_reference] = ACTIONS(984),
    [sym_numeric_character_reference] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(987),
    [anon_sym_RBRACK] = ACTIONS(990),
    [anon_sym_LT] = ACTIONS(993),
    [anon_sym_GT] = ACTIONS(996),
    [anon_sym_BANG] = ACTIONS(999),
    [anon_sym_DQUOTE] = ACTIONS(996),
    [anon_sym_POUND] = ACTIONS(996),
    [anon_sym_DOLLAR] = ACTIONS(996),
    [anon_sym_PERCENT] = ACTIONS(996),
    [anon_sym_AMP] = ACTIONS(1002),
    [anon_sym_SQUOTE] = ACTIONS(996),
    [anon_sym_STAR] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(996),
    [anon_sym_COMMA] = ACTIONS(996),
    [anon_sym_DASH] = ACTIONS(996),
    [anon_sym_DOT] = ACTIONS(996),
    [anon_sym_SLASH] = ACTIONS(996),
    [anon_sym_COLON] = ACTIONS(996),
    [anon_sym_SEMI] = ACTIONS(996),
    [anon_sym_EQ] = ACTIONS(996),
    [anon_sym_QMARK] = ACTIONS(996),
    [anon_sym_AT] = ACTIONS(996),
    [anon_sym_BSLASH] = ACTIONS(1005),
    [anon_sym_CARET] = ACTIONS(996),
    [anon_sym__] = ACTIONS(996),
    [anon_sym_BQUOTE] = ACTIONS(996),
    [anon_sym_LBRACE] = ACTIONS(996),
    [anon_sym_PIPE] = ACTIONS(996),
    [anon_sym_RBRACE] = ACTIONS(996),
    [anon_sym_TILDE] = ACTIONS(996),
    [anon_sym_LPAREN] = ACTIONS(996),
    [anon_sym_RPAREN] = ACTIONS(996),
    [sym__newline_token] = ACTIONS(1008),
    [sym_uri_autolink] = ACTIONS(984),
    [sym_email_autolink] = ACTIONS(984),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1011),
    [anon_sym_LT_QMARK] = ACTIONS(1014),
    [aux_sym__declaration_token1] = ACTIONS(1017),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1020),
    [sym__whitespace_ge_2] = ACTIONS(1023),
    [aux_sym__whitespace_token1] = ACTIONS(1026),
    [sym__word_no_digit] = ACTIONS(984),
    [sym__digits] = ACTIONS(984),
    [sym__code_span_start] = ACTIONS(1029),
    [sym__emphasis_open_star] = ACTIONS(1032),
    [sym__emphasis_open_underscore] = ACTIONS(1035),
    [sym__strikethrough_open] = ACTIONS(1038),
    [sym__strikethrough_close] = ACTIONS(1041),
    [sym__latex_span_start] = ACTIONS(1043),
    [sym__unclosed_span] = ACTIONS(984),
  },
  [STATE(25)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(46),
    [sym_full_reference_link] = STATE(46),
    [sym_collapsed_reference_link] = STATE(46),
    [sym_inline_link] = STATE(46),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(46),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(46),
    [aux_sym__inline_no_star] = STATE(46),
    [sym__strikethrough] = STATE(46),
    [sym__emphasis_star] = STATE(491),
    [sym__strong_emphasis_star] = STATE(46),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(46),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(114),
    [sym_entity_reference] = ACTIONS(117),
    [sym_numeric_character_reference] = ACTIONS(117),
    [anon_sym_LBRACK] = ACTIONS(120),
    [anon_sym_RBRACK] = ACTIONS(123),
    [anon_sym_LT] = ACTIONS(126),
    [anon_sym_GT] = ACTIONS(129),
    [anon_sym_BANG] = ACTIONS(132),
    [anon_sym_DQUOTE] = ACTIONS(129),
    [anon_sym_POUND] = ACTIONS(129),
    [anon_sym_DOLLAR] = ACTIONS(129),
    [anon_sym_PERCENT] = ACTIONS(129),
    [anon_sym_AMP] = ACTIONS(135),
    [anon_sym_SQUOTE] = ACTIONS(129),
    [anon_sym_STAR] = ACTIONS(129),
    [anon_sym_PLUS] = ACTIONS(129),
    [anon_sym_COMMA] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(129),
    [anon_sym_DOT] = ACTIONS(129),
    [anon_sym_SLASH] = ACTIONS(129),
    [anon_sym_COLON] = ACTIONS(129),
    [anon_sym_SEMI] = ACTIONS(129),
    [anon_sym_EQ] = ACTIONS(129),
    [anon_sym_QMARK] = ACTIONS(129),
    [anon_sym_AT] = ACTIONS(129),
    [anon_sym_BSLASH] = ACTIONS(138),
    [anon_sym_CARET] = ACTIONS(129),
    [anon_sym__] = ACTIONS(129),
    [anon_sym_BQUOTE] = ACTIONS(129),
    [anon_sym_LBRACE] = ACTIONS(129),
    [anon_sym_PIPE] = ACTIONS(129),
    [anon_sym_RBRACE] = ACTIONS(129),
    [anon_sym_TILDE] = ACTIONS(129),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_RPAREN] = ACTIONS(129),
    [sym__newline_token] = ACTIONS(141),
    [sym_uri_autolink] = ACTIONS(117),
    [sym_email_autolink] = ACTIONS(117),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(144),
    [anon_sym_LT_QMARK] = ACTIONS(147),
    [aux_sym__declaration_token1] = ACTIONS(150),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(153),
    [sym__whitespace_ge_2] = ACTIONS(156),
    [aux_sym__whitespace_token1] = ACTIONS(159),
    [sym__word_no_digit] = ACTIONS(117),
    [sym__digits] = ACTIONS(117),
    [sym__code_span_start] = ACTIONS(162),
    [sym__emphasis_open_star] = ACTIONS(165),
    [sym__emphasis_open_underscore] = ACTIONS(168),
    [sym__last_token_punctuation] = ACTIONS(1046),
    [sym__strikethrough_open] = ACTIONS(173),
    [sym__latex_span_start] = ACTIONS(176),
    [sym__unclosed_span] = ACTIONS(117),
  },
  [STATE(26)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(47),
    [sym_full_reference_link] = STATE(47),
    [sym_collapsed_reference_link] = STATE(47),
    [sym_inline_link] = STATE(47),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(47),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(47),
    [aux_sym__inline_no_underscore] = STATE(47),
    [sym__strikethrough] = STATE(47),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(47),
    [sym__emphasis_underscore] = STATE(493),
    [sym__strong_emphasis_underscore] = STATE(47),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(49),
    [sym_entity_reference] = ACTIONS(52),
    [sym_numeric_character_reference] = ACTIONS(52),
    [anon_sym_LBRACK] = ACTIONS(55),
    [anon_sym_RBRACK] = ACTIONS(58),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_GT] = ACTIONS(64),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(64),
    [anon_sym_POUND] = ACTIONS(64),
    [anon_sym_DOLLAR] = ACTIONS(64),
    [anon_sym_PERCENT] = ACTIONS(64),
    [anon_sym_AMP] = ACTIONS(70),
    [anon_sym_SQUOTE] = ACTIONS(64),
    [anon_sym_STAR] = ACTIONS(64),
    [anon_sym_PLUS] = ACTIONS(64),
    [anon_sym_COMMA] = ACTIONS(64),
    [anon_sym_DASH] = ACTIONS(64),
    [anon_sym_DOT] = ACTIONS(64),
    [anon_sym_SLASH] = ACTIONS(64),
    [anon_sym_COLON] = ACTIONS(64),
    [anon_sym_SEMI] = ACTIONS(64),
    [anon_sym_EQ] = ACTIONS(64),
    [anon_sym_QMARK] = ACTIONS(64),
    [anon_sym_AT] = ACTIONS(64),
    [anon_sym_BSLASH] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(64),
    [anon_sym__] = ACTIONS(64),
    [anon_sym_BQUOTE] = ACTIONS(64),
    [anon_sym_LBRACE] = ACTIONS(64),
    [anon_sym_PIPE] = ACTIONS(64),
    [anon_sym_RBRACE] = ACTIONS(64),
    [anon_sym_TILDE] = ACTIONS(64),
    [anon_sym_LPAREN] = ACTIONS(64),
    [anon_sym_RPAREN] = ACTIONS(64),
    [sym__newline_token] = ACTIONS(76),
    [sym_uri_autolink] = ACTIONS(52),
    [sym_email_autolink] = ACTIONS(52),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(79),
    [anon_sym_LT_QMARK] = ACTIONS(82),
    [aux_sym__declaration_token1] = ACTIONS(85),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(88),
    [sym__whitespace_ge_2] = ACTIONS(91),
    [aux_sym__whitespace_token1] = ACTIONS(94),
    [sym__word_no_digit] = ACTIONS(52),
    [sym__digits] = ACTIONS(52),
    [sym__code_span_start] = ACTIONS(97),
    [sym__emphasis_open_star] = ACTIONS(100),
    [sym__emphasis_open_underscore] = ACTIONS(103),
    [sym__last_token_punctuation] = ACTIONS(1048),
    [sym__strikethrough_open] = ACTIONS(108),
    [sym__latex_span_start] = ACTIONS(111),
    [sym__unclosed_span] = ACTIONS(52),
  },
  [STATE(27)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(48),
    [sym_full_reference_link] = STATE(48),
    [sym_collapsed_reference_link] = STATE(48),
    [sym_inline_link] = STATE(48),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(48),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(48),
    [aux_sym__inline_no_tilde] = STATE(48),
    [sym__strikethrough] = STATE(48),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(48),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(48),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(311),
    [sym_entity_reference] = ACTIONS(314),
    [sym_numeric_character_reference] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(317),
    [anon_sym_RBRACK] = ACTIONS(320),
    [anon_sym_LT] = ACTIONS(323),
    [anon_sym_GT] = ACTIONS(326),
    [anon_sym_BANG] = ACTIONS(329),
    [anon_sym_DQUOTE] = ACTIONS(326),
    [anon_sym_POUND] = ACTIONS(326),
    [anon_sym_DOLLAR] = ACTIONS(326),
    [anon_sym_PERCENT] = ACTIONS(326),
    [anon_sym_AMP] = ACTIONS(332),
    [anon_sym_SQUOTE] = ACTIONS(326),
    [anon_sym_STAR] = ACTIONS(326),
    [anon_sym_PLUS] = ACTIONS(326),
    [anon_sym_COMMA] = ACTIONS(326),
    [anon_sym_DASH] = ACTIONS(326),
    [anon_sym_DOT] = ACTIONS(326),
    [anon_sym_SLASH] = ACTIONS(326),
    [anon_sym_COLON] = ACTIONS(326),
    [anon_sym_SEMI] = ACTIONS(326),
    [anon_sym_EQ] = ACTIONS(326),
    [anon_sym_QMARK] = ACTIONS(326),
    [anon_sym_AT] = ACTIONS(326),
    [anon_sym_BSLASH] = ACTIONS(335),
    [anon_sym_CARET] = ACTIONS(326),
    [anon_sym__] = ACTIONS(326),
    [anon_sym_BQUOTE] = ACTIONS(326),
    [anon_sym_LBRACE] = ACTIONS(326),
    [anon_sym_PIPE] = ACTIONS(326),
    [anon_sym_RBRACE] = ACTIONS(326),
    [anon_sym_TILDE] = ACTIONS(326),
    [anon_sym_LPAREN] = ACTIONS(326),
    [anon_sym_RPAREN] = ACTIONS(326),
    [sym__newline_token] = ACTIONS(338),
    [sym_uri_autolink] = ACTIONS(314),
    [sym_email_autolink] = ACTIONS(314),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(341),
    [anon_sym_LT_QMARK] = ACTIONS(344),
    [aux_sym__declaration_token1] = ACTIONS(347),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(350),
    [sym__whitespace_ge_2] = ACTIONS(353),
    [aux_sym__whitespace_token1] = ACTIONS(356),
    [sym__word_no_digit] = ACTIONS(314),
    [sym__digits] = ACTIONS(314),
    [sym__code_span_start] = ACTIONS(359),
    [sym__emphasis_open_star] = ACTIONS(362),
    [sym__emphasis_open_underscore] = ACTIONS(365),
    [sym__last_token_punctuation] = ACTIONS(1050),
    [sym__strikethrough_open] = ACTIONS(370),
    [sym__latex_span_start] = ACTIONS(373),
    [sym__unclosed_span] = ACTIONS(314),
  },
  [STATE(28)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1052),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(29)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1054),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(30)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1056),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(31)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1058),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(32)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1060),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(33)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1062),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(34)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1064),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(35)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1066),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(36)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1068),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(37)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1070),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(38)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1072),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(39)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1074),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(40)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1076),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(41)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1078),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(42)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1080),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(43)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1082),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(44)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1084),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(45)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1086),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(46)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1088),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(47)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1090),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(48)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1092),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(49)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(20),
    [sym_full_reference_link] = STATE(20),
    [sym_collapsed_reference_link] = STATE(20),
    [sym_inline_link] = STATE(20),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(20),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(20),
    [aux_sym__inline_no_star] = STATE(20),
    [sym__strikethrough] = STATE(20),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(20),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(20),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__emphasis_close_star] = ACTIONS(1094),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(50)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1096),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(51)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(24),
    [sym_full_reference_link] = STATE(24),
    [sym_collapsed_reference_link] = STATE(24),
    [sym_inline_link] = STATE(24),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(24),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(24),
    [aux_sym__inline_no_tilde] = STATE(24),
    [sym__strikethrough] = STATE(24),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(24),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(24),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__strikethrough_close] = ACTIONS(1098),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(52)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(22),
    [sym_full_reference_link] = STATE(22),
    [sym_collapsed_reference_link] = STATE(22),
    [sym_inline_link] = STATE(22),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(22),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(22),
    [aux_sym__inline_no_underscore] = STATE(22),
    [sym__strikethrough] = STATE(22),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(22),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(22),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__emphasis_close_underscore] = ACTIONS(1100),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(53)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(33),
    [sym_full_reference_link] = STATE(33),
    [sym_collapsed_reference_link] = STATE(33),
    [sym_inline_link] = STATE(33),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(33),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(33),
    [aux_sym__inline_no_underscore] = STATE(33),
    [sym__strikethrough] = STATE(33),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(33),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(33),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(54)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(67),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(55)] = {
    [sym_backslash_escape] = STATE(155),
    [sym_code_span] = STATE(155),
    [sym_latex_block] = STATE(155),
    [sym__link_text] = STATE(1124),
    [sym__link_text_non_empty] = STATE(469),
    [sym_shortcut_link] = STATE(529),
    [sym_full_reference_link] = STATE(529),
    [sym_collapsed_reference_link] = STATE(529),
    [sym_inline_link] = STATE(529),
    [sym_image] = STATE(155),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(155),
    [sym__whitespace] = STATE(155),
    [sym__word] = STATE(155),
    [sym__soft_line_break] = STATE(155),
    [sym__inline_base] = STATE(529),
    [sym__text_base] = STATE(155),
    [sym__inline_element] = STATE(529),
    [aux_sym__inline] = STATE(13),
    [sym__strikethrough] = STATE(529),
    [sym__emphasis_star] = STATE(534),
    [sym__strong_emphasis_star] = STATE(529),
    [sym__emphasis_underscore] = STATE(534),
    [sym__strong_emphasis_underscore] = STATE(529),
    [aux_sym__inline_base_repeat1] = STATE(155),
    [sym__backslash_escape] = ACTIONS(3),
    [sym_entity_reference] = ACTIONS(5),
    [sym_numeric_character_reference] = ACTIONS(5),
    [anon_sym_LBRACK] = ACTIONS(7),
    [anon_sym_RBRACK] = ACTIONS(9),
    [anon_sym_LT] = ACTIONS(11),
    [anon_sym_GT] = ACTIONS(13),
    [anon_sym_BANG] = ACTIONS(15),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_POUND] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(13),
    [anon_sym_PERCENT] = ACTIONS(13),
    [anon_sym_AMP] = ACTIONS(17),
    [anon_sym_SQUOTE] = ACTIONS(13),
    [anon_sym_STAR] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(13),
    [anon_sym_COMMA] = ACTIONS(13),
    [anon_sym_DASH] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(13),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_COLON] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(13),
    [anon_sym_QMARK] = ACTIONS(13),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_BSLASH] = ACTIONS(19),
    [anon_sym_CARET] = ACTIONS(13),
    [anon_sym__] = ACTIONS(13),
    [anon_sym_BQUOTE] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_PIPE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(13),
    [anon_sym_TILDE] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(13),
    [sym__newline_token] = ACTIONS(21),
    [sym_uri_autolink] = ACTIONS(5),
    [sym_email_autolink] = ACTIONS(5),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(23),
    [anon_sym_LT_QMARK] = ACTIONS(25),
    [aux_sym__declaration_token1] = ACTIONS(27),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(29),
    [sym__whitespace_ge_2] = ACTIONS(31),
    [aux_sym__whitespace_token1] = ACTIONS(33),
    [sym__word_no_digit] = ACTIONS(5),
    [sym__digits] = ACTIONS(5),
    [sym__code_span_start] = ACTIONS(35),
    [sym__emphasis_open_star] = ACTIONS(37),
    [sym__emphasis_open_underscore] = ACTIONS(39),
    [sym__strikethrough_open] = ACTIONS(43),
    [sym__latex_span_start] = ACTIONS(45),
    [sym__unclosed_span] = ACTIONS(5),
  },
  [STATE(56)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(23),
    [sym_full_reference_link] = STATE(23),
    [sym_collapsed_reference_link] = STATE(23),
    [sym_inline_link] = STATE(23),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(23),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(23),
    [aux_sym__inline_no_tilde] = STATE(23),
    [sym__strikethrough] = STATE(23),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(23),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(23),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(57)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1144),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(58)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1146),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(59)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(21),
    [sym_full_reference_link] = STATE(21),
    [sym_collapsed_reference_link] = STATE(21),
    [sym_inline_link] = STATE(21),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(21),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(21),
    [aux_sym__inline_no_underscore] = STATE(21),
    [sym__strikethrough] = STATE(21),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(21),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(21),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(60)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(19),
    [sym_full_reference_link] = STATE(19),
    [sym_collapsed_reference_link] = STATE(19),
    [sym_inline_link] = STATE(19),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(19),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(19),
    [aux_sym__inline_no_star] = STATE(19),
    [sym__strikethrough] = STATE(19),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(19),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(19),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(61)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1148),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(62)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(58),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(63)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1150),
    [sym_entity_reference] = ACTIONS(1153),
    [sym_numeric_character_reference] = ACTIONS(1153),
    [anon_sym_LBRACK] = ACTIONS(1156),
    [anon_sym_RBRACK] = ACTIONS(1159),
    [anon_sym_LT] = ACTIONS(1162),
    [anon_sym_GT] = ACTIONS(1165),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1165),
    [anon_sym_POUND] = ACTIONS(1165),
    [anon_sym_DOLLAR] = ACTIONS(1165),
    [anon_sym_PERCENT] = ACTIONS(1165),
    [anon_sym_AMP] = ACTIONS(1171),
    [anon_sym_SQUOTE] = ACTIONS(1165),
    [anon_sym_STAR] = ACTIONS(1165),
    [anon_sym_PLUS] = ACTIONS(1165),
    [anon_sym_COMMA] = ACTIONS(1165),
    [anon_sym_DASH] = ACTIONS(1165),
    [anon_sym_DOT] = ACTIONS(1165),
    [anon_sym_SLASH] = ACTIONS(1165),
    [anon_sym_COLON] = ACTIONS(1165),
    [anon_sym_SEMI] = ACTIONS(1165),
    [anon_sym_EQ] = ACTIONS(1165),
    [anon_sym_QMARK] = ACTIONS(1165),
    [anon_sym_AT] = ACTIONS(1165),
    [anon_sym_BSLASH] = ACTIONS(1174),
    [anon_sym_CARET] = ACTIONS(1165),
    [anon_sym__] = ACTIONS(1165),
    [anon_sym_BQUOTE] = ACTIONS(1165),
    [anon_sym_LBRACE] = ACTIONS(1165),
    [anon_sym_PIPE] = ACTIONS(1165),
    [anon_sym_RBRACE] = ACTIONS(1165),
    [anon_sym_TILDE] = ACTIONS(1165),
    [anon_sym_LPAREN] = ACTIONS(1165),
    [anon_sym_RPAREN] = ACTIONS(1165),
    [sym__newline_token] = ACTIONS(1177),
    [sym_uri_autolink] = ACTIONS(1153),
    [sym_email_autolink] = ACTIONS(1153),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1180),
    [anon_sym_LT_QMARK] = ACTIONS(1183),
    [aux_sym__declaration_token1] = ACTIONS(1186),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1189),
    [sym__whitespace_ge_2] = ACTIONS(1192),
    [aux_sym__whitespace_token1] = ACTIONS(1195),
    [sym__word_no_digit] = ACTIONS(1153),
    [sym__digits] = ACTIONS(1153),
    [sym__code_span_start] = ACTIONS(1198),
    [sym__emphasis_open_star] = ACTIONS(1201),
    [sym__emphasis_open_underscore] = ACTIONS(1204),
    [sym__strikethrough_open] = ACTIONS(1207),
    [sym__latex_span_start] = ACTIONS(1210),
    [sym__unclosed_span] = ACTIONS(1153),
  },
  [STATE(64)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1213),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(65)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1215),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(66)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1217),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(67)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1219),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(68)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(32),
    [sym_full_reference_link] = STATE(32),
    [sym_collapsed_reference_link] = STATE(32),
    [sym_inline_link] = STATE(32),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(32),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(32),
    [aux_sym__inline_no_star] = STATE(32),
    [sym__strikethrough] = STATE(32),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(32),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(32),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(69)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(63),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1221),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(70)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(34),
    [sym_full_reference_link] = STATE(34),
    [sym_collapsed_reference_link] = STATE(34),
    [sym_inline_link] = STATE(34),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(34),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(34),
    [aux_sym__inline_no_tilde] = STATE(34),
    [sym__strikethrough] = STATE(34),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(34),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(34),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(71)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(69),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(72)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(37),
    [sym_full_reference_link] = STATE(37),
    [sym_collapsed_reference_link] = STATE(37),
    [sym_inline_link] = STATE(37),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(37),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(37),
    [aux_sym__inline_no_star] = STATE(37),
    [sym__strikethrough] = STATE(37),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(37),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(37),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(73)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(38),
    [sym_full_reference_link] = STATE(38),
    [sym_collapsed_reference_link] = STATE(38),
    [sym_inline_link] = STATE(38),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(38),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(38),
    [aux_sym__inline_no_underscore] = STATE(38),
    [sym__strikethrough] = STATE(38),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(38),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(38),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(74)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(39),
    [sym_full_reference_link] = STATE(39),
    [sym_collapsed_reference_link] = STATE(39),
    [sym_inline_link] = STATE(39),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(39),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(39),
    [aux_sym__inline_no_tilde] = STATE(39),
    [sym__strikethrough] = STATE(39),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(39),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(39),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(75)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(61),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(76)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(43),
    [sym_full_reference_link] = STATE(43),
    [sym_collapsed_reference_link] = STATE(43),
    [sym_inline_link] = STATE(43),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(43),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(43),
    [aux_sym__inline_no_star] = STATE(43),
    [sym__strikethrough] = STATE(43),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(43),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(43),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(77)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(44),
    [sym_full_reference_link] = STATE(44),
    [sym_collapsed_reference_link] = STATE(44),
    [sym_inline_link] = STATE(44),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(44),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(44),
    [aux_sym__inline_no_underscore] = STATE(44),
    [sym__strikethrough] = STATE(44),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(44),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(44),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(78)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(45),
    [sym_full_reference_link] = STATE(45),
    [sym_collapsed_reference_link] = STATE(45),
    [sym_inline_link] = STATE(45),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(45),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(45),
    [aux_sym__inline_no_tilde] = STATE(45),
    [sym__strikethrough] = STATE(45),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(45),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(45),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(79)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(57),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(80)] = {
    [sym_backslash_escape] = STATE(148),
    [sym_code_span] = STATE(148),
    [sym_latex_block] = STATE(148),
    [sym__link_text] = STATE(1120),
    [sym__link_text_non_empty] = STATE(367),
    [sym_shortcut_link] = STATE(49),
    [sym_full_reference_link] = STATE(49),
    [sym_collapsed_reference_link] = STATE(49),
    [sym_inline_link] = STATE(49),
    [sym_image] = STATE(148),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(148),
    [sym__whitespace] = STATE(148),
    [sym__word] = STATE(148),
    [sym__soft_line_break] = STATE(148),
    [sym__inline_base] = STATE(49),
    [sym__text_base] = STATE(148),
    [sym__inline_element_no_star] = STATE(49),
    [aux_sym__inline_no_star] = STATE(49),
    [sym__strikethrough] = STATE(49),
    [sym__emphasis_star] = STATE(425),
    [sym__strong_emphasis_star] = STATE(49),
    [sym__emphasis_underscore] = STATE(425),
    [sym__strong_emphasis_underscore] = STATE(49),
    [aux_sym__inline_base_repeat1] = STATE(148),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(642),
    [sym_numeric_character_reference] = ACTIONS(642),
    [anon_sym_LBRACK] = ACTIONS(644),
    [anon_sym_RBRACK] = ACTIONS(646),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(642),
    [sym_email_autolink] = ACTIONS(642),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(642),
    [sym__digits] = ACTIONS(642),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(674),
    [sym__emphasis_open_underscore] = ACTIONS(676),
    [sym__strikethrough_open] = ACTIONS(680),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(642),
  },
  [STATE(81)] = {
    [sym_backslash_escape] = STATE(150),
    [sym_code_span] = STATE(150),
    [sym_latex_block] = STATE(150),
    [sym__link_text] = STATE(1117),
    [sym__link_text_non_empty] = STATE(449),
    [sym_shortcut_link] = STATE(50),
    [sym_full_reference_link] = STATE(50),
    [sym_collapsed_reference_link] = STATE(50),
    [sym_inline_link] = STATE(50),
    [sym_image] = STATE(150),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(150),
    [sym__whitespace] = STATE(150),
    [sym__word] = STATE(150),
    [sym__soft_line_break] = STATE(150),
    [sym__inline_base] = STATE(50),
    [sym__text_base] = STATE(150),
    [sym__inline_element_no_underscore] = STATE(50),
    [aux_sym__inline_no_underscore] = STATE(50),
    [sym__strikethrough] = STATE(50),
    [sym__emphasis_star] = STATE(453),
    [sym__strong_emphasis_star] = STATE(50),
    [sym__emphasis_underscore] = STATE(453),
    [sym__strong_emphasis_underscore] = STATE(50),
    [aux_sym__inline_base_repeat1] = STATE(150),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(686),
    [sym_numeric_character_reference] = ACTIONS(686),
    [anon_sym_LBRACK] = ACTIONS(688),
    [anon_sym_RBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(686),
    [sym_email_autolink] = ACTIONS(686),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(686),
    [sym__digits] = ACTIONS(686),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(718),
    [sym__emphasis_open_underscore] = ACTIONS(720),
    [sym__strikethrough_open] = ACTIONS(724),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(686),
  },
  [STATE(82)] = {
    [sym_backslash_escape] = STATE(154),
    [sym_code_span] = STATE(154),
    [sym_latex_block] = STATE(154),
    [sym__link_text] = STATE(1121),
    [sym__link_text_non_empty] = STATE(288),
    [sym_shortcut_link] = STATE(51),
    [sym_full_reference_link] = STATE(51),
    [sym_collapsed_reference_link] = STATE(51),
    [sym_inline_link] = STATE(51),
    [sym_image] = STATE(154),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(154),
    [sym__whitespace] = STATE(154),
    [sym__word] = STATE(154),
    [sym__soft_line_break] = STATE(154),
    [sym__inline_base] = STATE(51),
    [sym__text_base] = STATE(154),
    [sym__inline_element_no_tilde] = STATE(51),
    [aux_sym__inline_no_tilde] = STATE(51),
    [sym__strikethrough] = STATE(51),
    [sym__emphasis_star] = STATE(482),
    [sym__strong_emphasis_star] = STATE(51),
    [sym__emphasis_underscore] = STATE(482),
    [sym__strong_emphasis_underscore] = STATE(51),
    [aux_sym__inline_base_repeat1] = STATE(154),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(736),
    [sym_numeric_character_reference] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_RBRACK] = ACTIONS(740),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(736),
    [sym_email_autolink] = ACTIONS(736),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(736),
    [sym__digits] = ACTIONS(736),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(768),
    [sym__emphasis_open_underscore] = ACTIONS(770),
    [sym__strikethrough_open] = ACTIONS(774),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(736),
  },
  [STATE(83)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(64),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(84)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(65),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(85)] = {
    [sym_backslash_escape] = STATE(156),
    [sym_code_span] = STATE(156),
    [sym_latex_block] = STATE(156),
    [sym__link_text] = STATE(1116),
    [sym__link_text_non_empty] = STATE(540),
    [sym_shortcut_link] = STATE(537),
    [sym_full_reference_link] = STATE(537),
    [sym_collapsed_reference_link] = STATE(537),
    [sym_inline_link] = STATE(537),
    [sym_image] = STATE(156),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(156),
    [sym__whitespace] = STATE(156),
    [sym__word] = STATE(156),
    [sym__soft_line_break] = STATE(156),
    [sym__inline_base] = STATE(537),
    [sym__text_base] = STATE(156),
    [sym__inline_element] = STATE(537),
    [aux_sym__inline] = STATE(66),
    [sym__strikethrough] = STATE(537),
    [sym__emphasis_star] = STATE(538),
    [sym__strong_emphasis_star] = STATE(537),
    [sym__emphasis_underscore] = STATE(538),
    [sym__strong_emphasis_underscore] = STATE(537),
    [aux_sym__inline_base_repeat1] = STATE(156),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(1104),
    [sym_numeric_character_reference] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(1108),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(1104),
    [sym_email_autolink] = ACTIONS(1104),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(1104),
    [sym__digits] = ACTIONS(1104),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(1136),
    [sym__emphasis_open_underscore] = ACTIONS(1138),
    [sym__strikethrough_open] = ACTIONS(1140),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(1104),
  },
  [STATE(86)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(110),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1223),
    [sym_entity_reference] = ACTIONS(1226),
    [sym_numeric_character_reference] = ACTIONS(1226),
    [anon_sym_LBRACK] = ACTIONS(236),
    [anon_sym_RBRACK] = ACTIONS(1229),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1235),
    [anon_sym_BANG] = ACTIONS(1238),
    [anon_sym_DQUOTE] = ACTIONS(1235),
    [anon_sym_POUND] = ACTIONS(1235),
    [anon_sym_DOLLAR] = ACTIONS(1235),
    [anon_sym_PERCENT] = ACTIONS(1235),
    [anon_sym_AMP] = ACTIONS(1241),
    [anon_sym_SQUOTE] = ACTIONS(1235),
    [anon_sym_STAR] = ACTIONS(1235),
    [anon_sym_PLUS] = ACTIONS(1235),
    [anon_sym_COMMA] = ACTIONS(1235),
    [anon_sym_DASH] = ACTIONS(1235),
    [anon_sym_DOT] = ACTIONS(1235),
    [anon_sym_SLASH] = ACTIONS(1235),
    [anon_sym_COLON] = ACTIONS(1235),
    [anon_sym_SEMI] = ACTIONS(1235),
    [anon_sym_EQ] = ACTIONS(1235),
    [anon_sym_QMARK] = ACTIONS(1235),
    [anon_sym_AT] = ACTIONS(1235),
    [anon_sym_BSLASH] = ACTIONS(1244),
    [anon_sym_CARET] = ACTIONS(1235),
    [anon_sym__] = ACTIONS(1235),
    [anon_sym_BQUOTE] = ACTIONS(1235),
    [anon_sym_LBRACE] = ACTIONS(1235),
    [anon_sym_PIPE] = ACTIONS(1235),
    [anon_sym_RBRACE] = ACTIONS(1235),
    [anon_sym_TILDE] = ACTIONS(1235),
    [anon_sym_LPAREN] = ACTIONS(1235),
    [anon_sym_RPAREN] = ACTIONS(1235),
    [sym__newline_token] = ACTIONS(1247),
    [sym_uri_autolink] = ACTIONS(1226),
    [sym_email_autolink] = ACTIONS(1226),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1250),
    [anon_sym_LT_QMARK] = ACTIONS(1253),
    [aux_sym__declaration_token1] = ACTIONS(1256),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1259),
    [sym__whitespace_ge_2] = ACTIONS(1262),
    [aux_sym__whitespace_token1] = ACTIONS(1265),
    [sym__word_no_digit] = ACTIONS(1226),
    [sym__digits] = ACTIONS(1226),
    [sym__code_span_start] = ACTIONS(1268),
    [sym__emphasis_open_star] = ACTIONS(1271),
    [sym__emphasis_open_underscore] = ACTIONS(1274),
    [sym__emphasis_close_star] = ACTIONS(236),
    [sym__last_token_punctuation] = ACTIONS(1277),
    [sym__strikethrough_open] = ACTIONS(1279),
    [sym__latex_span_start] = ACTIONS(1282),
    [sym__unclosed_span] = ACTIONS(1226),
  },
  [STATE(87)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(125),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1285),
    [sym_entity_reference] = ACTIONS(1288),
    [sym_numeric_character_reference] = ACTIONS(1288),
    [anon_sym_LBRACK] = ACTIONS(570),
    [anon_sym_RBRACK] = ACTIONS(1291),
    [anon_sym_LT] = ACTIONS(1294),
    [anon_sym_GT] = ACTIONS(1297),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(1297),
    [anon_sym_DOLLAR] = ACTIONS(1297),
    [anon_sym_PERCENT] = ACTIONS(1297),
    [anon_sym_AMP] = ACTIONS(1303),
    [anon_sym_SQUOTE] = ACTIONS(1297),
    [anon_sym_STAR] = ACTIONS(1297),
    [anon_sym_PLUS] = ACTIONS(1297),
    [anon_sym_COMMA] = ACTIONS(1297),
    [anon_sym_DASH] = ACTIONS(1297),
    [anon_sym_DOT] = ACTIONS(1297),
    [anon_sym_SLASH] = ACTIONS(1297),
    [anon_sym_COLON] = ACTIONS(1297),
    [anon_sym_SEMI] = ACTIONS(1297),
    [anon_sym_EQ] = ACTIONS(1297),
    [anon_sym_QMARK] = ACTIONS(1297),
    [anon_sym_AT] = ACTIONS(1297),
    [anon_sym_BSLASH] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1297),
    [anon_sym__] = ACTIONS(1297),
    [anon_sym_BQUOTE] = ACTIONS(1297),
    [anon_sym_LBRACE] = ACTIONS(1297),
    [anon_sym_PIPE] = ACTIONS(1297),
    [anon_sym_RBRACE] = ACTIONS(1297),
    [anon_sym_TILDE] = ACTIONS(1297),
    [anon_sym_LPAREN] = ACTIONS(1297),
    [anon_sym_RPAREN] = ACTIONS(1297),
    [sym__newline_token] = ACTIONS(1309),
    [sym_uri_autolink] = ACTIONS(1288),
    [sym_email_autolink] = ACTIONS(1288),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1312),
    [anon_sym_LT_QMARK] = ACTIONS(1315),
    [aux_sym__declaration_token1] = ACTIONS(1318),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1321),
    [sym__whitespace_ge_2] = ACTIONS(1324),
    [aux_sym__whitespace_token1] = ACTIONS(1327),
    [sym__word_no_digit] = ACTIONS(1288),
    [sym__digits] = ACTIONS(1288),
    [sym__code_span_start] = ACTIONS(1330),
    [sym__emphasis_open_star] = ACTIONS(1333),
    [sym__emphasis_open_underscore] = ACTIONS(1336),
    [sym__last_token_punctuation] = ACTIONS(1339),
    [sym__strikethrough_open] = ACTIONS(1341),
    [sym__strikethrough_close] = ACTIONS(570),
    [sym__latex_span_start] = ACTIONS(1344),
    [sym__unclosed_span] = ACTIONS(1288),
  },
  [STATE(88)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(134),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [ts_builtin_sym_end] = ACTIONS(47),
    [sym__backslash_escape] = ACTIONS(1347),
    [sym_entity_reference] = ACTIONS(1350),
    [sym_numeric_character_reference] = ACTIONS(1350),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_RBRACK] = ACTIONS(1353),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1359),
    [anon_sym_POUND] = ACTIONS(1359),
    [anon_sym_DOLLAR] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1365),
    [anon_sym_SQUOTE] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_COLON] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_EQ] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_BSLASH] = ACTIONS(1368),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym__] = ACTIONS(1359),
    [anon_sym_BQUOTE] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_RBRACE] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_RPAREN] = ACTIONS(1359),
    [sym__newline_token] = ACTIONS(1371),
    [sym_uri_autolink] = ACTIONS(1350),
    [sym_email_autolink] = ACTIONS(1350),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1374),
    [anon_sym_LT_QMARK] = ACTIONS(1377),
    [aux_sym__declaration_token1] = ACTIONS(1380),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1383),
    [sym__whitespace_ge_2] = ACTIONS(1386),
    [aux_sym__whitespace_token1] = ACTIONS(1389),
    [sym__word_no_digit] = ACTIONS(1350),
    [sym__digits] = ACTIONS(1350),
    [sym__code_span_start] = ACTIONS(1392),
    [sym__emphasis_open_star] = ACTIONS(1395),
    [sym__emphasis_open_underscore] = ACTIONS(1398),
    [sym__last_token_punctuation] = ACTIONS(1401),
    [sym__strikethrough_open] = ACTIONS(1403),
    [sym__latex_span_start] = ACTIONS(1406),
    [sym__unclosed_span] = ACTIONS(1350),
  },
  [STATE(89)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(118),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1409),
    [sym_entity_reference] = ACTIONS(1412),
    [sym_numeric_character_reference] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(433),
    [anon_sym_RBRACK] = ACTIONS(1415),
    [anon_sym_LT] = ACTIONS(1418),
    [anon_sym_GT] = ACTIONS(1421),
    [anon_sym_BANG] = ACTIONS(1424),
    [anon_sym_DQUOTE] = ACTIONS(1421),
    [anon_sym_POUND] = ACTIONS(1421),
    [anon_sym_DOLLAR] = ACTIONS(1421),
    [anon_sym_PERCENT] = ACTIONS(1421),
    [anon_sym_AMP] = ACTIONS(1427),
    [anon_sym_SQUOTE] = ACTIONS(1421),
    [anon_sym_STAR] = ACTIONS(1421),
    [anon_sym_PLUS] = ACTIONS(1421),
    [anon_sym_COMMA] = ACTIONS(1421),
    [anon_sym_DASH] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1421),
    [anon_sym_SLASH] = ACTIONS(1421),
    [anon_sym_COLON] = ACTIONS(1421),
    [anon_sym_SEMI] = ACTIONS(1421),
    [anon_sym_EQ] = ACTIONS(1421),
    [anon_sym_QMARK] = ACTIONS(1421),
    [anon_sym_AT] = ACTIONS(1421),
    [anon_sym_BSLASH] = ACTIONS(1430),
    [anon_sym_CARET] = ACTIONS(1421),
    [anon_sym__] = ACTIONS(1421),
    [anon_sym_BQUOTE] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1421),
    [anon_sym_PIPE] = ACTIONS(1421),
    [anon_sym_RBRACE] = ACTIONS(1421),
    [anon_sym_TILDE] = ACTIONS(1421),
    [anon_sym_LPAREN] = ACTIONS(1421),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [sym__newline_token] = ACTIONS(1433),
    [sym_uri_autolink] = ACTIONS(1412),
    [sym_email_autolink] = ACTIONS(1412),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1436),
    [anon_sym_LT_QMARK] = ACTIONS(1439),
    [aux_sym__declaration_token1] = ACTIONS(1442),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1445),
    [sym__whitespace_ge_2] = ACTIONS(1448),
    [aux_sym__whitespace_token1] = ACTIONS(1451),
    [sym__word_no_digit] = ACTIONS(1412),
    [sym__digits] = ACTIONS(1412),
    [sym__code_span_start] = ACTIONS(1454),
    [sym__emphasis_open_star] = ACTIONS(1457),
    [sym__emphasis_open_underscore] = ACTIONS(1460),
    [sym__emphasis_close_underscore] = ACTIONS(433),
    [sym__last_token_punctuation] = ACTIONS(1463),
    [sym__strikethrough_open] = ACTIONS(1465),
    [sym__latex_span_start] = ACTIONS(1468),
    [sym__unclosed_span] = ACTIONS(1412),
  },
  [STATE(90)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(132),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1347),
    [sym_entity_reference] = ACTIONS(1350),
    [sym_numeric_character_reference] = ACTIONS(1350),
    [anon_sym_LBRACK] = ACTIONS(47),
    [anon_sym_RBRACK] = ACTIONS(1353),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1359),
    [anon_sym_POUND] = ACTIONS(1359),
    [anon_sym_DOLLAR] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1365),
    [anon_sym_SQUOTE] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_COLON] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_EQ] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_BSLASH] = ACTIONS(1368),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym__] = ACTIONS(1359),
    [anon_sym_BQUOTE] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_RBRACE] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_RPAREN] = ACTIONS(1359),
    [sym__newline_token] = ACTIONS(1371),
    [sym_uri_autolink] = ACTIONS(1350),
    [sym_email_autolink] = ACTIONS(1350),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1374),
    [anon_sym_LT_QMARK] = ACTIONS(1377),
    [aux_sym__declaration_token1] = ACTIONS(1380),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1383),
    [sym__whitespace_ge_2] = ACTIONS(1386),
    [aux_sym__whitespace_token1] = ACTIONS(1389),
    [sym__word_no_digit] = ACTIONS(1350),
    [sym__digits] = ACTIONS(1350),
    [sym__code_span_start] = ACTIONS(1392),
    [sym__emphasis_open_star] = ACTIONS(1395),
    [sym__emphasis_open_underscore] = ACTIONS(1398),
    [sym__last_token_punctuation] = ACTIONS(1471),
    [sym__strikethrough_open] = ACTIONS(1403),
    [sym__latex_span_start] = ACTIONS(1406),
    [sym__unclosed_span] = ACTIONS(1350),
  },
  [STATE(91)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(101),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(101),
    [aux_sym__inline_no_tilde_no_link] = STATE(101),
    [sym__strikethrough_no_link] = STATE(101),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(101),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(101),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(1473),
    [sym_entity_reference] = ACTIONS(1476),
    [sym_numeric_character_reference] = ACTIONS(1476),
    [anon_sym_RBRACK] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1484),
    [anon_sym_BANG] = ACTIONS(1487),
    [anon_sym_DQUOTE] = ACTIONS(1484),
    [anon_sym_POUND] = ACTIONS(1484),
    [anon_sym_DOLLAR] = ACTIONS(1484),
    [anon_sym_PERCENT] = ACTIONS(1484),
    [anon_sym_AMP] = ACTIONS(1490),
    [anon_sym_SQUOTE] = ACTIONS(1484),
    [anon_sym_STAR] = ACTIONS(1484),
    [anon_sym_PLUS] = ACTIONS(1484),
    [anon_sym_COMMA] = ACTIONS(1484),
    [anon_sym_DASH] = ACTIONS(1484),
    [anon_sym_DOT] = ACTIONS(1484),
    [anon_sym_SLASH] = ACTIONS(1484),
    [anon_sym_COLON] = ACTIONS(1484),
    [anon_sym_SEMI] = ACTIONS(1484),
    [anon_sym_EQ] = ACTIONS(1484),
    [anon_sym_QMARK] = ACTIONS(1484),
    [anon_sym_AT] = ACTIONS(1484),
    [anon_sym_BSLASH] = ACTIONS(1493),
    [anon_sym_CARET] = ACTIONS(1484),
    [anon_sym__] = ACTIONS(1484),
    [anon_sym_BQUOTE] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(1484),
    [anon_sym_PIPE] = ACTIONS(1484),
    [anon_sym_RBRACE] = ACTIONS(1484),
    [anon_sym_TILDE] = ACTIONS(1484),
    [anon_sym_LPAREN] = ACTIONS(1484),
    [anon_sym_RPAREN] = ACTIONS(1484),
    [sym__newline_token] = ACTIONS(1496),
    [sym_uri_autolink] = ACTIONS(1476),
    [sym_email_autolink] = ACTIONS(1476),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1499),
    [anon_sym_LT_QMARK] = ACTIONS(1502),
    [aux_sym__declaration_token1] = ACTIONS(1505),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1508),
    [sym__whitespace_ge_2] = ACTIONS(1511),
    [aux_sym__whitespace_token1] = ACTIONS(1514),
    [sym__word_no_digit] = ACTIONS(1476),
    [sym__digits] = ACTIONS(1476),
    [sym__code_span_start] = ACTIONS(1517),
    [sym__emphasis_open_star] = ACTIONS(1520),
    [sym__emphasis_open_underscore] = ACTIONS(1523),
    [sym__last_token_punctuation] = ACTIONS(1526),
    [sym__strikethrough_open] = ACTIONS(1528),
    [sym__latex_span_start] = ACTIONS(1531),
    [sym__unclosed_span] = ACTIONS(1476),
  },
  [STATE(92)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(126),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(126),
    [aux_sym__inline_no_star_no_link] = STATE(126),
    [sym__strikethrough_no_link] = STATE(126),
    [sym__emphasis_star_no_link] = STATE(639),
    [sym__strong_emphasis_star_no_link] = STATE(126),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(126),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(1534),
    [sym_entity_reference] = ACTIONS(1537),
    [sym_numeric_character_reference] = ACTIONS(1537),
    [anon_sym_LT] = ACTIONS(1540),
    [anon_sym_GT] = ACTIONS(1543),
    [anon_sym_BANG] = ACTIONS(1546),
    [anon_sym_DQUOTE] = ACTIONS(1543),
    [anon_sym_POUND] = ACTIONS(1543),
    [anon_sym_DOLLAR] = ACTIONS(1543),
    [anon_sym_PERCENT] = ACTIONS(1543),
    [anon_sym_AMP] = ACTIONS(1549),
    [anon_sym_SQUOTE] = ACTIONS(1543),
    [anon_sym_STAR] = ACTIONS(1543),
    [anon_sym_PLUS] = ACTIONS(1543),
    [anon_sym_COMMA] = ACTIONS(1543),
    [anon_sym_DASH] = ACTIONS(1543),
    [anon_sym_DOT] = ACTIONS(1543),
    [anon_sym_SLASH] = ACTIONS(1543),
    [anon_sym_COLON] = ACTIONS(1543),
    [anon_sym_SEMI] = ACTIONS(1543),
    [anon_sym_EQ] = ACTIONS(1543),
    [anon_sym_QMARK] = ACTIONS(1543),
    [anon_sym_AT] = ACTIONS(1543),
    [anon_sym_BSLASH] = ACTIONS(1552),
    [anon_sym_CARET] = ACTIONS(1543),
    [anon_sym__] = ACTIONS(1543),
    [anon_sym_BQUOTE] = ACTIONS(1543),
    [anon_sym_LBRACE] = ACTIONS(1543),
    [anon_sym_PIPE] = ACTIONS(1543),
    [anon_sym_RBRACE] = ACTIONS(1543),
    [anon_sym_TILDE] = ACTIONS(1543),
    [anon_sym_LPAREN] = ACTIONS(1543),
    [anon_sym_RPAREN] = ACTIONS(1543),
    [sym__newline_token] = ACTIONS(1555),
    [sym_uri_autolink] = ACTIONS(1537),
    [sym_email_autolink] = ACTIONS(1537),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1558),
    [anon_sym_LT_QMARK] = ACTIONS(1561),
    [aux_sym__declaration_token1] = ACTIONS(1564),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1567),
    [sym__whitespace_ge_2] = ACTIONS(1570),
    [aux_sym__whitespace_token1] = ACTIONS(1573),
    [sym__word_no_digit] = ACTIONS(1537),
    [sym__digits] = ACTIONS(1537),
    [sym__code_span_start] = ACTIONS(1576),
    [sym__emphasis_open_star] = ACTIONS(1579),
    [sym__emphasis_open_underscore] = ACTIONS(1582),
    [sym__last_token_punctuation] = ACTIONS(1585),
    [sym__strikethrough_open] = ACTIONS(1587),
    [sym__strikethrough_close] = ACTIONS(1590),
    [sym__latex_span_start] = ACTIONS(1592),
    [sym__unclosed_span] = ACTIONS(1537),
  },
  [STATE(93)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(121),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(121),
    [aux_sym__inline_no_tilde_no_link] = STATE(121),
    [sym__strikethrough_no_link] = STATE(121),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(121),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(121),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(1595),
    [sym_entity_reference] = ACTIONS(1598),
    [sym_numeric_character_reference] = ACTIONS(1598),
    [anon_sym_LT] = ACTIONS(1601),
    [anon_sym_GT] = ACTIONS(1604),
    [anon_sym_BANG] = ACTIONS(1607),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [anon_sym_POUND] = ACTIONS(1604),
    [anon_sym_DOLLAR] = ACTIONS(1604),
    [anon_sym_PERCENT] = ACTIONS(1604),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_STAR] = ACTIONS(1604),
    [anon_sym_PLUS] = ACTIONS(1604),
    [anon_sym_COMMA] = ACTIONS(1604),
    [anon_sym_DASH] = ACTIONS(1604),
    [anon_sym_DOT] = ACTIONS(1604),
    [anon_sym_SLASH] = ACTIONS(1604),
    [anon_sym_COLON] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1604),
    [anon_sym_EQ] = ACTIONS(1604),
    [anon_sym_QMARK] = ACTIONS(1604),
    [anon_sym_AT] = ACTIONS(1604),
    [anon_sym_BSLASH] = ACTIONS(1613),
    [anon_sym_CARET] = ACTIONS(1604),
    [anon_sym__] = ACTIONS(1604),
    [anon_sym_BQUOTE] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(1604),
    [anon_sym_PIPE] = ACTIONS(1604),
    [anon_sym_RBRACE] = ACTIONS(1604),
    [anon_sym_TILDE] = ACTIONS(1604),
    [anon_sym_LPAREN] = ACTIONS(1604),
    [anon_sym_RPAREN] = ACTIONS(1604),
    [sym__newline_token] = ACTIONS(1616),
    [sym_uri_autolink] = ACTIONS(1598),
    [sym_email_autolink] = ACTIONS(1598),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1619),
    [anon_sym_LT_QMARK] = ACTIONS(1622),
    [aux_sym__declaration_token1] = ACTIONS(1625),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1628),
    [sym__whitespace_ge_2] = ACTIONS(1631),
    [aux_sym__whitespace_token1] = ACTIONS(1634),
    [sym__word_no_digit] = ACTIONS(1598),
    [sym__digits] = ACTIONS(1598),
    [sym__code_span_start] = ACTIONS(1637),
    [sym__emphasis_open_star] = ACTIONS(1640),
    [sym__emphasis_open_underscore] = ACTIONS(1643),
    [sym__emphasis_close_underscore] = ACTIONS(1646),
    [sym__last_token_punctuation] = ACTIONS(1648),
    [sym__strikethrough_open] = ACTIONS(1650),
    [sym__latex_span_start] = ACTIONS(1653),
    [sym__unclosed_span] = ACTIONS(1598),
  },
  [STATE(94)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(119),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(119),
    [aux_sym__inline_no_star_no_link] = STATE(119),
    [sym__strikethrough_no_link] = STATE(119),
    [sym__emphasis_star_no_link] = STATE(629),
    [sym__strong_emphasis_star_no_link] = STATE(119),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(119),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(1656),
    [sym_entity_reference] = ACTIONS(1659),
    [sym_numeric_character_reference] = ACTIONS(1659),
    [anon_sym_LT] = ACTIONS(1662),
    [anon_sym_GT] = ACTIONS(1665),
    [anon_sym_BANG] = ACTIONS(1668),
    [anon_sym_DQUOTE] = ACTIONS(1665),
    [anon_sym_POUND] = ACTIONS(1665),
    [anon_sym_DOLLAR] = ACTIONS(1665),
    [anon_sym_PERCENT] = ACTIONS(1665),
    [anon_sym_AMP] = ACTIONS(1671),
    [anon_sym_SQUOTE] = ACTIONS(1665),
    [anon_sym_STAR] = ACTIONS(1665),
    [anon_sym_PLUS] = ACTIONS(1665),
    [anon_sym_COMMA] = ACTIONS(1665),
    [anon_sym_DASH] = ACTIONS(1665),
    [anon_sym_DOT] = ACTIONS(1665),
    [anon_sym_SLASH] = ACTIONS(1665),
    [anon_sym_COLON] = ACTIONS(1665),
    [anon_sym_SEMI] = ACTIONS(1665),
    [anon_sym_EQ] = ACTIONS(1665),
    [anon_sym_QMARK] = ACTIONS(1665),
    [anon_sym_AT] = ACTIONS(1665),
    [anon_sym_BSLASH] = ACTIONS(1674),
    [anon_sym_CARET] = ACTIONS(1665),
    [anon_sym__] = ACTIONS(1665),
    [anon_sym_BQUOTE] = ACTIONS(1665),
    [anon_sym_LBRACE] = ACTIONS(1665),
    [anon_sym_PIPE] = ACTIONS(1665),
    [anon_sym_RBRACE] = ACTIONS(1665),
    [anon_sym_TILDE] = ACTIONS(1665),
    [anon_sym_LPAREN] = ACTIONS(1665),
    [anon_sym_RPAREN] = ACTIONS(1665),
    [sym__newline_token] = ACTIONS(1677),
    [sym_uri_autolink] = ACTIONS(1659),
    [sym_email_autolink] = ACTIONS(1659),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1680),
    [anon_sym_LT_QMARK] = ACTIONS(1683),
    [aux_sym__declaration_token1] = ACTIONS(1686),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1689),
    [sym__whitespace_ge_2] = ACTIONS(1692),
    [aux_sym__whitespace_token1] = ACTIONS(1695),
    [sym__word_no_digit] = ACTIONS(1659),
    [sym__digits] = ACTIONS(1659),
    [sym__code_span_start] = ACTIONS(1698),
    [sym__emphasis_open_star] = ACTIONS(1701),
    [sym__emphasis_open_underscore] = ACTIONS(1704),
    [sym__emphasis_close_underscore] = ACTIONS(1646),
    [sym__last_token_punctuation] = ACTIONS(1707),
    [sym__strikethrough_open] = ACTIONS(1709),
    [sym__latex_span_start] = ACTIONS(1712),
    [sym__unclosed_span] = ACTIONS(1659),
  },
  [STATE(95)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(127),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(127),
    [aux_sym__inline_no_underscore_no_link] = STATE(127),
    [sym__strikethrough_no_link] = STATE(127),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(127),
    [sym__emphasis_underscore_no_link] = STATE(640),
    [sym__strong_emphasis_underscore_no_link] = STATE(127),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(1715),
    [sym_entity_reference] = ACTIONS(1718),
    [sym_numeric_character_reference] = ACTIONS(1718),
    [anon_sym_LT] = ACTIONS(1721),
    [anon_sym_GT] = ACTIONS(1724),
    [anon_sym_BANG] = ACTIONS(1727),
    [anon_sym_DQUOTE] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_DOLLAR] = ACTIONS(1724),
    [anon_sym_PERCENT] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1730),
    [anon_sym_SQUOTE] = ACTIONS(1724),
    [anon_sym_STAR] = ACTIONS(1724),
    [anon_sym_PLUS] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1724),
    [anon_sym_DASH] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1724),
    [anon_sym_SLASH] = ACTIONS(1724),
    [anon_sym_COLON] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_EQ] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_AT] = ACTIONS(1724),
    [anon_sym_BSLASH] = ACTIONS(1733),
    [anon_sym_CARET] = ACTIONS(1724),
    [anon_sym__] = ACTIONS(1724),
    [anon_sym_BQUOTE] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_RBRACE] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1724),
    [anon_sym_RPAREN] = ACTIONS(1724),
    [sym__newline_token] = ACTIONS(1736),
    [sym_uri_autolink] = ACTIONS(1718),
    [sym_email_autolink] = ACTIONS(1718),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1739),
    [anon_sym_LT_QMARK] = ACTIONS(1742),
    [aux_sym__declaration_token1] = ACTIONS(1745),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1748),
    [sym__whitespace_ge_2] = ACTIONS(1751),
    [aux_sym__whitespace_token1] = ACTIONS(1754),
    [sym__word_no_digit] = ACTIONS(1718),
    [sym__digits] = ACTIONS(1718),
    [sym__code_span_start] = ACTIONS(1757),
    [sym__emphasis_open_star] = ACTIONS(1760),
    [sym__emphasis_open_underscore] = ACTIONS(1763),
    [sym__last_token_punctuation] = ACTIONS(1766),
    [sym__strikethrough_open] = ACTIONS(1768),
    [sym__strikethrough_close] = ACTIONS(1590),
    [sym__latex_span_start] = ACTIONS(1771),
    [sym__unclosed_span] = ACTIONS(1718),
  },
  [STATE(96)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(113),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(113),
    [aux_sym__inline_no_tilde_no_link] = STATE(113),
    [sym__strikethrough_no_link] = STATE(113),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(113),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(113),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(1774),
    [sym_entity_reference] = ACTIONS(1777),
    [sym_numeric_character_reference] = ACTIONS(1777),
    [anon_sym_LT] = ACTIONS(1780),
    [anon_sym_GT] = ACTIONS(1783),
    [anon_sym_BANG] = ACTIONS(1786),
    [anon_sym_DQUOTE] = ACTIONS(1783),
    [anon_sym_POUND] = ACTIONS(1783),
    [anon_sym_DOLLAR] = ACTIONS(1783),
    [anon_sym_PERCENT] = ACTIONS(1783),
    [anon_sym_AMP] = ACTIONS(1789),
    [anon_sym_SQUOTE] = ACTIONS(1783),
    [anon_sym_STAR] = ACTIONS(1783),
    [anon_sym_PLUS] = ACTIONS(1783),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_DASH] = ACTIONS(1783),
    [anon_sym_DOT] = ACTIONS(1783),
    [anon_sym_SLASH] = ACTIONS(1783),
    [anon_sym_COLON] = ACTIONS(1783),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_EQ] = ACTIONS(1783),
    [anon_sym_QMARK] = ACTIONS(1783),
    [anon_sym_AT] = ACTIONS(1783),
    [anon_sym_BSLASH] = ACTIONS(1792),
    [anon_sym_CARET] = ACTIONS(1783),
    [anon_sym__] = ACTIONS(1783),
    [anon_sym_BQUOTE] = ACTIONS(1783),
    [anon_sym_LBRACE] = ACTIONS(1783),
    [anon_sym_PIPE] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [anon_sym_TILDE] = ACTIONS(1783),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_RPAREN] = ACTIONS(1783),
    [sym__newline_token] = ACTIONS(1795),
    [sym_uri_autolink] = ACTIONS(1777),
    [sym_email_autolink] = ACTIONS(1777),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1798),
    [anon_sym_LT_QMARK] = ACTIONS(1801),
    [aux_sym__declaration_token1] = ACTIONS(1804),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1807),
    [sym__whitespace_ge_2] = ACTIONS(1810),
    [aux_sym__whitespace_token1] = ACTIONS(1813),
    [sym__word_no_digit] = ACTIONS(1777),
    [sym__digits] = ACTIONS(1777),
    [sym__code_span_start] = ACTIONS(1816),
    [sym__emphasis_open_star] = ACTIONS(1819),
    [sym__emphasis_open_underscore] = ACTIONS(1822),
    [sym__emphasis_close_star] = ACTIONS(1825),
    [sym__last_token_punctuation] = ACTIONS(1827),
    [sym__strikethrough_open] = ACTIONS(1829),
    [sym__latex_span_start] = ACTIONS(1832),
    [sym__unclosed_span] = ACTIONS(1777),
  },
  [STATE(97)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(112),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(112),
    [aux_sym__inline_no_underscore_no_link] = STATE(112),
    [sym__strikethrough_no_link] = STATE(112),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(112),
    [sym__emphasis_underscore_no_link] = STATE(628),
    [sym__strong_emphasis_underscore_no_link] = STATE(112),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(1835),
    [sym_entity_reference] = ACTIONS(1838),
    [sym_numeric_character_reference] = ACTIONS(1838),
    [anon_sym_LT] = ACTIONS(1841),
    [anon_sym_GT] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1847),
    [anon_sym_DQUOTE] = ACTIONS(1844),
    [anon_sym_POUND] = ACTIONS(1844),
    [anon_sym_DOLLAR] = ACTIONS(1844),
    [anon_sym_PERCENT] = ACTIONS(1844),
    [anon_sym_AMP] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1844),
    [anon_sym_PLUS] = ACTIONS(1844),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_DASH] = ACTIONS(1844),
    [anon_sym_DOT] = ACTIONS(1844),
    [anon_sym_SLASH] = ACTIONS(1844),
    [anon_sym_COLON] = ACTIONS(1844),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_EQ] = ACTIONS(1844),
    [anon_sym_QMARK] = ACTIONS(1844),
    [anon_sym_AT] = ACTIONS(1844),
    [anon_sym_BSLASH] = ACTIONS(1853),
    [anon_sym_CARET] = ACTIONS(1844),
    [anon_sym__] = ACTIONS(1844),
    [anon_sym_BQUOTE] = ACTIONS(1844),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [anon_sym_TILDE] = ACTIONS(1844),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_RPAREN] = ACTIONS(1844),
    [sym__newline_token] = ACTIONS(1856),
    [sym_uri_autolink] = ACTIONS(1838),
    [sym_email_autolink] = ACTIONS(1838),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1859),
    [anon_sym_LT_QMARK] = ACTIONS(1862),
    [aux_sym__declaration_token1] = ACTIONS(1865),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1868),
    [sym__whitespace_ge_2] = ACTIONS(1871),
    [aux_sym__whitespace_token1] = ACTIONS(1874),
    [sym__word_no_digit] = ACTIONS(1838),
    [sym__digits] = ACTIONS(1838),
    [sym__code_span_start] = ACTIONS(1877),
    [sym__emphasis_open_star] = ACTIONS(1880),
    [sym__emphasis_open_underscore] = ACTIONS(1883),
    [sym__emphasis_close_star] = ACTIONS(1825),
    [sym__last_token_punctuation] = ACTIONS(1886),
    [sym__strikethrough_open] = ACTIONS(1888),
    [sym__latex_span_start] = ACTIONS(1891),
    [sym__unclosed_span] = ACTIONS(1838),
  },
  [STATE(98)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(109),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(109),
    [aux_sym__inline_no_star_no_link] = STATE(109),
    [sym__strikethrough_no_link] = STATE(109),
    [sym__emphasis_star_no_link] = STATE(621),
    [sym__strong_emphasis_star_no_link] = STATE(109),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(109),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(1894),
    [sym_entity_reference] = ACTIONS(1897),
    [sym_numeric_character_reference] = ACTIONS(1897),
    [anon_sym_RBRACK] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1900),
    [anon_sym_GT] = ACTIONS(1903),
    [anon_sym_BANG] = ACTIONS(1906),
    [anon_sym_DQUOTE] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_DOLLAR] = ACTIONS(1903),
    [anon_sym_PERCENT] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1909),
    [anon_sym_SQUOTE] = ACTIONS(1903),
    [anon_sym_STAR] = ACTIONS(1903),
    [anon_sym_PLUS] = ACTIONS(1903),
    [anon_sym_COMMA] = ACTIONS(1903),
    [anon_sym_DASH] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1903),
    [anon_sym_SLASH] = ACTIONS(1903),
    [anon_sym_COLON] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1903),
    [anon_sym_EQ] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_AT] = ACTIONS(1903),
    [anon_sym_BSLASH] = ACTIONS(1912),
    [anon_sym_CARET] = ACTIONS(1903),
    [anon_sym__] = ACTIONS(1903),
    [anon_sym_BQUOTE] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_RBRACE] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1903),
    [anon_sym_RPAREN] = ACTIONS(1903),
    [sym__newline_token] = ACTIONS(1915),
    [sym_uri_autolink] = ACTIONS(1897),
    [sym_email_autolink] = ACTIONS(1897),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1918),
    [anon_sym_LT_QMARK] = ACTIONS(1921),
    [aux_sym__declaration_token1] = ACTIONS(1924),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1927),
    [sym__whitespace_ge_2] = ACTIONS(1930),
    [aux_sym__whitespace_token1] = ACTIONS(1933),
    [sym__word_no_digit] = ACTIONS(1897),
    [sym__digits] = ACTIONS(1897),
    [sym__code_span_start] = ACTIONS(1936),
    [sym__emphasis_open_star] = ACTIONS(1939),
    [sym__emphasis_open_underscore] = ACTIONS(1942),
    [sym__last_token_punctuation] = ACTIONS(1945),
    [sym__strikethrough_open] = ACTIONS(1947),
    [sym__latex_span_start] = ACTIONS(1950),
    [sym__unclosed_span] = ACTIONS(1897),
  },
  [STATE(99)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(117),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(117),
    [aux_sym__inline_no_underscore_no_link] = STATE(117),
    [sym__strikethrough_no_link] = STATE(117),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(117),
    [sym__emphasis_underscore_no_link] = STATE(626),
    [sym__strong_emphasis_underscore_no_link] = STATE(117),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(1953),
    [sym_entity_reference] = ACTIONS(1956),
    [sym_numeric_character_reference] = ACTIONS(1956),
    [anon_sym_RBRACK] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1959),
    [anon_sym_GT] = ACTIONS(1962),
    [anon_sym_BANG] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1962),
    [anon_sym_POUND] = ACTIONS(1962),
    [anon_sym_DOLLAR] = ACTIONS(1962),
    [anon_sym_PERCENT] = ACTIONS(1962),
    [anon_sym_AMP] = ACTIONS(1968),
    [anon_sym_SQUOTE] = ACTIONS(1962),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_PLUS] = ACTIONS(1962),
    [anon_sym_COMMA] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_DOT] = ACTIONS(1962),
    [anon_sym_SLASH] = ACTIONS(1962),
    [anon_sym_COLON] = ACTIONS(1962),
    [anon_sym_SEMI] = ACTIONS(1962),
    [anon_sym_EQ] = ACTIONS(1962),
    [anon_sym_QMARK] = ACTIONS(1962),
    [anon_sym_AT] = ACTIONS(1962),
    [anon_sym_BSLASH] = ACTIONS(1971),
    [anon_sym_CARET] = ACTIONS(1962),
    [anon_sym__] = ACTIONS(1962),
    [anon_sym_BQUOTE] = ACTIONS(1962),
    [anon_sym_LBRACE] = ACTIONS(1962),
    [anon_sym_PIPE] = ACTIONS(1962),
    [anon_sym_RBRACE] = ACTIONS(1962),
    [anon_sym_TILDE] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1962),
    [anon_sym_RPAREN] = ACTIONS(1962),
    [sym__newline_token] = ACTIONS(1974),
    [sym_uri_autolink] = ACTIONS(1956),
    [sym_email_autolink] = ACTIONS(1956),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1977),
    [anon_sym_LT_QMARK] = ACTIONS(1980),
    [aux_sym__declaration_token1] = ACTIONS(1983),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1986),
    [sym__whitespace_ge_2] = ACTIONS(1989),
    [aux_sym__whitespace_token1] = ACTIONS(1992),
    [sym__word_no_digit] = ACTIONS(1956),
    [sym__digits] = ACTIONS(1956),
    [sym__code_span_start] = ACTIONS(1995),
    [sym__emphasis_open_star] = ACTIONS(1998),
    [sym__emphasis_open_underscore] = ACTIONS(2001),
    [sym__last_token_punctuation] = ACTIONS(2004),
    [sym__strikethrough_open] = ACTIONS(2006),
    [sym__latex_span_start] = ACTIONS(2009),
    [sym__unclosed_span] = ACTIONS(1956),
  },
  [STATE(100)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(120),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(120),
    [aux_sym__inline_no_underscore_no_link] = STATE(120),
    [sym__strikethrough_no_link] = STATE(120),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(120),
    [sym__emphasis_underscore_no_link] = STATE(630),
    [sym__strong_emphasis_underscore_no_link] = STATE(120),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__last_token_punctuation] = ACTIONS(2026),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(101)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2046),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(102)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(2048),
    [sym_entity_reference] = ACTIONS(2051),
    [sym_numeric_character_reference] = ACTIONS(2051),
    [anon_sym_RBRACK] = ACTIONS(2054),
    [anon_sym_LT] = ACTIONS(2056),
    [anon_sym_GT] = ACTIONS(2059),
    [anon_sym_BANG] = ACTIONS(2062),
    [anon_sym_DQUOTE] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_DOLLAR] = ACTIONS(2059),
    [anon_sym_PERCENT] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2065),
    [anon_sym_SQUOTE] = ACTIONS(2059),
    [anon_sym_STAR] = ACTIONS(2059),
    [anon_sym_PLUS] = ACTIONS(2059),
    [anon_sym_COMMA] = ACTIONS(2059),
    [anon_sym_DASH] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2059),
    [anon_sym_SLASH] = ACTIONS(2059),
    [anon_sym_COLON] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2059),
    [anon_sym_EQ] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_AT] = ACTIONS(2059),
    [anon_sym_BSLASH] = ACTIONS(2068),
    [anon_sym_CARET] = ACTIONS(2059),
    [anon_sym__] = ACTIONS(2059),
    [anon_sym_BQUOTE] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_RBRACE] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2059),
    [anon_sym_RPAREN] = ACTIONS(2059),
    [sym__newline_token] = ACTIONS(2071),
    [sym_uri_autolink] = ACTIONS(2051),
    [sym_email_autolink] = ACTIONS(2051),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2074),
    [anon_sym_LT_QMARK] = ACTIONS(2077),
    [aux_sym__declaration_token1] = ACTIONS(2080),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2083),
    [sym__whitespace_ge_2] = ACTIONS(2086),
    [aux_sym__whitespace_token1] = ACTIONS(2089),
    [sym__word_no_digit] = ACTIONS(2051),
    [sym__digits] = ACTIONS(2051),
    [sym__code_span_start] = ACTIONS(2092),
    [sym__emphasis_open_star] = ACTIONS(2095),
    [sym__emphasis_open_underscore] = ACTIONS(2098),
    [sym__strikethrough_open] = ACTIONS(2101),
    [sym__latex_span_start] = ACTIONS(2104),
    [sym__unclosed_span] = ACTIONS(2051),
  },
  [STATE(103)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2121),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(104)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(111),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(111),
    [aux_sym__inline_no_star_no_link] = STATE(111),
    [sym__strikethrough_no_link] = STATE(111),
    [sym__emphasis_star_no_link] = STATE(627),
    [sym__strong_emphasis_star_no_link] = STATE(111),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(111),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__last_token_punctuation] = ACTIONS(2125),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(105)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2127),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(106)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(2129),
    [sym_entity_reference] = ACTIONS(2132),
    [sym_numeric_character_reference] = ACTIONS(2132),
    [anon_sym_LT] = ACTIONS(2135),
    [anon_sym_GT] = ACTIONS(2138),
    [anon_sym_BANG] = ACTIONS(2141),
    [anon_sym_DQUOTE] = ACTIONS(2138),
    [anon_sym_POUND] = ACTIONS(2138),
    [anon_sym_DOLLAR] = ACTIONS(2138),
    [anon_sym_PERCENT] = ACTIONS(2138),
    [anon_sym_AMP] = ACTIONS(2144),
    [anon_sym_SQUOTE] = ACTIONS(2138),
    [anon_sym_STAR] = ACTIONS(2138),
    [anon_sym_PLUS] = ACTIONS(2138),
    [anon_sym_COMMA] = ACTIONS(2138),
    [anon_sym_DASH] = ACTIONS(2138),
    [anon_sym_DOT] = ACTIONS(2138),
    [anon_sym_SLASH] = ACTIONS(2138),
    [anon_sym_COLON] = ACTIONS(2138),
    [anon_sym_SEMI] = ACTIONS(2138),
    [anon_sym_EQ] = ACTIONS(2138),
    [anon_sym_QMARK] = ACTIONS(2138),
    [anon_sym_AT] = ACTIONS(2138),
    [anon_sym_BSLASH] = ACTIONS(2147),
    [anon_sym_CARET] = ACTIONS(2138),
    [anon_sym__] = ACTIONS(2138),
    [anon_sym_BQUOTE] = ACTIONS(2138),
    [anon_sym_LBRACE] = ACTIONS(2138),
    [anon_sym_PIPE] = ACTIONS(2138),
    [anon_sym_RBRACE] = ACTIONS(2138),
    [anon_sym_TILDE] = ACTIONS(2138),
    [anon_sym_LPAREN] = ACTIONS(2138),
    [anon_sym_RPAREN] = ACTIONS(2138),
    [sym__newline_token] = ACTIONS(2150),
    [sym_uri_autolink] = ACTIONS(2132),
    [sym_email_autolink] = ACTIONS(2132),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2153),
    [anon_sym_LT_QMARK] = ACTIONS(2156),
    [aux_sym__declaration_token1] = ACTIONS(2159),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2162),
    [sym__whitespace_ge_2] = ACTIONS(2165),
    [aux_sym__whitespace_token1] = ACTIONS(2168),
    [sym__word_no_digit] = ACTIONS(2132),
    [sym__digits] = ACTIONS(2132),
    [sym__code_span_start] = ACTIONS(2171),
    [sym__emphasis_open_star] = ACTIONS(2174),
    [sym__emphasis_open_underscore] = ACTIONS(2177),
    [sym__emphasis_close_underscore] = ACTIONS(2180),
    [sym__strikethrough_open] = ACTIONS(2182),
    [sym__latex_span_start] = ACTIONS(2185),
    [sym__unclosed_span] = ACTIONS(2132),
  },
  [STATE(107)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2188),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(108)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(2190),
    [sym_entity_reference] = ACTIONS(2193),
    [sym_numeric_character_reference] = ACTIONS(2193),
    [anon_sym_LT] = ACTIONS(2196),
    [anon_sym_GT] = ACTIONS(2199),
    [anon_sym_BANG] = ACTIONS(2202),
    [anon_sym_DQUOTE] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_DOLLAR] = ACTIONS(2199),
    [anon_sym_PERCENT] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2205),
    [anon_sym_SQUOTE] = ACTIONS(2199),
    [anon_sym_STAR] = ACTIONS(2199),
    [anon_sym_PLUS] = ACTIONS(2199),
    [anon_sym_COMMA] = ACTIONS(2199),
    [anon_sym_DASH] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2199),
    [anon_sym_SLASH] = ACTIONS(2199),
    [anon_sym_COLON] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2199),
    [anon_sym_EQ] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_AT] = ACTIONS(2199),
    [anon_sym_BSLASH] = ACTIONS(2208),
    [anon_sym_CARET] = ACTIONS(2199),
    [anon_sym__] = ACTIONS(2199),
    [anon_sym_BQUOTE] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_RBRACE] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2199),
    [anon_sym_RPAREN] = ACTIONS(2199),
    [sym__newline_token] = ACTIONS(2211),
    [sym_uri_autolink] = ACTIONS(2193),
    [sym_email_autolink] = ACTIONS(2193),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2214),
    [anon_sym_LT_QMARK] = ACTIONS(2217),
    [aux_sym__declaration_token1] = ACTIONS(2220),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2223),
    [sym__whitespace_ge_2] = ACTIONS(2226),
    [aux_sym__whitespace_token1] = ACTIONS(2229),
    [sym__word_no_digit] = ACTIONS(2193),
    [sym__digits] = ACTIONS(2193),
    [sym__code_span_start] = ACTIONS(2232),
    [sym__emphasis_open_star] = ACTIONS(2235),
    [sym__emphasis_open_underscore] = ACTIONS(2238),
    [sym__strikethrough_open] = ACTIONS(2241),
    [sym__strikethrough_close] = ACTIONS(2244),
    [sym__latex_span_start] = ACTIONS(2246),
    [sym__unclosed_span] = ACTIONS(2193),
  },
  [STATE(109)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2249),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(110)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2251),
    [sym_numeric_character_reference] = ACTIONS(2251),
    [anon_sym_RBRACK] = ACTIONS(2253),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2251),
    [sym_email_autolink] = ACTIONS(2251),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2251),
    [sym__digits] = ACTIONS(2251),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2263),
    [sym__emphasis_open_underscore] = ACTIONS(2265),
    [sym__strikethrough_open] = ACTIONS(2267),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2251),
  },
  [STATE(111)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2269),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(112)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2271),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(113)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2273),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(114)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2275),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(115)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2277),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(116)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2279),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(117)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2281),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(118)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2251),
    [sym_numeric_character_reference] = ACTIONS(2251),
    [anon_sym_RBRACK] = ACTIONS(2283),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2251),
    [sym_email_autolink] = ACTIONS(2251),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2251),
    [sym__digits] = ACTIONS(2251),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2263),
    [sym__emphasis_open_underscore] = ACTIONS(2265),
    [sym__strikethrough_open] = ACTIONS(2267),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2251),
  },
  [STATE(119)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2285),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(120)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2287),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(121)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2289),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(122)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2291),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(123)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2293),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(124)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2295),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(125)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2251),
    [sym_numeric_character_reference] = ACTIONS(2251),
    [anon_sym_RBRACK] = ACTIONS(2297),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2251),
    [sym_email_autolink] = ACTIONS(2251),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2251),
    [sym__digits] = ACTIONS(2251),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2263),
    [sym__emphasis_open_underscore] = ACTIONS(2265),
    [sym__strikethrough_open] = ACTIONS(2267),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2251),
  },
  [STATE(126)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2299),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(127)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2301),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(128)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2303),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(129)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__emphasis_close_star] = ACTIONS(2305),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(130)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(106),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(106),
    [aux_sym__inline_no_underscore_no_link] = STATE(106),
    [sym__strikethrough_no_link] = STATE(106),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(106),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(106),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__emphasis_close_underscore] = ACTIONS(2307),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(131)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(108),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(108),
    [aux_sym__inline_no_tilde_no_link] = STATE(108),
    [sym__strikethrough_no_link] = STATE(108),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(108),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(108),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__strikethrough_close] = ACTIONS(2309),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(132)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2251),
    [sym_numeric_character_reference] = ACTIONS(2251),
    [anon_sym_RBRACK] = ACTIONS(2311),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2251),
    [sym_email_autolink] = ACTIONS(2251),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2251),
    [sym__digits] = ACTIONS(2251),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2263),
    [sym__emphasis_open_underscore] = ACTIONS(2265),
    [sym__strikethrough_open] = ACTIONS(2267),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2251),
  },
  [STATE(133)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(128),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(128),
    [aux_sym__inline_no_tilde_no_link] = STATE(128),
    [sym__strikethrough_no_link] = STATE(128),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(128),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(128),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__last_token_punctuation] = ACTIONS(2313),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(134)] = {
    [sym_backslash_escape] = STATE(165),
    [sym_code_span] = STATE(165),
    [sym_latex_block] = STATE(165),
    [sym_image] = STATE(165),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(165),
    [sym__whitespace] = STATE(165),
    [sym__word] = STATE(165),
    [sym__soft_line_break] = STATE(165),
    [sym__inline_base] = STATE(606),
    [sym__text_base] = STATE(165),
    [sym__inline_element_no_link] = STATE(606),
    [aux_sym__inline_no_link] = STATE(102),
    [sym__strikethrough_no_link] = STATE(606),
    [sym__emphasis_star_no_link] = STATE(642),
    [sym__strong_emphasis_star_no_link] = STATE(606),
    [sym__emphasis_underscore_no_link] = STATE(642),
    [sym__strong_emphasis_underscore_no_link] = STATE(606),
    [aux_sym__inline_base_repeat1] = STATE(165),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2251),
    [sym_numeric_character_reference] = ACTIONS(2251),
    [anon_sym_RBRACK] = ACTIONS(2315),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2251),
    [sym_email_autolink] = ACTIONS(2251),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2251),
    [sym__digits] = ACTIONS(2251),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2263),
    [sym__emphasis_open_underscore] = ACTIONS(2265),
    [sym__strikethrough_open] = ACTIONS(2267),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2251),
  },
  [STATE(135)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(135),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(135),
    [aux_sym__inline_no_star_no_link] = STATE(135),
    [sym__strikethrough_no_link] = STATE(135),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(135),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(135),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(2317),
    [sym_entity_reference] = ACTIONS(2320),
    [sym_numeric_character_reference] = ACTIONS(2320),
    [anon_sym_LT] = ACTIONS(2323),
    [anon_sym_GT] = ACTIONS(2326),
    [anon_sym_BANG] = ACTIONS(2329),
    [anon_sym_DQUOTE] = ACTIONS(2326),
    [anon_sym_POUND] = ACTIONS(2326),
    [anon_sym_DOLLAR] = ACTIONS(2326),
    [anon_sym_PERCENT] = ACTIONS(2326),
    [anon_sym_AMP] = ACTIONS(2332),
    [anon_sym_SQUOTE] = ACTIONS(2326),
    [anon_sym_STAR] = ACTIONS(2326),
    [anon_sym_PLUS] = ACTIONS(2326),
    [anon_sym_COMMA] = ACTIONS(2326),
    [anon_sym_DASH] = ACTIONS(2326),
    [anon_sym_DOT] = ACTIONS(2326),
    [anon_sym_SLASH] = ACTIONS(2326),
    [anon_sym_COLON] = ACTIONS(2326),
    [anon_sym_SEMI] = ACTIONS(2326),
    [anon_sym_EQ] = ACTIONS(2326),
    [anon_sym_QMARK] = ACTIONS(2326),
    [anon_sym_AT] = ACTIONS(2326),
    [anon_sym_BSLASH] = ACTIONS(2335),
    [anon_sym_CARET] = ACTIONS(2326),
    [anon_sym__] = ACTIONS(2326),
    [anon_sym_BQUOTE] = ACTIONS(2326),
    [anon_sym_LBRACE] = ACTIONS(2326),
    [anon_sym_PIPE] = ACTIONS(2326),
    [anon_sym_RBRACE] = ACTIONS(2326),
    [anon_sym_TILDE] = ACTIONS(2326),
    [anon_sym_LPAREN] = ACTIONS(2326),
    [anon_sym_RPAREN] = ACTIONS(2326),
    [sym__newline_token] = ACTIONS(2338),
    [sym_uri_autolink] = ACTIONS(2320),
    [sym_email_autolink] = ACTIONS(2320),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2341),
    [anon_sym_LT_QMARK] = ACTIONS(2344),
    [aux_sym__declaration_token1] = ACTIONS(2347),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2350),
    [sym__whitespace_ge_2] = ACTIONS(2353),
    [aux_sym__whitespace_token1] = ACTIONS(2356),
    [sym__word_no_digit] = ACTIONS(2320),
    [sym__digits] = ACTIONS(2320),
    [sym__code_span_start] = ACTIONS(2359),
    [sym__emphasis_open_star] = ACTIONS(2362),
    [sym__emphasis_open_underscore] = ACTIONS(2365),
    [sym__emphasis_close_star] = ACTIONS(2368),
    [sym__strikethrough_open] = ACTIONS(2370),
    [sym__latex_span_start] = ACTIONS(2373),
    [sym__unclosed_span] = ACTIONS(2320),
  },
  [STATE(136)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(105),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(105),
    [aux_sym__inline_no_underscore_no_link] = STATE(105),
    [sym__strikethrough_no_link] = STATE(105),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(105),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(105),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(137)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(103),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(103),
    [aux_sym__inline_no_star_no_link] = STATE(103),
    [sym__strikethrough_no_link] = STATE(103),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(103),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(103),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(138)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(129),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(129),
    [aux_sym__inline_no_star_no_link] = STATE(129),
    [sym__strikethrough_no_link] = STATE(129),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(129),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(129),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(139)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(124),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(124),
    [aux_sym__inline_no_tilde_no_link] = STATE(124),
    [sym__strikethrough_no_link] = STATE(124),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(124),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(124),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(140)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(107),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(107),
    [aux_sym__inline_no_tilde_no_link] = STATE(107),
    [sym__strikethrough_no_link] = STATE(107),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(107),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(107),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(141)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(116),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(116),
    [aux_sym__inline_no_tilde_no_link] = STATE(116),
    [sym__strikethrough_no_link] = STATE(116),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(116),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(116),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(142)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(115),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(115),
    [aux_sym__inline_no_underscore_no_link] = STATE(115),
    [sym__strikethrough_no_link] = STATE(115),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(115),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(115),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(143)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(114),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(114),
    [aux_sym__inline_no_star_no_link] = STATE(114),
    [sym__strikethrough_no_link] = STATE(114),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(114),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(114),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(144)] = {
    [sym_backslash_escape] = STATE(162),
    [sym_code_span] = STATE(162),
    [sym_latex_block] = STATE(162),
    [sym_image] = STATE(162),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(162),
    [sym__whitespace] = STATE(162),
    [sym__word] = STATE(162),
    [sym__soft_line_break] = STATE(162),
    [sym__inline_base] = STATE(122),
    [sym__text_base] = STATE(162),
    [sym__inline_element_no_star_no_link] = STATE(122),
    [aux_sym__inline_no_star_no_link] = STATE(122),
    [sym__strikethrough_no_link] = STATE(122),
    [sym__emphasis_star_no_link] = STATE(600),
    [sym__strong_emphasis_star_no_link] = STATE(122),
    [sym__emphasis_underscore_no_link] = STATE(600),
    [sym__strong_emphasis_underscore_no_link] = STATE(122),
    [aux_sym__inline_base_repeat1] = STATE(162),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2107),
    [sym_numeric_character_reference] = ACTIONS(2107),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2107),
    [sym_email_autolink] = ACTIONS(2107),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2107),
    [sym__digits] = ACTIONS(2107),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2117),
    [sym__emphasis_open_underscore] = ACTIONS(2119),
    [sym__strikethrough_open] = ACTIONS(2123),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2107),
  },
  [STATE(145)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(123),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(123),
    [aux_sym__inline_no_underscore_no_link] = STATE(123),
    [sym__strikethrough_no_link] = STATE(123),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(123),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(123),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(146)] = {
    [sym_backslash_escape] = STATE(159),
    [sym_code_span] = STATE(159),
    [sym_latex_block] = STATE(159),
    [sym_image] = STATE(159),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(159),
    [sym__whitespace] = STATE(159),
    [sym__word] = STATE(159),
    [sym__soft_line_break] = STATE(159),
    [sym__inline_base] = STATE(130),
    [sym__text_base] = STATE(159),
    [sym__inline_element_no_underscore_no_link] = STATE(130),
    [aux_sym__inline_no_underscore_no_link] = STATE(130),
    [sym__strikethrough_no_link] = STATE(130),
    [sym__emphasis_star_no_link] = STATE(622),
    [sym__strong_emphasis_star_no_link] = STATE(130),
    [sym__emphasis_underscore_no_link] = STATE(622),
    [sym__strong_emphasis_underscore_no_link] = STATE(130),
    [aux_sym__inline_base_repeat1] = STATE(159),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2012),
    [sym_numeric_character_reference] = ACTIONS(2012),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2012),
    [sym_email_autolink] = ACTIONS(2012),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2012),
    [sym__digits] = ACTIONS(2012),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2022),
    [sym__emphasis_open_underscore] = ACTIONS(2024),
    [sym__strikethrough_open] = ACTIONS(2028),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2012),
  },
  [STATE(147)] = {
    [sym_backslash_escape] = STATE(163),
    [sym_code_span] = STATE(163),
    [sym_latex_block] = STATE(163),
    [sym_image] = STATE(163),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(163),
    [sym__whitespace] = STATE(163),
    [sym__word] = STATE(163),
    [sym__soft_line_break] = STATE(163),
    [sym__inline_base] = STATE(131),
    [sym__text_base] = STATE(163),
    [sym__inline_element_no_tilde_no_link] = STATE(131),
    [aux_sym__inline_no_tilde_no_link] = STATE(131),
    [sym__strikethrough_no_link] = STATE(131),
    [sym__emphasis_star_no_link] = STATE(605),
    [sym__strong_emphasis_star_no_link] = STATE(131),
    [sym__emphasis_underscore_no_link] = STATE(605),
    [sym__strong_emphasis_underscore_no_link] = STATE(131),
    [aux_sym__inline_base_repeat1] = STATE(163),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2030),
    [sym_numeric_character_reference] = ACTIONS(2030),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2030),
    [sym_email_autolink] = ACTIONS(2030),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2030),
    [sym__digits] = ACTIONS(2030),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2040),
    [sym__emphasis_open_underscore] = ACTIONS(2042),
    [sym__strikethrough_open] = ACTIONS(2044),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2030),
  },
  [STATE(148)] = {
    [sym_backslash_escape] = STATE(149),
    [sym_code_span] = STATE(149),
    [sym_latex_block] = STATE(149),
    [sym_image] = STATE(149),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(149),
    [sym__whitespace] = STATE(149),
    [sym__word] = STATE(149),
    [sym__soft_line_break] = STATE(149),
    [sym__text_base] = STATE(149),
    [aux_sym__inline_base_repeat1] = STATE(149),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2376),
    [sym_numeric_character_reference] = ACTIONS(2376),
    [anon_sym_LBRACK] = ACTIONS(2378),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2376),
    [sym_email_autolink] = ACTIONS(2376),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(660),
    [anon_sym_LT_QMARK] = ACTIONS(662),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(666),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2376),
    [sym__digits] = ACTIONS(2376),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__emphasis_close_star] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2376),
  },
  [STATE(149)] = {
    [sym_backslash_escape] = STATE(149),
    [sym_code_span] = STATE(149),
    [sym_latex_block] = STATE(149),
    [sym_image] = STATE(149),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(385),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(149),
    [sym__whitespace] = STATE(149),
    [sym__word] = STATE(149),
    [sym__soft_line_break] = STATE(149),
    [sym__text_base] = STATE(149),
    [aux_sym__inline_base_repeat1] = STATE(149),
    [sym__backslash_escape] = ACTIONS(2380),
    [sym_entity_reference] = ACTIONS(2383),
    [sym_numeric_character_reference] = ACTIONS(2383),
    [anon_sym_LBRACK] = ACTIONS(2386),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2388),
    [anon_sym_GT] = ACTIONS(2391),
    [anon_sym_BANG] = ACTIONS(2394),
    [anon_sym_DQUOTE] = ACTIONS(2391),
    [anon_sym_POUND] = ACTIONS(2391),
    [anon_sym_DOLLAR] = ACTIONS(2391),
    [anon_sym_PERCENT] = ACTIONS(2391),
    [anon_sym_AMP] = ACTIONS(2397),
    [anon_sym_SQUOTE] = ACTIONS(2391),
    [anon_sym_STAR] = ACTIONS(2391),
    [anon_sym_PLUS] = ACTIONS(2391),
    [anon_sym_COMMA] = ACTIONS(2391),
    [anon_sym_DASH] = ACTIONS(2391),
    [anon_sym_DOT] = ACTIONS(2391),
    [anon_sym_SLASH] = ACTIONS(2391),
    [anon_sym_COLON] = ACTIONS(2391),
    [anon_sym_SEMI] = ACTIONS(2391),
    [anon_sym_EQ] = ACTIONS(2391),
    [anon_sym_QMARK] = ACTIONS(2391),
    [anon_sym_AT] = ACTIONS(2391),
    [anon_sym_BSLASH] = ACTIONS(2400),
    [anon_sym_CARET] = ACTIONS(2391),
    [anon_sym__] = ACTIONS(2391),
    [anon_sym_BQUOTE] = ACTIONS(2391),
    [anon_sym_LBRACE] = ACTIONS(2391),
    [anon_sym_PIPE] = ACTIONS(2391),
    [anon_sym_RBRACE] = ACTIONS(2391),
    [anon_sym_TILDE] = ACTIONS(2391),
    [anon_sym_LPAREN] = ACTIONS(2391),
    [anon_sym_RPAREN] = ACTIONS(2391),
    [sym__newline_token] = ACTIONS(2403),
    [sym_uri_autolink] = ACTIONS(2383),
    [sym_email_autolink] = ACTIONS(2383),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2406),
    [anon_sym_LT_QMARK] = ACTIONS(2409),
    [aux_sym__declaration_token1] = ACTIONS(2412),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2415),
    [sym__whitespace_ge_2] = ACTIONS(2418),
    [aux_sym__whitespace_token1] = ACTIONS(2421),
    [sym__word_no_digit] = ACTIONS(2383),
    [sym__digits] = ACTIONS(2383),
    [sym__code_span_start] = ACTIONS(2424),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__emphasis_close_star] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2427),
    [sym__unclosed_span] = ACTIONS(2383),
  },
  [STATE(150)] = {
    [sym_backslash_escape] = STATE(151),
    [sym_code_span] = STATE(151),
    [sym_latex_block] = STATE(151),
    [sym_image] = STATE(151),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(151),
    [sym__whitespace] = STATE(151),
    [sym__word] = STATE(151),
    [sym__soft_line_break] = STATE(151),
    [sym__text_base] = STATE(151),
    [aux_sym__inline_base_repeat1] = STATE(151),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2430),
    [sym_numeric_character_reference] = ACTIONS(2430),
    [anon_sym_LBRACK] = ACTIONS(2378),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2430),
    [sym_email_autolink] = ACTIONS(2430),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(704),
    [anon_sym_LT_QMARK] = ACTIONS(706),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(710),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2430),
    [sym__digits] = ACTIONS(2430),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__emphasis_close_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2430),
  },
  [STATE(151)] = {
    [sym_backslash_escape] = STATE(151),
    [sym_code_span] = STATE(151),
    [sym_latex_block] = STATE(151),
    [sym_image] = STATE(151),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(455),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(151),
    [sym__whitespace] = STATE(151),
    [sym__word] = STATE(151),
    [sym__soft_line_break] = STATE(151),
    [sym__text_base] = STATE(151),
    [aux_sym__inline_base_repeat1] = STATE(151),
    [sym__backslash_escape] = ACTIONS(2432),
    [sym_entity_reference] = ACTIONS(2435),
    [sym_numeric_character_reference] = ACTIONS(2435),
    [anon_sym_LBRACK] = ACTIONS(2386),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2438),
    [anon_sym_GT] = ACTIONS(2441),
    [anon_sym_BANG] = ACTIONS(2444),
    [anon_sym_DQUOTE] = ACTIONS(2441),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_DOLLAR] = ACTIONS(2441),
    [anon_sym_PERCENT] = ACTIONS(2441),
    [anon_sym_AMP] = ACTIONS(2447),
    [anon_sym_SQUOTE] = ACTIONS(2441),
    [anon_sym_STAR] = ACTIONS(2441),
    [anon_sym_PLUS] = ACTIONS(2441),
    [anon_sym_COMMA] = ACTIONS(2441),
    [anon_sym_DASH] = ACTIONS(2441),
    [anon_sym_DOT] = ACTIONS(2441),
    [anon_sym_SLASH] = ACTIONS(2441),
    [anon_sym_COLON] = ACTIONS(2441),
    [anon_sym_SEMI] = ACTIONS(2441),
    [anon_sym_EQ] = ACTIONS(2441),
    [anon_sym_QMARK] = ACTIONS(2441),
    [anon_sym_AT] = ACTIONS(2441),
    [anon_sym_BSLASH] = ACTIONS(2450),
    [anon_sym_CARET] = ACTIONS(2441),
    [anon_sym__] = ACTIONS(2441),
    [anon_sym_BQUOTE] = ACTIONS(2441),
    [anon_sym_LBRACE] = ACTIONS(2441),
    [anon_sym_PIPE] = ACTIONS(2441),
    [anon_sym_RBRACE] = ACTIONS(2441),
    [anon_sym_TILDE] = ACTIONS(2441),
    [anon_sym_LPAREN] = ACTIONS(2441),
    [anon_sym_RPAREN] = ACTIONS(2441),
    [sym__newline_token] = ACTIONS(2453),
    [sym_uri_autolink] = ACTIONS(2435),
    [sym_email_autolink] = ACTIONS(2435),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2456),
    [anon_sym_LT_QMARK] = ACTIONS(2459),
    [aux_sym__declaration_token1] = ACTIONS(2462),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2465),
    [sym__whitespace_ge_2] = ACTIONS(2468),
    [aux_sym__whitespace_token1] = ACTIONS(2471),
    [sym__word_no_digit] = ACTIONS(2435),
    [sym__digits] = ACTIONS(2435),
    [sym__code_span_start] = ACTIONS(2474),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__emphasis_close_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2477),
    [sym__unclosed_span] = ACTIONS(2435),
  },
  [STATE(152)] = {
    [sym_backslash_escape] = STATE(152),
    [sym_code_span] = STATE(152),
    [sym_latex_block] = STATE(152),
    [sym_image] = STATE(152),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(152),
    [sym__whitespace] = STATE(152),
    [sym__word] = STATE(152),
    [sym__soft_line_break] = STATE(152),
    [sym__text_base] = STATE(152),
    [aux_sym__inline_base_repeat1] = STATE(152),
    [ts_builtin_sym_end] = ACTIONS(2386),
    [sym__backslash_escape] = ACTIONS(2480),
    [sym_entity_reference] = ACTIONS(2483),
    [sym_numeric_character_reference] = ACTIONS(2483),
    [anon_sym_LBRACK] = ACTIONS(2386),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2486),
    [anon_sym_GT] = ACTIONS(2489),
    [anon_sym_BANG] = ACTIONS(2492),
    [anon_sym_DQUOTE] = ACTIONS(2489),
    [anon_sym_POUND] = ACTIONS(2489),
    [anon_sym_DOLLAR] = ACTIONS(2489),
    [anon_sym_PERCENT] = ACTIONS(2489),
    [anon_sym_AMP] = ACTIONS(2495),
    [anon_sym_SQUOTE] = ACTIONS(2489),
    [anon_sym_STAR] = ACTIONS(2489),
    [anon_sym_PLUS] = ACTIONS(2489),
    [anon_sym_COMMA] = ACTIONS(2489),
    [anon_sym_DASH] = ACTIONS(2489),
    [anon_sym_DOT] = ACTIONS(2489),
    [anon_sym_SLASH] = ACTIONS(2489),
    [anon_sym_COLON] = ACTIONS(2489),
    [anon_sym_SEMI] = ACTIONS(2489),
    [anon_sym_EQ] = ACTIONS(2489),
    [anon_sym_QMARK] = ACTIONS(2489),
    [anon_sym_AT] = ACTIONS(2489),
    [anon_sym_BSLASH] = ACTIONS(2498),
    [anon_sym_CARET] = ACTIONS(2489),
    [anon_sym__] = ACTIONS(2489),
    [anon_sym_BQUOTE] = ACTIONS(2489),
    [anon_sym_LBRACE] = ACTIONS(2489),
    [anon_sym_PIPE] = ACTIONS(2489),
    [anon_sym_RBRACE] = ACTIONS(2489),
    [anon_sym_TILDE] = ACTIONS(2489),
    [anon_sym_LPAREN] = ACTIONS(2489),
    [anon_sym_RPAREN] = ACTIONS(2489),
    [sym__newline_token] = ACTIONS(2501),
    [sym_uri_autolink] = ACTIONS(2483),
    [sym_email_autolink] = ACTIONS(2483),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2504),
    [anon_sym_LT_QMARK] = ACTIONS(2507),
    [aux_sym__declaration_token1] = ACTIONS(2510),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2513),
    [sym__whitespace_ge_2] = ACTIONS(2516),
    [aux_sym__whitespace_token1] = ACTIONS(2519),
    [sym__word_no_digit] = ACTIONS(2483),
    [sym__digits] = ACTIONS(2483),
    [sym__code_span_start] = ACTIONS(2522),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2525),
    [sym__unclosed_span] = ACTIONS(2483),
  },
  [STATE(153)] = {
    [sym_backslash_escape] = STATE(153),
    [sym_code_span] = STATE(153),
    [sym_latex_block] = STATE(153),
    [sym_image] = STATE(153),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(153),
    [sym__whitespace] = STATE(153),
    [sym__word] = STATE(153),
    [sym__soft_line_break] = STATE(153),
    [sym__text_base] = STATE(153),
    [aux_sym__inline_base_repeat1] = STATE(153),
    [sym__backslash_escape] = ACTIONS(2528),
    [sym_entity_reference] = ACTIONS(2531),
    [sym_numeric_character_reference] = ACTIONS(2531),
    [anon_sym_LBRACK] = ACTIONS(2386),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2537),
    [anon_sym_BANG] = ACTIONS(2540),
    [anon_sym_DQUOTE] = ACTIONS(2537),
    [anon_sym_POUND] = ACTIONS(2537),
    [anon_sym_DOLLAR] = ACTIONS(2537),
    [anon_sym_PERCENT] = ACTIONS(2537),
    [anon_sym_AMP] = ACTIONS(2543),
    [anon_sym_SQUOTE] = ACTIONS(2537),
    [anon_sym_STAR] = ACTIONS(2537),
    [anon_sym_PLUS] = ACTIONS(2537),
    [anon_sym_COMMA] = ACTIONS(2537),
    [anon_sym_DASH] = ACTIONS(2537),
    [anon_sym_DOT] = ACTIONS(2537),
    [anon_sym_SLASH] = ACTIONS(2537),
    [anon_sym_COLON] = ACTIONS(2537),
    [anon_sym_SEMI] = ACTIONS(2537),
    [anon_sym_EQ] = ACTIONS(2537),
    [anon_sym_QMARK] = ACTIONS(2537),
    [anon_sym_AT] = ACTIONS(2537),
    [anon_sym_BSLASH] = ACTIONS(2546),
    [anon_sym_CARET] = ACTIONS(2537),
    [anon_sym__] = ACTIONS(2537),
    [anon_sym_BQUOTE] = ACTIONS(2537),
    [anon_sym_LBRACE] = ACTIONS(2537),
    [anon_sym_PIPE] = ACTIONS(2537),
    [anon_sym_RBRACE] = ACTIONS(2537),
    [anon_sym_TILDE] = ACTIONS(2537),
    [anon_sym_LPAREN] = ACTIONS(2537),
    [anon_sym_RPAREN] = ACTIONS(2537),
    [sym__newline_token] = ACTIONS(2549),
    [sym_uri_autolink] = ACTIONS(2531),
    [sym_email_autolink] = ACTIONS(2531),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2552),
    [anon_sym_LT_QMARK] = ACTIONS(2555),
    [aux_sym__declaration_token1] = ACTIONS(2558),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2561),
    [sym__whitespace_ge_2] = ACTIONS(2564),
    [aux_sym__whitespace_token1] = ACTIONS(2567),
    [sym__word_no_digit] = ACTIONS(2531),
    [sym__digits] = ACTIONS(2531),
    [sym__code_span_start] = ACTIONS(2570),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__strikethrough_close] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2573),
    [sym__unclosed_span] = ACTIONS(2531),
  },
  [STATE(154)] = {
    [sym_backslash_escape] = STATE(153),
    [sym_code_span] = STATE(153),
    [sym_latex_block] = STATE(153),
    [sym_image] = STATE(153),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(290),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(153),
    [sym__whitespace] = STATE(153),
    [sym__word] = STATE(153),
    [sym__soft_line_break] = STATE(153),
    [sym__text_base] = STATE(153),
    [aux_sym__inline_base_repeat1] = STATE(153),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2576),
    [sym_numeric_character_reference] = ACTIONS(2576),
    [anon_sym_LBRACK] = ACTIONS(2378),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2576),
    [sym_email_autolink] = ACTIONS(2576),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(754),
    [anon_sym_LT_QMARK] = ACTIONS(756),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(760),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2576),
    [sym__digits] = ACTIONS(2576),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__strikethrough_close] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2576),
  },
  [STATE(155)] = {
    [sym_backslash_escape] = STATE(152),
    [sym_code_span] = STATE(152),
    [sym_latex_block] = STATE(152),
    [sym_image] = STATE(152),
    [sym__image_inline_link] = STATE(471),
    [sym__image_shortcut_link] = STATE(471),
    [sym__image_full_reference_link] = STATE(471),
    [sym__image_collapsed_reference_link] = STATE(471),
    [sym__image_description] = STATE(1122),
    [sym__image_description_non_empty] = STATE(478),
    [sym__html_tag] = STATE(505),
    [sym__open_tag] = STATE(522),
    [sym__closing_tag] = STATE(522),
    [sym__html_comment] = STATE(522),
    [sym__processing_instruction] = STATE(522),
    [sym__declaration] = STATE(522),
    [sym__cdata_section] = STATE(522),
    [sym_hard_line_break] = STATE(152),
    [sym__whitespace] = STATE(152),
    [sym__word] = STATE(152),
    [sym__soft_line_break] = STATE(152),
    [sym__text_base] = STATE(152),
    [aux_sym__inline_base_repeat1] = STATE(152),
    [ts_builtin_sym_end] = ACTIONS(2378),
    [sym__backslash_escape] = ACTIONS(3),
    [sym_entity_reference] = ACTIONS(2578),
    [sym_numeric_character_reference] = ACTIONS(2578),
    [anon_sym_LBRACK] = ACTIONS(2378),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(11),
    [anon_sym_GT] = ACTIONS(13),
    [anon_sym_BANG] = ACTIONS(15),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_POUND] = ACTIONS(13),
    [anon_sym_DOLLAR] = ACTIONS(13),
    [anon_sym_PERCENT] = ACTIONS(13),
    [anon_sym_AMP] = ACTIONS(17),
    [anon_sym_SQUOTE] = ACTIONS(13),
    [anon_sym_STAR] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(13),
    [anon_sym_COMMA] = ACTIONS(13),
    [anon_sym_DASH] = ACTIONS(13),
    [anon_sym_DOT] = ACTIONS(13),
    [anon_sym_SLASH] = ACTIONS(13),
    [anon_sym_COLON] = ACTIONS(13),
    [anon_sym_SEMI] = ACTIONS(13),
    [anon_sym_EQ] = ACTIONS(13),
    [anon_sym_QMARK] = ACTIONS(13),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_BSLASH] = ACTIONS(19),
    [anon_sym_CARET] = ACTIONS(13),
    [anon_sym__] = ACTIONS(13),
    [anon_sym_BQUOTE] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(13),
    [anon_sym_PIPE] = ACTIONS(13),
    [anon_sym_RBRACE] = ACTIONS(13),
    [anon_sym_TILDE] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(13),
    [sym__newline_token] = ACTIONS(21),
    [sym_uri_autolink] = ACTIONS(2578),
    [sym_email_autolink] = ACTIONS(2578),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(23),
    [anon_sym_LT_QMARK] = ACTIONS(25),
    [aux_sym__declaration_token1] = ACTIONS(27),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(29),
    [sym__whitespace_ge_2] = ACTIONS(31),
    [aux_sym__whitespace_token1] = ACTIONS(33),
    [sym__word_no_digit] = ACTIONS(2578),
    [sym__digits] = ACTIONS(2578),
    [sym__code_span_start] = ACTIONS(35),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(45),
    [sym__unclosed_span] = ACTIONS(2578),
  },
  [STATE(156)] = {
    [sym_backslash_escape] = STATE(157),
    [sym_code_span] = STATE(157),
    [sym_latex_block] = STATE(157),
    [sym_image] = STATE(157),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(157),
    [sym__whitespace] = STATE(157),
    [sym__word] = STATE(157),
    [sym__soft_line_break] = STATE(157),
    [sym__text_base] = STATE(157),
    [aux_sym__inline_base_repeat1] = STATE(157),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2580),
    [sym_numeric_character_reference] = ACTIONS(2580),
    [anon_sym_LBRACK] = ACTIONS(2378),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2580),
    [sym_email_autolink] = ACTIONS(2580),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(1122),
    [anon_sym_LT_QMARK] = ACTIONS(1124),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(1128),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2580),
    [sym__digits] = ACTIONS(2580),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2580),
  },
  [STATE(157)] = {
    [sym_backslash_escape] = STATE(157),
    [sym_code_span] = STATE(157),
    [sym_latex_block] = STATE(157),
    [sym_image] = STATE(157),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(564),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(157),
    [sym__whitespace] = STATE(157),
    [sym__word] = STATE(157),
    [sym__soft_line_break] = STATE(157),
    [sym__text_base] = STATE(157),
    [aux_sym__inline_base_repeat1] = STATE(157),
    [sym__backslash_escape] = ACTIONS(2582),
    [sym_entity_reference] = ACTIONS(2585),
    [sym_numeric_character_reference] = ACTIONS(2585),
    [anon_sym_LBRACK] = ACTIONS(2386),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2588),
    [anon_sym_GT] = ACTIONS(2591),
    [anon_sym_BANG] = ACTIONS(2594),
    [anon_sym_DQUOTE] = ACTIONS(2591),
    [anon_sym_POUND] = ACTIONS(2591),
    [anon_sym_DOLLAR] = ACTIONS(2591),
    [anon_sym_PERCENT] = ACTIONS(2591),
    [anon_sym_AMP] = ACTIONS(2597),
    [anon_sym_SQUOTE] = ACTIONS(2591),
    [anon_sym_STAR] = ACTIONS(2591),
    [anon_sym_PLUS] = ACTIONS(2591),
    [anon_sym_COMMA] = ACTIONS(2591),
    [anon_sym_DASH] = ACTIONS(2591),
    [anon_sym_DOT] = ACTIONS(2591),
    [anon_sym_SLASH] = ACTIONS(2591),
    [anon_sym_COLON] = ACTIONS(2591),
    [anon_sym_SEMI] = ACTIONS(2591),
    [anon_sym_EQ] = ACTIONS(2591),
    [anon_sym_QMARK] = ACTIONS(2591),
    [anon_sym_AT] = ACTIONS(2591),
    [anon_sym_BSLASH] = ACTIONS(2600),
    [anon_sym_CARET] = ACTIONS(2591),
    [anon_sym__] = ACTIONS(2591),
    [anon_sym_BQUOTE] = ACTIONS(2591),
    [anon_sym_LBRACE] = ACTIONS(2591),
    [anon_sym_PIPE] = ACTIONS(2591),
    [anon_sym_RBRACE] = ACTIONS(2591),
    [anon_sym_TILDE] = ACTIONS(2591),
    [anon_sym_LPAREN] = ACTIONS(2591),
    [anon_sym_RPAREN] = ACTIONS(2591),
    [sym__newline_token] = ACTIONS(2603),
    [sym_uri_autolink] = ACTIONS(2585),
    [sym_email_autolink] = ACTIONS(2585),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2606),
    [anon_sym_LT_QMARK] = ACTIONS(2609),
    [aux_sym__declaration_token1] = ACTIONS(2612),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2615),
    [sym__whitespace_ge_2] = ACTIONS(2618),
    [aux_sym__whitespace_token1] = ACTIONS(2621),
    [sym__word_no_digit] = ACTIONS(2585),
    [sym__digits] = ACTIONS(2585),
    [sym__code_span_start] = ACTIONS(2624),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2627),
    [sym__unclosed_span] = ACTIONS(2585),
  },
  [STATE(158)] = {
    [sym_backslash_escape] = STATE(158),
    [sym_code_span] = STATE(158),
    [sym_latex_block] = STATE(158),
    [sym_image] = STATE(158),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(158),
    [sym__whitespace] = STATE(158),
    [sym__word] = STATE(158),
    [sym__soft_line_break] = STATE(158),
    [sym__text_base] = STATE(158),
    [aux_sym__inline_base_repeat1] = STATE(158),
    [sym__backslash_escape] = ACTIONS(2432),
    [sym_entity_reference] = ACTIONS(2630),
    [sym_numeric_character_reference] = ACTIONS(2630),
    [anon_sym_LT] = ACTIONS(2438),
    [anon_sym_GT] = ACTIONS(2441),
    [anon_sym_BANG] = ACTIONS(2633),
    [anon_sym_DQUOTE] = ACTIONS(2441),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_DOLLAR] = ACTIONS(2441),
    [anon_sym_PERCENT] = ACTIONS(2441),
    [anon_sym_AMP] = ACTIONS(2447),
    [anon_sym_SQUOTE] = ACTIONS(2441),
    [anon_sym_STAR] = ACTIONS(2441),
    [anon_sym_PLUS] = ACTIONS(2441),
    [anon_sym_COMMA] = ACTIONS(2441),
    [anon_sym_DASH] = ACTIONS(2441),
    [anon_sym_DOT] = ACTIONS(2441),
    [anon_sym_SLASH] = ACTIONS(2441),
    [anon_sym_COLON] = ACTIONS(2441),
    [anon_sym_SEMI] = ACTIONS(2441),
    [anon_sym_EQ] = ACTIONS(2441),
    [anon_sym_QMARK] = ACTIONS(2441),
    [anon_sym_AT] = ACTIONS(2441),
    [anon_sym_BSLASH] = ACTIONS(2450),
    [anon_sym_CARET] = ACTIONS(2441),
    [anon_sym__] = ACTIONS(2441),
    [anon_sym_BQUOTE] = ACTIONS(2441),
    [anon_sym_LBRACE] = ACTIONS(2441),
    [anon_sym_PIPE] = ACTIONS(2441),
    [anon_sym_RBRACE] = ACTIONS(2441),
    [anon_sym_TILDE] = ACTIONS(2441),
    [anon_sym_LPAREN] = ACTIONS(2441),
    [anon_sym_RPAREN] = ACTIONS(2441),
    [sym__newline_token] = ACTIONS(2453),
    [sym_uri_autolink] = ACTIONS(2630),
    [sym_email_autolink] = ACTIONS(2630),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2636),
    [anon_sym_LT_QMARK] = ACTIONS(2639),
    [aux_sym__declaration_token1] = ACTIONS(2462),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2642),
    [sym__whitespace_ge_2] = ACTIONS(2468),
    [aux_sym__whitespace_token1] = ACTIONS(2471),
    [sym__word_no_digit] = ACTIONS(2630),
    [sym__digits] = ACTIONS(2630),
    [sym__code_span_start] = ACTIONS(2474),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__emphasis_close_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2477),
    [sym__unclosed_span] = ACTIONS(2630),
  },
  [STATE(159)] = {
    [sym_backslash_escape] = STATE(158),
    [sym_code_span] = STATE(158),
    [sym_latex_block] = STATE(158),
    [sym_image] = STATE(158),
    [sym__image_inline_link] = STATE(450),
    [sym__image_shortcut_link] = STATE(450),
    [sym__image_full_reference_link] = STATE(450),
    [sym__image_collapsed_reference_link] = STATE(450),
    [sym__image_description] = STATE(1119),
    [sym__image_description_non_empty] = STATE(576),
    [sym__html_tag] = STATE(456),
    [sym__open_tag] = STATE(292),
    [sym__closing_tag] = STATE(292),
    [sym__html_comment] = STATE(292),
    [sym__processing_instruction] = STATE(292),
    [sym__declaration] = STATE(292),
    [sym__cdata_section] = STATE(292),
    [sym_hard_line_break] = STATE(158),
    [sym__whitespace] = STATE(158),
    [sym__word] = STATE(158),
    [sym__soft_line_break] = STATE(158),
    [sym__text_base] = STATE(158),
    [aux_sym__inline_base_repeat1] = STATE(158),
    [sym__backslash_escape] = ACTIONS(684),
    [sym_entity_reference] = ACTIONS(2645),
    [sym_numeric_character_reference] = ACTIONS(2645),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(2014),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_POUND] = ACTIONS(694),
    [anon_sym_DOLLAR] = ACTIONS(694),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(698),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_STAR] = ACTIONS(694),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_DOT] = ACTIONS(694),
    [anon_sym_SLASH] = ACTIONS(694),
    [anon_sym_COLON] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_EQ] = ACTIONS(694),
    [anon_sym_QMARK] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_BSLASH] = ACTIONS(700),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym__] = ACTIONS(694),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_RPAREN] = ACTIONS(694),
    [sym__newline_token] = ACTIONS(702),
    [sym_uri_autolink] = ACTIONS(2645),
    [sym_email_autolink] = ACTIONS(2645),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2016),
    [anon_sym_LT_QMARK] = ACTIONS(2018),
    [aux_sym__declaration_token1] = ACTIONS(708),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2020),
    [sym__whitespace_ge_2] = ACTIONS(712),
    [aux_sym__whitespace_token1] = ACTIONS(714),
    [sym__word_no_digit] = ACTIONS(2645),
    [sym__digits] = ACTIONS(2645),
    [sym__code_span_start] = ACTIONS(716),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__emphasis_close_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(726),
    [sym__unclosed_span] = ACTIONS(2645),
  },
  [STATE(160)] = {
    [sym_backslash_escape] = STATE(160),
    [sym_code_span] = STATE(160),
    [sym_latex_block] = STATE(160),
    [sym_image] = STATE(160),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(160),
    [sym__whitespace] = STATE(160),
    [sym__word] = STATE(160),
    [sym__soft_line_break] = STATE(160),
    [sym__text_base] = STATE(160),
    [aux_sym__inline_base_repeat1] = STATE(160),
    [sym__backslash_escape] = ACTIONS(2528),
    [sym_entity_reference] = ACTIONS(2647),
    [sym_numeric_character_reference] = ACTIONS(2647),
    [anon_sym_LT] = ACTIONS(2534),
    [anon_sym_GT] = ACTIONS(2537),
    [anon_sym_BANG] = ACTIONS(2650),
    [anon_sym_DQUOTE] = ACTIONS(2537),
    [anon_sym_POUND] = ACTIONS(2537),
    [anon_sym_DOLLAR] = ACTIONS(2537),
    [anon_sym_PERCENT] = ACTIONS(2537),
    [anon_sym_AMP] = ACTIONS(2543),
    [anon_sym_SQUOTE] = ACTIONS(2537),
    [anon_sym_STAR] = ACTIONS(2537),
    [anon_sym_PLUS] = ACTIONS(2537),
    [anon_sym_COMMA] = ACTIONS(2537),
    [anon_sym_DASH] = ACTIONS(2537),
    [anon_sym_DOT] = ACTIONS(2537),
    [anon_sym_SLASH] = ACTIONS(2537),
    [anon_sym_COLON] = ACTIONS(2537),
    [anon_sym_SEMI] = ACTIONS(2537),
    [anon_sym_EQ] = ACTIONS(2537),
    [anon_sym_QMARK] = ACTIONS(2537),
    [anon_sym_AT] = ACTIONS(2537),
    [anon_sym_BSLASH] = ACTIONS(2546),
    [anon_sym_CARET] = ACTIONS(2537),
    [anon_sym__] = ACTIONS(2537),
    [anon_sym_BQUOTE] = ACTIONS(2537),
    [anon_sym_LBRACE] = ACTIONS(2537),
    [anon_sym_PIPE] = ACTIONS(2537),
    [anon_sym_RBRACE] = ACTIONS(2537),
    [anon_sym_TILDE] = ACTIONS(2537),
    [anon_sym_LPAREN] = ACTIONS(2537),
    [anon_sym_RPAREN] = ACTIONS(2537),
    [sym__newline_token] = ACTIONS(2549),
    [sym_uri_autolink] = ACTIONS(2647),
    [sym_email_autolink] = ACTIONS(2647),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2653),
    [anon_sym_LT_QMARK] = ACTIONS(2656),
    [aux_sym__declaration_token1] = ACTIONS(2558),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2659),
    [sym__whitespace_ge_2] = ACTIONS(2564),
    [aux_sym__whitespace_token1] = ACTIONS(2567),
    [sym__word_no_digit] = ACTIONS(2647),
    [sym__digits] = ACTIONS(2647),
    [sym__code_span_start] = ACTIONS(2570),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__strikethrough_close] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2573),
    [sym__unclosed_span] = ACTIONS(2647),
  },
  [STATE(161)] = {
    [sym_backslash_escape] = STATE(161),
    [sym_code_span] = STATE(161),
    [sym_latex_block] = STATE(161),
    [sym_image] = STATE(161),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(161),
    [sym__whitespace] = STATE(161),
    [sym__word] = STATE(161),
    [sym__soft_line_break] = STATE(161),
    [sym__text_base] = STATE(161),
    [aux_sym__inline_base_repeat1] = STATE(161),
    [sym__backslash_escape] = ACTIONS(2380),
    [sym_entity_reference] = ACTIONS(2662),
    [sym_numeric_character_reference] = ACTIONS(2662),
    [anon_sym_LT] = ACTIONS(2388),
    [anon_sym_GT] = ACTIONS(2391),
    [anon_sym_BANG] = ACTIONS(2665),
    [anon_sym_DQUOTE] = ACTIONS(2391),
    [anon_sym_POUND] = ACTIONS(2391),
    [anon_sym_DOLLAR] = ACTIONS(2391),
    [anon_sym_PERCENT] = ACTIONS(2391),
    [anon_sym_AMP] = ACTIONS(2397),
    [anon_sym_SQUOTE] = ACTIONS(2391),
    [anon_sym_STAR] = ACTIONS(2391),
    [anon_sym_PLUS] = ACTIONS(2391),
    [anon_sym_COMMA] = ACTIONS(2391),
    [anon_sym_DASH] = ACTIONS(2391),
    [anon_sym_DOT] = ACTIONS(2391),
    [anon_sym_SLASH] = ACTIONS(2391),
    [anon_sym_COLON] = ACTIONS(2391),
    [anon_sym_SEMI] = ACTIONS(2391),
    [anon_sym_EQ] = ACTIONS(2391),
    [anon_sym_QMARK] = ACTIONS(2391),
    [anon_sym_AT] = ACTIONS(2391),
    [anon_sym_BSLASH] = ACTIONS(2400),
    [anon_sym_CARET] = ACTIONS(2391),
    [anon_sym__] = ACTIONS(2391),
    [anon_sym_BQUOTE] = ACTIONS(2391),
    [anon_sym_LBRACE] = ACTIONS(2391),
    [anon_sym_PIPE] = ACTIONS(2391),
    [anon_sym_RBRACE] = ACTIONS(2391),
    [anon_sym_TILDE] = ACTIONS(2391),
    [anon_sym_LPAREN] = ACTIONS(2391),
    [anon_sym_RPAREN] = ACTIONS(2391),
    [sym__newline_token] = ACTIONS(2403),
    [sym_uri_autolink] = ACTIONS(2662),
    [sym_email_autolink] = ACTIONS(2662),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2668),
    [anon_sym_LT_QMARK] = ACTIONS(2671),
    [aux_sym__declaration_token1] = ACTIONS(2412),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2674),
    [sym__whitespace_ge_2] = ACTIONS(2418),
    [aux_sym__whitespace_token1] = ACTIONS(2421),
    [sym__word_no_digit] = ACTIONS(2662),
    [sym__digits] = ACTIONS(2662),
    [sym__code_span_start] = ACTIONS(2424),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__emphasis_close_star] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2427),
    [sym__unclosed_span] = ACTIONS(2662),
  },
  [STATE(162)] = {
    [sym_backslash_escape] = STATE(161),
    [sym_code_span] = STATE(161),
    [sym_latex_block] = STATE(161),
    [sym_image] = STATE(161),
    [sym__image_inline_link] = STATE(384),
    [sym__image_shortcut_link] = STATE(384),
    [sym__image_full_reference_link] = STATE(384),
    [sym__image_collapsed_reference_link] = STATE(384),
    [sym__image_description] = STATE(1125),
    [sym__image_description_non_empty] = STATE(575),
    [sym__html_tag] = STATE(386),
    [sym__open_tag] = STATE(457),
    [sym__closing_tag] = STATE(457),
    [sym__html_comment] = STATE(457),
    [sym__processing_instruction] = STATE(457),
    [sym__declaration] = STATE(457),
    [sym__cdata_section] = STATE(457),
    [sym_hard_line_break] = STATE(161),
    [sym__whitespace] = STATE(161),
    [sym__word] = STATE(161),
    [sym__soft_line_break] = STATE(161),
    [sym__text_base] = STATE(161),
    [aux_sym__inline_base_repeat1] = STATE(161),
    [sym__backslash_escape] = ACTIONS(640),
    [sym_entity_reference] = ACTIONS(2677),
    [sym_numeric_character_reference] = ACTIONS(2677),
    [anon_sym_LT] = ACTIONS(648),
    [anon_sym_GT] = ACTIONS(650),
    [anon_sym_BANG] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(650),
    [anon_sym_POUND] = ACTIONS(650),
    [anon_sym_DOLLAR] = ACTIONS(650),
    [anon_sym_PERCENT] = ACTIONS(650),
    [anon_sym_AMP] = ACTIONS(654),
    [anon_sym_SQUOTE] = ACTIONS(650),
    [anon_sym_STAR] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(650),
    [anon_sym_COMMA] = ACTIONS(650),
    [anon_sym_DASH] = ACTIONS(650),
    [anon_sym_DOT] = ACTIONS(650),
    [anon_sym_SLASH] = ACTIONS(650),
    [anon_sym_COLON] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(650),
    [anon_sym_EQ] = ACTIONS(650),
    [anon_sym_QMARK] = ACTIONS(650),
    [anon_sym_AT] = ACTIONS(650),
    [anon_sym_BSLASH] = ACTIONS(656),
    [anon_sym_CARET] = ACTIONS(650),
    [anon_sym__] = ACTIONS(650),
    [anon_sym_BQUOTE] = ACTIONS(650),
    [anon_sym_LBRACE] = ACTIONS(650),
    [anon_sym_PIPE] = ACTIONS(650),
    [anon_sym_RBRACE] = ACTIONS(650),
    [anon_sym_TILDE] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(650),
    [sym__newline_token] = ACTIONS(658),
    [sym_uri_autolink] = ACTIONS(2677),
    [sym_email_autolink] = ACTIONS(2677),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2111),
    [anon_sym_LT_QMARK] = ACTIONS(2113),
    [aux_sym__declaration_token1] = ACTIONS(664),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2115),
    [sym__whitespace_ge_2] = ACTIONS(668),
    [aux_sym__whitespace_token1] = ACTIONS(670),
    [sym__word_no_digit] = ACTIONS(2677),
    [sym__digits] = ACTIONS(2677),
    [sym__code_span_start] = ACTIONS(672),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__emphasis_close_star] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(682),
    [sym__unclosed_span] = ACTIONS(2677),
  },
  [STATE(163)] = {
    [sym_backslash_escape] = STATE(160),
    [sym_code_span] = STATE(160),
    [sym_latex_block] = STATE(160),
    [sym_image] = STATE(160),
    [sym__image_inline_link] = STATE(289),
    [sym__image_shortcut_link] = STATE(289),
    [sym__image_full_reference_link] = STATE(289),
    [sym__image_collapsed_reference_link] = STATE(289),
    [sym__image_description] = STATE(1123),
    [sym__image_description_non_empty] = STATE(598),
    [sym__html_tag] = STATE(291),
    [sym__open_tag] = STATE(359),
    [sym__closing_tag] = STATE(359),
    [sym__html_comment] = STATE(359),
    [sym__processing_instruction] = STATE(359),
    [sym__declaration] = STATE(359),
    [sym__cdata_section] = STATE(359),
    [sym_hard_line_break] = STATE(160),
    [sym__whitespace] = STATE(160),
    [sym__word] = STATE(160),
    [sym__soft_line_break] = STATE(160),
    [sym__text_base] = STATE(160),
    [aux_sym__inline_base_repeat1] = STATE(160),
    [sym__backslash_escape] = ACTIONS(734),
    [sym_entity_reference] = ACTIONS(2679),
    [sym_numeric_character_reference] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(742),
    [anon_sym_GT] = ACTIONS(744),
    [anon_sym_BANG] = ACTIONS(2032),
    [anon_sym_DQUOTE] = ACTIONS(744),
    [anon_sym_POUND] = ACTIONS(744),
    [anon_sym_DOLLAR] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_SQUOTE] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(744),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_COMMA] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_DOT] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_COLON] = ACTIONS(744),
    [anon_sym_SEMI] = ACTIONS(744),
    [anon_sym_EQ] = ACTIONS(744),
    [anon_sym_QMARK] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(744),
    [anon_sym_BSLASH] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(744),
    [anon_sym__] = ACTIONS(744),
    [anon_sym_BQUOTE] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(744),
    [anon_sym_PIPE] = ACTIONS(744),
    [anon_sym_RBRACE] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(744),
    [anon_sym_RPAREN] = ACTIONS(744),
    [sym__newline_token] = ACTIONS(752),
    [sym_uri_autolink] = ACTIONS(2679),
    [sym_email_autolink] = ACTIONS(2679),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2034),
    [anon_sym_LT_QMARK] = ACTIONS(2036),
    [aux_sym__declaration_token1] = ACTIONS(758),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2038),
    [sym__whitespace_ge_2] = ACTIONS(762),
    [aux_sym__whitespace_token1] = ACTIONS(764),
    [sym__word_no_digit] = ACTIONS(2679),
    [sym__digits] = ACTIONS(2679),
    [sym__code_span_start] = ACTIONS(766),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__strikethrough_close] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(776),
    [sym__unclosed_span] = ACTIONS(2679),
  },
  [STATE(164)] = {
    [sym_backslash_escape] = STATE(164),
    [sym_code_span] = STATE(164),
    [sym_latex_block] = STATE(164),
    [sym_image] = STATE(164),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(164),
    [sym__whitespace] = STATE(164),
    [sym__word] = STATE(164),
    [sym__soft_line_break] = STATE(164),
    [sym__text_base] = STATE(164),
    [aux_sym__inline_base_repeat1] = STATE(164),
    [sym__backslash_escape] = ACTIONS(2582),
    [sym_entity_reference] = ACTIONS(2681),
    [sym_numeric_character_reference] = ACTIONS(2681),
    [anon_sym_RBRACK] = ACTIONS(2386),
    [anon_sym_LT] = ACTIONS(2588),
    [anon_sym_GT] = ACTIONS(2591),
    [anon_sym_BANG] = ACTIONS(2684),
    [anon_sym_DQUOTE] = ACTIONS(2591),
    [anon_sym_POUND] = ACTIONS(2591),
    [anon_sym_DOLLAR] = ACTIONS(2591),
    [anon_sym_PERCENT] = ACTIONS(2591),
    [anon_sym_AMP] = ACTIONS(2597),
    [anon_sym_SQUOTE] = ACTIONS(2591),
    [anon_sym_STAR] = ACTIONS(2591),
    [anon_sym_PLUS] = ACTIONS(2591),
    [anon_sym_COMMA] = ACTIONS(2591),
    [anon_sym_DASH] = ACTIONS(2591),
    [anon_sym_DOT] = ACTIONS(2591),
    [anon_sym_SLASH] = ACTIONS(2591),
    [anon_sym_COLON] = ACTIONS(2591),
    [anon_sym_SEMI] = ACTIONS(2591),
    [anon_sym_EQ] = ACTIONS(2591),
    [anon_sym_QMARK] = ACTIONS(2591),
    [anon_sym_AT] = ACTIONS(2591),
    [anon_sym_BSLASH] = ACTIONS(2600),
    [anon_sym_CARET] = ACTIONS(2591),
    [anon_sym__] = ACTIONS(2591),
    [anon_sym_BQUOTE] = ACTIONS(2591),
    [anon_sym_LBRACE] = ACTIONS(2591),
    [anon_sym_PIPE] = ACTIONS(2591),
    [anon_sym_RBRACE] = ACTIONS(2591),
    [anon_sym_TILDE] = ACTIONS(2591),
    [anon_sym_LPAREN] = ACTIONS(2591),
    [anon_sym_RPAREN] = ACTIONS(2591),
    [sym__newline_token] = ACTIONS(2603),
    [sym_uri_autolink] = ACTIONS(2681),
    [sym_email_autolink] = ACTIONS(2681),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_QMARK] = ACTIONS(2690),
    [aux_sym__declaration_token1] = ACTIONS(2612),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2693),
    [sym__whitespace_ge_2] = ACTIONS(2618),
    [aux_sym__whitespace_token1] = ACTIONS(2621),
    [sym__word_no_digit] = ACTIONS(2681),
    [sym__digits] = ACTIONS(2681),
    [sym__code_span_start] = ACTIONS(2624),
    [sym__emphasis_open_star] = ACTIONS(2386),
    [sym__emphasis_open_underscore] = ACTIONS(2386),
    [sym__strikethrough_open] = ACTIONS(2386),
    [sym__latex_span_start] = ACTIONS(2627),
    [sym__unclosed_span] = ACTIONS(2681),
  },
  [STATE(165)] = {
    [sym_backslash_escape] = STATE(164),
    [sym_code_span] = STATE(164),
    [sym_latex_block] = STATE(164),
    [sym_image] = STATE(164),
    [sym__image_inline_link] = STATE(541),
    [sym__image_shortcut_link] = STATE(541),
    [sym__image_full_reference_link] = STATE(541),
    [sym__image_collapsed_reference_link] = STATE(541),
    [sym__image_description] = STATE(1118),
    [sym__image_description_non_empty] = STATE(536),
    [sym__html_tag] = STATE(542),
    [sym__open_tag] = STATE(577),
    [sym__closing_tag] = STATE(577),
    [sym__html_comment] = STATE(577),
    [sym__processing_instruction] = STATE(577),
    [sym__declaration] = STATE(577),
    [sym__cdata_section] = STATE(577),
    [sym_hard_line_break] = STATE(164),
    [sym__whitespace] = STATE(164),
    [sym__word] = STATE(164),
    [sym__soft_line_break] = STATE(164),
    [sym__text_base] = STATE(164),
    [aux_sym__inline_base_repeat1] = STATE(164),
    [sym__backslash_escape] = ACTIONS(1102),
    [sym_entity_reference] = ACTIONS(2696),
    [sym_numeric_character_reference] = ACTIONS(2696),
    [anon_sym_RBRACK] = ACTIONS(2378),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(1112),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_DOLLAR] = ACTIONS(1112),
    [anon_sym_PERCENT] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_SQUOTE] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_COMMA] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_DOT] = ACTIONS(1112),
    [anon_sym_SLASH] = ACTIONS(1112),
    [anon_sym_COLON] = ACTIONS(1112),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_EQ] = ACTIONS(1112),
    [anon_sym_QMARK] = ACTIONS(1112),
    [anon_sym_AT] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_CARET] = ACTIONS(1112),
    [anon_sym__] = ACTIONS(1112),
    [anon_sym_BQUOTE] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_RPAREN] = ACTIONS(1112),
    [sym__newline_token] = ACTIONS(1120),
    [sym_uri_autolink] = ACTIONS(2696),
    [sym_email_autolink] = ACTIONS(2696),
    [anon_sym_LT_BANG_DASH_DASH] = ACTIONS(2257),
    [anon_sym_LT_QMARK] = ACTIONS(2259),
    [aux_sym__declaration_token1] = ACTIONS(1126),
    [anon_sym_LT_BANG_LBRACKCDATA_LBRACK] = ACTIONS(2261),
    [sym__whitespace_ge_2] = ACTIONS(1130),
    [aux_sym__whitespace_token1] = ACTIONS(1132),
    [sym__word_no_digit] = ACTIONS(2696),
    [sym__digits] = ACTIONS(2696),
    [sym__code_span_start] = ACTIONS(1134),
    [sym__emphasis_open_star] = ACTIONS(2378),
    [sym__emphasis_open_underscore] = ACTIONS(2378),
    [sym__strikethrough_open] = ACTIONS(2378),
    [sym__latex_span_start] = ACTIONS(1142),
    [sym__unclosed_span] = ACTIONS(2696),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 11,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2713), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2700), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2721), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(786), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 26,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [81] = 10,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2724), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(789), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [160] = 12,
    ACTIONS(2729), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2742), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2750), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(787), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [243] = 11,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2753), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2700), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2755), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(764), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 26,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [324] = 12,
    ACTIONS(2729), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2758), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2760), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(765), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [407] = 11,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2724), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2700), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2726), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(789), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 26,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [488] = 12,
    ACTIONS(2729), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2763), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2765), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(790), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [571] = 10,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2713), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2721), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(786), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [650] = 10,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2768), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2770), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(769), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [729] = 10,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2742), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2750), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(787), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 28,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [808] = 10,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2773), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2775), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(757), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 28,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [887] = 10,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2763), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2765), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(790), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 28,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [966] = 10,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2753), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2755), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(764), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1045] = 10,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2758), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2760), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(765), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 28,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1124] = 11,
    ACTIONS(2784), 1,
      anon_sym_DASH,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2790), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2798), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2781), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(763), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2778), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1204] = 11,
    ACTIONS(2729), 1,
      anon_sym_LBRACK,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2801), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2803), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(818), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 12,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1284] = 10,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2801), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2803), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2733), 4,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(818), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 12,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 28,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1362] = 11,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2812), 1,
      anon_sym_DASH,
    ACTIONS(2815), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2817), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2809), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(785), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2806), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1442] = 12,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2812), 1,
      anon_sym_DASH,
    ACTIONS(2815), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2817), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2820), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2809), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(785), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2806), 25,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1524] = 11,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2828), 1,
      anon_sym_DASH,
    ACTIONS(2831), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2833), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2825), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(825), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2822), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1604] = 12,
    ACTIONS(2784), 1,
      anon_sym_DASH,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2790), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2798), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2836), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2781), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(763), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2778), 25,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1686] = 12,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2846), 1,
      anon_sym_DASH,
    ACTIONS(2849), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2838), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2851), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2840), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(827), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2843), 25,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1768] = 11,
    ACTIONS(2700), 1,
      anon_sym_LBRACK,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2854), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2856), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(817), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 12,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 27,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1848] = 11,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2846), 1,
      anon_sym_DASH,
    ACTIONS(2849), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2851), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2840), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(827), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 14,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2843), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [1928] = 10,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2854), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2856), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2702), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(817), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 12,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2006] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2883), 1,
      sym__latex_span_close,
    ACTIONS(2861), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(193), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2093] = 12,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2885), 1,
      anon_sym_LBRACK,
    ACTIONS(2893), 1,
      anon_sym_DASH,
    ACTIONS(2896), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2898), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2890), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(816), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2887), 26,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2174] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2903), 1,
      sym__latex_span_close,
    ACTIONS(2901), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2261] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2907), 1,
      sym__latex_span_close,
    ACTIONS(2905), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(195), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2348] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2909), 1,
      sym__latex_span_close,
    ACTIONS(2901), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2435] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2913), 1,
      sym__latex_span_close,
    ACTIONS(2911), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(198), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2522] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2917), 1,
      sym__latex_span_close,
    ACTIONS(2915), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(202), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2609] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2919), 1,
      sym__latex_span_close,
    ACTIONS(2901), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2696] = 11,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2893), 1,
      anon_sym_DASH,
    ACTIONS(2896), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2898), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2711), 3,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2890), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(816), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 13,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2887), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2775] = 15,
    ACTIONS(2921), 1,
      sym__backslash_escape,
    ACTIONS(2927), 1,
      anon_sym_LT,
    ACTIONS(2933), 1,
      anon_sym_BSLASH,
    ACTIONS(2936), 1,
      sym__newline_token,
    ACTIONS(2939), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2942), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2945), 1,
      aux_sym__declaration_token1,
    ACTIONS(2948), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2951), 1,
      sym__whitespace_ge_2,
    ACTIONS(2954), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2957), 1,
      sym__latex_span_close,
    ACTIONS(2924), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2930), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2862] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2961), 1,
      sym__latex_span_close,
    ACTIONS(2959), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(203), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [2949] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2963), 1,
      sym__latex_span_close,
    ACTIONS(2901), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [3036] = 15,
    ACTIONS(2859), 1,
      sym__backslash_escape,
    ACTIONS(2863), 1,
      anon_sym_LT,
    ACTIONS(2867), 1,
      anon_sym_BSLASH,
    ACTIONS(2869), 1,
      sym__newline_token,
    ACTIONS(2871), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(2873), 1,
      anon_sym_LT_QMARK,
    ACTIONS(2875), 1,
      aux_sym__declaration_token1,
    ACTIONS(2877), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2879), 1,
      sym__whitespace_ge_2,
    ACTIONS(2881), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2965), 1,
      sym__latex_span_close,
    ACTIONS(2901), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(722), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    STATE(200), 7,
      sym_backslash_escape,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_latex_block_repeat1,
    ACTIONS(2865), 28,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [3123] = 6,
    ACTIONS(2967), 1,
      anon_sym_SLASH,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(2973), 1,
      sym__last_token_punctuation,
    STATE(959), 1,
      sym__tag_name,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [3191] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(833), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2711), 6,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [3259] = 5,
    ACTIONS(702), 1,
      sym__newline_token,
    ACTIONS(2988), 1,
      sym__last_token_whitespace,
    STATE(463), 1,
      sym__soft_line_break,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [3325] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(836), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2711), 6,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [3393] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(2990), 1,
      anon_sym_SLASH,
    ACTIONS(2993), 1,
      sym__last_token_punctuation,
    STATE(945), 1,
      sym__tag_name,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [3461] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(2995), 1,
      anon_sym_SLASH,
    ACTIONS(2998), 1,
      sym__last_token_punctuation,
    STATE(948), 1,
      sym__tag_name,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [3529] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(3000), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      sym__last_token_punctuation,
    STATE(956), 1,
      sym__tag_name,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [3597] = 5,
    ACTIONS(21), 1,
      sym__newline_token,
    ACTIONS(3005), 1,
      sym__last_token_whitespace,
    STATE(506), 1,
      sym__soft_line_break,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [3663] = 5,
    ACTIONS(752), 1,
      sym__newline_token,
    ACTIONS(2973), 1,
      sym__last_token_punctuation,
    STATE(297), 1,
      sym__soft_line_break,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [3729] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(835), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2711), 6,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [3797] = 5,
    ACTIONS(752), 1,
      sym__newline_token,
    ACTIONS(3007), 1,
      sym__last_token_whitespace,
    STATE(297), 1,
      sym__soft_line_break,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [3863] = 5,
    ACTIONS(21), 1,
      sym__newline_token,
    ACTIONS(2998), 1,
      sym__last_token_punctuation,
    STATE(506), 1,
      sym__soft_line_break,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [3929] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(843), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2711), 6,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [3997] = 5,
    ACTIONS(702), 1,
      sym__newline_token,
    ACTIONS(3003), 1,
      sym__last_token_punctuation,
    STATE(463), 1,
      sym__soft_line_break,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4063] = 5,
    ACTIONS(658), 1,
      sym__newline_token,
    ACTIONS(2993), 1,
      sym__last_token_punctuation,
    STATE(393), 1,
      sym__soft_line_break,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4129] = 5,
    ACTIONS(658), 1,
      sym__newline_token,
    ACTIONS(3009), 1,
      sym__last_token_whitespace,
    STATE(393), 1,
      sym__soft_line_break,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4195] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3031), 1,
      sym__code_span_close,
    ACTIONS(3011), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4276] = 3,
    ACTIONS(3003), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4337] = 4,
    ACTIONS(47), 1,
      anon_sym_RBRACK,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4400] = 4,
    ACTIONS(2973), 1,
      sym__last_token_punctuation,
    ACTIONS(3037), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4463] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3040), 1,
      sym__code_span_close,
    ACTIONS(3011), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4544] = 3,
    ACTIONS(3046), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4605] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3050), 1,
      sym__code_span_close,
    ACTIONS(3048), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(228), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4686] = 4,
    ACTIONS(47), 1,
      anon_sym_RBRACK,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4749] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3052), 1,
      sym__code_span_close,
    ACTIONS(3011), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4830] = 3,
    ACTIONS(3005), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [4891] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3056), 1,
      sym__code_span_close,
    ACTIONS(3054), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(220), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4972] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3058), 1,
      sym__code_span_close,
    ACTIONS(3011), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [5053] = 3,
    ACTIONS(2998), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5114] = 3,
    ACTIONS(1339), 1,
      sym__last_token_punctuation,
    ACTIONS(3060), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(570), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5175] = 3,
    ACTIONS(1401), 1,
      sym__last_token_punctuation,
    ACTIONS(3062), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(47), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5236] = 4,
    ACTIONS(2998), 1,
      sym__last_token_punctuation,
    ACTIONS(3064), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5299] = 3,
    ACTIONS(1463), 1,
      sym__last_token_punctuation,
    ACTIONS(3067), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(433), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5360] = 3,
    ACTIONS(2988), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5421] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5482] = 4,
    ACTIONS(2993), 1,
      sym__last_token_punctuation,
    ACTIONS(3069), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5545] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3074), 1,
      sym__code_span_close,
    ACTIONS(3072), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(224), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [5626] = 3,
    ACTIONS(3076), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5687] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3078), 1,
      sym__code_span_close,
    ACTIONS(3011), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [5768] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(856), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2711), 6,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
    ACTIONS(2698), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [5835] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5896] = 3,
    ACTIONS(3080), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [5957] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3084), 1,
      sym__code_span_close,
    ACTIONS(3082), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(231), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [6038] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6099] = 3,
    ACTIONS(2973), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6160] = 5,
    ACTIONS(1120), 1,
      sym__newline_token,
    ACTIONS(3086), 1,
      sym__last_token_punctuation,
    STATE(545), 1,
      sym__soft_line_break,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6225] = 3,
    ACTIONS(3088), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6286] = 4,
    ACTIONS(47), 1,
      anon_sym_RBRACK,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6349] = 4,
    ACTIONS(3003), 1,
      sym__last_token_punctuation,
    ACTIONS(3090), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6412] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(3086), 1,
      sym__last_token_punctuation,
    ACTIONS(3093), 1,
      anon_sym_SLASH,
    STATE(946), 1,
      sym__tag_name,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [6479] = 3,
    ACTIONS(3009), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6540] = 5,
    ACTIONS(1120), 1,
      sym__newline_token,
    ACTIONS(3096), 1,
      sym__last_token_whitespace,
    STATE(545), 1,
      sym__soft_line_break,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6605] = 3,
    ACTIONS(3007), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6666] = 3,
    ACTIONS(1277), 1,
      sym__last_token_punctuation,
    ACTIONS(3098), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(236), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6727] = 3,
    ACTIONS(2993), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6788] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [6849] = 13,
    ACTIONS(3103), 1,
      anon_sym_LT,
    ACTIONS(3109), 1,
      sym__newline_token,
    ACTIONS(3112), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3115), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3118), 1,
      aux_sym__declaration_token1,
    ACTIONS(3121), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3124), 1,
      sym__whitespace_ge_2,
    ACTIONS(3127), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3130), 1,
      sym__code_span_close,
    ACTIONS(3100), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(260), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3106), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [6930] = 13,
    ACTIONS(3013), 1,
      anon_sym_LT,
    ACTIONS(3017), 1,
      sym__newline_token,
    ACTIONS(3019), 1,
      anon_sym_LT_BANG_DASH_DASH,
    ACTIONS(3021), 1,
      anon_sym_LT_QMARK,
    ACTIONS(3023), 1,
      aux_sym__declaration_token1,
    ACTIONS(3025), 1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3027), 1,
      sym__whitespace_ge_2,
    ACTIONS(3029), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3134), 1,
      sym__code_span_close,
    ACTIONS(3132), 4,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      sym__word_no_digit,
      sym__digits,
    STATE(242), 6,
      sym__html_tag,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      aux_sym_code_span_repeat1,
    STATE(779), 6,
      sym__open_tag,
      sym__closing_tag,
      sym__html_comment,
      sym__processing_instruction,
      sym__declaration,
      sym__cdata_section,
    ACTIONS(3015), 29,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7011] = 2,
    ACTIONS(3138), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3136), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7069] = 2,
    ACTIONS(3142), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7127] = 2,
    ACTIONS(3146), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7185] = 2,
    ACTIONS(3150), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7243] = 2,
    ACTIONS(3154), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3152), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7301] = 2,
    ACTIONS(3138), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3136), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7359] = 2,
    ACTIONS(3158), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7417] = 2,
    ACTIONS(3162), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7475] = 2,
    ACTIONS(3166), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3164), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7533] = 2,
    ACTIONS(3170), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3168), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7591] = 2,
    ACTIONS(3174), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3172), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7649] = 2,
    ACTIONS(3178), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3176), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7707] = 2,
    ACTIONS(3182), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3180), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7765] = 2,
    ACTIONS(3186), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3184), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7823] = 2,
    ACTIONS(3190), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7881] = 2,
    ACTIONS(3194), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7939] = 2,
    ACTIONS(3198), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3196), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [7997] = 9,
    ACTIONS(3200), 1,
      sym__backslash_escape,
    ACTIONS(3206), 1,
      anon_sym_RBRACK,
    ACTIONS(3214), 1,
      sym__newline_token,
    ACTIONS(3217), 1,
      sym__whitespace_ge_2,
    ACTIONS(3220), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3208), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3203), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3211), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [8069] = 2,
    ACTIONS(3225), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3223), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8127] = 2,
    ACTIONS(3229), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3227), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8185] = 2,
    ACTIONS(3142), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8243] = 2,
    ACTIONS(3146), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8301] = 2,
    ACTIONS(3150), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8359] = 2,
    ACTIONS(3154), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3152), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8417] = 2,
    ACTIONS(3138), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3136), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8475] = 2,
    ACTIONS(3158), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8533] = 3,
    ACTIONS(3233), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3236), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3231), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8593] = 2,
    ACTIONS(3240), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3238), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8651] = 3,
    ACTIONS(3244), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8711] = 2,
    ACTIONS(3251), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3249), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8769] = 2,
    ACTIONS(3255), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8827] = 2,
    ACTIONS(3162), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8885] = 2,
    ACTIONS(3259), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3257), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [8943] = 2,
    ACTIONS(3263), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3261), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9001] = 2,
    ACTIONS(3267), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9059] = 2,
    ACTIONS(3271), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3269), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9117] = 2,
    ACTIONS(3275), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9175] = 2,
    ACTIONS(3279), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9233] = 2,
    ACTIONS(3283), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9291] = 2,
    ACTIONS(3287), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9349] = 2,
    ACTIONS(3291), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9407] = 2,
    ACTIONS(3295), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3293), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9465] = 2,
    ACTIONS(3299), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3297), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9523] = 2,
    ACTIONS(3166), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3164), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9581] = 2,
    ACTIONS(3303), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3301), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9639] = 2,
    ACTIONS(3307), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3305), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9697] = 2,
    ACTIONS(3311), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3309), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9755] = 2,
    ACTIONS(3315), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3313), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9813] = 2,
    ACTIONS(3319), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3317), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9871] = 2,
    ACTIONS(3323), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3321), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9929] = 2,
    ACTIONS(3327), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3325), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [9987] = 2,
    ACTIONS(3331), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3329), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10045] = 2,
    ACTIONS(3335), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10103] = 2,
    ACTIONS(3339), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10161] = 2,
    ACTIONS(3343), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10219] = 2,
    ACTIONS(3347), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10277] = 2,
    ACTIONS(3351), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10335] = 2,
    ACTIONS(3355), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3353), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10393] = 2,
    ACTIONS(3359), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3357), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10451] = 2,
    ACTIONS(3363), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3361), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10509] = 2,
    ACTIONS(3367), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3365), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10567] = 2,
    ACTIONS(3371), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3369), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10625] = 2,
    ACTIONS(3375), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3373), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10683] = 2,
    ACTIONS(3379), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3377), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10741] = 2,
    ACTIONS(3383), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10799] = 2,
    ACTIONS(3387), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10857] = 2,
    ACTIONS(3170), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3168), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10915] = 2,
    ACTIONS(3391), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [10973] = 2,
    ACTIONS(3194), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11031] = 2,
    ACTIONS(3198), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3196), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11089] = 2,
    ACTIONS(3225), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3223), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11147] = 2,
    ACTIONS(3229), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3227), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11205] = 2,
    ACTIONS(3142), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11263] = 2,
    ACTIONS(3146), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11321] = 2,
    ACTIONS(3150), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11379] = 2,
    ACTIONS(3154), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3152), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11437] = 2,
    ACTIONS(3174), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3172), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11495] = 2,
    ACTIONS(3158), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11553] = 2,
    ACTIONS(3162), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11611] = 2,
    ACTIONS(3166), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3164), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11669] = 2,
    ACTIONS(3170), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3168), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11727] = 2,
    ACTIONS(3174), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3172), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11785] = 2,
    ACTIONS(3178), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3176), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11843] = 2,
    ACTIONS(3182), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3180), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11901] = 2,
    ACTIONS(3186), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3184), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [11959] = 2,
    ACTIONS(3178), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3176), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12017] = 2,
    ACTIONS(3182), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3180), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12075] = 4,
    ACTIONS(3086), 1,
      sym__last_token_punctuation,
    ACTIONS(3393), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12137] = 2,
    ACTIONS(3186), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3184), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12195] = 2,
    ACTIONS(3335), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12253] = 4,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3395), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3062), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(47), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12315] = 2,
    ACTIONS(3190), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12373] = 2,
    ACTIONS(3190), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12431] = 2,
    ACTIONS(3339), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12489] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3062), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(47), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12549] = 2,
    ACTIONS(3343), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12607] = 2,
    ACTIONS(3347), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12665] = 2,
    ACTIONS(3255), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12723] = 2,
    ACTIONS(3351), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12781] = 2,
    ACTIONS(3267), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12839] = 2,
    ACTIONS(3275), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12897] = 2,
    ACTIONS(3279), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [12955] = 2,
    ACTIONS(3283), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13013] = 2,
    ACTIONS(3287), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13071] = 2,
    ACTIONS(3291), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13129] = 3,
    ACTIONS(3233), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3236), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3231), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13189] = 2,
    ACTIONS(3335), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13247] = 2,
    ACTIONS(3339), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13305] = 2,
    ACTIONS(3343), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13363] = 2,
    ACTIONS(3347), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13421] = 2,
    ACTIONS(3351), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13479] = 2,
    ACTIONS(3383), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13537] = 2,
    ACTIONS(3387), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13595] = 3,
    ACTIONS(1471), 1,
      sym__last_token_punctuation,
    ACTIONS(3035), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3033), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13655] = 2,
    ACTIONS(3391), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13713] = 2,
    ACTIONS(3194), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13771] = 2,
    ACTIONS(3142), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13829] = 2,
    ACTIONS(3146), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13887] = 2,
    ACTIONS(3150), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [13945] = 2,
    ACTIONS(3158), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14003] = 2,
    ACTIONS(3162), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14061] = 4,
    ACTIONS(2993), 1,
      sym__last_token_punctuation,
    ACTIONS(3397), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14123] = 2,
    ACTIONS(3240), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3238), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14181] = 3,
    ACTIONS(3244), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14241] = 2,
    ACTIONS(3251), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3249), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14299] = 2,
    ACTIONS(3355), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3353), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14357] = 3,
    ACTIONS(3086), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14417] = 4,
    ACTIONS(3003), 1,
      sym__last_token_punctuation,
    ACTIONS(3399), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14479] = 3,
    ACTIONS(3401), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14539] = 2,
    ACTIONS(3259), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3257), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14597] = 2,
    ACTIONS(3263), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3261), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14655] = 2,
    ACTIONS(3271), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3269), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14713] = 2,
    ACTIONS(3295), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3293), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14771] = 3,
    ACTIONS(3096), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14831] = 2,
    ACTIONS(3299), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3297), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [14889] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3407), 1,
      anon_sym_RBRACK,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3405), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14961] = 2,
    ACTIONS(3359), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3357), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15019] = 4,
    ACTIONS(2973), 1,
      sym__last_token_punctuation,
    ACTIONS(3419), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15081] = 2,
    ACTIONS(3303), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3301), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15139] = 2,
    ACTIONS(3307), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3305), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15197] = 2,
    ACTIONS(3311), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3309), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15255] = 2,
    ACTIONS(3315), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3313), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15313] = 2,
    ACTIONS(3319), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3317), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15371] = 2,
    ACTIONS(3323), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3321), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15429] = 4,
    ACTIONS(3086), 1,
      sym__last_token_punctuation,
    ACTIONS(3421), 1,
      anon_sym_LBRACK,
    ACTIONS(2711), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15491] = 2,
    ACTIONS(3327), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3325), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15549] = 2,
    ACTIONS(3331), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3329), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15607] = 2,
    ACTIONS(3355), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3353), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15665] = 2,
    ACTIONS(3359), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3357), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15723] = 2,
    ACTIONS(3363), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3361), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15781] = 2,
    ACTIONS(3367), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3365), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15839] = 2,
    ACTIONS(3371), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3369), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15897] = 2,
    ACTIONS(3426), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3424), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [15955] = 2,
    ACTIONS(3379), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3377), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16013] = 2,
    ACTIONS(3363), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3361), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16071] = 2,
    ACTIONS(3198), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3196), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16129] = 2,
    ACTIONS(3225), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3223), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16187] = 2,
    ACTIONS(3229), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3227), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16245] = 2,
    ACTIONS(3154), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3152), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16303] = 2,
    ACTIONS(3138), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3136), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16361] = 3,
    ACTIONS(3432), 1,
      sym__emphasis_close_star,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16421] = 3,
    ACTIONS(3435), 1,
      sym__emphasis_close_star,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16481] = 3,
    ACTIONS(3442), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16541] = 2,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16599] = 2,
    ACTIONS(3166), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3164), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16657] = 2,
    ACTIONS(3170), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3168), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16715] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3447), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(437), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3445), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [16787] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3449), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(437), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3445), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [16859] = 2,
    ACTIONS(3174), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3172), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16917] = 2,
    ACTIONS(3178), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3176), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [16975] = 2,
    ACTIONS(3182), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3180), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17033] = 2,
    ACTIONS(3186), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3184), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17091] = 2,
    ACTIONS(3190), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17149] = 2,
    ACTIONS(3367), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3365), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17207] = 2,
    ACTIONS(3371), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3369), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17265] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3451), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3405), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [17337] = 2,
    ACTIONS(3375), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3373), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17395] = 2,
    ACTIONS(3379), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3377), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17453] = 2,
    ACTIONS(3327), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3325), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17511] = 2,
    ACTIONS(3319), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3317), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17569] = 2,
    ACTIONS(3259), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3257), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17627] = 2,
    ACTIONS(3455), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3453), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17685] = 2,
    ACTIONS(3323), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3321), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17743] = 3,
    ACTIONS(3457), 1,
      sym__emphasis_close_star,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17803] = 2,
    ACTIONS(3331), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3329), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17861] = 3,
    ACTIONS(3460), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17921] = 2,
    ACTIONS(3383), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [17979] = 3,
    ACTIONS(3233), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3236), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3231), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18039] = 2,
    ACTIONS(3240), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3238), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18097] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3465), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(460), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3463), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [18169] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3467), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(460), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3463), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [18241] = 2,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18299] = 3,
    ACTIONS(3469), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18359] = 3,
    ACTIONS(3244), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18419] = 2,
    ACTIONS(3251), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3249), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18477] = 2,
    ACTIONS(3255), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18535] = 2,
    ACTIONS(3387), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18593] = 2,
    ACTIONS(3259), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3257), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18651] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3472), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3405), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [18723] = 2,
    ACTIONS(3263), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3261), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18781] = 2,
    ACTIONS(3267), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18839] = 2,
    ACTIONS(3271), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3269), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18897] = 2,
    ACTIONS(3275), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [18955] = 2,
    ACTIONS(3279), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19013] = 2,
    ACTIONS(3283), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19071] = 2,
    ACTIONS(3287), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19129] = 3,
    ACTIONS(3474), 1,
      sym__emphasis_close_star,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19189] = 3,
    ACTIONS(3233), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3236), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3231), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19249] = 3,
    ACTIONS(3477), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19309] = 2,
    ACTIONS(3240), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3238), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19367] = 2,
    ACTIONS(3291), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19425] = 2,
    ACTIONS(3295), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3293), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19483] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3482), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(483), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3480), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [19555] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3484), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(483), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3480), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [19627] = 2,
    ACTIONS(3299), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3297), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19685] = 2,
    ACTIONS(3315), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3313), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19743] = 3,
    ACTIONS(3244), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19803] = 2,
    ACTIONS(3303), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3301), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19861] = 2,
    ACTIONS(3307), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3305), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19919] = 2,
    ACTIONS(3311), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3309), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [19977] = 2,
    ACTIONS(3488), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3486), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20035] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3490), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3405), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [20107] = 2,
    ACTIONS(3315), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3313), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20165] = 2,
    ACTIONS(3319), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3317), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20223] = 2,
    ACTIONS(3323), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3321), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20281] = 2,
    ACTIONS(3327), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3325), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20339] = 2,
    ACTIONS(3331), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3329), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20397] = 2,
    ACTIONS(3335), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20455] = 2,
    ACTIONS(3263), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3261), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20513] = 3,
    ACTIONS(3492), 1,
      sym__emphasis_close_star,
    ACTIONS(3430), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3428), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20573] = 2,
    ACTIONS(3339), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20631] = 3,
    ACTIONS(3495), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3440), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3438), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20691] = 2,
    ACTIONS(3343), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20749] = 2,
    ACTIONS(3347), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20807] = 2,
    ACTIONS(3351), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [20865] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3500), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(503), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3498), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [20937] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3502), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(503), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3498), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [21009] = 2,
    ACTIONS(3355), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3353), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21067] = 2,
    ACTIONS(3506), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3504), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21125] = 2,
    ACTIONS(3359), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3357), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21183] = 2,
    ACTIONS(3363), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3361), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21241] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3508), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(279), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3405), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [21313] = 2,
    ACTIONS(3267), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21371] = 2,
    ACTIONS(3251), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3249), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21429] = 2,
    ACTIONS(3271), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3269), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21487] = 2,
    ACTIONS(3275), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21545] = 2,
    ACTIONS(3279), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21603] = 2,
    ACTIONS(3283), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21661] = 2,
    ACTIONS(3287), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21719] = 2,
    ACTIONS(3367), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3365), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21777] = 2,
    ACTIONS(3371), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3369), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21835] = 2,
    ACTIONS(3375), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3373), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21893] = 2,
    ACTIONS(3291), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [21951] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3512), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(397), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3510), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [22023] = 2,
    ACTIONS(3295), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3293), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22081] = 9,
    ACTIONS(3403), 1,
      sym__backslash_escape,
    ACTIONS(3413), 1,
      sym__newline_token,
    ACTIONS(3415), 1,
      sym__whitespace_ge_2,
    ACTIONS(3417), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3514), 1,
      anon_sym_RBRACK,
    ACTIONS(3409), 3,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(397), 7,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      sym__text_base,
      sym__text_inline_no_link,
      aux_sym_link_label_repeat1,
    ACTIONS(3510), 11,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3411), 27,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [22153] = 2,
    ACTIONS(3379), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3377), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22211] = 2,
    ACTIONS(3383), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22269] = 2,
    ACTIONS(3387), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22327] = 2,
    ACTIONS(3299), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3297), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22385] = 2,
    ACTIONS(3255), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22443] = 2,
    ACTIONS(3391), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22501] = 2,
    ACTIONS(3303), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3301), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22559] = 2,
    ACTIONS(3391), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22617] = 2,
    ACTIONS(3194), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22675] = 2,
    ACTIONS(3198), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3196), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22733] = 2,
    ACTIONS(3225), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3223), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22791] = 2,
    ACTIONS(3518), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3516), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22849] = 2,
    ACTIONS(3522), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3520), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22907] = 2,
    ACTIONS(3307), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3305), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [22965] = 2,
    ACTIONS(3229), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3227), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23023] = 2,
    ACTIONS(3311), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3309), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23081] = 2,
    ACTIONS(3526), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3524), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      ts_builtin_sym_end,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23139] = 2,
    ACTIONS(3375), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3373), 46,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23197] = 4,
    ACTIONS(3244), 1,
      anon_sym_LPAREN,
    ACTIONS(3528), 1,
      anon_sym_LBRACK,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23258] = 2,
    ACTIONS(3518), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3516), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23315] = 2,
    ACTIONS(3526), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3524), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23372] = 2,
    ACTIONS(3506), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3504), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23429] = 3,
    ACTIONS(3233), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3236), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3231), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23488] = 2,
    ACTIONS(3240), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3238), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23545] = 2,
    ACTIONS(3251), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3249), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23602] = 2,
    ACTIONS(3259), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3257), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23659] = 2,
    ACTIONS(3263), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3261), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23716] = 2,
    ACTIONS(3271), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3269), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23773] = 2,
    ACTIONS(3295), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3293), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23830] = 2,
    ACTIONS(3299), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3297), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23887] = 2,
    ACTIONS(3303), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3301), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [23944] = 2,
    ACTIONS(3307), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3305), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24001] = 2,
    ACTIONS(3311), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3309), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24058] = 2,
    ACTIONS(3315), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3313), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24115] = 2,
    ACTIONS(3319), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3317), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24172] = 2,
    ACTIONS(3323), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3321), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24229] = 2,
    ACTIONS(3327), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3325), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24286] = 2,
    ACTIONS(3331), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3329), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24343] = 2,
    ACTIONS(3355), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3353), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24400] = 2,
    ACTIONS(3359), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3357), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24457] = 2,
    ACTIONS(3363), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3361), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24514] = 2,
    ACTIONS(3367), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3365), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24571] = 2,
    ACTIONS(3371), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3369), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24628] = 2,
    ACTIONS(3375), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3373), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24685] = 2,
    ACTIONS(3379), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3377), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24742] = 2,
    ACTIONS(3198), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3196), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24799] = 3,
    ACTIONS(3244), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24858] = 2,
    ACTIONS(3229), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3227), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24915] = 2,
    ACTIONS(3154), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3152), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [24972] = 2,
    ACTIONS(3138), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3136), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25029] = 2,
    ACTIONS(3166), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3164), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25086] = 2,
    ACTIONS(3170), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3168), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25143] = 2,
    ACTIONS(3174), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3172), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25200] = 2,
    ACTIONS(3178), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3176), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25257] = 2,
    ACTIONS(3182), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3180), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25314] = 2,
    ACTIONS(3186), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3184), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25371] = 2,
    ACTIONS(3190), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25428] = 4,
    ACTIONS(3244), 1,
      anon_sym_LPAREN,
    ACTIONS(3528), 1,
      anon_sym_LBRACK,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25489] = 4,
    ACTIONS(3244), 1,
      anon_sym_LPAREN,
    ACTIONS(3528), 1,
      anon_sym_LBRACK,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25550] = 2,
    ACTIONS(3255), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25607] = 2,
    ACTIONS(3267), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25664] = 2,
    ACTIONS(3275), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25721] = 2,
    ACTIONS(3279), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25778] = 2,
    ACTIONS(3283), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25835] = 2,
    ACTIONS(3287), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25892] = 2,
    ACTIONS(3291), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [25949] = 2,
    ACTIONS(3335), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26006] = 2,
    ACTIONS(3339), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26063] = 2,
    ACTIONS(3343), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26120] = 2,
    ACTIONS(3347), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26177] = 2,
    ACTIONS(3351), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26234] = 2,
    ACTIONS(3383), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26291] = 2,
    ACTIONS(3387), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26348] = 2,
    ACTIONS(3391), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26405] = 2,
    ACTIONS(3194), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26462] = 2,
    ACTIONS(3142), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26519] = 2,
    ACTIONS(3146), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26576] = 2,
    ACTIONS(3150), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26633] = 2,
    ACTIONS(3158), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26690] = 2,
    ACTIONS(3162), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26747] = 4,
    ACTIONS(3244), 1,
      anon_sym_LPAREN,
    ACTIONS(3528), 1,
      anon_sym_LBRACK,
    ACTIONS(3247), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3242), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26808] = 2,
    ACTIONS(3225), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3223), 45,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26865] = 2,
    ACTIONS(3532), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3530), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26921] = 2,
    ACTIONS(3536), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3534), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [26977] = 2,
    ACTIONS(3540), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3538), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27033] = 2,
    ACTIONS(3544), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3542), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27089] = 2,
    ACTIONS(3548), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3546), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27145] = 2,
    ACTIONS(3552), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3550), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27201] = 2,
    ACTIONS(3556), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3554), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27257] = 2,
    ACTIONS(3560), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3558), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27313] = 2,
    ACTIONS(3564), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3562), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27369] = 2,
    ACTIONS(3568), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3566), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27425] = 2,
    ACTIONS(3572), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3570), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27481] = 2,
    ACTIONS(3548), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3546), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27537] = 2,
    ACTIONS(3560), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3558), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27593] = 2,
    ACTIONS(3564), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3562), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27649] = 2,
    ACTIONS(3568), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3566), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27705] = 2,
    ACTIONS(3560), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3558), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27761] = 2,
    ACTIONS(3564), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3562), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27817] = 2,
    ACTIONS(3568), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3566), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27873] = 2,
    ACTIONS(3572), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3570), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27929] = 2,
    ACTIONS(3548), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3546), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__strikethrough_close,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [27985] = 2,
    ACTIONS(3544), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3542), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28041] = 3,
    ACTIONS(3574), 1,
      sym__emphasis_close_star,
    ACTIONS(3532), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3530), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28099] = 2,
    ACTIONS(3578), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3576), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28155] = 2,
    ACTIONS(3536), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3534), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28211] = 2,
    ACTIONS(3540), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3538), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28267] = 2,
    ACTIONS(3544), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3542), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_star,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28323] = 3,
    ACTIONS(3580), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3578), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3576), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28381] = 3,
    ACTIONS(3582), 1,
      sym__emphasis_close_star,
    ACTIONS(3532), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3530), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28439] = 3,
    ACTIONS(3584), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3578), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3576), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28497] = 3,
    ACTIONS(3586), 1,
      sym__emphasis_close_star,
    ACTIONS(3532), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3530), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28555] = 3,
    ACTIONS(3588), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3578), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3576), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28613] = 2,
    ACTIONS(3536), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3534), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28669] = 2,
    ACTIONS(3540), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3538), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28725] = 2,
    ACTIONS(3544), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3542), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28781] = 2,
    ACTIONS(3560), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3558), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28837] = 2,
    ACTIONS(3564), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3562), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28893] = 2,
    ACTIONS(3568), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3566), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [28949] = 2,
    ACTIONS(3572), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3570), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29005] = 2,
    ACTIONS(3548), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3546), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29061] = 3,
    ACTIONS(3590), 1,
      sym__emphasis_close_star,
    ACTIONS(3532), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3530), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29119] = 3,
    ACTIONS(3592), 1,
      sym__emphasis_close_underscore,
    ACTIONS(3578), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3576), 43,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29177] = 2,
    ACTIONS(3536), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3534), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29233] = 2,
    ACTIONS(3596), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3594), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29289] = 2,
    ACTIONS(3540), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3538), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__emphasis_close_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29345] = 2,
    ACTIONS(3572), 7,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      aux_sym__whitespace_token1,
    ACTIONS(3570), 44,
      sym__code_span_start,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__strikethrough_open,
      sym__latex_span_start,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym_uri_autolink,
      sym_email_autolink,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [29401] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3614), 1,
      anon_sym_RPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    STATE(987), 1,
      sym_link_destination,
    STATE(988), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29482] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3622), 1,
      anon_sym_RPAREN,
    STATE(1040), 1,
      sym_link_destination,
    STATE(1041), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29563] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3624), 1,
      anon_sym_RPAREN,
    STATE(1069), 1,
      sym_link_destination,
    STATE(1071), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(651), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29644] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3626), 1,
      anon_sym_RPAREN,
    STATE(1075), 1,
      sym_link_destination,
    STATE(1076), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(652), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29725] = 9,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3628), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(3630), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2733), 3,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BSLASH,
    STATE(807), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 6,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 29,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [29792] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3633), 1,
      anon_sym_RPAREN,
    STATE(1010), 1,
      sym_link_title,
    STATE(1093), 1,
      sym_link_destination,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(657), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29873] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3635), 1,
      anon_sym_RPAREN,
    STATE(1019), 1,
      sym_link_destination,
    STATE(1079), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [29954] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3637), 1,
      anon_sym_RPAREN,
    STATE(984), 1,
      sym_link_destination,
    STATE(994), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30035] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3639), 1,
      anon_sym_RPAREN,
    STATE(983), 1,
      sym_link_title,
    STATE(990), 1,
      sym_link_destination,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(655), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30116] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3641), 1,
      anon_sym_RPAREN,
    STATE(985), 1,
      sym_link_destination,
    STATE(989), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(656), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30197] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3643), 1,
      anon_sym_RPAREN,
    STATE(1011), 1,
      sym_link_destination,
    STATE(1012), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30278] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3645), 1,
      anon_sym_RPAREN,
    STATE(1014), 1,
      sym_link_destination,
    STATE(1015), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30359] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3647), 1,
      anon_sym_RPAREN,
    STATE(1062), 1,
      sym_link_destination,
    STATE(1065), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30440] = 15,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3653), 1,
      anon_sym_DQUOTE,
    ACTIONS(3657), 1,
      anon_sym_LPAREN,
    ACTIONS(3659), 1,
      anon_sym_RPAREN,
    ACTIONS(3662), 1,
      sym__newline_token,
    ACTIONS(3665), 1,
      sym__whitespace_ge_2,
    ACTIONS(3668), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3671), 1,
      sym__last_token_punctuation,
    STATE(853), 1,
      sym__soft_line_break,
    ACTIONS(3655), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(687), 2,
      sym__whitespace,
      aux_sym_link_title_repeat1,
    STATE(694), 2,
      sym__link_destination_parenthesis,
      aux_sym_link_destination_repeat2,
    STATE(878), 2,
      sym_backslash_escape,
      sym__word,
    ACTIONS(3649), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3651), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30519] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3673), 1,
      anon_sym_RPAREN,
    STATE(1091), 1,
      sym_link_destination,
    STATE(1092), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(645), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30600] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3675), 1,
      anon_sym_RPAREN,
    STATE(1094), 1,
      sym_link_destination,
    STATE(1097), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(662), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30681] = 9,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3677), 1,
      anon_sym_QMARK_GT,
    ACTIONS(3679), 2,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(2702), 3,
      anon_sym_LT,
      anon_sym_QMARK,
      anon_sym_BSLASH,
    STATE(806), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 6,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [30748] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3682), 1,
      anon_sym_RPAREN,
    STATE(991), 1,
      sym_link_destination,
    STATE(992), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30829] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3684), 1,
      anon_sym_RPAREN,
    STATE(1073), 1,
      sym_link_destination,
    STATE(1080), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30910] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3686), 1,
      anon_sym_RPAREN,
    STATE(1086), 1,
      sym_link_destination,
    STATE(1100), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(663), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [30991] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3688), 1,
      anon_sym_RPAREN,
    STATE(1024), 1,
      sym_link_destination,
    STATE(1025), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(667), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [31072] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3690), 1,
      anon_sym_RPAREN,
    STATE(1027), 1,
      sym_link_destination,
    STATE(1028), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(646), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [31153] = 16,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3604), 1,
      anon_sym_LT,
    ACTIONS(3606), 1,
      anon_sym_DQUOTE,
    ACTIONS(3610), 1,
      anon_sym_SQUOTE,
    ACTIONS(3612), 1,
      anon_sym_LPAREN,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3692), 1,
      anon_sym_RPAREN,
    STATE(1036), 1,
      sym_link_destination,
    STATE(1037), 1,
      sym_link_title,
    ACTIONS(3608), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(699), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(3600), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3602), 25,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [31234] = 15,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3653), 1,
      anon_sym_SQUOTE,
    ACTIONS(3662), 1,
      sym__newline_token,
    ACTIONS(3665), 1,
      sym__whitespace_ge_2,
    ACTIONS(3668), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3671), 1,
      sym__last_token_punctuation,
    ACTIONS(3700), 1,
      anon_sym_LPAREN,
    ACTIONS(3702), 1,
      anon_sym_RPAREN,
    STATE(830), 1,
      sym__soft_line_break,
    ACTIONS(3698), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(691), 2,
      sym__whitespace,
      aux_sym_link_title_repeat2,
    STATE(694), 2,
      sym__link_destination_parenthesis,
      aux_sym_link_destination_repeat2,
    STATE(879), 2,
      sym_backslash_escape,
      sym__word,
    ACTIONS(3694), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3696), 27,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [31313] = 9,
    ACTIONS(2739), 1,
      sym__newline_token,
    ACTIONS(2744), 1,
      sym__whitespace_ge_2,
    ACTIONS(2747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3705), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2733), 2,
      anon_sym_RBRACK,
      anon_sym_LT,
    ACTIONS(3707), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(784), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 5,
      sym__code_span_close,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2736), 30,
      anon_sym_LBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [31379] = 3,
    ACTIONS(3710), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31433] = 3,
    ACTIONS(3712), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31487] = 13,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3722), 1,
      anon_sym_RPAREN,
    STATE(870), 1,
      sym__soft_line_break,
    ACTIONS(3718), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(693), 2,
      sym__whitespace,
      aux_sym_link_title_repeat3,
    STATE(832), 2,
      sym__link_destination_parenthesis,
      aux_sym_link_destination_repeat2,
    STATE(864), 2,
      sym_backslash_escape,
      sym__word,
    ACTIONS(3714), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(3716), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [31561] = 10,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3730), 1,
      anon_sym_DASH,
    ACTIONS(3733), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3727), 2,
      anon_sym_LT,
      anon_sym_BSLASH,
    ACTIONS(3735), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(805), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 7,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_GT,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3724), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [31629] = 9,
    ACTIONS(2708), 1,
      sym__newline_token,
    ACTIONS(2715), 1,
      sym__whitespace_ge_2,
    ACTIONS(2718), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3738), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 2,
      anon_sym_LT,
      anon_sym_QMARK,
    ACTIONS(3740), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(780), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2698), 5,
      sym__code_span_close,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(2705), 30,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [31695] = 3,
    ACTIONS(3743), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31749] = 2,
    ACTIONS(3275), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31800] = 2,
    ACTIONS(3190), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31851] = 2,
    ACTIONS(3291), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [31902] = 10,
    ACTIONS(2787), 1,
      sym__newline_token,
    ACTIONS(2792), 1,
      sym__whitespace_ge_2,
    ACTIONS(2795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3748), 1,
      anon_sym_LT,
    ACTIONS(3751), 1,
      anon_sym_DASH,
    ACTIONS(3754), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3756), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(767), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(2698), 6,
      sym__code_span_close,
      anon_sym_GT,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
    ACTIONS(3745), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [31969] = 2,
    ACTIONS(3267), 4,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 42,
      sym__emphasis_open_star,
      sym__emphasis_open_underscore,
      sym__unclosed_span,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [32020] = 10,
    ACTIONS(3759), 1,
      sym__backslash_escape,
    ACTIONS(3771), 1,
      anon_sym_SQUOTE,
    ACTIONS(3773), 1,
      sym__newline_token,
    ACTIONS(3776), 1,
      sym__whitespace_ge_2,
    ACTIONS(3779), 1,
      aux_sym__whitespace_token1,
    STATE(830), 1,
      sym__soft_line_break,
    ACTIONS(3768), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3762), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(681), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(3765), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32086] = 6,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    ACTIONS(2711), 2,
      anon_sym_LT,
      anon_sym_BSLASH,
    STATE(852), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2698), 38,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [32144] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(3782), 1,
      anon_sym_SLASH,
    ACTIONS(3785), 1,
      sym__last_token_punctuation,
    STATE(952), 1,
      sym__tag_name,
    ACTIONS(2711), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 38,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [32202] = 9,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3671), 1,
      sym__last_token_punctuation,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3793), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3787), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(694), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [32266] = 10,
    ACTIONS(3797), 1,
      sym__backslash_escape,
    ACTIONS(3809), 1,
      anon_sym_LPAREN,
    ACTIONS(3812), 1,
      anon_sym_RPAREN,
    ACTIONS(3817), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3819), 1,
      sym__last_token_punctuation,
    ACTIONS(3806), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3815), 2,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3800), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(832), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3803), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [32332] = 10,
    ACTIONS(3821), 1,
      sym__backslash_escape,
    ACTIONS(3830), 1,
      anon_sym_DQUOTE,
    ACTIONS(3835), 1,
      sym__newline_token,
    ACTIONS(3838), 1,
      sym__whitespace_ge_2,
    ACTIONS(3841), 1,
      aux_sym__whitespace_token1,
    STATE(853), 1,
      sym__soft_line_break,
    ACTIONS(3832), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3824), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(686), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(3827), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32398] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3848), 1,
      anon_sym_DQUOTE,
    STATE(853), 1,
      sym__soft_line_break,
    ACTIONS(3850), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3844), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(686), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(3846), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32464] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3854), 1,
      anon_sym_DQUOTE,
    STATE(853), 1,
      sym__soft_line_break,
    ACTIONS(3850), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3852), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(687), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(3846), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32530] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3854), 1,
      anon_sym_SQUOTE,
    STATE(830), 1,
      sym__soft_line_break,
    ACTIONS(3860), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3856), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(691), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(3858), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32596] = 10,
    ACTIONS(3862), 1,
      sym__backslash_escape,
    ACTIONS(3874), 1,
      anon_sym_LPAREN,
    ACTIONS(3877), 1,
      anon_sym_RPAREN,
    ACTIONS(3882), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3884), 1,
      sym__last_token_punctuation,
    ACTIONS(3871), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3880), 2,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3865), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(832), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3868), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [32662] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3848), 1,
      anon_sym_SQUOTE,
    STATE(830), 1,
      sym__soft_line_break,
    ACTIONS(3860), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3886), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(681), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(3858), 29,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [32728] = 3,
    ACTIONS(3888), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [32779] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3848), 1,
      anon_sym_RPAREN,
    STATE(870), 1,
      sym__soft_line_break,
    ACTIONS(3894), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3890), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(703), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(3892), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [32844] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3900), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3898), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3896), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(696), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [32905] = 3,
    ACTIONS(3902), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [32956] = 8,
    ACTIONS(3904), 1,
      sym__backslash_escape,
    ACTIONS(3916), 1,
      anon_sym_LPAREN,
    ACTIONS(3921), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3913), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3919), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3907), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(696), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3910), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33017] = 10,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3616), 1,
      sym__newline_token,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3854), 1,
      anon_sym_RPAREN,
    STATE(870), 1,
      sym__soft_line_break,
    ACTIONS(3894), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3923), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(693), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(3892), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33082] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3927), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3925), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3896), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(696), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33143] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3795), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3793), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3787), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(694), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33204] = 6,
    ACTIONS(2970), 1,
      sym__word_no_digit,
    ACTIONS(3929), 1,
      anon_sym_SLASH,
    ACTIONS(3932), 1,
      sym__last_token_punctuation,
    STATE(962), 1,
      sym__tag_name,
    ACTIONS(2711), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 38,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__digits,
  [33261] = 6,
    ACTIONS(2711), 1,
      anon_sym_LT,
    ACTIONS(2975), 1,
      sym__newline_token,
    ACTIONS(2978), 1,
      sym__whitespace_ge_2,
    ACTIONS(2981), 1,
      aux_sym__whitespace_token1,
    STATE(846), 2,
      sym__whitespace,
      sym__soft_line_break,
    ACTIONS(2698), 38,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__word_no_digit,
      sym__digits,
  [33318] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(3900), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3898), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(3934), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(698), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33379] = 10,
    ACTIONS(3936), 1,
      sym__backslash_escape,
    ACTIONS(3948), 1,
      anon_sym_RPAREN,
    ACTIONS(3950), 1,
      sym__newline_token,
    ACTIONS(3953), 1,
      sym__whitespace_ge_2,
    ACTIONS(3956), 1,
      aux_sym__whitespace_token1,
    STATE(870), 1,
      sym__soft_line_break,
    ACTIONS(3945), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3939), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(703), 4,
      sym_backslash_escape,
      sym__whitespace,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(3942), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [33444] = 3,
    ACTIONS(3785), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33495] = 2,
    ACTIONS(3190), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33543] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3965), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3973), 1,
      sym__last_token_punctuation,
    ACTIONS(3971), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(822), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [33605] = 2,
    ACTIONS(3142), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33653] = 2,
    ACTIONS(3146), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33701] = 2,
    ACTIONS(3158), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33749] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3975), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3979), 1,
      sym__last_token_punctuation,
    ACTIONS(3977), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(826), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [33811] = 2,
    ACTIONS(3162), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [33859] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3981), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3985), 1,
      sym__last_token_punctuation,
    ACTIONS(3983), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(766), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [33921] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3991), 1,
      anon_sym_GT,
    ACTIONS(3993), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3987), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(737), 5,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__whitespace,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(3989), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [33981] = 3,
    ACTIONS(3995), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34031] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3997), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4001), 1,
      sym__last_token_punctuation,
    ACTIONS(3999), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(808), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34093] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4003), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4007), 1,
      sym__last_token_punctuation,
    ACTIONS(4005), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(775), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34155] = 6,
    ACTIONS(4013), 1,
      sym__newline_token,
    ACTIONS(4016), 1,
      sym__whitespace_ge_2,
    ACTIONS(4019), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4011), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(717), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(4009), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__word_no_digit,
      sym__digits,
  [34211] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4022), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4026), 1,
      sym__last_token_punctuation,
    ACTIONS(4024), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(815), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34273] = 8,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3618), 1,
      sym__whitespace_ge_2,
    ACTIONS(3620), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4030), 1,
      anon_sym_GT,
    ACTIONS(3993), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4028), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(713), 5,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__whitespace,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(3989), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34333] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4032), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4036), 1,
      sym__last_token_punctuation,
    ACTIONS(4034), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(753), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34395] = 3,
    ACTIONS(4038), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34445] = 2,
    ACTIONS(3255), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34493] = 3,
    ACTIONS(3932), 1,
      sym__last_token_punctuation,
    ACTIONS(2711), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(2698), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34543] = 2,
    ACTIONS(3267), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34591] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4040), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4044), 1,
      sym__last_token_punctuation,
    ACTIONS(4042), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(774), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [34653] = 2,
    ACTIONS(3275), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34701] = 2,
    ACTIONS(3279), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34749] = 2,
    ACTIONS(3283), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34797] = 2,
    ACTIONS(3287), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34845] = 2,
    ACTIONS(3291), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34893] = 2,
    ACTIONS(3335), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [34941] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4046), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4050), 1,
      sym__last_token_punctuation,
    ACTIONS(4048), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(781), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35003] = 2,
    ACTIONS(3339), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35051] = 2,
    ACTIONS(3343), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35099] = 2,
    ACTIONS(3347), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35147] = 2,
    ACTIONS(3351), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35195] = 8,
    ACTIONS(4052), 1,
      sym__backslash_escape,
    ACTIONS(4061), 1,
      anon_sym_GT,
    ACTIONS(4066), 1,
      sym__whitespace_ge_2,
    ACTIONS(4069), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4063), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4055), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(737), 5,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__whitespace,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(4058), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35255] = 2,
    ACTIONS(3383), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35303] = 2,
    ACTIONS(3387), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35351] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4072), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4076), 1,
      sym__last_token_punctuation,
    ACTIONS(4074), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(796), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35413] = 2,
    ACTIONS(3391), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35461] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4078), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4082), 1,
      sym__last_token_punctuation,
    ACTIONS(4080), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(801), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35523] = 2,
    ACTIONS(3194), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35571] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4084), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4088), 1,
      sym__last_token_punctuation,
    ACTIONS(4086), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(810), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35633] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4090), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4094), 1,
      sym__last_token_punctuation,
    ACTIONS(4092), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(812), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35695] = 9,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4096), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4100), 1,
      sym__last_token_punctuation,
    ACTIONS(4098), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(820), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [35757] = 2,
    ACTIONS(3150), 3,
      anon_sym_LT,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 40,
      sym__latex_span_close,
      sym__backslash_escape,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35805] = 2,
    ACTIONS(3267), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3265), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35852] = 2,
    ACTIONS(3291), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35899] = 2,
    ACTIONS(3343), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3341), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [35946] = 8,
    ACTIONS(4105), 1,
      anon_sym_RBRACK,
    ACTIONS(4108), 1,
      sym__newline_token,
    ACTIONS(4111), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4113), 1,
      sym__whitespace_ge_2,
    ACTIONS(4116), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4119), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(4102), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36005] = 2,
    ACTIONS(3347), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3345), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [36052] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4122), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36111] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4122), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4126), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(755), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36170] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4128), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36229] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3997), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3999), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(808), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36288] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4132), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36347] = 2,
    ACTIONS(3351), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3349), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [36394] = 8,
    ACTIONS(4111), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4143), 1,
      anon_sym_QMARK,
    ACTIONS(4146), 1,
      sym__newline_token,
    ACTIONS(4149), 1,
      sym__whitespace_ge_2,
    ACTIONS(4152), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4155), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(4140), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36453] = 2,
    ACTIONS(3279), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3277), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [36500] = 2,
    ACTIONS(3391), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3389), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [36547] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4003), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4005), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(775), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36606] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4040), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4042), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(774), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36665] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4162), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36724] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4170), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36783] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4003), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36842] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3981), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3983), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(766), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36901] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4172), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [36960] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4174), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37019] = 2,
    ACTIONS(3383), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3381), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [37066] = 2,
    ACTIONS(3387), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3385), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [37113] = 2,
    ACTIONS(3283), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3281), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [37160] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4046), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4048), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(781), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37219] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4046), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37278] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4176), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37337] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4176), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4178), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(777), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37396] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4180), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37455] = 2,
    ACTIONS(3194), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3192), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [37502] = 2,
    ACTIONS(3255), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3253), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [37549] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4182), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37608] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4184), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37667] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4184), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4186), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(783), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37726] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4188), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37785] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4190), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37844] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4022), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4024), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(815), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37903] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4192), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [37962] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4194), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38021] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4196), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4198), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(768), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38080] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4200), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38139] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4202), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38198] = 8,
    ACTIONS(4207), 1,
      anon_sym_DASH,
    ACTIONS(4210), 1,
      sym__newline_token,
    ACTIONS(4213), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4215), 1,
      sym__whitespace_ge_2,
    ACTIONS(4218), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4221), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(4204), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38257] = 2,
    ACTIONS(3142), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3140), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38304] = 2,
    ACTIONS(3146), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3144), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38351] = 2,
    ACTIONS(3150), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3148), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38398] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4078), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4080), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(801), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38457] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4078), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38516] = 2,
    ACTIONS(3335), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3333), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38563] = 2,
    ACTIONS(3339), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3337), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38610] = 2,
    ACTIONS(3158), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3156), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38657] = 2,
    ACTIONS(3162), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3160), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38704] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4224), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38763] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4224), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4226), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(803), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38822] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4228), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38881] = 2,
    ACTIONS(3287), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3285), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [38928] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4084), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4086), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(810), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [38987] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4230), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39046] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4232), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39105] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4196), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39164] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4090), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4092), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(812), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39223] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4090), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39282] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4032), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4034), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(753), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39341] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4234), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39400] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4234), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4236), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(814), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39459] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4238), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39518] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4032), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39577] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4096), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4098), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(820), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39636] = 8,
    ACTIONS(4158), 1,
      anon_sym_QMARK,
    ACTIONS(4160), 1,
      sym__newline_token,
    ACTIONS(4164), 1,
      sym__whitespace_ge_2,
    ACTIONS(4166), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4240), 1,
      anon_sym_QMARK_GT,
    ACTIONS(4168), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(759), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2700), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39695] = 8,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(4130), 1,
      sym__newline_token,
    ACTIONS(4134), 1,
      sym__whitespace_ge_2,
    ACTIONS(4136), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4242), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(4138), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(751), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__processing_instruction_repeat1,
    ACTIONS(2729), 31,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39754] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3965), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3971), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(822), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39813] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3965), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39872] = 2,
    ACTIONS(3275), 2,
      anon_sym_LT,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 40,
      sym__code_span_close,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_LT_BANG_DASH_DASH,
      anon_sym_LT_QMARK,
      aux_sym__declaration_token1,
      anon_sym_LT_BANG_LBRACKCDATA_LBRACK,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [39919] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4244), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [39978] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4244), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4246), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(824), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40037] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4248), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40096] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3975), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(3977), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(826), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40155] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3997), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4124), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(791), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40214] = 8,
    ACTIONS(3961), 1,
      anon_sym_DASH,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4072), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(4074), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(796), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__html_comment_repeat1,
    ACTIONS(3959), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40273] = 7,
    ACTIONS(4252), 1,
      anon_sym_GT,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40329] = 3,
    ACTIONS(3884), 1,
      sym__last_token_punctuation,
    ACTIONS(3882), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3880), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [40377] = 4,
    ACTIONS(4262), 1,
      sym__newline_token,
    STATE(1151), 1,
      sym__soft_line_break,
    ACTIONS(3882), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3880), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [40427] = 6,
    ACTIONS(4271), 1,
      anon_sym_LPAREN,
    ACTIONS(4275), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4277), 1,
      sym__last_token_punctuation,
    ACTIONS(4268), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4273), 2,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(4265), 34,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__word_no_digit,
      sym__digits,
  [40481] = 7,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(4279), 1,
      anon_sym_RPAREN,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(3896), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(696), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [40537] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4281), 1,
      anon_sym_GT,
    ACTIONS(4283), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(834), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40593] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4285), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40649] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4287), 1,
      anon_sym_GT,
    ACTIONS(4289), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(839), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40705] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4291), 1,
      anon_sym_GT,
    ACTIONS(4293), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(861), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40761] = 3,
    ACTIONS(4295), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [40809] = 3,
    ACTIONS(4297), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [40857] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4299), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40913] = 7,
    ACTIONS(4304), 1,
      anon_sym_GT,
    ACTIONS(4306), 1,
      sym__newline_token,
    ACTIONS(4309), 1,
      sym__whitespace_ge_2,
    ACTIONS(4312), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4315), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4301), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40969] = 3,
    ACTIONS(4320), 1,
      sym__last_token_punctuation,
    ACTIONS(4318), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4271), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [41017] = 7,
    ACTIONS(3598), 1,
      sym__backslash_escape,
    ACTIONS(3720), 1,
      anon_sym_LPAREN,
    ACTIONS(4324), 1,
      anon_sym_RPAREN,
    ACTIONS(3791), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4322), 4,
      sym_entity_reference,
      sym_numeric_character_reference,
      sym__word_no_digit,
      sym__digits,
    STATE(832), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(3789), 28,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [41073] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4326), 1,
      anon_sym_GT,
    ACTIONS(4328), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(847), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41129] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4332), 1,
      anon_sym_DQUOTE,
    ACTIONS(4334), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(845), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat2,
    ACTIONS(4330), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41185] = 7,
    ACTIONS(4339), 1,
      anon_sym_DQUOTE,
    ACTIONS(4341), 1,
      sym__newline_token,
    ACTIONS(4344), 1,
      sym__whitespace_ge_2,
    ACTIONS(4347), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4350), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(845), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat2,
    ACTIONS(4336), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41241] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4353), 1,
      anon_sym_GT,
    ACTIONS(4355), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(855), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41297] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4357), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41353] = 7,
    ACTIONS(4362), 1,
      anon_sym_SQUOTE,
    ACTIONS(4364), 1,
      sym__newline_token,
    ACTIONS(4367), 1,
      sym__whitespace_ge_2,
    ACTIONS(4370), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4373), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(848), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat1,
    ACTIONS(4359), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41409] = 3,
    ACTIONS(4382), 1,
      sym__last_token_punctuation,
    ACTIONS(4379), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4376), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [41457] = 5,
    ACTIONS(4320), 1,
      sym__last_token_punctuation,
    ACTIONS(4387), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4318), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4384), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(4271), 34,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      sym__word_no_digit,
      sym__digits,
  [41509] = 3,
    ACTIONS(3819), 1,
      sym__last_token_punctuation,
    ACTIONS(3817), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3815), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [41557] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4390), 1,
      anon_sym_GT,
    ACTIONS(4392), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(828), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41613] = 4,
    ACTIONS(4394), 1,
      sym__newline_token,
    STATE(1141), 1,
      sym__soft_line_break,
    ACTIONS(3817), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3815), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [41663] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4332), 1,
      anon_sym_SQUOTE,
    ACTIONS(4399), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(848), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat1,
    ACTIONS(4397), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41719] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4401), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41775] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4403), 1,
      anon_sym_GT,
    ACTIONS(4405), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(857), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41831] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4407), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41887] = 3,
    ACTIONS(4415), 1,
      sym__last_token_punctuation,
    ACTIONS(4412), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4409), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [41935] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4417), 1,
      anon_sym_DQUOTE,
    ACTIONS(4419), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(844), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat2,
    ACTIONS(4330), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [41991] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4417), 1,
      anon_sym_SQUOTE,
    ACTIONS(4421), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(854), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__attribute_value_repeat1,
    ACTIONS(4397), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [42047] = 7,
    ACTIONS(4254), 1,
      sym__newline_token,
    ACTIONS(4256), 1,
      sym__whitespace_ge_2,
    ACTIONS(4258), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4423), 1,
      anon_sym_GT,
    ACTIONS(4260), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(840), 4,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
      aux_sym__declaration_repeat1,
    ACTIONS(4250), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [42103] = 5,
    ACTIONS(3919), 1,
      anon_sym_LPAREN,
    ACTIONS(4431), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3948), 2,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(4428), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4425), 34,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__word_no_digit,
      sym__digits,
  [42154] = 2,
    ACTIONS(4435), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4433), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42199] = 5,
    ACTIONS(4271), 1,
      anon_sym_LPAREN,
    ACTIONS(4275), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4268), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4273), 2,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(4265), 34,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__word_no_digit,
      sym__digits,
  [42250] = 2,
    ACTIONS(3921), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3919), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42295] = 2,
    ACTIONS(4439), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4437), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42340] = 2,
    ACTIONS(3275), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42385] = 2,
    ACTIONS(3291), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42430] = 4,
    ACTIONS(4444), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4439), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(4441), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
    ACTIONS(4437), 34,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      sym__word_no_digit,
      sym__digits,
  [42479] = 4,
    ACTIONS(4447), 1,
      sym__newline_token,
    STATE(1148), 1,
      sym__soft_line_break,
    ACTIONS(4275), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4273), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42528] = 2,
    ACTIONS(4453), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4450), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42573] = 2,
    ACTIONS(4456), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3830), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42618] = 2,
    ACTIONS(4461), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4458), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42663] = 2,
    ACTIONS(4464), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3771), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42708] = 2,
    ACTIONS(4468), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4466), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42753] = 3,
    ACTIONS(4470), 1,
      sym__last_token_punctuation,
    ACTIONS(4275), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4273), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42800] = 2,
    ACTIONS(3190), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3188), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42845] = 2,
    ACTIONS(4379), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4376), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42890] = 2,
    ACTIONS(4412), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4409), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42935] = 2,
    ACTIONS(4474), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4472), 37,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [42980] = 3,
    ACTIONS(4476), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43026] = 3,
    ACTIONS(4478), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43072] = 3,
    ACTIONS(4480), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43118] = 3,
    ACTIONS(4484), 1,
      sym__last_token_punctuation,
    ACTIONS(4482), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(4213), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43164] = 3,
    ACTIONS(4490), 1,
      sym__last_token_punctuation,
    ACTIONS(4488), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(4486), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43210] = 3,
    ACTIONS(4496), 1,
      sym__last_token_punctuation,
    ACTIONS(4494), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4492), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43256] = 3,
    ACTIONS(4498), 1,
      sym__last_token_punctuation,
    ACTIONS(4488), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(4486), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43302] = 3,
    ACTIONS(4500), 1,
      sym__last_token_whitespace,
    ACTIONS(2986), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(2984), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43348] = 6,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4504), 2,
      sym__word_no_digit,
      sym__digits,
    STATE(902), 3,
      sym__whitespace,
      sym__word,
      sym__soft_line_break,
    ACTIONS(4502), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [43400] = 3,
    ACTIONS(4504), 1,
      sym__last_token_punctuation,
    ACTIONS(4508), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(4506), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43446] = 3,
    ACTIONS(4510), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43492] = 2,
    ACTIONS(4514), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4512), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43536] = 2,
    ACTIONS(4431), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(3948), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43580] = 3,
    ACTIONS(4516), 1,
      sym__last_token_whitespace,
    ACTIONS(3044), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(3042), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43626] = 3,
    ACTIONS(4520), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4522), 1,
      sym__last_token_punctuation,
    ACTIONS(4518), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43671] = 2,
    ACTIONS(4526), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(4524), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43714] = 3,
    ACTIONS(4530), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4532), 1,
      sym__last_token_punctuation,
    ACTIONS(4528), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43759] = 2,
    ACTIONS(3275), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43802] = 2,
    ACTIONS(3291), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43845] = 2,
    ACTIONS(3275), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43888] = 2,
    ACTIONS(4536), 3,
      anon_sym_AMP,
      anon_sym_BSLASH,
      aux_sym__whitespace_token1,
    ACTIONS(4534), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43931] = 2,
    ACTIONS(4482), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(4213), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [43974] = 2,
    ACTIONS(3291), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44017] = 3,
    ACTIONS(2986), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4538), 1,
      sym__last_token_whitespace,
    ACTIONS(2984), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44062] = 2,
    ACTIONS(4540), 2,
      anon_sym_RBRACK,
      aux_sym__whitespace_token1,
    ACTIONS(4111), 36,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_RBRACK_RBRACK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44105] = 2,
    ACTIONS(3275), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44148] = 2,
    ACTIONS(3291), 2,
      anon_sym_DASH,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_DASH_DASH_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44191] = 2,
    ACTIONS(4540), 2,
      anon_sym_QMARK,
      aux_sym__whitespace_token1,
    ACTIONS(4111), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      anon_sym_QMARK_GT,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44234] = 3,
    ACTIONS(3044), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4542), 1,
      sym__last_token_whitespace,
    ACTIONS(3042), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44279] = 3,
    ACTIONS(4546), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4548), 1,
      sym__last_token_punctuation,
    ACTIONS(4544), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44324] = 2,
    ACTIONS(3291), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44366] = 2,
    ACTIONS(4550), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4362), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44408] = 2,
    ACTIONS(4552), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4304), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44450] = 2,
    ACTIONS(4554), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4339), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44492] = 2,
    ACTIONS(3275), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 36,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
      sym__word_no_digit,
      sym__digits,
  [44534] = 1,
    ACTIONS(4556), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44568] = 1,
    ACTIONS(4558), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44602] = 1,
    ACTIONS(4560), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44636] = 1,
    ACTIONS(4562), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44670] = 1,
    ACTIONS(4564), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44704] = 1,
    ACTIONS(4566), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44738] = 1,
    ACTIONS(4568), 31,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44772] = 5,
    ACTIONS(4570), 1,
      sym__newline_token,
    ACTIONS(4573), 1,
      sym__whitespace_ge_2,
    ACTIONS(4576), 1,
      aux_sym__whitespace_token1,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
    ACTIONS(4009), 6,
      anon_sym_GT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [44795] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4585), 1,
      anon_sym_RPAREN,
    STATE(1081), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44825] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4587), 1,
      anon_sym_RPAREN,
    STATE(1016), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44855] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4589), 1,
      anon_sym_RPAREN,
    STATE(1021), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44885] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4591), 1,
      anon_sym_RPAREN,
    STATE(1067), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44915] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4593), 1,
      anon_sym_RPAREN,
    STATE(1085), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44945] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4595), 1,
      anon_sym_RPAREN,
    STATE(1042), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [44975] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4597), 1,
      anon_sym_RPAREN,
    STATE(1044), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45005] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4599), 1,
      anon_sym_RPAREN,
    STATE(1048), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45035] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4601), 1,
      anon_sym_RPAREN,
    STATE(1050), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45065] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4603), 1,
      anon_sym_RPAREN,
    STATE(1018), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45095] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4605), 1,
      anon_sym_RPAREN,
    STATE(993), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45125] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4607), 1,
      anon_sym_RPAREN,
    STATE(1001), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45155] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4609), 1,
      anon_sym_RPAREN,
    STATE(996), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45185] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4611), 1,
      anon_sym_RPAREN,
    STATE(1005), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45215] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4613), 1,
      anon_sym_RPAREN,
    STATE(1096), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45245] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4615), 1,
      anon_sym_RPAREN,
    STATE(1099), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45275] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4617), 1,
      anon_sym_RPAREN,
    STATE(999), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45305] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4619), 1,
      anon_sym_RPAREN,
    STATE(1061), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45335] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4621), 1,
      anon_sym_RPAREN,
    STATE(1026), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45365] = 9,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4579), 1,
      anon_sym_DQUOTE,
    ACTIONS(4581), 1,
      anon_sym_SQUOTE,
    ACTIONS(4583), 1,
      anon_sym_LPAREN,
    ACTIONS(4623), 1,
      anon_sym_RPAREN,
    STATE(1031), 1,
      sym_link_title,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45395] = 7,
    ACTIONS(4625), 1,
      anon_sym_GT,
    ACTIONS(4627), 1,
      anon_sym_SLASH,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(980), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45420] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4635), 1,
      anon_sym_GT,
    ACTIONS(4637), 1,
      anon_sym_SLASH,
    STATE(944), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(970), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45445] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4639), 1,
      anon_sym_GT,
    ACTIONS(4641), 1,
      anon_sym_SLASH,
    STATE(947), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(978), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45470] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4643), 1,
      anon_sym_GT,
    ACTIONS(4645), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(971), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45495] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4647), 1,
      anon_sym_GT,
    ACTIONS(4649), 1,
      anon_sym_SLASH,
    STATE(958), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(974), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45520] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4651), 1,
      anon_sym_GT,
    ACTIONS(4653), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(975), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45545] = 6,
    ACTIONS(4657), 1,
      sym__newline_token,
    ACTIONS(4660), 1,
      sym__whitespace_ge_2,
    ACTIONS(4663), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4655), 2,
      anon_sym_GT,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(1054), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45568] = 8,
    ACTIONS(4666), 1,
      anon_sym_DQUOTE,
    ACTIONS(4668), 1,
      anon_sym_SQUOTE,
    ACTIONS(4670), 1,
      sym__newline_token,
    ACTIONS(4672), 1,
      aux_sym__attribute_value_token1,
    ACTIONS(4674), 1,
      sym__whitespace_ge_2,
    ACTIONS(4676), 1,
      aux_sym__whitespace_token1,
    STATE(1107), 1,
      sym__attribute_value,
    STATE(953), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45595] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4678), 1,
      anon_sym_GT,
    ACTIONS(4680), 1,
      anon_sym_SLASH,
    STATE(960), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(963), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45620] = 8,
    ACTIONS(4666), 1,
      anon_sym_DQUOTE,
    ACTIONS(4668), 1,
      anon_sym_SQUOTE,
    ACTIONS(4670), 1,
      sym__newline_token,
    ACTIONS(4674), 1,
      sym__whitespace_ge_2,
    ACTIONS(4676), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4682), 1,
      aux_sym__attribute_value_token1,
    STATE(1108), 1,
      sym__attribute_value,
    STATE(969), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45647] = 8,
    ACTIONS(4666), 1,
      anon_sym_DQUOTE,
    ACTIONS(4668), 1,
      anon_sym_SQUOTE,
    ACTIONS(4670), 1,
      sym__newline_token,
    ACTIONS(4674), 1,
      sym__whitespace_ge_2,
    ACTIONS(4676), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4682), 1,
      aux_sym__attribute_value_token1,
    STATE(1108), 1,
      sym__attribute_value,
    STATE(955), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45674] = 8,
    ACTIONS(4666), 1,
      anon_sym_DQUOTE,
    ACTIONS(4668), 1,
      anon_sym_SQUOTE,
    ACTIONS(4670), 1,
      sym__newline_token,
    ACTIONS(4674), 1,
      sym__whitespace_ge_2,
    ACTIONS(4676), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4684), 1,
      aux_sym__attribute_value_token1,
    STATE(1109), 1,
      sym__attribute_value,
    STATE(969), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45701] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4686), 1,
      anon_sym_GT,
    ACTIONS(4688), 1,
      anon_sym_SLASH,
    STATE(949), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(972), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45726] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4690), 1,
      anon_sym_GT,
    ACTIONS(4692), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(979), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45751] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4694), 1,
      anon_sym_GT,
    ACTIONS(4696), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(966), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45776] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4698), 1,
      anon_sym_GT,
    ACTIONS(4700), 1,
      anon_sym_SLASH,
    STATE(957), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(976), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45801] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4702), 1,
      anon_sym_GT,
    ACTIONS(4704), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(964), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45826] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4706), 1,
      anon_sym_GT,
    ACTIONS(4708), 1,
      anon_sym_SLASH,
    STATE(950), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(965), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45851] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4710), 1,
      anon_sym_GT,
    ACTIONS(4712), 1,
      anon_sym_SLASH,
    STATE(961), 2,
      sym__attribute,
      aux_sym__open_tag_repeat1,
    STATE(977), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45876] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4702), 1,
      anon_sym_GT,
    ACTIONS(4704), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45900] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4716), 1,
      anon_sym_GT,
    ACTIONS(4718), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45924] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4720), 1,
      anon_sym_GT,
    ACTIONS(4722), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45948] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4724), 1,
      anon_sym_GT,
    ACTIONS(4726), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [45972] = 4,
    ACTIONS(4732), 1,
      aux_sym__whitespace_token1,
    STATE(973), 1,
      aux_sym__tag_name_repeat1,
    ACTIONS(4730), 3,
      anon_sym_DASH,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(4728), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [45990] = 4,
    ACTIONS(4739), 1,
      aux_sym__whitespace_token1,
    STATE(968), 1,
      aux_sym__tag_name_repeat1,
    ACTIONS(4736), 3,
      anon_sym_DASH,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(4734), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [46008] = 5,
    ACTIONS(4741), 1,
      sym__newline_token,
    ACTIONS(4744), 1,
      sym__whitespace_ge_2,
    ACTIONS(4747), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4009), 3,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      aux_sym__attribute_value_token1,
    STATE(969), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46028] = 7,
    ACTIONS(4625), 1,
      anon_sym_GT,
    ACTIONS(4627), 1,
      anon_sym_SLASH,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46052] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4750), 1,
      anon_sym_GT,
    ACTIONS(4752), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46076] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4651), 1,
      anon_sym_GT,
    ACTIONS(4653), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46100] = 4,
    ACTIONS(4758), 1,
      aux_sym__whitespace_token1,
    STATE(968), 1,
      aux_sym__tag_name_repeat1,
    ACTIONS(4756), 3,
      anon_sym_DASH,
      sym__word_no_digit,
      sym__digits,
    ACTIONS(4754), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [46118] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4694), 1,
      anon_sym_GT,
    ACTIONS(4696), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46142] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4760), 1,
      anon_sym_GT,
    ACTIONS(4762), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46166] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4690), 1,
      anon_sym_GT,
    ACTIONS(4692), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46190] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4706), 1,
      anon_sym_GT,
    ACTIONS(4708), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46214] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4643), 1,
      anon_sym_GT,
    ACTIONS(4645), 1,
      anon_sym_SLASH,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46238] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4764), 1,
      anon_sym_GT,
    ACTIONS(4766), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46262] = 7,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    ACTIONS(4768), 1,
      anon_sym_GT,
    ACTIONS(4770), 1,
      anon_sym_SLASH,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46286] = 5,
    ACTIONS(4772), 1,
      sym__newline_token,
    ACTIONS(4775), 1,
      sym__whitespace_ge_2,
    ACTIONS(4778), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4009), 3,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46306] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4591), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46324] = 5,
    ACTIONS(3643), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1009), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46342] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4587), 1,
      anon_sym_RPAREN,
    STATE(927), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46360] = 5,
    ACTIONS(3645), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(926), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46378] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4605), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46396] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4605), 1,
      anon_sym_RPAREN,
    STATE(940), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46414] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4605), 1,
      anon_sym_RPAREN,
    STATE(995), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46432] = 5,
    ACTIONS(3645), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1013), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46450] = 5,
    ACTIONS(3643), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(933), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46468] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4609), 1,
      anon_sym_RPAREN,
    STATE(935), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46486] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4609), 1,
      anon_sym_RPAREN,
    STATE(997), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46504] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4617), 1,
      anon_sym_RPAREN,
    STATE(998), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46522] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4587), 1,
      anon_sym_RPAREN,
    STATE(982), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46540] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4617), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46558] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4607), 1,
      anon_sym_RPAREN,
    STATE(1000), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46576] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4607), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46594] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4781), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46612] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4781), 1,
      anon_sym_RPAREN,
    STATE(1002), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46630] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4783), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46648] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4783), 1,
      anon_sym_RPAREN,
    STATE(1003), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46666] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4785), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46684] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4787), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46702] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4789), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46720] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4619), 1,
      anon_sym_RPAREN,
    STATE(1038), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46738] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4791), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46756] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4611), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46774] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4619), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46792] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4603), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46810] = 5,
    ACTIONS(3647), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1056), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46828] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4603), 1,
      anon_sym_RPAREN,
    STATE(942), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46846] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4603), 1,
      anon_sym_RPAREN,
    STATE(1020), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46864] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4589), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46882] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4589), 1,
      anon_sym_RPAREN,
    STATE(943), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46900] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4589), 1,
      anon_sym_RPAREN,
    STATE(1022), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46918] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4591), 1,
      anon_sym_RPAREN,
    STATE(1066), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46936] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4793), 1,
      anon_sym_GT,
    STATE(1030), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46954] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4621), 1,
      anon_sym_RPAREN,
    STATE(1023), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46972] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4611), 1,
      anon_sym_RPAREN,
    STATE(941), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [46990] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4621), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47008] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4623), 1,
      anon_sym_RPAREN,
    STATE(1029), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47026] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4623), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47044] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4795), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47062] = 5,
    ACTIONS(3692), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(929), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47080] = 5,
    ACTIONS(3692), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1035), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47098] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4795), 1,
      anon_sym_RPAREN,
    STATE(1032), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47116] = 5,
    ACTIONS(3622), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(930), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47134] = 5,
    ACTIONS(3622), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1039), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47152] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4797), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47170] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4799), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47188] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4797), 1,
      anon_sym_RPAREN,
    STATE(1033), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47206] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4801), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47224] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4803), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47242] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4805), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47260] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4595), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47278] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4595), 1,
      anon_sym_RPAREN,
    STATE(931), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47296] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4595), 1,
      anon_sym_RPAREN,
    STATE(1043), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47314] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4807), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47332] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4597), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47350] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4597), 1,
      anon_sym_RPAREN,
    STATE(932), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47368] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4597), 1,
      anon_sym_RPAREN,
    STATE(1046), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47386] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4599), 1,
      anon_sym_RPAREN,
    STATE(1047), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47404] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4599), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47422] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4601), 1,
      anon_sym_RPAREN,
    STATE(1049), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47440] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4809), 1,
      anon_sym_EQ,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47458] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4601), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47476] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4811), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47494] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4811), 1,
      anon_sym_RPAREN,
    STATE(1051), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47512] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4813), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47530] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4813), 1,
      anon_sym_RPAREN,
    STATE(1052), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47548] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4815), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47566] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4817), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47584] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4819), 1,
      anon_sym_EQ,
    STATE(1045), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47602] = 5,
    ACTIONS(4629), 1,
      sym__newline_token,
    ACTIONS(4631), 1,
      sym__whitespace_ge_2,
    ACTIONS(4633), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4714), 1,
      sym__attribute_name,
    STATE(981), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47620] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4821), 1,
      anon_sym_GT,
    STATE(1034), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47638] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4585), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47656] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4823), 1,
      anon_sym_GT,
    STATE(1089), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47674] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4825), 1,
      anon_sym_GT,
    STATE(1063), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47692] = 3,
    ACTIONS(3044), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4827), 1,
      sym__last_token_whitespace,
    ACTIONS(3042), 5,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__attribute_name,
      sym__whitespace_ge_2,
  [47706] = 3,
    ACTIONS(2986), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4829), 1,
      sym__last_token_whitespace,
    ACTIONS(2984), 5,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__attribute_name,
      sym__whitespace_ge_2,
  [47720] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4807), 1,
      anon_sym_RPAREN,
    STATE(1068), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47738] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4585), 1,
      anon_sym_RPAREN,
    STATE(938), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47756] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4831), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47774] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4833), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47792] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4585), 1,
      anon_sym_RPAREN,
    STATE(1083), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47810] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4835), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47828] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4835), 1,
      anon_sym_RPAREN,
    STATE(1078), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47846] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4837), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47864] = 5,
    ACTIONS(3635), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(937), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47882] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4593), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47900] = 5,
    ACTIONS(3635), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1007), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47918] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4839), 1,
      anon_sym_GT,
    STATE(1077), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47936] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4593), 1,
      anon_sym_RPAREN,
    STATE(939), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47954] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4841), 1,
      anon_sym_GT,
    STATE(1004), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47972] = 5,
    ACTIONS(3637), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(925), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [47990] = 5,
    ACTIONS(3637), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1090), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48008] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4843), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48026] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4845), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48044] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4611), 1,
      anon_sym_RPAREN,
    STATE(1008), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48062] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4593), 1,
      anon_sym_RPAREN,
    STATE(1087), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48080] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4613), 1,
      anon_sym_RPAREN,
    STATE(1095), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48098] = 3,
    ACTIONS(2986), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4847), 1,
      sym__last_token_whitespace,
    ACTIONS(2984), 5,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym__newline_token,
      aux_sym__attribute_value_token1,
      sym__whitespace_ge_2,
  [48112] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4613), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48130] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4849), 1,
      anon_sym_GT,
    STATE(1101), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48148] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4615), 1,
      anon_sym_RPAREN,
    STATE(1098), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48166] = 5,
    ACTIONS(3684), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(928), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48184] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4615), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48202] = 3,
    ACTIONS(3044), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4851), 1,
      sym__last_token_whitespace,
    ACTIONS(3042), 5,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym__newline_token,
      aux_sym__attribute_value_token1,
      sym__whitespace_ge_2,
  [48216] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4853), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48234] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4587), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48252] = 5,
    ACTIONS(3614), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(934), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48270] = 5,
    ACTIONS(3614), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(986), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48288] = 5,
    ACTIONS(3647), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(924), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48306] = 5,
    ACTIONS(3682), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(936), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48324] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4855), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48342] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4855), 1,
      anon_sym_RPAREN,
    STATE(1006), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48360] = 5,
    ACTIONS(3682), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1102), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48378] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4857), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48396] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4857), 1,
      anon_sym_RPAREN,
    STATE(1064), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48414] = 5,
    ACTIONS(3684), 1,
      anon_sym_RPAREN,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    STATE(1070), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48432] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4859), 1,
      anon_sym_GT,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48450] = 5,
    ACTIONS(3963), 1,
      sym__newline_token,
    ACTIONS(3967), 1,
      sym__whitespace_ge_2,
    ACTIONS(3969), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4609), 1,
      anon_sym_RPAREN,
    STATE(923), 3,
      sym__whitespace,
      sym__soft_line_break,
      aux_sym_inline_link_repeat1,
  [48468] = 2,
    ACTIONS(3275), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 5,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__attribute_name,
      sym__whitespace_ge_2,
  [48479] = 2,
    ACTIONS(3291), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 5,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym__newline_token,
      aux_sym__attribute_value_token1,
      sym__whitespace_ge_2,
  [48490] = 2,
    ACTIONS(3291), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3289), 5,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__attribute_name,
      sym__whitespace_ge_2,
  [48501] = 2,
    ACTIONS(3275), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3273), 5,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym__newline_token,
      aux_sym__attribute_value_token1,
      sym__whitespace_ge_2,
  [48512] = 2,
    ACTIONS(4863), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4861), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48522] = 2,
    ACTIONS(4867), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4865), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48532] = 2,
    ACTIONS(4871), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4869), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48542] = 2,
    ACTIONS(4875), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4873), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48552] = 2,
    ACTIONS(4879), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4877), 4,
      anon_sym_GT,
      anon_sym_SLASH,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48562] = 2,
    ACTIONS(3900), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3898), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48571] = 2,
    ACTIONS(3927), 1,
      aux_sym__whitespace_token1,
    ACTIONS(3925), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48580] = 2,
    ACTIONS(4883), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4881), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48589] = 2,
    ACTIONS(4887), 1,
      aux_sym__whitespace_token1,
    ACTIONS(4885), 3,
      anon_sym_RPAREN,
      sym__newline_token,
      sym__whitespace_ge_2,
  [48598] = 3,
    ACTIONS(4889), 1,
      anon_sym_LBRACK,
    ACTIONS(4891), 1,
      anon_sym_LPAREN,
    STATE(546), 1,
      sym_link_label,
  [48608] = 3,
    ACTIONS(4893), 1,
      anon_sym_LBRACK,
    ACTIONS(4895), 1,
      anon_sym_LPAREN,
    STATE(473), 1,
      sym_link_label,
  [48618] = 3,
    ACTIONS(4897), 1,
      anon_sym_LBRACK,
    ACTIONS(4899), 1,
      anon_sym_LPAREN,
    STATE(547), 1,
      sym_link_label,
  [48628] = 3,
    ACTIONS(4901), 1,
      anon_sym_LBRACK,
    ACTIONS(4903), 1,
      anon_sym_LPAREN,
    STATE(476), 1,
      sym_link_label,
  [48638] = 3,
    ACTIONS(4905), 1,
      anon_sym_LBRACK,
    ACTIONS(4907), 1,
      anon_sym_LPAREN,
    STATE(394), 1,
      sym_link_label,
  [48648] = 3,
    ACTIONS(4909), 1,
      anon_sym_LBRACK,
    ACTIONS(4911), 1,
      anon_sym_LPAREN,
    STATE(303), 1,
      sym_link_label,
  [48658] = 3,
    ACTIONS(4913), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      anon_sym_LPAREN,
    STATE(521), 1,
      sym_link_label,
  [48668] = 3,
    ACTIONS(4917), 1,
      anon_sym_LBRACK,
    ACTIONS(4919), 1,
      anon_sym_LPAREN,
    STATE(304), 1,
      sym_link_label,
  [48678] = 3,
    ACTIONS(4921), 1,
      anon_sym_LBRACK,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(516), 1,
      sym_link_label,
  [48688] = 3,
    ACTIONS(4925), 1,
      anon_sym_LBRACK,
    ACTIONS(4927), 1,
      anon_sym_LPAREN,
    STATE(396), 1,
      sym_link_label,
  [48698] = 2,
    ACTIONS(3042), 1,
      sym__trigger_error,
    ACTIONS(4929), 1,
      sym__last_token_whitespace,
  [48705] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1057), 1,
      sym__tag_name,
  [48712] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1055), 1,
      sym__tag_name,
  [48719] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1074), 1,
      sym__tag_name,
  [48726] = 1,
    ACTIONS(4933), 2,
      anon_sym_LBRACK,
      anon_sym_LPAREN,
  [48731] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1017), 1,
      sym__tag_name,
  [48738] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1084), 1,
      sym__tag_name,
  [48745] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1058), 1,
      sym__tag_name,
  [48752] = 2,
    ACTIONS(4931), 1,
      sym__word_no_digit,
    STATE(1072), 1,
      sym__tag_name,
  [48759] = 1,
    ACTIONS(4935), 1,
      anon_sym_GT,
  [48763] = 1,
    ACTIONS(4937), 1,
      anon_sym_GT,
  [48767] = 1,
    ACTIONS(4720), 1,
      anon_sym_GT,
  [48771] = 1,
    ACTIONS(4702), 1,
      anon_sym_GT,
  [48775] = 1,
    ACTIONS(4694), 1,
      anon_sym_GT,
  [48779] = 1,
    ACTIONS(4716), 1,
      anon_sym_GT,
  [48783] = 1,
    ACTIONS(4939), 1,
      sym__trigger_error,
  [48787] = 1,
    ACTIONS(4625), 1,
      anon_sym_GT,
  [48791] = 1,
    ACTIONS(4706), 1,
      anon_sym_GT,
  [48795] = 1,
    ACTIONS(4651), 1,
      anon_sym_GT,
  [48799] = 1,
    ACTIONS(4941), 1,
      ts_builtin_sym_end,
  [48803] = 1,
    ACTIONS(4943), 1,
      anon_sym_GT,
  [48807] = 1,
    ACTIONS(4945), 1,
      anon_sym_GT,
  [48811] = 1,
    ACTIONS(4947), 1,
      sym__trigger_error,
  [48815] = 1,
    ACTIONS(4768), 1,
      anon_sym_GT,
  [48819] = 1,
    ACTIONS(4750), 1,
      anon_sym_GT,
  [48823] = 1,
    ACTIONS(4949), 1,
      sym__trigger_error,
  [48827] = 1,
    ACTIONS(4764), 1,
      anon_sym_GT,
  [48831] = 1,
    ACTIONS(4690), 1,
      anon_sym_GT,
  [48835] = 1,
    ACTIONS(3273), 1,
      sym__trigger_error,
  [48839] = 1,
    ACTIONS(4643), 1,
      anon_sym_GT,
  [48843] = 1,
    ACTIONS(4951), 1,
      anon_sym_GT,
  [48847] = 1,
    ACTIONS(4953), 1,
      anon_sym_GT,
  [48851] = 1,
    ACTIONS(4724), 1,
      anon_sym_GT,
  [48855] = 1,
    ACTIONS(4955), 1,
      anon_sym_GT,
  [48859] = 1,
    ACTIONS(4760), 1,
      anon_sym_GT,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(166)] = 0,
  [SMALL_STATE(167)] = 81,
  [SMALL_STATE(168)] = 160,
  [SMALL_STATE(169)] = 243,
  [SMALL_STATE(170)] = 324,
  [SMALL_STATE(171)] = 407,
  [SMALL_STATE(172)] = 488,
  [SMALL_STATE(173)] = 571,
  [SMALL_STATE(174)] = 650,
  [SMALL_STATE(175)] = 729,
  [SMALL_STATE(176)] = 808,
  [SMALL_STATE(177)] = 887,
  [SMALL_STATE(178)] = 966,
  [SMALL_STATE(179)] = 1045,
  [SMALL_STATE(180)] = 1124,
  [SMALL_STATE(181)] = 1204,
  [SMALL_STATE(182)] = 1284,
  [SMALL_STATE(183)] = 1362,
  [SMALL_STATE(184)] = 1442,
  [SMALL_STATE(185)] = 1524,
  [SMALL_STATE(186)] = 1604,
  [SMALL_STATE(187)] = 1686,
  [SMALL_STATE(188)] = 1768,
  [SMALL_STATE(189)] = 1848,
  [SMALL_STATE(190)] = 1928,
  [SMALL_STATE(191)] = 2006,
  [SMALL_STATE(192)] = 2093,
  [SMALL_STATE(193)] = 2174,
  [SMALL_STATE(194)] = 2261,
  [SMALL_STATE(195)] = 2348,
  [SMALL_STATE(196)] = 2435,
  [SMALL_STATE(197)] = 2522,
  [SMALL_STATE(198)] = 2609,
  [SMALL_STATE(199)] = 2696,
  [SMALL_STATE(200)] = 2775,
  [SMALL_STATE(201)] = 2862,
  [SMALL_STATE(202)] = 2949,
  [SMALL_STATE(203)] = 3036,
  [SMALL_STATE(204)] = 3123,
  [SMALL_STATE(205)] = 3191,
  [SMALL_STATE(206)] = 3259,
  [SMALL_STATE(207)] = 3325,
  [SMALL_STATE(208)] = 3393,
  [SMALL_STATE(209)] = 3461,
  [SMALL_STATE(210)] = 3529,
  [SMALL_STATE(211)] = 3597,
  [SMALL_STATE(212)] = 3663,
  [SMALL_STATE(213)] = 3729,
  [SMALL_STATE(214)] = 3797,
  [SMALL_STATE(215)] = 3863,
  [SMALL_STATE(216)] = 3929,
  [SMALL_STATE(217)] = 3997,
  [SMALL_STATE(218)] = 4063,
  [SMALL_STATE(219)] = 4129,
  [SMALL_STATE(220)] = 4195,
  [SMALL_STATE(221)] = 4276,
  [SMALL_STATE(222)] = 4337,
  [SMALL_STATE(223)] = 4400,
  [SMALL_STATE(224)] = 4463,
  [SMALL_STATE(225)] = 4544,
  [SMALL_STATE(226)] = 4605,
  [SMALL_STATE(227)] = 4686,
  [SMALL_STATE(228)] = 4749,
  [SMALL_STATE(229)] = 4830,
  [SMALL_STATE(230)] = 4891,
  [SMALL_STATE(231)] = 4972,
  [SMALL_STATE(232)] = 5053,
  [SMALL_STATE(233)] = 5114,
  [SMALL_STATE(234)] = 5175,
  [SMALL_STATE(235)] = 5236,
  [SMALL_STATE(236)] = 5299,
  [SMALL_STATE(237)] = 5360,
  [SMALL_STATE(238)] = 5421,
  [SMALL_STATE(239)] = 5482,
  [SMALL_STATE(240)] = 5545,
  [SMALL_STATE(241)] = 5626,
  [SMALL_STATE(242)] = 5687,
  [SMALL_STATE(243)] = 5768,
  [SMALL_STATE(244)] = 5835,
  [SMALL_STATE(245)] = 5896,
  [SMALL_STATE(246)] = 5957,
  [SMALL_STATE(247)] = 6038,
  [SMALL_STATE(248)] = 6099,
  [SMALL_STATE(249)] = 6160,
  [SMALL_STATE(250)] = 6225,
  [SMALL_STATE(251)] = 6286,
  [SMALL_STATE(252)] = 6349,
  [SMALL_STATE(253)] = 6412,
  [SMALL_STATE(254)] = 6479,
  [SMALL_STATE(255)] = 6540,
  [SMALL_STATE(256)] = 6605,
  [SMALL_STATE(257)] = 6666,
  [SMALL_STATE(258)] = 6727,
  [SMALL_STATE(259)] = 6788,
  [SMALL_STATE(260)] = 6849,
  [SMALL_STATE(261)] = 6930,
  [SMALL_STATE(262)] = 7011,
  [SMALL_STATE(263)] = 7069,
  [SMALL_STATE(264)] = 7127,
  [SMALL_STATE(265)] = 7185,
  [SMALL_STATE(266)] = 7243,
  [SMALL_STATE(267)] = 7301,
  [SMALL_STATE(268)] = 7359,
  [SMALL_STATE(269)] = 7417,
  [SMALL_STATE(270)] = 7475,
  [SMALL_STATE(271)] = 7533,
  [SMALL_STATE(272)] = 7591,
  [SMALL_STATE(273)] = 7649,
  [SMALL_STATE(274)] = 7707,
  [SMALL_STATE(275)] = 7765,
  [SMALL_STATE(276)] = 7823,
  [SMALL_STATE(277)] = 7881,
  [SMALL_STATE(278)] = 7939,
  [SMALL_STATE(279)] = 7997,
  [SMALL_STATE(280)] = 8069,
  [SMALL_STATE(281)] = 8127,
  [SMALL_STATE(282)] = 8185,
  [SMALL_STATE(283)] = 8243,
  [SMALL_STATE(284)] = 8301,
  [SMALL_STATE(285)] = 8359,
  [SMALL_STATE(286)] = 8417,
  [SMALL_STATE(287)] = 8475,
  [SMALL_STATE(288)] = 8533,
  [SMALL_STATE(289)] = 8593,
  [SMALL_STATE(290)] = 8651,
  [SMALL_STATE(291)] = 8711,
  [SMALL_STATE(292)] = 8769,
  [SMALL_STATE(293)] = 8827,
  [SMALL_STATE(294)] = 8885,
  [SMALL_STATE(295)] = 8943,
  [SMALL_STATE(296)] = 9001,
  [SMALL_STATE(297)] = 9059,
  [SMALL_STATE(298)] = 9117,
  [SMALL_STATE(299)] = 9175,
  [SMALL_STATE(300)] = 9233,
  [SMALL_STATE(301)] = 9291,
  [SMALL_STATE(302)] = 9349,
  [SMALL_STATE(303)] = 9407,
  [SMALL_STATE(304)] = 9465,
  [SMALL_STATE(305)] = 9523,
  [SMALL_STATE(306)] = 9581,
  [SMALL_STATE(307)] = 9639,
  [SMALL_STATE(308)] = 9697,
  [SMALL_STATE(309)] = 9755,
  [SMALL_STATE(310)] = 9813,
  [SMALL_STATE(311)] = 9871,
  [SMALL_STATE(312)] = 9929,
  [SMALL_STATE(313)] = 9987,
  [SMALL_STATE(314)] = 10045,
  [SMALL_STATE(315)] = 10103,
  [SMALL_STATE(316)] = 10161,
  [SMALL_STATE(317)] = 10219,
  [SMALL_STATE(318)] = 10277,
  [SMALL_STATE(319)] = 10335,
  [SMALL_STATE(320)] = 10393,
  [SMALL_STATE(321)] = 10451,
  [SMALL_STATE(322)] = 10509,
  [SMALL_STATE(323)] = 10567,
  [SMALL_STATE(324)] = 10625,
  [SMALL_STATE(325)] = 10683,
  [SMALL_STATE(326)] = 10741,
  [SMALL_STATE(327)] = 10799,
  [SMALL_STATE(328)] = 10857,
  [SMALL_STATE(329)] = 10915,
  [SMALL_STATE(330)] = 10973,
  [SMALL_STATE(331)] = 11031,
  [SMALL_STATE(332)] = 11089,
  [SMALL_STATE(333)] = 11147,
  [SMALL_STATE(334)] = 11205,
  [SMALL_STATE(335)] = 11263,
  [SMALL_STATE(336)] = 11321,
  [SMALL_STATE(337)] = 11379,
  [SMALL_STATE(338)] = 11437,
  [SMALL_STATE(339)] = 11495,
  [SMALL_STATE(340)] = 11553,
  [SMALL_STATE(341)] = 11611,
  [SMALL_STATE(342)] = 11669,
  [SMALL_STATE(343)] = 11727,
  [SMALL_STATE(344)] = 11785,
  [SMALL_STATE(345)] = 11843,
  [SMALL_STATE(346)] = 11901,
  [SMALL_STATE(347)] = 11959,
  [SMALL_STATE(348)] = 12017,
  [SMALL_STATE(349)] = 12075,
  [SMALL_STATE(350)] = 12137,
  [SMALL_STATE(351)] = 12195,
  [SMALL_STATE(352)] = 12253,
  [SMALL_STATE(353)] = 12315,
  [SMALL_STATE(354)] = 12373,
  [SMALL_STATE(355)] = 12431,
  [SMALL_STATE(356)] = 12489,
  [SMALL_STATE(357)] = 12549,
  [SMALL_STATE(358)] = 12607,
  [SMALL_STATE(359)] = 12665,
  [SMALL_STATE(360)] = 12723,
  [SMALL_STATE(361)] = 12781,
  [SMALL_STATE(362)] = 12839,
  [SMALL_STATE(363)] = 12897,
  [SMALL_STATE(364)] = 12955,
  [SMALL_STATE(365)] = 13013,
  [SMALL_STATE(366)] = 13071,
  [SMALL_STATE(367)] = 13129,
  [SMALL_STATE(368)] = 13189,
  [SMALL_STATE(369)] = 13247,
  [SMALL_STATE(370)] = 13305,
  [SMALL_STATE(371)] = 13363,
  [SMALL_STATE(372)] = 13421,
  [SMALL_STATE(373)] = 13479,
  [SMALL_STATE(374)] = 13537,
  [SMALL_STATE(375)] = 13595,
  [SMALL_STATE(376)] = 13655,
  [SMALL_STATE(377)] = 13713,
  [SMALL_STATE(378)] = 13771,
  [SMALL_STATE(379)] = 13829,
  [SMALL_STATE(380)] = 13887,
  [SMALL_STATE(381)] = 13945,
  [SMALL_STATE(382)] = 14003,
  [SMALL_STATE(383)] = 14061,
  [SMALL_STATE(384)] = 14123,
  [SMALL_STATE(385)] = 14181,
  [SMALL_STATE(386)] = 14241,
  [SMALL_STATE(387)] = 14299,
  [SMALL_STATE(388)] = 14357,
  [SMALL_STATE(389)] = 14417,
  [SMALL_STATE(390)] = 14479,
  [SMALL_STATE(391)] = 14539,
  [SMALL_STATE(392)] = 14597,
  [SMALL_STATE(393)] = 14655,
  [SMALL_STATE(394)] = 14713,
  [SMALL_STATE(395)] = 14771,
  [SMALL_STATE(396)] = 14831,
  [SMALL_STATE(397)] = 14889,
  [SMALL_STATE(398)] = 14961,
  [SMALL_STATE(399)] = 15019,
  [SMALL_STATE(400)] = 15081,
  [SMALL_STATE(401)] = 15139,
  [SMALL_STATE(402)] = 15197,
  [SMALL_STATE(403)] = 15255,
  [SMALL_STATE(404)] = 15313,
  [SMALL_STATE(405)] = 15371,
  [SMALL_STATE(406)] = 15429,
  [SMALL_STATE(407)] = 15491,
  [SMALL_STATE(408)] = 15549,
  [SMALL_STATE(409)] = 15607,
  [SMALL_STATE(410)] = 15665,
  [SMALL_STATE(411)] = 15723,
  [SMALL_STATE(412)] = 15781,
  [SMALL_STATE(413)] = 15839,
  [SMALL_STATE(414)] = 15897,
  [SMALL_STATE(415)] = 15955,
  [SMALL_STATE(416)] = 16013,
  [SMALL_STATE(417)] = 16071,
  [SMALL_STATE(418)] = 16129,
  [SMALL_STATE(419)] = 16187,
  [SMALL_STATE(420)] = 16245,
  [SMALL_STATE(421)] = 16303,
  [SMALL_STATE(422)] = 16361,
  [SMALL_STATE(423)] = 16421,
  [SMALL_STATE(424)] = 16481,
  [SMALL_STATE(425)] = 16541,
  [SMALL_STATE(426)] = 16599,
  [SMALL_STATE(427)] = 16657,
  [SMALL_STATE(428)] = 16715,
  [SMALL_STATE(429)] = 16787,
  [SMALL_STATE(430)] = 16859,
  [SMALL_STATE(431)] = 16917,
  [SMALL_STATE(432)] = 16975,
  [SMALL_STATE(433)] = 17033,
  [SMALL_STATE(434)] = 17091,
  [SMALL_STATE(435)] = 17149,
  [SMALL_STATE(436)] = 17207,
  [SMALL_STATE(437)] = 17265,
  [SMALL_STATE(438)] = 17337,
  [SMALL_STATE(439)] = 17395,
  [SMALL_STATE(440)] = 17453,
  [SMALL_STATE(441)] = 17511,
  [SMALL_STATE(442)] = 17569,
  [SMALL_STATE(443)] = 17627,
  [SMALL_STATE(444)] = 17685,
  [SMALL_STATE(445)] = 17743,
  [SMALL_STATE(446)] = 17803,
  [SMALL_STATE(447)] = 17861,
  [SMALL_STATE(448)] = 17921,
  [SMALL_STATE(449)] = 17979,
  [SMALL_STATE(450)] = 18039,
  [SMALL_STATE(451)] = 18097,
  [SMALL_STATE(452)] = 18169,
  [SMALL_STATE(453)] = 18241,
  [SMALL_STATE(454)] = 18299,
  [SMALL_STATE(455)] = 18359,
  [SMALL_STATE(456)] = 18419,
  [SMALL_STATE(457)] = 18477,
  [SMALL_STATE(458)] = 18535,
  [SMALL_STATE(459)] = 18593,
  [SMALL_STATE(460)] = 18651,
  [SMALL_STATE(461)] = 18723,
  [SMALL_STATE(462)] = 18781,
  [SMALL_STATE(463)] = 18839,
  [SMALL_STATE(464)] = 18897,
  [SMALL_STATE(465)] = 18955,
  [SMALL_STATE(466)] = 19013,
  [SMALL_STATE(467)] = 19071,
  [SMALL_STATE(468)] = 19129,
  [SMALL_STATE(469)] = 19189,
  [SMALL_STATE(470)] = 19249,
  [SMALL_STATE(471)] = 19309,
  [SMALL_STATE(472)] = 19367,
  [SMALL_STATE(473)] = 19425,
  [SMALL_STATE(474)] = 19483,
  [SMALL_STATE(475)] = 19555,
  [SMALL_STATE(476)] = 19627,
  [SMALL_STATE(477)] = 19685,
  [SMALL_STATE(478)] = 19743,
  [SMALL_STATE(479)] = 19803,
  [SMALL_STATE(480)] = 19861,
  [SMALL_STATE(481)] = 19919,
  [SMALL_STATE(482)] = 19977,
  [SMALL_STATE(483)] = 20035,
  [SMALL_STATE(484)] = 20107,
  [SMALL_STATE(485)] = 20165,
  [SMALL_STATE(486)] = 20223,
  [SMALL_STATE(487)] = 20281,
  [SMALL_STATE(488)] = 20339,
  [SMALL_STATE(489)] = 20397,
  [SMALL_STATE(490)] = 20455,
  [SMALL_STATE(491)] = 20513,
  [SMALL_STATE(492)] = 20573,
  [SMALL_STATE(493)] = 20631,
  [SMALL_STATE(494)] = 20691,
  [SMALL_STATE(495)] = 20749,
  [SMALL_STATE(496)] = 20807,
  [SMALL_STATE(497)] = 20865,
  [SMALL_STATE(498)] = 20937,
  [SMALL_STATE(499)] = 21009,
  [SMALL_STATE(500)] = 21067,
  [SMALL_STATE(501)] = 21125,
  [SMALL_STATE(502)] = 21183,
  [SMALL_STATE(503)] = 21241,
  [SMALL_STATE(504)] = 21313,
  [SMALL_STATE(505)] = 21371,
  [SMALL_STATE(506)] = 21429,
  [SMALL_STATE(507)] = 21487,
  [SMALL_STATE(508)] = 21545,
  [SMALL_STATE(509)] = 21603,
  [SMALL_STATE(510)] = 21661,
  [SMALL_STATE(511)] = 21719,
  [SMALL_STATE(512)] = 21777,
  [SMALL_STATE(513)] = 21835,
  [SMALL_STATE(514)] = 21893,
  [SMALL_STATE(515)] = 21951,
  [SMALL_STATE(516)] = 22023,
  [SMALL_STATE(517)] = 22081,
  [SMALL_STATE(518)] = 22153,
  [SMALL_STATE(519)] = 22211,
  [SMALL_STATE(520)] = 22269,
  [SMALL_STATE(521)] = 22327,
  [SMALL_STATE(522)] = 22385,
  [SMALL_STATE(523)] = 22443,
  [SMALL_STATE(524)] = 22501,
  [SMALL_STATE(525)] = 22559,
  [SMALL_STATE(526)] = 22617,
  [SMALL_STATE(527)] = 22675,
  [SMALL_STATE(528)] = 22733,
  [SMALL_STATE(529)] = 22791,
  [SMALL_STATE(530)] = 22849,
  [SMALL_STATE(531)] = 22907,
  [SMALL_STATE(532)] = 22965,
  [SMALL_STATE(533)] = 23023,
  [SMALL_STATE(534)] = 23081,
  [SMALL_STATE(535)] = 23139,
  [SMALL_STATE(536)] = 23197,
  [SMALL_STATE(537)] = 23258,
  [SMALL_STATE(538)] = 23315,
  [SMALL_STATE(539)] = 23372,
  [SMALL_STATE(540)] = 23429,
  [SMALL_STATE(541)] = 23488,
  [SMALL_STATE(542)] = 23545,
  [SMALL_STATE(543)] = 23602,
  [SMALL_STATE(544)] = 23659,
  [SMALL_STATE(545)] = 23716,
  [SMALL_STATE(546)] = 23773,
  [SMALL_STATE(547)] = 23830,
  [SMALL_STATE(548)] = 23887,
  [SMALL_STATE(549)] = 23944,
  [SMALL_STATE(550)] = 24001,
  [SMALL_STATE(551)] = 24058,
  [SMALL_STATE(552)] = 24115,
  [SMALL_STATE(553)] = 24172,
  [SMALL_STATE(554)] = 24229,
  [SMALL_STATE(555)] = 24286,
  [SMALL_STATE(556)] = 24343,
  [SMALL_STATE(557)] = 24400,
  [SMALL_STATE(558)] = 24457,
  [SMALL_STATE(559)] = 24514,
  [SMALL_STATE(560)] = 24571,
  [SMALL_STATE(561)] = 24628,
  [SMALL_STATE(562)] = 24685,
  [SMALL_STATE(563)] = 24742,
  [SMALL_STATE(564)] = 24799,
  [SMALL_STATE(565)] = 24858,
  [SMALL_STATE(566)] = 24915,
  [SMALL_STATE(567)] = 24972,
  [SMALL_STATE(568)] = 25029,
  [SMALL_STATE(569)] = 25086,
  [SMALL_STATE(570)] = 25143,
  [SMALL_STATE(571)] = 25200,
  [SMALL_STATE(572)] = 25257,
  [SMALL_STATE(573)] = 25314,
  [SMALL_STATE(574)] = 25371,
  [SMALL_STATE(575)] = 25428,
  [SMALL_STATE(576)] = 25489,
  [SMALL_STATE(577)] = 25550,
  [SMALL_STATE(578)] = 25607,
  [SMALL_STATE(579)] = 25664,
  [SMALL_STATE(580)] = 25721,
  [SMALL_STATE(581)] = 25778,
  [SMALL_STATE(582)] = 25835,
  [SMALL_STATE(583)] = 25892,
  [SMALL_STATE(584)] = 25949,
  [SMALL_STATE(585)] = 26006,
  [SMALL_STATE(586)] = 26063,
  [SMALL_STATE(587)] = 26120,
  [SMALL_STATE(588)] = 26177,
  [SMALL_STATE(589)] = 26234,
  [SMALL_STATE(590)] = 26291,
  [SMALL_STATE(591)] = 26348,
  [SMALL_STATE(592)] = 26405,
  [SMALL_STATE(593)] = 26462,
  [SMALL_STATE(594)] = 26519,
  [SMALL_STATE(595)] = 26576,
  [SMALL_STATE(596)] = 26633,
  [SMALL_STATE(597)] = 26690,
  [SMALL_STATE(598)] = 26747,
  [SMALL_STATE(599)] = 26808,
  [SMALL_STATE(600)] = 26865,
  [SMALL_STATE(601)] = 26921,
  [SMALL_STATE(602)] = 26977,
  [SMALL_STATE(603)] = 27033,
  [SMALL_STATE(604)] = 27089,
  [SMALL_STATE(605)] = 27145,
  [SMALL_STATE(606)] = 27201,
  [SMALL_STATE(607)] = 27257,
  [SMALL_STATE(608)] = 27313,
  [SMALL_STATE(609)] = 27369,
  [SMALL_STATE(610)] = 27425,
  [SMALL_STATE(611)] = 27481,
  [SMALL_STATE(612)] = 27537,
  [SMALL_STATE(613)] = 27593,
  [SMALL_STATE(614)] = 27649,
  [SMALL_STATE(615)] = 27705,
  [SMALL_STATE(616)] = 27761,
  [SMALL_STATE(617)] = 27817,
  [SMALL_STATE(618)] = 27873,
  [SMALL_STATE(619)] = 27929,
  [SMALL_STATE(620)] = 27985,
  [SMALL_STATE(621)] = 28041,
  [SMALL_STATE(622)] = 28099,
  [SMALL_STATE(623)] = 28155,
  [SMALL_STATE(624)] = 28211,
  [SMALL_STATE(625)] = 28267,
  [SMALL_STATE(626)] = 28323,
  [SMALL_STATE(627)] = 28381,
  [SMALL_STATE(628)] = 28439,
  [SMALL_STATE(629)] = 28497,
  [SMALL_STATE(630)] = 28555,
  [SMALL_STATE(631)] = 28613,
  [SMALL_STATE(632)] = 28669,
  [SMALL_STATE(633)] = 28725,
  [SMALL_STATE(634)] = 28781,
  [SMALL_STATE(635)] = 28837,
  [SMALL_STATE(636)] = 28893,
  [SMALL_STATE(637)] = 28949,
  [SMALL_STATE(638)] = 29005,
  [SMALL_STATE(639)] = 29061,
  [SMALL_STATE(640)] = 29119,
  [SMALL_STATE(641)] = 29177,
  [SMALL_STATE(642)] = 29233,
  [SMALL_STATE(643)] = 29289,
  [SMALL_STATE(644)] = 29345,
  [SMALL_STATE(645)] = 29401,
  [SMALL_STATE(646)] = 29482,
  [SMALL_STATE(647)] = 29563,
  [SMALL_STATE(648)] = 29644,
  [SMALL_STATE(649)] = 29725,
  [SMALL_STATE(650)] = 29792,
  [SMALL_STATE(651)] = 29873,
  [SMALL_STATE(652)] = 29954,
  [SMALL_STATE(653)] = 30035,
  [SMALL_STATE(654)] = 30116,
  [SMALL_STATE(655)] = 30197,
  [SMALL_STATE(656)] = 30278,
  [SMALL_STATE(657)] = 30359,
  [SMALL_STATE(658)] = 30440,
  [SMALL_STATE(659)] = 30519,
  [SMALL_STATE(660)] = 30600,
  [SMALL_STATE(661)] = 30681,
  [SMALL_STATE(662)] = 30748,
  [SMALL_STATE(663)] = 30829,
  [SMALL_STATE(664)] = 30910,
  [SMALL_STATE(665)] = 30991,
  [SMALL_STATE(666)] = 31072,
  [SMALL_STATE(667)] = 31153,
  [SMALL_STATE(668)] = 31234,
  [SMALL_STATE(669)] = 31313,
  [SMALL_STATE(670)] = 31379,
  [SMALL_STATE(671)] = 31433,
  [SMALL_STATE(672)] = 31487,
  [SMALL_STATE(673)] = 31561,
  [SMALL_STATE(674)] = 31629,
  [SMALL_STATE(675)] = 31695,
  [SMALL_STATE(676)] = 31749,
  [SMALL_STATE(677)] = 31800,
  [SMALL_STATE(678)] = 31851,
  [SMALL_STATE(679)] = 31902,
  [SMALL_STATE(680)] = 31969,
  [SMALL_STATE(681)] = 32020,
  [SMALL_STATE(682)] = 32086,
  [SMALL_STATE(683)] = 32144,
  [SMALL_STATE(684)] = 32202,
  [SMALL_STATE(685)] = 32266,
  [SMALL_STATE(686)] = 32332,
  [SMALL_STATE(687)] = 32398,
  [SMALL_STATE(688)] = 32464,
  [SMALL_STATE(689)] = 32530,
  [SMALL_STATE(690)] = 32596,
  [SMALL_STATE(691)] = 32662,
  [SMALL_STATE(692)] = 32728,
  [SMALL_STATE(693)] = 32779,
  [SMALL_STATE(694)] = 32844,
  [SMALL_STATE(695)] = 32905,
  [SMALL_STATE(696)] = 32956,
  [SMALL_STATE(697)] = 33017,
  [SMALL_STATE(698)] = 33082,
  [SMALL_STATE(699)] = 33143,
  [SMALL_STATE(700)] = 33204,
  [SMALL_STATE(701)] = 33261,
  [SMALL_STATE(702)] = 33318,
  [SMALL_STATE(703)] = 33379,
  [SMALL_STATE(704)] = 33444,
  [SMALL_STATE(705)] = 33495,
  [SMALL_STATE(706)] = 33543,
  [SMALL_STATE(707)] = 33605,
  [SMALL_STATE(708)] = 33653,
  [SMALL_STATE(709)] = 33701,
  [SMALL_STATE(710)] = 33749,
  [SMALL_STATE(711)] = 33811,
  [SMALL_STATE(712)] = 33859,
  [SMALL_STATE(713)] = 33921,
  [SMALL_STATE(714)] = 33981,
  [SMALL_STATE(715)] = 34031,
  [SMALL_STATE(716)] = 34093,
  [SMALL_STATE(717)] = 34155,
  [SMALL_STATE(718)] = 34211,
  [SMALL_STATE(719)] = 34273,
  [SMALL_STATE(720)] = 34333,
  [SMALL_STATE(721)] = 34395,
  [SMALL_STATE(722)] = 34445,
  [SMALL_STATE(723)] = 34493,
  [SMALL_STATE(724)] = 34543,
  [SMALL_STATE(725)] = 34591,
  [SMALL_STATE(726)] = 34653,
  [SMALL_STATE(727)] = 34701,
  [SMALL_STATE(728)] = 34749,
  [SMALL_STATE(729)] = 34797,
  [SMALL_STATE(730)] = 34845,
  [SMALL_STATE(731)] = 34893,
  [SMALL_STATE(732)] = 34941,
  [SMALL_STATE(733)] = 35003,
  [SMALL_STATE(734)] = 35051,
  [SMALL_STATE(735)] = 35099,
  [SMALL_STATE(736)] = 35147,
  [SMALL_STATE(737)] = 35195,
  [SMALL_STATE(738)] = 35255,
  [SMALL_STATE(739)] = 35303,
  [SMALL_STATE(740)] = 35351,
  [SMALL_STATE(741)] = 35413,
  [SMALL_STATE(742)] = 35461,
  [SMALL_STATE(743)] = 35523,
  [SMALL_STATE(744)] = 35571,
  [SMALL_STATE(745)] = 35633,
  [SMALL_STATE(746)] = 35695,
  [SMALL_STATE(747)] = 35757,
  [SMALL_STATE(748)] = 35805,
  [SMALL_STATE(749)] = 35852,
  [SMALL_STATE(750)] = 35899,
  [SMALL_STATE(751)] = 35946,
  [SMALL_STATE(752)] = 36005,
  [SMALL_STATE(753)] = 36052,
  [SMALL_STATE(754)] = 36111,
  [SMALL_STATE(755)] = 36170,
  [SMALL_STATE(756)] = 36229,
  [SMALL_STATE(757)] = 36288,
  [SMALL_STATE(758)] = 36347,
  [SMALL_STATE(759)] = 36394,
  [SMALL_STATE(760)] = 36453,
  [SMALL_STATE(761)] = 36500,
  [SMALL_STATE(762)] = 36547,
  [SMALL_STATE(763)] = 36606,
  [SMALL_STATE(764)] = 36665,
  [SMALL_STATE(765)] = 36724,
  [SMALL_STATE(766)] = 36783,
  [SMALL_STATE(767)] = 36842,
  [SMALL_STATE(768)] = 36901,
  [SMALL_STATE(769)] = 36960,
  [SMALL_STATE(770)] = 37019,
  [SMALL_STATE(771)] = 37066,
  [SMALL_STATE(772)] = 37113,
  [SMALL_STATE(773)] = 37160,
  [SMALL_STATE(774)] = 37219,
  [SMALL_STATE(775)] = 37278,
  [SMALL_STATE(776)] = 37337,
  [SMALL_STATE(777)] = 37396,
  [SMALL_STATE(778)] = 37455,
  [SMALL_STATE(779)] = 37502,
  [SMALL_STATE(780)] = 37549,
  [SMALL_STATE(781)] = 37608,
  [SMALL_STATE(782)] = 37667,
  [SMALL_STATE(783)] = 37726,
  [SMALL_STATE(784)] = 37785,
  [SMALL_STATE(785)] = 37844,
  [SMALL_STATE(786)] = 37903,
  [SMALL_STATE(787)] = 37962,
  [SMALL_STATE(788)] = 38021,
  [SMALL_STATE(789)] = 38080,
  [SMALL_STATE(790)] = 38139,
  [SMALL_STATE(791)] = 38198,
  [SMALL_STATE(792)] = 38257,
  [SMALL_STATE(793)] = 38304,
  [SMALL_STATE(794)] = 38351,
  [SMALL_STATE(795)] = 38398,
  [SMALL_STATE(796)] = 38457,
  [SMALL_STATE(797)] = 38516,
  [SMALL_STATE(798)] = 38563,
  [SMALL_STATE(799)] = 38610,
  [SMALL_STATE(800)] = 38657,
  [SMALL_STATE(801)] = 38704,
  [SMALL_STATE(802)] = 38763,
  [SMALL_STATE(803)] = 38822,
  [SMALL_STATE(804)] = 38881,
  [SMALL_STATE(805)] = 38928,
  [SMALL_STATE(806)] = 38987,
  [SMALL_STATE(807)] = 39046,
  [SMALL_STATE(808)] = 39105,
  [SMALL_STATE(809)] = 39164,
  [SMALL_STATE(810)] = 39223,
  [SMALL_STATE(811)] = 39282,
  [SMALL_STATE(812)] = 39341,
  [SMALL_STATE(813)] = 39400,
  [SMALL_STATE(814)] = 39459,
  [SMALL_STATE(815)] = 39518,
  [SMALL_STATE(816)] = 39577,
  [SMALL_STATE(817)] = 39636,
  [SMALL_STATE(818)] = 39695,
  [SMALL_STATE(819)] = 39754,
  [SMALL_STATE(820)] = 39813,
  [SMALL_STATE(821)] = 39872,
  [SMALL_STATE(822)] = 39919,
  [SMALL_STATE(823)] = 39978,
  [SMALL_STATE(824)] = 40037,
  [SMALL_STATE(825)] = 40096,
  [SMALL_STATE(826)] = 40155,
  [SMALL_STATE(827)] = 40214,
  [SMALL_STATE(828)] = 40273,
  [SMALL_STATE(829)] = 40329,
  [SMALL_STATE(830)] = 40377,
  [SMALL_STATE(831)] = 40427,
  [SMALL_STATE(832)] = 40481,
  [SMALL_STATE(833)] = 40537,
  [SMALL_STATE(834)] = 40593,
  [SMALL_STATE(835)] = 40649,
  [SMALL_STATE(836)] = 40705,
  [SMALL_STATE(837)] = 40761,
  [SMALL_STATE(838)] = 40809,
  [SMALL_STATE(839)] = 40857,
  [SMALL_STATE(840)] = 40913,
  [SMALL_STATE(841)] = 40969,
  [SMALL_STATE(842)] = 41017,
  [SMALL_STATE(843)] = 41073,
  [SMALL_STATE(844)] = 41129,
  [SMALL_STATE(845)] = 41185,
  [SMALL_STATE(846)] = 41241,
  [SMALL_STATE(847)] = 41297,
  [SMALL_STATE(848)] = 41353,
  [SMALL_STATE(849)] = 41409,
  [SMALL_STATE(850)] = 41457,
  [SMALL_STATE(851)] = 41509,
  [SMALL_STATE(852)] = 41557,
  [SMALL_STATE(853)] = 41613,
  [SMALL_STATE(854)] = 41663,
  [SMALL_STATE(855)] = 41719,
  [SMALL_STATE(856)] = 41775,
  [SMALL_STATE(857)] = 41831,
  [SMALL_STATE(858)] = 41887,
  [SMALL_STATE(859)] = 41935,
  [SMALL_STATE(860)] = 41991,
  [SMALL_STATE(861)] = 42047,
  [SMALL_STATE(862)] = 42103,
  [SMALL_STATE(863)] = 42154,
  [SMALL_STATE(864)] = 42199,
  [SMALL_STATE(865)] = 42250,
  [SMALL_STATE(866)] = 42295,
  [SMALL_STATE(867)] = 42340,
  [SMALL_STATE(868)] = 42385,
  [SMALL_STATE(869)] = 42430,
  [SMALL_STATE(870)] = 42479,
  [SMALL_STATE(871)] = 42528,
  [SMALL_STATE(872)] = 42573,
  [SMALL_STATE(873)] = 42618,
  [SMALL_STATE(874)] = 42663,
  [SMALL_STATE(875)] = 42708,
  [SMALL_STATE(876)] = 42753,
  [SMALL_STATE(877)] = 42800,
  [SMALL_STATE(878)] = 42845,
  [SMALL_STATE(879)] = 42890,
  [SMALL_STATE(880)] = 42935,
  [SMALL_STATE(881)] = 42980,
  [SMALL_STATE(882)] = 43026,
  [SMALL_STATE(883)] = 43072,
  [SMALL_STATE(884)] = 43118,
  [SMALL_STATE(885)] = 43164,
  [SMALL_STATE(886)] = 43210,
  [SMALL_STATE(887)] = 43256,
  [SMALL_STATE(888)] = 43302,
  [SMALL_STATE(889)] = 43348,
  [SMALL_STATE(890)] = 43400,
  [SMALL_STATE(891)] = 43446,
  [SMALL_STATE(892)] = 43492,
  [SMALL_STATE(893)] = 43536,
  [SMALL_STATE(894)] = 43580,
  [SMALL_STATE(895)] = 43626,
  [SMALL_STATE(896)] = 43671,
  [SMALL_STATE(897)] = 43714,
  [SMALL_STATE(898)] = 43759,
  [SMALL_STATE(899)] = 43802,
  [SMALL_STATE(900)] = 43845,
  [SMALL_STATE(901)] = 43888,
  [SMALL_STATE(902)] = 43931,
  [SMALL_STATE(903)] = 43974,
  [SMALL_STATE(904)] = 44017,
  [SMALL_STATE(905)] = 44062,
  [SMALL_STATE(906)] = 44105,
  [SMALL_STATE(907)] = 44148,
  [SMALL_STATE(908)] = 44191,
  [SMALL_STATE(909)] = 44234,
  [SMALL_STATE(910)] = 44279,
  [SMALL_STATE(911)] = 44324,
  [SMALL_STATE(912)] = 44366,
  [SMALL_STATE(913)] = 44408,
  [SMALL_STATE(914)] = 44450,
  [SMALL_STATE(915)] = 44492,
  [SMALL_STATE(916)] = 44534,
  [SMALL_STATE(917)] = 44568,
  [SMALL_STATE(918)] = 44602,
  [SMALL_STATE(919)] = 44636,
  [SMALL_STATE(920)] = 44670,
  [SMALL_STATE(921)] = 44704,
  [SMALL_STATE(922)] = 44738,
  [SMALL_STATE(923)] = 44772,
  [SMALL_STATE(924)] = 44795,
  [SMALL_STATE(925)] = 44825,
  [SMALL_STATE(926)] = 44855,
  [SMALL_STATE(927)] = 44885,
  [SMALL_STATE(928)] = 44915,
  [SMALL_STATE(929)] = 44945,
  [SMALL_STATE(930)] = 44975,
  [SMALL_STATE(931)] = 45005,
  [SMALL_STATE(932)] = 45035,
  [SMALL_STATE(933)] = 45065,
  [SMALL_STATE(934)] = 45095,
  [SMALL_STATE(935)] = 45125,
  [SMALL_STATE(936)] = 45155,
  [SMALL_STATE(937)] = 45185,
  [SMALL_STATE(938)] = 45215,
  [SMALL_STATE(939)] = 45245,
  [SMALL_STATE(940)] = 45275,
  [SMALL_STATE(941)] = 45305,
  [SMALL_STATE(942)] = 45335,
  [SMALL_STATE(943)] = 45365,
  [SMALL_STATE(944)] = 45395,
  [SMALL_STATE(945)] = 45420,
  [SMALL_STATE(946)] = 45445,
  [SMALL_STATE(947)] = 45470,
  [SMALL_STATE(948)] = 45495,
  [SMALL_STATE(949)] = 45520,
  [SMALL_STATE(950)] = 45545,
  [SMALL_STATE(951)] = 45568,
  [SMALL_STATE(952)] = 45595,
  [SMALL_STATE(953)] = 45620,
  [SMALL_STATE(954)] = 45647,
  [SMALL_STATE(955)] = 45674,
  [SMALL_STATE(956)] = 45701,
  [SMALL_STATE(957)] = 45726,
  [SMALL_STATE(958)] = 45751,
  [SMALL_STATE(959)] = 45776,
  [SMALL_STATE(960)] = 45801,
  [SMALL_STATE(961)] = 45826,
  [SMALL_STATE(962)] = 45851,
  [SMALL_STATE(963)] = 45876,
  [SMALL_STATE(964)] = 45900,
  [SMALL_STATE(965)] = 45924,
  [SMALL_STATE(966)] = 45948,
  [SMALL_STATE(967)] = 45972,
  [SMALL_STATE(968)] = 45990,
  [SMALL_STATE(969)] = 46008,
  [SMALL_STATE(970)] = 46028,
  [SMALL_STATE(971)] = 46052,
  [SMALL_STATE(972)] = 46076,
  [SMALL_STATE(973)] = 46100,
  [SMALL_STATE(974)] = 46118,
  [SMALL_STATE(975)] = 46142,
  [SMALL_STATE(976)] = 46166,
  [SMALL_STATE(977)] = 46190,
  [SMALL_STATE(978)] = 46214,
  [SMALL_STATE(979)] = 46238,
  [SMALL_STATE(980)] = 46262,
  [SMALL_STATE(981)] = 46286,
  [SMALL_STATE(982)] = 46306,
  [SMALL_STATE(983)] = 46324,
  [SMALL_STATE(984)] = 46342,
  [SMALL_STATE(985)] = 46360,
  [SMALL_STATE(986)] = 46378,
  [SMALL_STATE(987)] = 46396,
  [SMALL_STATE(988)] = 46414,
  [SMALL_STATE(989)] = 46432,
  [SMALL_STATE(990)] = 46450,
  [SMALL_STATE(991)] = 46468,
  [SMALL_STATE(992)] = 46486,
  [SMALL_STATE(993)] = 46504,
  [SMALL_STATE(994)] = 46522,
  [SMALL_STATE(995)] = 46540,
  [SMALL_STATE(996)] = 46558,
  [SMALL_STATE(997)] = 46576,
  [SMALL_STATE(998)] = 46594,
  [SMALL_STATE(999)] = 46612,
  [SMALL_STATE(1000)] = 46630,
  [SMALL_STATE(1001)] = 46648,
  [SMALL_STATE(1002)] = 46666,
  [SMALL_STATE(1003)] = 46684,
  [SMALL_STATE(1004)] = 46702,
  [SMALL_STATE(1005)] = 46720,
  [SMALL_STATE(1006)] = 46738,
  [SMALL_STATE(1007)] = 46756,
  [SMALL_STATE(1008)] = 46774,
  [SMALL_STATE(1009)] = 46792,
  [SMALL_STATE(1010)] = 46810,
  [SMALL_STATE(1011)] = 46828,
  [SMALL_STATE(1012)] = 46846,
  [SMALL_STATE(1013)] = 46864,
  [SMALL_STATE(1014)] = 46882,
  [SMALL_STATE(1015)] = 46900,
  [SMALL_STATE(1016)] = 46918,
  [SMALL_STATE(1017)] = 46936,
  [SMALL_STATE(1018)] = 46954,
  [SMALL_STATE(1019)] = 46972,
  [SMALL_STATE(1020)] = 46990,
  [SMALL_STATE(1021)] = 47008,
  [SMALL_STATE(1022)] = 47026,
  [SMALL_STATE(1023)] = 47044,
  [SMALL_STATE(1024)] = 47062,
  [SMALL_STATE(1025)] = 47080,
  [SMALL_STATE(1026)] = 47098,
  [SMALL_STATE(1027)] = 47116,
  [SMALL_STATE(1028)] = 47134,
  [SMALL_STATE(1029)] = 47152,
  [SMALL_STATE(1030)] = 47170,
  [SMALL_STATE(1031)] = 47188,
  [SMALL_STATE(1032)] = 47206,
  [SMALL_STATE(1033)] = 47224,
  [SMALL_STATE(1034)] = 47242,
  [SMALL_STATE(1035)] = 47260,
  [SMALL_STATE(1036)] = 47278,
  [SMALL_STATE(1037)] = 47296,
  [SMALL_STATE(1038)] = 47314,
  [SMALL_STATE(1039)] = 47332,
  [SMALL_STATE(1040)] = 47350,
  [SMALL_STATE(1041)] = 47368,
  [SMALL_STATE(1042)] = 47386,
  [SMALL_STATE(1043)] = 47404,
  [SMALL_STATE(1044)] = 47422,
  [SMALL_STATE(1045)] = 47440,
  [SMALL_STATE(1046)] = 47458,
  [SMALL_STATE(1047)] = 47476,
  [SMALL_STATE(1048)] = 47494,
  [SMALL_STATE(1049)] = 47512,
  [SMALL_STATE(1050)] = 47530,
  [SMALL_STATE(1051)] = 47548,
  [SMALL_STATE(1052)] = 47566,
  [SMALL_STATE(1053)] = 47584,
  [SMALL_STATE(1054)] = 47602,
  [SMALL_STATE(1055)] = 47620,
  [SMALL_STATE(1056)] = 47638,
  [SMALL_STATE(1057)] = 47656,
  [SMALL_STATE(1058)] = 47674,
  [SMALL_STATE(1059)] = 47692,
  [SMALL_STATE(1060)] = 47706,
  [SMALL_STATE(1061)] = 47720,
  [SMALL_STATE(1062)] = 47738,
  [SMALL_STATE(1063)] = 47756,
  [SMALL_STATE(1064)] = 47774,
  [SMALL_STATE(1065)] = 47792,
  [SMALL_STATE(1066)] = 47810,
  [SMALL_STATE(1067)] = 47828,
  [SMALL_STATE(1068)] = 47846,
  [SMALL_STATE(1069)] = 47864,
  [SMALL_STATE(1070)] = 47882,
  [SMALL_STATE(1071)] = 47900,
  [SMALL_STATE(1072)] = 47918,
  [SMALL_STATE(1073)] = 47936,
  [SMALL_STATE(1074)] = 47954,
  [SMALL_STATE(1075)] = 47972,
  [SMALL_STATE(1076)] = 47990,
  [SMALL_STATE(1077)] = 48008,
  [SMALL_STATE(1078)] = 48026,
  [SMALL_STATE(1079)] = 48044,
  [SMALL_STATE(1080)] = 48062,
  [SMALL_STATE(1081)] = 48080,
  [SMALL_STATE(1082)] = 48098,
  [SMALL_STATE(1083)] = 48112,
  [SMALL_STATE(1084)] = 48130,
  [SMALL_STATE(1085)] = 48148,
  [SMALL_STATE(1086)] = 48166,
  [SMALL_STATE(1087)] = 48184,
  [SMALL_STATE(1088)] = 48202,
  [SMALL_STATE(1089)] = 48216,
  [SMALL_STATE(1090)] = 48234,
  [SMALL_STATE(1091)] = 48252,
  [SMALL_STATE(1092)] = 48270,
  [SMALL_STATE(1093)] = 48288,
  [SMALL_STATE(1094)] = 48306,
  [SMALL_STATE(1095)] = 48324,
  [SMALL_STATE(1096)] = 48342,
  [SMALL_STATE(1097)] = 48360,
  [SMALL_STATE(1098)] = 48378,
  [SMALL_STATE(1099)] = 48396,
  [SMALL_STATE(1100)] = 48414,
  [SMALL_STATE(1101)] = 48432,
  [SMALL_STATE(1102)] = 48450,
  [SMALL_STATE(1103)] = 48468,
  [SMALL_STATE(1104)] = 48479,
  [SMALL_STATE(1105)] = 48490,
  [SMALL_STATE(1106)] = 48501,
  [SMALL_STATE(1107)] = 48512,
  [SMALL_STATE(1108)] = 48522,
  [SMALL_STATE(1109)] = 48532,
  [SMALL_STATE(1110)] = 48542,
  [SMALL_STATE(1111)] = 48552,
  [SMALL_STATE(1112)] = 48562,
  [SMALL_STATE(1113)] = 48571,
  [SMALL_STATE(1114)] = 48580,
  [SMALL_STATE(1115)] = 48589,
  [SMALL_STATE(1116)] = 48598,
  [SMALL_STATE(1117)] = 48608,
  [SMALL_STATE(1118)] = 48618,
  [SMALL_STATE(1119)] = 48628,
  [SMALL_STATE(1120)] = 48638,
  [SMALL_STATE(1121)] = 48648,
  [SMALL_STATE(1122)] = 48658,
  [SMALL_STATE(1123)] = 48668,
  [SMALL_STATE(1124)] = 48678,
  [SMALL_STATE(1125)] = 48688,
  [SMALL_STATE(1126)] = 48698,
  [SMALL_STATE(1127)] = 48705,
  [SMALL_STATE(1128)] = 48712,
  [SMALL_STATE(1129)] = 48719,
  [SMALL_STATE(1130)] = 48726,
  [SMALL_STATE(1131)] = 48731,
  [SMALL_STATE(1132)] = 48738,
  [SMALL_STATE(1133)] = 48745,
  [SMALL_STATE(1134)] = 48752,
  [SMALL_STATE(1135)] = 48759,
  [SMALL_STATE(1136)] = 48763,
  [SMALL_STATE(1137)] = 48767,
  [SMALL_STATE(1138)] = 48771,
  [SMALL_STATE(1139)] = 48775,
  [SMALL_STATE(1140)] = 48779,
  [SMALL_STATE(1141)] = 48783,
  [SMALL_STATE(1142)] = 48787,
  [SMALL_STATE(1143)] = 48791,
  [SMALL_STATE(1144)] = 48795,
  [SMALL_STATE(1145)] = 48799,
  [SMALL_STATE(1146)] = 48803,
  [SMALL_STATE(1147)] = 48807,
  [SMALL_STATE(1148)] = 48811,
  [SMALL_STATE(1149)] = 48815,
  [SMALL_STATE(1150)] = 48819,
  [SMALL_STATE(1151)] = 48823,
  [SMALL_STATE(1152)] = 48827,
  [SMALL_STATE(1153)] = 48831,
  [SMALL_STATE(1154)] = 48835,
  [SMALL_STATE(1155)] = 48839,
  [SMALL_STATE(1156)] = 48843,
  [SMALL_STATE(1157)] = 48847,
  [SMALL_STATE(1158)] = 48851,
  [SMALL_STATE(1159)] = 48855,
  [SMALL_STATE(1160)] = 48859,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(232),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(215),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(185),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(207),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [47] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0),
  [49] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(434),
  [52] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(150),
  [55] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(89),
  [58] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(236),
  [61] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(210),
  [64] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(221),
  [67] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(252),
  [70] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(221),
  [73] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(217),
  [76] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(225),
  [79] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(180),
  [82] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(178),
  [85] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(213),
  [88] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(179),
  [91] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(206),
  [94] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(237),
  [97] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(240),
  [100] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(7),
  [103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(12),
  [106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(8),
  [111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(194),
  [114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(354),
  [117] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(148),
  [120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(86),
  [123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(257),
  [126] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(208),
  [129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(258),
  [132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(239),
  [135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(258),
  [138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(218),
  [141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(241),
  [144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(183),
  [147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(173),
  [150] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(205),
  [153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(175),
  [156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(219),
  [159] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(254),
  [162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(230),
  [165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(15),
  [168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(4),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(5),
  [176] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(191),
  [179] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(434),
  [182] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(150),
  [185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(89),
  [188] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(236),
  [191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(210),
  [194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(221),
  [197] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(252),
  [200] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(221),
  [203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(217),
  [206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(225),
  [209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(180),
  [212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(178),
  [215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(213),
  [218] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(179),
  [221] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(206),
  [224] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(237),
  [227] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(240),
  [230] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(7),
  [233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(12),
  [236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0),
  [238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [240] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(8),
  [243] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(194),
  [246] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(276),
  [249] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(154),
  [252] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(87),
  [255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(233),
  [258] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(204),
  [261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(248),
  [264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(223),
  [267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(248),
  [270] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(212),
  [273] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(250),
  [276] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(189),
  [279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(167),
  [282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(216),
  [285] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(177),
  [288] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(214),
  [291] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(256),
  [294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(226),
  [297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(9),
  [300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(10),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(16),
  [308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(196),
  [311] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(276),
  [314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(154),
  [317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(87),
  [320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(233),
  [323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(204),
  [326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(248),
  [329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(223),
  [332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(248),
  [335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(212),
  [338] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(250),
  [341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(189),
  [344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(167),
  [347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(216),
  [350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(177),
  [353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(214),
  [356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(256),
  [359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(226),
  [362] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(9),
  [365] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(10),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(16),
  [373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(196),
  [376] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(354),
  [379] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(148),
  [382] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(86),
  [385] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(257),
  [388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(208),
  [391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(258),
  [394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(239),
  [397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(258),
  [400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(218),
  [403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(241),
  [406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(183),
  [409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(173),
  [412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(205),
  [415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(175),
  [418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(219),
  [421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(254),
  [424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(230),
  [427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(15),
  [430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(4),
  [433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0),
  [435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [437] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(5),
  [440] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(191),
  [443] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(276),
  [446] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(154),
  [449] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(87),
  [452] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(233),
  [455] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(204),
  [458] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(248),
  [461] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(223),
  [464] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(248),
  [467] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(212),
  [470] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(250),
  [473] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(189),
  [476] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(167),
  [479] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(216),
  [482] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(177),
  [485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(214),
  [488] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(256),
  [491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(226),
  [494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(9),
  [497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(10),
  [500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [502] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(16),
  [505] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(196),
  [508] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(354),
  [511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(148),
  [514] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(86),
  [517] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(257),
  [520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(208),
  [523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(258),
  [526] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(239),
  [529] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(258),
  [532] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(218),
  [535] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(241),
  [538] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(183),
  [541] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(173),
  [544] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(205),
  [547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(175),
  [550] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(219),
  [553] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(254),
  [556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(230),
  [559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(15),
  [562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(4),
  [565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [567] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(5),
  [570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0),
  [572] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(191),
  [575] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(434),
  [578] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(150),
  [581] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(89),
  [584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(236),
  [587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(210),
  [590] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(221),
  [593] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(252),
  [596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(221),
  [599] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(217),
  [602] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(225),
  [605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(180),
  [608] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(178),
  [611] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(213),
  [614] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(179),
  [617] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(206),
  [620] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(237),
  [623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(240),
  [626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(7),
  [629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(12),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(8),
  [637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(194),
  [640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(208),
  [650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(183),
  [662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(254),
  [672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline, 2, 0, 0),
  [730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline, 1, 0, 0),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(212),
  [752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(189),
  [756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(256),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0),
  [782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(353),
  [785] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(155),
  [788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(88),
  [791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(234),
  [794] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(209),
  [797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(232),
  [800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(235),
  [803] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(232),
  [806] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(215),
  [809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(245),
  [812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(185),
  [815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(174),
  [818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(207),
  [821] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(176),
  [824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(211),
  [827] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(229),
  [830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(246),
  [833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(3),
  [836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(2),
  [839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(6),
  [842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(197),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [847] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(354),
  [850] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(148),
  [853] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(86),
  [856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(257),
  [859] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(208),
  [862] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(258),
  [865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(239),
  [868] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(258),
  [871] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(218),
  [874] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(241),
  [877] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(183),
  [880] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(173),
  [883] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(205),
  [886] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(175),
  [889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(219),
  [892] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(254),
  [895] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(230),
  [898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(15),
  [901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(4),
  [904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0),
  [906] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(5),
  [909] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star, 2, 0, 0), SHIFT_REPEAT(191),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(434),
  [917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(150),
  [920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(89),
  [923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(236),
  [926] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(210),
  [929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(221),
  [932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(252),
  [935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(221),
  [938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(217),
  [941] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(225),
  [944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(180),
  [947] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(178),
  [950] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(213),
  [953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(179),
  [956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(206),
  [959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(237),
  [962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(240),
  [965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(7),
  [968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(12),
  [971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0),
  [973] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(8),
  [976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore, 2, 0, 0), SHIFT_REPEAT(194),
  [979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(276),
  [984] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(154),
  [987] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(87),
  [990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(233),
  [993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(204),
  [996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(248),
  [999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(223),
  [1002] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(248),
  [1005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(212),
  [1008] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(250),
  [1011] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(189),
  [1014] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(167),
  [1017] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(216),
  [1020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(177),
  [1023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(214),
  [1026] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(256),
  [1029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(226),
  [1032] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(9),
  [1035] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(10),
  [1038] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(16),
  [1041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0),
  [1043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde, 2, 0, 0), SHIFT_REPEAT(196),
  [1046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [1048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [1050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [1052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [1054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [1056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [1058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [1060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [1064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [1066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [1068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [1070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [1072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [1074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [1076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [1078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [1080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [1082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [1084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [1086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [1088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [1090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [1092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [1094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [1096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [1098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [1100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [1102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [1104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [1106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [1108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [1110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [1112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [1114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [1116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [1118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [1120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [1122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [1124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(190),
  [1126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [1128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [1130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [1134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [1138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [1140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [1144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [1150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(574),
  [1153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(156),
  [1156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(90),
  [1159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(356),
  [1162] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(253),
  [1165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(388),
  [1168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(406),
  [1171] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(388),
  [1174] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(249),
  [1177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(390),
  [1180] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(199),
  [1183] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(190),
  [1186] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(243),
  [1189] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(182),
  [1192] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(255),
  [1195] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(395),
  [1198] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(261),
  [1201] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(25),
  [1204] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(26),
  [1207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(27),
  [1210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline, 2, 0, 0), SHIFT_REPEAT(201),
  [1213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [1215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [1217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [1219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [1223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(574),
  [1226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(165),
  [1229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(1130),
  [1232] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(253),
  [1235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(388),
  [1238] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(349),
  [1241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(388),
  [1244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(249),
  [1247] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(390),
  [1250] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(192),
  [1253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(188),
  [1256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(243),
  [1259] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(181),
  [1262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(255),
  [1265] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(395),
  [1268] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(261),
  [1271] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(98),
  [1274] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(99),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [1279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(91),
  [1282] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 0), SHIFT(201),
  [1285] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(574),
  [1288] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(165),
  [1291] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(1130),
  [1294] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(253),
  [1297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(388),
  [1300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(349),
  [1303] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(388),
  [1306] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(249),
  [1309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(390),
  [1312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(192),
  [1315] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(188),
  [1318] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(243),
  [1321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(181),
  [1324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(255),
  [1327] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(395),
  [1330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(261),
  [1333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(98),
  [1336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(99),
  [1339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [1341] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(91),
  [1344] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0), SHIFT(201),
  [1347] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(574),
  [1350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(165),
  [1353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(1130),
  [1356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(253),
  [1359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(388),
  [1362] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(349),
  [1365] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(388),
  [1368] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(249),
  [1371] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(390),
  [1374] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(192),
  [1377] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(188),
  [1380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(243),
  [1383] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(181),
  [1386] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(255),
  [1389] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(395),
  [1392] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(261),
  [1395] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(98),
  [1398] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(99),
  [1401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [1403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(91),
  [1406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 0), SHIFT(201),
  [1409] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(574),
  [1412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(165),
  [1415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(1130),
  [1418] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(253),
  [1421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(388),
  [1424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(349),
  [1427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(388),
  [1430] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(249),
  [1433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(390),
  [1436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(192),
  [1439] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(188),
  [1442] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(243),
  [1445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(181),
  [1448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(255),
  [1451] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(395),
  [1454] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(261),
  [1457] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(98),
  [1460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(99),
  [1463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [1465] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(91),
  [1468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0), SHIFT(201),
  [1471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [1473] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(276),
  [1476] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(163),
  [1479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0),
  [1481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(204),
  [1484] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(248),
  [1487] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(399),
  [1490] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(248),
  [1493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(212),
  [1496] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(250),
  [1499] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(187),
  [1502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(171),
  [1505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(216),
  [1508] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(172),
  [1511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(214),
  [1514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(256),
  [1517] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(226),
  [1520] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(92),
  [1523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(95),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(133),
  [1531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(196),
  [1534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(354),
  [1537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(162),
  [1540] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(208),
  [1543] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(258),
  [1546] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(383),
  [1549] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(258),
  [1552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(218),
  [1555] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(241),
  [1558] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(184),
  [1561] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(166),
  [1564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(205),
  [1567] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(168),
  [1570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(219),
  [1573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(254),
  [1576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(230),
  [1579] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(104),
  [1582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(97),
  [1585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [1587] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(96),
  [1590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0),
  [1592] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(191),
  [1595] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(276),
  [1598] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(163),
  [1601] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(204),
  [1604] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(248),
  [1607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(399),
  [1610] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(248),
  [1613] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(212),
  [1616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(250),
  [1619] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(187),
  [1622] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(171),
  [1625] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(216),
  [1628] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(172),
  [1631] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(214),
  [1634] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(256),
  [1637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(226),
  [1640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(92),
  [1643] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(95),
  [1646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0),
  [1648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [1650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(133),
  [1653] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(196),
  [1656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(354),
  [1659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(162),
  [1662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(208),
  [1665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(258),
  [1668] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(383),
  [1671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(258),
  [1674] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(218),
  [1677] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(241),
  [1680] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(184),
  [1683] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(166),
  [1686] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(205),
  [1689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(168),
  [1692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(219),
  [1695] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(254),
  [1698] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(230),
  [1701] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(104),
  [1704] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(97),
  [1707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [1709] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(96),
  [1712] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 0), SHIFT(191),
  [1715] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(434),
  [1718] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(159),
  [1721] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(210),
  [1724] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(221),
  [1727] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(389),
  [1730] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(221),
  [1733] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(217),
  [1736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(225),
  [1739] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(186),
  [1742] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(169),
  [1745] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(213),
  [1748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(170),
  [1751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(206),
  [1754] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(237),
  [1757] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(240),
  [1760] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(94),
  [1763] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(100),
  [1766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [1768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(93),
  [1771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 0), SHIFT(194),
  [1774] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(276),
  [1777] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(163),
  [1780] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(204),
  [1783] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(248),
  [1786] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(399),
  [1789] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(248),
  [1792] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(212),
  [1795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(250),
  [1798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(187),
  [1801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(171),
  [1804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(216),
  [1807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(172),
  [1810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(214),
  [1813] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(256),
  [1816] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(226),
  [1819] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(92),
  [1822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(95),
  [1825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [1829] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(133),
  [1832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(196),
  [1835] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(434),
  [1838] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(159),
  [1841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(210),
  [1844] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(221),
  [1847] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(389),
  [1850] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(221),
  [1853] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(217),
  [1856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(225),
  [1859] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(186),
  [1862] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(169),
  [1865] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(213),
  [1868] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(170),
  [1871] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(206),
  [1874] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(237),
  [1877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(240),
  [1880] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(94),
  [1883] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(100),
  [1886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [1888] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(93),
  [1891] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 0), SHIFT(194),
  [1894] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(354),
  [1897] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(162),
  [1900] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(208),
  [1903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(258),
  [1906] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(383),
  [1909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(258),
  [1912] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(218),
  [1915] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(241),
  [1918] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(184),
  [1921] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(166),
  [1924] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(205),
  [1927] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(168),
  [1930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(219),
  [1933] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(254),
  [1936] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(230),
  [1939] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(104),
  [1942] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(97),
  [1945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [1947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(96),
  [1950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(191),
  [1953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(434),
  [1956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(159),
  [1959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(210),
  [1962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(221),
  [1965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(389),
  [1968] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(221),
  [1971] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(217),
  [1974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(225),
  [1977] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(186),
  [1980] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(169),
  [1983] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(213),
  [1986] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(170),
  [1989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(206),
  [1992] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(237),
  [1995] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(240),
  [1998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(94),
  [2001] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(100),
  [2004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [2006] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(93),
  [2009] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 0), SHIFT(194),
  [2012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [2014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [2016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [2018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [2020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [2022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [2024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [2026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [2028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [2030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [2032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(187),
  [2036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(171),
  [2038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [2040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [2042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [2044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [2046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [2048] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(574),
  [2051] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(165),
  [2054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0),
  [2056] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(253),
  [2059] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(388),
  [2062] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(349),
  [2065] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(388),
  [2068] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(249),
  [2071] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(390),
  [2074] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(192),
  [2077] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(188),
  [2080] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(243),
  [2083] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(181),
  [2086] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(255),
  [2089] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(395),
  [2092] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(261),
  [2095] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(98),
  [2098] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(99),
  [2101] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(91),
  [2104] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 2, 0, 0), SHIFT_REPEAT(201),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [2109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [2111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(184),
  [2113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [2117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [2121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [2123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [2125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [2127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [2129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(434),
  [2132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(159),
  [2135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(210),
  [2138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(221),
  [2141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(389),
  [2144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(221),
  [2147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(217),
  [2150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(225),
  [2153] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(186),
  [2156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(169),
  [2159] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(213),
  [2162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(170),
  [2165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(206),
  [2168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(237),
  [2171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(240),
  [2174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(94),
  [2177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(100),
  [2180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0),
  [2182] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(93),
  [2185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_underscore_no_link, 2, 0, 0), SHIFT_REPEAT(194),
  [2188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [2190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(276),
  [2193] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(163),
  [2196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(204),
  [2199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(248),
  [2202] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(399),
  [2205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(248),
  [2208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(212),
  [2211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(250),
  [2214] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(187),
  [2217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(171),
  [2220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(216),
  [2223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(172),
  [2226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(214),
  [2229] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(256),
  [2232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(226),
  [2235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(92),
  [2238] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(95),
  [2241] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(133),
  [2244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0),
  [2246] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_tilde_no_link, 2, 0, 0), SHIFT_REPEAT(196),
  [2249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [2251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [2253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [2255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [2259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(188),
  [2261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [2263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [2265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [2269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [2271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [2275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [2279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [2281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [2283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [2285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [2287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [2289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [2293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [2297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [2303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [2305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [2309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [2311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [2313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [2317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(354),
  [2320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(162),
  [2323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(208),
  [2326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(258),
  [2329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(383),
  [2332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(258),
  [2335] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(218),
  [2338] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(241),
  [2341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(184),
  [2344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(166),
  [2347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(205),
  [2350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(168),
  [2353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(219),
  [2356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(254),
  [2359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(230),
  [2362] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(104),
  [2365] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(97),
  [2368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0),
  [2370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(96),
  [2373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_no_star_no_link, 2, 0, 0), SHIFT_REPEAT(191),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [2378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_base, 1, 0, 0),
  [2380] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [2383] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(149),
  [2386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0),
  [2388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(208),
  [2391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(258),
  [2394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [2397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(258),
  [2400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(218),
  [2403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(241),
  [2406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(183),
  [2409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [2412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(205),
  [2415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(175),
  [2418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(219),
  [2421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(254),
  [2424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(230),
  [2427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(191),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [2432] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [2435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(151),
  [2438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(210),
  [2441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(221),
  [2444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(252),
  [2447] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(221),
  [2450] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(217),
  [2453] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(225),
  [2456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(180),
  [2459] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(178),
  [2462] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(213),
  [2465] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(179),
  [2468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(206),
  [2471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(237),
  [2474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(240),
  [2477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(194),
  [2480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [2483] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(152),
  [2486] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(209),
  [2489] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(232),
  [2492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(235),
  [2495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(232),
  [2498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(215),
  [2501] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(245),
  [2504] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(185),
  [2507] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [2510] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(207),
  [2513] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(176),
  [2516] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(211),
  [2519] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [2522] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(246),
  [2525] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [2528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(276),
  [2531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(153),
  [2534] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(204),
  [2537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [2540] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [2543] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [2546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(212),
  [2549] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(250),
  [2552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(189),
  [2555] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(167),
  [2558] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(216),
  [2561] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(177),
  [2564] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(214),
  [2567] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(256),
  [2570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(226),
  [2573] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(196),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [2582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(574),
  [2585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(157),
  [2588] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(253),
  [2591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(388),
  [2594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [2597] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(388),
  [2600] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(249),
  [2603] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(390),
  [2606] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [2609] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(190),
  [2612] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(243),
  [2615] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(182),
  [2618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(255),
  [2621] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(395),
  [2624] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(261),
  [2627] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [2630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(158),
  [2633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(389),
  [2636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(186),
  [2639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(169),
  [2642] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(170),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [2647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(160),
  [2650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(399),
  [2653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(187),
  [2656] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(171),
  [2659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(172),
  [2662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(161),
  [2665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(383),
  [2668] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(184),
  [2671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(166),
  [2674] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(168),
  [2677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [2679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [2681] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(164),
  [2684] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(349),
  [2687] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(192),
  [2690] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(188),
  [2693] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 2, 0, 0), SHIFT_REPEAT(181),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [2698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0),
  [2700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [2702] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(887),
  [2705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(887),
  [2708] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(891),
  [2711] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [2715] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(881),
  [2718] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(881),
  [2721] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(786),
  [2724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [2726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(789),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [2731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(885),
  [2733] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(885),
  [2736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(885),
  [2739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(894),
  [2742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [2744] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(883),
  [2747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(883),
  [2750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(787),
  [2753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(764),
  [2758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [2760] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(765),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [2765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(790),
  [2768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [2770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(769),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [2775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(757),
  [2778] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(725),
  [2781] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(725),
  [2784] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(917),
  [2787] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(882),
  [2790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [2792] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(888),
  [2795] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(888),
  [2798] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(763),
  [2801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [2803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(818),
  [2806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(718),
  [2809] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(718),
  [2812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(918),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [2817] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(785),
  [2820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [2822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(710),
  [2825] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(710),
  [2828] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(916),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [2833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(825),
  [2836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [2838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [2840] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(740),
  [2843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(740),
  [2846] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(920),
  [2849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [2851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(827),
  [2854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [2856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(817),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [2863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [2867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [2869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [2871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [2875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [2877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [2879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [2881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(695),
  [2883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [2887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(746),
  [2890] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(746),
  [2893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(922),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [2898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(816),
  [2901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [2903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [2905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [2907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [2909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [2911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [2917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [2919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [2921] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(705),
  [2924] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(200),
  [2927] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(683),
  [2930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [2933] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [2936] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(692),
  [2939] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(673),
  [2942] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(661),
  [2945] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(682),
  [2948] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(649),
  [2951] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(695),
  [2954] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0), SHIFT_REPEAT(695),
  [2957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_latex_block_repeat1, 2, 0, 0),
  [2959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [2963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [2967] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1131),
  [2970] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(967),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2975] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(909),
  [2978] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(904),
  [2981] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(904),
  [2984] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__whitespace, 1, 0, 0),
  [2986] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__whitespace, 1, 0, 0),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [2990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1129),
  [2993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [2995] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1128),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [3000] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1132),
  [3003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [3005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [3007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [3009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [3011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [3013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [3015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [3017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [3019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [3021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [3023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [3025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [3027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [3029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [3031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [3033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_description_non_empty, 4, 0, 5),
  [3035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_description_non_empty, 4, 0, 5),
  [3037] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(79),
  [3040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__soft_line_break, 1, 0, 0),
  [3044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__soft_line_break, 1, 0, 0),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [3060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 0),
  [3062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 0),
  [3064] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(62),
  [3067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 0),
  [3069] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(71),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [3090] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(75),
  [3093] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1134),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [3098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 0),
  [3100] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(260),
  [3103] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(700),
  [3106] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(723),
  [3109] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(721),
  [3112] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(679),
  [3115] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(674),
  [3118] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(701),
  [3121] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(669),
  [3124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(714),
  [3127] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0), SHIFT_REPEAT(714),
  [3130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_code_span_repeat1, 2, 0, 0),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [3136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 5, 10, 0),
  [3138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 5, 10, 0),
  [3140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__closing_tag, 5, 100, 0),
  [3142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__closing_tag, 5, 100, 0),
  [3144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__open_tag, 5, 100, 0),
  [3146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__open_tag, 5, 100, 0),
  [3148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_comment, 5, 100, 0),
  [3150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_comment, 5, 100, 0),
  [3152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 5, 10, 0),
  [3154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 5, 10, 0),
  [3156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__open_tag, 6, 100, 0),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__open_tag, 6, 100, 0),
  [3160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_comment, 6, 100, 0),
  [3162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_comment, 6, 100, 0),
  [3164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 6, 10, 0),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 6, 10, 0),
  [3168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 6, 10, 0),
  [3170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 6, 10, 0),
  [3172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 7, 10, 0),
  [3174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 7, 10, 0),
  [3176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 7, 10, 0),
  [3178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 7, 10, 0),
  [3180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 8, 10, 0),
  [3182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 8, 10, 0),
  [3184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 8, 10, 0),
  [3186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 8, 10, 0),
  [3188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_backslash_escape, 1, 0, 0),
  [3190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_backslash_escape, 1, 0, 0),
  [3192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration, 4, 100, 0),
  [3194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration, 4, 100, 0),
  [3196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_label, 3, 0, 0),
  [3198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_label, 3, 0, 0),
  [3200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(677),
  [3203] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(279),
  [3206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0),
  [3208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(675),
  [3211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(675),
  [3214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(670),
  [3217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(671),
  [3220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(671),
  [3223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 4, 10, 0),
  [3225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 4, 10, 0),
  [3227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 4, 10, 0),
  [3229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 4, 10, 0),
  [3231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shortcut_link, 1, 10, 0),
  [3233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__link_text, 1, 10, 0), REDUCE(sym_shortcut_link, 1, 10, 0),
  [3236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shortcut_link, 1, 10, 0),
  [3238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_image, 1, 0, 0),
  [3240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_image, 1, 0, 0),
  [3242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_shortcut_link, 1, 30, 0),
  [3244] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__image_shortcut_link, 1, 30, 0), REDUCE(sym__image_description, 1, 30, 0),
  [3247] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_shortcut_link, 1, 30, 0),
  [3249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_base_repeat1, 1, 0, 1),
  [3251] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__inline_base_repeat1, 1, 0, 1),
  [3253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_tag, 1, 0, 0),
  [3255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_tag, 1, 0, 0),
  [3257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_code_span, 2, 0, 0),
  [3259] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_code_span, 2, 0, 0),
  [3261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_latex_block, 2, 0, 0),
  [3263] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_latex_block, 2, 0, 0),
  [3265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_base, 2, 0, 0),
  [3267] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_base, 2, 0, 0),
  [3269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hard_line_break, 2, 0, 0),
  [3271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hard_line_break, 2, 0, 0),
  [3273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__soft_line_break, 2, 0, 0),
  [3275] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__soft_line_break, 2, 0, 0),
  [3277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_comment, 2, 100, 0),
  [3279] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_comment, 2, 100, 0),
  [3281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__processing_instruction, 2, 100, 0),
  [3283] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__processing_instruction, 2, 100, 0),
  [3285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__cdata_section, 2, 100, 0),
  [3287] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__cdata_section, 2, 100, 0),
  [3289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__whitespace, 2, 0, 0),
  [3291] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__whitespace, 2, 0, 0),
  [3293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_full_reference_link, 2, 20, 0),
  [3295] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_full_reference_link, 2, 20, 0),
  [3297] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_full_reference_link, 2, 10, 0),
  [3299] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_full_reference_link, 2, 10, 0),
  [3301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_code_span, 3, 0, 0),
  [3303] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_code_span, 3, 0, 0),
  [3305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_star, 3, 1, 3),
  [3307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_star, 3, 1, 3),
  [3309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strong_emphasis_star, 3, 2, 3),
  [3311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strong_emphasis_star, 3, 2, 3),
  [3313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_underscore, 3, 1, 3),
  [3315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_underscore, 3, 1, 3),
  [3317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strong_emphasis_underscore, 3, 2, 3),
  [3319] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strong_emphasis_underscore, 3, 2, 3),
  [3321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strikethrough, 3, 1, 3),
  [3323] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strikethrough, 3, 1, 3),
  [3325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_latex_block, 3, 0, 0),
  [3327] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_latex_block, 3, 0, 0),
  [3329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_text_non_empty, 3, 0, 4),
  [3331] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__link_text_non_empty, 3, 0, 4),
  [3333] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__open_tag, 3, 100, 0),
  [3335] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__open_tag, 3, 100, 0),
  [3337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_comment, 3, 100, 0),
  [3339] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_comment, 3, 100, 0),
  [3341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__processing_instruction, 3, 100, 0),
  [3343] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__processing_instruction, 3, 100, 0),
  [3345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration, 3, 100, 0),
  [3347] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration, 3, 100, 0),
  [3349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__cdata_section, 3, 100, 0),
  [3351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__cdata_section, 3, 100, 0),
  [3353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_collapsed_reference_link, 3, 10, 0),
  [3355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_collapsed_reference_link, 3, 10, 0),
  [3357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inline_link, 3, 10, 0),
  [3359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inline_link, 3, 10, 0),
  [3361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_collapsed_reference_link, 3, 10, 0),
  [3363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_collapsed_reference_link, 3, 10, 0),
  [3365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_inline_link, 3, 10, 0),
  [3367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__image_inline_link, 3, 10, 0),
  [3369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_star, 4, 1, 3),
  [3371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_star, 4, 1, 3),
  [3373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_underscore, 4, 1, 3),
  [3375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_underscore, 4, 1, 3),
  [3377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strikethrough, 4, 1, 3),
  [3379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strikethrough, 4, 1, 3),
  [3381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__closing_tag, 4, 100, 0),
  [3383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__closing_tag, 4, 100, 0),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__open_tag, 4, 100, 0),
  [3387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__open_tag, 4, 100, 0),
  [3389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_comment, 4, 100, 0),
  [3391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_comment, 4, 100, 0),
  [3393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [3395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_description, 3, 30, 0),
  [3397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [3399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [3401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [3403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [3405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [3407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [3409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(675),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [3417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(671),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [3421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(83),
  [3424] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 2, 0, 0),
  [3426] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 2, 0, 0),
  [3428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2),
  [3430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_star, 1, 0, 2),
  [3432] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2), SHIFT(402),
  [3435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2), SHIFT(533),
  [3438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2),
  [3440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2),
  [3442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2), SHIFT(404),
  [3445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [3447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [3449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [3451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [3453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 2, 0, 0),
  [3455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 2, 0, 0),
  [3457] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2), SHIFT(481),
  [3460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2), SHIFT(485),
  [3463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [3465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [3467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [3469] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2), SHIFT(441),
  [3472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [3474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2), SHIFT(308),
  [3477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2), SHIFT(310),
  [3480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [3482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 2),
  [3488] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_tilde, 1, 0, 2),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [3492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_star, 1, 0, 2), SHIFT(550),
  [3495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__inline_element_no_underscore, 1, 0, 2), SHIFT(552),
  [3498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [3502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element, 2, 0, 0),
  [3506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element, 2, 0, 0),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [3514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [3516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline, 1, 0, 0),
  [3518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__inline, 1, 0, 0),
  [3520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_star, 2, 0, 0),
  [3522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_star, 2, 0, 0),
  [3524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element, 1, 0, 2),
  [3526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element, 1, 0, 2),
  [3528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__image_description, 1, 30, 0),
  [3530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 2),
  [3532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_star_no_link, 1, 0, 2),
  [3534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_star_no_link, 4, 1, 3),
  [3536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_star_no_link, 4, 1, 3),
  [3538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_underscore_no_link, 4, 1, 3),
  [3540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_underscore_no_link, 4, 1, 3),
  [3542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strikethrough_no_link, 4, 1, 3),
  [3544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strikethrough_no_link, 4, 1, 3),
  [3546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strikethrough_no_link, 3, 1, 3),
  [3548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strikethrough_no_link, 3, 1, 3),
  [3550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 2),
  [3552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_tilde_no_link, 1, 0, 2),
  [3554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_no_link, 1, 0, 0),
  [3556] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__inline_no_link, 1, 0, 0),
  [3558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_star_no_link, 3, 1, 3),
  [3560] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_star_no_link, 3, 1, 3),
  [3562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strong_emphasis_star_no_link, 3, 2, 3),
  [3564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strong_emphasis_star_no_link, 3, 2, 3),
  [3566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__emphasis_underscore_no_link, 3, 1, 3),
  [3568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__emphasis_underscore_no_link, 3, 1, 3),
  [3570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__strong_emphasis_underscore_no_link, 3, 2, 3),
  [3572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__strong_emphasis_underscore_no_link, 3, 2, 3),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [3576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 2),
  [3578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_underscore_no_link, 1, 0, 2),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [3588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [3590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [3594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_element_no_link, 1, 0, 2),
  [3596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_element_no_link, 1, 0, 2),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [3600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [3602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [3604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [3606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [3608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(684),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [3620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [3622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [3624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [3626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [3630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(807),
  [3633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [3635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [3637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [3639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [3641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [3643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [3645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [3647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [3649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [3651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [3653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [3655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(849),
  [3657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [3659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0), SHIFT(851),
  [3662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0), SHIFT(837),
  [3665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0), SHIFT(838),
  [3668] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_link_destination, 1, 10, 0), SHIFT(838),
  [3671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [3673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [3675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [3677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [3679] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(806),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [3688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [3692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [3698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(858),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [3702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0), SHIFT(829),
  [3705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [3707] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(784),
  [3710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [3716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [3718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(831),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [3724] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(744),
  [3727] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(744),
  [3730] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(921),
  [3733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [3735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(805),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [3740] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(780),
  [3743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [3745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(712),
  [3748] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(712),
  [3751] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(919),
  [3754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [3756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(767),
  [3759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(877),
  [3762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(681),
  [3765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(829),
  [3768] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(829),
  [3771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [3773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(837),
  [3776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(838),
  [3779] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(838),
  [3782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1133),
  [3785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [3787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [3789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [3791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [3793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0),
  [3795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_destination, 1, 10, 0),
  [3797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(877),
  [3800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(832),
  [3803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(841),
  [3806] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(841),
  [3809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(842),
  [3812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT(866),
  [3815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [3817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [3819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [3821] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(877),
  [3824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(686),
  [3827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(851),
  [3830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [3832] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(851),
  [3835] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(837),
  [3838] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [3841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [3844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [3846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [3848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [3850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [3858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [3860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [3862] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(877),
  [3865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(832),
  [3868] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(841),
  [3871] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(841),
  [3874] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(842),
  [3877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT(866),
  [3880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [3882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [3888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [3890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [3892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [3894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [3896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [3898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 2, 10, 0),
  [3900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_destination, 2, 10, 0),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [3904] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(877),
  [3907] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(696),
  [3910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(841),
  [3913] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(841),
  [3916] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(842),
  [3919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0),
  [3921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0),
  [3923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [3925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 3, 10, 0),
  [3927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_destination, 3, 10, 0),
  [3929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__text_base, 1, 0, 0), SHIFT(1127),
  [3932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [3934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [3936] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(877),
  [3939] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(703),
  [3942] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(876),
  [3945] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(876),
  [3948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [3950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(837),
  [3953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(838),
  [3956] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(838),
  [3959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [3961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [3963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [3967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [3969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [3971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [3973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [3975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [3977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [3979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [3981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [3983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [3985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [3987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [3989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [3993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(886),
  [3995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [3997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [3999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [4001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [4003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [4005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [4009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0),
  [4011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0),
  [4013] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(837),
  [4016] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [4019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [4022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [4026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [4032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [4034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [4036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [4038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [4044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [4052] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(877),
  [4055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(737),
  [4058] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [4061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0),
  [4063] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [4066] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [4069] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(838),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [4074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [4076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [4078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [4094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [4096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [4098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [4102] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(885),
  [4105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(885),
  [4108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(894),
  [4111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0),
  [4113] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(883),
  [4116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(883),
  [4119] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(751),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [4124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(894),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [4136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(883),
  [4138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [4140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(887),
  [4143] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(887),
  [4146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(891),
  [4149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(881),
  [4152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(881),
  [4155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0), SHIFT_REPEAT(759),
  [4158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [4166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [4194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [4196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [4198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [4200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [4202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [4204] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(890),
  [4207] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(889),
  [4210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(882),
  [4213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0),
  [4215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [4218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [4221] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(791),
  [4224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [4226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [4228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [4230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [4232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [4234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [4238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [4244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [4246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [4248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [4250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [4252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [4254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [4256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [4258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [4260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [4262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT_REPEAT(1126),
  [4265] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [4268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [4271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0),
  [4273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [4275] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [4279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [4283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [4285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [4287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [4293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [4295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [4297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [4299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [4301] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(895),
  [4304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0),
  [4306] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(909),
  [4309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [4312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [4315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(840),
  [4318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0),
  [4320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [4322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [4324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [4326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [4328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [4332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [4336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0), SHIFT_REPEAT(910),
  [4339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0),
  [4341] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0), SHIFT_REPEAT(909),
  [4344] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0), SHIFT_REPEAT(904),
  [4347] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0), SHIFT_REPEAT(904),
  [4350] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0), SHIFT_REPEAT(845),
  [4353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [4355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [4357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [4359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0), SHIFT_REPEAT(897),
  [4362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0),
  [4364] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0), SHIFT_REPEAT(909),
  [4367] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [4370] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0), SHIFT_REPEAT(904),
  [4373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0), SHIFT_REPEAT(848),
  [4376] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [4379] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [4382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [4384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(sym_link_title, 2, 0, 0),
  [4387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(sym_link_title, 2, 0, 0),
  [4390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [4392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [4394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT_REPEAT(1126),
  [4397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [4403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [4405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [4407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [4409] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [4412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 1, 0, 0), REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [4415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [4417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [4419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [4421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [4423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [4425] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [4428] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [4431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [4433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 3, 0, 0),
  [4435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 3, 0, 0),
  [4437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0),
  [4439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0),
  [4441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0), REDUCE(sym_link_title, 2, 0, 0),
  [4444] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0), REDUCE(sym_link_title, 2, 0, 0),
  [4447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0), SHIFT_REPEAT(1126),
  [4450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [4453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [4456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [4458] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [4461] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [4464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [4466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_destination_parenthesis, 3, 0, 0),
  [4468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__link_destination_parenthesis, 3, 0, 0),
  [4470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [4472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 3, 0, 0),
  [4474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 3, 0, 0),
  [4476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [4478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [4480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [4482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_comment_repeat1, 2, 0, 0),
  [4484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [4486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__processing_instruction_repeat1, 1, 0, 0),
  [4488] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 1, 0, 0),
  [4490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [4492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_no_angle, 1, 0, 0),
  [4494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_no_angle, 1, 0, 0),
  [4496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [4498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [4504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [4506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 1, 0, 0),
  [4508] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_comment_repeat1, 1, 0, 0),
  [4510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [4512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 3, 0, 0),
  [4514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 3, 0, 0),
  [4516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [4518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__declaration_repeat1, 1, 0, 0),
  [4520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__declaration_repeat1, 1, 0, 0),
  [4522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [4524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_comment_repeat1, 3, 0, 0),
  [4526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_comment_repeat1, 3, 0, 0),
  [4528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat1, 1, 0, 0),
  [4530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat1, 1, 0, 0),
  [4532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [4534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_no_angle, 2, 0, 0),
  [4536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_no_angle, 2, 0, 0),
  [4538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [4540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__processing_instruction_repeat1, 2, 0, 0),
  [4542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [4544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__attribute_value_repeat2, 1, 0, 0),
  [4546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat2, 1, 0, 0),
  [4548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [4550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat1, 2, 0, 0),
  [4552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__declaration_repeat1, 2, 0, 0),
  [4554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__attribute_value_repeat2, 2, 0, 0),
  [4556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [4558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [4560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [4562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [4564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [4566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [4568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [4570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(882),
  [4573] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [4576] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [4579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [4581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [4585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [4587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [4589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [4591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [4593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [4595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [4597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [4599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [4601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [4603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [4605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [4607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [4609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [4615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [4617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [4619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [4621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [4623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [4625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [4627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [4629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [4633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1060),
  [4635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [4641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [4643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [4645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [4655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__open_tag_repeat1, 2, 0, 0),
  [4657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__open_tag_repeat1, 2, 0, 0), SHIFT_REPEAT(1059),
  [4660] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__open_tag_repeat1, 2, 0, 0), SHIFT_REPEAT(1060),
  [4663] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__open_tag_repeat1, 2, 0, 0), SHIFT_REPEAT(1060),
  [4666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [4668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [4670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1088),
  [4672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [4674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1082),
  [4676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [4678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [4680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1138),
  [4682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [4684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [4686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [4688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [4690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [4692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [4694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [4696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [4698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [4700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [4702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [4704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1140),
  [4706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [4708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1137),
  [4710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [4712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [4714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [4716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [4718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [4720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [4722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [4724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [4726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1157),
  [4728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tag_name, 1, 0, 0),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [4732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tag_name, 1, 0, 0),
  [4734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__tag_name_repeat1, 2, 0, 0),
  [4736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__tag_name_repeat1, 2, 0, 0), SHIFT_REPEAT(968),
  [4739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__tag_name_repeat1, 2, 0, 0),
  [4741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1088),
  [4744] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1082),
  [4747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1082),
  [4750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [4752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [4754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tag_name, 2, 0, 0),
  [4756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [4758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tag_name, 2, 0, 0),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [4762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [4766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [4768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [4770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [4772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1059),
  [4775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1060),
  [4778] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_inline_link_repeat1, 2, 0, 0), SHIFT_REPEAT(1060),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [4783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [4785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [4789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [4793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [4797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [4799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [4803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [4809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [4811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [4813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [4815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [4817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [4819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [4821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [4823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [4825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [4827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [4831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [4833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [4835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [4837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [4839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [4843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [4845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [4847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [4849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [4853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [4855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [4859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [4861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 4, 0, 0),
  [4863] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 4, 0, 0),
  [4865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 5, 0, 0),
  [4867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 5, 0, 0),
  [4869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 6, 0, 0),
  [4871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 6, 0, 0),
  [4873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_value, 3, 0, 0),
  [4875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_value, 3, 0, 0),
  [4877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_value, 2, 0, 0),
  [4879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_value, 2, 0, 0),
  [4881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_title, 2, 0, 0),
  [4883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_title, 2, 0, 0),
  [4885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_title, 3, 0, 0),
  [4887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_title, 3, 0, 0),
  [4889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [4891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [4893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [4895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [4897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [4899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [4901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [4903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [4905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [4907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [4909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [4911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [4913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [4915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [4919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [4921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [4923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [4925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [4927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [4929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [4931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [4933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_text, 2, 10, 0),
  [4935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [4937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [4941] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [4947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [4949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [4951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [4955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__error = 0,
  ts_external_token__trigger_error = 1,
  ts_external_token__code_span_start = 2,
  ts_external_token__code_span_close = 3,
  ts_external_token__emphasis_open_star = 4,
  ts_external_token__emphasis_open_underscore = 5,
  ts_external_token__emphasis_close_star = 6,
  ts_external_token__emphasis_close_underscore = 7,
  ts_external_token__last_token_whitespace = 8,
  ts_external_token__last_token_punctuation = 9,
  ts_external_token__strikethrough_open = 10,
  ts_external_token__strikethrough_close = 11,
  ts_external_token__latex_span_start = 12,
  ts_external_token__latex_span_close = 13,
  ts_external_token__unclosed_span = 14,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__error] = sym__error,
  [ts_external_token__trigger_error] = sym__trigger_error,
  [ts_external_token__code_span_start] = sym__code_span_start,
  [ts_external_token__code_span_close] = sym__code_span_close,
  [ts_external_token__emphasis_open_star] = sym__emphasis_open_star,
  [ts_external_token__emphasis_open_underscore] = sym__emphasis_open_underscore,
  [ts_external_token__emphasis_close_star] = sym__emphasis_close_star,
  [ts_external_token__emphasis_close_underscore] = sym__emphasis_close_underscore,
  [ts_external_token__last_token_whitespace] = sym__last_token_whitespace,
  [ts_external_token__last_token_punctuation] = sym__last_token_punctuation,
  [ts_external_token__strikethrough_open] = sym__strikethrough_open,
  [ts_external_token__strikethrough_close] = sym__strikethrough_close,
  [ts_external_token__latex_span_start] = sym__latex_span_start,
  [ts_external_token__latex_span_close] = sym__latex_span_close,
  [ts_external_token__unclosed_span] = sym__unclosed_span,
};

static const bool ts_external_scanner_states[27][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__error] = true,
    [ts_external_token__trigger_error] = true,
    [ts_external_token__code_span_start] = true,
    [ts_external_token__code_span_close] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_star] = true,
    [ts_external_token__emphasis_close_underscore] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__strikethrough_close] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__latex_span_close] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [2] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [3] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [4] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_star] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [5] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_underscore] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [6] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__strikethrough_close] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [7] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_star] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [8] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [9] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__strikethrough_close] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [10] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_underscore] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [11] = {
    [ts_external_token__latex_span_close] = true,
  },
  [12] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_underscore] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [13] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__strikethrough_close] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [14] = {
    [ts_external_token__code_span_start] = true,
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__emphasis_close_star] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__strikethrough_open] = true,
    [ts_external_token__latex_span_start] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [15] = {
    [ts_external_token__code_span_close] = true,
  },
  [16] = {
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [17] = {
    [ts_external_token__last_token_punctuation] = true,
  },
  [18] = {
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [19] = {
    [ts_external_token__emphasis_open_star] = true,
    [ts_external_token__emphasis_open_underscore] = true,
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__unclosed_span] = true,
  },
  [20] = {
    [ts_external_token__last_token_punctuation] = true,
    [ts_external_token__latex_span_close] = true,
  },
  [21] = {
    [ts_external_token__last_token_whitespace] = true,
    [ts_external_token__latex_span_close] = true,
  },
  [22] = {
    [ts_external_token__code_span_close] = true,
    [ts_external_token__last_token_punctuation] = true,
  },
  [23] = {
    [ts_external_token__code_span_close] = true,
    [ts_external_token__last_token_whitespace] = true,
  },
  [24] = {
    [ts_external_token__last_token_whitespace] = true,
  },
  [25] = {
    [ts_external_token__trigger_error] = true,
    [ts_external_token__last_token_whitespace] = true,
  },
  [26] = {
    [ts_external_token__trigger_error] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_markdown_inline_external_scanner_create(void);
void tree_sitter_markdown_inline_external_scanner_destroy(void *);
bool tree_sitter_markdown_inline_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_markdown_inline_external_scanner_serialize(void *, char *);
void tree_sitter_markdown_inline_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_markdown_inline(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_markdown_inline_external_scanner_create,
      tree_sitter_markdown_inline_external_scanner_destroy,
      tree_sitter_markdown_inline_external_scanner_scan,
      tree_sitter_markdown_inline_external_scanner_serialize,
      tree_sitter_markdown_inline_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "markdown_inline",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 5,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
