/* Automatically @generated by tree-sitter v0.25.4 (726dcd1e872149d95de581589fc408fb8ea9cb0b) */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 925
#define LARGE_STATE_COUNT 351
#define SYMBOL_COUNT 203
#define ALIAS_COUNT 4
#define TOKEN_COUNT 90
#define EXTERNAL_TOKEN_COUNT 47
#define FIELD_COUNT 1
#define MAX_ALIAS_SEQUENCE_LENGTH 13
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 14
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym__backslash_escape = 1,
  sym_entity_reference = 2,
  sym_numeric_character_reference = 3,
  anon_sym_LBRACK = 4,
  anon_sym_RBRACK = 5,
  anon_sym_LT = 6,
  anon_sym_GT = 7,
  anon_sym_BANG = 8,
  anon_sym_DQUOTE = 9,
  anon_sym_POUND = 10,
  anon_sym_DOLLAR = 11,
  anon_sym_PERCENT = 12,
  anon_sym_AMP = 13,
  anon_sym_SQUOTE = 14,
  anon_sym_STAR = 15,
  anon_sym_PLUS = 16,
  anon_sym_COMMA = 17,
  anon_sym_DASH = 18,
  anon_sym_DOT = 19,
  anon_sym_SLASH = 20,
  anon_sym_COLON = 21,
  anon_sym_SEMI = 22,
  anon_sym_EQ = 23,
  anon_sym_QMARK = 24,
  anon_sym_AT = 25,
  anon_sym_BSLASH = 26,
  anon_sym_CARET = 27,
  anon_sym__ = 28,
  anon_sym_BQUOTE = 29,
  anon_sym_LBRACE = 30,
  anon_sym_PIPE = 31,
  anon_sym_RBRACE = 32,
  anon_sym_TILDE = 33,
  anon_sym_LPAREN = 34,
  anon_sym_RPAREN = 35,
  anon_sym_DASH_DASH_GT = 36,
  anon_sym_QMARK_GT = 37,
  anon_sym_RBRACK_RBRACK_GT = 38,
  aux_sym__word_token1 = 39,
  aux_sym__word_token2 = 40,
  aux_sym__word_token3 = 41,
  sym__whitespace = 42,
  sym__line_ending = 43,
  sym__soft_line_ending = 44,
  sym__block_close = 45,
  sym_block_continuation = 46,
  sym__block_quote_start = 47,
  sym__indented_chunk_start = 48,
  sym_atx_h1_marker = 49,
  sym_atx_h2_marker = 50,
  sym_atx_h3_marker = 51,
  sym_atx_h4_marker = 52,
  sym_atx_h5_marker = 53,
  sym_atx_h6_marker = 54,
  sym_setext_h1_underline = 55,
  sym_setext_h2_underline = 56,
  sym__thematic_break = 57,
  sym__list_marker_minus = 58,
  sym__list_marker_plus = 59,
  sym__list_marker_star = 60,
  sym__list_marker_parenthesis = 61,
  sym__list_marker_dot = 62,
  sym__list_marker_minus_dont_interrupt = 63,
  sym__list_marker_plus_dont_interrupt = 64,
  sym__list_marker_star_dont_interrupt = 65,
  sym__list_marker_parenthesis_dont_interrupt = 66,
  sym__list_marker_dot_dont_interrupt = 67,
  sym__fenced_code_block_start_backtick = 68,
  sym__fenced_code_block_start_tilde = 69,
  sym__blank_line_start = 70,
  sym__fenced_code_block_end_backtick = 71,
  sym__fenced_code_block_end_tilde = 72,
  sym__html_block_1_start = 73,
  sym__html_block_1_end = 74,
  sym__html_block_2_start = 75,
  sym__html_block_3_start = 76,
  sym__html_block_4_start = 77,
  sym__html_block_5_start = 78,
  sym__html_block_6_start = 79,
  sym__html_block_7_start = 80,
  sym__close_block = 81,
  sym__no_indented_chunk = 82,
  sym__error = 83,
  sym__trigger_error = 84,
  sym__eof = 85,
  sym_minus_metadata = 86,
  sym_plus_metadata = 87,
  sym__pipe_table_start = 88,
  sym__pipe_table_line_ending = 89,
  sym_document = 90,
  sym_backslash_escape = 91,
  sym_link_label = 92,
  sym_link_destination = 93,
  sym__link_destination_parenthesis = 94,
  sym__text_no_angle = 95,
  sym_link_title = 96,
  sym__last_token_punctuation = 97,
  sym__block = 98,
  sym__block_not_section = 99,
  sym_section = 100,
  sym__section1 = 101,
  sym__section2 = 102,
  sym__section3 = 103,
  sym__section4 = 104,
  sym__section5 = 105,
  sym__section6 = 106,
  sym_thematic_break = 107,
  sym__atx_heading1 = 108,
  sym__atx_heading2 = 109,
  sym__atx_heading3 = 110,
  sym__atx_heading4 = 111,
  sym__atx_heading5 = 112,
  sym__atx_heading6 = 113,
  sym__atx_heading_content = 114,
  sym__setext_heading1 = 115,
  sym__setext_heading2 = 116,
  sym_indented_code_block = 117,
  sym__indented_chunk = 118,
  sym_fenced_code_block = 119,
  sym_code_fence_content = 120,
  sym_info_string = 121,
  sym_language = 122,
  sym_html_block = 123,
  sym__html_block_1 = 124,
  sym__html_block_2 = 125,
  sym__html_block_3 = 126,
  sym__html_block_4 = 127,
  sym__html_block_5 = 128,
  sym__html_block_6 = 129,
  sym__html_block_7 = 130,
  sym_link_reference_definition = 131,
  sym__text_inline_no_link = 132,
  sym_paragraph = 133,
  sym__blank_line = 134,
  sym_block_quote = 135,
  sym_list = 136,
  sym__list_plus = 137,
  sym__list_minus = 138,
  sym__list_star = 139,
  sym__list_dot = 140,
  sym__list_parenthesis = 141,
  sym_list_marker_plus = 142,
  sym_list_marker_minus = 143,
  sym_list_marker_star = 144,
  sym_list_marker_dot = 145,
  sym_list_marker_parenthesis = 146,
  sym__list_item_plus = 147,
  sym__list_item_minus = 148,
  sym__list_item_star = 149,
  sym__list_item_dot = 150,
  sym__list_item_parenthesis = 151,
  sym__list_item_content = 152,
  sym__newline = 153,
  sym__soft_line_break = 154,
  sym__line = 155,
  sym__word = 156,
  sym_task_list_marker_checked = 157,
  sym_task_list_marker_unchecked = 158,
  sym_pipe_table = 159,
  sym__pipe_table_newline = 160,
  sym_pipe_table_delimiter_row = 161,
  sym_pipe_table_delimiter_cell = 162,
  sym_pipe_table_row = 163,
  sym_pipe_table_cell = 164,
  aux_sym_document_repeat1 = 165,
  aux_sym_document_repeat2 = 166,
  aux_sym_link_label_repeat1 = 167,
  aux_sym_link_destination_repeat1 = 168,
  aux_sym_link_destination_repeat2 = 169,
  aux_sym_link_title_repeat1 = 170,
  aux_sym_link_title_repeat2 = 171,
  aux_sym_link_title_repeat3 = 172,
  aux_sym__section1_repeat1 = 173,
  aux_sym__section2_repeat1 = 174,
  aux_sym__section3_repeat1 = 175,
  aux_sym__section4_repeat1 = 176,
  aux_sym__section5_repeat1 = 177,
  aux_sym_indented_code_block_repeat1 = 178,
  aux_sym__indented_chunk_repeat1 = 179,
  aux_sym_code_fence_content_repeat1 = 180,
  aux_sym_info_string_repeat1 = 181,
  aux_sym_info_string_repeat2 = 182,
  aux_sym_language_repeat1 = 183,
  aux_sym__html_block_1_repeat1 = 184,
  aux_sym__html_block_2_repeat1 = 185,
  aux_sym__html_block_3_repeat1 = 186,
  aux_sym__html_block_4_repeat1 = 187,
  aux_sym__html_block_5_repeat1 = 188,
  aux_sym__html_block_6_repeat1 = 189,
  aux_sym_paragraph_repeat1 = 190,
  aux_sym_block_quote_repeat1 = 191,
  aux_sym__list_plus_repeat1 = 192,
  aux_sym__list_minus_repeat1 = 193,
  aux_sym__list_star_repeat1 = 194,
  aux_sym__list_dot_repeat1 = 195,
  aux_sym__list_parenthesis_repeat1 = 196,
  aux_sym__line_repeat1 = 197,
  aux_sym_pipe_table_repeat1 = 198,
  aux_sym_pipe_table_delimiter_row_repeat1 = 199,
  aux_sym_pipe_table_delimiter_cell_repeat1 = 200,
  aux_sym_pipe_table_row_repeat1 = 201,
  aux_sym_pipe_table_cell_repeat1 = 202,
  alias_sym_inline = 203,
  alias_sym_pipe_table_align_left = 204,
  alias_sym_pipe_table_align_right = 205,
  alias_sym_pipe_table_header = 206,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__backslash_escape] = "_backslash_escape",
  [sym_entity_reference] = "entity_reference",
  [sym_numeric_character_reference] = "numeric_character_reference",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_BANG] = "!",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_POUND] = "#",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_PERCENT] = "%",
  [anon_sym_AMP] = "&",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_STAR] = "*",
  [anon_sym_PLUS] = "+",
  [anon_sym_COMMA] = ",",
  [anon_sym_DASH] = "-",
  [anon_sym_DOT] = ".",
  [anon_sym_SLASH] = "/",
  [anon_sym_COLON] = ":",
  [anon_sym_SEMI] = ";",
  [anon_sym_EQ] = "=",
  [anon_sym_QMARK] = "\?",
  [anon_sym_AT] = "@",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_CARET] = "^",
  [anon_sym__] = "_",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_LBRACE] = "{",
  [anon_sym_PIPE] = "|",
  [anon_sym_RBRACE] = "}",
  [anon_sym_TILDE] = "~",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_DASH_DASH_GT] = "-->",
  [anon_sym_QMARK_GT] = "\?>",
  [anon_sym_RBRACK_RBRACK_GT] = "]]>",
  [aux_sym__word_token1] = "_word_token1",
  [aux_sym__word_token2] = "_word_token2",
  [aux_sym__word_token3] = "_word_token3",
  [sym__whitespace] = "_whitespace",
  [sym__line_ending] = "_line_ending",
  [sym__soft_line_ending] = "_soft_line_ending",
  [sym__block_close] = "_block_close",
  [sym_block_continuation] = "block_continuation",
  [sym__block_quote_start] = "block_quote_marker",
  [sym__indented_chunk_start] = "_indented_chunk_start",
  [sym_atx_h1_marker] = "atx_h1_marker",
  [sym_atx_h2_marker] = "atx_h2_marker",
  [sym_atx_h3_marker] = "atx_h3_marker",
  [sym_atx_h4_marker] = "atx_h4_marker",
  [sym_atx_h5_marker] = "atx_h5_marker",
  [sym_atx_h6_marker] = "atx_h6_marker",
  [sym_setext_h1_underline] = "setext_h1_underline",
  [sym_setext_h2_underline] = "setext_h2_underline",
  [sym__thematic_break] = "_thematic_break",
  [sym__list_marker_minus] = "_list_marker_minus",
  [sym__list_marker_plus] = "_list_marker_plus",
  [sym__list_marker_star] = "_list_marker_star",
  [sym__list_marker_parenthesis] = "_list_marker_parenthesis",
  [sym__list_marker_dot] = "_list_marker_dot",
  [sym__list_marker_minus_dont_interrupt] = "_list_marker_minus_dont_interrupt",
  [sym__list_marker_plus_dont_interrupt] = "_list_marker_plus_dont_interrupt",
  [sym__list_marker_star_dont_interrupt] = "_list_marker_star_dont_interrupt",
  [sym__list_marker_parenthesis_dont_interrupt] = "_list_marker_parenthesis_dont_interrupt",
  [sym__list_marker_dot_dont_interrupt] = "_list_marker_dot_dont_interrupt",
  [sym__fenced_code_block_start_backtick] = "fenced_code_block_delimiter",
  [sym__fenced_code_block_start_tilde] = "fenced_code_block_delimiter",
  [sym__blank_line_start] = "_blank_line_start",
  [sym__fenced_code_block_end_backtick] = "fenced_code_block_delimiter",
  [sym__fenced_code_block_end_tilde] = "fenced_code_block_delimiter",
  [sym__html_block_1_start] = "_html_block_1_start",
  [sym__html_block_1_end] = "_html_block_1_end",
  [sym__html_block_2_start] = "_html_block_2_start",
  [sym__html_block_3_start] = "_html_block_3_start",
  [sym__html_block_4_start] = "_html_block_4_start",
  [sym__html_block_5_start] = "_html_block_5_start",
  [sym__html_block_6_start] = "_html_block_6_start",
  [sym__html_block_7_start] = "_html_block_7_start",
  [sym__close_block] = "_close_block",
  [sym__no_indented_chunk] = "_no_indented_chunk",
  [sym__error] = "_error",
  [sym__trigger_error] = "_trigger_error",
  [sym__eof] = "_eof",
  [sym_minus_metadata] = "minus_metadata",
  [sym_plus_metadata] = "plus_metadata",
  [sym__pipe_table_start] = "_pipe_table_start",
  [sym__pipe_table_line_ending] = "_pipe_table_line_ending",
  [sym_document] = "document",
  [sym_backslash_escape] = "backslash_escape",
  [sym_link_label] = "link_label",
  [sym_link_destination] = "link_destination",
  [sym__link_destination_parenthesis] = "_link_destination_parenthesis",
  [sym__text_no_angle] = "_text_no_angle",
  [sym_link_title] = "link_title",
  [sym__last_token_punctuation] = "_last_token_punctuation",
  [sym__block] = "_block",
  [sym__block_not_section] = "_block_not_section",
  [sym_section] = "section",
  [sym__section1] = "_section1",
  [sym__section2] = "_section2",
  [sym__section3] = "_section3",
  [sym__section4] = "_section4",
  [sym__section5] = "_section5",
  [sym__section6] = "_section6",
  [sym_thematic_break] = "thematic_break",
  [sym__atx_heading1] = "atx_heading",
  [sym__atx_heading2] = "atx_heading",
  [sym__atx_heading3] = "atx_heading",
  [sym__atx_heading4] = "atx_heading",
  [sym__atx_heading5] = "atx_heading",
  [sym__atx_heading6] = "atx_heading",
  [sym__atx_heading_content] = "_atx_heading_content",
  [sym__setext_heading1] = "setext_heading",
  [sym__setext_heading2] = "setext_heading",
  [sym_indented_code_block] = "indented_code_block",
  [sym__indented_chunk] = "_indented_chunk",
  [sym_fenced_code_block] = "fenced_code_block",
  [sym_code_fence_content] = "code_fence_content",
  [sym_info_string] = "info_string",
  [sym_language] = "language",
  [sym_html_block] = "html_block",
  [sym__html_block_1] = "_html_block_1",
  [sym__html_block_2] = "_html_block_2",
  [sym__html_block_3] = "_html_block_3",
  [sym__html_block_4] = "_html_block_4",
  [sym__html_block_5] = "_html_block_5",
  [sym__html_block_6] = "_html_block_6",
  [sym__html_block_7] = "_html_block_7",
  [sym_link_reference_definition] = "link_reference_definition",
  [sym__text_inline_no_link] = "_text_inline_no_link",
  [sym_paragraph] = "paragraph",
  [sym__blank_line] = "_blank_line",
  [sym_block_quote] = "block_quote",
  [sym_list] = "list",
  [sym__list_plus] = "_list_plus",
  [sym__list_minus] = "_list_minus",
  [sym__list_star] = "_list_star",
  [sym__list_dot] = "_list_dot",
  [sym__list_parenthesis] = "_list_parenthesis",
  [sym_list_marker_plus] = "list_marker_plus",
  [sym_list_marker_minus] = "list_marker_minus",
  [sym_list_marker_star] = "list_marker_star",
  [sym_list_marker_dot] = "list_marker_dot",
  [sym_list_marker_parenthesis] = "list_marker_parenthesis",
  [sym__list_item_plus] = "list_item",
  [sym__list_item_minus] = "list_item",
  [sym__list_item_star] = "list_item",
  [sym__list_item_dot] = "list_item",
  [sym__list_item_parenthesis] = "list_item",
  [sym__list_item_content] = "_list_item_content",
  [sym__newline] = "_newline",
  [sym__soft_line_break] = "_soft_line_break",
  [sym__line] = "_line",
  [sym__word] = "_word",
  [sym_task_list_marker_checked] = "task_list_marker_checked",
  [sym_task_list_marker_unchecked] = "task_list_marker_unchecked",
  [sym_pipe_table] = "pipe_table",
  [sym__pipe_table_newline] = "_pipe_table_newline",
  [sym_pipe_table_delimiter_row] = "pipe_table_delimiter_row",
  [sym_pipe_table_delimiter_cell] = "pipe_table_delimiter_cell",
  [sym_pipe_table_row] = "pipe_table_row",
  [sym_pipe_table_cell] = "pipe_table_cell",
  [aux_sym_document_repeat1] = "document_repeat1",
  [aux_sym_document_repeat2] = "document_repeat2",
  [aux_sym_link_label_repeat1] = "link_label_repeat1",
  [aux_sym_link_destination_repeat1] = "link_destination_repeat1",
  [aux_sym_link_destination_repeat2] = "link_destination_repeat2",
  [aux_sym_link_title_repeat1] = "link_title_repeat1",
  [aux_sym_link_title_repeat2] = "link_title_repeat2",
  [aux_sym_link_title_repeat3] = "link_title_repeat3",
  [aux_sym__section1_repeat1] = "_section1_repeat1",
  [aux_sym__section2_repeat1] = "_section2_repeat1",
  [aux_sym__section3_repeat1] = "_section3_repeat1",
  [aux_sym__section4_repeat1] = "_section4_repeat1",
  [aux_sym__section5_repeat1] = "_section5_repeat1",
  [aux_sym_indented_code_block_repeat1] = "indented_code_block_repeat1",
  [aux_sym__indented_chunk_repeat1] = "_indented_chunk_repeat1",
  [aux_sym_code_fence_content_repeat1] = "code_fence_content_repeat1",
  [aux_sym_info_string_repeat1] = "info_string_repeat1",
  [aux_sym_info_string_repeat2] = "info_string_repeat2",
  [aux_sym_language_repeat1] = "language_repeat1",
  [aux_sym__html_block_1_repeat1] = "_html_block_1_repeat1",
  [aux_sym__html_block_2_repeat1] = "_html_block_2_repeat1",
  [aux_sym__html_block_3_repeat1] = "_html_block_3_repeat1",
  [aux_sym__html_block_4_repeat1] = "_html_block_4_repeat1",
  [aux_sym__html_block_5_repeat1] = "_html_block_5_repeat1",
  [aux_sym__html_block_6_repeat1] = "_html_block_6_repeat1",
  [aux_sym_paragraph_repeat1] = "paragraph_repeat1",
  [aux_sym_block_quote_repeat1] = "block_quote_repeat1",
  [aux_sym__list_plus_repeat1] = "_list_plus_repeat1",
  [aux_sym__list_minus_repeat1] = "_list_minus_repeat1",
  [aux_sym__list_star_repeat1] = "_list_star_repeat1",
  [aux_sym__list_dot_repeat1] = "_list_dot_repeat1",
  [aux_sym__list_parenthesis_repeat1] = "_list_parenthesis_repeat1",
  [aux_sym__line_repeat1] = "_line_repeat1",
  [aux_sym_pipe_table_repeat1] = "pipe_table_repeat1",
  [aux_sym_pipe_table_delimiter_row_repeat1] = "pipe_table_delimiter_row_repeat1",
  [aux_sym_pipe_table_delimiter_cell_repeat1] = "pipe_table_delimiter_cell_repeat1",
  [aux_sym_pipe_table_row_repeat1] = "pipe_table_row_repeat1",
  [aux_sym_pipe_table_cell_repeat1] = "pipe_table_cell_repeat1",
  [alias_sym_inline] = "inline",
  [alias_sym_pipe_table_align_left] = "pipe_table_align_left",
  [alias_sym_pipe_table_align_right] = "pipe_table_align_right",
  [alias_sym_pipe_table_header] = "pipe_table_header",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__backslash_escape] = sym__backslash_escape,
  [sym_entity_reference] = sym_entity_reference,
  [sym_numeric_character_reference] = sym_numeric_character_reference,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym__] = anon_sym__,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_DASH_DASH_GT] = anon_sym_DASH_DASH_GT,
  [anon_sym_QMARK_GT] = anon_sym_QMARK_GT,
  [anon_sym_RBRACK_RBRACK_GT] = anon_sym_RBRACK_RBRACK_GT,
  [aux_sym__word_token1] = aux_sym__word_token1,
  [aux_sym__word_token2] = aux_sym__word_token2,
  [aux_sym__word_token3] = aux_sym__word_token3,
  [sym__whitespace] = sym__whitespace,
  [sym__line_ending] = sym__line_ending,
  [sym__soft_line_ending] = sym__soft_line_ending,
  [sym__block_close] = sym__block_close,
  [sym_block_continuation] = sym_block_continuation,
  [sym__block_quote_start] = sym__block_quote_start,
  [sym__indented_chunk_start] = sym__indented_chunk_start,
  [sym_atx_h1_marker] = sym_atx_h1_marker,
  [sym_atx_h2_marker] = sym_atx_h2_marker,
  [sym_atx_h3_marker] = sym_atx_h3_marker,
  [sym_atx_h4_marker] = sym_atx_h4_marker,
  [sym_atx_h5_marker] = sym_atx_h5_marker,
  [sym_atx_h6_marker] = sym_atx_h6_marker,
  [sym_setext_h1_underline] = sym_setext_h1_underline,
  [sym_setext_h2_underline] = sym_setext_h2_underline,
  [sym__thematic_break] = sym__thematic_break,
  [sym__list_marker_minus] = sym__list_marker_minus,
  [sym__list_marker_plus] = sym__list_marker_plus,
  [sym__list_marker_star] = sym__list_marker_star,
  [sym__list_marker_parenthesis] = sym__list_marker_parenthesis,
  [sym__list_marker_dot] = sym__list_marker_dot,
  [sym__list_marker_minus_dont_interrupt] = sym__list_marker_minus_dont_interrupt,
  [sym__list_marker_plus_dont_interrupt] = sym__list_marker_plus_dont_interrupt,
  [sym__list_marker_star_dont_interrupt] = sym__list_marker_star_dont_interrupt,
  [sym__list_marker_parenthesis_dont_interrupt] = sym__list_marker_parenthesis_dont_interrupt,
  [sym__list_marker_dot_dont_interrupt] = sym__list_marker_dot_dont_interrupt,
  [sym__fenced_code_block_start_backtick] = sym__fenced_code_block_start_backtick,
  [sym__fenced_code_block_start_tilde] = sym__fenced_code_block_start_backtick,
  [sym__blank_line_start] = sym__blank_line_start,
  [sym__fenced_code_block_end_backtick] = sym__fenced_code_block_start_backtick,
  [sym__fenced_code_block_end_tilde] = sym__fenced_code_block_start_backtick,
  [sym__html_block_1_start] = sym__html_block_1_start,
  [sym__html_block_1_end] = sym__html_block_1_end,
  [sym__html_block_2_start] = sym__html_block_2_start,
  [sym__html_block_3_start] = sym__html_block_3_start,
  [sym__html_block_4_start] = sym__html_block_4_start,
  [sym__html_block_5_start] = sym__html_block_5_start,
  [sym__html_block_6_start] = sym__html_block_6_start,
  [sym__html_block_7_start] = sym__html_block_7_start,
  [sym__close_block] = sym__close_block,
  [sym__no_indented_chunk] = sym__no_indented_chunk,
  [sym__error] = sym__error,
  [sym__trigger_error] = sym__trigger_error,
  [sym__eof] = sym__eof,
  [sym_minus_metadata] = sym_minus_metadata,
  [sym_plus_metadata] = sym_plus_metadata,
  [sym__pipe_table_start] = sym__pipe_table_start,
  [sym__pipe_table_line_ending] = sym__pipe_table_line_ending,
  [sym_document] = sym_document,
  [sym_backslash_escape] = sym_backslash_escape,
  [sym_link_label] = sym_link_label,
  [sym_link_destination] = sym_link_destination,
  [sym__link_destination_parenthesis] = sym__link_destination_parenthesis,
  [sym__text_no_angle] = sym__text_no_angle,
  [sym_link_title] = sym_link_title,
  [sym__last_token_punctuation] = sym__last_token_punctuation,
  [sym__block] = sym__block,
  [sym__block_not_section] = sym__block_not_section,
  [sym_section] = sym_section,
  [sym__section1] = sym__section1,
  [sym__section2] = sym__section2,
  [sym__section3] = sym__section3,
  [sym__section4] = sym__section4,
  [sym__section5] = sym__section5,
  [sym__section6] = sym__section6,
  [sym_thematic_break] = sym_thematic_break,
  [sym__atx_heading1] = sym__atx_heading1,
  [sym__atx_heading2] = sym__atx_heading1,
  [sym__atx_heading3] = sym__atx_heading1,
  [sym__atx_heading4] = sym__atx_heading1,
  [sym__atx_heading5] = sym__atx_heading1,
  [sym__atx_heading6] = sym__atx_heading1,
  [sym__atx_heading_content] = sym__atx_heading_content,
  [sym__setext_heading1] = sym__setext_heading1,
  [sym__setext_heading2] = sym__setext_heading1,
  [sym_indented_code_block] = sym_indented_code_block,
  [sym__indented_chunk] = sym__indented_chunk,
  [sym_fenced_code_block] = sym_fenced_code_block,
  [sym_code_fence_content] = sym_code_fence_content,
  [sym_info_string] = sym_info_string,
  [sym_language] = sym_language,
  [sym_html_block] = sym_html_block,
  [sym__html_block_1] = sym__html_block_1,
  [sym__html_block_2] = sym__html_block_2,
  [sym__html_block_3] = sym__html_block_3,
  [sym__html_block_4] = sym__html_block_4,
  [sym__html_block_5] = sym__html_block_5,
  [sym__html_block_6] = sym__html_block_6,
  [sym__html_block_7] = sym__html_block_7,
  [sym_link_reference_definition] = sym_link_reference_definition,
  [sym__text_inline_no_link] = sym__text_inline_no_link,
  [sym_paragraph] = sym_paragraph,
  [sym__blank_line] = sym__blank_line,
  [sym_block_quote] = sym_block_quote,
  [sym_list] = sym_list,
  [sym__list_plus] = sym__list_plus,
  [sym__list_minus] = sym__list_minus,
  [sym__list_star] = sym__list_star,
  [sym__list_dot] = sym__list_dot,
  [sym__list_parenthesis] = sym__list_parenthesis,
  [sym_list_marker_plus] = sym_list_marker_plus,
  [sym_list_marker_minus] = sym_list_marker_minus,
  [sym_list_marker_star] = sym_list_marker_star,
  [sym_list_marker_dot] = sym_list_marker_dot,
  [sym_list_marker_parenthesis] = sym_list_marker_parenthesis,
  [sym__list_item_plus] = sym__list_item_plus,
  [sym__list_item_minus] = sym__list_item_plus,
  [sym__list_item_star] = sym__list_item_plus,
  [sym__list_item_dot] = sym__list_item_plus,
  [sym__list_item_parenthesis] = sym__list_item_plus,
  [sym__list_item_content] = sym__list_item_content,
  [sym__newline] = sym__newline,
  [sym__soft_line_break] = sym__soft_line_break,
  [sym__line] = sym__line,
  [sym__word] = sym__word,
  [sym_task_list_marker_checked] = sym_task_list_marker_checked,
  [sym_task_list_marker_unchecked] = sym_task_list_marker_unchecked,
  [sym_pipe_table] = sym_pipe_table,
  [sym__pipe_table_newline] = sym__pipe_table_newline,
  [sym_pipe_table_delimiter_row] = sym_pipe_table_delimiter_row,
  [sym_pipe_table_delimiter_cell] = sym_pipe_table_delimiter_cell,
  [sym_pipe_table_row] = sym_pipe_table_row,
  [sym_pipe_table_cell] = sym_pipe_table_cell,
  [aux_sym_document_repeat1] = aux_sym_document_repeat1,
  [aux_sym_document_repeat2] = aux_sym_document_repeat2,
  [aux_sym_link_label_repeat1] = aux_sym_link_label_repeat1,
  [aux_sym_link_destination_repeat1] = aux_sym_link_destination_repeat1,
  [aux_sym_link_destination_repeat2] = aux_sym_link_destination_repeat2,
  [aux_sym_link_title_repeat1] = aux_sym_link_title_repeat1,
  [aux_sym_link_title_repeat2] = aux_sym_link_title_repeat2,
  [aux_sym_link_title_repeat3] = aux_sym_link_title_repeat3,
  [aux_sym__section1_repeat1] = aux_sym__section1_repeat1,
  [aux_sym__section2_repeat1] = aux_sym__section2_repeat1,
  [aux_sym__section3_repeat1] = aux_sym__section3_repeat1,
  [aux_sym__section4_repeat1] = aux_sym__section4_repeat1,
  [aux_sym__section5_repeat1] = aux_sym__section5_repeat1,
  [aux_sym_indented_code_block_repeat1] = aux_sym_indented_code_block_repeat1,
  [aux_sym__indented_chunk_repeat1] = aux_sym__indented_chunk_repeat1,
  [aux_sym_code_fence_content_repeat1] = aux_sym_code_fence_content_repeat1,
  [aux_sym_info_string_repeat1] = aux_sym_info_string_repeat1,
  [aux_sym_info_string_repeat2] = aux_sym_info_string_repeat2,
  [aux_sym_language_repeat1] = aux_sym_language_repeat1,
  [aux_sym__html_block_1_repeat1] = aux_sym__html_block_1_repeat1,
  [aux_sym__html_block_2_repeat1] = aux_sym__html_block_2_repeat1,
  [aux_sym__html_block_3_repeat1] = aux_sym__html_block_3_repeat1,
  [aux_sym__html_block_4_repeat1] = aux_sym__html_block_4_repeat1,
  [aux_sym__html_block_5_repeat1] = aux_sym__html_block_5_repeat1,
  [aux_sym__html_block_6_repeat1] = aux_sym__html_block_6_repeat1,
  [aux_sym_paragraph_repeat1] = aux_sym_paragraph_repeat1,
  [aux_sym_block_quote_repeat1] = aux_sym_block_quote_repeat1,
  [aux_sym__list_plus_repeat1] = aux_sym__list_plus_repeat1,
  [aux_sym__list_minus_repeat1] = aux_sym__list_minus_repeat1,
  [aux_sym__list_star_repeat1] = aux_sym__list_star_repeat1,
  [aux_sym__list_dot_repeat1] = aux_sym__list_dot_repeat1,
  [aux_sym__list_parenthesis_repeat1] = aux_sym__list_parenthesis_repeat1,
  [aux_sym__line_repeat1] = aux_sym__line_repeat1,
  [aux_sym_pipe_table_repeat1] = aux_sym_pipe_table_repeat1,
  [aux_sym_pipe_table_delimiter_row_repeat1] = aux_sym_pipe_table_delimiter_row_repeat1,
  [aux_sym_pipe_table_delimiter_cell_repeat1] = aux_sym_pipe_table_delimiter_cell_repeat1,
  [aux_sym_pipe_table_row_repeat1] = aux_sym_pipe_table_row_repeat1,
  [aux_sym_pipe_table_cell_repeat1] = aux_sym_pipe_table_cell_repeat1,
  [alias_sym_inline] = alias_sym_inline,
  [alias_sym_pipe_table_align_left] = alias_sym_pipe_table_align_left,
  [alias_sym_pipe_table_align_right] = alias_sym_pipe_table_align_right,
  [alias_sym_pipe_table_header] = alias_sym_pipe_table_header,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__backslash_escape] = {
    .visible = false,
    .named = true,
  },
  [sym_entity_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_numeric_character_reference] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK_RBRACK_GT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__word_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__word_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__word_token3] = {
    .visible = false,
    .named = false,
  },
  [sym__whitespace] = {
    .visible = false,
    .named = true,
  },
  [sym__line_ending] = {
    .visible = false,
    .named = true,
  },
  [sym__soft_line_ending] = {
    .visible = false,
    .named = true,
  },
  [sym__block_close] = {
    .visible = false,
    .named = true,
  },
  [sym_block_continuation] = {
    .visible = true,
    .named = true,
  },
  [sym__block_quote_start] = {
    .visible = true,
    .named = true,
  },
  [sym__indented_chunk_start] = {
    .visible = false,
    .named = true,
  },
  [sym_atx_h1_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_atx_h2_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_atx_h3_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_atx_h4_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_atx_h5_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_atx_h6_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_setext_h1_underline] = {
    .visible = true,
    .named = true,
  },
  [sym_setext_h2_underline] = {
    .visible = true,
    .named = true,
  },
  [sym__thematic_break] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_minus] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_plus] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_star] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_dot] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_minus_dont_interrupt] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_plus_dont_interrupt] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_star_dont_interrupt] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_parenthesis_dont_interrupt] = {
    .visible = false,
    .named = true,
  },
  [sym__list_marker_dot_dont_interrupt] = {
    .visible = false,
    .named = true,
  },
  [sym__fenced_code_block_start_backtick] = {
    .visible = true,
    .named = true,
  },
  [sym__fenced_code_block_start_tilde] = {
    .visible = true,
    .named = true,
  },
  [sym__blank_line_start] = {
    .visible = false,
    .named = true,
  },
  [sym__fenced_code_block_end_backtick] = {
    .visible = true,
    .named = true,
  },
  [sym__fenced_code_block_end_tilde] = {
    .visible = true,
    .named = true,
  },
  [sym__html_block_1_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_1_end] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_2_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_3_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_4_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_5_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_6_start] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_7_start] = {
    .visible = false,
    .named = true,
  },
  [sym__close_block] = {
    .visible = false,
    .named = true,
  },
  [sym__no_indented_chunk] = {
    .visible = false,
    .named = true,
  },
  [sym__error] = {
    .visible = false,
    .named = true,
  },
  [sym__trigger_error] = {
    .visible = false,
    .named = true,
  },
  [sym__eof] = {
    .visible = false,
    .named = true,
  },
  [sym_minus_metadata] = {
    .visible = true,
    .named = true,
  },
  [sym_plus_metadata] = {
    .visible = true,
    .named = true,
  },
  [sym__pipe_table_start] = {
    .visible = false,
    .named = true,
  },
  [sym__pipe_table_line_ending] = {
    .visible = false,
    .named = true,
  },
  [sym_document] = {
    .visible = true,
    .named = true,
  },
  [sym_backslash_escape] = {
    .visible = true,
    .named = true,
  },
  [sym_link_label] = {
    .visible = true,
    .named = true,
  },
  [sym_link_destination] = {
    .visible = true,
    .named = true,
  },
  [sym__link_destination_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym__text_no_angle] = {
    .visible = false,
    .named = true,
  },
  [sym_link_title] = {
    .visible = true,
    .named = true,
  },
  [sym__last_token_punctuation] = {
    .visible = false,
    .named = true,
  },
  [sym__block] = {
    .visible = false,
    .named = true,
  },
  [sym__block_not_section] = {
    .visible = false,
    .named = true,
  },
  [sym_section] = {
    .visible = true,
    .named = true,
  },
  [sym__section1] = {
    .visible = false,
    .named = true,
  },
  [sym__section2] = {
    .visible = false,
    .named = true,
  },
  [sym__section3] = {
    .visible = false,
    .named = true,
  },
  [sym__section4] = {
    .visible = false,
    .named = true,
  },
  [sym__section5] = {
    .visible = false,
    .named = true,
  },
  [sym__section6] = {
    .visible = false,
    .named = true,
  },
  [sym_thematic_break] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading1] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading2] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading3] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading4] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading5] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading6] = {
    .visible = true,
    .named = true,
  },
  [sym__atx_heading_content] = {
    .visible = false,
    .named = true,
  },
  [sym__setext_heading1] = {
    .visible = true,
    .named = true,
  },
  [sym__setext_heading2] = {
    .visible = true,
    .named = true,
  },
  [sym_indented_code_block] = {
    .visible = true,
    .named = true,
  },
  [sym__indented_chunk] = {
    .visible = false,
    .named = true,
  },
  [sym_fenced_code_block] = {
    .visible = true,
    .named = true,
  },
  [sym_code_fence_content] = {
    .visible = true,
    .named = true,
  },
  [sym_info_string] = {
    .visible = true,
    .named = true,
  },
  [sym_language] = {
    .visible = true,
    .named = true,
  },
  [sym_html_block] = {
    .visible = true,
    .named = true,
  },
  [sym__html_block_1] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_2] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_3] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_4] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_5] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_6] = {
    .visible = false,
    .named = true,
  },
  [sym__html_block_7] = {
    .visible = false,
    .named = true,
  },
  [sym_link_reference_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__text_inline_no_link] = {
    .visible = false,
    .named = true,
  },
  [sym_paragraph] = {
    .visible = true,
    .named = true,
  },
  [sym__blank_line] = {
    .visible = false,
    .named = true,
  },
  [sym_block_quote] = {
    .visible = true,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym__list_plus] = {
    .visible = false,
    .named = true,
  },
  [sym__list_minus] = {
    .visible = false,
    .named = true,
  },
  [sym__list_star] = {
    .visible = false,
    .named = true,
  },
  [sym__list_dot] = {
    .visible = false,
    .named = true,
  },
  [sym__list_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym_list_marker_plus] = {
    .visible = true,
    .named = true,
  },
  [sym_list_marker_minus] = {
    .visible = true,
    .named = true,
  },
  [sym_list_marker_star] = {
    .visible = true,
    .named = true,
  },
  [sym_list_marker_dot] = {
    .visible = true,
    .named = true,
  },
  [sym_list_marker_parenthesis] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_plus] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_minus] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_star] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_dot] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_parenthesis] = {
    .visible = true,
    .named = true,
  },
  [sym__list_item_content] = {
    .visible = false,
    .named = true,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym__soft_line_break] = {
    .visible = false,
    .named = true,
  },
  [sym__line] = {
    .visible = false,
    .named = true,
  },
  [sym__word] = {
    .visible = false,
    .named = true,
  },
  [sym_task_list_marker_checked] = {
    .visible = true,
    .named = true,
  },
  [sym_task_list_marker_unchecked] = {
    .visible = true,
    .named = true,
  },
  [sym_pipe_table] = {
    .visible = true,
    .named = true,
  },
  [sym__pipe_table_newline] = {
    .visible = false,
    .named = true,
  },
  [sym_pipe_table_delimiter_row] = {
    .visible = true,
    .named = true,
  },
  [sym_pipe_table_delimiter_cell] = {
    .visible = true,
    .named = true,
  },
  [sym_pipe_table_row] = {
    .visible = true,
    .named = true,
  },
  [sym_pipe_table_cell] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_document_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_document_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_label_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_destination_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_destination_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_link_title_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__section1_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__section2_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__section3_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__section4_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__section5_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_indented_code_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__indented_chunk_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_code_fence_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_info_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_info_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_language_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_1_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_2_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_3_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_4_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_5_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__html_block_6_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_paragraph_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_quote_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_plus_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_minus_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_star_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_dot_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_parenthesis_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__line_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipe_table_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipe_table_delimiter_row_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipe_table_delimiter_cell_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipe_table_row_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pipe_table_cell_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_inline] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_pipe_table_align_left] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_pipe_table_align_right] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_pipe_table_header] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_heading_content = 1,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_heading_content] = "heading_content",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [6] = {.index = 2, .length = 1},
  [7] = {.index = 3, .length = 1},
  [9] = {.index = 1, .length = 1},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_heading_content, 0, .inherited = true},
  [1] =
    {field_heading_content, 0},
  [2] =
    {field_heading_content, 1},
  [3] =
    {field_heading_content, 1, .inherited = true},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [2] = {
    [0] = sym_section,
  },
  [3] = {
    [0] = alias_sym_inline,
  },
  [4] = {
    [1] = sym_section,
  },
  [5] = {
    [0] = alias_sym_inline,
  },
  [6] = {
    [1] = alias_sym_inline,
  },
  [8] = {
    [0] = sym_pipe_table_cell,
  },
  [10] = {
    [0] = alias_sym_pipe_table_align_left,
  },
  [11] = {
    [1] = alias_sym_pipe_table_header,
  },
  [12] = {
    [1] = alias_sym_pipe_table_align_right,
  },
  [13] = {
    [0] = alias_sym_pipe_table_align_left,
    [2] = alias_sym_pipe_table_align_right,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__section2, 2,
    sym__section2,
    sym_section,
  sym__section3, 2,
    sym__section3,
    sym_section,
  sym__section4, 2,
    sym__section4,
    sym_section,
  sym__section5, 2,
    sym__section5,
    sym_section,
  sym__section6, 2,
    sym__section6,
    sym_section,
  sym__line, 2,
    sym__line,
    alias_sym_inline,
  sym_pipe_table_row, 2,
    sym_pipe_table_row,
    alias_sym_pipe_table_header,
  aux_sym_document_repeat1, 2,
    aux_sym_document_repeat1,
    sym_section,
  aux_sym_paragraph_repeat1, 2,
    aux_sym_paragraph_repeat1,
    alias_sym_inline,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 3,
  [8] = 4,
  [9] = 2,
  [10] = 6,
  [11] = 5,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 12,
  [18] = 13,
  [19] = 14,
  [20] = 15,
  [21] = 16,
  [22] = 22,
  [23] = 22,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 24,
  [33] = 33,
  [34] = 29,
  [35] = 35,
  [36] = 33,
  [37] = 37,
  [38] = 37,
  [39] = 39,
  [40] = 39,
  [41] = 41,
  [42] = 41,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 45,
  [47] = 44,
  [48] = 43,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 50,
  [53] = 49,
  [54] = 51,
  [55] = 55,
  [56] = 56,
  [57] = 56,
  [58] = 55,
  [59] = 59,
  [60] = 59,
  [61] = 61,
  [62] = 62,
  [63] = 61,
  [64] = 64,
  [65] = 64,
  [66] = 62,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 67,
  [71] = 68,
  [72] = 69,
  [73] = 73,
  [74] = 74,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 73,
  [81] = 81,
  [82] = 74,
  [83] = 83,
  [84] = 84,
  [85] = 85,
  [86] = 83,
  [87] = 75,
  [88] = 76,
  [89] = 77,
  [90] = 84,
  [91] = 85,
  [92] = 92,
  [93] = 93,
  [94] = 93,
  [95] = 95,
  [96] = 78,
  [97] = 92,
  [98] = 79,
  [99] = 81,
  [100] = 95,
  [101] = 101,
  [102] = 101,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 107,
  [109] = 106,
  [110] = 110,
  [111] = 111,
  [112] = 110,
  [113] = 111,
  [114] = 114,
  [115] = 103,
  [116] = 104,
  [117] = 105,
  [118] = 114,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 122,
  [123] = 123,
  [124] = 124,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 130,
  [136] = 131,
  [137] = 119,
  [138] = 133,
  [139] = 134,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 143,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 93,
  [148] = 148,
  [149] = 120,
  [150] = 121,
  [151] = 124,
  [152] = 125,
  [153] = 132,
  [154] = 154,
  [155] = 145,
  [156] = 156,
  [157] = 122,
  [158] = 158,
  [159] = 123,
  [160] = 158,
  [161] = 107,
  [162] = 107,
  [163] = 163,
  [164] = 164,
  [165] = 93,
  [166] = 148,
  [167] = 156,
  [168] = 126,
  [169] = 154,
  [170] = 170,
  [171] = 171,
  [172] = 163,
  [173] = 164,
  [174] = 141,
  [175] = 127,
  [176] = 142,
  [177] = 128,
  [178] = 146,
  [179] = 129,
  [180] = 140,
  [181] = 143,
  [182] = 144,
  [183] = 171,
  [184] = 184,
  [185] = 185,
  [186] = 186,
  [187] = 187,
  [188] = 188,
  [189] = 189,
  [190] = 190,
  [191] = 191,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 196,
  [197] = 197,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 207,
  [208] = 208,
  [209] = 209,
  [210] = 210,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 184,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 226,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 203,
  [234] = 234,
  [235] = 235,
  [236] = 228,
  [237] = 204,
  [238] = 230,
  [239] = 239,
  [240] = 232,
  [241] = 206,
  [242] = 234,
  [243] = 188,
  [244] = 244,
  [245] = 110,
  [246] = 200,
  [247] = 247,
  [248] = 202,
  [249] = 205,
  [250] = 207,
  [251] = 209,
  [252] = 210,
  [253] = 211,
  [254] = 212,
  [255] = 221,
  [256] = 222,
  [257] = 208,
  [258] = 223,
  [259] = 224,
  [260] = 213,
  [261] = 227,
  [262] = 214,
  [263] = 231,
  [264] = 264,
  [265] = 235,
  [266] = 215,
  [267] = 267,
  [268] = 268,
  [269] = 244,
  [270] = 247,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 264,
  [277] = 277,
  [278] = 267,
  [279] = 268,
  [280] = 280,
  [281] = 272,
  [282] = 273,
  [283] = 274,
  [284] = 275,
  [285] = 277,
  [286] = 286,
  [287] = 287,
  [288] = 286,
  [289] = 289,
  [290] = 290,
  [291] = 287,
  [292] = 292,
  [293] = 216,
  [294] = 294,
  [295] = 217,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 289,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 305,
  [306] = 306,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 218,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 219,
  [315] = 271,
  [316] = 185,
  [317] = 186,
  [318] = 187,
  [319] = 290,
  [320] = 189,
  [321] = 190,
  [322] = 191,
  [323] = 192,
  [324] = 193,
  [325] = 194,
  [326] = 111,
  [327] = 292,
  [328] = 294,
  [329] = 158,
  [330] = 296,
  [331] = 158,
  [332] = 297,
  [333] = 298,
  [334] = 299,
  [335] = 110,
  [336] = 300,
  [337] = 302,
  [338] = 303,
  [339] = 195,
  [340] = 304,
  [341] = 305,
  [342] = 306,
  [343] = 307,
  [344] = 308,
  [345] = 309,
  [346] = 220,
  [347] = 311,
  [348] = 312,
  [349] = 313,
  [350] = 280,
  [351] = 351,
  [352] = 351,
  [353] = 353,
  [354] = 354,
  [355] = 355,
  [356] = 354,
  [357] = 355,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 366,
  [368] = 368,
  [369] = 369,
  [370] = 364,
  [371] = 371,
  [372] = 372,
  [373] = 362,
  [374] = 374,
  [375] = 362,
  [376] = 365,
  [377] = 369,
  [378] = 371,
  [379] = 379,
  [380] = 380,
  [381] = 368,
  [382] = 382,
  [383] = 383,
  [384] = 359,
  [385] = 385,
  [386] = 386,
  [387] = 382,
  [388] = 388,
  [389] = 389,
  [390] = 390,
  [391] = 391,
  [392] = 388,
  [393] = 393,
  [394] = 394,
  [395] = 395,
  [396] = 396,
  [397] = 397,
  [398] = 398,
  [399] = 399,
  [400] = 400,
  [401] = 401,
  [402] = 402,
  [403] = 389,
  [404] = 404,
  [405] = 405,
  [406] = 406,
  [407] = 396,
  [408] = 408,
  [409] = 409,
  [410] = 410,
  [411] = 411,
  [412] = 410,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 406,
  [417] = 408,
  [418] = 409,
  [419] = 411,
  [420] = 420,
  [421] = 395,
  [422] = 398,
  [423] = 423,
  [424] = 424,
  [425] = 405,
  [426] = 420,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 427,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 438,
  [439] = 439,
  [440] = 428,
  [441] = 429,
  [442] = 437,
  [443] = 443,
  [444] = 444,
  [445] = 433,
  [446] = 446,
  [447] = 430,
  [448] = 436,
  [449] = 449,
  [450] = 438,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 453,
  [456] = 456,
  [457] = 454,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 459,
  [470] = 470,
  [471] = 468,
  [472] = 472,
  [473] = 461,
  [474] = 460,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 458,
  [483] = 483,
  [484] = 484,
  [485] = 480,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 486,
  [490] = 470,
  [491] = 360,
  [492] = 472,
  [493] = 484,
  [494] = 481,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 466,
  [500] = 467,
  [501] = 466,
  [502] = 107,
  [503] = 503,
  [504] = 504,
  [505] = 467,
  [506] = 466,
  [507] = 467,
  [508] = 431,
  [509] = 466,
  [510] = 510,
  [511] = 511,
  [512] = 512,
  [513] = 466,
  [514] = 514,
  [515] = 467,
  [516] = 516,
  [517] = 517,
  [518] = 466,
  [519] = 467,
  [520] = 466,
  [521] = 521,
  [522] = 467,
  [523] = 523,
  [524] = 521,
  [525] = 525,
  [526] = 510,
  [527] = 512,
  [528] = 511,
  [529] = 467,
  [530] = 530,
  [531] = 531,
  [532] = 466,
  [533] = 158,
  [534] = 495,
  [535] = 497,
  [536] = 462,
  [537] = 107,
  [538] = 93,
  [539] = 497,
  [540] = 463,
  [541] = 477,
  [542] = 478,
  [543] = 543,
  [544] = 93,
  [545] = 431,
  [546] = 93,
  [547] = 93,
  [548] = 93,
  [549] = 479,
  [550] = 93,
  [551] = 551,
  [552] = 552,
  [553] = 93,
  [554] = 554,
  [555] = 475,
  [556] = 556,
  [557] = 557,
  [558] = 467,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 464,
  [564] = 465,
  [565] = 483,
  [566] = 566,
  [567] = 551,
  [568] = 517,
  [569] = 551,
  [570] = 467,
  [571] = 110,
  [572] = 572,
  [573] = 110,
  [574] = 574,
  [575] = 110,
  [576] = 576,
  [577] = 107,
  [578] = 551,
  [579] = 495,
  [580] = 110,
  [581] = 581,
  [582] = 551,
  [583] = 466,
  [584] = 551,
  [585] = 497,
  [586] = 586,
  [587] = 498,
  [588] = 93,
  [589] = 589,
  [590] = 110,
  [591] = 591,
  [592] = 462,
  [593] = 463,
  [594] = 594,
  [595] = 594,
  [596] = 596,
  [597] = 597,
  [598] = 551,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 110,
  [603] = 603,
  [604] = 504,
  [605] = 551,
  [606] = 464,
  [607] = 465,
  [608] = 608,
  [609] = 110,
  [610] = 158,
  [611] = 611,
  [612] = 514,
  [613] = 613,
  [614] = 614,
  [615] = 516,
  [616] = 613,
  [617] = 617,
  [618] = 158,
  [619] = 551,
  [620] = 614,
  [621] = 110,
  [622] = 622,
  [623] = 617,
  [624] = 624,
  [625] = 543,
  [626] = 626,
  [627] = 611,
  [628] = 496,
  [629] = 608,
  [630] = 596,
  [631] = 551,
  [632] = 574,
  [633] = 543,
  [634] = 586,
  [635] = 635,
  [636] = 636,
  [637] = 637,
  [638] = 638,
  [639] = 639,
  [640] = 640,
  [641] = 641,
  [642] = 642,
  [643] = 643,
  [644] = 644,
  [645] = 645,
  [646] = 646,
  [647] = 646,
  [648] = 648,
  [649] = 649,
  [650] = 650,
  [651] = 651,
  [652] = 652,
  [653] = 653,
  [654] = 649,
  [655] = 655,
  [656] = 655,
  [657] = 657,
  [658] = 653,
  [659] = 657,
  [660] = 650,
  [661] = 196,
  [662] = 662,
  [663] = 663,
  [664] = 664,
  [665] = 665,
  [666] = 666,
  [667] = 667,
  [668] = 107,
  [669] = 669,
  [670] = 670,
  [671] = 671,
  [672] = 672,
  [673] = 673,
  [674] = 674,
  [675] = 675,
  [676] = 676,
  [677] = 677,
  [678] = 675,
  [679] = 679,
  [680] = 667,
  [681] = 664,
  [682] = 682,
  [683] = 671,
  [684] = 674,
  [685] = 669,
  [686] = 665,
  [687] = 666,
  [688] = 682,
  [689] = 670,
  [690] = 663,
  [691] = 673,
  [692] = 692,
  [693] = 693,
  [694] = 694,
  [695] = 695,
  [696] = 696,
  [697] = 697,
  [698] = 648,
  [699] = 699,
  [700] = 700,
  [701] = 701,
  [702] = 158,
  [703] = 703,
  [704] = 704,
  [705] = 705,
  [706] = 706,
  [707] = 707,
  [708] = 93,
  [709] = 709,
  [710] = 710,
  [711] = 711,
  [712] = 712,
  [713] = 713,
  [714] = 644,
  [715] = 715,
  [716] = 645,
  [717] = 717,
  [718] = 718,
  [719] = 719,
  [720] = 720,
  [721] = 703,
  [722] = 697,
  [723] = 706,
  [724] = 724,
  [725] = 725,
  [726] = 726,
  [727] = 727,
  [728] = 728,
  [729] = 725,
  [730] = 717,
  [731] = 720,
  [732] = 726,
  [733] = 733,
  [734] = 734,
  [735] = 735,
  [736] = 733,
  [737] = 727,
  [738] = 704,
  [739] = 739,
  [740] = 699,
  [741] = 710,
  [742] = 728,
  [743] = 724,
  [744] = 744,
  [745] = 745,
  [746] = 746,
  [747] = 747,
  [748] = 748,
  [749] = 749,
  [750] = 750,
  [751] = 751,
  [752] = 752,
  [753] = 110,
  [754] = 754,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 758,
  [759] = 759,
  [760] = 760,
  [761] = 745,
  [762] = 762,
  [763] = 763,
  [764] = 764,
  [765] = 756,
  [766] = 766,
  [767] = 767,
  [768] = 754,
  [769] = 769,
  [770] = 758,
  [771] = 766,
  [772] = 748,
  [773] = 773,
  [774] = 757,
  [775] = 775,
  [776] = 700,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 712,
  [781] = 778,
  [782] = 779,
  [783] = 711,
  [784] = 784,
  [785] = 718,
  [786] = 786,
  [787] = 734,
  [788] = 784,
  [789] = 715,
  [790] = 779,
  [791] = 719,
  [792] = 777,
  [793] = 793,
  [794] = 794,
  [795] = 701,
  [796] = 796,
  [797] = 797,
  [798] = 796,
  [799] = 799,
  [800] = 800,
  [801] = 801,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 763,
  [806] = 806,
  [807] = 764,
  [808] = 801,
  [809] = 809,
  [810] = 802,
  [811] = 811,
  [812] = 812,
  [813] = 799,
  [814] = 814,
  [815] = 815,
  [816] = 735,
  [817] = 817,
  [818] = 818,
  [819] = 107,
  [820] = 820,
  [821] = 760,
  [822] = 814,
  [823] = 800,
  [824] = 824,
  [825] = 818,
  [826] = 767,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 824,
  [831] = 831,
  [832] = 828,
  [833] = 746,
  [834] = 93,
  [835] = 835,
  [836] = 836,
  [837] = 93,
  [838] = 803,
  [839] = 839,
  [840] = 840,
  [841] = 836,
  [842] = 842,
  [843] = 817,
  [844] = 762,
  [845] = 835,
  [846] = 797,
  [847] = 747,
  [848] = 769,
  [849] = 829,
  [850] = 811,
  [851] = 812,
  [852] = 827,
  [853] = 831,
  [854] = 815,
  [855] = 806,
  [856] = 693,
  [857] = 820,
  [858] = 842,
  [859] = 840,
  [860] = 860,
  [861] = 861,
  [862] = 861,
  [863] = 863,
  [864] = 864,
  [865] = 220,
  [866] = 866,
  [867] = 867,
  [868] = 868,
  [869] = 863,
  [870] = 870,
  [871] = 860,
  [872] = 872,
  [873] = 873,
  [874] = 110,
  [875] = 875,
  [876] = 876,
  [877] = 877,
  [878] = 878,
  [879] = 879,
  [880] = 880,
  [881] = 881,
  [882] = 879,
  [883] = 873,
  [884] = 884,
  [885] = 884,
  [886] = 110,
  [887] = 887,
  [888] = 888,
  [889] = 889,
  [890] = 890,
  [891] = 888,
  [892] = 892,
  [893] = 887,
  [894] = 889,
  [895] = 895,
  [896] = 892,
  [897] = 868,
  [898] = 158,
  [899] = 890,
  [900] = 900,
  [901] = 864,
  [902] = 902,
  [903] = 903,
  [904] = 904,
  [905] = 905,
  [906] = 906,
  [907] = 907,
  [908] = 908,
  [909] = 909,
  [910] = 910,
  [911] = 908,
  [912] = 881,
  [913] = 913,
  [914] = 914,
  [915] = 915,
  [916] = 905,
  [917] = 877,
  [918] = 918,
  [919] = 904,
  [920] = 906,
  [921] = 921,
  [922] = 921,
  [923] = 909,
  [924] = 918,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(2031);
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2046,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2052,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2059,
        '@', 2060,
        '[', 2035,
        '\\', 2062,
        ']', 2037,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2046,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2051,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2058,
        '@', 2060,
        '[', 2035,
        '\\', 2062,
        ']', 2036,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2045,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2052,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2058,
        '@', 2060,
        '[', 2035,
        '\\', 2061,
        ']', 2036,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2045,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2051,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2059,
        '@', 2060,
        '[', 2035,
        '\\', 2061,
        ']', 2036,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2045,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2051,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2058,
        '@', 2060,
        '[', 2035,
        '\\', 2062,
        ']', 2036,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2045,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2051,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2058,
        '@', 2060,
        '[', 2035,
        '\\', 2061,
        ']', 2037,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 6:
      if (lookahead == '1') ADVANCE(2010);
      if (lookahead == '3') ADVANCE(9);
      END_STATE();
    case 7:
      if (lookahead == '1') ADVANCE(2024);
      if (lookahead == ';') ADVANCE(2033);
      END_STATE();
    case 8:
      if (lookahead == '1') ADVANCE(234);
      if (lookahead == '2') ADVANCE(2011);
      if (lookahead == '3') ADVANCE(231);
      if (lookahead == '4') ADVANCE(11);
      if (lookahead == '5') ADVANCE(2012);
      if (lookahead == '7') ADVANCE(12);
      END_STATE();
    case 9:
      if (lookahead == '4') ADVANCE(13);
      END_STATE();
    case 10:
      if (lookahead == '4') ADVANCE(13);
      if (lookahead == 'f') ADVANCE(1521);
      END_STATE();
    case 11:
      if (lookahead == '5') ADVANCE(13);
      END_STATE();
    case 12:
      if (lookahead == '8') ADVANCE(13);
      END_STATE();
    case 13:
      if (lookahead == ';') ADVANCE(2033);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        ';', 2033,
        'A', 446,
        'B', 436,
        'E', 141,
        'H', 399,
        'a', 672,
        'b', 437,
        'c', 464,
        'd', 656,
        'e', 138,
        'f', 1114,
        'g', 28,
        'h', 482,
        'j', 607,
        'l', 64,
        'm', 1097,
        'n', 318,
        'o', 407,
        'p', 486,
        'r', 440,
        's', 387,
        't', 91,
        'u', 1631,
        'v', 924,
      );
      END_STATE();
    case 15:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1615);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'q') ADVANCE(1932);
      if (lookahead == 's') ADVANCE(831);
      if (lookahead == 'x') ADVANCE(1156);
      END_STATE();
    case 16:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'A') ADVANCE(1734);
      END_STATE();
    case 17:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'V') ADVANCE(925);
      END_STATE();
    case 18:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      END_STATE();
    case 19:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'D') ADVANCE(1505);
      END_STATE();
    case 20:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'E') ADVANCE(1593);
      END_STATE();
    case 21:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'L') ADVANCE(927);
      END_STATE();
    case 22:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'R') ADVANCE(1178);
      END_STATE();
    case 23:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'U') ADVANCE(1549);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        ';', 2033,
        'C', 1516,
        'D', 1490,
        'E', 1225,
        'G', 1750,
        'H', 1943,
        'L', 899,
        'N', 866,
        'P', 1711,
        'R', 900,
        'S', 1594,
        'T', 1103,
        'V', 939,
      );
      END_STATE();
    case 25:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'C') ADVANCE(425);
      END_STATE();
    case 26:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'D') ADVANCE(1451);
      if (lookahead == 'E') ADVANCE(1593);
      END_STATE();
    case 27:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'D') ADVANCE(300);
      END_STATE();
    case 28:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(13);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        ';', 2033,
        'E', 13,
        'a', 1530,
        'c', 1921,
        'e', 90,
        'i', 1352,
        'n', 317,
        'o', 735,
        's', 1108,
        'u', 1662,
      );
      END_STATE();
    case 30:
      ADVANCE_MAP(
        ';', 2033,
        'E', 13,
        'd', 1451,
        'e', 101,
        'm', 1928,
        'n', 2013,
        'p', 1301,
        'r', 436,
        's', 879,
      );
      END_STATE();
    case 31:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'v') ADVANCE(2026);
      END_STATE();
    case 32:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'e') ADVANCE(186);
      END_STATE();
    case 33:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'i') ADVANCE(734);
      if (lookahead == 'o') ADVANCE(1756);
      if (lookahead == 'p') ADVANCE(1682);
      END_STATE();
    case 34:
      ADVANCE_MAP(
        ';', 2033,
        'E', 158,
        'a', 674,
        'b', 1602,
        'c', 1144,
        'd', 1451,
        'e', 159,
        'f', 1603,
        'g', 141,
        'i', 1337,
        'j', 607,
        'l', 230,
        'n', 318,
        'o', 1533,
        'r', 390,
        's', 642,
        't', 92,
        'v', 924,
      );
      END_STATE();
    case 35:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      END_STATE();
    case 36:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      if (lookahead == 'F') ADVANCE(1950);
      if (lookahead == 'G') ADVANCE(1743);
      if (lookahead == 'L') ADVANCE(825);
      if (lookahead == 'S') ADVANCE(1311);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 37:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      if (lookahead == 'F') ADVANCE(1950);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 38:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      if (lookahead == 'G') ADVANCE(1743);
      if (lookahead == 'L') ADVANCE(825);
      if (lookahead == 'S') ADVANCE(1311);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 39:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      if (lookahead == 'S') ADVANCE(1311);
      END_STATE();
    case 40:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E') ADVANCE(1593);
      if (lookahead == 'S') ADVANCE(1311);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 41:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'G') ADVANCE(13);
      END_STATE();
    case 42:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'H') ADVANCE(13);
      END_STATE();
    case 43:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'I') ADVANCE(1433);
      if (lookahead == 'S') ADVANCE(1914);
      if (lookahead == 'U') ADVANCE(1414);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        ';', 2033,
        'J', 607,
        'a', 673,
        'c', 466,
        'e', 956,
        'f', 1603,
        'l', 118,
        'm', 1096,
        'o', 1381,
        's', 639,
        'T', 13,
        't', 13,
      );
      END_STATE();
    case 45:
      ADVANCE_MAP(
        ';', 2033,
        'J', 607,
        'a', 1353,
        'b', 1602,
        'c', 835,
        'd', 1451,
        'f', 1603,
        'o', 1533,
        'r', 851,
        's', 631,
        'T', 13,
        'g', 13,
        't', 13,
      );
      END_STATE();
    case 46:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'L') ADVANCE(825);
      END_STATE();
    case 47:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'N') ADVANCE(13);
      END_STATE();
    case 48:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'P') ADVANCE(13);
      END_STATE();
    case 49:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'P') ADVANCE(1301);
      END_STATE();
    case 50:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'T') ADVANCE(13);
      END_STATE();
    case 51:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 52:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'Y') ADVANCE(13);
      END_STATE();
    case 53:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(637);
      if (lookahead == 'p') ADVANCE(1267);
      END_STATE();
    case 54:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(988);
      if (lookahead == 'o') ADVANCE(1866);
      END_STATE();
    case 55:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(2028);
      END_STATE();
    case 56:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(631);
      if (lookahead == 'l') ADVANCE(13);
      END_STATE();
    case 57:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1603);
      END_STATE();
    case 58:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(645);
      if (lookahead == 'p') ADVANCE(13);
      END_STATE();
    case 59:
      ADVANCE_MAP(
        ';', 2033,
        'a', 1531,
        'c', 1921,
        'e', 110,
        'i', 1670,
        'n', 317,
        'p', 1526,
        's', 1108,
        'E', 13,
        'y', 13,
      );
      END_STATE();
    case 60:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(166);
      if (lookahead == 's') ADVANCE(1109);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 61:
      ADVANCE_MAP(
        ';', 2033,
        'a', 1530,
        'b', 137,
        'f', 1756,
        'h', 1192,
        'l', 1530,
        'p', 1211,
        's', 1108,
        't', 1211,
        'c', 13,
        'w', 13,
      );
      END_STATE();
    case 62:
      ADVANCE_MAP(
        ';', 2033,
        'a', 703,
        'c', 1086,
        'd', 2022,
        'm', 170,
        's', 1108,
        't', 1974,
        'b', 13,
        'e', 13,
      );
      END_STATE();
    case 63:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(703);
      if (lookahead == 'i') ADVANCE(734);
      if (lookahead == 'o') ADVANCE(1756);
      if (lookahead == 'p') ADVANCE(1676);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 64:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'c') ADVANCE(1502);
      if (lookahead == 'h') ADVANCE(501);
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 65:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'd') ADVANCE(133);
      if (lookahead == 'i') ADVANCE(990);
      if (lookahead == 'o') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(1266);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 66:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1756);
      END_STATE();
    case 67:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1767);
      if (lookahead == 'c') ADVANCE(1086);
      if (lookahead == 'd') ADVANCE(1458);
      END_STATE();
    case 68:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1211);
      END_STATE();
    case 69:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'e') ADVANCE(774);
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 70:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1216);
      if (lookahead == 'c') ADVANCE(1811);
      if (lookahead == 'g') ADVANCE(13);
      END_STATE();
    case 71:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1356);
      if (lookahead == 'b') ADVANCE(1725);
      if (lookahead == 'c') ADVANCE(429);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 72:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1356);
      if (lookahead == 's') ADVANCE(1266);
      if (lookahead == 'd' ||
          lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 73:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(696);
      END_STATE();
    case 74:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1651);
      if (lookahead == 'c') ADVANCE(1043);
      if (lookahead == 'o') ADVANCE(1690);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 75:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1236);
      END_STATE();
    case 76:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1619);
      if (lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 77:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1565);
      if (lookahead == 'c') ADVANCE(1941);
      if (lookahead == 'e') ADVANCE(1585);
      if (lookahead == 'n') ADVANCE(523);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 78:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1116);
      if (lookahead == 'e') ADVANCE(195);
      END_STATE();
    case 79:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a') ADVANCE(1310);
      if (lookahead == 's') ADVANCE(1211);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 80:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(1451);
      if (lookahead == 'c') ADVANCE(1086);
      if (lookahead == 'f') ADVANCE(179);
      END_STATE();
    case 81:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(13);
      if (lookahead == 'd') ADVANCE(217);
      END_STATE();
    case 82:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(13);
      if (lookahead == 'h') ADVANCE(1803);
      END_STATE();
    case 83:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(57);
      END_STATE();
    case 84:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(57);
      if (lookahead == 'd') ADVANCE(13);
      END_STATE();
    case 85:
      ADVANCE_MAP(
        ';', 2033,
        'b', 137,
        'f', 1756,
        'h', 1192,
        'l', 1530,
        'p', 1211,
        's', 1108,
        't', 1211,
      );
      END_STATE();
    case 86:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(399);
      if (lookahead == 'e') ADVANCE(1585);
      END_STATE();
    case 87:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b') ADVANCE(1678);
      if (lookahead == 'c') ADVANCE(429);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'o') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 88:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(13);
      END_STATE();
    case 89:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(121);
      if (lookahead == 'f') ADVANCE(1419);
      if (lookahead == 'm') ADVANCE(1075);
      if (lookahead == 's') ADVANCE(703);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 90:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(77);
      END_STATE();
    case 91:
      ADVANCE_MAP(
        ';', 2033,
        'c', 616,
        'd', 1451,
        'h', 1652,
        'i', 1350,
        'l', 436,
        'q', 1933,
        'r', 359,
      );
      END_STATE();
    case 92:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(616);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'l') ADVANCE(358);
      if (lookahead == 'q') ADVANCE(1933);
      if (lookahead == 'r') ADVANCE(424);
      END_STATE();
    case 93:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(120);
      END_STATE();
    case 94:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(615);
      if (lookahead == 'd') ADVANCE(1495);
      if (lookahead == 'l') ADVANCE(128);
      END_STATE();
    case 95:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(615);
      if (lookahead == 'd') ADVANCE(1496);
      if (lookahead == 'g') ADVANCE(128);
      if (lookahead == 's') ADVANCE(522);
      END_STATE();
    case 96:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(453);
      if (lookahead == 'e') ADVANCE(1021);
      if (lookahead == 'l') ADVANCE(508);
      if (lookahead == 'p') ADVANCE(1663);
      END_STATE();
    case 97:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(1086);
      if (lookahead == 'w') ADVANCE(13);
      END_STATE();
    case 98:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(1921);
      if (lookahead == 'e') ADVANCE(93);
      END_STATE();
    case 99:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(1921);
      if (lookahead == 'e' ||
          lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 100:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c') ADVANCE(543);
      if (lookahead == 'f') ADVANCE(1131);
      if (lookahead == 'o') ADVANCE(729);
      if (lookahead == 't') ADVANCE(96);
      END_STATE();
    case 101:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1451);
      END_STATE();
    case 102:
      ADVANCE_MAP(
        ';', 2033,
        'd', 1451,
        'e', 186,
        'g', 28,
        'l', 28,
        'n', 779,
        'p', 1301,
        'r', 436,
      );
      END_STATE();
    case 103:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 's') ADVANCE(219);
      if (lookahead == 'E' ||
          lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 104:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(13);
      END_STATE();
    case 105:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(779);
      END_STATE();
    case 106:
      ADVANCE_MAP(
        ';', 2033,
        'd', 1452,
        'e', 101,
        'h', 1769,
        'l', 436,
        'm', 1928,
        'n', 2013,
        'p', 1301,
        's', 879,
      );
      if (('1' <= lookahead && lookahead <= '3') ||
          lookahead == 'E') ADVANCE(13);
      END_STATE();
    case 107:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1444);
      if (lookahead == 'l') ADVANCE(839);
      if (lookahead == 'r') ADVANCE(1139);
      END_STATE();
    case 108:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1444);
      if (lookahead == 'l') ADVANCE(839);
      if (lookahead == 'u') ADVANCE(1530);
      END_STATE();
    case 109:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1444);
      if (lookahead == 'l') ADVANCE(932);
      if (lookahead == 'q') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(1182);
      END_STATE();
    case 110:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'd') ADVANCE(1116);
      END_STATE();
    case 111:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 112:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(779);
      if (lookahead == 'm') ADVANCE(1786);
      if (lookahead == 'r') ADVANCE(1840);
      if (lookahead == 's') ADVANCE(1553);
      if (lookahead == 'z') ADVANCE(436);
      END_STATE();
    case 113:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 's') ADVANCE(867);
      END_STATE();
    case 114:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1603);
      END_STATE();
    case 115:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1586);
      if (lookahead == 'm') ADVANCE(1168);
      if (lookahead == 'p') ADVANCE(1301);
      if (lookahead == 's') ADVANCE(1600);
      END_STATE();
    case 116:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1811);
      END_STATE();
    case 117:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1014);
      END_STATE();
    case 118:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(959);
      END_STATE();
    case 119:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(186);
      END_STATE();
    case 120:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1585);
      END_STATE();
    case 121:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1585);
      if (lookahead == 'l') ADVANCE(800);
      END_STATE();
    case 122:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(195);
      END_STATE();
    case 123:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(705);
      if (lookahead == 'i') ADVANCE(1340);
      if (lookahead == 'o') ADVANCE(758);
      END_STATE();
    case 124:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(182);
      END_STATE();
    case 125:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1590);
      END_STATE();
    case 126:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1590);
      if (lookahead == 'n') ADVANCE(844);
      END_STATE();
    case 127:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(452);
      END_STATE();
    case 128:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1756);
      END_STATE();
    case 129:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1787);
      END_STATE();
    case 130:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1411);
      if (lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 131:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1688);
      if (lookahead == 's') ADVANCE(815);
      END_STATE();
    case 132:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1644);
      if (lookahead == 's') ADVANCE(1530);
      END_STATE();
    case 133:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1654);
      if (lookahead == 'f' ||
          lookahead == 'm') ADVANCE(13);
      END_STATE();
    case 134:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e') ADVANCE(1643);
      END_STATE();
    case 135:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 136:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(1440);
      if (lookahead == 'y') ADVANCE(196);
      END_STATE();
    case 137:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(1756);
      END_STATE();
    case 138:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(1816);
      if (lookahead == 'g') ADVANCE(13);
      if (lookahead == 'q') ADVANCE(187);
      if (lookahead == 's') ADVANCE(95);
      END_STATE();
    case 139:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(1357);
      if (lookahead == 'l') ADVANCE(801);
      END_STATE();
    case 140:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f') ADVANCE(1842);
      if (lookahead == 'q') ADVANCE(187);
      if (lookahead == 's') ADVANCE(195);
      END_STATE();
    case 141:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'g') ADVANCE(13);
      END_STATE();
    case 142:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'g') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(1832);
      if (lookahead == 'm') ADVANCE(1564);
      END_STATE();
    case 143:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'g') ADVANCE(1811);
      END_STATE();
    case 144:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'g') ADVANCE(779);
      END_STATE();
    case 145:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'h') ADVANCE(13);
      END_STATE();
    case 146:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'h') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(1482);
      END_STATE();
    case 147:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'h') ADVANCE(1441);
      END_STATE();
    case 148:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1089);
      if (lookahead == 'n') ADVANCE(971);
      if (lookahead == 'o') ADVANCE(1832);
      END_STATE();
    case 149:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(753);
      if (lookahead == 'o') ADVANCE(1374);
      END_STATE();
    case 150:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1633);
      if (lookahead == 'u') ADVANCE(1848);
      if (lookahead == 'E' ||
          lookahead == 'd' ||
          lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 151:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1633);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 152:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1391);
      if (lookahead == 'p') ADVANCE(505);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 153:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1870);
      END_STATE();
    case 154:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(718);
      END_STATE();
    case 155:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1366);
      if (lookahead == 'n') ADVANCE(1099);
      END_STATE();
    case 156:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1293);
      END_STATE();
    case 157:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'i') ADVANCE(1430);
      if (lookahead == 'l') ADVANCE(13);
      if (lookahead == 's') ADVANCE(101);
      END_STATE();
    case 158:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(13);
      END_STATE();
    case 159:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(13);
      if (lookahead == 'q') ADVANCE(187);
      if (lookahead == 's') ADVANCE(94);
      END_STATE();
    case 160:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1869);
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 161:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1811);
      END_STATE();
    case 162:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1482);
      END_STATE();
    case 163:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1482);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 164:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(779);
      if (lookahead == 'd' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 165:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1873);
      if (lookahead == 'm') ADVANCE(425);
      END_STATE();
    case 166:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'l') ADVANCE(1252);
      END_STATE();
    case 167:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'm') ADVANCE(13);
      END_STATE();
    case 168:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'm') ADVANCE(132);
      END_STATE();
    case 169:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'm') ADVANCE(459);
      END_STATE();
    case 170:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'n') ADVANCE(13);
      END_STATE();
    case 171:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'n') ADVANCE(478);
      END_STATE();
    case 172:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(13);
      END_STATE();
    case 173:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(191);
      END_STATE();
    case 174:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(158);
      END_STATE();
    case 175:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(949);
      END_STATE();
    case 176:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1980);
      END_STATE();
    case 177:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1969);
      END_STATE();
    case 178:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(703);
      END_STATE();
    case 179:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1657);
      END_STATE();
    case 180:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1886);
      END_STATE();
    case 181:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1396);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 182:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'o') ADVANCE(1421);
      END_STATE();
    case 183:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'p') ADVANCE(13);
      END_STATE();
    case 184:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(155);
      END_STATE();
    case 185:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'p') ADVANCE(1682);
      END_STATE();
    case 186:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'q') ADVANCE(13);
      END_STATE();
    case 187:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'q') ADVANCE(13);
      if (lookahead == 's') ADVANCE(1296);
      END_STATE();
    case 188:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'q') ADVANCE(187);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 189:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'q') ADVANCE(186);
      END_STATE();
    case 190:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'q') ADVANCE(1932);
      END_STATE();
    case 191:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 192:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'r') ADVANCE(417);
      if (lookahead == 's') ADVANCE(101);
      END_STATE();
    case 193:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'r') ADVANCE(1124);
      END_STATE();
    case 194:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'r') ADVANCE(1102);
      END_STATE();
    case 195:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 196:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(1603);
      END_STATE();
    case 197:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(104);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 198:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(84);
      END_STATE();
    case 199:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(2002);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 200:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(1316);
      END_STATE();
    case 201:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(1211);
      END_STATE();
    case 202:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(1874);
      END_STATE();
    case 203:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(815);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 204:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(869);
      END_STATE();
    case 205:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(882);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 206:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 's') ADVANCE(1936);
      END_STATE();
    case 207:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 208:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 209:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(1441);
      END_STATE();
    case 210:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(134);
      END_STATE();
    case 211:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(587);
      END_STATE();
    case 212:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(1211);
      END_STATE();
    case 213:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(685);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 214:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 215:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(1465);
      END_STATE();
    case 216:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 't') ADVANCE(1124);
      END_STATE();
    case 217:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 218:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'u') ADVANCE(1094);
      END_STATE();
    case 219:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 220:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'v') ADVANCE(2026);
      END_STATE();
    case 221:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'v') ADVANCE(576);
      END_STATE();
    case 222:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'v') ADVANCE(881);
      END_STATE();
    case 223:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'a' ||
          lookahead == 'h') ADVANCE(13);
      END_STATE();
    case 224:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'b' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 225:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'c' ||
          lookahead == 'w') ADVANCE(13);
      END_STATE();
    case 226:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e' ||
          lookahead == 'g') ADVANCE(13);
      END_STATE();
    case 227:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e' ||
          lookahead == 'l') ADVANCE(13);
      END_STATE();
    case 228:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'f' ||
          lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 229:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 230:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'E' ||
          lookahead == 'a' ||
          lookahead == 'j') ADVANCE(13);
      END_STATE();
    case 231:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == '4' ||
          lookahead == '5' ||
          lookahead == '8') ADVANCE(13);
      END_STATE();
    case 232:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'D' ||
          lookahead == 'U' ||
          lookahead == 'd' ||
          lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 233:
      if (lookahead == ';') ADVANCE(2033);
      if (lookahead == 'H' ||
          lookahead == 'L' ||
          lookahead == 'R' ||
          lookahead == 'h' ||
          lookahead == 'l' ||
          lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 234:
      if (lookahead == ';') ADVANCE(2033);
      if (('2' <= lookahead && lookahead <= '6') ||
          lookahead == '8') ADVANCE(13);
      END_STATE();
    case 235:
      if (lookahead == ';') ADVANCE(2034);
      END_STATE();
    case 236:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(237);
      END_STATE();
    case 237:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(235);
      END_STATE();
    case 238:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 239:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(238);
      END_STATE();
    case 240:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(239);
      END_STATE();
    case 241:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(240);
      END_STATE();
    case 242:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(235);
      END_STATE();
    case 243:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      END_STATE();
    case 244:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(243);
      END_STATE();
    case 245:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 246:
      if (lookahead == ';') ADVANCE(2034);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(245);
      END_STATE();
    case 247:
      if (lookahead == '>') ADVANCE(2072);
      END_STATE();
    case 248:
      if (lookahead == '>') ADVANCE(2074);
      END_STATE();
    case 249:
      ADVANCE_MAP(
        'A', 363,
        'a', 678,
        'c', 1142,
        'f', 1603,
        'i', 1231,
        'o', 1536,
        's', 652,
        'u', 1329,
      );
      END_STATE();
    case 250:
      ADVANCE_MAP(
        'A', 446,
        'B', 436,
        'H', 399,
        'a', 625,
        'b', 437,
        'c', 464,
        'd', 655,
        'e', 70,
        'f', 1114,
        'h', 483,
        'i', 1004,
        'l', 442,
        'm', 1514,
        'n', 1330,
        'o', 408,
        'p', 490,
        'r', 436,
        's', 388,
        't', 1070,
        'u', 1256,
        'x', 13,
      );
      END_STATE();
    case 251:
      ADVANCE_MAP(
        'A', 607,
        'I', 607,
        'U', 607,
        'a', 665,
        'c', 1144,
        'f', 1603,
        'o', 1533,
        's', 631,
        'u', 1338,
      );
      END_STATE();
    case 252:
      if (lookahead == 'A') ADVANCE(303);
      END_STATE();
    case 253:
      ADVANCE_MAP(
        'A', 1428,
        'C', 941,
        'D', 1446,
        'F', 1255,
        'R', 1177,
        'T', 886,
        'U', 1544,
        'V', 920,
        'a', 1734,
        'r', 1170,
      );
      END_STATE();
    case 254:
      ADVANCE_MAP(
        'A', 1622,
        'B', 493,
        'D', 432,
        'a', 1388,
        'c', 1992,
        'd', 432,
        'e', 792,
        'f', 1603,
        'l', 1869,
        'n', 1785,
        'o', 1533,
        'p', 1672,
        'r', 1869,
        's', 654,
        'z', 1117,
      );
      END_STATE();
    case 255:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'd') ADVANCE(1603);
      if (lookahead == 'e') ADVANCE(140);
      if (lookahead == 's') ADVANCE(1108);
      if (lookahead == 't') ADVANCE(194);
      END_STATE();
    case 256:
      ADVANCE_MAP(
        'A', 1622,
        'H', 399,
        'a', 987,
        'b', 1208,
        'c', 469,
        'd', 54,
        'e', 142,
        'f', 1115,
        'h', 476,
        'i', 447,
        'j', 607,
        'l', 621,
        'o', 1259,
        'r', 583,
        's', 623,
        't', 733,
        'u', 441,
        'w', 451,
        'z', 612,
      );
      END_STATE();
    case 257:
      ADVANCE_MAP(
        'A', 1622,
        'H', 399,
        'a', 666,
        'b', 1625,
        'c', 1123,
        'd', 438,
        'f', 1115,
        'g', 1664,
        'h', 477,
        'l', 714,
        'm', 56,
        'o', 991,
        'p', 548,
        'r', 713,
        's', 631,
        't', 730,
        'u', 443,
        'w', 451,
      );
      END_STATE();
    case 258:
      ADVANCE_MAP(
        'A', 1622,
        'a', 1130,
        'b', 399,
        'c', 1142,
        'e', 530,
        'f', 1603,
        'k', 1757,
        'o', 444,
        's', 644,
        'y', 585,
      );
      END_STATE();
    case 259:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1622);
      END_STATE();
    case 260:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'p') ADVANCE(399);
      END_STATE();
    case 261:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1615);
      if (lookahead == 'c') ADVANCE(207);
      if (lookahead == 'm') ADVANCE(1076);
      if (lookahead == 's') ADVANCE(1975);
      if (lookahead == 't') ADVANCE(1324);
      if (lookahead == 'x') ADVANCE(1811);
      END_STATE();
    case 262:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1615);
      if (lookahead == 'n') ADVANCE(1975);
      END_STATE();
    case 263:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1615);
      if (lookahead == 'n') ADVANCE(830);
      END_STATE();
    case 264:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'a') ADVANCE(1714);
      if (lookahead == 'i') ADVANCE(1007);
      if (lookahead == 't') ADVANCE(1697);
      END_STATE();
    case 265:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 't') ADVANCE(193);
      END_STATE();
    case 266:
      if (lookahead == 'A') ADVANCE(1622);
      if (lookahead == 't') ADVANCE(1692);
      END_STATE();
    case 267:
      ADVANCE_MAP(
        'A', 1737,
        'D', 1505,
        'E', 1596,
        'T', 871,
        'a', 1734,
        'd', 1512,
        'p', 880,
        's', 1093,
      );
      END_STATE();
    case 268:
      if (lookahead == 'A') ADVANCE(671);
      END_STATE();
    case 269:
      if (lookahead == 'A') ADVANCE(671);
      if (lookahead == 'D') ADVANCE(1449);
      if (lookahead == 'G') ADVANCE(1655);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 270:
      if (lookahead == 'A') ADVANCE(1253);
      END_STATE();
    case 271:
      ADVANCE_MAP(
        'A', 1429,
        'C', 941,
        'D', 1446,
        'F', 1255,
        'T', 886,
        'U', 1544,
        'V', 920,
        'a', 1734,
      );
      END_STATE();
    case 272:
      if (lookahead == 'A') ADVANCE(1734);
      END_STATE();
    case 273:
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'D') ADVANCE(1505);
      END_STATE();
    case 274:
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'R') ADVANCE(1178);
      END_STATE();
    case 275:
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'R') ADVANCE(1178);
      if (lookahead == 'T') ADVANCE(856);
      END_STATE();
    case 276:
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'T') ADVANCE(856);
      END_STATE();
    case 277:
      if (lookahead == 'A') ADVANCE(1734);
      if (lookahead == 'V') ADVANCE(925);
      END_STATE();
    case 278:
      if (lookahead == 'A') ADVANCE(1738);
      if (lookahead == 'B') ADVANCE(1602);
      if (lookahead == 'L') ADVANCE(918);
      if (lookahead == 'R') ADVANCE(1176);
      if (lookahead == 'T') ADVANCE(871);
      if (lookahead == 'a') ADVANCE(1734);
      END_STATE();
    case 279:
      if (lookahead == 'B') ADVANCE(402);
      if (lookahead == 'P') ADVANCE(551);
      END_STATE();
    case 280:
      ADVANCE_MAP(
        'B', 436,
        'E', 41,
        'a', 675,
        'c', 466,
        'e', 222,
        'f', 1603,
        'h', 1441,
        'i', 985,
        'o', 1539,
        'r', 1170,
        's', 638,
        'u', 1246,
      );
      END_STATE();
    case 281:
      if (lookahead == 'B') ADVANCE(399);
      END_STATE();
    case 282:
      if (lookahead == 'B') ADVANCE(399);
      if (lookahead == 'L') ADVANCE(1125);
      if (lookahead == 'S') ADVANCE(913);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 283:
      if (lookahead == 'B') ADVANCE(1735);
      END_STATE();
    case 284:
      if (lookahead == 'B') ADVANCE(1727);
      if (lookahead == 'n') ADVANCE(285);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(24);
      END_STATE();
    case 285:
      if (lookahead == 'B') ADVANCE(1732);
      END_STATE();
    case 286:
      if (lookahead == 'C') ADVANCE(336);
      if (lookahead == 'c') ADVANCE(1992);
      END_STATE();
    case 287:
      if (lookahead == 'C') ADVANCE(425);
      END_STATE();
    case 288:
      if (lookahead == 'C') ADVANCE(1506);
      if (lookahead == 'T') ADVANCE(1157);
      END_STATE();
    case 289:
      if (lookahead == 'C') ADVANCE(1183);
      END_STATE();
    case 290:
      if (lookahead == 'C') ADVANCE(1278);
      END_STATE();
    case 291:
      if (lookahead == 'C') ADVANCE(1940);
      END_STATE();
    case 292:
      if (lookahead == 'C') ADVANCE(1522);
      END_STATE();
    case 293:
      if (lookahead == 'C') ADVANCE(1522);
      if (lookahead == 'D') ADVANCE(1448);
      if (lookahead == 'L') ADVANCE(930);
      if (lookahead == 'R') ADVANCE(1180);
      if (lookahead == 'U') ADVANCE(1545);
      if (lookahead == 'V') ADVANCE(939);
      END_STATE();
    case 294:
      if (lookahead == 'D') ADVANCE(1451);
      END_STATE();
    case 295:
      if (lookahead == 'D') ADVANCE(1451);
      if (lookahead == 'M') ADVANCE(1168);
      if (lookahead == 'P') ADVANCE(1301);
      if (lookahead == 'T') ADVANCE(1157);
      END_STATE();
    case 296:
      if (lookahead == 'D') ADVANCE(1451);
      if (lookahead == 'a') ADVANCE(1622);
      END_STATE();
    case 297:
      if (lookahead == 'D') ADVANCE(1451);
      if (lookahead == 'o') ADVANCE(1811);
      END_STATE();
    case 298:
      if (lookahead == 'D') ADVANCE(1451);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 299:
      ADVANCE_MAP(
        'D', 180,
        'J', 607,
        'S', 607,
        'Z', 607,
        'a', 989,
        'c', 469,
        'e', 1212,
        'f', 1603,
        'i', 411,
        'o', 1537,
        's', 652,
      );
      END_STATE();
    case 300:
      if (lookahead == 'D') ADVANCE(13);
      END_STATE();
    case 301:
      ADVANCE_MAP(
        'D', 297,
        'a', 669,
        'c', 467,
        'd', 1451,
        'e', 13,
        'f', 298,
        'g', 192,
        'l', 157,
        'm', 398,
        'n', 974,
        'o', 991,
        'p', 480,
        'q', 662,
        'r', 296,
        's', 632,
        't', 223,
        'u', 1318,
        'x', 690,
      );
      END_STATE();
    case 302:
      ADVANCE_MAP(
        'D', 294,
        'a', 617,
        'c', 1507,
        'd', 432,
        'e', 515,
        'f', 1603,
        'h', 1441,
        'i', 706,
        'l', 670,
        'n', 1561,
        'o', 770,
        'p', 13,
        's', 651,
        'u', 165,
      );
      END_STATE();
    case 303:
      if (lookahead == 'D') ADVANCE(316);
      END_STATE();
    case 304:
      ADVANCE_MAP(
        'D', 2027,
        'H', 232,
        'U', 2027,
        'V', 233,
        'b', 1456,
        'd', 2027,
        'h', 232,
        'm', 1168,
        'p', 1301,
        't', 1157,
        'u', 2027,
        'v', 233,
      );
      END_STATE();
    case 305:
      if (lookahead == 'D') ADVANCE(607);
      END_STATE();
    case 306:
      ADVANCE_MAP(
        'D', 432,
        'H', 436,
        'a', 1530,
        'd', 432,
        'g', 2019,
        'i', 1375,
        'l', 265,
        'r', 266,
        's', 1108,
      );
      END_STATE();
    case 307:
      ADVANCE_MAP(
        'D', 432,
        'b', 399,
        'c', 1992,
        'd', 435,
        'e', 782,
        'f', 1603,
        'o', 1533,
        's', 631,
        'v', 750,
      );
      END_STATE();
    case 308:
      if (lookahead == 'D') ADVANCE(432);
      if (lookahead == 'd') ADVANCE(432);
      END_STATE();
    case 309:
      if (lookahead == 'D') ADVANCE(1505);
      if (lookahead == 'L') ADVANCE(927);
      if (lookahead == 'R') ADVANCE(1178);
      if (lookahead == 'U') ADVANCE(1549);
      END_STATE();
    case 310:
      if (lookahead == 'D') ADVANCE(1107);
      END_STATE();
    case 311:
      if (lookahead == 'D') ADVANCE(873);
      END_STATE();
    case 312:
      if (lookahead == 'D') ADVANCE(1513);
      if (lookahead == 'E') ADVANCE(1593);
      END_STATE();
    case 313:
      if (lookahead == 'D') ADVANCE(1517);
      if (lookahead == 'T') ADVANCE(892);
      if (lookahead == 'V') ADVANCE(920);
      END_STATE();
    case 314:
      if (lookahead == 'D') ADVANCE(1527);
      if (lookahead == 'Q') ADVANCE(1951);
      END_STATE();
    case 315:
      ADVANCE_MAP(
        'E', 1251,
        'M', 48,
        'a', 665,
        'b', 1602,
        'c', 1123,
        'f', 1603,
        'g', 1664,
        'l', 1532,
        'm', 394,
        'n', 734,
        'o', 991,
        'p', 1557,
        'r', 1098,
        's', 649,
        't', 1145,
        'u', 1317,
      );
      END_STATE();
    case 316:
      if (lookahead == 'E') ADVANCE(13);
      END_STATE();
    case 317:
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'a') ADVANCE(1530);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 318:
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'a') ADVANCE(1548);
      if (lookahead == 'e') ADVANCE(189);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 319:
      if (lookahead == 'E') ADVANCE(13);
      if (lookahead == 'e') ADVANCE(188);
      if (lookahead == 's') ADVANCE(1108);
      if (lookahead == 't') ADVANCE(191);
      END_STATE();
    case 320:
      if (lookahead == 'E') ADVANCE(1224);
      if (lookahead == 'U') ADVANCE(1580);
      END_STATE();
    case 321:
      ADVANCE_MAP(
        'E', 607,
        'J', 1257,
        'O', 607,
        'a', 665,
        'c', 1123,
        'd', 1451,
        'f', 1603,
        'g', 1664,
        'm', 53,
        'n', 1839,
        'o', 992,
        's', 631,
        't', 1171,
        'u', 1200,
      );
      END_STATE();
    case 322:
      if (lookahead == 'E') ADVANCE(1223);
      END_STATE();
    case 323:
      ADVANCE_MAP(
        'E', 1257,
        'a', 665,
        'c', 1123,
        'd', 605,
        'f', 1603,
        'g', 1664,
        'm', 396,
        'o', 1533,
        'p', 843,
        'r', 13,
        's', 646,
        't', 1092,
        'u', 1317,
        'v', 858,
      );
      END_STATE();
    case 324:
      if (lookahead == 'E') ADVANCE(1596);
      END_STATE();
    case 325:
      if (lookahead == 'E') ADVANCE(1593);
      END_STATE();
    case 326:
      if (lookahead == 'E') ADVANCE(1598);
      if (lookahead == 'F') ADVANCE(1950);
      if (lookahead == 'G') ADVANCE(1743);
      if (lookahead == 'L') ADVANCE(825);
      if (lookahead == 'S') ADVANCE(1311);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 327:
      if (lookahead == 'E') ADVANCE(1599);
      if (lookahead == 'F') ADVANCE(1950);
      if (lookahead == 'G') ADVANCE(1743);
      if (lookahead == 'L') ADVANCE(825);
      if (lookahead == 'S') ADVANCE(1311);
      if (lookahead == 'T') ADVANCE(1171);
      END_STATE();
    case 328:
      if (lookahead == 'F') ADVANCE(371);
      END_STATE();
    case 329:
      if (lookahead == 'F') ADVANCE(1944);
      END_STATE();
    case 330:
      if (lookahead == 'G') ADVANCE(13);
      END_STATE();
    case 331:
      ADVANCE_MAP(
        'G', 975,
        'L', 906,
        'R', 1170,
        'V', 308,
        'a', 589,
        'b', 1758,
        'c', 419,
        'd', 432,
        'e', 15,
        'f', 1603,
        'g', 319,
        'h', 260,
        'i', 197,
        'j', 607,
        'l', 255,
        'm', 1075,
        'o', 184,
        'p', 488,
        'r', 264,
        's', 624,
        't', 995,
        'u', 168,
        'v', 306,
        'w', 263,
      );
      END_STATE();
    case 332:
      if (lookahead == 'G') ADVANCE(1743);
      END_STATE();
    case 333:
      if (lookahead == 'G') ADVANCE(1752);
      if (lookahead == 'L') ADVANCE(908);
      END_STATE();
    case 334:
      ADVANCE_MAP(
        'H', 286,
        'O', 328,
        'a', 671,
        'c', 69,
        'f', 1603,
        'h', 1508,
        'i', 993,
        'm', 513,
        'o', 1533,
        'q', 1629,
        's', 631,
        't', 399,
        'u', 573,
      );
      END_STATE();
    case 335:
      ADVANCE_MAP(
        'H', 356,
        'R', 252,
        'S', 340,
        'a', 2016,
        'c', 466,
        'f', 1603,
        'h', 814,
        'i', 1240,
        'o', 1533,
        'r', 1150,
        's', 652,
      );
      END_STATE();
    case 336:
      if (lookahead == 'H') ADVANCE(607);
      END_STATE();
    case 337:
      if (lookahead == 'H') ADVANCE(607);
      if (lookahead == 'J') ADVANCE(607);
      if (lookahead == 'a') ADVANCE(1554);
      if (lookahead == 'c') ADVANCE(836);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 338:
      ADVANCE_MAP(
        'H', 607,
        'O', 357,
        'a', 677,
        'c', 465,
        'd', 1451,
        'e', 767,
        'f', 1603,
        'h', 1076,
        'i', 1650,
        'l', 1455,
        'o', 1295,
        'r', 1457,
        's', 631,
        'u', 1540,
      );
      END_STATE();
    case 339:
      ADVANCE_MAP(
        'H', 607,
        'a', 671,
        'c', 469,
        'd', 1451,
        'e', 1660,
        'f', 1603,
        'o', 1533,
        's', 631,
      );
      END_STATE();
    case 340:
      if (lookahead == 'H') ADVANCE(607);
      if (lookahead == 'c') ADVANCE(1992);
      END_STATE();
    case 341:
      if (lookahead == 'H') ADVANCE(1952);
      END_STATE();
    case 342:
      if (lookahead == 'I') ADVANCE(13);
      END_STATE();
    case 343:
      if (lookahead == 'I') ADVANCE(1351);
      END_STATE();
    case 344:
      if (lookahead == 'I') ADVANCE(1418);
      END_STATE();
    case 345:
      ADVANCE_MAP(
        'J', 607,
        'a', 671,
        'c', 466,
        'e', 996,
        'f', 1603,
        'o', 284,
        's', 631,
        't', 1145,
        'u', 13,
      );
      END_STATE();
    case 346:
      if (lookahead == 'L') ADVANCE(825);
      END_STATE();
    case 347:
      if (lookahead == 'L') ADVANCE(1125);
      END_STATE();
    case 348:
      if (lookahead == 'L') ADVANCE(923);
      if (lookahead == 'R') ADVANCE(1178);
      END_STATE();
    case 349:
      if (lookahead == 'L') ADVANCE(923);
      if (lookahead == 'R') ADVANCE(1178);
      if (lookahead == 'l') ADVANCE(905);
      if (lookahead == 'r') ADVANCE(1170);
      END_STATE();
    case 350:
      if (lookahead == 'L') ADVANCE(927);
      if (lookahead == 'R') ADVANCE(1178);
      END_STATE();
    case 351:
      if (lookahead == 'M') ADVANCE(909);
      if (lookahead == 'T') ADVANCE(1058);
      if (lookahead == 'V') ADVANCE(897);
      END_STATE();
    case 352:
      if (lookahead == 'M') ADVANCE(1168);
      END_STATE();
    case 353:
      ADVANCE_MAP(
        'N', 1451,
        'a', 661,
        'b', 1646,
        'c', 1472,
        'd', 1592,
        'e', 726,
        'f', 1603,
        'i', 976,
        'k', 531,
        'l', 412,
        'n', 791,
        'o', 1538,
        'p', 1673,
        'r', 789,
        's', 633,
        'u', 1258,
      );
      END_STATE();
    case 354:
      ADVANCE_MAP(
        'N', 330,
        'T', 42,
        'a', 665,
        'c', 468,
        'd', 1451,
        'f', 1603,
        'g', 1664,
        'l', 910,
        'm', 397,
        'o', 991,
        'p', 1807,
        'q', 1904,
        's', 641,
        't', 385,
        'u', 1317,
        'x', 1101,
      );
      END_STATE();
    case 355:
      if (lookahead == 'O') ADVANCE(50);
      END_STATE();
    case 356:
      if (lookahead == 'O') ADVANCE(362);
      END_STATE();
    case 357:
      if (lookahead == 'P') ADVANCE(52);
      END_STATE();
    case 358:
      if (lookahead == 'P') ADVANCE(399);
      END_STATE();
    case 359:
      if (lookahead == 'P') ADVANCE(399);
      if (lookahead == 'i') ADVANCE(229);
      END_STATE();
    case 360:
      if (lookahead == 'P') ADVANCE(1301);
      END_STATE();
    case 361:
      if (lookahead == 'Q') ADVANCE(1951);
      END_STATE();
    case 362:
      if (lookahead == 'R') ADVANCE(47);
      END_STATE();
    case 363:
      if (lookahead == 'R') ADVANCE(305);
      END_STATE();
    case 364:
      if (lookahead == 'R') ADVANCE(1181);
      if (lookahead == 'T') ADVANCE(892);
      if (lookahead == 'V') ADVANCE(920);
      END_STATE();
    case 365:
      ADVANCE_MAP(
        'S', 13,
        'a', 668,
        'c', 1123,
        'd', 433,
        'e', 1257,
        'f', 704,
        'g', 1483,
        'h', 580,
        'i', 1370,
        'l', 439,
        'm', 395,
        'o', 1533,
        'p', 400,
        'r', 65,
        's', 647,
        't', 1140,
        'u', 1317,
        'v', 578,
      );
      END_STATE();
    case 366:
      if (lookahead == 'S') ADVANCE(1355);
      END_STATE();
    case 367:
      if (lookahead == 'S') ADVANCE(1355);
      if (lookahead == 'V') ADVANCE(895);
      END_STATE();
    case 368:
      if (lookahead == 'S') ADVANCE(1573);
      END_STATE();
    case 369:
      if (lookahead == 'S') ADVANCE(1914);
      END_STATE();
    case 370:
      if (lookahead == 'S') ADVANCE(1600);
      END_STATE();
    case 371:
      if (lookahead == 'T') ADVANCE(607);
      END_STATE();
    case 372:
      if (lookahead == 'T') ADVANCE(1072);
      END_STATE();
    case 373:
      if (lookahead == 'T') ADVANCE(1053);
      END_STATE();
    case 374:
      if (lookahead == 'T') ADVANCE(892);
      if (lookahead == 'V') ADVANCE(920);
      END_STATE();
    case 375:
      if (lookahead == 'T') ADVANCE(1754);
      END_STATE();
    case 376:
      if (lookahead == 'U') ADVANCE(355);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 377:
      if (lookahead == 'V') ADVANCE(939);
      END_STATE();
    case 378:
      if (lookahead == 'V') ADVANCE(925);
      END_STATE();
    case 379:
      if (lookahead == 'W') ADVANCE(1127);
      END_STATE();
    case 380:
      if (lookahead == ']') ADVANCE(2077);
      END_STATE();
    case 381:
      if (lookahead == ']') ADVANCE(2076);
      END_STATE();
    case 382:
      ADVANCE_MAP(
        'a', 665,
        'b', 1602,
        'c', 150,
        'e', 1251,
        'f', 191,
        'g', 1664,
        'l', 790,
        'm', 58,
        'n', 738,
        'o', 991,
        'p', 63,
        'r', 1098,
        's', 650,
        't', 1145,
        'u', 1317,
        'w', 686,
      );
      END_STATE();
    case 383:
      ADVANCE_MAP(
        'a', 665,
        'c', 151,
        'e', 614,
        'f', 2020,
        'g', 1664,
        'i', 148,
        'j', 1257,
        'm', 391,
        'n', 100,
        'o', 611,
        'p', 1663,
        'q', 1919,
        's', 643,
        't', 156,
        'u', 1200,
      );
      END_STATE();
    case 384:
      ADVANCE_MAP(
        'a', 657,
        'c', 1992,
        'e', 692,
        'f', 1603,
        'o', 1533,
        'r', 788,
        's', 631,
        'u', 1328,
      );
      END_STATE();
    case 385:
      if (lookahead == 'a') ADVANCE(13);
      END_STATE();
    case 386:
      if (lookahead == 'a') ADVANCE(191);
      END_STATE();
    case 387:
      if (lookahead == 'a') ADVANCE(1592);
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'h') ADVANCE(13);
      if (lookahead == 'i') ADVANCE(1322);
      if (lookahead == 'q') ADVANCE(572);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 388:
      if (lookahead == 'a') ADVANCE(1592);
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'h') ADVANCE(13);
      if (lookahead == 'q') ADVANCE(572);
      END_STATE();
    case 389:
      ADVANCE_MAP(
        'a', 676,
        'c', 414,
        'd', 1451,
        'e', 760,
        'f', 1603,
        'h', 609,
        'i', 1607,
        'l', 1909,
        'o', 1305,
        'r', 445,
        's', 653,
        't', 729,
        'u', 777,
        'w', 686,
        'y', 1248,
      );
      END_STATE();
    case 390:
      if (lookahead == 'a') ADVANCE(1966);
      END_STATE();
    case 391:
      if (lookahead == 'a') ADVANCE(636);
      if (lookahead == 'o') ADVANCE(949);
      if (lookahead == 'p') ADVANCE(778);
      END_STATE();
    case 392:
      if (lookahead == 'a') ADVANCE(1265);
      if (lookahead == 'e') ADVANCE(1578);
      if (lookahead == 'i') ADVANCE(736);
      if (lookahead == 't') ADVANCE(122);
      END_STATE();
    case 393:
      if (lookahead == 'a') ADVANCE(228);
      END_STATE();
    case 394:
      if (lookahead == 'a') ADVANCE(631);
      END_STATE();
    case 395:
      if (lookahead == 'a') ADVANCE(631);
      if (lookahead == 'e') ADVANCE(984);
      if (lookahead == 'i') ADVANCE(698);
      END_STATE();
    case 396:
      if (lookahead == 'a') ADVANCE(631);
      if (lookahead == 'e') ADVANCE(984);
      if (lookahead == 'i') ADVANCE(697);
      END_STATE();
    case 397:
      if (lookahead == 'a') ADVANCE(631);
      if (lookahead == 'p') ADVANCE(1838);
      END_STATE();
    case 398:
      if (lookahead == 'a') ADVANCE(631);
      if (lookahead == 'p') ADVANCE(1847);
      if (lookahead == 's') ADVANCE(1543);
      END_STATE();
    case 399:
      if (lookahead == 'a') ADVANCE(1603);
      END_STATE();
    case 400:
      if (lookahead == 'a') ADVANCE(1603);
      if (lookahead == 'e') ADVANCE(1651);
      if (lookahead == 'l') ADVANCE(1918);
      END_STATE();
    case 401:
      if (lookahead == 'a') ADVANCE(1603);
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(1918);
      END_STATE();
    case 402:
      if (lookahead == 'a') ADVANCE(1603);
      if (lookahead == 'r') ADVANCE(481);
      END_STATE();
    case 403:
      ADVANCE_MAP(
        'a', 667,
        'b', 1625,
        'c', 1123,
        'd', 605,
        'f', 1603,
        'g', 1664,
        'm', 394,
        'n', 748,
        'o', 991,
        'p', 267,
        'r', 1113,
        's', 631,
        't', 1171,
        'u', 1317,
      );
      END_STATE();
    case 404:
      if (lookahead == 'a') ADVANCE(1370);
      END_STATE();
    case 405:
      if (lookahead == 'a') ADVANCE(219);
      END_STATE();
    case 406:
      ADVANCE_MAP(
        'a', 1728,
        'b', 1657,
        'c', 466,
        'd', 1451,
        'e', 1306,
        'f', 1603,
        'h', 820,
        'i', 1294,
        'o', 812,
        'p', 1673,
        'r', 518,
        's', 622,
        'w', 1080,
      );
      END_STATE();
    case 407:
      ADVANCE_MAP(
        'a', 1379,
        'b', 1657,
        'n', 982,
        'o', 1579,
        'p', 401,
        't', 1157,
        'w', 449,
        'z', 130,
      );
      END_STATE();
    case 408:
      if (lookahead == 'a') ADVANCE(1379);
      if (lookahead == 'b') ADVANCE(1657);
      if (lookahead == 'p') ADVANCE(401);
      if (lookahead == 't') ADVANCE(1157);
      END_STATE();
    case 409:
      if (lookahead == 'a') ADVANCE(1811);
      END_STATE();
    case 410:
      if (lookahead == 'a') ADVANCE(1811);
      if (lookahead == 'l') ADVANCE(1105);
      if (lookahead == 't') ADVANCE(1385);
      END_STATE();
    case 411:
      if (lookahead == 'a') ADVANCE(720);
      if (lookahead == 'f') ADVANCE(960);
      END_STATE();
    case 412:
      if (lookahead == 'a') ADVANCE(687);
      if (lookahead == 'k') ADVANCE(6);
      if (lookahead == 'o') ADVANCE(680);
      END_STATE();
    case 413:
      if (lookahead == 'a') ADVANCE(104);
      END_STATE();
    case 414:
      if (lookahead == 'a') ADVANCE(1555);
      if (lookahead == 'e') ADVANCE(759);
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'u') ADVANCE(1563);
      END_STATE();
    case 415:
      if (lookahead == 'a') ADVANCE(771);
      END_STATE();
    case 416:
      if (lookahead == 'a') ADVANCE(199);
      END_STATE();
    case 417:
      if (lookahead == 'a') ADVANCE(1964);
      END_STATE();
    case 418:
      if (lookahead == 'a') ADVANCE(627);
      if (lookahead == 'o') ADVANCE(1980);
      END_STATE();
    case 419:
      if (lookahead == 'a') ADVANCE(1531);
      if (lookahead == 'e') ADVANCE(774);
      if (lookahead == 'o') ADVANCE(1392);
      if (lookahead == 'u') ADVANCE(1530);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 420:
      if (lookahead == 'a') ADVANCE(207);
      END_STATE();
    case 421:
      ADVANCE_MAP(
        'a', 626,
        'c', 1144,
        'e', 170,
        'f', 1603,
        'i', 607,
        'o', 1533,
        's', 631,
        'u', 613,
      );
      END_STATE();
    case 422:
      ADVANCE_MAP(
        'a', 1653,
        'c', 1992,
        'f', 1603,
        'h', 1076,
        'i', 13,
        'l', 1913,
        'o', 1141,
        'r', 123,
        's', 640,
      );
      END_STATE();
    case 423:
      if (lookahead == 'a') ADVANCE(1554);
      END_STATE();
    case 424:
      if (lookahead == 'a') ADVANCE(1566);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'e') ADVANCE(1587);
      if (lookahead == 'l') ADVANCE(825);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 425:
      if (lookahead == 'a') ADVANCE(1530);
      END_STATE();
    case 426:
      ADVANCE_MAP(
        'a', 1530,
        'c', 1992,
        'e', 755,
        'f', 1603,
        'i', 1409,
        'o', 1533,
        's', 631,
        'u', 13,
      );
      END_STATE();
    case 427:
      if (lookahead == 'a') ADVANCE(1530);
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'u') ADVANCE(1530);
      END_STATE();
    case 428:
      if (lookahead == 'a') ADVANCE(1530);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 429:
      if (lookahead == 'a') ADVANCE(1530);
      if (lookahead == 'u') ADVANCE(1530);
      END_STATE();
    case 430:
      if (lookahead == 'a') ADVANCE(973);
      END_STATE();
    case 431:
      ADVANCE_MAP(
        'a', 1313,
        'c', 1992,
        'e', 1343,
        'f', 1138,
        'i', 1257,
        'j', 1257,
        'l', 410,
        'n', 1443,
        'o', 1534,
        'p', 521,
        'r', 418,
        's', 631,
      );
      END_STATE();
    case 432:
      if (lookahead == 'a') ADVANCE(1777);
      END_STATE();
    case 433:
      if (lookahead == 'a') ADVANCE(1777);
      if (lookahead == 'b') ADVANCE(1291);
      if (lookahead == 'i') ADVANCE(1961);
      if (lookahead == 'o') ADVANCE(1811);
      if (lookahead == 's') ADVANCE(1493);
      END_STATE();
    case 434:
      if (lookahead == 'a') ADVANCE(1192);
      END_STATE();
    case 435:
      if (lookahead == 'a') ADVANCE(1780);
      END_STATE();
    case 436:
      if (lookahead == 'a') ADVANCE(1622);
      END_STATE();
    case 437:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'b') ADVANCE(1657);
      if (lookahead == 'r') ADVANCE(471);
      END_STATE();
    case 438:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'b') ADVANCE(1291);
      if (lookahead == 'h') ADVANCE(399);
      END_STATE();
    case 439:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'c') ADVANCE(1087);
      if (lookahead == 'i') ADVANCE(1391);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 440:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'c') ADVANCE(1502);
      if (lookahead == 'h') ADVANCE(502);
      if (lookahead == 'm') ADVANCE(13);
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 441:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'h') ADVANCE(399);
      END_STATE();
    case 442:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'h') ADVANCE(399);
      if (lookahead == 'm') ADVANCE(13);
      END_STATE();
    case 443:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'm') ADVANCE(158);
      END_STATE();
    case 444:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'm') ADVANCE(1853);
      if (lookahead == 'o') ADVANCE(1195);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'r') ADVANCE(578);
      END_STATE();
    case 445:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'o') ADVANCE(1782);
      END_STATE();
    case 446:
      if (lookahead == 'a') ADVANCE(1622);
      if (lookahead == 'r') ADVANCE(1603);
      if (lookahead == 't') ADVANCE(532);
      END_STATE();
    case 447:
      if (lookahead == 'a') ADVANCE(1320);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'g') ADVANCE(514);
      if (lookahead == 's') ADVANCE(1120);
      if (lookahead == 'v') ADVANCE(149);
      END_STATE();
    case 448:
      if (lookahead == 'a') ADVANCE(2003);
      END_STATE();
    case 449:
      if (lookahead == 'a') ADVANCE(1767);
      if (lookahead == 'b') ADVANCE(399);
      END_STATE();
    case 450:
      if (lookahead == 'a') ADVANCE(1767);
      if (lookahead == 'c') ADVANCE(1142);
      if (lookahead == 'd') ADVANCE(432);
      if (lookahead == 'R' ||
          lookahead == 'S') ADVANCE(13);
      END_STATE();
    case 451:
      if (lookahead == 'a') ADVANCE(1390);
      END_STATE();
    case 452:
      if (lookahead == 'a') ADVANCE(1850);
      END_STATE();
    case 453:
      if (lookahead == 'a') ADVANCE(1211);
      END_STATE();
    case 454:
      ADVANCE_MAP(
        'a', 1609,
        'c', 1992,
        'e', 1610,
        'f', 1603,
        'h', 1090,
        'i', 213,
        'l', 470,
        'm', 13,
        'o', 1151,
        'r', 29,
        's', 640,
        'u', 1389,
      );
      END_STATE();
    case 455:
      if (lookahead == 'a') ADVANCE(1844);
      if (lookahead == 'e') ADVANCE(1791);
      if (lookahead == 'o') ADVANCE(207);
      END_STATE();
    case 456:
      if (lookahead == 'a') ADVANCE(1035);
      END_STATE();
    case 457:
      if (lookahead == 'a') ADVANCE(1209);
      END_STATE();
    case 458:
      if (lookahead == 'a') ADVANCE(1234);
      if (lookahead == 'l') ADVANCE(1125);
      if (lookahead == 's') ADVANCE(1938);
      END_STATE();
    case 459:
      if (lookahead == 'a') ADVANCE(1657);
      END_STATE();
    case 460:
      if (lookahead == 'a') ADVANCE(1770);
      END_STATE();
    case 461:
      if (lookahead == 'a') ADVANCE(1357);
      END_STATE();
    case 462:
      if (lookahead == 'a') ADVANCE(615);
      END_STATE();
    case 463:
      if (lookahead == 'a') ADVANCE(1391);
      END_STATE();
    case 464:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'e') ADVANCE(775);
      if (lookahead == 'u') ADVANCE(571);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 465:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'e') ADVANCE(759);
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'o') ADVANCE(1419);
      END_STATE();
    case 466:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'e') ADVANCE(774);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 467:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'i') ADVANCE(1633);
      if (lookahead == 'o') ADVANCE(1241);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 468:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'i') ADVANCE(1633);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 469:
      if (lookahead == 'a') ADVANCE(1656);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 470:
      if (lookahead == 'a') ADVANCE(1364);
      if (lookahead == 'u') ADVANCE(1761);
      END_STATE();
    case 471:
      if (lookahead == 'a') ADVANCE(629);
      if (lookahead == 'k') ADVANCE(783);
      END_STATE();
    case 472:
      if (lookahead == 'a') ADVANCE(1552);
      if (lookahead == 'u') ADVANCE(1552);
      END_STATE();
    case 473:
      if (lookahead == 'a') ADVANCE(740);
      END_STATE();
    case 474:
      if (lookahead == 'a') ADVANCE(1228);
      END_STATE();
    case 475:
      if (lookahead == 'a') ADVANCE(1228);
      if (lookahead == 'i') ADVANCE(1273);
      END_STATE();
    case 476:
      if (lookahead == 'a') ADVANCE(1612);
      END_STATE();
    case 477:
      if (lookahead == 'a') ADVANCE(1612);
      if (lookahead == 'b') ADVANCE(1242);
      END_STATE();
    case 478:
      if (lookahead == 'a') ADVANCE(1247);
      END_STATE();
    case 479:
      if (lookahead == 'a') ADVANCE(1247);
      if (lookahead == 'e') ADVANCE(1767);
      if (lookahead == 'i') ADVANCE(1963);
      END_STATE();
    case 480:
      if (lookahead == 'a') ADVANCE(1647);
      if (lookahead == 'l') ADVANCE(1918);
      if (lookahead == 's') ADVANCE(1083);
      END_STATE();
    case 481:
      if (lookahead == 'a') ADVANCE(630);
      END_STATE();
    case 482:
      if (lookahead == 'a') ADVANCE(1642);
      if (lookahead == 'b') ADVANCE(1242);
      END_STATE();
    case 483:
      if (lookahead == 'a') ADVANCE(1642);
      if (lookahead == 'o') ADVANCE(219);
      END_STATE();
    case 484:
      if (lookahead == 'a') ADVANCE(1574);
      END_STATE();
    case 485:
      if (lookahead == 'a') ADVANCE(715);
      END_STATE();
    case 486:
      if (lookahead == 'a') ADVANCE(1640);
      END_STATE();
    case 487:
      if (lookahead == 'a') ADVANCE(1239);
      END_STATE();
    case 488:
      if (lookahead == 'a') ADVANCE(1632);
      if (lookahead == 'o') ADVANCE(1307);
      if (lookahead == 'r') ADVANCE(98);
      END_STATE();
    case 489:
      if (lookahead == 'a') ADVANCE(1235);
      END_STATE();
    case 490:
      if (lookahead == 'a') ADVANCE(1627);
      if (lookahead == 'p') ADVANCE(1526);
      END_STATE();
    case 491:
      if (lookahead == 'a') ADVANCE(1613);
      if (lookahead == 'r') ADVANCE(562);
      END_STATE();
    case 492:
      if (lookahead == 'a') ADVANCE(1218);
      END_STATE();
    case 493:
      if (lookahead == 'a') ADVANCE(1626);
      END_STATE();
    case 494:
      if (lookahead == 'a') ADVANCE(1226);
      END_STATE();
    case 495:
      if (lookahead == 'a') ADVANCE(1213);
      END_STATE();
    case 496:
      if (lookahead == 'a') ADVANCE(1219);
      END_STATE();
    case 497:
      if (lookahead == 'a') ADVANCE(1214);
      END_STATE();
    case 498:
      if (lookahead == 'a') ADVANCE(1680);
      END_STATE();
    case 499:
      if (lookahead == 'a') ADVANCE(1232);
      END_STATE();
    case 500:
      if (lookahead == 'a') ADVANCE(1220);
      END_STATE();
    case 501:
      if (lookahead == 'a') ADVANCE(1601);
      END_STATE();
    case 502:
      if (lookahead == 'a') ADVANCE(1630);
      END_STATE();
    case 503:
      if (lookahead == 'a') ADVANCE(1606);
      END_STATE();
    case 504:
      if (lookahead == 'a') ADVANCE(1684);
      END_STATE();
    case 505:
      if (lookahead == 'a') ADVANCE(1628);
      END_STATE();
    case 506:
      if (lookahead == 'a') ADVANCE(1868);
      END_STATE();
    case 507:
      if (lookahead == 'a') ADVANCE(1695);
      END_STATE();
    case 508:
      if (lookahead == 'a') ADVANCE(1668);
      END_STATE();
    case 509:
      if (lookahead == 'a') ADVANCE(1659);
      END_STATE();
    case 510:
      if (lookahead == 'a') ADVANCE(1665);
      END_STATE();
    case 511:
      ADVANCE_MAP(
        'a', 671,
        'b', 1592,
        'c', 59,
        'd', 1480,
        'e', 261,
        'f', 1623,
        'h', 74,
        'i', 1012,
        'l', 436,
        'm', 392,
        'o', 957,
        'p', 415,
        'q', 659,
        'r', 436,
        's', 634,
        't', 491,
        'u', 574,
        'w', 262,
        'z', 1251,
      );
      END_STATE();
    case 512:
      ADVANCE_MAP(
        'a', 671,
        'c', 469,
        'd', 1451,
        'e', 916,
        'f', 1603,
        'h', 607,
        'i', 1006,
        'o', 1533,
        's', 631,
        'w', 1191,
      );
      END_STATE();
    case 513:
      if (lookahead == 'a') ADVANCE(1263);
      END_STATE();
    case 514:
      if (lookahead == 'a') ADVANCE(1332);
      END_STATE();
    case 515:
      if (lookahead == 'a') ADVANCE(1789);
      END_STATE();
    case 516:
      ADVANCE_MAP(
        'a', 1558,
        'c', 836,
        'f', 1603,
        'g', 1667,
        'h', 607,
        'j', 607,
        'o', 1533,
        's', 631,
      );
      END_STATE();
    case 517:
      if (lookahead == 'a') ADVANCE(709);
      END_STATE();
    case 518:
      if (lookahead == 'a') ADVANCE(756);
      if (lookahead == 'i') ADVANCE(563);
      if (lookahead == 'p') ADVANCE(818);
      END_STATE();
    case 519:
      if (lookahead == 'a') ADVANCE(1716);
      END_STATE();
    case 520:
      if (lookahead == 'a') ADVANCE(1865);
      END_STATE();
    case 521:
      if (lookahead == 'a') ADVANCE(1683);
      END_STATE();
    case 522:
      if (lookahead == 'a') ADVANCE(1565);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'e') ADVANCE(1588);
      if (lookahead == 'g') ADVANCE(1831);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 523:
      if (lookahead == 'a') ADVANCE(1565);
      if (lookahead == 'e') ADVANCE(1591);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 524:
      if (lookahead == 'a') ADVANCE(1565);
      if (lookahead == 's') ADVANCE(1108);
      END_STATE();
    case 525:
      if (lookahead == 'a') ADVANCE(695);
      END_STATE();
    case 526:
      if (lookahead == 'a') ADVANCE(1717);
      if (lookahead == 'l') ADVANCE(1994);
      if (lookahead == 'r') ADVANCE(787);
      if (lookahead == 'v') ADVANCE(936);
      END_STATE();
    case 527:
      if (lookahead == 'a') ADVANCE(1402);
      END_STATE();
    case 528:
      if (lookahead == 'a') ADVANCE(1253);
      if (lookahead == 'k') ADVANCE(219);
      END_STATE();
    case 529:
      if (lookahead == 'a') ADVANCE(1875);
      END_STATE();
    case 530:
      if (lookahead == 'a') ADVANCE(1729);
      if (lookahead == 'l') ADVANCE(1268);
      if (lookahead == 'r') ADVANCE(679);
      END_STATE();
    case 531:
      if (lookahead == 'a') ADVANCE(1679);
      END_STATE();
    case 532:
      if (lookahead == 'a') ADVANCE(1116);
      END_STATE();
    case 533:
      if (lookahead == 'a') ADVANCE(1116);
      if (lookahead == 'i') ADVANCE(1453);
      END_STATE();
    case 534:
      if (lookahead == 'a') ADVANCE(1929);
      END_STATE();
    case 535:
      if (lookahead == 'a') ADVANCE(1275);
      END_STATE();
    case 536:
      if (lookahead == 'a') ADVANCE(1720);
      END_STATE();
    case 537:
      if (lookahead == 'a') ADVANCE(1931);
      END_STATE();
    case 538:
      if (lookahead == 'a') ADVANCE(1883);
      END_STATE();
    case 539:
      if (lookahead == 'a') ADVANCE(1403);
      END_STATE();
    case 540:
      if (lookahead == 'a') ADVANCE(1722);
      END_STATE();
    case 541:
      if (lookahead == 'a') ADVANCE(1260);
      END_STATE();
    case 542:
      if (lookahead == 'a') ADVANCE(1712);
      END_STATE();
    case 543:
      if (lookahead == 'a') ADVANCE(1671);
      END_STATE();
    case 544:
      if (lookahead == 'a') ADVANCE(1718);
      END_STATE();
    case 545:
      if (lookahead == 'a') ADVANCE(1719);
      END_STATE();
    case 546:
      if (lookahead == 'a') ADVANCE(1310);
      END_STATE();
    case 547:
      if (lookahead == 'a') ADVANCE(1734);
      END_STATE();
    case 548:
      if (lookahead == 'a') ADVANCE(1734);
      if (lookahead == 'd') ADVANCE(1512);
      if (lookahead == 'h') ADVANCE(504);
      if (lookahead == 'l') ADVANCE(1918);
      if (lookahead == 's') ADVANCE(1085);
      if (lookahead == 'u') ADVANCE(1584);
      END_STATE();
    case 549:
      if (lookahead == 'a') ADVANCE(1734);
      if (lookahead == 'd') ADVANCE(1528);
      if (lookahead == 'h') ADVANCE(504);
      END_STATE();
    case 550:
      if (lookahead == 'a') ADVANCE(1734);
      if (lookahead == 'r') ADVANCE(1170);
      END_STATE();
    case 551:
      if (lookahead == 'a') ADVANCE(1745);
      END_STATE();
    case 552:
      if (lookahead == 'a') ADVANCE(1891);
      END_STATE();
    case 553:
      if (lookahead == 'a') ADVANCE(1739);
      if (lookahead == 'h') ADVANCE(564);
      if (lookahead == 'l') ADVANCE(928);
      if (lookahead == 'r') ADVANCE(1185);
      if (lookahead == 's') ADVANCE(1595);
      if (lookahead == 't') ADVANCE(1073);
      END_STATE();
    case 554:
      if (lookahead == 'a') ADVANCE(1739);
      if (lookahead == 'h') ADVANCE(564);
      if (lookahead == 'l') ADVANCE(948);
      if (lookahead == 'r') ADVANCE(1179);
      if (lookahead == 't') ADVANCE(1073);
      END_STATE();
    case 555:
      if (lookahead == 'a') ADVANCE(1893);
      END_STATE();
    case 556:
      if (lookahead == 'a') ADVANCE(1740);
      END_STATE();
    case 557:
      if (lookahead == 'a') ADVANCE(1740);
      if (lookahead == 'd') ADVANCE(450);
      END_STATE();
    case 558:
      if (lookahead == 'a') ADVANCE(1741);
      END_STATE();
    case 559:
      if (lookahead == 'a') ADVANCE(1741);
      if (lookahead == 'h') ADVANCE(566);
      END_STATE();
    case 560:
      if (lookahead == 'a') ADVANCE(1895);
      END_STATE();
    case 561:
      if (lookahead == 'a') ADVANCE(1742);
      if (lookahead == 'h') ADVANCE(566);
      if (lookahead == 's') ADVANCE(1595);
      END_STATE();
    case 562:
      if (lookahead == 'a') ADVANCE(1175);
      if (lookahead == 'n') ADVANCE(1756);
      END_STATE();
    case 563:
      if (lookahead == 'a') ADVANCE(1424);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'm') ADVANCE(1168);
      if (lookahead == 'p') ADVANCE(1301);
      if (lookahead == 's') ADVANCE(571);
      if (lookahead == 't') ADVANCE(1153);
      END_STATE();
    case 564:
      if (lookahead == 'a') ADVANCE(1747);
      END_STATE();
    case 565:
      if (lookahead == 'a') ADVANCE(1431);
      END_STATE();
    case 566:
      if (lookahead == 'a') ADVANCE(1749);
      END_STATE();
    case 567:
      if (lookahead == 'a') ADVANCE(1432);
      END_STATE();
    case 568:
      if (lookahead == 'a') ADVANCE(1434);
      END_STATE();
    case 569:
      if (lookahead == 'a') ADVANCE(1435);
      END_STATE();
    case 570:
      if (lookahead == 'a') ADVANCE(1436);
      END_STATE();
    case 571:
      if (lookahead == 'b') ADVANCE(13);
      END_STATE();
    case 572:
      if (lookahead == 'b') ADVANCE(13);
      if (lookahead == 'u') ADVANCE(1442);
      END_STATE();
    case 573:
      if (lookahead == 'b') ADVANCE(204);
      if (lookahead == 'c') ADVANCE(717);
      if (lookahead == 'm') ADVANCE(13);
      if (lookahead == 'p') ADVANCE(131);
      END_STATE();
    case 574:
      if (lookahead == 'b') ADVANCE(30);
      if (lookahead == 'c') ADVANCE(628);
      if (lookahead == 'm') ADVANCE(13);
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 'p') ADVANCE(106);
      END_STATE();
    case 575:
      if (lookahead == 'b') ADVANCE(205);
      if (lookahead == 'c') ADVANCE(684);
      if (lookahead == 'p') ADVANCE(205);
      END_STATE();
    case 576:
      if (lookahead == 'b') ADVANCE(104);
      END_STATE();
    case 577:
      if (lookahead == 'b') ADVANCE(113);
      if (lookahead == 'p') ADVANCE(113);
      END_STATE();
    case 578:
      if (lookahead == 'b') ADVANCE(399);
      END_STATE();
    case 579:
      if (lookahead == 'b') ADVANCE(399);
      if (lookahead == 'g') ADVANCE(822);
      END_STATE();
    case 580:
      if (lookahead == 'b') ADVANCE(399);
      if (lookahead == 'm') ADVANCE(13);
      END_STATE();
    case 581:
      if (lookahead == 'b') ADVANCE(399);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 582:
      if (lookahead == 'b') ADVANCE(399);
      if (lookahead == 't') ADVANCE(154);
      if (lookahead == 'y') ADVANCE(372);
      END_STATE();
    case 583:
      if (lookahead == 'b') ADVANCE(1207);
      if (lookahead == 'c') ADVANCE(1501);
      END_STATE();
    case 584:
      if (lookahead == 'b') ADVANCE(111);
      if (lookahead == 'p') ADVANCE(111);
      END_STATE();
    case 585:
      if (lookahead == 'b') ADVANCE(1947);
      if (lookahead == 'p') ADVANCE(1068);
      END_STATE();
    case 586:
      if (lookahead == 'b') ADVANCE(744);
      END_STATE();
    case 587:
      if (lookahead == 'b') ADVANCE(1657);
      END_STATE();
    case 588:
      if (lookahead == 'b') ADVANCE(779);
      if (lookahead == 'p') ADVANCE(779);
      END_STATE();
    case 589:
      if (lookahead == 'b') ADVANCE(1221);
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 'p') ADVANCE(33);
      if (lookahead == 't') ADVANCE(1937);
      END_STATE();
    case 590:
      if (lookahead == 'b') ADVANCE(1762);
      END_STATE();
    case 591:
      if (lookahead == 'b') ADVANCE(386);
      END_STATE();
    case 592:
      if (lookahead == 'b') ADVANCE(931);
      END_STATE();
    case 593:
      if (lookahead == 'b') ADVANCE(1365);
      if (lookahead == 'p') ADVANCE(1365);
      END_STATE();
    case 594:
      if (lookahead == 'b') ADVANCE(1806);
      if (lookahead == 'c') ADVANCE(716);
      if (lookahead == 'p') ADVANCE(883);
      END_STATE();
    case 595:
      if (lookahead == 'b') ADVANCE(1806);
      if (lookahead == 'p') ADVANCE(883);
      END_STATE();
    case 596:
      if (lookahead == 'b') ADVANCE(1277);
      END_STATE();
    case 597:
      if (lookahead == 'b') ADVANCE(1280);
      END_STATE();
    case 598:
      if (lookahead == 'b') ADVANCE(1723);
      END_STATE();
    case 599:
      if (lookahead == 'b') ADVANCE(1283);
      END_STATE();
    case 600:
      if (lookahead == 'b') ADVANCE(1284);
      END_STATE();
    case 601:
      if (lookahead == 'b') ADVANCE(1298);
      END_STATE();
    case 602:
      if (lookahead == 'b') ADVANCE(510);
      END_STATE();
    case 603:
      if (lookahead == 'b') ADVANCE(1289);
      END_STATE();
    case 604:
      if (lookahead == 'b') ADVANCE(1290);
      END_STATE();
    case 605:
      if (lookahead == 'b') ADVANCE(1291);
      END_STATE();
    case 606:
      if (lookahead == 'b') ADVANCE(1809);
      if (lookahead == 'p') ADVANCE(1809);
      END_STATE();
    case 607:
      if (lookahead == 'c') ADVANCE(1992);
      END_STATE();
    case 608:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'e') ADVANCE(1966);
      END_STATE();
    case 609:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'e') ADVANCE(689);
      if (lookahead == 'i') ADVANCE(13);
      END_STATE();
    case 610:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(1302);
      if (lookahead == 'o') ADVANCE(1535);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 611:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'g') ADVANCE(1482);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 612:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'i') ADVANCE(1006);
      END_STATE();
    case 613:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'm') ADVANCE(158);
      END_STATE();
    case 614:
      if (lookahead == 'c') ADVANCE(1992);
      if (lookahead == 'x') ADVANCE(618);
      END_STATE();
    case 615:
      if (lookahead == 'c') ADVANCE(13);
      END_STATE();
    case 616:
      if (lookahead == 'c') ADVANCE(13);
      if (lookahead == 'i') ADVANCE(1603);
      END_STATE();
    case 617:
      if (lookahead == 'c') ADVANCE(191);
      if (lookahead == 'l') ADVANCE(784);
      if (lookahead == 'p') ADVANCE(202);
      if (lookahead == 'r') ADVANCE(1204);
      END_STATE();
    case 618:
      if (lookahead == 'c') ADVANCE(158);
      END_STATE();
    case 619:
      ADVANCE_MAP(
        'c', 427,
        'd', 1869,
        'f', 1603,
        'h', 259,
        'i', 13,
        'l', 259,
        'm', 425,
        'n', 1112,
        'o', 731,
        'r', 259,
        's', 648,
        'u', 1562,
        'v', 856,
        'w', 819,
      );
      END_STATE();
    case 620:
      if (lookahead == 'c') ADVANCE(427);
      if (lookahead == 'o') ADVANCE(732);
      if (lookahead == 's') ADVANCE(1597);
      if (lookahead == 't') ADVANCE(1753);
      if (lookahead == 'u') ADVANCE(1561);
      if (lookahead == 'v') ADVANCE(856);
      if (lookahead == 'w') ADVANCE(819);
      END_STATE();
    case 621:
      if (lookahead == 'c') ADVANCE(1501);
      END_STATE();
    case 622:
      if (lookahead == 'c') ADVANCE(2023);
      if (lookahead == 'h') ADVANCE(607);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 623:
      if (lookahead == 'c') ADVANCE(2023);
      if (lookahead == 'o') ADVANCE(1211);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 624:
      if (lookahead == 'c') ADVANCE(99);
      if (lookahead == 'h') ADVANCE(1519);
      if (lookahead == 'i') ADVANCE(1333);
      if (lookahead == 'm') ADVANCE(1075);
      if (lookahead == 'p') ADVANCE(399);
      if (lookahead == 'q') ADVANCE(1790);
      if (lookahead == 'u') ADVANCE(575);
      END_STATE();
    case 625:
      if (lookahead == 'c') ADVANCE(785);
      if (lookahead == 'd') ADVANCE(1122);
      if (lookahead == 'e') ADVANCE(1334);
      if (lookahead == 'n') ADVANCE(983);
      if (lookahead == 'q') ADVANCE(1907);
      if (lookahead == 'r') ADVANCE(1616);
      if (lookahead == 't') ADVANCE(533);
      END_STATE();
    case 626:
      if (lookahead == 'c') ADVANCE(1916);
      END_STATE();
    case 627:
      if (lookahead == 'c') ADVANCE(8);
      if (lookahead == 's') ADVANCE(1211);
      END_STATE();
    case 628:
      if (lookahead == 'c') ADVANCE(77);
      END_STATE();
    case 629:
      if (lookahead == 'c') ADVANCE(2018);
      END_STATE();
    case 630:
      if (lookahead == 'c') ADVANCE(780);
      END_STATE();
    case 631:
      if (lookahead == 'c') ADVANCE(1603);
      END_STATE();
    case 632:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'i') ADVANCE(1316);
      END_STATE();
    case 633:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'e') ADVANCE(1325);
      if (lookahead == 'i') ADVANCE(1326);
      if (lookahead == 'o') ADVANCE(1215);
      END_STATE();
    case 634:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'e') ADVANCE(1859);
      if (lookahead == 'm') ADVANCE(1165);
      if (lookahead == 't') ADVANCE(503);
      END_STATE();
    case 635:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'e') ADVANCE(1636);
      END_STATE();
    case 636:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'g') ADVANCE(781);
      if (lookahead == 't') ADVANCE(1036);
      END_STATE();
    case 637:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'g') ADVANCE(1135);
      END_STATE();
    case 638:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'h') ADVANCE(13);
      END_STATE();
    case 639:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'h') ADVANCE(13);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 640:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(13);
      END_STATE();
    case 641:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(1316);
      END_STATE();
    case 642:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(1321);
      END_STATE();
    case 643:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(1363);
      END_STATE();
    case 644:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'l') ADVANCE(432);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 645:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'l') ADVANCE(973);
      END_STATE();
    case 646:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'l') ADVANCE(460);
      END_STATE();
    case 647:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'l') ADVANCE(460);
      if (lookahead == 'o') ADVANCE(1211);
      END_STATE();
    case 648:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'q') ADVANCE(700);
      END_STATE();
    case 649:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(1118);
      END_STATE();
    case 650:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 't') ADVANCE(13);
      if (lookahead == 'y') ADVANCE(1341);
      END_STATE();
    case 651:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 't') ADVANCE(1571);
      END_STATE();
    case 652:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 't') ADVANCE(1674);
      END_STATE();
    case 653:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'u') ADVANCE(584);
      END_STATE();
    case 654:
      if (lookahead == 'c') ADVANCE(1603);
      if (lookahead == 'u') ADVANCE(593);
      END_STATE();
    case 655:
      if (lookahead == 'c') ADVANCE(385);
      if (lookahead == 'l') ADVANCE(765);
      if (lookahead == 'q') ADVANCE(1927);
      if (lookahead == 's') ADVANCE(1036);
      END_STATE();
    case 656:
      if (lookahead == 'c') ADVANCE(385);
      if (lookahead == 'q') ADVANCE(1927);
      if (lookahead == 'r') ADVANCE(766);
      if (lookahead == 's') ADVANCE(1036);
      END_STATE();
    case 657:
      if (lookahead == 'c') ADVANCE(1203);
      if (lookahead == 'r') ADVANCE(1962);
      END_STATE();
    case 658:
      if (lookahead == 'c') ADVANCE(1370);
      if (lookahead == 'i') ADVANCE(1440);
      if (lookahead == 'm') ADVANCE(1116);
      if (lookahead == 'p') ADVANCE(13);
      if (lookahead == 't') ADVANCE(872);
      END_STATE();
    case 659:
      if (lookahead == 'c') ADVANCE(472);
      if (lookahead == 's') ADVANCE(1908);
      if (lookahead == 'u') ADVANCE(76);
      END_STATE();
    case 660:
      if (lookahead == 'c') ADVANCE(1811);
      END_STATE();
    case 661:
      if (lookahead == 'c') ADVANCE(1194);
      if (lookahead == 'r') ADVANCE(1965);
      END_STATE();
    case 662:
      if (lookahead == 'c') ADVANCE(1143);
      if (lookahead == 's') ADVANCE(1110);
      if (lookahead == 'u') ADVANCE(479);
      if (lookahead == 'v') ADVANCE(1578);
      END_STATE();
    case 663:
      if (lookahead == 'c') ADVANCE(1142);
      if (lookahead == 'e') ADVANCE(739);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 'p') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(127);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 664:
      if (lookahead == 'c') ADVANCE(1142);
      if (lookahead == 'e') ADVANCE(768);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 665:
      if (lookahead == 'c') ADVANCE(1915);
      END_STATE();
    case 666:
      if (lookahead == 'c') ADVANCE(1915);
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 667:
      if (lookahead == 'c') ADVANCE(1915);
      if (lookahead == 'r') ADVANCE(1639);
      END_STATE();
    case 668:
      if (lookahead == 'c') ADVANCE(1915);
      if (lookahead == 's') ADVANCE(1811);
      END_STATE();
    case 669:
      if (lookahead == 'c') ADVANCE(1915);
      if (lookahead == 's') ADVANCE(1865);
      END_STATE();
    case 670:
      if (lookahead == 'c') ADVANCE(1530);
      if (lookahead == 'd') ADVANCE(1603);
      END_STATE();
    case 671:
      if (lookahead == 'c') ADVANCE(1948);
      END_STATE();
    case 672:
      ADVANCE_MAP(
        'c', 1948,
        'e', 1334,
        'g', 1721,
        'm', 586,
        'n', 981,
        'p', 13,
        'q', 1907,
        'r', 1614,
        't', 78,
      );
      END_STATE();
    case 673:
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'm') ADVANCE(586);
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 'p') ADVANCE(1300);
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 674:
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'm') ADVANCE(1344);
      if (lookahead == 'p') ADVANCE(13);
      END_STATE();
    case 675:
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 'r') ADVANCE(1649);
      END_STATE();
    case 676:
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'p') ADVANCE(71);
      if (lookahead == 'r') ADVANCE(816);
      END_STATE();
    case 677:
      if (lookahead == 'c') ADVANCE(1948);
      if (lookahead == 'p') ADVANCE(153);
      if (lookahead == 'y') ADVANCE(1261);
      END_STATE();
    case 678:
      if (lookahead == 'c') ADVANCE(842);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 679:
      if (lookahead == 'c') ADVANCE(1482);
      END_STATE();
    case 680:
      if (lookahead == 'c') ADVANCE(1192);
      END_STATE();
    case 681:
      if (lookahead == 'c') ADVANCE(1201);
      END_STATE();
    case 682:
      if (lookahead == 'c') ADVANCE(1201);
      if (lookahead == 's') ADVANCE(793);
      END_STATE();
    case 683:
      if (lookahead == 'c') ADVANCE(1193);
      if (lookahead == 'n') ADVANCE(368);
      END_STATE();
    case 684:
      if (lookahead == 'c') ADVANCE(120);
      END_STATE();
    case 685:
      if (lookahead == 'c') ADVANCE(1069);
      END_STATE();
    case 686:
      if (lookahead == 'c') ADVANCE(1462);
      if (lookahead == 'i') ADVANCE(1370);
      END_STATE();
    case 687:
      if (lookahead == 'c') ADVANCE(1196);
      if (lookahead == 'n') ADVANCE(1192);
      END_STATE();
    case 688:
      if (lookahead == 'c') ADVANCE(425);
      END_STATE();
    case 689:
      if (lookahead == 'c') ADVANCE(1197);
      END_STATE();
    case 690:
      if (lookahead == 'c') ADVANCE(1211);
      if (lookahead == 'i') ADVANCE(1767);
      if (lookahead == 'p') ADVANCE(911);
      END_STATE();
    case 691:
      if (lookahead == 'c') ADVANCE(1206);
      if (lookahead == 'n') ADVANCE(1776);
      END_STATE();
    case 692:
      if (lookahead == 'c') ADVANCE(534);
      if (lookahead == 'r') ADVANCE(1394);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 693:
      if (lookahead == 'c') ADVANCE(1198);
      if (lookahead == 'k') ADVANCE(1961);
      END_STATE();
    case 694:
      if (lookahead == 'c') ADVANCE(615);
      END_STATE();
    case 695:
      if (lookahead == 'c') ADVANCE(779);
      END_STATE();
    case 696:
      if (lookahead == 'c') ADVANCE(1050);
      END_STATE();
    case 697:
      if (lookahead == 'c') ADVANCE(1656);
      END_STATE();
    case 698:
      if (lookahead == 'c') ADVANCE(1656);
      if (lookahead == 'd') ADVANCE(13);
      if (lookahead == 'n') ADVANCE(1918);
      END_STATE();
    case 699:
      if (lookahead == 'c') ADVANCE(1776);
      END_STATE();
    case 700:
      if (lookahead == 'c') ADVANCE(1911);
      END_STATE();
    case 701:
      if (lookahead == 'c') ADVANCE(1810);
      END_STATE();
    case 702:
      if (lookahead == 'c') ADVANCE(453);
      END_STATE();
    case 703:
      if (lookahead == 'c') ADVANCE(1086);
      END_STATE();
    case 704:
      if (lookahead == 'c') ADVANCE(1086);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 705:
      if (lookahead == 'c') ADVANCE(919);
      END_STATE();
    case 706:
      if (lookahead == 'c') ADVANCE(1645);
      if (lookahead == 'd') ADVANCE(67);
      if (lookahead == 'n') ADVANCE(1926);
      END_STATE();
    case 707:
      if (lookahead == 'c') ADVANCE(1897);
      END_STATE();
    case 708:
      if (lookahead == 'c') ADVANCE(1884);
      END_STATE();
    case 709:
      if (lookahead == 'c') ADVANCE(915);
      END_STATE();
    case 710:
      if (lookahead == 'c') ADVANCE(1868);
      END_STATE();
    case 711:
      if (lookahead == 'c') ADVANCE(1144);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'm') ADVANCE(452);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(635);
      if (lookahead == 'u') ADVANCE(1199);
      END_STATE();
    case 712:
      if (lookahead == 'c') ADVANCE(1144);
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(635);
      if (lookahead == 'u') ADVANCE(1199);
      END_STATE();
    case 713:
      if (lookahead == 'c') ADVANCE(1503);
      if (lookahead == 'i') ADVANCE(1378);
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 714:
      if (lookahead == 'c') ADVANCE(1503);
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 715:
      if (lookahead == 'c') ADVANCE(1205);
      END_STATE();
    case 716:
      if (lookahead == 'c') ADVANCE(885);
      END_STATE();
    case 717:
      if (lookahead == 'c') ADVANCE(885);
      if (lookahead == 'h') ADVANCE(373);
      END_STATE();
    case 718:
      if (lookahead == 'c') ADVANCE(492);
      END_STATE();
    case 719:
      if (lookahead == 'c') ADVANCE(1276);
      END_STATE();
    case 720:
      if (lookahead == 'c') ADVANCE(1700);
      if (lookahead == 'm') ADVANCE(1488);
      END_STATE();
    case 721:
      if (lookahead == 'c') ADVANCE(496);
      END_STATE();
    case 722:
      if (lookahead == 'c') ADVANCE(1883);
      END_STATE();
    case 723:
      if (lookahead == 'c') ADVANCE(1471);
      if (lookahead == 'e') ADVANCE(1529);
      if (lookahead == 'p') ADVANCE(1673);
      if (lookahead == 's') ADVANCE(1128);
      END_STATE();
    case 724:
      if (lookahead == 'c') ADVANCE(1260);
      END_STATE();
    case 725:
      if (lookahead == 'c') ADVANCE(499);
      END_STATE();
    case 726:
      if (lookahead == 'c') ADVANCE(537);
      if (lookahead == 'm') ADVANCE(1564);
      if (lookahead == 'p') ADVANCE(1771);
      if (lookahead == 'r') ADVANCE(1417);
      if (lookahead == 't') ADVANCE(1979);
      END_STATE();
    case 727:
      if (lookahead == 'c') ADVANCE(544);
      END_STATE();
    case 728:
      if (lookahead == 'c') ADVANCE(943);
      END_STATE();
    case 729:
      if (lookahead == 'd') ADVANCE(1451);
      END_STATE();
    case 730:
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'i') ADVANCE(1293);
      if (lookahead == 'r') ADVANCE(1082);
      END_STATE();
    case 731:
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'p') ADVANCE(950);
      if (lookahead == 't') ADVANCE(1153);
      END_STATE();
    case 732:
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'p') ADVANCE(1301);
      if (lookahead == 't') ADVANCE(1157);
      END_STATE();
    case 733:
      if (lookahead == 'd') ADVANCE(1451);
      if (lookahead == 'r') ADVANCE(1082);
      END_STATE();
    case 734:
      if (lookahead == 'd') ADVANCE(13);
      END_STATE();
    case 735:
      if (lookahead == 'd') ADVANCE(13);
      if (lookahead == 'f') ADVANCE(458);
      if (lookahead == 'p') ADVANCE(209);
      END_STATE();
    case 736:
      if (lookahead == 'd') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(779);
      END_STATE();
    case 737:
      if (lookahead == 'd') ADVANCE(13);
      if (lookahead == 'u') ADVANCE(158);
      END_STATE();
    case 738:
      if (lookahead == 'd') ADVANCE(72);
      if (lookahead == 'g') ADVANCE(112);
      END_STATE();
    case 739:
      if (lookahead == 'd') ADVANCE(579);
      if (lookahead == 'i') ADVANCE(874);
      END_STATE();
    case 740:
      if (lookahead == 'd') ADVANCE(1297);
      END_STATE();
    case 741:
      if (lookahead == 'd') ADVANCE(367);
      END_STATE();
    case 742:
      if (lookahead == 'd') ADVANCE(343);
      END_STATE();
    case 743:
      if (lookahead == 'd') ADVANCE(333);
      END_STATE();
    case 744:
      if (lookahead == 'd') ADVANCE(385);
      END_STATE();
    case 745:
      if (lookahead == 'd') ADVANCE(607);
      if (lookahead == 'r') ADVANCE(97);
      END_STATE();
    case 746:
      if (lookahead == 'd') ADVANCE(144);
      END_STATE();
    case 747:
      if (lookahead == 'd') ADVANCE(55);
      END_STATE();
    case 748:
      if (lookahead == 'd') ADVANCE(858);
      if (lookahead == 'i') ADVANCE(1486);
      END_STATE();
    case 749:
      if (lookahead == 'd') ADVANCE(111);
      END_STATE();
    case 750:
      if (lookahead == 'd') ADVANCE(432);
      END_STATE();
    case 751:
      if (lookahead == 'd') ADVANCE(451);
      END_STATE();
    case 752:
      if (lookahead == 'd') ADVANCE(206);
      END_STATE();
    case 753:
      if (lookahead == 'd') ADVANCE(124);
      END_STATE();
    case 754:
      if (lookahead == 'd') ADVANCE(1119);
      END_STATE();
    case 755:
      if (lookahead == 'd') ADVANCE(1119);
      if (lookahead == 'l') ADVANCE(1309);
      END_STATE();
    case 756:
      if (lookahead == 'd') ADVANCE(779);
      END_STATE();
    case 757:
      if (lookahead == 'd') ADVANCE(1923);
      END_STATE();
    case 758:
      if (lookahead == 'd') ADVANCE(1923);
      if (lookahead == 'p') ADVANCE(1525);
      END_STATE();
    case 759:
      if (lookahead == 'd') ADVANCE(1077);
      END_STATE();
    case 760:
      if (lookahead == 'd') ADVANCE(1077);
      if (lookahead == 'm') ADVANCE(1564);
      if (lookahead == 'n') ADVANCE(210);
      END_STATE();
    case 761:
      if (lookahead == 'd') ADVANCE(1802);
      if (lookahead == 'u') ADVANCE(1046);
      END_STATE();
    case 762:
      if (lookahead == 'd') ADVANCE(794);
      END_STATE();
    case 763:
      if (lookahead == 'd') ADVANCE(1765);
      END_STATE();
    case 764:
      if (lookahead == 'd') ADVANCE(1858);
      END_STATE();
    case 765:
      if (lookahead == 'd') ADVANCE(1046);
      END_STATE();
    case 766:
      if (lookahead == 'd') ADVANCE(1046);
      if (lookahead == 'u') ADVANCE(1802);
      END_STATE();
    case 767:
      if (lookahead == 'd') ADVANCE(1152);
      if (lookahead == 'n') ADVANCE(1887);
      END_STATE();
    case 768:
      if (lookahead == 'd') ADVANCE(998);
      END_STATE();
    case 769:
      if (lookahead == 'd') ADVANCE(1444);
      if (lookahead == 'u') ADVANCE(1530);
      END_STATE();
    case 770:
      if (lookahead == 'd') ADVANCE(863);
      if (lookahead == 'p') ADVANCE(949);
      END_STATE();
    case 771:
      if (lookahead == 'd') ADVANCE(859);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 772:
      if (lookahead == 'd') ADVANCE(876);
      END_STATE();
    case 773:
      if (lookahead == 'd') ADVANCE(1467);
      END_STATE();
    case 774:
      if (lookahead == 'd') ADVANCE(1116);
      END_STATE();
    case 775:
      if (lookahead == 'd') ADVANCE(1116);
      if (lookahead == 'i') ADVANCE(1211);
      END_STATE();
    case 776:
      if (lookahead == 'd') ADVANCE(878);
      END_STATE();
    case 777:
      if (lookahead == 'd') ADVANCE(498);
      if (lookahead == 'e') ADVANCE(1546);
      if (lookahead == 'l') ADVANCE(519);
      if (lookahead == 'p') ADVANCE(87);
      if (lookahead == 'r') ADVANCE(526);
      if (lookahead == 'v') ADVANCE(856);
      if (lookahead == 'w') ADVANCE(778);
      END_STATE();
    case 778:
      if (lookahead == 'e') ADVANCE(734);
      END_STATE();
    case 779:
      if (lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 780:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'k') ADVANCE(815);
      END_STATE();
    case 781:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(1125);
      if (lookahead == 'p') ADVANCE(505);
      END_STATE();
    case 782:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(582);
      END_STATE();
    case 783:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 's') ADVANCE(1217);
      END_STATE();
    case 784:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 't') ADVANCE(129);
      END_STATE();
    case 785:
      if (lookahead == 'e') ADVANCE(13);
      if (lookahead == 'u') ADVANCE(1860);
      END_STATE();
    case 786:
      if (lookahead == 'e') ADVANCE(294);
      END_STATE();
    case 787:
      if (lookahead == 'e') ADVANCE(170);
      END_STATE();
    case 788:
      if (lookahead == 'e') ADVANCE(1966);
      END_STATE();
    case 789:
      if (lookahead == 'e') ADVANCE(1966);
      if (lookahead == 'v') ADVANCE(591);
      END_STATE();
    case 790:
      if (lookahead == 'e') ADVANCE(958);
      if (lookahead == 'p') ADVANCE(1039);
      END_STATE();
    case 791:
      if (lookahead == 'e') ADVANCE(190);
      if (lookahead == 'o') ADVANCE(1811);
      END_STATE();
    case 792:
      if (lookahead == 'e') ADVANCE(86);
      if (lookahead == 'l') ADVANCE(1268);
      if (lookahead == 'r') ADVANCE(581);
      END_STATE();
    case 793:
      if (lookahead == 'e') ADVANCE(291);
      END_STATE();
    case 794:
      if (lookahead == 'e') ADVANCE(37);
      END_STATE();
    case 795:
      if (lookahead == 'e') ADVANCE(16);
      END_STATE();
    case 796:
      if (lookahead == 'e') ADVANCE(10);
      END_STATE();
    case 797:
      if (lookahead == 'e') ADVANCE(295);
      END_STATE();
    case 798:
      if (lookahead == 'e') ADVANCE(293);
      END_STATE();
    case 799:
      if (lookahead == 'e') ADVANCE(43);
      END_STATE();
    case 800:
      if (lookahead == 'e') ADVANCE(557);
      END_STATE();
    case 801:
      if (lookahead == 'e') ADVANCE(1347);
      END_STATE();
    case 802:
      if (lookahead == 'e') ADVANCE(17);
      END_STATE();
    case 803:
      if (lookahead == 'e') ADVANCE(320);
      END_STATE();
    case 804:
      if (lookahead == 'e') ADVANCE(351);
      END_STATE();
    case 805:
      if (lookahead == 'e') ADVANCE(109);
      END_STATE();
    case 806:
      if (lookahead == 'e') ADVANCE(288);
      END_STATE();
    case 807:
      if (lookahead == 'e') ADVANCE(378);
      END_STATE();
    case 808:
      if (lookahead == 'e') ADVANCE(20);
      END_STATE();
    case 809:
      if (lookahead == 'e') ADVANCE(107);
      END_STATE();
    case 810:
      if (lookahead == 'e') ADVANCE(361);
      END_STATE();
    case 811:
      if (lookahead == 'e') ADVANCE(1603);
      END_STATE();
    case 812:
      if (lookahead == 'e') ADVANCE(385);
      if (lookahead == 'p') ADVANCE(80);
      if (lookahead == 's') ADVANCE(385);
      END_STATE();
    case 813:
      if (lookahead == 'e') ADVANCE(1370);
      END_STATE();
    case 814:
      if (lookahead == 'e') ADVANCE(1689);
      if (lookahead == 'i') ADVANCE(683);
      END_STATE();
    case 815:
      if (lookahead == 'e') ADVANCE(1811);
      END_STATE();
    case 816:
      if (lookahead == 'e') ADVANCE(1811);
      if (lookahead == 'o') ADVANCE(1357);
      END_STATE();
    case 817:
      if (lookahead == 'e') ADVANCE(283);
      END_STATE();
    case 818:
      if (lookahead == 'e') ADVANCE(2008);
      END_STATE();
    case 819:
      if (lookahead == 'e') ADVANCE(768);
      END_STATE();
    case 820:
      if (lookahead == 'e') ADVANCE(1705);
      if (lookahead == 'i') ADVANCE(691);
      if (lookahead == 'k') ADVANCE(428);
      if (lookahead == 'o') ADVANCE(1605);
      END_STATE();
    case 821:
      if (lookahead == 'e') ADVANCE(959);
      END_STATE();
    case 822:
      if (lookahead == 'e') ADVANCE(186);
      END_STATE();
    case 823:
      if (lookahead == 'e') ADVANCE(322);
      END_STATE();
    case 824:
      if (lookahead == 'e') ADVANCE(369);
      END_STATE();
    case 825:
      if (lookahead == 'e') ADVANCE(1782);
      END_STATE();
    case 826:
      if (lookahead == 'e') ADVANCE(268);
      END_STATE();
    case 827:
      if (lookahead == 'e') ADVANCE(1585);
      END_STATE();
    case 828:
      if (lookahead == 'e') ADVANCE(311);
      END_STATE();
    case 829:
      if (lookahead == 'e') ADVANCE(769);
      END_STATE();
    case 830:
      if (lookahead == 'e') ADVANCE(399);
      END_STATE();
    case 831:
      if (lookahead == 'e') ADVANCE(399);
      if (lookahead == 'i') ADVANCE(1316);
      END_STATE();
    case 832:
      if (lookahead == 'e') ADVANCE(963);
      END_STATE();
    case 833:
      if (lookahead == 'e') ADVANCE(195);
      END_STATE();
    case 834:
      if (lookahead == 'e') ADVANCE(292);
      END_STATE();
    case 835:
      if (lookahead == 'e') ADVANCE(774);
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 836:
      if (lookahead == 'e') ADVANCE(774);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 837:
      if (lookahead == 'e') ADVANCE(2004);
      END_STATE();
    case 838:
      if (lookahead == 'e') ADVANCE(1589);
      if (lookahead == 'v') ADVANCE(856);
      if (lookahead == 'w') ADVANCE(819);
      END_STATE();
    case 839:
      if (lookahead == 'e') ADVANCE(954);
      END_STATE();
    case 840:
      if (lookahead == 'e') ADVANCE(198);
      END_STATE();
    case 841:
      if (lookahead == 'e') ADVANCE(1591);
      END_STATE();
    case 842:
      if (lookahead == 'e') ADVANCE(1192);
      END_STATE();
    case 843:
      if (lookahead == 'e') ADVANCE(1360);
      END_STATE();
    case 844:
      if (lookahead == 'e') ADVANCE(1590);
      END_STATE();
    case 845:
      if (lookahead == 'e') ADVANCE(1832);
      END_STATE();
    case 846:
      if (lookahead == 'e') ADVANCE(1756);
      END_STATE();
    case 847:
      if (lookahead == 'e') ADVANCE(1767);
      END_STATE();
    case 848:
      if (lookahead == 'e') ADVANCE(1850);
      END_STATE();
    case 849:
      if (lookahead == 'e') ADVANCE(1211);
      END_STATE();
    case 850:
      if (lookahead == 'e') ADVANCE(741);
      END_STATE();
    case 851:
      if (lookahead == 'e') ADVANCE(552);
      END_STATE();
    case 852:
      if (lookahead == 'e') ADVANCE(743);
      END_STATE();
    case 853:
      if (lookahead == 'e') ADVANCE(1357);
      END_STATE();
    case 854:
      if (lookahead == 'e') ADVANCE(615);
      END_STATE();
    case 855:
      if (lookahead == 'e') ADVANCE(746);
      END_STATE();
    case 856:
      if (lookahead == 'e') ADVANCE(779);
      END_STATE();
    case 857:
      if (lookahead == 'e') ADVANCE(763);
      END_STATE();
    case 858:
      if (lookahead == 'e') ADVANCE(1611);
      END_STATE();
    case 859:
      if (lookahead == 'e') ADVANCE(1762);
      END_STATE();
    case 860:
      if (lookahead == 'e') ADVANCE(1529);
      if (lookahead == 'k') ADVANCE(423);
      if (lookahead == 'n') ADVANCE(1497);
      if (lookahead == 'p') ADVANCE(1045);
      if (lookahead == 'r') ADVANCE(147);
      if (lookahead == 's') ADVANCE(1091);
      if (lookahead == 't') ADVANCE(1062);
      END_STATE();
    case 861:
      if (lookahead == 'e') ADVANCE(1529);
      if (lookahead == 'p') ADVANCE(1044);
      END_STATE();
    case 862:
      if (lookahead == 'e') ADVANCE(1773);
      END_STATE();
    case 863:
      if (lookahead == 'e') ADVANCE(1247);
      END_STATE();
    case 864:
      if (lookahead == 'e') ADVANCE(722);
      END_STATE();
    case 865:
      if (lookahead == 'e') ADVANCE(1411);
      END_STATE();
    case 866:
      if (lookahead == 'e') ADVANCE(1804);
      END_STATE();
    case 867:
      if (lookahead == 'e') ADVANCE(1856);
      END_STATE();
    case 868:
      if (lookahead == 'e') ADVANCE(1775);
      END_STATE();
    case 869:
      if (lookahead == 'e') ADVANCE(1834);
      END_STATE();
    case 870:
      if (lookahead == 'e') ADVANCE(434);
      END_STATE();
    case 871:
      if (lookahead == 'e') ADVANCE(795);
      END_STATE();
    case 872:
      if (lookahead == 'e') ADVANCE(1383);
      END_STATE();
    case 873:
      if (lookahead == 'e') ADVANCE(1274);
      END_STATE();
    case 874:
      if (lookahead == 'e') ADVANCE(1651);
      END_STATE();
    case 875:
      if (lookahead == 'e') ADVANCE(473);
      END_STATE();
    case 876:
      if (lookahead == 'e') ADVANCE(1765);
      END_STATE();
    case 877:
      if (lookahead == 'e') ADVANCE(457);
      END_STATE();
    case 878:
      if (lookahead == 'e') ADVANCE(1766);
      END_STATE();
    case 879:
      if (lookahead == 'e') ADVANCE(1821);
      if (lookahead == 'i') ADVANCE(1316);
      if (lookahead == 'u') ADVANCE(2015);
      END_STATE();
    case 880:
      if (lookahead == 'e') ADVANCE(1617);
      END_STATE();
    case 881:
      if (lookahead == 'e') ADVANCE(1724);
      END_STATE();
    case 882:
      if (lookahead == 'e') ADVANCE(1857);
      END_STATE();
    case 883:
      if (lookahead == 'e') ADVANCE(1688);
      END_STATE();
    case 884:
      if (lookahead == 'e') ADVANCE(1864);
      END_STATE();
    case 885:
      if (lookahead == 'e') ADVANCE(857);
      END_STATE();
    case 886:
      if (lookahead == 'e') ADVANCE(802);
      if (lookahead == 'r') ADVANCE(1188);
      END_STATE();
    case 887:
      if (lookahead == 'e') ADVANCE(1604);
      END_STATE();
    case 888:
      if (lookahead == 'e') ADVANCE(1698);
      if (lookahead == 'i') ADVANCE(1370);
      END_STATE();
    case 889:
      if (lookahead == 'e') ADVANCE(1648);
      END_STATE();
    case 890:
      if (lookahead == 'e') ADVANCE(1726);
      END_STATE();
    case 891:
      if (lookahead == 'e') ADVANCE(1746);
      END_STATE();
    case 892:
      if (lookahead == 'e') ADVANCE(807);
      END_STATE();
    case 893:
      if (lookahead == 'e') ADVANCE(1661);
      END_STATE();
    case 894:
      if (lookahead == 'e') ADVANCE(933);
      END_STATE();
    case 895:
      if (lookahead == 'e') ADVANCE(1658);
      END_STATE();
    case 896:
      if (lookahead == 'e') ADVANCE(1621);
      END_STATE();
    case 897:
      if (lookahead == 'e') ADVANCE(1666);
      END_STATE();
    case 898:
      if (lookahead == 'e') ADVANCE(1624);
      END_STATE();
    case 899:
      if (lookahead == 'e') ADVANCE(964);
      END_STATE();
    case 900:
      if (lookahead == 'e') ADVANCE(1968);
      if (lookahead == 'i') ADVANCE(1011);
      END_STATE();
    case 901:
      if (lookahead == 'e') ADVANCE(1281);
      END_STATE();
    case 902:
      if (lookahead == 'e') ADVANCE(377);
      END_STATE();
    case 903:
      if (lookahead == 'e') ADVANCE(751);
      END_STATE();
    case 904:
      if (lookahead == 'e') ADVANCE(531);
      if (lookahead == 'w') ADVANCE(531);
      END_STATE();
    case 905:
      if (lookahead == 'e') ADVANCE(961);
      END_STATE();
    case 906:
      if (lookahead == 'e') ADVANCE(961);
      if (lookahead == 'l') ADVANCE(13);
      if (lookahead == 't') ADVANCE(219);
      END_STATE();
    case 907:
      if (lookahead == 'e') ADVANCE(853);
      END_STATE();
    case 908:
      if (lookahead == 'e') ADVANCE(1793);
      END_STATE();
    case 909:
      if (lookahead == 'e') ADVANCE(754);
      END_STATE();
    case 910:
      if (lookahead == 'e') ADVANCE(1346);
      END_STATE();
    case 911:
      if (lookahead == 'e') ADVANCE(707);
      if (lookahead == 'o') ADVANCE(1427);
      END_STATE();
    case 912:
      if (lookahead == 'e') ADVANCE(520);
      END_STATE();
    case 913:
      if (lookahead == 'e') ADVANCE(1582);
      END_STATE();
    case 914:
      if (lookahead == 'e') ADVANCE(1686);
      END_STATE();
    case 915:
      if (lookahead == 'e') ADVANCE(1879);
      END_STATE();
    case 916:
      if (lookahead == 'e') ADVANCE(1879);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 917:
      if (lookahead == 'e') ADVANCE(602);
      END_STATE();
    case 918:
      if (lookahead == 'e') ADVANCE(965);
      END_STATE();
    case 919:
      if (lookahead == 'e') ADVANCE(772);
      END_STATE();
    case 920:
      if (lookahead == 'e') ADVANCE(708);
      END_STATE();
    case 921:
      if (lookahead == 'e') ADVANCE(1783);
      END_STATE();
    case 922:
      if (lookahead == 'e') ADVANCE(1569);
      END_STATE();
    case 923:
      if (lookahead == 'e') ADVANCE(966);
      END_STATE();
    case 924:
      if (lookahead == 'e') ADVANCE(1685);
      if (lookahead == 'n') ADVANCE(316);
      END_STATE();
    case 925:
      if (lookahead == 'e') ADVANCE(710);
      END_STATE();
    case 926:
      if (lookahead == 'e') ADVANCE(1437);
      END_STATE();
    case 927:
      if (lookahead == 'e') ADVANCE(967);
      END_STATE();
    case 928:
      if (lookahead == 'e') ADVANCE(968);
      END_STATE();
    case 929:
      if (lookahead == 'e') ADVANCE(1420);
      END_STATE();
    case 930:
      if (lookahead == 'e') ADVANCE(969);
      if (lookahead == 'o') ADVANCE(1397);
      END_STATE();
    case 931:
      if (lookahead == 'e') ADVANCE(1704);
      END_STATE();
    case 932:
      if (lookahead == 'e') ADVANCE(962);
      END_STATE();
    case 933:
      if (lookahead == 'e') ADVANCE(1898);
      END_STATE();
    case 934:
      if (lookahead == 'e') ADVANCE(1751);
      END_STATE();
    case 935:
      if (lookahead == 'e') ADVANCE(1405);
      END_STATE();
    case 936:
      if (lookahead == 'e') ADVANCE(556);
      END_STATE();
    case 937:
      if (lookahead == 'e') ADVANCE(1315);
      END_STATE();
    case 938:
      if (lookahead == 'e') ADVANCE(1013);
      END_STATE();
    case 939:
      if (lookahead == 'e') ADVANCE(1748);
      END_STATE();
    case 940:
      if (lookahead == 'e') ADVANCE(1885);
      END_STATE();
    case 941:
      if (lookahead == 'e') ADVANCE(1164);
      END_STATE();
    case 942:
      if (lookahead == 'e') ADVANCE(1731);
      END_STATE();
    case 943:
      if (lookahead == 'e') ADVANCE(776);
      END_STATE();
    case 944:
      if (lookahead == 'e') ADVANCE(555);
      END_STATE();
    case 945:
      if (lookahead == 'e') ADVANCE(1380);
      END_STATE();
    case 946:
      if (lookahead == 'e') ADVANCE(560);
      END_STATE();
    case 947:
      if (lookahead == 'e') ADVANCE(728);
      END_STATE();
    case 948:
      if (lookahead == 'e') ADVANCE(972);
      END_STATE();
    case 949:
      if (lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 950:
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(1918);
      END_STATE();
    case 951:
      if (lookahead == 'f') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(1477);
      END_STATE();
    case 952:
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(13);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 's') ADVANCE(631);
      END_STATE();
    case 953:
      if (lookahead == 'f') ADVANCE(1603);
      if (lookahead == 'i') ADVANCE(1370);
      if (lookahead == 'o') ADVANCE(1533);
      if (lookahead == 'p') ADVANCE(1673);
      if (lookahead == 's') ADVANCE(631);
      if (lookahead == 'u') ADVANCE(455);
      END_STATE();
    case 954:
      if (lookahead == 'f') ADVANCE(1811);
      END_STATE();
    case 955:
      if (lookahead == 'f') ADVANCE(960);
      END_STATE();
    case 956:
      if (lookahead == 'f') ADVANCE(1814);
      if (lookahead == 's') ADVANCE(1759);
      END_STATE();
    case 957:
      if (lookahead == 'f') ADVANCE(1837);
      if (lookahead == 'l') ADVANCE(83);
      if (lookahead == 'p') ADVANCE(949);
      END_STATE();
    case 958:
      if (lookahead == 'f') ADVANCE(1778);
      if (lookahead == 'p') ADVANCE(1036);
      END_STATE();
    case 959:
      if (lookahead == 'f') ADVANCE(1902);
      END_STATE();
    case 960:
      if (lookahead == 'f') ADVANCE(934);
      END_STATE();
    case 961:
      if (lookahead == 'f') ADVANCE(1842);
      END_STATE();
    case 962:
      if (lookahead == 'f') ADVANCE(1856);
      END_STATE();
    case 963:
      if (lookahead == 'f') ADVANCE(1521);
      END_STATE();
    case 964:
      if (lookahead == 'f') ADVANCE(1845);
      if (lookahead == 's') ADVANCE(1764);
      END_STATE();
    case 965:
      if (lookahead == 'f') ADVANCE(1823);
      END_STATE();
    case 966:
      if (lookahead == 'f') ADVANCE(1825);
      END_STATE();
    case 967:
      if (lookahead == 'f') ADVANCE(1841);
      END_STATE();
    case 968:
      if (lookahead == 'f') ADVANCE(1881);
      END_STATE();
    case 969:
      if (lookahead == 'f') ADVANCE(1829);
      END_STATE();
    case 970:
      if (lookahead == 'f') ADVANCE(1481);
      END_STATE();
    case 971:
      if (lookahead == 'f') ADVANCE(1120);
      END_STATE();
    case 972:
      if (lookahead == 'f') ADVANCE(1901);
      END_STATE();
    case 973:
      if (lookahead == 'g') ADVANCE(13);
      END_STATE();
    case 974:
      if (lookahead == 'g') ADVANCE(13);
      if (lookahead == 's') ADVANCE(1530);
      END_STATE();
    case 975:
      if (lookahead == 'g') ADVANCE(13);
      if (lookahead == 't') ADVANCE(219);
      END_STATE();
    case 976:
      if (lookahead == 'g') ADVANCE(620);
      END_STATE();
    case 977:
      if (lookahead == 'g') ADVANCE(101);
      END_STATE();
    case 978:
      if (lookahead == 'g') ADVANCE(101);
      if (lookahead == 'i') ADVANCE(1370);
      END_STATE();
    case 979:
      if (lookahead == 'g') ADVANCE(349);
      END_STATE();
    case 980:
      if (lookahead == 'g') ADVANCE(368);
      END_STATE();
    case 981:
      if (lookahead == 'g') ADVANCE(105);
      END_STATE();
    case 982:
      if (lookahead == 'g') ADVANCE(1314);
      END_STATE();
    case 983:
      if (lookahead == 'g') ADVANCE(164);
      END_STATE();
    case 984:
      if (lookahead == 'g') ADVANCE(385);
      END_STATE();
    case 985:
      if (lookahead == 'g') ADVANCE(1052);
      END_STATE();
    case 986:
      if (lookahead == 'g') ADVANCE(2007);
      END_STATE();
    case 987:
      if (lookahead == 'g') ADVANCE(999);
      if (lookahead == 'l') ADVANCE(848);
      if (lookahead == 'r') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(1040);
      END_STATE();
    case 988:
      if (lookahead == 'g') ADVANCE(999);
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 989:
      if (lookahead == 'g') ADVANCE(999);
      if (lookahead == 'r') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(1042);
      END_STATE();
    case 990:
      if (lookahead == 'g') ADVANCE(1443);
      END_STATE();
    case 991:
      if (lookahead == 'g') ADVANCE(1482);
      if (lookahead == 'p') ADVANCE(949);
      END_STATE();
    case 992:
      if (lookahead == 'g') ADVANCE(1482);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 993:
      if (lookahead == 'g') ADVANCE(1323);
      END_STATE();
    case 994:
      if (lookahead == 'g') ADVANCE(1041);
      END_STATE();
    case 995:
      if (lookahead == 'g') ADVANCE(1211);
      if (lookahead == 'i') ADVANCE(1243);
      if (lookahead == 'l') ADVANCE(973);
      if (lookahead == 'r') ADVANCE(1184);
      END_STATE();
    case 996:
      if (lookahead == 'g') ADVANCE(529);
      if (lookahead == 's') ADVANCE(1867);
      if (lookahead == 'w') ADVANCE(347);
      END_STATE();
    case 997:
      if (lookahead == 'g') ADVANCE(1357);
      END_STATE();
    case 998:
      if (lookahead == 'g') ADVANCE(779);
      END_STATE();
    case 999:
      if (lookahead == 'g') ADVANCE(811);
      END_STATE();
    case 1000:
      if (lookahead == 'g') ADVANCE(815);
      END_STATE();
    case 1001:
      if (lookahead == 'g') ADVANCE(1730);
      END_STATE();
    case 1002:
      if (lookahead == 'g') ADVANCE(1730);
      if (lookahead == 'i') ADVANCE(1370);
      if (lookahead == 't') ADVANCE(1511);
      END_STATE();
    case 1003:
      if (lookahead == 'g') ADVANCE(1628);
      END_STATE();
    case 1004:
      if (lookahead == 'g') ADVANCE(1055);
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 's') ADVANCE(1154);
      END_STATE();
    case 1005:
      if (lookahead == 'g') ADVANCE(773);
      END_STATE();
    case 1006:
      if (lookahead == 'g') ADVANCE(1637);
      END_STATE();
    case 1007:
      if (lookahead == 'g') ADVANCE(1051);
      END_STATE();
    case 1008:
      if (lookahead == 'g') ADVANCE(1831);
      END_STATE();
    case 1009:
      if (lookahead == 'g') ADVANCE(1831);
      if (lookahead == 'l') ADVANCE(825);
      END_STATE();
    case 1010:
      if (lookahead == 'g') ADVANCE(1831);
      if (lookahead == 'q') ADVANCE(1008);
      END_STATE();
    case 1011:
      if (lookahead == 'g') ADVANCE(1056);
      END_STATE();
    case 1012:
      if (lookahead == 'g') ADVANCE(1349);
      if (lookahead == 'm') ADVANCE(102);
      END_STATE();
    case 1013:
      if (lookahead == 'g') ADVANCE(1681);
      END_STATE();
    case 1014:
      if (lookahead == 'g') ADVANCE(1681);
      if (lookahead == 'r') ADVANCE(1798);
      END_STATE();
    case 1015:
      if (lookahead == 'g') ADVANCE(1057);
      END_STATE();
    case 1016:
      if (lookahead == 'g') ADVANCE(1059);
      END_STATE();
    case 1017:
      if (lookahead == 'g') ADVANCE(1060);
      END_STATE();
    case 1018:
      if (lookahead == 'g') ADVANCE(1061);
      END_STATE();
    case 1019:
      if (lookahead == 'g') ADVANCE(1260);
      END_STATE();
    case 1020:
      if (lookahead == 'g') ADVANCE(1063);
      END_STATE();
    case 1021:
      if (lookahead == 'g') ADVANCE(893);
      if (lookahead == 'r') ADVANCE(702);
      END_STATE();
    case 1022:
      if (lookahead == 'g') ADVANCE(1064);
      END_STATE();
    case 1023:
      if (lookahead == 'g') ADVANCE(1065);
      END_STATE();
    case 1024:
      if (lookahead == 'g') ADVANCE(1282);
      END_STATE();
    case 1025:
      if (lookahead == 'g') ADVANCE(1054);
      END_STATE();
    case 1026:
      if (lookahead == 'g') ADVANCE(1284);
      END_STATE();
    case 1027:
      if (lookahead == 'g') ADVANCE(1285);
      END_STATE();
    case 1028:
      if (lookahead == 'g') ADVANCE(1286);
      END_STATE();
    case 1029:
      if (lookahead == 'g') ADVANCE(1299);
      END_STATE();
    case 1030:
      if (lookahead == 'g') ADVANCE(1287);
      END_STATE();
    case 1031:
      if (lookahead == 'g') ADVANCE(1288);
      END_STATE();
    case 1032:
      if (lookahead == 'g') ADVANCE(348);
      END_STATE();
    case 1033:
      if (lookahead == 'g') ADVANCE(547);
      END_STATE();
    case 1034:
      if (lookahead == 'g') ADVANCE(1074);
      END_STATE();
    case 1035:
      if (lookahead == 'h') ADVANCE(734);
      END_STATE();
    case 1036:
      if (lookahead == 'h') ADVANCE(13);
      END_STATE();
    case 1037:
      if (lookahead == 'h') ADVANCE(158);
      END_STATE();
    case 1038:
      if (lookahead == 'h') ADVANCE(368);
      END_STATE();
    case 1039:
      if (lookahead == 'h') ADVANCE(385);
      END_STATE();
    case 1040:
      if (lookahead == 'h') ADVANCE(219);
      END_STATE();
    case 1041:
      if (lookahead == 'h') ADVANCE(1811);
      END_STATE();
    case 1042:
      if (lookahead == 'h') ADVANCE(1961);
      END_STATE();
    case 1043:
      if (lookahead == 'h') ADVANCE(607);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1044:
      if (lookahead == 'h') ADVANCE(1076);
      END_STATE();
    case 1045:
      if (lookahead == 'h') ADVANCE(1076);
      if (lookahead == 'i') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(1510);
      END_STATE();
    case 1046:
      if (lookahead == 'h') ADVANCE(399);
      END_STATE();
    case 1047:
      if (lookahead == 'h') ADVANCE(1530);
      END_STATE();
    case 1048:
      if (lookahead == 'h') ADVANCE(1192);
      END_STATE();
    case 1049:
      if (lookahead == 'h') ADVANCE(1192);
      if (lookahead == 'r') ADVANCE(177);
      END_STATE();
    case 1050:
      if (lookahead == 'h') ADVANCE(779);
      END_STATE();
    case 1051:
      if (lookahead == 'h') ADVANCE(1902);
      END_STATE();
    case 1052:
      if (lookahead == 'h') ADVANCE(1818);
      END_STATE();
    case 1053:
      if (lookahead == 'h') ADVANCE(409);
      END_STATE();
    case 1054:
      if (lookahead == 'h') ADVANCE(1856);
      END_STATE();
    case 1055:
      if (lookahead == 'h') ADVANCE(1820);
      END_STATE();
    case 1056:
      if (lookahead == 'h') ADVANCE(1845);
      END_STATE();
    case 1057:
      if (lookahead == 'h') ADVANCE(1833);
      END_STATE();
    case 1058:
      if (lookahead == 'h') ADVANCE(1079);
      END_STATE();
    case 1059:
      if (lookahead == 'h') ADVANCE(1824);
      END_STATE();
    case 1060:
      if (lookahead == 'h') ADVANCE(1826);
      END_STATE();
    case 1061:
      if (lookahead == 'h') ADVANCE(1841);
      END_STATE();
    case 1062:
      if (lookahead == 'h') ADVANCE(845);
      if (lookahead == 'r') ADVANCE(1187);
      END_STATE();
    case 1063:
      if (lookahead == 'h') ADVANCE(1828);
      END_STATE();
    case 1064:
      if (lookahead == 'h') ADVANCE(1830);
      END_STATE();
    case 1065:
      if (lookahead == 'h') ADVANCE(1827);
      END_STATE();
    case 1066:
      if (lookahead == 'h') ADVANCE(921);
      END_STATE();
    case 1067:
      if (lookahead == 'h') ADVANCE(1113);
      END_STATE();
    case 1068:
      if (lookahead == 'h') ADVANCE(853);
      END_STATE();
    case 1069:
      if (lookahead == 'h') ADVANCE(970);
      END_STATE();
    case 1070:
      if (lookahead == 'h') ADVANCE(1652);
      if (lookahead == 'i') ADVANCE(1350);
      if (lookahead == 'r') ADVANCE(1084);
      END_STATE();
    case 1071:
      if (lookahead == 'h') ADVANCE(875);
      END_STATE();
    case 1072:
      if (lookahead == 'h') ADVANCE(1137);
      END_STATE();
    case 1073:
      if (lookahead == 'h') ADVANCE(1733);
      END_STATE();
    case 1074:
      if (lookahead == 'h') ADVANCE(1901);
      END_STATE();
    case 1075:
      if (lookahead == 'i') ADVANCE(734);
      END_STATE();
    case 1076:
      if (lookahead == 'i') ADVANCE(13);
      END_STATE();
    case 1077:
      if (lookahead == 'i') ADVANCE(158);
      END_STATE();
    case 1078:
      if (lookahead == 'i') ADVANCE(141);
      END_STATE();
    case 1079:
      if (lookahead == 'i') ADVANCE(683);
      END_STATE();
    case 1080:
      if (lookahead == 'i') ADVANCE(1990);
      if (lookahead == 'o') ADVANCE(1071);
      END_STATE();
    case 1081:
      if (lookahead == 'i') ADVANCE(2006);
      END_STATE();
    case 1082:
      if (lookahead == 'i') ADVANCE(135);
      END_STATE();
    case 1083:
      if (lookahead == 'i') ADVANCE(163);
      END_STATE();
    case 1084:
      if (lookahead == 'i') ADVANCE(160);
      END_STATE();
    case 1085:
      if (lookahead == 'i') ADVANCE(146);
      END_STATE();
    case 1086:
      if (lookahead == 'i') ADVANCE(1603);
      END_STATE();
    case 1087:
      if (lookahead == 'i') ADVANCE(1603);
      if (lookahead == 'r') ADVANCE(1457);
      END_STATE();
    case 1088:
      if (lookahead == 'i') ADVANCE(1370);
      END_STATE();
    case 1089:
      if (lookahead == 'i') ADVANCE(1370);
      if (lookahead == 'n') ADVANCE(1811);
      END_STATE();
    case 1090:
      if (lookahead == 'i') ADVANCE(219);
      if (lookahead == 'm') ADVANCE(1348);
      if (lookahead == 'o') ADVANCE(1391);
      END_STATE();
    case 1091:
      if (lookahead == 'i') ADVANCE(993);
      if (lookahead == 'u') ADVANCE(606);
      END_STATE();
    case 1092:
      if (lookahead == 'i') ADVANCE(1244);
      END_STATE();
    case 1093:
      if (lookahead == 'i') ADVANCE(162);
      END_STATE();
    case 1094:
      if (lookahead == 'i') ADVANCE(1811);
      END_STATE();
    case 1095:
      if (lookahead == 'i') ADVANCE(1961);
      END_STATE();
    case 1096:
      if (lookahead == 'i') ADVANCE(729);
      END_STATE();
    case 1097:
      if (lookahead == 'i') ADVANCE(729);
      if (lookahead == 'o') ADVANCE(1945);
      END_STATE();
    case 1098:
      if (lookahead == 'i') ADVANCE(1358);
      END_STATE();
    case 1099:
      if (lookahead == 'i') ADVANCE(220);
      END_STATE();
    case 1100:
      if (lookahead == 'i') ADVANCE(1805);
      END_STATE();
    case 1101:
      if (lookahead == 'i') ADVANCE(1805);
      if (lookahead == 'p') ADVANCE(1485);
      END_STATE();
    case 1102:
      if (lookahead == 'i') ADVANCE(111);
      END_STATE();
    case 1103:
      if (lookahead == 'i') ADVANCE(1240);
      END_STATE();
    case 1104:
      if (lookahead == 'i') ADVANCE(1530);
      END_STATE();
    case 1105:
      if (lookahead == 'i') ADVANCE(973);
      END_STATE();
    case 1106:
      if (lookahead == 'i') ADVANCE(973);
      if (lookahead == 'l') ADVANCE(1105);
      END_STATE();
    case 1107:
      if (lookahead == 'i') ADVANCE(955);
      END_STATE();
    case 1108:
      if (lookahead == 'i') ADVANCE(1316);
      END_STATE();
    case 1109:
      if (lookahead == 'i') ADVANCE(1316);
      if (lookahead == 'l') ADVANCE(13);
      END_STATE();
    case 1110:
      if (lookahead == 'i') ADVANCE(1316);
      if (lookahead == 'l') ADVANCE(527);
      END_STATE();
    case 1111:
      if (lookahead == 'i') ADVANCE(1482);
      END_STATE();
    case 1112:
      if (lookahead == 'i') ADVANCE(1756);
      END_STATE();
    case 1113:
      if (lookahead == 'i') ADVANCE(1378);
      END_STATE();
    case 1114:
      if (lookahead == 'i') ADVANCE(1781);
      if (lookahead == 'l') ADVANCE(1491);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 1115:
      if (lookahead == 'i') ADVANCE(1781);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 1116:
      if (lookahead == 'i') ADVANCE(1211);
      END_STATE();
    case 1117:
      if (lookahead == 'i') ADVANCE(986);
      END_STATE();
    case 1118:
      if (lookahead == 'i') ADVANCE(997);
      END_STATE();
    case 1119:
      if (lookahead == 'i') ADVANCE(1922);
      END_STATE();
    case 1120:
      if (lookahead == 'i') ADVANCE(1357);
      END_STATE();
    case 1121:
      if (lookahead == 'i') ADVANCE(1241);
      END_STATE();
    case 1122:
      if (lookahead == 'i') ADVANCE(615);
      END_STATE();
    case 1123:
      if (lookahead == 'i') ADVANCE(1633);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1124:
      if (lookahead == 'i') ADVANCE(779);
      END_STATE();
    case 1125:
      if (lookahead == 'i') ADVANCE(1391);
      END_STATE();
    case 1126:
      if (lookahead == 'i') ADVANCE(541);
      END_STATE();
    case 1127:
      if (lookahead == 'i') ADVANCE(764);
      END_STATE();
    case 1128:
      if (lookahead == 'i') ADVANCE(1335);
      END_STATE();
    case 1129:
      if (lookahead == 'i') ADVANCE(1033);
      END_STATE();
    case 1130:
      if (lookahead == 'i') ADVANCE(1677);
      if (lookahead == 'l') ADVANCE(949);
      if (lookahead == 'm') ADVANCE(1132);
      if (lookahead == 'r') ADVANCE(745);
      END_STATE();
    case 1131:
      if (lookahead == 'i') ADVANCE(1382);
      END_STATE();
    case 1132:
      if (lookahead == 'i') ADVANCE(1227);
      END_STATE();
    case 1133:
      if (lookahead == 'i') ADVANCE(846);
      END_STATE();
    case 1134:
      if (lookahead == 'i') ADVANCE(1917);
      END_STATE();
    case 1135:
      if (lookahead == 'i') ADVANCE(1422);
      END_STATE();
    case 1136:
      if (lookahead == 'i') ADVANCE(1415);
      END_STATE();
    case 1137:
      if (lookahead == 'i') ADVANCE(1361);
      END_STATE();
    case 1138:
      if (lookahead == 'i') ADVANCE(1257);
      if (lookahead == 'l') ADVANCE(1106);
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 1139:
      if (lookahead == 'i') ADVANCE(994);
      END_STATE();
    case 1140:
      if (lookahead == 'i') ADVANCE(1245);
      END_STATE();
    case 1141:
      if (lookahead == 'i') ADVANCE(1393);
      if (lookahead == 'p') ADVANCE(949);
      END_STATE();
    case 1142:
      if (lookahead == 'i') ADVANCE(1670);
      END_STATE();
    case 1143:
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'o') ADVANCE(1241);
      END_STATE();
    case 1144:
      if (lookahead == 'i') ADVANCE(1670);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1145:
      if (lookahead == 'i') ADVANCE(1243);
      END_STATE();
    case 1146:
      if (lookahead == 'i') ADVANCE(598);
      END_STATE();
    case 1147:
      if (lookahead == 'i') ADVANCE(489);
      END_STATE();
    case 1148:
      if (lookahead == 'i') ADVANCE(1796);
      END_STATE();
    case 1149:
      if (lookahead == 'i') ADVANCE(1967);
      END_STATE();
    case 1150:
      if (lookahead == 'i') ADVANCE(1576);
      END_STATE();
    case 1151:
      if (lookahead == 'i') ADVANCE(1401);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'u') ADVANCE(1372);
      END_STATE();
    case 1152:
      if (lookahead == 'i') ADVANCE(1262);
      END_STATE();
    case 1153:
      if (lookahead == 'i') ADVANCE(1340);
      END_STATE();
    case 1154:
      if (lookahead == 'i') ADVANCE(1395);
      END_STATE();
    case 1155:
      if (lookahead == 'i') ADVANCE(1800);
      END_STATE();
    case 1156:
      if (lookahead == 'i') ADVANCE(1795);
      END_STATE();
    case 1157:
      if (lookahead == 'i') ADVANCE(1350);
      END_STATE();
    case 1158:
      if (lookahead == 'i') ADVANCE(1336);
      END_STATE();
    case 1159:
      if (lookahead == 'i') ADVANCE(1876);
      END_STATE();
    case 1160:
      if (lookahead == 'i') ADVANCE(1273);
      END_STATE();
    case 1161:
      if (lookahead == 'i') ADVANCE(1399);
      END_STATE();
    case 1162:
      if (lookahead == 'i') ADVANCE(890);
      END_STATE();
    case 1163:
      if (lookahead == 'i') ADVANCE(497);
      END_STATE();
    case 1164:
      if (lookahead == 'i') ADVANCE(1292);
      END_STATE();
    case 1165:
      if (lookahead == 'i') ADVANCE(1260);
      END_STATE();
    case 1166:
      if (lookahead == 'i') ADVANCE(1499);
      END_STATE();
    case 1167:
      if (lookahead == 'i') ADVANCE(1478);
      END_STATE();
    case 1168:
      if (lookahead == 'i') ADVANCE(1416);
      END_STATE();
    case 1169:
      if (lookahead == 'i') ADVANCE(1416);
      if (lookahead == 'n') ADVANCE(13);
      END_STATE();
    case 1170:
      if (lookahead == 'i') ADVANCE(1007);
      END_STATE();
    case 1171:
      if (lookahead == 'i') ADVANCE(1293);
      END_STATE();
    case 1172:
      if (lookahead == 'i') ADVANCE(721);
      END_STATE();
    case 1173:
      if (lookahead == 'i') ADVANCE(725);
      END_STATE();
    case 1174:
      if (lookahead == 'i') ADVANCE(599);
      END_STATE();
    case 1175:
      if (lookahead == 'i') ADVANCE(1015);
      END_STATE();
    case 1176:
      if (lookahead == 'i') ADVANCE(1016);
      END_STATE();
    case 1177:
      if (lookahead == 'i') ADVANCE(1017);
      END_STATE();
    case 1178:
      if (lookahead == 'i') ADVANCE(1018);
      END_STATE();
    case 1179:
      if (lookahead == 'i') ADVANCE(1020);
      END_STATE();
    case 1180:
      if (lookahead == 'i') ADVANCE(1022);
      END_STATE();
    case 1181:
      if (lookahead == 'i') ADVANCE(1023);
      END_STATE();
    case 1182:
      if (lookahead == 'i') ADVANCE(1025);
      END_STATE();
    case 1183:
      if (lookahead == 'i') ADVANCE(1744);
      END_STATE();
    case 1184:
      if (lookahead == 'i') ADVANCE(565);
      END_STATE();
    case 1185:
      if (lookahead == 'i') ADVANCE(1034);
      END_STATE();
    case 1186:
      if (lookahead == 'i') ADVANCE(567);
      END_STATE();
    case 1187:
      if (lookahead == 'i') ADVANCE(568);
      END_STATE();
    case 1188:
      if (lookahead == 'i') ADVANCE(569);
      END_STATE();
    case 1189:
      if (lookahead == 'i') ADVANCE(570);
      END_STATE();
    case 1190:
      if (lookahead == 'j') ADVANCE(13);
      END_STATE();
    case 1191:
      if (lookahead == 'j') ADVANCE(13);
      if (lookahead == 'n') ADVANCE(1190);
      END_STATE();
    case 1192:
      if (lookahead == 'k') ADVANCE(13);
      END_STATE();
    case 1193:
      if (lookahead == 'k') ADVANCE(368);
      END_STATE();
    case 1194:
      if (lookahead == 'k') ADVANCE(723);
      END_STATE();
    case 1195:
      if (lookahead == 'k') ADVANCE(1297);
      END_STATE();
    case 1196:
      if (lookahead == 'k') ADVANCE(1270);
      END_STATE();
    case 1197:
      if (lookahead == 'k') ADVANCE(169);
      END_STATE();
    case 1198:
      if (lookahead == 'k') ADVANCE(145);
      END_STATE();
    case 1199:
      if (lookahead == 'k') ADVANCE(607);
      END_STATE();
    case 1200:
      if (lookahead == 'k') ADVANCE(607);
      if (lookahead == 'm') ADVANCE(158);
      END_STATE();
    case 1201:
      if (lookahead == 'k') ADVANCE(1984);
      END_STATE();
    case 1202:
      if (lookahead == 'k') ADVANCE(211);
      END_STATE();
    case 1203:
      if (lookahead == 'k') ADVANCE(1792);
      END_STATE();
    case 1204:
      if (lookahead == 'k') ADVANCE(811);
      END_STATE();
    case 1205:
      if (lookahead == 'k') ADVANCE(815);
      END_STATE();
    case 1206:
      if (lookahead == 'k') ADVANCE(524);
      END_STATE();
    case 1207:
      if (lookahead == 'k') ADVANCE(531);
      END_STATE();
    case 1208:
      if (lookahead == 'k') ADVANCE(531);
      if (lookahead == 'l') ADVANCE(462);
      END_STATE();
    case 1209:
      if (lookahead == 'k') ADVANCE(1161);
      END_STATE();
    case 1210:
      if (lookahead == 'l') ADVANCE(734);
      END_STATE();
    case 1211:
      if (lookahead == 'l') ADVANCE(13);
      END_STATE();
    case 1212:
      if (lookahead == 'l') ADVANCE(208);
      END_STATE();
    case 1213:
      if (lookahead == 'l') ADVANCE(347);
      END_STATE();
    case 1214:
      if (lookahead == 'l') ADVANCE(316);
      END_STATE();
    case 1215:
      if (lookahead == 'l') ADVANCE(82);
      END_STATE();
    case 1216:
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 1217:
      if (lookahead == 'l') ADVANCE(2017);
      END_STATE();
    case 1218:
      if (lookahead == 'l') ADVANCE(282);
      END_STATE();
    case 1219:
      if (lookahead == 'l') ADVANCE(269);
      END_STATE();
    case 1220:
      if (lookahead == 'l') ADVANCE(46);
      END_STATE();
    case 1221:
      if (lookahead == 'l') ADVANCE(385);
      END_STATE();
    case 1222:
      if (lookahead == 'l') ADVANCE(1996);
      END_STATE();
    case 1223:
      if (lookahead == 'l') ADVANCE(910);
      END_STATE();
    case 1224:
      if (lookahead == 'l') ADVANCE(910);
      if (lookahead == 'q') ADVANCE(1935);
      END_STATE();
    case 1225:
      if (lookahead == 'l') ADVANCE(910);
      if (lookahead == 'q') ADVANCE(1934);
      if (lookahead == 'x') ADVANCE(1100);
      END_STATE();
    case 1226:
      if (lookahead == 'l') ADVANCE(332);
      END_STATE();
    case 1227:
      if (lookahead == 'l') ADVANCE(1811);
      END_STATE();
    case 1228:
      if (lookahead == 'l') ADVANCE(51);
      END_STATE();
    case 1229:
      if (lookahead == 'l') ADVANCE(325);
      END_STATE();
    case 1230:
      if (lookahead == 'l') ADVANCE(289);
      END_STATE();
    case 1231:
      if (lookahead == 'l') ADVANCE(592);
      END_STATE();
    case 1232:
      if (lookahead == 'l') ADVANCE(281);
      END_STATE();
    case 1233:
      if (lookahead == 'l') ADVANCE(370);
      END_STATE();
    case 1234:
      if (lookahead == 'l') ADVANCE(399);
      END_STATE();
    case 1235:
      if (lookahead == 'l') ADVANCE(300);
      END_STATE();
    case 1236:
      if (lookahead == 'l') ADVANCE(195);
      END_STATE();
    case 1237:
      if (lookahead == 'l') ADVANCE(432);
      END_STATE();
    case 1238:
      if (lookahead == 'l') ADVANCE(116);
      END_STATE();
    case 1239:
      if (lookahead == 'l') ADVANCE(310);
      END_STATE();
    case 1240:
      if (lookahead == 'l') ADVANCE(762);
      END_STATE();
    case 1241:
      if (lookahead == 'l') ADVANCE(1482);
      END_STATE();
    case 1242:
      if (lookahead == 'l') ADVANCE(1192);
      END_STATE();
    case 1243:
      if (lookahead == 'l') ADVANCE(749);
      END_STATE();
    case 1244:
      if (lookahead == 'l') ADVANCE(749);
      if (lookahead == 'm') ADVANCE(846);
      END_STATE();
    case 1245:
      if (lookahead == 'l') ADVANCE(749);
      if (lookahead == 'm') ADVANCE(868);
      END_STATE();
    case 1246:
      if (lookahead == 'l') ADVANCE(828);
      END_STATE();
    case 1247:
      if (lookahead == 'l') ADVANCE(1756);
      END_STATE();
    case 1248:
      if (lookahead == 'l') ADVANCE(701);
      END_STATE();
    case 1249:
      if (lookahead == 'l') ADVANCE(1112);
      END_STATE();
    case 1250:
      if (lookahead == 'l') ADVANCE(1995);
      END_STATE();
    case 1251:
      if (lookahead == 'l') ADVANCE(1078);
      END_STATE();
    case 1252:
      if (lookahead == 'l') ADVANCE(849);
      END_STATE();
    case 1253:
      if (lookahead == 'l') ADVANCE(1211);
      END_STATE();
    case 1254:
      if (lookahead == 'l') ADVANCE(2005);
      END_STATE();
    case 1255:
      if (lookahead == 'l') ADVANCE(1491);
      END_STATE();
    case 1256:
      if (lookahead == 'l') ADVANCE(1942);
      END_STATE();
    case 1257:
      if (lookahead == 'l') ADVANCE(1105);
      END_STATE();
    case 1258:
      if (lookahead == 'l') ADVANCE(1238);
      if (lookahead == 'm') ADVANCE(1542);
      END_STATE();
    case 1259:
      if (lookahead == 'l') ADVANCE(1234);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(115);
      if (lookahead == 'u') ADVANCE(597);
      if (lookahead == 'w') ADVANCE(1362);
      END_STATE();
    case 1260:
      if (lookahead == 'l') ADVANCE(779);
      END_STATE();
    case 1261:
      if (lookahead == 'l') ADVANCE(837);
      END_STATE();
    case 1262:
      if (lookahead == 'l') ADVANCE(1221);
      END_STATE();
    case 1263:
      if (lookahead == 'l') ADVANCE(1230);
      END_STATE();
    case 1264:
      if (lookahead == 'l') ADVANCE(850);
      END_STATE();
    case 1265:
      if (lookahead == 'l') ADVANCE(1312);
      if (lookahead == 's') ADVANCE(1047);
      END_STATE();
    case 1266:
      if (lookahead == 'l') ADVANCE(1509);
      END_STATE();
    case 1267:
      if (lookahead == 'l') ADVANCE(1133);
      END_STATE();
    case 1268:
      if (lookahead == 'l') ADVANCE(1104);
      END_STATE();
    case 1269:
      if (lookahead == 'l') ADVANCE(1229);
      END_STATE();
    case 1270:
      if (lookahead == 'l') ADVANCE(1460);
      if (lookahead == 's') ADVANCE(1600);
      if (lookahead == 't') ADVANCE(1755);
      END_STATE();
    case 1271:
      if (lookahead == 'l') ADVANCE(825);
      END_STATE();
    case 1272:
      if (lookahead == 'l') ADVANCE(825);
      if (lookahead == 'q') ADVANCE(1271);
      END_STATE();
    case 1273:
      if (lookahead == 'l') ADVANCE(1146);
      END_STATE();
    case 1274:
      if (lookahead == 'l') ADVANCE(448);
      END_STATE();
    case 1275:
      if (lookahead == 'l') ADVANCE(1233);
      END_STATE();
    case 1276:
      if (lookahead == 'l') ADVANCE(797);
      END_STATE();
    case 1277:
      if (lookahead == 'l') ADVANCE(798);
      END_STATE();
    case 1278:
      if (lookahead == 'l') ADVANCE(1494);
      END_STATE();
    case 1279:
      if (lookahead == 'l') ADVANCE(786);
      END_STATE();
    case 1280:
      if (lookahead == 'l') ADVANCE(917);
      END_STATE();
    case 1281:
      if (lookahead == 'l') ADVANCE(839);
      if (lookahead == 'r') ADVANCE(1139);
      END_STATE();
    case 1282:
      if (lookahead == 'l') ADVANCE(805);
      END_STATE();
    case 1283:
      if (lookahead == 'l') ADVANCE(806);
      END_STATE();
    case 1284:
      if (lookahead == 'l') ADVANCE(817);
      END_STATE();
    case 1285:
      if (lookahead == 'l') ADVANCE(937);
      END_STATE();
    case 1286:
      if (lookahead == 'l') ADVANCE(829);
      END_STATE();
    case 1287:
      if (lookahead == 'l') ADVANCE(808);
      END_STATE();
    case 1288:
      if (lookahead == 'l') ADVANCE(809);
      END_STATE();
    case 1289:
      if (lookahead == 'l') ADVANCE(810);
      END_STATE();
    case 1290:
      if (lookahead == 'l') ADVANCE(826);
      END_STATE();
    case 1291:
      if (lookahead == 'l') ADVANCE(462);
      END_STATE();
    case 1292:
      if (lookahead == 'l') ADVANCE(1113);
      END_STATE();
    case 1293:
      if (lookahead == 'l') ADVANCE(756);
      END_STATE();
    case 1294:
      if (lookahead == 'l') ADVANCE(756);
      if (lookahead == 'm') ADVANCE(840);
      if (lookahead == 'n') ADVANCE(1811);
      END_STATE();
    case 1295:
      if (lookahead == 'l') ADVANCE(1484);
      if (lookahead == 'n') ADVANCE(1002);
      if (lookahead == 'p') ADVANCE(951);
      if (lookahead == 'u') ADVANCE(1423);
      END_STATE();
    case 1296:
      if (lookahead == 'l') ADVANCE(404);
      END_STATE();
    case 1297:
      if (lookahead == 'l') ADVANCE(821);
      if (lookahead == 'r') ADVANCE(1170);
      END_STATE();
    case 1298:
      if (lookahead == 'l') ADVANCE(902);
      END_STATE();
    case 1299:
      if (lookahead == 'l') ADVANCE(901);
      END_STATE();
    case 1300:
      if (lookahead == 'l') ADVANCE(517);
      END_STATE();
    case 1301:
      if (lookahead == 'l') ADVANCE(1918);
      END_STATE();
    case 1302:
      if (lookahead == 'l') ADVANCE(1264);
      END_STATE();
    case 1303:
      if (lookahead == 'l') ADVANCE(463);
      END_STATE();
    case 1304:
      if (lookahead == 'l') ADVANCE(1154);
      END_STATE();
    case 1305:
      if (lookahead == 'l') ADVANCE(1489);
      if (lookahead == 'm') ADVANCE(1345);
      if (lookahead == 'n') ADVANCE(978);
      if (lookahead == 'p') ADVANCE(136);
      END_STATE();
    case 1306:
      if (lookahead == 'l') ADVANCE(1702);
      END_STATE();
    case 1307:
      if (lookahead == 'l') ADVANCE(1088);
      END_STATE();
    case 1308:
      if (lookahead == 'l') ADVANCE(1249);
      END_STATE();
    case 1309:
      if (lookahead == 'l') ADVANCE(1136);
      END_STATE();
    case 1310:
      if (lookahead == 'l') ADVANCE(1252);
      END_STATE();
    case 1311:
      if (lookahead == 'l') ADVANCE(539);
      END_STATE();
    case 1312:
      if (lookahead == 'l') ADVANCE(1808);
      END_STATE();
    case 1313:
      if (lookahead == 'l') ADVANCE(1304);
      END_STATE();
    case 1314:
      if (lookahead == 'l') ADVANCE(905);
      if (lookahead == 'm') ADVANCE(484);
      if (lookahead == 'r') ADVANCE(1170);
      END_STATE();
    case 1315:
      if (lookahead == 'l') ADVANCE(932);
      if (lookahead == 'r') ADVANCE(1182);
      END_STATE();
    case 1316:
      if (lookahead == 'm') ADVANCE(13);
      END_STATE();
    case 1317:
      if (lookahead == 'm') ADVANCE(158);
      END_STATE();
    case 1318:
      if (lookahead == 'm') ADVANCE(158);
      if (lookahead == 'r') ADVANCE(1441);
      END_STATE();
    case 1319:
      if (lookahead == 'm') ADVANCE(368);
      END_STATE();
    case 1320:
      if (lookahead == 'm') ADVANCE(181);
      END_STATE();
    case 1321:
      if (lookahead == 'm') ADVANCE(227);
      END_STATE();
    case 1322:
      if (lookahead == 'm') ADVANCE(226);
      END_STATE();
    case 1323:
      if (lookahead == 'm') ADVANCE(385);
      END_STATE();
    case 1324:
      if (lookahead == 'm') ADVANCE(1169);
      END_STATE();
    case 1325:
      if (lookahead == 'm') ADVANCE(1076);
      END_STATE();
    case 1326:
      if (lookahead == 'm') ADVANCE(111);
      END_STATE();
    case 1327:
      if (lookahead == 'm') ADVANCE(1530);
      END_STATE();
    case 1328:
      if (lookahead == 'm') ADVANCE(1560);
      END_STATE();
    case 1329:
      if (lookahead == 'm') ADVANCE(1541);
      END_STATE();
    case 1330:
      if (lookahead == 'm') ADVANCE(1075);
      END_STATE();
    case 1331:
      if (lookahead == 'm') ADVANCE(1075);
      if (lookahead == 'p') ADVANCE(536);
      END_STATE();
    case 1332:
      if (lookahead == 'm') ADVANCE(1323);
      END_STATE();
    case 1333:
      if (lookahead == 'm') ADVANCE(119);
      END_STATE();
    case 1334:
      if (lookahead == 'm') ADVANCE(1564);
      END_STATE();
    case 1335:
      if (lookahead == 'm') ADVANCE(120);
      END_STATE();
    case 1336:
      if (lookahead == 'm') ADVANCE(425);
      END_STATE();
    case 1337:
      if (lookahead == 'm') ADVANCE(849);
      END_STATE();
    case 1338:
      if (lookahead == 'm') ADVANCE(1211);
      END_STATE();
    case 1339:
      if (lookahead == 'm') ADVANCE(1357);
      END_STATE();
    case 1340:
      if (lookahead == 'm') ADVANCE(779);
      END_STATE();
    case 1341:
      if (lookahead == 'm') ADVANCE(1556);
      END_STATE();
    case 1342:
      if (lookahead == 'm') ADVANCE(1551);
      END_STATE();
    case 1343:
      if (lookahead == 'm') ADVANCE(541);
      END_STATE();
    case 1344:
      if (lookahead == 'm') ADVANCE(413);
      END_STATE();
    case 1345:
      if (lookahead == 'm') ADVANCE(420);
      if (lookahead == 'p') ADVANCE(139);
      END_STATE();
    case 1346:
      if (lookahead == 'm') ADVANCE(813);
      END_STATE();
    case 1347:
      if (lookahead == 'm') ADVANCE(813);
      if (lookahead == 'x') ADVANCE(846);
      END_STATE();
    case 1348:
      if (lookahead == 'm') ADVANCE(409);
      END_STATE();
    case 1349:
      if (lookahead == 'm') ADVANCE(393);
      END_STATE();
    case 1350:
      if (lookahead == 'm') ADVANCE(846);
      END_STATE();
    case 1351:
      if (lookahead == 'm') ADVANCE(1572);
      END_STATE();
    case 1352:
      if (lookahead == 'm') ADVANCE(833);
      END_STATE();
    case 1353:
      if (lookahead == 'm') ADVANCE(1344);
      END_STATE();
    case 1354:
      if (lookahead == 'm') ADVANCE(1168);
      END_STATE();
    case 1355:
      if (lookahead == 'm') ADVANCE(535);
      END_STATE();
    case 1356:
      if (lookahead == 'n') ADVANCE(734);
      END_STATE();
    case 1357:
      if (lookahead == 'n') ADVANCE(13);
      END_STATE();
    case 1358:
      if (lookahead == 'n') ADVANCE(141);
      END_STATE();
    case 1359:
      if (lookahead == 'n') ADVANCE(278);
      END_STATE();
    case 1360:
      if (lookahead == 'n') ADVANCE(291);
      END_STATE();
    case 1361:
      if (lookahead == 'n') ADVANCE(368);
      END_STATE();
    case 1362:
      if (lookahead == 'n') ADVANCE(549);
      END_STATE();
    case 1363:
      if (lookahead == 'n') ADVANCE(103);
      END_STATE();
    case 1364:
      if (lookahead == 'n') ADVANCE(693);
      END_STATE();
    case 1365:
      if (lookahead == 'n') ADVANCE(2013);
      END_STATE();
    case 1366:
      if (lookahead == 'n') ADVANCE(31);
      END_STATE();
    case 1367:
      if (lookahead == 'n') ADVANCE(374);
      END_STATE();
    case 1368:
      if (lookahead == 'n') ADVANCE(378);
      END_STATE();
    case 1369:
      if (lookahead == 'n') ADVANCE(49);
      END_STATE();
    case 1370:
      if (lookahead == 'n') ADVANCE(1811);
      END_STATE();
    case 1371:
      if (lookahead == 'n') ADVANCE(341);
      END_STATE();
    case 1372:
      if (lookahead == 'n') ADVANCE(104);
      END_STATE();
    case 1373:
      if (lookahead == 'n') ADVANCE(272);
      END_STATE();
    case 1374:
      if (lookahead == 'n') ADVANCE(1989);
      END_STATE();
    case 1375:
      if (lookahead == 'n') ADVANCE(971);
      END_STATE();
    case 1376:
      if (lookahead == 'n') ADVANCE(769);
      END_STATE();
    case 1377:
      if (lookahead == 'n') ADVANCE(111);
      END_STATE();
    case 1378:
      if (lookahead == 'n') ADVANCE(973);
      END_STATE();
    case 1379:
      if (lookahead == 'n') ADVANCE(973);
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 1380:
      if (lookahead == 'n') ADVANCE(1871);
      END_STATE();
    case 1381:
      if (lookahead == 'n') ADVANCE(979);
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'w') ADVANCE(880);
      END_STATE();
    case 1382:
      if (lookahead == 'n') ADVANCE(216);
      END_STATE();
    case 1383:
      if (lookahead == 'n') ADVANCE(1192);
      END_STATE();
    case 1384:
      if (lookahead == 'n') ADVANCE(119);
      END_STATE();
    case 1385:
      if (lookahead == 'n') ADVANCE(1756);
      END_STATE();
    case 1386:
      if (lookahead == 'n') ADVANCE(742);
      END_STATE();
    case 1387:
      if (lookahead == 'n') ADVANCE(114);
      END_STATE();
    case 1388:
      if (lookahead == 'n') ADVANCE(1003);
      if (lookahead == 'r') ADVANCE(860);
      END_STATE();
    case 1389:
      if (lookahead == 'n') ADVANCE(699);
      END_STATE();
    case 1390:
      if (lookahead == 'n') ADVANCE(1019);
      END_STATE();
    case 1391:
      if (lookahead == 'n') ADVANCE(779);
      END_STATE();
    case 1392:
      if (lookahead == 'n') ADVANCE(977);
      END_STATE();
    case 1393:
      if (lookahead == 'n') ADVANCE(727);
      END_STATE();
    case 1394:
      if (lookahead == 'n') ADVANCE(1487);
      END_STATE();
    case 1395:
      if (lookahead == 'n') ADVANCE(1005);
      END_STATE();
    case 1396:
      if (lookahead == 'n') ADVANCE(752);
      END_STATE();
    case 1397:
      if (lookahead == 'n') ADVANCE(1032);
      END_STATE();
    case 1398:
      if (lookahead == 'n') ADVANCE(811);
      END_STATE();
    case 1399:
      if (lookahead == 'n') ADVANCE(980);
      END_STATE();
    case 1400:
      if (lookahead == 'n') ADVANCE(722);
      END_STATE();
    case 1401:
      if (lookahead == 'n') ADVANCE(1889);
      END_STATE();
    case 1402:
      if (lookahead == 'n') ADVANCE(1822);
      END_STATE();
    case 1403:
      if (lookahead == 'n') ADVANCE(1836);
      END_STATE();
    case 1404:
      if (lookahead == 'n') ADVANCE(1861);
      END_STATE();
    case 1405:
      if (lookahead == 'n') ADVANCE(1863);
      END_STATE();
    case 1406:
      if (lookahead == 'n') ADVANCE(841);
      END_STATE();
    case 1407:
      if (lookahead == 'n') ADVANCE(844);
      END_STATE();
    case 1408:
      if (lookahead == 'n') ADVANCE(1281);
      END_STATE();
    case 1409:
      if (lookahead == 'n') ADVANCE(1920);
      END_STATE();
    case 1410:
      if (lookahead == 'n') ADVANCE(68);
      END_STATE();
    case 1411:
      if (lookahead == 'n') ADVANCE(998);
      END_STATE();
    case 1412:
      if (lookahead == 'n') ADVANCE(1001);
      END_STATE();
    case 1413:
      if (lookahead == 'n') ADVANCE(926);
      END_STATE();
    case 1414:
      if (lookahead == 'n') ADVANCE(1111);
      END_STATE();
    case 1415:
      if (lookahead == 'n') ADVANCE(1879);
      END_STATE();
    case 1416:
      if (lookahead == 'n') ADVANCE(1918);
      END_STATE();
    case 1417:
      if (lookahead == 'n') ADVANCE(1492);
      END_STATE();
    case 1418:
      if (lookahead == 'n') ADVANCE(1880);
      END_STATE();
    case 1419:
      if (lookahead == 'n') ADVANCE(1088);
      END_STATE();
    case 1420:
      if (lookahead == 'n') ADVANCE(1878);
      END_STATE();
    case 1421:
      if (lookahead == 'n') ADVANCE(1898);
      END_STATE();
    case 1422:
      if (lookahead == 'n') ADVANCE(509);
      END_STATE();
    case 1423:
      if (lookahead == 'n') ADVANCE(1890);
      END_STATE();
    case 1424:
      if (lookahead == 'n') ADVANCE(1024);
      END_STATE();
    case 1425:
      if (lookahead == 'n') ADVANCE(1888);
      END_STATE();
    case 1426:
      if (lookahead == 'n') ADVANCE(1167);
      END_STATE();
    case 1427:
      if (lookahead == 'n') ADVANCE(929);
      END_STATE();
    case 1428:
      if (lookahead == 'n') ADVANCE(1026);
      if (lookahead == 'r') ADVANCE(1706);
      END_STATE();
    case 1429:
      if (lookahead == 'n') ADVANCE(1026);
      if (lookahead == 'r') ADVANCE(1710);
      END_STATE();
    case 1430:
      if (lookahead == 'n') ADVANCE(1892);
      END_STATE();
    case 1431:
      if (lookahead == 'n') ADVANCE(1027);
      END_STATE();
    case 1432:
      if (lookahead == 'n') ADVANCE(1028);
      END_STATE();
    case 1433:
      if (lookahead == 'n') ADVANCE(1894);
      END_STATE();
    case 1434:
      if (lookahead == 'n') ADVANCE(1029);
      END_STATE();
    case 1435:
      if (lookahead == 'n') ADVANCE(1030);
      END_STATE();
    case 1436:
      if (lookahead == 'n') ADVANCE(1031);
      END_STATE();
    case 1437:
      if (lookahead == 'n') ADVANCE(1899);
      END_STATE();
    case 1438:
      if (lookahead == 'n') ADVANCE(558);
      END_STATE();
    case 1439:
      if (lookahead == 'n') ADVANCE(547);
      END_STATE();
    case 1440:
      if (lookahead == 'o') ADVANCE(734);
      END_STATE();
    case 1441:
      if (lookahead == 'o') ADVANCE(13);
      END_STATE();
    case 1442:
      if (lookahead == 'o') ADVANCE(191);
      END_STATE();
    case 1443:
      if (lookahead == 'o') ADVANCE(949);
      END_STATE();
    case 1444:
      if (lookahead == 'o') ADVANCE(1980);
      END_STATE();
    case 1445:
      if (lookahead == 'o') ADVANCE(379);
      END_STATE();
    case 1446:
      if (lookahead == 'o') ADVANCE(1959);
      END_STATE();
    case 1447:
      if (lookahead == 'o') ADVANCE(108);
      END_STATE();
    case 1448:
      if (lookahead == 'o') ADVANCE(1813);
      END_STATE();
    case 1449:
      if (lookahead == 'o') ADVANCE(1812);
      END_STATE();
    case 1450:
      if (lookahead == 'o') ADVANCE(1603);
      END_STATE();
    case 1451:
      if (lookahead == 'o') ADVANCE(1811);
      END_STATE();
    case 1452:
      if (lookahead == 'o') ADVANCE(1811);
      if (lookahead == 's') ADVANCE(1912);
      END_STATE();
    case 1453:
      if (lookahead == 'o') ADVANCE(171);
      END_STATE();
    case 1454:
      if (lookahead == 'o') ADVANCE(1991);
      END_STATE();
    case 1455:
      if (lookahead == 'o') ADVANCE(682);
      END_STATE();
    case 1456:
      if (lookahead == 'o') ADVANCE(1989);
      END_STATE();
    case 1457:
      if (lookahead == 'o') ADVANCE(1782);
      END_STATE();
    case 1458:
      if (lookahead == 'o') ADVANCE(207);
      END_STATE();
    case 1459:
      if (lookahead == 'o') ADVANCE(1969);
      END_STATE();
    case 1460:
      if (lookahead == 'o') ADVANCE(2009);
      END_STATE();
    case 1461:
      if (lookahead == 'o') ADVANCE(1530);
      END_STATE();
    case 1462:
      if (lookahead == 'o') ADVANCE(1419);
      END_STATE();
    case 1463:
      if (lookahead == 'o') ADVANCE(1970);
      END_STATE();
    case 1464:
      if (lookahead == 'o') ADVANCE(1971);
      END_STATE();
    case 1465:
      if (lookahead == 'o') ADVANCE(1316);
      END_STATE();
    case 1466:
      if (lookahead == 'o') ADVANCE(1192);
      END_STATE();
    case 1467:
      if (lookahead == 'o') ADVANCE(1866);
      END_STATE();
    case 1468:
      if (lookahead == 'o') ADVANCE(1972);
      END_STATE();
    case 1469:
      if (lookahead == 'o') ADVANCE(1985);
      END_STATE();
    case 1470:
      if (lookahead == 'o') ADVANCE(1756);
      END_STATE();
    case 1471:
      if (lookahead == 'o') ADVANCE(1378);
      END_STATE();
    case 1472:
      if (lookahead == 'o') ADVANCE(1378);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1473:
      if (lookahead == 'o') ADVANCE(1986);
      END_STATE();
    case 1474:
      if (lookahead == 'o') ADVANCE(1978);
      END_STATE();
    case 1475:
      if (lookahead == 'o') ADVANCE(1973);
      END_STATE();
    case 1476:
      if (lookahead == 'o') ADVANCE(1211);
      if (lookahead == 'u') ADVANCE(571);
      END_STATE();
    case 1477:
      if (lookahead == 'o') ADVANCE(757);
      END_STATE();
    case 1478:
      if (lookahead == 'o') ADVANCE(1385);
      END_STATE();
    case 1479:
      if (lookahead == 'o') ADVANCE(1976);
      END_STATE();
    case 1480:
      if (lookahead == 'o') ADVANCE(1817);
      END_STATE();
    case 1481:
      if (lookahead == 'o') ADVANCE(1657);
      END_STATE();
    case 1482:
      if (lookahead == 'o') ADVANCE(1357);
      END_STATE();
    case 1483:
      if (lookahead == 'o') ADVANCE(1357);
      if (lookahead == 'r') ADVANCE(417);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 1484:
      if (lookahead == 'o') ADVANCE(1377);
      END_STATE();
    case 1485:
      if (lookahead == 'o') ADVANCE(1413);
      END_STATE();
    case 1486:
      if (lookahead == 'o') ADVANCE(1369);
      END_STATE();
    case 1487:
      if (lookahead == 'o') ADVANCE(1955);
      END_STATE();
    case 1488:
      if (lookahead == 'o') ADVANCE(1356);
      END_STATE();
    case 1489:
      if (lookahead == 'o') ADVANCE(1384);
      END_STATE();
    case 1490:
      if (lookahead == 'o') ADVANCE(1957);
      END_STATE();
    case 1491:
      if (lookahead == 'o') ADVANCE(1450);
      END_STATE();
    case 1492:
      if (lookahead == 'o') ADVANCE(1903);
      END_STATE();
    case 1493:
      if (lookahead == 'o') ADVANCE(1210);
      END_STATE();
    case 1494:
      if (lookahead == 'o') ADVANCE(681);
      END_STATE();
    case 1495:
      if (lookahead == 'o') ADVANCE(1852);
      END_STATE();
    case 1496:
      if (lookahead == 'o') ADVANCE(1854);
      END_STATE();
    case 1497:
      if (lookahead == 'o') ADVANCE(1896);
      END_STATE();
    case 1498:
      if (lookahead == 'o') ADVANCE(1408);
      END_STATE();
    case 1499:
      if (lookahead == 'o') ADVANCE(1410);
      END_STATE();
    case 1500:
      if (lookahead == 'o') ADVANCE(1376);
      END_STATE();
    case 1501:
      if (lookahead == 'o') ADVANCE(1669);
      if (lookahead == 'r') ADVANCE(1461);
      END_STATE();
    case 1502:
      if (lookahead == 'o') ADVANCE(1694);
      END_STATE();
    case 1503:
      if (lookahead == 'o') ADVANCE(1696);
      if (lookahead == 'r') ADVANCE(1461);
      END_STATE();
    case 1504:
      if (lookahead == 'o') ADVANCE(1620);
      END_STATE();
    case 1505:
      if (lookahead == 'o') ADVANCE(1981);
      END_STATE();
    case 1506:
      if (lookahead == 'o') ADVANCE(1332);
      END_STATE();
    case 1507:
      if (lookahead == 'o') ADVANCE(1332);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1508:
      if (lookahead == 'o') ADVANCE(1675);
      END_STATE();
    case 1509:
      if (lookahead == 'o') ADVANCE(1559);
      END_STATE();
    case 1510:
      if (lookahead == 'o') ADVANCE(1570);
      END_STATE();
    case 1511:
      if (lookahead == 'o') ADVANCE(1939);
      END_STATE();
    case 1512:
      if (lookahead == 'o') ADVANCE(1987);
      END_STATE();
    case 1513:
      if (lookahead == 'o') ADVANCE(1982);
      END_STATE();
    case 1514:
      if (lookahead == 'o') ADVANCE(1945);
      END_STATE();
    case 1515:
      if (lookahead == 'o') ADVANCE(1860);
      END_STATE();
    case 1516:
      if (lookahead == 'o') ADVANCE(1412);
      if (lookahead == 'u') ADVANCE(1550);
      END_STATE();
    case 1517:
      if (lookahead == 'o') ADVANCE(1983);
      END_STATE();
    case 1518:
      if (lookahead == 'o') ADVANCE(1425);
      END_STATE();
    case 1519:
      if (lookahead == 'o') ADVANCE(1690);
      END_STATE();
    case 1520:
      if (lookahead == 'o') ADVANCE(1498);
      END_STATE();
    case 1521:
      if (lookahead == 'o') ADVANCE(1671);
      END_STATE();
    case 1522:
      if (lookahead == 'o') ADVANCE(1404);
      END_STATE();
    case 1523:
      if (lookahead == 'o') ADVANCE(1478);
      END_STATE();
    case 1524:
      if (lookahead == 'o') ADVANCE(1500);
      END_STATE();
    case 1525:
      if (lookahead == 'o') ADVANCE(1736);
      END_STATE();
    case 1526:
      if (lookahead == 'o') ADVANCE(1307);
      END_STATE();
    case 1527:
      if (lookahead == 'o') ADVANCE(1960);
      END_STATE();
    case 1528:
      if (lookahead == 'o') ADVANCE(1988);
      END_STATE();
    case 1529:
      if (lookahead == 'p') ADVANCE(1807);
      END_STATE();
    case 1530:
      if (lookahead == 'p') ADVANCE(13);
      END_STATE();
    case 1531:
      if (lookahead == 'p') ADVANCE(13);
      if (lookahead == 'r') ADVANCE(1482);
      END_STATE();
    case 1532:
      if (lookahead == 'p') ADVANCE(1039);
      END_STATE();
    case 1533:
      if (lookahead == 'p') ADVANCE(949);
      END_STATE();
    case 1534:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'r') ADVANCE(528);
      END_STATE();
    case 1535:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'r') ADVANCE(270);
      if (lookahead == 'u') ADVANCE(1687);
      END_STATE();
    case 1536:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'r') ADVANCE(1081);
      END_STATE();
    case 1537:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(26);
      if (lookahead == 'u') ADVANCE(596);
      if (lookahead == 'w') ADVANCE(1359);
      END_STATE();
    case 1538:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 't') ADVANCE(215);
      if (lookahead == 'w') ADVANCE(1872);
      if (lookahead == 'x') ADVANCE(304);
      END_STATE();
    case 1539:
      if (lookahead == 'p') ADVANCE(949);
      if (lookahead == 'u') ADVANCE(1386);
      END_STATE();
    case 1540:
      if (lookahead == 'p') ADVANCE(25);
      END_STATE();
    case 1541:
      if (lookahead == 'p') ADVANCE(312);
      END_STATE();
    case 1542:
      if (lookahead == 'p') ADVANCE(32);
      END_STATE();
    case 1543:
      if (lookahead == 'p') ADVANCE(7);
      END_STATE();
    case 1544:
      if (lookahead == 'p') ADVANCE(313);
      END_STATE();
    case 1545:
      if (lookahead == 'p') ADVANCE(273);
      END_STATE();
    case 1546:
      if (lookahead == 'p') ADVANCE(1603);
      if (lookahead == 's') ADVANCE(615);
      END_STATE();
    case 1547:
      if (lookahead == 'p') ADVANCE(385);
      END_STATE();
    case 1548:
      if (lookahead == 'p') ADVANCE(185);
      END_STATE();
    case 1549:
      if (lookahead == 'p') ADVANCE(272);
      END_STATE();
    case 1550:
      if (lookahead == 'p') ADVANCE(287);
      END_STATE();
    case 1551:
      if (lookahead == 'p') ADVANCE(111);
      END_STATE();
    case 1552:
      if (lookahead == 'p') ADVANCE(195);
      END_STATE();
    case 1553:
      if (lookahead == 'p') ADVANCE(1036);
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 1554:
      if (lookahead == 'p') ADVANCE(1547);
      END_STATE();
    case 1555:
      if (lookahead == 'p') ADVANCE(1756);
      if (lookahead == 'r') ADVANCE(1482);
      END_STATE();
    case 1556:
      if (lookahead == 'p') ADVANCE(120);
      END_STATE();
    case 1557:
      if (lookahead == 'p') ADVANCE(1222);
      END_STATE();
    case 1558:
      if (lookahead == 'p') ADVANCE(1567);
      END_STATE();
    case 1559:
      if (lookahead == 'p') ADVANCE(779);
      END_STATE();
    case 1560:
      if (lookahead == 'p') ADVANCE(827);
      END_STATE();
    case 1561:
      if (lookahead == 'p') ADVANCE(1301);
      END_STATE();
    case 1562:
      if (lookahead == 'p') ADVANCE(1301);
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 1563:
      if (lookahead == 'p') ADVANCE(1779);
      END_STATE();
    case 1564:
      if (lookahead == 'p') ADVANCE(1851);
      END_STATE();
    case 1565:
      if (lookahead == 'p') ADVANCE(1577);
      END_STATE();
    case 1566:
      if (lookahead == 'p') ADVANCE(1577);
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 1567:
      if (lookahead == 'p') ADVANCE(405);
      END_STATE();
    case 1568:
      if (lookahead == 'p') ADVANCE(1520);
      END_STATE();
    case 1569:
      if (lookahead == 'p') ADVANCE(1303);
      END_STATE();
    case 1570:
      if (lookahead == 'p') ADVANCE(1843);
      END_STATE();
    case 1571:
      if (lookahead == 'p') ADVANCE(1470);
      END_STATE();
    case 1572:
      if (lookahead == 'p') ADVANCE(1267);
      END_STATE();
    case 1573:
      if (lookahead == 'p') ADVANCE(525);
      END_STATE();
    case 1574:
      if (lookahead == 'p') ADVANCE(1797);
      END_STATE();
    case 1575:
      if (lookahead == 'p') ADVANCE(1702);
      if (lookahead == 's') ADVANCE(1930);
      END_STATE();
    case 1576:
      if (lookahead == 'p') ADVANCE(1279);
      END_STATE();
    case 1577:
      if (lookahead == 'p') ADVANCE(1682);
      END_STATE();
    case 1578:
      if (lookahead == 'p') ADVANCE(507);
      END_STATE();
    case 1579:
      if (lookahead == 'p') ADVANCE(556);
      END_STATE();
    case 1580:
      if (lookahead == 'p') ADVANCE(324);
      END_STATE();
    case 1581:
      if (lookahead == 'p') ADVANCE(1524);
      END_STATE();
    case 1582:
      if (lookahead == 'p') ADVANCE(540);
      END_STATE();
    case 1583:
      if (lookahead == 'p') ADVANCE(1523);
      END_STATE();
    case 1584:
      if (lookahead == 'p') ADVANCE(558);
      END_STATE();
    case 1585:
      if (lookahead == 'q') ADVANCE(13);
      END_STATE();
    case 1586:
      if (lookahead == 'q') ADVANCE(101);
      END_STATE();
    case 1587:
      if (lookahead == 'q') ADVANCE(1272);
      END_STATE();
    case 1588:
      if (lookahead == 'q') ADVANCE(1010);
      END_STATE();
    case 1589:
      if (lookahead == 'q') ADVANCE(1575);
      END_STATE();
    case 1590:
      if (lookahead == 'q') ADVANCE(186);
      END_STATE();
    case 1591:
      if (lookahead == 'q') ADVANCE(1585);
      END_STATE();
    case 1592:
      if (lookahead == 'q') ADVANCE(1906);
      END_STATE();
    case 1593:
      if (lookahead == 'q') ADVANCE(1925);
      END_STATE();
    case 1594:
      if (lookahead == 'q') ADVANCE(1958);
      if (lookahead == 'u') ADVANCE(594);
      END_STATE();
    case 1595:
      if (lookahead == 'q') ADVANCE(1946);
      END_STATE();
    case 1596:
      if (lookahead == 'q') ADVANCE(1935);
      END_STATE();
    case 1597:
      if (lookahead == 'q') ADVANCE(700);
      if (lookahead == 't') ADVANCE(399);
      END_STATE();
    case 1598:
      if (lookahead == 'q') ADVANCE(1949);
      END_STATE();
    case 1599:
      if (lookahead == 'q') ADVANCE(1953);
      END_STATE();
    case 1600:
      if (lookahead == 'q') ADVANCE(1956);
      END_STATE();
    case 1601:
      if (lookahead == 'r') ADVANCE(734);
      END_STATE();
    case 1602:
      if (lookahead == 'r') ADVANCE(788);
      END_STATE();
    case 1603:
      if (lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 1604:
      if (lookahead == 'r') ADVANCE(294);
      END_STATE();
    case 1605:
      if (lookahead == 'r') ADVANCE(170);
      END_STATE();
    case 1606:
      if (lookahead == 'r') ADVANCE(949);
      END_STATE();
    case 1607:
      if (lookahead == 'r') ADVANCE(89);
      END_STATE();
    case 1608:
      if (lookahead == 'r') ADVANCE(183);
      END_STATE();
    case 1609:
      if (lookahead == 'r') ADVANCE(60);
      END_STATE();
    case 1610:
      if (lookahead == 'r') ADVANCE(658);
      END_STATE();
    case 1611:
      if (lookahead == 'r') ADVANCE(279);
      END_STATE();
    case 1612:
      if (lookahead == 'r') ADVANCE(2021);
      END_STATE();
    case 1613:
      if (lookahead == 'r') ADVANCE(135);
      END_STATE();
    case 1614:
      if (lookahead == 'r') ADVANCE(85);
      END_STATE();
    case 1615:
      if (lookahead == 'r') ADVANCE(1049);
      END_STATE();
    case 1616:
      if (lookahead == 'r') ADVANCE(61);
      END_STATE();
    case 1617:
      if (lookahead == 'r') ADVANCE(350);
      END_STATE();
    case 1618:
      if (lookahead == 'r') ADVANCE(225);
      END_STATE();
    case 1619:
      if (lookahead == 'r') ADVANCE(2025);
      END_STATE();
    case 1620:
      if (lookahead == 'r') ADVANCE(18);
      END_STATE();
    case 1621:
      if (lookahead == 'r') ADVANCE(36);
      END_STATE();
    case 1622:
      if (lookahead == 'r') ADVANCE(1603);
      END_STATE();
    case 1623:
      if (lookahead == 'r') ADVANCE(176);
      END_STATE();
    case 1624:
      if (lookahead == 'r') ADVANCE(332);
      END_STATE();
    case 1625:
      if (lookahead == 'r') ADVANCE(608);
      END_STATE();
    case 1626:
      if (lookahead == 'r') ADVANCE(219);
      END_STATE();
    case 1627:
      if (lookahead == 'r') ADVANCE(143);
      END_STATE();
    case 1628:
      if (lookahead == 'r') ADVANCE(1811);
      END_STATE();
    case 1629:
      if (lookahead == 'r') ADVANCE(1811);
      if (lookahead == 'u') ADVANCE(542);
      END_STATE();
    case 1630:
      if (lookahead == 'r') ADVANCE(104);
      END_STATE();
    case 1631:
      if (lookahead == 'r') ADVANCE(761);
      END_STATE();
    case 1632:
      if (lookahead == 'r') ADVANCE(79);
      END_STATE();
    case 1633:
      if (lookahead == 'r') ADVANCE(88);
      END_STATE();
    case 1634:
      if (lookahead == 'r') ADVANCE(75);
      END_STATE();
    case 1635:
      if (lookahead == 'r') ADVANCE(344);
      END_STATE();
    case 1636:
      if (lookahead == 'r') ADVANCE(607);
      END_STATE();
    case 1637:
      if (lookahead == 'r') ADVANCE(436);
      END_STATE();
    case 1638:
      if (lookahead == 'r') ADVANCE(1076);
      END_STATE();
    case 1639:
      if (lookahead == 'r') ADVANCE(178);
      END_STATE();
    case 1640:
      if (lookahead == 'r') ADVANCE(161);
      END_STATE();
    case 1641:
      if (lookahead == 'r') ADVANCE(167);
      END_STATE();
    case 1642:
      if (lookahead == 'r') ADVANCE(737);
      END_STATE();
    case 1643:
      if (lookahead == 'r') ADVANCE(729);
      END_STATE();
    case 1644:
      if (lookahead == 'r') ADVANCE(1441);
      END_STATE();
    case 1645:
      if (lookahead == 'r') ADVANCE(172);
      END_STATE();
    case 1646:
      if (lookahead == 'r') ADVANCE(1202);
      END_STATE();
    case 1647:
      if (lookahead == 'r') ADVANCE(201);
      END_STATE();
    case 1648:
      if (lookahead == 'r') ADVANCE(290);
      END_STATE();
    case 1649:
      if (lookahead == 'r') ADVANCE(212);
      END_STATE();
    case 1650:
      if (lookahead == 'r') ADVANCE(719);
      END_STATE();
    case 1651:
      if (lookahead == 'r') ADVANCE(1530);
      END_STATE();
    case 1652:
      if (lookahead == 'r') ADVANCE(856);
      END_STATE();
    case 1653:
      if (lookahead == 'r') ADVANCE(1871);
      END_STATE();
    case 1654:
      if (lookahead == 'r') ADVANCE(175);
      END_STATE();
    case 1655:
      if (lookahead == 'r') ADVANCE(390);
      END_STATE();
    case 1656:
      if (lookahead == 'r') ADVANCE(1482);
      END_STATE();
    case 1657:
      if (lookahead == 'r') ADVANCE(1192);
      END_STATE();
    case 1658:
      if (lookahead == 'r') ADVANCE(1999);
      END_STATE();
    case 1659:
      if (lookahead == 'r') ADVANCE(2001);
      END_STATE();
    case 1660:
      if (lookahead == 'r') ADVANCE(1445);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 1661:
      if (lookahead == 'r') ADVANCE(1756);
      END_STATE();
    case 1662:
      if (lookahead == 'r') ADVANCE(849);
      END_STATE();
    case 1663:
      if (lookahead == 'r') ADVANCE(1440);
      END_STATE();
    case 1664:
      if (lookahead == 'r') ADVANCE(417);
      END_STATE();
    case 1665:
      if (lookahead == 'r') ADVANCE(1977);
      END_STATE();
    case 1666:
      if (lookahead == 'r') ADVANCE(2000);
      END_STATE();
    case 1667:
      if (lookahead == 'r') ADVANCE(907);
      END_STATE();
    case 1668:
      if (lookahead == 'r') ADVANCE(1048);
      END_STATE();
    case 1669:
      if (lookahead == 'r') ADVANCE(1357);
      END_STATE();
    case 1670:
      if (lookahead == 'r') ADVANCE(615);
      END_STATE();
    case 1671:
      if (lookahead == 'r') ADVANCE(779);
      END_STATE();
    case 1672:
      if (lookahead == 'r') ADVANCE(1461);
      END_STATE();
    case 1673:
      if (lookahead == 'r') ADVANCE(1153);
      END_STATE();
    case 1674:
      if (lookahead == 'r') ADVANCE(1466);
      END_STATE();
    case 1675:
      if (lookahead == 'r') ADVANCE(1819);
      END_STATE();
    case 1676:
      if (lookahead == 'r') ADVANCE(1454);
      END_STATE();
    case 1677:
      if (lookahead == 'r') ADVANCE(1776);
      END_STATE();
    case 1678:
      if (lookahead == 'r') ADVANCE(688);
      END_STATE();
    case 1679:
      if (lookahead == 'r') ADVANCE(1459);
      END_STATE();
    case 1680:
      if (lookahead == 'r') ADVANCE(1612);
      END_STATE();
    case 1681:
      if (lookahead == 'r') ADVANCE(453);
      END_STATE();
    case 1682:
      if (lookahead == 'r') ADVANCE(1456);
      END_STATE();
    case 1683:
      if (lookahead == 'r') ADVANCE(1889);
      END_STATE();
    case 1684:
      if (lookahead == 'r') ADVANCE(1568);
      END_STATE();
    case 1685:
      if (lookahead == 'r') ADVANCE(1882);
      END_STATE();
    case 1686:
      if (lookahead == 'r') ADVANCE(1798);
      END_STATE();
    case 1687:
      if (lookahead == 'r') ADVANCE(1162);
      END_STATE();
    case 1688:
      if (lookahead == 'r') ADVANCE(1806);
      END_STATE();
    case 1689:
      if (lookahead == 'r') ADVANCE(832);
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 1690:
      if (lookahead == 'r') ADVANCE(1835);
      END_STATE();
    case 1691:
      if (lookahead == 'r') ADVANCE(456);
      END_STATE();
    case 1692:
      if (lookahead == 'r') ADVANCE(1124);
      END_STATE();
    case 1693:
      if (lookahead == 'r') ADVANCE(1463);
      END_STATE();
    case 1694:
      if (lookahead == 'r') ADVANCE(1398);
      END_STATE();
    case 1695:
      if (lookahead == 'r') ADVANCE(1784);
      END_STATE();
    case 1696:
      if (lookahead == 'r') ADVANCE(1387);
      END_STATE();
    case 1697:
      if (lookahead == 'r') ADVANCE(1102);
      END_STATE();
    case 1698:
      if (lookahead == 'r') ADVANCE(1426);
      END_STATE();
    case 1699:
      if (lookahead == 'r') ADVANCE(1250);
      END_STATE();
    case 1700:
      if (lookahead == 'r') ADVANCE(1159);
      END_STATE();
    case 1701:
      if (lookahead == 'r') ADVANCE(1254);
      END_STATE();
    case 1702:
      if (lookahead == 'r') ADVANCE(854);
      END_STATE();
    case 1703:
      if (lookahead == 'r') ADVANCE(1464);
      END_STATE();
    case 1704:
      if (lookahead == 'r') ADVANCE(1815);
      END_STATE();
    case 1705:
      if (lookahead == 'r') ADVANCE(796);
      if (lookahead == 't') ADVANCE(416);
      END_STATE();
    case 1706:
      if (lookahead == 'r') ADVANCE(1468);
      END_STATE();
    case 1707:
      if (lookahead == 'r') ADVANCE(1469);
      END_STATE();
    case 1708:
      if (lookahead == 'r') ADVANCE(1473);
      END_STATE();
    case 1709:
      if (lookahead == 'r') ADVANCE(1474);
      END_STATE();
    case 1710:
      if (lookahead == 'r') ADVANCE(1475);
      END_STATE();
    case 1711:
      if (lookahead == 'r') ADVANCE(947);
      END_STATE();
    case 1712:
      if (lookahead == 'r') ADVANCE(799);
      END_STATE();
    case 1713:
      if (lookahead == 'r') ADVANCE(1479);
      END_STATE();
    case 1714:
      if (lookahead == 'r') ADVANCE(1618);
      END_STATE();
    case 1715:
      if (lookahead == 'r') ADVANCE(903);
      END_STATE();
    case 1716:
      if (lookahead == 'r') ADVANCE(1608);
      END_STATE();
    case 1717:
      if (lookahead == 'r') ADVANCE(1641);
      END_STATE();
    case 1718:
      if (lookahead == 'r') ADVANCE(922);
      END_STATE();
    case 1719:
      if (lookahead == 'r') ADVANCE(824);
      END_STATE();
    case 1720:
      if (lookahead == 'r') ADVANCE(546);
      END_STATE();
    case 1721:
      if (lookahead == 'r') ADVANCE(461);
      END_STATE();
    case 1722:
      if (lookahead == 'r') ADVANCE(506);
      END_STATE();
    case 1723:
      if (lookahead == 'r') ADVANCE(1134);
      END_STATE();
    case 1724:
      if (lookahead == 'r') ADVANCE(1799);
      END_STATE();
    case 1725:
      if (lookahead == 'r') ADVANCE(700);
      END_STATE();
    case 1726:
      if (lookahead == 'r') ADVANCE(1879);
      END_STATE();
    case 1727:
      if (lookahead == 'r') ADVANCE(870);
      END_STATE();
    case 1728:
      if (lookahead == 'r') ADVANCE(1000);
      if (lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 1729:
      if (lookahead == 'r') ADVANCE(1862);
      END_STATE();
    case 1730:
      if (lookahead == 'r') ADVANCE(1924);
      END_STATE();
    case 1731:
      if (lookahead == 'r') ADVANCE(1801);
      END_STATE();
    case 1732:
      if (lookahead == 'r') ADVANCE(877);
      END_STATE();
    case 1733:
      if (lookahead == 'r') ADVANCE(894);
      END_STATE();
    case 1734:
      if (lookahead == 'r') ADVANCE(1679);
      END_STATE();
    case 1735:
      if (lookahead == 'r') ADVANCE(485);
      END_STATE();
    case 1736:
      if (lookahead == 'r') ADVANCE(1877);
      END_STATE();
    case 1737:
      if (lookahead == 'r') ADVANCE(1693);
      END_STATE();
    case 1738:
      if (lookahead == 'r') ADVANCE(1703);
      END_STATE();
    case 1739:
      if (lookahead == 'r') ADVANCE(1707);
      END_STATE();
    case 1740:
      if (lookahead == 'r') ADVANCE(1708);
      END_STATE();
    case 1741:
      if (lookahead == 'r') ADVANCE(1709);
      END_STATE();
    case 1742:
      if (lookahead == 'r') ADVANCE(1713);
      END_STATE();
    case 1743:
      if (lookahead == 'r') ADVANCE(912);
      END_STATE();
    case 1744:
      if (lookahead == 'r') ADVANCE(724);
      END_STATE();
    case 1745:
      if (lookahead == 'r') ADVANCE(935);
      END_STATE();
    case 1746:
      if (lookahead == 'r') ADVANCE(327);
      END_STATE();
    case 1747:
      if (lookahead == 'r') ADVANCE(1581);
      END_STATE();
    case 1748:
      if (lookahead == 'r') ADVANCE(1900);
      END_STATE();
    case 1749:
      if (lookahead == 'r') ADVANCE(1583);
      END_STATE();
    case 1750:
      if (lookahead == 'r') ADVANCE(944);
      END_STATE();
    case 1751:
      if (lookahead == 'r') ADVANCE(945);
      END_STATE();
    case 1752:
      if (lookahead == 'r') ADVANCE(946);
      END_STATE();
    case 1753:
      if (lookahead == 'r') ADVANCE(1186);
      END_STATE();
    case 1754:
      if (lookahead == 'r') ADVANCE(1188);
      END_STATE();
    case 1755:
      if (lookahead == 'r') ADVANCE(1189);
      END_STATE();
    case 1756:
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 1757:
      if (lookahead == 's') ADVANCE(904);
      END_STATE();
    case 1758:
      if (lookahead == 's') ADVANCE(183);
      if (lookahead == 'u') ADVANCE(1342);
      END_STATE();
    case 1759:
      if (lookahead == 's') ADVANCE(326);
      END_STATE();
    case 1760:
      if (lookahead == 's') ADVANCE(352);
      END_STATE();
    case 1761:
      if (lookahead == 's') ADVANCE(62);
      END_STATE();
    case 1762:
      if (lookahead == 's') ADVANCE(218);
      END_STATE();
    case 1763:
      if (lookahead == 's') ADVANCE(81);
      END_STATE();
    case 1764:
      if (lookahead == 's') ADVANCE(38);
      END_STATE();
    case 1765:
      if (lookahead == 's') ADVANCE(40);
      END_STATE();
    case 1766:
      if (lookahead == 's') ADVANCE(39);
      END_STATE();
    case 1767:
      if (lookahead == 's') ADVANCE(1811);
      END_STATE();
    case 1768:
      if (lookahead == 's') ADVANCE(346);
      END_STATE();
    case 1769:
      if (lookahead == 's') ADVANCE(1476);
      END_STATE();
    case 1770:
      if (lookahead == 's') ADVANCE(145);
      END_STATE();
    case 1771:
      if (lookahead == 's') ADVANCE(1076);
      END_STATE();
    case 1772:
      if (lookahead == 's') ADVANCE(360);
      END_STATE();
    case 1773:
      if (lookahead == 's') ADVANCE(207);
      END_STATE();
    case 1774:
      if (lookahead == 's') ADVANCE(111);
      END_STATE();
    case 1775:
      if (lookahead == 's') ADVANCE(66);
      END_STATE();
    case 1776:
      if (lookahead == 's') ADVANCE(1530);
      END_STATE();
    case 1777:
      if (lookahead == 's') ADVANCE(1036);
      END_STATE();
    case 1778:
      if (lookahead == 's') ADVANCE(2002);
      END_STATE();
    case 1779:
      if (lookahead == 's') ADVANCE(200);
      END_STATE();
    case 1780:
      if (lookahead == 's') ADVANCE(1037);
      END_STATE();
    case 1781:
      if (lookahead == 's') ADVANCE(1041);
      END_STATE();
    case 1782:
      if (lookahead == 's') ADVANCE(1756);
      END_STATE();
    case 1783:
      if (lookahead == 's') ADVANCE(1112);
      END_STATE();
    case 1784:
      if (lookahead == 's') ADVANCE(1211);
      END_STATE();
    case 1785:
      if (lookahead == 's') ADVANCE(1905);
      END_STATE();
    case 1786:
      if (lookahead == 's') ADVANCE(747);
      END_STATE();
    case 1787:
      if (lookahead == 's') ADVANCE(779);
      END_STATE();
    case 1788:
      if (lookahead == 's') ADVANCE(827);
      END_STATE();
    case 1789:
      if (lookahead == 's') ADVANCE(1954);
      END_STATE();
    case 1790:
      if (lookahead == 's') ADVANCE(1910);
      END_STATE();
    case 1791:
      if (lookahead == 's') ADVANCE(1856);
      END_STATE();
    case 1792:
      if (lookahead == 's') ADVANCE(1237);
      END_STATE();
    case 1793:
      if (lookahead == 's') ADVANCE(1768);
      END_STATE();
    case 1794:
      if (lookahead == 's') ADVANCE(1846);
      END_STATE();
    case 1795:
      if (lookahead == 's') ADVANCE(1849);
      END_STATE();
    case 1796:
      if (lookahead == 's') ADVANCE(1174);
      END_STATE();
    case 1797:
      if (lookahead == 's') ADVANCE(1843);
      END_STATE();
    case 1798:
      if (lookahead == 's') ADVANCE(864);
      END_STATE();
    case 1799:
      if (lookahead == 's') ADVANCE(803);
      END_STATE();
    case 1800:
      if (lookahead == 's') ADVANCE(834);
      END_STATE();
    case 1801:
      if (lookahead == 's') ADVANCE(823);
      END_STATE();
    case 1802:
      if (lookahead == 's') ADVANCE(1046);
      END_STATE();
    case 1803:
      if (lookahead == 's') ADVANCE(1912);
      END_STATE();
    case 1804:
      if (lookahead == 's') ADVANCE(1867);
      END_STATE();
    case 1805:
      if (lookahead == 's') ADVANCE(1855);
      END_STATE();
    case 1806:
      if (lookahead == 's') ADVANCE(869);
      END_STATE();
    case 1807:
      if (lookahead == 's') ADVANCE(1121);
      END_STATE();
    case 1808:
      if (lookahead == 's') ADVANCE(884);
      END_STATE();
    case 1809:
      if (lookahead == 's') ADVANCE(940);
      END_STATE();
    case 1810:
      if (lookahead == 't') ADVANCE(1992);
      END_STATE();
    case 1811:
      if (lookahead == 't') ADVANCE(13);
      END_STATE();
    case 1812:
      if (lookahead == 't') ADVANCE(13);
      if (lookahead == 'u') ADVANCE(604);
      END_STATE();
    case 1813:
      if (lookahead == 't') ADVANCE(13);
      if (lookahead == 'w') ADVANCE(1373);
      END_STATE();
    case 1814:
      if (lookahead == 't') ADVANCE(253);
      END_STATE();
    case 1815:
      if (lookahead == 't') ADVANCE(368);
      END_STATE();
    case 1816:
      if (lookahead == 't') ADVANCE(554);
      END_STATE();
    case 1817:
      if (lookahead == 't') ADVANCE(224);
      END_STATE();
    case 1818:
      if (lookahead == 't') ADVANCE(271);
      END_STATE();
    case 1819:
      if (lookahead == 't') ADVANCE(309);
      END_STATE();
    case 1820:
      if (lookahead == 't') ADVANCE(553);
      END_STATE();
    case 1821:
      if (lookahead == 't') ADVANCE(126);
      END_STATE();
    case 1822:
      if (lookahead == 't') ADVANCE(1009);
      END_STATE();
    case 1823:
      if (lookahead == 't') ADVANCE(364);
      END_STATE();
    case 1824:
      if (lookahead == 't') ADVANCE(374);
      END_STATE();
    case 1825:
      if (lookahead == 't') ADVANCE(274);
      END_STATE();
    case 1826:
      if (lookahead == 't') ADVANCE(277);
      END_STATE();
    case 1827:
      if (lookahead == 't') ADVANCE(378);
      END_STATE();
    case 1828:
      if (lookahead == 't') ADVANCE(561);
      END_STATE();
    case 1829:
      if (lookahead == 't') ADVANCE(275);
      END_STATE();
    case 1830:
      if (lookahead == 't') ADVANCE(276);
      END_STATE();
    case 1831:
      if (lookahead == 't') ADVANCE(1603);
      END_STATE();
    case 1832:
      if (lookahead == 't') ADVANCE(385);
      END_STATE();
    case 1833:
      if (lookahead == 't') ADVANCE(861);
      END_STATE();
    case 1834:
      if (lookahead == 't') ADVANCE(35);
      END_STATE();
    case 1835:
      if (lookahead == 't') ADVANCE(1331);
      END_STATE();
    case 1836:
      if (lookahead == 't') ADVANCE(325);
      END_STATE();
    case 1837:
      if (lookahead == 't') ADVANCE(607);
      END_STATE();
    case 1838:
      if (lookahead == 't') ADVANCE(1993);
      END_STATE();
    case 1839:
      if (lookahead == 't') ADVANCE(117);
      if (lookahead == 'v') ADVANCE(1148);
      END_STATE();
    case 1840:
      if (lookahead == 't') ADVANCE(221);
      END_STATE();
    case 1841:
      if (lookahead == 't') ADVANCE(272);
      END_STATE();
    case 1842:
      if (lookahead == 't') ADVANCE(550);
      END_STATE();
    case 1843:
      if (lookahead == 't') ADVANCE(1441);
      END_STATE();
    case 1844:
      if (lookahead == 't') ADVANCE(888);
      END_STATE();
    case 1845:
      if (lookahead == 't') ADVANCE(375);
      END_STATE();
    case 1846:
      if (lookahead == 't') ADVANCE(73);
      END_STATE();
    case 1847:
      if (lookahead == 't') ADVANCE(1997);
      END_STATE();
    case 1848:
      if (lookahead == 't') ADVANCE(111);
      END_STATE();
    case 1849:
      if (lookahead == 't') ADVANCE(195);
      END_STATE();
    case 1850:
      if (lookahead == 't') ADVANCE(1036);
      END_STATE();
    case 1851:
      if (lookahead == 't') ADVANCE(1998);
      END_STATE();
    case 1852:
      if (lookahead == 't') ADVANCE(174);
      END_STATE();
    case 1853:
      if (lookahead == 't') ADVANCE(1041);
      END_STATE();
    case 1854:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 1855:
      if (lookahead == 't') ADVANCE(1756);
      END_STATE();
    case 1856:
      if (lookahead == 't') ADVANCE(120);
      END_STATE();
    case 1857:
      if (lookahead == 't') ADVANCE(125);
      END_STATE();
    case 1858:
      if (lookahead == 't') ADVANCE(1038);
      END_STATE();
    case 1859:
      if (lookahead == 't') ADVANCE(1339);
      END_STATE();
    case 1860:
      if (lookahead == 't') ADVANCE(779);
      END_STATE();
    case 1861:
      if (lookahead == 't') ADVANCE(1511);
      END_STATE();
    case 1862:
      if (lookahead == 't') ADVANCE(1762);
      END_STATE();
    case 1863:
      if (lookahead == 't') ADVANCE(1066);
      END_STATE();
    case 1864:
      if (lookahead == 't') ADVANCE(1354);
      END_STATE();
    case 1865:
      if (lookahead == 't') ADVANCE(811);
      END_STATE();
    case 1866:
      if (lookahead == 't') ADVANCE(1788);
      END_STATE();
    case 1867:
      if (lookahead == 't') ADVANCE(852);
      END_STATE();
    case 1868:
      if (lookahead == 't') ADVANCE(1450);
      END_STATE();
    case 1869:
      if (lookahead == 't') ADVANCE(1638);
      END_STATE();
    case 1870:
      if (lookahead == 't') ADVANCE(487);
      END_STATE();
    case 1871:
      if (lookahead == 't') ADVANCE(1147);
      END_STATE();
    case 1872:
      if (lookahead == 't') ADVANCE(1124);
      END_STATE();
    case 1873:
      if (lookahead == 't') ADVANCE(1158);
      END_STATE();
    case 1874:
      if (lookahead == 't') ADVANCE(1447);
      END_STATE();
    case 1875:
      if (lookahead == 't') ADVANCE(1149);
      END_STATE();
    case 1876:
      if (lookahead == 't') ADVANCE(1172);
      END_STATE();
    case 1877:
      if (lookahead == 't') ADVANCE(1166);
      END_STATE();
    case 1878:
      if (lookahead == 't') ADVANCE(1126);
      END_STATE();
    case 1879:
      if (lookahead == 't') ADVANCE(1606);
      END_STATE();
    case 1880:
      if (lookahead == 't') ADVANCE(938);
      END_STATE();
    case 1881:
      if (lookahead == 't') ADVANCE(559);
      END_STATE();
    case 1882:
      if (lookahead == 't') ADVANCE(1406);
      END_STATE();
    case 1883:
      if (lookahead == 't') ADVANCE(1111);
      END_STATE();
    case 1884:
      if (lookahead == 't') ADVANCE(1504);
      END_STATE();
    case 1885:
      if (lookahead == 't') ADVANCE(1407);
      END_STATE();
    case 1886:
      if (lookahead == 't') ADVANCE(1691);
      END_STATE();
    case 1887:
      if (lookahead == 't') ADVANCE(887);
      END_STATE();
    case 1888:
      if (lookahead == 't') ADVANCE(495);
      END_STATE();
    case 1889:
      if (lookahead == 't') ADVANCE(1088);
      END_STATE();
    case 1890:
      if (lookahead == 't') ADVANCE(889);
      END_STATE();
    case 1891:
      if (lookahead == 't') ADVANCE(891);
      END_STATE();
    case 1892:
      if (lookahead == 't') ADVANCE(893);
      END_STATE();
    case 1893:
      if (lookahead == 't') ADVANCE(896);
      END_STATE();
    case 1894:
      if (lookahead == 't') ADVANCE(914);
      END_STATE();
    case 1895:
      if (lookahead == 't') ADVANCE(898);
      END_STATE();
    case 1896:
      if (lookahead == 't') ADVANCE(1067);
      END_STATE();
    case 1897:
      if (lookahead == 't') ADVANCE(538);
      END_STATE();
    case 1898:
      if (lookahead == 't') ADVANCE(1157);
      END_STATE();
    case 1899:
      if (lookahead == 't') ADVANCE(1163);
      END_STATE();
    case 1900:
      if (lookahead == 't') ADVANCE(1173);
      END_STATE();
    case 1901:
      if (lookahead == 't') ADVANCE(558);
      END_STATE();
    case 1902:
      if (lookahead == 't') ADVANCE(547);
      END_STATE();
    case 1903:
      if (lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 1904:
      if (lookahead == 'u') ADVANCE(475);
      END_STATE();
    case 1905:
      if (lookahead == 'u') ADVANCE(2015);
      END_STATE();
    case 1906:
      if (lookahead == 'u') ADVANCE(1441);
      END_STATE();
    case 1907:
      if (lookahead == 'u') ADVANCE(172);
      END_STATE();
    case 1908:
      if (lookahead == 'u') ADVANCE(577);
      END_STATE();
    case 1909:
      if (lookahead == 'u') ADVANCE(590);
      END_STATE();
    case 1910:
      if (lookahead == 'u') ADVANCE(588);
      END_STATE();
    case 1911:
      if (lookahead == 'u') ADVANCE(1530);
      END_STATE();
    case 1912:
      if (lookahead == 'u') ADVANCE(571);
      END_STATE();
    case 1913:
      if (lookahead == 'u') ADVANCE(1760);
      END_STATE();
    case 1914:
      if (lookahead == 'u') ADVANCE(595);
      END_STATE();
    case 1915:
      if (lookahead == 'u') ADVANCE(1848);
      END_STATE();
    case 1916:
      if (lookahead == 'u') ADVANCE(1848);
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1917:
      if (lookahead == 'u') ADVANCE(1316);
      END_STATE();
    case 1918:
      if (lookahead == 'u') ADVANCE(1756);
      END_STATE();
    case 1919:
      if (lookahead == 'u') ADVANCE(862);
      END_STATE();
    case 1920:
      if (lookahead == 'u') ADVANCE(1772);
      END_STATE();
    case 1921:
      if (lookahead == 'u') ADVANCE(779);
      END_STATE();
    case 1922:
      if (lookahead == 'u') ADVANCE(1319);
      END_STATE();
    case 1923:
      if (lookahead == 'u') ADVANCE(660);
      END_STATE();
    case 1924:
      if (lookahead == 'u') ADVANCE(813);
      END_STATE();
    case 1925:
      if (lookahead == 'u') ADVANCE(453);
      END_STATE();
    case 1926:
      if (lookahead == 'u') ADVANCE(1763);
      END_STATE();
    case 1927:
      if (lookahead == 'u') ADVANCE(1442);
      END_STATE();
    case 1928:
      if (lookahead == 'u') ADVANCE(1227);
      END_STATE();
    case 1929:
      if (lookahead == 'u') ADVANCE(1787);
      END_STATE();
    case 1930:
      if (lookahead == 'u') ADVANCE(694);
      END_STATE();
    case 1931:
      if (lookahead == 'u') ADVANCE(1774);
      END_STATE();
    case 1932:
      if (lookahead == 'u') ADVANCE(1095);
      END_STATE();
    case 1933:
      if (lookahead == 'u') ADVANCE(847);
      END_STATE();
    case 1934:
      if (lookahead == 'u') ADVANCE(474);
      END_STATE();
    case 1935:
      if (lookahead == 'u') ADVANCE(1160);
      END_STATE();
    case 1936:
      if (lookahead == 'u') ADVANCE(1094);
      END_STATE();
    case 1937:
      if (lookahead == 'u') ADVANCE(1634);
      END_STATE();
    case 1938:
      if (lookahead == 'u') ADVANCE(1606);
      END_STATE();
    case 1939:
      if (lookahead == 'u') ADVANCE(1635);
      END_STATE();
    case 1940:
      if (lookahead == 'u') ADVANCE(1699);
      END_STATE();
    case 1941:
      if (lookahead == 'u') ADVANCE(1701);
      END_STATE();
    case 1942:
      if (lookahead == 'u') ADVANCE(1046);
      END_STATE();
    case 1943:
      if (lookahead == 'u') ADVANCE(1329);
      END_STATE();
    case 1944:
      if (lookahead == 'u') ADVANCE(1400);
      END_STATE();
    case 1945:
      if (lookahead == 'u') ADVANCE(1794);
      END_STATE();
    case 1946:
      if (lookahead == 'u') ADVANCE(1129);
      END_STATE();
    case 1947:
      if (lookahead == 'u') ADVANCE(1253);
      END_STATE();
    case 1948:
      if (lookahead == 'u') ADVANCE(1860);
      END_STATE();
    case 1949:
      if (lookahead == 'u') ADVANCE(494);
      END_STATE();
    case 1950:
      if (lookahead == 'u') ADVANCE(1269);
      END_STATE();
    case 1951:
      if (lookahead == 'u') ADVANCE(1515);
      END_STATE();
    case 1952:
      if (lookahead == 'u') ADVANCE(1327);
      END_STATE();
    case 1953:
      if (lookahead == 'u') ADVANCE(500);
      END_STATE();
    case 1954:
      if (lookahead == 'u') ADVANCE(1715);
      END_STATE();
    case 1955:
      if (lookahead == 'u') ADVANCE(1308);
      END_STATE();
    case 1956:
      if (lookahead == 'u') ADVANCE(543);
      END_STATE();
    case 1957:
      if (lookahead == 'u') ADVANCE(601);
      END_STATE();
    case 1958:
      if (lookahead == 'u') ADVANCE(545);
      END_STATE();
    case 1959:
      if (lookahead == 'u') ADVANCE(600);
      if (lookahead == 'w') ADVANCE(1367);
      END_STATE();
    case 1960:
      if (lookahead == 'u') ADVANCE(603);
      END_STATE();
    case 1961:
      if (lookahead == 'v') ADVANCE(13);
      END_STATE();
    case 1962:
      if (lookahead == 'v') ADVANCE(13);
      if (lookahead == 'w') ADVANCE(778);
      END_STATE();
    case 1963:
      if (lookahead == 'v') ADVANCE(27);
      END_STATE();
    case 1964:
      if (lookahead == 'v') ADVANCE(111);
      END_STATE();
    case 1965:
      if (lookahead == 'v') ADVANCE(856);
      if (lookahead == 'w') ADVANCE(855);
      END_STATE();
    case 1966:
      if (lookahead == 'v') ADVANCE(779);
      END_STATE();
    case 1967:
      if (lookahead == 'v') ADVANCE(804);
      END_STATE();
    case 1968:
      if (lookahead == 'v') ADVANCE(942);
      END_STATE();
    case 1969:
      if (lookahead == 'w') ADVANCE(13);
      END_STATE();
    case 1970:
      if (lookahead == 'w') ADVANCE(19);
      END_STATE();
    case 1971:
      if (lookahead == 'w') ADVANCE(23);
      END_STATE();
    case 1972:
      if (lookahead == 'w') ADVANCE(22);
      END_STATE();
    case 1973:
      if (lookahead == 'w') ADVANCE(21);
      END_STATE();
    case 1974:
      if (lookahead == 'w') ADVANCE(1441);
      END_STATE();
    case 1975:
      if (lookahead == 'w') ADVANCE(399);
      END_STATE();
    case 1976:
      if (lookahead == 'w') ADVANCE(195);
      END_STATE();
    case 1977:
      if (lookahead == 'w') ADVANCE(819);
      END_STATE();
    case 1978:
      if (lookahead == 'w') ADVANCE(1756);
      END_STATE();
    case 1979:
      if (lookahead == 'w') ADVANCE(907);
      if (lookahead == 'a' ||
          lookahead == 'h') ADVANCE(13);
      END_STATE();
    case 1980:
      if (lookahead == 'w') ADVANCE(1357);
      END_STATE();
    case 1981:
      if (lookahead == 'w') ADVANCE(1373);
      END_STATE();
    case 1982:
      if (lookahead == 'w') ADVANCE(1371);
      END_STATE();
    case 1983:
      if (lookahead == 'w') ADVANCE(1368);
      END_STATE();
    case 1984:
      if (lookahead == 'w') ADVANCE(1155);
      END_STATE();
    case 1985:
      if (lookahead == 'w') ADVANCE(214);
      END_STATE();
    case 1986:
      if (lookahead == 'w') ADVANCE(1281);
      END_STATE();
    case 1987:
      if (lookahead == 'w') ADVANCE(1439);
      END_STATE();
    case 1988:
      if (lookahead == 'w') ADVANCE(1438);
      END_STATE();
    case 1989:
      if (lookahead == 'x') ADVANCE(13);
      END_STATE();
    case 1990:
      if (lookahead == 'x') ADVANCE(1811);
      END_STATE();
    case 1991:
      if (lookahead == 'x') ADVANCE(120);
      END_STATE();
    case 1992:
      if (lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 1993:
      if (lookahead == 'y') ADVANCE(367);
      END_STATE();
    case 1994:
      if (lookahead == 'y') ADVANCE(838);
      END_STATE();
    case 1995:
      if (lookahead == 'y') ADVANCE(314);
      END_STATE();
    case 1996:
      if (lookahead == 'y') ADVANCE(329);
      END_STATE();
    case 1997:
      if (lookahead == 'y') ADVANCE(203);
      END_STATE();
    case 1998:
      if (lookahead == 'y') ADVANCE(1961);
      END_STATE();
    case 1999:
      if (lookahead == 'y') ADVANCE(366);
      END_STATE();
    case 2000:
      if (lookahead == 'y') ADVANCE(372);
      END_STATE();
    case 2001:
      if (lookahead == 'y') ADVANCE(342);
      END_STATE();
    case 2002:
      if (lookahead == 'y') ADVANCE(1316);
      END_STATE();
    case 2003:
      if (lookahead == 'y') ADVANCE(778);
      END_STATE();
    case 2004:
      if (lookahead == 'y') ADVANCE(1756);
      END_STATE();
    case 2005:
      if (lookahead == 'y') ADVANCE(827);
      END_STATE();
    case 2006:
      if (lookahead == 'z') ADVANCE(1518);
      END_STATE();
    case 2007:
      if (lookahead == 'z') ADVANCE(430);
      END_STATE();
    case 2008:
      if (lookahead == 'z') ADVANCE(1134);
      END_STATE();
    case 2009:
      if (lookahead == 'z') ADVANCE(865);
      END_STATE();
    case 2010:
      if (lookahead == '2' ||
          lookahead == '4') ADVANCE(13);
      END_STATE();
    case 2011:
      if (lookahead == '3' ||
          lookahead == '5') ADVANCE(13);
      END_STATE();
    case 2012:
      if (lookahead == '6' ||
          lookahead == '8') ADVANCE(13);
      END_STATE();
    case 2013:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(13);
      END_STATE();
    case 2014:
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(2029);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(241);
      END_STATE();
    case 2015:
      if (lookahead == 'b' ||
          lookahead == 'p') ADVANCE(13);
      END_STATE();
    case 2016:
      if (lookahead == 'b' ||
          lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 2017:
      if (lookahead == 'd' ||
          lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 2018:
      if (lookahead == 'e' ||
          lookahead == 'k') ADVANCE(13);
      END_STATE();
    case 2019:
      if (lookahead == 'e' ||
          lookahead == 't') ADVANCE(13);
      END_STATE();
    case 2020:
      if (lookahead == 'f' ||
          lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 2021:
      if (lookahead == 'l' ||
          lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 2022:
      if (lookahead == 'o' ||
          lookahead == 'u') ADVANCE(13);
      END_STATE();
    case 2023:
      if (lookahead == 'r' ||
          lookahead == 'y') ADVANCE(13);
      END_STATE();
    case 2024:
      if (lookahead == '3' ||
          lookahead == '4') ADVANCE(13);
      END_STATE();
    case 2025:
      if (lookahead == 'e' ||
          lookahead == 'f') ADVANCE(13);
      END_STATE();
    case 2026:
      if (('a' <= lookahead && lookahead <= 'c')) ADVANCE(13);
      END_STATE();
    case 2027:
      if (lookahead == 'L' ||
          lookahead == 'R' ||
          lookahead == 'l' ||
          lookahead == 'r') ADVANCE(13);
      END_STATE();
    case 2028:
      if (('a' <= lookahead && lookahead <= 'h')) ADVANCE(13);
      END_STATE();
    case 2029:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(246);
      END_STATE();
    case 2030:
      if (eof) ADVANCE(2031);
      ADVANCE_MAP(
        '!', 2040,
        '"', 2041,
        '#', 2042,
        '$', 2043,
        '%', 2044,
        '&', 2045,
        '\'', 2047,
        '(', 2070,
        ')', 2071,
        '*', 2048,
        '+', 2049,
        ',', 2050,
        '-', 2051,
        '.', 2053,
        '/', 2054,
        ':', 2055,
        ';', 2056,
        '<', 2038,
        '=', 2057,
        '>', 2039,
        '?', 2058,
        '@', 2060,
        '[', 2035,
        '\\', 2061,
        ']', 2036,
        '^', 2063,
        '_', 2064,
        '`', 2065,
        '{', 2066,
        '|', 2067,
        '}', 2068,
        '~', 2069,
        '\t', 2078,
        ' ', 2078,
      );
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(2075);
      END_STATE();
    case 2031:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 2032:
      ACCEPT_TOKEN(sym__backslash_escape);
      END_STATE();
    case 2033:
      ACCEPT_TOKEN(sym_entity_reference);
      END_STATE();
    case 2034:
      ACCEPT_TOKEN(sym_numeric_character_reference);
      END_STATE();
    case 2035:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(380);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(381);
      END_STATE();
    case 2036:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 2037:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      if (lookahead == ']') ADVANCE(248);
      END_STATE();
    case 2038:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 2039:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 2040:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 2041:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 2042:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 2043:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 2044:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 2045:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 2046:
      ACCEPT_TOKEN(anon_sym_AMP);
      ADVANCE_MAP(
        '#', 2014,
        'A', 315,
        'B', 384,
        'C', 338,
        'D', 299,
        'E', 354,
        'F', 610,
        'G', 45,
        'H', 249,
        'I', 321,
        'J', 712,
        'K', 337,
        'L', 44,
        'M', 426,
        'N', 345,
        'O', 323,
        'P', 422,
        'Q', 376,
        'R', 280,
        'S', 334,
        'T', 335,
        'U', 403,
        'V', 307,
        'W', 664,
        'X', 952,
        'Y', 251,
        'Z', 339,
        'a', 382,
        'b', 353,
        'c', 389,
        'd', 256,
        'e', 301,
        'f', 431,
        'g', 34,
        'h', 258,
        'i', 383,
        'j', 711,
        'k', 516,
        'l', 14,
        'm', 302,
        'n', 331,
        'o', 365,
        'p', 454,
        'q', 953,
        'r', 250,
        's', 511,
        't', 406,
        'u', 257,
        'v', 254,
        'w', 663,
        'x', 619,
        'y', 421,
        'z', 512,
      );
      END_STATE();
    case 2047:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 2048:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 2049:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 2050:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 2051:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 2052:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(247);
      END_STATE();
    case 2053:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 2054:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 2055:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 2056:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 2057:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 2058:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 2059:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '>') ADVANCE(2073);
      END_STATE();
    case 2060:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 2061:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 2062:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (('!' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          ('[' <= lookahead && lookahead <= '`') ||
          ('{' <= lookahead && lookahead <= '~')) ADVANCE(2032);
      END_STATE();
    case 2063:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 2064:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2065:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 2066:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 2067:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 2068:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 2069:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 2070:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 2071:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 2072:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_GT);
      END_STATE();
    case 2073:
      ACCEPT_TOKEN(anon_sym_QMARK_GT);
      END_STATE();
    case 2074:
      ACCEPT_TOKEN(anon_sym_RBRACK_RBRACK_GT);
      END_STATE();
    case 2075:
      ACCEPT_TOKEN(aux_sym__word_token1);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          (lookahead < ' ' || '/' < lookahead) &&
          (lookahead < ':' || '@' < lookahead) &&
          (lookahead < '[' || '`' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(2075);
      END_STATE();
    case 2076:
      ACCEPT_TOKEN(aux_sym__word_token2);
      END_STATE();
    case 2077:
      ACCEPT_TOKEN(aux_sym__word_token3);
      END_STATE();
    case 2078:
      ACCEPT_TOKEN(sym__whitespace);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(2078);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 2030, .external_lex_state = 2},
  [2] = {.lex_state = 2030, .external_lex_state = 3},
  [3] = {.lex_state = 2030, .external_lex_state = 3},
  [4] = {.lex_state = 2030, .external_lex_state = 3},
  [5] = {.lex_state = 2030, .external_lex_state = 3},
  [6] = {.lex_state = 2030, .external_lex_state = 3},
  [7] = {.lex_state = 2030, .external_lex_state = 3},
  [8] = {.lex_state = 2030, .external_lex_state = 3},
  [9] = {.lex_state = 2030, .external_lex_state = 3},
  [10] = {.lex_state = 2030, .external_lex_state = 3},
  [11] = {.lex_state = 2030, .external_lex_state = 3},
  [12] = {.lex_state = 2030, .external_lex_state = 4},
  [13] = {.lex_state = 2030, .external_lex_state = 4},
  [14] = {.lex_state = 2030, .external_lex_state = 4},
  [15] = {.lex_state = 2030, .external_lex_state = 4},
  [16] = {.lex_state = 2030, .external_lex_state = 4},
  [17] = {.lex_state = 2030, .external_lex_state = 4},
  [18] = {.lex_state = 2030, .external_lex_state = 4},
  [19] = {.lex_state = 2030, .external_lex_state = 4},
  [20] = {.lex_state = 2030, .external_lex_state = 4},
  [21] = {.lex_state = 2030, .external_lex_state = 4},
  [22] = {.lex_state = 2030, .external_lex_state = 5},
  [23] = {.lex_state = 2030, .external_lex_state = 5},
  [24] = {.lex_state = 2030, .external_lex_state = 6},
  [25] = {.lex_state = 2030, .external_lex_state = 6},
  [26] = {.lex_state = 2030, .external_lex_state = 6},
  [27] = {.lex_state = 2030, .external_lex_state = 6},
  [28] = {.lex_state = 2030, .external_lex_state = 6},
  [29] = {.lex_state = 2030, .external_lex_state = 6},
  [30] = {.lex_state = 2030, .external_lex_state = 4},
  [31] = {.lex_state = 2030, .external_lex_state = 4},
  [32] = {.lex_state = 2030, .external_lex_state = 6},
  [33] = {.lex_state = 2030, .external_lex_state = 6},
  [34] = {.lex_state = 2030, .external_lex_state = 6},
  [35] = {.lex_state = 2030, .external_lex_state = 4},
  [36] = {.lex_state = 2030, .external_lex_state = 6},
  [37] = {.lex_state = 2030, .external_lex_state = 6},
  [38] = {.lex_state = 2030, .external_lex_state = 4},
  [39] = {.lex_state = 2030, .external_lex_state = 4},
  [40] = {.lex_state = 2030, .external_lex_state = 6},
  [41] = {.lex_state = 2030, .external_lex_state = 4},
  [42] = {.lex_state = 2030, .external_lex_state = 6},
  [43] = {.lex_state = 2030, .external_lex_state = 4},
  [44] = {.lex_state = 2030, .external_lex_state = 4},
  [45] = {.lex_state = 2030, .external_lex_state = 4},
  [46] = {.lex_state = 2030, .external_lex_state = 6},
  [47] = {.lex_state = 2030, .external_lex_state = 6},
  [48] = {.lex_state = 2030, .external_lex_state = 6},
  [49] = {.lex_state = 2030, .external_lex_state = 4},
  [50] = {.lex_state = 2030, .external_lex_state = 4},
  [51] = {.lex_state = 2030, .external_lex_state = 6},
  [52] = {.lex_state = 2030, .external_lex_state = 6},
  [53] = {.lex_state = 2030, .external_lex_state = 6},
  [54] = {.lex_state = 2030, .external_lex_state = 4},
  [55] = {.lex_state = 2030, .external_lex_state = 4},
  [56] = {.lex_state = 2030, .external_lex_state = 4},
  [57] = {.lex_state = 2030, .external_lex_state = 6},
  [58] = {.lex_state = 2030, .external_lex_state = 6},
  [59] = {.lex_state = 2030, .external_lex_state = 4},
  [60] = {.lex_state = 2030, .external_lex_state = 6},
  [61] = {.lex_state = 2030, .external_lex_state = 4},
  [62] = {.lex_state = 2030, .external_lex_state = 6},
  [63] = {.lex_state = 2030, .external_lex_state = 6},
  [64] = {.lex_state = 2030, .external_lex_state = 4},
  [65] = {.lex_state = 2030, .external_lex_state = 6},
  [66] = {.lex_state = 2030, .external_lex_state = 4},
  [67] = {.lex_state = 2030, .external_lex_state = 4},
  [68] = {.lex_state = 2030, .external_lex_state = 6},
  [69] = {.lex_state = 2030, .external_lex_state = 6},
  [70] = {.lex_state = 2030, .external_lex_state = 6},
  [71] = {.lex_state = 2030, .external_lex_state = 4},
  [72] = {.lex_state = 2030, .external_lex_state = 4},
  [73] = {.lex_state = 2030, .external_lex_state = 4},
  [74] = {.lex_state = 2030, .external_lex_state = 4},
  [75] = {.lex_state = 2030, .external_lex_state = 4},
  [76] = {.lex_state = 2030, .external_lex_state = 4},
  [77] = {.lex_state = 2030, .external_lex_state = 4},
  [78] = {.lex_state = 2030, .external_lex_state = 6},
  [79] = {.lex_state = 2030, .external_lex_state = 6},
  [80] = {.lex_state = 2030, .external_lex_state = 6},
  [81] = {.lex_state = 2030, .external_lex_state = 6},
  [82] = {.lex_state = 2030, .external_lex_state = 6},
  [83] = {.lex_state = 2030, .external_lex_state = 6},
  [84] = {.lex_state = 2030, .external_lex_state = 4},
  [85] = {.lex_state = 2030, .external_lex_state = 4},
  [86] = {.lex_state = 2030, .external_lex_state = 4},
  [87] = {.lex_state = 2030, .external_lex_state = 6},
  [88] = {.lex_state = 2030, .external_lex_state = 6},
  [89] = {.lex_state = 2030, .external_lex_state = 6},
  [90] = {.lex_state = 2030, .external_lex_state = 6},
  [91] = {.lex_state = 2030, .external_lex_state = 6},
  [92] = {.lex_state = 2030, .external_lex_state = 6},
  [93] = {.lex_state = 2030, .external_lex_state = 7},
  [94] = {.lex_state = 2030, .external_lex_state = 8},
  [95] = {.lex_state = 2030, .external_lex_state = 4},
  [96] = {.lex_state = 2030, .external_lex_state = 4},
  [97] = {.lex_state = 2030, .external_lex_state = 4},
  [98] = {.lex_state = 2030, .external_lex_state = 4},
  [99] = {.lex_state = 2030, .external_lex_state = 4},
  [100] = {.lex_state = 2030, .external_lex_state = 6},
  [101] = {.lex_state = 2030, .external_lex_state = 9},
  [102] = {.lex_state = 2030, .external_lex_state = 10},
  [103] = {.lex_state = 2030, .external_lex_state = 10},
  [104] = {.lex_state = 2030, .external_lex_state = 10},
  [105] = {.lex_state = 2030, .external_lex_state = 10},
  [106] = {.lex_state = 2030, .external_lex_state = 10},
  [107] = {.lex_state = 2030, .external_lex_state = 11},
  [108] = {.lex_state = 2030, .external_lex_state = 12},
  [109] = {.lex_state = 2030, .external_lex_state = 9},
  [110] = {.lex_state = 2030, .external_lex_state = 13},
  [111] = {.lex_state = 2030, .external_lex_state = 14},
  [112] = {.lex_state = 2030, .external_lex_state = 14},
  [113] = {.lex_state = 2030, .external_lex_state = 13},
  [114] = {.lex_state = 2030, .external_lex_state = 13},
  [115] = {.lex_state = 2030, .external_lex_state = 9},
  [116] = {.lex_state = 2030, .external_lex_state = 9},
  [117] = {.lex_state = 2030, .external_lex_state = 9},
  [118] = {.lex_state = 2030, .external_lex_state = 14},
  [119] = {.lex_state = 2030, .external_lex_state = 5},
  [120] = {.lex_state = 2030, .external_lex_state = 3},
  [121] = {.lex_state = 2030, .external_lex_state = 3},
  [122] = {.lex_state = 2030, .external_lex_state = 3},
  [123] = {.lex_state = 2030, .external_lex_state = 3},
  [124] = {.lex_state = 2030, .external_lex_state = 3},
  [125] = {.lex_state = 2030, .external_lex_state = 3},
  [126] = {.lex_state = 2030, .external_lex_state = 5},
  [127] = {.lex_state = 2030, .external_lex_state = 5},
  [128] = {.lex_state = 2030, .external_lex_state = 5},
  [129] = {.lex_state = 2030, .external_lex_state = 5},
  [130] = {.lex_state = 2030, .external_lex_state = 5},
  [131] = {.lex_state = 2030, .external_lex_state = 5},
  [132] = {.lex_state = 2030, .external_lex_state = 3},
  [133] = {.lex_state = 2030, .external_lex_state = 5},
  [134] = {.lex_state = 2030, .external_lex_state = 5},
  [135] = {.lex_state = 2030, .external_lex_state = 3},
  [136] = {.lex_state = 2030, .external_lex_state = 3},
  [137] = {.lex_state = 2030, .external_lex_state = 3},
  [138] = {.lex_state = 2030, .external_lex_state = 3},
  [139] = {.lex_state = 2030, .external_lex_state = 3},
  [140] = {.lex_state = 2030, .external_lex_state = 5},
  [141] = {.lex_state = 2030, .external_lex_state = 5},
  [142] = {.lex_state = 2030, .external_lex_state = 5},
  [143] = {.lex_state = 2030, .external_lex_state = 5},
  [144] = {.lex_state = 2030, .external_lex_state = 5},
  [145] = {.lex_state = 2030, .external_lex_state = 5},
  [146] = {.lex_state = 2030, .external_lex_state = 5},
  [147] = {.lex_state = 2030, .external_lex_state = 3},
  [148] = {.lex_state = 2030, .external_lex_state = 5},
  [149] = {.lex_state = 2030, .external_lex_state = 5},
  [150] = {.lex_state = 2030, .external_lex_state = 5},
  [151] = {.lex_state = 2030, .external_lex_state = 5},
  [152] = {.lex_state = 2030, .external_lex_state = 5},
  [153] = {.lex_state = 2030, .external_lex_state = 5},
  [154] = {.lex_state = 2030, .external_lex_state = 5},
  [155] = {.lex_state = 2030, .external_lex_state = 3},
  [156] = {.lex_state = 2030, .external_lex_state = 5},
  [157] = {.lex_state = 2030, .external_lex_state = 5},
  [158] = {.lex_state = 2030, .external_lex_state = 9},
  [159] = {.lex_state = 2030, .external_lex_state = 5},
  [160] = {.lex_state = 2030, .external_lex_state = 10},
  [161] = {.lex_state = 2030, .external_lex_state = 3},
  [162] = {.lex_state = 2030, .external_lex_state = 5},
  [163] = {.lex_state = 2030, .external_lex_state = 5},
  [164] = {.lex_state = 2030, .external_lex_state = 5},
  [165] = {.lex_state = 2030, .external_lex_state = 5},
  [166] = {.lex_state = 2030, .external_lex_state = 3},
  [167] = {.lex_state = 2030, .external_lex_state = 3},
  [168] = {.lex_state = 2030, .external_lex_state = 3},
  [169] = {.lex_state = 2030, .external_lex_state = 3},
  [170] = {.lex_state = 2030, .external_lex_state = 6},
  [171] = {.lex_state = 2030, .external_lex_state = 3},
  [172] = {.lex_state = 2030, .external_lex_state = 3},
  [173] = {.lex_state = 2030, .external_lex_state = 3},
  [174] = {.lex_state = 2030, .external_lex_state = 3},
  [175] = {.lex_state = 2030, .external_lex_state = 3},
  [176] = {.lex_state = 2030, .external_lex_state = 3},
  [177] = {.lex_state = 2030, .external_lex_state = 3},
  [178] = {.lex_state = 2030, .external_lex_state = 3},
  [179] = {.lex_state = 2030, .external_lex_state = 3},
  [180] = {.lex_state = 2030, .external_lex_state = 3},
  [181] = {.lex_state = 2030, .external_lex_state = 3},
  [182] = {.lex_state = 2030, .external_lex_state = 3},
  [183] = {.lex_state = 2030, .external_lex_state = 5},
  [184] = {.lex_state = 2030, .external_lex_state = 6},
  [185] = {.lex_state = 2030, .external_lex_state = 4},
  [186] = {.lex_state = 2030, .external_lex_state = 4},
  [187] = {.lex_state = 2030, .external_lex_state = 4},
  [188] = {.lex_state = 2030, .external_lex_state = 4},
  [189] = {.lex_state = 2030, .external_lex_state = 4},
  [190] = {.lex_state = 2030, .external_lex_state = 4},
  [191] = {.lex_state = 2030, .external_lex_state = 4},
  [192] = {.lex_state = 2030, .external_lex_state = 4},
  [193] = {.lex_state = 2030, .external_lex_state = 4},
  [194] = {.lex_state = 2030, .external_lex_state = 4},
  [195] = {.lex_state = 2030, .external_lex_state = 6},
  [196] = {.lex_state = 2030, .external_lex_state = 6},
  [197] = {.lex_state = 2030, .external_lex_state = 3},
  [198] = {.lex_state = 2030, .external_lex_state = 3},
  [199] = {.lex_state = 2030, .external_lex_state = 3},
  [200] = {.lex_state = 2030, .external_lex_state = 4},
  [201] = {.lex_state = 2030, .external_lex_state = 3},
  [202] = {.lex_state = 2030, .external_lex_state = 4},
  [203] = {.lex_state = 2030, .external_lex_state = 6},
  [204] = {.lex_state = 2030, .external_lex_state = 6},
  [205] = {.lex_state = 2030, .external_lex_state = 4},
  [206] = {.lex_state = 2030, .external_lex_state = 6},
  [207] = {.lex_state = 2030, .external_lex_state = 4},
  [208] = {.lex_state = 2030, .external_lex_state = 6},
  [209] = {.lex_state = 2030, .external_lex_state = 4},
  [210] = {.lex_state = 2030, .external_lex_state = 4},
  [211] = {.lex_state = 2030, .external_lex_state = 4},
  [212] = {.lex_state = 2030, .external_lex_state = 4},
  [213] = {.lex_state = 2030, .external_lex_state = 6},
  [214] = {.lex_state = 2030, .external_lex_state = 6},
  [215] = {.lex_state = 2030, .external_lex_state = 6},
  [216] = {.lex_state = 2030, .external_lex_state = 6},
  [217] = {.lex_state = 2030, .external_lex_state = 6},
  [218] = {.lex_state = 2030, .external_lex_state = 6},
  [219] = {.lex_state = 2030, .external_lex_state = 6},
  [220] = {.lex_state = 2030, .external_lex_state = 6},
  [221] = {.lex_state = 2030, .external_lex_state = 4},
  [222] = {.lex_state = 2030, .external_lex_state = 4},
  [223] = {.lex_state = 2030, .external_lex_state = 4},
  [224] = {.lex_state = 2030, .external_lex_state = 4},
  [225] = {.lex_state = 2030, .external_lex_state = 4},
  [226] = {.lex_state = 2030, .external_lex_state = 6},
  [227] = {.lex_state = 2030, .external_lex_state = 4},
  [228] = {.lex_state = 2030, .external_lex_state = 6},
  [229] = {.lex_state = 2030, .external_lex_state = 4},
  [230] = {.lex_state = 2030, .external_lex_state = 6},
  [231] = {.lex_state = 2030, .external_lex_state = 4},
  [232] = {.lex_state = 2030, .external_lex_state = 6},
  [233] = {.lex_state = 2030, .external_lex_state = 4},
  [234] = {.lex_state = 2030, .external_lex_state = 6},
  [235] = {.lex_state = 2030, .external_lex_state = 4},
  [236] = {.lex_state = 2030, .external_lex_state = 4},
  [237] = {.lex_state = 2030, .external_lex_state = 4},
  [238] = {.lex_state = 2030, .external_lex_state = 4},
  [239] = {.lex_state = 2030, .external_lex_state = 3},
  [240] = {.lex_state = 2030, .external_lex_state = 4},
  [241] = {.lex_state = 2030, .external_lex_state = 4},
  [242] = {.lex_state = 2030, .external_lex_state = 4},
  [243] = {.lex_state = 2030, .external_lex_state = 6},
  [244] = {.lex_state = 2030, .external_lex_state = 4},
  [245] = {.lex_state = 2030, .external_lex_state = 4},
  [246] = {.lex_state = 2030, .external_lex_state = 6},
  [247] = {.lex_state = 2030, .external_lex_state = 4},
  [248] = {.lex_state = 2030, .external_lex_state = 6},
  [249] = {.lex_state = 2030, .external_lex_state = 6},
  [250] = {.lex_state = 2030, .external_lex_state = 6},
  [251] = {.lex_state = 2030, .external_lex_state = 6},
  [252] = {.lex_state = 2030, .external_lex_state = 6},
  [253] = {.lex_state = 2030, .external_lex_state = 6},
  [254] = {.lex_state = 2030, .external_lex_state = 6},
  [255] = {.lex_state = 2030, .external_lex_state = 6},
  [256] = {.lex_state = 2030, .external_lex_state = 6},
  [257] = {.lex_state = 2030, .external_lex_state = 4},
  [258] = {.lex_state = 2030, .external_lex_state = 6},
  [259] = {.lex_state = 2030, .external_lex_state = 6},
  [260] = {.lex_state = 2030, .external_lex_state = 4},
  [261] = {.lex_state = 2030, .external_lex_state = 6},
  [262] = {.lex_state = 2030, .external_lex_state = 4},
  [263] = {.lex_state = 2030, .external_lex_state = 6},
  [264] = {.lex_state = 2030, .external_lex_state = 4},
  [265] = {.lex_state = 2030, .external_lex_state = 6},
  [266] = {.lex_state = 2030, .external_lex_state = 4},
  [267] = {.lex_state = 2030, .external_lex_state = 4},
  [268] = {.lex_state = 2030, .external_lex_state = 4},
  [269] = {.lex_state = 2030, .external_lex_state = 6},
  [270] = {.lex_state = 2030, .external_lex_state = 6},
  [271] = {.lex_state = 2030, .external_lex_state = 4},
  [272] = {.lex_state = 2030, .external_lex_state = 4},
  [273] = {.lex_state = 2030, .external_lex_state = 4},
  [274] = {.lex_state = 2030, .external_lex_state = 4},
  [275] = {.lex_state = 2030, .external_lex_state = 4},
  [276] = {.lex_state = 2030, .external_lex_state = 6},
  [277] = {.lex_state = 2030, .external_lex_state = 4},
  [278] = {.lex_state = 2030, .external_lex_state = 6},
  [279] = {.lex_state = 2030, .external_lex_state = 6},
  [280] = {.lex_state = 2030, .external_lex_state = 6},
  [281] = {.lex_state = 2030, .external_lex_state = 6},
  [282] = {.lex_state = 2030, .external_lex_state = 6},
  [283] = {.lex_state = 2030, .external_lex_state = 6},
  [284] = {.lex_state = 2030, .external_lex_state = 6},
  [285] = {.lex_state = 2030, .external_lex_state = 6},
  [286] = {.lex_state = 2030, .external_lex_state = 6},
  [287] = {.lex_state = 2030, .external_lex_state = 6},
  [288] = {.lex_state = 2030, .external_lex_state = 4},
  [289] = {.lex_state = 2030, .external_lex_state = 6},
  [290] = {.lex_state = 2030, .external_lex_state = 6},
  [291] = {.lex_state = 2030, .external_lex_state = 4},
  [292] = {.lex_state = 2030, .external_lex_state = 6},
  [293] = {.lex_state = 2030, .external_lex_state = 4},
  [294] = {.lex_state = 2030, .external_lex_state = 6},
  [295] = {.lex_state = 2030, .external_lex_state = 4},
  [296] = {.lex_state = 2030, .external_lex_state = 6},
  [297] = {.lex_state = 2030, .external_lex_state = 6},
  [298] = {.lex_state = 2030, .external_lex_state = 6},
  [299] = {.lex_state = 2030, .external_lex_state = 6},
  [300] = {.lex_state = 2030, .external_lex_state = 6},
  [301] = {.lex_state = 2030, .external_lex_state = 4},
  [302] = {.lex_state = 2030, .external_lex_state = 6},
  [303] = {.lex_state = 2030, .external_lex_state = 6},
  [304] = {.lex_state = 2030, .external_lex_state = 6},
  [305] = {.lex_state = 2030, .external_lex_state = 6},
  [306] = {.lex_state = 2030, .external_lex_state = 6},
  [307] = {.lex_state = 2030, .external_lex_state = 6},
  [308] = {.lex_state = 2030, .external_lex_state = 6},
  [309] = {.lex_state = 2030, .external_lex_state = 6},
  [310] = {.lex_state = 2030, .external_lex_state = 4},
  [311] = {.lex_state = 2030, .external_lex_state = 6},
  [312] = {.lex_state = 2030, .external_lex_state = 6},
  [313] = {.lex_state = 2030, .external_lex_state = 6},
  [314] = {.lex_state = 2030, .external_lex_state = 4},
  [315] = {.lex_state = 2030, .external_lex_state = 6},
  [316] = {.lex_state = 2030, .external_lex_state = 6},
  [317] = {.lex_state = 2030, .external_lex_state = 6},
  [318] = {.lex_state = 2030, .external_lex_state = 6},
  [319] = {.lex_state = 2030, .external_lex_state = 4},
  [320] = {.lex_state = 2030, .external_lex_state = 6},
  [321] = {.lex_state = 2030, .external_lex_state = 6},
  [322] = {.lex_state = 2030, .external_lex_state = 6},
  [323] = {.lex_state = 2030, .external_lex_state = 6},
  [324] = {.lex_state = 2030, .external_lex_state = 6},
  [325] = {.lex_state = 2030, .external_lex_state = 6},
  [326] = {.lex_state = 2030, .external_lex_state = 6},
  [327] = {.lex_state = 2030, .external_lex_state = 4},
  [328] = {.lex_state = 2030, .external_lex_state = 4},
  [329] = {.lex_state = 2030, .external_lex_state = 4},
  [330] = {.lex_state = 2030, .external_lex_state = 4},
  [331] = {.lex_state = 2030, .external_lex_state = 6},
  [332] = {.lex_state = 2030, .external_lex_state = 4},
  [333] = {.lex_state = 2030, .external_lex_state = 4},
  [334] = {.lex_state = 2030, .external_lex_state = 4},
  [335] = {.lex_state = 2030, .external_lex_state = 6},
  [336] = {.lex_state = 2030, .external_lex_state = 4},
  [337] = {.lex_state = 2030, .external_lex_state = 4},
  [338] = {.lex_state = 2030, .external_lex_state = 4},
  [339] = {.lex_state = 2030, .external_lex_state = 4},
  [340] = {.lex_state = 2030, .external_lex_state = 4},
  [341] = {.lex_state = 2030, .external_lex_state = 4},
  [342] = {.lex_state = 2030, .external_lex_state = 4},
  [343] = {.lex_state = 2030, .external_lex_state = 4},
  [344] = {.lex_state = 2030, .external_lex_state = 4},
  [345] = {.lex_state = 2030, .external_lex_state = 4},
  [346] = {.lex_state = 2030, .external_lex_state = 4},
  [347] = {.lex_state = 2030, .external_lex_state = 4},
  [348] = {.lex_state = 2030, .external_lex_state = 4},
  [349] = {.lex_state = 2030, .external_lex_state = 4},
  [350] = {.lex_state = 2030, .external_lex_state = 4},
  [351] = {.lex_state = 2030, .external_lex_state = 15},
  [352] = {.lex_state = 2030, .external_lex_state = 15},
  [353] = {.lex_state = 1, .external_lex_state = 16},
  [354] = {.lex_state = 1, .external_lex_state = 17},
  [355] = {.lex_state = 1, .external_lex_state = 17},
  [356] = {.lex_state = 1, .external_lex_state = 17},
  [357] = {.lex_state = 1, .external_lex_state = 17},
  [358] = {.lex_state = 1, .external_lex_state = 16},
  [359] = {.lex_state = 2030, .external_lex_state = 18},
  [360] = {.lex_state = 1, .external_lex_state = 16},
  [361] = {.lex_state = 1, .external_lex_state = 16},
  [362] = {.lex_state = 2030, .external_lex_state = 16},
  [363] = {.lex_state = 1, .external_lex_state = 16},
  [364] = {.lex_state = 2030, .external_lex_state = 18},
  [365] = {.lex_state = 2030, .external_lex_state = 19},
  [366] = {.lex_state = 2030, .external_lex_state = 18},
  [367] = {.lex_state = 2030, .external_lex_state = 18},
  [368] = {.lex_state = 1, .external_lex_state = 17},
  [369] = {.lex_state = 2030, .external_lex_state = 19},
  [370] = {.lex_state = 2030, .external_lex_state = 18},
  [371] = {.lex_state = 2030, .external_lex_state = 19},
  [372] = {.lex_state = 1, .external_lex_state = 17},
  [373] = {.lex_state = 2030, .external_lex_state = 16},
  [374] = {.lex_state = 1, .external_lex_state = 17},
  [375] = {.lex_state = 2030, .external_lex_state = 16},
  [376] = {.lex_state = 2030, .external_lex_state = 19},
  [377] = {.lex_state = 2030, .external_lex_state = 19},
  [378] = {.lex_state = 2030, .external_lex_state = 19},
  [379] = {.lex_state = 1, .external_lex_state = 17},
  [380] = {.lex_state = 1, .external_lex_state = 17},
  [381] = {.lex_state = 1, .external_lex_state = 17},
  [382] = {.lex_state = 1, .external_lex_state = 17},
  [383] = {.lex_state = 1, .external_lex_state = 17},
  [384] = {.lex_state = 2030, .external_lex_state = 18},
  [385] = {.lex_state = 1, .external_lex_state = 16},
  [386] = {.lex_state = 1, .external_lex_state = 16},
  [387] = {.lex_state = 1, .external_lex_state = 17},
  [388] = {.lex_state = 2, .external_lex_state = 20},
  [389] = {.lex_state = 1, .external_lex_state = 21},
  [390] = {.lex_state = 1, .external_lex_state = 21},
  [391] = {.lex_state = 1, .external_lex_state = 21},
  [392] = {.lex_state = 2, .external_lex_state = 20},
  [393] = {.lex_state = 1, .external_lex_state = 21},
  [394] = {.lex_state = 1, .external_lex_state = 21},
  [395] = {.lex_state = 3, .external_lex_state = 20},
  [396] = {.lex_state = 2030, .external_lex_state = 19},
  [397] = {.lex_state = 2030, .external_lex_state = 22},
  [398] = {.lex_state = 5, .external_lex_state = 20},
  [399] = {.lex_state = 2, .external_lex_state = 20},
  [400] = {.lex_state = 3, .external_lex_state = 20},
  [401] = {.lex_state = 5, .external_lex_state = 20},
  [402] = {.lex_state = 1, .external_lex_state = 21},
  [403] = {.lex_state = 1, .external_lex_state = 21},
  [404] = {.lex_state = 1, .external_lex_state = 21},
  [405] = {.lex_state = 1, .external_lex_state = 21},
  [406] = {.lex_state = 2030, .external_lex_state = 22},
  [407] = {.lex_state = 2030, .external_lex_state = 18},
  [408] = {.lex_state = 2, .external_lex_state = 20},
  [409] = {.lex_state = 3, .external_lex_state = 20},
  [410] = {.lex_state = 2030, .external_lex_state = 19},
  [411] = {.lex_state = 5, .external_lex_state = 20},
  [412] = {.lex_state = 2030, .external_lex_state = 18},
  [413] = {.lex_state = 2030, .external_lex_state = 16},
  [414] = {.lex_state = 1, .external_lex_state = 17},
  [415] = {.lex_state = 4, .external_lex_state = 23},
  [416] = {.lex_state = 2030, .external_lex_state = 22},
  [417] = {.lex_state = 2, .external_lex_state = 20},
  [418] = {.lex_state = 3, .external_lex_state = 20},
  [419] = {.lex_state = 5, .external_lex_state = 20},
  [420] = {.lex_state = 2030, .external_lex_state = 22},
  [421] = {.lex_state = 3, .external_lex_state = 20},
  [422] = {.lex_state = 5, .external_lex_state = 20},
  [423] = {.lex_state = 1, .external_lex_state = 21},
  [424] = {.lex_state = 1, .external_lex_state = 21},
  [425] = {.lex_state = 1, .external_lex_state = 21},
  [426] = {.lex_state = 2030, .external_lex_state = 22},
  [427] = {.lex_state = 2030, .external_lex_state = 20},
  [428] = {.lex_state = 2030, .external_lex_state = 20},
  [429] = {.lex_state = 2030, .external_lex_state = 20},
  [430] = {.lex_state = 2030, .external_lex_state = 20},
  [431] = {.lex_state = 4, .external_lex_state = 23},
  [432] = {.lex_state = 1, .external_lex_state = 21},
  [433] = {.lex_state = 2030, .external_lex_state = 20},
  [434] = {.lex_state = 2030, .external_lex_state = 20},
  [435] = {.lex_state = 2030, .external_lex_state = 21},
  [436] = {.lex_state = 2030, .external_lex_state = 20},
  [437] = {.lex_state = 2030, .external_lex_state = 20},
  [438] = {.lex_state = 2030, .external_lex_state = 20},
  [439] = {.lex_state = 2030, .external_lex_state = 20},
  [440] = {.lex_state = 2030, .external_lex_state = 20},
  [441] = {.lex_state = 2030, .external_lex_state = 20},
  [442] = {.lex_state = 2030, .external_lex_state = 20},
  [443] = {.lex_state = 1, .external_lex_state = 21},
  [444] = {.lex_state = 2030, .external_lex_state = 20},
  [445] = {.lex_state = 2030, .external_lex_state = 20},
  [446] = {.lex_state = 1, .external_lex_state = 21},
  [447] = {.lex_state = 2030, .external_lex_state = 20},
  [448] = {.lex_state = 2030, .external_lex_state = 20},
  [449] = {.lex_state = 1, .external_lex_state = 17},
  [450] = {.lex_state = 2030, .external_lex_state = 20},
  [451] = {.lex_state = 1, .external_lex_state = 17},
  [452] = {.lex_state = 1, .external_lex_state = 21},
  [453] = {.lex_state = 1, .external_lex_state = 21},
  [454] = {.lex_state = 1, .external_lex_state = 21},
  [455] = {.lex_state = 1, .external_lex_state = 21},
  [456] = {.lex_state = 2030, .external_lex_state = 20},
  [457] = {.lex_state = 1, .external_lex_state = 21},
  [458] = {.lex_state = 1},
  [459] = {.lex_state = 2030, .external_lex_state = 17},
  [460] = {.lex_state = 2030, .external_lex_state = 17},
  [461] = {.lex_state = 2030, .external_lex_state = 17},
  [462] = {.lex_state = 4, .external_lex_state = 23},
  [463] = {.lex_state = 4, .external_lex_state = 23},
  [464] = {.lex_state = 4, .external_lex_state = 23},
  [465] = {.lex_state = 4, .external_lex_state = 23},
  [466] = {.lex_state = 1, .external_lex_state = 17},
  [467] = {.lex_state = 1, .external_lex_state = 17},
  [468] = {.lex_state = 2030, .external_lex_state = 17},
  [469] = {.lex_state = 2030, .external_lex_state = 17},
  [470] = {.lex_state = 2030, .external_lex_state = 17},
  [471] = {.lex_state = 2030, .external_lex_state = 17},
  [472] = {.lex_state = 2030, .external_lex_state = 17},
  [473] = {.lex_state = 2030, .external_lex_state = 17},
  [474] = {.lex_state = 2030, .external_lex_state = 17},
  [475] = {.lex_state = 4, .external_lex_state = 23},
  [476] = {.lex_state = 1},
  [477] = {.lex_state = 4, .external_lex_state = 23},
  [478] = {.lex_state = 4, .external_lex_state = 23},
  [479] = {.lex_state = 4, .external_lex_state = 23},
  [480] = {.lex_state = 1},
  [481] = {.lex_state = 1},
  [482] = {.lex_state = 1},
  [483] = {.lex_state = 4, .external_lex_state = 23},
  [484] = {.lex_state = 1},
  [485] = {.lex_state = 1},
  [486] = {.lex_state = 1},
  [487] = {.lex_state = 1},
  [488] = {.lex_state = 1},
  [489] = {.lex_state = 1},
  [490] = {.lex_state = 2030, .external_lex_state = 17},
  [491] = {.lex_state = 1},
  [492] = {.lex_state = 2030, .external_lex_state = 17},
  [493] = {.lex_state = 1},
  [494] = {.lex_state = 1},
  [495] = {.lex_state = 4, .external_lex_state = 23},
  [496] = {.lex_state = 1, .external_lex_state = 16},
  [497] = {.lex_state = 1, .external_lex_state = 16},
  [498] = {.lex_state = 4, .external_lex_state = 23},
  [499] = {.lex_state = 2030, .external_lex_state = 19},
  [500] = {.lex_state = 2030, .external_lex_state = 19},
  [501] = {.lex_state = 2030, .external_lex_state = 18},
  [502] = {.lex_state = 1, .external_lex_state = 24},
  [503] = {.lex_state = 4, .external_lex_state = 25},
  [504] = {.lex_state = 4, .external_lex_state = 23},
  [505] = {.lex_state = 2030, .external_lex_state = 18},
  [506] = {.lex_state = 2030, .external_lex_state = 16},
  [507] = {.lex_state = 2030, .external_lex_state = 16},
  [508] = {.lex_state = 4, .external_lex_state = 17},
  [509] = {.lex_state = 2030, .external_lex_state = 22},
  [510] = {.lex_state = 1},
  [511] = {.lex_state = 4},
  [512] = {.lex_state = 1},
  [513] = {.lex_state = 2, .external_lex_state = 20},
  [514] = {.lex_state = 1, .external_lex_state = 16},
  [515] = {.lex_state = 2, .external_lex_state = 20},
  [516] = {.lex_state = 1, .external_lex_state = 16},
  [517] = {.lex_state = 4, .external_lex_state = 23},
  [518] = {.lex_state = 3, .external_lex_state = 20},
  [519] = {.lex_state = 3, .external_lex_state = 20},
  [520] = {.lex_state = 5, .external_lex_state = 20},
  [521] = {.lex_state = 1},
  [522] = {.lex_state = 5, .external_lex_state = 20},
  [523] = {.lex_state = 1, .external_lex_state = 21},
  [524] = {.lex_state = 1},
  [525] = {.lex_state = 1, .external_lex_state = 21},
  [526] = {.lex_state = 1},
  [527] = {.lex_state = 1},
  [528] = {.lex_state = 4},
  [529] = {.lex_state = 2030, .external_lex_state = 22},
  [530] = {.lex_state = 1, .external_lex_state = 21},
  [531] = {.lex_state = 1, .external_lex_state = 21},
  [532] = {.lex_state = 2030, .external_lex_state = 20},
  [533] = {.lex_state = 1, .external_lex_state = 21},
  [534] = {.lex_state = 4, .external_lex_state = 17},
  [535] = {.lex_state = 1, .external_lex_state = 17},
  [536] = {.lex_state = 4, .external_lex_state = 17},
  [537] = {.lex_state = 2030, .external_lex_state = 26},
  [538] = {.lex_state = 2030, .external_lex_state = 27},
  [539] = {.lex_state = 1, .external_lex_state = 21},
  [540] = {.lex_state = 4, .external_lex_state = 17},
  [541] = {.lex_state = 4, .external_lex_state = 17},
  [542] = {.lex_state = 4, .external_lex_state = 17},
  [543] = {.lex_state = 4, .external_lex_state = 23},
  [544] = {.lex_state = 2030, .external_lex_state = 28},
  [545] = {.lex_state = 4},
  [546] = {.lex_state = 2030, .external_lex_state = 29},
  [547] = {.lex_state = 2, .external_lex_state = 30},
  [548] = {.lex_state = 3, .external_lex_state = 30},
  [549] = {.lex_state = 4, .external_lex_state = 17},
  [550] = {.lex_state = 5, .external_lex_state = 30},
  [551] = {.lex_state = 1, .external_lex_state = 17},
  [552] = {.lex_state = 1, .external_lex_state = 21},
  [553] = {.lex_state = 2030, .external_lex_state = 31},
  [554] = {.lex_state = 1, .external_lex_state = 21},
  [555] = {.lex_state = 4, .external_lex_state = 17},
  [556] = {.lex_state = 2030, .external_lex_state = 32},
  [557] = {.lex_state = 1, .external_lex_state = 17},
  [558] = {.lex_state = 2030, .external_lex_state = 20},
  [559] = {.lex_state = 2030, .external_lex_state = 17},
  [560] = {.lex_state = 1, .external_lex_state = 17},
  [561] = {.lex_state = 2030, .external_lex_state = 33},
  [562] = {.lex_state = 1, .external_lex_state = 21},
  [563] = {.lex_state = 4, .external_lex_state = 17},
  [564] = {.lex_state = 4, .external_lex_state = 17},
  [565] = {.lex_state = 4, .external_lex_state = 17},
  [566] = {.lex_state = 4, .external_lex_state = 23},
  [567] = {.lex_state = 2030, .external_lex_state = 22},
  [568] = {.lex_state = 4, .external_lex_state = 17},
  [569] = {.lex_state = 5, .external_lex_state = 20},
  [570] = {.lex_state = 2030, .external_lex_state = 17},
  [571] = {.lex_state = 3, .external_lex_state = 20},
  [572] = {.lex_state = 1, .external_lex_state = 21},
  [573] = {.lex_state = 2030, .external_lex_state = 19},
  [574] = {.lex_state = 4, .external_lex_state = 23},
  [575] = {.lex_state = 5, .external_lex_state = 20},
  [576] = {.lex_state = 1, .external_lex_state = 21},
  [577] = {.lex_state = 1, .external_lex_state = 34},
  [578] = {.lex_state = 2030, .external_lex_state = 19},
  [579] = {.lex_state = 4},
  [580] = {.lex_state = 2030, .external_lex_state = 33},
  [581] = {.lex_state = 1, .external_lex_state = 21},
  [582] = {.lex_state = 2030, .external_lex_state = 18},
  [583] = {.lex_state = 2030, .external_lex_state = 17},
  [584] = {.lex_state = 2030, .external_lex_state = 16},
  [585] = {.lex_state = 1},
  [586] = {.lex_state = 4, .external_lex_state = 23},
  [587] = {.lex_state = 4, .external_lex_state = 17},
  [588] = {.lex_state = 2030, .external_lex_state = 30},
  [589] = {.lex_state = 2030, .external_lex_state = 22},
  [590] = {.lex_state = 2030, .external_lex_state = 18},
  [591] = {.lex_state = 2, .external_lex_state = 20},
  [592] = {.lex_state = 4},
  [593] = {.lex_state = 4},
  [594] = {.lex_state = 4},
  [595] = {.lex_state = 4},
  [596] = {.lex_state = 4, .external_lex_state = 23},
  [597] = {.lex_state = 3, .external_lex_state = 20},
  [598] = {.lex_state = 2, .external_lex_state = 20},
  [599] = {.lex_state = 2030, .external_lex_state = 16},
  [600] = {.lex_state = 2030, .external_lex_state = 16},
  [601] = {.lex_state = 2030, .external_lex_state = 16},
  [602] = {.lex_state = 2030, .external_lex_state = 22},
  [603] = {.lex_state = 5, .external_lex_state = 20},
  [604] = {.lex_state = 4, .external_lex_state = 17},
  [605] = {.lex_state = 3, .external_lex_state = 20},
  [606] = {.lex_state = 4},
  [607] = {.lex_state = 4},
  [608] = {.lex_state = 4, .external_lex_state = 23},
  [609] = {.lex_state = 2, .external_lex_state = 20},
  [610] = {.lex_state = 2030, .external_lex_state = 16},
  [611] = {.lex_state = 4},
  [612] = {.lex_state = 1},
  [613] = {.lex_state = 4},
  [614] = {.lex_state = 4},
  [615] = {.lex_state = 1},
  [616] = {.lex_state = 4},
  [617] = {.lex_state = 4},
  [618] = {.lex_state = 1},
  [619] = {.lex_state = 2030, .external_lex_state = 20},
  [620] = {.lex_state = 4},
  [621] = {.lex_state = 2030, .external_lex_state = 20},
  [622] = {.lex_state = 2030, .external_lex_state = 20},
  [623] = {.lex_state = 4},
  [624] = {.lex_state = 1},
  [625] = {.lex_state = 4, .external_lex_state = 17},
  [626] = {.lex_state = 2030, .external_lex_state = 20},
  [627] = {.lex_state = 4},
  [628] = {.lex_state = 1},
  [629] = {.lex_state = 4, .external_lex_state = 17},
  [630] = {.lex_state = 4, .external_lex_state = 17},
  [631] = {.lex_state = 2030, .external_lex_state = 17},
  [632] = {.lex_state = 4, .external_lex_state = 17},
  [633] = {.lex_state = 4},
  [634] = {.lex_state = 4, .external_lex_state = 17},
  [635] = {.lex_state = 0, .external_lex_state = 35},
  [636] = {.lex_state = 0, .external_lex_state = 35},
  [637] = {.lex_state = 0, .external_lex_state = 35},
  [638] = {.lex_state = 0, .external_lex_state = 35},
  [639] = {.lex_state = 0, .external_lex_state = 35},
  [640] = {.lex_state = 2030, .external_lex_state = 23},
  [641] = {.lex_state = 2030, .external_lex_state = 23},
  [642] = {.lex_state = 2030, .external_lex_state = 23},
  [643] = {.lex_state = 2030, .external_lex_state = 23},
  [644] = {.lex_state = 2030, .external_lex_state = 23},
  [645] = {.lex_state = 2030, .external_lex_state = 23},
  [646] = {.lex_state = 2030},
  [647] = {.lex_state = 2030},
  [648] = {.lex_state = 2030, .external_lex_state = 23},
  [649] = {.lex_state = 0, .external_lex_state = 36},
  [650] = {.lex_state = 0, .external_lex_state = 36},
  [651] = {.lex_state = 2030, .external_lex_state = 23},
  [652] = {.lex_state = 2030, .external_lex_state = 23},
  [653] = {.lex_state = 0, .external_lex_state = 36},
  [654] = {.lex_state = 0, .external_lex_state = 36},
  [655] = {.lex_state = 0, .external_lex_state = 36},
  [656] = {.lex_state = 0, .external_lex_state = 36},
  [657] = {.lex_state = 0, .external_lex_state = 36},
  [658] = {.lex_state = 0, .external_lex_state = 36},
  [659] = {.lex_state = 0, .external_lex_state = 36},
  [660] = {.lex_state = 0, .external_lex_state = 36},
  [661] = {.lex_state = 0, .external_lex_state = 35},
  [662] = {.lex_state = 2030, .external_lex_state = 23},
  [663] = {.lex_state = 0, .external_lex_state = 37},
  [664] = {.lex_state = 0, .external_lex_state = 16},
  [665] = {.lex_state = 0, .external_lex_state = 37},
  [666] = {.lex_state = 0, .external_lex_state = 37},
  [667] = {.lex_state = 0, .external_lex_state = 16},
  [668] = {.lex_state = 0, .external_lex_state = 38},
  [669] = {.lex_state = 0, .external_lex_state = 16},
  [670] = {.lex_state = 0, .external_lex_state = 16},
  [671] = {.lex_state = 0, .external_lex_state = 37},
  [672] = {.lex_state = 2030},
  [673] = {.lex_state = 0, .external_lex_state = 37},
  [674] = {.lex_state = 0, .external_lex_state = 23},
  [675] = {.lex_state = 0, .external_lex_state = 23},
  [676] = {.lex_state = 2030, .external_lex_state = 23},
  [677] = {.lex_state = 2030},
  [678] = {.lex_state = 0, .external_lex_state = 23},
  [679] = {.lex_state = 2030, .external_lex_state = 23},
  [680] = {.lex_state = 0, .external_lex_state = 16},
  [681] = {.lex_state = 0, .external_lex_state = 16},
  [682] = {.lex_state = 0, .external_lex_state = 16},
  [683] = {.lex_state = 0, .external_lex_state = 37},
  [684] = {.lex_state = 0, .external_lex_state = 23},
  [685] = {.lex_state = 0, .external_lex_state = 16},
  [686] = {.lex_state = 0, .external_lex_state = 37},
  [687] = {.lex_state = 0, .external_lex_state = 37},
  [688] = {.lex_state = 0, .external_lex_state = 16},
  [689] = {.lex_state = 0, .external_lex_state = 16},
  [690] = {.lex_state = 0, .external_lex_state = 37},
  [691] = {.lex_state = 0, .external_lex_state = 37},
  [692] = {.lex_state = 2030, .external_lex_state = 23},
  [693] = {.lex_state = 0, .external_lex_state = 23},
  [694] = {.lex_state = 0, .external_lex_state = 23},
  [695] = {.lex_state = 0, .external_lex_state = 23},
  [696] = {.lex_state = 0, .external_lex_state = 23},
  [697] = {.lex_state = 0, .external_lex_state = 16},
  [698] = {.lex_state = 2030},
  [699] = {.lex_state = 0, .external_lex_state = 37},
  [700] = {.lex_state = 0, .external_lex_state = 23},
  [701] = {.lex_state = 0, .external_lex_state = 23},
  [702] = {.lex_state = 0, .external_lex_state = 37},
  [703] = {.lex_state = 0, .external_lex_state = 16},
  [704] = {.lex_state = 0, .external_lex_state = 37},
  [705] = {.lex_state = 0, .external_lex_state = 23},
  [706] = {.lex_state = 0, .external_lex_state = 16},
  [707] = {.lex_state = 2030},
  [708] = {.lex_state = 2030, .external_lex_state = 34},
  [709] = {.lex_state = 0, .external_lex_state = 23},
  [710] = {.lex_state = 0, .external_lex_state = 37},
  [711] = {.lex_state = 0, .external_lex_state = 23},
  [712] = {.lex_state = 0, .external_lex_state = 23},
  [713] = {.lex_state = 0, .external_lex_state = 23},
  [714] = {.lex_state = 2030},
  [715] = {.lex_state = 0, .external_lex_state = 23},
  [716] = {.lex_state = 2030},
  [717] = {.lex_state = 0, .external_lex_state = 37},
  [718] = {.lex_state = 0, .external_lex_state = 23},
  [719] = {.lex_state = 0, .external_lex_state = 23},
  [720] = {.lex_state = 0, .external_lex_state = 16},
  [721] = {.lex_state = 0, .external_lex_state = 16},
  [722] = {.lex_state = 0, .external_lex_state = 16},
  [723] = {.lex_state = 0, .external_lex_state = 16},
  [724] = {.lex_state = 0, .external_lex_state = 16},
  [725] = {.lex_state = 0, .external_lex_state = 16},
  [726] = {.lex_state = 0, .external_lex_state = 16},
  [727] = {.lex_state = 0, .external_lex_state = 16},
  [728] = {.lex_state = 0, .external_lex_state = 37},
  [729] = {.lex_state = 0, .external_lex_state = 16},
  [730] = {.lex_state = 0, .external_lex_state = 37},
  [731] = {.lex_state = 0, .external_lex_state = 16},
  [732] = {.lex_state = 0, .external_lex_state = 16},
  [733] = {.lex_state = 0, .external_lex_state = 37},
  [734] = {.lex_state = 0, .external_lex_state = 23},
  [735] = {.lex_state = 0, .external_lex_state = 23},
  [736] = {.lex_state = 0, .external_lex_state = 37},
  [737] = {.lex_state = 0, .external_lex_state = 16},
  [738] = {.lex_state = 0, .external_lex_state = 37},
  [739] = {.lex_state = 0, .external_lex_state = 23},
  [740] = {.lex_state = 0, .external_lex_state = 37},
  [741] = {.lex_state = 0, .external_lex_state = 37},
  [742] = {.lex_state = 0, .external_lex_state = 37},
  [743] = {.lex_state = 0, .external_lex_state = 16},
  [744] = {.lex_state = 0, .external_lex_state = 16},
  [745] = {.lex_state = 0},
  [746] = {.lex_state = 0, .external_lex_state = 23},
  [747] = {.lex_state = 0, .external_lex_state = 23},
  [748] = {.lex_state = 0},
  [749] = {.lex_state = 0, .external_lex_state = 23},
  [750] = {.lex_state = 0, .external_lex_state = 23},
  [751] = {.lex_state = 0, .external_lex_state = 23},
  [752] = {.lex_state = 2030},
  [753] = {.lex_state = 2030},
  [754] = {.lex_state = 0},
  [755] = {.lex_state = 0, .external_lex_state = 16},
  [756] = {.lex_state = 0},
  [757] = {.lex_state = 0},
  [758] = {.lex_state = 0},
  [759] = {.lex_state = 0, .external_lex_state = 23},
  [760] = {.lex_state = 0, .external_lex_state = 23},
  [761] = {.lex_state = 0},
  [762] = {.lex_state = 0, .external_lex_state = 23},
  [763] = {.lex_state = 0, .external_lex_state = 23},
  [764] = {.lex_state = 0, .external_lex_state = 23},
  [765] = {.lex_state = 0},
  [766] = {.lex_state = 0},
  [767] = {.lex_state = 0, .external_lex_state = 23},
  [768] = {.lex_state = 0},
  [769] = {.lex_state = 0, .external_lex_state = 23},
  [770] = {.lex_state = 0},
  [771] = {.lex_state = 0},
  [772] = {.lex_state = 0},
  [773] = {.lex_state = 0, .external_lex_state = 23},
  [774] = {.lex_state = 0},
  [775] = {.lex_state = 0, .external_lex_state = 23},
  [776] = {.lex_state = 0, .external_lex_state = 17},
  [777] = {.lex_state = 0, .external_lex_state = 39},
  [778] = {.lex_state = 0, .external_lex_state = 39},
  [779] = {.lex_state = 0, .external_lex_state = 39},
  [780] = {.lex_state = 0, .external_lex_state = 17},
  [781] = {.lex_state = 0, .external_lex_state = 39},
  [782] = {.lex_state = 0, .external_lex_state = 39},
  [783] = {.lex_state = 0, .external_lex_state = 17},
  [784] = {.lex_state = 0, .external_lex_state = 39},
  [785] = {.lex_state = 0, .external_lex_state = 17},
  [786] = {.lex_state = 0, .external_lex_state = 16},
  [787] = {.lex_state = 0, .external_lex_state = 17},
  [788] = {.lex_state = 0, .external_lex_state = 39},
  [789] = {.lex_state = 0, .external_lex_state = 17},
  [790] = {.lex_state = 0, .external_lex_state = 39},
  [791] = {.lex_state = 0, .external_lex_state = 17},
  [792] = {.lex_state = 0, .external_lex_state = 39},
  [793] = {.lex_state = 0, .external_lex_state = 16},
  [794] = {.lex_state = 0, .external_lex_state = 23},
  [795] = {.lex_state = 0, .external_lex_state = 17},
  [796] = {.lex_state = 0, .external_lex_state = 17},
  [797] = {.lex_state = 0, .external_lex_state = 17},
  [798] = {.lex_state = 0, .external_lex_state = 17},
  [799] = {.lex_state = 0, .external_lex_state = 17},
  [800] = {.lex_state = 0, .external_lex_state = 17},
  [801] = {.lex_state = 0, .external_lex_state = 17},
  [802] = {.lex_state = 0, .external_lex_state = 17},
  [803] = {.lex_state = 0, .external_lex_state = 17},
  [804] = {.lex_state = 0},
  [805] = {.lex_state = 0, .external_lex_state = 17},
  [806] = {.lex_state = 0},
  [807] = {.lex_state = 0, .external_lex_state = 17},
  [808] = {.lex_state = 0, .external_lex_state = 17},
  [809] = {.lex_state = 0},
  [810] = {.lex_state = 0, .external_lex_state = 17},
  [811] = {.lex_state = 0, .external_lex_state = 40},
  [812] = {.lex_state = 0, .external_lex_state = 41},
  [813] = {.lex_state = 0, .external_lex_state = 17},
  [814] = {.lex_state = 0, .external_lex_state = 40},
  [815] = {.lex_state = 0, .external_lex_state = 41},
  [816] = {.lex_state = 0},
  [817] = {.lex_state = 0, .external_lex_state = 40},
  [818] = {.lex_state = 0, .external_lex_state = 41},
  [819] = {.lex_state = 0, .external_lex_state = 42},
  [820] = {.lex_state = 0, .external_lex_state = 17},
  [821] = {.lex_state = 0, .external_lex_state = 17},
  [822] = {.lex_state = 0, .external_lex_state = 40},
  [823] = {.lex_state = 0, .external_lex_state = 17},
  [824] = {.lex_state = 0},
  [825] = {.lex_state = 0, .external_lex_state = 41},
  [826] = {.lex_state = 0, .external_lex_state = 17},
  [827] = {.lex_state = 2030},
  [828] = {.lex_state = 0, .external_lex_state = 17},
  [829] = {.lex_state = 0, .external_lex_state = 17},
  [830] = {.lex_state = 0},
  [831] = {.lex_state = 0, .external_lex_state = 17},
  [832] = {.lex_state = 0, .external_lex_state = 17},
  [833] = {.lex_state = 0, .external_lex_state = 17},
  [834] = {.lex_state = 0, .external_lex_state = 43},
  [835] = {.lex_state = 0, .external_lex_state = 17},
  [836] = {.lex_state = 0, .external_lex_state = 17},
  [837] = {.lex_state = 0, .external_lex_state = 44},
  [838] = {.lex_state = 0, .external_lex_state = 17},
  [839] = {.lex_state = 0, .external_lex_state = 44},
  [840] = {.lex_state = 0, .external_lex_state = 17},
  [841] = {.lex_state = 0, .external_lex_state = 17},
  [842] = {.lex_state = 0, .external_lex_state = 17},
  [843] = {.lex_state = 0, .external_lex_state = 40},
  [844] = {.lex_state = 0, .external_lex_state = 17},
  [845] = {.lex_state = 0, .external_lex_state = 17},
  [846] = {.lex_state = 0, .external_lex_state = 17},
  [847] = {.lex_state = 0, .external_lex_state = 17},
  [848] = {.lex_state = 0, .external_lex_state = 17},
  [849] = {.lex_state = 0, .external_lex_state = 17},
  [850] = {.lex_state = 0, .external_lex_state = 40},
  [851] = {.lex_state = 0, .external_lex_state = 41},
  [852] = {.lex_state = 2030},
  [853] = {.lex_state = 0, .external_lex_state = 17},
  [854] = {.lex_state = 0, .external_lex_state = 41},
  [855] = {.lex_state = 0},
  [856] = {.lex_state = 0},
  [857] = {.lex_state = 0, .external_lex_state = 17},
  [858] = {.lex_state = 0, .external_lex_state = 17},
  [859] = {.lex_state = 0, .external_lex_state = 17},
  [860] = {.lex_state = 0, .external_lex_state = 45},
  [861] = {.lex_state = 0, .external_lex_state = 46},
  [862] = {.lex_state = 0, .external_lex_state = 46},
  [863] = {.lex_state = 0, .external_lex_state = 46},
  [864] = {.lex_state = 0},
  [865] = {.lex_state = 0, .external_lex_state = 45},
  [866] = {.lex_state = 0},
  [867] = {.lex_state = 0, .external_lex_state = 45},
  [868] = {.lex_state = 0, .external_lex_state = 46},
  [869] = {.lex_state = 0, .external_lex_state = 46},
  [870] = {.lex_state = 0},
  [871] = {.lex_state = 0, .external_lex_state = 45},
  [872] = {.lex_state = 0, .external_lex_state = 47},
  [873] = {.lex_state = 0, .external_lex_state = 46},
  [874] = {.lex_state = 0, .external_lex_state = 45},
  [875] = {.lex_state = 0},
  [876] = {.lex_state = 0, .external_lex_state = 46},
  [877] = {.lex_state = 0, .external_lex_state = 46},
  [878] = {.lex_state = 0, .external_lex_state = 45},
  [879] = {.lex_state = 0, .external_lex_state = 46},
  [880] = {.lex_state = 0, .external_lex_state = 47},
  [881] = {.lex_state = 0, .external_lex_state = 46},
  [882] = {.lex_state = 0, .external_lex_state = 46},
  [883] = {.lex_state = 0, .external_lex_state = 46},
  [884] = {.lex_state = 0, .external_lex_state = 46},
  [885] = {.lex_state = 0, .external_lex_state = 46},
  [886] = {.lex_state = 0, .external_lex_state = 46},
  [887] = {.lex_state = 0},
  [888] = {.lex_state = 0, .external_lex_state = 46},
  [889] = {.lex_state = 0, .external_lex_state = 46},
  [890] = {.lex_state = 0, .external_lex_state = 46},
  [891] = {.lex_state = 0, .external_lex_state = 46},
  [892] = {.lex_state = 0, .external_lex_state = 46},
  [893] = {.lex_state = 0},
  [894] = {.lex_state = 0, .external_lex_state = 46},
  [895] = {.lex_state = 0, .external_lex_state = 45},
  [896] = {.lex_state = 0, .external_lex_state = 46},
  [897] = {.lex_state = 0, .external_lex_state = 46},
  [898] = {.lex_state = 0, .external_lex_state = 47},
  [899] = {.lex_state = 0, .external_lex_state = 46},
  [900] = {.lex_state = 0, .external_lex_state = 45},
  [901] = {.lex_state = 0},
  [902] = {.lex_state = 0, .external_lex_state = 45},
  [903] = {.lex_state = 0, .external_lex_state = 45},
  [904] = {.lex_state = 0, .external_lex_state = 45},
  [905] = {.lex_state = 0},
  [906] = {.lex_state = 0, .external_lex_state = 45},
  [907] = {.lex_state = 0, .external_lex_state = 47},
  [908] = {.lex_state = 0, .external_lex_state = 46},
  [909] = {.lex_state = 0, .external_lex_state = 46},
  [910] = {.lex_state = 0},
  [911] = {.lex_state = 0, .external_lex_state = 46},
  [912] = {.lex_state = 0, .external_lex_state = 46},
  [913] = {.lex_state = 0, .external_lex_state = 17},
  [914] = {.lex_state = 0, .external_lex_state = 17},
  [915] = {.lex_state = 0},
  [916] = {.lex_state = 0},
  [917] = {.lex_state = 0, .external_lex_state = 46},
  [918] = {.lex_state = 0},
  [919] = {.lex_state = 0, .external_lex_state = 45},
  [920] = {.lex_state = 0, .external_lex_state = 45},
  [921] = {.lex_state = 0, .external_lex_state = 45},
  [922] = {.lex_state = 0, .external_lex_state = 45},
  [923] = {.lex_state = 0, .external_lex_state = 46},
  [924] = {.lex_state = 0},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__backslash_escape] = ACTIONS(1),
    [sym_entity_reference] = ACTIONS(1),
    [sym_numeric_character_reference] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_DASH_DASH_GT] = ACTIONS(1),
    [anon_sym_QMARK_GT] = ACTIONS(1),
    [anon_sym_RBRACK_RBRACK_GT] = ACTIONS(1),
    [aux_sym__word_token1] = ACTIONS(1),
    [aux_sym__word_token2] = ACTIONS(1),
    [aux_sym__word_token3] = ACTIONS(1),
    [sym__whitespace] = ACTIONS(1),
    [sym__line_ending] = ACTIONS(1),
    [sym__soft_line_ending] = ACTIONS(1),
    [sym__block_close] = ACTIONS(1),
    [sym_block_continuation] = ACTIONS(1),
    [sym__block_quote_start] = ACTIONS(1),
    [sym__indented_chunk_start] = ACTIONS(1),
    [sym_atx_h1_marker] = ACTIONS(1),
    [sym_atx_h2_marker] = ACTIONS(1),
    [sym_atx_h3_marker] = ACTIONS(1),
    [sym_atx_h4_marker] = ACTIONS(1),
    [sym_atx_h5_marker] = ACTIONS(1),
    [sym_atx_h6_marker] = ACTIONS(1),
    [sym_setext_h1_underline] = ACTIONS(1),
    [sym_setext_h2_underline] = ACTIONS(1),
    [sym__thematic_break] = ACTIONS(1),
    [sym__list_marker_minus] = ACTIONS(1),
    [sym__list_marker_plus] = ACTIONS(1),
    [sym__list_marker_star] = ACTIONS(1),
    [sym__list_marker_parenthesis] = ACTIONS(1),
    [sym__list_marker_dot] = ACTIONS(1),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1),
    [sym__blank_line_start] = ACTIONS(1),
    [sym__fenced_code_block_end_backtick] = ACTIONS(1),
    [sym__fenced_code_block_end_tilde] = ACTIONS(1),
    [sym__html_block_1_start] = ACTIONS(1),
    [sym__html_block_1_end] = ACTIONS(1),
    [sym__html_block_2_start] = ACTIONS(1),
    [sym__html_block_3_start] = ACTIONS(1),
    [sym__html_block_4_start] = ACTIONS(1),
    [sym__html_block_5_start] = ACTIONS(1),
    [sym__html_block_6_start] = ACTIONS(1),
    [sym__html_block_7_start] = ACTIONS(1),
    [sym__close_block] = ACTIONS(1),
    [sym__no_indented_chunk] = ACTIONS(1),
    [sym__error] = ACTIONS(1),
    [sym__trigger_error] = ACTIONS(1),
    [sym__eof] = ACTIONS(1),
    [sym_minus_metadata] = ACTIONS(1),
    [sym_plus_metadata] = ACTIONS(1),
    [sym__pipe_table_start] = ACTIONS(1),
    [sym__pipe_table_line_ending] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_document] = STATE(875),
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_section] = STATE(635),
    [sym__section1] = STATE(661),
    [sym__section2] = STATE(661),
    [sym__section3] = STATE(661),
    [sym__section4] = STATE(661),
    [sym__section5] = STATE(661),
    [sym__section6] = STATE(661),
    [sym_thematic_break] = STATE(339),
    [sym__atx_heading1] = STATE(38),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(30),
    [aux_sym_document_repeat2] = STATE(635),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(3),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(17),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym_minus_metadata] = ACTIONS(61),
    [sym_plus_metadata] = ACTIONS(61),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(2)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(879),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(71),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(3)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(917),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(113),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(4)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(912),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(115),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(5)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(891),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(117),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(6)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(884),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(119),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(7)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(877),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(121),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(8)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(881),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(123),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(9)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(882),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(125),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(10)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(885),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(127),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(11)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(888),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym_block_continuation] = ACTIONS(129),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(12)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(868),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(13)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(909),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(14)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(892),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(15)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(862),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(16)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(869),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(17)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(897),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(18)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(923),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(19)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(896),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(20)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(861),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(21)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(27),
    [sym__block_not_section] = STATE(27),
    [sym_section] = STATE(27),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(27),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(27),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(27),
    [sym_html_block] = STATE(27),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(27),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(170),
    [sym_block_quote] = STATE(27),
    [sym_list] = STATE(27),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__list_item_content] = STATE(863),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_task_list_marker_checked] = STATE(915),
    [sym_task_list_marker_unchecked] = STATE(915),
    [sym_pipe_table] = STATE(27),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(27),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(65),
    [aux_sym__word_token3] = ACTIONS(67),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(22)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(24),
    [sym__block_not_section] = STATE(24),
    [sym_section] = STATE(24),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(24),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(24),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(24),
    [sym_html_block] = STATE(24),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(24),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(24),
    [sym_block_quote] = STATE(24),
    [sym_list] = STATE(24),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(24),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(24),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(131),
    [sym_block_continuation] = ACTIONS(133),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(23)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(32),
    [sym__block_not_section] = STATE(32),
    [sym_section] = STATE(32),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(32),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(32),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(32),
    [sym_html_block] = STATE(32),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(32),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(32),
    [sym_block_quote] = STATE(32),
    [sym_list] = STATE(32),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(32),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(32),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(135),
    [sym_block_continuation] = ACTIONS(137),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(24)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(139),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(25)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(141),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(26)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(25),
    [sym__block_not_section] = STATE(25),
    [sym_section] = STATE(25),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(25),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(25),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(25),
    [sym_html_block] = STATE(25),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(25),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(25),
    [sym_block_quote] = STATE(25),
    [sym_list] = STATE(25),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(25),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(25),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(143),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(27)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(145),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(28)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(147),
    [anon_sym_RBRACK] = ACTIONS(150),
    [anon_sym_LT] = ACTIONS(150),
    [anon_sym_GT] = ACTIONS(150),
    [anon_sym_BANG] = ACTIONS(150),
    [anon_sym_DQUOTE] = ACTIONS(150),
    [anon_sym_POUND] = ACTIONS(150),
    [anon_sym_DOLLAR] = ACTIONS(150),
    [anon_sym_PERCENT] = ACTIONS(150),
    [anon_sym_AMP] = ACTIONS(150),
    [anon_sym_SQUOTE] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(150),
    [anon_sym_PLUS] = ACTIONS(150),
    [anon_sym_COMMA] = ACTIONS(150),
    [anon_sym_DASH] = ACTIONS(150),
    [anon_sym_DOT] = ACTIONS(150),
    [anon_sym_SLASH] = ACTIONS(150),
    [anon_sym_COLON] = ACTIONS(150),
    [anon_sym_SEMI] = ACTIONS(150),
    [anon_sym_EQ] = ACTIONS(150),
    [anon_sym_QMARK] = ACTIONS(150),
    [anon_sym_AT] = ACTIONS(150),
    [anon_sym_BSLASH] = ACTIONS(150),
    [anon_sym_CARET] = ACTIONS(150),
    [anon_sym__] = ACTIONS(150),
    [anon_sym_BQUOTE] = ACTIONS(150),
    [anon_sym_LBRACE] = ACTIONS(150),
    [anon_sym_PIPE] = ACTIONS(150),
    [anon_sym_RBRACE] = ACTIONS(150),
    [anon_sym_TILDE] = ACTIONS(150),
    [anon_sym_LPAREN] = ACTIONS(150),
    [anon_sym_RPAREN] = ACTIONS(150),
    [aux_sym__word_token1] = ACTIONS(150),
    [aux_sym__word_token2] = ACTIONS(150),
    [aux_sym__word_token3] = ACTIONS(150),
    [sym__whitespace] = ACTIONS(153),
    [sym__soft_line_ending] = ACTIONS(156),
    [sym__block_close] = ACTIONS(159),
    [sym__block_quote_start] = ACTIONS(161),
    [sym__indented_chunk_start] = ACTIONS(164),
    [sym_atx_h1_marker] = ACTIONS(167),
    [sym_atx_h2_marker] = ACTIONS(170),
    [sym_atx_h3_marker] = ACTIONS(173),
    [sym_atx_h4_marker] = ACTIONS(176),
    [sym_atx_h5_marker] = ACTIONS(179),
    [sym_atx_h6_marker] = ACTIONS(182),
    [sym__thematic_break] = ACTIONS(185),
    [sym__list_marker_minus] = ACTIONS(188),
    [sym__list_marker_plus] = ACTIONS(191),
    [sym__list_marker_star] = ACTIONS(194),
    [sym__list_marker_parenthesis] = ACTIONS(197),
    [sym__list_marker_dot] = ACTIONS(200),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(188),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(191),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(194),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(197),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(200),
    [sym__fenced_code_block_start_backtick] = ACTIONS(203),
    [sym__fenced_code_block_start_tilde] = ACTIONS(206),
    [sym__blank_line_start] = ACTIONS(209),
    [sym__html_block_1_start] = ACTIONS(212),
    [sym__html_block_2_start] = ACTIONS(215),
    [sym__html_block_3_start] = ACTIONS(218),
    [sym__html_block_4_start] = ACTIONS(221),
    [sym__html_block_5_start] = ACTIONS(224),
    [sym__html_block_6_start] = ACTIONS(227),
    [sym__html_block_7_start] = ACTIONS(230),
    [sym__pipe_table_start] = ACTIONS(233),
  },
  [STATE(29)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(236),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(30)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_section] = STATE(638),
    [sym__section1] = STATE(661),
    [sym__section2] = STATE(661),
    [sym__section3] = STATE(661),
    [sym__section4] = STATE(661),
    [sym__section5] = STATE(661),
    [sym__section6] = STATE(661),
    [sym_thematic_break] = STATE(339),
    [sym__atx_heading1] = STATE(38),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(71),
    [aux_sym_document_repeat2] = STATE(638),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(238),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(17),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(31)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_section] = STATE(637),
    [sym__section1] = STATE(661),
    [sym__section2] = STATE(661),
    [sym__section3] = STATE(661),
    [sym__section4] = STATE(661),
    [sym__section5] = STATE(661),
    [sym__section6] = STATE(661),
    [sym_thematic_break] = STATE(339),
    [sym__atx_heading1] = STATE(38),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(71),
    [aux_sym_document_repeat2] = STATE(637),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(240),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(17),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(32)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(242),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(33)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(34),
    [sym__block_not_section] = STATE(34),
    [sym_section] = STATE(34),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(34),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(34),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(34),
    [sym_html_block] = STATE(34),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(34),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(34),
    [sym_block_quote] = STATE(34),
    [sym_list] = STATE(34),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(34),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(34),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(242),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(34)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(28),
    [sym__block_not_section] = STATE(28),
    [sym_section] = STATE(28),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(28),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(28),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(28),
    [sym_html_block] = STATE(28),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(28),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(28),
    [sym_block_quote] = STATE(28),
    [sym_list] = STATE(28),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(28),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(28),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(244),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(35)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_section] = STATE(636),
    [sym__section1] = STATE(661),
    [sym__section2] = STATE(661),
    [sym__section3] = STATE(661),
    [sym__section4] = STATE(661),
    [sym__section5] = STATE(661),
    [sym__section6] = STATE(661),
    [sym_thematic_break] = STATE(339),
    [sym__atx_heading1] = STATE(38),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(31),
    [aux_sym_document_repeat2] = STATE(636),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(246),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(17),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(36)] = {
    [sym_link_label] = STATE(916),
    [sym__block] = STATE(29),
    [sym__block_not_section] = STATE(29),
    [sym_section] = STATE(29),
    [sym__section1] = STATE(196),
    [sym__section2] = STATE(196),
    [sym__section3] = STATE(196),
    [sym__section4] = STATE(196),
    [sym__section5] = STATE(196),
    [sym__section6] = STATE(196),
    [sym_thematic_break] = STATE(29),
    [sym__atx_heading1] = STATE(37),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(29),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(29),
    [sym_html_block] = STATE(29),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(29),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(29),
    [sym_block_quote] = STATE(29),
    [sym_list] = STATE(29),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(29),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym_block_quote_repeat1] = STATE(29),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(139),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(77),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(37)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(42),
    [sym__section2] = STATE(226),
    [sym__section3] = STATE(226),
    [sym__section4] = STATE(226),
    [sym__section5] = STATE(226),
    [sym__section6] = STATE(226),
    [sym_thematic_break] = STATE(42),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(42),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(42),
    [sym_html_block] = STATE(42),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(42),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(42),
    [sym_block_quote] = STATE(42),
    [sym_list] = STATE(42),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(42),
    [aux_sym__section1_repeat1] = STATE(42),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(248),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(248),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(38)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(41),
    [sym__section2] = STATE(229),
    [sym__section3] = STATE(229),
    [sym__section4] = STATE(229),
    [sym__section5] = STATE(229),
    [sym__section6] = STATE(229),
    [sym_thematic_break] = STATE(41),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(41),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(41),
    [sym_html_block] = STATE(41),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(41),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(41),
    [sym_block_quote] = STATE(41),
    [sym_list] = STATE(41),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(41),
    [aux_sym__section1_repeat1] = STATE(41),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(248),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(248),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(39)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(39),
    [sym__section2] = STATE(229),
    [sym__section3] = STATE(229),
    [sym__section4] = STATE(229),
    [sym__section5] = STATE(229),
    [sym__section6] = STATE(229),
    [sym_thematic_break] = STATE(39),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(39),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(39),
    [sym_html_block] = STATE(39),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(39),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(39),
    [sym_block_quote] = STATE(39),
    [sym_list] = STATE(39),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(39),
    [aux_sym__section1_repeat1] = STATE(39),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(250),
    [anon_sym_LBRACK] = ACTIONS(252),
    [anon_sym_RBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(255),
    [anon_sym_GT] = ACTIONS(255),
    [anon_sym_BANG] = ACTIONS(255),
    [anon_sym_DQUOTE] = ACTIONS(255),
    [anon_sym_POUND] = ACTIONS(255),
    [anon_sym_DOLLAR] = ACTIONS(255),
    [anon_sym_PERCENT] = ACTIONS(255),
    [anon_sym_AMP] = ACTIONS(255),
    [anon_sym_SQUOTE] = ACTIONS(255),
    [anon_sym_STAR] = ACTIONS(255),
    [anon_sym_PLUS] = ACTIONS(255),
    [anon_sym_COMMA] = ACTIONS(255),
    [anon_sym_DASH] = ACTIONS(255),
    [anon_sym_DOT] = ACTIONS(255),
    [anon_sym_SLASH] = ACTIONS(255),
    [anon_sym_COLON] = ACTIONS(255),
    [anon_sym_SEMI] = ACTIONS(255),
    [anon_sym_EQ] = ACTIONS(255),
    [anon_sym_QMARK] = ACTIONS(255),
    [anon_sym_AT] = ACTIONS(255),
    [anon_sym_BSLASH] = ACTIONS(255),
    [anon_sym_CARET] = ACTIONS(255),
    [anon_sym__] = ACTIONS(255),
    [anon_sym_BQUOTE] = ACTIONS(255),
    [anon_sym_LBRACE] = ACTIONS(255),
    [anon_sym_PIPE] = ACTIONS(255),
    [anon_sym_RBRACE] = ACTIONS(255),
    [anon_sym_TILDE] = ACTIONS(255),
    [anon_sym_LPAREN] = ACTIONS(255),
    [anon_sym_RPAREN] = ACTIONS(255),
    [aux_sym__word_token1] = ACTIONS(255),
    [aux_sym__word_token2] = ACTIONS(255),
    [aux_sym__word_token3] = ACTIONS(255),
    [sym__whitespace] = ACTIONS(258),
    [sym__soft_line_ending] = ACTIONS(261),
    [sym__block_quote_start] = ACTIONS(264),
    [sym__indented_chunk_start] = ACTIONS(267),
    [sym_atx_h1_marker] = ACTIONS(250),
    [sym_atx_h2_marker] = ACTIONS(270),
    [sym_atx_h3_marker] = ACTIONS(273),
    [sym_atx_h4_marker] = ACTIONS(276),
    [sym_atx_h5_marker] = ACTIONS(279),
    [sym_atx_h6_marker] = ACTIONS(282),
    [sym__thematic_break] = ACTIONS(285),
    [sym__list_marker_minus] = ACTIONS(288),
    [sym__list_marker_plus] = ACTIONS(291),
    [sym__list_marker_star] = ACTIONS(294),
    [sym__list_marker_parenthesis] = ACTIONS(297),
    [sym__list_marker_dot] = ACTIONS(300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(288),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(291),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(297),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(303),
    [sym__fenced_code_block_start_tilde] = ACTIONS(306),
    [sym__blank_line_start] = ACTIONS(309),
    [sym__html_block_1_start] = ACTIONS(312),
    [sym__html_block_2_start] = ACTIONS(315),
    [sym__html_block_3_start] = ACTIONS(318),
    [sym__html_block_4_start] = ACTIONS(321),
    [sym__html_block_5_start] = ACTIONS(324),
    [sym__html_block_6_start] = ACTIONS(327),
    [sym__html_block_7_start] = ACTIONS(330),
    [sym__pipe_table_start] = ACTIONS(333),
  },
  [STATE(40)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(40),
    [sym__section2] = STATE(226),
    [sym__section3] = STATE(226),
    [sym__section4] = STATE(226),
    [sym__section5] = STATE(226),
    [sym__section6] = STATE(226),
    [sym_thematic_break] = STATE(40),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(40),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(40),
    [sym_html_block] = STATE(40),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(40),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(40),
    [sym_block_quote] = STATE(40),
    [sym_list] = STATE(40),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(40),
    [aux_sym__section1_repeat1] = STATE(40),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(252),
    [anon_sym_RBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(255),
    [anon_sym_GT] = ACTIONS(255),
    [anon_sym_BANG] = ACTIONS(255),
    [anon_sym_DQUOTE] = ACTIONS(255),
    [anon_sym_POUND] = ACTIONS(255),
    [anon_sym_DOLLAR] = ACTIONS(255),
    [anon_sym_PERCENT] = ACTIONS(255),
    [anon_sym_AMP] = ACTIONS(255),
    [anon_sym_SQUOTE] = ACTIONS(255),
    [anon_sym_STAR] = ACTIONS(255),
    [anon_sym_PLUS] = ACTIONS(255),
    [anon_sym_COMMA] = ACTIONS(255),
    [anon_sym_DASH] = ACTIONS(255),
    [anon_sym_DOT] = ACTIONS(255),
    [anon_sym_SLASH] = ACTIONS(255),
    [anon_sym_COLON] = ACTIONS(255),
    [anon_sym_SEMI] = ACTIONS(255),
    [anon_sym_EQ] = ACTIONS(255),
    [anon_sym_QMARK] = ACTIONS(255),
    [anon_sym_AT] = ACTIONS(255),
    [anon_sym_BSLASH] = ACTIONS(255),
    [anon_sym_CARET] = ACTIONS(255),
    [anon_sym__] = ACTIONS(255),
    [anon_sym_BQUOTE] = ACTIONS(255),
    [anon_sym_LBRACE] = ACTIONS(255),
    [anon_sym_PIPE] = ACTIONS(255),
    [anon_sym_RBRACE] = ACTIONS(255),
    [anon_sym_TILDE] = ACTIONS(255),
    [anon_sym_LPAREN] = ACTIONS(255),
    [anon_sym_RPAREN] = ACTIONS(255),
    [aux_sym__word_token1] = ACTIONS(255),
    [aux_sym__word_token2] = ACTIONS(255),
    [aux_sym__word_token3] = ACTIONS(255),
    [sym__whitespace] = ACTIONS(336),
    [sym__soft_line_ending] = ACTIONS(261),
    [sym__block_close] = ACTIONS(250),
    [sym__block_quote_start] = ACTIONS(339),
    [sym__indented_chunk_start] = ACTIONS(342),
    [sym_atx_h1_marker] = ACTIONS(250),
    [sym_atx_h2_marker] = ACTIONS(345),
    [sym_atx_h3_marker] = ACTIONS(348),
    [sym_atx_h4_marker] = ACTIONS(351),
    [sym_atx_h5_marker] = ACTIONS(354),
    [sym_atx_h6_marker] = ACTIONS(357),
    [sym__thematic_break] = ACTIONS(360),
    [sym__list_marker_minus] = ACTIONS(288),
    [sym__list_marker_plus] = ACTIONS(291),
    [sym__list_marker_star] = ACTIONS(294),
    [sym__list_marker_parenthesis] = ACTIONS(297),
    [sym__list_marker_dot] = ACTIONS(300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(288),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(291),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(297),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(363),
    [sym__fenced_code_block_start_tilde] = ACTIONS(366),
    [sym__blank_line_start] = ACTIONS(369),
    [sym__html_block_1_start] = ACTIONS(372),
    [sym__html_block_2_start] = ACTIONS(375),
    [sym__html_block_3_start] = ACTIONS(378),
    [sym__html_block_4_start] = ACTIONS(381),
    [sym__html_block_5_start] = ACTIONS(384),
    [sym__html_block_6_start] = ACTIONS(387),
    [sym__html_block_7_start] = ACTIONS(390),
    [sym__pipe_table_start] = ACTIONS(393),
  },
  [STATE(41)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(39),
    [sym__section2] = STATE(229),
    [sym__section3] = STATE(229),
    [sym__section4] = STATE(229),
    [sym__section5] = STATE(229),
    [sym__section6] = STATE(229),
    [sym_thematic_break] = STATE(39),
    [sym__atx_heading2] = STATE(45),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(39),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(39),
    [sym_html_block] = STATE(39),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(39),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(39),
    [sym_block_quote] = STATE(39),
    [sym_list] = STATE(39),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(39),
    [aux_sym__section1_repeat1] = STATE(39),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(396),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(396),
    [sym_atx_h2_marker] = ACTIONS(19),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(42)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(40),
    [sym__section2] = STATE(226),
    [sym__section3] = STATE(226),
    [sym__section4] = STATE(226),
    [sym__section5] = STATE(226),
    [sym__section6] = STATE(226),
    [sym_thematic_break] = STATE(40),
    [sym__atx_heading2] = STATE(46),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(40),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(40),
    [sym_html_block] = STATE(40),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(40),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(40),
    [sym_block_quote] = STATE(40),
    [sym_list] = STATE(40),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(40),
    [aux_sym__section1_repeat1] = STATE(40),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(396),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(396),
    [sym_atx_h2_marker] = ACTIONS(79),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(43)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(43),
    [sym__section3] = STATE(236),
    [sym__section4] = STATE(236),
    [sym__section5] = STATE(236),
    [sym__section6] = STATE(236),
    [sym_thematic_break] = STATE(43),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(43),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(43),
    [sym_html_block] = STATE(43),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(43),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(43),
    [sym_block_quote] = STATE(43),
    [sym_list] = STATE(43),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(43),
    [aux_sym__section2_repeat1] = STATE(43),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(398),
    [anon_sym_LBRACK] = ACTIONS(400),
    [anon_sym_RBRACK] = ACTIONS(403),
    [anon_sym_LT] = ACTIONS(403),
    [anon_sym_GT] = ACTIONS(403),
    [anon_sym_BANG] = ACTIONS(403),
    [anon_sym_DQUOTE] = ACTIONS(403),
    [anon_sym_POUND] = ACTIONS(403),
    [anon_sym_DOLLAR] = ACTIONS(403),
    [anon_sym_PERCENT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(403),
    [anon_sym_SQUOTE] = ACTIONS(403),
    [anon_sym_STAR] = ACTIONS(403),
    [anon_sym_PLUS] = ACTIONS(403),
    [anon_sym_COMMA] = ACTIONS(403),
    [anon_sym_DASH] = ACTIONS(403),
    [anon_sym_DOT] = ACTIONS(403),
    [anon_sym_SLASH] = ACTIONS(403),
    [anon_sym_COLON] = ACTIONS(403),
    [anon_sym_SEMI] = ACTIONS(403),
    [anon_sym_EQ] = ACTIONS(403),
    [anon_sym_QMARK] = ACTIONS(403),
    [anon_sym_AT] = ACTIONS(403),
    [anon_sym_BSLASH] = ACTIONS(403),
    [anon_sym_CARET] = ACTIONS(403),
    [anon_sym__] = ACTIONS(403),
    [anon_sym_BQUOTE] = ACTIONS(403),
    [anon_sym_LBRACE] = ACTIONS(403),
    [anon_sym_PIPE] = ACTIONS(403),
    [anon_sym_RBRACE] = ACTIONS(403),
    [anon_sym_TILDE] = ACTIONS(403),
    [anon_sym_LPAREN] = ACTIONS(403),
    [anon_sym_RPAREN] = ACTIONS(403),
    [aux_sym__word_token1] = ACTIONS(403),
    [aux_sym__word_token2] = ACTIONS(403),
    [aux_sym__word_token3] = ACTIONS(403),
    [sym__whitespace] = ACTIONS(406),
    [sym__soft_line_ending] = ACTIONS(409),
    [sym__block_quote_start] = ACTIONS(412),
    [sym__indented_chunk_start] = ACTIONS(415),
    [sym_atx_h1_marker] = ACTIONS(398),
    [sym_atx_h2_marker] = ACTIONS(398),
    [sym_atx_h3_marker] = ACTIONS(418),
    [sym_atx_h4_marker] = ACTIONS(421),
    [sym_atx_h5_marker] = ACTIONS(424),
    [sym_atx_h6_marker] = ACTIONS(427),
    [sym__thematic_break] = ACTIONS(430),
    [sym__list_marker_minus] = ACTIONS(433),
    [sym__list_marker_plus] = ACTIONS(436),
    [sym__list_marker_star] = ACTIONS(439),
    [sym__list_marker_parenthesis] = ACTIONS(442),
    [sym__list_marker_dot] = ACTIONS(445),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(433),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(436),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(439),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(442),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(445),
    [sym__fenced_code_block_start_backtick] = ACTIONS(448),
    [sym__fenced_code_block_start_tilde] = ACTIONS(451),
    [sym__blank_line_start] = ACTIONS(454),
    [sym__html_block_1_start] = ACTIONS(457),
    [sym__html_block_2_start] = ACTIONS(460),
    [sym__html_block_3_start] = ACTIONS(463),
    [sym__html_block_4_start] = ACTIONS(466),
    [sym__html_block_5_start] = ACTIONS(469),
    [sym__html_block_6_start] = ACTIONS(472),
    [sym__html_block_7_start] = ACTIONS(475),
    [sym__pipe_table_start] = ACTIONS(478),
  },
  [STATE(44)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(43),
    [sym__section3] = STATE(236),
    [sym__section4] = STATE(236),
    [sym__section5] = STATE(236),
    [sym__section6] = STATE(236),
    [sym_thematic_break] = STATE(43),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(43),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(43),
    [sym_html_block] = STATE(43),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(43),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(43),
    [sym_block_quote] = STATE(43),
    [sym_list] = STATE(43),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(43),
    [aux_sym__section2_repeat1] = STATE(43),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(481),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(481),
    [sym_atx_h2_marker] = ACTIONS(481),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(45)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(44),
    [sym__section3] = STATE(236),
    [sym__section4] = STATE(236),
    [sym__section5] = STATE(236),
    [sym__section6] = STATE(236),
    [sym_thematic_break] = STATE(44),
    [sym__atx_heading3] = STATE(54),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(44),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(44),
    [sym_html_block] = STATE(44),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(44),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(44),
    [sym_block_quote] = STATE(44),
    [sym_list] = STATE(44),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(44),
    [aux_sym__section2_repeat1] = STATE(44),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(483),
    [sym_atx_h2_marker] = ACTIONS(483),
    [sym_atx_h3_marker] = ACTIONS(21),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(46)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(47),
    [sym__section3] = STATE(228),
    [sym__section4] = STATE(228),
    [sym__section5] = STATE(228),
    [sym__section6] = STATE(228),
    [sym_thematic_break] = STATE(47),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(47),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(47),
    [sym_html_block] = STATE(47),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(47),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(47),
    [sym_block_quote] = STATE(47),
    [sym_list] = STATE(47),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(47),
    [aux_sym__section2_repeat1] = STATE(47),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(483),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(483),
    [sym_atx_h2_marker] = ACTIONS(483),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(47)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(48),
    [sym__section3] = STATE(228),
    [sym__section4] = STATE(228),
    [sym__section5] = STATE(228),
    [sym__section6] = STATE(228),
    [sym_thematic_break] = STATE(48),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(48),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(48),
    [sym_html_block] = STATE(48),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(48),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(48),
    [sym_block_quote] = STATE(48),
    [sym_list] = STATE(48),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(48),
    [aux_sym__section2_repeat1] = STATE(48),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(481),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(481),
    [sym_atx_h2_marker] = ACTIONS(481),
    [sym_atx_h3_marker] = ACTIONS(81),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(48)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(48),
    [sym__section3] = STATE(228),
    [sym__section4] = STATE(228),
    [sym__section5] = STATE(228),
    [sym__section6] = STATE(228),
    [sym_thematic_break] = STATE(48),
    [sym__atx_heading3] = STATE(51),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(48),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(48),
    [sym_html_block] = STATE(48),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(48),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(48),
    [sym_block_quote] = STATE(48),
    [sym_list] = STATE(48),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(48),
    [aux_sym__section2_repeat1] = STATE(48),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(400),
    [anon_sym_RBRACK] = ACTIONS(403),
    [anon_sym_LT] = ACTIONS(403),
    [anon_sym_GT] = ACTIONS(403),
    [anon_sym_BANG] = ACTIONS(403),
    [anon_sym_DQUOTE] = ACTIONS(403),
    [anon_sym_POUND] = ACTIONS(403),
    [anon_sym_DOLLAR] = ACTIONS(403),
    [anon_sym_PERCENT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(403),
    [anon_sym_SQUOTE] = ACTIONS(403),
    [anon_sym_STAR] = ACTIONS(403),
    [anon_sym_PLUS] = ACTIONS(403),
    [anon_sym_COMMA] = ACTIONS(403),
    [anon_sym_DASH] = ACTIONS(403),
    [anon_sym_DOT] = ACTIONS(403),
    [anon_sym_SLASH] = ACTIONS(403),
    [anon_sym_COLON] = ACTIONS(403),
    [anon_sym_SEMI] = ACTIONS(403),
    [anon_sym_EQ] = ACTIONS(403),
    [anon_sym_QMARK] = ACTIONS(403),
    [anon_sym_AT] = ACTIONS(403),
    [anon_sym_BSLASH] = ACTIONS(403),
    [anon_sym_CARET] = ACTIONS(403),
    [anon_sym__] = ACTIONS(403),
    [anon_sym_BQUOTE] = ACTIONS(403),
    [anon_sym_LBRACE] = ACTIONS(403),
    [anon_sym_PIPE] = ACTIONS(403),
    [anon_sym_RBRACE] = ACTIONS(403),
    [anon_sym_TILDE] = ACTIONS(403),
    [anon_sym_LPAREN] = ACTIONS(403),
    [anon_sym_RPAREN] = ACTIONS(403),
    [aux_sym__word_token1] = ACTIONS(403),
    [aux_sym__word_token2] = ACTIONS(403),
    [aux_sym__word_token3] = ACTIONS(403),
    [sym__whitespace] = ACTIONS(485),
    [sym__soft_line_ending] = ACTIONS(409),
    [sym__block_close] = ACTIONS(398),
    [sym__block_quote_start] = ACTIONS(488),
    [sym__indented_chunk_start] = ACTIONS(491),
    [sym_atx_h1_marker] = ACTIONS(398),
    [sym_atx_h2_marker] = ACTIONS(398),
    [sym_atx_h3_marker] = ACTIONS(494),
    [sym_atx_h4_marker] = ACTIONS(497),
    [sym_atx_h5_marker] = ACTIONS(500),
    [sym_atx_h6_marker] = ACTIONS(503),
    [sym__thematic_break] = ACTIONS(506),
    [sym__list_marker_minus] = ACTIONS(433),
    [sym__list_marker_plus] = ACTIONS(436),
    [sym__list_marker_star] = ACTIONS(439),
    [sym__list_marker_parenthesis] = ACTIONS(442),
    [sym__list_marker_dot] = ACTIONS(445),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(433),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(436),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(439),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(442),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(445),
    [sym__fenced_code_block_start_backtick] = ACTIONS(509),
    [sym__fenced_code_block_start_tilde] = ACTIONS(512),
    [sym__blank_line_start] = ACTIONS(515),
    [sym__html_block_1_start] = ACTIONS(518),
    [sym__html_block_2_start] = ACTIONS(521),
    [sym__html_block_3_start] = ACTIONS(524),
    [sym__html_block_4_start] = ACTIONS(527),
    [sym__html_block_5_start] = ACTIONS(530),
    [sym__html_block_6_start] = ACTIONS(533),
    [sym__html_block_7_start] = ACTIONS(536),
    [sym__pipe_table_start] = ACTIONS(539),
  },
  [STATE(49)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(49),
    [sym__section4] = STATE(238),
    [sym__section5] = STATE(238),
    [sym__section6] = STATE(238),
    [sym_thematic_break] = STATE(49),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(49),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(49),
    [sym_html_block] = STATE(49),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(49),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(49),
    [sym_block_quote] = STATE(49),
    [sym_list] = STATE(49),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(49),
    [aux_sym__section3_repeat1] = STATE(49),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(542),
    [anon_sym_LBRACK] = ACTIONS(544),
    [anon_sym_RBRACK] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(547),
    [anon_sym_GT] = ACTIONS(547),
    [anon_sym_BANG] = ACTIONS(547),
    [anon_sym_DQUOTE] = ACTIONS(547),
    [anon_sym_POUND] = ACTIONS(547),
    [anon_sym_DOLLAR] = ACTIONS(547),
    [anon_sym_PERCENT] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(547),
    [anon_sym_SQUOTE] = ACTIONS(547),
    [anon_sym_STAR] = ACTIONS(547),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_COMMA] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_DOT] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(547),
    [anon_sym_COLON] = ACTIONS(547),
    [anon_sym_SEMI] = ACTIONS(547),
    [anon_sym_EQ] = ACTIONS(547),
    [anon_sym_QMARK] = ACTIONS(547),
    [anon_sym_AT] = ACTIONS(547),
    [anon_sym_BSLASH] = ACTIONS(547),
    [anon_sym_CARET] = ACTIONS(547),
    [anon_sym__] = ACTIONS(547),
    [anon_sym_BQUOTE] = ACTIONS(547),
    [anon_sym_LBRACE] = ACTIONS(547),
    [anon_sym_PIPE] = ACTIONS(547),
    [anon_sym_RBRACE] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_LPAREN] = ACTIONS(547),
    [anon_sym_RPAREN] = ACTIONS(547),
    [aux_sym__word_token1] = ACTIONS(547),
    [aux_sym__word_token2] = ACTIONS(547),
    [aux_sym__word_token3] = ACTIONS(547),
    [sym__whitespace] = ACTIONS(550),
    [sym__soft_line_ending] = ACTIONS(553),
    [sym__block_quote_start] = ACTIONS(556),
    [sym__indented_chunk_start] = ACTIONS(559),
    [sym_atx_h1_marker] = ACTIONS(542),
    [sym_atx_h2_marker] = ACTIONS(542),
    [sym_atx_h3_marker] = ACTIONS(542),
    [sym_atx_h4_marker] = ACTIONS(562),
    [sym_atx_h5_marker] = ACTIONS(565),
    [sym_atx_h6_marker] = ACTIONS(568),
    [sym__thematic_break] = ACTIONS(571),
    [sym__list_marker_minus] = ACTIONS(574),
    [sym__list_marker_plus] = ACTIONS(577),
    [sym__list_marker_star] = ACTIONS(580),
    [sym__list_marker_parenthesis] = ACTIONS(583),
    [sym__list_marker_dot] = ACTIONS(586),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(574),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(577),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(580),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(583),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(586),
    [sym__fenced_code_block_start_backtick] = ACTIONS(589),
    [sym__fenced_code_block_start_tilde] = ACTIONS(592),
    [sym__blank_line_start] = ACTIONS(595),
    [sym__html_block_1_start] = ACTIONS(598),
    [sym__html_block_2_start] = ACTIONS(601),
    [sym__html_block_3_start] = ACTIONS(604),
    [sym__html_block_4_start] = ACTIONS(607),
    [sym__html_block_5_start] = ACTIONS(610),
    [sym__html_block_6_start] = ACTIONS(613),
    [sym__html_block_7_start] = ACTIONS(616),
    [sym__pipe_table_start] = ACTIONS(619),
  },
  [STATE(50)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(49),
    [sym__section4] = STATE(238),
    [sym__section5] = STATE(238),
    [sym__section6] = STATE(238),
    [sym_thematic_break] = STATE(49),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(49),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(49),
    [sym_html_block] = STATE(49),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(49),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(49),
    [sym_block_quote] = STATE(49),
    [sym_list] = STATE(49),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(49),
    [aux_sym__section3_repeat1] = STATE(49),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(622),
    [sym_atx_h2_marker] = ACTIONS(622),
    [sym_atx_h3_marker] = ACTIONS(622),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(51)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(52),
    [sym__section4] = STATE(230),
    [sym__section5] = STATE(230),
    [sym__section6] = STATE(230),
    [sym_thematic_break] = STATE(52),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(52),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(52),
    [sym_html_block] = STATE(52),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(52),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(52),
    [sym_block_quote] = STATE(52),
    [sym_list] = STATE(52),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(52),
    [aux_sym__section3_repeat1] = STATE(52),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(624),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(624),
    [sym_atx_h2_marker] = ACTIONS(624),
    [sym_atx_h3_marker] = ACTIONS(624),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(52)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(53),
    [sym__section4] = STATE(230),
    [sym__section5] = STATE(230),
    [sym__section6] = STATE(230),
    [sym_thematic_break] = STATE(53),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(53),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(53),
    [sym_html_block] = STATE(53),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(53),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(53),
    [sym_block_quote] = STATE(53),
    [sym_list] = STATE(53),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(53),
    [aux_sym__section3_repeat1] = STATE(53),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(622),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(622),
    [sym_atx_h2_marker] = ACTIONS(622),
    [sym_atx_h3_marker] = ACTIONS(622),
    [sym_atx_h4_marker] = ACTIONS(83),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(53)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(53),
    [sym__section4] = STATE(230),
    [sym__section5] = STATE(230),
    [sym__section6] = STATE(230),
    [sym_thematic_break] = STATE(53),
    [sym__atx_heading4] = STATE(58),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(53),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(53),
    [sym_html_block] = STATE(53),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(53),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(53),
    [sym_block_quote] = STATE(53),
    [sym_list] = STATE(53),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(53),
    [aux_sym__section3_repeat1] = STATE(53),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(544),
    [anon_sym_RBRACK] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(547),
    [anon_sym_GT] = ACTIONS(547),
    [anon_sym_BANG] = ACTIONS(547),
    [anon_sym_DQUOTE] = ACTIONS(547),
    [anon_sym_POUND] = ACTIONS(547),
    [anon_sym_DOLLAR] = ACTIONS(547),
    [anon_sym_PERCENT] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(547),
    [anon_sym_SQUOTE] = ACTIONS(547),
    [anon_sym_STAR] = ACTIONS(547),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_COMMA] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_DOT] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(547),
    [anon_sym_COLON] = ACTIONS(547),
    [anon_sym_SEMI] = ACTIONS(547),
    [anon_sym_EQ] = ACTIONS(547),
    [anon_sym_QMARK] = ACTIONS(547),
    [anon_sym_AT] = ACTIONS(547),
    [anon_sym_BSLASH] = ACTIONS(547),
    [anon_sym_CARET] = ACTIONS(547),
    [anon_sym__] = ACTIONS(547),
    [anon_sym_BQUOTE] = ACTIONS(547),
    [anon_sym_LBRACE] = ACTIONS(547),
    [anon_sym_PIPE] = ACTIONS(547),
    [anon_sym_RBRACE] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_LPAREN] = ACTIONS(547),
    [anon_sym_RPAREN] = ACTIONS(547),
    [aux_sym__word_token1] = ACTIONS(547),
    [aux_sym__word_token2] = ACTIONS(547),
    [aux_sym__word_token3] = ACTIONS(547),
    [sym__whitespace] = ACTIONS(626),
    [sym__soft_line_ending] = ACTIONS(553),
    [sym__block_close] = ACTIONS(542),
    [sym__block_quote_start] = ACTIONS(629),
    [sym__indented_chunk_start] = ACTIONS(632),
    [sym_atx_h1_marker] = ACTIONS(542),
    [sym_atx_h2_marker] = ACTIONS(542),
    [sym_atx_h3_marker] = ACTIONS(542),
    [sym_atx_h4_marker] = ACTIONS(635),
    [sym_atx_h5_marker] = ACTIONS(638),
    [sym_atx_h6_marker] = ACTIONS(641),
    [sym__thematic_break] = ACTIONS(644),
    [sym__list_marker_minus] = ACTIONS(574),
    [sym__list_marker_plus] = ACTIONS(577),
    [sym__list_marker_star] = ACTIONS(580),
    [sym__list_marker_parenthesis] = ACTIONS(583),
    [sym__list_marker_dot] = ACTIONS(586),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(574),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(577),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(580),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(583),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(586),
    [sym__fenced_code_block_start_backtick] = ACTIONS(647),
    [sym__fenced_code_block_start_tilde] = ACTIONS(650),
    [sym__blank_line_start] = ACTIONS(653),
    [sym__html_block_1_start] = ACTIONS(656),
    [sym__html_block_2_start] = ACTIONS(659),
    [sym__html_block_3_start] = ACTIONS(662),
    [sym__html_block_4_start] = ACTIONS(665),
    [sym__html_block_5_start] = ACTIONS(668),
    [sym__html_block_6_start] = ACTIONS(671),
    [sym__html_block_7_start] = ACTIONS(674),
    [sym__pipe_table_start] = ACTIONS(677),
  },
  [STATE(54)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(50),
    [sym__section4] = STATE(238),
    [sym__section5] = STATE(238),
    [sym__section6] = STATE(238),
    [sym_thematic_break] = STATE(50),
    [sym__atx_heading4] = STATE(55),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(50),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(50),
    [sym_html_block] = STATE(50),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(50),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(50),
    [sym_block_quote] = STATE(50),
    [sym_list] = STATE(50),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(50),
    [aux_sym__section3_repeat1] = STATE(50),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(624),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(624),
    [sym_atx_h2_marker] = ACTIONS(624),
    [sym_atx_h3_marker] = ACTIONS(624),
    [sym_atx_h4_marker] = ACTIONS(23),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(55)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(59),
    [sym__section5] = STATE(240),
    [sym__section6] = STATE(240),
    [sym_thematic_break] = STATE(59),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(59),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(59),
    [sym_html_block] = STATE(59),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(59),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(59),
    [sym_block_quote] = STATE(59),
    [sym_list] = STATE(59),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(59),
    [aux_sym__section4_repeat1] = STATE(59),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(680),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(680),
    [sym_atx_h2_marker] = ACTIONS(680),
    [sym_atx_h3_marker] = ACTIONS(680),
    [sym_atx_h4_marker] = ACTIONS(680),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(56)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(56),
    [sym__section5] = STATE(240),
    [sym__section6] = STATE(240),
    [sym_thematic_break] = STATE(56),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(56),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(56),
    [sym_html_block] = STATE(56),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(56),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(56),
    [sym_block_quote] = STATE(56),
    [sym_list] = STATE(56),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(56),
    [aux_sym__section4_repeat1] = STATE(56),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(682),
    [anon_sym_LBRACK] = ACTIONS(684),
    [anon_sym_RBRACK] = ACTIONS(687),
    [anon_sym_LT] = ACTIONS(687),
    [anon_sym_GT] = ACTIONS(687),
    [anon_sym_BANG] = ACTIONS(687),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_POUND] = ACTIONS(687),
    [anon_sym_DOLLAR] = ACTIONS(687),
    [anon_sym_PERCENT] = ACTIONS(687),
    [anon_sym_AMP] = ACTIONS(687),
    [anon_sym_SQUOTE] = ACTIONS(687),
    [anon_sym_STAR] = ACTIONS(687),
    [anon_sym_PLUS] = ACTIONS(687),
    [anon_sym_COMMA] = ACTIONS(687),
    [anon_sym_DASH] = ACTIONS(687),
    [anon_sym_DOT] = ACTIONS(687),
    [anon_sym_SLASH] = ACTIONS(687),
    [anon_sym_COLON] = ACTIONS(687),
    [anon_sym_SEMI] = ACTIONS(687),
    [anon_sym_EQ] = ACTIONS(687),
    [anon_sym_QMARK] = ACTIONS(687),
    [anon_sym_AT] = ACTIONS(687),
    [anon_sym_BSLASH] = ACTIONS(687),
    [anon_sym_CARET] = ACTIONS(687),
    [anon_sym__] = ACTIONS(687),
    [anon_sym_BQUOTE] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(687),
    [anon_sym_PIPE] = ACTIONS(687),
    [anon_sym_RBRACE] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(687),
    [anon_sym_LPAREN] = ACTIONS(687),
    [anon_sym_RPAREN] = ACTIONS(687),
    [aux_sym__word_token1] = ACTIONS(687),
    [aux_sym__word_token2] = ACTIONS(687),
    [aux_sym__word_token3] = ACTIONS(687),
    [sym__whitespace] = ACTIONS(690),
    [sym__soft_line_ending] = ACTIONS(693),
    [sym__block_quote_start] = ACTIONS(696),
    [sym__indented_chunk_start] = ACTIONS(699),
    [sym_atx_h1_marker] = ACTIONS(682),
    [sym_atx_h2_marker] = ACTIONS(682),
    [sym_atx_h3_marker] = ACTIONS(682),
    [sym_atx_h4_marker] = ACTIONS(682),
    [sym_atx_h5_marker] = ACTIONS(702),
    [sym_atx_h6_marker] = ACTIONS(705),
    [sym__thematic_break] = ACTIONS(708),
    [sym__list_marker_minus] = ACTIONS(711),
    [sym__list_marker_plus] = ACTIONS(714),
    [sym__list_marker_star] = ACTIONS(717),
    [sym__list_marker_parenthesis] = ACTIONS(720),
    [sym__list_marker_dot] = ACTIONS(723),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(711),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(714),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(717),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(720),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(723),
    [sym__fenced_code_block_start_backtick] = ACTIONS(726),
    [sym__fenced_code_block_start_tilde] = ACTIONS(729),
    [sym__blank_line_start] = ACTIONS(732),
    [sym__html_block_1_start] = ACTIONS(735),
    [sym__html_block_2_start] = ACTIONS(738),
    [sym__html_block_3_start] = ACTIONS(741),
    [sym__html_block_4_start] = ACTIONS(744),
    [sym__html_block_5_start] = ACTIONS(747),
    [sym__html_block_6_start] = ACTIONS(750),
    [sym__html_block_7_start] = ACTIONS(753),
    [sym__pipe_table_start] = ACTIONS(756),
  },
  [STATE(57)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(57),
    [sym__section5] = STATE(232),
    [sym__section6] = STATE(232),
    [sym_thematic_break] = STATE(57),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(57),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(57),
    [sym_html_block] = STATE(57),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(57),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(57),
    [sym_block_quote] = STATE(57),
    [sym_list] = STATE(57),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(57),
    [aux_sym__section4_repeat1] = STATE(57),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(684),
    [anon_sym_RBRACK] = ACTIONS(687),
    [anon_sym_LT] = ACTIONS(687),
    [anon_sym_GT] = ACTIONS(687),
    [anon_sym_BANG] = ACTIONS(687),
    [anon_sym_DQUOTE] = ACTIONS(687),
    [anon_sym_POUND] = ACTIONS(687),
    [anon_sym_DOLLAR] = ACTIONS(687),
    [anon_sym_PERCENT] = ACTIONS(687),
    [anon_sym_AMP] = ACTIONS(687),
    [anon_sym_SQUOTE] = ACTIONS(687),
    [anon_sym_STAR] = ACTIONS(687),
    [anon_sym_PLUS] = ACTIONS(687),
    [anon_sym_COMMA] = ACTIONS(687),
    [anon_sym_DASH] = ACTIONS(687),
    [anon_sym_DOT] = ACTIONS(687),
    [anon_sym_SLASH] = ACTIONS(687),
    [anon_sym_COLON] = ACTIONS(687),
    [anon_sym_SEMI] = ACTIONS(687),
    [anon_sym_EQ] = ACTIONS(687),
    [anon_sym_QMARK] = ACTIONS(687),
    [anon_sym_AT] = ACTIONS(687),
    [anon_sym_BSLASH] = ACTIONS(687),
    [anon_sym_CARET] = ACTIONS(687),
    [anon_sym__] = ACTIONS(687),
    [anon_sym_BQUOTE] = ACTIONS(687),
    [anon_sym_LBRACE] = ACTIONS(687),
    [anon_sym_PIPE] = ACTIONS(687),
    [anon_sym_RBRACE] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(687),
    [anon_sym_LPAREN] = ACTIONS(687),
    [anon_sym_RPAREN] = ACTIONS(687),
    [aux_sym__word_token1] = ACTIONS(687),
    [aux_sym__word_token2] = ACTIONS(687),
    [aux_sym__word_token3] = ACTIONS(687),
    [sym__whitespace] = ACTIONS(759),
    [sym__soft_line_ending] = ACTIONS(693),
    [sym__block_close] = ACTIONS(682),
    [sym__block_quote_start] = ACTIONS(762),
    [sym__indented_chunk_start] = ACTIONS(765),
    [sym_atx_h1_marker] = ACTIONS(682),
    [sym_atx_h2_marker] = ACTIONS(682),
    [sym_atx_h3_marker] = ACTIONS(682),
    [sym_atx_h4_marker] = ACTIONS(682),
    [sym_atx_h5_marker] = ACTIONS(768),
    [sym_atx_h6_marker] = ACTIONS(771),
    [sym__thematic_break] = ACTIONS(774),
    [sym__list_marker_minus] = ACTIONS(711),
    [sym__list_marker_plus] = ACTIONS(714),
    [sym__list_marker_star] = ACTIONS(717),
    [sym__list_marker_parenthesis] = ACTIONS(720),
    [sym__list_marker_dot] = ACTIONS(723),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(711),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(714),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(717),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(720),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(723),
    [sym__fenced_code_block_start_backtick] = ACTIONS(777),
    [sym__fenced_code_block_start_tilde] = ACTIONS(780),
    [sym__blank_line_start] = ACTIONS(783),
    [sym__html_block_1_start] = ACTIONS(786),
    [sym__html_block_2_start] = ACTIONS(789),
    [sym__html_block_3_start] = ACTIONS(792),
    [sym__html_block_4_start] = ACTIONS(795),
    [sym__html_block_5_start] = ACTIONS(798),
    [sym__html_block_6_start] = ACTIONS(801),
    [sym__html_block_7_start] = ACTIONS(804),
    [sym__pipe_table_start] = ACTIONS(807),
  },
  [STATE(58)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(60),
    [sym__section5] = STATE(232),
    [sym__section6] = STATE(232),
    [sym_thematic_break] = STATE(60),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(60),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(60),
    [sym_html_block] = STATE(60),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(60),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(60),
    [sym_block_quote] = STATE(60),
    [sym_list] = STATE(60),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(60),
    [aux_sym__section4_repeat1] = STATE(60),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(680),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(680),
    [sym_atx_h2_marker] = ACTIONS(680),
    [sym_atx_h3_marker] = ACTIONS(680),
    [sym_atx_h4_marker] = ACTIONS(680),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(59)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(56),
    [sym__section5] = STATE(240),
    [sym__section6] = STATE(240),
    [sym_thematic_break] = STATE(56),
    [sym__atx_heading5] = STATE(66),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(56),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(56),
    [sym_html_block] = STATE(56),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(56),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(56),
    [sym_block_quote] = STATE(56),
    [sym_list] = STATE(56),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(56),
    [aux_sym__section4_repeat1] = STATE(56),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(810),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(810),
    [sym_atx_h2_marker] = ACTIONS(810),
    [sym_atx_h3_marker] = ACTIONS(810),
    [sym_atx_h4_marker] = ACTIONS(810),
    [sym_atx_h5_marker] = ACTIONS(25),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(60)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(57),
    [sym__section5] = STATE(232),
    [sym__section6] = STATE(232),
    [sym_thematic_break] = STATE(57),
    [sym__atx_heading5] = STATE(62),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(57),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(57),
    [sym_html_block] = STATE(57),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(57),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(57),
    [sym_block_quote] = STATE(57),
    [sym_list] = STATE(57),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(57),
    [aux_sym__section4_repeat1] = STATE(57),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(810),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(810),
    [sym_atx_h2_marker] = ACTIONS(810),
    [sym_atx_h3_marker] = ACTIONS(810),
    [sym_atx_h4_marker] = ACTIONS(810),
    [sym_atx_h5_marker] = ACTIONS(85),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(61)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(64),
    [sym__section6] = STATE(242),
    [sym_thematic_break] = STATE(64),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(64),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(64),
    [sym_html_block] = STATE(64),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(64),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(64),
    [sym_block_quote] = STATE(64),
    [sym_list] = STATE(64),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(64),
    [aux_sym__section5_repeat1] = STATE(64),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(812),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(812),
    [sym_atx_h2_marker] = ACTIONS(812),
    [sym_atx_h3_marker] = ACTIONS(812),
    [sym_atx_h4_marker] = ACTIONS(812),
    [sym_atx_h5_marker] = ACTIONS(812),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(62)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(63),
    [sym__section6] = STATE(234),
    [sym_thematic_break] = STATE(63),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(63),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(63),
    [sym_html_block] = STATE(63),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(63),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(63),
    [sym_block_quote] = STATE(63),
    [sym_list] = STATE(63),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(63),
    [aux_sym__section5_repeat1] = STATE(63),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(814),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(814),
    [sym_atx_h2_marker] = ACTIONS(814),
    [sym_atx_h3_marker] = ACTIONS(814),
    [sym_atx_h4_marker] = ACTIONS(814),
    [sym_atx_h5_marker] = ACTIONS(814),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(63)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(65),
    [sym__section6] = STATE(234),
    [sym_thematic_break] = STATE(65),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(65),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(65),
    [sym_html_block] = STATE(65),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(65),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(65),
    [sym_block_quote] = STATE(65),
    [sym_list] = STATE(65),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(65),
    [aux_sym__section5_repeat1] = STATE(65),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(812),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(812),
    [sym_atx_h2_marker] = ACTIONS(812),
    [sym_atx_h3_marker] = ACTIONS(812),
    [sym_atx_h4_marker] = ACTIONS(812),
    [sym_atx_h5_marker] = ACTIONS(812),
    [sym_atx_h6_marker] = ACTIONS(87),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(64)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(64),
    [sym__section6] = STATE(242),
    [sym_thematic_break] = STATE(64),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(64),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(64),
    [sym_html_block] = STATE(64),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(64),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(64),
    [sym_block_quote] = STATE(64),
    [sym_list] = STATE(64),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(64),
    [aux_sym__section5_repeat1] = STATE(64),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(816),
    [anon_sym_LBRACK] = ACTIONS(818),
    [anon_sym_RBRACK] = ACTIONS(821),
    [anon_sym_LT] = ACTIONS(821),
    [anon_sym_GT] = ACTIONS(821),
    [anon_sym_BANG] = ACTIONS(821),
    [anon_sym_DQUOTE] = ACTIONS(821),
    [anon_sym_POUND] = ACTIONS(821),
    [anon_sym_DOLLAR] = ACTIONS(821),
    [anon_sym_PERCENT] = ACTIONS(821),
    [anon_sym_AMP] = ACTIONS(821),
    [anon_sym_SQUOTE] = ACTIONS(821),
    [anon_sym_STAR] = ACTIONS(821),
    [anon_sym_PLUS] = ACTIONS(821),
    [anon_sym_COMMA] = ACTIONS(821),
    [anon_sym_DASH] = ACTIONS(821),
    [anon_sym_DOT] = ACTIONS(821),
    [anon_sym_SLASH] = ACTIONS(821),
    [anon_sym_COLON] = ACTIONS(821),
    [anon_sym_SEMI] = ACTIONS(821),
    [anon_sym_EQ] = ACTIONS(821),
    [anon_sym_QMARK] = ACTIONS(821),
    [anon_sym_AT] = ACTIONS(821),
    [anon_sym_BSLASH] = ACTIONS(821),
    [anon_sym_CARET] = ACTIONS(821),
    [anon_sym__] = ACTIONS(821),
    [anon_sym_BQUOTE] = ACTIONS(821),
    [anon_sym_LBRACE] = ACTIONS(821),
    [anon_sym_PIPE] = ACTIONS(821),
    [anon_sym_RBRACE] = ACTIONS(821),
    [anon_sym_TILDE] = ACTIONS(821),
    [anon_sym_LPAREN] = ACTIONS(821),
    [anon_sym_RPAREN] = ACTIONS(821),
    [aux_sym__word_token1] = ACTIONS(821),
    [aux_sym__word_token2] = ACTIONS(821),
    [aux_sym__word_token3] = ACTIONS(821),
    [sym__whitespace] = ACTIONS(824),
    [sym__soft_line_ending] = ACTIONS(827),
    [sym__block_quote_start] = ACTIONS(830),
    [sym__indented_chunk_start] = ACTIONS(833),
    [sym_atx_h1_marker] = ACTIONS(816),
    [sym_atx_h2_marker] = ACTIONS(816),
    [sym_atx_h3_marker] = ACTIONS(816),
    [sym_atx_h4_marker] = ACTIONS(816),
    [sym_atx_h5_marker] = ACTIONS(816),
    [sym_atx_h6_marker] = ACTIONS(836),
    [sym__thematic_break] = ACTIONS(839),
    [sym__list_marker_minus] = ACTIONS(842),
    [sym__list_marker_plus] = ACTIONS(845),
    [sym__list_marker_star] = ACTIONS(848),
    [sym__list_marker_parenthesis] = ACTIONS(851),
    [sym__list_marker_dot] = ACTIONS(854),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(842),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(845),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(848),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(851),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(854),
    [sym__fenced_code_block_start_backtick] = ACTIONS(857),
    [sym__fenced_code_block_start_tilde] = ACTIONS(860),
    [sym__blank_line_start] = ACTIONS(863),
    [sym__html_block_1_start] = ACTIONS(866),
    [sym__html_block_2_start] = ACTIONS(869),
    [sym__html_block_3_start] = ACTIONS(872),
    [sym__html_block_4_start] = ACTIONS(875),
    [sym__html_block_5_start] = ACTIONS(878),
    [sym__html_block_6_start] = ACTIONS(881),
    [sym__html_block_7_start] = ACTIONS(884),
    [sym__pipe_table_start] = ACTIONS(887),
  },
  [STATE(65)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(65),
    [sym__section6] = STATE(234),
    [sym_thematic_break] = STATE(65),
    [sym__atx_heading6] = STATE(69),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(65),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(65),
    [sym_html_block] = STATE(65),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(65),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(65),
    [sym_block_quote] = STATE(65),
    [sym_list] = STATE(65),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(65),
    [aux_sym__section5_repeat1] = STATE(65),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(818),
    [anon_sym_RBRACK] = ACTIONS(821),
    [anon_sym_LT] = ACTIONS(821),
    [anon_sym_GT] = ACTIONS(821),
    [anon_sym_BANG] = ACTIONS(821),
    [anon_sym_DQUOTE] = ACTIONS(821),
    [anon_sym_POUND] = ACTIONS(821),
    [anon_sym_DOLLAR] = ACTIONS(821),
    [anon_sym_PERCENT] = ACTIONS(821),
    [anon_sym_AMP] = ACTIONS(821),
    [anon_sym_SQUOTE] = ACTIONS(821),
    [anon_sym_STAR] = ACTIONS(821),
    [anon_sym_PLUS] = ACTIONS(821),
    [anon_sym_COMMA] = ACTIONS(821),
    [anon_sym_DASH] = ACTIONS(821),
    [anon_sym_DOT] = ACTIONS(821),
    [anon_sym_SLASH] = ACTIONS(821),
    [anon_sym_COLON] = ACTIONS(821),
    [anon_sym_SEMI] = ACTIONS(821),
    [anon_sym_EQ] = ACTIONS(821),
    [anon_sym_QMARK] = ACTIONS(821),
    [anon_sym_AT] = ACTIONS(821),
    [anon_sym_BSLASH] = ACTIONS(821),
    [anon_sym_CARET] = ACTIONS(821),
    [anon_sym__] = ACTIONS(821),
    [anon_sym_BQUOTE] = ACTIONS(821),
    [anon_sym_LBRACE] = ACTIONS(821),
    [anon_sym_PIPE] = ACTIONS(821),
    [anon_sym_RBRACE] = ACTIONS(821),
    [anon_sym_TILDE] = ACTIONS(821),
    [anon_sym_LPAREN] = ACTIONS(821),
    [anon_sym_RPAREN] = ACTIONS(821),
    [aux_sym__word_token1] = ACTIONS(821),
    [aux_sym__word_token2] = ACTIONS(821),
    [aux_sym__word_token3] = ACTIONS(821),
    [sym__whitespace] = ACTIONS(890),
    [sym__soft_line_ending] = ACTIONS(827),
    [sym__block_close] = ACTIONS(816),
    [sym__block_quote_start] = ACTIONS(893),
    [sym__indented_chunk_start] = ACTIONS(896),
    [sym_atx_h1_marker] = ACTIONS(816),
    [sym_atx_h2_marker] = ACTIONS(816),
    [sym_atx_h3_marker] = ACTIONS(816),
    [sym_atx_h4_marker] = ACTIONS(816),
    [sym_atx_h5_marker] = ACTIONS(816),
    [sym_atx_h6_marker] = ACTIONS(899),
    [sym__thematic_break] = ACTIONS(902),
    [sym__list_marker_minus] = ACTIONS(842),
    [sym__list_marker_plus] = ACTIONS(845),
    [sym__list_marker_star] = ACTIONS(848),
    [sym__list_marker_parenthesis] = ACTIONS(851),
    [sym__list_marker_dot] = ACTIONS(854),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(842),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(845),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(848),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(851),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(854),
    [sym__fenced_code_block_start_backtick] = ACTIONS(905),
    [sym__fenced_code_block_start_tilde] = ACTIONS(908),
    [sym__blank_line_start] = ACTIONS(911),
    [sym__html_block_1_start] = ACTIONS(914),
    [sym__html_block_2_start] = ACTIONS(917),
    [sym__html_block_3_start] = ACTIONS(920),
    [sym__html_block_4_start] = ACTIONS(923),
    [sym__html_block_5_start] = ACTIONS(926),
    [sym__html_block_6_start] = ACTIONS(929),
    [sym__html_block_7_start] = ACTIONS(932),
    [sym__pipe_table_start] = ACTIONS(935),
  },
  [STATE(66)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(61),
    [sym__section6] = STATE(242),
    [sym_thematic_break] = STATE(61),
    [sym__atx_heading6] = STATE(72),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(61),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(61),
    [sym_html_block] = STATE(61),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(61),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(61),
    [sym_block_quote] = STATE(61),
    [sym_list] = STATE(61),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(61),
    [aux_sym__section5_repeat1] = STATE(61),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(814),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(814),
    [sym_atx_h2_marker] = ACTIONS(814),
    [sym_atx_h3_marker] = ACTIONS(814),
    [sym_atx_h4_marker] = ACTIONS(814),
    [sym_atx_h5_marker] = ACTIONS(814),
    [sym_atx_h6_marker] = ACTIONS(27),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(67)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_thematic_break] = STATE(339),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(71),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(938),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(938),
    [sym_atx_h2_marker] = ACTIONS(938),
    [sym_atx_h3_marker] = ACTIONS(938),
    [sym_atx_h4_marker] = ACTIONS(938),
    [sym_atx_h5_marker] = ACTIONS(938),
    [sym_atx_h6_marker] = ACTIONS(938),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(68)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(195),
    [sym_thematic_break] = STATE(195),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(195),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(195),
    [sym_html_block] = STATE(195),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(195),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(195),
    [sym_block_quote] = STATE(195),
    [sym_list] = STATE(195),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(195),
    [aux_sym_document_repeat1] = STATE(68),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(940),
    [anon_sym_RBRACK] = ACTIONS(943),
    [anon_sym_LT] = ACTIONS(943),
    [anon_sym_GT] = ACTIONS(943),
    [anon_sym_BANG] = ACTIONS(943),
    [anon_sym_DQUOTE] = ACTIONS(943),
    [anon_sym_POUND] = ACTIONS(943),
    [anon_sym_DOLLAR] = ACTIONS(943),
    [anon_sym_PERCENT] = ACTIONS(943),
    [anon_sym_AMP] = ACTIONS(943),
    [anon_sym_SQUOTE] = ACTIONS(943),
    [anon_sym_STAR] = ACTIONS(943),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_COMMA] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(943),
    [anon_sym_DOT] = ACTIONS(943),
    [anon_sym_SLASH] = ACTIONS(943),
    [anon_sym_COLON] = ACTIONS(943),
    [anon_sym_SEMI] = ACTIONS(943),
    [anon_sym_EQ] = ACTIONS(943),
    [anon_sym_QMARK] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(943),
    [anon_sym_BSLASH] = ACTIONS(943),
    [anon_sym_CARET] = ACTIONS(943),
    [anon_sym__] = ACTIONS(943),
    [anon_sym_BQUOTE] = ACTIONS(943),
    [anon_sym_LBRACE] = ACTIONS(943),
    [anon_sym_PIPE] = ACTIONS(943),
    [anon_sym_RBRACE] = ACTIONS(943),
    [anon_sym_TILDE] = ACTIONS(943),
    [anon_sym_LPAREN] = ACTIONS(943),
    [anon_sym_RPAREN] = ACTIONS(943),
    [aux_sym__word_token1] = ACTIONS(943),
    [aux_sym__word_token2] = ACTIONS(943),
    [aux_sym__word_token3] = ACTIONS(943),
    [sym__whitespace] = ACTIONS(946),
    [sym__soft_line_ending] = ACTIONS(949),
    [sym__block_close] = ACTIONS(952),
    [sym__block_quote_start] = ACTIONS(954),
    [sym__indented_chunk_start] = ACTIONS(957),
    [sym_atx_h1_marker] = ACTIONS(952),
    [sym_atx_h2_marker] = ACTIONS(952),
    [sym_atx_h3_marker] = ACTIONS(952),
    [sym_atx_h4_marker] = ACTIONS(952),
    [sym_atx_h5_marker] = ACTIONS(952),
    [sym_atx_h6_marker] = ACTIONS(952),
    [sym__thematic_break] = ACTIONS(960),
    [sym__list_marker_minus] = ACTIONS(963),
    [sym__list_marker_plus] = ACTIONS(966),
    [sym__list_marker_star] = ACTIONS(969),
    [sym__list_marker_parenthesis] = ACTIONS(972),
    [sym__list_marker_dot] = ACTIONS(975),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(963),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(966),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(969),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(972),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(975),
    [sym__fenced_code_block_start_backtick] = ACTIONS(978),
    [sym__fenced_code_block_start_tilde] = ACTIONS(981),
    [sym__blank_line_start] = ACTIONS(984),
    [sym__html_block_1_start] = ACTIONS(987),
    [sym__html_block_2_start] = ACTIONS(990),
    [sym__html_block_3_start] = ACTIONS(993),
    [sym__html_block_4_start] = ACTIONS(996),
    [sym__html_block_5_start] = ACTIONS(999),
    [sym__html_block_6_start] = ACTIONS(1002),
    [sym__html_block_7_start] = ACTIONS(1005),
    [sym__pipe_table_start] = ACTIONS(1008),
  },
  [STATE(69)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(195),
    [sym_thematic_break] = STATE(195),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(195),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(195),
    [sym_html_block] = STATE(195),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(195),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(195),
    [sym_block_quote] = STATE(195),
    [sym_list] = STATE(195),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(195),
    [aux_sym_document_repeat1] = STATE(70),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(1011),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(1011),
    [sym_atx_h2_marker] = ACTIONS(1011),
    [sym_atx_h3_marker] = ACTIONS(1011),
    [sym_atx_h4_marker] = ACTIONS(1011),
    [sym_atx_h5_marker] = ACTIONS(1011),
    [sym_atx_h6_marker] = ACTIONS(1011),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(70)] = {
    [sym_link_label] = STATE(916),
    [sym__block_not_section] = STATE(195),
    [sym_thematic_break] = STATE(195),
    [sym__setext_heading1] = STATE(203),
    [sym__setext_heading2] = STATE(204),
    [sym_indented_code_block] = STATE(195),
    [sym__indented_chunk] = STATE(78),
    [sym_fenced_code_block] = STATE(195),
    [sym_html_block] = STATE(195),
    [sym__html_block_1] = STATE(206),
    [sym__html_block_2] = STATE(206),
    [sym__html_block_3] = STATE(206),
    [sym__html_block_4] = STATE(206),
    [sym__html_block_5] = STATE(206),
    [sym__html_block_6] = STATE(206),
    [sym__html_block_7] = STATE(206),
    [sym_link_reference_definition] = STATE(195),
    [sym_paragraph] = STATE(118),
    [sym__blank_line] = STATE(195),
    [sym_block_quote] = STATE(195),
    [sym_list] = STATE(195),
    [sym__list_plus] = STATE(208),
    [sym__list_minus] = STATE(208),
    [sym__list_star] = STATE(208),
    [sym__list_dot] = STATE(208),
    [sym__list_parenthesis] = STATE(208),
    [sym_list_marker_plus] = STATE(7),
    [sym_list_marker_minus] = STATE(8),
    [sym_list_marker_star] = STATE(9),
    [sym_list_marker_dot] = STATE(10),
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_plus] = STATE(79),
    [sym__list_item_minus] = STATE(80),
    [sym__list_item_star] = STATE(81),
    [sym__list_item_dot] = STATE(82),
    [sym__list_item_parenthesis] = STATE(83),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(195),
    [aux_sym_document_repeat1] = STATE(68),
    [aux_sym_paragraph_repeat1] = STATE(362),
    [aux_sym__list_plus_repeat1] = STATE(79),
    [aux_sym__list_minus_repeat1] = STATE(80),
    [aux_sym__list_star_repeat1] = STATE(81),
    [aux_sym__list_dot_repeat1] = STATE(82),
    [aux_sym__list_parenthesis_repeat1] = STATE(83),
    [aux_sym__line_repeat1] = STATE(506),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(69),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_close] = ACTIONS(938),
    [sym__block_quote_start] = ACTIONS(73),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(938),
    [sym_atx_h2_marker] = ACTIONS(938),
    [sym_atx_h3_marker] = ACTIONS(938),
    [sym_atx_h4_marker] = ACTIONS(938),
    [sym_atx_h5_marker] = ACTIONS(938),
    [sym_atx_h6_marker] = ACTIONS(938),
    [sym__thematic_break] = ACTIONS(89),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(91),
    [sym__fenced_code_block_start_tilde] = ACTIONS(93),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(97),
    [sym__html_block_2_start] = ACTIONS(99),
    [sym__html_block_3_start] = ACTIONS(101),
    [sym__html_block_4_start] = ACTIONS(103),
    [sym__html_block_5_start] = ACTIONS(105),
    [sym__html_block_6_start] = ACTIONS(107),
    [sym__html_block_7_start] = ACTIONS(109),
    [sym__pipe_table_start] = ACTIONS(111),
  },
  [STATE(71)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_thematic_break] = STATE(339),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(71),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(952),
    [anon_sym_LBRACK] = ACTIONS(940),
    [anon_sym_RBRACK] = ACTIONS(943),
    [anon_sym_LT] = ACTIONS(943),
    [anon_sym_GT] = ACTIONS(943),
    [anon_sym_BANG] = ACTIONS(943),
    [anon_sym_DQUOTE] = ACTIONS(943),
    [anon_sym_POUND] = ACTIONS(943),
    [anon_sym_DOLLAR] = ACTIONS(943),
    [anon_sym_PERCENT] = ACTIONS(943),
    [anon_sym_AMP] = ACTIONS(943),
    [anon_sym_SQUOTE] = ACTIONS(943),
    [anon_sym_STAR] = ACTIONS(943),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_COMMA] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(943),
    [anon_sym_DOT] = ACTIONS(943),
    [anon_sym_SLASH] = ACTIONS(943),
    [anon_sym_COLON] = ACTIONS(943),
    [anon_sym_SEMI] = ACTIONS(943),
    [anon_sym_EQ] = ACTIONS(943),
    [anon_sym_QMARK] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(943),
    [anon_sym_BSLASH] = ACTIONS(943),
    [anon_sym_CARET] = ACTIONS(943),
    [anon_sym__] = ACTIONS(943),
    [anon_sym_BQUOTE] = ACTIONS(943),
    [anon_sym_LBRACE] = ACTIONS(943),
    [anon_sym_PIPE] = ACTIONS(943),
    [anon_sym_RBRACE] = ACTIONS(943),
    [anon_sym_TILDE] = ACTIONS(943),
    [anon_sym_LPAREN] = ACTIONS(943),
    [anon_sym_RPAREN] = ACTIONS(943),
    [aux_sym__word_token1] = ACTIONS(943),
    [aux_sym__word_token2] = ACTIONS(943),
    [aux_sym__word_token3] = ACTIONS(943),
    [sym__whitespace] = ACTIONS(1013),
    [sym__soft_line_ending] = ACTIONS(949),
    [sym__block_quote_start] = ACTIONS(1016),
    [sym__indented_chunk_start] = ACTIONS(1019),
    [sym_atx_h1_marker] = ACTIONS(952),
    [sym_atx_h2_marker] = ACTIONS(952),
    [sym_atx_h3_marker] = ACTIONS(952),
    [sym_atx_h4_marker] = ACTIONS(952),
    [sym_atx_h5_marker] = ACTIONS(952),
    [sym_atx_h6_marker] = ACTIONS(952),
    [sym__thematic_break] = ACTIONS(1022),
    [sym__list_marker_minus] = ACTIONS(963),
    [sym__list_marker_plus] = ACTIONS(966),
    [sym__list_marker_star] = ACTIONS(969),
    [sym__list_marker_parenthesis] = ACTIONS(972),
    [sym__list_marker_dot] = ACTIONS(975),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(963),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(966),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(969),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(972),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(975),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1025),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1028),
    [sym__blank_line_start] = ACTIONS(1031),
    [sym__html_block_1_start] = ACTIONS(1034),
    [sym__html_block_2_start] = ACTIONS(1037),
    [sym__html_block_3_start] = ACTIONS(1040),
    [sym__html_block_4_start] = ACTIONS(1043),
    [sym__html_block_5_start] = ACTIONS(1046),
    [sym__html_block_6_start] = ACTIONS(1049),
    [sym__html_block_7_start] = ACTIONS(1052),
    [sym__pipe_table_start] = ACTIONS(1055),
  },
  [STATE(72)] = {
    [sym_link_label] = STATE(905),
    [sym__block_not_section] = STATE(339),
    [sym_thematic_break] = STATE(339),
    [sym__setext_heading1] = STATE(233),
    [sym__setext_heading2] = STATE(237),
    [sym_indented_code_block] = STATE(339),
    [sym__indented_chunk] = STATE(96),
    [sym_fenced_code_block] = STATE(339),
    [sym_html_block] = STATE(339),
    [sym__html_block_1] = STATE(241),
    [sym__html_block_2] = STATE(241),
    [sym__html_block_3] = STATE(241),
    [sym__html_block_4] = STATE(241),
    [sym__html_block_5] = STATE(241),
    [sym__html_block_6] = STATE(241),
    [sym__html_block_7] = STATE(241),
    [sym_link_reference_definition] = STATE(339),
    [sym_paragraph] = STATE(114),
    [sym__blank_line] = STATE(339),
    [sym_block_quote] = STATE(339),
    [sym_list] = STATE(339),
    [sym__list_plus] = STATE(257),
    [sym__list_minus] = STATE(257),
    [sym__list_star] = STATE(257),
    [sym__list_dot] = STATE(257),
    [sym__list_parenthesis] = STATE(257),
    [sym_list_marker_plus] = STATE(3),
    [sym_list_marker_minus] = STATE(4),
    [sym_list_marker_star] = STATE(2),
    [sym_list_marker_dot] = STATE(6),
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_plus] = STATE(98),
    [sym__list_item_minus] = STATE(73),
    [sym__list_item_star] = STATE(99),
    [sym__list_item_dot] = STATE(74),
    [sym__list_item_parenthesis] = STATE(86),
    [sym__soft_line_break] = STATE(601),
    [sym__line] = STATE(601),
    [sym__word] = STATE(506),
    [sym_pipe_table] = STATE(339),
    [aux_sym_document_repeat1] = STATE(67),
    [aux_sym_paragraph_repeat1] = STATE(375),
    [aux_sym__list_plus_repeat1] = STATE(98),
    [aux_sym__list_minus_repeat1] = STATE(73),
    [aux_sym__list_star_repeat1] = STATE(99),
    [aux_sym__list_dot_repeat1] = STATE(74),
    [aux_sym__list_parenthesis_repeat1] = STATE(86),
    [aux_sym__line_repeat1] = STATE(506),
    [ts_builtin_sym_end] = ACTIONS(1011),
    [anon_sym_LBRACK] = ACTIONS(5),
    [anon_sym_RBRACK] = ACTIONS(7),
    [anon_sym_LT] = ACTIONS(7),
    [anon_sym_GT] = ACTIONS(7),
    [anon_sym_BANG] = ACTIONS(7),
    [anon_sym_DQUOTE] = ACTIONS(7),
    [anon_sym_POUND] = ACTIONS(7),
    [anon_sym_DOLLAR] = ACTIONS(7),
    [anon_sym_PERCENT] = ACTIONS(7),
    [anon_sym_AMP] = ACTIONS(7),
    [anon_sym_SQUOTE] = ACTIONS(7),
    [anon_sym_STAR] = ACTIONS(7),
    [anon_sym_PLUS] = ACTIONS(7),
    [anon_sym_COMMA] = ACTIONS(7),
    [anon_sym_DASH] = ACTIONS(7),
    [anon_sym_DOT] = ACTIONS(7),
    [anon_sym_SLASH] = ACTIONS(7),
    [anon_sym_COLON] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(7),
    [anon_sym_EQ] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(7),
    [anon_sym_AT] = ACTIONS(7),
    [anon_sym_BSLASH] = ACTIONS(7),
    [anon_sym_CARET] = ACTIONS(7),
    [anon_sym__] = ACTIONS(7),
    [anon_sym_BQUOTE] = ACTIONS(7),
    [anon_sym_LBRACE] = ACTIONS(7),
    [anon_sym_PIPE] = ACTIONS(7),
    [anon_sym_RBRACE] = ACTIONS(7),
    [anon_sym_TILDE] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(7),
    [anon_sym_RPAREN] = ACTIONS(7),
    [aux_sym__word_token1] = ACTIONS(7),
    [aux_sym__word_token2] = ACTIONS(7),
    [aux_sym__word_token3] = ACTIONS(7),
    [sym__whitespace] = ACTIONS(9),
    [sym__soft_line_ending] = ACTIONS(11),
    [sym__block_quote_start] = ACTIONS(13),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(1011),
    [sym_atx_h2_marker] = ACTIONS(1011),
    [sym_atx_h3_marker] = ACTIONS(1011),
    [sym_atx_h4_marker] = ACTIONS(1011),
    [sym_atx_h5_marker] = ACTIONS(1011),
    [sym_atx_h6_marker] = ACTIONS(1011),
    [sym__thematic_break] = ACTIONS(29),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(41),
    [sym__fenced_code_block_start_tilde] = ACTIONS(43),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(47),
    [sym__html_block_2_start] = ACTIONS(49),
    [sym__html_block_3_start] = ACTIONS(51),
    [sym__html_block_4_start] = ACTIONS(53),
    [sym__html_block_5_start] = ACTIONS(55),
    [sym__html_block_6_start] = ACTIONS(57),
    [sym__html_block_7_start] = ACTIONS(59),
    [sym__pipe_table_start] = ACTIONS(63),
  },
  [STATE(73)] = {
    [sym_list_marker_minus] = STATE(4),
    [sym__list_item_minus] = STATE(76),
    [aux_sym__list_minus_repeat1] = STATE(76),
    [ts_builtin_sym_end] = ACTIONS(1058),
    [anon_sym_LBRACK] = ACTIONS(1060),
    [anon_sym_RBRACK] = ACTIONS(1058),
    [anon_sym_LT] = ACTIONS(1058),
    [anon_sym_GT] = ACTIONS(1058),
    [anon_sym_BANG] = ACTIONS(1058),
    [anon_sym_DQUOTE] = ACTIONS(1058),
    [anon_sym_POUND] = ACTIONS(1058),
    [anon_sym_DOLLAR] = ACTIONS(1058),
    [anon_sym_PERCENT] = ACTIONS(1058),
    [anon_sym_AMP] = ACTIONS(1058),
    [anon_sym_SQUOTE] = ACTIONS(1058),
    [anon_sym_STAR] = ACTIONS(1058),
    [anon_sym_PLUS] = ACTIONS(1058),
    [anon_sym_COMMA] = ACTIONS(1058),
    [anon_sym_DASH] = ACTIONS(1058),
    [anon_sym_DOT] = ACTIONS(1058),
    [anon_sym_SLASH] = ACTIONS(1058),
    [anon_sym_COLON] = ACTIONS(1058),
    [anon_sym_SEMI] = ACTIONS(1058),
    [anon_sym_EQ] = ACTIONS(1058),
    [anon_sym_QMARK] = ACTIONS(1058),
    [anon_sym_AT] = ACTIONS(1058),
    [anon_sym_BSLASH] = ACTIONS(1058),
    [anon_sym_CARET] = ACTIONS(1058),
    [anon_sym__] = ACTIONS(1058),
    [anon_sym_BQUOTE] = ACTIONS(1058),
    [anon_sym_LBRACE] = ACTIONS(1058),
    [anon_sym_PIPE] = ACTIONS(1058),
    [anon_sym_RBRACE] = ACTIONS(1058),
    [anon_sym_TILDE] = ACTIONS(1058),
    [anon_sym_LPAREN] = ACTIONS(1058),
    [anon_sym_RPAREN] = ACTIONS(1058),
    [aux_sym__word_token1] = ACTIONS(1058),
    [aux_sym__word_token2] = ACTIONS(1058),
    [aux_sym__word_token3] = ACTIONS(1058),
    [sym__whitespace] = ACTIONS(1058),
    [sym__soft_line_ending] = ACTIONS(1058),
    [sym__block_quote_start] = ACTIONS(1058),
    [sym__indented_chunk_start] = ACTIONS(1058),
    [sym_atx_h1_marker] = ACTIONS(1058),
    [sym_atx_h2_marker] = ACTIONS(1058),
    [sym_atx_h3_marker] = ACTIONS(1058),
    [sym_atx_h4_marker] = ACTIONS(1058),
    [sym_atx_h5_marker] = ACTIONS(1058),
    [sym_atx_h6_marker] = ACTIONS(1058),
    [sym__thematic_break] = ACTIONS(1058),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(1058),
    [sym__list_marker_star] = ACTIONS(1058),
    [sym__list_marker_parenthesis] = ACTIONS(1058),
    [sym__list_marker_dot] = ACTIONS(1058),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1058),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1058),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1058),
    [sym__blank_line_start] = ACTIONS(1058),
    [sym__html_block_1_start] = ACTIONS(1058),
    [sym__html_block_2_start] = ACTIONS(1058),
    [sym__html_block_3_start] = ACTIONS(1058),
    [sym__html_block_4_start] = ACTIONS(1058),
    [sym__html_block_5_start] = ACTIONS(1058),
    [sym__html_block_6_start] = ACTIONS(1058),
    [sym__html_block_7_start] = ACTIONS(1058),
    [sym__pipe_table_start] = ACTIONS(1058),
  },
  [STATE(74)] = {
    [sym_list_marker_dot] = STATE(6),
    [sym__list_item_dot] = STATE(84),
    [aux_sym__list_dot_repeat1] = STATE(84),
    [ts_builtin_sym_end] = ACTIONS(1062),
    [anon_sym_LBRACK] = ACTIONS(1064),
    [anon_sym_RBRACK] = ACTIONS(1062),
    [anon_sym_LT] = ACTIONS(1062),
    [anon_sym_GT] = ACTIONS(1062),
    [anon_sym_BANG] = ACTIONS(1062),
    [anon_sym_DQUOTE] = ACTIONS(1062),
    [anon_sym_POUND] = ACTIONS(1062),
    [anon_sym_DOLLAR] = ACTIONS(1062),
    [anon_sym_PERCENT] = ACTIONS(1062),
    [anon_sym_AMP] = ACTIONS(1062),
    [anon_sym_SQUOTE] = ACTIONS(1062),
    [anon_sym_STAR] = ACTIONS(1062),
    [anon_sym_PLUS] = ACTIONS(1062),
    [anon_sym_COMMA] = ACTIONS(1062),
    [anon_sym_DASH] = ACTIONS(1062),
    [anon_sym_DOT] = ACTIONS(1062),
    [anon_sym_SLASH] = ACTIONS(1062),
    [anon_sym_COLON] = ACTIONS(1062),
    [anon_sym_SEMI] = ACTIONS(1062),
    [anon_sym_EQ] = ACTIONS(1062),
    [anon_sym_QMARK] = ACTIONS(1062),
    [anon_sym_AT] = ACTIONS(1062),
    [anon_sym_BSLASH] = ACTIONS(1062),
    [anon_sym_CARET] = ACTIONS(1062),
    [anon_sym__] = ACTIONS(1062),
    [anon_sym_BQUOTE] = ACTIONS(1062),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_PIPE] = ACTIONS(1062),
    [anon_sym_RBRACE] = ACTIONS(1062),
    [anon_sym_TILDE] = ACTIONS(1062),
    [anon_sym_LPAREN] = ACTIONS(1062),
    [anon_sym_RPAREN] = ACTIONS(1062),
    [aux_sym__word_token1] = ACTIONS(1062),
    [aux_sym__word_token2] = ACTIONS(1062),
    [aux_sym__word_token3] = ACTIONS(1062),
    [sym__whitespace] = ACTIONS(1062),
    [sym__soft_line_ending] = ACTIONS(1062),
    [sym__block_quote_start] = ACTIONS(1062),
    [sym__indented_chunk_start] = ACTIONS(1062),
    [sym_atx_h1_marker] = ACTIONS(1062),
    [sym_atx_h2_marker] = ACTIONS(1062),
    [sym_atx_h3_marker] = ACTIONS(1062),
    [sym_atx_h4_marker] = ACTIONS(1062),
    [sym_atx_h5_marker] = ACTIONS(1062),
    [sym_atx_h6_marker] = ACTIONS(1062),
    [sym__thematic_break] = ACTIONS(1062),
    [sym__list_marker_minus] = ACTIONS(1062),
    [sym__list_marker_plus] = ACTIONS(1062),
    [sym__list_marker_star] = ACTIONS(1062),
    [sym__list_marker_parenthesis] = ACTIONS(1062),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1062),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1062),
    [sym__blank_line_start] = ACTIONS(1062),
    [sym__html_block_1_start] = ACTIONS(1062),
    [sym__html_block_2_start] = ACTIONS(1062),
    [sym__html_block_3_start] = ACTIONS(1062),
    [sym__html_block_4_start] = ACTIONS(1062),
    [sym__html_block_5_start] = ACTIONS(1062),
    [sym__html_block_6_start] = ACTIONS(1062),
    [sym__html_block_7_start] = ACTIONS(1062),
    [sym__pipe_table_start] = ACTIONS(1062),
  },
  [STATE(75)] = {
    [sym_list_marker_plus] = STATE(3),
    [sym__list_item_plus] = STATE(75),
    [aux_sym__list_plus_repeat1] = STATE(75),
    [ts_builtin_sym_end] = ACTIONS(1066),
    [anon_sym_LBRACK] = ACTIONS(1068),
    [anon_sym_RBRACK] = ACTIONS(1066),
    [anon_sym_LT] = ACTIONS(1066),
    [anon_sym_GT] = ACTIONS(1066),
    [anon_sym_BANG] = ACTIONS(1066),
    [anon_sym_DQUOTE] = ACTIONS(1066),
    [anon_sym_POUND] = ACTIONS(1066),
    [anon_sym_DOLLAR] = ACTIONS(1066),
    [anon_sym_PERCENT] = ACTIONS(1066),
    [anon_sym_AMP] = ACTIONS(1066),
    [anon_sym_SQUOTE] = ACTIONS(1066),
    [anon_sym_STAR] = ACTIONS(1066),
    [anon_sym_PLUS] = ACTIONS(1066),
    [anon_sym_COMMA] = ACTIONS(1066),
    [anon_sym_DASH] = ACTIONS(1066),
    [anon_sym_DOT] = ACTIONS(1066),
    [anon_sym_SLASH] = ACTIONS(1066),
    [anon_sym_COLON] = ACTIONS(1066),
    [anon_sym_SEMI] = ACTIONS(1066),
    [anon_sym_EQ] = ACTIONS(1066),
    [anon_sym_QMARK] = ACTIONS(1066),
    [anon_sym_AT] = ACTIONS(1066),
    [anon_sym_BSLASH] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(1066),
    [anon_sym__] = ACTIONS(1066),
    [anon_sym_BQUOTE] = ACTIONS(1066),
    [anon_sym_LBRACE] = ACTIONS(1066),
    [anon_sym_PIPE] = ACTIONS(1066),
    [anon_sym_RBRACE] = ACTIONS(1066),
    [anon_sym_TILDE] = ACTIONS(1066),
    [anon_sym_LPAREN] = ACTIONS(1066),
    [anon_sym_RPAREN] = ACTIONS(1066),
    [aux_sym__word_token1] = ACTIONS(1066),
    [aux_sym__word_token2] = ACTIONS(1066),
    [aux_sym__word_token3] = ACTIONS(1066),
    [sym__whitespace] = ACTIONS(1066),
    [sym__soft_line_ending] = ACTIONS(1066),
    [sym__block_quote_start] = ACTIONS(1066),
    [sym__indented_chunk_start] = ACTIONS(1066),
    [sym_atx_h1_marker] = ACTIONS(1066),
    [sym_atx_h2_marker] = ACTIONS(1066),
    [sym_atx_h3_marker] = ACTIONS(1066),
    [sym_atx_h4_marker] = ACTIONS(1066),
    [sym_atx_h5_marker] = ACTIONS(1066),
    [sym_atx_h6_marker] = ACTIONS(1066),
    [sym__thematic_break] = ACTIONS(1066),
    [sym__list_marker_minus] = ACTIONS(1066),
    [sym__list_marker_plus] = ACTIONS(1070),
    [sym__list_marker_star] = ACTIONS(1066),
    [sym__list_marker_parenthesis] = ACTIONS(1066),
    [sym__list_marker_dot] = ACTIONS(1066),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1070),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1066),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1066),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1066),
    [sym__blank_line_start] = ACTIONS(1066),
    [sym__html_block_1_start] = ACTIONS(1066),
    [sym__html_block_2_start] = ACTIONS(1066),
    [sym__html_block_3_start] = ACTIONS(1066),
    [sym__html_block_4_start] = ACTIONS(1066),
    [sym__html_block_5_start] = ACTIONS(1066),
    [sym__html_block_6_start] = ACTIONS(1066),
    [sym__html_block_7_start] = ACTIONS(1066),
    [sym__pipe_table_start] = ACTIONS(1066),
  },
  [STATE(76)] = {
    [sym_list_marker_minus] = STATE(4),
    [sym__list_item_minus] = STATE(76),
    [aux_sym__list_minus_repeat1] = STATE(76),
    [ts_builtin_sym_end] = ACTIONS(1073),
    [anon_sym_LBRACK] = ACTIONS(1075),
    [anon_sym_RBRACK] = ACTIONS(1073),
    [anon_sym_LT] = ACTIONS(1073),
    [anon_sym_GT] = ACTIONS(1073),
    [anon_sym_BANG] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1073),
    [anon_sym_POUND] = ACTIONS(1073),
    [anon_sym_DOLLAR] = ACTIONS(1073),
    [anon_sym_PERCENT] = ACTIONS(1073),
    [anon_sym_AMP] = ACTIONS(1073),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_STAR] = ACTIONS(1073),
    [anon_sym_PLUS] = ACTIONS(1073),
    [anon_sym_COMMA] = ACTIONS(1073),
    [anon_sym_DASH] = ACTIONS(1073),
    [anon_sym_DOT] = ACTIONS(1073),
    [anon_sym_SLASH] = ACTIONS(1073),
    [anon_sym_COLON] = ACTIONS(1073),
    [anon_sym_SEMI] = ACTIONS(1073),
    [anon_sym_EQ] = ACTIONS(1073),
    [anon_sym_QMARK] = ACTIONS(1073),
    [anon_sym_AT] = ACTIONS(1073),
    [anon_sym_BSLASH] = ACTIONS(1073),
    [anon_sym_CARET] = ACTIONS(1073),
    [anon_sym__] = ACTIONS(1073),
    [anon_sym_BQUOTE] = ACTIONS(1073),
    [anon_sym_LBRACE] = ACTIONS(1073),
    [anon_sym_PIPE] = ACTIONS(1073),
    [anon_sym_RBRACE] = ACTIONS(1073),
    [anon_sym_TILDE] = ACTIONS(1073),
    [anon_sym_LPAREN] = ACTIONS(1073),
    [anon_sym_RPAREN] = ACTIONS(1073),
    [aux_sym__word_token1] = ACTIONS(1073),
    [aux_sym__word_token2] = ACTIONS(1073),
    [aux_sym__word_token3] = ACTIONS(1073),
    [sym__whitespace] = ACTIONS(1073),
    [sym__soft_line_ending] = ACTIONS(1073),
    [sym__block_quote_start] = ACTIONS(1073),
    [sym__indented_chunk_start] = ACTIONS(1073),
    [sym_atx_h1_marker] = ACTIONS(1073),
    [sym_atx_h2_marker] = ACTIONS(1073),
    [sym_atx_h3_marker] = ACTIONS(1073),
    [sym_atx_h4_marker] = ACTIONS(1073),
    [sym_atx_h5_marker] = ACTIONS(1073),
    [sym_atx_h6_marker] = ACTIONS(1073),
    [sym__thematic_break] = ACTIONS(1073),
    [sym__list_marker_minus] = ACTIONS(1077),
    [sym__list_marker_plus] = ACTIONS(1073),
    [sym__list_marker_star] = ACTIONS(1073),
    [sym__list_marker_parenthesis] = ACTIONS(1073),
    [sym__list_marker_dot] = ACTIONS(1073),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1077),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1073),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1073),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1073),
    [sym__blank_line_start] = ACTIONS(1073),
    [sym__html_block_1_start] = ACTIONS(1073),
    [sym__html_block_2_start] = ACTIONS(1073),
    [sym__html_block_3_start] = ACTIONS(1073),
    [sym__html_block_4_start] = ACTIONS(1073),
    [sym__html_block_5_start] = ACTIONS(1073),
    [sym__html_block_6_start] = ACTIONS(1073),
    [sym__html_block_7_start] = ACTIONS(1073),
    [sym__pipe_table_start] = ACTIONS(1073),
  },
  [STATE(77)] = {
    [sym_list_marker_star] = STATE(2),
    [sym__list_item_star] = STATE(77),
    [aux_sym__list_star_repeat1] = STATE(77),
    [ts_builtin_sym_end] = ACTIONS(1080),
    [anon_sym_LBRACK] = ACTIONS(1082),
    [anon_sym_RBRACK] = ACTIONS(1080),
    [anon_sym_LT] = ACTIONS(1080),
    [anon_sym_GT] = ACTIONS(1080),
    [anon_sym_BANG] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1080),
    [anon_sym_POUND] = ACTIONS(1080),
    [anon_sym_DOLLAR] = ACTIONS(1080),
    [anon_sym_PERCENT] = ACTIONS(1080),
    [anon_sym_AMP] = ACTIONS(1080),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_STAR] = ACTIONS(1080),
    [anon_sym_PLUS] = ACTIONS(1080),
    [anon_sym_COMMA] = ACTIONS(1080),
    [anon_sym_DASH] = ACTIONS(1080),
    [anon_sym_DOT] = ACTIONS(1080),
    [anon_sym_SLASH] = ACTIONS(1080),
    [anon_sym_COLON] = ACTIONS(1080),
    [anon_sym_SEMI] = ACTIONS(1080),
    [anon_sym_EQ] = ACTIONS(1080),
    [anon_sym_QMARK] = ACTIONS(1080),
    [anon_sym_AT] = ACTIONS(1080),
    [anon_sym_BSLASH] = ACTIONS(1080),
    [anon_sym_CARET] = ACTIONS(1080),
    [anon_sym__] = ACTIONS(1080),
    [anon_sym_BQUOTE] = ACTIONS(1080),
    [anon_sym_LBRACE] = ACTIONS(1080),
    [anon_sym_PIPE] = ACTIONS(1080),
    [anon_sym_RBRACE] = ACTIONS(1080),
    [anon_sym_TILDE] = ACTIONS(1080),
    [anon_sym_LPAREN] = ACTIONS(1080),
    [anon_sym_RPAREN] = ACTIONS(1080),
    [aux_sym__word_token1] = ACTIONS(1080),
    [aux_sym__word_token2] = ACTIONS(1080),
    [aux_sym__word_token3] = ACTIONS(1080),
    [sym__whitespace] = ACTIONS(1080),
    [sym__soft_line_ending] = ACTIONS(1080),
    [sym__block_quote_start] = ACTIONS(1080),
    [sym__indented_chunk_start] = ACTIONS(1080),
    [sym_atx_h1_marker] = ACTIONS(1080),
    [sym_atx_h2_marker] = ACTIONS(1080),
    [sym_atx_h3_marker] = ACTIONS(1080),
    [sym_atx_h4_marker] = ACTIONS(1080),
    [sym_atx_h5_marker] = ACTIONS(1080),
    [sym_atx_h6_marker] = ACTIONS(1080),
    [sym__thematic_break] = ACTIONS(1080),
    [sym__list_marker_minus] = ACTIONS(1080),
    [sym__list_marker_plus] = ACTIONS(1080),
    [sym__list_marker_star] = ACTIONS(1084),
    [sym__list_marker_parenthesis] = ACTIONS(1080),
    [sym__list_marker_dot] = ACTIONS(1080),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1084),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1080),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1080),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1080),
    [sym__blank_line_start] = ACTIONS(1080),
    [sym__html_block_1_start] = ACTIONS(1080),
    [sym__html_block_2_start] = ACTIONS(1080),
    [sym__html_block_3_start] = ACTIONS(1080),
    [sym__html_block_4_start] = ACTIONS(1080),
    [sym__html_block_5_start] = ACTIONS(1080),
    [sym__html_block_6_start] = ACTIONS(1080),
    [sym__html_block_7_start] = ACTIONS(1080),
    [sym__pipe_table_start] = ACTIONS(1080),
  },
  [STATE(78)] = {
    [sym__indented_chunk] = STATE(100),
    [sym__blank_line] = STATE(100),
    [aux_sym_indented_code_block_repeat1] = STATE(100),
    [anon_sym_LBRACK] = ACTIONS(1087),
    [anon_sym_RBRACK] = ACTIONS(1089),
    [anon_sym_LT] = ACTIONS(1089),
    [anon_sym_GT] = ACTIONS(1089),
    [anon_sym_BANG] = ACTIONS(1089),
    [anon_sym_DQUOTE] = ACTIONS(1089),
    [anon_sym_POUND] = ACTIONS(1089),
    [anon_sym_DOLLAR] = ACTIONS(1089),
    [anon_sym_PERCENT] = ACTIONS(1089),
    [anon_sym_AMP] = ACTIONS(1089),
    [anon_sym_SQUOTE] = ACTIONS(1089),
    [anon_sym_STAR] = ACTIONS(1089),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_COMMA] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_DOT] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1089),
    [anon_sym_COLON] = ACTIONS(1089),
    [anon_sym_SEMI] = ACTIONS(1089),
    [anon_sym_EQ] = ACTIONS(1089),
    [anon_sym_QMARK] = ACTIONS(1089),
    [anon_sym_AT] = ACTIONS(1089),
    [anon_sym_BSLASH] = ACTIONS(1089),
    [anon_sym_CARET] = ACTIONS(1089),
    [anon_sym__] = ACTIONS(1089),
    [anon_sym_BQUOTE] = ACTIONS(1089),
    [anon_sym_LBRACE] = ACTIONS(1089),
    [anon_sym_PIPE] = ACTIONS(1089),
    [anon_sym_RBRACE] = ACTIONS(1089),
    [anon_sym_TILDE] = ACTIONS(1089),
    [anon_sym_LPAREN] = ACTIONS(1089),
    [anon_sym_RPAREN] = ACTIONS(1089),
    [aux_sym__word_token1] = ACTIONS(1089),
    [aux_sym__word_token2] = ACTIONS(1089),
    [aux_sym__word_token3] = ACTIONS(1089),
    [sym__whitespace] = ACTIONS(1089),
    [sym__soft_line_ending] = ACTIONS(1089),
    [sym__block_close] = ACTIONS(1089),
    [sym__block_quote_start] = ACTIONS(1089),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(1089),
    [sym_atx_h2_marker] = ACTIONS(1089),
    [sym_atx_h3_marker] = ACTIONS(1089),
    [sym_atx_h4_marker] = ACTIONS(1089),
    [sym_atx_h5_marker] = ACTIONS(1089),
    [sym_atx_h6_marker] = ACTIONS(1089),
    [sym__thematic_break] = ACTIONS(1089),
    [sym__list_marker_minus] = ACTIONS(1089),
    [sym__list_marker_plus] = ACTIONS(1089),
    [sym__list_marker_star] = ACTIONS(1089),
    [sym__list_marker_parenthesis] = ACTIONS(1089),
    [sym__list_marker_dot] = ACTIONS(1089),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1089),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1089),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1089),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(1089),
    [sym__html_block_2_start] = ACTIONS(1089),
    [sym__html_block_3_start] = ACTIONS(1089),
    [sym__html_block_4_start] = ACTIONS(1089),
    [sym__html_block_5_start] = ACTIONS(1089),
    [sym__html_block_6_start] = ACTIONS(1089),
    [sym__html_block_7_start] = ACTIONS(1089),
    [sym__pipe_table_start] = ACTIONS(1089),
  },
  [STATE(79)] = {
    [sym_list_marker_plus] = STATE(7),
    [sym__list_item_plus] = STATE(87),
    [aux_sym__list_plus_repeat1] = STATE(87),
    [anon_sym_LBRACK] = ACTIONS(1091),
    [anon_sym_RBRACK] = ACTIONS(1093),
    [anon_sym_LT] = ACTIONS(1093),
    [anon_sym_GT] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1093),
    [anon_sym_DQUOTE] = ACTIONS(1093),
    [anon_sym_POUND] = ACTIONS(1093),
    [anon_sym_DOLLAR] = ACTIONS(1093),
    [anon_sym_PERCENT] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1093),
    [anon_sym_SQUOTE] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(1093),
    [anon_sym_PLUS] = ACTIONS(1093),
    [anon_sym_COMMA] = ACTIONS(1093),
    [anon_sym_DASH] = ACTIONS(1093),
    [anon_sym_DOT] = ACTIONS(1093),
    [anon_sym_SLASH] = ACTIONS(1093),
    [anon_sym_COLON] = ACTIONS(1093),
    [anon_sym_SEMI] = ACTIONS(1093),
    [anon_sym_EQ] = ACTIONS(1093),
    [anon_sym_QMARK] = ACTIONS(1093),
    [anon_sym_AT] = ACTIONS(1093),
    [anon_sym_BSLASH] = ACTIONS(1093),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym__] = ACTIONS(1093),
    [anon_sym_BQUOTE] = ACTIONS(1093),
    [anon_sym_LBRACE] = ACTIONS(1093),
    [anon_sym_PIPE] = ACTIONS(1093),
    [anon_sym_RBRACE] = ACTIONS(1093),
    [anon_sym_TILDE] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1093),
    [aux_sym__word_token1] = ACTIONS(1093),
    [aux_sym__word_token2] = ACTIONS(1093),
    [aux_sym__word_token3] = ACTIONS(1093),
    [sym__whitespace] = ACTIONS(1093),
    [sym__soft_line_ending] = ACTIONS(1093),
    [sym__block_close] = ACTIONS(1093),
    [sym__block_quote_start] = ACTIONS(1093),
    [sym__indented_chunk_start] = ACTIONS(1093),
    [sym_atx_h1_marker] = ACTIONS(1093),
    [sym_atx_h2_marker] = ACTIONS(1093),
    [sym_atx_h3_marker] = ACTIONS(1093),
    [sym_atx_h4_marker] = ACTIONS(1093),
    [sym_atx_h5_marker] = ACTIONS(1093),
    [sym_atx_h6_marker] = ACTIONS(1093),
    [sym__thematic_break] = ACTIONS(1093),
    [sym__list_marker_minus] = ACTIONS(1093),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(1093),
    [sym__list_marker_parenthesis] = ACTIONS(1093),
    [sym__list_marker_dot] = ACTIONS(1093),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1093),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1093),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1093),
    [sym__blank_line_start] = ACTIONS(1093),
    [sym__html_block_1_start] = ACTIONS(1093),
    [sym__html_block_2_start] = ACTIONS(1093),
    [sym__html_block_3_start] = ACTIONS(1093),
    [sym__html_block_4_start] = ACTIONS(1093),
    [sym__html_block_5_start] = ACTIONS(1093),
    [sym__html_block_6_start] = ACTIONS(1093),
    [sym__html_block_7_start] = ACTIONS(1093),
    [sym__pipe_table_start] = ACTIONS(1093),
  },
  [STATE(80)] = {
    [sym_list_marker_minus] = STATE(8),
    [sym__list_item_minus] = STATE(88),
    [aux_sym__list_minus_repeat1] = STATE(88),
    [anon_sym_LBRACK] = ACTIONS(1060),
    [anon_sym_RBRACK] = ACTIONS(1058),
    [anon_sym_LT] = ACTIONS(1058),
    [anon_sym_GT] = ACTIONS(1058),
    [anon_sym_BANG] = ACTIONS(1058),
    [anon_sym_DQUOTE] = ACTIONS(1058),
    [anon_sym_POUND] = ACTIONS(1058),
    [anon_sym_DOLLAR] = ACTIONS(1058),
    [anon_sym_PERCENT] = ACTIONS(1058),
    [anon_sym_AMP] = ACTIONS(1058),
    [anon_sym_SQUOTE] = ACTIONS(1058),
    [anon_sym_STAR] = ACTIONS(1058),
    [anon_sym_PLUS] = ACTIONS(1058),
    [anon_sym_COMMA] = ACTIONS(1058),
    [anon_sym_DASH] = ACTIONS(1058),
    [anon_sym_DOT] = ACTIONS(1058),
    [anon_sym_SLASH] = ACTIONS(1058),
    [anon_sym_COLON] = ACTIONS(1058),
    [anon_sym_SEMI] = ACTIONS(1058),
    [anon_sym_EQ] = ACTIONS(1058),
    [anon_sym_QMARK] = ACTIONS(1058),
    [anon_sym_AT] = ACTIONS(1058),
    [anon_sym_BSLASH] = ACTIONS(1058),
    [anon_sym_CARET] = ACTIONS(1058),
    [anon_sym__] = ACTIONS(1058),
    [anon_sym_BQUOTE] = ACTIONS(1058),
    [anon_sym_LBRACE] = ACTIONS(1058),
    [anon_sym_PIPE] = ACTIONS(1058),
    [anon_sym_RBRACE] = ACTIONS(1058),
    [anon_sym_TILDE] = ACTIONS(1058),
    [anon_sym_LPAREN] = ACTIONS(1058),
    [anon_sym_RPAREN] = ACTIONS(1058),
    [aux_sym__word_token1] = ACTIONS(1058),
    [aux_sym__word_token2] = ACTIONS(1058),
    [aux_sym__word_token3] = ACTIONS(1058),
    [sym__whitespace] = ACTIONS(1058),
    [sym__soft_line_ending] = ACTIONS(1058),
    [sym__block_close] = ACTIONS(1058),
    [sym__block_quote_start] = ACTIONS(1058),
    [sym__indented_chunk_start] = ACTIONS(1058),
    [sym_atx_h1_marker] = ACTIONS(1058),
    [sym_atx_h2_marker] = ACTIONS(1058),
    [sym_atx_h3_marker] = ACTIONS(1058),
    [sym_atx_h4_marker] = ACTIONS(1058),
    [sym_atx_h5_marker] = ACTIONS(1058),
    [sym_atx_h6_marker] = ACTIONS(1058),
    [sym__thematic_break] = ACTIONS(1058),
    [sym__list_marker_minus] = ACTIONS(31),
    [sym__list_marker_plus] = ACTIONS(1058),
    [sym__list_marker_star] = ACTIONS(1058),
    [sym__list_marker_parenthesis] = ACTIONS(1058),
    [sym__list_marker_dot] = ACTIONS(1058),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(31),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1058),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1058),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1058),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1058),
    [sym__blank_line_start] = ACTIONS(1058),
    [sym__html_block_1_start] = ACTIONS(1058),
    [sym__html_block_2_start] = ACTIONS(1058),
    [sym__html_block_3_start] = ACTIONS(1058),
    [sym__html_block_4_start] = ACTIONS(1058),
    [sym__html_block_5_start] = ACTIONS(1058),
    [sym__html_block_6_start] = ACTIONS(1058),
    [sym__html_block_7_start] = ACTIONS(1058),
    [sym__pipe_table_start] = ACTIONS(1058),
  },
  [STATE(81)] = {
    [sym_list_marker_star] = STATE(9),
    [sym__list_item_star] = STATE(89),
    [aux_sym__list_star_repeat1] = STATE(89),
    [anon_sym_LBRACK] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1097),
    [anon_sym_LT] = ACTIONS(1097),
    [anon_sym_GT] = ACTIONS(1097),
    [anon_sym_BANG] = ACTIONS(1097),
    [anon_sym_DQUOTE] = ACTIONS(1097),
    [anon_sym_POUND] = ACTIONS(1097),
    [anon_sym_DOLLAR] = ACTIONS(1097),
    [anon_sym_PERCENT] = ACTIONS(1097),
    [anon_sym_AMP] = ACTIONS(1097),
    [anon_sym_SQUOTE] = ACTIONS(1097),
    [anon_sym_STAR] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1097),
    [anon_sym_COMMA] = ACTIONS(1097),
    [anon_sym_DASH] = ACTIONS(1097),
    [anon_sym_DOT] = ACTIONS(1097),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_COLON] = ACTIONS(1097),
    [anon_sym_SEMI] = ACTIONS(1097),
    [anon_sym_EQ] = ACTIONS(1097),
    [anon_sym_QMARK] = ACTIONS(1097),
    [anon_sym_AT] = ACTIONS(1097),
    [anon_sym_BSLASH] = ACTIONS(1097),
    [anon_sym_CARET] = ACTIONS(1097),
    [anon_sym__] = ACTIONS(1097),
    [anon_sym_BQUOTE] = ACTIONS(1097),
    [anon_sym_LBRACE] = ACTIONS(1097),
    [anon_sym_PIPE] = ACTIONS(1097),
    [anon_sym_RBRACE] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1097),
    [anon_sym_LPAREN] = ACTIONS(1097),
    [anon_sym_RPAREN] = ACTIONS(1097),
    [aux_sym__word_token1] = ACTIONS(1097),
    [aux_sym__word_token2] = ACTIONS(1097),
    [aux_sym__word_token3] = ACTIONS(1097),
    [sym__whitespace] = ACTIONS(1097),
    [sym__soft_line_ending] = ACTIONS(1097),
    [sym__block_close] = ACTIONS(1097),
    [sym__block_quote_start] = ACTIONS(1097),
    [sym__indented_chunk_start] = ACTIONS(1097),
    [sym_atx_h1_marker] = ACTIONS(1097),
    [sym_atx_h2_marker] = ACTIONS(1097),
    [sym_atx_h3_marker] = ACTIONS(1097),
    [sym_atx_h4_marker] = ACTIONS(1097),
    [sym_atx_h5_marker] = ACTIONS(1097),
    [sym_atx_h6_marker] = ACTIONS(1097),
    [sym__thematic_break] = ACTIONS(1097),
    [sym__list_marker_minus] = ACTIONS(1097),
    [sym__list_marker_plus] = ACTIONS(1097),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(1097),
    [sym__list_marker_dot] = ACTIONS(1097),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1097),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1097),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1097),
    [sym__blank_line_start] = ACTIONS(1097),
    [sym__html_block_1_start] = ACTIONS(1097),
    [sym__html_block_2_start] = ACTIONS(1097),
    [sym__html_block_3_start] = ACTIONS(1097),
    [sym__html_block_4_start] = ACTIONS(1097),
    [sym__html_block_5_start] = ACTIONS(1097),
    [sym__html_block_6_start] = ACTIONS(1097),
    [sym__html_block_7_start] = ACTIONS(1097),
    [sym__pipe_table_start] = ACTIONS(1097),
  },
  [STATE(82)] = {
    [sym_list_marker_dot] = STATE(10),
    [sym__list_item_dot] = STATE(90),
    [aux_sym__list_dot_repeat1] = STATE(90),
    [anon_sym_LBRACK] = ACTIONS(1064),
    [anon_sym_RBRACK] = ACTIONS(1062),
    [anon_sym_LT] = ACTIONS(1062),
    [anon_sym_GT] = ACTIONS(1062),
    [anon_sym_BANG] = ACTIONS(1062),
    [anon_sym_DQUOTE] = ACTIONS(1062),
    [anon_sym_POUND] = ACTIONS(1062),
    [anon_sym_DOLLAR] = ACTIONS(1062),
    [anon_sym_PERCENT] = ACTIONS(1062),
    [anon_sym_AMP] = ACTIONS(1062),
    [anon_sym_SQUOTE] = ACTIONS(1062),
    [anon_sym_STAR] = ACTIONS(1062),
    [anon_sym_PLUS] = ACTIONS(1062),
    [anon_sym_COMMA] = ACTIONS(1062),
    [anon_sym_DASH] = ACTIONS(1062),
    [anon_sym_DOT] = ACTIONS(1062),
    [anon_sym_SLASH] = ACTIONS(1062),
    [anon_sym_COLON] = ACTIONS(1062),
    [anon_sym_SEMI] = ACTIONS(1062),
    [anon_sym_EQ] = ACTIONS(1062),
    [anon_sym_QMARK] = ACTIONS(1062),
    [anon_sym_AT] = ACTIONS(1062),
    [anon_sym_BSLASH] = ACTIONS(1062),
    [anon_sym_CARET] = ACTIONS(1062),
    [anon_sym__] = ACTIONS(1062),
    [anon_sym_BQUOTE] = ACTIONS(1062),
    [anon_sym_LBRACE] = ACTIONS(1062),
    [anon_sym_PIPE] = ACTIONS(1062),
    [anon_sym_RBRACE] = ACTIONS(1062),
    [anon_sym_TILDE] = ACTIONS(1062),
    [anon_sym_LPAREN] = ACTIONS(1062),
    [anon_sym_RPAREN] = ACTIONS(1062),
    [aux_sym__word_token1] = ACTIONS(1062),
    [aux_sym__word_token2] = ACTIONS(1062),
    [aux_sym__word_token3] = ACTIONS(1062),
    [sym__whitespace] = ACTIONS(1062),
    [sym__soft_line_ending] = ACTIONS(1062),
    [sym__block_close] = ACTIONS(1062),
    [sym__block_quote_start] = ACTIONS(1062),
    [sym__indented_chunk_start] = ACTIONS(1062),
    [sym_atx_h1_marker] = ACTIONS(1062),
    [sym_atx_h2_marker] = ACTIONS(1062),
    [sym_atx_h3_marker] = ACTIONS(1062),
    [sym_atx_h4_marker] = ACTIONS(1062),
    [sym_atx_h5_marker] = ACTIONS(1062),
    [sym_atx_h6_marker] = ACTIONS(1062),
    [sym__thematic_break] = ACTIONS(1062),
    [sym__list_marker_minus] = ACTIONS(1062),
    [sym__list_marker_plus] = ACTIONS(1062),
    [sym__list_marker_star] = ACTIONS(1062),
    [sym__list_marker_parenthesis] = ACTIONS(1062),
    [sym__list_marker_dot] = ACTIONS(39),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1062),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(39),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1062),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1062),
    [sym__blank_line_start] = ACTIONS(1062),
    [sym__html_block_1_start] = ACTIONS(1062),
    [sym__html_block_2_start] = ACTIONS(1062),
    [sym__html_block_3_start] = ACTIONS(1062),
    [sym__html_block_4_start] = ACTIONS(1062),
    [sym__html_block_5_start] = ACTIONS(1062),
    [sym__html_block_6_start] = ACTIONS(1062),
    [sym__html_block_7_start] = ACTIONS(1062),
    [sym__pipe_table_start] = ACTIONS(1062),
  },
  [STATE(83)] = {
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_parenthesis] = STATE(91),
    [aux_sym__list_parenthesis_repeat1] = STATE(91),
    [anon_sym_LBRACK] = ACTIONS(1099),
    [anon_sym_RBRACK] = ACTIONS(1101),
    [anon_sym_LT] = ACTIONS(1101),
    [anon_sym_GT] = ACTIONS(1101),
    [anon_sym_BANG] = ACTIONS(1101),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_POUND] = ACTIONS(1101),
    [anon_sym_DOLLAR] = ACTIONS(1101),
    [anon_sym_PERCENT] = ACTIONS(1101),
    [anon_sym_AMP] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1101),
    [anon_sym_STAR] = ACTIONS(1101),
    [anon_sym_PLUS] = ACTIONS(1101),
    [anon_sym_COMMA] = ACTIONS(1101),
    [anon_sym_DASH] = ACTIONS(1101),
    [anon_sym_DOT] = ACTIONS(1101),
    [anon_sym_SLASH] = ACTIONS(1101),
    [anon_sym_COLON] = ACTIONS(1101),
    [anon_sym_SEMI] = ACTIONS(1101),
    [anon_sym_EQ] = ACTIONS(1101),
    [anon_sym_QMARK] = ACTIONS(1101),
    [anon_sym_AT] = ACTIONS(1101),
    [anon_sym_BSLASH] = ACTIONS(1101),
    [anon_sym_CARET] = ACTIONS(1101),
    [anon_sym__] = ACTIONS(1101),
    [anon_sym_BQUOTE] = ACTIONS(1101),
    [anon_sym_LBRACE] = ACTIONS(1101),
    [anon_sym_PIPE] = ACTIONS(1101),
    [anon_sym_RBRACE] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1101),
    [anon_sym_RPAREN] = ACTIONS(1101),
    [aux_sym__word_token1] = ACTIONS(1101),
    [aux_sym__word_token2] = ACTIONS(1101),
    [aux_sym__word_token3] = ACTIONS(1101),
    [sym__whitespace] = ACTIONS(1101),
    [sym__soft_line_ending] = ACTIONS(1101),
    [sym__block_close] = ACTIONS(1101),
    [sym__block_quote_start] = ACTIONS(1101),
    [sym__indented_chunk_start] = ACTIONS(1101),
    [sym_atx_h1_marker] = ACTIONS(1101),
    [sym_atx_h2_marker] = ACTIONS(1101),
    [sym_atx_h3_marker] = ACTIONS(1101),
    [sym_atx_h4_marker] = ACTIONS(1101),
    [sym_atx_h5_marker] = ACTIONS(1101),
    [sym_atx_h6_marker] = ACTIONS(1101),
    [sym__thematic_break] = ACTIONS(1101),
    [sym__list_marker_minus] = ACTIONS(1101),
    [sym__list_marker_plus] = ACTIONS(1101),
    [sym__list_marker_star] = ACTIONS(1101),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(1101),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1101),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1101),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1101),
    [sym__blank_line_start] = ACTIONS(1101),
    [sym__html_block_1_start] = ACTIONS(1101),
    [sym__html_block_2_start] = ACTIONS(1101),
    [sym__html_block_3_start] = ACTIONS(1101),
    [sym__html_block_4_start] = ACTIONS(1101),
    [sym__html_block_5_start] = ACTIONS(1101),
    [sym__html_block_6_start] = ACTIONS(1101),
    [sym__html_block_7_start] = ACTIONS(1101),
    [sym__pipe_table_start] = ACTIONS(1101),
  },
  [STATE(84)] = {
    [sym_list_marker_dot] = STATE(6),
    [sym__list_item_dot] = STATE(84),
    [aux_sym__list_dot_repeat1] = STATE(84),
    [ts_builtin_sym_end] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(1103),
    [anon_sym_LT] = ACTIONS(1103),
    [anon_sym_GT] = ACTIONS(1103),
    [anon_sym_BANG] = ACTIONS(1103),
    [anon_sym_DQUOTE] = ACTIONS(1103),
    [anon_sym_POUND] = ACTIONS(1103),
    [anon_sym_DOLLAR] = ACTIONS(1103),
    [anon_sym_PERCENT] = ACTIONS(1103),
    [anon_sym_AMP] = ACTIONS(1103),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_STAR] = ACTIONS(1103),
    [anon_sym_PLUS] = ACTIONS(1103),
    [anon_sym_COMMA] = ACTIONS(1103),
    [anon_sym_DASH] = ACTIONS(1103),
    [anon_sym_DOT] = ACTIONS(1103),
    [anon_sym_SLASH] = ACTIONS(1103),
    [anon_sym_COLON] = ACTIONS(1103),
    [anon_sym_SEMI] = ACTIONS(1103),
    [anon_sym_EQ] = ACTIONS(1103),
    [anon_sym_QMARK] = ACTIONS(1103),
    [anon_sym_AT] = ACTIONS(1103),
    [anon_sym_BSLASH] = ACTIONS(1103),
    [anon_sym_CARET] = ACTIONS(1103),
    [anon_sym__] = ACTIONS(1103),
    [anon_sym_BQUOTE] = ACTIONS(1103),
    [anon_sym_LBRACE] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(1103),
    [anon_sym_RBRACE] = ACTIONS(1103),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1103),
    [aux_sym__word_token1] = ACTIONS(1103),
    [aux_sym__word_token2] = ACTIONS(1103),
    [aux_sym__word_token3] = ACTIONS(1103),
    [sym__whitespace] = ACTIONS(1103),
    [sym__soft_line_ending] = ACTIONS(1103),
    [sym__block_quote_start] = ACTIONS(1103),
    [sym__indented_chunk_start] = ACTIONS(1103),
    [sym_atx_h1_marker] = ACTIONS(1103),
    [sym_atx_h2_marker] = ACTIONS(1103),
    [sym_atx_h3_marker] = ACTIONS(1103),
    [sym_atx_h4_marker] = ACTIONS(1103),
    [sym_atx_h5_marker] = ACTIONS(1103),
    [sym_atx_h6_marker] = ACTIONS(1103),
    [sym__thematic_break] = ACTIONS(1103),
    [sym__list_marker_minus] = ACTIONS(1103),
    [sym__list_marker_plus] = ACTIONS(1103),
    [sym__list_marker_star] = ACTIONS(1103),
    [sym__list_marker_parenthesis] = ACTIONS(1103),
    [sym__list_marker_dot] = ACTIONS(1107),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1107),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1103),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1103),
    [sym__blank_line_start] = ACTIONS(1103),
    [sym__html_block_1_start] = ACTIONS(1103),
    [sym__html_block_2_start] = ACTIONS(1103),
    [sym__html_block_3_start] = ACTIONS(1103),
    [sym__html_block_4_start] = ACTIONS(1103),
    [sym__html_block_5_start] = ACTIONS(1103),
    [sym__html_block_6_start] = ACTIONS(1103),
    [sym__html_block_7_start] = ACTIONS(1103),
    [sym__pipe_table_start] = ACTIONS(1103),
  },
  [STATE(85)] = {
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_parenthesis] = STATE(85),
    [aux_sym__list_parenthesis_repeat1] = STATE(85),
    [ts_builtin_sym_end] = ACTIONS(1110),
    [anon_sym_LBRACK] = ACTIONS(1112),
    [anon_sym_RBRACK] = ACTIONS(1110),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1110),
    [anon_sym_BANG] = ACTIONS(1110),
    [anon_sym_DQUOTE] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(1110),
    [anon_sym_DOLLAR] = ACTIONS(1110),
    [anon_sym_PERCENT] = ACTIONS(1110),
    [anon_sym_AMP] = ACTIONS(1110),
    [anon_sym_SQUOTE] = ACTIONS(1110),
    [anon_sym_STAR] = ACTIONS(1110),
    [anon_sym_PLUS] = ACTIONS(1110),
    [anon_sym_COMMA] = ACTIONS(1110),
    [anon_sym_DASH] = ACTIONS(1110),
    [anon_sym_DOT] = ACTIONS(1110),
    [anon_sym_SLASH] = ACTIONS(1110),
    [anon_sym_COLON] = ACTIONS(1110),
    [anon_sym_SEMI] = ACTIONS(1110),
    [anon_sym_EQ] = ACTIONS(1110),
    [anon_sym_QMARK] = ACTIONS(1110),
    [anon_sym_AT] = ACTIONS(1110),
    [anon_sym_BSLASH] = ACTIONS(1110),
    [anon_sym_CARET] = ACTIONS(1110),
    [anon_sym__] = ACTIONS(1110),
    [anon_sym_BQUOTE] = ACTIONS(1110),
    [anon_sym_LBRACE] = ACTIONS(1110),
    [anon_sym_PIPE] = ACTIONS(1110),
    [anon_sym_RBRACE] = ACTIONS(1110),
    [anon_sym_TILDE] = ACTIONS(1110),
    [anon_sym_LPAREN] = ACTIONS(1110),
    [anon_sym_RPAREN] = ACTIONS(1110),
    [aux_sym__word_token1] = ACTIONS(1110),
    [aux_sym__word_token2] = ACTIONS(1110),
    [aux_sym__word_token3] = ACTIONS(1110),
    [sym__whitespace] = ACTIONS(1110),
    [sym__soft_line_ending] = ACTIONS(1110),
    [sym__block_quote_start] = ACTIONS(1110),
    [sym__indented_chunk_start] = ACTIONS(1110),
    [sym_atx_h1_marker] = ACTIONS(1110),
    [sym_atx_h2_marker] = ACTIONS(1110),
    [sym_atx_h3_marker] = ACTIONS(1110),
    [sym_atx_h4_marker] = ACTIONS(1110),
    [sym_atx_h5_marker] = ACTIONS(1110),
    [sym_atx_h6_marker] = ACTIONS(1110),
    [sym__thematic_break] = ACTIONS(1110),
    [sym__list_marker_minus] = ACTIONS(1110),
    [sym__list_marker_plus] = ACTIONS(1110),
    [sym__list_marker_star] = ACTIONS(1110),
    [sym__list_marker_parenthesis] = ACTIONS(1114),
    [sym__list_marker_dot] = ACTIONS(1110),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1114),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1110),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1110),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1110),
    [sym__blank_line_start] = ACTIONS(1110),
    [sym__html_block_1_start] = ACTIONS(1110),
    [sym__html_block_2_start] = ACTIONS(1110),
    [sym__html_block_3_start] = ACTIONS(1110),
    [sym__html_block_4_start] = ACTIONS(1110),
    [sym__html_block_5_start] = ACTIONS(1110),
    [sym__html_block_6_start] = ACTIONS(1110),
    [sym__html_block_7_start] = ACTIONS(1110),
    [sym__pipe_table_start] = ACTIONS(1110),
  },
  [STATE(86)] = {
    [sym_list_marker_parenthesis] = STATE(5),
    [sym__list_item_parenthesis] = STATE(85),
    [aux_sym__list_parenthesis_repeat1] = STATE(85),
    [ts_builtin_sym_end] = ACTIONS(1101),
    [anon_sym_LBRACK] = ACTIONS(1099),
    [anon_sym_RBRACK] = ACTIONS(1101),
    [anon_sym_LT] = ACTIONS(1101),
    [anon_sym_GT] = ACTIONS(1101),
    [anon_sym_BANG] = ACTIONS(1101),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_POUND] = ACTIONS(1101),
    [anon_sym_DOLLAR] = ACTIONS(1101),
    [anon_sym_PERCENT] = ACTIONS(1101),
    [anon_sym_AMP] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1101),
    [anon_sym_STAR] = ACTIONS(1101),
    [anon_sym_PLUS] = ACTIONS(1101),
    [anon_sym_COMMA] = ACTIONS(1101),
    [anon_sym_DASH] = ACTIONS(1101),
    [anon_sym_DOT] = ACTIONS(1101),
    [anon_sym_SLASH] = ACTIONS(1101),
    [anon_sym_COLON] = ACTIONS(1101),
    [anon_sym_SEMI] = ACTIONS(1101),
    [anon_sym_EQ] = ACTIONS(1101),
    [anon_sym_QMARK] = ACTIONS(1101),
    [anon_sym_AT] = ACTIONS(1101),
    [anon_sym_BSLASH] = ACTIONS(1101),
    [anon_sym_CARET] = ACTIONS(1101),
    [anon_sym__] = ACTIONS(1101),
    [anon_sym_BQUOTE] = ACTIONS(1101),
    [anon_sym_LBRACE] = ACTIONS(1101),
    [anon_sym_PIPE] = ACTIONS(1101),
    [anon_sym_RBRACE] = ACTIONS(1101),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1101),
    [anon_sym_RPAREN] = ACTIONS(1101),
    [aux_sym__word_token1] = ACTIONS(1101),
    [aux_sym__word_token2] = ACTIONS(1101),
    [aux_sym__word_token3] = ACTIONS(1101),
    [sym__whitespace] = ACTIONS(1101),
    [sym__soft_line_ending] = ACTIONS(1101),
    [sym__block_quote_start] = ACTIONS(1101),
    [sym__indented_chunk_start] = ACTIONS(1101),
    [sym_atx_h1_marker] = ACTIONS(1101),
    [sym_atx_h2_marker] = ACTIONS(1101),
    [sym_atx_h3_marker] = ACTIONS(1101),
    [sym_atx_h4_marker] = ACTIONS(1101),
    [sym_atx_h5_marker] = ACTIONS(1101),
    [sym_atx_h6_marker] = ACTIONS(1101),
    [sym__thematic_break] = ACTIONS(1101),
    [sym__list_marker_minus] = ACTIONS(1101),
    [sym__list_marker_plus] = ACTIONS(1101),
    [sym__list_marker_star] = ACTIONS(1101),
    [sym__list_marker_parenthesis] = ACTIONS(37),
    [sym__list_marker_dot] = ACTIONS(1101),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1101),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(37),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1101),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1101),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1101),
    [sym__blank_line_start] = ACTIONS(1101),
    [sym__html_block_1_start] = ACTIONS(1101),
    [sym__html_block_2_start] = ACTIONS(1101),
    [sym__html_block_3_start] = ACTIONS(1101),
    [sym__html_block_4_start] = ACTIONS(1101),
    [sym__html_block_5_start] = ACTIONS(1101),
    [sym__html_block_6_start] = ACTIONS(1101),
    [sym__html_block_7_start] = ACTIONS(1101),
    [sym__pipe_table_start] = ACTIONS(1101),
  },
  [STATE(87)] = {
    [sym_list_marker_plus] = STATE(7),
    [sym__list_item_plus] = STATE(87),
    [aux_sym__list_plus_repeat1] = STATE(87),
    [anon_sym_LBRACK] = ACTIONS(1068),
    [anon_sym_RBRACK] = ACTIONS(1066),
    [anon_sym_LT] = ACTIONS(1066),
    [anon_sym_GT] = ACTIONS(1066),
    [anon_sym_BANG] = ACTIONS(1066),
    [anon_sym_DQUOTE] = ACTIONS(1066),
    [anon_sym_POUND] = ACTIONS(1066),
    [anon_sym_DOLLAR] = ACTIONS(1066),
    [anon_sym_PERCENT] = ACTIONS(1066),
    [anon_sym_AMP] = ACTIONS(1066),
    [anon_sym_SQUOTE] = ACTIONS(1066),
    [anon_sym_STAR] = ACTIONS(1066),
    [anon_sym_PLUS] = ACTIONS(1066),
    [anon_sym_COMMA] = ACTIONS(1066),
    [anon_sym_DASH] = ACTIONS(1066),
    [anon_sym_DOT] = ACTIONS(1066),
    [anon_sym_SLASH] = ACTIONS(1066),
    [anon_sym_COLON] = ACTIONS(1066),
    [anon_sym_SEMI] = ACTIONS(1066),
    [anon_sym_EQ] = ACTIONS(1066),
    [anon_sym_QMARK] = ACTIONS(1066),
    [anon_sym_AT] = ACTIONS(1066),
    [anon_sym_BSLASH] = ACTIONS(1066),
    [anon_sym_CARET] = ACTIONS(1066),
    [anon_sym__] = ACTIONS(1066),
    [anon_sym_BQUOTE] = ACTIONS(1066),
    [anon_sym_LBRACE] = ACTIONS(1066),
    [anon_sym_PIPE] = ACTIONS(1066),
    [anon_sym_RBRACE] = ACTIONS(1066),
    [anon_sym_TILDE] = ACTIONS(1066),
    [anon_sym_LPAREN] = ACTIONS(1066),
    [anon_sym_RPAREN] = ACTIONS(1066),
    [aux_sym__word_token1] = ACTIONS(1066),
    [aux_sym__word_token2] = ACTIONS(1066),
    [aux_sym__word_token3] = ACTIONS(1066),
    [sym__whitespace] = ACTIONS(1066),
    [sym__soft_line_ending] = ACTIONS(1066),
    [sym__block_close] = ACTIONS(1066),
    [sym__block_quote_start] = ACTIONS(1066),
    [sym__indented_chunk_start] = ACTIONS(1066),
    [sym_atx_h1_marker] = ACTIONS(1066),
    [sym_atx_h2_marker] = ACTIONS(1066),
    [sym_atx_h3_marker] = ACTIONS(1066),
    [sym_atx_h4_marker] = ACTIONS(1066),
    [sym_atx_h5_marker] = ACTIONS(1066),
    [sym_atx_h6_marker] = ACTIONS(1066),
    [sym__thematic_break] = ACTIONS(1066),
    [sym__list_marker_minus] = ACTIONS(1066),
    [sym__list_marker_plus] = ACTIONS(1070),
    [sym__list_marker_star] = ACTIONS(1066),
    [sym__list_marker_parenthesis] = ACTIONS(1066),
    [sym__list_marker_dot] = ACTIONS(1066),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1070),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1066),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1066),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1066),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1066),
    [sym__blank_line_start] = ACTIONS(1066),
    [sym__html_block_1_start] = ACTIONS(1066),
    [sym__html_block_2_start] = ACTIONS(1066),
    [sym__html_block_3_start] = ACTIONS(1066),
    [sym__html_block_4_start] = ACTIONS(1066),
    [sym__html_block_5_start] = ACTIONS(1066),
    [sym__html_block_6_start] = ACTIONS(1066),
    [sym__html_block_7_start] = ACTIONS(1066),
    [sym__pipe_table_start] = ACTIONS(1066),
  },
  [STATE(88)] = {
    [sym_list_marker_minus] = STATE(8),
    [sym__list_item_minus] = STATE(88),
    [aux_sym__list_minus_repeat1] = STATE(88),
    [anon_sym_LBRACK] = ACTIONS(1075),
    [anon_sym_RBRACK] = ACTIONS(1073),
    [anon_sym_LT] = ACTIONS(1073),
    [anon_sym_GT] = ACTIONS(1073),
    [anon_sym_BANG] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1073),
    [anon_sym_POUND] = ACTIONS(1073),
    [anon_sym_DOLLAR] = ACTIONS(1073),
    [anon_sym_PERCENT] = ACTIONS(1073),
    [anon_sym_AMP] = ACTIONS(1073),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_STAR] = ACTIONS(1073),
    [anon_sym_PLUS] = ACTIONS(1073),
    [anon_sym_COMMA] = ACTIONS(1073),
    [anon_sym_DASH] = ACTIONS(1073),
    [anon_sym_DOT] = ACTIONS(1073),
    [anon_sym_SLASH] = ACTIONS(1073),
    [anon_sym_COLON] = ACTIONS(1073),
    [anon_sym_SEMI] = ACTIONS(1073),
    [anon_sym_EQ] = ACTIONS(1073),
    [anon_sym_QMARK] = ACTIONS(1073),
    [anon_sym_AT] = ACTIONS(1073),
    [anon_sym_BSLASH] = ACTIONS(1073),
    [anon_sym_CARET] = ACTIONS(1073),
    [anon_sym__] = ACTIONS(1073),
    [anon_sym_BQUOTE] = ACTIONS(1073),
    [anon_sym_LBRACE] = ACTIONS(1073),
    [anon_sym_PIPE] = ACTIONS(1073),
    [anon_sym_RBRACE] = ACTIONS(1073),
    [anon_sym_TILDE] = ACTIONS(1073),
    [anon_sym_LPAREN] = ACTIONS(1073),
    [anon_sym_RPAREN] = ACTIONS(1073),
    [aux_sym__word_token1] = ACTIONS(1073),
    [aux_sym__word_token2] = ACTIONS(1073),
    [aux_sym__word_token3] = ACTIONS(1073),
    [sym__whitespace] = ACTIONS(1073),
    [sym__soft_line_ending] = ACTIONS(1073),
    [sym__block_close] = ACTIONS(1073),
    [sym__block_quote_start] = ACTIONS(1073),
    [sym__indented_chunk_start] = ACTIONS(1073),
    [sym_atx_h1_marker] = ACTIONS(1073),
    [sym_atx_h2_marker] = ACTIONS(1073),
    [sym_atx_h3_marker] = ACTIONS(1073),
    [sym_atx_h4_marker] = ACTIONS(1073),
    [sym_atx_h5_marker] = ACTIONS(1073),
    [sym_atx_h6_marker] = ACTIONS(1073),
    [sym__thematic_break] = ACTIONS(1073),
    [sym__list_marker_minus] = ACTIONS(1077),
    [sym__list_marker_plus] = ACTIONS(1073),
    [sym__list_marker_star] = ACTIONS(1073),
    [sym__list_marker_parenthesis] = ACTIONS(1073),
    [sym__list_marker_dot] = ACTIONS(1073),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1077),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1073),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1073),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1073),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1073),
    [sym__blank_line_start] = ACTIONS(1073),
    [sym__html_block_1_start] = ACTIONS(1073),
    [sym__html_block_2_start] = ACTIONS(1073),
    [sym__html_block_3_start] = ACTIONS(1073),
    [sym__html_block_4_start] = ACTIONS(1073),
    [sym__html_block_5_start] = ACTIONS(1073),
    [sym__html_block_6_start] = ACTIONS(1073),
    [sym__html_block_7_start] = ACTIONS(1073),
    [sym__pipe_table_start] = ACTIONS(1073),
  },
  [STATE(89)] = {
    [sym_list_marker_star] = STATE(9),
    [sym__list_item_star] = STATE(89),
    [aux_sym__list_star_repeat1] = STATE(89),
    [anon_sym_LBRACK] = ACTIONS(1082),
    [anon_sym_RBRACK] = ACTIONS(1080),
    [anon_sym_LT] = ACTIONS(1080),
    [anon_sym_GT] = ACTIONS(1080),
    [anon_sym_BANG] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1080),
    [anon_sym_POUND] = ACTIONS(1080),
    [anon_sym_DOLLAR] = ACTIONS(1080),
    [anon_sym_PERCENT] = ACTIONS(1080),
    [anon_sym_AMP] = ACTIONS(1080),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_STAR] = ACTIONS(1080),
    [anon_sym_PLUS] = ACTIONS(1080),
    [anon_sym_COMMA] = ACTIONS(1080),
    [anon_sym_DASH] = ACTIONS(1080),
    [anon_sym_DOT] = ACTIONS(1080),
    [anon_sym_SLASH] = ACTIONS(1080),
    [anon_sym_COLON] = ACTIONS(1080),
    [anon_sym_SEMI] = ACTIONS(1080),
    [anon_sym_EQ] = ACTIONS(1080),
    [anon_sym_QMARK] = ACTIONS(1080),
    [anon_sym_AT] = ACTIONS(1080),
    [anon_sym_BSLASH] = ACTIONS(1080),
    [anon_sym_CARET] = ACTIONS(1080),
    [anon_sym__] = ACTIONS(1080),
    [anon_sym_BQUOTE] = ACTIONS(1080),
    [anon_sym_LBRACE] = ACTIONS(1080),
    [anon_sym_PIPE] = ACTIONS(1080),
    [anon_sym_RBRACE] = ACTIONS(1080),
    [anon_sym_TILDE] = ACTIONS(1080),
    [anon_sym_LPAREN] = ACTIONS(1080),
    [anon_sym_RPAREN] = ACTIONS(1080),
    [aux_sym__word_token1] = ACTIONS(1080),
    [aux_sym__word_token2] = ACTIONS(1080),
    [aux_sym__word_token3] = ACTIONS(1080),
    [sym__whitespace] = ACTIONS(1080),
    [sym__soft_line_ending] = ACTIONS(1080),
    [sym__block_close] = ACTIONS(1080),
    [sym__block_quote_start] = ACTIONS(1080),
    [sym__indented_chunk_start] = ACTIONS(1080),
    [sym_atx_h1_marker] = ACTIONS(1080),
    [sym_atx_h2_marker] = ACTIONS(1080),
    [sym_atx_h3_marker] = ACTIONS(1080),
    [sym_atx_h4_marker] = ACTIONS(1080),
    [sym_atx_h5_marker] = ACTIONS(1080),
    [sym_atx_h6_marker] = ACTIONS(1080),
    [sym__thematic_break] = ACTIONS(1080),
    [sym__list_marker_minus] = ACTIONS(1080),
    [sym__list_marker_plus] = ACTIONS(1080),
    [sym__list_marker_star] = ACTIONS(1084),
    [sym__list_marker_parenthesis] = ACTIONS(1080),
    [sym__list_marker_dot] = ACTIONS(1080),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1084),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1080),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1080),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1080),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1080),
    [sym__blank_line_start] = ACTIONS(1080),
    [sym__html_block_1_start] = ACTIONS(1080),
    [sym__html_block_2_start] = ACTIONS(1080),
    [sym__html_block_3_start] = ACTIONS(1080),
    [sym__html_block_4_start] = ACTIONS(1080),
    [sym__html_block_5_start] = ACTIONS(1080),
    [sym__html_block_6_start] = ACTIONS(1080),
    [sym__html_block_7_start] = ACTIONS(1080),
    [sym__pipe_table_start] = ACTIONS(1080),
  },
  [STATE(90)] = {
    [sym_list_marker_dot] = STATE(10),
    [sym__list_item_dot] = STATE(90),
    [aux_sym__list_dot_repeat1] = STATE(90),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(1103),
    [anon_sym_LT] = ACTIONS(1103),
    [anon_sym_GT] = ACTIONS(1103),
    [anon_sym_BANG] = ACTIONS(1103),
    [anon_sym_DQUOTE] = ACTIONS(1103),
    [anon_sym_POUND] = ACTIONS(1103),
    [anon_sym_DOLLAR] = ACTIONS(1103),
    [anon_sym_PERCENT] = ACTIONS(1103),
    [anon_sym_AMP] = ACTIONS(1103),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_STAR] = ACTIONS(1103),
    [anon_sym_PLUS] = ACTIONS(1103),
    [anon_sym_COMMA] = ACTIONS(1103),
    [anon_sym_DASH] = ACTIONS(1103),
    [anon_sym_DOT] = ACTIONS(1103),
    [anon_sym_SLASH] = ACTIONS(1103),
    [anon_sym_COLON] = ACTIONS(1103),
    [anon_sym_SEMI] = ACTIONS(1103),
    [anon_sym_EQ] = ACTIONS(1103),
    [anon_sym_QMARK] = ACTIONS(1103),
    [anon_sym_AT] = ACTIONS(1103),
    [anon_sym_BSLASH] = ACTIONS(1103),
    [anon_sym_CARET] = ACTIONS(1103),
    [anon_sym__] = ACTIONS(1103),
    [anon_sym_BQUOTE] = ACTIONS(1103),
    [anon_sym_LBRACE] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(1103),
    [anon_sym_RBRACE] = ACTIONS(1103),
    [anon_sym_TILDE] = ACTIONS(1103),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1103),
    [aux_sym__word_token1] = ACTIONS(1103),
    [aux_sym__word_token2] = ACTIONS(1103),
    [aux_sym__word_token3] = ACTIONS(1103),
    [sym__whitespace] = ACTIONS(1103),
    [sym__soft_line_ending] = ACTIONS(1103),
    [sym__block_close] = ACTIONS(1103),
    [sym__block_quote_start] = ACTIONS(1103),
    [sym__indented_chunk_start] = ACTIONS(1103),
    [sym_atx_h1_marker] = ACTIONS(1103),
    [sym_atx_h2_marker] = ACTIONS(1103),
    [sym_atx_h3_marker] = ACTIONS(1103),
    [sym_atx_h4_marker] = ACTIONS(1103),
    [sym_atx_h5_marker] = ACTIONS(1103),
    [sym_atx_h6_marker] = ACTIONS(1103),
    [sym__thematic_break] = ACTIONS(1103),
    [sym__list_marker_minus] = ACTIONS(1103),
    [sym__list_marker_plus] = ACTIONS(1103),
    [sym__list_marker_star] = ACTIONS(1103),
    [sym__list_marker_parenthesis] = ACTIONS(1103),
    [sym__list_marker_dot] = ACTIONS(1107),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1103),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1107),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1103),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1103),
    [sym__blank_line_start] = ACTIONS(1103),
    [sym__html_block_1_start] = ACTIONS(1103),
    [sym__html_block_2_start] = ACTIONS(1103),
    [sym__html_block_3_start] = ACTIONS(1103),
    [sym__html_block_4_start] = ACTIONS(1103),
    [sym__html_block_5_start] = ACTIONS(1103),
    [sym__html_block_6_start] = ACTIONS(1103),
    [sym__html_block_7_start] = ACTIONS(1103),
    [sym__pipe_table_start] = ACTIONS(1103),
  },
  [STATE(91)] = {
    [sym_list_marker_parenthesis] = STATE(11),
    [sym__list_item_parenthesis] = STATE(91),
    [aux_sym__list_parenthesis_repeat1] = STATE(91),
    [anon_sym_LBRACK] = ACTIONS(1112),
    [anon_sym_RBRACK] = ACTIONS(1110),
    [anon_sym_LT] = ACTIONS(1110),
    [anon_sym_GT] = ACTIONS(1110),
    [anon_sym_BANG] = ACTIONS(1110),
    [anon_sym_DQUOTE] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(1110),
    [anon_sym_DOLLAR] = ACTIONS(1110),
    [anon_sym_PERCENT] = ACTIONS(1110),
    [anon_sym_AMP] = ACTIONS(1110),
    [anon_sym_SQUOTE] = ACTIONS(1110),
    [anon_sym_STAR] = ACTIONS(1110),
    [anon_sym_PLUS] = ACTIONS(1110),
    [anon_sym_COMMA] = ACTIONS(1110),
    [anon_sym_DASH] = ACTIONS(1110),
    [anon_sym_DOT] = ACTIONS(1110),
    [anon_sym_SLASH] = ACTIONS(1110),
    [anon_sym_COLON] = ACTIONS(1110),
    [anon_sym_SEMI] = ACTIONS(1110),
    [anon_sym_EQ] = ACTIONS(1110),
    [anon_sym_QMARK] = ACTIONS(1110),
    [anon_sym_AT] = ACTIONS(1110),
    [anon_sym_BSLASH] = ACTIONS(1110),
    [anon_sym_CARET] = ACTIONS(1110),
    [anon_sym__] = ACTIONS(1110),
    [anon_sym_BQUOTE] = ACTIONS(1110),
    [anon_sym_LBRACE] = ACTIONS(1110),
    [anon_sym_PIPE] = ACTIONS(1110),
    [anon_sym_RBRACE] = ACTIONS(1110),
    [anon_sym_TILDE] = ACTIONS(1110),
    [anon_sym_LPAREN] = ACTIONS(1110),
    [anon_sym_RPAREN] = ACTIONS(1110),
    [aux_sym__word_token1] = ACTIONS(1110),
    [aux_sym__word_token2] = ACTIONS(1110),
    [aux_sym__word_token3] = ACTIONS(1110),
    [sym__whitespace] = ACTIONS(1110),
    [sym__soft_line_ending] = ACTIONS(1110),
    [sym__block_close] = ACTIONS(1110),
    [sym__block_quote_start] = ACTIONS(1110),
    [sym__indented_chunk_start] = ACTIONS(1110),
    [sym_atx_h1_marker] = ACTIONS(1110),
    [sym_atx_h2_marker] = ACTIONS(1110),
    [sym_atx_h3_marker] = ACTIONS(1110),
    [sym_atx_h4_marker] = ACTIONS(1110),
    [sym_atx_h5_marker] = ACTIONS(1110),
    [sym_atx_h6_marker] = ACTIONS(1110),
    [sym__thematic_break] = ACTIONS(1110),
    [sym__list_marker_minus] = ACTIONS(1110),
    [sym__list_marker_plus] = ACTIONS(1110),
    [sym__list_marker_star] = ACTIONS(1110),
    [sym__list_marker_parenthesis] = ACTIONS(1114),
    [sym__list_marker_dot] = ACTIONS(1110),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1110),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1114),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1110),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1110),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1110),
    [sym__blank_line_start] = ACTIONS(1110),
    [sym__html_block_1_start] = ACTIONS(1110),
    [sym__html_block_2_start] = ACTIONS(1110),
    [sym__html_block_3_start] = ACTIONS(1110),
    [sym__html_block_4_start] = ACTIONS(1110),
    [sym__html_block_5_start] = ACTIONS(1110),
    [sym__html_block_6_start] = ACTIONS(1110),
    [sym__html_block_7_start] = ACTIONS(1110),
    [sym__pipe_table_start] = ACTIONS(1110),
  },
  [STATE(92)] = {
    [sym__indented_chunk] = STATE(92),
    [sym__blank_line] = STATE(92),
    [aux_sym_indented_code_block_repeat1] = STATE(92),
    [anon_sym_LBRACK] = ACTIONS(1117),
    [anon_sym_RBRACK] = ACTIONS(1119),
    [anon_sym_LT] = ACTIONS(1119),
    [anon_sym_GT] = ACTIONS(1119),
    [anon_sym_BANG] = ACTIONS(1119),
    [anon_sym_DQUOTE] = ACTIONS(1119),
    [anon_sym_POUND] = ACTIONS(1119),
    [anon_sym_DOLLAR] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1119),
    [anon_sym_AMP] = ACTIONS(1119),
    [anon_sym_SQUOTE] = ACTIONS(1119),
    [anon_sym_STAR] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1119),
    [anon_sym_COMMA] = ACTIONS(1119),
    [anon_sym_DASH] = ACTIONS(1119),
    [anon_sym_DOT] = ACTIONS(1119),
    [anon_sym_SLASH] = ACTIONS(1119),
    [anon_sym_COLON] = ACTIONS(1119),
    [anon_sym_SEMI] = ACTIONS(1119),
    [anon_sym_EQ] = ACTIONS(1119),
    [anon_sym_QMARK] = ACTIONS(1119),
    [anon_sym_AT] = ACTIONS(1119),
    [anon_sym_BSLASH] = ACTIONS(1119),
    [anon_sym_CARET] = ACTIONS(1119),
    [anon_sym__] = ACTIONS(1119),
    [anon_sym_BQUOTE] = ACTIONS(1119),
    [anon_sym_LBRACE] = ACTIONS(1119),
    [anon_sym_PIPE] = ACTIONS(1119),
    [anon_sym_RBRACE] = ACTIONS(1119),
    [anon_sym_TILDE] = ACTIONS(1119),
    [anon_sym_LPAREN] = ACTIONS(1119),
    [anon_sym_RPAREN] = ACTIONS(1119),
    [aux_sym__word_token1] = ACTIONS(1119),
    [aux_sym__word_token2] = ACTIONS(1119),
    [aux_sym__word_token3] = ACTIONS(1119),
    [sym__whitespace] = ACTIONS(1119),
    [sym__soft_line_ending] = ACTIONS(1119),
    [sym__block_close] = ACTIONS(1119),
    [sym__block_quote_start] = ACTIONS(1119),
    [sym__indented_chunk_start] = ACTIONS(1121),
    [sym_atx_h1_marker] = ACTIONS(1119),
    [sym_atx_h2_marker] = ACTIONS(1119),
    [sym_atx_h3_marker] = ACTIONS(1119),
    [sym_atx_h4_marker] = ACTIONS(1119),
    [sym_atx_h5_marker] = ACTIONS(1119),
    [sym_atx_h6_marker] = ACTIONS(1119),
    [sym__thematic_break] = ACTIONS(1119),
    [sym__list_marker_minus] = ACTIONS(1119),
    [sym__list_marker_plus] = ACTIONS(1119),
    [sym__list_marker_star] = ACTIONS(1119),
    [sym__list_marker_parenthesis] = ACTIONS(1119),
    [sym__list_marker_dot] = ACTIONS(1119),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1119),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1119),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1119),
    [sym__blank_line_start] = ACTIONS(1124),
    [sym__html_block_1_start] = ACTIONS(1119),
    [sym__html_block_2_start] = ACTIONS(1119),
    [sym__html_block_3_start] = ACTIONS(1119),
    [sym__html_block_4_start] = ACTIONS(1119),
    [sym__html_block_5_start] = ACTIONS(1119),
    [sym__html_block_6_start] = ACTIONS(1119),
    [sym__html_block_7_start] = ACTIONS(1119),
    [sym__pipe_table_start] = ACTIONS(1119),
  },
  [STATE(93)] = {
    [ts_builtin_sym_end] = ACTIONS(1127),
    [anon_sym_LBRACK] = ACTIONS(1129),
    [anon_sym_RBRACK] = ACTIONS(1127),
    [anon_sym_LT] = ACTIONS(1127),
    [anon_sym_GT] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_DQUOTE] = ACTIONS(1127),
    [anon_sym_POUND] = ACTIONS(1127),
    [anon_sym_DOLLAR] = ACTIONS(1127),
    [anon_sym_PERCENT] = ACTIONS(1127),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_SQUOTE] = ACTIONS(1127),
    [anon_sym_STAR] = ACTIONS(1127),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_COMMA] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_DOT] = ACTIONS(1127),
    [anon_sym_SLASH] = ACTIONS(1127),
    [anon_sym_COLON] = ACTIONS(1127),
    [anon_sym_SEMI] = ACTIONS(1127),
    [anon_sym_EQ] = ACTIONS(1127),
    [anon_sym_QMARK] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1127),
    [anon_sym_BSLASH] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym__] = ACTIONS(1127),
    [anon_sym_BQUOTE] = ACTIONS(1127),
    [anon_sym_LBRACE] = ACTIONS(1127),
    [anon_sym_PIPE] = ACTIONS(1127),
    [anon_sym_RBRACE] = ACTIONS(1127),
    [anon_sym_TILDE] = ACTIONS(1127),
    [anon_sym_LPAREN] = ACTIONS(1127),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [aux_sym__word_token1] = ACTIONS(1127),
    [aux_sym__word_token2] = ACTIONS(1127),
    [aux_sym__word_token3] = ACTIONS(1127),
    [sym__whitespace] = ACTIONS(1127),
    [sym__soft_line_ending] = ACTIONS(1127),
    [sym_block_continuation] = ACTIONS(1131),
    [sym__block_quote_start] = ACTIONS(1127),
    [sym__indented_chunk_start] = ACTIONS(1127),
    [sym_atx_h1_marker] = ACTIONS(1127),
    [sym_atx_h2_marker] = ACTIONS(1127),
    [sym_atx_h3_marker] = ACTIONS(1127),
    [sym_atx_h4_marker] = ACTIONS(1127),
    [sym_atx_h5_marker] = ACTIONS(1127),
    [sym_atx_h6_marker] = ACTIONS(1127),
    [sym_setext_h1_underline] = ACTIONS(1127),
    [sym_setext_h2_underline] = ACTIONS(1127),
    [sym__thematic_break] = ACTIONS(1127),
    [sym__list_marker_minus] = ACTIONS(1127),
    [sym__list_marker_plus] = ACTIONS(1127),
    [sym__list_marker_star] = ACTIONS(1127),
    [sym__list_marker_parenthesis] = ACTIONS(1127),
    [sym__list_marker_dot] = ACTIONS(1127),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1127),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1127),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1127),
    [sym__blank_line_start] = ACTIONS(1127),
    [sym__html_block_1_start] = ACTIONS(1127),
    [sym__html_block_2_start] = ACTIONS(1127),
    [sym__html_block_3_start] = ACTIONS(1127),
    [sym__html_block_4_start] = ACTIONS(1127),
    [sym__html_block_5_start] = ACTIONS(1127),
    [sym__html_block_6_start] = ACTIONS(1127),
    [sym__html_block_7_start] = ACTIONS(1127),
    [sym__pipe_table_start] = ACTIONS(1127),
  },
  [STATE(94)] = {
    [anon_sym_LBRACK] = ACTIONS(1129),
    [anon_sym_RBRACK] = ACTIONS(1127),
    [anon_sym_LT] = ACTIONS(1127),
    [anon_sym_GT] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_DQUOTE] = ACTIONS(1127),
    [anon_sym_POUND] = ACTIONS(1127),
    [anon_sym_DOLLAR] = ACTIONS(1127),
    [anon_sym_PERCENT] = ACTIONS(1127),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_SQUOTE] = ACTIONS(1127),
    [anon_sym_STAR] = ACTIONS(1127),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_COMMA] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_DOT] = ACTIONS(1127),
    [anon_sym_SLASH] = ACTIONS(1127),
    [anon_sym_COLON] = ACTIONS(1127),
    [anon_sym_SEMI] = ACTIONS(1127),
    [anon_sym_EQ] = ACTIONS(1127),
    [anon_sym_QMARK] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1127),
    [anon_sym_BSLASH] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym__] = ACTIONS(1127),
    [anon_sym_BQUOTE] = ACTIONS(1127),
    [anon_sym_LBRACE] = ACTIONS(1127),
    [anon_sym_PIPE] = ACTIONS(1127),
    [anon_sym_RBRACE] = ACTIONS(1127),
    [anon_sym_TILDE] = ACTIONS(1127),
    [anon_sym_LPAREN] = ACTIONS(1127),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [aux_sym__word_token1] = ACTIONS(1127),
    [aux_sym__word_token2] = ACTIONS(1127),
    [aux_sym__word_token3] = ACTIONS(1127),
    [sym__whitespace] = ACTIONS(1127),
    [sym__soft_line_ending] = ACTIONS(1127),
    [sym__block_close] = ACTIONS(1127),
    [sym_block_continuation] = ACTIONS(1133),
    [sym__block_quote_start] = ACTIONS(1127),
    [sym__indented_chunk_start] = ACTIONS(1127),
    [sym_atx_h1_marker] = ACTIONS(1127),
    [sym_atx_h2_marker] = ACTIONS(1127),
    [sym_atx_h3_marker] = ACTIONS(1127),
    [sym_atx_h4_marker] = ACTIONS(1127),
    [sym_atx_h5_marker] = ACTIONS(1127),
    [sym_atx_h6_marker] = ACTIONS(1127),
    [sym_setext_h1_underline] = ACTIONS(1127),
    [sym_setext_h2_underline] = ACTIONS(1127),
    [sym__thematic_break] = ACTIONS(1127),
    [sym__list_marker_minus] = ACTIONS(1127),
    [sym__list_marker_plus] = ACTIONS(1127),
    [sym__list_marker_star] = ACTIONS(1127),
    [sym__list_marker_parenthesis] = ACTIONS(1127),
    [sym__list_marker_dot] = ACTIONS(1127),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1127),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1127),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1127),
    [sym__blank_line_start] = ACTIONS(1127),
    [sym__html_block_1_start] = ACTIONS(1127),
    [sym__html_block_2_start] = ACTIONS(1127),
    [sym__html_block_3_start] = ACTIONS(1127),
    [sym__html_block_4_start] = ACTIONS(1127),
    [sym__html_block_5_start] = ACTIONS(1127),
    [sym__html_block_6_start] = ACTIONS(1127),
    [sym__html_block_7_start] = ACTIONS(1127),
    [sym__pipe_table_start] = ACTIONS(1127),
  },
  [STATE(95)] = {
    [sym__indented_chunk] = STATE(97),
    [sym__blank_line] = STATE(97),
    [aux_sym_indented_code_block_repeat1] = STATE(97),
    [ts_builtin_sym_end] = ACTIONS(1135),
    [anon_sym_LBRACK] = ACTIONS(1137),
    [anon_sym_RBRACK] = ACTIONS(1135),
    [anon_sym_LT] = ACTIONS(1135),
    [anon_sym_GT] = ACTIONS(1135),
    [anon_sym_BANG] = ACTIONS(1135),
    [anon_sym_DQUOTE] = ACTIONS(1135),
    [anon_sym_POUND] = ACTIONS(1135),
    [anon_sym_DOLLAR] = ACTIONS(1135),
    [anon_sym_PERCENT] = ACTIONS(1135),
    [anon_sym_AMP] = ACTIONS(1135),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_STAR] = ACTIONS(1135),
    [anon_sym_PLUS] = ACTIONS(1135),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_DASH] = ACTIONS(1135),
    [anon_sym_DOT] = ACTIONS(1135),
    [anon_sym_SLASH] = ACTIONS(1135),
    [anon_sym_COLON] = ACTIONS(1135),
    [anon_sym_SEMI] = ACTIONS(1135),
    [anon_sym_EQ] = ACTIONS(1135),
    [anon_sym_QMARK] = ACTIONS(1135),
    [anon_sym_AT] = ACTIONS(1135),
    [anon_sym_BSLASH] = ACTIONS(1135),
    [anon_sym_CARET] = ACTIONS(1135),
    [anon_sym__] = ACTIONS(1135),
    [anon_sym_BQUOTE] = ACTIONS(1135),
    [anon_sym_LBRACE] = ACTIONS(1135),
    [anon_sym_PIPE] = ACTIONS(1135),
    [anon_sym_RBRACE] = ACTIONS(1135),
    [anon_sym_TILDE] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1135),
    [anon_sym_RPAREN] = ACTIONS(1135),
    [aux_sym__word_token1] = ACTIONS(1135),
    [aux_sym__word_token2] = ACTIONS(1135),
    [aux_sym__word_token3] = ACTIONS(1135),
    [sym__whitespace] = ACTIONS(1135),
    [sym__soft_line_ending] = ACTIONS(1135),
    [sym__block_quote_start] = ACTIONS(1135),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(1135),
    [sym_atx_h2_marker] = ACTIONS(1135),
    [sym_atx_h3_marker] = ACTIONS(1135),
    [sym_atx_h4_marker] = ACTIONS(1135),
    [sym_atx_h5_marker] = ACTIONS(1135),
    [sym_atx_h6_marker] = ACTIONS(1135),
    [sym__thematic_break] = ACTIONS(1135),
    [sym__list_marker_minus] = ACTIONS(1135),
    [sym__list_marker_plus] = ACTIONS(1135),
    [sym__list_marker_star] = ACTIONS(1135),
    [sym__list_marker_parenthesis] = ACTIONS(1135),
    [sym__list_marker_dot] = ACTIONS(1135),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1135),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1135),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1135),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(1135),
    [sym__html_block_2_start] = ACTIONS(1135),
    [sym__html_block_3_start] = ACTIONS(1135),
    [sym__html_block_4_start] = ACTIONS(1135),
    [sym__html_block_5_start] = ACTIONS(1135),
    [sym__html_block_6_start] = ACTIONS(1135),
    [sym__html_block_7_start] = ACTIONS(1135),
    [sym__pipe_table_start] = ACTIONS(1135),
  },
  [STATE(96)] = {
    [sym__indented_chunk] = STATE(95),
    [sym__blank_line] = STATE(95),
    [aux_sym_indented_code_block_repeat1] = STATE(95),
    [ts_builtin_sym_end] = ACTIONS(1089),
    [anon_sym_LBRACK] = ACTIONS(1087),
    [anon_sym_RBRACK] = ACTIONS(1089),
    [anon_sym_LT] = ACTIONS(1089),
    [anon_sym_GT] = ACTIONS(1089),
    [anon_sym_BANG] = ACTIONS(1089),
    [anon_sym_DQUOTE] = ACTIONS(1089),
    [anon_sym_POUND] = ACTIONS(1089),
    [anon_sym_DOLLAR] = ACTIONS(1089),
    [anon_sym_PERCENT] = ACTIONS(1089),
    [anon_sym_AMP] = ACTIONS(1089),
    [anon_sym_SQUOTE] = ACTIONS(1089),
    [anon_sym_STAR] = ACTIONS(1089),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_COMMA] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_DOT] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1089),
    [anon_sym_COLON] = ACTIONS(1089),
    [anon_sym_SEMI] = ACTIONS(1089),
    [anon_sym_EQ] = ACTIONS(1089),
    [anon_sym_QMARK] = ACTIONS(1089),
    [anon_sym_AT] = ACTIONS(1089),
    [anon_sym_BSLASH] = ACTIONS(1089),
    [anon_sym_CARET] = ACTIONS(1089),
    [anon_sym__] = ACTIONS(1089),
    [anon_sym_BQUOTE] = ACTIONS(1089),
    [anon_sym_LBRACE] = ACTIONS(1089),
    [anon_sym_PIPE] = ACTIONS(1089),
    [anon_sym_RBRACE] = ACTIONS(1089),
    [anon_sym_TILDE] = ACTIONS(1089),
    [anon_sym_LPAREN] = ACTIONS(1089),
    [anon_sym_RPAREN] = ACTIONS(1089),
    [aux_sym__word_token1] = ACTIONS(1089),
    [aux_sym__word_token2] = ACTIONS(1089),
    [aux_sym__word_token3] = ACTIONS(1089),
    [sym__whitespace] = ACTIONS(1089),
    [sym__soft_line_ending] = ACTIONS(1089),
    [sym__block_quote_start] = ACTIONS(1089),
    [sym__indented_chunk_start] = ACTIONS(15),
    [sym_atx_h1_marker] = ACTIONS(1089),
    [sym_atx_h2_marker] = ACTIONS(1089),
    [sym_atx_h3_marker] = ACTIONS(1089),
    [sym_atx_h4_marker] = ACTIONS(1089),
    [sym_atx_h5_marker] = ACTIONS(1089),
    [sym_atx_h6_marker] = ACTIONS(1089),
    [sym__thematic_break] = ACTIONS(1089),
    [sym__list_marker_minus] = ACTIONS(1089),
    [sym__list_marker_plus] = ACTIONS(1089),
    [sym__list_marker_star] = ACTIONS(1089),
    [sym__list_marker_parenthesis] = ACTIONS(1089),
    [sym__list_marker_dot] = ACTIONS(1089),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1089),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1089),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1089),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1089),
    [sym__blank_line_start] = ACTIONS(45),
    [sym__html_block_1_start] = ACTIONS(1089),
    [sym__html_block_2_start] = ACTIONS(1089),
    [sym__html_block_3_start] = ACTIONS(1089),
    [sym__html_block_4_start] = ACTIONS(1089),
    [sym__html_block_5_start] = ACTIONS(1089),
    [sym__html_block_6_start] = ACTIONS(1089),
    [sym__html_block_7_start] = ACTIONS(1089),
    [sym__pipe_table_start] = ACTIONS(1089),
  },
  [STATE(97)] = {
    [sym__indented_chunk] = STATE(97),
    [sym__blank_line] = STATE(97),
    [aux_sym_indented_code_block_repeat1] = STATE(97),
    [ts_builtin_sym_end] = ACTIONS(1119),
    [anon_sym_LBRACK] = ACTIONS(1117),
    [anon_sym_RBRACK] = ACTIONS(1119),
    [anon_sym_LT] = ACTIONS(1119),
    [anon_sym_GT] = ACTIONS(1119),
    [anon_sym_BANG] = ACTIONS(1119),
    [anon_sym_DQUOTE] = ACTIONS(1119),
    [anon_sym_POUND] = ACTIONS(1119),
    [anon_sym_DOLLAR] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1119),
    [anon_sym_AMP] = ACTIONS(1119),
    [anon_sym_SQUOTE] = ACTIONS(1119),
    [anon_sym_STAR] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1119),
    [anon_sym_COMMA] = ACTIONS(1119),
    [anon_sym_DASH] = ACTIONS(1119),
    [anon_sym_DOT] = ACTIONS(1119),
    [anon_sym_SLASH] = ACTIONS(1119),
    [anon_sym_COLON] = ACTIONS(1119),
    [anon_sym_SEMI] = ACTIONS(1119),
    [anon_sym_EQ] = ACTIONS(1119),
    [anon_sym_QMARK] = ACTIONS(1119),
    [anon_sym_AT] = ACTIONS(1119),
    [anon_sym_BSLASH] = ACTIONS(1119),
    [anon_sym_CARET] = ACTIONS(1119),
    [anon_sym__] = ACTIONS(1119),
    [anon_sym_BQUOTE] = ACTIONS(1119),
    [anon_sym_LBRACE] = ACTIONS(1119),
    [anon_sym_PIPE] = ACTIONS(1119),
    [anon_sym_RBRACE] = ACTIONS(1119),
    [anon_sym_TILDE] = ACTIONS(1119),
    [anon_sym_LPAREN] = ACTIONS(1119),
    [anon_sym_RPAREN] = ACTIONS(1119),
    [aux_sym__word_token1] = ACTIONS(1119),
    [aux_sym__word_token2] = ACTIONS(1119),
    [aux_sym__word_token3] = ACTIONS(1119),
    [sym__whitespace] = ACTIONS(1119),
    [sym__soft_line_ending] = ACTIONS(1119),
    [sym__block_quote_start] = ACTIONS(1119),
    [sym__indented_chunk_start] = ACTIONS(1139),
    [sym_atx_h1_marker] = ACTIONS(1119),
    [sym_atx_h2_marker] = ACTIONS(1119),
    [sym_atx_h3_marker] = ACTIONS(1119),
    [sym_atx_h4_marker] = ACTIONS(1119),
    [sym_atx_h5_marker] = ACTIONS(1119),
    [sym_atx_h6_marker] = ACTIONS(1119),
    [sym__thematic_break] = ACTIONS(1119),
    [sym__list_marker_minus] = ACTIONS(1119),
    [sym__list_marker_plus] = ACTIONS(1119),
    [sym__list_marker_star] = ACTIONS(1119),
    [sym__list_marker_parenthesis] = ACTIONS(1119),
    [sym__list_marker_dot] = ACTIONS(1119),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1119),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1119),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1119),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1119),
    [sym__blank_line_start] = ACTIONS(1142),
    [sym__html_block_1_start] = ACTIONS(1119),
    [sym__html_block_2_start] = ACTIONS(1119),
    [sym__html_block_3_start] = ACTIONS(1119),
    [sym__html_block_4_start] = ACTIONS(1119),
    [sym__html_block_5_start] = ACTIONS(1119),
    [sym__html_block_6_start] = ACTIONS(1119),
    [sym__html_block_7_start] = ACTIONS(1119),
    [sym__pipe_table_start] = ACTIONS(1119),
  },
  [STATE(98)] = {
    [sym_list_marker_plus] = STATE(3),
    [sym__list_item_plus] = STATE(75),
    [aux_sym__list_plus_repeat1] = STATE(75),
    [ts_builtin_sym_end] = ACTIONS(1093),
    [anon_sym_LBRACK] = ACTIONS(1091),
    [anon_sym_RBRACK] = ACTIONS(1093),
    [anon_sym_LT] = ACTIONS(1093),
    [anon_sym_GT] = ACTIONS(1093),
    [anon_sym_BANG] = ACTIONS(1093),
    [anon_sym_DQUOTE] = ACTIONS(1093),
    [anon_sym_POUND] = ACTIONS(1093),
    [anon_sym_DOLLAR] = ACTIONS(1093),
    [anon_sym_PERCENT] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1093),
    [anon_sym_SQUOTE] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(1093),
    [anon_sym_PLUS] = ACTIONS(1093),
    [anon_sym_COMMA] = ACTIONS(1093),
    [anon_sym_DASH] = ACTIONS(1093),
    [anon_sym_DOT] = ACTIONS(1093),
    [anon_sym_SLASH] = ACTIONS(1093),
    [anon_sym_COLON] = ACTIONS(1093),
    [anon_sym_SEMI] = ACTIONS(1093),
    [anon_sym_EQ] = ACTIONS(1093),
    [anon_sym_QMARK] = ACTIONS(1093),
    [anon_sym_AT] = ACTIONS(1093),
    [anon_sym_BSLASH] = ACTIONS(1093),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym__] = ACTIONS(1093),
    [anon_sym_BQUOTE] = ACTIONS(1093),
    [anon_sym_LBRACE] = ACTIONS(1093),
    [anon_sym_PIPE] = ACTIONS(1093),
    [anon_sym_RBRACE] = ACTIONS(1093),
    [anon_sym_TILDE] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1093),
    [aux_sym__word_token1] = ACTIONS(1093),
    [aux_sym__word_token2] = ACTIONS(1093),
    [aux_sym__word_token3] = ACTIONS(1093),
    [sym__whitespace] = ACTIONS(1093),
    [sym__soft_line_ending] = ACTIONS(1093),
    [sym__block_quote_start] = ACTIONS(1093),
    [sym__indented_chunk_start] = ACTIONS(1093),
    [sym_atx_h1_marker] = ACTIONS(1093),
    [sym_atx_h2_marker] = ACTIONS(1093),
    [sym_atx_h3_marker] = ACTIONS(1093),
    [sym_atx_h4_marker] = ACTIONS(1093),
    [sym_atx_h5_marker] = ACTIONS(1093),
    [sym_atx_h6_marker] = ACTIONS(1093),
    [sym__thematic_break] = ACTIONS(1093),
    [sym__list_marker_minus] = ACTIONS(1093),
    [sym__list_marker_plus] = ACTIONS(33),
    [sym__list_marker_star] = ACTIONS(1093),
    [sym__list_marker_parenthesis] = ACTIONS(1093),
    [sym__list_marker_dot] = ACTIONS(1093),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(33),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1093),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1093),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1093),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1093),
    [sym__blank_line_start] = ACTIONS(1093),
    [sym__html_block_1_start] = ACTIONS(1093),
    [sym__html_block_2_start] = ACTIONS(1093),
    [sym__html_block_3_start] = ACTIONS(1093),
    [sym__html_block_4_start] = ACTIONS(1093),
    [sym__html_block_5_start] = ACTIONS(1093),
    [sym__html_block_6_start] = ACTIONS(1093),
    [sym__html_block_7_start] = ACTIONS(1093),
    [sym__pipe_table_start] = ACTIONS(1093),
  },
  [STATE(99)] = {
    [sym_list_marker_star] = STATE(2),
    [sym__list_item_star] = STATE(77),
    [aux_sym__list_star_repeat1] = STATE(77),
    [ts_builtin_sym_end] = ACTIONS(1097),
    [anon_sym_LBRACK] = ACTIONS(1095),
    [anon_sym_RBRACK] = ACTIONS(1097),
    [anon_sym_LT] = ACTIONS(1097),
    [anon_sym_GT] = ACTIONS(1097),
    [anon_sym_BANG] = ACTIONS(1097),
    [anon_sym_DQUOTE] = ACTIONS(1097),
    [anon_sym_POUND] = ACTIONS(1097),
    [anon_sym_DOLLAR] = ACTIONS(1097),
    [anon_sym_PERCENT] = ACTIONS(1097),
    [anon_sym_AMP] = ACTIONS(1097),
    [anon_sym_SQUOTE] = ACTIONS(1097),
    [anon_sym_STAR] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1097),
    [anon_sym_COMMA] = ACTIONS(1097),
    [anon_sym_DASH] = ACTIONS(1097),
    [anon_sym_DOT] = ACTIONS(1097),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_COLON] = ACTIONS(1097),
    [anon_sym_SEMI] = ACTIONS(1097),
    [anon_sym_EQ] = ACTIONS(1097),
    [anon_sym_QMARK] = ACTIONS(1097),
    [anon_sym_AT] = ACTIONS(1097),
    [anon_sym_BSLASH] = ACTIONS(1097),
    [anon_sym_CARET] = ACTIONS(1097),
    [anon_sym__] = ACTIONS(1097),
    [anon_sym_BQUOTE] = ACTIONS(1097),
    [anon_sym_LBRACE] = ACTIONS(1097),
    [anon_sym_PIPE] = ACTIONS(1097),
    [anon_sym_RBRACE] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1097),
    [anon_sym_LPAREN] = ACTIONS(1097),
    [anon_sym_RPAREN] = ACTIONS(1097),
    [aux_sym__word_token1] = ACTIONS(1097),
    [aux_sym__word_token2] = ACTIONS(1097),
    [aux_sym__word_token3] = ACTIONS(1097),
    [sym__whitespace] = ACTIONS(1097),
    [sym__soft_line_ending] = ACTIONS(1097),
    [sym__block_quote_start] = ACTIONS(1097),
    [sym__indented_chunk_start] = ACTIONS(1097),
    [sym_atx_h1_marker] = ACTIONS(1097),
    [sym_atx_h2_marker] = ACTIONS(1097),
    [sym_atx_h3_marker] = ACTIONS(1097),
    [sym_atx_h4_marker] = ACTIONS(1097),
    [sym_atx_h5_marker] = ACTIONS(1097),
    [sym_atx_h6_marker] = ACTIONS(1097),
    [sym__thematic_break] = ACTIONS(1097),
    [sym__list_marker_minus] = ACTIONS(1097),
    [sym__list_marker_plus] = ACTIONS(1097),
    [sym__list_marker_star] = ACTIONS(35),
    [sym__list_marker_parenthesis] = ACTIONS(1097),
    [sym__list_marker_dot] = ACTIONS(1097),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(35),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1097),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1097),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1097),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1097),
    [sym__blank_line_start] = ACTIONS(1097),
    [sym__html_block_1_start] = ACTIONS(1097),
    [sym__html_block_2_start] = ACTIONS(1097),
    [sym__html_block_3_start] = ACTIONS(1097),
    [sym__html_block_4_start] = ACTIONS(1097),
    [sym__html_block_5_start] = ACTIONS(1097),
    [sym__html_block_6_start] = ACTIONS(1097),
    [sym__html_block_7_start] = ACTIONS(1097),
    [sym__pipe_table_start] = ACTIONS(1097),
  },
  [STATE(100)] = {
    [sym__indented_chunk] = STATE(92),
    [sym__blank_line] = STATE(92),
    [aux_sym_indented_code_block_repeat1] = STATE(92),
    [anon_sym_LBRACK] = ACTIONS(1137),
    [anon_sym_RBRACK] = ACTIONS(1135),
    [anon_sym_LT] = ACTIONS(1135),
    [anon_sym_GT] = ACTIONS(1135),
    [anon_sym_BANG] = ACTIONS(1135),
    [anon_sym_DQUOTE] = ACTIONS(1135),
    [anon_sym_POUND] = ACTIONS(1135),
    [anon_sym_DOLLAR] = ACTIONS(1135),
    [anon_sym_PERCENT] = ACTIONS(1135),
    [anon_sym_AMP] = ACTIONS(1135),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_STAR] = ACTIONS(1135),
    [anon_sym_PLUS] = ACTIONS(1135),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_DASH] = ACTIONS(1135),
    [anon_sym_DOT] = ACTIONS(1135),
    [anon_sym_SLASH] = ACTIONS(1135),
    [anon_sym_COLON] = ACTIONS(1135),
    [anon_sym_SEMI] = ACTIONS(1135),
    [anon_sym_EQ] = ACTIONS(1135),
    [anon_sym_QMARK] = ACTIONS(1135),
    [anon_sym_AT] = ACTIONS(1135),
    [anon_sym_BSLASH] = ACTIONS(1135),
    [anon_sym_CARET] = ACTIONS(1135),
    [anon_sym__] = ACTIONS(1135),
    [anon_sym_BQUOTE] = ACTIONS(1135),
    [anon_sym_LBRACE] = ACTIONS(1135),
    [anon_sym_PIPE] = ACTIONS(1135),
    [anon_sym_RBRACE] = ACTIONS(1135),
    [anon_sym_TILDE] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1135),
    [anon_sym_RPAREN] = ACTIONS(1135),
    [aux_sym__word_token1] = ACTIONS(1135),
    [aux_sym__word_token2] = ACTIONS(1135),
    [aux_sym__word_token3] = ACTIONS(1135),
    [sym__whitespace] = ACTIONS(1135),
    [sym__soft_line_ending] = ACTIONS(1135),
    [sym__block_close] = ACTIONS(1135),
    [sym__block_quote_start] = ACTIONS(1135),
    [sym__indented_chunk_start] = ACTIONS(75),
    [sym_atx_h1_marker] = ACTIONS(1135),
    [sym_atx_h2_marker] = ACTIONS(1135),
    [sym_atx_h3_marker] = ACTIONS(1135),
    [sym_atx_h4_marker] = ACTIONS(1135),
    [sym_atx_h5_marker] = ACTIONS(1135),
    [sym_atx_h6_marker] = ACTIONS(1135),
    [sym__thematic_break] = ACTIONS(1135),
    [sym__list_marker_minus] = ACTIONS(1135),
    [sym__list_marker_plus] = ACTIONS(1135),
    [sym__list_marker_star] = ACTIONS(1135),
    [sym__list_marker_parenthesis] = ACTIONS(1135),
    [sym__list_marker_dot] = ACTIONS(1135),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1135),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1135),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1135),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1135),
    [sym__blank_line_start] = ACTIONS(95),
    [sym__html_block_1_start] = ACTIONS(1135),
    [sym__html_block_2_start] = ACTIONS(1135),
    [sym__html_block_3_start] = ACTIONS(1135),
    [sym__html_block_4_start] = ACTIONS(1135),
    [sym__html_block_5_start] = ACTIONS(1135),
    [sym__html_block_6_start] = ACTIONS(1135),
    [sym__html_block_7_start] = ACTIONS(1135),
    [sym__pipe_table_start] = ACTIONS(1135),
  },
  [STATE(101)] = {
    [sym_link_title] = STATE(731),
    [ts_builtin_sym_end] = ACTIONS(1145),
    [anon_sym_LBRACK] = ACTIONS(1147),
    [anon_sym_RBRACK] = ACTIONS(1145),
    [anon_sym_LT] = ACTIONS(1145),
    [anon_sym_GT] = ACTIONS(1145),
    [anon_sym_BANG] = ACTIONS(1145),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [anon_sym_POUND] = ACTIONS(1145),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [anon_sym_PERCENT] = ACTIONS(1145),
    [anon_sym_AMP] = ACTIONS(1145),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_STAR] = ACTIONS(1145),
    [anon_sym_PLUS] = ACTIONS(1145),
    [anon_sym_COMMA] = ACTIONS(1145),
    [anon_sym_DASH] = ACTIONS(1145),
    [anon_sym_DOT] = ACTIONS(1145),
    [anon_sym_SLASH] = ACTIONS(1145),
    [anon_sym_COLON] = ACTIONS(1145),
    [anon_sym_SEMI] = ACTIONS(1145),
    [anon_sym_EQ] = ACTIONS(1145),
    [anon_sym_QMARK] = ACTIONS(1145),
    [anon_sym_AT] = ACTIONS(1145),
    [anon_sym_BSLASH] = ACTIONS(1145),
    [anon_sym_CARET] = ACTIONS(1145),
    [anon_sym__] = ACTIONS(1145),
    [anon_sym_BQUOTE] = ACTIONS(1145),
    [anon_sym_LBRACE] = ACTIONS(1145),
    [anon_sym_PIPE] = ACTIONS(1145),
    [anon_sym_RBRACE] = ACTIONS(1145),
    [anon_sym_TILDE] = ACTIONS(1145),
    [anon_sym_LPAREN] = ACTIONS(1155),
    [anon_sym_RPAREN] = ACTIONS(1145),
    [aux_sym__word_token1] = ACTIONS(1145),
    [aux_sym__word_token2] = ACTIONS(1145),
    [aux_sym__word_token3] = ACTIONS(1145),
    [sym__whitespace] = ACTIONS(1158),
    [sym__soft_line_ending] = ACTIONS(1145),
    [sym__block_quote_start] = ACTIONS(1145),
    [sym__indented_chunk_start] = ACTIONS(1145),
    [sym_atx_h1_marker] = ACTIONS(1145),
    [sym_atx_h2_marker] = ACTIONS(1145),
    [sym_atx_h3_marker] = ACTIONS(1145),
    [sym_atx_h4_marker] = ACTIONS(1145),
    [sym_atx_h5_marker] = ACTIONS(1145),
    [sym_atx_h6_marker] = ACTIONS(1145),
    [sym__thematic_break] = ACTIONS(1145),
    [sym__list_marker_minus] = ACTIONS(1145),
    [sym__list_marker_plus] = ACTIONS(1145),
    [sym__list_marker_star] = ACTIONS(1145),
    [sym__list_marker_parenthesis] = ACTIONS(1145),
    [sym__list_marker_dot] = ACTIONS(1145),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1145),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1145),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1145),
    [sym__blank_line_start] = ACTIONS(1145),
    [sym__html_block_1_start] = ACTIONS(1145),
    [sym__html_block_2_start] = ACTIONS(1145),
    [sym__html_block_3_start] = ACTIONS(1145),
    [sym__html_block_4_start] = ACTIONS(1145),
    [sym__html_block_5_start] = ACTIONS(1145),
    [sym__html_block_6_start] = ACTIONS(1145),
    [sym__html_block_7_start] = ACTIONS(1145),
    [sym__no_indented_chunk] = ACTIONS(1161),
    [sym__pipe_table_start] = ACTIONS(1145),
  },
  [STATE(102)] = {
    [sym_link_title] = STATE(720),
    [anon_sym_LBRACK] = ACTIONS(1147),
    [anon_sym_RBRACK] = ACTIONS(1145),
    [anon_sym_LT] = ACTIONS(1145),
    [anon_sym_GT] = ACTIONS(1145),
    [anon_sym_BANG] = ACTIONS(1145),
    [anon_sym_DQUOTE] = ACTIONS(1149),
    [anon_sym_POUND] = ACTIONS(1145),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [anon_sym_PERCENT] = ACTIONS(1145),
    [anon_sym_AMP] = ACTIONS(1145),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_STAR] = ACTIONS(1145),
    [anon_sym_PLUS] = ACTIONS(1145),
    [anon_sym_COMMA] = ACTIONS(1145),
    [anon_sym_DASH] = ACTIONS(1145),
    [anon_sym_DOT] = ACTIONS(1145),
    [anon_sym_SLASH] = ACTIONS(1145),
    [anon_sym_COLON] = ACTIONS(1145),
    [anon_sym_SEMI] = ACTIONS(1145),
    [anon_sym_EQ] = ACTIONS(1145),
    [anon_sym_QMARK] = ACTIONS(1145),
    [anon_sym_AT] = ACTIONS(1145),
    [anon_sym_BSLASH] = ACTIONS(1145),
    [anon_sym_CARET] = ACTIONS(1145),
    [anon_sym__] = ACTIONS(1145),
    [anon_sym_BQUOTE] = ACTIONS(1145),
    [anon_sym_LBRACE] = ACTIONS(1145),
    [anon_sym_PIPE] = ACTIONS(1145),
    [anon_sym_RBRACE] = ACTIONS(1145),
    [anon_sym_TILDE] = ACTIONS(1145),
    [anon_sym_LPAREN] = ACTIONS(1155),
    [anon_sym_RPAREN] = ACTIONS(1145),
    [aux_sym__word_token1] = ACTIONS(1145),
    [aux_sym__word_token2] = ACTIONS(1145),
    [aux_sym__word_token3] = ACTIONS(1145),
    [sym__whitespace] = ACTIONS(1163),
    [sym__soft_line_ending] = ACTIONS(1145),
    [sym__block_close] = ACTIONS(1145),
    [sym__block_quote_start] = ACTIONS(1145),
    [sym__indented_chunk_start] = ACTIONS(1145),
    [sym_atx_h1_marker] = ACTIONS(1145),
    [sym_atx_h2_marker] = ACTIONS(1145),
    [sym_atx_h3_marker] = ACTIONS(1145),
    [sym_atx_h4_marker] = ACTIONS(1145),
    [sym_atx_h5_marker] = ACTIONS(1145),
    [sym_atx_h6_marker] = ACTIONS(1145),
    [sym__thematic_break] = ACTIONS(1145),
    [sym__list_marker_minus] = ACTIONS(1145),
    [sym__list_marker_plus] = ACTIONS(1145),
    [sym__list_marker_star] = ACTIONS(1145),
    [sym__list_marker_parenthesis] = ACTIONS(1145),
    [sym__list_marker_dot] = ACTIONS(1145),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1145),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1145),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1145),
    [sym__blank_line_start] = ACTIONS(1145),
    [sym__html_block_1_start] = ACTIONS(1145),
    [sym__html_block_2_start] = ACTIONS(1145),
    [sym__html_block_3_start] = ACTIONS(1145),
    [sym__html_block_4_start] = ACTIONS(1145),
    [sym__html_block_5_start] = ACTIONS(1145),
    [sym__html_block_6_start] = ACTIONS(1145),
    [sym__html_block_7_start] = ACTIONS(1145),
    [sym__no_indented_chunk] = ACTIONS(1166),
    [sym__pipe_table_start] = ACTIONS(1145),
  },
  [STATE(103)] = {
    [sym_link_title] = STATE(721),
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_RBRACK] = ACTIONS(1170),
    [anon_sym_LT] = ACTIONS(1170),
    [anon_sym_GT] = ACTIONS(1170),
    [anon_sym_BANG] = ACTIONS(1170),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [anon_sym_POUND] = ACTIONS(1170),
    [anon_sym_DOLLAR] = ACTIONS(1170),
    [anon_sym_PERCENT] = ACTIONS(1170),
    [anon_sym_AMP] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1175),
    [anon_sym_STAR] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_COMMA] = ACTIONS(1170),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_DOT] = ACTIONS(1170),
    [anon_sym_SLASH] = ACTIONS(1170),
    [anon_sym_COLON] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1170),
    [anon_sym_EQ] = ACTIONS(1170),
    [anon_sym_QMARK] = ACTIONS(1170),
    [anon_sym_AT] = ACTIONS(1170),
    [anon_sym_BSLASH] = ACTIONS(1170),
    [anon_sym_CARET] = ACTIONS(1170),
    [anon_sym__] = ACTIONS(1170),
    [anon_sym_BQUOTE] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1170),
    [anon_sym_PIPE] = ACTIONS(1170),
    [anon_sym_RBRACE] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1170),
    [anon_sym_LPAREN] = ACTIONS(1178),
    [anon_sym_RPAREN] = ACTIONS(1170),
    [aux_sym__word_token1] = ACTIONS(1170),
    [aux_sym__word_token2] = ACTIONS(1170),
    [aux_sym__word_token3] = ACTIONS(1170),
    [sym__whitespace] = ACTIONS(1181),
    [sym__soft_line_ending] = ACTIONS(1170),
    [sym__block_close] = ACTIONS(1170),
    [sym__block_quote_start] = ACTIONS(1170),
    [sym__indented_chunk_start] = ACTIONS(1170),
    [sym_atx_h1_marker] = ACTIONS(1170),
    [sym_atx_h2_marker] = ACTIONS(1170),
    [sym_atx_h3_marker] = ACTIONS(1170),
    [sym_atx_h4_marker] = ACTIONS(1170),
    [sym_atx_h5_marker] = ACTIONS(1170),
    [sym_atx_h6_marker] = ACTIONS(1170),
    [sym__thematic_break] = ACTIONS(1170),
    [sym__list_marker_minus] = ACTIONS(1170),
    [sym__list_marker_plus] = ACTIONS(1170),
    [sym__list_marker_star] = ACTIONS(1170),
    [sym__list_marker_parenthesis] = ACTIONS(1170),
    [sym__list_marker_dot] = ACTIONS(1170),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1170),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1170),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1170),
    [sym__blank_line_start] = ACTIONS(1170),
    [sym__html_block_1_start] = ACTIONS(1170),
    [sym__html_block_2_start] = ACTIONS(1170),
    [sym__html_block_3_start] = ACTIONS(1170),
    [sym__html_block_4_start] = ACTIONS(1170),
    [sym__html_block_5_start] = ACTIONS(1170),
    [sym__html_block_6_start] = ACTIONS(1170),
    [sym__html_block_7_start] = ACTIONS(1170),
    [sym__no_indented_chunk] = ACTIONS(1184),
    [sym__pipe_table_start] = ACTIONS(1170),
  },
  [STATE(104)] = {
    [sym_link_title] = STATE(722),
    [anon_sym_LBRACK] = ACTIONS(1186),
    [anon_sym_RBRACK] = ACTIONS(1188),
    [anon_sym_LT] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1190),
    [anon_sym_POUND] = ACTIONS(1188),
    [anon_sym_DOLLAR] = ACTIONS(1188),
    [anon_sym_PERCENT] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1193),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_PLUS] = ACTIONS(1188),
    [anon_sym_COMMA] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_SLASH] = ACTIONS(1188),
    [anon_sym_COLON] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym_EQ] = ACTIONS(1188),
    [anon_sym_QMARK] = ACTIONS(1188),
    [anon_sym_AT] = ACTIONS(1188),
    [anon_sym_BSLASH] = ACTIONS(1188),
    [anon_sym_CARET] = ACTIONS(1188),
    [anon_sym__] = ACTIONS(1188),
    [anon_sym_BQUOTE] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_PIPE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_LPAREN] = ACTIONS(1196),
    [anon_sym_RPAREN] = ACTIONS(1188),
    [aux_sym__word_token1] = ACTIONS(1188),
    [aux_sym__word_token2] = ACTIONS(1188),
    [aux_sym__word_token3] = ACTIONS(1188),
    [sym__whitespace] = ACTIONS(1199),
    [sym__soft_line_ending] = ACTIONS(1188),
    [sym__block_close] = ACTIONS(1188),
    [sym__block_quote_start] = ACTIONS(1188),
    [sym__indented_chunk_start] = ACTIONS(1188),
    [sym_atx_h1_marker] = ACTIONS(1188),
    [sym_atx_h2_marker] = ACTIONS(1188),
    [sym_atx_h3_marker] = ACTIONS(1188),
    [sym_atx_h4_marker] = ACTIONS(1188),
    [sym_atx_h5_marker] = ACTIONS(1188),
    [sym_atx_h6_marker] = ACTIONS(1188),
    [sym__thematic_break] = ACTIONS(1188),
    [sym__list_marker_minus] = ACTIONS(1188),
    [sym__list_marker_plus] = ACTIONS(1188),
    [sym__list_marker_star] = ACTIONS(1188),
    [sym__list_marker_parenthesis] = ACTIONS(1188),
    [sym__list_marker_dot] = ACTIONS(1188),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1188),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1188),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1188),
    [sym__blank_line_start] = ACTIONS(1188),
    [sym__html_block_1_start] = ACTIONS(1188),
    [sym__html_block_2_start] = ACTIONS(1188),
    [sym__html_block_3_start] = ACTIONS(1188),
    [sym__html_block_4_start] = ACTIONS(1188),
    [sym__html_block_5_start] = ACTIONS(1188),
    [sym__html_block_6_start] = ACTIONS(1188),
    [sym__html_block_7_start] = ACTIONS(1188),
    [sym__no_indented_chunk] = ACTIONS(1202),
    [sym__pipe_table_start] = ACTIONS(1188),
  },
  [STATE(105)] = {
    [sym_link_title] = STATE(723),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_RBRACK] = ACTIONS(1206),
    [anon_sym_LT] = ACTIONS(1206),
    [anon_sym_GT] = ACTIONS(1206),
    [anon_sym_BANG] = ACTIONS(1206),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [anon_sym_POUND] = ACTIONS(1206),
    [anon_sym_DOLLAR] = ACTIONS(1206),
    [anon_sym_PERCENT] = ACTIONS(1206),
    [anon_sym_AMP] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1211),
    [anon_sym_STAR] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_COMMA] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_SLASH] = ACTIONS(1206),
    [anon_sym_COLON] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1206),
    [anon_sym_EQ] = ACTIONS(1206),
    [anon_sym_QMARK] = ACTIONS(1206),
    [anon_sym_AT] = ACTIONS(1206),
    [anon_sym_BSLASH] = ACTIONS(1206),
    [anon_sym_CARET] = ACTIONS(1206),
    [anon_sym__] = ACTIONS(1206),
    [anon_sym_BQUOTE] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1206),
    [anon_sym_PIPE] = ACTIONS(1206),
    [anon_sym_RBRACE] = ACTIONS(1206),
    [anon_sym_TILDE] = ACTIONS(1206),
    [anon_sym_LPAREN] = ACTIONS(1214),
    [anon_sym_RPAREN] = ACTIONS(1206),
    [aux_sym__word_token1] = ACTIONS(1206),
    [aux_sym__word_token2] = ACTIONS(1206),
    [aux_sym__word_token3] = ACTIONS(1206),
    [sym__whitespace] = ACTIONS(1217),
    [sym__soft_line_ending] = ACTIONS(1206),
    [sym__block_close] = ACTIONS(1206),
    [sym__block_quote_start] = ACTIONS(1206),
    [sym__indented_chunk_start] = ACTIONS(1206),
    [sym_atx_h1_marker] = ACTIONS(1206),
    [sym_atx_h2_marker] = ACTIONS(1206),
    [sym_atx_h3_marker] = ACTIONS(1206),
    [sym_atx_h4_marker] = ACTIONS(1206),
    [sym_atx_h5_marker] = ACTIONS(1206),
    [sym_atx_h6_marker] = ACTIONS(1206),
    [sym__thematic_break] = ACTIONS(1206),
    [sym__list_marker_minus] = ACTIONS(1206),
    [sym__list_marker_plus] = ACTIONS(1206),
    [sym__list_marker_star] = ACTIONS(1206),
    [sym__list_marker_parenthesis] = ACTIONS(1206),
    [sym__list_marker_dot] = ACTIONS(1206),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1206),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1206),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1206),
    [sym__blank_line_start] = ACTIONS(1206),
    [sym__html_block_1_start] = ACTIONS(1206),
    [sym__html_block_2_start] = ACTIONS(1206),
    [sym__html_block_3_start] = ACTIONS(1206),
    [sym__html_block_4_start] = ACTIONS(1206),
    [sym__html_block_5_start] = ACTIONS(1206),
    [sym__html_block_6_start] = ACTIONS(1206),
    [sym__html_block_7_start] = ACTIONS(1206),
    [sym__no_indented_chunk] = ACTIONS(1220),
    [sym__pipe_table_start] = ACTIONS(1206),
  },
  [STATE(106)] = {
    [sym_link_title] = STATE(724),
    [anon_sym_LBRACK] = ACTIONS(1222),
    [anon_sym_RBRACK] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1224),
    [anon_sym_GT] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1226),
    [anon_sym_POUND] = ACTIONS(1224),
    [anon_sym_DOLLAR] = ACTIONS(1224),
    [anon_sym_PERCENT] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1229),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_PLUS] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_SLASH] = ACTIONS(1224),
    [anon_sym_COLON] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_EQ] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1224),
    [anon_sym_BSLASH] = ACTIONS(1224),
    [anon_sym_CARET] = ACTIONS(1224),
    [anon_sym__] = ACTIONS(1224),
    [anon_sym_BQUOTE] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_PIPE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1232),
    [anon_sym_RPAREN] = ACTIONS(1224),
    [aux_sym__word_token1] = ACTIONS(1224),
    [aux_sym__word_token2] = ACTIONS(1224),
    [aux_sym__word_token3] = ACTIONS(1224),
    [sym__whitespace] = ACTIONS(1235),
    [sym__soft_line_ending] = ACTIONS(1224),
    [sym__block_close] = ACTIONS(1224),
    [sym__block_quote_start] = ACTIONS(1224),
    [sym__indented_chunk_start] = ACTIONS(1224),
    [sym_atx_h1_marker] = ACTIONS(1224),
    [sym_atx_h2_marker] = ACTIONS(1224),
    [sym_atx_h3_marker] = ACTIONS(1224),
    [sym_atx_h4_marker] = ACTIONS(1224),
    [sym_atx_h5_marker] = ACTIONS(1224),
    [sym_atx_h6_marker] = ACTIONS(1224),
    [sym__thematic_break] = ACTIONS(1224),
    [sym__list_marker_minus] = ACTIONS(1224),
    [sym__list_marker_plus] = ACTIONS(1224),
    [sym__list_marker_star] = ACTIONS(1224),
    [sym__list_marker_parenthesis] = ACTIONS(1224),
    [sym__list_marker_dot] = ACTIONS(1224),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1224),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1224),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1224),
    [sym__blank_line_start] = ACTIONS(1224),
    [sym__html_block_1_start] = ACTIONS(1224),
    [sym__html_block_2_start] = ACTIONS(1224),
    [sym__html_block_3_start] = ACTIONS(1224),
    [sym__html_block_4_start] = ACTIONS(1224),
    [sym__html_block_5_start] = ACTIONS(1224),
    [sym__html_block_6_start] = ACTIONS(1224),
    [sym__html_block_7_start] = ACTIONS(1224),
    [sym__no_indented_chunk] = ACTIONS(1238),
    [sym__pipe_table_start] = ACTIONS(1224),
  },
  [STATE(107)] = {
    [ts_builtin_sym_end] = ACTIONS(1240),
    [anon_sym_LBRACK] = ACTIONS(1242),
    [anon_sym_RBRACK] = ACTIONS(1240),
    [anon_sym_LT] = ACTIONS(1240),
    [anon_sym_GT] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [anon_sym_POUND] = ACTIONS(1240),
    [anon_sym_DOLLAR] = ACTIONS(1240),
    [anon_sym_PERCENT] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_PLUS] = ACTIONS(1240),
    [anon_sym_COMMA] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_DOT] = ACTIONS(1240),
    [anon_sym_SLASH] = ACTIONS(1240),
    [anon_sym_COLON] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym_EQ] = ACTIONS(1240),
    [anon_sym_QMARK] = ACTIONS(1240),
    [anon_sym_AT] = ACTIONS(1240),
    [anon_sym_BSLASH] = ACTIONS(1240),
    [anon_sym_CARET] = ACTIONS(1240),
    [anon_sym__] = ACTIONS(1240),
    [anon_sym_BQUOTE] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_PIPE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_LPAREN] = ACTIONS(1240),
    [anon_sym_RPAREN] = ACTIONS(1240),
    [aux_sym__word_token1] = ACTIONS(1240),
    [aux_sym__word_token2] = ACTIONS(1240),
    [aux_sym__word_token3] = ACTIONS(1240),
    [sym__whitespace] = ACTIONS(1240),
    [sym__soft_line_ending] = ACTIONS(1240),
    [sym_block_continuation] = ACTIONS(1244),
    [sym__block_quote_start] = ACTIONS(1240),
    [sym__indented_chunk_start] = ACTIONS(1240),
    [sym_atx_h1_marker] = ACTIONS(1240),
    [sym_atx_h2_marker] = ACTIONS(1240),
    [sym_atx_h3_marker] = ACTIONS(1240),
    [sym_atx_h4_marker] = ACTIONS(1240),
    [sym_atx_h5_marker] = ACTIONS(1240),
    [sym_atx_h6_marker] = ACTIONS(1240),
    [sym__thematic_break] = ACTIONS(1240),
    [sym__list_marker_minus] = ACTIONS(1240),
    [sym__list_marker_plus] = ACTIONS(1240),
    [sym__list_marker_star] = ACTIONS(1240),
    [sym__list_marker_parenthesis] = ACTIONS(1240),
    [sym__list_marker_dot] = ACTIONS(1240),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1240),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1240),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1240),
    [sym__blank_line_start] = ACTIONS(1240),
    [sym__html_block_1_start] = ACTIONS(1240),
    [sym__html_block_2_start] = ACTIONS(1240),
    [sym__html_block_3_start] = ACTIONS(1240),
    [sym__html_block_4_start] = ACTIONS(1240),
    [sym__html_block_5_start] = ACTIONS(1240),
    [sym__html_block_6_start] = ACTIONS(1240),
    [sym__html_block_7_start] = ACTIONS(1240),
    [sym__no_indented_chunk] = ACTIONS(1240),
    [sym__pipe_table_start] = ACTIONS(1240),
  },
  [STATE(108)] = {
    [anon_sym_LBRACK] = ACTIONS(1242),
    [anon_sym_RBRACK] = ACTIONS(1240),
    [anon_sym_LT] = ACTIONS(1240),
    [anon_sym_GT] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [anon_sym_POUND] = ACTIONS(1240),
    [anon_sym_DOLLAR] = ACTIONS(1240),
    [anon_sym_PERCENT] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_PLUS] = ACTIONS(1240),
    [anon_sym_COMMA] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_DOT] = ACTIONS(1240),
    [anon_sym_SLASH] = ACTIONS(1240),
    [anon_sym_COLON] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym_EQ] = ACTIONS(1240),
    [anon_sym_QMARK] = ACTIONS(1240),
    [anon_sym_AT] = ACTIONS(1240),
    [anon_sym_BSLASH] = ACTIONS(1240),
    [anon_sym_CARET] = ACTIONS(1240),
    [anon_sym__] = ACTIONS(1240),
    [anon_sym_BQUOTE] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_PIPE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_LPAREN] = ACTIONS(1240),
    [anon_sym_RPAREN] = ACTIONS(1240),
    [aux_sym__word_token1] = ACTIONS(1240),
    [aux_sym__word_token2] = ACTIONS(1240),
    [aux_sym__word_token3] = ACTIONS(1240),
    [sym__whitespace] = ACTIONS(1240),
    [sym__soft_line_ending] = ACTIONS(1240),
    [sym__block_close] = ACTIONS(1240),
    [sym_block_continuation] = ACTIONS(1246),
    [sym__block_quote_start] = ACTIONS(1240),
    [sym__indented_chunk_start] = ACTIONS(1240),
    [sym_atx_h1_marker] = ACTIONS(1240),
    [sym_atx_h2_marker] = ACTIONS(1240),
    [sym_atx_h3_marker] = ACTIONS(1240),
    [sym_atx_h4_marker] = ACTIONS(1240),
    [sym_atx_h5_marker] = ACTIONS(1240),
    [sym_atx_h6_marker] = ACTIONS(1240),
    [sym__thematic_break] = ACTIONS(1240),
    [sym__list_marker_minus] = ACTIONS(1240),
    [sym__list_marker_plus] = ACTIONS(1240),
    [sym__list_marker_star] = ACTIONS(1240),
    [sym__list_marker_parenthesis] = ACTIONS(1240),
    [sym__list_marker_dot] = ACTIONS(1240),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1240),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1240),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1240),
    [sym__blank_line_start] = ACTIONS(1240),
    [sym__html_block_1_start] = ACTIONS(1240),
    [sym__html_block_2_start] = ACTIONS(1240),
    [sym__html_block_3_start] = ACTIONS(1240),
    [sym__html_block_4_start] = ACTIONS(1240),
    [sym__html_block_5_start] = ACTIONS(1240),
    [sym__html_block_6_start] = ACTIONS(1240),
    [sym__html_block_7_start] = ACTIONS(1240),
    [sym__no_indented_chunk] = ACTIONS(1240),
    [sym__pipe_table_start] = ACTIONS(1240),
  },
  [STATE(109)] = {
    [sym_link_title] = STATE(743),
    [ts_builtin_sym_end] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1222),
    [anon_sym_RBRACK] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1224),
    [anon_sym_GT] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1226),
    [anon_sym_POUND] = ACTIONS(1224),
    [anon_sym_DOLLAR] = ACTIONS(1224),
    [anon_sym_PERCENT] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1229),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_PLUS] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_SLASH] = ACTIONS(1224),
    [anon_sym_COLON] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_EQ] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1224),
    [anon_sym_BSLASH] = ACTIONS(1224),
    [anon_sym_CARET] = ACTIONS(1224),
    [anon_sym__] = ACTIONS(1224),
    [anon_sym_BQUOTE] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_PIPE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1232),
    [anon_sym_RPAREN] = ACTIONS(1224),
    [aux_sym__word_token1] = ACTIONS(1224),
    [aux_sym__word_token2] = ACTIONS(1224),
    [aux_sym__word_token3] = ACTIONS(1224),
    [sym__whitespace] = ACTIONS(1248),
    [sym__soft_line_ending] = ACTIONS(1224),
    [sym__block_quote_start] = ACTIONS(1224),
    [sym__indented_chunk_start] = ACTIONS(1224),
    [sym_atx_h1_marker] = ACTIONS(1224),
    [sym_atx_h2_marker] = ACTIONS(1224),
    [sym_atx_h3_marker] = ACTIONS(1224),
    [sym_atx_h4_marker] = ACTIONS(1224),
    [sym_atx_h5_marker] = ACTIONS(1224),
    [sym_atx_h6_marker] = ACTIONS(1224),
    [sym__thematic_break] = ACTIONS(1224),
    [sym__list_marker_minus] = ACTIONS(1224),
    [sym__list_marker_plus] = ACTIONS(1224),
    [sym__list_marker_star] = ACTIONS(1224),
    [sym__list_marker_parenthesis] = ACTIONS(1224),
    [sym__list_marker_dot] = ACTIONS(1224),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1224),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1224),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1224),
    [sym__blank_line_start] = ACTIONS(1224),
    [sym__html_block_1_start] = ACTIONS(1224),
    [sym__html_block_2_start] = ACTIONS(1224),
    [sym__html_block_3_start] = ACTIONS(1224),
    [sym__html_block_4_start] = ACTIONS(1224),
    [sym__html_block_5_start] = ACTIONS(1224),
    [sym__html_block_6_start] = ACTIONS(1224),
    [sym__html_block_7_start] = ACTIONS(1224),
    [sym__no_indented_chunk] = ACTIONS(1251),
    [sym__pipe_table_start] = ACTIONS(1224),
  },
  [STATE(110)] = {
    [ts_builtin_sym_end] = ACTIONS(1253),
    [anon_sym_LBRACK] = ACTIONS(1255),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_LT] = ACTIONS(1253),
    [anon_sym_GT] = ACTIONS(1253),
    [anon_sym_BANG] = ACTIONS(1253),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [anon_sym_POUND] = ACTIONS(1253),
    [anon_sym_DOLLAR] = ACTIONS(1253),
    [anon_sym_PERCENT] = ACTIONS(1253),
    [anon_sym_AMP] = ACTIONS(1253),
    [anon_sym_SQUOTE] = ACTIONS(1253),
    [anon_sym_STAR] = ACTIONS(1253),
    [anon_sym_PLUS] = ACTIONS(1253),
    [anon_sym_COMMA] = ACTIONS(1253),
    [anon_sym_DASH] = ACTIONS(1253),
    [anon_sym_DOT] = ACTIONS(1253),
    [anon_sym_SLASH] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1253),
    [anon_sym_SEMI] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1253),
    [anon_sym_AT] = ACTIONS(1253),
    [anon_sym_BSLASH] = ACTIONS(1253),
    [anon_sym_CARET] = ACTIONS(1253),
    [anon_sym__] = ACTIONS(1253),
    [anon_sym_BQUOTE] = ACTIONS(1253),
    [anon_sym_LBRACE] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1253),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_TILDE] = ACTIONS(1253),
    [anon_sym_LPAREN] = ACTIONS(1253),
    [anon_sym_RPAREN] = ACTIONS(1253),
    [aux_sym__word_token1] = ACTIONS(1253),
    [aux_sym__word_token2] = ACTIONS(1253),
    [aux_sym__word_token3] = ACTIONS(1253),
    [sym__whitespace] = ACTIONS(1253),
    [sym__soft_line_ending] = ACTIONS(1253),
    [sym__block_quote_start] = ACTIONS(1253),
    [sym__indented_chunk_start] = ACTIONS(1253),
    [sym_atx_h1_marker] = ACTIONS(1253),
    [sym_atx_h2_marker] = ACTIONS(1253),
    [sym_atx_h3_marker] = ACTIONS(1253),
    [sym_atx_h4_marker] = ACTIONS(1253),
    [sym_atx_h5_marker] = ACTIONS(1253),
    [sym_atx_h6_marker] = ACTIONS(1253),
    [sym_setext_h1_underline] = ACTIONS(1253),
    [sym_setext_h2_underline] = ACTIONS(1253),
    [sym__thematic_break] = ACTIONS(1253),
    [sym__list_marker_minus] = ACTIONS(1253),
    [sym__list_marker_plus] = ACTIONS(1253),
    [sym__list_marker_star] = ACTIONS(1253),
    [sym__list_marker_parenthesis] = ACTIONS(1253),
    [sym__list_marker_dot] = ACTIONS(1253),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1253),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1253),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1253),
    [sym__blank_line_start] = ACTIONS(1253),
    [sym__html_block_1_start] = ACTIONS(1253),
    [sym__html_block_2_start] = ACTIONS(1253),
    [sym__html_block_3_start] = ACTIONS(1253),
    [sym__html_block_4_start] = ACTIONS(1253),
    [sym__html_block_5_start] = ACTIONS(1253),
    [sym__html_block_6_start] = ACTIONS(1253),
    [sym__html_block_7_start] = ACTIONS(1253),
    [sym__pipe_table_start] = ACTIONS(1253),
  },
  [STATE(111)] = {
    [anon_sym_LBRACK] = ACTIONS(1257),
    [anon_sym_RBRACK] = ACTIONS(1259),
    [anon_sym_LT] = ACTIONS(1259),
    [anon_sym_GT] = ACTIONS(1259),
    [anon_sym_BANG] = ACTIONS(1259),
    [anon_sym_DQUOTE] = ACTIONS(1259),
    [anon_sym_POUND] = ACTIONS(1259),
    [anon_sym_DOLLAR] = ACTIONS(1259),
    [anon_sym_PERCENT] = ACTIONS(1259),
    [anon_sym_AMP] = ACTIONS(1259),
    [anon_sym_SQUOTE] = ACTIONS(1259),
    [anon_sym_STAR] = ACTIONS(1259),
    [anon_sym_PLUS] = ACTIONS(1259),
    [anon_sym_COMMA] = ACTIONS(1259),
    [anon_sym_DASH] = ACTIONS(1259),
    [anon_sym_DOT] = ACTIONS(1259),
    [anon_sym_SLASH] = ACTIONS(1259),
    [anon_sym_COLON] = ACTIONS(1259),
    [anon_sym_SEMI] = ACTIONS(1259),
    [anon_sym_EQ] = ACTIONS(1259),
    [anon_sym_QMARK] = ACTIONS(1259),
    [anon_sym_AT] = ACTIONS(1259),
    [anon_sym_BSLASH] = ACTIONS(1259),
    [anon_sym_CARET] = ACTIONS(1259),
    [anon_sym__] = ACTIONS(1259),
    [anon_sym_BQUOTE] = ACTIONS(1259),
    [anon_sym_LBRACE] = ACTIONS(1259),
    [anon_sym_PIPE] = ACTIONS(1259),
    [anon_sym_RBRACE] = ACTIONS(1259),
    [anon_sym_TILDE] = ACTIONS(1259),
    [anon_sym_LPAREN] = ACTIONS(1259),
    [anon_sym_RPAREN] = ACTIONS(1259),
    [aux_sym__word_token1] = ACTIONS(1259),
    [aux_sym__word_token2] = ACTIONS(1259),
    [aux_sym__word_token3] = ACTIONS(1259),
    [sym__whitespace] = ACTIONS(1259),
    [sym__soft_line_ending] = ACTIONS(1259),
    [sym__block_close] = ACTIONS(1259),
    [sym__block_quote_start] = ACTIONS(1259),
    [sym__indented_chunk_start] = ACTIONS(1259),
    [sym_atx_h1_marker] = ACTIONS(1259),
    [sym_atx_h2_marker] = ACTIONS(1259),
    [sym_atx_h3_marker] = ACTIONS(1259),
    [sym_atx_h4_marker] = ACTIONS(1259),
    [sym_atx_h5_marker] = ACTIONS(1259),
    [sym_atx_h6_marker] = ACTIONS(1259),
    [sym_setext_h1_underline] = ACTIONS(1259),
    [sym_setext_h2_underline] = ACTIONS(1259),
    [sym__thematic_break] = ACTIONS(1259),
    [sym__list_marker_minus] = ACTIONS(1259),
    [sym__list_marker_plus] = ACTIONS(1259),
    [sym__list_marker_star] = ACTIONS(1259),
    [sym__list_marker_parenthesis] = ACTIONS(1259),
    [sym__list_marker_dot] = ACTIONS(1259),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1259),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1259),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1259),
    [sym__blank_line_start] = ACTIONS(1259),
    [sym__html_block_1_start] = ACTIONS(1259),
    [sym__html_block_2_start] = ACTIONS(1259),
    [sym__html_block_3_start] = ACTIONS(1259),
    [sym__html_block_4_start] = ACTIONS(1259),
    [sym__html_block_5_start] = ACTIONS(1259),
    [sym__html_block_6_start] = ACTIONS(1259),
    [sym__html_block_7_start] = ACTIONS(1259),
    [sym__pipe_table_start] = ACTIONS(1259),
  },
  [STATE(112)] = {
    [anon_sym_LBRACK] = ACTIONS(1255),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_LT] = ACTIONS(1253),
    [anon_sym_GT] = ACTIONS(1253),
    [anon_sym_BANG] = ACTIONS(1253),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [anon_sym_POUND] = ACTIONS(1253),
    [anon_sym_DOLLAR] = ACTIONS(1253),
    [anon_sym_PERCENT] = ACTIONS(1253),
    [anon_sym_AMP] = ACTIONS(1253),
    [anon_sym_SQUOTE] = ACTIONS(1253),
    [anon_sym_STAR] = ACTIONS(1253),
    [anon_sym_PLUS] = ACTIONS(1253),
    [anon_sym_COMMA] = ACTIONS(1253),
    [anon_sym_DASH] = ACTIONS(1253),
    [anon_sym_DOT] = ACTIONS(1253),
    [anon_sym_SLASH] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1253),
    [anon_sym_SEMI] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1253),
    [anon_sym_AT] = ACTIONS(1253),
    [anon_sym_BSLASH] = ACTIONS(1253),
    [anon_sym_CARET] = ACTIONS(1253),
    [anon_sym__] = ACTIONS(1253),
    [anon_sym_BQUOTE] = ACTIONS(1253),
    [anon_sym_LBRACE] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1253),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_TILDE] = ACTIONS(1253),
    [anon_sym_LPAREN] = ACTIONS(1253),
    [anon_sym_RPAREN] = ACTIONS(1253),
    [aux_sym__word_token1] = ACTIONS(1253),
    [aux_sym__word_token2] = ACTIONS(1253),
    [aux_sym__word_token3] = ACTIONS(1253),
    [sym__whitespace] = ACTIONS(1253),
    [sym__soft_line_ending] = ACTIONS(1253),
    [sym__block_close] = ACTIONS(1253),
    [sym__block_quote_start] = ACTIONS(1253),
    [sym__indented_chunk_start] = ACTIONS(1253),
    [sym_atx_h1_marker] = ACTIONS(1253),
    [sym_atx_h2_marker] = ACTIONS(1253),
    [sym_atx_h3_marker] = ACTIONS(1253),
    [sym_atx_h4_marker] = ACTIONS(1253),
    [sym_atx_h5_marker] = ACTIONS(1253),
    [sym_atx_h6_marker] = ACTIONS(1253),
    [sym_setext_h1_underline] = ACTIONS(1253),
    [sym_setext_h2_underline] = ACTIONS(1253),
    [sym__thematic_break] = ACTIONS(1253),
    [sym__list_marker_minus] = ACTIONS(1253),
    [sym__list_marker_plus] = ACTIONS(1253),
    [sym__list_marker_star] = ACTIONS(1253),
    [sym__list_marker_parenthesis] = ACTIONS(1253),
    [sym__list_marker_dot] = ACTIONS(1253),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1253),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1253),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1253),
    [sym__blank_line_start] = ACTIONS(1253),
    [sym__html_block_1_start] = ACTIONS(1253),
    [sym__html_block_2_start] = ACTIONS(1253),
    [sym__html_block_3_start] = ACTIONS(1253),
    [sym__html_block_4_start] = ACTIONS(1253),
    [sym__html_block_5_start] = ACTIONS(1253),
    [sym__html_block_6_start] = ACTIONS(1253),
    [sym__html_block_7_start] = ACTIONS(1253),
    [sym__pipe_table_start] = ACTIONS(1253),
  },
  [STATE(113)] = {
    [ts_builtin_sym_end] = ACTIONS(1259),
    [anon_sym_LBRACK] = ACTIONS(1257),
    [anon_sym_RBRACK] = ACTIONS(1259),
    [anon_sym_LT] = ACTIONS(1259),
    [anon_sym_GT] = ACTIONS(1259),
    [anon_sym_BANG] = ACTIONS(1259),
    [anon_sym_DQUOTE] = ACTIONS(1259),
    [anon_sym_POUND] = ACTIONS(1259),
    [anon_sym_DOLLAR] = ACTIONS(1259),
    [anon_sym_PERCENT] = ACTIONS(1259),
    [anon_sym_AMP] = ACTIONS(1259),
    [anon_sym_SQUOTE] = ACTIONS(1259),
    [anon_sym_STAR] = ACTIONS(1259),
    [anon_sym_PLUS] = ACTIONS(1259),
    [anon_sym_COMMA] = ACTIONS(1259),
    [anon_sym_DASH] = ACTIONS(1259),
    [anon_sym_DOT] = ACTIONS(1259),
    [anon_sym_SLASH] = ACTIONS(1259),
    [anon_sym_COLON] = ACTIONS(1259),
    [anon_sym_SEMI] = ACTIONS(1259),
    [anon_sym_EQ] = ACTIONS(1259),
    [anon_sym_QMARK] = ACTIONS(1259),
    [anon_sym_AT] = ACTIONS(1259),
    [anon_sym_BSLASH] = ACTIONS(1259),
    [anon_sym_CARET] = ACTIONS(1259),
    [anon_sym__] = ACTIONS(1259),
    [anon_sym_BQUOTE] = ACTIONS(1259),
    [anon_sym_LBRACE] = ACTIONS(1259),
    [anon_sym_PIPE] = ACTIONS(1259),
    [anon_sym_RBRACE] = ACTIONS(1259),
    [anon_sym_TILDE] = ACTIONS(1259),
    [anon_sym_LPAREN] = ACTIONS(1259),
    [anon_sym_RPAREN] = ACTIONS(1259),
    [aux_sym__word_token1] = ACTIONS(1259),
    [aux_sym__word_token2] = ACTIONS(1259),
    [aux_sym__word_token3] = ACTIONS(1259),
    [sym__whitespace] = ACTIONS(1259),
    [sym__soft_line_ending] = ACTIONS(1259),
    [sym__block_quote_start] = ACTIONS(1259),
    [sym__indented_chunk_start] = ACTIONS(1259),
    [sym_atx_h1_marker] = ACTIONS(1259),
    [sym_atx_h2_marker] = ACTIONS(1259),
    [sym_atx_h3_marker] = ACTIONS(1259),
    [sym_atx_h4_marker] = ACTIONS(1259),
    [sym_atx_h5_marker] = ACTIONS(1259),
    [sym_atx_h6_marker] = ACTIONS(1259),
    [sym_setext_h1_underline] = ACTIONS(1259),
    [sym_setext_h2_underline] = ACTIONS(1259),
    [sym__thematic_break] = ACTIONS(1259),
    [sym__list_marker_minus] = ACTIONS(1259),
    [sym__list_marker_plus] = ACTIONS(1259),
    [sym__list_marker_star] = ACTIONS(1259),
    [sym__list_marker_parenthesis] = ACTIONS(1259),
    [sym__list_marker_dot] = ACTIONS(1259),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1259),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1259),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1259),
    [sym__blank_line_start] = ACTIONS(1259),
    [sym__html_block_1_start] = ACTIONS(1259),
    [sym__html_block_2_start] = ACTIONS(1259),
    [sym__html_block_3_start] = ACTIONS(1259),
    [sym__html_block_4_start] = ACTIONS(1259),
    [sym__html_block_5_start] = ACTIONS(1259),
    [sym__html_block_6_start] = ACTIONS(1259),
    [sym__html_block_7_start] = ACTIONS(1259),
    [sym__pipe_table_start] = ACTIONS(1259),
  },
  [STATE(114)] = {
    [ts_builtin_sym_end] = ACTIONS(1261),
    [anon_sym_LBRACK] = ACTIONS(1263),
    [anon_sym_RBRACK] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_BANG] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1261),
    [anon_sym_POUND] = ACTIONS(1261),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_SQUOTE] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_COMMA] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_DOT] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_COLON] = ACTIONS(1261),
    [anon_sym_SEMI] = ACTIONS(1261),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_AT] = ACTIONS(1261),
    [anon_sym_BSLASH] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym__] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_LBRACE] = ACTIONS(1261),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_RBRACE] = ACTIONS(1261),
    [anon_sym_TILDE] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_RPAREN] = ACTIONS(1261),
    [aux_sym__word_token1] = ACTIONS(1261),
    [aux_sym__word_token2] = ACTIONS(1261),
    [aux_sym__word_token3] = ACTIONS(1261),
    [sym__whitespace] = ACTIONS(1261),
    [sym__soft_line_ending] = ACTIONS(1261),
    [sym__block_quote_start] = ACTIONS(1261),
    [sym__indented_chunk_start] = ACTIONS(1261),
    [sym_atx_h1_marker] = ACTIONS(1261),
    [sym_atx_h2_marker] = ACTIONS(1261),
    [sym_atx_h3_marker] = ACTIONS(1261),
    [sym_atx_h4_marker] = ACTIONS(1261),
    [sym_atx_h5_marker] = ACTIONS(1261),
    [sym_atx_h6_marker] = ACTIONS(1261),
    [sym_setext_h1_underline] = ACTIONS(1265),
    [sym_setext_h2_underline] = ACTIONS(1267),
    [sym__thematic_break] = ACTIONS(1261),
    [sym__list_marker_minus] = ACTIONS(1261),
    [sym__list_marker_plus] = ACTIONS(1261),
    [sym__list_marker_star] = ACTIONS(1261),
    [sym__list_marker_parenthesis] = ACTIONS(1261),
    [sym__list_marker_dot] = ACTIONS(1261),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1261),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1261),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1261),
    [sym__blank_line_start] = ACTIONS(1261),
    [sym__html_block_1_start] = ACTIONS(1261),
    [sym__html_block_2_start] = ACTIONS(1261),
    [sym__html_block_3_start] = ACTIONS(1261),
    [sym__html_block_4_start] = ACTIONS(1261),
    [sym__html_block_5_start] = ACTIONS(1261),
    [sym__html_block_6_start] = ACTIONS(1261),
    [sym__html_block_7_start] = ACTIONS(1261),
    [sym__pipe_table_start] = ACTIONS(1261),
  },
  [STATE(115)] = {
    [sym_link_title] = STATE(703),
    [ts_builtin_sym_end] = ACTIONS(1170),
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_RBRACK] = ACTIONS(1170),
    [anon_sym_LT] = ACTIONS(1170),
    [anon_sym_GT] = ACTIONS(1170),
    [anon_sym_BANG] = ACTIONS(1170),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [anon_sym_POUND] = ACTIONS(1170),
    [anon_sym_DOLLAR] = ACTIONS(1170),
    [anon_sym_PERCENT] = ACTIONS(1170),
    [anon_sym_AMP] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1175),
    [anon_sym_STAR] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_COMMA] = ACTIONS(1170),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_DOT] = ACTIONS(1170),
    [anon_sym_SLASH] = ACTIONS(1170),
    [anon_sym_COLON] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1170),
    [anon_sym_EQ] = ACTIONS(1170),
    [anon_sym_QMARK] = ACTIONS(1170),
    [anon_sym_AT] = ACTIONS(1170),
    [anon_sym_BSLASH] = ACTIONS(1170),
    [anon_sym_CARET] = ACTIONS(1170),
    [anon_sym__] = ACTIONS(1170),
    [anon_sym_BQUOTE] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1170),
    [anon_sym_PIPE] = ACTIONS(1170),
    [anon_sym_RBRACE] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1170),
    [anon_sym_LPAREN] = ACTIONS(1178),
    [anon_sym_RPAREN] = ACTIONS(1170),
    [aux_sym__word_token1] = ACTIONS(1170),
    [aux_sym__word_token2] = ACTIONS(1170),
    [aux_sym__word_token3] = ACTIONS(1170),
    [sym__whitespace] = ACTIONS(1269),
    [sym__soft_line_ending] = ACTIONS(1170),
    [sym__block_quote_start] = ACTIONS(1170),
    [sym__indented_chunk_start] = ACTIONS(1170),
    [sym_atx_h1_marker] = ACTIONS(1170),
    [sym_atx_h2_marker] = ACTIONS(1170),
    [sym_atx_h3_marker] = ACTIONS(1170),
    [sym_atx_h4_marker] = ACTIONS(1170),
    [sym_atx_h5_marker] = ACTIONS(1170),
    [sym_atx_h6_marker] = ACTIONS(1170),
    [sym__thematic_break] = ACTIONS(1170),
    [sym__list_marker_minus] = ACTIONS(1170),
    [sym__list_marker_plus] = ACTIONS(1170),
    [sym__list_marker_star] = ACTIONS(1170),
    [sym__list_marker_parenthesis] = ACTIONS(1170),
    [sym__list_marker_dot] = ACTIONS(1170),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1170),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1170),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1170),
    [sym__blank_line_start] = ACTIONS(1170),
    [sym__html_block_1_start] = ACTIONS(1170),
    [sym__html_block_2_start] = ACTIONS(1170),
    [sym__html_block_3_start] = ACTIONS(1170),
    [sym__html_block_4_start] = ACTIONS(1170),
    [sym__html_block_5_start] = ACTIONS(1170),
    [sym__html_block_6_start] = ACTIONS(1170),
    [sym__html_block_7_start] = ACTIONS(1170),
    [sym__no_indented_chunk] = ACTIONS(1272),
    [sym__pipe_table_start] = ACTIONS(1170),
  },
  [STATE(116)] = {
    [sym_link_title] = STATE(697),
    [ts_builtin_sym_end] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(1186),
    [anon_sym_RBRACK] = ACTIONS(1188),
    [anon_sym_LT] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1190),
    [anon_sym_POUND] = ACTIONS(1188),
    [anon_sym_DOLLAR] = ACTIONS(1188),
    [anon_sym_PERCENT] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1193),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_PLUS] = ACTIONS(1188),
    [anon_sym_COMMA] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_SLASH] = ACTIONS(1188),
    [anon_sym_COLON] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym_EQ] = ACTIONS(1188),
    [anon_sym_QMARK] = ACTIONS(1188),
    [anon_sym_AT] = ACTIONS(1188),
    [anon_sym_BSLASH] = ACTIONS(1188),
    [anon_sym_CARET] = ACTIONS(1188),
    [anon_sym__] = ACTIONS(1188),
    [anon_sym_BQUOTE] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_PIPE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_LPAREN] = ACTIONS(1196),
    [anon_sym_RPAREN] = ACTIONS(1188),
    [aux_sym__word_token1] = ACTIONS(1188),
    [aux_sym__word_token2] = ACTIONS(1188),
    [aux_sym__word_token3] = ACTIONS(1188),
    [sym__whitespace] = ACTIONS(1274),
    [sym__soft_line_ending] = ACTIONS(1188),
    [sym__block_quote_start] = ACTIONS(1188),
    [sym__indented_chunk_start] = ACTIONS(1188),
    [sym_atx_h1_marker] = ACTIONS(1188),
    [sym_atx_h2_marker] = ACTIONS(1188),
    [sym_atx_h3_marker] = ACTIONS(1188),
    [sym_atx_h4_marker] = ACTIONS(1188),
    [sym_atx_h5_marker] = ACTIONS(1188),
    [sym_atx_h6_marker] = ACTIONS(1188),
    [sym__thematic_break] = ACTIONS(1188),
    [sym__list_marker_minus] = ACTIONS(1188),
    [sym__list_marker_plus] = ACTIONS(1188),
    [sym__list_marker_star] = ACTIONS(1188),
    [sym__list_marker_parenthesis] = ACTIONS(1188),
    [sym__list_marker_dot] = ACTIONS(1188),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1188),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1188),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1188),
    [sym__blank_line_start] = ACTIONS(1188),
    [sym__html_block_1_start] = ACTIONS(1188),
    [sym__html_block_2_start] = ACTIONS(1188),
    [sym__html_block_3_start] = ACTIONS(1188),
    [sym__html_block_4_start] = ACTIONS(1188),
    [sym__html_block_5_start] = ACTIONS(1188),
    [sym__html_block_6_start] = ACTIONS(1188),
    [sym__html_block_7_start] = ACTIONS(1188),
    [sym__no_indented_chunk] = ACTIONS(1277),
    [sym__pipe_table_start] = ACTIONS(1188),
  },
  [STATE(117)] = {
    [sym_link_title] = STATE(706),
    [ts_builtin_sym_end] = ACTIONS(1206),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_RBRACK] = ACTIONS(1206),
    [anon_sym_LT] = ACTIONS(1206),
    [anon_sym_GT] = ACTIONS(1206),
    [anon_sym_BANG] = ACTIONS(1206),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [anon_sym_POUND] = ACTIONS(1206),
    [anon_sym_DOLLAR] = ACTIONS(1206),
    [anon_sym_PERCENT] = ACTIONS(1206),
    [anon_sym_AMP] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1211),
    [anon_sym_STAR] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_COMMA] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_SLASH] = ACTIONS(1206),
    [anon_sym_COLON] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1206),
    [anon_sym_EQ] = ACTIONS(1206),
    [anon_sym_QMARK] = ACTIONS(1206),
    [anon_sym_AT] = ACTIONS(1206),
    [anon_sym_BSLASH] = ACTIONS(1206),
    [anon_sym_CARET] = ACTIONS(1206),
    [anon_sym__] = ACTIONS(1206),
    [anon_sym_BQUOTE] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1206),
    [anon_sym_PIPE] = ACTIONS(1206),
    [anon_sym_RBRACE] = ACTIONS(1206),
    [anon_sym_TILDE] = ACTIONS(1206),
    [anon_sym_LPAREN] = ACTIONS(1214),
    [anon_sym_RPAREN] = ACTIONS(1206),
    [aux_sym__word_token1] = ACTIONS(1206),
    [aux_sym__word_token2] = ACTIONS(1206),
    [aux_sym__word_token3] = ACTIONS(1206),
    [sym__whitespace] = ACTIONS(1279),
    [sym__soft_line_ending] = ACTIONS(1206),
    [sym__block_quote_start] = ACTIONS(1206),
    [sym__indented_chunk_start] = ACTIONS(1206),
    [sym_atx_h1_marker] = ACTIONS(1206),
    [sym_atx_h2_marker] = ACTIONS(1206),
    [sym_atx_h3_marker] = ACTIONS(1206),
    [sym_atx_h4_marker] = ACTIONS(1206),
    [sym_atx_h5_marker] = ACTIONS(1206),
    [sym_atx_h6_marker] = ACTIONS(1206),
    [sym__thematic_break] = ACTIONS(1206),
    [sym__list_marker_minus] = ACTIONS(1206),
    [sym__list_marker_plus] = ACTIONS(1206),
    [sym__list_marker_star] = ACTIONS(1206),
    [sym__list_marker_parenthesis] = ACTIONS(1206),
    [sym__list_marker_dot] = ACTIONS(1206),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1206),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1206),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1206),
    [sym__blank_line_start] = ACTIONS(1206),
    [sym__html_block_1_start] = ACTIONS(1206),
    [sym__html_block_2_start] = ACTIONS(1206),
    [sym__html_block_3_start] = ACTIONS(1206),
    [sym__html_block_4_start] = ACTIONS(1206),
    [sym__html_block_5_start] = ACTIONS(1206),
    [sym__html_block_6_start] = ACTIONS(1206),
    [sym__html_block_7_start] = ACTIONS(1206),
    [sym__no_indented_chunk] = ACTIONS(1282),
    [sym__pipe_table_start] = ACTIONS(1206),
  },
  [STATE(118)] = {
    [anon_sym_LBRACK] = ACTIONS(1263),
    [anon_sym_RBRACK] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_BANG] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1261),
    [anon_sym_POUND] = ACTIONS(1261),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_SQUOTE] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_COMMA] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_DOT] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_COLON] = ACTIONS(1261),
    [anon_sym_SEMI] = ACTIONS(1261),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_AT] = ACTIONS(1261),
    [anon_sym_BSLASH] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym__] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_LBRACE] = ACTIONS(1261),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_RBRACE] = ACTIONS(1261),
    [anon_sym_TILDE] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_RPAREN] = ACTIONS(1261),
    [aux_sym__word_token1] = ACTIONS(1261),
    [aux_sym__word_token2] = ACTIONS(1261),
    [aux_sym__word_token3] = ACTIONS(1261),
    [sym__whitespace] = ACTIONS(1261),
    [sym__soft_line_ending] = ACTIONS(1261),
    [sym__block_close] = ACTIONS(1261),
    [sym__block_quote_start] = ACTIONS(1261),
    [sym__indented_chunk_start] = ACTIONS(1261),
    [sym_atx_h1_marker] = ACTIONS(1261),
    [sym_atx_h2_marker] = ACTIONS(1261),
    [sym_atx_h3_marker] = ACTIONS(1261),
    [sym_atx_h4_marker] = ACTIONS(1261),
    [sym_atx_h5_marker] = ACTIONS(1261),
    [sym_atx_h6_marker] = ACTIONS(1261),
    [sym_setext_h1_underline] = ACTIONS(1284),
    [sym_setext_h2_underline] = ACTIONS(1286),
    [sym__thematic_break] = ACTIONS(1261),
    [sym__list_marker_minus] = ACTIONS(1261),
    [sym__list_marker_plus] = ACTIONS(1261),
    [sym__list_marker_star] = ACTIONS(1261),
    [sym__list_marker_parenthesis] = ACTIONS(1261),
    [sym__list_marker_dot] = ACTIONS(1261),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1261),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1261),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1261),
    [sym__blank_line_start] = ACTIONS(1261),
    [sym__html_block_1_start] = ACTIONS(1261),
    [sym__html_block_2_start] = ACTIONS(1261),
    [sym__html_block_3_start] = ACTIONS(1261),
    [sym__html_block_4_start] = ACTIONS(1261),
    [sym__html_block_5_start] = ACTIONS(1261),
    [sym__html_block_6_start] = ACTIONS(1261),
    [sym__html_block_7_start] = ACTIONS(1261),
    [sym__pipe_table_start] = ACTIONS(1261),
  },
  [STATE(119)] = {
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_RBRACK] = ACTIONS(1290),
    [anon_sym_LT] = ACTIONS(1290),
    [anon_sym_GT] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [anon_sym_DQUOTE] = ACTIONS(1290),
    [anon_sym_POUND] = ACTIONS(1290),
    [anon_sym_DOLLAR] = ACTIONS(1290),
    [anon_sym_PERCENT] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_SLASH] = ACTIONS(1290),
    [anon_sym_COLON] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_EQ] = ACTIONS(1290),
    [anon_sym_QMARK] = ACTIONS(1290),
    [anon_sym_AT] = ACTIONS(1290),
    [anon_sym_BSLASH] = ACTIONS(1290),
    [anon_sym_CARET] = ACTIONS(1290),
    [anon_sym__] = ACTIONS(1290),
    [anon_sym_BQUOTE] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_PIPE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_RPAREN] = ACTIONS(1290),
    [aux_sym__word_token1] = ACTIONS(1290),
    [aux_sym__word_token2] = ACTIONS(1290),
    [aux_sym__word_token3] = ACTIONS(1290),
    [sym__whitespace] = ACTIONS(1290),
    [sym__soft_line_ending] = ACTIONS(1290),
    [sym__block_close] = ACTIONS(1290),
    [sym_block_continuation] = ACTIONS(1292),
    [sym__block_quote_start] = ACTIONS(1290),
    [sym__indented_chunk_start] = ACTIONS(1290),
    [sym_atx_h1_marker] = ACTIONS(1290),
    [sym_atx_h2_marker] = ACTIONS(1290),
    [sym_atx_h3_marker] = ACTIONS(1290),
    [sym_atx_h4_marker] = ACTIONS(1290),
    [sym_atx_h5_marker] = ACTIONS(1290),
    [sym_atx_h6_marker] = ACTIONS(1290),
    [sym__thematic_break] = ACTIONS(1290),
    [sym__list_marker_minus] = ACTIONS(1290),
    [sym__list_marker_plus] = ACTIONS(1290),
    [sym__list_marker_star] = ACTIONS(1290),
    [sym__list_marker_parenthesis] = ACTIONS(1290),
    [sym__list_marker_dot] = ACTIONS(1290),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1290),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1290),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1290),
    [sym__blank_line_start] = ACTIONS(1290),
    [sym__html_block_1_start] = ACTIONS(1290),
    [sym__html_block_2_start] = ACTIONS(1290),
    [sym__html_block_3_start] = ACTIONS(1290),
    [sym__html_block_4_start] = ACTIONS(1290),
    [sym__html_block_5_start] = ACTIONS(1290),
    [sym__html_block_6_start] = ACTIONS(1290),
    [sym__html_block_7_start] = ACTIONS(1290),
    [sym__pipe_table_start] = ACTIONS(1290),
  },
  [STATE(120)] = {
    [ts_builtin_sym_end] = ACTIONS(1294),
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_RBRACK] = ACTIONS(1294),
    [anon_sym_LT] = ACTIONS(1294),
    [anon_sym_GT] = ACTIONS(1294),
    [anon_sym_BANG] = ACTIONS(1294),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [anon_sym_POUND] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1294),
    [anon_sym_PERCENT] = ACTIONS(1294),
    [anon_sym_AMP] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_COMMA] = ACTIONS(1294),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_DOT] = ACTIONS(1294),
    [anon_sym_SLASH] = ACTIONS(1294),
    [anon_sym_COLON] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1294),
    [anon_sym_EQ] = ACTIONS(1294),
    [anon_sym_QMARK] = ACTIONS(1294),
    [anon_sym_AT] = ACTIONS(1294),
    [anon_sym_BSLASH] = ACTIONS(1294),
    [anon_sym_CARET] = ACTIONS(1294),
    [anon_sym__] = ACTIONS(1294),
    [anon_sym_BQUOTE] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1294),
    [anon_sym_PIPE] = ACTIONS(1294),
    [anon_sym_RBRACE] = ACTIONS(1294),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(1294),
    [anon_sym_RPAREN] = ACTIONS(1294),
    [aux_sym__word_token1] = ACTIONS(1294),
    [aux_sym__word_token2] = ACTIONS(1294),
    [aux_sym__word_token3] = ACTIONS(1294),
    [sym__whitespace] = ACTIONS(1294),
    [sym__soft_line_ending] = ACTIONS(1294),
    [sym_block_continuation] = ACTIONS(1298),
    [sym__block_quote_start] = ACTIONS(1294),
    [sym__indented_chunk_start] = ACTIONS(1294),
    [sym_atx_h1_marker] = ACTIONS(1294),
    [sym_atx_h2_marker] = ACTIONS(1294),
    [sym_atx_h3_marker] = ACTIONS(1294),
    [sym_atx_h4_marker] = ACTIONS(1294),
    [sym_atx_h5_marker] = ACTIONS(1294),
    [sym_atx_h6_marker] = ACTIONS(1294),
    [sym__thematic_break] = ACTIONS(1294),
    [sym__list_marker_minus] = ACTIONS(1294),
    [sym__list_marker_plus] = ACTIONS(1294),
    [sym__list_marker_star] = ACTIONS(1294),
    [sym__list_marker_parenthesis] = ACTIONS(1294),
    [sym__list_marker_dot] = ACTIONS(1294),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1294),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1294),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1294),
    [sym__blank_line_start] = ACTIONS(1294),
    [sym__html_block_1_start] = ACTIONS(1294),
    [sym__html_block_2_start] = ACTIONS(1294),
    [sym__html_block_3_start] = ACTIONS(1294),
    [sym__html_block_4_start] = ACTIONS(1294),
    [sym__html_block_5_start] = ACTIONS(1294),
    [sym__html_block_6_start] = ACTIONS(1294),
    [sym__html_block_7_start] = ACTIONS(1294),
    [sym__pipe_table_start] = ACTIONS(1294),
  },
  [STATE(121)] = {
    [ts_builtin_sym_end] = ACTIONS(1300),
    [anon_sym_LBRACK] = ACTIONS(1302),
    [anon_sym_RBRACK] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [anon_sym_POUND] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_AT] = ACTIONS(1300),
    [anon_sym_BSLASH] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym__] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_RPAREN] = ACTIONS(1300),
    [aux_sym__word_token1] = ACTIONS(1300),
    [aux_sym__word_token2] = ACTIONS(1300),
    [aux_sym__word_token3] = ACTIONS(1300),
    [sym__whitespace] = ACTIONS(1300),
    [sym__soft_line_ending] = ACTIONS(1300),
    [sym_block_continuation] = ACTIONS(1304),
    [sym__block_quote_start] = ACTIONS(1300),
    [sym__indented_chunk_start] = ACTIONS(1300),
    [sym_atx_h1_marker] = ACTIONS(1300),
    [sym_atx_h2_marker] = ACTIONS(1300),
    [sym_atx_h3_marker] = ACTIONS(1300),
    [sym_atx_h4_marker] = ACTIONS(1300),
    [sym_atx_h5_marker] = ACTIONS(1300),
    [sym_atx_h6_marker] = ACTIONS(1300),
    [sym__thematic_break] = ACTIONS(1300),
    [sym__list_marker_minus] = ACTIONS(1300),
    [sym__list_marker_plus] = ACTIONS(1300),
    [sym__list_marker_star] = ACTIONS(1300),
    [sym__list_marker_parenthesis] = ACTIONS(1300),
    [sym__list_marker_dot] = ACTIONS(1300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1300),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1300),
    [sym__blank_line_start] = ACTIONS(1300),
    [sym__html_block_1_start] = ACTIONS(1300),
    [sym__html_block_2_start] = ACTIONS(1300),
    [sym__html_block_3_start] = ACTIONS(1300),
    [sym__html_block_4_start] = ACTIONS(1300),
    [sym__html_block_5_start] = ACTIONS(1300),
    [sym__html_block_6_start] = ACTIONS(1300),
    [sym__html_block_7_start] = ACTIONS(1300),
    [sym__pipe_table_start] = ACTIONS(1300),
  },
  [STATE(122)] = {
    [ts_builtin_sym_end] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [anon_sym_RBRACK] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_POUND] = ACTIONS(1306),
    [anon_sym_DOLLAR] = ACTIONS(1306),
    [anon_sym_PERCENT] = ACTIONS(1306),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_COLON] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1306),
    [anon_sym_EQ] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_BSLASH] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1306),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_BQUOTE] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1306),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_RBRACE] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1306),
    [anon_sym_RPAREN] = ACTIONS(1306),
    [aux_sym__word_token1] = ACTIONS(1306),
    [aux_sym__word_token2] = ACTIONS(1306),
    [aux_sym__word_token3] = ACTIONS(1306),
    [sym__whitespace] = ACTIONS(1306),
    [sym__soft_line_ending] = ACTIONS(1306),
    [sym_block_continuation] = ACTIONS(1310),
    [sym__block_quote_start] = ACTIONS(1306),
    [sym__indented_chunk_start] = ACTIONS(1306),
    [sym_atx_h1_marker] = ACTIONS(1306),
    [sym_atx_h2_marker] = ACTIONS(1306),
    [sym_atx_h3_marker] = ACTIONS(1306),
    [sym_atx_h4_marker] = ACTIONS(1306),
    [sym_atx_h5_marker] = ACTIONS(1306),
    [sym_atx_h6_marker] = ACTIONS(1306),
    [sym__thematic_break] = ACTIONS(1306),
    [sym__list_marker_minus] = ACTIONS(1306),
    [sym__list_marker_plus] = ACTIONS(1306),
    [sym__list_marker_star] = ACTIONS(1306),
    [sym__list_marker_parenthesis] = ACTIONS(1306),
    [sym__list_marker_dot] = ACTIONS(1306),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1306),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1306),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1306),
    [sym__blank_line_start] = ACTIONS(1306),
    [sym__html_block_1_start] = ACTIONS(1306),
    [sym__html_block_2_start] = ACTIONS(1306),
    [sym__html_block_3_start] = ACTIONS(1306),
    [sym__html_block_4_start] = ACTIONS(1306),
    [sym__html_block_5_start] = ACTIONS(1306),
    [sym__html_block_6_start] = ACTIONS(1306),
    [sym__html_block_7_start] = ACTIONS(1306),
    [sym__pipe_table_start] = ACTIONS(1306),
  },
  [STATE(123)] = {
    [ts_builtin_sym_end] = ACTIONS(1312),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [anon_sym_RBRACK] = ACTIONS(1312),
    [anon_sym_LT] = ACTIONS(1312),
    [anon_sym_GT] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [anon_sym_POUND] = ACTIONS(1312),
    [anon_sym_DOLLAR] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_PLUS] = ACTIONS(1312),
    [anon_sym_COMMA] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_DOT] = ACTIONS(1312),
    [anon_sym_SLASH] = ACTIONS(1312),
    [anon_sym_COLON] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym_EQ] = ACTIONS(1312),
    [anon_sym_QMARK] = ACTIONS(1312),
    [anon_sym_AT] = ACTIONS(1312),
    [anon_sym_BSLASH] = ACTIONS(1312),
    [anon_sym_CARET] = ACTIONS(1312),
    [anon_sym__] = ACTIONS(1312),
    [anon_sym_BQUOTE] = ACTIONS(1312),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_PIPE] = ACTIONS(1312),
    [anon_sym_RBRACE] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_LPAREN] = ACTIONS(1312),
    [anon_sym_RPAREN] = ACTIONS(1312),
    [aux_sym__word_token1] = ACTIONS(1312),
    [aux_sym__word_token2] = ACTIONS(1312),
    [aux_sym__word_token3] = ACTIONS(1312),
    [sym__whitespace] = ACTIONS(1312),
    [sym__soft_line_ending] = ACTIONS(1312),
    [sym_block_continuation] = ACTIONS(1316),
    [sym__block_quote_start] = ACTIONS(1312),
    [sym__indented_chunk_start] = ACTIONS(1312),
    [sym_atx_h1_marker] = ACTIONS(1312),
    [sym_atx_h2_marker] = ACTIONS(1312),
    [sym_atx_h3_marker] = ACTIONS(1312),
    [sym_atx_h4_marker] = ACTIONS(1312),
    [sym_atx_h5_marker] = ACTIONS(1312),
    [sym_atx_h6_marker] = ACTIONS(1312),
    [sym__thematic_break] = ACTIONS(1312),
    [sym__list_marker_minus] = ACTIONS(1312),
    [sym__list_marker_plus] = ACTIONS(1312),
    [sym__list_marker_star] = ACTIONS(1312),
    [sym__list_marker_parenthesis] = ACTIONS(1312),
    [sym__list_marker_dot] = ACTIONS(1312),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1312),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1312),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1312),
    [sym__blank_line_start] = ACTIONS(1312),
    [sym__html_block_1_start] = ACTIONS(1312),
    [sym__html_block_2_start] = ACTIONS(1312),
    [sym__html_block_3_start] = ACTIONS(1312),
    [sym__html_block_4_start] = ACTIONS(1312),
    [sym__html_block_5_start] = ACTIONS(1312),
    [sym__html_block_6_start] = ACTIONS(1312),
    [sym__html_block_7_start] = ACTIONS(1312),
    [sym__pipe_table_start] = ACTIONS(1312),
  },
  [STATE(124)] = {
    [ts_builtin_sym_end] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_RBRACK] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1318),
    [anon_sym_GT] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1318),
    [anon_sym_BSLASH] = ACTIONS(1318),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym__] = ACTIONS(1318),
    [anon_sym_BQUOTE] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_RPAREN] = ACTIONS(1318),
    [aux_sym__word_token1] = ACTIONS(1318),
    [aux_sym__word_token2] = ACTIONS(1318),
    [aux_sym__word_token3] = ACTIONS(1318),
    [sym__whitespace] = ACTIONS(1318),
    [sym__soft_line_ending] = ACTIONS(1318),
    [sym_block_continuation] = ACTIONS(1322),
    [sym__block_quote_start] = ACTIONS(1318),
    [sym__indented_chunk_start] = ACTIONS(1318),
    [sym_atx_h1_marker] = ACTIONS(1318),
    [sym_atx_h2_marker] = ACTIONS(1318),
    [sym_atx_h3_marker] = ACTIONS(1318),
    [sym_atx_h4_marker] = ACTIONS(1318),
    [sym_atx_h5_marker] = ACTIONS(1318),
    [sym_atx_h6_marker] = ACTIONS(1318),
    [sym__thematic_break] = ACTIONS(1318),
    [sym__list_marker_minus] = ACTIONS(1318),
    [sym__list_marker_plus] = ACTIONS(1318),
    [sym__list_marker_star] = ACTIONS(1318),
    [sym__list_marker_parenthesis] = ACTIONS(1318),
    [sym__list_marker_dot] = ACTIONS(1318),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1318),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1318),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1318),
    [sym__blank_line_start] = ACTIONS(1318),
    [sym__html_block_1_start] = ACTIONS(1318),
    [sym__html_block_2_start] = ACTIONS(1318),
    [sym__html_block_3_start] = ACTIONS(1318),
    [sym__html_block_4_start] = ACTIONS(1318),
    [sym__html_block_5_start] = ACTIONS(1318),
    [sym__html_block_6_start] = ACTIONS(1318),
    [sym__html_block_7_start] = ACTIONS(1318),
    [sym__pipe_table_start] = ACTIONS(1318),
  },
  [STATE(125)] = {
    [ts_builtin_sym_end] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1326),
    [anon_sym_RBRACK] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_POUND] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_COLON] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_BSLASH] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_RPAREN] = ACTIONS(1324),
    [aux_sym__word_token1] = ACTIONS(1324),
    [aux_sym__word_token2] = ACTIONS(1324),
    [aux_sym__word_token3] = ACTIONS(1324),
    [sym__whitespace] = ACTIONS(1324),
    [sym__soft_line_ending] = ACTIONS(1324),
    [sym_block_continuation] = ACTIONS(1328),
    [sym__block_quote_start] = ACTIONS(1324),
    [sym__indented_chunk_start] = ACTIONS(1324),
    [sym_atx_h1_marker] = ACTIONS(1324),
    [sym_atx_h2_marker] = ACTIONS(1324),
    [sym_atx_h3_marker] = ACTIONS(1324),
    [sym_atx_h4_marker] = ACTIONS(1324),
    [sym_atx_h5_marker] = ACTIONS(1324),
    [sym_atx_h6_marker] = ACTIONS(1324),
    [sym__thematic_break] = ACTIONS(1324),
    [sym__list_marker_minus] = ACTIONS(1324),
    [sym__list_marker_plus] = ACTIONS(1324),
    [sym__list_marker_star] = ACTIONS(1324),
    [sym__list_marker_parenthesis] = ACTIONS(1324),
    [sym__list_marker_dot] = ACTIONS(1324),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1324),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1324),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1324),
    [sym__blank_line_start] = ACTIONS(1324),
    [sym__html_block_1_start] = ACTIONS(1324),
    [sym__html_block_2_start] = ACTIONS(1324),
    [sym__html_block_3_start] = ACTIONS(1324),
    [sym__html_block_4_start] = ACTIONS(1324),
    [sym__html_block_5_start] = ACTIONS(1324),
    [sym__html_block_6_start] = ACTIONS(1324),
    [sym__html_block_7_start] = ACTIONS(1324),
    [sym__pipe_table_start] = ACTIONS(1324),
  },
  [STATE(126)] = {
    [anon_sym_LBRACK] = ACTIONS(1330),
    [anon_sym_RBRACK] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_COLON] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1332),
    [anon_sym_BSLASH] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym__] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [aux_sym__word_token1] = ACTIONS(1332),
    [aux_sym__word_token2] = ACTIONS(1332),
    [aux_sym__word_token3] = ACTIONS(1332),
    [sym__whitespace] = ACTIONS(1332),
    [sym__soft_line_ending] = ACTIONS(1332),
    [sym__block_close] = ACTIONS(1332),
    [sym_block_continuation] = ACTIONS(1334),
    [sym__block_quote_start] = ACTIONS(1332),
    [sym__indented_chunk_start] = ACTIONS(1332),
    [sym_atx_h1_marker] = ACTIONS(1332),
    [sym_atx_h2_marker] = ACTIONS(1332),
    [sym_atx_h3_marker] = ACTIONS(1332),
    [sym_atx_h4_marker] = ACTIONS(1332),
    [sym_atx_h5_marker] = ACTIONS(1332),
    [sym_atx_h6_marker] = ACTIONS(1332),
    [sym__thematic_break] = ACTIONS(1332),
    [sym__list_marker_minus] = ACTIONS(1332),
    [sym__list_marker_plus] = ACTIONS(1332),
    [sym__list_marker_star] = ACTIONS(1332),
    [sym__list_marker_parenthesis] = ACTIONS(1332),
    [sym__list_marker_dot] = ACTIONS(1332),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1332),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1332),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1332),
    [sym__blank_line_start] = ACTIONS(1332),
    [sym__html_block_1_start] = ACTIONS(1332),
    [sym__html_block_2_start] = ACTIONS(1332),
    [sym__html_block_3_start] = ACTIONS(1332),
    [sym__html_block_4_start] = ACTIONS(1332),
    [sym__html_block_5_start] = ACTIONS(1332),
    [sym__html_block_6_start] = ACTIONS(1332),
    [sym__html_block_7_start] = ACTIONS(1332),
    [sym__pipe_table_start] = ACTIONS(1332),
  },
  [STATE(127)] = {
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_RBRACK] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_POUND] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1338),
    [anon_sym_PERCENT] = ACTIONS(1338),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_COLON] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1338),
    [anon_sym_EQ] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_BSLASH] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1338),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_BQUOTE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1338),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_RBRACE] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1338),
    [anon_sym_RPAREN] = ACTIONS(1338),
    [aux_sym__word_token1] = ACTIONS(1338),
    [aux_sym__word_token2] = ACTIONS(1338),
    [aux_sym__word_token3] = ACTIONS(1338),
    [sym__whitespace] = ACTIONS(1338),
    [sym__soft_line_ending] = ACTIONS(1338),
    [sym__block_close] = ACTIONS(1338),
    [sym_block_continuation] = ACTIONS(1340),
    [sym__block_quote_start] = ACTIONS(1338),
    [sym__indented_chunk_start] = ACTIONS(1338),
    [sym_atx_h1_marker] = ACTIONS(1338),
    [sym_atx_h2_marker] = ACTIONS(1338),
    [sym_atx_h3_marker] = ACTIONS(1338),
    [sym_atx_h4_marker] = ACTIONS(1338),
    [sym_atx_h5_marker] = ACTIONS(1338),
    [sym_atx_h6_marker] = ACTIONS(1338),
    [sym__thematic_break] = ACTIONS(1338),
    [sym__list_marker_minus] = ACTIONS(1338),
    [sym__list_marker_plus] = ACTIONS(1338),
    [sym__list_marker_star] = ACTIONS(1338),
    [sym__list_marker_parenthesis] = ACTIONS(1338),
    [sym__list_marker_dot] = ACTIONS(1338),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1338),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1338),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1338),
    [sym__blank_line_start] = ACTIONS(1338),
    [sym__html_block_1_start] = ACTIONS(1338),
    [sym__html_block_2_start] = ACTIONS(1338),
    [sym__html_block_3_start] = ACTIONS(1338),
    [sym__html_block_4_start] = ACTIONS(1338),
    [sym__html_block_5_start] = ACTIONS(1338),
    [sym__html_block_6_start] = ACTIONS(1338),
    [sym__html_block_7_start] = ACTIONS(1338),
    [sym__pipe_table_start] = ACTIONS(1338),
  },
  [STATE(128)] = {
    [anon_sym_LBRACK] = ACTIONS(1342),
    [anon_sym_RBRACK] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [anon_sym_POUND] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_SQUOTE] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_COLON] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_BSLASH] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym__] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_RPAREN] = ACTIONS(1344),
    [aux_sym__word_token1] = ACTIONS(1344),
    [aux_sym__word_token2] = ACTIONS(1344),
    [aux_sym__word_token3] = ACTIONS(1344),
    [sym__whitespace] = ACTIONS(1344),
    [sym__soft_line_ending] = ACTIONS(1344),
    [sym__block_close] = ACTIONS(1344),
    [sym_block_continuation] = ACTIONS(1346),
    [sym__block_quote_start] = ACTIONS(1344),
    [sym__indented_chunk_start] = ACTIONS(1344),
    [sym_atx_h1_marker] = ACTIONS(1344),
    [sym_atx_h2_marker] = ACTIONS(1344),
    [sym_atx_h3_marker] = ACTIONS(1344),
    [sym_atx_h4_marker] = ACTIONS(1344),
    [sym_atx_h5_marker] = ACTIONS(1344),
    [sym_atx_h6_marker] = ACTIONS(1344),
    [sym__thematic_break] = ACTIONS(1344),
    [sym__list_marker_minus] = ACTIONS(1344),
    [sym__list_marker_plus] = ACTIONS(1344),
    [sym__list_marker_star] = ACTIONS(1344),
    [sym__list_marker_parenthesis] = ACTIONS(1344),
    [sym__list_marker_dot] = ACTIONS(1344),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1344),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1344),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1344),
    [sym__blank_line_start] = ACTIONS(1344),
    [sym__html_block_1_start] = ACTIONS(1344),
    [sym__html_block_2_start] = ACTIONS(1344),
    [sym__html_block_3_start] = ACTIONS(1344),
    [sym__html_block_4_start] = ACTIONS(1344),
    [sym__html_block_5_start] = ACTIONS(1344),
    [sym__html_block_6_start] = ACTIONS(1344),
    [sym__html_block_7_start] = ACTIONS(1344),
    [sym__pipe_table_start] = ACTIONS(1344),
  },
  [STATE(129)] = {
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_RBRACK] = ACTIONS(1350),
    [anon_sym_LT] = ACTIONS(1350),
    [anon_sym_GT] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [anon_sym_POUND] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_PLUS] = ACTIONS(1350),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_EQ] = ACTIONS(1350),
    [anon_sym_QMARK] = ACTIONS(1350),
    [anon_sym_AT] = ACTIONS(1350),
    [anon_sym_BSLASH] = ACTIONS(1350),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym__] = ACTIONS(1350),
    [anon_sym_BQUOTE] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_RPAREN] = ACTIONS(1350),
    [aux_sym__word_token1] = ACTIONS(1350),
    [aux_sym__word_token2] = ACTIONS(1350),
    [aux_sym__word_token3] = ACTIONS(1350),
    [sym__whitespace] = ACTIONS(1350),
    [sym__soft_line_ending] = ACTIONS(1350),
    [sym__block_close] = ACTIONS(1350),
    [sym_block_continuation] = ACTIONS(1352),
    [sym__block_quote_start] = ACTIONS(1350),
    [sym__indented_chunk_start] = ACTIONS(1350),
    [sym_atx_h1_marker] = ACTIONS(1350),
    [sym_atx_h2_marker] = ACTIONS(1350),
    [sym_atx_h3_marker] = ACTIONS(1350),
    [sym_atx_h4_marker] = ACTIONS(1350),
    [sym_atx_h5_marker] = ACTIONS(1350),
    [sym_atx_h6_marker] = ACTIONS(1350),
    [sym__thematic_break] = ACTIONS(1350),
    [sym__list_marker_minus] = ACTIONS(1350),
    [sym__list_marker_plus] = ACTIONS(1350),
    [sym__list_marker_star] = ACTIONS(1350),
    [sym__list_marker_parenthesis] = ACTIONS(1350),
    [sym__list_marker_dot] = ACTIONS(1350),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1350),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1350),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1350),
    [sym__blank_line_start] = ACTIONS(1350),
    [sym__html_block_1_start] = ACTIONS(1350),
    [sym__html_block_2_start] = ACTIONS(1350),
    [sym__html_block_3_start] = ACTIONS(1350),
    [sym__html_block_4_start] = ACTIONS(1350),
    [sym__html_block_5_start] = ACTIONS(1350),
    [sym__html_block_6_start] = ACTIONS(1350),
    [sym__html_block_7_start] = ACTIONS(1350),
    [sym__pipe_table_start] = ACTIONS(1350),
  },
  [STATE(130)] = {
    [anon_sym_LBRACK] = ACTIONS(1354),
    [anon_sym_RBRACK] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [anon_sym_POUND] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_SQUOTE] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_COMMA] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_DOT] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_COLON] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_AT] = ACTIONS(1356),
    [anon_sym_BSLASH] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym__] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_TILDE] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_RPAREN] = ACTIONS(1356),
    [aux_sym__word_token1] = ACTIONS(1356),
    [aux_sym__word_token2] = ACTIONS(1356),
    [aux_sym__word_token3] = ACTIONS(1356),
    [sym__whitespace] = ACTIONS(1356),
    [sym__soft_line_ending] = ACTIONS(1356),
    [sym__block_close] = ACTIONS(1356),
    [sym_block_continuation] = ACTIONS(1358),
    [sym__block_quote_start] = ACTIONS(1356),
    [sym__indented_chunk_start] = ACTIONS(1356),
    [sym_atx_h1_marker] = ACTIONS(1356),
    [sym_atx_h2_marker] = ACTIONS(1356),
    [sym_atx_h3_marker] = ACTIONS(1356),
    [sym_atx_h4_marker] = ACTIONS(1356),
    [sym_atx_h5_marker] = ACTIONS(1356),
    [sym_atx_h6_marker] = ACTIONS(1356),
    [sym__thematic_break] = ACTIONS(1356),
    [sym__list_marker_minus] = ACTIONS(1356),
    [sym__list_marker_plus] = ACTIONS(1356),
    [sym__list_marker_star] = ACTIONS(1356),
    [sym__list_marker_parenthesis] = ACTIONS(1356),
    [sym__list_marker_dot] = ACTIONS(1356),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1356),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1356),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1356),
    [sym__blank_line_start] = ACTIONS(1356),
    [sym__html_block_1_start] = ACTIONS(1356),
    [sym__html_block_2_start] = ACTIONS(1356),
    [sym__html_block_3_start] = ACTIONS(1356),
    [sym__html_block_4_start] = ACTIONS(1356),
    [sym__html_block_5_start] = ACTIONS(1356),
    [sym__html_block_6_start] = ACTIONS(1356),
    [sym__html_block_7_start] = ACTIONS(1356),
    [sym__pipe_table_start] = ACTIONS(1356),
  },
  [STATE(131)] = {
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_RBRACK] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_POUND] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_COLON] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_BSLASH] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_RPAREN] = ACTIONS(1362),
    [aux_sym__word_token1] = ACTIONS(1362),
    [aux_sym__word_token2] = ACTIONS(1362),
    [aux_sym__word_token3] = ACTIONS(1362),
    [sym__whitespace] = ACTIONS(1362),
    [sym__soft_line_ending] = ACTIONS(1362),
    [sym__block_close] = ACTIONS(1362),
    [sym_block_continuation] = ACTIONS(1364),
    [sym__block_quote_start] = ACTIONS(1362),
    [sym__indented_chunk_start] = ACTIONS(1362),
    [sym_atx_h1_marker] = ACTIONS(1362),
    [sym_atx_h2_marker] = ACTIONS(1362),
    [sym_atx_h3_marker] = ACTIONS(1362),
    [sym_atx_h4_marker] = ACTIONS(1362),
    [sym_atx_h5_marker] = ACTIONS(1362),
    [sym_atx_h6_marker] = ACTIONS(1362),
    [sym__thematic_break] = ACTIONS(1362),
    [sym__list_marker_minus] = ACTIONS(1362),
    [sym__list_marker_plus] = ACTIONS(1362),
    [sym__list_marker_star] = ACTIONS(1362),
    [sym__list_marker_parenthesis] = ACTIONS(1362),
    [sym__list_marker_dot] = ACTIONS(1362),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1362),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1362),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1362),
    [sym__blank_line_start] = ACTIONS(1362),
    [sym__html_block_1_start] = ACTIONS(1362),
    [sym__html_block_2_start] = ACTIONS(1362),
    [sym__html_block_3_start] = ACTIONS(1362),
    [sym__html_block_4_start] = ACTIONS(1362),
    [sym__html_block_5_start] = ACTIONS(1362),
    [sym__html_block_6_start] = ACTIONS(1362),
    [sym__html_block_7_start] = ACTIONS(1362),
    [sym__pipe_table_start] = ACTIONS(1362),
  },
  [STATE(132)] = {
    [ts_builtin_sym_end] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [anon_sym_RBRACK] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_POUND] = ACTIONS(1366),
    [anon_sym_DOLLAR] = ACTIONS(1366),
    [anon_sym_PERCENT] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_COLON] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym_EQ] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_BSLASH] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1366),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_BQUOTE] = ACTIONS(1366),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_RBRACE] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1366),
    [anon_sym_RPAREN] = ACTIONS(1366),
    [aux_sym__word_token1] = ACTIONS(1366),
    [aux_sym__word_token2] = ACTIONS(1366),
    [aux_sym__word_token3] = ACTIONS(1366),
    [sym__whitespace] = ACTIONS(1366),
    [sym__soft_line_ending] = ACTIONS(1366),
    [sym_block_continuation] = ACTIONS(1370),
    [sym__block_quote_start] = ACTIONS(1366),
    [sym__indented_chunk_start] = ACTIONS(1366),
    [sym_atx_h1_marker] = ACTIONS(1366),
    [sym_atx_h2_marker] = ACTIONS(1366),
    [sym_atx_h3_marker] = ACTIONS(1366),
    [sym_atx_h4_marker] = ACTIONS(1366),
    [sym_atx_h5_marker] = ACTIONS(1366),
    [sym_atx_h6_marker] = ACTIONS(1366),
    [sym__thematic_break] = ACTIONS(1366),
    [sym__list_marker_minus] = ACTIONS(1366),
    [sym__list_marker_plus] = ACTIONS(1366),
    [sym__list_marker_star] = ACTIONS(1366),
    [sym__list_marker_parenthesis] = ACTIONS(1366),
    [sym__list_marker_dot] = ACTIONS(1366),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1366),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1366),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1366),
    [sym__blank_line_start] = ACTIONS(1366),
    [sym__html_block_1_start] = ACTIONS(1366),
    [sym__html_block_2_start] = ACTIONS(1366),
    [sym__html_block_3_start] = ACTIONS(1366),
    [sym__html_block_4_start] = ACTIONS(1366),
    [sym__html_block_5_start] = ACTIONS(1366),
    [sym__html_block_6_start] = ACTIONS(1366),
    [sym__html_block_7_start] = ACTIONS(1366),
    [sym__pipe_table_start] = ACTIONS(1366),
  },
  [STATE(133)] = {
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_RBRACK] = ACTIONS(1374),
    [anon_sym_LT] = ACTIONS(1374),
    [anon_sym_GT] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1374),
    [anon_sym_POUND] = ACTIONS(1374),
    [anon_sym_DOLLAR] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1374),
    [anon_sym_PLUS] = ACTIONS(1374),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_DASH] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1374),
    [anon_sym_COLON] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_EQ] = ACTIONS(1374),
    [anon_sym_QMARK] = ACTIONS(1374),
    [anon_sym_AT] = ACTIONS(1374),
    [anon_sym_BSLASH] = ACTIONS(1374),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym__] = ACTIONS(1374),
    [anon_sym_BQUOTE] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_RPAREN] = ACTIONS(1374),
    [aux_sym__word_token1] = ACTIONS(1374),
    [aux_sym__word_token2] = ACTIONS(1374),
    [aux_sym__word_token3] = ACTIONS(1374),
    [sym__whitespace] = ACTIONS(1374),
    [sym__soft_line_ending] = ACTIONS(1374),
    [sym__block_close] = ACTIONS(1374),
    [sym_block_continuation] = ACTIONS(1376),
    [sym__block_quote_start] = ACTIONS(1374),
    [sym__indented_chunk_start] = ACTIONS(1374),
    [sym_atx_h1_marker] = ACTIONS(1374),
    [sym_atx_h2_marker] = ACTIONS(1374),
    [sym_atx_h3_marker] = ACTIONS(1374),
    [sym_atx_h4_marker] = ACTIONS(1374),
    [sym_atx_h5_marker] = ACTIONS(1374),
    [sym_atx_h6_marker] = ACTIONS(1374),
    [sym__thematic_break] = ACTIONS(1374),
    [sym__list_marker_minus] = ACTIONS(1374),
    [sym__list_marker_plus] = ACTIONS(1374),
    [sym__list_marker_star] = ACTIONS(1374),
    [sym__list_marker_parenthesis] = ACTIONS(1374),
    [sym__list_marker_dot] = ACTIONS(1374),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1374),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1374),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1374),
    [sym__blank_line_start] = ACTIONS(1374),
    [sym__html_block_1_start] = ACTIONS(1374),
    [sym__html_block_2_start] = ACTIONS(1374),
    [sym__html_block_3_start] = ACTIONS(1374),
    [sym__html_block_4_start] = ACTIONS(1374),
    [sym__html_block_5_start] = ACTIONS(1374),
    [sym__html_block_6_start] = ACTIONS(1374),
    [sym__html_block_7_start] = ACTIONS(1374),
    [sym__pipe_table_start] = ACTIONS(1374),
  },
  [STATE(134)] = {
    [anon_sym_LBRACK] = ACTIONS(1378),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_POUND] = ACTIONS(1380),
    [anon_sym_DOLLAR] = ACTIONS(1380),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_BSLASH] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_BQUOTE] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [aux_sym__word_token1] = ACTIONS(1380),
    [aux_sym__word_token2] = ACTIONS(1380),
    [aux_sym__word_token3] = ACTIONS(1380),
    [sym__whitespace] = ACTIONS(1380),
    [sym__soft_line_ending] = ACTIONS(1380),
    [sym__block_close] = ACTIONS(1380),
    [sym_block_continuation] = ACTIONS(1382),
    [sym__block_quote_start] = ACTIONS(1380),
    [sym__indented_chunk_start] = ACTIONS(1380),
    [sym_atx_h1_marker] = ACTIONS(1380),
    [sym_atx_h2_marker] = ACTIONS(1380),
    [sym_atx_h3_marker] = ACTIONS(1380),
    [sym_atx_h4_marker] = ACTIONS(1380),
    [sym_atx_h5_marker] = ACTIONS(1380),
    [sym_atx_h6_marker] = ACTIONS(1380),
    [sym__thematic_break] = ACTIONS(1380),
    [sym__list_marker_minus] = ACTIONS(1380),
    [sym__list_marker_plus] = ACTIONS(1380),
    [sym__list_marker_star] = ACTIONS(1380),
    [sym__list_marker_parenthesis] = ACTIONS(1380),
    [sym__list_marker_dot] = ACTIONS(1380),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1380),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1380),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1380),
    [sym__blank_line_start] = ACTIONS(1380),
    [sym__html_block_1_start] = ACTIONS(1380),
    [sym__html_block_2_start] = ACTIONS(1380),
    [sym__html_block_3_start] = ACTIONS(1380),
    [sym__html_block_4_start] = ACTIONS(1380),
    [sym__html_block_5_start] = ACTIONS(1380),
    [sym__html_block_6_start] = ACTIONS(1380),
    [sym__html_block_7_start] = ACTIONS(1380),
    [sym__pipe_table_start] = ACTIONS(1380),
  },
  [STATE(135)] = {
    [ts_builtin_sym_end] = ACTIONS(1356),
    [anon_sym_LBRACK] = ACTIONS(1354),
    [anon_sym_RBRACK] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [anon_sym_POUND] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_SQUOTE] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_COMMA] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_DOT] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_COLON] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_AT] = ACTIONS(1356),
    [anon_sym_BSLASH] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym__] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_TILDE] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_RPAREN] = ACTIONS(1356),
    [aux_sym__word_token1] = ACTIONS(1356),
    [aux_sym__word_token2] = ACTIONS(1356),
    [aux_sym__word_token3] = ACTIONS(1356),
    [sym__whitespace] = ACTIONS(1356),
    [sym__soft_line_ending] = ACTIONS(1356),
    [sym_block_continuation] = ACTIONS(1384),
    [sym__block_quote_start] = ACTIONS(1356),
    [sym__indented_chunk_start] = ACTIONS(1356),
    [sym_atx_h1_marker] = ACTIONS(1356),
    [sym_atx_h2_marker] = ACTIONS(1356),
    [sym_atx_h3_marker] = ACTIONS(1356),
    [sym_atx_h4_marker] = ACTIONS(1356),
    [sym_atx_h5_marker] = ACTIONS(1356),
    [sym_atx_h6_marker] = ACTIONS(1356),
    [sym__thematic_break] = ACTIONS(1356),
    [sym__list_marker_minus] = ACTIONS(1356),
    [sym__list_marker_plus] = ACTIONS(1356),
    [sym__list_marker_star] = ACTIONS(1356),
    [sym__list_marker_parenthesis] = ACTIONS(1356),
    [sym__list_marker_dot] = ACTIONS(1356),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1356),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1356),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1356),
    [sym__blank_line_start] = ACTIONS(1356),
    [sym__html_block_1_start] = ACTIONS(1356),
    [sym__html_block_2_start] = ACTIONS(1356),
    [sym__html_block_3_start] = ACTIONS(1356),
    [sym__html_block_4_start] = ACTIONS(1356),
    [sym__html_block_5_start] = ACTIONS(1356),
    [sym__html_block_6_start] = ACTIONS(1356),
    [sym__html_block_7_start] = ACTIONS(1356),
    [sym__pipe_table_start] = ACTIONS(1356),
  },
  [STATE(136)] = {
    [ts_builtin_sym_end] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_RBRACK] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_POUND] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_COLON] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_BSLASH] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_RPAREN] = ACTIONS(1362),
    [aux_sym__word_token1] = ACTIONS(1362),
    [aux_sym__word_token2] = ACTIONS(1362),
    [aux_sym__word_token3] = ACTIONS(1362),
    [sym__whitespace] = ACTIONS(1362),
    [sym__soft_line_ending] = ACTIONS(1362),
    [sym_block_continuation] = ACTIONS(1386),
    [sym__block_quote_start] = ACTIONS(1362),
    [sym__indented_chunk_start] = ACTIONS(1362),
    [sym_atx_h1_marker] = ACTIONS(1362),
    [sym_atx_h2_marker] = ACTIONS(1362),
    [sym_atx_h3_marker] = ACTIONS(1362),
    [sym_atx_h4_marker] = ACTIONS(1362),
    [sym_atx_h5_marker] = ACTIONS(1362),
    [sym_atx_h6_marker] = ACTIONS(1362),
    [sym__thematic_break] = ACTIONS(1362),
    [sym__list_marker_minus] = ACTIONS(1362),
    [sym__list_marker_plus] = ACTIONS(1362),
    [sym__list_marker_star] = ACTIONS(1362),
    [sym__list_marker_parenthesis] = ACTIONS(1362),
    [sym__list_marker_dot] = ACTIONS(1362),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1362),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1362),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1362),
    [sym__blank_line_start] = ACTIONS(1362),
    [sym__html_block_1_start] = ACTIONS(1362),
    [sym__html_block_2_start] = ACTIONS(1362),
    [sym__html_block_3_start] = ACTIONS(1362),
    [sym__html_block_4_start] = ACTIONS(1362),
    [sym__html_block_5_start] = ACTIONS(1362),
    [sym__html_block_6_start] = ACTIONS(1362),
    [sym__html_block_7_start] = ACTIONS(1362),
    [sym__pipe_table_start] = ACTIONS(1362),
  },
  [STATE(137)] = {
    [ts_builtin_sym_end] = ACTIONS(1290),
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_RBRACK] = ACTIONS(1290),
    [anon_sym_LT] = ACTIONS(1290),
    [anon_sym_GT] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [anon_sym_DQUOTE] = ACTIONS(1290),
    [anon_sym_POUND] = ACTIONS(1290),
    [anon_sym_DOLLAR] = ACTIONS(1290),
    [anon_sym_PERCENT] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_SLASH] = ACTIONS(1290),
    [anon_sym_COLON] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_EQ] = ACTIONS(1290),
    [anon_sym_QMARK] = ACTIONS(1290),
    [anon_sym_AT] = ACTIONS(1290),
    [anon_sym_BSLASH] = ACTIONS(1290),
    [anon_sym_CARET] = ACTIONS(1290),
    [anon_sym__] = ACTIONS(1290),
    [anon_sym_BQUOTE] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_PIPE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_RPAREN] = ACTIONS(1290),
    [aux_sym__word_token1] = ACTIONS(1290),
    [aux_sym__word_token2] = ACTIONS(1290),
    [aux_sym__word_token3] = ACTIONS(1290),
    [sym__whitespace] = ACTIONS(1290),
    [sym__soft_line_ending] = ACTIONS(1290),
    [sym_block_continuation] = ACTIONS(1388),
    [sym__block_quote_start] = ACTIONS(1290),
    [sym__indented_chunk_start] = ACTIONS(1290),
    [sym_atx_h1_marker] = ACTIONS(1290),
    [sym_atx_h2_marker] = ACTIONS(1290),
    [sym_atx_h3_marker] = ACTIONS(1290),
    [sym_atx_h4_marker] = ACTIONS(1290),
    [sym_atx_h5_marker] = ACTIONS(1290),
    [sym_atx_h6_marker] = ACTIONS(1290),
    [sym__thematic_break] = ACTIONS(1290),
    [sym__list_marker_minus] = ACTIONS(1290),
    [sym__list_marker_plus] = ACTIONS(1290),
    [sym__list_marker_star] = ACTIONS(1290),
    [sym__list_marker_parenthesis] = ACTIONS(1290),
    [sym__list_marker_dot] = ACTIONS(1290),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1290),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1290),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1290),
    [sym__blank_line_start] = ACTIONS(1290),
    [sym__html_block_1_start] = ACTIONS(1290),
    [sym__html_block_2_start] = ACTIONS(1290),
    [sym__html_block_3_start] = ACTIONS(1290),
    [sym__html_block_4_start] = ACTIONS(1290),
    [sym__html_block_5_start] = ACTIONS(1290),
    [sym__html_block_6_start] = ACTIONS(1290),
    [sym__html_block_7_start] = ACTIONS(1290),
    [sym__pipe_table_start] = ACTIONS(1290),
  },
  [STATE(138)] = {
    [ts_builtin_sym_end] = ACTIONS(1374),
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_RBRACK] = ACTIONS(1374),
    [anon_sym_LT] = ACTIONS(1374),
    [anon_sym_GT] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1374),
    [anon_sym_POUND] = ACTIONS(1374),
    [anon_sym_DOLLAR] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1374),
    [anon_sym_PLUS] = ACTIONS(1374),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_DASH] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1374),
    [anon_sym_COLON] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_EQ] = ACTIONS(1374),
    [anon_sym_QMARK] = ACTIONS(1374),
    [anon_sym_AT] = ACTIONS(1374),
    [anon_sym_BSLASH] = ACTIONS(1374),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym__] = ACTIONS(1374),
    [anon_sym_BQUOTE] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_RPAREN] = ACTIONS(1374),
    [aux_sym__word_token1] = ACTIONS(1374),
    [aux_sym__word_token2] = ACTIONS(1374),
    [aux_sym__word_token3] = ACTIONS(1374),
    [sym__whitespace] = ACTIONS(1374),
    [sym__soft_line_ending] = ACTIONS(1374),
    [sym_block_continuation] = ACTIONS(1390),
    [sym__block_quote_start] = ACTIONS(1374),
    [sym__indented_chunk_start] = ACTIONS(1374),
    [sym_atx_h1_marker] = ACTIONS(1374),
    [sym_atx_h2_marker] = ACTIONS(1374),
    [sym_atx_h3_marker] = ACTIONS(1374),
    [sym_atx_h4_marker] = ACTIONS(1374),
    [sym_atx_h5_marker] = ACTIONS(1374),
    [sym_atx_h6_marker] = ACTIONS(1374),
    [sym__thematic_break] = ACTIONS(1374),
    [sym__list_marker_minus] = ACTIONS(1374),
    [sym__list_marker_plus] = ACTIONS(1374),
    [sym__list_marker_star] = ACTIONS(1374),
    [sym__list_marker_parenthesis] = ACTIONS(1374),
    [sym__list_marker_dot] = ACTIONS(1374),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1374),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1374),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1374),
    [sym__blank_line_start] = ACTIONS(1374),
    [sym__html_block_1_start] = ACTIONS(1374),
    [sym__html_block_2_start] = ACTIONS(1374),
    [sym__html_block_3_start] = ACTIONS(1374),
    [sym__html_block_4_start] = ACTIONS(1374),
    [sym__html_block_5_start] = ACTIONS(1374),
    [sym__html_block_6_start] = ACTIONS(1374),
    [sym__html_block_7_start] = ACTIONS(1374),
    [sym__pipe_table_start] = ACTIONS(1374),
  },
  [STATE(139)] = {
    [ts_builtin_sym_end] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_POUND] = ACTIONS(1380),
    [anon_sym_DOLLAR] = ACTIONS(1380),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_BSLASH] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_BQUOTE] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [aux_sym__word_token1] = ACTIONS(1380),
    [aux_sym__word_token2] = ACTIONS(1380),
    [aux_sym__word_token3] = ACTIONS(1380),
    [sym__whitespace] = ACTIONS(1380),
    [sym__soft_line_ending] = ACTIONS(1380),
    [sym_block_continuation] = ACTIONS(1392),
    [sym__block_quote_start] = ACTIONS(1380),
    [sym__indented_chunk_start] = ACTIONS(1380),
    [sym_atx_h1_marker] = ACTIONS(1380),
    [sym_atx_h2_marker] = ACTIONS(1380),
    [sym_atx_h3_marker] = ACTIONS(1380),
    [sym_atx_h4_marker] = ACTIONS(1380),
    [sym_atx_h5_marker] = ACTIONS(1380),
    [sym_atx_h6_marker] = ACTIONS(1380),
    [sym__thematic_break] = ACTIONS(1380),
    [sym__list_marker_minus] = ACTIONS(1380),
    [sym__list_marker_plus] = ACTIONS(1380),
    [sym__list_marker_star] = ACTIONS(1380),
    [sym__list_marker_parenthesis] = ACTIONS(1380),
    [sym__list_marker_dot] = ACTIONS(1380),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1380),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1380),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1380),
    [sym__blank_line_start] = ACTIONS(1380),
    [sym__html_block_1_start] = ACTIONS(1380),
    [sym__html_block_2_start] = ACTIONS(1380),
    [sym__html_block_3_start] = ACTIONS(1380),
    [sym__html_block_4_start] = ACTIONS(1380),
    [sym__html_block_5_start] = ACTIONS(1380),
    [sym__html_block_6_start] = ACTIONS(1380),
    [sym__html_block_7_start] = ACTIONS(1380),
    [sym__pipe_table_start] = ACTIONS(1380),
  },
  [STATE(140)] = {
    [anon_sym_LBRACK] = ACTIONS(1394),
    [anon_sym_RBRACK] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_POUND] = ACTIONS(1396),
    [anon_sym_DOLLAR] = ACTIONS(1396),
    [anon_sym_PERCENT] = ACTIONS(1396),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_COLON] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1396),
    [anon_sym_EQ] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_BSLASH] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1396),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_BQUOTE] = ACTIONS(1396),
    [anon_sym_LBRACE] = ACTIONS(1396),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_RBRACE] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1396),
    [anon_sym_RPAREN] = ACTIONS(1396),
    [aux_sym__word_token1] = ACTIONS(1396),
    [aux_sym__word_token2] = ACTIONS(1396),
    [aux_sym__word_token3] = ACTIONS(1396),
    [sym__whitespace] = ACTIONS(1396),
    [sym__soft_line_ending] = ACTIONS(1396),
    [sym__block_close] = ACTIONS(1396),
    [sym_block_continuation] = ACTIONS(1398),
    [sym__block_quote_start] = ACTIONS(1396),
    [sym__indented_chunk_start] = ACTIONS(1396),
    [sym_atx_h1_marker] = ACTIONS(1396),
    [sym_atx_h2_marker] = ACTIONS(1396),
    [sym_atx_h3_marker] = ACTIONS(1396),
    [sym_atx_h4_marker] = ACTIONS(1396),
    [sym_atx_h5_marker] = ACTIONS(1396),
    [sym_atx_h6_marker] = ACTIONS(1396),
    [sym__thematic_break] = ACTIONS(1396),
    [sym__list_marker_minus] = ACTIONS(1396),
    [sym__list_marker_plus] = ACTIONS(1396),
    [sym__list_marker_star] = ACTIONS(1396),
    [sym__list_marker_parenthesis] = ACTIONS(1396),
    [sym__list_marker_dot] = ACTIONS(1396),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1396),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1396),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1396),
    [sym__blank_line_start] = ACTIONS(1396),
    [sym__html_block_1_start] = ACTIONS(1396),
    [sym__html_block_2_start] = ACTIONS(1396),
    [sym__html_block_3_start] = ACTIONS(1396),
    [sym__html_block_4_start] = ACTIONS(1396),
    [sym__html_block_5_start] = ACTIONS(1396),
    [sym__html_block_6_start] = ACTIONS(1396),
    [sym__html_block_7_start] = ACTIONS(1396),
    [sym__pipe_table_start] = ACTIONS(1396),
  },
  [STATE(141)] = {
    [anon_sym_LBRACK] = ACTIONS(1400),
    [anon_sym_RBRACK] = ACTIONS(1402),
    [anon_sym_LT] = ACTIONS(1402),
    [anon_sym_GT] = ACTIONS(1402),
    [anon_sym_BANG] = ACTIONS(1402),
    [anon_sym_DQUOTE] = ACTIONS(1402),
    [anon_sym_POUND] = ACTIONS(1402),
    [anon_sym_DOLLAR] = ACTIONS(1402),
    [anon_sym_PERCENT] = ACTIONS(1402),
    [anon_sym_AMP] = ACTIONS(1402),
    [anon_sym_SQUOTE] = ACTIONS(1402),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_COMMA] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_DOT] = ACTIONS(1402),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_COLON] = ACTIONS(1402),
    [anon_sym_SEMI] = ACTIONS(1402),
    [anon_sym_EQ] = ACTIONS(1402),
    [anon_sym_QMARK] = ACTIONS(1402),
    [anon_sym_AT] = ACTIONS(1402),
    [anon_sym_BSLASH] = ACTIONS(1402),
    [anon_sym_CARET] = ACTIONS(1402),
    [anon_sym__] = ACTIONS(1402),
    [anon_sym_BQUOTE] = ACTIONS(1402),
    [anon_sym_LBRACE] = ACTIONS(1402),
    [anon_sym_PIPE] = ACTIONS(1402),
    [anon_sym_RBRACE] = ACTIONS(1402),
    [anon_sym_TILDE] = ACTIONS(1402),
    [anon_sym_LPAREN] = ACTIONS(1402),
    [anon_sym_RPAREN] = ACTIONS(1402),
    [aux_sym__word_token1] = ACTIONS(1402),
    [aux_sym__word_token2] = ACTIONS(1402),
    [aux_sym__word_token3] = ACTIONS(1402),
    [sym__whitespace] = ACTIONS(1402),
    [sym__soft_line_ending] = ACTIONS(1402),
    [sym__block_close] = ACTIONS(1402),
    [sym_block_continuation] = ACTIONS(1404),
    [sym__block_quote_start] = ACTIONS(1402),
    [sym__indented_chunk_start] = ACTIONS(1402),
    [sym_atx_h1_marker] = ACTIONS(1402),
    [sym_atx_h2_marker] = ACTIONS(1402),
    [sym_atx_h3_marker] = ACTIONS(1402),
    [sym_atx_h4_marker] = ACTIONS(1402),
    [sym_atx_h5_marker] = ACTIONS(1402),
    [sym_atx_h6_marker] = ACTIONS(1402),
    [sym__thematic_break] = ACTIONS(1402),
    [sym__list_marker_minus] = ACTIONS(1402),
    [sym__list_marker_plus] = ACTIONS(1402),
    [sym__list_marker_star] = ACTIONS(1402),
    [sym__list_marker_parenthesis] = ACTIONS(1402),
    [sym__list_marker_dot] = ACTIONS(1402),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1402),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1402),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1402),
    [sym__blank_line_start] = ACTIONS(1402),
    [sym__html_block_1_start] = ACTIONS(1402),
    [sym__html_block_2_start] = ACTIONS(1402),
    [sym__html_block_3_start] = ACTIONS(1402),
    [sym__html_block_4_start] = ACTIONS(1402),
    [sym__html_block_5_start] = ACTIONS(1402),
    [sym__html_block_6_start] = ACTIONS(1402),
    [sym__html_block_7_start] = ACTIONS(1402),
    [sym__pipe_table_start] = ACTIONS(1402),
  },
  [STATE(142)] = {
    [anon_sym_LBRACK] = ACTIONS(1406),
    [anon_sym_RBRACK] = ACTIONS(1408),
    [anon_sym_LT] = ACTIONS(1408),
    [anon_sym_GT] = ACTIONS(1408),
    [anon_sym_BANG] = ACTIONS(1408),
    [anon_sym_DQUOTE] = ACTIONS(1408),
    [anon_sym_POUND] = ACTIONS(1408),
    [anon_sym_DOLLAR] = ACTIONS(1408),
    [anon_sym_PERCENT] = ACTIONS(1408),
    [anon_sym_AMP] = ACTIONS(1408),
    [anon_sym_SQUOTE] = ACTIONS(1408),
    [anon_sym_STAR] = ACTIONS(1408),
    [anon_sym_PLUS] = ACTIONS(1408),
    [anon_sym_COMMA] = ACTIONS(1408),
    [anon_sym_DASH] = ACTIONS(1408),
    [anon_sym_DOT] = ACTIONS(1408),
    [anon_sym_SLASH] = ACTIONS(1408),
    [anon_sym_COLON] = ACTIONS(1408),
    [anon_sym_SEMI] = ACTIONS(1408),
    [anon_sym_EQ] = ACTIONS(1408),
    [anon_sym_QMARK] = ACTIONS(1408),
    [anon_sym_AT] = ACTIONS(1408),
    [anon_sym_BSLASH] = ACTIONS(1408),
    [anon_sym_CARET] = ACTIONS(1408),
    [anon_sym__] = ACTIONS(1408),
    [anon_sym_BQUOTE] = ACTIONS(1408),
    [anon_sym_LBRACE] = ACTIONS(1408),
    [anon_sym_PIPE] = ACTIONS(1408),
    [anon_sym_RBRACE] = ACTIONS(1408),
    [anon_sym_TILDE] = ACTIONS(1408),
    [anon_sym_LPAREN] = ACTIONS(1408),
    [anon_sym_RPAREN] = ACTIONS(1408),
    [aux_sym__word_token1] = ACTIONS(1408),
    [aux_sym__word_token2] = ACTIONS(1408),
    [aux_sym__word_token3] = ACTIONS(1408),
    [sym__whitespace] = ACTIONS(1408),
    [sym__soft_line_ending] = ACTIONS(1408),
    [sym__block_close] = ACTIONS(1408),
    [sym_block_continuation] = ACTIONS(1410),
    [sym__block_quote_start] = ACTIONS(1408),
    [sym__indented_chunk_start] = ACTIONS(1408),
    [sym_atx_h1_marker] = ACTIONS(1408),
    [sym_atx_h2_marker] = ACTIONS(1408),
    [sym_atx_h3_marker] = ACTIONS(1408),
    [sym_atx_h4_marker] = ACTIONS(1408),
    [sym_atx_h5_marker] = ACTIONS(1408),
    [sym_atx_h6_marker] = ACTIONS(1408),
    [sym__thematic_break] = ACTIONS(1408),
    [sym__list_marker_minus] = ACTIONS(1408),
    [sym__list_marker_plus] = ACTIONS(1408),
    [sym__list_marker_star] = ACTIONS(1408),
    [sym__list_marker_parenthesis] = ACTIONS(1408),
    [sym__list_marker_dot] = ACTIONS(1408),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1408),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1408),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1408),
    [sym__blank_line_start] = ACTIONS(1408),
    [sym__html_block_1_start] = ACTIONS(1408),
    [sym__html_block_2_start] = ACTIONS(1408),
    [sym__html_block_3_start] = ACTIONS(1408),
    [sym__html_block_4_start] = ACTIONS(1408),
    [sym__html_block_5_start] = ACTIONS(1408),
    [sym__html_block_6_start] = ACTIONS(1408),
    [sym__html_block_7_start] = ACTIONS(1408),
    [sym__pipe_table_start] = ACTIONS(1408),
  },
  [STATE(143)] = {
    [anon_sym_LBRACK] = ACTIONS(1412),
    [anon_sym_RBRACK] = ACTIONS(1414),
    [anon_sym_LT] = ACTIONS(1414),
    [anon_sym_GT] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1414),
    [anon_sym_POUND] = ACTIONS(1414),
    [anon_sym_DOLLAR] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1414),
    [anon_sym_SQUOTE] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1414),
    [anon_sym_PLUS] = ACTIONS(1414),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_DASH] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1414),
    [anon_sym_COLON] = ACTIONS(1414),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_EQ] = ACTIONS(1414),
    [anon_sym_QMARK] = ACTIONS(1414),
    [anon_sym_AT] = ACTIONS(1414),
    [anon_sym_BSLASH] = ACTIONS(1414),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym__] = ACTIONS(1414),
    [anon_sym_BQUOTE] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_RPAREN] = ACTIONS(1414),
    [aux_sym__word_token1] = ACTIONS(1414),
    [aux_sym__word_token2] = ACTIONS(1414),
    [aux_sym__word_token3] = ACTIONS(1414),
    [sym__whitespace] = ACTIONS(1414),
    [sym__soft_line_ending] = ACTIONS(1414),
    [sym__block_close] = ACTIONS(1414),
    [sym_block_continuation] = ACTIONS(1416),
    [sym__block_quote_start] = ACTIONS(1414),
    [sym__indented_chunk_start] = ACTIONS(1414),
    [sym_atx_h1_marker] = ACTIONS(1414),
    [sym_atx_h2_marker] = ACTIONS(1414),
    [sym_atx_h3_marker] = ACTIONS(1414),
    [sym_atx_h4_marker] = ACTIONS(1414),
    [sym_atx_h5_marker] = ACTIONS(1414),
    [sym_atx_h6_marker] = ACTIONS(1414),
    [sym__thematic_break] = ACTIONS(1414),
    [sym__list_marker_minus] = ACTIONS(1414),
    [sym__list_marker_plus] = ACTIONS(1414),
    [sym__list_marker_star] = ACTIONS(1414),
    [sym__list_marker_parenthesis] = ACTIONS(1414),
    [sym__list_marker_dot] = ACTIONS(1414),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1414),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1414),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1414),
    [sym__blank_line_start] = ACTIONS(1414),
    [sym__html_block_1_start] = ACTIONS(1414),
    [sym__html_block_2_start] = ACTIONS(1414),
    [sym__html_block_3_start] = ACTIONS(1414),
    [sym__html_block_4_start] = ACTIONS(1414),
    [sym__html_block_5_start] = ACTIONS(1414),
    [sym__html_block_6_start] = ACTIONS(1414),
    [sym__html_block_7_start] = ACTIONS(1414),
    [sym__pipe_table_start] = ACTIONS(1414),
  },
  [STATE(144)] = {
    [anon_sym_LBRACK] = ACTIONS(1418),
    [anon_sym_RBRACK] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_POUND] = ACTIONS(1420),
    [anon_sym_DOLLAR] = ACTIONS(1420),
    [anon_sym_PERCENT] = ACTIONS(1420),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_COLON] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1420),
    [anon_sym_EQ] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_BSLASH] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1420),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_BQUOTE] = ACTIONS(1420),
    [anon_sym_LBRACE] = ACTIONS(1420),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_RBRACE] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1420),
    [anon_sym_RPAREN] = ACTIONS(1420),
    [aux_sym__word_token1] = ACTIONS(1420),
    [aux_sym__word_token2] = ACTIONS(1420),
    [aux_sym__word_token3] = ACTIONS(1420),
    [sym__whitespace] = ACTIONS(1420),
    [sym__soft_line_ending] = ACTIONS(1420),
    [sym__block_close] = ACTIONS(1420),
    [sym_block_continuation] = ACTIONS(1422),
    [sym__block_quote_start] = ACTIONS(1420),
    [sym__indented_chunk_start] = ACTIONS(1420),
    [sym_atx_h1_marker] = ACTIONS(1420),
    [sym_atx_h2_marker] = ACTIONS(1420),
    [sym_atx_h3_marker] = ACTIONS(1420),
    [sym_atx_h4_marker] = ACTIONS(1420),
    [sym_atx_h5_marker] = ACTIONS(1420),
    [sym_atx_h6_marker] = ACTIONS(1420),
    [sym__thematic_break] = ACTIONS(1420),
    [sym__list_marker_minus] = ACTIONS(1420),
    [sym__list_marker_plus] = ACTIONS(1420),
    [sym__list_marker_star] = ACTIONS(1420),
    [sym__list_marker_parenthesis] = ACTIONS(1420),
    [sym__list_marker_dot] = ACTIONS(1420),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1420),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1420),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1420),
    [sym__blank_line_start] = ACTIONS(1420),
    [sym__html_block_1_start] = ACTIONS(1420),
    [sym__html_block_2_start] = ACTIONS(1420),
    [sym__html_block_3_start] = ACTIONS(1420),
    [sym__html_block_4_start] = ACTIONS(1420),
    [sym__html_block_5_start] = ACTIONS(1420),
    [sym__html_block_6_start] = ACTIONS(1420),
    [sym__html_block_7_start] = ACTIONS(1420),
    [sym__pipe_table_start] = ACTIONS(1420),
  },
  [STATE(145)] = {
    [anon_sym_LBRACK] = ACTIONS(1424),
    [anon_sym_RBRACK] = ACTIONS(1426),
    [anon_sym_LT] = ACTIONS(1426),
    [anon_sym_GT] = ACTIONS(1426),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_DQUOTE] = ACTIONS(1426),
    [anon_sym_POUND] = ACTIONS(1426),
    [anon_sym_DOLLAR] = ACTIONS(1426),
    [anon_sym_PERCENT] = ACTIONS(1426),
    [anon_sym_AMP] = ACTIONS(1426),
    [anon_sym_SQUOTE] = ACTIONS(1426),
    [anon_sym_STAR] = ACTIONS(1426),
    [anon_sym_PLUS] = ACTIONS(1426),
    [anon_sym_COMMA] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1426),
    [anon_sym_DOT] = ACTIONS(1426),
    [anon_sym_SLASH] = ACTIONS(1426),
    [anon_sym_COLON] = ACTIONS(1426),
    [anon_sym_SEMI] = ACTIONS(1426),
    [anon_sym_EQ] = ACTIONS(1426),
    [anon_sym_QMARK] = ACTIONS(1426),
    [anon_sym_AT] = ACTIONS(1426),
    [anon_sym_BSLASH] = ACTIONS(1426),
    [anon_sym_CARET] = ACTIONS(1426),
    [anon_sym__] = ACTIONS(1426),
    [anon_sym_BQUOTE] = ACTIONS(1426),
    [anon_sym_LBRACE] = ACTIONS(1426),
    [anon_sym_PIPE] = ACTIONS(1426),
    [anon_sym_RBRACE] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_LPAREN] = ACTIONS(1426),
    [anon_sym_RPAREN] = ACTIONS(1426),
    [aux_sym__word_token1] = ACTIONS(1426),
    [aux_sym__word_token2] = ACTIONS(1426),
    [aux_sym__word_token3] = ACTIONS(1426),
    [sym__whitespace] = ACTIONS(1426),
    [sym__soft_line_ending] = ACTIONS(1426),
    [sym__block_close] = ACTIONS(1426),
    [sym_block_continuation] = ACTIONS(1428),
    [sym__block_quote_start] = ACTIONS(1426),
    [sym__indented_chunk_start] = ACTIONS(1426),
    [sym_atx_h1_marker] = ACTIONS(1426),
    [sym_atx_h2_marker] = ACTIONS(1426),
    [sym_atx_h3_marker] = ACTIONS(1426),
    [sym_atx_h4_marker] = ACTIONS(1426),
    [sym_atx_h5_marker] = ACTIONS(1426),
    [sym_atx_h6_marker] = ACTIONS(1426),
    [sym__thematic_break] = ACTIONS(1426),
    [sym__list_marker_minus] = ACTIONS(1426),
    [sym__list_marker_plus] = ACTIONS(1426),
    [sym__list_marker_star] = ACTIONS(1426),
    [sym__list_marker_parenthesis] = ACTIONS(1426),
    [sym__list_marker_dot] = ACTIONS(1426),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1426),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1426),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1426),
    [sym__blank_line_start] = ACTIONS(1426),
    [sym__html_block_1_start] = ACTIONS(1426),
    [sym__html_block_2_start] = ACTIONS(1426),
    [sym__html_block_3_start] = ACTIONS(1426),
    [sym__html_block_4_start] = ACTIONS(1426),
    [sym__html_block_5_start] = ACTIONS(1426),
    [sym__html_block_6_start] = ACTIONS(1426),
    [sym__html_block_7_start] = ACTIONS(1426),
    [sym__pipe_table_start] = ACTIONS(1426),
  },
  [STATE(146)] = {
    [anon_sym_LBRACK] = ACTIONS(1430),
    [anon_sym_RBRACK] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_POUND] = ACTIONS(1432),
    [anon_sym_DOLLAR] = ACTIONS(1432),
    [anon_sym_PERCENT] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_COLON] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1432),
    [anon_sym_EQ] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_BSLASH] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1432),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_BQUOTE] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1432),
    [anon_sym_RPAREN] = ACTIONS(1432),
    [aux_sym__word_token1] = ACTIONS(1432),
    [aux_sym__word_token2] = ACTIONS(1432),
    [aux_sym__word_token3] = ACTIONS(1432),
    [sym__whitespace] = ACTIONS(1432),
    [sym__soft_line_ending] = ACTIONS(1432),
    [sym__block_close] = ACTIONS(1432),
    [sym_block_continuation] = ACTIONS(1434),
    [sym__block_quote_start] = ACTIONS(1432),
    [sym__indented_chunk_start] = ACTIONS(1432),
    [sym_atx_h1_marker] = ACTIONS(1432),
    [sym_atx_h2_marker] = ACTIONS(1432),
    [sym_atx_h3_marker] = ACTIONS(1432),
    [sym_atx_h4_marker] = ACTIONS(1432),
    [sym_atx_h5_marker] = ACTIONS(1432),
    [sym_atx_h6_marker] = ACTIONS(1432),
    [sym__thematic_break] = ACTIONS(1432),
    [sym__list_marker_minus] = ACTIONS(1432),
    [sym__list_marker_plus] = ACTIONS(1432),
    [sym__list_marker_star] = ACTIONS(1432),
    [sym__list_marker_parenthesis] = ACTIONS(1432),
    [sym__list_marker_dot] = ACTIONS(1432),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1432),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1432),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1432),
    [sym__blank_line_start] = ACTIONS(1432),
    [sym__html_block_1_start] = ACTIONS(1432),
    [sym__html_block_2_start] = ACTIONS(1432),
    [sym__html_block_3_start] = ACTIONS(1432),
    [sym__html_block_4_start] = ACTIONS(1432),
    [sym__html_block_5_start] = ACTIONS(1432),
    [sym__html_block_6_start] = ACTIONS(1432),
    [sym__html_block_7_start] = ACTIONS(1432),
    [sym__pipe_table_start] = ACTIONS(1432),
  },
  [STATE(147)] = {
    [ts_builtin_sym_end] = ACTIONS(1127),
    [anon_sym_LBRACK] = ACTIONS(1129),
    [anon_sym_RBRACK] = ACTIONS(1127),
    [anon_sym_LT] = ACTIONS(1127),
    [anon_sym_GT] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_DQUOTE] = ACTIONS(1127),
    [anon_sym_POUND] = ACTIONS(1127),
    [anon_sym_DOLLAR] = ACTIONS(1127),
    [anon_sym_PERCENT] = ACTIONS(1127),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_SQUOTE] = ACTIONS(1127),
    [anon_sym_STAR] = ACTIONS(1127),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_COMMA] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_DOT] = ACTIONS(1127),
    [anon_sym_SLASH] = ACTIONS(1127),
    [anon_sym_COLON] = ACTIONS(1127),
    [anon_sym_SEMI] = ACTIONS(1127),
    [anon_sym_EQ] = ACTIONS(1127),
    [anon_sym_QMARK] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1127),
    [anon_sym_BSLASH] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym__] = ACTIONS(1127),
    [anon_sym_BQUOTE] = ACTIONS(1127),
    [anon_sym_LBRACE] = ACTIONS(1127),
    [anon_sym_PIPE] = ACTIONS(1127),
    [anon_sym_RBRACE] = ACTIONS(1127),
    [anon_sym_TILDE] = ACTIONS(1127),
    [anon_sym_LPAREN] = ACTIONS(1127),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [aux_sym__word_token1] = ACTIONS(1127),
    [aux_sym__word_token2] = ACTIONS(1127),
    [aux_sym__word_token3] = ACTIONS(1127),
    [sym__whitespace] = ACTIONS(1127),
    [sym__soft_line_ending] = ACTIONS(1127),
    [sym_block_continuation] = ACTIONS(1436),
    [sym__block_quote_start] = ACTIONS(1127),
    [sym__indented_chunk_start] = ACTIONS(1127),
    [sym_atx_h1_marker] = ACTIONS(1127),
    [sym_atx_h2_marker] = ACTIONS(1127),
    [sym_atx_h3_marker] = ACTIONS(1127),
    [sym_atx_h4_marker] = ACTIONS(1127),
    [sym_atx_h5_marker] = ACTIONS(1127),
    [sym_atx_h6_marker] = ACTIONS(1127),
    [sym__thematic_break] = ACTIONS(1127),
    [sym__list_marker_minus] = ACTIONS(1127),
    [sym__list_marker_plus] = ACTIONS(1127),
    [sym__list_marker_star] = ACTIONS(1127),
    [sym__list_marker_parenthesis] = ACTIONS(1127),
    [sym__list_marker_dot] = ACTIONS(1127),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1127),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1127),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1127),
    [sym__blank_line_start] = ACTIONS(1127),
    [sym__html_block_1_start] = ACTIONS(1127),
    [sym__html_block_2_start] = ACTIONS(1127),
    [sym__html_block_3_start] = ACTIONS(1127),
    [sym__html_block_4_start] = ACTIONS(1127),
    [sym__html_block_5_start] = ACTIONS(1127),
    [sym__html_block_6_start] = ACTIONS(1127),
    [sym__html_block_7_start] = ACTIONS(1127),
    [sym__pipe_table_start] = ACTIONS(1127),
  },
  [STATE(148)] = {
    [anon_sym_LBRACK] = ACTIONS(1438),
    [anon_sym_RBRACK] = ACTIONS(1440),
    [anon_sym_LT] = ACTIONS(1440),
    [anon_sym_GT] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_DQUOTE] = ACTIONS(1440),
    [anon_sym_POUND] = ACTIONS(1440),
    [anon_sym_DOLLAR] = ACTIONS(1440),
    [anon_sym_PERCENT] = ACTIONS(1440),
    [anon_sym_AMP] = ACTIONS(1440),
    [anon_sym_SQUOTE] = ACTIONS(1440),
    [anon_sym_STAR] = ACTIONS(1440),
    [anon_sym_PLUS] = ACTIONS(1440),
    [anon_sym_COMMA] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_DOT] = ACTIONS(1440),
    [anon_sym_SLASH] = ACTIONS(1440),
    [anon_sym_COLON] = ACTIONS(1440),
    [anon_sym_SEMI] = ACTIONS(1440),
    [anon_sym_EQ] = ACTIONS(1440),
    [anon_sym_QMARK] = ACTIONS(1440),
    [anon_sym_AT] = ACTIONS(1440),
    [anon_sym_BSLASH] = ACTIONS(1440),
    [anon_sym_CARET] = ACTIONS(1440),
    [anon_sym__] = ACTIONS(1440),
    [anon_sym_BQUOTE] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_PIPE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [anon_sym_TILDE] = ACTIONS(1440),
    [anon_sym_LPAREN] = ACTIONS(1440),
    [anon_sym_RPAREN] = ACTIONS(1440),
    [aux_sym__word_token1] = ACTIONS(1440),
    [aux_sym__word_token2] = ACTIONS(1440),
    [aux_sym__word_token3] = ACTIONS(1440),
    [sym__whitespace] = ACTIONS(1440),
    [sym__soft_line_ending] = ACTIONS(1440),
    [sym__block_close] = ACTIONS(1440),
    [sym_block_continuation] = ACTIONS(1442),
    [sym__block_quote_start] = ACTIONS(1440),
    [sym__indented_chunk_start] = ACTIONS(1440),
    [sym_atx_h1_marker] = ACTIONS(1440),
    [sym_atx_h2_marker] = ACTIONS(1440),
    [sym_atx_h3_marker] = ACTIONS(1440),
    [sym_atx_h4_marker] = ACTIONS(1440),
    [sym_atx_h5_marker] = ACTIONS(1440),
    [sym_atx_h6_marker] = ACTIONS(1440),
    [sym__thematic_break] = ACTIONS(1440),
    [sym__list_marker_minus] = ACTIONS(1440),
    [sym__list_marker_plus] = ACTIONS(1440),
    [sym__list_marker_star] = ACTIONS(1440),
    [sym__list_marker_parenthesis] = ACTIONS(1440),
    [sym__list_marker_dot] = ACTIONS(1440),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1440),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1440),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1440),
    [sym__blank_line_start] = ACTIONS(1440),
    [sym__html_block_1_start] = ACTIONS(1440),
    [sym__html_block_2_start] = ACTIONS(1440),
    [sym__html_block_3_start] = ACTIONS(1440),
    [sym__html_block_4_start] = ACTIONS(1440),
    [sym__html_block_5_start] = ACTIONS(1440),
    [sym__html_block_6_start] = ACTIONS(1440),
    [sym__html_block_7_start] = ACTIONS(1440),
    [sym__pipe_table_start] = ACTIONS(1440),
  },
  [STATE(149)] = {
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_RBRACK] = ACTIONS(1294),
    [anon_sym_LT] = ACTIONS(1294),
    [anon_sym_GT] = ACTIONS(1294),
    [anon_sym_BANG] = ACTIONS(1294),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [anon_sym_POUND] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1294),
    [anon_sym_PERCENT] = ACTIONS(1294),
    [anon_sym_AMP] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_COMMA] = ACTIONS(1294),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_DOT] = ACTIONS(1294),
    [anon_sym_SLASH] = ACTIONS(1294),
    [anon_sym_COLON] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1294),
    [anon_sym_EQ] = ACTIONS(1294),
    [anon_sym_QMARK] = ACTIONS(1294),
    [anon_sym_AT] = ACTIONS(1294),
    [anon_sym_BSLASH] = ACTIONS(1294),
    [anon_sym_CARET] = ACTIONS(1294),
    [anon_sym__] = ACTIONS(1294),
    [anon_sym_BQUOTE] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1294),
    [anon_sym_PIPE] = ACTIONS(1294),
    [anon_sym_RBRACE] = ACTIONS(1294),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(1294),
    [anon_sym_RPAREN] = ACTIONS(1294),
    [aux_sym__word_token1] = ACTIONS(1294),
    [aux_sym__word_token2] = ACTIONS(1294),
    [aux_sym__word_token3] = ACTIONS(1294),
    [sym__whitespace] = ACTIONS(1294),
    [sym__soft_line_ending] = ACTIONS(1294),
    [sym__block_close] = ACTIONS(1294),
    [sym_block_continuation] = ACTIONS(1444),
    [sym__block_quote_start] = ACTIONS(1294),
    [sym__indented_chunk_start] = ACTIONS(1294),
    [sym_atx_h1_marker] = ACTIONS(1294),
    [sym_atx_h2_marker] = ACTIONS(1294),
    [sym_atx_h3_marker] = ACTIONS(1294),
    [sym_atx_h4_marker] = ACTIONS(1294),
    [sym_atx_h5_marker] = ACTIONS(1294),
    [sym_atx_h6_marker] = ACTIONS(1294),
    [sym__thematic_break] = ACTIONS(1294),
    [sym__list_marker_minus] = ACTIONS(1294),
    [sym__list_marker_plus] = ACTIONS(1294),
    [sym__list_marker_star] = ACTIONS(1294),
    [sym__list_marker_parenthesis] = ACTIONS(1294),
    [sym__list_marker_dot] = ACTIONS(1294),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1294),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1294),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1294),
    [sym__blank_line_start] = ACTIONS(1294),
    [sym__html_block_1_start] = ACTIONS(1294),
    [sym__html_block_2_start] = ACTIONS(1294),
    [sym__html_block_3_start] = ACTIONS(1294),
    [sym__html_block_4_start] = ACTIONS(1294),
    [sym__html_block_5_start] = ACTIONS(1294),
    [sym__html_block_6_start] = ACTIONS(1294),
    [sym__html_block_7_start] = ACTIONS(1294),
    [sym__pipe_table_start] = ACTIONS(1294),
  },
  [STATE(150)] = {
    [anon_sym_LBRACK] = ACTIONS(1302),
    [anon_sym_RBRACK] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [anon_sym_POUND] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_AT] = ACTIONS(1300),
    [anon_sym_BSLASH] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym__] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_RPAREN] = ACTIONS(1300),
    [aux_sym__word_token1] = ACTIONS(1300),
    [aux_sym__word_token2] = ACTIONS(1300),
    [aux_sym__word_token3] = ACTIONS(1300),
    [sym__whitespace] = ACTIONS(1300),
    [sym__soft_line_ending] = ACTIONS(1300),
    [sym__block_close] = ACTIONS(1300),
    [sym_block_continuation] = ACTIONS(1446),
    [sym__block_quote_start] = ACTIONS(1300),
    [sym__indented_chunk_start] = ACTIONS(1300),
    [sym_atx_h1_marker] = ACTIONS(1300),
    [sym_atx_h2_marker] = ACTIONS(1300),
    [sym_atx_h3_marker] = ACTIONS(1300),
    [sym_atx_h4_marker] = ACTIONS(1300),
    [sym_atx_h5_marker] = ACTIONS(1300),
    [sym_atx_h6_marker] = ACTIONS(1300),
    [sym__thematic_break] = ACTIONS(1300),
    [sym__list_marker_minus] = ACTIONS(1300),
    [sym__list_marker_plus] = ACTIONS(1300),
    [sym__list_marker_star] = ACTIONS(1300),
    [sym__list_marker_parenthesis] = ACTIONS(1300),
    [sym__list_marker_dot] = ACTIONS(1300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1300),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1300),
    [sym__blank_line_start] = ACTIONS(1300),
    [sym__html_block_1_start] = ACTIONS(1300),
    [sym__html_block_2_start] = ACTIONS(1300),
    [sym__html_block_3_start] = ACTIONS(1300),
    [sym__html_block_4_start] = ACTIONS(1300),
    [sym__html_block_5_start] = ACTIONS(1300),
    [sym__html_block_6_start] = ACTIONS(1300),
    [sym__html_block_7_start] = ACTIONS(1300),
    [sym__pipe_table_start] = ACTIONS(1300),
  },
  [STATE(151)] = {
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_RBRACK] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1318),
    [anon_sym_GT] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1318),
    [anon_sym_BSLASH] = ACTIONS(1318),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym__] = ACTIONS(1318),
    [anon_sym_BQUOTE] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_RPAREN] = ACTIONS(1318),
    [aux_sym__word_token1] = ACTIONS(1318),
    [aux_sym__word_token2] = ACTIONS(1318),
    [aux_sym__word_token3] = ACTIONS(1318),
    [sym__whitespace] = ACTIONS(1318),
    [sym__soft_line_ending] = ACTIONS(1318),
    [sym__block_close] = ACTIONS(1318),
    [sym_block_continuation] = ACTIONS(1448),
    [sym__block_quote_start] = ACTIONS(1318),
    [sym__indented_chunk_start] = ACTIONS(1318),
    [sym_atx_h1_marker] = ACTIONS(1318),
    [sym_atx_h2_marker] = ACTIONS(1318),
    [sym_atx_h3_marker] = ACTIONS(1318),
    [sym_atx_h4_marker] = ACTIONS(1318),
    [sym_atx_h5_marker] = ACTIONS(1318),
    [sym_atx_h6_marker] = ACTIONS(1318),
    [sym__thematic_break] = ACTIONS(1318),
    [sym__list_marker_minus] = ACTIONS(1318),
    [sym__list_marker_plus] = ACTIONS(1318),
    [sym__list_marker_star] = ACTIONS(1318),
    [sym__list_marker_parenthesis] = ACTIONS(1318),
    [sym__list_marker_dot] = ACTIONS(1318),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1318),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1318),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1318),
    [sym__blank_line_start] = ACTIONS(1318),
    [sym__html_block_1_start] = ACTIONS(1318),
    [sym__html_block_2_start] = ACTIONS(1318),
    [sym__html_block_3_start] = ACTIONS(1318),
    [sym__html_block_4_start] = ACTIONS(1318),
    [sym__html_block_5_start] = ACTIONS(1318),
    [sym__html_block_6_start] = ACTIONS(1318),
    [sym__html_block_7_start] = ACTIONS(1318),
    [sym__pipe_table_start] = ACTIONS(1318),
  },
  [STATE(152)] = {
    [anon_sym_LBRACK] = ACTIONS(1326),
    [anon_sym_RBRACK] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_POUND] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_COLON] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_BSLASH] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_RPAREN] = ACTIONS(1324),
    [aux_sym__word_token1] = ACTIONS(1324),
    [aux_sym__word_token2] = ACTIONS(1324),
    [aux_sym__word_token3] = ACTIONS(1324),
    [sym__whitespace] = ACTIONS(1324),
    [sym__soft_line_ending] = ACTIONS(1324),
    [sym__block_close] = ACTIONS(1324),
    [sym_block_continuation] = ACTIONS(1450),
    [sym__block_quote_start] = ACTIONS(1324),
    [sym__indented_chunk_start] = ACTIONS(1324),
    [sym_atx_h1_marker] = ACTIONS(1324),
    [sym_atx_h2_marker] = ACTIONS(1324),
    [sym_atx_h3_marker] = ACTIONS(1324),
    [sym_atx_h4_marker] = ACTIONS(1324),
    [sym_atx_h5_marker] = ACTIONS(1324),
    [sym_atx_h6_marker] = ACTIONS(1324),
    [sym__thematic_break] = ACTIONS(1324),
    [sym__list_marker_minus] = ACTIONS(1324),
    [sym__list_marker_plus] = ACTIONS(1324),
    [sym__list_marker_star] = ACTIONS(1324),
    [sym__list_marker_parenthesis] = ACTIONS(1324),
    [sym__list_marker_dot] = ACTIONS(1324),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1324),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1324),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1324),
    [sym__blank_line_start] = ACTIONS(1324),
    [sym__html_block_1_start] = ACTIONS(1324),
    [sym__html_block_2_start] = ACTIONS(1324),
    [sym__html_block_3_start] = ACTIONS(1324),
    [sym__html_block_4_start] = ACTIONS(1324),
    [sym__html_block_5_start] = ACTIONS(1324),
    [sym__html_block_6_start] = ACTIONS(1324),
    [sym__html_block_7_start] = ACTIONS(1324),
    [sym__pipe_table_start] = ACTIONS(1324),
  },
  [STATE(153)] = {
    [anon_sym_LBRACK] = ACTIONS(1368),
    [anon_sym_RBRACK] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_POUND] = ACTIONS(1366),
    [anon_sym_DOLLAR] = ACTIONS(1366),
    [anon_sym_PERCENT] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_COLON] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym_EQ] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_BSLASH] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1366),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_BQUOTE] = ACTIONS(1366),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_RBRACE] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1366),
    [anon_sym_RPAREN] = ACTIONS(1366),
    [aux_sym__word_token1] = ACTIONS(1366),
    [aux_sym__word_token2] = ACTIONS(1366),
    [aux_sym__word_token3] = ACTIONS(1366),
    [sym__whitespace] = ACTIONS(1366),
    [sym__soft_line_ending] = ACTIONS(1366),
    [sym__block_close] = ACTIONS(1366),
    [sym_block_continuation] = ACTIONS(1452),
    [sym__block_quote_start] = ACTIONS(1366),
    [sym__indented_chunk_start] = ACTIONS(1366),
    [sym_atx_h1_marker] = ACTIONS(1366),
    [sym_atx_h2_marker] = ACTIONS(1366),
    [sym_atx_h3_marker] = ACTIONS(1366),
    [sym_atx_h4_marker] = ACTIONS(1366),
    [sym_atx_h5_marker] = ACTIONS(1366),
    [sym_atx_h6_marker] = ACTIONS(1366),
    [sym__thematic_break] = ACTIONS(1366),
    [sym__list_marker_minus] = ACTIONS(1366),
    [sym__list_marker_plus] = ACTIONS(1366),
    [sym__list_marker_star] = ACTIONS(1366),
    [sym__list_marker_parenthesis] = ACTIONS(1366),
    [sym__list_marker_dot] = ACTIONS(1366),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1366),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1366),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1366),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1366),
    [sym__blank_line_start] = ACTIONS(1366),
    [sym__html_block_1_start] = ACTIONS(1366),
    [sym__html_block_2_start] = ACTIONS(1366),
    [sym__html_block_3_start] = ACTIONS(1366),
    [sym__html_block_4_start] = ACTIONS(1366),
    [sym__html_block_5_start] = ACTIONS(1366),
    [sym__html_block_6_start] = ACTIONS(1366),
    [sym__html_block_7_start] = ACTIONS(1366),
    [sym__pipe_table_start] = ACTIONS(1366),
  },
  [STATE(154)] = {
    [anon_sym_LBRACK] = ACTIONS(1454),
    [anon_sym_RBRACK] = ACTIONS(1456),
    [anon_sym_LT] = ACTIONS(1456),
    [anon_sym_GT] = ACTIONS(1456),
    [anon_sym_BANG] = ACTIONS(1456),
    [anon_sym_DQUOTE] = ACTIONS(1456),
    [anon_sym_POUND] = ACTIONS(1456),
    [anon_sym_DOLLAR] = ACTIONS(1456),
    [anon_sym_PERCENT] = ACTIONS(1456),
    [anon_sym_AMP] = ACTIONS(1456),
    [anon_sym_SQUOTE] = ACTIONS(1456),
    [anon_sym_STAR] = ACTIONS(1456),
    [anon_sym_PLUS] = ACTIONS(1456),
    [anon_sym_COMMA] = ACTIONS(1456),
    [anon_sym_DASH] = ACTIONS(1456),
    [anon_sym_DOT] = ACTIONS(1456),
    [anon_sym_SLASH] = ACTIONS(1456),
    [anon_sym_COLON] = ACTIONS(1456),
    [anon_sym_SEMI] = ACTIONS(1456),
    [anon_sym_EQ] = ACTIONS(1456),
    [anon_sym_QMARK] = ACTIONS(1456),
    [anon_sym_AT] = ACTIONS(1456),
    [anon_sym_BSLASH] = ACTIONS(1456),
    [anon_sym_CARET] = ACTIONS(1456),
    [anon_sym__] = ACTIONS(1456),
    [anon_sym_BQUOTE] = ACTIONS(1456),
    [anon_sym_LBRACE] = ACTIONS(1456),
    [anon_sym_PIPE] = ACTIONS(1456),
    [anon_sym_RBRACE] = ACTIONS(1456),
    [anon_sym_TILDE] = ACTIONS(1456),
    [anon_sym_LPAREN] = ACTIONS(1456),
    [anon_sym_RPAREN] = ACTIONS(1456),
    [aux_sym__word_token1] = ACTIONS(1456),
    [aux_sym__word_token2] = ACTIONS(1456),
    [aux_sym__word_token3] = ACTIONS(1456),
    [sym__whitespace] = ACTIONS(1456),
    [sym__soft_line_ending] = ACTIONS(1456),
    [sym__block_close] = ACTIONS(1456),
    [sym_block_continuation] = ACTIONS(1458),
    [sym__block_quote_start] = ACTIONS(1456),
    [sym__indented_chunk_start] = ACTIONS(1456),
    [sym_atx_h1_marker] = ACTIONS(1456),
    [sym_atx_h2_marker] = ACTIONS(1456),
    [sym_atx_h3_marker] = ACTIONS(1456),
    [sym_atx_h4_marker] = ACTIONS(1456),
    [sym_atx_h5_marker] = ACTIONS(1456),
    [sym_atx_h6_marker] = ACTIONS(1456),
    [sym__thematic_break] = ACTIONS(1456),
    [sym__list_marker_minus] = ACTIONS(1456),
    [sym__list_marker_plus] = ACTIONS(1456),
    [sym__list_marker_star] = ACTIONS(1456),
    [sym__list_marker_parenthesis] = ACTIONS(1456),
    [sym__list_marker_dot] = ACTIONS(1456),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1456),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1456),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1456),
    [sym__blank_line_start] = ACTIONS(1456),
    [sym__html_block_1_start] = ACTIONS(1456),
    [sym__html_block_2_start] = ACTIONS(1456),
    [sym__html_block_3_start] = ACTIONS(1456),
    [sym__html_block_4_start] = ACTIONS(1456),
    [sym__html_block_5_start] = ACTIONS(1456),
    [sym__html_block_6_start] = ACTIONS(1456),
    [sym__html_block_7_start] = ACTIONS(1456),
    [sym__pipe_table_start] = ACTIONS(1456),
  },
  [STATE(155)] = {
    [ts_builtin_sym_end] = ACTIONS(1426),
    [anon_sym_LBRACK] = ACTIONS(1424),
    [anon_sym_RBRACK] = ACTIONS(1426),
    [anon_sym_LT] = ACTIONS(1426),
    [anon_sym_GT] = ACTIONS(1426),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_DQUOTE] = ACTIONS(1426),
    [anon_sym_POUND] = ACTIONS(1426),
    [anon_sym_DOLLAR] = ACTIONS(1426),
    [anon_sym_PERCENT] = ACTIONS(1426),
    [anon_sym_AMP] = ACTIONS(1426),
    [anon_sym_SQUOTE] = ACTIONS(1426),
    [anon_sym_STAR] = ACTIONS(1426),
    [anon_sym_PLUS] = ACTIONS(1426),
    [anon_sym_COMMA] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1426),
    [anon_sym_DOT] = ACTIONS(1426),
    [anon_sym_SLASH] = ACTIONS(1426),
    [anon_sym_COLON] = ACTIONS(1426),
    [anon_sym_SEMI] = ACTIONS(1426),
    [anon_sym_EQ] = ACTIONS(1426),
    [anon_sym_QMARK] = ACTIONS(1426),
    [anon_sym_AT] = ACTIONS(1426),
    [anon_sym_BSLASH] = ACTIONS(1426),
    [anon_sym_CARET] = ACTIONS(1426),
    [anon_sym__] = ACTIONS(1426),
    [anon_sym_BQUOTE] = ACTIONS(1426),
    [anon_sym_LBRACE] = ACTIONS(1426),
    [anon_sym_PIPE] = ACTIONS(1426),
    [anon_sym_RBRACE] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_LPAREN] = ACTIONS(1426),
    [anon_sym_RPAREN] = ACTIONS(1426),
    [aux_sym__word_token1] = ACTIONS(1426),
    [aux_sym__word_token2] = ACTIONS(1426),
    [aux_sym__word_token3] = ACTIONS(1426),
    [sym__whitespace] = ACTIONS(1426),
    [sym__soft_line_ending] = ACTIONS(1426),
    [sym_block_continuation] = ACTIONS(1460),
    [sym__block_quote_start] = ACTIONS(1426),
    [sym__indented_chunk_start] = ACTIONS(1426),
    [sym_atx_h1_marker] = ACTIONS(1426),
    [sym_atx_h2_marker] = ACTIONS(1426),
    [sym_atx_h3_marker] = ACTIONS(1426),
    [sym_atx_h4_marker] = ACTIONS(1426),
    [sym_atx_h5_marker] = ACTIONS(1426),
    [sym_atx_h6_marker] = ACTIONS(1426),
    [sym__thematic_break] = ACTIONS(1426),
    [sym__list_marker_minus] = ACTIONS(1426),
    [sym__list_marker_plus] = ACTIONS(1426),
    [sym__list_marker_star] = ACTIONS(1426),
    [sym__list_marker_parenthesis] = ACTIONS(1426),
    [sym__list_marker_dot] = ACTIONS(1426),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1426),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1426),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1426),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1426),
    [sym__blank_line_start] = ACTIONS(1426),
    [sym__html_block_1_start] = ACTIONS(1426),
    [sym__html_block_2_start] = ACTIONS(1426),
    [sym__html_block_3_start] = ACTIONS(1426),
    [sym__html_block_4_start] = ACTIONS(1426),
    [sym__html_block_5_start] = ACTIONS(1426),
    [sym__html_block_6_start] = ACTIONS(1426),
    [sym__html_block_7_start] = ACTIONS(1426),
    [sym__pipe_table_start] = ACTIONS(1426),
  },
  [STATE(156)] = {
    [anon_sym_LBRACK] = ACTIONS(1462),
    [anon_sym_RBRACK] = ACTIONS(1464),
    [anon_sym_LT] = ACTIONS(1464),
    [anon_sym_GT] = ACTIONS(1464),
    [anon_sym_BANG] = ACTIONS(1464),
    [anon_sym_DQUOTE] = ACTIONS(1464),
    [anon_sym_POUND] = ACTIONS(1464),
    [anon_sym_DOLLAR] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1464),
    [anon_sym_AMP] = ACTIONS(1464),
    [anon_sym_SQUOTE] = ACTIONS(1464),
    [anon_sym_STAR] = ACTIONS(1464),
    [anon_sym_PLUS] = ACTIONS(1464),
    [anon_sym_COMMA] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_DOT] = ACTIONS(1464),
    [anon_sym_SLASH] = ACTIONS(1464),
    [anon_sym_COLON] = ACTIONS(1464),
    [anon_sym_SEMI] = ACTIONS(1464),
    [anon_sym_EQ] = ACTIONS(1464),
    [anon_sym_QMARK] = ACTIONS(1464),
    [anon_sym_AT] = ACTIONS(1464),
    [anon_sym_BSLASH] = ACTIONS(1464),
    [anon_sym_CARET] = ACTIONS(1464),
    [anon_sym__] = ACTIONS(1464),
    [anon_sym_BQUOTE] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [anon_sym_PIPE] = ACTIONS(1464),
    [anon_sym_RBRACE] = ACTIONS(1464),
    [anon_sym_TILDE] = ACTIONS(1464),
    [anon_sym_LPAREN] = ACTIONS(1464),
    [anon_sym_RPAREN] = ACTIONS(1464),
    [aux_sym__word_token1] = ACTIONS(1464),
    [aux_sym__word_token2] = ACTIONS(1464),
    [aux_sym__word_token3] = ACTIONS(1464),
    [sym__whitespace] = ACTIONS(1464),
    [sym__soft_line_ending] = ACTIONS(1464),
    [sym__block_close] = ACTIONS(1464),
    [sym_block_continuation] = ACTIONS(1466),
    [sym__block_quote_start] = ACTIONS(1464),
    [sym__indented_chunk_start] = ACTIONS(1464),
    [sym_atx_h1_marker] = ACTIONS(1464),
    [sym_atx_h2_marker] = ACTIONS(1464),
    [sym_atx_h3_marker] = ACTIONS(1464),
    [sym_atx_h4_marker] = ACTIONS(1464),
    [sym_atx_h5_marker] = ACTIONS(1464),
    [sym_atx_h6_marker] = ACTIONS(1464),
    [sym__thematic_break] = ACTIONS(1464),
    [sym__list_marker_minus] = ACTIONS(1464),
    [sym__list_marker_plus] = ACTIONS(1464),
    [sym__list_marker_star] = ACTIONS(1464),
    [sym__list_marker_parenthesis] = ACTIONS(1464),
    [sym__list_marker_dot] = ACTIONS(1464),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1464),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1464),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1464),
    [sym__blank_line_start] = ACTIONS(1464),
    [sym__html_block_1_start] = ACTIONS(1464),
    [sym__html_block_2_start] = ACTIONS(1464),
    [sym__html_block_3_start] = ACTIONS(1464),
    [sym__html_block_4_start] = ACTIONS(1464),
    [sym__html_block_5_start] = ACTIONS(1464),
    [sym__html_block_6_start] = ACTIONS(1464),
    [sym__html_block_7_start] = ACTIONS(1464),
    [sym__pipe_table_start] = ACTIONS(1464),
  },
  [STATE(157)] = {
    [anon_sym_LBRACK] = ACTIONS(1308),
    [anon_sym_RBRACK] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_POUND] = ACTIONS(1306),
    [anon_sym_DOLLAR] = ACTIONS(1306),
    [anon_sym_PERCENT] = ACTIONS(1306),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_COLON] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1306),
    [anon_sym_EQ] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_BSLASH] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1306),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_BQUOTE] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1306),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_RBRACE] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1306),
    [anon_sym_RPAREN] = ACTIONS(1306),
    [aux_sym__word_token1] = ACTIONS(1306),
    [aux_sym__word_token2] = ACTIONS(1306),
    [aux_sym__word_token3] = ACTIONS(1306),
    [sym__whitespace] = ACTIONS(1306),
    [sym__soft_line_ending] = ACTIONS(1306),
    [sym__block_close] = ACTIONS(1306),
    [sym_block_continuation] = ACTIONS(1468),
    [sym__block_quote_start] = ACTIONS(1306),
    [sym__indented_chunk_start] = ACTIONS(1306),
    [sym_atx_h1_marker] = ACTIONS(1306),
    [sym_atx_h2_marker] = ACTIONS(1306),
    [sym_atx_h3_marker] = ACTIONS(1306),
    [sym_atx_h4_marker] = ACTIONS(1306),
    [sym_atx_h5_marker] = ACTIONS(1306),
    [sym_atx_h6_marker] = ACTIONS(1306),
    [sym__thematic_break] = ACTIONS(1306),
    [sym__list_marker_minus] = ACTIONS(1306),
    [sym__list_marker_plus] = ACTIONS(1306),
    [sym__list_marker_star] = ACTIONS(1306),
    [sym__list_marker_parenthesis] = ACTIONS(1306),
    [sym__list_marker_dot] = ACTIONS(1306),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1306),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1306),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1306),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1306),
    [sym__blank_line_start] = ACTIONS(1306),
    [sym__html_block_1_start] = ACTIONS(1306),
    [sym__html_block_2_start] = ACTIONS(1306),
    [sym__html_block_3_start] = ACTIONS(1306),
    [sym__html_block_4_start] = ACTIONS(1306),
    [sym__html_block_5_start] = ACTIONS(1306),
    [sym__html_block_6_start] = ACTIONS(1306),
    [sym__html_block_7_start] = ACTIONS(1306),
    [sym__pipe_table_start] = ACTIONS(1306),
  },
  [STATE(158)] = {
    [ts_builtin_sym_end] = ACTIONS(1470),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_RBRACK] = ACTIONS(1470),
    [anon_sym_LT] = ACTIONS(1470),
    [anon_sym_GT] = ACTIONS(1470),
    [anon_sym_BANG] = ACTIONS(1470),
    [anon_sym_DQUOTE] = ACTIONS(1470),
    [anon_sym_POUND] = ACTIONS(1470),
    [anon_sym_DOLLAR] = ACTIONS(1470),
    [anon_sym_PERCENT] = ACTIONS(1470),
    [anon_sym_AMP] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_STAR] = ACTIONS(1470),
    [anon_sym_PLUS] = ACTIONS(1470),
    [anon_sym_COMMA] = ACTIONS(1470),
    [anon_sym_DASH] = ACTIONS(1470),
    [anon_sym_DOT] = ACTIONS(1470),
    [anon_sym_SLASH] = ACTIONS(1470),
    [anon_sym_COLON] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1470),
    [anon_sym_EQ] = ACTIONS(1470),
    [anon_sym_QMARK] = ACTIONS(1470),
    [anon_sym_AT] = ACTIONS(1470),
    [anon_sym_BSLASH] = ACTIONS(1470),
    [anon_sym_CARET] = ACTIONS(1470),
    [anon_sym__] = ACTIONS(1470),
    [anon_sym_BQUOTE] = ACTIONS(1470),
    [anon_sym_LBRACE] = ACTIONS(1470),
    [anon_sym_PIPE] = ACTIONS(1470),
    [anon_sym_RBRACE] = ACTIONS(1470),
    [anon_sym_TILDE] = ACTIONS(1470),
    [anon_sym_LPAREN] = ACTIONS(1470),
    [anon_sym_RPAREN] = ACTIONS(1470),
    [aux_sym__word_token1] = ACTIONS(1470),
    [aux_sym__word_token2] = ACTIONS(1470),
    [aux_sym__word_token3] = ACTIONS(1470),
    [sym__whitespace] = ACTIONS(1470),
    [sym__soft_line_ending] = ACTIONS(1470),
    [sym__block_quote_start] = ACTIONS(1470),
    [sym__indented_chunk_start] = ACTIONS(1470),
    [sym_atx_h1_marker] = ACTIONS(1470),
    [sym_atx_h2_marker] = ACTIONS(1470),
    [sym_atx_h3_marker] = ACTIONS(1470),
    [sym_atx_h4_marker] = ACTIONS(1470),
    [sym_atx_h5_marker] = ACTIONS(1470),
    [sym_atx_h6_marker] = ACTIONS(1470),
    [sym__thematic_break] = ACTIONS(1470),
    [sym__list_marker_minus] = ACTIONS(1470),
    [sym__list_marker_plus] = ACTIONS(1470),
    [sym__list_marker_star] = ACTIONS(1470),
    [sym__list_marker_parenthesis] = ACTIONS(1470),
    [sym__list_marker_dot] = ACTIONS(1470),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1470),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1470),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1470),
    [sym__blank_line_start] = ACTIONS(1470),
    [sym__html_block_1_start] = ACTIONS(1470),
    [sym__html_block_2_start] = ACTIONS(1470),
    [sym__html_block_3_start] = ACTIONS(1470),
    [sym__html_block_4_start] = ACTIONS(1470),
    [sym__html_block_5_start] = ACTIONS(1470),
    [sym__html_block_6_start] = ACTIONS(1470),
    [sym__html_block_7_start] = ACTIONS(1470),
    [sym__no_indented_chunk] = ACTIONS(1470),
    [sym__pipe_table_start] = ACTIONS(1470),
  },
  [STATE(159)] = {
    [anon_sym_LBRACK] = ACTIONS(1314),
    [anon_sym_RBRACK] = ACTIONS(1312),
    [anon_sym_LT] = ACTIONS(1312),
    [anon_sym_GT] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [anon_sym_POUND] = ACTIONS(1312),
    [anon_sym_DOLLAR] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_PLUS] = ACTIONS(1312),
    [anon_sym_COMMA] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_DOT] = ACTIONS(1312),
    [anon_sym_SLASH] = ACTIONS(1312),
    [anon_sym_COLON] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym_EQ] = ACTIONS(1312),
    [anon_sym_QMARK] = ACTIONS(1312),
    [anon_sym_AT] = ACTIONS(1312),
    [anon_sym_BSLASH] = ACTIONS(1312),
    [anon_sym_CARET] = ACTIONS(1312),
    [anon_sym__] = ACTIONS(1312),
    [anon_sym_BQUOTE] = ACTIONS(1312),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_PIPE] = ACTIONS(1312),
    [anon_sym_RBRACE] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_LPAREN] = ACTIONS(1312),
    [anon_sym_RPAREN] = ACTIONS(1312),
    [aux_sym__word_token1] = ACTIONS(1312),
    [aux_sym__word_token2] = ACTIONS(1312),
    [aux_sym__word_token3] = ACTIONS(1312),
    [sym__whitespace] = ACTIONS(1312),
    [sym__soft_line_ending] = ACTIONS(1312),
    [sym__block_close] = ACTIONS(1312),
    [sym_block_continuation] = ACTIONS(1474),
    [sym__block_quote_start] = ACTIONS(1312),
    [sym__indented_chunk_start] = ACTIONS(1312),
    [sym_atx_h1_marker] = ACTIONS(1312),
    [sym_atx_h2_marker] = ACTIONS(1312),
    [sym_atx_h3_marker] = ACTIONS(1312),
    [sym_atx_h4_marker] = ACTIONS(1312),
    [sym_atx_h5_marker] = ACTIONS(1312),
    [sym_atx_h6_marker] = ACTIONS(1312),
    [sym__thematic_break] = ACTIONS(1312),
    [sym__list_marker_minus] = ACTIONS(1312),
    [sym__list_marker_plus] = ACTIONS(1312),
    [sym__list_marker_star] = ACTIONS(1312),
    [sym__list_marker_parenthesis] = ACTIONS(1312),
    [sym__list_marker_dot] = ACTIONS(1312),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1312),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1312),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1312),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1312),
    [sym__blank_line_start] = ACTIONS(1312),
    [sym__html_block_1_start] = ACTIONS(1312),
    [sym__html_block_2_start] = ACTIONS(1312),
    [sym__html_block_3_start] = ACTIONS(1312),
    [sym__html_block_4_start] = ACTIONS(1312),
    [sym__html_block_5_start] = ACTIONS(1312),
    [sym__html_block_6_start] = ACTIONS(1312),
    [sym__html_block_7_start] = ACTIONS(1312),
    [sym__pipe_table_start] = ACTIONS(1312),
  },
  [STATE(160)] = {
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_RBRACK] = ACTIONS(1470),
    [anon_sym_LT] = ACTIONS(1470),
    [anon_sym_GT] = ACTIONS(1470),
    [anon_sym_BANG] = ACTIONS(1470),
    [anon_sym_DQUOTE] = ACTIONS(1470),
    [anon_sym_POUND] = ACTIONS(1470),
    [anon_sym_DOLLAR] = ACTIONS(1470),
    [anon_sym_PERCENT] = ACTIONS(1470),
    [anon_sym_AMP] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_STAR] = ACTIONS(1470),
    [anon_sym_PLUS] = ACTIONS(1470),
    [anon_sym_COMMA] = ACTIONS(1470),
    [anon_sym_DASH] = ACTIONS(1470),
    [anon_sym_DOT] = ACTIONS(1470),
    [anon_sym_SLASH] = ACTIONS(1470),
    [anon_sym_COLON] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1470),
    [anon_sym_EQ] = ACTIONS(1470),
    [anon_sym_QMARK] = ACTIONS(1470),
    [anon_sym_AT] = ACTIONS(1470),
    [anon_sym_BSLASH] = ACTIONS(1470),
    [anon_sym_CARET] = ACTIONS(1470),
    [anon_sym__] = ACTIONS(1470),
    [anon_sym_BQUOTE] = ACTIONS(1470),
    [anon_sym_LBRACE] = ACTIONS(1470),
    [anon_sym_PIPE] = ACTIONS(1470),
    [anon_sym_RBRACE] = ACTIONS(1470),
    [anon_sym_TILDE] = ACTIONS(1470),
    [anon_sym_LPAREN] = ACTIONS(1470),
    [anon_sym_RPAREN] = ACTIONS(1470),
    [aux_sym__word_token1] = ACTIONS(1470),
    [aux_sym__word_token2] = ACTIONS(1470),
    [aux_sym__word_token3] = ACTIONS(1470),
    [sym__whitespace] = ACTIONS(1470),
    [sym__soft_line_ending] = ACTIONS(1470),
    [sym__block_close] = ACTIONS(1470),
    [sym__block_quote_start] = ACTIONS(1470),
    [sym__indented_chunk_start] = ACTIONS(1470),
    [sym_atx_h1_marker] = ACTIONS(1470),
    [sym_atx_h2_marker] = ACTIONS(1470),
    [sym_atx_h3_marker] = ACTIONS(1470),
    [sym_atx_h4_marker] = ACTIONS(1470),
    [sym_atx_h5_marker] = ACTIONS(1470),
    [sym_atx_h6_marker] = ACTIONS(1470),
    [sym__thematic_break] = ACTIONS(1470),
    [sym__list_marker_minus] = ACTIONS(1470),
    [sym__list_marker_plus] = ACTIONS(1470),
    [sym__list_marker_star] = ACTIONS(1470),
    [sym__list_marker_parenthesis] = ACTIONS(1470),
    [sym__list_marker_dot] = ACTIONS(1470),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1470),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1470),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1470),
    [sym__blank_line_start] = ACTIONS(1470),
    [sym__html_block_1_start] = ACTIONS(1470),
    [sym__html_block_2_start] = ACTIONS(1470),
    [sym__html_block_3_start] = ACTIONS(1470),
    [sym__html_block_4_start] = ACTIONS(1470),
    [sym__html_block_5_start] = ACTIONS(1470),
    [sym__html_block_6_start] = ACTIONS(1470),
    [sym__html_block_7_start] = ACTIONS(1470),
    [sym__no_indented_chunk] = ACTIONS(1470),
    [sym__pipe_table_start] = ACTIONS(1470),
  },
  [STATE(161)] = {
    [ts_builtin_sym_end] = ACTIONS(1240),
    [anon_sym_LBRACK] = ACTIONS(1242),
    [anon_sym_RBRACK] = ACTIONS(1240),
    [anon_sym_LT] = ACTIONS(1240),
    [anon_sym_GT] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [anon_sym_POUND] = ACTIONS(1240),
    [anon_sym_DOLLAR] = ACTIONS(1240),
    [anon_sym_PERCENT] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_PLUS] = ACTIONS(1240),
    [anon_sym_COMMA] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_DOT] = ACTIONS(1240),
    [anon_sym_SLASH] = ACTIONS(1240),
    [anon_sym_COLON] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym_EQ] = ACTIONS(1240),
    [anon_sym_QMARK] = ACTIONS(1240),
    [anon_sym_AT] = ACTIONS(1240),
    [anon_sym_BSLASH] = ACTIONS(1240),
    [anon_sym_CARET] = ACTIONS(1240),
    [anon_sym__] = ACTIONS(1240),
    [anon_sym_BQUOTE] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_PIPE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_LPAREN] = ACTIONS(1240),
    [anon_sym_RPAREN] = ACTIONS(1240),
    [aux_sym__word_token1] = ACTIONS(1240),
    [aux_sym__word_token2] = ACTIONS(1240),
    [aux_sym__word_token3] = ACTIONS(1240),
    [sym__whitespace] = ACTIONS(1240),
    [sym__soft_line_ending] = ACTIONS(1240),
    [sym_block_continuation] = ACTIONS(1476),
    [sym__block_quote_start] = ACTIONS(1240),
    [sym__indented_chunk_start] = ACTIONS(1240),
    [sym_atx_h1_marker] = ACTIONS(1240),
    [sym_atx_h2_marker] = ACTIONS(1240),
    [sym_atx_h3_marker] = ACTIONS(1240),
    [sym_atx_h4_marker] = ACTIONS(1240),
    [sym_atx_h5_marker] = ACTIONS(1240),
    [sym_atx_h6_marker] = ACTIONS(1240),
    [sym__thematic_break] = ACTIONS(1240),
    [sym__list_marker_minus] = ACTIONS(1240),
    [sym__list_marker_plus] = ACTIONS(1240),
    [sym__list_marker_star] = ACTIONS(1240),
    [sym__list_marker_parenthesis] = ACTIONS(1240),
    [sym__list_marker_dot] = ACTIONS(1240),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1240),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1240),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1240),
    [sym__blank_line_start] = ACTIONS(1240),
    [sym__html_block_1_start] = ACTIONS(1240),
    [sym__html_block_2_start] = ACTIONS(1240),
    [sym__html_block_3_start] = ACTIONS(1240),
    [sym__html_block_4_start] = ACTIONS(1240),
    [sym__html_block_5_start] = ACTIONS(1240),
    [sym__html_block_6_start] = ACTIONS(1240),
    [sym__html_block_7_start] = ACTIONS(1240),
    [sym__pipe_table_start] = ACTIONS(1240),
  },
  [STATE(162)] = {
    [anon_sym_LBRACK] = ACTIONS(1242),
    [anon_sym_RBRACK] = ACTIONS(1240),
    [anon_sym_LT] = ACTIONS(1240),
    [anon_sym_GT] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [anon_sym_POUND] = ACTIONS(1240),
    [anon_sym_DOLLAR] = ACTIONS(1240),
    [anon_sym_PERCENT] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_PLUS] = ACTIONS(1240),
    [anon_sym_COMMA] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_DOT] = ACTIONS(1240),
    [anon_sym_SLASH] = ACTIONS(1240),
    [anon_sym_COLON] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym_EQ] = ACTIONS(1240),
    [anon_sym_QMARK] = ACTIONS(1240),
    [anon_sym_AT] = ACTIONS(1240),
    [anon_sym_BSLASH] = ACTIONS(1240),
    [anon_sym_CARET] = ACTIONS(1240),
    [anon_sym__] = ACTIONS(1240),
    [anon_sym_BQUOTE] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_PIPE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_LPAREN] = ACTIONS(1240),
    [anon_sym_RPAREN] = ACTIONS(1240),
    [aux_sym__word_token1] = ACTIONS(1240),
    [aux_sym__word_token2] = ACTIONS(1240),
    [aux_sym__word_token3] = ACTIONS(1240),
    [sym__whitespace] = ACTIONS(1240),
    [sym__soft_line_ending] = ACTIONS(1240),
    [sym__block_close] = ACTIONS(1240),
    [sym_block_continuation] = ACTIONS(1478),
    [sym__block_quote_start] = ACTIONS(1240),
    [sym__indented_chunk_start] = ACTIONS(1240),
    [sym_atx_h1_marker] = ACTIONS(1240),
    [sym_atx_h2_marker] = ACTIONS(1240),
    [sym_atx_h3_marker] = ACTIONS(1240),
    [sym_atx_h4_marker] = ACTIONS(1240),
    [sym_atx_h5_marker] = ACTIONS(1240),
    [sym_atx_h6_marker] = ACTIONS(1240),
    [sym__thematic_break] = ACTIONS(1240),
    [sym__list_marker_minus] = ACTIONS(1240),
    [sym__list_marker_plus] = ACTIONS(1240),
    [sym__list_marker_star] = ACTIONS(1240),
    [sym__list_marker_parenthesis] = ACTIONS(1240),
    [sym__list_marker_dot] = ACTIONS(1240),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1240),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1240),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1240),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1240),
    [sym__blank_line_start] = ACTIONS(1240),
    [sym__html_block_1_start] = ACTIONS(1240),
    [sym__html_block_2_start] = ACTIONS(1240),
    [sym__html_block_3_start] = ACTIONS(1240),
    [sym__html_block_4_start] = ACTIONS(1240),
    [sym__html_block_5_start] = ACTIONS(1240),
    [sym__html_block_6_start] = ACTIONS(1240),
    [sym__html_block_7_start] = ACTIONS(1240),
    [sym__pipe_table_start] = ACTIONS(1240),
  },
  [STATE(163)] = {
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_RBRACK] = ACTIONS(1482),
    [anon_sym_LT] = ACTIONS(1482),
    [anon_sym_GT] = ACTIONS(1482),
    [anon_sym_BANG] = ACTIONS(1482),
    [anon_sym_DQUOTE] = ACTIONS(1482),
    [anon_sym_POUND] = ACTIONS(1482),
    [anon_sym_DOLLAR] = ACTIONS(1482),
    [anon_sym_PERCENT] = ACTIONS(1482),
    [anon_sym_AMP] = ACTIONS(1482),
    [anon_sym_SQUOTE] = ACTIONS(1482),
    [anon_sym_STAR] = ACTIONS(1482),
    [anon_sym_PLUS] = ACTIONS(1482),
    [anon_sym_COMMA] = ACTIONS(1482),
    [anon_sym_DASH] = ACTIONS(1482),
    [anon_sym_DOT] = ACTIONS(1482),
    [anon_sym_SLASH] = ACTIONS(1482),
    [anon_sym_COLON] = ACTIONS(1482),
    [anon_sym_SEMI] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(1482),
    [anon_sym_QMARK] = ACTIONS(1482),
    [anon_sym_AT] = ACTIONS(1482),
    [anon_sym_BSLASH] = ACTIONS(1482),
    [anon_sym_CARET] = ACTIONS(1482),
    [anon_sym__] = ACTIONS(1482),
    [anon_sym_BQUOTE] = ACTIONS(1482),
    [anon_sym_LBRACE] = ACTIONS(1482),
    [anon_sym_PIPE] = ACTIONS(1482),
    [anon_sym_RBRACE] = ACTIONS(1482),
    [anon_sym_TILDE] = ACTIONS(1482),
    [anon_sym_LPAREN] = ACTIONS(1482),
    [anon_sym_RPAREN] = ACTIONS(1482),
    [aux_sym__word_token1] = ACTIONS(1482),
    [aux_sym__word_token2] = ACTIONS(1482),
    [aux_sym__word_token3] = ACTIONS(1482),
    [sym__whitespace] = ACTIONS(1482),
    [sym__soft_line_ending] = ACTIONS(1482),
    [sym__block_close] = ACTIONS(1482),
    [sym_block_continuation] = ACTIONS(1484),
    [sym__block_quote_start] = ACTIONS(1482),
    [sym__indented_chunk_start] = ACTIONS(1482),
    [sym_atx_h1_marker] = ACTIONS(1482),
    [sym_atx_h2_marker] = ACTIONS(1482),
    [sym_atx_h3_marker] = ACTIONS(1482),
    [sym_atx_h4_marker] = ACTIONS(1482),
    [sym_atx_h5_marker] = ACTIONS(1482),
    [sym_atx_h6_marker] = ACTIONS(1482),
    [sym__thematic_break] = ACTIONS(1482),
    [sym__list_marker_minus] = ACTIONS(1482),
    [sym__list_marker_plus] = ACTIONS(1482),
    [sym__list_marker_star] = ACTIONS(1482),
    [sym__list_marker_parenthesis] = ACTIONS(1482),
    [sym__list_marker_dot] = ACTIONS(1482),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1482),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1482),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1482),
    [sym__blank_line_start] = ACTIONS(1482),
    [sym__html_block_1_start] = ACTIONS(1482),
    [sym__html_block_2_start] = ACTIONS(1482),
    [sym__html_block_3_start] = ACTIONS(1482),
    [sym__html_block_4_start] = ACTIONS(1482),
    [sym__html_block_5_start] = ACTIONS(1482),
    [sym__html_block_6_start] = ACTIONS(1482),
    [sym__html_block_7_start] = ACTIONS(1482),
    [sym__pipe_table_start] = ACTIONS(1482),
  },
  [STATE(164)] = {
    [anon_sym_LBRACK] = ACTIONS(1486),
    [anon_sym_RBRACK] = ACTIONS(1488),
    [anon_sym_LT] = ACTIONS(1488),
    [anon_sym_GT] = ACTIONS(1488),
    [anon_sym_BANG] = ACTIONS(1488),
    [anon_sym_DQUOTE] = ACTIONS(1488),
    [anon_sym_POUND] = ACTIONS(1488),
    [anon_sym_DOLLAR] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_AMP] = ACTIONS(1488),
    [anon_sym_SQUOTE] = ACTIONS(1488),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_PLUS] = ACTIONS(1488),
    [anon_sym_COMMA] = ACTIONS(1488),
    [anon_sym_DASH] = ACTIONS(1488),
    [anon_sym_DOT] = ACTIONS(1488),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_COLON] = ACTIONS(1488),
    [anon_sym_SEMI] = ACTIONS(1488),
    [anon_sym_EQ] = ACTIONS(1488),
    [anon_sym_QMARK] = ACTIONS(1488),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_BSLASH] = ACTIONS(1488),
    [anon_sym_CARET] = ACTIONS(1488),
    [anon_sym__] = ACTIONS(1488),
    [anon_sym_BQUOTE] = ACTIONS(1488),
    [anon_sym_LBRACE] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1488),
    [anon_sym_RBRACE] = ACTIONS(1488),
    [anon_sym_TILDE] = ACTIONS(1488),
    [anon_sym_LPAREN] = ACTIONS(1488),
    [anon_sym_RPAREN] = ACTIONS(1488),
    [aux_sym__word_token1] = ACTIONS(1488),
    [aux_sym__word_token2] = ACTIONS(1488),
    [aux_sym__word_token3] = ACTIONS(1488),
    [sym__whitespace] = ACTIONS(1488),
    [sym__soft_line_ending] = ACTIONS(1488),
    [sym__block_close] = ACTIONS(1488),
    [sym_block_continuation] = ACTIONS(1490),
    [sym__block_quote_start] = ACTIONS(1488),
    [sym__indented_chunk_start] = ACTIONS(1488),
    [sym_atx_h1_marker] = ACTIONS(1488),
    [sym_atx_h2_marker] = ACTIONS(1488),
    [sym_atx_h3_marker] = ACTIONS(1488),
    [sym_atx_h4_marker] = ACTIONS(1488),
    [sym_atx_h5_marker] = ACTIONS(1488),
    [sym_atx_h6_marker] = ACTIONS(1488),
    [sym__thematic_break] = ACTIONS(1488),
    [sym__list_marker_minus] = ACTIONS(1488),
    [sym__list_marker_plus] = ACTIONS(1488),
    [sym__list_marker_star] = ACTIONS(1488),
    [sym__list_marker_parenthesis] = ACTIONS(1488),
    [sym__list_marker_dot] = ACTIONS(1488),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1488),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1488),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1488),
    [sym__blank_line_start] = ACTIONS(1488),
    [sym__html_block_1_start] = ACTIONS(1488),
    [sym__html_block_2_start] = ACTIONS(1488),
    [sym__html_block_3_start] = ACTIONS(1488),
    [sym__html_block_4_start] = ACTIONS(1488),
    [sym__html_block_5_start] = ACTIONS(1488),
    [sym__html_block_6_start] = ACTIONS(1488),
    [sym__html_block_7_start] = ACTIONS(1488),
    [sym__pipe_table_start] = ACTIONS(1488),
  },
  [STATE(165)] = {
    [anon_sym_LBRACK] = ACTIONS(1129),
    [anon_sym_RBRACK] = ACTIONS(1127),
    [anon_sym_LT] = ACTIONS(1127),
    [anon_sym_GT] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_DQUOTE] = ACTIONS(1127),
    [anon_sym_POUND] = ACTIONS(1127),
    [anon_sym_DOLLAR] = ACTIONS(1127),
    [anon_sym_PERCENT] = ACTIONS(1127),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_SQUOTE] = ACTIONS(1127),
    [anon_sym_STAR] = ACTIONS(1127),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_COMMA] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_DOT] = ACTIONS(1127),
    [anon_sym_SLASH] = ACTIONS(1127),
    [anon_sym_COLON] = ACTIONS(1127),
    [anon_sym_SEMI] = ACTIONS(1127),
    [anon_sym_EQ] = ACTIONS(1127),
    [anon_sym_QMARK] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1127),
    [anon_sym_BSLASH] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym__] = ACTIONS(1127),
    [anon_sym_BQUOTE] = ACTIONS(1127),
    [anon_sym_LBRACE] = ACTIONS(1127),
    [anon_sym_PIPE] = ACTIONS(1127),
    [anon_sym_RBRACE] = ACTIONS(1127),
    [anon_sym_TILDE] = ACTIONS(1127),
    [anon_sym_LPAREN] = ACTIONS(1127),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [aux_sym__word_token1] = ACTIONS(1127),
    [aux_sym__word_token2] = ACTIONS(1127),
    [aux_sym__word_token3] = ACTIONS(1127),
    [sym__whitespace] = ACTIONS(1127),
    [sym__soft_line_ending] = ACTIONS(1127),
    [sym__block_close] = ACTIONS(1127),
    [sym_block_continuation] = ACTIONS(1492),
    [sym__block_quote_start] = ACTIONS(1127),
    [sym__indented_chunk_start] = ACTIONS(1127),
    [sym_atx_h1_marker] = ACTIONS(1127),
    [sym_atx_h2_marker] = ACTIONS(1127),
    [sym_atx_h3_marker] = ACTIONS(1127),
    [sym_atx_h4_marker] = ACTIONS(1127),
    [sym_atx_h5_marker] = ACTIONS(1127),
    [sym_atx_h6_marker] = ACTIONS(1127),
    [sym__thematic_break] = ACTIONS(1127),
    [sym__list_marker_minus] = ACTIONS(1127),
    [sym__list_marker_plus] = ACTIONS(1127),
    [sym__list_marker_star] = ACTIONS(1127),
    [sym__list_marker_parenthesis] = ACTIONS(1127),
    [sym__list_marker_dot] = ACTIONS(1127),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1127),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1127),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1127),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1127),
    [sym__blank_line_start] = ACTIONS(1127),
    [sym__html_block_1_start] = ACTIONS(1127),
    [sym__html_block_2_start] = ACTIONS(1127),
    [sym__html_block_3_start] = ACTIONS(1127),
    [sym__html_block_4_start] = ACTIONS(1127),
    [sym__html_block_5_start] = ACTIONS(1127),
    [sym__html_block_6_start] = ACTIONS(1127),
    [sym__html_block_7_start] = ACTIONS(1127),
    [sym__pipe_table_start] = ACTIONS(1127),
  },
  [STATE(166)] = {
    [ts_builtin_sym_end] = ACTIONS(1440),
    [anon_sym_LBRACK] = ACTIONS(1438),
    [anon_sym_RBRACK] = ACTIONS(1440),
    [anon_sym_LT] = ACTIONS(1440),
    [anon_sym_GT] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_DQUOTE] = ACTIONS(1440),
    [anon_sym_POUND] = ACTIONS(1440),
    [anon_sym_DOLLAR] = ACTIONS(1440),
    [anon_sym_PERCENT] = ACTIONS(1440),
    [anon_sym_AMP] = ACTIONS(1440),
    [anon_sym_SQUOTE] = ACTIONS(1440),
    [anon_sym_STAR] = ACTIONS(1440),
    [anon_sym_PLUS] = ACTIONS(1440),
    [anon_sym_COMMA] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_DOT] = ACTIONS(1440),
    [anon_sym_SLASH] = ACTIONS(1440),
    [anon_sym_COLON] = ACTIONS(1440),
    [anon_sym_SEMI] = ACTIONS(1440),
    [anon_sym_EQ] = ACTIONS(1440),
    [anon_sym_QMARK] = ACTIONS(1440),
    [anon_sym_AT] = ACTIONS(1440),
    [anon_sym_BSLASH] = ACTIONS(1440),
    [anon_sym_CARET] = ACTIONS(1440),
    [anon_sym__] = ACTIONS(1440),
    [anon_sym_BQUOTE] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_PIPE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [anon_sym_TILDE] = ACTIONS(1440),
    [anon_sym_LPAREN] = ACTIONS(1440),
    [anon_sym_RPAREN] = ACTIONS(1440),
    [aux_sym__word_token1] = ACTIONS(1440),
    [aux_sym__word_token2] = ACTIONS(1440),
    [aux_sym__word_token3] = ACTIONS(1440),
    [sym__whitespace] = ACTIONS(1440),
    [sym__soft_line_ending] = ACTIONS(1440),
    [sym_block_continuation] = ACTIONS(1494),
    [sym__block_quote_start] = ACTIONS(1440),
    [sym__indented_chunk_start] = ACTIONS(1440),
    [sym_atx_h1_marker] = ACTIONS(1440),
    [sym_atx_h2_marker] = ACTIONS(1440),
    [sym_atx_h3_marker] = ACTIONS(1440),
    [sym_atx_h4_marker] = ACTIONS(1440),
    [sym_atx_h5_marker] = ACTIONS(1440),
    [sym_atx_h6_marker] = ACTIONS(1440),
    [sym__thematic_break] = ACTIONS(1440),
    [sym__list_marker_minus] = ACTIONS(1440),
    [sym__list_marker_plus] = ACTIONS(1440),
    [sym__list_marker_star] = ACTIONS(1440),
    [sym__list_marker_parenthesis] = ACTIONS(1440),
    [sym__list_marker_dot] = ACTIONS(1440),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1440),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1440),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1440),
    [sym__blank_line_start] = ACTIONS(1440),
    [sym__html_block_1_start] = ACTIONS(1440),
    [sym__html_block_2_start] = ACTIONS(1440),
    [sym__html_block_3_start] = ACTIONS(1440),
    [sym__html_block_4_start] = ACTIONS(1440),
    [sym__html_block_5_start] = ACTIONS(1440),
    [sym__html_block_6_start] = ACTIONS(1440),
    [sym__html_block_7_start] = ACTIONS(1440),
    [sym__pipe_table_start] = ACTIONS(1440),
  },
  [STATE(167)] = {
    [ts_builtin_sym_end] = ACTIONS(1464),
    [anon_sym_LBRACK] = ACTIONS(1462),
    [anon_sym_RBRACK] = ACTIONS(1464),
    [anon_sym_LT] = ACTIONS(1464),
    [anon_sym_GT] = ACTIONS(1464),
    [anon_sym_BANG] = ACTIONS(1464),
    [anon_sym_DQUOTE] = ACTIONS(1464),
    [anon_sym_POUND] = ACTIONS(1464),
    [anon_sym_DOLLAR] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1464),
    [anon_sym_AMP] = ACTIONS(1464),
    [anon_sym_SQUOTE] = ACTIONS(1464),
    [anon_sym_STAR] = ACTIONS(1464),
    [anon_sym_PLUS] = ACTIONS(1464),
    [anon_sym_COMMA] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_DOT] = ACTIONS(1464),
    [anon_sym_SLASH] = ACTIONS(1464),
    [anon_sym_COLON] = ACTIONS(1464),
    [anon_sym_SEMI] = ACTIONS(1464),
    [anon_sym_EQ] = ACTIONS(1464),
    [anon_sym_QMARK] = ACTIONS(1464),
    [anon_sym_AT] = ACTIONS(1464),
    [anon_sym_BSLASH] = ACTIONS(1464),
    [anon_sym_CARET] = ACTIONS(1464),
    [anon_sym__] = ACTIONS(1464),
    [anon_sym_BQUOTE] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [anon_sym_PIPE] = ACTIONS(1464),
    [anon_sym_RBRACE] = ACTIONS(1464),
    [anon_sym_TILDE] = ACTIONS(1464),
    [anon_sym_LPAREN] = ACTIONS(1464),
    [anon_sym_RPAREN] = ACTIONS(1464),
    [aux_sym__word_token1] = ACTIONS(1464),
    [aux_sym__word_token2] = ACTIONS(1464),
    [aux_sym__word_token3] = ACTIONS(1464),
    [sym__whitespace] = ACTIONS(1464),
    [sym__soft_line_ending] = ACTIONS(1464),
    [sym_block_continuation] = ACTIONS(1496),
    [sym__block_quote_start] = ACTIONS(1464),
    [sym__indented_chunk_start] = ACTIONS(1464),
    [sym_atx_h1_marker] = ACTIONS(1464),
    [sym_atx_h2_marker] = ACTIONS(1464),
    [sym_atx_h3_marker] = ACTIONS(1464),
    [sym_atx_h4_marker] = ACTIONS(1464),
    [sym_atx_h5_marker] = ACTIONS(1464),
    [sym_atx_h6_marker] = ACTIONS(1464),
    [sym__thematic_break] = ACTIONS(1464),
    [sym__list_marker_minus] = ACTIONS(1464),
    [sym__list_marker_plus] = ACTIONS(1464),
    [sym__list_marker_star] = ACTIONS(1464),
    [sym__list_marker_parenthesis] = ACTIONS(1464),
    [sym__list_marker_dot] = ACTIONS(1464),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1464),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1464),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1464),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1464),
    [sym__blank_line_start] = ACTIONS(1464),
    [sym__html_block_1_start] = ACTIONS(1464),
    [sym__html_block_2_start] = ACTIONS(1464),
    [sym__html_block_3_start] = ACTIONS(1464),
    [sym__html_block_4_start] = ACTIONS(1464),
    [sym__html_block_5_start] = ACTIONS(1464),
    [sym__html_block_6_start] = ACTIONS(1464),
    [sym__html_block_7_start] = ACTIONS(1464),
    [sym__pipe_table_start] = ACTIONS(1464),
  },
  [STATE(168)] = {
    [ts_builtin_sym_end] = ACTIONS(1332),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [anon_sym_RBRACK] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_COLON] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1332),
    [anon_sym_BSLASH] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym__] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [aux_sym__word_token1] = ACTIONS(1332),
    [aux_sym__word_token2] = ACTIONS(1332),
    [aux_sym__word_token3] = ACTIONS(1332),
    [sym__whitespace] = ACTIONS(1332),
    [sym__soft_line_ending] = ACTIONS(1332),
    [sym_block_continuation] = ACTIONS(1498),
    [sym__block_quote_start] = ACTIONS(1332),
    [sym__indented_chunk_start] = ACTIONS(1332),
    [sym_atx_h1_marker] = ACTIONS(1332),
    [sym_atx_h2_marker] = ACTIONS(1332),
    [sym_atx_h3_marker] = ACTIONS(1332),
    [sym_atx_h4_marker] = ACTIONS(1332),
    [sym_atx_h5_marker] = ACTIONS(1332),
    [sym_atx_h6_marker] = ACTIONS(1332),
    [sym__thematic_break] = ACTIONS(1332),
    [sym__list_marker_minus] = ACTIONS(1332),
    [sym__list_marker_plus] = ACTIONS(1332),
    [sym__list_marker_star] = ACTIONS(1332),
    [sym__list_marker_parenthesis] = ACTIONS(1332),
    [sym__list_marker_dot] = ACTIONS(1332),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1332),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1332),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1332),
    [sym__blank_line_start] = ACTIONS(1332),
    [sym__html_block_1_start] = ACTIONS(1332),
    [sym__html_block_2_start] = ACTIONS(1332),
    [sym__html_block_3_start] = ACTIONS(1332),
    [sym__html_block_4_start] = ACTIONS(1332),
    [sym__html_block_5_start] = ACTIONS(1332),
    [sym__html_block_6_start] = ACTIONS(1332),
    [sym__html_block_7_start] = ACTIONS(1332),
    [sym__pipe_table_start] = ACTIONS(1332),
  },
  [STATE(169)] = {
    [ts_builtin_sym_end] = ACTIONS(1456),
    [anon_sym_LBRACK] = ACTIONS(1454),
    [anon_sym_RBRACK] = ACTIONS(1456),
    [anon_sym_LT] = ACTIONS(1456),
    [anon_sym_GT] = ACTIONS(1456),
    [anon_sym_BANG] = ACTIONS(1456),
    [anon_sym_DQUOTE] = ACTIONS(1456),
    [anon_sym_POUND] = ACTIONS(1456),
    [anon_sym_DOLLAR] = ACTIONS(1456),
    [anon_sym_PERCENT] = ACTIONS(1456),
    [anon_sym_AMP] = ACTIONS(1456),
    [anon_sym_SQUOTE] = ACTIONS(1456),
    [anon_sym_STAR] = ACTIONS(1456),
    [anon_sym_PLUS] = ACTIONS(1456),
    [anon_sym_COMMA] = ACTIONS(1456),
    [anon_sym_DASH] = ACTIONS(1456),
    [anon_sym_DOT] = ACTIONS(1456),
    [anon_sym_SLASH] = ACTIONS(1456),
    [anon_sym_COLON] = ACTIONS(1456),
    [anon_sym_SEMI] = ACTIONS(1456),
    [anon_sym_EQ] = ACTIONS(1456),
    [anon_sym_QMARK] = ACTIONS(1456),
    [anon_sym_AT] = ACTIONS(1456),
    [anon_sym_BSLASH] = ACTIONS(1456),
    [anon_sym_CARET] = ACTIONS(1456),
    [anon_sym__] = ACTIONS(1456),
    [anon_sym_BQUOTE] = ACTIONS(1456),
    [anon_sym_LBRACE] = ACTIONS(1456),
    [anon_sym_PIPE] = ACTIONS(1456),
    [anon_sym_RBRACE] = ACTIONS(1456),
    [anon_sym_TILDE] = ACTIONS(1456),
    [anon_sym_LPAREN] = ACTIONS(1456),
    [anon_sym_RPAREN] = ACTIONS(1456),
    [aux_sym__word_token1] = ACTIONS(1456),
    [aux_sym__word_token2] = ACTIONS(1456),
    [aux_sym__word_token3] = ACTIONS(1456),
    [sym__whitespace] = ACTIONS(1456),
    [sym__soft_line_ending] = ACTIONS(1456),
    [sym_block_continuation] = ACTIONS(1500),
    [sym__block_quote_start] = ACTIONS(1456),
    [sym__indented_chunk_start] = ACTIONS(1456),
    [sym_atx_h1_marker] = ACTIONS(1456),
    [sym_atx_h2_marker] = ACTIONS(1456),
    [sym_atx_h3_marker] = ACTIONS(1456),
    [sym_atx_h4_marker] = ACTIONS(1456),
    [sym_atx_h5_marker] = ACTIONS(1456),
    [sym_atx_h6_marker] = ACTIONS(1456),
    [sym__thematic_break] = ACTIONS(1456),
    [sym__list_marker_minus] = ACTIONS(1456),
    [sym__list_marker_plus] = ACTIONS(1456),
    [sym__list_marker_star] = ACTIONS(1456),
    [sym__list_marker_parenthesis] = ACTIONS(1456),
    [sym__list_marker_dot] = ACTIONS(1456),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1456),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1456),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1456),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1456),
    [sym__blank_line_start] = ACTIONS(1456),
    [sym__html_block_1_start] = ACTIONS(1456),
    [sym__html_block_2_start] = ACTIONS(1456),
    [sym__html_block_3_start] = ACTIONS(1456),
    [sym__html_block_4_start] = ACTIONS(1456),
    [sym__html_block_5_start] = ACTIONS(1456),
    [sym__html_block_6_start] = ACTIONS(1456),
    [sym__html_block_7_start] = ACTIONS(1456),
    [sym__pipe_table_start] = ACTIONS(1456),
  },
  [STATE(170)] = {
    [sym__blank_line] = STATE(902),
    [anon_sym_LBRACK] = ACTIONS(1263),
    [anon_sym_RBRACK] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(1261),
    [anon_sym_GT] = ACTIONS(1261),
    [anon_sym_BANG] = ACTIONS(1261),
    [anon_sym_DQUOTE] = ACTIONS(1261),
    [anon_sym_POUND] = ACTIONS(1261),
    [anon_sym_DOLLAR] = ACTIONS(1261),
    [anon_sym_PERCENT] = ACTIONS(1261),
    [anon_sym_AMP] = ACTIONS(1261),
    [anon_sym_SQUOTE] = ACTIONS(1261),
    [anon_sym_STAR] = ACTIONS(1261),
    [anon_sym_PLUS] = ACTIONS(1261),
    [anon_sym_COMMA] = ACTIONS(1261),
    [anon_sym_DASH] = ACTIONS(1261),
    [anon_sym_DOT] = ACTIONS(1261),
    [anon_sym_SLASH] = ACTIONS(1261),
    [anon_sym_COLON] = ACTIONS(1261),
    [anon_sym_SEMI] = ACTIONS(1261),
    [anon_sym_EQ] = ACTIONS(1261),
    [anon_sym_QMARK] = ACTIONS(1261),
    [anon_sym_AT] = ACTIONS(1261),
    [anon_sym_BSLASH] = ACTIONS(1261),
    [anon_sym_CARET] = ACTIONS(1261),
    [anon_sym__] = ACTIONS(1261),
    [anon_sym_BQUOTE] = ACTIONS(1261),
    [anon_sym_LBRACE] = ACTIONS(1261),
    [anon_sym_PIPE] = ACTIONS(1261),
    [anon_sym_RBRACE] = ACTIONS(1261),
    [anon_sym_TILDE] = ACTIONS(1261),
    [anon_sym_LPAREN] = ACTIONS(1261),
    [anon_sym_RPAREN] = ACTIONS(1261),
    [aux_sym__word_token1] = ACTIONS(1261),
    [aux_sym__word_token2] = ACTIONS(1261),
    [aux_sym__word_token3] = ACTIONS(1261),
    [sym__whitespace] = ACTIONS(1261),
    [sym__soft_line_ending] = ACTIONS(1261),
    [sym__block_close] = ACTIONS(1261),
    [sym__block_quote_start] = ACTIONS(1261),
    [sym__indented_chunk_start] = ACTIONS(1261),
    [sym_atx_h1_marker] = ACTIONS(1261),
    [sym_atx_h2_marker] = ACTIONS(1261),
    [sym_atx_h3_marker] = ACTIONS(1261),
    [sym_atx_h4_marker] = ACTIONS(1261),
    [sym_atx_h5_marker] = ACTIONS(1261),
    [sym_atx_h6_marker] = ACTIONS(1261),
    [sym__thematic_break] = ACTIONS(1261),
    [sym__list_marker_minus] = ACTIONS(1261),
    [sym__list_marker_plus] = ACTIONS(1261),
    [sym__list_marker_star] = ACTIONS(1261),
    [sym__list_marker_parenthesis] = ACTIONS(1261),
    [sym__list_marker_dot] = ACTIONS(1261),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1261),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1261),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1261),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1261),
    [sym__blank_line_start] = ACTIONS(1502),
    [sym__html_block_1_start] = ACTIONS(1261),
    [sym__html_block_2_start] = ACTIONS(1261),
    [sym__html_block_3_start] = ACTIONS(1261),
    [sym__html_block_4_start] = ACTIONS(1261),
    [sym__html_block_5_start] = ACTIONS(1261),
    [sym__html_block_6_start] = ACTIONS(1261),
    [sym__html_block_7_start] = ACTIONS(1261),
    [sym__pipe_table_start] = ACTIONS(1261),
  },
  [STATE(171)] = {
    [ts_builtin_sym_end] = ACTIONS(1504),
    [anon_sym_LBRACK] = ACTIONS(1506),
    [anon_sym_RBRACK] = ACTIONS(1504),
    [anon_sym_LT] = ACTIONS(1504),
    [anon_sym_GT] = ACTIONS(1504),
    [anon_sym_BANG] = ACTIONS(1504),
    [anon_sym_DQUOTE] = ACTIONS(1504),
    [anon_sym_POUND] = ACTIONS(1504),
    [anon_sym_DOLLAR] = ACTIONS(1504),
    [anon_sym_PERCENT] = ACTIONS(1504),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_SQUOTE] = ACTIONS(1504),
    [anon_sym_STAR] = ACTIONS(1504),
    [anon_sym_PLUS] = ACTIONS(1504),
    [anon_sym_COMMA] = ACTIONS(1504),
    [anon_sym_DASH] = ACTIONS(1504),
    [anon_sym_DOT] = ACTIONS(1504),
    [anon_sym_SLASH] = ACTIONS(1504),
    [anon_sym_COLON] = ACTIONS(1504),
    [anon_sym_SEMI] = ACTIONS(1504),
    [anon_sym_EQ] = ACTIONS(1504),
    [anon_sym_QMARK] = ACTIONS(1504),
    [anon_sym_AT] = ACTIONS(1504),
    [anon_sym_BSLASH] = ACTIONS(1504),
    [anon_sym_CARET] = ACTIONS(1504),
    [anon_sym__] = ACTIONS(1504),
    [anon_sym_BQUOTE] = ACTIONS(1504),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_PIPE] = ACTIONS(1504),
    [anon_sym_RBRACE] = ACTIONS(1504),
    [anon_sym_TILDE] = ACTIONS(1504),
    [anon_sym_LPAREN] = ACTIONS(1504),
    [anon_sym_RPAREN] = ACTIONS(1504),
    [aux_sym__word_token1] = ACTIONS(1504),
    [aux_sym__word_token2] = ACTIONS(1504),
    [aux_sym__word_token3] = ACTIONS(1504),
    [sym__whitespace] = ACTIONS(1504),
    [sym__soft_line_ending] = ACTIONS(1504),
    [sym_block_continuation] = ACTIONS(1508),
    [sym__block_quote_start] = ACTIONS(1504),
    [sym__indented_chunk_start] = ACTIONS(1504),
    [sym_atx_h1_marker] = ACTIONS(1504),
    [sym_atx_h2_marker] = ACTIONS(1504),
    [sym_atx_h3_marker] = ACTIONS(1504),
    [sym_atx_h4_marker] = ACTIONS(1504),
    [sym_atx_h5_marker] = ACTIONS(1504),
    [sym_atx_h6_marker] = ACTIONS(1504),
    [sym__thematic_break] = ACTIONS(1504),
    [sym__list_marker_minus] = ACTIONS(1504),
    [sym__list_marker_plus] = ACTIONS(1504),
    [sym__list_marker_star] = ACTIONS(1504),
    [sym__list_marker_parenthesis] = ACTIONS(1504),
    [sym__list_marker_dot] = ACTIONS(1504),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1504),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1504),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1504),
    [sym__blank_line_start] = ACTIONS(1504),
    [sym__html_block_1_start] = ACTIONS(1504),
    [sym__html_block_2_start] = ACTIONS(1504),
    [sym__html_block_3_start] = ACTIONS(1504),
    [sym__html_block_4_start] = ACTIONS(1504),
    [sym__html_block_5_start] = ACTIONS(1504),
    [sym__html_block_6_start] = ACTIONS(1504),
    [sym__html_block_7_start] = ACTIONS(1504),
    [sym__pipe_table_start] = ACTIONS(1504),
  },
  [STATE(172)] = {
    [ts_builtin_sym_end] = ACTIONS(1482),
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_RBRACK] = ACTIONS(1482),
    [anon_sym_LT] = ACTIONS(1482),
    [anon_sym_GT] = ACTIONS(1482),
    [anon_sym_BANG] = ACTIONS(1482),
    [anon_sym_DQUOTE] = ACTIONS(1482),
    [anon_sym_POUND] = ACTIONS(1482),
    [anon_sym_DOLLAR] = ACTIONS(1482),
    [anon_sym_PERCENT] = ACTIONS(1482),
    [anon_sym_AMP] = ACTIONS(1482),
    [anon_sym_SQUOTE] = ACTIONS(1482),
    [anon_sym_STAR] = ACTIONS(1482),
    [anon_sym_PLUS] = ACTIONS(1482),
    [anon_sym_COMMA] = ACTIONS(1482),
    [anon_sym_DASH] = ACTIONS(1482),
    [anon_sym_DOT] = ACTIONS(1482),
    [anon_sym_SLASH] = ACTIONS(1482),
    [anon_sym_COLON] = ACTIONS(1482),
    [anon_sym_SEMI] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(1482),
    [anon_sym_QMARK] = ACTIONS(1482),
    [anon_sym_AT] = ACTIONS(1482),
    [anon_sym_BSLASH] = ACTIONS(1482),
    [anon_sym_CARET] = ACTIONS(1482),
    [anon_sym__] = ACTIONS(1482),
    [anon_sym_BQUOTE] = ACTIONS(1482),
    [anon_sym_LBRACE] = ACTIONS(1482),
    [anon_sym_PIPE] = ACTIONS(1482),
    [anon_sym_RBRACE] = ACTIONS(1482),
    [anon_sym_TILDE] = ACTIONS(1482),
    [anon_sym_LPAREN] = ACTIONS(1482),
    [anon_sym_RPAREN] = ACTIONS(1482),
    [aux_sym__word_token1] = ACTIONS(1482),
    [aux_sym__word_token2] = ACTIONS(1482),
    [aux_sym__word_token3] = ACTIONS(1482),
    [sym__whitespace] = ACTIONS(1482),
    [sym__soft_line_ending] = ACTIONS(1482),
    [sym_block_continuation] = ACTIONS(1510),
    [sym__block_quote_start] = ACTIONS(1482),
    [sym__indented_chunk_start] = ACTIONS(1482),
    [sym_atx_h1_marker] = ACTIONS(1482),
    [sym_atx_h2_marker] = ACTIONS(1482),
    [sym_atx_h3_marker] = ACTIONS(1482),
    [sym_atx_h4_marker] = ACTIONS(1482),
    [sym_atx_h5_marker] = ACTIONS(1482),
    [sym_atx_h6_marker] = ACTIONS(1482),
    [sym__thematic_break] = ACTIONS(1482),
    [sym__list_marker_minus] = ACTIONS(1482),
    [sym__list_marker_plus] = ACTIONS(1482),
    [sym__list_marker_star] = ACTIONS(1482),
    [sym__list_marker_parenthesis] = ACTIONS(1482),
    [sym__list_marker_dot] = ACTIONS(1482),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1482),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1482),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1482),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1482),
    [sym__blank_line_start] = ACTIONS(1482),
    [sym__html_block_1_start] = ACTIONS(1482),
    [sym__html_block_2_start] = ACTIONS(1482),
    [sym__html_block_3_start] = ACTIONS(1482),
    [sym__html_block_4_start] = ACTIONS(1482),
    [sym__html_block_5_start] = ACTIONS(1482),
    [sym__html_block_6_start] = ACTIONS(1482),
    [sym__html_block_7_start] = ACTIONS(1482),
    [sym__pipe_table_start] = ACTIONS(1482),
  },
  [STATE(173)] = {
    [ts_builtin_sym_end] = ACTIONS(1488),
    [anon_sym_LBRACK] = ACTIONS(1486),
    [anon_sym_RBRACK] = ACTIONS(1488),
    [anon_sym_LT] = ACTIONS(1488),
    [anon_sym_GT] = ACTIONS(1488),
    [anon_sym_BANG] = ACTIONS(1488),
    [anon_sym_DQUOTE] = ACTIONS(1488),
    [anon_sym_POUND] = ACTIONS(1488),
    [anon_sym_DOLLAR] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_AMP] = ACTIONS(1488),
    [anon_sym_SQUOTE] = ACTIONS(1488),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_PLUS] = ACTIONS(1488),
    [anon_sym_COMMA] = ACTIONS(1488),
    [anon_sym_DASH] = ACTIONS(1488),
    [anon_sym_DOT] = ACTIONS(1488),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_COLON] = ACTIONS(1488),
    [anon_sym_SEMI] = ACTIONS(1488),
    [anon_sym_EQ] = ACTIONS(1488),
    [anon_sym_QMARK] = ACTIONS(1488),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_BSLASH] = ACTIONS(1488),
    [anon_sym_CARET] = ACTIONS(1488),
    [anon_sym__] = ACTIONS(1488),
    [anon_sym_BQUOTE] = ACTIONS(1488),
    [anon_sym_LBRACE] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1488),
    [anon_sym_RBRACE] = ACTIONS(1488),
    [anon_sym_TILDE] = ACTIONS(1488),
    [anon_sym_LPAREN] = ACTIONS(1488),
    [anon_sym_RPAREN] = ACTIONS(1488),
    [aux_sym__word_token1] = ACTIONS(1488),
    [aux_sym__word_token2] = ACTIONS(1488),
    [aux_sym__word_token3] = ACTIONS(1488),
    [sym__whitespace] = ACTIONS(1488),
    [sym__soft_line_ending] = ACTIONS(1488),
    [sym_block_continuation] = ACTIONS(1512),
    [sym__block_quote_start] = ACTIONS(1488),
    [sym__indented_chunk_start] = ACTIONS(1488),
    [sym_atx_h1_marker] = ACTIONS(1488),
    [sym_atx_h2_marker] = ACTIONS(1488),
    [sym_atx_h3_marker] = ACTIONS(1488),
    [sym_atx_h4_marker] = ACTIONS(1488),
    [sym_atx_h5_marker] = ACTIONS(1488),
    [sym_atx_h6_marker] = ACTIONS(1488),
    [sym__thematic_break] = ACTIONS(1488),
    [sym__list_marker_minus] = ACTIONS(1488),
    [sym__list_marker_plus] = ACTIONS(1488),
    [sym__list_marker_star] = ACTIONS(1488),
    [sym__list_marker_parenthesis] = ACTIONS(1488),
    [sym__list_marker_dot] = ACTIONS(1488),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1488),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1488),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1488),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1488),
    [sym__blank_line_start] = ACTIONS(1488),
    [sym__html_block_1_start] = ACTIONS(1488),
    [sym__html_block_2_start] = ACTIONS(1488),
    [sym__html_block_3_start] = ACTIONS(1488),
    [sym__html_block_4_start] = ACTIONS(1488),
    [sym__html_block_5_start] = ACTIONS(1488),
    [sym__html_block_6_start] = ACTIONS(1488),
    [sym__html_block_7_start] = ACTIONS(1488),
    [sym__pipe_table_start] = ACTIONS(1488),
  },
  [STATE(174)] = {
    [ts_builtin_sym_end] = ACTIONS(1402),
    [anon_sym_LBRACK] = ACTIONS(1400),
    [anon_sym_RBRACK] = ACTIONS(1402),
    [anon_sym_LT] = ACTIONS(1402),
    [anon_sym_GT] = ACTIONS(1402),
    [anon_sym_BANG] = ACTIONS(1402),
    [anon_sym_DQUOTE] = ACTIONS(1402),
    [anon_sym_POUND] = ACTIONS(1402),
    [anon_sym_DOLLAR] = ACTIONS(1402),
    [anon_sym_PERCENT] = ACTIONS(1402),
    [anon_sym_AMP] = ACTIONS(1402),
    [anon_sym_SQUOTE] = ACTIONS(1402),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_COMMA] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_DOT] = ACTIONS(1402),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_COLON] = ACTIONS(1402),
    [anon_sym_SEMI] = ACTIONS(1402),
    [anon_sym_EQ] = ACTIONS(1402),
    [anon_sym_QMARK] = ACTIONS(1402),
    [anon_sym_AT] = ACTIONS(1402),
    [anon_sym_BSLASH] = ACTIONS(1402),
    [anon_sym_CARET] = ACTIONS(1402),
    [anon_sym__] = ACTIONS(1402),
    [anon_sym_BQUOTE] = ACTIONS(1402),
    [anon_sym_LBRACE] = ACTIONS(1402),
    [anon_sym_PIPE] = ACTIONS(1402),
    [anon_sym_RBRACE] = ACTIONS(1402),
    [anon_sym_TILDE] = ACTIONS(1402),
    [anon_sym_LPAREN] = ACTIONS(1402),
    [anon_sym_RPAREN] = ACTIONS(1402),
    [aux_sym__word_token1] = ACTIONS(1402),
    [aux_sym__word_token2] = ACTIONS(1402),
    [aux_sym__word_token3] = ACTIONS(1402),
    [sym__whitespace] = ACTIONS(1402),
    [sym__soft_line_ending] = ACTIONS(1402),
    [sym_block_continuation] = ACTIONS(1514),
    [sym__block_quote_start] = ACTIONS(1402),
    [sym__indented_chunk_start] = ACTIONS(1402),
    [sym_atx_h1_marker] = ACTIONS(1402),
    [sym_atx_h2_marker] = ACTIONS(1402),
    [sym_atx_h3_marker] = ACTIONS(1402),
    [sym_atx_h4_marker] = ACTIONS(1402),
    [sym_atx_h5_marker] = ACTIONS(1402),
    [sym_atx_h6_marker] = ACTIONS(1402),
    [sym__thematic_break] = ACTIONS(1402),
    [sym__list_marker_minus] = ACTIONS(1402),
    [sym__list_marker_plus] = ACTIONS(1402),
    [sym__list_marker_star] = ACTIONS(1402),
    [sym__list_marker_parenthesis] = ACTIONS(1402),
    [sym__list_marker_dot] = ACTIONS(1402),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1402),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1402),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1402),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1402),
    [sym__blank_line_start] = ACTIONS(1402),
    [sym__html_block_1_start] = ACTIONS(1402),
    [sym__html_block_2_start] = ACTIONS(1402),
    [sym__html_block_3_start] = ACTIONS(1402),
    [sym__html_block_4_start] = ACTIONS(1402),
    [sym__html_block_5_start] = ACTIONS(1402),
    [sym__html_block_6_start] = ACTIONS(1402),
    [sym__html_block_7_start] = ACTIONS(1402),
    [sym__pipe_table_start] = ACTIONS(1402),
  },
  [STATE(175)] = {
    [ts_builtin_sym_end] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_RBRACK] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_POUND] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1338),
    [anon_sym_PERCENT] = ACTIONS(1338),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_COLON] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1338),
    [anon_sym_EQ] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_BSLASH] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1338),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_BQUOTE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1338),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_RBRACE] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1338),
    [anon_sym_RPAREN] = ACTIONS(1338),
    [aux_sym__word_token1] = ACTIONS(1338),
    [aux_sym__word_token2] = ACTIONS(1338),
    [aux_sym__word_token3] = ACTIONS(1338),
    [sym__whitespace] = ACTIONS(1338),
    [sym__soft_line_ending] = ACTIONS(1338),
    [sym_block_continuation] = ACTIONS(1516),
    [sym__block_quote_start] = ACTIONS(1338),
    [sym__indented_chunk_start] = ACTIONS(1338),
    [sym_atx_h1_marker] = ACTIONS(1338),
    [sym_atx_h2_marker] = ACTIONS(1338),
    [sym_atx_h3_marker] = ACTIONS(1338),
    [sym_atx_h4_marker] = ACTIONS(1338),
    [sym_atx_h5_marker] = ACTIONS(1338),
    [sym_atx_h6_marker] = ACTIONS(1338),
    [sym__thematic_break] = ACTIONS(1338),
    [sym__list_marker_minus] = ACTIONS(1338),
    [sym__list_marker_plus] = ACTIONS(1338),
    [sym__list_marker_star] = ACTIONS(1338),
    [sym__list_marker_parenthesis] = ACTIONS(1338),
    [sym__list_marker_dot] = ACTIONS(1338),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1338),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1338),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1338),
    [sym__blank_line_start] = ACTIONS(1338),
    [sym__html_block_1_start] = ACTIONS(1338),
    [sym__html_block_2_start] = ACTIONS(1338),
    [sym__html_block_3_start] = ACTIONS(1338),
    [sym__html_block_4_start] = ACTIONS(1338),
    [sym__html_block_5_start] = ACTIONS(1338),
    [sym__html_block_6_start] = ACTIONS(1338),
    [sym__html_block_7_start] = ACTIONS(1338),
    [sym__pipe_table_start] = ACTIONS(1338),
  },
  [STATE(176)] = {
    [ts_builtin_sym_end] = ACTIONS(1408),
    [anon_sym_LBRACK] = ACTIONS(1406),
    [anon_sym_RBRACK] = ACTIONS(1408),
    [anon_sym_LT] = ACTIONS(1408),
    [anon_sym_GT] = ACTIONS(1408),
    [anon_sym_BANG] = ACTIONS(1408),
    [anon_sym_DQUOTE] = ACTIONS(1408),
    [anon_sym_POUND] = ACTIONS(1408),
    [anon_sym_DOLLAR] = ACTIONS(1408),
    [anon_sym_PERCENT] = ACTIONS(1408),
    [anon_sym_AMP] = ACTIONS(1408),
    [anon_sym_SQUOTE] = ACTIONS(1408),
    [anon_sym_STAR] = ACTIONS(1408),
    [anon_sym_PLUS] = ACTIONS(1408),
    [anon_sym_COMMA] = ACTIONS(1408),
    [anon_sym_DASH] = ACTIONS(1408),
    [anon_sym_DOT] = ACTIONS(1408),
    [anon_sym_SLASH] = ACTIONS(1408),
    [anon_sym_COLON] = ACTIONS(1408),
    [anon_sym_SEMI] = ACTIONS(1408),
    [anon_sym_EQ] = ACTIONS(1408),
    [anon_sym_QMARK] = ACTIONS(1408),
    [anon_sym_AT] = ACTIONS(1408),
    [anon_sym_BSLASH] = ACTIONS(1408),
    [anon_sym_CARET] = ACTIONS(1408),
    [anon_sym__] = ACTIONS(1408),
    [anon_sym_BQUOTE] = ACTIONS(1408),
    [anon_sym_LBRACE] = ACTIONS(1408),
    [anon_sym_PIPE] = ACTIONS(1408),
    [anon_sym_RBRACE] = ACTIONS(1408),
    [anon_sym_TILDE] = ACTIONS(1408),
    [anon_sym_LPAREN] = ACTIONS(1408),
    [anon_sym_RPAREN] = ACTIONS(1408),
    [aux_sym__word_token1] = ACTIONS(1408),
    [aux_sym__word_token2] = ACTIONS(1408),
    [aux_sym__word_token3] = ACTIONS(1408),
    [sym__whitespace] = ACTIONS(1408),
    [sym__soft_line_ending] = ACTIONS(1408),
    [sym_block_continuation] = ACTIONS(1518),
    [sym__block_quote_start] = ACTIONS(1408),
    [sym__indented_chunk_start] = ACTIONS(1408),
    [sym_atx_h1_marker] = ACTIONS(1408),
    [sym_atx_h2_marker] = ACTIONS(1408),
    [sym_atx_h3_marker] = ACTIONS(1408),
    [sym_atx_h4_marker] = ACTIONS(1408),
    [sym_atx_h5_marker] = ACTIONS(1408),
    [sym_atx_h6_marker] = ACTIONS(1408),
    [sym__thematic_break] = ACTIONS(1408),
    [sym__list_marker_minus] = ACTIONS(1408),
    [sym__list_marker_plus] = ACTIONS(1408),
    [sym__list_marker_star] = ACTIONS(1408),
    [sym__list_marker_parenthesis] = ACTIONS(1408),
    [sym__list_marker_dot] = ACTIONS(1408),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1408),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1408),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1408),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1408),
    [sym__blank_line_start] = ACTIONS(1408),
    [sym__html_block_1_start] = ACTIONS(1408),
    [sym__html_block_2_start] = ACTIONS(1408),
    [sym__html_block_3_start] = ACTIONS(1408),
    [sym__html_block_4_start] = ACTIONS(1408),
    [sym__html_block_5_start] = ACTIONS(1408),
    [sym__html_block_6_start] = ACTIONS(1408),
    [sym__html_block_7_start] = ACTIONS(1408),
    [sym__pipe_table_start] = ACTIONS(1408),
  },
  [STATE(177)] = {
    [ts_builtin_sym_end] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1342),
    [anon_sym_RBRACK] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [anon_sym_POUND] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_SQUOTE] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_COLON] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_BSLASH] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym__] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_RPAREN] = ACTIONS(1344),
    [aux_sym__word_token1] = ACTIONS(1344),
    [aux_sym__word_token2] = ACTIONS(1344),
    [aux_sym__word_token3] = ACTIONS(1344),
    [sym__whitespace] = ACTIONS(1344),
    [sym__soft_line_ending] = ACTIONS(1344),
    [sym_block_continuation] = ACTIONS(1520),
    [sym__block_quote_start] = ACTIONS(1344),
    [sym__indented_chunk_start] = ACTIONS(1344),
    [sym_atx_h1_marker] = ACTIONS(1344),
    [sym_atx_h2_marker] = ACTIONS(1344),
    [sym_atx_h3_marker] = ACTIONS(1344),
    [sym_atx_h4_marker] = ACTIONS(1344),
    [sym_atx_h5_marker] = ACTIONS(1344),
    [sym_atx_h6_marker] = ACTIONS(1344),
    [sym__thematic_break] = ACTIONS(1344),
    [sym__list_marker_minus] = ACTIONS(1344),
    [sym__list_marker_plus] = ACTIONS(1344),
    [sym__list_marker_star] = ACTIONS(1344),
    [sym__list_marker_parenthesis] = ACTIONS(1344),
    [sym__list_marker_dot] = ACTIONS(1344),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1344),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1344),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1344),
    [sym__blank_line_start] = ACTIONS(1344),
    [sym__html_block_1_start] = ACTIONS(1344),
    [sym__html_block_2_start] = ACTIONS(1344),
    [sym__html_block_3_start] = ACTIONS(1344),
    [sym__html_block_4_start] = ACTIONS(1344),
    [sym__html_block_5_start] = ACTIONS(1344),
    [sym__html_block_6_start] = ACTIONS(1344),
    [sym__html_block_7_start] = ACTIONS(1344),
    [sym__pipe_table_start] = ACTIONS(1344),
  },
  [STATE(178)] = {
    [ts_builtin_sym_end] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1430),
    [anon_sym_RBRACK] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_POUND] = ACTIONS(1432),
    [anon_sym_DOLLAR] = ACTIONS(1432),
    [anon_sym_PERCENT] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_COLON] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1432),
    [anon_sym_EQ] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_BSLASH] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1432),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_BQUOTE] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1432),
    [anon_sym_RPAREN] = ACTIONS(1432),
    [aux_sym__word_token1] = ACTIONS(1432),
    [aux_sym__word_token2] = ACTIONS(1432),
    [aux_sym__word_token3] = ACTIONS(1432),
    [sym__whitespace] = ACTIONS(1432),
    [sym__soft_line_ending] = ACTIONS(1432),
    [sym_block_continuation] = ACTIONS(1522),
    [sym__block_quote_start] = ACTIONS(1432),
    [sym__indented_chunk_start] = ACTIONS(1432),
    [sym_atx_h1_marker] = ACTIONS(1432),
    [sym_atx_h2_marker] = ACTIONS(1432),
    [sym_atx_h3_marker] = ACTIONS(1432),
    [sym_atx_h4_marker] = ACTIONS(1432),
    [sym_atx_h5_marker] = ACTIONS(1432),
    [sym_atx_h6_marker] = ACTIONS(1432),
    [sym__thematic_break] = ACTIONS(1432),
    [sym__list_marker_minus] = ACTIONS(1432),
    [sym__list_marker_plus] = ACTIONS(1432),
    [sym__list_marker_star] = ACTIONS(1432),
    [sym__list_marker_parenthesis] = ACTIONS(1432),
    [sym__list_marker_dot] = ACTIONS(1432),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1432),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1432),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1432),
    [sym__blank_line_start] = ACTIONS(1432),
    [sym__html_block_1_start] = ACTIONS(1432),
    [sym__html_block_2_start] = ACTIONS(1432),
    [sym__html_block_3_start] = ACTIONS(1432),
    [sym__html_block_4_start] = ACTIONS(1432),
    [sym__html_block_5_start] = ACTIONS(1432),
    [sym__html_block_6_start] = ACTIONS(1432),
    [sym__html_block_7_start] = ACTIONS(1432),
    [sym__pipe_table_start] = ACTIONS(1432),
  },
  [STATE(179)] = {
    [ts_builtin_sym_end] = ACTIONS(1350),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_RBRACK] = ACTIONS(1350),
    [anon_sym_LT] = ACTIONS(1350),
    [anon_sym_GT] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [anon_sym_POUND] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_PLUS] = ACTIONS(1350),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_EQ] = ACTIONS(1350),
    [anon_sym_QMARK] = ACTIONS(1350),
    [anon_sym_AT] = ACTIONS(1350),
    [anon_sym_BSLASH] = ACTIONS(1350),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym__] = ACTIONS(1350),
    [anon_sym_BQUOTE] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_RPAREN] = ACTIONS(1350),
    [aux_sym__word_token1] = ACTIONS(1350),
    [aux_sym__word_token2] = ACTIONS(1350),
    [aux_sym__word_token3] = ACTIONS(1350),
    [sym__whitespace] = ACTIONS(1350),
    [sym__soft_line_ending] = ACTIONS(1350),
    [sym_block_continuation] = ACTIONS(1524),
    [sym__block_quote_start] = ACTIONS(1350),
    [sym__indented_chunk_start] = ACTIONS(1350),
    [sym_atx_h1_marker] = ACTIONS(1350),
    [sym_atx_h2_marker] = ACTIONS(1350),
    [sym_atx_h3_marker] = ACTIONS(1350),
    [sym_atx_h4_marker] = ACTIONS(1350),
    [sym_atx_h5_marker] = ACTIONS(1350),
    [sym_atx_h6_marker] = ACTIONS(1350),
    [sym__thematic_break] = ACTIONS(1350),
    [sym__list_marker_minus] = ACTIONS(1350),
    [sym__list_marker_plus] = ACTIONS(1350),
    [sym__list_marker_star] = ACTIONS(1350),
    [sym__list_marker_parenthesis] = ACTIONS(1350),
    [sym__list_marker_dot] = ACTIONS(1350),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1350),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1350),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1350),
    [sym__blank_line_start] = ACTIONS(1350),
    [sym__html_block_1_start] = ACTIONS(1350),
    [sym__html_block_2_start] = ACTIONS(1350),
    [sym__html_block_3_start] = ACTIONS(1350),
    [sym__html_block_4_start] = ACTIONS(1350),
    [sym__html_block_5_start] = ACTIONS(1350),
    [sym__html_block_6_start] = ACTIONS(1350),
    [sym__html_block_7_start] = ACTIONS(1350),
    [sym__pipe_table_start] = ACTIONS(1350),
  },
  [STATE(180)] = {
    [ts_builtin_sym_end] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [anon_sym_RBRACK] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_POUND] = ACTIONS(1396),
    [anon_sym_DOLLAR] = ACTIONS(1396),
    [anon_sym_PERCENT] = ACTIONS(1396),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_COLON] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1396),
    [anon_sym_EQ] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_BSLASH] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1396),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_BQUOTE] = ACTIONS(1396),
    [anon_sym_LBRACE] = ACTIONS(1396),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_RBRACE] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1396),
    [anon_sym_RPAREN] = ACTIONS(1396),
    [aux_sym__word_token1] = ACTIONS(1396),
    [aux_sym__word_token2] = ACTIONS(1396),
    [aux_sym__word_token3] = ACTIONS(1396),
    [sym__whitespace] = ACTIONS(1396),
    [sym__soft_line_ending] = ACTIONS(1396),
    [sym_block_continuation] = ACTIONS(1526),
    [sym__block_quote_start] = ACTIONS(1396),
    [sym__indented_chunk_start] = ACTIONS(1396),
    [sym_atx_h1_marker] = ACTIONS(1396),
    [sym_atx_h2_marker] = ACTIONS(1396),
    [sym_atx_h3_marker] = ACTIONS(1396),
    [sym_atx_h4_marker] = ACTIONS(1396),
    [sym_atx_h5_marker] = ACTIONS(1396),
    [sym_atx_h6_marker] = ACTIONS(1396),
    [sym__thematic_break] = ACTIONS(1396),
    [sym__list_marker_minus] = ACTIONS(1396),
    [sym__list_marker_plus] = ACTIONS(1396),
    [sym__list_marker_star] = ACTIONS(1396),
    [sym__list_marker_parenthesis] = ACTIONS(1396),
    [sym__list_marker_dot] = ACTIONS(1396),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1396),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1396),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1396),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1396),
    [sym__blank_line_start] = ACTIONS(1396),
    [sym__html_block_1_start] = ACTIONS(1396),
    [sym__html_block_2_start] = ACTIONS(1396),
    [sym__html_block_3_start] = ACTIONS(1396),
    [sym__html_block_4_start] = ACTIONS(1396),
    [sym__html_block_5_start] = ACTIONS(1396),
    [sym__html_block_6_start] = ACTIONS(1396),
    [sym__html_block_7_start] = ACTIONS(1396),
    [sym__pipe_table_start] = ACTIONS(1396),
  },
  [STATE(181)] = {
    [ts_builtin_sym_end] = ACTIONS(1414),
    [anon_sym_LBRACK] = ACTIONS(1412),
    [anon_sym_RBRACK] = ACTIONS(1414),
    [anon_sym_LT] = ACTIONS(1414),
    [anon_sym_GT] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1414),
    [anon_sym_POUND] = ACTIONS(1414),
    [anon_sym_DOLLAR] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1414),
    [anon_sym_SQUOTE] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1414),
    [anon_sym_PLUS] = ACTIONS(1414),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_DASH] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1414),
    [anon_sym_COLON] = ACTIONS(1414),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_EQ] = ACTIONS(1414),
    [anon_sym_QMARK] = ACTIONS(1414),
    [anon_sym_AT] = ACTIONS(1414),
    [anon_sym_BSLASH] = ACTIONS(1414),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym__] = ACTIONS(1414),
    [anon_sym_BQUOTE] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_RPAREN] = ACTIONS(1414),
    [aux_sym__word_token1] = ACTIONS(1414),
    [aux_sym__word_token2] = ACTIONS(1414),
    [aux_sym__word_token3] = ACTIONS(1414),
    [sym__whitespace] = ACTIONS(1414),
    [sym__soft_line_ending] = ACTIONS(1414),
    [sym_block_continuation] = ACTIONS(1528),
    [sym__block_quote_start] = ACTIONS(1414),
    [sym__indented_chunk_start] = ACTIONS(1414),
    [sym_atx_h1_marker] = ACTIONS(1414),
    [sym_atx_h2_marker] = ACTIONS(1414),
    [sym_atx_h3_marker] = ACTIONS(1414),
    [sym_atx_h4_marker] = ACTIONS(1414),
    [sym_atx_h5_marker] = ACTIONS(1414),
    [sym_atx_h6_marker] = ACTIONS(1414),
    [sym__thematic_break] = ACTIONS(1414),
    [sym__list_marker_minus] = ACTIONS(1414),
    [sym__list_marker_plus] = ACTIONS(1414),
    [sym__list_marker_star] = ACTIONS(1414),
    [sym__list_marker_parenthesis] = ACTIONS(1414),
    [sym__list_marker_dot] = ACTIONS(1414),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1414),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1414),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1414),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1414),
    [sym__blank_line_start] = ACTIONS(1414),
    [sym__html_block_1_start] = ACTIONS(1414),
    [sym__html_block_2_start] = ACTIONS(1414),
    [sym__html_block_3_start] = ACTIONS(1414),
    [sym__html_block_4_start] = ACTIONS(1414),
    [sym__html_block_5_start] = ACTIONS(1414),
    [sym__html_block_6_start] = ACTIONS(1414),
    [sym__html_block_7_start] = ACTIONS(1414),
    [sym__pipe_table_start] = ACTIONS(1414),
  },
  [STATE(182)] = {
    [ts_builtin_sym_end] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [anon_sym_RBRACK] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_POUND] = ACTIONS(1420),
    [anon_sym_DOLLAR] = ACTIONS(1420),
    [anon_sym_PERCENT] = ACTIONS(1420),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_COLON] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1420),
    [anon_sym_EQ] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_BSLASH] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1420),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_BQUOTE] = ACTIONS(1420),
    [anon_sym_LBRACE] = ACTIONS(1420),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_RBRACE] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1420),
    [anon_sym_RPAREN] = ACTIONS(1420),
    [aux_sym__word_token1] = ACTIONS(1420),
    [aux_sym__word_token2] = ACTIONS(1420),
    [aux_sym__word_token3] = ACTIONS(1420),
    [sym__whitespace] = ACTIONS(1420),
    [sym__soft_line_ending] = ACTIONS(1420),
    [sym_block_continuation] = ACTIONS(1530),
    [sym__block_quote_start] = ACTIONS(1420),
    [sym__indented_chunk_start] = ACTIONS(1420),
    [sym_atx_h1_marker] = ACTIONS(1420),
    [sym_atx_h2_marker] = ACTIONS(1420),
    [sym_atx_h3_marker] = ACTIONS(1420),
    [sym_atx_h4_marker] = ACTIONS(1420),
    [sym_atx_h5_marker] = ACTIONS(1420),
    [sym_atx_h6_marker] = ACTIONS(1420),
    [sym__thematic_break] = ACTIONS(1420),
    [sym__list_marker_minus] = ACTIONS(1420),
    [sym__list_marker_plus] = ACTIONS(1420),
    [sym__list_marker_star] = ACTIONS(1420),
    [sym__list_marker_parenthesis] = ACTIONS(1420),
    [sym__list_marker_dot] = ACTIONS(1420),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1420),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1420),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1420),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1420),
    [sym__blank_line_start] = ACTIONS(1420),
    [sym__html_block_1_start] = ACTIONS(1420),
    [sym__html_block_2_start] = ACTIONS(1420),
    [sym__html_block_3_start] = ACTIONS(1420),
    [sym__html_block_4_start] = ACTIONS(1420),
    [sym__html_block_5_start] = ACTIONS(1420),
    [sym__html_block_6_start] = ACTIONS(1420),
    [sym__html_block_7_start] = ACTIONS(1420),
    [sym__pipe_table_start] = ACTIONS(1420),
  },
  [STATE(183)] = {
    [anon_sym_LBRACK] = ACTIONS(1506),
    [anon_sym_RBRACK] = ACTIONS(1504),
    [anon_sym_LT] = ACTIONS(1504),
    [anon_sym_GT] = ACTIONS(1504),
    [anon_sym_BANG] = ACTIONS(1504),
    [anon_sym_DQUOTE] = ACTIONS(1504),
    [anon_sym_POUND] = ACTIONS(1504),
    [anon_sym_DOLLAR] = ACTIONS(1504),
    [anon_sym_PERCENT] = ACTIONS(1504),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_SQUOTE] = ACTIONS(1504),
    [anon_sym_STAR] = ACTIONS(1504),
    [anon_sym_PLUS] = ACTIONS(1504),
    [anon_sym_COMMA] = ACTIONS(1504),
    [anon_sym_DASH] = ACTIONS(1504),
    [anon_sym_DOT] = ACTIONS(1504),
    [anon_sym_SLASH] = ACTIONS(1504),
    [anon_sym_COLON] = ACTIONS(1504),
    [anon_sym_SEMI] = ACTIONS(1504),
    [anon_sym_EQ] = ACTIONS(1504),
    [anon_sym_QMARK] = ACTIONS(1504),
    [anon_sym_AT] = ACTIONS(1504),
    [anon_sym_BSLASH] = ACTIONS(1504),
    [anon_sym_CARET] = ACTIONS(1504),
    [anon_sym__] = ACTIONS(1504),
    [anon_sym_BQUOTE] = ACTIONS(1504),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_PIPE] = ACTIONS(1504),
    [anon_sym_RBRACE] = ACTIONS(1504),
    [anon_sym_TILDE] = ACTIONS(1504),
    [anon_sym_LPAREN] = ACTIONS(1504),
    [anon_sym_RPAREN] = ACTIONS(1504),
    [aux_sym__word_token1] = ACTIONS(1504),
    [aux_sym__word_token2] = ACTIONS(1504),
    [aux_sym__word_token3] = ACTIONS(1504),
    [sym__whitespace] = ACTIONS(1504),
    [sym__soft_line_ending] = ACTIONS(1504),
    [sym__block_close] = ACTIONS(1504),
    [sym_block_continuation] = ACTIONS(1532),
    [sym__block_quote_start] = ACTIONS(1504),
    [sym__indented_chunk_start] = ACTIONS(1504),
    [sym_atx_h1_marker] = ACTIONS(1504),
    [sym_atx_h2_marker] = ACTIONS(1504),
    [sym_atx_h3_marker] = ACTIONS(1504),
    [sym_atx_h4_marker] = ACTIONS(1504),
    [sym_atx_h5_marker] = ACTIONS(1504),
    [sym_atx_h6_marker] = ACTIONS(1504),
    [sym__thematic_break] = ACTIONS(1504),
    [sym__list_marker_minus] = ACTIONS(1504),
    [sym__list_marker_plus] = ACTIONS(1504),
    [sym__list_marker_star] = ACTIONS(1504),
    [sym__list_marker_parenthesis] = ACTIONS(1504),
    [sym__list_marker_dot] = ACTIONS(1504),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1504),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1504),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1504),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1504),
    [sym__blank_line_start] = ACTIONS(1504),
    [sym__html_block_1_start] = ACTIONS(1504),
    [sym__html_block_2_start] = ACTIONS(1504),
    [sym__html_block_3_start] = ACTIONS(1504),
    [sym__html_block_4_start] = ACTIONS(1504),
    [sym__html_block_5_start] = ACTIONS(1504),
    [sym__html_block_6_start] = ACTIONS(1504),
    [sym__html_block_7_start] = ACTIONS(1504),
    [sym__pipe_table_start] = ACTIONS(1504),
  },
  [STATE(184)] = {
    [anon_sym_LBRACK] = ACTIONS(1534),
    [anon_sym_RBRACK] = ACTIONS(1536),
    [anon_sym_LT] = ACTIONS(1536),
    [anon_sym_GT] = ACTIONS(1536),
    [anon_sym_BANG] = ACTIONS(1536),
    [anon_sym_DQUOTE] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(1536),
    [anon_sym_DOLLAR] = ACTIONS(1536),
    [anon_sym_PERCENT] = ACTIONS(1536),
    [anon_sym_AMP] = ACTIONS(1536),
    [anon_sym_SQUOTE] = ACTIONS(1536),
    [anon_sym_STAR] = ACTIONS(1536),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_DOT] = ACTIONS(1536),
    [anon_sym_SLASH] = ACTIONS(1536),
    [anon_sym_COLON] = ACTIONS(1536),
    [anon_sym_SEMI] = ACTIONS(1536),
    [anon_sym_EQ] = ACTIONS(1536),
    [anon_sym_QMARK] = ACTIONS(1536),
    [anon_sym_AT] = ACTIONS(1536),
    [anon_sym_BSLASH] = ACTIONS(1536),
    [anon_sym_CARET] = ACTIONS(1536),
    [anon_sym__] = ACTIONS(1536),
    [anon_sym_BQUOTE] = ACTIONS(1536),
    [anon_sym_LBRACE] = ACTIONS(1536),
    [anon_sym_PIPE] = ACTIONS(1536),
    [anon_sym_RBRACE] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1536),
    [anon_sym_LPAREN] = ACTIONS(1536),
    [anon_sym_RPAREN] = ACTIONS(1536),
    [aux_sym__word_token1] = ACTIONS(1536),
    [aux_sym__word_token2] = ACTIONS(1536),
    [aux_sym__word_token3] = ACTIONS(1536),
    [sym__whitespace] = ACTIONS(1536),
    [sym__soft_line_ending] = ACTIONS(1536),
    [sym__block_close] = ACTIONS(1536),
    [sym__block_quote_start] = ACTIONS(1536),
    [sym__indented_chunk_start] = ACTIONS(1536),
    [sym_atx_h1_marker] = ACTIONS(1536),
    [sym_atx_h2_marker] = ACTIONS(1536),
    [sym_atx_h3_marker] = ACTIONS(1536),
    [sym_atx_h4_marker] = ACTIONS(1536),
    [sym_atx_h5_marker] = ACTIONS(1536),
    [sym_atx_h6_marker] = ACTIONS(1536),
    [sym__thematic_break] = ACTIONS(1536),
    [sym__list_marker_minus] = ACTIONS(1536),
    [sym__list_marker_plus] = ACTIONS(1536),
    [sym__list_marker_star] = ACTIONS(1536),
    [sym__list_marker_parenthesis] = ACTIONS(1536),
    [sym__list_marker_dot] = ACTIONS(1536),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1536),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1536),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1536),
    [sym__blank_line_start] = ACTIONS(1536),
    [sym__html_block_1_start] = ACTIONS(1536),
    [sym__html_block_2_start] = ACTIONS(1536),
    [sym__html_block_3_start] = ACTIONS(1536),
    [sym__html_block_4_start] = ACTIONS(1536),
    [sym__html_block_5_start] = ACTIONS(1536),
    [sym__html_block_6_start] = ACTIONS(1536),
    [sym__html_block_7_start] = ACTIONS(1536),
    [sym__pipe_table_start] = ACTIONS(1536),
  },
  [STATE(185)] = {
    [ts_builtin_sym_end] = ACTIONS(1538),
    [anon_sym_LBRACK] = ACTIONS(1540),
    [anon_sym_RBRACK] = ACTIONS(1538),
    [anon_sym_LT] = ACTIONS(1538),
    [anon_sym_GT] = ACTIONS(1538),
    [anon_sym_BANG] = ACTIONS(1538),
    [anon_sym_DQUOTE] = ACTIONS(1538),
    [anon_sym_POUND] = ACTIONS(1538),
    [anon_sym_DOLLAR] = ACTIONS(1538),
    [anon_sym_PERCENT] = ACTIONS(1538),
    [anon_sym_AMP] = ACTIONS(1538),
    [anon_sym_SQUOTE] = ACTIONS(1538),
    [anon_sym_STAR] = ACTIONS(1538),
    [anon_sym_PLUS] = ACTIONS(1538),
    [anon_sym_COMMA] = ACTIONS(1538),
    [anon_sym_DASH] = ACTIONS(1538),
    [anon_sym_DOT] = ACTIONS(1538),
    [anon_sym_SLASH] = ACTIONS(1538),
    [anon_sym_COLON] = ACTIONS(1538),
    [anon_sym_SEMI] = ACTIONS(1538),
    [anon_sym_EQ] = ACTIONS(1538),
    [anon_sym_QMARK] = ACTIONS(1538),
    [anon_sym_AT] = ACTIONS(1538),
    [anon_sym_BSLASH] = ACTIONS(1538),
    [anon_sym_CARET] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(1538),
    [anon_sym_BQUOTE] = ACTIONS(1538),
    [anon_sym_LBRACE] = ACTIONS(1538),
    [anon_sym_PIPE] = ACTIONS(1538),
    [anon_sym_RBRACE] = ACTIONS(1538),
    [anon_sym_TILDE] = ACTIONS(1538),
    [anon_sym_LPAREN] = ACTIONS(1538),
    [anon_sym_RPAREN] = ACTIONS(1538),
    [aux_sym__word_token1] = ACTIONS(1538),
    [aux_sym__word_token2] = ACTIONS(1538),
    [aux_sym__word_token3] = ACTIONS(1538),
    [sym__whitespace] = ACTIONS(1538),
    [sym__soft_line_ending] = ACTIONS(1538),
    [sym__block_quote_start] = ACTIONS(1538),
    [sym__indented_chunk_start] = ACTIONS(1538),
    [sym_atx_h1_marker] = ACTIONS(1538),
    [sym_atx_h2_marker] = ACTIONS(1538),
    [sym_atx_h3_marker] = ACTIONS(1538),
    [sym_atx_h4_marker] = ACTIONS(1538),
    [sym_atx_h5_marker] = ACTIONS(1538),
    [sym_atx_h6_marker] = ACTIONS(1538),
    [sym__thematic_break] = ACTIONS(1538),
    [sym__list_marker_minus] = ACTIONS(1538),
    [sym__list_marker_plus] = ACTIONS(1538),
    [sym__list_marker_star] = ACTIONS(1538),
    [sym__list_marker_parenthesis] = ACTIONS(1538),
    [sym__list_marker_dot] = ACTIONS(1538),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1538),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1538),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1538),
    [sym__blank_line_start] = ACTIONS(1538),
    [sym__html_block_1_start] = ACTIONS(1538),
    [sym__html_block_2_start] = ACTIONS(1538),
    [sym__html_block_3_start] = ACTIONS(1538),
    [sym__html_block_4_start] = ACTIONS(1538),
    [sym__html_block_5_start] = ACTIONS(1538),
    [sym__html_block_6_start] = ACTIONS(1538),
    [sym__html_block_7_start] = ACTIONS(1538),
    [sym__pipe_table_start] = ACTIONS(1538),
  },
  [STATE(186)] = {
    [ts_builtin_sym_end] = ACTIONS(1542),
    [anon_sym_LBRACK] = ACTIONS(1544),
    [anon_sym_RBRACK] = ACTIONS(1542),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_BANG] = ACTIONS(1542),
    [anon_sym_DQUOTE] = ACTIONS(1542),
    [anon_sym_POUND] = ACTIONS(1542),
    [anon_sym_DOLLAR] = ACTIONS(1542),
    [anon_sym_PERCENT] = ACTIONS(1542),
    [anon_sym_AMP] = ACTIONS(1542),
    [anon_sym_SQUOTE] = ACTIONS(1542),
    [anon_sym_STAR] = ACTIONS(1542),
    [anon_sym_PLUS] = ACTIONS(1542),
    [anon_sym_COMMA] = ACTIONS(1542),
    [anon_sym_DASH] = ACTIONS(1542),
    [anon_sym_DOT] = ACTIONS(1542),
    [anon_sym_SLASH] = ACTIONS(1542),
    [anon_sym_COLON] = ACTIONS(1542),
    [anon_sym_SEMI] = ACTIONS(1542),
    [anon_sym_EQ] = ACTIONS(1542),
    [anon_sym_QMARK] = ACTIONS(1542),
    [anon_sym_AT] = ACTIONS(1542),
    [anon_sym_BSLASH] = ACTIONS(1542),
    [anon_sym_CARET] = ACTIONS(1542),
    [anon_sym__] = ACTIONS(1542),
    [anon_sym_BQUOTE] = ACTIONS(1542),
    [anon_sym_LBRACE] = ACTIONS(1542),
    [anon_sym_PIPE] = ACTIONS(1542),
    [anon_sym_RBRACE] = ACTIONS(1542),
    [anon_sym_TILDE] = ACTIONS(1542),
    [anon_sym_LPAREN] = ACTIONS(1542),
    [anon_sym_RPAREN] = ACTIONS(1542),
    [aux_sym__word_token1] = ACTIONS(1542),
    [aux_sym__word_token2] = ACTIONS(1542),
    [aux_sym__word_token3] = ACTIONS(1542),
    [sym__whitespace] = ACTIONS(1542),
    [sym__soft_line_ending] = ACTIONS(1542),
    [sym__block_quote_start] = ACTIONS(1542),
    [sym__indented_chunk_start] = ACTIONS(1542),
    [sym_atx_h1_marker] = ACTIONS(1542),
    [sym_atx_h2_marker] = ACTIONS(1542),
    [sym_atx_h3_marker] = ACTIONS(1542),
    [sym_atx_h4_marker] = ACTIONS(1542),
    [sym_atx_h5_marker] = ACTIONS(1542),
    [sym_atx_h6_marker] = ACTIONS(1542),
    [sym__thematic_break] = ACTIONS(1542),
    [sym__list_marker_minus] = ACTIONS(1542),
    [sym__list_marker_plus] = ACTIONS(1542),
    [sym__list_marker_star] = ACTIONS(1542),
    [sym__list_marker_parenthesis] = ACTIONS(1542),
    [sym__list_marker_dot] = ACTIONS(1542),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1542),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1542),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1542),
    [sym__blank_line_start] = ACTIONS(1542),
    [sym__html_block_1_start] = ACTIONS(1542),
    [sym__html_block_2_start] = ACTIONS(1542),
    [sym__html_block_3_start] = ACTIONS(1542),
    [sym__html_block_4_start] = ACTIONS(1542),
    [sym__html_block_5_start] = ACTIONS(1542),
    [sym__html_block_6_start] = ACTIONS(1542),
    [sym__html_block_7_start] = ACTIONS(1542),
    [sym__pipe_table_start] = ACTIONS(1542),
  },
  [STATE(187)] = {
    [ts_builtin_sym_end] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1222),
    [anon_sym_RBRACK] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1224),
    [anon_sym_GT] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [anon_sym_POUND] = ACTIONS(1224),
    [anon_sym_DOLLAR] = ACTIONS(1224),
    [anon_sym_PERCENT] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_PLUS] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_SLASH] = ACTIONS(1224),
    [anon_sym_COLON] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_EQ] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1224),
    [anon_sym_BSLASH] = ACTIONS(1224),
    [anon_sym_CARET] = ACTIONS(1224),
    [anon_sym__] = ACTIONS(1224),
    [anon_sym_BQUOTE] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_PIPE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_RPAREN] = ACTIONS(1224),
    [aux_sym__word_token1] = ACTIONS(1224),
    [aux_sym__word_token2] = ACTIONS(1224),
    [aux_sym__word_token3] = ACTIONS(1224),
    [sym__whitespace] = ACTIONS(1224),
    [sym__soft_line_ending] = ACTIONS(1224),
    [sym__block_quote_start] = ACTIONS(1224),
    [sym__indented_chunk_start] = ACTIONS(1224),
    [sym_atx_h1_marker] = ACTIONS(1224),
    [sym_atx_h2_marker] = ACTIONS(1224),
    [sym_atx_h3_marker] = ACTIONS(1224),
    [sym_atx_h4_marker] = ACTIONS(1224),
    [sym_atx_h5_marker] = ACTIONS(1224),
    [sym_atx_h6_marker] = ACTIONS(1224),
    [sym__thematic_break] = ACTIONS(1224),
    [sym__list_marker_minus] = ACTIONS(1224),
    [sym__list_marker_plus] = ACTIONS(1224),
    [sym__list_marker_star] = ACTIONS(1224),
    [sym__list_marker_parenthesis] = ACTIONS(1224),
    [sym__list_marker_dot] = ACTIONS(1224),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1224),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1224),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1224),
    [sym__blank_line_start] = ACTIONS(1224),
    [sym__html_block_1_start] = ACTIONS(1224),
    [sym__html_block_2_start] = ACTIONS(1224),
    [sym__html_block_3_start] = ACTIONS(1224),
    [sym__html_block_4_start] = ACTIONS(1224),
    [sym__html_block_5_start] = ACTIONS(1224),
    [sym__html_block_6_start] = ACTIONS(1224),
    [sym__html_block_7_start] = ACTIONS(1224),
    [sym__pipe_table_start] = ACTIONS(1224),
  },
  [STATE(188)] = {
    [ts_builtin_sym_end] = ACTIONS(1332),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [anon_sym_RBRACK] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_COLON] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1332),
    [anon_sym_BSLASH] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym__] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [aux_sym__word_token1] = ACTIONS(1332),
    [aux_sym__word_token2] = ACTIONS(1332),
    [aux_sym__word_token3] = ACTIONS(1332),
    [sym__whitespace] = ACTIONS(1332),
    [sym__soft_line_ending] = ACTIONS(1332),
    [sym__block_quote_start] = ACTIONS(1332),
    [sym__indented_chunk_start] = ACTIONS(1332),
    [sym_atx_h1_marker] = ACTIONS(1332),
    [sym_atx_h2_marker] = ACTIONS(1332),
    [sym_atx_h3_marker] = ACTIONS(1332),
    [sym_atx_h4_marker] = ACTIONS(1332),
    [sym_atx_h5_marker] = ACTIONS(1332),
    [sym_atx_h6_marker] = ACTIONS(1332),
    [sym__thematic_break] = ACTIONS(1332),
    [sym__list_marker_minus] = ACTIONS(1332),
    [sym__list_marker_plus] = ACTIONS(1332),
    [sym__list_marker_star] = ACTIONS(1332),
    [sym__list_marker_parenthesis] = ACTIONS(1332),
    [sym__list_marker_dot] = ACTIONS(1332),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1332),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1332),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1332),
    [sym__blank_line_start] = ACTIONS(1332),
    [sym__html_block_1_start] = ACTIONS(1332),
    [sym__html_block_2_start] = ACTIONS(1332),
    [sym__html_block_3_start] = ACTIONS(1332),
    [sym__html_block_4_start] = ACTIONS(1332),
    [sym__html_block_5_start] = ACTIONS(1332),
    [sym__html_block_6_start] = ACTIONS(1332),
    [sym__html_block_7_start] = ACTIONS(1332),
    [sym__pipe_table_start] = ACTIONS(1332),
  },
  [STATE(189)] = {
    [ts_builtin_sym_end] = ACTIONS(1546),
    [anon_sym_LBRACK] = ACTIONS(1548),
    [anon_sym_RBRACK] = ACTIONS(1546),
    [anon_sym_LT] = ACTIONS(1546),
    [anon_sym_GT] = ACTIONS(1546),
    [anon_sym_BANG] = ACTIONS(1546),
    [anon_sym_DQUOTE] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_DOLLAR] = ACTIONS(1546),
    [anon_sym_PERCENT] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1546),
    [anon_sym_STAR] = ACTIONS(1546),
    [anon_sym_PLUS] = ACTIONS(1546),
    [anon_sym_COMMA] = ACTIONS(1546),
    [anon_sym_DASH] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1546),
    [anon_sym_SLASH] = ACTIONS(1546),
    [anon_sym_COLON] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1546),
    [anon_sym_EQ] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_AT] = ACTIONS(1546),
    [anon_sym_BSLASH] = ACTIONS(1546),
    [anon_sym_CARET] = ACTIONS(1546),
    [anon_sym__] = ACTIONS(1546),
    [anon_sym_BQUOTE] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_RBRACE] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1546),
    [anon_sym_RPAREN] = ACTIONS(1546),
    [aux_sym__word_token1] = ACTIONS(1546),
    [aux_sym__word_token2] = ACTIONS(1546),
    [aux_sym__word_token3] = ACTIONS(1546),
    [sym__whitespace] = ACTIONS(1546),
    [sym__soft_line_ending] = ACTIONS(1546),
    [sym__block_quote_start] = ACTIONS(1546),
    [sym__indented_chunk_start] = ACTIONS(1546),
    [sym_atx_h1_marker] = ACTIONS(1546),
    [sym_atx_h2_marker] = ACTIONS(1546),
    [sym_atx_h3_marker] = ACTIONS(1546),
    [sym_atx_h4_marker] = ACTIONS(1546),
    [sym_atx_h5_marker] = ACTIONS(1546),
    [sym_atx_h6_marker] = ACTIONS(1546),
    [sym__thematic_break] = ACTIONS(1546),
    [sym__list_marker_minus] = ACTIONS(1546),
    [sym__list_marker_plus] = ACTIONS(1546),
    [sym__list_marker_star] = ACTIONS(1546),
    [sym__list_marker_parenthesis] = ACTIONS(1546),
    [sym__list_marker_dot] = ACTIONS(1546),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1546),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1546),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1546),
    [sym__blank_line_start] = ACTIONS(1546),
    [sym__html_block_1_start] = ACTIONS(1546),
    [sym__html_block_2_start] = ACTIONS(1546),
    [sym__html_block_3_start] = ACTIONS(1546),
    [sym__html_block_4_start] = ACTIONS(1546),
    [sym__html_block_5_start] = ACTIONS(1546),
    [sym__html_block_6_start] = ACTIONS(1546),
    [sym__html_block_7_start] = ACTIONS(1546),
    [sym__pipe_table_start] = ACTIONS(1546),
  },
  [STATE(190)] = {
    [ts_builtin_sym_end] = ACTIONS(1550),
    [anon_sym_LBRACK] = ACTIONS(1552),
    [anon_sym_RBRACK] = ACTIONS(1550),
    [anon_sym_LT] = ACTIONS(1550),
    [anon_sym_GT] = ACTIONS(1550),
    [anon_sym_BANG] = ACTIONS(1550),
    [anon_sym_DQUOTE] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_DOLLAR] = ACTIONS(1550),
    [anon_sym_PERCENT] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1550),
    [anon_sym_STAR] = ACTIONS(1550),
    [anon_sym_PLUS] = ACTIONS(1550),
    [anon_sym_COMMA] = ACTIONS(1550),
    [anon_sym_DASH] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1550),
    [anon_sym_SLASH] = ACTIONS(1550),
    [anon_sym_COLON] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_EQ] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_AT] = ACTIONS(1550),
    [anon_sym_BSLASH] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1550),
    [anon_sym__] = ACTIONS(1550),
    [anon_sym_BQUOTE] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_RBRACE] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1550),
    [anon_sym_RPAREN] = ACTIONS(1550),
    [aux_sym__word_token1] = ACTIONS(1550),
    [aux_sym__word_token2] = ACTIONS(1550),
    [aux_sym__word_token3] = ACTIONS(1550),
    [sym__whitespace] = ACTIONS(1550),
    [sym__soft_line_ending] = ACTIONS(1550),
    [sym__block_quote_start] = ACTIONS(1550),
    [sym__indented_chunk_start] = ACTIONS(1550),
    [sym_atx_h1_marker] = ACTIONS(1550),
    [sym_atx_h2_marker] = ACTIONS(1550),
    [sym_atx_h3_marker] = ACTIONS(1550),
    [sym_atx_h4_marker] = ACTIONS(1550),
    [sym_atx_h5_marker] = ACTIONS(1550),
    [sym_atx_h6_marker] = ACTIONS(1550),
    [sym__thematic_break] = ACTIONS(1550),
    [sym__list_marker_minus] = ACTIONS(1550),
    [sym__list_marker_plus] = ACTIONS(1550),
    [sym__list_marker_star] = ACTIONS(1550),
    [sym__list_marker_parenthesis] = ACTIONS(1550),
    [sym__list_marker_dot] = ACTIONS(1550),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1550),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1550),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1550),
    [sym__blank_line_start] = ACTIONS(1550),
    [sym__html_block_1_start] = ACTIONS(1550),
    [sym__html_block_2_start] = ACTIONS(1550),
    [sym__html_block_3_start] = ACTIONS(1550),
    [sym__html_block_4_start] = ACTIONS(1550),
    [sym__html_block_5_start] = ACTIONS(1550),
    [sym__html_block_6_start] = ACTIONS(1550),
    [sym__html_block_7_start] = ACTIONS(1550),
    [sym__pipe_table_start] = ACTIONS(1550),
  },
  [STATE(191)] = {
    [ts_builtin_sym_end] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1556),
    [anon_sym_RBRACK] = ACTIONS(1554),
    [anon_sym_LT] = ACTIONS(1554),
    [anon_sym_GT] = ACTIONS(1554),
    [anon_sym_BANG] = ACTIONS(1554),
    [anon_sym_DQUOTE] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_DOLLAR] = ACTIONS(1554),
    [anon_sym_PERCENT] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1554),
    [anon_sym_STAR] = ACTIONS(1554),
    [anon_sym_PLUS] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1554),
    [anon_sym_SLASH] = ACTIONS(1554),
    [anon_sym_COLON] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_EQ] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_AT] = ACTIONS(1554),
    [anon_sym_BSLASH] = ACTIONS(1554),
    [anon_sym_CARET] = ACTIONS(1554),
    [anon_sym__] = ACTIONS(1554),
    [anon_sym_BQUOTE] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_RBRACE] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1554),
    [anon_sym_RPAREN] = ACTIONS(1554),
    [aux_sym__word_token1] = ACTIONS(1554),
    [aux_sym__word_token2] = ACTIONS(1554),
    [aux_sym__word_token3] = ACTIONS(1554),
    [sym__whitespace] = ACTIONS(1554),
    [sym__soft_line_ending] = ACTIONS(1554),
    [sym__block_quote_start] = ACTIONS(1554),
    [sym__indented_chunk_start] = ACTIONS(1554),
    [sym_atx_h1_marker] = ACTIONS(1554),
    [sym_atx_h2_marker] = ACTIONS(1554),
    [sym_atx_h3_marker] = ACTIONS(1554),
    [sym_atx_h4_marker] = ACTIONS(1554),
    [sym_atx_h5_marker] = ACTIONS(1554),
    [sym_atx_h6_marker] = ACTIONS(1554),
    [sym__thematic_break] = ACTIONS(1554),
    [sym__list_marker_minus] = ACTIONS(1554),
    [sym__list_marker_plus] = ACTIONS(1554),
    [sym__list_marker_star] = ACTIONS(1554),
    [sym__list_marker_parenthesis] = ACTIONS(1554),
    [sym__list_marker_dot] = ACTIONS(1554),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1554),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1554),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1554),
    [sym__blank_line_start] = ACTIONS(1554),
    [sym__html_block_1_start] = ACTIONS(1554),
    [sym__html_block_2_start] = ACTIONS(1554),
    [sym__html_block_3_start] = ACTIONS(1554),
    [sym__html_block_4_start] = ACTIONS(1554),
    [sym__html_block_5_start] = ACTIONS(1554),
    [sym__html_block_6_start] = ACTIONS(1554),
    [sym__html_block_7_start] = ACTIONS(1554),
    [sym__pipe_table_start] = ACTIONS(1554),
  },
  [STATE(192)] = {
    [ts_builtin_sym_end] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1560),
    [anon_sym_RBRACK] = ACTIONS(1558),
    [anon_sym_LT] = ACTIONS(1558),
    [anon_sym_GT] = ACTIONS(1558),
    [anon_sym_BANG] = ACTIONS(1558),
    [anon_sym_DQUOTE] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [anon_sym_PERCENT] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1558),
    [anon_sym_STAR] = ACTIONS(1558),
    [anon_sym_PLUS] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1558),
    [anon_sym_DASH] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1558),
    [anon_sym_SLASH] = ACTIONS(1558),
    [anon_sym_COLON] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_EQ] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_AT] = ACTIONS(1558),
    [anon_sym_BSLASH] = ACTIONS(1558),
    [anon_sym_CARET] = ACTIONS(1558),
    [anon_sym__] = ACTIONS(1558),
    [anon_sym_BQUOTE] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_RBRACE] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1558),
    [anon_sym_RPAREN] = ACTIONS(1558),
    [aux_sym__word_token1] = ACTIONS(1558),
    [aux_sym__word_token2] = ACTIONS(1558),
    [aux_sym__word_token3] = ACTIONS(1558),
    [sym__whitespace] = ACTIONS(1558),
    [sym__soft_line_ending] = ACTIONS(1558),
    [sym__block_quote_start] = ACTIONS(1558),
    [sym__indented_chunk_start] = ACTIONS(1558),
    [sym_atx_h1_marker] = ACTIONS(1558),
    [sym_atx_h2_marker] = ACTIONS(1558),
    [sym_atx_h3_marker] = ACTIONS(1558),
    [sym_atx_h4_marker] = ACTIONS(1558),
    [sym_atx_h5_marker] = ACTIONS(1558),
    [sym_atx_h6_marker] = ACTIONS(1558),
    [sym__thematic_break] = ACTIONS(1558),
    [sym__list_marker_minus] = ACTIONS(1558),
    [sym__list_marker_plus] = ACTIONS(1558),
    [sym__list_marker_star] = ACTIONS(1558),
    [sym__list_marker_parenthesis] = ACTIONS(1558),
    [sym__list_marker_dot] = ACTIONS(1558),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1558),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1558),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1558),
    [sym__blank_line_start] = ACTIONS(1558),
    [sym__html_block_1_start] = ACTIONS(1558),
    [sym__html_block_2_start] = ACTIONS(1558),
    [sym__html_block_3_start] = ACTIONS(1558),
    [sym__html_block_4_start] = ACTIONS(1558),
    [sym__html_block_5_start] = ACTIONS(1558),
    [sym__html_block_6_start] = ACTIONS(1558),
    [sym__html_block_7_start] = ACTIONS(1558),
    [sym__pipe_table_start] = ACTIONS(1558),
  },
  [STATE(193)] = {
    [ts_builtin_sym_end] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1564),
    [anon_sym_RBRACK] = ACTIONS(1562),
    [anon_sym_LT] = ACTIONS(1562),
    [anon_sym_GT] = ACTIONS(1562),
    [anon_sym_BANG] = ACTIONS(1562),
    [anon_sym_DQUOTE] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_DOLLAR] = ACTIONS(1562),
    [anon_sym_PERCENT] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1562),
    [anon_sym_STAR] = ACTIONS(1562),
    [anon_sym_PLUS] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1562),
    [anon_sym_DASH] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1562),
    [anon_sym_SLASH] = ACTIONS(1562),
    [anon_sym_COLON] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_EQ] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_AT] = ACTIONS(1562),
    [anon_sym_BSLASH] = ACTIONS(1562),
    [anon_sym_CARET] = ACTIONS(1562),
    [anon_sym__] = ACTIONS(1562),
    [anon_sym_BQUOTE] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_RBRACE] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1562),
    [anon_sym_RPAREN] = ACTIONS(1562),
    [aux_sym__word_token1] = ACTIONS(1562),
    [aux_sym__word_token2] = ACTIONS(1562),
    [aux_sym__word_token3] = ACTIONS(1562),
    [sym__whitespace] = ACTIONS(1562),
    [sym__soft_line_ending] = ACTIONS(1562),
    [sym__block_quote_start] = ACTIONS(1562),
    [sym__indented_chunk_start] = ACTIONS(1562),
    [sym_atx_h1_marker] = ACTIONS(1562),
    [sym_atx_h2_marker] = ACTIONS(1562),
    [sym_atx_h3_marker] = ACTIONS(1562),
    [sym_atx_h4_marker] = ACTIONS(1562),
    [sym_atx_h5_marker] = ACTIONS(1562),
    [sym_atx_h6_marker] = ACTIONS(1562),
    [sym__thematic_break] = ACTIONS(1562),
    [sym__list_marker_minus] = ACTIONS(1562),
    [sym__list_marker_plus] = ACTIONS(1562),
    [sym__list_marker_star] = ACTIONS(1562),
    [sym__list_marker_parenthesis] = ACTIONS(1562),
    [sym__list_marker_dot] = ACTIONS(1562),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1562),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1562),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1562),
    [sym__blank_line_start] = ACTIONS(1562),
    [sym__html_block_1_start] = ACTIONS(1562),
    [sym__html_block_2_start] = ACTIONS(1562),
    [sym__html_block_3_start] = ACTIONS(1562),
    [sym__html_block_4_start] = ACTIONS(1562),
    [sym__html_block_5_start] = ACTIONS(1562),
    [sym__html_block_6_start] = ACTIONS(1562),
    [sym__html_block_7_start] = ACTIONS(1562),
    [sym__pipe_table_start] = ACTIONS(1562),
  },
  [STATE(194)] = {
    [ts_builtin_sym_end] = ACTIONS(1566),
    [anon_sym_LBRACK] = ACTIONS(1568),
    [anon_sym_RBRACK] = ACTIONS(1566),
    [anon_sym_LT] = ACTIONS(1566),
    [anon_sym_GT] = ACTIONS(1566),
    [anon_sym_BANG] = ACTIONS(1566),
    [anon_sym_DQUOTE] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_DOLLAR] = ACTIONS(1566),
    [anon_sym_PERCENT] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1566),
    [anon_sym_PLUS] = ACTIONS(1566),
    [anon_sym_COMMA] = ACTIONS(1566),
    [anon_sym_DASH] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1566),
    [anon_sym_SLASH] = ACTIONS(1566),
    [anon_sym_COLON] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1566),
    [anon_sym_EQ] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_AT] = ACTIONS(1566),
    [anon_sym_BSLASH] = ACTIONS(1566),
    [anon_sym_CARET] = ACTIONS(1566),
    [anon_sym__] = ACTIONS(1566),
    [anon_sym_BQUOTE] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_RBRACE] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1566),
    [anon_sym_RPAREN] = ACTIONS(1566),
    [aux_sym__word_token1] = ACTIONS(1566),
    [aux_sym__word_token2] = ACTIONS(1566),
    [aux_sym__word_token3] = ACTIONS(1566),
    [sym__whitespace] = ACTIONS(1566),
    [sym__soft_line_ending] = ACTIONS(1566),
    [sym__block_quote_start] = ACTIONS(1566),
    [sym__indented_chunk_start] = ACTIONS(1566),
    [sym_atx_h1_marker] = ACTIONS(1566),
    [sym_atx_h2_marker] = ACTIONS(1566),
    [sym_atx_h3_marker] = ACTIONS(1566),
    [sym_atx_h4_marker] = ACTIONS(1566),
    [sym_atx_h5_marker] = ACTIONS(1566),
    [sym_atx_h6_marker] = ACTIONS(1566),
    [sym__thematic_break] = ACTIONS(1566),
    [sym__list_marker_minus] = ACTIONS(1566),
    [sym__list_marker_plus] = ACTIONS(1566),
    [sym__list_marker_star] = ACTIONS(1566),
    [sym__list_marker_parenthesis] = ACTIONS(1566),
    [sym__list_marker_dot] = ACTIONS(1566),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1566),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1566),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1566),
    [sym__blank_line_start] = ACTIONS(1566),
    [sym__html_block_1_start] = ACTIONS(1566),
    [sym__html_block_2_start] = ACTIONS(1566),
    [sym__html_block_3_start] = ACTIONS(1566),
    [sym__html_block_4_start] = ACTIONS(1566),
    [sym__html_block_5_start] = ACTIONS(1566),
    [sym__html_block_6_start] = ACTIONS(1566),
    [sym__html_block_7_start] = ACTIONS(1566),
    [sym__pipe_table_start] = ACTIONS(1566),
  },
  [STATE(195)] = {
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_LT] = ACTIONS(1572),
    [anon_sym_GT] = ACTIONS(1572),
    [anon_sym_BANG] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [anon_sym_POUND] = ACTIONS(1572),
    [anon_sym_DOLLAR] = ACTIONS(1572),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_AMP] = ACTIONS(1572),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_STAR] = ACTIONS(1572),
    [anon_sym_PLUS] = ACTIONS(1572),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_DASH] = ACTIONS(1572),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1572),
    [anon_sym_COLON] = ACTIONS(1572),
    [anon_sym_SEMI] = ACTIONS(1572),
    [anon_sym_EQ] = ACTIONS(1572),
    [anon_sym_QMARK] = ACTIONS(1572),
    [anon_sym_AT] = ACTIONS(1572),
    [anon_sym_BSLASH] = ACTIONS(1572),
    [anon_sym_CARET] = ACTIONS(1572),
    [anon_sym__] = ACTIONS(1572),
    [anon_sym_BQUOTE] = ACTIONS(1572),
    [anon_sym_LBRACE] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1572),
    [anon_sym_RBRACE] = ACTIONS(1572),
    [anon_sym_TILDE] = ACTIONS(1572),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_RPAREN] = ACTIONS(1572),
    [aux_sym__word_token1] = ACTIONS(1572),
    [aux_sym__word_token2] = ACTIONS(1572),
    [aux_sym__word_token3] = ACTIONS(1572),
    [sym__whitespace] = ACTIONS(1572),
    [sym__soft_line_ending] = ACTIONS(1572),
    [sym__block_close] = ACTIONS(1572),
    [sym__block_quote_start] = ACTIONS(1572),
    [sym__indented_chunk_start] = ACTIONS(1572),
    [sym_atx_h1_marker] = ACTIONS(1572),
    [sym_atx_h2_marker] = ACTIONS(1572),
    [sym_atx_h3_marker] = ACTIONS(1572),
    [sym_atx_h4_marker] = ACTIONS(1572),
    [sym_atx_h5_marker] = ACTIONS(1572),
    [sym_atx_h6_marker] = ACTIONS(1572),
    [sym__thematic_break] = ACTIONS(1572),
    [sym__list_marker_minus] = ACTIONS(1572),
    [sym__list_marker_plus] = ACTIONS(1572),
    [sym__list_marker_star] = ACTIONS(1572),
    [sym__list_marker_parenthesis] = ACTIONS(1572),
    [sym__list_marker_dot] = ACTIONS(1572),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1572),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1572),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1572),
    [sym__blank_line_start] = ACTIONS(1572),
    [sym__html_block_1_start] = ACTIONS(1572),
    [sym__html_block_2_start] = ACTIONS(1572),
    [sym__html_block_3_start] = ACTIONS(1572),
    [sym__html_block_4_start] = ACTIONS(1572),
    [sym__html_block_5_start] = ACTIONS(1572),
    [sym__html_block_6_start] = ACTIONS(1572),
    [sym__html_block_7_start] = ACTIONS(1572),
    [sym__pipe_table_start] = ACTIONS(1572),
  },
  [STATE(196)] = {
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1576),
    [anon_sym_LT] = ACTIONS(1576),
    [anon_sym_GT] = ACTIONS(1576),
    [anon_sym_BANG] = ACTIONS(1576),
    [anon_sym_DQUOTE] = ACTIONS(1576),
    [anon_sym_POUND] = ACTIONS(1576),
    [anon_sym_DOLLAR] = ACTIONS(1576),
    [anon_sym_PERCENT] = ACTIONS(1576),
    [anon_sym_AMP] = ACTIONS(1576),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_STAR] = ACTIONS(1576),
    [anon_sym_PLUS] = ACTIONS(1576),
    [anon_sym_COMMA] = ACTIONS(1576),
    [anon_sym_DASH] = ACTIONS(1576),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_SLASH] = ACTIONS(1576),
    [anon_sym_COLON] = ACTIONS(1576),
    [anon_sym_SEMI] = ACTIONS(1576),
    [anon_sym_EQ] = ACTIONS(1576),
    [anon_sym_QMARK] = ACTIONS(1576),
    [anon_sym_AT] = ACTIONS(1576),
    [anon_sym_BSLASH] = ACTIONS(1576),
    [anon_sym_CARET] = ACTIONS(1576),
    [anon_sym__] = ACTIONS(1576),
    [anon_sym_BQUOTE] = ACTIONS(1576),
    [anon_sym_LBRACE] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1576),
    [anon_sym_RBRACE] = ACTIONS(1576),
    [anon_sym_TILDE] = ACTIONS(1576),
    [anon_sym_LPAREN] = ACTIONS(1576),
    [anon_sym_RPAREN] = ACTIONS(1576),
    [aux_sym__word_token1] = ACTIONS(1576),
    [aux_sym__word_token2] = ACTIONS(1576),
    [aux_sym__word_token3] = ACTIONS(1576),
    [sym__whitespace] = ACTIONS(1576),
    [sym__soft_line_ending] = ACTIONS(1576),
    [sym__block_close] = ACTIONS(1576),
    [sym__block_quote_start] = ACTIONS(1576),
    [sym__indented_chunk_start] = ACTIONS(1576),
    [sym_atx_h1_marker] = ACTIONS(1576),
    [sym_atx_h2_marker] = ACTIONS(1576),
    [sym_atx_h3_marker] = ACTIONS(1576),
    [sym_atx_h4_marker] = ACTIONS(1576),
    [sym_atx_h5_marker] = ACTIONS(1576),
    [sym_atx_h6_marker] = ACTIONS(1576),
    [sym__thematic_break] = ACTIONS(1576),
    [sym__list_marker_minus] = ACTIONS(1576),
    [sym__list_marker_plus] = ACTIONS(1576),
    [sym__list_marker_star] = ACTIONS(1576),
    [sym__list_marker_parenthesis] = ACTIONS(1576),
    [sym__list_marker_dot] = ACTIONS(1576),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1576),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1576),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1576),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1576),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1576),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1576),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1576),
    [sym__blank_line_start] = ACTIONS(1576),
    [sym__html_block_1_start] = ACTIONS(1576),
    [sym__html_block_2_start] = ACTIONS(1576),
    [sym__html_block_3_start] = ACTIONS(1576),
    [sym__html_block_4_start] = ACTIONS(1576),
    [sym__html_block_5_start] = ACTIONS(1576),
    [sym__html_block_6_start] = ACTIONS(1576),
    [sym__html_block_7_start] = ACTIONS(1576),
    [sym__pipe_table_start] = ACTIONS(1576),
  },
  [STATE(197)] = {
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1580),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_BANG] = ACTIONS(1580),
    [anon_sym_DQUOTE] = ACTIONS(1580),
    [anon_sym_POUND] = ACTIONS(1580),
    [anon_sym_DOLLAR] = ACTIONS(1580),
    [anon_sym_PERCENT] = ACTIONS(1580),
    [anon_sym_AMP] = ACTIONS(1580),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_STAR] = ACTIONS(1580),
    [anon_sym_PLUS] = ACTIONS(1580),
    [anon_sym_COMMA] = ACTIONS(1580),
    [anon_sym_DASH] = ACTIONS(1580),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_SLASH] = ACTIONS(1580),
    [anon_sym_COLON] = ACTIONS(1580),
    [anon_sym_SEMI] = ACTIONS(1580),
    [anon_sym_EQ] = ACTIONS(1580),
    [anon_sym_QMARK] = ACTIONS(1580),
    [anon_sym_AT] = ACTIONS(1580),
    [anon_sym_BSLASH] = ACTIONS(1580),
    [anon_sym_CARET] = ACTIONS(1580),
    [anon_sym__] = ACTIONS(1580),
    [anon_sym_BQUOTE] = ACTIONS(1580),
    [anon_sym_LBRACE] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1580),
    [anon_sym_RBRACE] = ACTIONS(1580),
    [anon_sym_TILDE] = ACTIONS(1580),
    [anon_sym_LPAREN] = ACTIONS(1580),
    [anon_sym_RPAREN] = ACTIONS(1580),
    [aux_sym__word_token1] = ACTIONS(1580),
    [aux_sym__word_token2] = ACTIONS(1580),
    [aux_sym__word_token3] = ACTIONS(1580),
    [sym__whitespace] = ACTIONS(1580),
    [sym__soft_line_ending] = ACTIONS(1580),
    [sym_block_continuation] = ACTIONS(1580),
    [sym__block_quote_start] = ACTIONS(1580),
    [sym__indented_chunk_start] = ACTIONS(1580),
    [sym_atx_h1_marker] = ACTIONS(1580),
    [sym_atx_h2_marker] = ACTIONS(1580),
    [sym_atx_h3_marker] = ACTIONS(1580),
    [sym_atx_h4_marker] = ACTIONS(1580),
    [sym_atx_h5_marker] = ACTIONS(1580),
    [sym_atx_h6_marker] = ACTIONS(1580),
    [sym__thematic_break] = ACTIONS(1580),
    [sym__list_marker_minus] = ACTIONS(1580),
    [sym__list_marker_plus] = ACTIONS(1580),
    [sym__list_marker_star] = ACTIONS(1580),
    [sym__list_marker_parenthesis] = ACTIONS(1580),
    [sym__list_marker_dot] = ACTIONS(1580),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1580),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1580),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1580),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1580),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1580),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1580),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1580),
    [sym__blank_line_start] = ACTIONS(1580),
    [sym__html_block_1_start] = ACTIONS(1580),
    [sym__html_block_2_start] = ACTIONS(1580),
    [sym__html_block_3_start] = ACTIONS(1580),
    [sym__html_block_4_start] = ACTIONS(1580),
    [sym__html_block_5_start] = ACTIONS(1580),
    [sym__html_block_6_start] = ACTIONS(1580),
    [sym__html_block_7_start] = ACTIONS(1580),
    [sym__pipe_table_start] = ACTIONS(1580),
  },
  [STATE(198)] = {
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1584),
    [anon_sym_LT] = ACTIONS(1584),
    [anon_sym_GT] = ACTIONS(1584),
    [anon_sym_BANG] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1584),
    [anon_sym_POUND] = ACTIONS(1584),
    [anon_sym_DOLLAR] = ACTIONS(1584),
    [anon_sym_PERCENT] = ACTIONS(1584),
    [anon_sym_AMP] = ACTIONS(1584),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_STAR] = ACTIONS(1584),
    [anon_sym_PLUS] = ACTIONS(1584),
    [anon_sym_COMMA] = ACTIONS(1584),
    [anon_sym_DASH] = ACTIONS(1584),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_SLASH] = ACTIONS(1584),
    [anon_sym_COLON] = ACTIONS(1584),
    [anon_sym_SEMI] = ACTIONS(1584),
    [anon_sym_EQ] = ACTIONS(1584),
    [anon_sym_QMARK] = ACTIONS(1584),
    [anon_sym_AT] = ACTIONS(1584),
    [anon_sym_BSLASH] = ACTIONS(1584),
    [anon_sym_CARET] = ACTIONS(1584),
    [anon_sym__] = ACTIONS(1584),
    [anon_sym_BQUOTE] = ACTIONS(1584),
    [anon_sym_LBRACE] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1584),
    [anon_sym_RBRACE] = ACTIONS(1584),
    [anon_sym_TILDE] = ACTIONS(1584),
    [anon_sym_LPAREN] = ACTIONS(1584),
    [anon_sym_RPAREN] = ACTIONS(1584),
    [aux_sym__word_token1] = ACTIONS(1584),
    [aux_sym__word_token2] = ACTIONS(1584),
    [aux_sym__word_token3] = ACTIONS(1584),
    [sym__whitespace] = ACTIONS(1584),
    [sym__soft_line_ending] = ACTIONS(1584),
    [sym_block_continuation] = ACTIONS(1584),
    [sym__block_quote_start] = ACTIONS(1584),
    [sym__indented_chunk_start] = ACTIONS(1584),
    [sym_atx_h1_marker] = ACTIONS(1584),
    [sym_atx_h2_marker] = ACTIONS(1584),
    [sym_atx_h3_marker] = ACTIONS(1584),
    [sym_atx_h4_marker] = ACTIONS(1584),
    [sym_atx_h5_marker] = ACTIONS(1584),
    [sym_atx_h6_marker] = ACTIONS(1584),
    [sym__thematic_break] = ACTIONS(1584),
    [sym__list_marker_minus] = ACTIONS(1584),
    [sym__list_marker_plus] = ACTIONS(1584),
    [sym__list_marker_star] = ACTIONS(1584),
    [sym__list_marker_parenthesis] = ACTIONS(1584),
    [sym__list_marker_dot] = ACTIONS(1584),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1584),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1584),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1584),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1584),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1584),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1584),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1584),
    [sym__blank_line_start] = ACTIONS(1584),
    [sym__html_block_1_start] = ACTIONS(1584),
    [sym__html_block_2_start] = ACTIONS(1584),
    [sym__html_block_3_start] = ACTIONS(1584),
    [sym__html_block_4_start] = ACTIONS(1584),
    [sym__html_block_5_start] = ACTIONS(1584),
    [sym__html_block_6_start] = ACTIONS(1584),
    [sym__html_block_7_start] = ACTIONS(1584),
    [sym__pipe_table_start] = ACTIONS(1584),
  },
  [STATE(199)] = {
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_RBRACK] = ACTIONS(1588),
    [anon_sym_LT] = ACTIONS(1588),
    [anon_sym_GT] = ACTIONS(1588),
    [anon_sym_BANG] = ACTIONS(1588),
    [anon_sym_DQUOTE] = ACTIONS(1588),
    [anon_sym_POUND] = ACTIONS(1588),
    [anon_sym_DOLLAR] = ACTIONS(1588),
    [anon_sym_PERCENT] = ACTIONS(1588),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_STAR] = ACTIONS(1588),
    [anon_sym_PLUS] = ACTIONS(1588),
    [anon_sym_COMMA] = ACTIONS(1588),
    [anon_sym_DASH] = ACTIONS(1588),
    [anon_sym_DOT] = ACTIONS(1588),
    [anon_sym_SLASH] = ACTIONS(1588),
    [anon_sym_COLON] = ACTIONS(1588),
    [anon_sym_SEMI] = ACTIONS(1588),
    [anon_sym_EQ] = ACTIONS(1588),
    [anon_sym_QMARK] = ACTIONS(1588),
    [anon_sym_AT] = ACTIONS(1588),
    [anon_sym_BSLASH] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1588),
    [anon_sym__] = ACTIONS(1588),
    [anon_sym_BQUOTE] = ACTIONS(1588),
    [anon_sym_LBRACE] = ACTIONS(1588),
    [anon_sym_PIPE] = ACTIONS(1588),
    [anon_sym_RBRACE] = ACTIONS(1588),
    [anon_sym_TILDE] = ACTIONS(1588),
    [anon_sym_LPAREN] = ACTIONS(1588),
    [anon_sym_RPAREN] = ACTIONS(1588),
    [aux_sym__word_token1] = ACTIONS(1588),
    [aux_sym__word_token2] = ACTIONS(1588),
    [aux_sym__word_token3] = ACTIONS(1588),
    [sym__whitespace] = ACTIONS(1588),
    [sym__soft_line_ending] = ACTIONS(1588),
    [sym_block_continuation] = ACTIONS(1588),
    [sym__block_quote_start] = ACTIONS(1588),
    [sym__indented_chunk_start] = ACTIONS(1588),
    [sym_atx_h1_marker] = ACTIONS(1588),
    [sym_atx_h2_marker] = ACTIONS(1588),
    [sym_atx_h3_marker] = ACTIONS(1588),
    [sym_atx_h4_marker] = ACTIONS(1588),
    [sym_atx_h5_marker] = ACTIONS(1588),
    [sym_atx_h6_marker] = ACTIONS(1588),
    [sym__thematic_break] = ACTIONS(1588),
    [sym__list_marker_minus] = ACTIONS(1588),
    [sym__list_marker_plus] = ACTIONS(1588),
    [sym__list_marker_star] = ACTIONS(1588),
    [sym__list_marker_parenthesis] = ACTIONS(1588),
    [sym__list_marker_dot] = ACTIONS(1588),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1588),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1588),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1588),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1588),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1588),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1588),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1588),
    [sym__blank_line_start] = ACTIONS(1588),
    [sym__html_block_1_start] = ACTIONS(1588),
    [sym__html_block_2_start] = ACTIONS(1588),
    [sym__html_block_3_start] = ACTIONS(1588),
    [sym__html_block_4_start] = ACTIONS(1588),
    [sym__html_block_5_start] = ACTIONS(1588),
    [sym__html_block_6_start] = ACTIONS(1588),
    [sym__html_block_7_start] = ACTIONS(1588),
    [sym__pipe_table_start] = ACTIONS(1588),
  },
  [STATE(200)] = {
    [ts_builtin_sym_end] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_RBRACK] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_POUND] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1338),
    [anon_sym_PERCENT] = ACTIONS(1338),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_COLON] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1338),
    [anon_sym_EQ] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_BSLASH] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1338),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_BQUOTE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1338),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_RBRACE] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1338),
    [anon_sym_RPAREN] = ACTIONS(1338),
    [aux_sym__word_token1] = ACTIONS(1338),
    [aux_sym__word_token2] = ACTIONS(1338),
    [aux_sym__word_token3] = ACTIONS(1338),
    [sym__whitespace] = ACTIONS(1338),
    [sym__soft_line_ending] = ACTIONS(1338),
    [sym__block_quote_start] = ACTIONS(1338),
    [sym__indented_chunk_start] = ACTIONS(1338),
    [sym_atx_h1_marker] = ACTIONS(1338),
    [sym_atx_h2_marker] = ACTIONS(1338),
    [sym_atx_h3_marker] = ACTIONS(1338),
    [sym_atx_h4_marker] = ACTIONS(1338),
    [sym_atx_h5_marker] = ACTIONS(1338),
    [sym_atx_h6_marker] = ACTIONS(1338),
    [sym__thematic_break] = ACTIONS(1338),
    [sym__list_marker_minus] = ACTIONS(1338),
    [sym__list_marker_plus] = ACTIONS(1338),
    [sym__list_marker_star] = ACTIONS(1338),
    [sym__list_marker_parenthesis] = ACTIONS(1338),
    [sym__list_marker_dot] = ACTIONS(1338),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1338),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1338),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1338),
    [sym__blank_line_start] = ACTIONS(1338),
    [sym__html_block_1_start] = ACTIONS(1338),
    [sym__html_block_2_start] = ACTIONS(1338),
    [sym__html_block_3_start] = ACTIONS(1338),
    [sym__html_block_4_start] = ACTIONS(1338),
    [sym__html_block_5_start] = ACTIONS(1338),
    [sym__html_block_6_start] = ACTIONS(1338),
    [sym__html_block_7_start] = ACTIONS(1338),
    [sym__pipe_table_start] = ACTIONS(1338),
  },
  [STATE(201)] = {
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_RBRACK] = ACTIONS(1592),
    [anon_sym_LT] = ACTIONS(1592),
    [anon_sym_GT] = ACTIONS(1592),
    [anon_sym_BANG] = ACTIONS(1592),
    [anon_sym_DQUOTE] = ACTIONS(1592),
    [anon_sym_POUND] = ACTIONS(1592),
    [anon_sym_DOLLAR] = ACTIONS(1592),
    [anon_sym_PERCENT] = ACTIONS(1592),
    [anon_sym_AMP] = ACTIONS(1592),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_STAR] = ACTIONS(1592),
    [anon_sym_PLUS] = ACTIONS(1592),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_DASH] = ACTIONS(1592),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_SLASH] = ACTIONS(1592),
    [anon_sym_COLON] = ACTIONS(1592),
    [anon_sym_SEMI] = ACTIONS(1592),
    [anon_sym_EQ] = ACTIONS(1592),
    [anon_sym_QMARK] = ACTIONS(1592),
    [anon_sym_AT] = ACTIONS(1592),
    [anon_sym_BSLASH] = ACTIONS(1592),
    [anon_sym_CARET] = ACTIONS(1592),
    [anon_sym__] = ACTIONS(1592),
    [anon_sym_BQUOTE] = ACTIONS(1592),
    [anon_sym_LBRACE] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_RBRACE] = ACTIONS(1592),
    [anon_sym_TILDE] = ACTIONS(1592),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_RPAREN] = ACTIONS(1592),
    [aux_sym__word_token1] = ACTIONS(1592),
    [aux_sym__word_token2] = ACTIONS(1592),
    [aux_sym__word_token3] = ACTIONS(1592),
    [sym__whitespace] = ACTIONS(1592),
    [sym__soft_line_ending] = ACTIONS(1592),
    [sym_block_continuation] = ACTIONS(1592),
    [sym__block_quote_start] = ACTIONS(1592),
    [sym__indented_chunk_start] = ACTIONS(1592),
    [sym_atx_h1_marker] = ACTIONS(1592),
    [sym_atx_h2_marker] = ACTIONS(1592),
    [sym_atx_h3_marker] = ACTIONS(1592),
    [sym_atx_h4_marker] = ACTIONS(1592),
    [sym_atx_h5_marker] = ACTIONS(1592),
    [sym_atx_h6_marker] = ACTIONS(1592),
    [sym__thematic_break] = ACTIONS(1592),
    [sym__list_marker_minus] = ACTIONS(1592),
    [sym__list_marker_plus] = ACTIONS(1592),
    [sym__list_marker_star] = ACTIONS(1592),
    [sym__list_marker_parenthesis] = ACTIONS(1592),
    [sym__list_marker_dot] = ACTIONS(1592),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1592),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1592),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1592),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1592),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1592),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1592),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1592),
    [sym__blank_line_start] = ACTIONS(1592),
    [sym__html_block_1_start] = ACTIONS(1592),
    [sym__html_block_2_start] = ACTIONS(1592),
    [sym__html_block_3_start] = ACTIONS(1592),
    [sym__html_block_4_start] = ACTIONS(1592),
    [sym__html_block_5_start] = ACTIONS(1592),
    [sym__html_block_6_start] = ACTIONS(1592),
    [sym__html_block_7_start] = ACTIONS(1592),
    [sym__pipe_table_start] = ACTIONS(1592),
  },
  [STATE(202)] = {
    [ts_builtin_sym_end] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1596),
    [anon_sym_RBRACK] = ACTIONS(1594),
    [anon_sym_LT] = ACTIONS(1594),
    [anon_sym_GT] = ACTIONS(1594),
    [anon_sym_BANG] = ACTIONS(1594),
    [anon_sym_DQUOTE] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_DOLLAR] = ACTIONS(1594),
    [anon_sym_PERCENT] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1594),
    [anon_sym_STAR] = ACTIONS(1594),
    [anon_sym_PLUS] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1594),
    [anon_sym_SLASH] = ACTIONS(1594),
    [anon_sym_COLON] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_EQ] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_AT] = ACTIONS(1594),
    [anon_sym_BSLASH] = ACTIONS(1594),
    [anon_sym_CARET] = ACTIONS(1594),
    [anon_sym__] = ACTIONS(1594),
    [anon_sym_BQUOTE] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_RBRACE] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1594),
    [anon_sym_RPAREN] = ACTIONS(1594),
    [aux_sym__word_token1] = ACTIONS(1594),
    [aux_sym__word_token2] = ACTIONS(1594),
    [aux_sym__word_token3] = ACTIONS(1594),
    [sym__whitespace] = ACTIONS(1594),
    [sym__soft_line_ending] = ACTIONS(1594),
    [sym__block_quote_start] = ACTIONS(1594),
    [sym__indented_chunk_start] = ACTIONS(1594),
    [sym_atx_h1_marker] = ACTIONS(1594),
    [sym_atx_h2_marker] = ACTIONS(1594),
    [sym_atx_h3_marker] = ACTIONS(1594),
    [sym_atx_h4_marker] = ACTIONS(1594),
    [sym_atx_h5_marker] = ACTIONS(1594),
    [sym_atx_h6_marker] = ACTIONS(1594),
    [sym__thematic_break] = ACTIONS(1594),
    [sym__list_marker_minus] = ACTIONS(1594),
    [sym__list_marker_plus] = ACTIONS(1594),
    [sym__list_marker_star] = ACTIONS(1594),
    [sym__list_marker_parenthesis] = ACTIONS(1594),
    [sym__list_marker_dot] = ACTIONS(1594),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1594),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1594),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1594),
    [sym__blank_line_start] = ACTIONS(1594),
    [sym__html_block_1_start] = ACTIONS(1594),
    [sym__html_block_2_start] = ACTIONS(1594),
    [sym__html_block_3_start] = ACTIONS(1594),
    [sym__html_block_4_start] = ACTIONS(1594),
    [sym__html_block_5_start] = ACTIONS(1594),
    [sym__html_block_6_start] = ACTIONS(1594),
    [sym__html_block_7_start] = ACTIONS(1594),
    [sym__pipe_table_start] = ACTIONS(1594),
  },
  [STATE(203)] = {
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_LT] = ACTIONS(1600),
    [anon_sym_GT] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [anon_sym_POUND] = ACTIONS(1600),
    [anon_sym_DOLLAR] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1600),
    [anon_sym_AMP] = ACTIONS(1600),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_STAR] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_DASH] = ACTIONS(1600),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_SLASH] = ACTIONS(1600),
    [anon_sym_COLON] = ACTIONS(1600),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_EQ] = ACTIONS(1600),
    [anon_sym_QMARK] = ACTIONS(1600),
    [anon_sym_AT] = ACTIONS(1600),
    [anon_sym_BSLASH] = ACTIONS(1600),
    [anon_sym_CARET] = ACTIONS(1600),
    [anon_sym__] = ACTIONS(1600),
    [anon_sym_BQUOTE] = ACTIONS(1600),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [aux_sym__word_token1] = ACTIONS(1600),
    [aux_sym__word_token2] = ACTIONS(1600),
    [aux_sym__word_token3] = ACTIONS(1600),
    [sym__whitespace] = ACTIONS(1600),
    [sym__soft_line_ending] = ACTIONS(1600),
    [sym__block_close] = ACTIONS(1600),
    [sym__block_quote_start] = ACTIONS(1600),
    [sym__indented_chunk_start] = ACTIONS(1600),
    [sym_atx_h1_marker] = ACTIONS(1600),
    [sym_atx_h2_marker] = ACTIONS(1600),
    [sym_atx_h3_marker] = ACTIONS(1600),
    [sym_atx_h4_marker] = ACTIONS(1600),
    [sym_atx_h5_marker] = ACTIONS(1600),
    [sym_atx_h6_marker] = ACTIONS(1600),
    [sym__thematic_break] = ACTIONS(1600),
    [sym__list_marker_minus] = ACTIONS(1600),
    [sym__list_marker_plus] = ACTIONS(1600),
    [sym__list_marker_star] = ACTIONS(1600),
    [sym__list_marker_parenthesis] = ACTIONS(1600),
    [sym__list_marker_dot] = ACTIONS(1600),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1600),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1600),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1600),
    [sym__blank_line_start] = ACTIONS(1600),
    [sym__html_block_1_start] = ACTIONS(1600),
    [sym__html_block_2_start] = ACTIONS(1600),
    [sym__html_block_3_start] = ACTIONS(1600),
    [sym__html_block_4_start] = ACTIONS(1600),
    [sym__html_block_5_start] = ACTIONS(1600),
    [sym__html_block_6_start] = ACTIONS(1600),
    [sym__html_block_7_start] = ACTIONS(1600),
    [sym__pipe_table_start] = ACTIONS(1600),
  },
  [STATE(204)] = {
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_LT] = ACTIONS(1600),
    [anon_sym_GT] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [anon_sym_POUND] = ACTIONS(1600),
    [anon_sym_DOLLAR] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1600),
    [anon_sym_AMP] = ACTIONS(1600),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_STAR] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_DASH] = ACTIONS(1600),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_SLASH] = ACTIONS(1600),
    [anon_sym_COLON] = ACTIONS(1600),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_EQ] = ACTIONS(1600),
    [anon_sym_QMARK] = ACTIONS(1600),
    [anon_sym_AT] = ACTIONS(1600),
    [anon_sym_BSLASH] = ACTIONS(1600),
    [anon_sym_CARET] = ACTIONS(1600),
    [anon_sym__] = ACTIONS(1600),
    [anon_sym_BQUOTE] = ACTIONS(1600),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [aux_sym__word_token1] = ACTIONS(1600),
    [aux_sym__word_token2] = ACTIONS(1600),
    [aux_sym__word_token3] = ACTIONS(1600),
    [sym__whitespace] = ACTIONS(1600),
    [sym__soft_line_ending] = ACTIONS(1600),
    [sym__block_close] = ACTIONS(1600),
    [sym__block_quote_start] = ACTIONS(1600),
    [sym__indented_chunk_start] = ACTIONS(1600),
    [sym_atx_h1_marker] = ACTIONS(1600),
    [sym_atx_h2_marker] = ACTIONS(1600),
    [sym_atx_h3_marker] = ACTIONS(1600),
    [sym_atx_h4_marker] = ACTIONS(1600),
    [sym_atx_h5_marker] = ACTIONS(1600),
    [sym_atx_h6_marker] = ACTIONS(1600),
    [sym__thematic_break] = ACTIONS(1600),
    [sym__list_marker_minus] = ACTIONS(1600),
    [sym__list_marker_plus] = ACTIONS(1600),
    [sym__list_marker_star] = ACTIONS(1600),
    [sym__list_marker_parenthesis] = ACTIONS(1600),
    [sym__list_marker_dot] = ACTIONS(1600),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1600),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1600),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1600),
    [sym__blank_line_start] = ACTIONS(1600),
    [sym__html_block_1_start] = ACTIONS(1600),
    [sym__html_block_2_start] = ACTIONS(1600),
    [sym__html_block_3_start] = ACTIONS(1600),
    [sym__html_block_4_start] = ACTIONS(1600),
    [sym__html_block_5_start] = ACTIONS(1600),
    [sym__html_block_6_start] = ACTIONS(1600),
    [sym__html_block_7_start] = ACTIONS(1600),
    [sym__pipe_table_start] = ACTIONS(1600),
  },
  [STATE(205)] = {
    [ts_builtin_sym_end] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1604),
    [anon_sym_RBRACK] = ACTIONS(1602),
    [anon_sym_LT] = ACTIONS(1602),
    [anon_sym_GT] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_DOLLAR] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1602),
    [anon_sym_STAR] = ACTIONS(1602),
    [anon_sym_PLUS] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1602),
    [anon_sym_DASH] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1602),
    [anon_sym_SLASH] = ACTIONS(1602),
    [anon_sym_COLON] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_EQ] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_AT] = ACTIONS(1602),
    [anon_sym_BSLASH] = ACTIONS(1602),
    [anon_sym_CARET] = ACTIONS(1602),
    [anon_sym__] = ACTIONS(1602),
    [anon_sym_BQUOTE] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_RPAREN] = ACTIONS(1602),
    [aux_sym__word_token1] = ACTIONS(1602),
    [aux_sym__word_token2] = ACTIONS(1602),
    [aux_sym__word_token3] = ACTIONS(1602),
    [sym__whitespace] = ACTIONS(1602),
    [sym__soft_line_ending] = ACTIONS(1602),
    [sym__block_quote_start] = ACTIONS(1602),
    [sym__indented_chunk_start] = ACTIONS(1602),
    [sym_atx_h1_marker] = ACTIONS(1602),
    [sym_atx_h2_marker] = ACTIONS(1602),
    [sym_atx_h3_marker] = ACTIONS(1602),
    [sym_atx_h4_marker] = ACTIONS(1602),
    [sym_atx_h5_marker] = ACTIONS(1602),
    [sym_atx_h6_marker] = ACTIONS(1602),
    [sym__thematic_break] = ACTIONS(1602),
    [sym__list_marker_minus] = ACTIONS(1602),
    [sym__list_marker_plus] = ACTIONS(1602),
    [sym__list_marker_star] = ACTIONS(1602),
    [sym__list_marker_parenthesis] = ACTIONS(1602),
    [sym__list_marker_dot] = ACTIONS(1602),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1602),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1602),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1602),
    [sym__blank_line_start] = ACTIONS(1602),
    [sym__html_block_1_start] = ACTIONS(1602),
    [sym__html_block_2_start] = ACTIONS(1602),
    [sym__html_block_3_start] = ACTIONS(1602),
    [sym__html_block_4_start] = ACTIONS(1602),
    [sym__html_block_5_start] = ACTIONS(1602),
    [sym__html_block_6_start] = ACTIONS(1602),
    [sym__html_block_7_start] = ACTIONS(1602),
    [sym__pipe_table_start] = ACTIONS(1602),
  },
  [STATE(206)] = {
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_LT] = ACTIONS(1608),
    [anon_sym_GT] = ACTIONS(1608),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_POUND] = ACTIONS(1608),
    [anon_sym_DOLLAR] = ACTIONS(1608),
    [anon_sym_PERCENT] = ACTIONS(1608),
    [anon_sym_AMP] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_STAR] = ACTIONS(1608),
    [anon_sym_PLUS] = ACTIONS(1608),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_DASH] = ACTIONS(1608),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_SLASH] = ACTIONS(1608),
    [anon_sym_COLON] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1608),
    [anon_sym_EQ] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_BSLASH] = ACTIONS(1608),
    [anon_sym_CARET] = ACTIONS(1608),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_BQUOTE] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1608),
    [anon_sym_RBRACE] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_RPAREN] = ACTIONS(1608),
    [aux_sym__word_token1] = ACTIONS(1608),
    [aux_sym__word_token2] = ACTIONS(1608),
    [aux_sym__word_token3] = ACTIONS(1608),
    [sym__whitespace] = ACTIONS(1608),
    [sym__soft_line_ending] = ACTIONS(1608),
    [sym__block_close] = ACTIONS(1608),
    [sym__block_quote_start] = ACTIONS(1608),
    [sym__indented_chunk_start] = ACTIONS(1608),
    [sym_atx_h1_marker] = ACTIONS(1608),
    [sym_atx_h2_marker] = ACTIONS(1608),
    [sym_atx_h3_marker] = ACTIONS(1608),
    [sym_atx_h4_marker] = ACTIONS(1608),
    [sym_atx_h5_marker] = ACTIONS(1608),
    [sym_atx_h6_marker] = ACTIONS(1608),
    [sym__thematic_break] = ACTIONS(1608),
    [sym__list_marker_minus] = ACTIONS(1608),
    [sym__list_marker_plus] = ACTIONS(1608),
    [sym__list_marker_star] = ACTIONS(1608),
    [sym__list_marker_parenthesis] = ACTIONS(1608),
    [sym__list_marker_dot] = ACTIONS(1608),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1608),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1608),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1608),
    [sym__blank_line_start] = ACTIONS(1608),
    [sym__html_block_1_start] = ACTIONS(1608),
    [sym__html_block_2_start] = ACTIONS(1608),
    [sym__html_block_3_start] = ACTIONS(1608),
    [sym__html_block_4_start] = ACTIONS(1608),
    [sym__html_block_5_start] = ACTIONS(1608),
    [sym__html_block_6_start] = ACTIONS(1608),
    [sym__html_block_7_start] = ACTIONS(1608),
    [sym__pipe_table_start] = ACTIONS(1608),
  },
  [STATE(207)] = {
    [ts_builtin_sym_end] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1612),
    [anon_sym_RBRACK] = ACTIONS(1610),
    [anon_sym_LT] = ACTIONS(1610),
    [anon_sym_GT] = ACTIONS(1610),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_DOLLAR] = ACTIONS(1610),
    [anon_sym_PERCENT] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_STAR] = ACTIONS(1610),
    [anon_sym_PLUS] = ACTIONS(1610),
    [anon_sym_COMMA] = ACTIONS(1610),
    [anon_sym_DASH] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1610),
    [anon_sym_SLASH] = ACTIONS(1610),
    [anon_sym_COLON] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1610),
    [anon_sym_EQ] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_BSLASH] = ACTIONS(1610),
    [anon_sym_CARET] = ACTIONS(1610),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_BQUOTE] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_RBRACE] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1610),
    [anon_sym_RPAREN] = ACTIONS(1610),
    [aux_sym__word_token1] = ACTIONS(1610),
    [aux_sym__word_token2] = ACTIONS(1610),
    [aux_sym__word_token3] = ACTIONS(1610),
    [sym__whitespace] = ACTIONS(1610),
    [sym__soft_line_ending] = ACTIONS(1610),
    [sym__block_quote_start] = ACTIONS(1610),
    [sym__indented_chunk_start] = ACTIONS(1610),
    [sym_atx_h1_marker] = ACTIONS(1610),
    [sym_atx_h2_marker] = ACTIONS(1610),
    [sym_atx_h3_marker] = ACTIONS(1610),
    [sym_atx_h4_marker] = ACTIONS(1610),
    [sym_atx_h5_marker] = ACTIONS(1610),
    [sym_atx_h6_marker] = ACTIONS(1610),
    [sym__thematic_break] = ACTIONS(1610),
    [sym__list_marker_minus] = ACTIONS(1610),
    [sym__list_marker_plus] = ACTIONS(1610),
    [sym__list_marker_star] = ACTIONS(1610),
    [sym__list_marker_parenthesis] = ACTIONS(1610),
    [sym__list_marker_dot] = ACTIONS(1610),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1610),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1610),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1610),
    [sym__blank_line_start] = ACTIONS(1610),
    [sym__html_block_1_start] = ACTIONS(1610),
    [sym__html_block_2_start] = ACTIONS(1610),
    [sym__html_block_3_start] = ACTIONS(1610),
    [sym__html_block_4_start] = ACTIONS(1610),
    [sym__html_block_5_start] = ACTIONS(1610),
    [sym__html_block_6_start] = ACTIONS(1610),
    [sym__html_block_7_start] = ACTIONS(1610),
    [sym__pipe_table_start] = ACTIONS(1610),
  },
  [STATE(208)] = {
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_LT] = ACTIONS(1616),
    [anon_sym_GT] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [anon_sym_POUND] = ACTIONS(1616),
    [anon_sym_DOLLAR] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1616),
    [anon_sym_AMP] = ACTIONS(1616),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_STAR] = ACTIONS(1616),
    [anon_sym_PLUS] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_DASH] = ACTIONS(1616),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_SLASH] = ACTIONS(1616),
    [anon_sym_COLON] = ACTIONS(1616),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_EQ] = ACTIONS(1616),
    [anon_sym_QMARK] = ACTIONS(1616),
    [anon_sym_AT] = ACTIONS(1616),
    [anon_sym_BSLASH] = ACTIONS(1616),
    [anon_sym_CARET] = ACTIONS(1616),
    [anon_sym__] = ACTIONS(1616),
    [anon_sym_BQUOTE] = ACTIONS(1616),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [aux_sym__word_token1] = ACTIONS(1616),
    [aux_sym__word_token2] = ACTIONS(1616),
    [aux_sym__word_token3] = ACTIONS(1616),
    [sym__whitespace] = ACTIONS(1616),
    [sym__soft_line_ending] = ACTIONS(1616),
    [sym__block_close] = ACTIONS(1616),
    [sym__block_quote_start] = ACTIONS(1616),
    [sym__indented_chunk_start] = ACTIONS(1616),
    [sym_atx_h1_marker] = ACTIONS(1616),
    [sym_atx_h2_marker] = ACTIONS(1616),
    [sym_atx_h3_marker] = ACTIONS(1616),
    [sym_atx_h4_marker] = ACTIONS(1616),
    [sym_atx_h5_marker] = ACTIONS(1616),
    [sym_atx_h6_marker] = ACTIONS(1616),
    [sym__thematic_break] = ACTIONS(1616),
    [sym__list_marker_minus] = ACTIONS(1616),
    [sym__list_marker_plus] = ACTIONS(1616),
    [sym__list_marker_star] = ACTIONS(1616),
    [sym__list_marker_parenthesis] = ACTIONS(1616),
    [sym__list_marker_dot] = ACTIONS(1616),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1616),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1616),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1616),
    [sym__blank_line_start] = ACTIONS(1616),
    [sym__html_block_1_start] = ACTIONS(1616),
    [sym__html_block_2_start] = ACTIONS(1616),
    [sym__html_block_3_start] = ACTIONS(1616),
    [sym__html_block_4_start] = ACTIONS(1616),
    [sym__html_block_5_start] = ACTIONS(1616),
    [sym__html_block_6_start] = ACTIONS(1616),
    [sym__html_block_7_start] = ACTIONS(1616),
    [sym__pipe_table_start] = ACTIONS(1616),
  },
  [STATE(209)] = {
    [ts_builtin_sym_end] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1620),
    [anon_sym_RBRACK] = ACTIONS(1618),
    [anon_sym_LT] = ACTIONS(1618),
    [anon_sym_GT] = ACTIONS(1618),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_DOLLAR] = ACTIONS(1618),
    [anon_sym_PERCENT] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_COMMA] = ACTIONS(1618),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1618),
    [anon_sym_SLASH] = ACTIONS(1618),
    [anon_sym_COLON] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1618),
    [anon_sym_EQ] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_BSLASH] = ACTIONS(1618),
    [anon_sym_CARET] = ACTIONS(1618),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_BQUOTE] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_RBRACE] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1618),
    [anon_sym_RPAREN] = ACTIONS(1618),
    [aux_sym__word_token1] = ACTIONS(1618),
    [aux_sym__word_token2] = ACTIONS(1618),
    [aux_sym__word_token3] = ACTIONS(1618),
    [sym__whitespace] = ACTIONS(1618),
    [sym__soft_line_ending] = ACTIONS(1618),
    [sym__block_quote_start] = ACTIONS(1618),
    [sym__indented_chunk_start] = ACTIONS(1618),
    [sym_atx_h1_marker] = ACTIONS(1618),
    [sym_atx_h2_marker] = ACTIONS(1618),
    [sym_atx_h3_marker] = ACTIONS(1618),
    [sym_atx_h4_marker] = ACTIONS(1618),
    [sym_atx_h5_marker] = ACTIONS(1618),
    [sym_atx_h6_marker] = ACTIONS(1618),
    [sym__thematic_break] = ACTIONS(1618),
    [sym__list_marker_minus] = ACTIONS(1618),
    [sym__list_marker_plus] = ACTIONS(1618),
    [sym__list_marker_star] = ACTIONS(1618),
    [sym__list_marker_parenthesis] = ACTIONS(1618),
    [sym__list_marker_dot] = ACTIONS(1618),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1618),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1618),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1618),
    [sym__blank_line_start] = ACTIONS(1618),
    [sym__html_block_1_start] = ACTIONS(1618),
    [sym__html_block_2_start] = ACTIONS(1618),
    [sym__html_block_3_start] = ACTIONS(1618),
    [sym__html_block_4_start] = ACTIONS(1618),
    [sym__html_block_5_start] = ACTIONS(1618),
    [sym__html_block_6_start] = ACTIONS(1618),
    [sym__html_block_7_start] = ACTIONS(1618),
    [sym__pipe_table_start] = ACTIONS(1618),
  },
  [STATE(210)] = {
    [ts_builtin_sym_end] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1624),
    [anon_sym_RBRACK] = ACTIONS(1622),
    [anon_sym_LT] = ACTIONS(1622),
    [anon_sym_GT] = ACTIONS(1622),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_DOLLAR] = ACTIONS(1622),
    [anon_sym_PERCENT] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_STAR] = ACTIONS(1622),
    [anon_sym_PLUS] = ACTIONS(1622),
    [anon_sym_COMMA] = ACTIONS(1622),
    [anon_sym_DASH] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1622),
    [anon_sym_SLASH] = ACTIONS(1622),
    [anon_sym_COLON] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym_EQ] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_BSLASH] = ACTIONS(1622),
    [anon_sym_CARET] = ACTIONS(1622),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_BQUOTE] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_RBRACE] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1622),
    [anon_sym_RPAREN] = ACTIONS(1622),
    [aux_sym__word_token1] = ACTIONS(1622),
    [aux_sym__word_token2] = ACTIONS(1622),
    [aux_sym__word_token3] = ACTIONS(1622),
    [sym__whitespace] = ACTIONS(1622),
    [sym__soft_line_ending] = ACTIONS(1622),
    [sym__block_quote_start] = ACTIONS(1622),
    [sym__indented_chunk_start] = ACTIONS(1622),
    [sym_atx_h1_marker] = ACTIONS(1622),
    [sym_atx_h2_marker] = ACTIONS(1622),
    [sym_atx_h3_marker] = ACTIONS(1622),
    [sym_atx_h4_marker] = ACTIONS(1622),
    [sym_atx_h5_marker] = ACTIONS(1622),
    [sym_atx_h6_marker] = ACTIONS(1622),
    [sym__thematic_break] = ACTIONS(1622),
    [sym__list_marker_minus] = ACTIONS(1622),
    [sym__list_marker_plus] = ACTIONS(1622),
    [sym__list_marker_star] = ACTIONS(1622),
    [sym__list_marker_parenthesis] = ACTIONS(1622),
    [sym__list_marker_dot] = ACTIONS(1622),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1622),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1622),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1622),
    [sym__blank_line_start] = ACTIONS(1622),
    [sym__html_block_1_start] = ACTIONS(1622),
    [sym__html_block_2_start] = ACTIONS(1622),
    [sym__html_block_3_start] = ACTIONS(1622),
    [sym__html_block_4_start] = ACTIONS(1622),
    [sym__html_block_5_start] = ACTIONS(1622),
    [sym__html_block_6_start] = ACTIONS(1622),
    [sym__html_block_7_start] = ACTIONS(1622),
    [sym__pipe_table_start] = ACTIONS(1622),
  },
  [STATE(211)] = {
    [ts_builtin_sym_end] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1628),
    [anon_sym_RBRACK] = ACTIONS(1626),
    [anon_sym_LT] = ACTIONS(1626),
    [anon_sym_GT] = ACTIONS(1626),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_DOLLAR] = ACTIONS(1626),
    [anon_sym_PERCENT] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_STAR] = ACTIONS(1626),
    [anon_sym_PLUS] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1626),
    [anon_sym_DASH] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1626),
    [anon_sym_SLASH] = ACTIONS(1626),
    [anon_sym_COLON] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_EQ] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_BSLASH] = ACTIONS(1626),
    [anon_sym_CARET] = ACTIONS(1626),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_BQUOTE] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_RBRACE] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1626),
    [anon_sym_RPAREN] = ACTIONS(1626),
    [aux_sym__word_token1] = ACTIONS(1626),
    [aux_sym__word_token2] = ACTIONS(1626),
    [aux_sym__word_token3] = ACTIONS(1626),
    [sym__whitespace] = ACTIONS(1626),
    [sym__soft_line_ending] = ACTIONS(1626),
    [sym__block_quote_start] = ACTIONS(1626),
    [sym__indented_chunk_start] = ACTIONS(1626),
    [sym_atx_h1_marker] = ACTIONS(1626),
    [sym_atx_h2_marker] = ACTIONS(1626),
    [sym_atx_h3_marker] = ACTIONS(1626),
    [sym_atx_h4_marker] = ACTIONS(1626),
    [sym_atx_h5_marker] = ACTIONS(1626),
    [sym_atx_h6_marker] = ACTIONS(1626),
    [sym__thematic_break] = ACTIONS(1626),
    [sym__list_marker_minus] = ACTIONS(1626),
    [sym__list_marker_plus] = ACTIONS(1626),
    [sym__list_marker_star] = ACTIONS(1626),
    [sym__list_marker_parenthesis] = ACTIONS(1626),
    [sym__list_marker_dot] = ACTIONS(1626),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1626),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1626),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1626),
    [sym__blank_line_start] = ACTIONS(1626),
    [sym__html_block_1_start] = ACTIONS(1626),
    [sym__html_block_2_start] = ACTIONS(1626),
    [sym__html_block_3_start] = ACTIONS(1626),
    [sym__html_block_4_start] = ACTIONS(1626),
    [sym__html_block_5_start] = ACTIONS(1626),
    [sym__html_block_6_start] = ACTIONS(1626),
    [sym__html_block_7_start] = ACTIONS(1626),
    [sym__pipe_table_start] = ACTIONS(1626),
  },
  [STATE(212)] = {
    [ts_builtin_sym_end] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1632),
    [anon_sym_RBRACK] = ACTIONS(1630),
    [anon_sym_LT] = ACTIONS(1630),
    [anon_sym_GT] = ACTIONS(1630),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_DOLLAR] = ACTIONS(1630),
    [anon_sym_PERCENT] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_STAR] = ACTIONS(1630),
    [anon_sym_PLUS] = ACTIONS(1630),
    [anon_sym_COMMA] = ACTIONS(1630),
    [anon_sym_DASH] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1630),
    [anon_sym_SLASH] = ACTIONS(1630),
    [anon_sym_COLON] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1630),
    [anon_sym_EQ] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_BSLASH] = ACTIONS(1630),
    [anon_sym_CARET] = ACTIONS(1630),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_BQUOTE] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_RBRACE] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1630),
    [anon_sym_RPAREN] = ACTIONS(1630),
    [aux_sym__word_token1] = ACTIONS(1630),
    [aux_sym__word_token2] = ACTIONS(1630),
    [aux_sym__word_token3] = ACTIONS(1630),
    [sym__whitespace] = ACTIONS(1630),
    [sym__soft_line_ending] = ACTIONS(1630),
    [sym__block_quote_start] = ACTIONS(1630),
    [sym__indented_chunk_start] = ACTIONS(1630),
    [sym_atx_h1_marker] = ACTIONS(1630),
    [sym_atx_h2_marker] = ACTIONS(1630),
    [sym_atx_h3_marker] = ACTIONS(1630),
    [sym_atx_h4_marker] = ACTIONS(1630),
    [sym_atx_h5_marker] = ACTIONS(1630),
    [sym_atx_h6_marker] = ACTIONS(1630),
    [sym__thematic_break] = ACTIONS(1630),
    [sym__list_marker_minus] = ACTIONS(1630),
    [sym__list_marker_plus] = ACTIONS(1630),
    [sym__list_marker_star] = ACTIONS(1630),
    [sym__list_marker_parenthesis] = ACTIONS(1630),
    [sym__list_marker_dot] = ACTIONS(1630),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1630),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1630),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1630),
    [sym__blank_line_start] = ACTIONS(1630),
    [sym__html_block_1_start] = ACTIONS(1630),
    [sym__html_block_2_start] = ACTIONS(1630),
    [sym__html_block_3_start] = ACTIONS(1630),
    [sym__html_block_4_start] = ACTIONS(1630),
    [sym__html_block_5_start] = ACTIONS(1630),
    [sym__html_block_6_start] = ACTIONS(1630),
    [sym__html_block_7_start] = ACTIONS(1630),
    [sym__pipe_table_start] = ACTIONS(1630),
  },
  [STATE(213)] = {
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_LT] = ACTIONS(1636),
    [anon_sym_GT] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [anon_sym_POUND] = ACTIONS(1636),
    [anon_sym_DOLLAR] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1636),
    [anon_sym_AMP] = ACTIONS(1636),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_STAR] = ACTIONS(1636),
    [anon_sym_PLUS] = ACTIONS(1636),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_DASH] = ACTIONS(1636),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_SLASH] = ACTIONS(1636),
    [anon_sym_COLON] = ACTIONS(1636),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_EQ] = ACTIONS(1636),
    [anon_sym_QMARK] = ACTIONS(1636),
    [anon_sym_AT] = ACTIONS(1636),
    [anon_sym_BSLASH] = ACTIONS(1636),
    [anon_sym_CARET] = ACTIONS(1636),
    [anon_sym__] = ACTIONS(1636),
    [anon_sym_BQUOTE] = ACTIONS(1636),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_RPAREN] = ACTIONS(1636),
    [aux_sym__word_token1] = ACTIONS(1636),
    [aux_sym__word_token2] = ACTIONS(1636),
    [aux_sym__word_token3] = ACTIONS(1636),
    [sym__whitespace] = ACTIONS(1636),
    [sym__soft_line_ending] = ACTIONS(1636),
    [sym__block_close] = ACTIONS(1636),
    [sym__block_quote_start] = ACTIONS(1636),
    [sym__indented_chunk_start] = ACTIONS(1636),
    [sym_atx_h1_marker] = ACTIONS(1636),
    [sym_atx_h2_marker] = ACTIONS(1636),
    [sym_atx_h3_marker] = ACTIONS(1636),
    [sym_atx_h4_marker] = ACTIONS(1636),
    [sym_atx_h5_marker] = ACTIONS(1636),
    [sym_atx_h6_marker] = ACTIONS(1636),
    [sym__thematic_break] = ACTIONS(1636),
    [sym__list_marker_minus] = ACTIONS(1636),
    [sym__list_marker_plus] = ACTIONS(1636),
    [sym__list_marker_star] = ACTIONS(1636),
    [sym__list_marker_parenthesis] = ACTIONS(1636),
    [sym__list_marker_dot] = ACTIONS(1636),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1636),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1636),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1636),
    [sym__blank_line_start] = ACTIONS(1636),
    [sym__html_block_1_start] = ACTIONS(1636),
    [sym__html_block_2_start] = ACTIONS(1636),
    [sym__html_block_3_start] = ACTIONS(1636),
    [sym__html_block_4_start] = ACTIONS(1636),
    [sym__html_block_5_start] = ACTIONS(1636),
    [sym__html_block_6_start] = ACTIONS(1636),
    [sym__html_block_7_start] = ACTIONS(1636),
    [sym__pipe_table_start] = ACTIONS(1636),
  },
  [STATE(214)] = {
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_LT] = ACTIONS(1640),
    [anon_sym_GT] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [anon_sym_POUND] = ACTIONS(1640),
    [anon_sym_DOLLAR] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1640),
    [anon_sym_AMP] = ACTIONS(1640),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_STAR] = ACTIONS(1640),
    [anon_sym_PLUS] = ACTIONS(1640),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_DASH] = ACTIONS(1640),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_SLASH] = ACTIONS(1640),
    [anon_sym_COLON] = ACTIONS(1640),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_EQ] = ACTIONS(1640),
    [anon_sym_QMARK] = ACTIONS(1640),
    [anon_sym_AT] = ACTIONS(1640),
    [anon_sym_BSLASH] = ACTIONS(1640),
    [anon_sym_CARET] = ACTIONS(1640),
    [anon_sym__] = ACTIONS(1640),
    [anon_sym_BQUOTE] = ACTIONS(1640),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_RPAREN] = ACTIONS(1640),
    [aux_sym__word_token1] = ACTIONS(1640),
    [aux_sym__word_token2] = ACTIONS(1640),
    [aux_sym__word_token3] = ACTIONS(1640),
    [sym__whitespace] = ACTIONS(1640),
    [sym__soft_line_ending] = ACTIONS(1640),
    [sym__block_close] = ACTIONS(1640),
    [sym__block_quote_start] = ACTIONS(1640),
    [sym__indented_chunk_start] = ACTIONS(1640),
    [sym_atx_h1_marker] = ACTIONS(1640),
    [sym_atx_h2_marker] = ACTIONS(1640),
    [sym_atx_h3_marker] = ACTIONS(1640),
    [sym_atx_h4_marker] = ACTIONS(1640),
    [sym_atx_h5_marker] = ACTIONS(1640),
    [sym_atx_h6_marker] = ACTIONS(1640),
    [sym__thematic_break] = ACTIONS(1640),
    [sym__list_marker_minus] = ACTIONS(1640),
    [sym__list_marker_plus] = ACTIONS(1640),
    [sym__list_marker_star] = ACTIONS(1640),
    [sym__list_marker_parenthesis] = ACTIONS(1640),
    [sym__list_marker_dot] = ACTIONS(1640),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1640),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1640),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1640),
    [sym__blank_line_start] = ACTIONS(1640),
    [sym__html_block_1_start] = ACTIONS(1640),
    [sym__html_block_2_start] = ACTIONS(1640),
    [sym__html_block_3_start] = ACTIONS(1640),
    [sym__html_block_4_start] = ACTIONS(1640),
    [sym__html_block_5_start] = ACTIONS(1640),
    [sym__html_block_6_start] = ACTIONS(1640),
    [sym__html_block_7_start] = ACTIONS(1640),
    [sym__pipe_table_start] = ACTIONS(1640),
  },
  [STATE(215)] = {
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_LT] = ACTIONS(1644),
    [anon_sym_GT] = ACTIONS(1644),
    [anon_sym_BANG] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [anon_sym_POUND] = ACTIONS(1644),
    [anon_sym_DOLLAR] = ACTIONS(1644),
    [anon_sym_PERCENT] = ACTIONS(1644),
    [anon_sym_AMP] = ACTIONS(1644),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_STAR] = ACTIONS(1644),
    [anon_sym_PLUS] = ACTIONS(1644),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_DASH] = ACTIONS(1644),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_SLASH] = ACTIONS(1644),
    [anon_sym_COLON] = ACTIONS(1644),
    [anon_sym_SEMI] = ACTIONS(1644),
    [anon_sym_EQ] = ACTIONS(1644),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_AT] = ACTIONS(1644),
    [anon_sym_BSLASH] = ACTIONS(1644),
    [anon_sym_CARET] = ACTIONS(1644),
    [anon_sym__] = ACTIONS(1644),
    [anon_sym_BQUOTE] = ACTIONS(1644),
    [anon_sym_LBRACE] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1644),
    [anon_sym_RBRACE] = ACTIONS(1644),
    [anon_sym_TILDE] = ACTIONS(1644),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_RPAREN] = ACTIONS(1644),
    [aux_sym__word_token1] = ACTIONS(1644),
    [aux_sym__word_token2] = ACTIONS(1644),
    [aux_sym__word_token3] = ACTIONS(1644),
    [sym__whitespace] = ACTIONS(1644),
    [sym__soft_line_ending] = ACTIONS(1644),
    [sym__block_close] = ACTIONS(1644),
    [sym__block_quote_start] = ACTIONS(1644),
    [sym__indented_chunk_start] = ACTIONS(1644),
    [sym_atx_h1_marker] = ACTIONS(1644),
    [sym_atx_h2_marker] = ACTIONS(1644),
    [sym_atx_h3_marker] = ACTIONS(1644),
    [sym_atx_h4_marker] = ACTIONS(1644),
    [sym_atx_h5_marker] = ACTIONS(1644),
    [sym_atx_h6_marker] = ACTIONS(1644),
    [sym__thematic_break] = ACTIONS(1644),
    [sym__list_marker_minus] = ACTIONS(1644),
    [sym__list_marker_plus] = ACTIONS(1644),
    [sym__list_marker_star] = ACTIONS(1644),
    [sym__list_marker_parenthesis] = ACTIONS(1644),
    [sym__list_marker_dot] = ACTIONS(1644),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1644),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1644),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1644),
    [sym__blank_line_start] = ACTIONS(1644),
    [sym__html_block_1_start] = ACTIONS(1644),
    [sym__html_block_2_start] = ACTIONS(1644),
    [sym__html_block_3_start] = ACTIONS(1644),
    [sym__html_block_4_start] = ACTIONS(1644),
    [sym__html_block_5_start] = ACTIONS(1644),
    [sym__html_block_6_start] = ACTIONS(1644),
    [sym__html_block_7_start] = ACTIONS(1644),
    [sym__pipe_table_start] = ACTIONS(1644),
  },
  [STATE(216)] = {
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_LT] = ACTIONS(1648),
    [anon_sym_GT] = ACTIONS(1648),
    [anon_sym_BANG] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [anon_sym_POUND] = ACTIONS(1648),
    [anon_sym_DOLLAR] = ACTIONS(1648),
    [anon_sym_PERCENT] = ACTIONS(1648),
    [anon_sym_AMP] = ACTIONS(1648),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_STAR] = ACTIONS(1648),
    [anon_sym_PLUS] = ACTIONS(1648),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_DASH] = ACTIONS(1648),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_SLASH] = ACTIONS(1648),
    [anon_sym_COLON] = ACTIONS(1648),
    [anon_sym_SEMI] = ACTIONS(1648),
    [anon_sym_EQ] = ACTIONS(1648),
    [anon_sym_QMARK] = ACTIONS(1648),
    [anon_sym_AT] = ACTIONS(1648),
    [anon_sym_BSLASH] = ACTIONS(1648),
    [anon_sym_CARET] = ACTIONS(1648),
    [anon_sym__] = ACTIONS(1648),
    [anon_sym_BQUOTE] = ACTIONS(1648),
    [anon_sym_LBRACE] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1648),
    [anon_sym_RBRACE] = ACTIONS(1648),
    [anon_sym_TILDE] = ACTIONS(1648),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_RPAREN] = ACTIONS(1648),
    [aux_sym__word_token1] = ACTIONS(1648),
    [aux_sym__word_token2] = ACTIONS(1648),
    [aux_sym__word_token3] = ACTIONS(1648),
    [sym__whitespace] = ACTIONS(1648),
    [sym__soft_line_ending] = ACTIONS(1648),
    [sym__block_close] = ACTIONS(1648),
    [sym__block_quote_start] = ACTIONS(1648),
    [sym__indented_chunk_start] = ACTIONS(1648),
    [sym_atx_h1_marker] = ACTIONS(1648),
    [sym_atx_h2_marker] = ACTIONS(1648),
    [sym_atx_h3_marker] = ACTIONS(1648),
    [sym_atx_h4_marker] = ACTIONS(1648),
    [sym_atx_h5_marker] = ACTIONS(1648),
    [sym_atx_h6_marker] = ACTIONS(1648),
    [sym__thematic_break] = ACTIONS(1648),
    [sym__list_marker_minus] = ACTIONS(1648),
    [sym__list_marker_plus] = ACTIONS(1648),
    [sym__list_marker_star] = ACTIONS(1648),
    [sym__list_marker_parenthesis] = ACTIONS(1648),
    [sym__list_marker_dot] = ACTIONS(1648),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1648),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1648),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1648),
    [sym__blank_line_start] = ACTIONS(1648),
    [sym__html_block_1_start] = ACTIONS(1648),
    [sym__html_block_2_start] = ACTIONS(1648),
    [sym__html_block_3_start] = ACTIONS(1648),
    [sym__html_block_4_start] = ACTIONS(1648),
    [sym__html_block_5_start] = ACTIONS(1648),
    [sym__html_block_6_start] = ACTIONS(1648),
    [sym__html_block_7_start] = ACTIONS(1648),
    [sym__pipe_table_start] = ACTIONS(1648),
  },
  [STATE(217)] = {
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_LT] = ACTIONS(1652),
    [anon_sym_GT] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [anon_sym_POUND] = ACTIONS(1652),
    [anon_sym_DOLLAR] = ACTIONS(1652),
    [anon_sym_PERCENT] = ACTIONS(1652),
    [anon_sym_AMP] = ACTIONS(1652),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_STAR] = ACTIONS(1652),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(1652),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_SLASH] = ACTIONS(1652),
    [anon_sym_COLON] = ACTIONS(1652),
    [anon_sym_SEMI] = ACTIONS(1652),
    [anon_sym_EQ] = ACTIONS(1652),
    [anon_sym_QMARK] = ACTIONS(1652),
    [anon_sym_AT] = ACTIONS(1652),
    [anon_sym_BSLASH] = ACTIONS(1652),
    [anon_sym_CARET] = ACTIONS(1652),
    [anon_sym__] = ACTIONS(1652),
    [anon_sym_BQUOTE] = ACTIONS(1652),
    [anon_sym_LBRACE] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1652),
    [anon_sym_RBRACE] = ACTIONS(1652),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_RPAREN] = ACTIONS(1652),
    [aux_sym__word_token1] = ACTIONS(1652),
    [aux_sym__word_token2] = ACTIONS(1652),
    [aux_sym__word_token3] = ACTIONS(1652),
    [sym__whitespace] = ACTIONS(1652),
    [sym__soft_line_ending] = ACTIONS(1652),
    [sym__block_close] = ACTIONS(1652),
    [sym__block_quote_start] = ACTIONS(1652),
    [sym__indented_chunk_start] = ACTIONS(1652),
    [sym_atx_h1_marker] = ACTIONS(1652),
    [sym_atx_h2_marker] = ACTIONS(1652),
    [sym_atx_h3_marker] = ACTIONS(1652),
    [sym_atx_h4_marker] = ACTIONS(1652),
    [sym_atx_h5_marker] = ACTIONS(1652),
    [sym_atx_h6_marker] = ACTIONS(1652),
    [sym__thematic_break] = ACTIONS(1652),
    [sym__list_marker_minus] = ACTIONS(1652),
    [sym__list_marker_plus] = ACTIONS(1652),
    [sym__list_marker_star] = ACTIONS(1652),
    [sym__list_marker_parenthesis] = ACTIONS(1652),
    [sym__list_marker_dot] = ACTIONS(1652),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1652),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1652),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1652),
    [sym__blank_line_start] = ACTIONS(1652),
    [sym__html_block_1_start] = ACTIONS(1652),
    [sym__html_block_2_start] = ACTIONS(1652),
    [sym__html_block_3_start] = ACTIONS(1652),
    [sym__html_block_4_start] = ACTIONS(1652),
    [sym__html_block_5_start] = ACTIONS(1652),
    [sym__html_block_6_start] = ACTIONS(1652),
    [sym__html_block_7_start] = ACTIONS(1652),
    [sym__pipe_table_start] = ACTIONS(1652),
  },
  [STATE(218)] = {
    [anon_sym_LBRACK] = ACTIONS(1654),
    [anon_sym_RBRACK] = ACTIONS(1656),
    [anon_sym_LT] = ACTIONS(1656),
    [anon_sym_GT] = ACTIONS(1656),
    [anon_sym_BANG] = ACTIONS(1656),
    [anon_sym_DQUOTE] = ACTIONS(1656),
    [anon_sym_POUND] = ACTIONS(1656),
    [anon_sym_DOLLAR] = ACTIONS(1656),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_AMP] = ACTIONS(1656),
    [anon_sym_SQUOTE] = ACTIONS(1656),
    [anon_sym_STAR] = ACTIONS(1656),
    [anon_sym_PLUS] = ACTIONS(1656),
    [anon_sym_COMMA] = ACTIONS(1656),
    [anon_sym_DASH] = ACTIONS(1656),
    [anon_sym_DOT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1656),
    [anon_sym_COLON] = ACTIONS(1656),
    [anon_sym_SEMI] = ACTIONS(1656),
    [anon_sym_EQ] = ACTIONS(1656),
    [anon_sym_QMARK] = ACTIONS(1656),
    [anon_sym_AT] = ACTIONS(1656),
    [anon_sym_BSLASH] = ACTIONS(1656),
    [anon_sym_CARET] = ACTIONS(1656),
    [anon_sym__] = ACTIONS(1656),
    [anon_sym_BQUOTE] = ACTIONS(1656),
    [anon_sym_LBRACE] = ACTIONS(1656),
    [anon_sym_PIPE] = ACTIONS(1656),
    [anon_sym_RBRACE] = ACTIONS(1656),
    [anon_sym_TILDE] = ACTIONS(1656),
    [anon_sym_LPAREN] = ACTIONS(1656),
    [anon_sym_RPAREN] = ACTIONS(1656),
    [aux_sym__word_token1] = ACTIONS(1656),
    [aux_sym__word_token2] = ACTIONS(1656),
    [aux_sym__word_token3] = ACTIONS(1656),
    [sym__whitespace] = ACTIONS(1656),
    [sym__soft_line_ending] = ACTIONS(1656),
    [sym__block_close] = ACTIONS(1656),
    [sym__block_quote_start] = ACTIONS(1656),
    [sym__indented_chunk_start] = ACTIONS(1656),
    [sym_atx_h1_marker] = ACTIONS(1656),
    [sym_atx_h2_marker] = ACTIONS(1656),
    [sym_atx_h3_marker] = ACTIONS(1656),
    [sym_atx_h4_marker] = ACTIONS(1656),
    [sym_atx_h5_marker] = ACTIONS(1656),
    [sym_atx_h6_marker] = ACTIONS(1656),
    [sym__thematic_break] = ACTIONS(1656),
    [sym__list_marker_minus] = ACTIONS(1656),
    [sym__list_marker_plus] = ACTIONS(1656),
    [sym__list_marker_star] = ACTIONS(1656),
    [sym__list_marker_parenthesis] = ACTIONS(1656),
    [sym__list_marker_dot] = ACTIONS(1656),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1656),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1656),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1656),
    [sym__blank_line_start] = ACTIONS(1656),
    [sym__html_block_1_start] = ACTIONS(1656),
    [sym__html_block_2_start] = ACTIONS(1656),
    [sym__html_block_3_start] = ACTIONS(1656),
    [sym__html_block_4_start] = ACTIONS(1656),
    [sym__html_block_5_start] = ACTIONS(1656),
    [sym__html_block_6_start] = ACTIONS(1656),
    [sym__html_block_7_start] = ACTIONS(1656),
    [sym__pipe_table_start] = ACTIONS(1656),
  },
  [STATE(219)] = {
    [anon_sym_LBRACK] = ACTIONS(1658),
    [anon_sym_RBRACK] = ACTIONS(1660),
    [anon_sym_LT] = ACTIONS(1660),
    [anon_sym_GT] = ACTIONS(1660),
    [anon_sym_BANG] = ACTIONS(1660),
    [anon_sym_DQUOTE] = ACTIONS(1660),
    [anon_sym_POUND] = ACTIONS(1660),
    [anon_sym_DOLLAR] = ACTIONS(1660),
    [anon_sym_PERCENT] = ACTIONS(1660),
    [anon_sym_AMP] = ACTIONS(1660),
    [anon_sym_SQUOTE] = ACTIONS(1660),
    [anon_sym_STAR] = ACTIONS(1660),
    [anon_sym_PLUS] = ACTIONS(1660),
    [anon_sym_COMMA] = ACTIONS(1660),
    [anon_sym_DASH] = ACTIONS(1660),
    [anon_sym_DOT] = ACTIONS(1660),
    [anon_sym_SLASH] = ACTIONS(1660),
    [anon_sym_COLON] = ACTIONS(1660),
    [anon_sym_SEMI] = ACTIONS(1660),
    [anon_sym_EQ] = ACTIONS(1660),
    [anon_sym_QMARK] = ACTIONS(1660),
    [anon_sym_AT] = ACTIONS(1660),
    [anon_sym_BSLASH] = ACTIONS(1660),
    [anon_sym_CARET] = ACTIONS(1660),
    [anon_sym__] = ACTIONS(1660),
    [anon_sym_BQUOTE] = ACTIONS(1660),
    [anon_sym_LBRACE] = ACTIONS(1660),
    [anon_sym_PIPE] = ACTIONS(1660),
    [anon_sym_RBRACE] = ACTIONS(1660),
    [anon_sym_TILDE] = ACTIONS(1660),
    [anon_sym_LPAREN] = ACTIONS(1660),
    [anon_sym_RPAREN] = ACTIONS(1660),
    [aux_sym__word_token1] = ACTIONS(1660),
    [aux_sym__word_token2] = ACTIONS(1660),
    [aux_sym__word_token3] = ACTIONS(1660),
    [sym__whitespace] = ACTIONS(1660),
    [sym__soft_line_ending] = ACTIONS(1660),
    [sym__block_close] = ACTIONS(1660),
    [sym__block_quote_start] = ACTIONS(1660),
    [sym__indented_chunk_start] = ACTIONS(1660),
    [sym_atx_h1_marker] = ACTIONS(1660),
    [sym_atx_h2_marker] = ACTIONS(1660),
    [sym_atx_h3_marker] = ACTIONS(1660),
    [sym_atx_h4_marker] = ACTIONS(1660),
    [sym_atx_h5_marker] = ACTIONS(1660),
    [sym_atx_h6_marker] = ACTIONS(1660),
    [sym__thematic_break] = ACTIONS(1660),
    [sym__list_marker_minus] = ACTIONS(1660),
    [sym__list_marker_plus] = ACTIONS(1660),
    [sym__list_marker_star] = ACTIONS(1660),
    [sym__list_marker_parenthesis] = ACTIONS(1660),
    [sym__list_marker_dot] = ACTIONS(1660),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1660),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1660),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1660),
    [sym__blank_line_start] = ACTIONS(1660),
    [sym__html_block_1_start] = ACTIONS(1660),
    [sym__html_block_2_start] = ACTIONS(1660),
    [sym__html_block_3_start] = ACTIONS(1660),
    [sym__html_block_4_start] = ACTIONS(1660),
    [sym__html_block_5_start] = ACTIONS(1660),
    [sym__html_block_6_start] = ACTIONS(1660),
    [sym__html_block_7_start] = ACTIONS(1660),
    [sym__pipe_table_start] = ACTIONS(1660),
  },
  [STATE(220)] = {
    [anon_sym_LBRACK] = ACTIONS(1662),
    [anon_sym_RBRACK] = ACTIONS(1664),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_BANG] = ACTIONS(1664),
    [anon_sym_DQUOTE] = ACTIONS(1664),
    [anon_sym_POUND] = ACTIONS(1664),
    [anon_sym_DOLLAR] = ACTIONS(1664),
    [anon_sym_PERCENT] = ACTIONS(1664),
    [anon_sym_AMP] = ACTIONS(1664),
    [anon_sym_SQUOTE] = ACTIONS(1664),
    [anon_sym_STAR] = ACTIONS(1664),
    [anon_sym_PLUS] = ACTIONS(1664),
    [anon_sym_COMMA] = ACTIONS(1664),
    [anon_sym_DASH] = ACTIONS(1664),
    [anon_sym_DOT] = ACTIONS(1664),
    [anon_sym_SLASH] = ACTIONS(1664),
    [anon_sym_COLON] = ACTIONS(1664),
    [anon_sym_SEMI] = ACTIONS(1664),
    [anon_sym_EQ] = ACTIONS(1664),
    [anon_sym_QMARK] = ACTIONS(1664),
    [anon_sym_AT] = ACTIONS(1664),
    [anon_sym_BSLASH] = ACTIONS(1664),
    [anon_sym_CARET] = ACTIONS(1664),
    [anon_sym__] = ACTIONS(1664),
    [anon_sym_BQUOTE] = ACTIONS(1664),
    [anon_sym_LBRACE] = ACTIONS(1664),
    [anon_sym_PIPE] = ACTIONS(1664),
    [anon_sym_RBRACE] = ACTIONS(1664),
    [anon_sym_TILDE] = ACTIONS(1664),
    [anon_sym_LPAREN] = ACTIONS(1664),
    [anon_sym_RPAREN] = ACTIONS(1664),
    [aux_sym__word_token1] = ACTIONS(1664),
    [aux_sym__word_token2] = ACTIONS(1664),
    [aux_sym__word_token3] = ACTIONS(1664),
    [sym__whitespace] = ACTIONS(1664),
    [sym__soft_line_ending] = ACTIONS(1664),
    [sym__block_close] = ACTIONS(1664),
    [sym__block_quote_start] = ACTIONS(1664),
    [sym__indented_chunk_start] = ACTIONS(1664),
    [sym_atx_h1_marker] = ACTIONS(1664),
    [sym_atx_h2_marker] = ACTIONS(1664),
    [sym_atx_h3_marker] = ACTIONS(1664),
    [sym_atx_h4_marker] = ACTIONS(1664),
    [sym_atx_h5_marker] = ACTIONS(1664),
    [sym_atx_h6_marker] = ACTIONS(1664),
    [sym__thematic_break] = ACTIONS(1664),
    [sym__list_marker_minus] = ACTIONS(1664),
    [sym__list_marker_plus] = ACTIONS(1664),
    [sym__list_marker_star] = ACTIONS(1664),
    [sym__list_marker_parenthesis] = ACTIONS(1664),
    [sym__list_marker_dot] = ACTIONS(1664),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1664),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1664),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1664),
    [sym__blank_line_start] = ACTIONS(1664),
    [sym__html_block_1_start] = ACTIONS(1664),
    [sym__html_block_2_start] = ACTIONS(1664),
    [sym__html_block_3_start] = ACTIONS(1664),
    [sym__html_block_4_start] = ACTIONS(1664),
    [sym__html_block_5_start] = ACTIONS(1664),
    [sym__html_block_6_start] = ACTIONS(1664),
    [sym__html_block_7_start] = ACTIONS(1664),
    [sym__pipe_table_start] = ACTIONS(1664),
  },
  [STATE(221)] = {
    [ts_builtin_sym_end] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1342),
    [anon_sym_RBRACK] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [anon_sym_POUND] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_SQUOTE] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_COLON] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_BSLASH] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym__] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_RPAREN] = ACTIONS(1344),
    [aux_sym__word_token1] = ACTIONS(1344),
    [aux_sym__word_token2] = ACTIONS(1344),
    [aux_sym__word_token3] = ACTIONS(1344),
    [sym__whitespace] = ACTIONS(1344),
    [sym__soft_line_ending] = ACTIONS(1344),
    [sym__block_quote_start] = ACTIONS(1344),
    [sym__indented_chunk_start] = ACTIONS(1344),
    [sym_atx_h1_marker] = ACTIONS(1344),
    [sym_atx_h2_marker] = ACTIONS(1344),
    [sym_atx_h3_marker] = ACTIONS(1344),
    [sym_atx_h4_marker] = ACTIONS(1344),
    [sym_atx_h5_marker] = ACTIONS(1344),
    [sym_atx_h6_marker] = ACTIONS(1344),
    [sym__thematic_break] = ACTIONS(1344),
    [sym__list_marker_minus] = ACTIONS(1344),
    [sym__list_marker_plus] = ACTIONS(1344),
    [sym__list_marker_star] = ACTIONS(1344),
    [sym__list_marker_parenthesis] = ACTIONS(1344),
    [sym__list_marker_dot] = ACTIONS(1344),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1344),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1344),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1344),
    [sym__blank_line_start] = ACTIONS(1344),
    [sym__html_block_1_start] = ACTIONS(1344),
    [sym__html_block_2_start] = ACTIONS(1344),
    [sym__html_block_3_start] = ACTIONS(1344),
    [sym__html_block_4_start] = ACTIONS(1344),
    [sym__html_block_5_start] = ACTIONS(1344),
    [sym__html_block_6_start] = ACTIONS(1344),
    [sym__html_block_7_start] = ACTIONS(1344),
    [sym__pipe_table_start] = ACTIONS(1344),
  },
  [STATE(222)] = {
    [ts_builtin_sym_end] = ACTIONS(1350),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_RBRACK] = ACTIONS(1350),
    [anon_sym_LT] = ACTIONS(1350),
    [anon_sym_GT] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [anon_sym_POUND] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_PLUS] = ACTIONS(1350),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_EQ] = ACTIONS(1350),
    [anon_sym_QMARK] = ACTIONS(1350),
    [anon_sym_AT] = ACTIONS(1350),
    [anon_sym_BSLASH] = ACTIONS(1350),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym__] = ACTIONS(1350),
    [anon_sym_BQUOTE] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_RPAREN] = ACTIONS(1350),
    [aux_sym__word_token1] = ACTIONS(1350),
    [aux_sym__word_token2] = ACTIONS(1350),
    [aux_sym__word_token3] = ACTIONS(1350),
    [sym__whitespace] = ACTIONS(1350),
    [sym__soft_line_ending] = ACTIONS(1350),
    [sym__block_quote_start] = ACTIONS(1350),
    [sym__indented_chunk_start] = ACTIONS(1350),
    [sym_atx_h1_marker] = ACTIONS(1350),
    [sym_atx_h2_marker] = ACTIONS(1350),
    [sym_atx_h3_marker] = ACTIONS(1350),
    [sym_atx_h4_marker] = ACTIONS(1350),
    [sym_atx_h5_marker] = ACTIONS(1350),
    [sym_atx_h6_marker] = ACTIONS(1350),
    [sym__thematic_break] = ACTIONS(1350),
    [sym__list_marker_minus] = ACTIONS(1350),
    [sym__list_marker_plus] = ACTIONS(1350),
    [sym__list_marker_star] = ACTIONS(1350),
    [sym__list_marker_parenthesis] = ACTIONS(1350),
    [sym__list_marker_dot] = ACTIONS(1350),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1350),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1350),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1350),
    [sym__blank_line_start] = ACTIONS(1350),
    [sym__html_block_1_start] = ACTIONS(1350),
    [sym__html_block_2_start] = ACTIONS(1350),
    [sym__html_block_3_start] = ACTIONS(1350),
    [sym__html_block_4_start] = ACTIONS(1350),
    [sym__html_block_5_start] = ACTIONS(1350),
    [sym__html_block_6_start] = ACTIONS(1350),
    [sym__html_block_7_start] = ACTIONS(1350),
    [sym__pipe_table_start] = ACTIONS(1350),
  },
  [STATE(223)] = {
    [ts_builtin_sym_end] = ACTIONS(1356),
    [anon_sym_LBRACK] = ACTIONS(1354),
    [anon_sym_RBRACK] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [anon_sym_POUND] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_SQUOTE] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_COMMA] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_DOT] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_COLON] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_AT] = ACTIONS(1356),
    [anon_sym_BSLASH] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym__] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_TILDE] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_RPAREN] = ACTIONS(1356),
    [aux_sym__word_token1] = ACTIONS(1356),
    [aux_sym__word_token2] = ACTIONS(1356),
    [aux_sym__word_token3] = ACTIONS(1356),
    [sym__whitespace] = ACTIONS(1356),
    [sym__soft_line_ending] = ACTIONS(1356),
    [sym__block_quote_start] = ACTIONS(1356),
    [sym__indented_chunk_start] = ACTIONS(1356),
    [sym_atx_h1_marker] = ACTIONS(1356),
    [sym_atx_h2_marker] = ACTIONS(1356),
    [sym_atx_h3_marker] = ACTIONS(1356),
    [sym_atx_h4_marker] = ACTIONS(1356),
    [sym_atx_h5_marker] = ACTIONS(1356),
    [sym_atx_h6_marker] = ACTIONS(1356),
    [sym__thematic_break] = ACTIONS(1356),
    [sym__list_marker_minus] = ACTIONS(1356),
    [sym__list_marker_plus] = ACTIONS(1356),
    [sym__list_marker_star] = ACTIONS(1356),
    [sym__list_marker_parenthesis] = ACTIONS(1356),
    [sym__list_marker_dot] = ACTIONS(1356),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1356),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1356),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1356),
    [sym__blank_line_start] = ACTIONS(1356),
    [sym__html_block_1_start] = ACTIONS(1356),
    [sym__html_block_2_start] = ACTIONS(1356),
    [sym__html_block_3_start] = ACTIONS(1356),
    [sym__html_block_4_start] = ACTIONS(1356),
    [sym__html_block_5_start] = ACTIONS(1356),
    [sym__html_block_6_start] = ACTIONS(1356),
    [sym__html_block_7_start] = ACTIONS(1356),
    [sym__pipe_table_start] = ACTIONS(1356),
  },
  [STATE(224)] = {
    [ts_builtin_sym_end] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_RBRACK] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_POUND] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_COLON] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_BSLASH] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_RPAREN] = ACTIONS(1362),
    [aux_sym__word_token1] = ACTIONS(1362),
    [aux_sym__word_token2] = ACTIONS(1362),
    [aux_sym__word_token3] = ACTIONS(1362),
    [sym__whitespace] = ACTIONS(1362),
    [sym__soft_line_ending] = ACTIONS(1362),
    [sym__block_quote_start] = ACTIONS(1362),
    [sym__indented_chunk_start] = ACTIONS(1362),
    [sym_atx_h1_marker] = ACTIONS(1362),
    [sym_atx_h2_marker] = ACTIONS(1362),
    [sym_atx_h3_marker] = ACTIONS(1362),
    [sym_atx_h4_marker] = ACTIONS(1362),
    [sym_atx_h5_marker] = ACTIONS(1362),
    [sym_atx_h6_marker] = ACTIONS(1362),
    [sym__thematic_break] = ACTIONS(1362),
    [sym__list_marker_minus] = ACTIONS(1362),
    [sym__list_marker_plus] = ACTIONS(1362),
    [sym__list_marker_star] = ACTIONS(1362),
    [sym__list_marker_parenthesis] = ACTIONS(1362),
    [sym__list_marker_dot] = ACTIONS(1362),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1362),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1362),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1362),
    [sym__blank_line_start] = ACTIONS(1362),
    [sym__html_block_1_start] = ACTIONS(1362),
    [sym__html_block_2_start] = ACTIONS(1362),
    [sym__html_block_3_start] = ACTIONS(1362),
    [sym__html_block_4_start] = ACTIONS(1362),
    [sym__html_block_5_start] = ACTIONS(1362),
    [sym__html_block_6_start] = ACTIONS(1362),
    [sym__html_block_7_start] = ACTIONS(1362),
    [sym__pipe_table_start] = ACTIONS(1362),
  },
  [STATE(225)] = {
    [ts_builtin_sym_end] = ACTIONS(1536),
    [anon_sym_LBRACK] = ACTIONS(1534),
    [anon_sym_RBRACK] = ACTIONS(1536),
    [anon_sym_LT] = ACTIONS(1536),
    [anon_sym_GT] = ACTIONS(1536),
    [anon_sym_BANG] = ACTIONS(1536),
    [anon_sym_DQUOTE] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(1536),
    [anon_sym_DOLLAR] = ACTIONS(1536),
    [anon_sym_PERCENT] = ACTIONS(1536),
    [anon_sym_AMP] = ACTIONS(1536),
    [anon_sym_SQUOTE] = ACTIONS(1536),
    [anon_sym_STAR] = ACTIONS(1536),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_DOT] = ACTIONS(1536),
    [anon_sym_SLASH] = ACTIONS(1536),
    [anon_sym_COLON] = ACTIONS(1536),
    [anon_sym_SEMI] = ACTIONS(1536),
    [anon_sym_EQ] = ACTIONS(1536),
    [anon_sym_QMARK] = ACTIONS(1536),
    [anon_sym_AT] = ACTIONS(1536),
    [anon_sym_BSLASH] = ACTIONS(1536),
    [anon_sym_CARET] = ACTIONS(1536),
    [anon_sym__] = ACTIONS(1536),
    [anon_sym_BQUOTE] = ACTIONS(1536),
    [anon_sym_LBRACE] = ACTIONS(1536),
    [anon_sym_PIPE] = ACTIONS(1536),
    [anon_sym_RBRACE] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1536),
    [anon_sym_LPAREN] = ACTIONS(1536),
    [anon_sym_RPAREN] = ACTIONS(1536),
    [aux_sym__word_token1] = ACTIONS(1536),
    [aux_sym__word_token2] = ACTIONS(1536),
    [aux_sym__word_token3] = ACTIONS(1536),
    [sym__whitespace] = ACTIONS(1536),
    [sym__soft_line_ending] = ACTIONS(1536),
    [sym__block_quote_start] = ACTIONS(1536),
    [sym__indented_chunk_start] = ACTIONS(1536),
    [sym_atx_h1_marker] = ACTIONS(1536),
    [sym_atx_h2_marker] = ACTIONS(1536),
    [sym_atx_h3_marker] = ACTIONS(1536),
    [sym_atx_h4_marker] = ACTIONS(1536),
    [sym_atx_h5_marker] = ACTIONS(1536),
    [sym_atx_h6_marker] = ACTIONS(1536),
    [sym__thematic_break] = ACTIONS(1536),
    [sym__list_marker_minus] = ACTIONS(1536),
    [sym__list_marker_plus] = ACTIONS(1536),
    [sym__list_marker_star] = ACTIONS(1536),
    [sym__list_marker_parenthesis] = ACTIONS(1536),
    [sym__list_marker_dot] = ACTIONS(1536),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1536),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1536),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1536),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1536),
    [sym__blank_line_start] = ACTIONS(1536),
    [sym__html_block_1_start] = ACTIONS(1536),
    [sym__html_block_2_start] = ACTIONS(1536),
    [sym__html_block_3_start] = ACTIONS(1536),
    [sym__html_block_4_start] = ACTIONS(1536),
    [sym__html_block_5_start] = ACTIONS(1536),
    [sym__html_block_6_start] = ACTIONS(1536),
    [sym__html_block_7_start] = ACTIONS(1536),
    [sym__pipe_table_start] = ACTIONS(1536),
  },
  [STATE(226)] = {
    [anon_sym_LBRACK] = ACTIONS(1666),
    [anon_sym_RBRACK] = ACTIONS(1668),
    [anon_sym_LT] = ACTIONS(1668),
    [anon_sym_GT] = ACTIONS(1668),
    [anon_sym_BANG] = ACTIONS(1668),
    [anon_sym_DQUOTE] = ACTIONS(1668),
    [anon_sym_POUND] = ACTIONS(1668),
    [anon_sym_DOLLAR] = ACTIONS(1668),
    [anon_sym_PERCENT] = ACTIONS(1668),
    [anon_sym_AMP] = ACTIONS(1668),
    [anon_sym_SQUOTE] = ACTIONS(1668),
    [anon_sym_STAR] = ACTIONS(1668),
    [anon_sym_PLUS] = ACTIONS(1668),
    [anon_sym_COMMA] = ACTIONS(1668),
    [anon_sym_DASH] = ACTIONS(1668),
    [anon_sym_DOT] = ACTIONS(1668),
    [anon_sym_SLASH] = ACTIONS(1668),
    [anon_sym_COLON] = ACTIONS(1668),
    [anon_sym_SEMI] = ACTIONS(1668),
    [anon_sym_EQ] = ACTIONS(1668),
    [anon_sym_QMARK] = ACTIONS(1668),
    [anon_sym_AT] = ACTIONS(1668),
    [anon_sym_BSLASH] = ACTIONS(1668),
    [anon_sym_CARET] = ACTIONS(1668),
    [anon_sym__] = ACTIONS(1668),
    [anon_sym_BQUOTE] = ACTIONS(1668),
    [anon_sym_LBRACE] = ACTIONS(1668),
    [anon_sym_PIPE] = ACTIONS(1668),
    [anon_sym_RBRACE] = ACTIONS(1668),
    [anon_sym_TILDE] = ACTIONS(1668),
    [anon_sym_LPAREN] = ACTIONS(1668),
    [anon_sym_RPAREN] = ACTIONS(1668),
    [aux_sym__word_token1] = ACTIONS(1668),
    [aux_sym__word_token2] = ACTIONS(1668),
    [aux_sym__word_token3] = ACTIONS(1668),
    [sym__whitespace] = ACTIONS(1668),
    [sym__soft_line_ending] = ACTIONS(1668),
    [sym__block_close] = ACTIONS(1668),
    [sym__block_quote_start] = ACTIONS(1668),
    [sym__indented_chunk_start] = ACTIONS(1668),
    [sym_atx_h1_marker] = ACTIONS(1668),
    [sym_atx_h2_marker] = ACTIONS(1668),
    [sym_atx_h3_marker] = ACTIONS(1668),
    [sym_atx_h4_marker] = ACTIONS(1668),
    [sym_atx_h5_marker] = ACTIONS(1668),
    [sym_atx_h6_marker] = ACTIONS(1668),
    [sym__thematic_break] = ACTIONS(1668),
    [sym__list_marker_minus] = ACTIONS(1668),
    [sym__list_marker_plus] = ACTIONS(1668),
    [sym__list_marker_star] = ACTIONS(1668),
    [sym__list_marker_parenthesis] = ACTIONS(1668),
    [sym__list_marker_dot] = ACTIONS(1668),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1668),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1668),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1668),
    [sym__blank_line_start] = ACTIONS(1668),
    [sym__html_block_1_start] = ACTIONS(1668),
    [sym__html_block_2_start] = ACTIONS(1668),
    [sym__html_block_3_start] = ACTIONS(1668),
    [sym__html_block_4_start] = ACTIONS(1668),
    [sym__html_block_5_start] = ACTIONS(1668),
    [sym__html_block_6_start] = ACTIONS(1668),
    [sym__html_block_7_start] = ACTIONS(1668),
    [sym__pipe_table_start] = ACTIONS(1668),
  },
  [STATE(227)] = {
    [ts_builtin_sym_end] = ACTIONS(1290),
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_RBRACK] = ACTIONS(1290),
    [anon_sym_LT] = ACTIONS(1290),
    [anon_sym_GT] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [anon_sym_DQUOTE] = ACTIONS(1290),
    [anon_sym_POUND] = ACTIONS(1290),
    [anon_sym_DOLLAR] = ACTIONS(1290),
    [anon_sym_PERCENT] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_SLASH] = ACTIONS(1290),
    [anon_sym_COLON] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_EQ] = ACTIONS(1290),
    [anon_sym_QMARK] = ACTIONS(1290),
    [anon_sym_AT] = ACTIONS(1290),
    [anon_sym_BSLASH] = ACTIONS(1290),
    [anon_sym_CARET] = ACTIONS(1290),
    [anon_sym__] = ACTIONS(1290),
    [anon_sym_BQUOTE] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_PIPE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_RPAREN] = ACTIONS(1290),
    [aux_sym__word_token1] = ACTIONS(1290),
    [aux_sym__word_token2] = ACTIONS(1290),
    [aux_sym__word_token3] = ACTIONS(1290),
    [sym__whitespace] = ACTIONS(1290),
    [sym__soft_line_ending] = ACTIONS(1290),
    [sym__block_quote_start] = ACTIONS(1290),
    [sym__indented_chunk_start] = ACTIONS(1290),
    [sym_atx_h1_marker] = ACTIONS(1290),
    [sym_atx_h2_marker] = ACTIONS(1290),
    [sym_atx_h3_marker] = ACTIONS(1290),
    [sym_atx_h4_marker] = ACTIONS(1290),
    [sym_atx_h5_marker] = ACTIONS(1290),
    [sym_atx_h6_marker] = ACTIONS(1290),
    [sym__thematic_break] = ACTIONS(1290),
    [sym__list_marker_minus] = ACTIONS(1290),
    [sym__list_marker_plus] = ACTIONS(1290),
    [sym__list_marker_star] = ACTIONS(1290),
    [sym__list_marker_parenthesis] = ACTIONS(1290),
    [sym__list_marker_dot] = ACTIONS(1290),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1290),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1290),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1290),
    [sym__blank_line_start] = ACTIONS(1290),
    [sym__html_block_1_start] = ACTIONS(1290),
    [sym__html_block_2_start] = ACTIONS(1290),
    [sym__html_block_3_start] = ACTIONS(1290),
    [sym__html_block_4_start] = ACTIONS(1290),
    [sym__html_block_5_start] = ACTIONS(1290),
    [sym__html_block_6_start] = ACTIONS(1290),
    [sym__html_block_7_start] = ACTIONS(1290),
    [sym__pipe_table_start] = ACTIONS(1290),
  },
  [STATE(228)] = {
    [anon_sym_LBRACK] = ACTIONS(1670),
    [anon_sym_RBRACK] = ACTIONS(1672),
    [anon_sym_LT] = ACTIONS(1672),
    [anon_sym_GT] = ACTIONS(1672),
    [anon_sym_BANG] = ACTIONS(1672),
    [anon_sym_DQUOTE] = ACTIONS(1672),
    [anon_sym_POUND] = ACTIONS(1672),
    [anon_sym_DOLLAR] = ACTIONS(1672),
    [anon_sym_PERCENT] = ACTIONS(1672),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_SQUOTE] = ACTIONS(1672),
    [anon_sym_STAR] = ACTIONS(1672),
    [anon_sym_PLUS] = ACTIONS(1672),
    [anon_sym_COMMA] = ACTIONS(1672),
    [anon_sym_DASH] = ACTIONS(1672),
    [anon_sym_DOT] = ACTIONS(1672),
    [anon_sym_SLASH] = ACTIONS(1672),
    [anon_sym_COLON] = ACTIONS(1672),
    [anon_sym_SEMI] = ACTIONS(1672),
    [anon_sym_EQ] = ACTIONS(1672),
    [anon_sym_QMARK] = ACTIONS(1672),
    [anon_sym_AT] = ACTIONS(1672),
    [anon_sym_BSLASH] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1672),
    [anon_sym__] = ACTIONS(1672),
    [anon_sym_BQUOTE] = ACTIONS(1672),
    [anon_sym_LBRACE] = ACTIONS(1672),
    [anon_sym_PIPE] = ACTIONS(1672),
    [anon_sym_RBRACE] = ACTIONS(1672),
    [anon_sym_TILDE] = ACTIONS(1672),
    [anon_sym_LPAREN] = ACTIONS(1672),
    [anon_sym_RPAREN] = ACTIONS(1672),
    [aux_sym__word_token1] = ACTIONS(1672),
    [aux_sym__word_token2] = ACTIONS(1672),
    [aux_sym__word_token3] = ACTIONS(1672),
    [sym__whitespace] = ACTIONS(1672),
    [sym__soft_line_ending] = ACTIONS(1672),
    [sym__block_close] = ACTIONS(1672),
    [sym__block_quote_start] = ACTIONS(1672),
    [sym__indented_chunk_start] = ACTIONS(1672),
    [sym_atx_h1_marker] = ACTIONS(1672),
    [sym_atx_h2_marker] = ACTIONS(1672),
    [sym_atx_h3_marker] = ACTIONS(1672),
    [sym_atx_h4_marker] = ACTIONS(1672),
    [sym_atx_h5_marker] = ACTIONS(1672),
    [sym_atx_h6_marker] = ACTIONS(1672),
    [sym__thematic_break] = ACTIONS(1672),
    [sym__list_marker_minus] = ACTIONS(1672),
    [sym__list_marker_plus] = ACTIONS(1672),
    [sym__list_marker_star] = ACTIONS(1672),
    [sym__list_marker_parenthesis] = ACTIONS(1672),
    [sym__list_marker_dot] = ACTIONS(1672),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1672),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1672),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1672),
    [sym__blank_line_start] = ACTIONS(1672),
    [sym__html_block_1_start] = ACTIONS(1672),
    [sym__html_block_2_start] = ACTIONS(1672),
    [sym__html_block_3_start] = ACTIONS(1672),
    [sym__html_block_4_start] = ACTIONS(1672),
    [sym__html_block_5_start] = ACTIONS(1672),
    [sym__html_block_6_start] = ACTIONS(1672),
    [sym__html_block_7_start] = ACTIONS(1672),
    [sym__pipe_table_start] = ACTIONS(1672),
  },
  [STATE(229)] = {
    [ts_builtin_sym_end] = ACTIONS(1668),
    [anon_sym_LBRACK] = ACTIONS(1666),
    [anon_sym_RBRACK] = ACTIONS(1668),
    [anon_sym_LT] = ACTIONS(1668),
    [anon_sym_GT] = ACTIONS(1668),
    [anon_sym_BANG] = ACTIONS(1668),
    [anon_sym_DQUOTE] = ACTIONS(1668),
    [anon_sym_POUND] = ACTIONS(1668),
    [anon_sym_DOLLAR] = ACTIONS(1668),
    [anon_sym_PERCENT] = ACTIONS(1668),
    [anon_sym_AMP] = ACTIONS(1668),
    [anon_sym_SQUOTE] = ACTIONS(1668),
    [anon_sym_STAR] = ACTIONS(1668),
    [anon_sym_PLUS] = ACTIONS(1668),
    [anon_sym_COMMA] = ACTIONS(1668),
    [anon_sym_DASH] = ACTIONS(1668),
    [anon_sym_DOT] = ACTIONS(1668),
    [anon_sym_SLASH] = ACTIONS(1668),
    [anon_sym_COLON] = ACTIONS(1668),
    [anon_sym_SEMI] = ACTIONS(1668),
    [anon_sym_EQ] = ACTIONS(1668),
    [anon_sym_QMARK] = ACTIONS(1668),
    [anon_sym_AT] = ACTIONS(1668),
    [anon_sym_BSLASH] = ACTIONS(1668),
    [anon_sym_CARET] = ACTIONS(1668),
    [anon_sym__] = ACTIONS(1668),
    [anon_sym_BQUOTE] = ACTIONS(1668),
    [anon_sym_LBRACE] = ACTIONS(1668),
    [anon_sym_PIPE] = ACTIONS(1668),
    [anon_sym_RBRACE] = ACTIONS(1668),
    [anon_sym_TILDE] = ACTIONS(1668),
    [anon_sym_LPAREN] = ACTIONS(1668),
    [anon_sym_RPAREN] = ACTIONS(1668),
    [aux_sym__word_token1] = ACTIONS(1668),
    [aux_sym__word_token2] = ACTIONS(1668),
    [aux_sym__word_token3] = ACTIONS(1668),
    [sym__whitespace] = ACTIONS(1668),
    [sym__soft_line_ending] = ACTIONS(1668),
    [sym__block_quote_start] = ACTIONS(1668),
    [sym__indented_chunk_start] = ACTIONS(1668),
    [sym_atx_h1_marker] = ACTIONS(1668),
    [sym_atx_h2_marker] = ACTIONS(1668),
    [sym_atx_h3_marker] = ACTIONS(1668),
    [sym_atx_h4_marker] = ACTIONS(1668),
    [sym_atx_h5_marker] = ACTIONS(1668),
    [sym_atx_h6_marker] = ACTIONS(1668),
    [sym__thematic_break] = ACTIONS(1668),
    [sym__list_marker_minus] = ACTIONS(1668),
    [sym__list_marker_plus] = ACTIONS(1668),
    [sym__list_marker_star] = ACTIONS(1668),
    [sym__list_marker_parenthesis] = ACTIONS(1668),
    [sym__list_marker_dot] = ACTIONS(1668),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1668),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1668),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1668),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1668),
    [sym__blank_line_start] = ACTIONS(1668),
    [sym__html_block_1_start] = ACTIONS(1668),
    [sym__html_block_2_start] = ACTIONS(1668),
    [sym__html_block_3_start] = ACTIONS(1668),
    [sym__html_block_4_start] = ACTIONS(1668),
    [sym__html_block_5_start] = ACTIONS(1668),
    [sym__html_block_6_start] = ACTIONS(1668),
    [sym__html_block_7_start] = ACTIONS(1668),
    [sym__pipe_table_start] = ACTIONS(1668),
  },
  [STATE(230)] = {
    [anon_sym_LBRACK] = ACTIONS(1674),
    [anon_sym_RBRACK] = ACTIONS(1676),
    [anon_sym_LT] = ACTIONS(1676),
    [anon_sym_GT] = ACTIONS(1676),
    [anon_sym_BANG] = ACTIONS(1676),
    [anon_sym_DQUOTE] = ACTIONS(1676),
    [anon_sym_POUND] = ACTIONS(1676),
    [anon_sym_DOLLAR] = ACTIONS(1676),
    [anon_sym_PERCENT] = ACTIONS(1676),
    [anon_sym_AMP] = ACTIONS(1676),
    [anon_sym_SQUOTE] = ACTIONS(1676),
    [anon_sym_STAR] = ACTIONS(1676),
    [anon_sym_PLUS] = ACTIONS(1676),
    [anon_sym_COMMA] = ACTIONS(1676),
    [anon_sym_DASH] = ACTIONS(1676),
    [anon_sym_DOT] = ACTIONS(1676),
    [anon_sym_SLASH] = ACTIONS(1676),
    [anon_sym_COLON] = ACTIONS(1676),
    [anon_sym_SEMI] = ACTIONS(1676),
    [anon_sym_EQ] = ACTIONS(1676),
    [anon_sym_QMARK] = ACTIONS(1676),
    [anon_sym_AT] = ACTIONS(1676),
    [anon_sym_BSLASH] = ACTIONS(1676),
    [anon_sym_CARET] = ACTIONS(1676),
    [anon_sym__] = ACTIONS(1676),
    [anon_sym_BQUOTE] = ACTIONS(1676),
    [anon_sym_LBRACE] = ACTIONS(1676),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_RBRACE] = ACTIONS(1676),
    [anon_sym_TILDE] = ACTIONS(1676),
    [anon_sym_LPAREN] = ACTIONS(1676),
    [anon_sym_RPAREN] = ACTIONS(1676),
    [aux_sym__word_token1] = ACTIONS(1676),
    [aux_sym__word_token2] = ACTIONS(1676),
    [aux_sym__word_token3] = ACTIONS(1676),
    [sym__whitespace] = ACTIONS(1676),
    [sym__soft_line_ending] = ACTIONS(1676),
    [sym__block_close] = ACTIONS(1676),
    [sym__block_quote_start] = ACTIONS(1676),
    [sym__indented_chunk_start] = ACTIONS(1676),
    [sym_atx_h1_marker] = ACTIONS(1676),
    [sym_atx_h2_marker] = ACTIONS(1676),
    [sym_atx_h3_marker] = ACTIONS(1676),
    [sym_atx_h4_marker] = ACTIONS(1676),
    [sym_atx_h5_marker] = ACTIONS(1676),
    [sym_atx_h6_marker] = ACTIONS(1676),
    [sym__thematic_break] = ACTIONS(1676),
    [sym__list_marker_minus] = ACTIONS(1676),
    [sym__list_marker_plus] = ACTIONS(1676),
    [sym__list_marker_star] = ACTIONS(1676),
    [sym__list_marker_parenthesis] = ACTIONS(1676),
    [sym__list_marker_dot] = ACTIONS(1676),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1676),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1676),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1676),
    [sym__blank_line_start] = ACTIONS(1676),
    [sym__html_block_1_start] = ACTIONS(1676),
    [sym__html_block_2_start] = ACTIONS(1676),
    [sym__html_block_3_start] = ACTIONS(1676),
    [sym__html_block_4_start] = ACTIONS(1676),
    [sym__html_block_5_start] = ACTIONS(1676),
    [sym__html_block_6_start] = ACTIONS(1676),
    [sym__html_block_7_start] = ACTIONS(1676),
    [sym__pipe_table_start] = ACTIONS(1676),
  },
  [STATE(231)] = {
    [ts_builtin_sym_end] = ACTIONS(1374),
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_RBRACK] = ACTIONS(1374),
    [anon_sym_LT] = ACTIONS(1374),
    [anon_sym_GT] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1374),
    [anon_sym_POUND] = ACTIONS(1374),
    [anon_sym_DOLLAR] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1374),
    [anon_sym_PLUS] = ACTIONS(1374),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_DASH] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1374),
    [anon_sym_COLON] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_EQ] = ACTIONS(1374),
    [anon_sym_QMARK] = ACTIONS(1374),
    [anon_sym_AT] = ACTIONS(1374),
    [anon_sym_BSLASH] = ACTIONS(1374),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym__] = ACTIONS(1374),
    [anon_sym_BQUOTE] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_RPAREN] = ACTIONS(1374),
    [aux_sym__word_token1] = ACTIONS(1374),
    [aux_sym__word_token2] = ACTIONS(1374),
    [aux_sym__word_token3] = ACTIONS(1374),
    [sym__whitespace] = ACTIONS(1374),
    [sym__soft_line_ending] = ACTIONS(1374),
    [sym__block_quote_start] = ACTIONS(1374),
    [sym__indented_chunk_start] = ACTIONS(1374),
    [sym_atx_h1_marker] = ACTIONS(1374),
    [sym_atx_h2_marker] = ACTIONS(1374),
    [sym_atx_h3_marker] = ACTIONS(1374),
    [sym_atx_h4_marker] = ACTIONS(1374),
    [sym_atx_h5_marker] = ACTIONS(1374),
    [sym_atx_h6_marker] = ACTIONS(1374),
    [sym__thematic_break] = ACTIONS(1374),
    [sym__list_marker_minus] = ACTIONS(1374),
    [sym__list_marker_plus] = ACTIONS(1374),
    [sym__list_marker_star] = ACTIONS(1374),
    [sym__list_marker_parenthesis] = ACTIONS(1374),
    [sym__list_marker_dot] = ACTIONS(1374),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1374),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1374),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1374),
    [sym__blank_line_start] = ACTIONS(1374),
    [sym__html_block_1_start] = ACTIONS(1374),
    [sym__html_block_2_start] = ACTIONS(1374),
    [sym__html_block_3_start] = ACTIONS(1374),
    [sym__html_block_4_start] = ACTIONS(1374),
    [sym__html_block_5_start] = ACTIONS(1374),
    [sym__html_block_6_start] = ACTIONS(1374),
    [sym__html_block_7_start] = ACTIONS(1374),
    [sym__pipe_table_start] = ACTIONS(1374),
  },
  [STATE(232)] = {
    [anon_sym_LBRACK] = ACTIONS(1678),
    [anon_sym_RBRACK] = ACTIONS(1680),
    [anon_sym_LT] = ACTIONS(1680),
    [anon_sym_GT] = ACTIONS(1680),
    [anon_sym_BANG] = ACTIONS(1680),
    [anon_sym_DQUOTE] = ACTIONS(1680),
    [anon_sym_POUND] = ACTIONS(1680),
    [anon_sym_DOLLAR] = ACTIONS(1680),
    [anon_sym_PERCENT] = ACTIONS(1680),
    [anon_sym_AMP] = ACTIONS(1680),
    [anon_sym_SQUOTE] = ACTIONS(1680),
    [anon_sym_STAR] = ACTIONS(1680),
    [anon_sym_PLUS] = ACTIONS(1680),
    [anon_sym_COMMA] = ACTIONS(1680),
    [anon_sym_DASH] = ACTIONS(1680),
    [anon_sym_DOT] = ACTIONS(1680),
    [anon_sym_SLASH] = ACTIONS(1680),
    [anon_sym_COLON] = ACTIONS(1680),
    [anon_sym_SEMI] = ACTIONS(1680),
    [anon_sym_EQ] = ACTIONS(1680),
    [anon_sym_QMARK] = ACTIONS(1680),
    [anon_sym_AT] = ACTIONS(1680),
    [anon_sym_BSLASH] = ACTIONS(1680),
    [anon_sym_CARET] = ACTIONS(1680),
    [anon_sym__] = ACTIONS(1680),
    [anon_sym_BQUOTE] = ACTIONS(1680),
    [anon_sym_LBRACE] = ACTIONS(1680),
    [anon_sym_PIPE] = ACTIONS(1680),
    [anon_sym_RBRACE] = ACTIONS(1680),
    [anon_sym_TILDE] = ACTIONS(1680),
    [anon_sym_LPAREN] = ACTIONS(1680),
    [anon_sym_RPAREN] = ACTIONS(1680),
    [aux_sym__word_token1] = ACTIONS(1680),
    [aux_sym__word_token2] = ACTIONS(1680),
    [aux_sym__word_token3] = ACTIONS(1680),
    [sym__whitespace] = ACTIONS(1680),
    [sym__soft_line_ending] = ACTIONS(1680),
    [sym__block_close] = ACTIONS(1680),
    [sym__block_quote_start] = ACTIONS(1680),
    [sym__indented_chunk_start] = ACTIONS(1680),
    [sym_atx_h1_marker] = ACTIONS(1680),
    [sym_atx_h2_marker] = ACTIONS(1680),
    [sym_atx_h3_marker] = ACTIONS(1680),
    [sym_atx_h4_marker] = ACTIONS(1680),
    [sym_atx_h5_marker] = ACTIONS(1680),
    [sym_atx_h6_marker] = ACTIONS(1680),
    [sym__thematic_break] = ACTIONS(1680),
    [sym__list_marker_minus] = ACTIONS(1680),
    [sym__list_marker_plus] = ACTIONS(1680),
    [sym__list_marker_star] = ACTIONS(1680),
    [sym__list_marker_parenthesis] = ACTIONS(1680),
    [sym__list_marker_dot] = ACTIONS(1680),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1680),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1680),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1680),
    [sym__blank_line_start] = ACTIONS(1680),
    [sym__html_block_1_start] = ACTIONS(1680),
    [sym__html_block_2_start] = ACTIONS(1680),
    [sym__html_block_3_start] = ACTIONS(1680),
    [sym__html_block_4_start] = ACTIONS(1680),
    [sym__html_block_5_start] = ACTIONS(1680),
    [sym__html_block_6_start] = ACTIONS(1680),
    [sym__html_block_7_start] = ACTIONS(1680),
    [sym__pipe_table_start] = ACTIONS(1680),
  },
  [STATE(233)] = {
    [ts_builtin_sym_end] = ACTIONS(1600),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_LT] = ACTIONS(1600),
    [anon_sym_GT] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [anon_sym_POUND] = ACTIONS(1600),
    [anon_sym_DOLLAR] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1600),
    [anon_sym_AMP] = ACTIONS(1600),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_STAR] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_DASH] = ACTIONS(1600),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_SLASH] = ACTIONS(1600),
    [anon_sym_COLON] = ACTIONS(1600),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_EQ] = ACTIONS(1600),
    [anon_sym_QMARK] = ACTIONS(1600),
    [anon_sym_AT] = ACTIONS(1600),
    [anon_sym_BSLASH] = ACTIONS(1600),
    [anon_sym_CARET] = ACTIONS(1600),
    [anon_sym__] = ACTIONS(1600),
    [anon_sym_BQUOTE] = ACTIONS(1600),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [aux_sym__word_token1] = ACTIONS(1600),
    [aux_sym__word_token2] = ACTIONS(1600),
    [aux_sym__word_token3] = ACTIONS(1600),
    [sym__whitespace] = ACTIONS(1600),
    [sym__soft_line_ending] = ACTIONS(1600),
    [sym__block_quote_start] = ACTIONS(1600),
    [sym__indented_chunk_start] = ACTIONS(1600),
    [sym_atx_h1_marker] = ACTIONS(1600),
    [sym_atx_h2_marker] = ACTIONS(1600),
    [sym_atx_h3_marker] = ACTIONS(1600),
    [sym_atx_h4_marker] = ACTIONS(1600),
    [sym_atx_h5_marker] = ACTIONS(1600),
    [sym_atx_h6_marker] = ACTIONS(1600),
    [sym__thematic_break] = ACTIONS(1600),
    [sym__list_marker_minus] = ACTIONS(1600),
    [sym__list_marker_plus] = ACTIONS(1600),
    [sym__list_marker_star] = ACTIONS(1600),
    [sym__list_marker_parenthesis] = ACTIONS(1600),
    [sym__list_marker_dot] = ACTIONS(1600),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1600),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1600),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1600),
    [sym__blank_line_start] = ACTIONS(1600),
    [sym__html_block_1_start] = ACTIONS(1600),
    [sym__html_block_2_start] = ACTIONS(1600),
    [sym__html_block_3_start] = ACTIONS(1600),
    [sym__html_block_4_start] = ACTIONS(1600),
    [sym__html_block_5_start] = ACTIONS(1600),
    [sym__html_block_6_start] = ACTIONS(1600),
    [sym__html_block_7_start] = ACTIONS(1600),
    [sym__pipe_table_start] = ACTIONS(1600),
  },
  [STATE(234)] = {
    [anon_sym_LBRACK] = ACTIONS(1682),
    [anon_sym_RBRACK] = ACTIONS(1684),
    [anon_sym_LT] = ACTIONS(1684),
    [anon_sym_GT] = ACTIONS(1684),
    [anon_sym_BANG] = ACTIONS(1684),
    [anon_sym_DQUOTE] = ACTIONS(1684),
    [anon_sym_POUND] = ACTIONS(1684),
    [anon_sym_DOLLAR] = ACTIONS(1684),
    [anon_sym_PERCENT] = ACTIONS(1684),
    [anon_sym_AMP] = ACTIONS(1684),
    [anon_sym_SQUOTE] = ACTIONS(1684),
    [anon_sym_STAR] = ACTIONS(1684),
    [anon_sym_PLUS] = ACTIONS(1684),
    [anon_sym_COMMA] = ACTIONS(1684),
    [anon_sym_DASH] = ACTIONS(1684),
    [anon_sym_DOT] = ACTIONS(1684),
    [anon_sym_SLASH] = ACTIONS(1684),
    [anon_sym_COLON] = ACTIONS(1684),
    [anon_sym_SEMI] = ACTIONS(1684),
    [anon_sym_EQ] = ACTIONS(1684),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_AT] = ACTIONS(1684),
    [anon_sym_BSLASH] = ACTIONS(1684),
    [anon_sym_CARET] = ACTIONS(1684),
    [anon_sym__] = ACTIONS(1684),
    [anon_sym_BQUOTE] = ACTIONS(1684),
    [anon_sym_LBRACE] = ACTIONS(1684),
    [anon_sym_PIPE] = ACTIONS(1684),
    [anon_sym_RBRACE] = ACTIONS(1684),
    [anon_sym_TILDE] = ACTIONS(1684),
    [anon_sym_LPAREN] = ACTIONS(1684),
    [anon_sym_RPAREN] = ACTIONS(1684),
    [aux_sym__word_token1] = ACTIONS(1684),
    [aux_sym__word_token2] = ACTIONS(1684),
    [aux_sym__word_token3] = ACTIONS(1684),
    [sym__whitespace] = ACTIONS(1684),
    [sym__soft_line_ending] = ACTIONS(1684),
    [sym__block_close] = ACTIONS(1684),
    [sym__block_quote_start] = ACTIONS(1684),
    [sym__indented_chunk_start] = ACTIONS(1684),
    [sym_atx_h1_marker] = ACTIONS(1684),
    [sym_atx_h2_marker] = ACTIONS(1684),
    [sym_atx_h3_marker] = ACTIONS(1684),
    [sym_atx_h4_marker] = ACTIONS(1684),
    [sym_atx_h5_marker] = ACTIONS(1684),
    [sym_atx_h6_marker] = ACTIONS(1684),
    [sym__thematic_break] = ACTIONS(1684),
    [sym__list_marker_minus] = ACTIONS(1684),
    [sym__list_marker_plus] = ACTIONS(1684),
    [sym__list_marker_star] = ACTIONS(1684),
    [sym__list_marker_parenthesis] = ACTIONS(1684),
    [sym__list_marker_dot] = ACTIONS(1684),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1684),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1684),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1684),
    [sym__blank_line_start] = ACTIONS(1684),
    [sym__html_block_1_start] = ACTIONS(1684),
    [sym__html_block_2_start] = ACTIONS(1684),
    [sym__html_block_3_start] = ACTIONS(1684),
    [sym__html_block_4_start] = ACTIONS(1684),
    [sym__html_block_5_start] = ACTIONS(1684),
    [sym__html_block_6_start] = ACTIONS(1684),
    [sym__html_block_7_start] = ACTIONS(1684),
    [sym__pipe_table_start] = ACTIONS(1684),
  },
  [STATE(235)] = {
    [ts_builtin_sym_end] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_POUND] = ACTIONS(1380),
    [anon_sym_DOLLAR] = ACTIONS(1380),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_BSLASH] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_BQUOTE] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [aux_sym__word_token1] = ACTIONS(1380),
    [aux_sym__word_token2] = ACTIONS(1380),
    [aux_sym__word_token3] = ACTIONS(1380),
    [sym__whitespace] = ACTIONS(1380),
    [sym__soft_line_ending] = ACTIONS(1380),
    [sym__block_quote_start] = ACTIONS(1380),
    [sym__indented_chunk_start] = ACTIONS(1380),
    [sym_atx_h1_marker] = ACTIONS(1380),
    [sym_atx_h2_marker] = ACTIONS(1380),
    [sym_atx_h3_marker] = ACTIONS(1380),
    [sym_atx_h4_marker] = ACTIONS(1380),
    [sym_atx_h5_marker] = ACTIONS(1380),
    [sym_atx_h6_marker] = ACTIONS(1380),
    [sym__thematic_break] = ACTIONS(1380),
    [sym__list_marker_minus] = ACTIONS(1380),
    [sym__list_marker_plus] = ACTIONS(1380),
    [sym__list_marker_star] = ACTIONS(1380),
    [sym__list_marker_parenthesis] = ACTIONS(1380),
    [sym__list_marker_dot] = ACTIONS(1380),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1380),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1380),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1380),
    [sym__blank_line_start] = ACTIONS(1380),
    [sym__html_block_1_start] = ACTIONS(1380),
    [sym__html_block_2_start] = ACTIONS(1380),
    [sym__html_block_3_start] = ACTIONS(1380),
    [sym__html_block_4_start] = ACTIONS(1380),
    [sym__html_block_5_start] = ACTIONS(1380),
    [sym__html_block_6_start] = ACTIONS(1380),
    [sym__html_block_7_start] = ACTIONS(1380),
    [sym__pipe_table_start] = ACTIONS(1380),
  },
  [STATE(236)] = {
    [ts_builtin_sym_end] = ACTIONS(1672),
    [anon_sym_LBRACK] = ACTIONS(1670),
    [anon_sym_RBRACK] = ACTIONS(1672),
    [anon_sym_LT] = ACTIONS(1672),
    [anon_sym_GT] = ACTIONS(1672),
    [anon_sym_BANG] = ACTIONS(1672),
    [anon_sym_DQUOTE] = ACTIONS(1672),
    [anon_sym_POUND] = ACTIONS(1672),
    [anon_sym_DOLLAR] = ACTIONS(1672),
    [anon_sym_PERCENT] = ACTIONS(1672),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_SQUOTE] = ACTIONS(1672),
    [anon_sym_STAR] = ACTIONS(1672),
    [anon_sym_PLUS] = ACTIONS(1672),
    [anon_sym_COMMA] = ACTIONS(1672),
    [anon_sym_DASH] = ACTIONS(1672),
    [anon_sym_DOT] = ACTIONS(1672),
    [anon_sym_SLASH] = ACTIONS(1672),
    [anon_sym_COLON] = ACTIONS(1672),
    [anon_sym_SEMI] = ACTIONS(1672),
    [anon_sym_EQ] = ACTIONS(1672),
    [anon_sym_QMARK] = ACTIONS(1672),
    [anon_sym_AT] = ACTIONS(1672),
    [anon_sym_BSLASH] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1672),
    [anon_sym__] = ACTIONS(1672),
    [anon_sym_BQUOTE] = ACTIONS(1672),
    [anon_sym_LBRACE] = ACTIONS(1672),
    [anon_sym_PIPE] = ACTIONS(1672),
    [anon_sym_RBRACE] = ACTIONS(1672),
    [anon_sym_TILDE] = ACTIONS(1672),
    [anon_sym_LPAREN] = ACTIONS(1672),
    [anon_sym_RPAREN] = ACTIONS(1672),
    [aux_sym__word_token1] = ACTIONS(1672),
    [aux_sym__word_token2] = ACTIONS(1672),
    [aux_sym__word_token3] = ACTIONS(1672),
    [sym__whitespace] = ACTIONS(1672),
    [sym__soft_line_ending] = ACTIONS(1672),
    [sym__block_quote_start] = ACTIONS(1672),
    [sym__indented_chunk_start] = ACTIONS(1672),
    [sym_atx_h1_marker] = ACTIONS(1672),
    [sym_atx_h2_marker] = ACTIONS(1672),
    [sym_atx_h3_marker] = ACTIONS(1672),
    [sym_atx_h4_marker] = ACTIONS(1672),
    [sym_atx_h5_marker] = ACTIONS(1672),
    [sym_atx_h6_marker] = ACTIONS(1672),
    [sym__thematic_break] = ACTIONS(1672),
    [sym__list_marker_minus] = ACTIONS(1672),
    [sym__list_marker_plus] = ACTIONS(1672),
    [sym__list_marker_star] = ACTIONS(1672),
    [sym__list_marker_parenthesis] = ACTIONS(1672),
    [sym__list_marker_dot] = ACTIONS(1672),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1672),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1672),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1672),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1672),
    [sym__blank_line_start] = ACTIONS(1672),
    [sym__html_block_1_start] = ACTIONS(1672),
    [sym__html_block_2_start] = ACTIONS(1672),
    [sym__html_block_3_start] = ACTIONS(1672),
    [sym__html_block_4_start] = ACTIONS(1672),
    [sym__html_block_5_start] = ACTIONS(1672),
    [sym__html_block_6_start] = ACTIONS(1672),
    [sym__html_block_7_start] = ACTIONS(1672),
    [sym__pipe_table_start] = ACTIONS(1672),
  },
  [STATE(237)] = {
    [ts_builtin_sym_end] = ACTIONS(1600),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_LT] = ACTIONS(1600),
    [anon_sym_GT] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [anon_sym_POUND] = ACTIONS(1600),
    [anon_sym_DOLLAR] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1600),
    [anon_sym_AMP] = ACTIONS(1600),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_STAR] = ACTIONS(1600),
    [anon_sym_PLUS] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_DASH] = ACTIONS(1600),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_SLASH] = ACTIONS(1600),
    [anon_sym_COLON] = ACTIONS(1600),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_EQ] = ACTIONS(1600),
    [anon_sym_QMARK] = ACTIONS(1600),
    [anon_sym_AT] = ACTIONS(1600),
    [anon_sym_BSLASH] = ACTIONS(1600),
    [anon_sym_CARET] = ACTIONS(1600),
    [anon_sym__] = ACTIONS(1600),
    [anon_sym_BQUOTE] = ACTIONS(1600),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [aux_sym__word_token1] = ACTIONS(1600),
    [aux_sym__word_token2] = ACTIONS(1600),
    [aux_sym__word_token3] = ACTIONS(1600),
    [sym__whitespace] = ACTIONS(1600),
    [sym__soft_line_ending] = ACTIONS(1600),
    [sym__block_quote_start] = ACTIONS(1600),
    [sym__indented_chunk_start] = ACTIONS(1600),
    [sym_atx_h1_marker] = ACTIONS(1600),
    [sym_atx_h2_marker] = ACTIONS(1600),
    [sym_atx_h3_marker] = ACTIONS(1600),
    [sym_atx_h4_marker] = ACTIONS(1600),
    [sym_atx_h5_marker] = ACTIONS(1600),
    [sym_atx_h6_marker] = ACTIONS(1600),
    [sym__thematic_break] = ACTIONS(1600),
    [sym__list_marker_minus] = ACTIONS(1600),
    [sym__list_marker_plus] = ACTIONS(1600),
    [sym__list_marker_star] = ACTIONS(1600),
    [sym__list_marker_parenthesis] = ACTIONS(1600),
    [sym__list_marker_dot] = ACTIONS(1600),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1600),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1600),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1600),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1600),
    [sym__blank_line_start] = ACTIONS(1600),
    [sym__html_block_1_start] = ACTIONS(1600),
    [sym__html_block_2_start] = ACTIONS(1600),
    [sym__html_block_3_start] = ACTIONS(1600),
    [sym__html_block_4_start] = ACTIONS(1600),
    [sym__html_block_5_start] = ACTIONS(1600),
    [sym__html_block_6_start] = ACTIONS(1600),
    [sym__html_block_7_start] = ACTIONS(1600),
    [sym__pipe_table_start] = ACTIONS(1600),
  },
  [STATE(238)] = {
    [ts_builtin_sym_end] = ACTIONS(1676),
    [anon_sym_LBRACK] = ACTIONS(1674),
    [anon_sym_RBRACK] = ACTIONS(1676),
    [anon_sym_LT] = ACTIONS(1676),
    [anon_sym_GT] = ACTIONS(1676),
    [anon_sym_BANG] = ACTIONS(1676),
    [anon_sym_DQUOTE] = ACTIONS(1676),
    [anon_sym_POUND] = ACTIONS(1676),
    [anon_sym_DOLLAR] = ACTIONS(1676),
    [anon_sym_PERCENT] = ACTIONS(1676),
    [anon_sym_AMP] = ACTIONS(1676),
    [anon_sym_SQUOTE] = ACTIONS(1676),
    [anon_sym_STAR] = ACTIONS(1676),
    [anon_sym_PLUS] = ACTIONS(1676),
    [anon_sym_COMMA] = ACTIONS(1676),
    [anon_sym_DASH] = ACTIONS(1676),
    [anon_sym_DOT] = ACTIONS(1676),
    [anon_sym_SLASH] = ACTIONS(1676),
    [anon_sym_COLON] = ACTIONS(1676),
    [anon_sym_SEMI] = ACTIONS(1676),
    [anon_sym_EQ] = ACTIONS(1676),
    [anon_sym_QMARK] = ACTIONS(1676),
    [anon_sym_AT] = ACTIONS(1676),
    [anon_sym_BSLASH] = ACTIONS(1676),
    [anon_sym_CARET] = ACTIONS(1676),
    [anon_sym__] = ACTIONS(1676),
    [anon_sym_BQUOTE] = ACTIONS(1676),
    [anon_sym_LBRACE] = ACTIONS(1676),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_RBRACE] = ACTIONS(1676),
    [anon_sym_TILDE] = ACTIONS(1676),
    [anon_sym_LPAREN] = ACTIONS(1676),
    [anon_sym_RPAREN] = ACTIONS(1676),
    [aux_sym__word_token1] = ACTIONS(1676),
    [aux_sym__word_token2] = ACTIONS(1676),
    [aux_sym__word_token3] = ACTIONS(1676),
    [sym__whitespace] = ACTIONS(1676),
    [sym__soft_line_ending] = ACTIONS(1676),
    [sym__block_quote_start] = ACTIONS(1676),
    [sym__indented_chunk_start] = ACTIONS(1676),
    [sym_atx_h1_marker] = ACTIONS(1676),
    [sym_atx_h2_marker] = ACTIONS(1676),
    [sym_atx_h3_marker] = ACTIONS(1676),
    [sym_atx_h4_marker] = ACTIONS(1676),
    [sym_atx_h5_marker] = ACTIONS(1676),
    [sym_atx_h6_marker] = ACTIONS(1676),
    [sym__thematic_break] = ACTIONS(1676),
    [sym__list_marker_minus] = ACTIONS(1676),
    [sym__list_marker_plus] = ACTIONS(1676),
    [sym__list_marker_star] = ACTIONS(1676),
    [sym__list_marker_parenthesis] = ACTIONS(1676),
    [sym__list_marker_dot] = ACTIONS(1676),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1676),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1676),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1676),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1676),
    [sym__blank_line_start] = ACTIONS(1676),
    [sym__html_block_1_start] = ACTIONS(1676),
    [sym__html_block_2_start] = ACTIONS(1676),
    [sym__html_block_3_start] = ACTIONS(1676),
    [sym__html_block_4_start] = ACTIONS(1676),
    [sym__html_block_5_start] = ACTIONS(1676),
    [sym__html_block_6_start] = ACTIONS(1676),
    [sym__html_block_7_start] = ACTIONS(1676),
    [sym__pipe_table_start] = ACTIONS(1676),
  },
  [STATE(239)] = {
    [anon_sym_LBRACK] = ACTIONS(1686),
    [anon_sym_RBRACK] = ACTIONS(1688),
    [anon_sym_LT] = ACTIONS(1688),
    [anon_sym_GT] = ACTIONS(1688),
    [anon_sym_BANG] = ACTIONS(1688),
    [anon_sym_DQUOTE] = ACTIONS(1688),
    [anon_sym_POUND] = ACTIONS(1688),
    [anon_sym_DOLLAR] = ACTIONS(1688),
    [anon_sym_PERCENT] = ACTIONS(1688),
    [anon_sym_AMP] = ACTIONS(1688),
    [anon_sym_SQUOTE] = ACTIONS(1688),
    [anon_sym_STAR] = ACTIONS(1688),
    [anon_sym_PLUS] = ACTIONS(1688),
    [anon_sym_COMMA] = ACTIONS(1688),
    [anon_sym_DASH] = ACTIONS(1688),
    [anon_sym_DOT] = ACTIONS(1688),
    [anon_sym_SLASH] = ACTIONS(1688),
    [anon_sym_COLON] = ACTIONS(1688),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_EQ] = ACTIONS(1688),
    [anon_sym_QMARK] = ACTIONS(1688),
    [anon_sym_AT] = ACTIONS(1688),
    [anon_sym_BSLASH] = ACTIONS(1688),
    [anon_sym_CARET] = ACTIONS(1688),
    [anon_sym__] = ACTIONS(1688),
    [anon_sym_BQUOTE] = ACTIONS(1688),
    [anon_sym_LBRACE] = ACTIONS(1688),
    [anon_sym_PIPE] = ACTIONS(1688),
    [anon_sym_RBRACE] = ACTIONS(1688),
    [anon_sym_TILDE] = ACTIONS(1688),
    [anon_sym_LPAREN] = ACTIONS(1688),
    [anon_sym_RPAREN] = ACTIONS(1688),
    [aux_sym__word_token1] = ACTIONS(1688),
    [aux_sym__word_token2] = ACTIONS(1688),
    [aux_sym__word_token3] = ACTIONS(1688),
    [sym__whitespace] = ACTIONS(1688),
    [sym__soft_line_ending] = ACTIONS(1688),
    [sym_block_continuation] = ACTIONS(1688),
    [sym__block_quote_start] = ACTIONS(1688),
    [sym__indented_chunk_start] = ACTIONS(1688),
    [sym_atx_h1_marker] = ACTIONS(1688),
    [sym_atx_h2_marker] = ACTIONS(1688),
    [sym_atx_h3_marker] = ACTIONS(1688),
    [sym_atx_h4_marker] = ACTIONS(1688),
    [sym_atx_h5_marker] = ACTIONS(1688),
    [sym_atx_h6_marker] = ACTIONS(1688),
    [sym__thematic_break] = ACTIONS(1688),
    [sym__list_marker_minus] = ACTIONS(1688),
    [sym__list_marker_plus] = ACTIONS(1688),
    [sym__list_marker_star] = ACTIONS(1688),
    [sym__list_marker_parenthesis] = ACTIONS(1688),
    [sym__list_marker_dot] = ACTIONS(1688),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1688),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1688),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1688),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1688),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1688),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1688),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1688),
    [sym__blank_line_start] = ACTIONS(1688),
    [sym__html_block_1_start] = ACTIONS(1688),
    [sym__html_block_2_start] = ACTIONS(1688),
    [sym__html_block_3_start] = ACTIONS(1688),
    [sym__html_block_4_start] = ACTIONS(1688),
    [sym__html_block_5_start] = ACTIONS(1688),
    [sym__html_block_6_start] = ACTIONS(1688),
    [sym__html_block_7_start] = ACTIONS(1688),
    [sym__pipe_table_start] = ACTIONS(1688),
  },
  [STATE(240)] = {
    [ts_builtin_sym_end] = ACTIONS(1680),
    [anon_sym_LBRACK] = ACTIONS(1678),
    [anon_sym_RBRACK] = ACTIONS(1680),
    [anon_sym_LT] = ACTIONS(1680),
    [anon_sym_GT] = ACTIONS(1680),
    [anon_sym_BANG] = ACTIONS(1680),
    [anon_sym_DQUOTE] = ACTIONS(1680),
    [anon_sym_POUND] = ACTIONS(1680),
    [anon_sym_DOLLAR] = ACTIONS(1680),
    [anon_sym_PERCENT] = ACTIONS(1680),
    [anon_sym_AMP] = ACTIONS(1680),
    [anon_sym_SQUOTE] = ACTIONS(1680),
    [anon_sym_STAR] = ACTIONS(1680),
    [anon_sym_PLUS] = ACTIONS(1680),
    [anon_sym_COMMA] = ACTIONS(1680),
    [anon_sym_DASH] = ACTIONS(1680),
    [anon_sym_DOT] = ACTIONS(1680),
    [anon_sym_SLASH] = ACTIONS(1680),
    [anon_sym_COLON] = ACTIONS(1680),
    [anon_sym_SEMI] = ACTIONS(1680),
    [anon_sym_EQ] = ACTIONS(1680),
    [anon_sym_QMARK] = ACTIONS(1680),
    [anon_sym_AT] = ACTIONS(1680),
    [anon_sym_BSLASH] = ACTIONS(1680),
    [anon_sym_CARET] = ACTIONS(1680),
    [anon_sym__] = ACTIONS(1680),
    [anon_sym_BQUOTE] = ACTIONS(1680),
    [anon_sym_LBRACE] = ACTIONS(1680),
    [anon_sym_PIPE] = ACTIONS(1680),
    [anon_sym_RBRACE] = ACTIONS(1680),
    [anon_sym_TILDE] = ACTIONS(1680),
    [anon_sym_LPAREN] = ACTIONS(1680),
    [anon_sym_RPAREN] = ACTIONS(1680),
    [aux_sym__word_token1] = ACTIONS(1680),
    [aux_sym__word_token2] = ACTIONS(1680),
    [aux_sym__word_token3] = ACTIONS(1680),
    [sym__whitespace] = ACTIONS(1680),
    [sym__soft_line_ending] = ACTIONS(1680),
    [sym__block_quote_start] = ACTIONS(1680),
    [sym__indented_chunk_start] = ACTIONS(1680),
    [sym_atx_h1_marker] = ACTIONS(1680),
    [sym_atx_h2_marker] = ACTIONS(1680),
    [sym_atx_h3_marker] = ACTIONS(1680),
    [sym_atx_h4_marker] = ACTIONS(1680),
    [sym_atx_h5_marker] = ACTIONS(1680),
    [sym_atx_h6_marker] = ACTIONS(1680),
    [sym__thematic_break] = ACTIONS(1680),
    [sym__list_marker_minus] = ACTIONS(1680),
    [sym__list_marker_plus] = ACTIONS(1680),
    [sym__list_marker_star] = ACTIONS(1680),
    [sym__list_marker_parenthesis] = ACTIONS(1680),
    [sym__list_marker_dot] = ACTIONS(1680),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1680),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1680),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1680),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1680),
    [sym__blank_line_start] = ACTIONS(1680),
    [sym__html_block_1_start] = ACTIONS(1680),
    [sym__html_block_2_start] = ACTIONS(1680),
    [sym__html_block_3_start] = ACTIONS(1680),
    [sym__html_block_4_start] = ACTIONS(1680),
    [sym__html_block_5_start] = ACTIONS(1680),
    [sym__html_block_6_start] = ACTIONS(1680),
    [sym__html_block_7_start] = ACTIONS(1680),
    [sym__pipe_table_start] = ACTIONS(1680),
  },
  [STATE(241)] = {
    [ts_builtin_sym_end] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_LT] = ACTIONS(1608),
    [anon_sym_GT] = ACTIONS(1608),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_POUND] = ACTIONS(1608),
    [anon_sym_DOLLAR] = ACTIONS(1608),
    [anon_sym_PERCENT] = ACTIONS(1608),
    [anon_sym_AMP] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_STAR] = ACTIONS(1608),
    [anon_sym_PLUS] = ACTIONS(1608),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_DASH] = ACTIONS(1608),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_SLASH] = ACTIONS(1608),
    [anon_sym_COLON] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1608),
    [anon_sym_EQ] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_BSLASH] = ACTIONS(1608),
    [anon_sym_CARET] = ACTIONS(1608),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_BQUOTE] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1608),
    [anon_sym_RBRACE] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_RPAREN] = ACTIONS(1608),
    [aux_sym__word_token1] = ACTIONS(1608),
    [aux_sym__word_token2] = ACTIONS(1608),
    [aux_sym__word_token3] = ACTIONS(1608),
    [sym__whitespace] = ACTIONS(1608),
    [sym__soft_line_ending] = ACTIONS(1608),
    [sym__block_quote_start] = ACTIONS(1608),
    [sym__indented_chunk_start] = ACTIONS(1608),
    [sym_atx_h1_marker] = ACTIONS(1608),
    [sym_atx_h2_marker] = ACTIONS(1608),
    [sym_atx_h3_marker] = ACTIONS(1608),
    [sym_atx_h4_marker] = ACTIONS(1608),
    [sym_atx_h5_marker] = ACTIONS(1608),
    [sym_atx_h6_marker] = ACTIONS(1608),
    [sym__thematic_break] = ACTIONS(1608),
    [sym__list_marker_minus] = ACTIONS(1608),
    [sym__list_marker_plus] = ACTIONS(1608),
    [sym__list_marker_star] = ACTIONS(1608),
    [sym__list_marker_parenthesis] = ACTIONS(1608),
    [sym__list_marker_dot] = ACTIONS(1608),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1608),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1608),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1608),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1608),
    [sym__blank_line_start] = ACTIONS(1608),
    [sym__html_block_1_start] = ACTIONS(1608),
    [sym__html_block_2_start] = ACTIONS(1608),
    [sym__html_block_3_start] = ACTIONS(1608),
    [sym__html_block_4_start] = ACTIONS(1608),
    [sym__html_block_5_start] = ACTIONS(1608),
    [sym__html_block_6_start] = ACTIONS(1608),
    [sym__html_block_7_start] = ACTIONS(1608),
    [sym__pipe_table_start] = ACTIONS(1608),
  },
  [STATE(242)] = {
    [ts_builtin_sym_end] = ACTIONS(1684),
    [anon_sym_LBRACK] = ACTIONS(1682),
    [anon_sym_RBRACK] = ACTIONS(1684),
    [anon_sym_LT] = ACTIONS(1684),
    [anon_sym_GT] = ACTIONS(1684),
    [anon_sym_BANG] = ACTIONS(1684),
    [anon_sym_DQUOTE] = ACTIONS(1684),
    [anon_sym_POUND] = ACTIONS(1684),
    [anon_sym_DOLLAR] = ACTIONS(1684),
    [anon_sym_PERCENT] = ACTIONS(1684),
    [anon_sym_AMP] = ACTIONS(1684),
    [anon_sym_SQUOTE] = ACTIONS(1684),
    [anon_sym_STAR] = ACTIONS(1684),
    [anon_sym_PLUS] = ACTIONS(1684),
    [anon_sym_COMMA] = ACTIONS(1684),
    [anon_sym_DASH] = ACTIONS(1684),
    [anon_sym_DOT] = ACTIONS(1684),
    [anon_sym_SLASH] = ACTIONS(1684),
    [anon_sym_COLON] = ACTIONS(1684),
    [anon_sym_SEMI] = ACTIONS(1684),
    [anon_sym_EQ] = ACTIONS(1684),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_AT] = ACTIONS(1684),
    [anon_sym_BSLASH] = ACTIONS(1684),
    [anon_sym_CARET] = ACTIONS(1684),
    [anon_sym__] = ACTIONS(1684),
    [anon_sym_BQUOTE] = ACTIONS(1684),
    [anon_sym_LBRACE] = ACTIONS(1684),
    [anon_sym_PIPE] = ACTIONS(1684),
    [anon_sym_RBRACE] = ACTIONS(1684),
    [anon_sym_TILDE] = ACTIONS(1684),
    [anon_sym_LPAREN] = ACTIONS(1684),
    [anon_sym_RPAREN] = ACTIONS(1684),
    [aux_sym__word_token1] = ACTIONS(1684),
    [aux_sym__word_token2] = ACTIONS(1684),
    [aux_sym__word_token3] = ACTIONS(1684),
    [sym__whitespace] = ACTIONS(1684),
    [sym__soft_line_ending] = ACTIONS(1684),
    [sym__block_quote_start] = ACTIONS(1684),
    [sym__indented_chunk_start] = ACTIONS(1684),
    [sym_atx_h1_marker] = ACTIONS(1684),
    [sym_atx_h2_marker] = ACTIONS(1684),
    [sym_atx_h3_marker] = ACTIONS(1684),
    [sym_atx_h4_marker] = ACTIONS(1684),
    [sym_atx_h5_marker] = ACTIONS(1684),
    [sym_atx_h6_marker] = ACTIONS(1684),
    [sym__thematic_break] = ACTIONS(1684),
    [sym__list_marker_minus] = ACTIONS(1684),
    [sym__list_marker_plus] = ACTIONS(1684),
    [sym__list_marker_star] = ACTIONS(1684),
    [sym__list_marker_parenthesis] = ACTIONS(1684),
    [sym__list_marker_dot] = ACTIONS(1684),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1684),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1684),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1684),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1684),
    [sym__blank_line_start] = ACTIONS(1684),
    [sym__html_block_1_start] = ACTIONS(1684),
    [sym__html_block_2_start] = ACTIONS(1684),
    [sym__html_block_3_start] = ACTIONS(1684),
    [sym__html_block_4_start] = ACTIONS(1684),
    [sym__html_block_5_start] = ACTIONS(1684),
    [sym__html_block_6_start] = ACTIONS(1684),
    [sym__html_block_7_start] = ACTIONS(1684),
    [sym__pipe_table_start] = ACTIONS(1684),
  },
  [STATE(243)] = {
    [anon_sym_LBRACK] = ACTIONS(1330),
    [anon_sym_RBRACK] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_GT] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_DOLLAR] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1332),
    [anon_sym_COLON] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_EQ] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1332),
    [anon_sym_BSLASH] = ACTIONS(1332),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym__] = ACTIONS(1332),
    [anon_sym_BQUOTE] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [aux_sym__word_token1] = ACTIONS(1332),
    [aux_sym__word_token2] = ACTIONS(1332),
    [aux_sym__word_token3] = ACTIONS(1332),
    [sym__whitespace] = ACTIONS(1332),
    [sym__soft_line_ending] = ACTIONS(1332),
    [sym__block_close] = ACTIONS(1332),
    [sym__block_quote_start] = ACTIONS(1332),
    [sym__indented_chunk_start] = ACTIONS(1332),
    [sym_atx_h1_marker] = ACTIONS(1332),
    [sym_atx_h2_marker] = ACTIONS(1332),
    [sym_atx_h3_marker] = ACTIONS(1332),
    [sym_atx_h4_marker] = ACTIONS(1332),
    [sym_atx_h5_marker] = ACTIONS(1332),
    [sym_atx_h6_marker] = ACTIONS(1332),
    [sym__thematic_break] = ACTIONS(1332),
    [sym__list_marker_minus] = ACTIONS(1332),
    [sym__list_marker_plus] = ACTIONS(1332),
    [sym__list_marker_star] = ACTIONS(1332),
    [sym__list_marker_parenthesis] = ACTIONS(1332),
    [sym__list_marker_dot] = ACTIONS(1332),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1332),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1332),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1332),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1332),
    [sym__blank_line_start] = ACTIONS(1332),
    [sym__html_block_1_start] = ACTIONS(1332),
    [sym__html_block_2_start] = ACTIONS(1332),
    [sym__html_block_3_start] = ACTIONS(1332),
    [sym__html_block_4_start] = ACTIONS(1332),
    [sym__html_block_5_start] = ACTIONS(1332),
    [sym__html_block_6_start] = ACTIONS(1332),
    [sym__html_block_7_start] = ACTIONS(1332),
    [sym__pipe_table_start] = ACTIONS(1332),
  },
  [STATE(244)] = {
    [ts_builtin_sym_end] = ACTIONS(1690),
    [anon_sym_LBRACK] = ACTIONS(1692),
    [anon_sym_RBRACK] = ACTIONS(1690),
    [anon_sym_LT] = ACTIONS(1690),
    [anon_sym_GT] = ACTIONS(1690),
    [anon_sym_BANG] = ACTIONS(1690),
    [anon_sym_DQUOTE] = ACTIONS(1690),
    [anon_sym_POUND] = ACTIONS(1690),
    [anon_sym_DOLLAR] = ACTIONS(1690),
    [anon_sym_PERCENT] = ACTIONS(1690),
    [anon_sym_AMP] = ACTIONS(1690),
    [anon_sym_SQUOTE] = ACTIONS(1690),
    [anon_sym_STAR] = ACTIONS(1690),
    [anon_sym_PLUS] = ACTIONS(1690),
    [anon_sym_COMMA] = ACTIONS(1690),
    [anon_sym_DASH] = ACTIONS(1690),
    [anon_sym_DOT] = ACTIONS(1690),
    [anon_sym_SLASH] = ACTIONS(1690),
    [anon_sym_COLON] = ACTIONS(1690),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_EQ] = ACTIONS(1690),
    [anon_sym_QMARK] = ACTIONS(1690),
    [anon_sym_AT] = ACTIONS(1690),
    [anon_sym_BSLASH] = ACTIONS(1690),
    [anon_sym_CARET] = ACTIONS(1690),
    [anon_sym__] = ACTIONS(1690),
    [anon_sym_BQUOTE] = ACTIONS(1690),
    [anon_sym_LBRACE] = ACTIONS(1690),
    [anon_sym_PIPE] = ACTIONS(1690),
    [anon_sym_RBRACE] = ACTIONS(1690),
    [anon_sym_TILDE] = ACTIONS(1690),
    [anon_sym_LPAREN] = ACTIONS(1690),
    [anon_sym_RPAREN] = ACTIONS(1690),
    [aux_sym__word_token1] = ACTIONS(1690),
    [aux_sym__word_token2] = ACTIONS(1690),
    [aux_sym__word_token3] = ACTIONS(1690),
    [sym__whitespace] = ACTIONS(1690),
    [sym__soft_line_ending] = ACTIONS(1690),
    [sym__block_quote_start] = ACTIONS(1690),
    [sym__indented_chunk_start] = ACTIONS(1690),
    [sym_atx_h1_marker] = ACTIONS(1690),
    [sym_atx_h2_marker] = ACTIONS(1690),
    [sym_atx_h3_marker] = ACTIONS(1690),
    [sym_atx_h4_marker] = ACTIONS(1690),
    [sym_atx_h5_marker] = ACTIONS(1690),
    [sym_atx_h6_marker] = ACTIONS(1690),
    [sym__thematic_break] = ACTIONS(1690),
    [sym__list_marker_minus] = ACTIONS(1690),
    [sym__list_marker_plus] = ACTIONS(1690),
    [sym__list_marker_star] = ACTIONS(1690),
    [sym__list_marker_parenthesis] = ACTIONS(1690),
    [sym__list_marker_dot] = ACTIONS(1690),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1690),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1690),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1690),
    [sym__blank_line_start] = ACTIONS(1690),
    [sym__html_block_1_start] = ACTIONS(1690),
    [sym__html_block_2_start] = ACTIONS(1690),
    [sym__html_block_3_start] = ACTIONS(1690),
    [sym__html_block_4_start] = ACTIONS(1690),
    [sym__html_block_5_start] = ACTIONS(1690),
    [sym__html_block_6_start] = ACTIONS(1690),
    [sym__html_block_7_start] = ACTIONS(1690),
    [sym__pipe_table_start] = ACTIONS(1690),
  },
  [STATE(245)] = {
    [ts_builtin_sym_end] = ACTIONS(1253),
    [anon_sym_LBRACK] = ACTIONS(1255),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_LT] = ACTIONS(1253),
    [anon_sym_GT] = ACTIONS(1253),
    [anon_sym_BANG] = ACTIONS(1253),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [anon_sym_POUND] = ACTIONS(1253),
    [anon_sym_DOLLAR] = ACTIONS(1253),
    [anon_sym_PERCENT] = ACTIONS(1253),
    [anon_sym_AMP] = ACTIONS(1253),
    [anon_sym_SQUOTE] = ACTIONS(1253),
    [anon_sym_STAR] = ACTIONS(1253),
    [anon_sym_PLUS] = ACTIONS(1253),
    [anon_sym_COMMA] = ACTIONS(1253),
    [anon_sym_DASH] = ACTIONS(1253),
    [anon_sym_DOT] = ACTIONS(1253),
    [anon_sym_SLASH] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1253),
    [anon_sym_SEMI] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1253),
    [anon_sym_AT] = ACTIONS(1253),
    [anon_sym_BSLASH] = ACTIONS(1253),
    [anon_sym_CARET] = ACTIONS(1253),
    [anon_sym__] = ACTIONS(1253),
    [anon_sym_BQUOTE] = ACTIONS(1253),
    [anon_sym_LBRACE] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1253),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_TILDE] = ACTIONS(1253),
    [anon_sym_LPAREN] = ACTIONS(1253),
    [anon_sym_RPAREN] = ACTIONS(1253),
    [aux_sym__word_token1] = ACTIONS(1253),
    [aux_sym__word_token2] = ACTIONS(1253),
    [aux_sym__word_token3] = ACTIONS(1253),
    [sym__whitespace] = ACTIONS(1253),
    [sym__soft_line_ending] = ACTIONS(1253),
    [sym__block_quote_start] = ACTIONS(1253),
    [sym__indented_chunk_start] = ACTIONS(1253),
    [sym_atx_h1_marker] = ACTIONS(1253),
    [sym_atx_h2_marker] = ACTIONS(1253),
    [sym_atx_h3_marker] = ACTIONS(1253),
    [sym_atx_h4_marker] = ACTIONS(1253),
    [sym_atx_h5_marker] = ACTIONS(1253),
    [sym_atx_h6_marker] = ACTIONS(1253),
    [sym__thematic_break] = ACTIONS(1253),
    [sym__list_marker_minus] = ACTIONS(1253),
    [sym__list_marker_plus] = ACTIONS(1253),
    [sym__list_marker_star] = ACTIONS(1253),
    [sym__list_marker_parenthesis] = ACTIONS(1253),
    [sym__list_marker_dot] = ACTIONS(1253),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1253),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1253),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1253),
    [sym__blank_line_start] = ACTIONS(1253),
    [sym__html_block_1_start] = ACTIONS(1253),
    [sym__html_block_2_start] = ACTIONS(1253),
    [sym__html_block_3_start] = ACTIONS(1253),
    [sym__html_block_4_start] = ACTIONS(1253),
    [sym__html_block_5_start] = ACTIONS(1253),
    [sym__html_block_6_start] = ACTIONS(1253),
    [sym__html_block_7_start] = ACTIONS(1253),
    [sym__pipe_table_start] = ACTIONS(1253),
  },
  [STATE(246)] = {
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_RBRACK] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_POUND] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1338),
    [anon_sym_PERCENT] = ACTIONS(1338),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_COLON] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1338),
    [anon_sym_EQ] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_BSLASH] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1338),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_BQUOTE] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1338),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_RBRACE] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1338),
    [anon_sym_RPAREN] = ACTIONS(1338),
    [aux_sym__word_token1] = ACTIONS(1338),
    [aux_sym__word_token2] = ACTIONS(1338),
    [aux_sym__word_token3] = ACTIONS(1338),
    [sym__whitespace] = ACTIONS(1338),
    [sym__soft_line_ending] = ACTIONS(1338),
    [sym__block_close] = ACTIONS(1338),
    [sym__block_quote_start] = ACTIONS(1338),
    [sym__indented_chunk_start] = ACTIONS(1338),
    [sym_atx_h1_marker] = ACTIONS(1338),
    [sym_atx_h2_marker] = ACTIONS(1338),
    [sym_atx_h3_marker] = ACTIONS(1338),
    [sym_atx_h4_marker] = ACTIONS(1338),
    [sym_atx_h5_marker] = ACTIONS(1338),
    [sym_atx_h6_marker] = ACTIONS(1338),
    [sym__thematic_break] = ACTIONS(1338),
    [sym__list_marker_minus] = ACTIONS(1338),
    [sym__list_marker_plus] = ACTIONS(1338),
    [sym__list_marker_star] = ACTIONS(1338),
    [sym__list_marker_parenthesis] = ACTIONS(1338),
    [sym__list_marker_dot] = ACTIONS(1338),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1338),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1338),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1338),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1338),
    [sym__blank_line_start] = ACTIONS(1338),
    [sym__html_block_1_start] = ACTIONS(1338),
    [sym__html_block_2_start] = ACTIONS(1338),
    [sym__html_block_3_start] = ACTIONS(1338),
    [sym__html_block_4_start] = ACTIONS(1338),
    [sym__html_block_5_start] = ACTIONS(1338),
    [sym__html_block_6_start] = ACTIONS(1338),
    [sym__html_block_7_start] = ACTIONS(1338),
    [sym__pipe_table_start] = ACTIONS(1338),
  },
  [STATE(247)] = {
    [ts_builtin_sym_end] = ACTIONS(1694),
    [anon_sym_LBRACK] = ACTIONS(1696),
    [anon_sym_RBRACK] = ACTIONS(1694),
    [anon_sym_LT] = ACTIONS(1694),
    [anon_sym_GT] = ACTIONS(1694),
    [anon_sym_BANG] = ACTIONS(1694),
    [anon_sym_DQUOTE] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_DOLLAR] = ACTIONS(1694),
    [anon_sym_PERCENT] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1694),
    [anon_sym_STAR] = ACTIONS(1694),
    [anon_sym_PLUS] = ACTIONS(1694),
    [anon_sym_COMMA] = ACTIONS(1694),
    [anon_sym_DASH] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1694),
    [anon_sym_SLASH] = ACTIONS(1694),
    [anon_sym_COLON] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1694),
    [anon_sym_EQ] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_AT] = ACTIONS(1694),
    [anon_sym_BSLASH] = ACTIONS(1694),
    [anon_sym_CARET] = ACTIONS(1694),
    [anon_sym__] = ACTIONS(1694),
    [anon_sym_BQUOTE] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_RBRACE] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1694),
    [anon_sym_RPAREN] = ACTIONS(1694),
    [aux_sym__word_token1] = ACTIONS(1694),
    [aux_sym__word_token2] = ACTIONS(1694),
    [aux_sym__word_token3] = ACTIONS(1694),
    [sym__whitespace] = ACTIONS(1694),
    [sym__soft_line_ending] = ACTIONS(1694),
    [sym__block_quote_start] = ACTIONS(1694),
    [sym__indented_chunk_start] = ACTIONS(1694),
    [sym_atx_h1_marker] = ACTIONS(1694),
    [sym_atx_h2_marker] = ACTIONS(1694),
    [sym_atx_h3_marker] = ACTIONS(1694),
    [sym_atx_h4_marker] = ACTIONS(1694),
    [sym_atx_h5_marker] = ACTIONS(1694),
    [sym_atx_h6_marker] = ACTIONS(1694),
    [sym__thematic_break] = ACTIONS(1694),
    [sym__list_marker_minus] = ACTIONS(1694),
    [sym__list_marker_plus] = ACTIONS(1694),
    [sym__list_marker_star] = ACTIONS(1694),
    [sym__list_marker_parenthesis] = ACTIONS(1694),
    [sym__list_marker_dot] = ACTIONS(1694),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1694),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1694),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1694),
    [sym__blank_line_start] = ACTIONS(1694),
    [sym__html_block_1_start] = ACTIONS(1694),
    [sym__html_block_2_start] = ACTIONS(1694),
    [sym__html_block_3_start] = ACTIONS(1694),
    [sym__html_block_4_start] = ACTIONS(1694),
    [sym__html_block_5_start] = ACTIONS(1694),
    [sym__html_block_6_start] = ACTIONS(1694),
    [sym__html_block_7_start] = ACTIONS(1694),
    [sym__pipe_table_start] = ACTIONS(1694),
  },
  [STATE(248)] = {
    [anon_sym_LBRACK] = ACTIONS(1596),
    [anon_sym_RBRACK] = ACTIONS(1594),
    [anon_sym_LT] = ACTIONS(1594),
    [anon_sym_GT] = ACTIONS(1594),
    [anon_sym_BANG] = ACTIONS(1594),
    [anon_sym_DQUOTE] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_DOLLAR] = ACTIONS(1594),
    [anon_sym_PERCENT] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1594),
    [anon_sym_STAR] = ACTIONS(1594),
    [anon_sym_PLUS] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1594),
    [anon_sym_SLASH] = ACTIONS(1594),
    [anon_sym_COLON] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_EQ] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_AT] = ACTIONS(1594),
    [anon_sym_BSLASH] = ACTIONS(1594),
    [anon_sym_CARET] = ACTIONS(1594),
    [anon_sym__] = ACTIONS(1594),
    [anon_sym_BQUOTE] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_RBRACE] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1594),
    [anon_sym_RPAREN] = ACTIONS(1594),
    [aux_sym__word_token1] = ACTIONS(1594),
    [aux_sym__word_token2] = ACTIONS(1594),
    [aux_sym__word_token3] = ACTIONS(1594),
    [sym__whitespace] = ACTIONS(1594),
    [sym__soft_line_ending] = ACTIONS(1594),
    [sym__block_close] = ACTIONS(1594),
    [sym__block_quote_start] = ACTIONS(1594),
    [sym__indented_chunk_start] = ACTIONS(1594),
    [sym_atx_h1_marker] = ACTIONS(1594),
    [sym_atx_h2_marker] = ACTIONS(1594),
    [sym_atx_h3_marker] = ACTIONS(1594),
    [sym_atx_h4_marker] = ACTIONS(1594),
    [sym_atx_h5_marker] = ACTIONS(1594),
    [sym_atx_h6_marker] = ACTIONS(1594),
    [sym__thematic_break] = ACTIONS(1594),
    [sym__list_marker_minus] = ACTIONS(1594),
    [sym__list_marker_plus] = ACTIONS(1594),
    [sym__list_marker_star] = ACTIONS(1594),
    [sym__list_marker_parenthesis] = ACTIONS(1594),
    [sym__list_marker_dot] = ACTIONS(1594),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1594),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1594),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1594),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1594),
    [sym__blank_line_start] = ACTIONS(1594),
    [sym__html_block_1_start] = ACTIONS(1594),
    [sym__html_block_2_start] = ACTIONS(1594),
    [sym__html_block_3_start] = ACTIONS(1594),
    [sym__html_block_4_start] = ACTIONS(1594),
    [sym__html_block_5_start] = ACTIONS(1594),
    [sym__html_block_6_start] = ACTIONS(1594),
    [sym__html_block_7_start] = ACTIONS(1594),
    [sym__pipe_table_start] = ACTIONS(1594),
  },
  [STATE(249)] = {
    [anon_sym_LBRACK] = ACTIONS(1604),
    [anon_sym_RBRACK] = ACTIONS(1602),
    [anon_sym_LT] = ACTIONS(1602),
    [anon_sym_GT] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_DOLLAR] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1602),
    [anon_sym_STAR] = ACTIONS(1602),
    [anon_sym_PLUS] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1602),
    [anon_sym_DASH] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1602),
    [anon_sym_SLASH] = ACTIONS(1602),
    [anon_sym_COLON] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_EQ] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_AT] = ACTIONS(1602),
    [anon_sym_BSLASH] = ACTIONS(1602),
    [anon_sym_CARET] = ACTIONS(1602),
    [anon_sym__] = ACTIONS(1602),
    [anon_sym_BQUOTE] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_RPAREN] = ACTIONS(1602),
    [aux_sym__word_token1] = ACTIONS(1602),
    [aux_sym__word_token2] = ACTIONS(1602),
    [aux_sym__word_token3] = ACTIONS(1602),
    [sym__whitespace] = ACTIONS(1602),
    [sym__soft_line_ending] = ACTIONS(1602),
    [sym__block_close] = ACTIONS(1602),
    [sym__block_quote_start] = ACTIONS(1602),
    [sym__indented_chunk_start] = ACTIONS(1602),
    [sym_atx_h1_marker] = ACTIONS(1602),
    [sym_atx_h2_marker] = ACTIONS(1602),
    [sym_atx_h3_marker] = ACTIONS(1602),
    [sym_atx_h4_marker] = ACTIONS(1602),
    [sym_atx_h5_marker] = ACTIONS(1602),
    [sym_atx_h6_marker] = ACTIONS(1602),
    [sym__thematic_break] = ACTIONS(1602),
    [sym__list_marker_minus] = ACTIONS(1602),
    [sym__list_marker_plus] = ACTIONS(1602),
    [sym__list_marker_star] = ACTIONS(1602),
    [sym__list_marker_parenthesis] = ACTIONS(1602),
    [sym__list_marker_dot] = ACTIONS(1602),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1602),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1602),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1602),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1602),
    [sym__blank_line_start] = ACTIONS(1602),
    [sym__html_block_1_start] = ACTIONS(1602),
    [sym__html_block_2_start] = ACTIONS(1602),
    [sym__html_block_3_start] = ACTIONS(1602),
    [sym__html_block_4_start] = ACTIONS(1602),
    [sym__html_block_5_start] = ACTIONS(1602),
    [sym__html_block_6_start] = ACTIONS(1602),
    [sym__html_block_7_start] = ACTIONS(1602),
    [sym__pipe_table_start] = ACTIONS(1602),
  },
  [STATE(250)] = {
    [anon_sym_LBRACK] = ACTIONS(1612),
    [anon_sym_RBRACK] = ACTIONS(1610),
    [anon_sym_LT] = ACTIONS(1610),
    [anon_sym_GT] = ACTIONS(1610),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_DOLLAR] = ACTIONS(1610),
    [anon_sym_PERCENT] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_STAR] = ACTIONS(1610),
    [anon_sym_PLUS] = ACTIONS(1610),
    [anon_sym_COMMA] = ACTIONS(1610),
    [anon_sym_DASH] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1610),
    [anon_sym_SLASH] = ACTIONS(1610),
    [anon_sym_COLON] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1610),
    [anon_sym_EQ] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_BSLASH] = ACTIONS(1610),
    [anon_sym_CARET] = ACTIONS(1610),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_BQUOTE] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_RBRACE] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1610),
    [anon_sym_RPAREN] = ACTIONS(1610),
    [aux_sym__word_token1] = ACTIONS(1610),
    [aux_sym__word_token2] = ACTIONS(1610),
    [aux_sym__word_token3] = ACTIONS(1610),
    [sym__whitespace] = ACTIONS(1610),
    [sym__soft_line_ending] = ACTIONS(1610),
    [sym__block_close] = ACTIONS(1610),
    [sym__block_quote_start] = ACTIONS(1610),
    [sym__indented_chunk_start] = ACTIONS(1610),
    [sym_atx_h1_marker] = ACTIONS(1610),
    [sym_atx_h2_marker] = ACTIONS(1610),
    [sym_atx_h3_marker] = ACTIONS(1610),
    [sym_atx_h4_marker] = ACTIONS(1610),
    [sym_atx_h5_marker] = ACTIONS(1610),
    [sym_atx_h6_marker] = ACTIONS(1610),
    [sym__thematic_break] = ACTIONS(1610),
    [sym__list_marker_minus] = ACTIONS(1610),
    [sym__list_marker_plus] = ACTIONS(1610),
    [sym__list_marker_star] = ACTIONS(1610),
    [sym__list_marker_parenthesis] = ACTIONS(1610),
    [sym__list_marker_dot] = ACTIONS(1610),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1610),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1610),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1610),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1610),
    [sym__blank_line_start] = ACTIONS(1610),
    [sym__html_block_1_start] = ACTIONS(1610),
    [sym__html_block_2_start] = ACTIONS(1610),
    [sym__html_block_3_start] = ACTIONS(1610),
    [sym__html_block_4_start] = ACTIONS(1610),
    [sym__html_block_5_start] = ACTIONS(1610),
    [sym__html_block_6_start] = ACTIONS(1610),
    [sym__html_block_7_start] = ACTIONS(1610),
    [sym__pipe_table_start] = ACTIONS(1610),
  },
  [STATE(251)] = {
    [anon_sym_LBRACK] = ACTIONS(1620),
    [anon_sym_RBRACK] = ACTIONS(1618),
    [anon_sym_LT] = ACTIONS(1618),
    [anon_sym_GT] = ACTIONS(1618),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_DOLLAR] = ACTIONS(1618),
    [anon_sym_PERCENT] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1618),
    [anon_sym_PLUS] = ACTIONS(1618),
    [anon_sym_COMMA] = ACTIONS(1618),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1618),
    [anon_sym_SLASH] = ACTIONS(1618),
    [anon_sym_COLON] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1618),
    [anon_sym_EQ] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_BSLASH] = ACTIONS(1618),
    [anon_sym_CARET] = ACTIONS(1618),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_BQUOTE] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_RBRACE] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1618),
    [anon_sym_RPAREN] = ACTIONS(1618),
    [aux_sym__word_token1] = ACTIONS(1618),
    [aux_sym__word_token2] = ACTIONS(1618),
    [aux_sym__word_token3] = ACTIONS(1618),
    [sym__whitespace] = ACTIONS(1618),
    [sym__soft_line_ending] = ACTIONS(1618),
    [sym__block_close] = ACTIONS(1618),
    [sym__block_quote_start] = ACTIONS(1618),
    [sym__indented_chunk_start] = ACTIONS(1618),
    [sym_atx_h1_marker] = ACTIONS(1618),
    [sym_atx_h2_marker] = ACTIONS(1618),
    [sym_atx_h3_marker] = ACTIONS(1618),
    [sym_atx_h4_marker] = ACTIONS(1618),
    [sym_atx_h5_marker] = ACTIONS(1618),
    [sym_atx_h6_marker] = ACTIONS(1618),
    [sym__thematic_break] = ACTIONS(1618),
    [sym__list_marker_minus] = ACTIONS(1618),
    [sym__list_marker_plus] = ACTIONS(1618),
    [sym__list_marker_star] = ACTIONS(1618),
    [sym__list_marker_parenthesis] = ACTIONS(1618),
    [sym__list_marker_dot] = ACTIONS(1618),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1618),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1618),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1618),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1618),
    [sym__blank_line_start] = ACTIONS(1618),
    [sym__html_block_1_start] = ACTIONS(1618),
    [sym__html_block_2_start] = ACTIONS(1618),
    [sym__html_block_3_start] = ACTIONS(1618),
    [sym__html_block_4_start] = ACTIONS(1618),
    [sym__html_block_5_start] = ACTIONS(1618),
    [sym__html_block_6_start] = ACTIONS(1618),
    [sym__html_block_7_start] = ACTIONS(1618),
    [sym__pipe_table_start] = ACTIONS(1618),
  },
  [STATE(252)] = {
    [anon_sym_LBRACK] = ACTIONS(1624),
    [anon_sym_RBRACK] = ACTIONS(1622),
    [anon_sym_LT] = ACTIONS(1622),
    [anon_sym_GT] = ACTIONS(1622),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_DOLLAR] = ACTIONS(1622),
    [anon_sym_PERCENT] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_STAR] = ACTIONS(1622),
    [anon_sym_PLUS] = ACTIONS(1622),
    [anon_sym_COMMA] = ACTIONS(1622),
    [anon_sym_DASH] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1622),
    [anon_sym_SLASH] = ACTIONS(1622),
    [anon_sym_COLON] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym_EQ] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_BSLASH] = ACTIONS(1622),
    [anon_sym_CARET] = ACTIONS(1622),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_BQUOTE] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_RBRACE] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1622),
    [anon_sym_RPAREN] = ACTIONS(1622),
    [aux_sym__word_token1] = ACTIONS(1622),
    [aux_sym__word_token2] = ACTIONS(1622),
    [aux_sym__word_token3] = ACTIONS(1622),
    [sym__whitespace] = ACTIONS(1622),
    [sym__soft_line_ending] = ACTIONS(1622),
    [sym__block_close] = ACTIONS(1622),
    [sym__block_quote_start] = ACTIONS(1622),
    [sym__indented_chunk_start] = ACTIONS(1622),
    [sym_atx_h1_marker] = ACTIONS(1622),
    [sym_atx_h2_marker] = ACTIONS(1622),
    [sym_atx_h3_marker] = ACTIONS(1622),
    [sym_atx_h4_marker] = ACTIONS(1622),
    [sym_atx_h5_marker] = ACTIONS(1622),
    [sym_atx_h6_marker] = ACTIONS(1622),
    [sym__thematic_break] = ACTIONS(1622),
    [sym__list_marker_minus] = ACTIONS(1622),
    [sym__list_marker_plus] = ACTIONS(1622),
    [sym__list_marker_star] = ACTIONS(1622),
    [sym__list_marker_parenthesis] = ACTIONS(1622),
    [sym__list_marker_dot] = ACTIONS(1622),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1622),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1622),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1622),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1622),
    [sym__blank_line_start] = ACTIONS(1622),
    [sym__html_block_1_start] = ACTIONS(1622),
    [sym__html_block_2_start] = ACTIONS(1622),
    [sym__html_block_3_start] = ACTIONS(1622),
    [sym__html_block_4_start] = ACTIONS(1622),
    [sym__html_block_5_start] = ACTIONS(1622),
    [sym__html_block_6_start] = ACTIONS(1622),
    [sym__html_block_7_start] = ACTIONS(1622),
    [sym__pipe_table_start] = ACTIONS(1622),
  },
  [STATE(253)] = {
    [anon_sym_LBRACK] = ACTIONS(1628),
    [anon_sym_RBRACK] = ACTIONS(1626),
    [anon_sym_LT] = ACTIONS(1626),
    [anon_sym_GT] = ACTIONS(1626),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_DOLLAR] = ACTIONS(1626),
    [anon_sym_PERCENT] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_STAR] = ACTIONS(1626),
    [anon_sym_PLUS] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1626),
    [anon_sym_DASH] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1626),
    [anon_sym_SLASH] = ACTIONS(1626),
    [anon_sym_COLON] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_EQ] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_BSLASH] = ACTIONS(1626),
    [anon_sym_CARET] = ACTIONS(1626),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_BQUOTE] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_RBRACE] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1626),
    [anon_sym_RPAREN] = ACTIONS(1626),
    [aux_sym__word_token1] = ACTIONS(1626),
    [aux_sym__word_token2] = ACTIONS(1626),
    [aux_sym__word_token3] = ACTIONS(1626),
    [sym__whitespace] = ACTIONS(1626),
    [sym__soft_line_ending] = ACTIONS(1626),
    [sym__block_close] = ACTIONS(1626),
    [sym__block_quote_start] = ACTIONS(1626),
    [sym__indented_chunk_start] = ACTIONS(1626),
    [sym_atx_h1_marker] = ACTIONS(1626),
    [sym_atx_h2_marker] = ACTIONS(1626),
    [sym_atx_h3_marker] = ACTIONS(1626),
    [sym_atx_h4_marker] = ACTIONS(1626),
    [sym_atx_h5_marker] = ACTIONS(1626),
    [sym_atx_h6_marker] = ACTIONS(1626),
    [sym__thematic_break] = ACTIONS(1626),
    [sym__list_marker_minus] = ACTIONS(1626),
    [sym__list_marker_plus] = ACTIONS(1626),
    [sym__list_marker_star] = ACTIONS(1626),
    [sym__list_marker_parenthesis] = ACTIONS(1626),
    [sym__list_marker_dot] = ACTIONS(1626),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1626),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1626),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1626),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1626),
    [sym__blank_line_start] = ACTIONS(1626),
    [sym__html_block_1_start] = ACTIONS(1626),
    [sym__html_block_2_start] = ACTIONS(1626),
    [sym__html_block_3_start] = ACTIONS(1626),
    [sym__html_block_4_start] = ACTIONS(1626),
    [sym__html_block_5_start] = ACTIONS(1626),
    [sym__html_block_6_start] = ACTIONS(1626),
    [sym__html_block_7_start] = ACTIONS(1626),
    [sym__pipe_table_start] = ACTIONS(1626),
  },
  [STATE(254)] = {
    [anon_sym_LBRACK] = ACTIONS(1632),
    [anon_sym_RBRACK] = ACTIONS(1630),
    [anon_sym_LT] = ACTIONS(1630),
    [anon_sym_GT] = ACTIONS(1630),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_DOLLAR] = ACTIONS(1630),
    [anon_sym_PERCENT] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_STAR] = ACTIONS(1630),
    [anon_sym_PLUS] = ACTIONS(1630),
    [anon_sym_COMMA] = ACTIONS(1630),
    [anon_sym_DASH] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1630),
    [anon_sym_SLASH] = ACTIONS(1630),
    [anon_sym_COLON] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1630),
    [anon_sym_EQ] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_BSLASH] = ACTIONS(1630),
    [anon_sym_CARET] = ACTIONS(1630),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_BQUOTE] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_RBRACE] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1630),
    [anon_sym_RPAREN] = ACTIONS(1630),
    [aux_sym__word_token1] = ACTIONS(1630),
    [aux_sym__word_token2] = ACTIONS(1630),
    [aux_sym__word_token3] = ACTIONS(1630),
    [sym__whitespace] = ACTIONS(1630),
    [sym__soft_line_ending] = ACTIONS(1630),
    [sym__block_close] = ACTIONS(1630),
    [sym__block_quote_start] = ACTIONS(1630),
    [sym__indented_chunk_start] = ACTIONS(1630),
    [sym_atx_h1_marker] = ACTIONS(1630),
    [sym_atx_h2_marker] = ACTIONS(1630),
    [sym_atx_h3_marker] = ACTIONS(1630),
    [sym_atx_h4_marker] = ACTIONS(1630),
    [sym_atx_h5_marker] = ACTIONS(1630),
    [sym_atx_h6_marker] = ACTIONS(1630),
    [sym__thematic_break] = ACTIONS(1630),
    [sym__list_marker_minus] = ACTIONS(1630),
    [sym__list_marker_plus] = ACTIONS(1630),
    [sym__list_marker_star] = ACTIONS(1630),
    [sym__list_marker_parenthesis] = ACTIONS(1630),
    [sym__list_marker_dot] = ACTIONS(1630),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1630),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1630),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1630),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1630),
    [sym__blank_line_start] = ACTIONS(1630),
    [sym__html_block_1_start] = ACTIONS(1630),
    [sym__html_block_2_start] = ACTIONS(1630),
    [sym__html_block_3_start] = ACTIONS(1630),
    [sym__html_block_4_start] = ACTIONS(1630),
    [sym__html_block_5_start] = ACTIONS(1630),
    [sym__html_block_6_start] = ACTIONS(1630),
    [sym__html_block_7_start] = ACTIONS(1630),
    [sym__pipe_table_start] = ACTIONS(1630),
  },
  [STATE(255)] = {
    [anon_sym_LBRACK] = ACTIONS(1342),
    [anon_sym_RBRACK] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_DQUOTE] = ACTIONS(1344),
    [anon_sym_POUND] = ACTIONS(1344),
    [anon_sym_DOLLAR] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_SQUOTE] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_COLON] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_EQ] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_BSLASH] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym__] = ACTIONS(1344),
    [anon_sym_BQUOTE] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_RPAREN] = ACTIONS(1344),
    [aux_sym__word_token1] = ACTIONS(1344),
    [aux_sym__word_token2] = ACTIONS(1344),
    [aux_sym__word_token3] = ACTIONS(1344),
    [sym__whitespace] = ACTIONS(1344),
    [sym__soft_line_ending] = ACTIONS(1344),
    [sym__block_close] = ACTIONS(1344),
    [sym__block_quote_start] = ACTIONS(1344),
    [sym__indented_chunk_start] = ACTIONS(1344),
    [sym_atx_h1_marker] = ACTIONS(1344),
    [sym_atx_h2_marker] = ACTIONS(1344),
    [sym_atx_h3_marker] = ACTIONS(1344),
    [sym_atx_h4_marker] = ACTIONS(1344),
    [sym_atx_h5_marker] = ACTIONS(1344),
    [sym_atx_h6_marker] = ACTIONS(1344),
    [sym__thematic_break] = ACTIONS(1344),
    [sym__list_marker_minus] = ACTIONS(1344),
    [sym__list_marker_plus] = ACTIONS(1344),
    [sym__list_marker_star] = ACTIONS(1344),
    [sym__list_marker_parenthesis] = ACTIONS(1344),
    [sym__list_marker_dot] = ACTIONS(1344),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1344),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1344),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1344),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1344),
    [sym__blank_line_start] = ACTIONS(1344),
    [sym__html_block_1_start] = ACTIONS(1344),
    [sym__html_block_2_start] = ACTIONS(1344),
    [sym__html_block_3_start] = ACTIONS(1344),
    [sym__html_block_4_start] = ACTIONS(1344),
    [sym__html_block_5_start] = ACTIONS(1344),
    [sym__html_block_6_start] = ACTIONS(1344),
    [sym__html_block_7_start] = ACTIONS(1344),
    [sym__pipe_table_start] = ACTIONS(1344),
  },
  [STATE(256)] = {
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_RBRACK] = ACTIONS(1350),
    [anon_sym_LT] = ACTIONS(1350),
    [anon_sym_GT] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [anon_sym_POUND] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_PLUS] = ACTIONS(1350),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_EQ] = ACTIONS(1350),
    [anon_sym_QMARK] = ACTIONS(1350),
    [anon_sym_AT] = ACTIONS(1350),
    [anon_sym_BSLASH] = ACTIONS(1350),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym__] = ACTIONS(1350),
    [anon_sym_BQUOTE] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_RPAREN] = ACTIONS(1350),
    [aux_sym__word_token1] = ACTIONS(1350),
    [aux_sym__word_token2] = ACTIONS(1350),
    [aux_sym__word_token3] = ACTIONS(1350),
    [sym__whitespace] = ACTIONS(1350),
    [sym__soft_line_ending] = ACTIONS(1350),
    [sym__block_close] = ACTIONS(1350),
    [sym__block_quote_start] = ACTIONS(1350),
    [sym__indented_chunk_start] = ACTIONS(1350),
    [sym_atx_h1_marker] = ACTIONS(1350),
    [sym_atx_h2_marker] = ACTIONS(1350),
    [sym_atx_h3_marker] = ACTIONS(1350),
    [sym_atx_h4_marker] = ACTIONS(1350),
    [sym_atx_h5_marker] = ACTIONS(1350),
    [sym_atx_h6_marker] = ACTIONS(1350),
    [sym__thematic_break] = ACTIONS(1350),
    [sym__list_marker_minus] = ACTIONS(1350),
    [sym__list_marker_plus] = ACTIONS(1350),
    [sym__list_marker_star] = ACTIONS(1350),
    [sym__list_marker_parenthesis] = ACTIONS(1350),
    [sym__list_marker_dot] = ACTIONS(1350),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1350),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1350),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1350),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1350),
    [sym__blank_line_start] = ACTIONS(1350),
    [sym__html_block_1_start] = ACTIONS(1350),
    [sym__html_block_2_start] = ACTIONS(1350),
    [sym__html_block_3_start] = ACTIONS(1350),
    [sym__html_block_4_start] = ACTIONS(1350),
    [sym__html_block_5_start] = ACTIONS(1350),
    [sym__html_block_6_start] = ACTIONS(1350),
    [sym__html_block_7_start] = ACTIONS(1350),
    [sym__pipe_table_start] = ACTIONS(1350),
  },
  [STATE(257)] = {
    [ts_builtin_sym_end] = ACTIONS(1616),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_LT] = ACTIONS(1616),
    [anon_sym_GT] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [anon_sym_POUND] = ACTIONS(1616),
    [anon_sym_DOLLAR] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1616),
    [anon_sym_AMP] = ACTIONS(1616),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_STAR] = ACTIONS(1616),
    [anon_sym_PLUS] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_DASH] = ACTIONS(1616),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_SLASH] = ACTIONS(1616),
    [anon_sym_COLON] = ACTIONS(1616),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_EQ] = ACTIONS(1616),
    [anon_sym_QMARK] = ACTIONS(1616),
    [anon_sym_AT] = ACTIONS(1616),
    [anon_sym_BSLASH] = ACTIONS(1616),
    [anon_sym_CARET] = ACTIONS(1616),
    [anon_sym__] = ACTIONS(1616),
    [anon_sym_BQUOTE] = ACTIONS(1616),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [aux_sym__word_token1] = ACTIONS(1616),
    [aux_sym__word_token2] = ACTIONS(1616),
    [aux_sym__word_token3] = ACTIONS(1616),
    [sym__whitespace] = ACTIONS(1616),
    [sym__soft_line_ending] = ACTIONS(1616),
    [sym__block_quote_start] = ACTIONS(1616),
    [sym__indented_chunk_start] = ACTIONS(1616),
    [sym_atx_h1_marker] = ACTIONS(1616),
    [sym_atx_h2_marker] = ACTIONS(1616),
    [sym_atx_h3_marker] = ACTIONS(1616),
    [sym_atx_h4_marker] = ACTIONS(1616),
    [sym_atx_h5_marker] = ACTIONS(1616),
    [sym_atx_h6_marker] = ACTIONS(1616),
    [sym__thematic_break] = ACTIONS(1616),
    [sym__list_marker_minus] = ACTIONS(1616),
    [sym__list_marker_plus] = ACTIONS(1616),
    [sym__list_marker_star] = ACTIONS(1616),
    [sym__list_marker_parenthesis] = ACTIONS(1616),
    [sym__list_marker_dot] = ACTIONS(1616),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1616),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1616),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1616),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1616),
    [sym__blank_line_start] = ACTIONS(1616),
    [sym__html_block_1_start] = ACTIONS(1616),
    [sym__html_block_2_start] = ACTIONS(1616),
    [sym__html_block_3_start] = ACTIONS(1616),
    [sym__html_block_4_start] = ACTIONS(1616),
    [sym__html_block_5_start] = ACTIONS(1616),
    [sym__html_block_6_start] = ACTIONS(1616),
    [sym__html_block_7_start] = ACTIONS(1616),
    [sym__pipe_table_start] = ACTIONS(1616),
  },
  [STATE(258)] = {
    [anon_sym_LBRACK] = ACTIONS(1354),
    [anon_sym_RBRACK] = ACTIONS(1356),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_GT] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_DQUOTE] = ACTIONS(1356),
    [anon_sym_POUND] = ACTIONS(1356),
    [anon_sym_DOLLAR] = ACTIONS(1356),
    [anon_sym_PERCENT] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_SQUOTE] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_COMMA] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_DOT] = ACTIONS(1356),
    [anon_sym_SLASH] = ACTIONS(1356),
    [anon_sym_COLON] = ACTIONS(1356),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_EQ] = ACTIONS(1356),
    [anon_sym_QMARK] = ACTIONS(1356),
    [anon_sym_AT] = ACTIONS(1356),
    [anon_sym_BSLASH] = ACTIONS(1356),
    [anon_sym_CARET] = ACTIONS(1356),
    [anon_sym__] = ACTIONS(1356),
    [anon_sym_BQUOTE] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_TILDE] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_RPAREN] = ACTIONS(1356),
    [aux_sym__word_token1] = ACTIONS(1356),
    [aux_sym__word_token2] = ACTIONS(1356),
    [aux_sym__word_token3] = ACTIONS(1356),
    [sym__whitespace] = ACTIONS(1356),
    [sym__soft_line_ending] = ACTIONS(1356),
    [sym__block_close] = ACTIONS(1356),
    [sym__block_quote_start] = ACTIONS(1356),
    [sym__indented_chunk_start] = ACTIONS(1356),
    [sym_atx_h1_marker] = ACTIONS(1356),
    [sym_atx_h2_marker] = ACTIONS(1356),
    [sym_atx_h3_marker] = ACTIONS(1356),
    [sym_atx_h4_marker] = ACTIONS(1356),
    [sym_atx_h5_marker] = ACTIONS(1356),
    [sym_atx_h6_marker] = ACTIONS(1356),
    [sym__thematic_break] = ACTIONS(1356),
    [sym__list_marker_minus] = ACTIONS(1356),
    [sym__list_marker_plus] = ACTIONS(1356),
    [sym__list_marker_star] = ACTIONS(1356),
    [sym__list_marker_parenthesis] = ACTIONS(1356),
    [sym__list_marker_dot] = ACTIONS(1356),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1356),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1356),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1356),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1356),
    [sym__blank_line_start] = ACTIONS(1356),
    [sym__html_block_1_start] = ACTIONS(1356),
    [sym__html_block_2_start] = ACTIONS(1356),
    [sym__html_block_3_start] = ACTIONS(1356),
    [sym__html_block_4_start] = ACTIONS(1356),
    [sym__html_block_5_start] = ACTIONS(1356),
    [sym__html_block_6_start] = ACTIONS(1356),
    [sym__html_block_7_start] = ACTIONS(1356),
    [sym__pipe_table_start] = ACTIONS(1356),
  },
  [STATE(259)] = {
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_RBRACK] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_POUND] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [anon_sym_PERCENT] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_COLON] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym_EQ] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_BSLASH] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1362),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1362),
    [anon_sym_RPAREN] = ACTIONS(1362),
    [aux_sym__word_token1] = ACTIONS(1362),
    [aux_sym__word_token2] = ACTIONS(1362),
    [aux_sym__word_token3] = ACTIONS(1362),
    [sym__whitespace] = ACTIONS(1362),
    [sym__soft_line_ending] = ACTIONS(1362),
    [sym__block_close] = ACTIONS(1362),
    [sym__block_quote_start] = ACTIONS(1362),
    [sym__indented_chunk_start] = ACTIONS(1362),
    [sym_atx_h1_marker] = ACTIONS(1362),
    [sym_atx_h2_marker] = ACTIONS(1362),
    [sym_atx_h3_marker] = ACTIONS(1362),
    [sym_atx_h4_marker] = ACTIONS(1362),
    [sym_atx_h5_marker] = ACTIONS(1362),
    [sym_atx_h6_marker] = ACTIONS(1362),
    [sym__thematic_break] = ACTIONS(1362),
    [sym__list_marker_minus] = ACTIONS(1362),
    [sym__list_marker_plus] = ACTIONS(1362),
    [sym__list_marker_star] = ACTIONS(1362),
    [sym__list_marker_parenthesis] = ACTIONS(1362),
    [sym__list_marker_dot] = ACTIONS(1362),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1362),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1362),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1362),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1362),
    [sym__blank_line_start] = ACTIONS(1362),
    [sym__html_block_1_start] = ACTIONS(1362),
    [sym__html_block_2_start] = ACTIONS(1362),
    [sym__html_block_3_start] = ACTIONS(1362),
    [sym__html_block_4_start] = ACTIONS(1362),
    [sym__html_block_5_start] = ACTIONS(1362),
    [sym__html_block_6_start] = ACTIONS(1362),
    [sym__html_block_7_start] = ACTIONS(1362),
    [sym__pipe_table_start] = ACTIONS(1362),
  },
  [STATE(260)] = {
    [ts_builtin_sym_end] = ACTIONS(1636),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_LT] = ACTIONS(1636),
    [anon_sym_GT] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [anon_sym_POUND] = ACTIONS(1636),
    [anon_sym_DOLLAR] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1636),
    [anon_sym_AMP] = ACTIONS(1636),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_STAR] = ACTIONS(1636),
    [anon_sym_PLUS] = ACTIONS(1636),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_DASH] = ACTIONS(1636),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_SLASH] = ACTIONS(1636),
    [anon_sym_COLON] = ACTIONS(1636),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_EQ] = ACTIONS(1636),
    [anon_sym_QMARK] = ACTIONS(1636),
    [anon_sym_AT] = ACTIONS(1636),
    [anon_sym_BSLASH] = ACTIONS(1636),
    [anon_sym_CARET] = ACTIONS(1636),
    [anon_sym__] = ACTIONS(1636),
    [anon_sym_BQUOTE] = ACTIONS(1636),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_RPAREN] = ACTIONS(1636),
    [aux_sym__word_token1] = ACTIONS(1636),
    [aux_sym__word_token2] = ACTIONS(1636),
    [aux_sym__word_token3] = ACTIONS(1636),
    [sym__whitespace] = ACTIONS(1636),
    [sym__soft_line_ending] = ACTIONS(1636),
    [sym__block_quote_start] = ACTIONS(1636),
    [sym__indented_chunk_start] = ACTIONS(1636),
    [sym_atx_h1_marker] = ACTIONS(1636),
    [sym_atx_h2_marker] = ACTIONS(1636),
    [sym_atx_h3_marker] = ACTIONS(1636),
    [sym_atx_h4_marker] = ACTIONS(1636),
    [sym_atx_h5_marker] = ACTIONS(1636),
    [sym_atx_h6_marker] = ACTIONS(1636),
    [sym__thematic_break] = ACTIONS(1636),
    [sym__list_marker_minus] = ACTIONS(1636),
    [sym__list_marker_plus] = ACTIONS(1636),
    [sym__list_marker_star] = ACTIONS(1636),
    [sym__list_marker_parenthesis] = ACTIONS(1636),
    [sym__list_marker_dot] = ACTIONS(1636),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1636),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1636),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1636),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1636),
    [sym__blank_line_start] = ACTIONS(1636),
    [sym__html_block_1_start] = ACTIONS(1636),
    [sym__html_block_2_start] = ACTIONS(1636),
    [sym__html_block_3_start] = ACTIONS(1636),
    [sym__html_block_4_start] = ACTIONS(1636),
    [sym__html_block_5_start] = ACTIONS(1636),
    [sym__html_block_6_start] = ACTIONS(1636),
    [sym__html_block_7_start] = ACTIONS(1636),
    [sym__pipe_table_start] = ACTIONS(1636),
  },
  [STATE(261)] = {
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_RBRACK] = ACTIONS(1290),
    [anon_sym_LT] = ACTIONS(1290),
    [anon_sym_GT] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [anon_sym_DQUOTE] = ACTIONS(1290),
    [anon_sym_POUND] = ACTIONS(1290),
    [anon_sym_DOLLAR] = ACTIONS(1290),
    [anon_sym_PERCENT] = ACTIONS(1290),
    [anon_sym_AMP] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_COMMA] = ACTIONS(1290),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_DOT] = ACTIONS(1290),
    [anon_sym_SLASH] = ACTIONS(1290),
    [anon_sym_COLON] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [anon_sym_EQ] = ACTIONS(1290),
    [anon_sym_QMARK] = ACTIONS(1290),
    [anon_sym_AT] = ACTIONS(1290),
    [anon_sym_BSLASH] = ACTIONS(1290),
    [anon_sym_CARET] = ACTIONS(1290),
    [anon_sym__] = ACTIONS(1290),
    [anon_sym_BQUOTE] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_PIPE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(1290),
    [anon_sym_RPAREN] = ACTIONS(1290),
    [aux_sym__word_token1] = ACTIONS(1290),
    [aux_sym__word_token2] = ACTIONS(1290),
    [aux_sym__word_token3] = ACTIONS(1290),
    [sym__whitespace] = ACTIONS(1290),
    [sym__soft_line_ending] = ACTIONS(1290),
    [sym__block_close] = ACTIONS(1290),
    [sym__block_quote_start] = ACTIONS(1290),
    [sym__indented_chunk_start] = ACTIONS(1290),
    [sym_atx_h1_marker] = ACTIONS(1290),
    [sym_atx_h2_marker] = ACTIONS(1290),
    [sym_atx_h3_marker] = ACTIONS(1290),
    [sym_atx_h4_marker] = ACTIONS(1290),
    [sym_atx_h5_marker] = ACTIONS(1290),
    [sym_atx_h6_marker] = ACTIONS(1290),
    [sym__thematic_break] = ACTIONS(1290),
    [sym__list_marker_minus] = ACTIONS(1290),
    [sym__list_marker_plus] = ACTIONS(1290),
    [sym__list_marker_star] = ACTIONS(1290),
    [sym__list_marker_parenthesis] = ACTIONS(1290),
    [sym__list_marker_dot] = ACTIONS(1290),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1290),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1290),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1290),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1290),
    [sym__blank_line_start] = ACTIONS(1290),
    [sym__html_block_1_start] = ACTIONS(1290),
    [sym__html_block_2_start] = ACTIONS(1290),
    [sym__html_block_3_start] = ACTIONS(1290),
    [sym__html_block_4_start] = ACTIONS(1290),
    [sym__html_block_5_start] = ACTIONS(1290),
    [sym__html_block_6_start] = ACTIONS(1290),
    [sym__html_block_7_start] = ACTIONS(1290),
    [sym__pipe_table_start] = ACTIONS(1290),
  },
  [STATE(262)] = {
    [ts_builtin_sym_end] = ACTIONS(1640),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_LT] = ACTIONS(1640),
    [anon_sym_GT] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [anon_sym_POUND] = ACTIONS(1640),
    [anon_sym_DOLLAR] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1640),
    [anon_sym_AMP] = ACTIONS(1640),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_STAR] = ACTIONS(1640),
    [anon_sym_PLUS] = ACTIONS(1640),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_DASH] = ACTIONS(1640),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_SLASH] = ACTIONS(1640),
    [anon_sym_COLON] = ACTIONS(1640),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_EQ] = ACTIONS(1640),
    [anon_sym_QMARK] = ACTIONS(1640),
    [anon_sym_AT] = ACTIONS(1640),
    [anon_sym_BSLASH] = ACTIONS(1640),
    [anon_sym_CARET] = ACTIONS(1640),
    [anon_sym__] = ACTIONS(1640),
    [anon_sym_BQUOTE] = ACTIONS(1640),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_RPAREN] = ACTIONS(1640),
    [aux_sym__word_token1] = ACTIONS(1640),
    [aux_sym__word_token2] = ACTIONS(1640),
    [aux_sym__word_token3] = ACTIONS(1640),
    [sym__whitespace] = ACTIONS(1640),
    [sym__soft_line_ending] = ACTIONS(1640),
    [sym__block_quote_start] = ACTIONS(1640),
    [sym__indented_chunk_start] = ACTIONS(1640),
    [sym_atx_h1_marker] = ACTIONS(1640),
    [sym_atx_h2_marker] = ACTIONS(1640),
    [sym_atx_h3_marker] = ACTIONS(1640),
    [sym_atx_h4_marker] = ACTIONS(1640),
    [sym_atx_h5_marker] = ACTIONS(1640),
    [sym_atx_h6_marker] = ACTIONS(1640),
    [sym__thematic_break] = ACTIONS(1640),
    [sym__list_marker_minus] = ACTIONS(1640),
    [sym__list_marker_plus] = ACTIONS(1640),
    [sym__list_marker_star] = ACTIONS(1640),
    [sym__list_marker_parenthesis] = ACTIONS(1640),
    [sym__list_marker_dot] = ACTIONS(1640),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1640),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1640),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1640),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1640),
    [sym__blank_line_start] = ACTIONS(1640),
    [sym__html_block_1_start] = ACTIONS(1640),
    [sym__html_block_2_start] = ACTIONS(1640),
    [sym__html_block_3_start] = ACTIONS(1640),
    [sym__html_block_4_start] = ACTIONS(1640),
    [sym__html_block_5_start] = ACTIONS(1640),
    [sym__html_block_6_start] = ACTIONS(1640),
    [sym__html_block_7_start] = ACTIONS(1640),
    [sym__pipe_table_start] = ACTIONS(1640),
  },
  [STATE(263)] = {
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_RBRACK] = ACTIONS(1374),
    [anon_sym_LT] = ACTIONS(1374),
    [anon_sym_GT] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1374),
    [anon_sym_POUND] = ACTIONS(1374),
    [anon_sym_DOLLAR] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1374),
    [anon_sym_PLUS] = ACTIONS(1374),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_DASH] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1374),
    [anon_sym_COLON] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_EQ] = ACTIONS(1374),
    [anon_sym_QMARK] = ACTIONS(1374),
    [anon_sym_AT] = ACTIONS(1374),
    [anon_sym_BSLASH] = ACTIONS(1374),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym__] = ACTIONS(1374),
    [anon_sym_BQUOTE] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_RPAREN] = ACTIONS(1374),
    [aux_sym__word_token1] = ACTIONS(1374),
    [aux_sym__word_token2] = ACTIONS(1374),
    [aux_sym__word_token3] = ACTIONS(1374),
    [sym__whitespace] = ACTIONS(1374),
    [sym__soft_line_ending] = ACTIONS(1374),
    [sym__block_close] = ACTIONS(1374),
    [sym__block_quote_start] = ACTIONS(1374),
    [sym__indented_chunk_start] = ACTIONS(1374),
    [sym_atx_h1_marker] = ACTIONS(1374),
    [sym_atx_h2_marker] = ACTIONS(1374),
    [sym_atx_h3_marker] = ACTIONS(1374),
    [sym_atx_h4_marker] = ACTIONS(1374),
    [sym_atx_h5_marker] = ACTIONS(1374),
    [sym_atx_h6_marker] = ACTIONS(1374),
    [sym__thematic_break] = ACTIONS(1374),
    [sym__list_marker_minus] = ACTIONS(1374),
    [sym__list_marker_plus] = ACTIONS(1374),
    [sym__list_marker_star] = ACTIONS(1374),
    [sym__list_marker_parenthesis] = ACTIONS(1374),
    [sym__list_marker_dot] = ACTIONS(1374),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1374),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1374),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1374),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1374),
    [sym__blank_line_start] = ACTIONS(1374),
    [sym__html_block_1_start] = ACTIONS(1374),
    [sym__html_block_2_start] = ACTIONS(1374),
    [sym__html_block_3_start] = ACTIONS(1374),
    [sym__html_block_4_start] = ACTIONS(1374),
    [sym__html_block_5_start] = ACTIONS(1374),
    [sym__html_block_6_start] = ACTIONS(1374),
    [sym__html_block_7_start] = ACTIONS(1374),
    [sym__pipe_table_start] = ACTIONS(1374),
  },
  [STATE(264)] = {
    [ts_builtin_sym_end] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1430),
    [anon_sym_RBRACK] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_POUND] = ACTIONS(1432),
    [anon_sym_DOLLAR] = ACTIONS(1432),
    [anon_sym_PERCENT] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_COLON] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1432),
    [anon_sym_EQ] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_BSLASH] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1432),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_BQUOTE] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1432),
    [anon_sym_RPAREN] = ACTIONS(1432),
    [aux_sym__word_token1] = ACTIONS(1432),
    [aux_sym__word_token2] = ACTIONS(1432),
    [aux_sym__word_token3] = ACTIONS(1432),
    [sym__whitespace] = ACTIONS(1432),
    [sym__soft_line_ending] = ACTIONS(1432),
    [sym__block_quote_start] = ACTIONS(1432),
    [sym__indented_chunk_start] = ACTIONS(1432),
    [sym_atx_h1_marker] = ACTIONS(1432),
    [sym_atx_h2_marker] = ACTIONS(1432),
    [sym_atx_h3_marker] = ACTIONS(1432),
    [sym_atx_h4_marker] = ACTIONS(1432),
    [sym_atx_h5_marker] = ACTIONS(1432),
    [sym_atx_h6_marker] = ACTIONS(1432),
    [sym__thematic_break] = ACTIONS(1432),
    [sym__list_marker_minus] = ACTIONS(1432),
    [sym__list_marker_plus] = ACTIONS(1432),
    [sym__list_marker_star] = ACTIONS(1432),
    [sym__list_marker_parenthesis] = ACTIONS(1432),
    [sym__list_marker_dot] = ACTIONS(1432),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1432),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1432),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1432),
    [sym__blank_line_start] = ACTIONS(1432),
    [sym__html_block_1_start] = ACTIONS(1432),
    [sym__html_block_2_start] = ACTIONS(1432),
    [sym__html_block_3_start] = ACTIONS(1432),
    [sym__html_block_4_start] = ACTIONS(1432),
    [sym__html_block_5_start] = ACTIONS(1432),
    [sym__html_block_6_start] = ACTIONS(1432),
    [sym__html_block_7_start] = ACTIONS(1432),
    [sym__pipe_table_start] = ACTIONS(1432),
  },
  [STATE(265)] = {
    [anon_sym_LBRACK] = ACTIONS(1378),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_POUND] = ACTIONS(1380),
    [anon_sym_DOLLAR] = ACTIONS(1380),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_BSLASH] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_BQUOTE] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [aux_sym__word_token1] = ACTIONS(1380),
    [aux_sym__word_token2] = ACTIONS(1380),
    [aux_sym__word_token3] = ACTIONS(1380),
    [sym__whitespace] = ACTIONS(1380),
    [sym__soft_line_ending] = ACTIONS(1380),
    [sym__block_close] = ACTIONS(1380),
    [sym__block_quote_start] = ACTIONS(1380),
    [sym__indented_chunk_start] = ACTIONS(1380),
    [sym_atx_h1_marker] = ACTIONS(1380),
    [sym_atx_h2_marker] = ACTIONS(1380),
    [sym_atx_h3_marker] = ACTIONS(1380),
    [sym_atx_h4_marker] = ACTIONS(1380),
    [sym_atx_h5_marker] = ACTIONS(1380),
    [sym_atx_h6_marker] = ACTIONS(1380),
    [sym__thematic_break] = ACTIONS(1380),
    [sym__list_marker_minus] = ACTIONS(1380),
    [sym__list_marker_plus] = ACTIONS(1380),
    [sym__list_marker_star] = ACTIONS(1380),
    [sym__list_marker_parenthesis] = ACTIONS(1380),
    [sym__list_marker_dot] = ACTIONS(1380),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1380),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1380),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1380),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1380),
    [sym__blank_line_start] = ACTIONS(1380),
    [sym__html_block_1_start] = ACTIONS(1380),
    [sym__html_block_2_start] = ACTIONS(1380),
    [sym__html_block_3_start] = ACTIONS(1380),
    [sym__html_block_4_start] = ACTIONS(1380),
    [sym__html_block_5_start] = ACTIONS(1380),
    [sym__html_block_6_start] = ACTIONS(1380),
    [sym__html_block_7_start] = ACTIONS(1380),
    [sym__pipe_table_start] = ACTIONS(1380),
  },
  [STATE(266)] = {
    [ts_builtin_sym_end] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_LT] = ACTIONS(1644),
    [anon_sym_GT] = ACTIONS(1644),
    [anon_sym_BANG] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [anon_sym_POUND] = ACTIONS(1644),
    [anon_sym_DOLLAR] = ACTIONS(1644),
    [anon_sym_PERCENT] = ACTIONS(1644),
    [anon_sym_AMP] = ACTIONS(1644),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_STAR] = ACTIONS(1644),
    [anon_sym_PLUS] = ACTIONS(1644),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_DASH] = ACTIONS(1644),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_SLASH] = ACTIONS(1644),
    [anon_sym_COLON] = ACTIONS(1644),
    [anon_sym_SEMI] = ACTIONS(1644),
    [anon_sym_EQ] = ACTIONS(1644),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_AT] = ACTIONS(1644),
    [anon_sym_BSLASH] = ACTIONS(1644),
    [anon_sym_CARET] = ACTIONS(1644),
    [anon_sym__] = ACTIONS(1644),
    [anon_sym_BQUOTE] = ACTIONS(1644),
    [anon_sym_LBRACE] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1644),
    [anon_sym_RBRACE] = ACTIONS(1644),
    [anon_sym_TILDE] = ACTIONS(1644),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_RPAREN] = ACTIONS(1644),
    [aux_sym__word_token1] = ACTIONS(1644),
    [aux_sym__word_token2] = ACTIONS(1644),
    [aux_sym__word_token3] = ACTIONS(1644),
    [sym__whitespace] = ACTIONS(1644),
    [sym__soft_line_ending] = ACTIONS(1644),
    [sym__block_quote_start] = ACTIONS(1644),
    [sym__indented_chunk_start] = ACTIONS(1644),
    [sym_atx_h1_marker] = ACTIONS(1644),
    [sym_atx_h2_marker] = ACTIONS(1644),
    [sym_atx_h3_marker] = ACTIONS(1644),
    [sym_atx_h4_marker] = ACTIONS(1644),
    [sym_atx_h5_marker] = ACTIONS(1644),
    [sym_atx_h6_marker] = ACTIONS(1644),
    [sym__thematic_break] = ACTIONS(1644),
    [sym__list_marker_minus] = ACTIONS(1644),
    [sym__list_marker_plus] = ACTIONS(1644),
    [sym__list_marker_star] = ACTIONS(1644),
    [sym__list_marker_parenthesis] = ACTIONS(1644),
    [sym__list_marker_dot] = ACTIONS(1644),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1644),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1644),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1644),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1644),
    [sym__blank_line_start] = ACTIONS(1644),
    [sym__html_block_1_start] = ACTIONS(1644),
    [sym__html_block_2_start] = ACTIONS(1644),
    [sym__html_block_3_start] = ACTIONS(1644),
    [sym__html_block_4_start] = ACTIONS(1644),
    [sym__html_block_5_start] = ACTIONS(1644),
    [sym__html_block_6_start] = ACTIONS(1644),
    [sym__html_block_7_start] = ACTIONS(1644),
    [sym__pipe_table_start] = ACTIONS(1644),
  },
  [STATE(267)] = {
    [ts_builtin_sym_end] = ACTIONS(1698),
    [anon_sym_LBRACK] = ACTIONS(1700),
    [anon_sym_RBRACK] = ACTIONS(1698),
    [anon_sym_LT] = ACTIONS(1698),
    [anon_sym_GT] = ACTIONS(1698),
    [anon_sym_BANG] = ACTIONS(1698),
    [anon_sym_DQUOTE] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_DOLLAR] = ACTIONS(1698),
    [anon_sym_PERCENT] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1698),
    [anon_sym_STAR] = ACTIONS(1698),
    [anon_sym_PLUS] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1698),
    [anon_sym_DASH] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1698),
    [anon_sym_SLASH] = ACTIONS(1698),
    [anon_sym_COLON] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_EQ] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_AT] = ACTIONS(1698),
    [anon_sym_BSLASH] = ACTIONS(1698),
    [anon_sym_CARET] = ACTIONS(1698),
    [anon_sym__] = ACTIONS(1698),
    [anon_sym_BQUOTE] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_RBRACE] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1698),
    [anon_sym_RPAREN] = ACTIONS(1698),
    [aux_sym__word_token1] = ACTIONS(1698),
    [aux_sym__word_token2] = ACTIONS(1698),
    [aux_sym__word_token3] = ACTIONS(1698),
    [sym__whitespace] = ACTIONS(1698),
    [sym__soft_line_ending] = ACTIONS(1698),
    [sym__block_quote_start] = ACTIONS(1698),
    [sym__indented_chunk_start] = ACTIONS(1698),
    [sym_atx_h1_marker] = ACTIONS(1698),
    [sym_atx_h2_marker] = ACTIONS(1698),
    [sym_atx_h3_marker] = ACTIONS(1698),
    [sym_atx_h4_marker] = ACTIONS(1698),
    [sym_atx_h5_marker] = ACTIONS(1698),
    [sym_atx_h6_marker] = ACTIONS(1698),
    [sym__thematic_break] = ACTIONS(1698),
    [sym__list_marker_minus] = ACTIONS(1698),
    [sym__list_marker_plus] = ACTIONS(1698),
    [sym__list_marker_star] = ACTIONS(1698),
    [sym__list_marker_parenthesis] = ACTIONS(1698),
    [sym__list_marker_dot] = ACTIONS(1698),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1698),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1698),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1698),
    [sym__blank_line_start] = ACTIONS(1698),
    [sym__html_block_1_start] = ACTIONS(1698),
    [sym__html_block_2_start] = ACTIONS(1698),
    [sym__html_block_3_start] = ACTIONS(1698),
    [sym__html_block_4_start] = ACTIONS(1698),
    [sym__html_block_5_start] = ACTIONS(1698),
    [sym__html_block_6_start] = ACTIONS(1698),
    [sym__html_block_7_start] = ACTIONS(1698),
    [sym__pipe_table_start] = ACTIONS(1698),
  },
  [STATE(268)] = {
    [ts_builtin_sym_end] = ACTIONS(1702),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1702),
    [anon_sym_LT] = ACTIONS(1702),
    [anon_sym_GT] = ACTIONS(1702),
    [anon_sym_BANG] = ACTIONS(1702),
    [anon_sym_DQUOTE] = ACTIONS(1702),
    [anon_sym_POUND] = ACTIONS(1702),
    [anon_sym_DOLLAR] = ACTIONS(1702),
    [anon_sym_PERCENT] = ACTIONS(1702),
    [anon_sym_AMP] = ACTIONS(1702),
    [anon_sym_SQUOTE] = ACTIONS(1702),
    [anon_sym_STAR] = ACTIONS(1702),
    [anon_sym_PLUS] = ACTIONS(1702),
    [anon_sym_COMMA] = ACTIONS(1702),
    [anon_sym_DASH] = ACTIONS(1702),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_SLASH] = ACTIONS(1702),
    [anon_sym_COLON] = ACTIONS(1702),
    [anon_sym_SEMI] = ACTIONS(1702),
    [anon_sym_EQ] = ACTIONS(1702),
    [anon_sym_QMARK] = ACTIONS(1702),
    [anon_sym_AT] = ACTIONS(1702),
    [anon_sym_BSLASH] = ACTIONS(1702),
    [anon_sym_CARET] = ACTIONS(1702),
    [anon_sym__] = ACTIONS(1702),
    [anon_sym_BQUOTE] = ACTIONS(1702),
    [anon_sym_LBRACE] = ACTIONS(1702),
    [anon_sym_PIPE] = ACTIONS(1702),
    [anon_sym_RBRACE] = ACTIONS(1702),
    [anon_sym_TILDE] = ACTIONS(1702),
    [anon_sym_LPAREN] = ACTIONS(1702),
    [anon_sym_RPAREN] = ACTIONS(1702),
    [aux_sym__word_token1] = ACTIONS(1702),
    [aux_sym__word_token2] = ACTIONS(1702),
    [aux_sym__word_token3] = ACTIONS(1702),
    [sym__whitespace] = ACTIONS(1702),
    [sym__soft_line_ending] = ACTIONS(1702),
    [sym__block_quote_start] = ACTIONS(1702),
    [sym__indented_chunk_start] = ACTIONS(1702),
    [sym_atx_h1_marker] = ACTIONS(1702),
    [sym_atx_h2_marker] = ACTIONS(1702),
    [sym_atx_h3_marker] = ACTIONS(1702),
    [sym_atx_h4_marker] = ACTIONS(1702),
    [sym_atx_h5_marker] = ACTIONS(1702),
    [sym_atx_h6_marker] = ACTIONS(1702),
    [sym__thematic_break] = ACTIONS(1702),
    [sym__list_marker_minus] = ACTIONS(1702),
    [sym__list_marker_plus] = ACTIONS(1702),
    [sym__list_marker_star] = ACTIONS(1702),
    [sym__list_marker_parenthesis] = ACTIONS(1702),
    [sym__list_marker_dot] = ACTIONS(1702),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1702),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1702),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1702),
    [sym__blank_line_start] = ACTIONS(1702),
    [sym__html_block_1_start] = ACTIONS(1702),
    [sym__html_block_2_start] = ACTIONS(1702),
    [sym__html_block_3_start] = ACTIONS(1702),
    [sym__html_block_4_start] = ACTIONS(1702),
    [sym__html_block_5_start] = ACTIONS(1702),
    [sym__html_block_6_start] = ACTIONS(1702),
    [sym__html_block_7_start] = ACTIONS(1702),
    [sym__pipe_table_start] = ACTIONS(1702),
  },
  [STATE(269)] = {
    [anon_sym_LBRACK] = ACTIONS(1692),
    [anon_sym_RBRACK] = ACTIONS(1690),
    [anon_sym_LT] = ACTIONS(1690),
    [anon_sym_GT] = ACTIONS(1690),
    [anon_sym_BANG] = ACTIONS(1690),
    [anon_sym_DQUOTE] = ACTIONS(1690),
    [anon_sym_POUND] = ACTIONS(1690),
    [anon_sym_DOLLAR] = ACTIONS(1690),
    [anon_sym_PERCENT] = ACTIONS(1690),
    [anon_sym_AMP] = ACTIONS(1690),
    [anon_sym_SQUOTE] = ACTIONS(1690),
    [anon_sym_STAR] = ACTIONS(1690),
    [anon_sym_PLUS] = ACTIONS(1690),
    [anon_sym_COMMA] = ACTIONS(1690),
    [anon_sym_DASH] = ACTIONS(1690),
    [anon_sym_DOT] = ACTIONS(1690),
    [anon_sym_SLASH] = ACTIONS(1690),
    [anon_sym_COLON] = ACTIONS(1690),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_EQ] = ACTIONS(1690),
    [anon_sym_QMARK] = ACTIONS(1690),
    [anon_sym_AT] = ACTIONS(1690),
    [anon_sym_BSLASH] = ACTIONS(1690),
    [anon_sym_CARET] = ACTIONS(1690),
    [anon_sym__] = ACTIONS(1690),
    [anon_sym_BQUOTE] = ACTIONS(1690),
    [anon_sym_LBRACE] = ACTIONS(1690),
    [anon_sym_PIPE] = ACTIONS(1690),
    [anon_sym_RBRACE] = ACTIONS(1690),
    [anon_sym_TILDE] = ACTIONS(1690),
    [anon_sym_LPAREN] = ACTIONS(1690),
    [anon_sym_RPAREN] = ACTIONS(1690),
    [aux_sym__word_token1] = ACTIONS(1690),
    [aux_sym__word_token2] = ACTIONS(1690),
    [aux_sym__word_token3] = ACTIONS(1690),
    [sym__whitespace] = ACTIONS(1690),
    [sym__soft_line_ending] = ACTIONS(1690),
    [sym__block_close] = ACTIONS(1690),
    [sym__block_quote_start] = ACTIONS(1690),
    [sym__indented_chunk_start] = ACTIONS(1690),
    [sym_atx_h1_marker] = ACTIONS(1690),
    [sym_atx_h2_marker] = ACTIONS(1690),
    [sym_atx_h3_marker] = ACTIONS(1690),
    [sym_atx_h4_marker] = ACTIONS(1690),
    [sym_atx_h5_marker] = ACTIONS(1690),
    [sym_atx_h6_marker] = ACTIONS(1690),
    [sym__thematic_break] = ACTIONS(1690),
    [sym__list_marker_minus] = ACTIONS(1690),
    [sym__list_marker_plus] = ACTIONS(1690),
    [sym__list_marker_star] = ACTIONS(1690),
    [sym__list_marker_parenthesis] = ACTIONS(1690),
    [sym__list_marker_dot] = ACTIONS(1690),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1690),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1690),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1690),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1690),
    [sym__blank_line_start] = ACTIONS(1690),
    [sym__html_block_1_start] = ACTIONS(1690),
    [sym__html_block_2_start] = ACTIONS(1690),
    [sym__html_block_3_start] = ACTIONS(1690),
    [sym__html_block_4_start] = ACTIONS(1690),
    [sym__html_block_5_start] = ACTIONS(1690),
    [sym__html_block_6_start] = ACTIONS(1690),
    [sym__html_block_7_start] = ACTIONS(1690),
    [sym__pipe_table_start] = ACTIONS(1690),
  },
  [STATE(270)] = {
    [anon_sym_LBRACK] = ACTIONS(1696),
    [anon_sym_RBRACK] = ACTIONS(1694),
    [anon_sym_LT] = ACTIONS(1694),
    [anon_sym_GT] = ACTIONS(1694),
    [anon_sym_BANG] = ACTIONS(1694),
    [anon_sym_DQUOTE] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_DOLLAR] = ACTIONS(1694),
    [anon_sym_PERCENT] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1694),
    [anon_sym_STAR] = ACTIONS(1694),
    [anon_sym_PLUS] = ACTIONS(1694),
    [anon_sym_COMMA] = ACTIONS(1694),
    [anon_sym_DASH] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1694),
    [anon_sym_SLASH] = ACTIONS(1694),
    [anon_sym_COLON] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1694),
    [anon_sym_EQ] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_AT] = ACTIONS(1694),
    [anon_sym_BSLASH] = ACTIONS(1694),
    [anon_sym_CARET] = ACTIONS(1694),
    [anon_sym__] = ACTIONS(1694),
    [anon_sym_BQUOTE] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_RBRACE] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1694),
    [anon_sym_RPAREN] = ACTIONS(1694),
    [aux_sym__word_token1] = ACTIONS(1694),
    [aux_sym__word_token2] = ACTIONS(1694),
    [aux_sym__word_token3] = ACTIONS(1694),
    [sym__whitespace] = ACTIONS(1694),
    [sym__soft_line_ending] = ACTIONS(1694),
    [sym__block_close] = ACTIONS(1694),
    [sym__block_quote_start] = ACTIONS(1694),
    [sym__indented_chunk_start] = ACTIONS(1694),
    [sym_atx_h1_marker] = ACTIONS(1694),
    [sym_atx_h2_marker] = ACTIONS(1694),
    [sym_atx_h3_marker] = ACTIONS(1694),
    [sym_atx_h4_marker] = ACTIONS(1694),
    [sym_atx_h5_marker] = ACTIONS(1694),
    [sym_atx_h6_marker] = ACTIONS(1694),
    [sym__thematic_break] = ACTIONS(1694),
    [sym__list_marker_minus] = ACTIONS(1694),
    [sym__list_marker_plus] = ACTIONS(1694),
    [sym__list_marker_star] = ACTIONS(1694),
    [sym__list_marker_parenthesis] = ACTIONS(1694),
    [sym__list_marker_dot] = ACTIONS(1694),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1694),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1694),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1694),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1694),
    [sym__blank_line_start] = ACTIONS(1694),
    [sym__html_block_1_start] = ACTIONS(1694),
    [sym__html_block_2_start] = ACTIONS(1694),
    [sym__html_block_3_start] = ACTIONS(1694),
    [sym__html_block_4_start] = ACTIONS(1694),
    [sym__html_block_5_start] = ACTIONS(1694),
    [sym__html_block_6_start] = ACTIONS(1694),
    [sym__html_block_7_start] = ACTIONS(1694),
    [sym__pipe_table_start] = ACTIONS(1694),
  },
  [STATE(271)] = {
    [ts_builtin_sym_end] = ACTIONS(1706),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_LT] = ACTIONS(1706),
    [anon_sym_GT] = ACTIONS(1706),
    [anon_sym_BANG] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [anon_sym_POUND] = ACTIONS(1706),
    [anon_sym_DOLLAR] = ACTIONS(1706),
    [anon_sym_PERCENT] = ACTIONS(1706),
    [anon_sym_AMP] = ACTIONS(1706),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_STAR] = ACTIONS(1706),
    [anon_sym_PLUS] = ACTIONS(1706),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_DASH] = ACTIONS(1706),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_SLASH] = ACTIONS(1706),
    [anon_sym_COLON] = ACTIONS(1706),
    [anon_sym_SEMI] = ACTIONS(1706),
    [anon_sym_EQ] = ACTIONS(1706),
    [anon_sym_QMARK] = ACTIONS(1706),
    [anon_sym_AT] = ACTIONS(1706),
    [anon_sym_BSLASH] = ACTIONS(1706),
    [anon_sym_CARET] = ACTIONS(1706),
    [anon_sym__] = ACTIONS(1706),
    [anon_sym_BQUOTE] = ACTIONS(1706),
    [anon_sym_LBRACE] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1706),
    [anon_sym_RBRACE] = ACTIONS(1706),
    [anon_sym_TILDE] = ACTIONS(1706),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_RPAREN] = ACTIONS(1706),
    [aux_sym__word_token1] = ACTIONS(1706),
    [aux_sym__word_token2] = ACTIONS(1706),
    [aux_sym__word_token3] = ACTIONS(1706),
    [sym__whitespace] = ACTIONS(1706),
    [sym__soft_line_ending] = ACTIONS(1706),
    [sym__block_quote_start] = ACTIONS(1706),
    [sym__indented_chunk_start] = ACTIONS(1706),
    [sym_atx_h1_marker] = ACTIONS(1706),
    [sym_atx_h2_marker] = ACTIONS(1706),
    [sym_atx_h3_marker] = ACTIONS(1706),
    [sym_atx_h4_marker] = ACTIONS(1706),
    [sym_atx_h5_marker] = ACTIONS(1706),
    [sym_atx_h6_marker] = ACTIONS(1706),
    [sym__thematic_break] = ACTIONS(1706),
    [sym__list_marker_minus] = ACTIONS(1706),
    [sym__list_marker_plus] = ACTIONS(1706),
    [sym__list_marker_star] = ACTIONS(1706),
    [sym__list_marker_parenthesis] = ACTIONS(1706),
    [sym__list_marker_dot] = ACTIONS(1706),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1706),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1706),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1706),
    [sym__blank_line_start] = ACTIONS(1706),
    [sym__html_block_1_start] = ACTIONS(1706),
    [sym__html_block_2_start] = ACTIONS(1706),
    [sym__html_block_3_start] = ACTIONS(1706),
    [sym__html_block_4_start] = ACTIONS(1706),
    [sym__html_block_5_start] = ACTIONS(1706),
    [sym__html_block_6_start] = ACTIONS(1706),
    [sym__html_block_7_start] = ACTIONS(1706),
    [sym__pipe_table_start] = ACTIONS(1706),
  },
  [STATE(272)] = {
    [ts_builtin_sym_end] = ACTIONS(1710),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_LT] = ACTIONS(1710),
    [anon_sym_GT] = ACTIONS(1710),
    [anon_sym_BANG] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [anon_sym_POUND] = ACTIONS(1710),
    [anon_sym_DOLLAR] = ACTIONS(1710),
    [anon_sym_PERCENT] = ACTIONS(1710),
    [anon_sym_AMP] = ACTIONS(1710),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_STAR] = ACTIONS(1710),
    [anon_sym_PLUS] = ACTIONS(1710),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_DASH] = ACTIONS(1710),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_SLASH] = ACTIONS(1710),
    [anon_sym_COLON] = ACTIONS(1710),
    [anon_sym_SEMI] = ACTIONS(1710),
    [anon_sym_EQ] = ACTIONS(1710),
    [anon_sym_QMARK] = ACTIONS(1710),
    [anon_sym_AT] = ACTIONS(1710),
    [anon_sym_BSLASH] = ACTIONS(1710),
    [anon_sym_CARET] = ACTIONS(1710),
    [anon_sym__] = ACTIONS(1710),
    [anon_sym_BQUOTE] = ACTIONS(1710),
    [anon_sym_LBRACE] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1710),
    [anon_sym_RBRACE] = ACTIONS(1710),
    [anon_sym_TILDE] = ACTIONS(1710),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_RPAREN] = ACTIONS(1710),
    [aux_sym__word_token1] = ACTIONS(1710),
    [aux_sym__word_token2] = ACTIONS(1710),
    [aux_sym__word_token3] = ACTIONS(1710),
    [sym__whitespace] = ACTIONS(1710),
    [sym__soft_line_ending] = ACTIONS(1710),
    [sym__block_quote_start] = ACTIONS(1710),
    [sym__indented_chunk_start] = ACTIONS(1710),
    [sym_atx_h1_marker] = ACTIONS(1710),
    [sym_atx_h2_marker] = ACTIONS(1710),
    [sym_atx_h3_marker] = ACTIONS(1710),
    [sym_atx_h4_marker] = ACTIONS(1710),
    [sym_atx_h5_marker] = ACTIONS(1710),
    [sym_atx_h6_marker] = ACTIONS(1710),
    [sym__thematic_break] = ACTIONS(1710),
    [sym__list_marker_minus] = ACTIONS(1710),
    [sym__list_marker_plus] = ACTIONS(1710),
    [sym__list_marker_star] = ACTIONS(1710),
    [sym__list_marker_parenthesis] = ACTIONS(1710),
    [sym__list_marker_dot] = ACTIONS(1710),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1710),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1710),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1710),
    [sym__blank_line_start] = ACTIONS(1710),
    [sym__html_block_1_start] = ACTIONS(1710),
    [sym__html_block_2_start] = ACTIONS(1710),
    [sym__html_block_3_start] = ACTIONS(1710),
    [sym__html_block_4_start] = ACTIONS(1710),
    [sym__html_block_5_start] = ACTIONS(1710),
    [sym__html_block_6_start] = ACTIONS(1710),
    [sym__html_block_7_start] = ACTIONS(1710),
    [sym__pipe_table_start] = ACTIONS(1710),
  },
  [STATE(273)] = {
    [ts_builtin_sym_end] = ACTIONS(1714),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_LT] = ACTIONS(1714),
    [anon_sym_GT] = ACTIONS(1714),
    [anon_sym_BANG] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [anon_sym_POUND] = ACTIONS(1714),
    [anon_sym_DOLLAR] = ACTIONS(1714),
    [anon_sym_PERCENT] = ACTIONS(1714),
    [anon_sym_AMP] = ACTIONS(1714),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_STAR] = ACTIONS(1714),
    [anon_sym_PLUS] = ACTIONS(1714),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_DASH] = ACTIONS(1714),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_SLASH] = ACTIONS(1714),
    [anon_sym_COLON] = ACTIONS(1714),
    [anon_sym_SEMI] = ACTIONS(1714),
    [anon_sym_EQ] = ACTIONS(1714),
    [anon_sym_QMARK] = ACTIONS(1714),
    [anon_sym_AT] = ACTIONS(1714),
    [anon_sym_BSLASH] = ACTIONS(1714),
    [anon_sym_CARET] = ACTIONS(1714),
    [anon_sym__] = ACTIONS(1714),
    [anon_sym_BQUOTE] = ACTIONS(1714),
    [anon_sym_LBRACE] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1714),
    [anon_sym_RBRACE] = ACTIONS(1714),
    [anon_sym_TILDE] = ACTIONS(1714),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_RPAREN] = ACTIONS(1714),
    [aux_sym__word_token1] = ACTIONS(1714),
    [aux_sym__word_token2] = ACTIONS(1714),
    [aux_sym__word_token3] = ACTIONS(1714),
    [sym__whitespace] = ACTIONS(1714),
    [sym__soft_line_ending] = ACTIONS(1714),
    [sym__block_quote_start] = ACTIONS(1714),
    [sym__indented_chunk_start] = ACTIONS(1714),
    [sym_atx_h1_marker] = ACTIONS(1714),
    [sym_atx_h2_marker] = ACTIONS(1714),
    [sym_atx_h3_marker] = ACTIONS(1714),
    [sym_atx_h4_marker] = ACTIONS(1714),
    [sym_atx_h5_marker] = ACTIONS(1714),
    [sym_atx_h6_marker] = ACTIONS(1714),
    [sym__thematic_break] = ACTIONS(1714),
    [sym__list_marker_minus] = ACTIONS(1714),
    [sym__list_marker_plus] = ACTIONS(1714),
    [sym__list_marker_star] = ACTIONS(1714),
    [sym__list_marker_parenthesis] = ACTIONS(1714),
    [sym__list_marker_dot] = ACTIONS(1714),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1714),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1714),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1714),
    [sym__blank_line_start] = ACTIONS(1714),
    [sym__html_block_1_start] = ACTIONS(1714),
    [sym__html_block_2_start] = ACTIONS(1714),
    [sym__html_block_3_start] = ACTIONS(1714),
    [sym__html_block_4_start] = ACTIONS(1714),
    [sym__html_block_5_start] = ACTIONS(1714),
    [sym__html_block_6_start] = ACTIONS(1714),
    [sym__html_block_7_start] = ACTIONS(1714),
    [sym__pipe_table_start] = ACTIONS(1714),
  },
  [STATE(274)] = {
    [ts_builtin_sym_end] = ACTIONS(1718),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_LT] = ACTIONS(1718),
    [anon_sym_GT] = ACTIONS(1718),
    [anon_sym_BANG] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [anon_sym_POUND] = ACTIONS(1718),
    [anon_sym_DOLLAR] = ACTIONS(1718),
    [anon_sym_PERCENT] = ACTIONS(1718),
    [anon_sym_AMP] = ACTIONS(1718),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_STAR] = ACTIONS(1718),
    [anon_sym_PLUS] = ACTIONS(1718),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_DASH] = ACTIONS(1718),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_SLASH] = ACTIONS(1718),
    [anon_sym_COLON] = ACTIONS(1718),
    [anon_sym_SEMI] = ACTIONS(1718),
    [anon_sym_EQ] = ACTIONS(1718),
    [anon_sym_QMARK] = ACTIONS(1718),
    [anon_sym_AT] = ACTIONS(1718),
    [anon_sym_BSLASH] = ACTIONS(1718),
    [anon_sym_CARET] = ACTIONS(1718),
    [anon_sym__] = ACTIONS(1718),
    [anon_sym_BQUOTE] = ACTIONS(1718),
    [anon_sym_LBRACE] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1718),
    [anon_sym_RBRACE] = ACTIONS(1718),
    [anon_sym_TILDE] = ACTIONS(1718),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_RPAREN] = ACTIONS(1718),
    [aux_sym__word_token1] = ACTIONS(1718),
    [aux_sym__word_token2] = ACTIONS(1718),
    [aux_sym__word_token3] = ACTIONS(1718),
    [sym__whitespace] = ACTIONS(1718),
    [sym__soft_line_ending] = ACTIONS(1718),
    [sym__block_quote_start] = ACTIONS(1718),
    [sym__indented_chunk_start] = ACTIONS(1718),
    [sym_atx_h1_marker] = ACTIONS(1718),
    [sym_atx_h2_marker] = ACTIONS(1718),
    [sym_atx_h3_marker] = ACTIONS(1718),
    [sym_atx_h4_marker] = ACTIONS(1718),
    [sym_atx_h5_marker] = ACTIONS(1718),
    [sym_atx_h6_marker] = ACTIONS(1718),
    [sym__thematic_break] = ACTIONS(1718),
    [sym__list_marker_minus] = ACTIONS(1718),
    [sym__list_marker_plus] = ACTIONS(1718),
    [sym__list_marker_star] = ACTIONS(1718),
    [sym__list_marker_parenthesis] = ACTIONS(1718),
    [sym__list_marker_dot] = ACTIONS(1718),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1718),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1718),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1718),
    [sym__blank_line_start] = ACTIONS(1718),
    [sym__html_block_1_start] = ACTIONS(1718),
    [sym__html_block_2_start] = ACTIONS(1718),
    [sym__html_block_3_start] = ACTIONS(1718),
    [sym__html_block_4_start] = ACTIONS(1718),
    [sym__html_block_5_start] = ACTIONS(1718),
    [sym__html_block_6_start] = ACTIONS(1718),
    [sym__html_block_7_start] = ACTIONS(1718),
    [sym__pipe_table_start] = ACTIONS(1718),
  },
  [STATE(275)] = {
    [ts_builtin_sym_end] = ACTIONS(1722),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_LT] = ACTIONS(1722),
    [anon_sym_GT] = ACTIONS(1722),
    [anon_sym_BANG] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [anon_sym_POUND] = ACTIONS(1722),
    [anon_sym_DOLLAR] = ACTIONS(1722),
    [anon_sym_PERCENT] = ACTIONS(1722),
    [anon_sym_AMP] = ACTIONS(1722),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_STAR] = ACTIONS(1722),
    [anon_sym_PLUS] = ACTIONS(1722),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_DASH] = ACTIONS(1722),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_SLASH] = ACTIONS(1722),
    [anon_sym_COLON] = ACTIONS(1722),
    [anon_sym_SEMI] = ACTIONS(1722),
    [anon_sym_EQ] = ACTIONS(1722),
    [anon_sym_QMARK] = ACTIONS(1722),
    [anon_sym_AT] = ACTIONS(1722),
    [anon_sym_BSLASH] = ACTIONS(1722),
    [anon_sym_CARET] = ACTIONS(1722),
    [anon_sym__] = ACTIONS(1722),
    [anon_sym_BQUOTE] = ACTIONS(1722),
    [anon_sym_LBRACE] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1722),
    [anon_sym_RBRACE] = ACTIONS(1722),
    [anon_sym_TILDE] = ACTIONS(1722),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_RPAREN] = ACTIONS(1722),
    [aux_sym__word_token1] = ACTIONS(1722),
    [aux_sym__word_token2] = ACTIONS(1722),
    [aux_sym__word_token3] = ACTIONS(1722),
    [sym__whitespace] = ACTIONS(1722),
    [sym__soft_line_ending] = ACTIONS(1722),
    [sym__block_quote_start] = ACTIONS(1722),
    [sym__indented_chunk_start] = ACTIONS(1722),
    [sym_atx_h1_marker] = ACTIONS(1722),
    [sym_atx_h2_marker] = ACTIONS(1722),
    [sym_atx_h3_marker] = ACTIONS(1722),
    [sym_atx_h4_marker] = ACTIONS(1722),
    [sym_atx_h5_marker] = ACTIONS(1722),
    [sym_atx_h6_marker] = ACTIONS(1722),
    [sym__thematic_break] = ACTIONS(1722),
    [sym__list_marker_minus] = ACTIONS(1722),
    [sym__list_marker_plus] = ACTIONS(1722),
    [sym__list_marker_star] = ACTIONS(1722),
    [sym__list_marker_parenthesis] = ACTIONS(1722),
    [sym__list_marker_dot] = ACTIONS(1722),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1722),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1722),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1722),
    [sym__blank_line_start] = ACTIONS(1722),
    [sym__html_block_1_start] = ACTIONS(1722),
    [sym__html_block_2_start] = ACTIONS(1722),
    [sym__html_block_3_start] = ACTIONS(1722),
    [sym__html_block_4_start] = ACTIONS(1722),
    [sym__html_block_5_start] = ACTIONS(1722),
    [sym__html_block_6_start] = ACTIONS(1722),
    [sym__html_block_7_start] = ACTIONS(1722),
    [sym__pipe_table_start] = ACTIONS(1722),
  },
  [STATE(276)] = {
    [anon_sym_LBRACK] = ACTIONS(1430),
    [anon_sym_RBRACK] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_POUND] = ACTIONS(1432),
    [anon_sym_DOLLAR] = ACTIONS(1432),
    [anon_sym_PERCENT] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_COLON] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1432),
    [anon_sym_EQ] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_BSLASH] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1432),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_BQUOTE] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1432),
    [anon_sym_RPAREN] = ACTIONS(1432),
    [aux_sym__word_token1] = ACTIONS(1432),
    [aux_sym__word_token2] = ACTIONS(1432),
    [aux_sym__word_token3] = ACTIONS(1432),
    [sym__whitespace] = ACTIONS(1432),
    [sym__soft_line_ending] = ACTIONS(1432),
    [sym__block_close] = ACTIONS(1432),
    [sym__block_quote_start] = ACTIONS(1432),
    [sym__indented_chunk_start] = ACTIONS(1432),
    [sym_atx_h1_marker] = ACTIONS(1432),
    [sym_atx_h2_marker] = ACTIONS(1432),
    [sym_atx_h3_marker] = ACTIONS(1432),
    [sym_atx_h4_marker] = ACTIONS(1432),
    [sym_atx_h5_marker] = ACTIONS(1432),
    [sym_atx_h6_marker] = ACTIONS(1432),
    [sym__thematic_break] = ACTIONS(1432),
    [sym__list_marker_minus] = ACTIONS(1432),
    [sym__list_marker_plus] = ACTIONS(1432),
    [sym__list_marker_star] = ACTIONS(1432),
    [sym__list_marker_parenthesis] = ACTIONS(1432),
    [sym__list_marker_dot] = ACTIONS(1432),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1432),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1432),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1432),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1432),
    [sym__blank_line_start] = ACTIONS(1432),
    [sym__html_block_1_start] = ACTIONS(1432),
    [sym__html_block_2_start] = ACTIONS(1432),
    [sym__html_block_3_start] = ACTIONS(1432),
    [sym__html_block_4_start] = ACTIONS(1432),
    [sym__html_block_5_start] = ACTIONS(1432),
    [sym__html_block_6_start] = ACTIONS(1432),
    [sym__html_block_7_start] = ACTIONS(1432),
    [sym__pipe_table_start] = ACTIONS(1432),
  },
  [STATE(277)] = {
    [ts_builtin_sym_end] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_LT] = ACTIONS(1726),
    [anon_sym_GT] = ACTIONS(1726),
    [anon_sym_BANG] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [anon_sym_POUND] = ACTIONS(1726),
    [anon_sym_DOLLAR] = ACTIONS(1726),
    [anon_sym_PERCENT] = ACTIONS(1726),
    [anon_sym_AMP] = ACTIONS(1726),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_STAR] = ACTIONS(1726),
    [anon_sym_PLUS] = ACTIONS(1726),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_DASH] = ACTIONS(1726),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_SLASH] = ACTIONS(1726),
    [anon_sym_COLON] = ACTIONS(1726),
    [anon_sym_SEMI] = ACTIONS(1726),
    [anon_sym_EQ] = ACTIONS(1726),
    [anon_sym_QMARK] = ACTIONS(1726),
    [anon_sym_AT] = ACTIONS(1726),
    [anon_sym_BSLASH] = ACTIONS(1726),
    [anon_sym_CARET] = ACTIONS(1726),
    [anon_sym__] = ACTIONS(1726),
    [anon_sym_BQUOTE] = ACTIONS(1726),
    [anon_sym_LBRACE] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1726),
    [anon_sym_RBRACE] = ACTIONS(1726),
    [anon_sym_TILDE] = ACTIONS(1726),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_RPAREN] = ACTIONS(1726),
    [aux_sym__word_token1] = ACTIONS(1726),
    [aux_sym__word_token2] = ACTIONS(1726),
    [aux_sym__word_token3] = ACTIONS(1726),
    [sym__whitespace] = ACTIONS(1726),
    [sym__soft_line_ending] = ACTIONS(1726),
    [sym__block_quote_start] = ACTIONS(1726),
    [sym__indented_chunk_start] = ACTIONS(1726),
    [sym_atx_h1_marker] = ACTIONS(1726),
    [sym_atx_h2_marker] = ACTIONS(1726),
    [sym_atx_h3_marker] = ACTIONS(1726),
    [sym_atx_h4_marker] = ACTIONS(1726),
    [sym_atx_h5_marker] = ACTIONS(1726),
    [sym_atx_h6_marker] = ACTIONS(1726),
    [sym__thematic_break] = ACTIONS(1726),
    [sym__list_marker_minus] = ACTIONS(1726),
    [sym__list_marker_plus] = ACTIONS(1726),
    [sym__list_marker_star] = ACTIONS(1726),
    [sym__list_marker_parenthesis] = ACTIONS(1726),
    [sym__list_marker_dot] = ACTIONS(1726),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1726),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1726),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1726),
    [sym__blank_line_start] = ACTIONS(1726),
    [sym__html_block_1_start] = ACTIONS(1726),
    [sym__html_block_2_start] = ACTIONS(1726),
    [sym__html_block_3_start] = ACTIONS(1726),
    [sym__html_block_4_start] = ACTIONS(1726),
    [sym__html_block_5_start] = ACTIONS(1726),
    [sym__html_block_6_start] = ACTIONS(1726),
    [sym__html_block_7_start] = ACTIONS(1726),
    [sym__pipe_table_start] = ACTIONS(1726),
  },
  [STATE(278)] = {
    [anon_sym_LBRACK] = ACTIONS(1700),
    [anon_sym_RBRACK] = ACTIONS(1698),
    [anon_sym_LT] = ACTIONS(1698),
    [anon_sym_GT] = ACTIONS(1698),
    [anon_sym_BANG] = ACTIONS(1698),
    [anon_sym_DQUOTE] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_DOLLAR] = ACTIONS(1698),
    [anon_sym_PERCENT] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1698),
    [anon_sym_STAR] = ACTIONS(1698),
    [anon_sym_PLUS] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1698),
    [anon_sym_DASH] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1698),
    [anon_sym_SLASH] = ACTIONS(1698),
    [anon_sym_COLON] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_EQ] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_AT] = ACTIONS(1698),
    [anon_sym_BSLASH] = ACTIONS(1698),
    [anon_sym_CARET] = ACTIONS(1698),
    [anon_sym__] = ACTIONS(1698),
    [anon_sym_BQUOTE] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_RBRACE] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1698),
    [anon_sym_RPAREN] = ACTIONS(1698),
    [aux_sym__word_token1] = ACTIONS(1698),
    [aux_sym__word_token2] = ACTIONS(1698),
    [aux_sym__word_token3] = ACTIONS(1698),
    [sym__whitespace] = ACTIONS(1698),
    [sym__soft_line_ending] = ACTIONS(1698),
    [sym__block_close] = ACTIONS(1698),
    [sym__block_quote_start] = ACTIONS(1698),
    [sym__indented_chunk_start] = ACTIONS(1698),
    [sym_atx_h1_marker] = ACTIONS(1698),
    [sym_atx_h2_marker] = ACTIONS(1698),
    [sym_atx_h3_marker] = ACTIONS(1698),
    [sym_atx_h4_marker] = ACTIONS(1698),
    [sym_atx_h5_marker] = ACTIONS(1698),
    [sym_atx_h6_marker] = ACTIONS(1698),
    [sym__thematic_break] = ACTIONS(1698),
    [sym__list_marker_minus] = ACTIONS(1698),
    [sym__list_marker_plus] = ACTIONS(1698),
    [sym__list_marker_star] = ACTIONS(1698),
    [sym__list_marker_parenthesis] = ACTIONS(1698),
    [sym__list_marker_dot] = ACTIONS(1698),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1698),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1698),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1698),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1698),
    [sym__blank_line_start] = ACTIONS(1698),
    [sym__html_block_1_start] = ACTIONS(1698),
    [sym__html_block_2_start] = ACTIONS(1698),
    [sym__html_block_3_start] = ACTIONS(1698),
    [sym__html_block_4_start] = ACTIONS(1698),
    [sym__html_block_5_start] = ACTIONS(1698),
    [sym__html_block_6_start] = ACTIONS(1698),
    [sym__html_block_7_start] = ACTIONS(1698),
    [sym__pipe_table_start] = ACTIONS(1698),
  },
  [STATE(279)] = {
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1702),
    [anon_sym_LT] = ACTIONS(1702),
    [anon_sym_GT] = ACTIONS(1702),
    [anon_sym_BANG] = ACTIONS(1702),
    [anon_sym_DQUOTE] = ACTIONS(1702),
    [anon_sym_POUND] = ACTIONS(1702),
    [anon_sym_DOLLAR] = ACTIONS(1702),
    [anon_sym_PERCENT] = ACTIONS(1702),
    [anon_sym_AMP] = ACTIONS(1702),
    [anon_sym_SQUOTE] = ACTIONS(1702),
    [anon_sym_STAR] = ACTIONS(1702),
    [anon_sym_PLUS] = ACTIONS(1702),
    [anon_sym_COMMA] = ACTIONS(1702),
    [anon_sym_DASH] = ACTIONS(1702),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_SLASH] = ACTIONS(1702),
    [anon_sym_COLON] = ACTIONS(1702),
    [anon_sym_SEMI] = ACTIONS(1702),
    [anon_sym_EQ] = ACTIONS(1702),
    [anon_sym_QMARK] = ACTIONS(1702),
    [anon_sym_AT] = ACTIONS(1702),
    [anon_sym_BSLASH] = ACTIONS(1702),
    [anon_sym_CARET] = ACTIONS(1702),
    [anon_sym__] = ACTIONS(1702),
    [anon_sym_BQUOTE] = ACTIONS(1702),
    [anon_sym_LBRACE] = ACTIONS(1702),
    [anon_sym_PIPE] = ACTIONS(1702),
    [anon_sym_RBRACE] = ACTIONS(1702),
    [anon_sym_TILDE] = ACTIONS(1702),
    [anon_sym_LPAREN] = ACTIONS(1702),
    [anon_sym_RPAREN] = ACTIONS(1702),
    [aux_sym__word_token1] = ACTIONS(1702),
    [aux_sym__word_token2] = ACTIONS(1702),
    [aux_sym__word_token3] = ACTIONS(1702),
    [sym__whitespace] = ACTIONS(1702),
    [sym__soft_line_ending] = ACTIONS(1702),
    [sym__block_close] = ACTIONS(1702),
    [sym__block_quote_start] = ACTIONS(1702),
    [sym__indented_chunk_start] = ACTIONS(1702),
    [sym_atx_h1_marker] = ACTIONS(1702),
    [sym_atx_h2_marker] = ACTIONS(1702),
    [sym_atx_h3_marker] = ACTIONS(1702),
    [sym_atx_h4_marker] = ACTIONS(1702),
    [sym_atx_h5_marker] = ACTIONS(1702),
    [sym_atx_h6_marker] = ACTIONS(1702),
    [sym__thematic_break] = ACTIONS(1702),
    [sym__list_marker_minus] = ACTIONS(1702),
    [sym__list_marker_plus] = ACTIONS(1702),
    [sym__list_marker_star] = ACTIONS(1702),
    [sym__list_marker_parenthesis] = ACTIONS(1702),
    [sym__list_marker_dot] = ACTIONS(1702),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1702),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1702),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1702),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1702),
    [sym__blank_line_start] = ACTIONS(1702),
    [sym__html_block_1_start] = ACTIONS(1702),
    [sym__html_block_2_start] = ACTIONS(1702),
    [sym__html_block_3_start] = ACTIONS(1702),
    [sym__html_block_4_start] = ACTIONS(1702),
    [sym__html_block_5_start] = ACTIONS(1702),
    [sym__html_block_6_start] = ACTIONS(1702),
    [sym__html_block_7_start] = ACTIONS(1702),
    [sym__pipe_table_start] = ACTIONS(1702),
  },
  [STATE(280)] = {
    [anon_sym_LBRACK] = ACTIONS(1730),
    [anon_sym_RBRACK] = ACTIONS(1732),
    [anon_sym_LT] = ACTIONS(1732),
    [anon_sym_GT] = ACTIONS(1732),
    [anon_sym_BANG] = ACTIONS(1732),
    [anon_sym_DQUOTE] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_DOLLAR] = ACTIONS(1732),
    [anon_sym_PERCENT] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1732),
    [anon_sym_STAR] = ACTIONS(1732),
    [anon_sym_PLUS] = ACTIONS(1732),
    [anon_sym_COMMA] = ACTIONS(1732),
    [anon_sym_DASH] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1732),
    [anon_sym_SLASH] = ACTIONS(1732),
    [anon_sym_COLON] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym_EQ] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_AT] = ACTIONS(1732),
    [anon_sym_BSLASH] = ACTIONS(1732),
    [anon_sym_CARET] = ACTIONS(1732),
    [anon_sym__] = ACTIONS(1732),
    [anon_sym_BQUOTE] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_RBRACE] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1732),
    [anon_sym_RPAREN] = ACTIONS(1732),
    [aux_sym__word_token1] = ACTIONS(1732),
    [aux_sym__word_token2] = ACTIONS(1732),
    [aux_sym__word_token3] = ACTIONS(1732),
    [sym__whitespace] = ACTIONS(1732),
    [sym__soft_line_ending] = ACTIONS(1732),
    [sym__block_close] = ACTIONS(1732),
    [sym__block_quote_start] = ACTIONS(1732),
    [sym__indented_chunk_start] = ACTIONS(1732),
    [sym_atx_h1_marker] = ACTIONS(1732),
    [sym_atx_h2_marker] = ACTIONS(1732),
    [sym_atx_h3_marker] = ACTIONS(1732),
    [sym_atx_h4_marker] = ACTIONS(1732),
    [sym_atx_h5_marker] = ACTIONS(1732),
    [sym_atx_h6_marker] = ACTIONS(1732),
    [sym__thematic_break] = ACTIONS(1732),
    [sym__list_marker_minus] = ACTIONS(1732),
    [sym__list_marker_plus] = ACTIONS(1732),
    [sym__list_marker_star] = ACTIONS(1732),
    [sym__list_marker_parenthesis] = ACTIONS(1732),
    [sym__list_marker_dot] = ACTIONS(1732),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1732),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1732),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1732),
    [sym__blank_line_start] = ACTIONS(1732),
    [sym__html_block_1_start] = ACTIONS(1732),
    [sym__html_block_2_start] = ACTIONS(1732),
    [sym__html_block_3_start] = ACTIONS(1732),
    [sym__html_block_4_start] = ACTIONS(1732),
    [sym__html_block_5_start] = ACTIONS(1732),
    [sym__html_block_6_start] = ACTIONS(1732),
    [sym__html_block_7_start] = ACTIONS(1732),
    [sym__pipe_table_start] = ACTIONS(1732),
  },
  [STATE(281)] = {
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_LT] = ACTIONS(1710),
    [anon_sym_GT] = ACTIONS(1710),
    [anon_sym_BANG] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [anon_sym_POUND] = ACTIONS(1710),
    [anon_sym_DOLLAR] = ACTIONS(1710),
    [anon_sym_PERCENT] = ACTIONS(1710),
    [anon_sym_AMP] = ACTIONS(1710),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_STAR] = ACTIONS(1710),
    [anon_sym_PLUS] = ACTIONS(1710),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_DASH] = ACTIONS(1710),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_SLASH] = ACTIONS(1710),
    [anon_sym_COLON] = ACTIONS(1710),
    [anon_sym_SEMI] = ACTIONS(1710),
    [anon_sym_EQ] = ACTIONS(1710),
    [anon_sym_QMARK] = ACTIONS(1710),
    [anon_sym_AT] = ACTIONS(1710),
    [anon_sym_BSLASH] = ACTIONS(1710),
    [anon_sym_CARET] = ACTIONS(1710),
    [anon_sym__] = ACTIONS(1710),
    [anon_sym_BQUOTE] = ACTIONS(1710),
    [anon_sym_LBRACE] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1710),
    [anon_sym_RBRACE] = ACTIONS(1710),
    [anon_sym_TILDE] = ACTIONS(1710),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_RPAREN] = ACTIONS(1710),
    [aux_sym__word_token1] = ACTIONS(1710),
    [aux_sym__word_token2] = ACTIONS(1710),
    [aux_sym__word_token3] = ACTIONS(1710),
    [sym__whitespace] = ACTIONS(1710),
    [sym__soft_line_ending] = ACTIONS(1710),
    [sym__block_close] = ACTIONS(1710),
    [sym__block_quote_start] = ACTIONS(1710),
    [sym__indented_chunk_start] = ACTIONS(1710),
    [sym_atx_h1_marker] = ACTIONS(1710),
    [sym_atx_h2_marker] = ACTIONS(1710),
    [sym_atx_h3_marker] = ACTIONS(1710),
    [sym_atx_h4_marker] = ACTIONS(1710),
    [sym_atx_h5_marker] = ACTIONS(1710),
    [sym_atx_h6_marker] = ACTIONS(1710),
    [sym__thematic_break] = ACTIONS(1710),
    [sym__list_marker_minus] = ACTIONS(1710),
    [sym__list_marker_plus] = ACTIONS(1710),
    [sym__list_marker_star] = ACTIONS(1710),
    [sym__list_marker_parenthesis] = ACTIONS(1710),
    [sym__list_marker_dot] = ACTIONS(1710),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1710),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1710),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1710),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1710),
    [sym__blank_line_start] = ACTIONS(1710),
    [sym__html_block_1_start] = ACTIONS(1710),
    [sym__html_block_2_start] = ACTIONS(1710),
    [sym__html_block_3_start] = ACTIONS(1710),
    [sym__html_block_4_start] = ACTIONS(1710),
    [sym__html_block_5_start] = ACTIONS(1710),
    [sym__html_block_6_start] = ACTIONS(1710),
    [sym__html_block_7_start] = ACTIONS(1710),
    [sym__pipe_table_start] = ACTIONS(1710),
  },
  [STATE(282)] = {
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_LT] = ACTIONS(1714),
    [anon_sym_GT] = ACTIONS(1714),
    [anon_sym_BANG] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [anon_sym_POUND] = ACTIONS(1714),
    [anon_sym_DOLLAR] = ACTIONS(1714),
    [anon_sym_PERCENT] = ACTIONS(1714),
    [anon_sym_AMP] = ACTIONS(1714),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_STAR] = ACTIONS(1714),
    [anon_sym_PLUS] = ACTIONS(1714),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_DASH] = ACTIONS(1714),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_SLASH] = ACTIONS(1714),
    [anon_sym_COLON] = ACTIONS(1714),
    [anon_sym_SEMI] = ACTIONS(1714),
    [anon_sym_EQ] = ACTIONS(1714),
    [anon_sym_QMARK] = ACTIONS(1714),
    [anon_sym_AT] = ACTIONS(1714),
    [anon_sym_BSLASH] = ACTIONS(1714),
    [anon_sym_CARET] = ACTIONS(1714),
    [anon_sym__] = ACTIONS(1714),
    [anon_sym_BQUOTE] = ACTIONS(1714),
    [anon_sym_LBRACE] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1714),
    [anon_sym_RBRACE] = ACTIONS(1714),
    [anon_sym_TILDE] = ACTIONS(1714),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_RPAREN] = ACTIONS(1714),
    [aux_sym__word_token1] = ACTIONS(1714),
    [aux_sym__word_token2] = ACTIONS(1714),
    [aux_sym__word_token3] = ACTIONS(1714),
    [sym__whitespace] = ACTIONS(1714),
    [sym__soft_line_ending] = ACTIONS(1714),
    [sym__block_close] = ACTIONS(1714),
    [sym__block_quote_start] = ACTIONS(1714),
    [sym__indented_chunk_start] = ACTIONS(1714),
    [sym_atx_h1_marker] = ACTIONS(1714),
    [sym_atx_h2_marker] = ACTIONS(1714),
    [sym_atx_h3_marker] = ACTIONS(1714),
    [sym_atx_h4_marker] = ACTIONS(1714),
    [sym_atx_h5_marker] = ACTIONS(1714),
    [sym_atx_h6_marker] = ACTIONS(1714),
    [sym__thematic_break] = ACTIONS(1714),
    [sym__list_marker_minus] = ACTIONS(1714),
    [sym__list_marker_plus] = ACTIONS(1714),
    [sym__list_marker_star] = ACTIONS(1714),
    [sym__list_marker_parenthesis] = ACTIONS(1714),
    [sym__list_marker_dot] = ACTIONS(1714),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1714),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1714),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1714),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1714),
    [sym__blank_line_start] = ACTIONS(1714),
    [sym__html_block_1_start] = ACTIONS(1714),
    [sym__html_block_2_start] = ACTIONS(1714),
    [sym__html_block_3_start] = ACTIONS(1714),
    [sym__html_block_4_start] = ACTIONS(1714),
    [sym__html_block_5_start] = ACTIONS(1714),
    [sym__html_block_6_start] = ACTIONS(1714),
    [sym__html_block_7_start] = ACTIONS(1714),
    [sym__pipe_table_start] = ACTIONS(1714),
  },
  [STATE(283)] = {
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_LT] = ACTIONS(1718),
    [anon_sym_GT] = ACTIONS(1718),
    [anon_sym_BANG] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [anon_sym_POUND] = ACTIONS(1718),
    [anon_sym_DOLLAR] = ACTIONS(1718),
    [anon_sym_PERCENT] = ACTIONS(1718),
    [anon_sym_AMP] = ACTIONS(1718),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_STAR] = ACTIONS(1718),
    [anon_sym_PLUS] = ACTIONS(1718),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_DASH] = ACTIONS(1718),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_SLASH] = ACTIONS(1718),
    [anon_sym_COLON] = ACTIONS(1718),
    [anon_sym_SEMI] = ACTIONS(1718),
    [anon_sym_EQ] = ACTIONS(1718),
    [anon_sym_QMARK] = ACTIONS(1718),
    [anon_sym_AT] = ACTIONS(1718),
    [anon_sym_BSLASH] = ACTIONS(1718),
    [anon_sym_CARET] = ACTIONS(1718),
    [anon_sym__] = ACTIONS(1718),
    [anon_sym_BQUOTE] = ACTIONS(1718),
    [anon_sym_LBRACE] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1718),
    [anon_sym_RBRACE] = ACTIONS(1718),
    [anon_sym_TILDE] = ACTIONS(1718),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_RPAREN] = ACTIONS(1718),
    [aux_sym__word_token1] = ACTIONS(1718),
    [aux_sym__word_token2] = ACTIONS(1718),
    [aux_sym__word_token3] = ACTIONS(1718),
    [sym__whitespace] = ACTIONS(1718),
    [sym__soft_line_ending] = ACTIONS(1718),
    [sym__block_close] = ACTIONS(1718),
    [sym__block_quote_start] = ACTIONS(1718),
    [sym__indented_chunk_start] = ACTIONS(1718),
    [sym_atx_h1_marker] = ACTIONS(1718),
    [sym_atx_h2_marker] = ACTIONS(1718),
    [sym_atx_h3_marker] = ACTIONS(1718),
    [sym_atx_h4_marker] = ACTIONS(1718),
    [sym_atx_h5_marker] = ACTIONS(1718),
    [sym_atx_h6_marker] = ACTIONS(1718),
    [sym__thematic_break] = ACTIONS(1718),
    [sym__list_marker_minus] = ACTIONS(1718),
    [sym__list_marker_plus] = ACTIONS(1718),
    [sym__list_marker_star] = ACTIONS(1718),
    [sym__list_marker_parenthesis] = ACTIONS(1718),
    [sym__list_marker_dot] = ACTIONS(1718),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1718),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1718),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1718),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1718),
    [sym__blank_line_start] = ACTIONS(1718),
    [sym__html_block_1_start] = ACTIONS(1718),
    [sym__html_block_2_start] = ACTIONS(1718),
    [sym__html_block_3_start] = ACTIONS(1718),
    [sym__html_block_4_start] = ACTIONS(1718),
    [sym__html_block_5_start] = ACTIONS(1718),
    [sym__html_block_6_start] = ACTIONS(1718),
    [sym__html_block_7_start] = ACTIONS(1718),
    [sym__pipe_table_start] = ACTIONS(1718),
  },
  [STATE(284)] = {
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_LT] = ACTIONS(1722),
    [anon_sym_GT] = ACTIONS(1722),
    [anon_sym_BANG] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [anon_sym_POUND] = ACTIONS(1722),
    [anon_sym_DOLLAR] = ACTIONS(1722),
    [anon_sym_PERCENT] = ACTIONS(1722),
    [anon_sym_AMP] = ACTIONS(1722),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_STAR] = ACTIONS(1722),
    [anon_sym_PLUS] = ACTIONS(1722),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_DASH] = ACTIONS(1722),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_SLASH] = ACTIONS(1722),
    [anon_sym_COLON] = ACTIONS(1722),
    [anon_sym_SEMI] = ACTIONS(1722),
    [anon_sym_EQ] = ACTIONS(1722),
    [anon_sym_QMARK] = ACTIONS(1722),
    [anon_sym_AT] = ACTIONS(1722),
    [anon_sym_BSLASH] = ACTIONS(1722),
    [anon_sym_CARET] = ACTIONS(1722),
    [anon_sym__] = ACTIONS(1722),
    [anon_sym_BQUOTE] = ACTIONS(1722),
    [anon_sym_LBRACE] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1722),
    [anon_sym_RBRACE] = ACTIONS(1722),
    [anon_sym_TILDE] = ACTIONS(1722),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_RPAREN] = ACTIONS(1722),
    [aux_sym__word_token1] = ACTIONS(1722),
    [aux_sym__word_token2] = ACTIONS(1722),
    [aux_sym__word_token3] = ACTIONS(1722),
    [sym__whitespace] = ACTIONS(1722),
    [sym__soft_line_ending] = ACTIONS(1722),
    [sym__block_close] = ACTIONS(1722),
    [sym__block_quote_start] = ACTIONS(1722),
    [sym__indented_chunk_start] = ACTIONS(1722),
    [sym_atx_h1_marker] = ACTIONS(1722),
    [sym_atx_h2_marker] = ACTIONS(1722),
    [sym_atx_h3_marker] = ACTIONS(1722),
    [sym_atx_h4_marker] = ACTIONS(1722),
    [sym_atx_h5_marker] = ACTIONS(1722),
    [sym_atx_h6_marker] = ACTIONS(1722),
    [sym__thematic_break] = ACTIONS(1722),
    [sym__list_marker_minus] = ACTIONS(1722),
    [sym__list_marker_plus] = ACTIONS(1722),
    [sym__list_marker_star] = ACTIONS(1722),
    [sym__list_marker_parenthesis] = ACTIONS(1722),
    [sym__list_marker_dot] = ACTIONS(1722),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1722),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1722),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1722),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1722),
    [sym__blank_line_start] = ACTIONS(1722),
    [sym__html_block_1_start] = ACTIONS(1722),
    [sym__html_block_2_start] = ACTIONS(1722),
    [sym__html_block_3_start] = ACTIONS(1722),
    [sym__html_block_4_start] = ACTIONS(1722),
    [sym__html_block_5_start] = ACTIONS(1722),
    [sym__html_block_6_start] = ACTIONS(1722),
    [sym__html_block_7_start] = ACTIONS(1722),
    [sym__pipe_table_start] = ACTIONS(1722),
  },
  [STATE(285)] = {
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_LT] = ACTIONS(1726),
    [anon_sym_GT] = ACTIONS(1726),
    [anon_sym_BANG] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [anon_sym_POUND] = ACTIONS(1726),
    [anon_sym_DOLLAR] = ACTIONS(1726),
    [anon_sym_PERCENT] = ACTIONS(1726),
    [anon_sym_AMP] = ACTIONS(1726),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_STAR] = ACTIONS(1726),
    [anon_sym_PLUS] = ACTIONS(1726),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_DASH] = ACTIONS(1726),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_SLASH] = ACTIONS(1726),
    [anon_sym_COLON] = ACTIONS(1726),
    [anon_sym_SEMI] = ACTIONS(1726),
    [anon_sym_EQ] = ACTIONS(1726),
    [anon_sym_QMARK] = ACTIONS(1726),
    [anon_sym_AT] = ACTIONS(1726),
    [anon_sym_BSLASH] = ACTIONS(1726),
    [anon_sym_CARET] = ACTIONS(1726),
    [anon_sym__] = ACTIONS(1726),
    [anon_sym_BQUOTE] = ACTIONS(1726),
    [anon_sym_LBRACE] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1726),
    [anon_sym_RBRACE] = ACTIONS(1726),
    [anon_sym_TILDE] = ACTIONS(1726),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_RPAREN] = ACTIONS(1726),
    [aux_sym__word_token1] = ACTIONS(1726),
    [aux_sym__word_token2] = ACTIONS(1726),
    [aux_sym__word_token3] = ACTIONS(1726),
    [sym__whitespace] = ACTIONS(1726),
    [sym__soft_line_ending] = ACTIONS(1726),
    [sym__block_close] = ACTIONS(1726),
    [sym__block_quote_start] = ACTIONS(1726),
    [sym__indented_chunk_start] = ACTIONS(1726),
    [sym_atx_h1_marker] = ACTIONS(1726),
    [sym_atx_h2_marker] = ACTIONS(1726),
    [sym_atx_h3_marker] = ACTIONS(1726),
    [sym_atx_h4_marker] = ACTIONS(1726),
    [sym_atx_h5_marker] = ACTIONS(1726),
    [sym_atx_h6_marker] = ACTIONS(1726),
    [sym__thematic_break] = ACTIONS(1726),
    [sym__list_marker_minus] = ACTIONS(1726),
    [sym__list_marker_plus] = ACTIONS(1726),
    [sym__list_marker_star] = ACTIONS(1726),
    [sym__list_marker_parenthesis] = ACTIONS(1726),
    [sym__list_marker_dot] = ACTIONS(1726),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1726),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1726),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1726),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1726),
    [sym__blank_line_start] = ACTIONS(1726),
    [sym__html_block_1_start] = ACTIONS(1726),
    [sym__html_block_2_start] = ACTIONS(1726),
    [sym__html_block_3_start] = ACTIONS(1726),
    [sym__html_block_4_start] = ACTIONS(1726),
    [sym__html_block_5_start] = ACTIONS(1726),
    [sym__html_block_6_start] = ACTIONS(1726),
    [sym__html_block_7_start] = ACTIONS(1726),
    [sym__pipe_table_start] = ACTIONS(1726),
  },
  [STATE(286)] = {
    [anon_sym_LBRACK] = ACTIONS(1734),
    [anon_sym_RBRACK] = ACTIONS(1736),
    [anon_sym_LT] = ACTIONS(1736),
    [anon_sym_GT] = ACTIONS(1736),
    [anon_sym_BANG] = ACTIONS(1736),
    [anon_sym_DQUOTE] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_DOLLAR] = ACTIONS(1736),
    [anon_sym_PERCENT] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1736),
    [anon_sym_STAR] = ACTIONS(1736),
    [anon_sym_PLUS] = ACTIONS(1736),
    [anon_sym_COMMA] = ACTIONS(1736),
    [anon_sym_DASH] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1736),
    [anon_sym_SLASH] = ACTIONS(1736),
    [anon_sym_COLON] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym_EQ] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_AT] = ACTIONS(1736),
    [anon_sym_BSLASH] = ACTIONS(1736),
    [anon_sym_CARET] = ACTIONS(1736),
    [anon_sym__] = ACTIONS(1736),
    [anon_sym_BQUOTE] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_RBRACE] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1736),
    [anon_sym_RPAREN] = ACTIONS(1736),
    [aux_sym__word_token1] = ACTIONS(1736),
    [aux_sym__word_token2] = ACTIONS(1736),
    [aux_sym__word_token3] = ACTIONS(1736),
    [sym__whitespace] = ACTIONS(1736),
    [sym__soft_line_ending] = ACTIONS(1736),
    [sym__block_close] = ACTIONS(1736),
    [sym__block_quote_start] = ACTIONS(1736),
    [sym__indented_chunk_start] = ACTIONS(1736),
    [sym_atx_h1_marker] = ACTIONS(1736),
    [sym_atx_h2_marker] = ACTIONS(1736),
    [sym_atx_h3_marker] = ACTIONS(1736),
    [sym_atx_h4_marker] = ACTIONS(1736),
    [sym_atx_h5_marker] = ACTIONS(1736),
    [sym_atx_h6_marker] = ACTIONS(1736),
    [sym__thematic_break] = ACTIONS(1736),
    [sym__list_marker_minus] = ACTIONS(1736),
    [sym__list_marker_plus] = ACTIONS(1736),
    [sym__list_marker_star] = ACTIONS(1736),
    [sym__list_marker_parenthesis] = ACTIONS(1736),
    [sym__list_marker_dot] = ACTIONS(1736),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1736),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1736),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1736),
    [sym__blank_line_start] = ACTIONS(1736),
    [sym__html_block_1_start] = ACTIONS(1736),
    [sym__html_block_2_start] = ACTIONS(1736),
    [sym__html_block_3_start] = ACTIONS(1736),
    [sym__html_block_4_start] = ACTIONS(1736),
    [sym__html_block_5_start] = ACTIONS(1736),
    [sym__html_block_6_start] = ACTIONS(1736),
    [sym__html_block_7_start] = ACTIONS(1736),
    [sym__pipe_table_start] = ACTIONS(1736),
  },
  [STATE(287)] = {
    [anon_sym_LBRACK] = ACTIONS(1147),
    [anon_sym_RBRACK] = ACTIONS(1145),
    [anon_sym_LT] = ACTIONS(1145),
    [anon_sym_GT] = ACTIONS(1145),
    [anon_sym_BANG] = ACTIONS(1145),
    [anon_sym_DQUOTE] = ACTIONS(1145),
    [anon_sym_POUND] = ACTIONS(1145),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [anon_sym_PERCENT] = ACTIONS(1145),
    [anon_sym_AMP] = ACTIONS(1145),
    [anon_sym_SQUOTE] = ACTIONS(1145),
    [anon_sym_STAR] = ACTIONS(1145),
    [anon_sym_PLUS] = ACTIONS(1145),
    [anon_sym_COMMA] = ACTIONS(1145),
    [anon_sym_DASH] = ACTIONS(1145),
    [anon_sym_DOT] = ACTIONS(1145),
    [anon_sym_SLASH] = ACTIONS(1145),
    [anon_sym_COLON] = ACTIONS(1145),
    [anon_sym_SEMI] = ACTIONS(1145),
    [anon_sym_EQ] = ACTIONS(1145),
    [anon_sym_QMARK] = ACTIONS(1145),
    [anon_sym_AT] = ACTIONS(1145),
    [anon_sym_BSLASH] = ACTIONS(1145),
    [anon_sym_CARET] = ACTIONS(1145),
    [anon_sym__] = ACTIONS(1145),
    [anon_sym_BQUOTE] = ACTIONS(1145),
    [anon_sym_LBRACE] = ACTIONS(1145),
    [anon_sym_PIPE] = ACTIONS(1145),
    [anon_sym_RBRACE] = ACTIONS(1145),
    [anon_sym_TILDE] = ACTIONS(1145),
    [anon_sym_LPAREN] = ACTIONS(1145),
    [anon_sym_RPAREN] = ACTIONS(1145),
    [aux_sym__word_token1] = ACTIONS(1145),
    [aux_sym__word_token2] = ACTIONS(1145),
    [aux_sym__word_token3] = ACTIONS(1145),
    [sym__whitespace] = ACTIONS(1145),
    [sym__soft_line_ending] = ACTIONS(1145),
    [sym__block_close] = ACTIONS(1145),
    [sym__block_quote_start] = ACTIONS(1145),
    [sym__indented_chunk_start] = ACTIONS(1145),
    [sym_atx_h1_marker] = ACTIONS(1145),
    [sym_atx_h2_marker] = ACTIONS(1145),
    [sym_atx_h3_marker] = ACTIONS(1145),
    [sym_atx_h4_marker] = ACTIONS(1145),
    [sym_atx_h5_marker] = ACTIONS(1145),
    [sym_atx_h6_marker] = ACTIONS(1145),
    [sym__thematic_break] = ACTIONS(1145),
    [sym__list_marker_minus] = ACTIONS(1145),
    [sym__list_marker_plus] = ACTIONS(1145),
    [sym__list_marker_star] = ACTIONS(1145),
    [sym__list_marker_parenthesis] = ACTIONS(1145),
    [sym__list_marker_dot] = ACTIONS(1145),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1145),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1145),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1145),
    [sym__blank_line_start] = ACTIONS(1145),
    [sym__html_block_1_start] = ACTIONS(1145),
    [sym__html_block_2_start] = ACTIONS(1145),
    [sym__html_block_3_start] = ACTIONS(1145),
    [sym__html_block_4_start] = ACTIONS(1145),
    [sym__html_block_5_start] = ACTIONS(1145),
    [sym__html_block_6_start] = ACTIONS(1145),
    [sym__html_block_7_start] = ACTIONS(1145),
    [sym__pipe_table_start] = ACTIONS(1145),
  },
  [STATE(288)] = {
    [ts_builtin_sym_end] = ACTIONS(1736),
    [anon_sym_LBRACK] = ACTIONS(1734),
    [anon_sym_RBRACK] = ACTIONS(1736),
    [anon_sym_LT] = ACTIONS(1736),
    [anon_sym_GT] = ACTIONS(1736),
    [anon_sym_BANG] = ACTIONS(1736),
    [anon_sym_DQUOTE] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_DOLLAR] = ACTIONS(1736),
    [anon_sym_PERCENT] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1736),
    [anon_sym_STAR] = ACTIONS(1736),
    [anon_sym_PLUS] = ACTIONS(1736),
    [anon_sym_COMMA] = ACTIONS(1736),
    [anon_sym_DASH] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1736),
    [anon_sym_SLASH] = ACTIONS(1736),
    [anon_sym_COLON] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym_EQ] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_AT] = ACTIONS(1736),
    [anon_sym_BSLASH] = ACTIONS(1736),
    [anon_sym_CARET] = ACTIONS(1736),
    [anon_sym__] = ACTIONS(1736),
    [anon_sym_BQUOTE] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_RBRACE] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1736),
    [anon_sym_RPAREN] = ACTIONS(1736),
    [aux_sym__word_token1] = ACTIONS(1736),
    [aux_sym__word_token2] = ACTIONS(1736),
    [aux_sym__word_token3] = ACTIONS(1736),
    [sym__whitespace] = ACTIONS(1736),
    [sym__soft_line_ending] = ACTIONS(1736),
    [sym__block_quote_start] = ACTIONS(1736),
    [sym__indented_chunk_start] = ACTIONS(1736),
    [sym_atx_h1_marker] = ACTIONS(1736),
    [sym_atx_h2_marker] = ACTIONS(1736),
    [sym_atx_h3_marker] = ACTIONS(1736),
    [sym_atx_h4_marker] = ACTIONS(1736),
    [sym_atx_h5_marker] = ACTIONS(1736),
    [sym_atx_h6_marker] = ACTIONS(1736),
    [sym__thematic_break] = ACTIONS(1736),
    [sym__list_marker_minus] = ACTIONS(1736),
    [sym__list_marker_plus] = ACTIONS(1736),
    [sym__list_marker_star] = ACTIONS(1736),
    [sym__list_marker_parenthesis] = ACTIONS(1736),
    [sym__list_marker_dot] = ACTIONS(1736),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1736),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1736),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1736),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1736),
    [sym__blank_line_start] = ACTIONS(1736),
    [sym__html_block_1_start] = ACTIONS(1736),
    [sym__html_block_2_start] = ACTIONS(1736),
    [sym__html_block_3_start] = ACTIONS(1736),
    [sym__html_block_4_start] = ACTIONS(1736),
    [sym__html_block_5_start] = ACTIONS(1736),
    [sym__html_block_6_start] = ACTIONS(1736),
    [sym__html_block_7_start] = ACTIONS(1736),
    [sym__pipe_table_start] = ACTIONS(1736),
  },
  [STATE(289)] = {
    [anon_sym_LBRACK] = ACTIONS(1438),
    [anon_sym_RBRACK] = ACTIONS(1440),
    [anon_sym_LT] = ACTIONS(1440),
    [anon_sym_GT] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_DQUOTE] = ACTIONS(1440),
    [anon_sym_POUND] = ACTIONS(1440),
    [anon_sym_DOLLAR] = ACTIONS(1440),
    [anon_sym_PERCENT] = ACTIONS(1440),
    [anon_sym_AMP] = ACTIONS(1440),
    [anon_sym_SQUOTE] = ACTIONS(1440),
    [anon_sym_STAR] = ACTIONS(1440),
    [anon_sym_PLUS] = ACTIONS(1440),
    [anon_sym_COMMA] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_DOT] = ACTIONS(1440),
    [anon_sym_SLASH] = ACTIONS(1440),
    [anon_sym_COLON] = ACTIONS(1440),
    [anon_sym_SEMI] = ACTIONS(1440),
    [anon_sym_EQ] = ACTIONS(1440),
    [anon_sym_QMARK] = ACTIONS(1440),
    [anon_sym_AT] = ACTIONS(1440),
    [anon_sym_BSLASH] = ACTIONS(1440),
    [anon_sym_CARET] = ACTIONS(1440),
    [anon_sym__] = ACTIONS(1440),
    [anon_sym_BQUOTE] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_PIPE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [anon_sym_TILDE] = ACTIONS(1440),
    [anon_sym_LPAREN] = ACTIONS(1440),
    [anon_sym_RPAREN] = ACTIONS(1440),
    [aux_sym__word_token1] = ACTIONS(1440),
    [aux_sym__word_token2] = ACTIONS(1440),
    [aux_sym__word_token3] = ACTIONS(1440),
    [sym__whitespace] = ACTIONS(1440),
    [sym__soft_line_ending] = ACTIONS(1440),
    [sym__block_close] = ACTIONS(1440),
    [sym__block_quote_start] = ACTIONS(1440),
    [sym__indented_chunk_start] = ACTIONS(1440),
    [sym_atx_h1_marker] = ACTIONS(1440),
    [sym_atx_h2_marker] = ACTIONS(1440),
    [sym_atx_h3_marker] = ACTIONS(1440),
    [sym_atx_h4_marker] = ACTIONS(1440),
    [sym_atx_h5_marker] = ACTIONS(1440),
    [sym_atx_h6_marker] = ACTIONS(1440),
    [sym__thematic_break] = ACTIONS(1440),
    [sym__list_marker_minus] = ACTIONS(1440),
    [sym__list_marker_plus] = ACTIONS(1440),
    [sym__list_marker_star] = ACTIONS(1440),
    [sym__list_marker_parenthesis] = ACTIONS(1440),
    [sym__list_marker_dot] = ACTIONS(1440),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1440),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1440),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1440),
    [sym__blank_line_start] = ACTIONS(1440),
    [sym__html_block_1_start] = ACTIONS(1440),
    [sym__html_block_2_start] = ACTIONS(1440),
    [sym__html_block_3_start] = ACTIONS(1440),
    [sym__html_block_4_start] = ACTIONS(1440),
    [sym__html_block_5_start] = ACTIONS(1440),
    [sym__html_block_6_start] = ACTIONS(1440),
    [sym__html_block_7_start] = ACTIONS(1440),
    [sym__pipe_table_start] = ACTIONS(1440),
  },
  [STATE(290)] = {
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_RBRACK] = ACTIONS(1294),
    [anon_sym_LT] = ACTIONS(1294),
    [anon_sym_GT] = ACTIONS(1294),
    [anon_sym_BANG] = ACTIONS(1294),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [anon_sym_POUND] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1294),
    [anon_sym_PERCENT] = ACTIONS(1294),
    [anon_sym_AMP] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_COMMA] = ACTIONS(1294),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_DOT] = ACTIONS(1294),
    [anon_sym_SLASH] = ACTIONS(1294),
    [anon_sym_COLON] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1294),
    [anon_sym_EQ] = ACTIONS(1294),
    [anon_sym_QMARK] = ACTIONS(1294),
    [anon_sym_AT] = ACTIONS(1294),
    [anon_sym_BSLASH] = ACTIONS(1294),
    [anon_sym_CARET] = ACTIONS(1294),
    [anon_sym__] = ACTIONS(1294),
    [anon_sym_BQUOTE] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1294),
    [anon_sym_PIPE] = ACTIONS(1294),
    [anon_sym_RBRACE] = ACTIONS(1294),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(1294),
    [anon_sym_RPAREN] = ACTIONS(1294),
    [aux_sym__word_token1] = ACTIONS(1294),
    [aux_sym__word_token2] = ACTIONS(1294),
    [aux_sym__word_token3] = ACTIONS(1294),
    [sym__whitespace] = ACTIONS(1294),
    [sym__soft_line_ending] = ACTIONS(1294),
    [sym__block_close] = ACTIONS(1294),
    [sym__block_quote_start] = ACTIONS(1294),
    [sym__indented_chunk_start] = ACTIONS(1294),
    [sym_atx_h1_marker] = ACTIONS(1294),
    [sym_atx_h2_marker] = ACTIONS(1294),
    [sym_atx_h3_marker] = ACTIONS(1294),
    [sym_atx_h4_marker] = ACTIONS(1294),
    [sym_atx_h5_marker] = ACTIONS(1294),
    [sym_atx_h6_marker] = ACTIONS(1294),
    [sym__thematic_break] = ACTIONS(1294),
    [sym__list_marker_minus] = ACTIONS(1294),
    [sym__list_marker_plus] = ACTIONS(1294),
    [sym__list_marker_star] = ACTIONS(1294),
    [sym__list_marker_parenthesis] = ACTIONS(1294),
    [sym__list_marker_dot] = ACTIONS(1294),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1294),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1294),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1294),
    [sym__blank_line_start] = ACTIONS(1294),
    [sym__html_block_1_start] = ACTIONS(1294),
    [sym__html_block_2_start] = ACTIONS(1294),
    [sym__html_block_3_start] = ACTIONS(1294),
    [sym__html_block_4_start] = ACTIONS(1294),
    [sym__html_block_5_start] = ACTIONS(1294),
    [sym__html_block_6_start] = ACTIONS(1294),
    [sym__html_block_7_start] = ACTIONS(1294),
    [sym__pipe_table_start] = ACTIONS(1294),
  },
  [STATE(291)] = {
    [ts_builtin_sym_end] = ACTIONS(1145),
    [anon_sym_LBRACK] = ACTIONS(1147),
    [anon_sym_RBRACK] = ACTIONS(1145),
    [anon_sym_LT] = ACTIONS(1145),
    [anon_sym_GT] = ACTIONS(1145),
    [anon_sym_BANG] = ACTIONS(1145),
    [anon_sym_DQUOTE] = ACTIONS(1145),
    [anon_sym_POUND] = ACTIONS(1145),
    [anon_sym_DOLLAR] = ACTIONS(1145),
    [anon_sym_PERCENT] = ACTIONS(1145),
    [anon_sym_AMP] = ACTIONS(1145),
    [anon_sym_SQUOTE] = ACTIONS(1145),
    [anon_sym_STAR] = ACTIONS(1145),
    [anon_sym_PLUS] = ACTIONS(1145),
    [anon_sym_COMMA] = ACTIONS(1145),
    [anon_sym_DASH] = ACTIONS(1145),
    [anon_sym_DOT] = ACTIONS(1145),
    [anon_sym_SLASH] = ACTIONS(1145),
    [anon_sym_COLON] = ACTIONS(1145),
    [anon_sym_SEMI] = ACTIONS(1145),
    [anon_sym_EQ] = ACTIONS(1145),
    [anon_sym_QMARK] = ACTIONS(1145),
    [anon_sym_AT] = ACTIONS(1145),
    [anon_sym_BSLASH] = ACTIONS(1145),
    [anon_sym_CARET] = ACTIONS(1145),
    [anon_sym__] = ACTIONS(1145),
    [anon_sym_BQUOTE] = ACTIONS(1145),
    [anon_sym_LBRACE] = ACTIONS(1145),
    [anon_sym_PIPE] = ACTIONS(1145),
    [anon_sym_RBRACE] = ACTIONS(1145),
    [anon_sym_TILDE] = ACTIONS(1145),
    [anon_sym_LPAREN] = ACTIONS(1145),
    [anon_sym_RPAREN] = ACTIONS(1145),
    [aux_sym__word_token1] = ACTIONS(1145),
    [aux_sym__word_token2] = ACTIONS(1145),
    [aux_sym__word_token3] = ACTIONS(1145),
    [sym__whitespace] = ACTIONS(1145),
    [sym__soft_line_ending] = ACTIONS(1145),
    [sym__block_quote_start] = ACTIONS(1145),
    [sym__indented_chunk_start] = ACTIONS(1145),
    [sym_atx_h1_marker] = ACTIONS(1145),
    [sym_atx_h2_marker] = ACTIONS(1145),
    [sym_atx_h3_marker] = ACTIONS(1145),
    [sym_atx_h4_marker] = ACTIONS(1145),
    [sym_atx_h5_marker] = ACTIONS(1145),
    [sym_atx_h6_marker] = ACTIONS(1145),
    [sym__thematic_break] = ACTIONS(1145),
    [sym__list_marker_minus] = ACTIONS(1145),
    [sym__list_marker_plus] = ACTIONS(1145),
    [sym__list_marker_star] = ACTIONS(1145),
    [sym__list_marker_parenthesis] = ACTIONS(1145),
    [sym__list_marker_dot] = ACTIONS(1145),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1145),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1145),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1145),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1145),
    [sym__blank_line_start] = ACTIONS(1145),
    [sym__html_block_1_start] = ACTIONS(1145),
    [sym__html_block_2_start] = ACTIONS(1145),
    [sym__html_block_3_start] = ACTIONS(1145),
    [sym__html_block_4_start] = ACTIONS(1145),
    [sym__html_block_5_start] = ACTIONS(1145),
    [sym__html_block_6_start] = ACTIONS(1145),
    [sym__html_block_7_start] = ACTIONS(1145),
    [sym__pipe_table_start] = ACTIONS(1145),
  },
  [STATE(292)] = {
    [anon_sym_LBRACK] = ACTIONS(1302),
    [anon_sym_RBRACK] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [anon_sym_POUND] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_AT] = ACTIONS(1300),
    [anon_sym_BSLASH] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym__] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_RPAREN] = ACTIONS(1300),
    [aux_sym__word_token1] = ACTIONS(1300),
    [aux_sym__word_token2] = ACTIONS(1300),
    [aux_sym__word_token3] = ACTIONS(1300),
    [sym__whitespace] = ACTIONS(1300),
    [sym__soft_line_ending] = ACTIONS(1300),
    [sym__block_close] = ACTIONS(1300),
    [sym__block_quote_start] = ACTIONS(1300),
    [sym__indented_chunk_start] = ACTIONS(1300),
    [sym_atx_h1_marker] = ACTIONS(1300),
    [sym_atx_h2_marker] = ACTIONS(1300),
    [sym_atx_h3_marker] = ACTIONS(1300),
    [sym_atx_h4_marker] = ACTIONS(1300),
    [sym_atx_h5_marker] = ACTIONS(1300),
    [sym_atx_h6_marker] = ACTIONS(1300),
    [sym__thematic_break] = ACTIONS(1300),
    [sym__list_marker_minus] = ACTIONS(1300),
    [sym__list_marker_plus] = ACTIONS(1300),
    [sym__list_marker_star] = ACTIONS(1300),
    [sym__list_marker_parenthesis] = ACTIONS(1300),
    [sym__list_marker_dot] = ACTIONS(1300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1300),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1300),
    [sym__blank_line_start] = ACTIONS(1300),
    [sym__html_block_1_start] = ACTIONS(1300),
    [sym__html_block_2_start] = ACTIONS(1300),
    [sym__html_block_3_start] = ACTIONS(1300),
    [sym__html_block_4_start] = ACTIONS(1300),
    [sym__html_block_5_start] = ACTIONS(1300),
    [sym__html_block_6_start] = ACTIONS(1300),
    [sym__html_block_7_start] = ACTIONS(1300),
    [sym__pipe_table_start] = ACTIONS(1300),
  },
  [STATE(293)] = {
    [ts_builtin_sym_end] = ACTIONS(1648),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_LT] = ACTIONS(1648),
    [anon_sym_GT] = ACTIONS(1648),
    [anon_sym_BANG] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [anon_sym_POUND] = ACTIONS(1648),
    [anon_sym_DOLLAR] = ACTIONS(1648),
    [anon_sym_PERCENT] = ACTIONS(1648),
    [anon_sym_AMP] = ACTIONS(1648),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_STAR] = ACTIONS(1648),
    [anon_sym_PLUS] = ACTIONS(1648),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_DASH] = ACTIONS(1648),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_SLASH] = ACTIONS(1648),
    [anon_sym_COLON] = ACTIONS(1648),
    [anon_sym_SEMI] = ACTIONS(1648),
    [anon_sym_EQ] = ACTIONS(1648),
    [anon_sym_QMARK] = ACTIONS(1648),
    [anon_sym_AT] = ACTIONS(1648),
    [anon_sym_BSLASH] = ACTIONS(1648),
    [anon_sym_CARET] = ACTIONS(1648),
    [anon_sym__] = ACTIONS(1648),
    [anon_sym_BQUOTE] = ACTIONS(1648),
    [anon_sym_LBRACE] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1648),
    [anon_sym_RBRACE] = ACTIONS(1648),
    [anon_sym_TILDE] = ACTIONS(1648),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_RPAREN] = ACTIONS(1648),
    [aux_sym__word_token1] = ACTIONS(1648),
    [aux_sym__word_token2] = ACTIONS(1648),
    [aux_sym__word_token3] = ACTIONS(1648),
    [sym__whitespace] = ACTIONS(1648),
    [sym__soft_line_ending] = ACTIONS(1648),
    [sym__block_quote_start] = ACTIONS(1648),
    [sym__indented_chunk_start] = ACTIONS(1648),
    [sym_atx_h1_marker] = ACTIONS(1648),
    [sym_atx_h2_marker] = ACTIONS(1648),
    [sym_atx_h3_marker] = ACTIONS(1648),
    [sym_atx_h4_marker] = ACTIONS(1648),
    [sym_atx_h5_marker] = ACTIONS(1648),
    [sym_atx_h6_marker] = ACTIONS(1648),
    [sym__thematic_break] = ACTIONS(1648),
    [sym__list_marker_minus] = ACTIONS(1648),
    [sym__list_marker_plus] = ACTIONS(1648),
    [sym__list_marker_star] = ACTIONS(1648),
    [sym__list_marker_parenthesis] = ACTIONS(1648),
    [sym__list_marker_dot] = ACTIONS(1648),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1648),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1648),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1648),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1648),
    [sym__blank_line_start] = ACTIONS(1648),
    [sym__html_block_1_start] = ACTIONS(1648),
    [sym__html_block_2_start] = ACTIONS(1648),
    [sym__html_block_3_start] = ACTIONS(1648),
    [sym__html_block_4_start] = ACTIONS(1648),
    [sym__html_block_5_start] = ACTIONS(1648),
    [sym__html_block_6_start] = ACTIONS(1648),
    [sym__html_block_7_start] = ACTIONS(1648),
    [sym__pipe_table_start] = ACTIONS(1648),
  },
  [STATE(294)] = {
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_RBRACK] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1318),
    [anon_sym_GT] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1318),
    [anon_sym_BSLASH] = ACTIONS(1318),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym__] = ACTIONS(1318),
    [anon_sym_BQUOTE] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_RPAREN] = ACTIONS(1318),
    [aux_sym__word_token1] = ACTIONS(1318),
    [aux_sym__word_token2] = ACTIONS(1318),
    [aux_sym__word_token3] = ACTIONS(1318),
    [sym__whitespace] = ACTIONS(1318),
    [sym__soft_line_ending] = ACTIONS(1318),
    [sym__block_close] = ACTIONS(1318),
    [sym__block_quote_start] = ACTIONS(1318),
    [sym__indented_chunk_start] = ACTIONS(1318),
    [sym_atx_h1_marker] = ACTIONS(1318),
    [sym_atx_h2_marker] = ACTIONS(1318),
    [sym_atx_h3_marker] = ACTIONS(1318),
    [sym_atx_h4_marker] = ACTIONS(1318),
    [sym_atx_h5_marker] = ACTIONS(1318),
    [sym_atx_h6_marker] = ACTIONS(1318),
    [sym__thematic_break] = ACTIONS(1318),
    [sym__list_marker_minus] = ACTIONS(1318),
    [sym__list_marker_plus] = ACTIONS(1318),
    [sym__list_marker_star] = ACTIONS(1318),
    [sym__list_marker_parenthesis] = ACTIONS(1318),
    [sym__list_marker_dot] = ACTIONS(1318),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1318),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1318),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1318),
    [sym__blank_line_start] = ACTIONS(1318),
    [sym__html_block_1_start] = ACTIONS(1318),
    [sym__html_block_2_start] = ACTIONS(1318),
    [sym__html_block_3_start] = ACTIONS(1318),
    [sym__html_block_4_start] = ACTIONS(1318),
    [sym__html_block_5_start] = ACTIONS(1318),
    [sym__html_block_6_start] = ACTIONS(1318),
    [sym__html_block_7_start] = ACTIONS(1318),
    [sym__pipe_table_start] = ACTIONS(1318),
  },
  [STATE(295)] = {
    [ts_builtin_sym_end] = ACTIONS(1652),
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_LT] = ACTIONS(1652),
    [anon_sym_GT] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [anon_sym_POUND] = ACTIONS(1652),
    [anon_sym_DOLLAR] = ACTIONS(1652),
    [anon_sym_PERCENT] = ACTIONS(1652),
    [anon_sym_AMP] = ACTIONS(1652),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_STAR] = ACTIONS(1652),
    [anon_sym_PLUS] = ACTIONS(1652),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(1652),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_SLASH] = ACTIONS(1652),
    [anon_sym_COLON] = ACTIONS(1652),
    [anon_sym_SEMI] = ACTIONS(1652),
    [anon_sym_EQ] = ACTIONS(1652),
    [anon_sym_QMARK] = ACTIONS(1652),
    [anon_sym_AT] = ACTIONS(1652),
    [anon_sym_BSLASH] = ACTIONS(1652),
    [anon_sym_CARET] = ACTIONS(1652),
    [anon_sym__] = ACTIONS(1652),
    [anon_sym_BQUOTE] = ACTIONS(1652),
    [anon_sym_LBRACE] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1652),
    [anon_sym_RBRACE] = ACTIONS(1652),
    [anon_sym_TILDE] = ACTIONS(1652),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_RPAREN] = ACTIONS(1652),
    [aux_sym__word_token1] = ACTIONS(1652),
    [aux_sym__word_token2] = ACTIONS(1652),
    [aux_sym__word_token3] = ACTIONS(1652),
    [sym__whitespace] = ACTIONS(1652),
    [sym__soft_line_ending] = ACTIONS(1652),
    [sym__block_quote_start] = ACTIONS(1652),
    [sym__indented_chunk_start] = ACTIONS(1652),
    [sym_atx_h1_marker] = ACTIONS(1652),
    [sym_atx_h2_marker] = ACTIONS(1652),
    [sym_atx_h3_marker] = ACTIONS(1652),
    [sym_atx_h4_marker] = ACTIONS(1652),
    [sym_atx_h5_marker] = ACTIONS(1652),
    [sym_atx_h6_marker] = ACTIONS(1652),
    [sym__thematic_break] = ACTIONS(1652),
    [sym__list_marker_minus] = ACTIONS(1652),
    [sym__list_marker_plus] = ACTIONS(1652),
    [sym__list_marker_star] = ACTIONS(1652),
    [sym__list_marker_parenthesis] = ACTIONS(1652),
    [sym__list_marker_dot] = ACTIONS(1652),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1652),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1652),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1652),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1652),
    [sym__blank_line_start] = ACTIONS(1652),
    [sym__html_block_1_start] = ACTIONS(1652),
    [sym__html_block_2_start] = ACTIONS(1652),
    [sym__html_block_3_start] = ACTIONS(1652),
    [sym__html_block_4_start] = ACTIONS(1652),
    [sym__html_block_5_start] = ACTIONS(1652),
    [sym__html_block_6_start] = ACTIONS(1652),
    [sym__html_block_7_start] = ACTIONS(1652),
    [sym__pipe_table_start] = ACTIONS(1652),
  },
  [STATE(296)] = {
    [anon_sym_LBRACK] = ACTIONS(1326),
    [anon_sym_RBRACK] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_POUND] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_COLON] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_BSLASH] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_RPAREN] = ACTIONS(1324),
    [aux_sym__word_token1] = ACTIONS(1324),
    [aux_sym__word_token2] = ACTIONS(1324),
    [aux_sym__word_token3] = ACTIONS(1324),
    [sym__whitespace] = ACTIONS(1324),
    [sym__soft_line_ending] = ACTIONS(1324),
    [sym__block_close] = ACTIONS(1324),
    [sym__block_quote_start] = ACTIONS(1324),
    [sym__indented_chunk_start] = ACTIONS(1324),
    [sym_atx_h1_marker] = ACTIONS(1324),
    [sym_atx_h2_marker] = ACTIONS(1324),
    [sym_atx_h3_marker] = ACTIONS(1324),
    [sym_atx_h4_marker] = ACTIONS(1324),
    [sym_atx_h5_marker] = ACTIONS(1324),
    [sym_atx_h6_marker] = ACTIONS(1324),
    [sym__thematic_break] = ACTIONS(1324),
    [sym__list_marker_minus] = ACTIONS(1324),
    [sym__list_marker_plus] = ACTIONS(1324),
    [sym__list_marker_star] = ACTIONS(1324),
    [sym__list_marker_parenthesis] = ACTIONS(1324),
    [sym__list_marker_dot] = ACTIONS(1324),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1324),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1324),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1324),
    [sym__blank_line_start] = ACTIONS(1324),
    [sym__html_block_1_start] = ACTIONS(1324),
    [sym__html_block_2_start] = ACTIONS(1324),
    [sym__html_block_3_start] = ACTIONS(1324),
    [sym__html_block_4_start] = ACTIONS(1324),
    [sym__html_block_5_start] = ACTIONS(1324),
    [sym__html_block_6_start] = ACTIONS(1324),
    [sym__html_block_7_start] = ACTIONS(1324),
    [sym__pipe_table_start] = ACTIONS(1324),
  },
  [STATE(297)] = {
    [anon_sym_LBRACK] = ACTIONS(1738),
    [anon_sym_RBRACK] = ACTIONS(1740),
    [anon_sym_LT] = ACTIONS(1740),
    [anon_sym_GT] = ACTIONS(1740),
    [anon_sym_BANG] = ACTIONS(1740),
    [anon_sym_DQUOTE] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_DOLLAR] = ACTIONS(1740),
    [anon_sym_PERCENT] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1740),
    [anon_sym_STAR] = ACTIONS(1740),
    [anon_sym_PLUS] = ACTIONS(1740),
    [anon_sym_COMMA] = ACTIONS(1740),
    [anon_sym_DASH] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1740),
    [anon_sym_SLASH] = ACTIONS(1740),
    [anon_sym_COLON] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1740),
    [anon_sym_EQ] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_AT] = ACTIONS(1740),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_CARET] = ACTIONS(1740),
    [anon_sym__] = ACTIONS(1740),
    [anon_sym_BQUOTE] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_RBRACE] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1740),
    [anon_sym_RPAREN] = ACTIONS(1740),
    [aux_sym__word_token1] = ACTIONS(1740),
    [aux_sym__word_token2] = ACTIONS(1740),
    [aux_sym__word_token3] = ACTIONS(1740),
    [sym__whitespace] = ACTIONS(1740),
    [sym__soft_line_ending] = ACTIONS(1740),
    [sym__block_close] = ACTIONS(1740),
    [sym__block_quote_start] = ACTIONS(1740),
    [sym__indented_chunk_start] = ACTIONS(1740),
    [sym_atx_h1_marker] = ACTIONS(1740),
    [sym_atx_h2_marker] = ACTIONS(1740),
    [sym_atx_h3_marker] = ACTIONS(1740),
    [sym_atx_h4_marker] = ACTIONS(1740),
    [sym_atx_h5_marker] = ACTIONS(1740),
    [sym_atx_h6_marker] = ACTIONS(1740),
    [sym__thematic_break] = ACTIONS(1740),
    [sym__list_marker_minus] = ACTIONS(1740),
    [sym__list_marker_plus] = ACTIONS(1740),
    [sym__list_marker_star] = ACTIONS(1740),
    [sym__list_marker_parenthesis] = ACTIONS(1740),
    [sym__list_marker_dot] = ACTIONS(1740),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1740),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1740),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1740),
    [sym__blank_line_start] = ACTIONS(1740),
    [sym__html_block_1_start] = ACTIONS(1740),
    [sym__html_block_2_start] = ACTIONS(1740),
    [sym__html_block_3_start] = ACTIONS(1740),
    [sym__html_block_4_start] = ACTIONS(1740),
    [sym__html_block_5_start] = ACTIONS(1740),
    [sym__html_block_6_start] = ACTIONS(1740),
    [sym__html_block_7_start] = ACTIONS(1740),
    [sym__pipe_table_start] = ACTIONS(1740),
  },
  [STATE(298)] = {
    [anon_sym_LBRACK] = ACTIONS(1742),
    [anon_sym_RBRACK] = ACTIONS(1744),
    [anon_sym_LT] = ACTIONS(1744),
    [anon_sym_GT] = ACTIONS(1744),
    [anon_sym_BANG] = ACTIONS(1744),
    [anon_sym_DQUOTE] = ACTIONS(1744),
    [anon_sym_POUND] = ACTIONS(1744),
    [anon_sym_DOLLAR] = ACTIONS(1744),
    [anon_sym_PERCENT] = ACTIONS(1744),
    [anon_sym_AMP] = ACTIONS(1744),
    [anon_sym_SQUOTE] = ACTIONS(1744),
    [anon_sym_STAR] = ACTIONS(1744),
    [anon_sym_PLUS] = ACTIONS(1744),
    [anon_sym_COMMA] = ACTIONS(1744),
    [anon_sym_DASH] = ACTIONS(1744),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_SLASH] = ACTIONS(1744),
    [anon_sym_COLON] = ACTIONS(1744),
    [anon_sym_SEMI] = ACTIONS(1744),
    [anon_sym_EQ] = ACTIONS(1744),
    [anon_sym_QMARK] = ACTIONS(1744),
    [anon_sym_AT] = ACTIONS(1744),
    [anon_sym_BSLASH] = ACTIONS(1744),
    [anon_sym_CARET] = ACTIONS(1744),
    [anon_sym__] = ACTIONS(1744),
    [anon_sym_BQUOTE] = ACTIONS(1744),
    [anon_sym_LBRACE] = ACTIONS(1744),
    [anon_sym_PIPE] = ACTIONS(1744),
    [anon_sym_RBRACE] = ACTIONS(1744),
    [anon_sym_TILDE] = ACTIONS(1744),
    [anon_sym_LPAREN] = ACTIONS(1744),
    [anon_sym_RPAREN] = ACTIONS(1744),
    [aux_sym__word_token1] = ACTIONS(1744),
    [aux_sym__word_token2] = ACTIONS(1744),
    [aux_sym__word_token3] = ACTIONS(1744),
    [sym__whitespace] = ACTIONS(1744),
    [sym__soft_line_ending] = ACTIONS(1744),
    [sym__block_close] = ACTIONS(1744),
    [sym__block_quote_start] = ACTIONS(1744),
    [sym__indented_chunk_start] = ACTIONS(1744),
    [sym_atx_h1_marker] = ACTIONS(1744),
    [sym_atx_h2_marker] = ACTIONS(1744),
    [sym_atx_h3_marker] = ACTIONS(1744),
    [sym_atx_h4_marker] = ACTIONS(1744),
    [sym_atx_h5_marker] = ACTIONS(1744),
    [sym_atx_h6_marker] = ACTIONS(1744),
    [sym__thematic_break] = ACTIONS(1744),
    [sym__list_marker_minus] = ACTIONS(1744),
    [sym__list_marker_plus] = ACTIONS(1744),
    [sym__list_marker_star] = ACTIONS(1744),
    [sym__list_marker_parenthesis] = ACTIONS(1744),
    [sym__list_marker_dot] = ACTIONS(1744),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1744),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1744),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1744),
    [sym__blank_line_start] = ACTIONS(1744),
    [sym__html_block_1_start] = ACTIONS(1744),
    [sym__html_block_2_start] = ACTIONS(1744),
    [sym__html_block_3_start] = ACTIONS(1744),
    [sym__html_block_4_start] = ACTIONS(1744),
    [sym__html_block_5_start] = ACTIONS(1744),
    [sym__html_block_6_start] = ACTIONS(1744),
    [sym__html_block_7_start] = ACTIONS(1744),
    [sym__pipe_table_start] = ACTIONS(1744),
  },
  [STATE(299)] = {
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_LT] = ACTIONS(1748),
    [anon_sym_GT] = ACTIONS(1748),
    [anon_sym_BANG] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [anon_sym_POUND] = ACTIONS(1748),
    [anon_sym_DOLLAR] = ACTIONS(1748),
    [anon_sym_PERCENT] = ACTIONS(1748),
    [anon_sym_AMP] = ACTIONS(1748),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_STAR] = ACTIONS(1748),
    [anon_sym_PLUS] = ACTIONS(1748),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_DASH] = ACTIONS(1748),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_SLASH] = ACTIONS(1748),
    [anon_sym_COLON] = ACTIONS(1748),
    [anon_sym_SEMI] = ACTIONS(1748),
    [anon_sym_EQ] = ACTIONS(1748),
    [anon_sym_QMARK] = ACTIONS(1748),
    [anon_sym_AT] = ACTIONS(1748),
    [anon_sym_BSLASH] = ACTIONS(1748),
    [anon_sym_CARET] = ACTIONS(1748),
    [anon_sym__] = ACTIONS(1748),
    [anon_sym_BQUOTE] = ACTIONS(1748),
    [anon_sym_LBRACE] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1748),
    [anon_sym_RBRACE] = ACTIONS(1748),
    [anon_sym_TILDE] = ACTIONS(1748),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_RPAREN] = ACTIONS(1748),
    [aux_sym__word_token1] = ACTIONS(1748),
    [aux_sym__word_token2] = ACTIONS(1748),
    [aux_sym__word_token3] = ACTIONS(1748),
    [sym__whitespace] = ACTIONS(1748),
    [sym__soft_line_ending] = ACTIONS(1748),
    [sym__block_close] = ACTIONS(1748),
    [sym__block_quote_start] = ACTIONS(1748),
    [sym__indented_chunk_start] = ACTIONS(1748),
    [sym_atx_h1_marker] = ACTIONS(1748),
    [sym_atx_h2_marker] = ACTIONS(1748),
    [sym_atx_h3_marker] = ACTIONS(1748),
    [sym_atx_h4_marker] = ACTIONS(1748),
    [sym_atx_h5_marker] = ACTIONS(1748),
    [sym_atx_h6_marker] = ACTIONS(1748),
    [sym__thematic_break] = ACTIONS(1748),
    [sym__list_marker_minus] = ACTIONS(1748),
    [sym__list_marker_plus] = ACTIONS(1748),
    [sym__list_marker_star] = ACTIONS(1748),
    [sym__list_marker_parenthesis] = ACTIONS(1748),
    [sym__list_marker_dot] = ACTIONS(1748),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1748),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1748),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1748),
    [sym__blank_line_start] = ACTIONS(1748),
    [sym__html_block_1_start] = ACTIONS(1748),
    [sym__html_block_2_start] = ACTIONS(1748),
    [sym__html_block_3_start] = ACTIONS(1748),
    [sym__html_block_4_start] = ACTIONS(1748),
    [sym__html_block_5_start] = ACTIONS(1748),
    [sym__html_block_6_start] = ACTIONS(1748),
    [sym__html_block_7_start] = ACTIONS(1748),
    [sym__pipe_table_start] = ACTIONS(1748),
  },
  [STATE(300)] = {
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_RBRACK] = ACTIONS(1170),
    [anon_sym_LT] = ACTIONS(1170),
    [anon_sym_GT] = ACTIONS(1170),
    [anon_sym_BANG] = ACTIONS(1170),
    [anon_sym_DQUOTE] = ACTIONS(1170),
    [anon_sym_POUND] = ACTIONS(1170),
    [anon_sym_DOLLAR] = ACTIONS(1170),
    [anon_sym_PERCENT] = ACTIONS(1170),
    [anon_sym_AMP] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_COMMA] = ACTIONS(1170),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_DOT] = ACTIONS(1170),
    [anon_sym_SLASH] = ACTIONS(1170),
    [anon_sym_COLON] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1170),
    [anon_sym_EQ] = ACTIONS(1170),
    [anon_sym_QMARK] = ACTIONS(1170),
    [anon_sym_AT] = ACTIONS(1170),
    [anon_sym_BSLASH] = ACTIONS(1170),
    [anon_sym_CARET] = ACTIONS(1170),
    [anon_sym__] = ACTIONS(1170),
    [anon_sym_BQUOTE] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1170),
    [anon_sym_PIPE] = ACTIONS(1170),
    [anon_sym_RBRACE] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1170),
    [anon_sym_LPAREN] = ACTIONS(1170),
    [anon_sym_RPAREN] = ACTIONS(1170),
    [aux_sym__word_token1] = ACTIONS(1170),
    [aux_sym__word_token2] = ACTIONS(1170),
    [aux_sym__word_token3] = ACTIONS(1170),
    [sym__whitespace] = ACTIONS(1170),
    [sym__soft_line_ending] = ACTIONS(1170),
    [sym__block_close] = ACTIONS(1170),
    [sym__block_quote_start] = ACTIONS(1170),
    [sym__indented_chunk_start] = ACTIONS(1170),
    [sym_atx_h1_marker] = ACTIONS(1170),
    [sym_atx_h2_marker] = ACTIONS(1170),
    [sym_atx_h3_marker] = ACTIONS(1170),
    [sym_atx_h4_marker] = ACTIONS(1170),
    [sym_atx_h5_marker] = ACTIONS(1170),
    [sym_atx_h6_marker] = ACTIONS(1170),
    [sym__thematic_break] = ACTIONS(1170),
    [sym__list_marker_minus] = ACTIONS(1170),
    [sym__list_marker_plus] = ACTIONS(1170),
    [sym__list_marker_star] = ACTIONS(1170),
    [sym__list_marker_parenthesis] = ACTIONS(1170),
    [sym__list_marker_dot] = ACTIONS(1170),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1170),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1170),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1170),
    [sym__blank_line_start] = ACTIONS(1170),
    [sym__html_block_1_start] = ACTIONS(1170),
    [sym__html_block_2_start] = ACTIONS(1170),
    [sym__html_block_3_start] = ACTIONS(1170),
    [sym__html_block_4_start] = ACTIONS(1170),
    [sym__html_block_5_start] = ACTIONS(1170),
    [sym__html_block_6_start] = ACTIONS(1170),
    [sym__html_block_7_start] = ACTIONS(1170),
    [sym__pipe_table_start] = ACTIONS(1170),
  },
  [STATE(301)] = {
    [ts_builtin_sym_end] = ACTIONS(1440),
    [anon_sym_LBRACK] = ACTIONS(1438),
    [anon_sym_RBRACK] = ACTIONS(1440),
    [anon_sym_LT] = ACTIONS(1440),
    [anon_sym_GT] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_DQUOTE] = ACTIONS(1440),
    [anon_sym_POUND] = ACTIONS(1440),
    [anon_sym_DOLLAR] = ACTIONS(1440),
    [anon_sym_PERCENT] = ACTIONS(1440),
    [anon_sym_AMP] = ACTIONS(1440),
    [anon_sym_SQUOTE] = ACTIONS(1440),
    [anon_sym_STAR] = ACTIONS(1440),
    [anon_sym_PLUS] = ACTIONS(1440),
    [anon_sym_COMMA] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_DOT] = ACTIONS(1440),
    [anon_sym_SLASH] = ACTIONS(1440),
    [anon_sym_COLON] = ACTIONS(1440),
    [anon_sym_SEMI] = ACTIONS(1440),
    [anon_sym_EQ] = ACTIONS(1440),
    [anon_sym_QMARK] = ACTIONS(1440),
    [anon_sym_AT] = ACTIONS(1440),
    [anon_sym_BSLASH] = ACTIONS(1440),
    [anon_sym_CARET] = ACTIONS(1440),
    [anon_sym__] = ACTIONS(1440),
    [anon_sym_BQUOTE] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_PIPE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [anon_sym_TILDE] = ACTIONS(1440),
    [anon_sym_LPAREN] = ACTIONS(1440),
    [anon_sym_RPAREN] = ACTIONS(1440),
    [aux_sym__word_token1] = ACTIONS(1440),
    [aux_sym__word_token2] = ACTIONS(1440),
    [aux_sym__word_token3] = ACTIONS(1440),
    [sym__whitespace] = ACTIONS(1440),
    [sym__soft_line_ending] = ACTIONS(1440),
    [sym__block_quote_start] = ACTIONS(1440),
    [sym__indented_chunk_start] = ACTIONS(1440),
    [sym_atx_h1_marker] = ACTIONS(1440),
    [sym_atx_h2_marker] = ACTIONS(1440),
    [sym_atx_h3_marker] = ACTIONS(1440),
    [sym_atx_h4_marker] = ACTIONS(1440),
    [sym_atx_h5_marker] = ACTIONS(1440),
    [sym_atx_h6_marker] = ACTIONS(1440),
    [sym__thematic_break] = ACTIONS(1440),
    [sym__list_marker_minus] = ACTIONS(1440),
    [sym__list_marker_plus] = ACTIONS(1440),
    [sym__list_marker_star] = ACTIONS(1440),
    [sym__list_marker_parenthesis] = ACTIONS(1440),
    [sym__list_marker_dot] = ACTIONS(1440),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1440),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1440),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1440),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1440),
    [sym__blank_line_start] = ACTIONS(1440),
    [sym__html_block_1_start] = ACTIONS(1440),
    [sym__html_block_2_start] = ACTIONS(1440),
    [sym__html_block_3_start] = ACTIONS(1440),
    [sym__html_block_4_start] = ACTIONS(1440),
    [sym__html_block_5_start] = ACTIONS(1440),
    [sym__html_block_6_start] = ACTIONS(1440),
    [sym__html_block_7_start] = ACTIONS(1440),
    [sym__pipe_table_start] = ACTIONS(1440),
  },
  [STATE(302)] = {
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_LT] = ACTIONS(1752),
    [anon_sym_GT] = ACTIONS(1752),
    [anon_sym_BANG] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [anon_sym_POUND] = ACTIONS(1752),
    [anon_sym_DOLLAR] = ACTIONS(1752),
    [anon_sym_PERCENT] = ACTIONS(1752),
    [anon_sym_AMP] = ACTIONS(1752),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_STAR] = ACTIONS(1752),
    [anon_sym_PLUS] = ACTIONS(1752),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_DASH] = ACTIONS(1752),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_SLASH] = ACTIONS(1752),
    [anon_sym_COLON] = ACTIONS(1752),
    [anon_sym_SEMI] = ACTIONS(1752),
    [anon_sym_EQ] = ACTIONS(1752),
    [anon_sym_QMARK] = ACTIONS(1752),
    [anon_sym_AT] = ACTIONS(1752),
    [anon_sym_BSLASH] = ACTIONS(1752),
    [anon_sym_CARET] = ACTIONS(1752),
    [anon_sym__] = ACTIONS(1752),
    [anon_sym_BQUOTE] = ACTIONS(1752),
    [anon_sym_LBRACE] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1752),
    [anon_sym_RBRACE] = ACTIONS(1752),
    [anon_sym_TILDE] = ACTIONS(1752),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_RPAREN] = ACTIONS(1752),
    [aux_sym__word_token1] = ACTIONS(1752),
    [aux_sym__word_token2] = ACTIONS(1752),
    [aux_sym__word_token3] = ACTIONS(1752),
    [sym__whitespace] = ACTIONS(1752),
    [sym__soft_line_ending] = ACTIONS(1752),
    [sym__block_close] = ACTIONS(1752),
    [sym__block_quote_start] = ACTIONS(1752),
    [sym__indented_chunk_start] = ACTIONS(1752),
    [sym_atx_h1_marker] = ACTIONS(1752),
    [sym_atx_h2_marker] = ACTIONS(1752),
    [sym_atx_h3_marker] = ACTIONS(1752),
    [sym_atx_h4_marker] = ACTIONS(1752),
    [sym_atx_h5_marker] = ACTIONS(1752),
    [sym_atx_h6_marker] = ACTIONS(1752),
    [sym__thematic_break] = ACTIONS(1752),
    [sym__list_marker_minus] = ACTIONS(1752),
    [sym__list_marker_plus] = ACTIONS(1752),
    [sym__list_marker_star] = ACTIONS(1752),
    [sym__list_marker_parenthesis] = ACTIONS(1752),
    [sym__list_marker_dot] = ACTIONS(1752),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1752),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1752),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1752),
    [sym__blank_line_start] = ACTIONS(1752),
    [sym__html_block_1_start] = ACTIONS(1752),
    [sym__html_block_2_start] = ACTIONS(1752),
    [sym__html_block_3_start] = ACTIONS(1752),
    [sym__html_block_4_start] = ACTIONS(1752),
    [sym__html_block_5_start] = ACTIONS(1752),
    [sym__html_block_6_start] = ACTIONS(1752),
    [sym__html_block_7_start] = ACTIONS(1752),
    [sym__pipe_table_start] = ACTIONS(1752),
  },
  [STATE(303)] = {
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_LT] = ACTIONS(1756),
    [anon_sym_GT] = ACTIONS(1756),
    [anon_sym_BANG] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [anon_sym_POUND] = ACTIONS(1756),
    [anon_sym_DOLLAR] = ACTIONS(1756),
    [anon_sym_PERCENT] = ACTIONS(1756),
    [anon_sym_AMP] = ACTIONS(1756),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_STAR] = ACTIONS(1756),
    [anon_sym_PLUS] = ACTIONS(1756),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_DASH] = ACTIONS(1756),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_SLASH] = ACTIONS(1756),
    [anon_sym_COLON] = ACTIONS(1756),
    [anon_sym_SEMI] = ACTIONS(1756),
    [anon_sym_EQ] = ACTIONS(1756),
    [anon_sym_QMARK] = ACTIONS(1756),
    [anon_sym_AT] = ACTIONS(1756),
    [anon_sym_BSLASH] = ACTIONS(1756),
    [anon_sym_CARET] = ACTIONS(1756),
    [anon_sym__] = ACTIONS(1756),
    [anon_sym_BQUOTE] = ACTIONS(1756),
    [anon_sym_LBRACE] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1756),
    [anon_sym_RBRACE] = ACTIONS(1756),
    [anon_sym_TILDE] = ACTIONS(1756),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_RPAREN] = ACTIONS(1756),
    [aux_sym__word_token1] = ACTIONS(1756),
    [aux_sym__word_token2] = ACTIONS(1756),
    [aux_sym__word_token3] = ACTIONS(1756),
    [sym__whitespace] = ACTIONS(1756),
    [sym__soft_line_ending] = ACTIONS(1756),
    [sym__block_close] = ACTIONS(1756),
    [sym__block_quote_start] = ACTIONS(1756),
    [sym__indented_chunk_start] = ACTIONS(1756),
    [sym_atx_h1_marker] = ACTIONS(1756),
    [sym_atx_h2_marker] = ACTIONS(1756),
    [sym_atx_h3_marker] = ACTIONS(1756),
    [sym_atx_h4_marker] = ACTIONS(1756),
    [sym_atx_h5_marker] = ACTIONS(1756),
    [sym_atx_h6_marker] = ACTIONS(1756),
    [sym__thematic_break] = ACTIONS(1756),
    [sym__list_marker_minus] = ACTIONS(1756),
    [sym__list_marker_plus] = ACTIONS(1756),
    [sym__list_marker_star] = ACTIONS(1756),
    [sym__list_marker_parenthesis] = ACTIONS(1756),
    [sym__list_marker_dot] = ACTIONS(1756),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1756),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1756),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1756),
    [sym__blank_line_start] = ACTIONS(1756),
    [sym__html_block_1_start] = ACTIONS(1756),
    [sym__html_block_2_start] = ACTIONS(1756),
    [sym__html_block_3_start] = ACTIONS(1756),
    [sym__html_block_4_start] = ACTIONS(1756),
    [sym__html_block_5_start] = ACTIONS(1756),
    [sym__html_block_6_start] = ACTIONS(1756),
    [sym__html_block_7_start] = ACTIONS(1756),
    [sym__pipe_table_start] = ACTIONS(1756),
  },
  [STATE(304)] = {
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_LT] = ACTIONS(1760),
    [anon_sym_GT] = ACTIONS(1760),
    [anon_sym_BANG] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [anon_sym_POUND] = ACTIONS(1760),
    [anon_sym_DOLLAR] = ACTIONS(1760),
    [anon_sym_PERCENT] = ACTIONS(1760),
    [anon_sym_AMP] = ACTIONS(1760),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_STAR] = ACTIONS(1760),
    [anon_sym_PLUS] = ACTIONS(1760),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_DASH] = ACTIONS(1760),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_SLASH] = ACTIONS(1760),
    [anon_sym_COLON] = ACTIONS(1760),
    [anon_sym_SEMI] = ACTIONS(1760),
    [anon_sym_EQ] = ACTIONS(1760),
    [anon_sym_QMARK] = ACTIONS(1760),
    [anon_sym_AT] = ACTIONS(1760),
    [anon_sym_BSLASH] = ACTIONS(1760),
    [anon_sym_CARET] = ACTIONS(1760),
    [anon_sym__] = ACTIONS(1760),
    [anon_sym_BQUOTE] = ACTIONS(1760),
    [anon_sym_LBRACE] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1760),
    [anon_sym_RBRACE] = ACTIONS(1760),
    [anon_sym_TILDE] = ACTIONS(1760),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_RPAREN] = ACTIONS(1760),
    [aux_sym__word_token1] = ACTIONS(1760),
    [aux_sym__word_token2] = ACTIONS(1760),
    [aux_sym__word_token3] = ACTIONS(1760),
    [sym__whitespace] = ACTIONS(1760),
    [sym__soft_line_ending] = ACTIONS(1760),
    [sym__block_close] = ACTIONS(1760),
    [sym__block_quote_start] = ACTIONS(1760),
    [sym__indented_chunk_start] = ACTIONS(1760),
    [sym_atx_h1_marker] = ACTIONS(1760),
    [sym_atx_h2_marker] = ACTIONS(1760),
    [sym_atx_h3_marker] = ACTIONS(1760),
    [sym_atx_h4_marker] = ACTIONS(1760),
    [sym_atx_h5_marker] = ACTIONS(1760),
    [sym_atx_h6_marker] = ACTIONS(1760),
    [sym__thematic_break] = ACTIONS(1760),
    [sym__list_marker_minus] = ACTIONS(1760),
    [sym__list_marker_plus] = ACTIONS(1760),
    [sym__list_marker_star] = ACTIONS(1760),
    [sym__list_marker_parenthesis] = ACTIONS(1760),
    [sym__list_marker_dot] = ACTIONS(1760),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1760),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1760),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1760),
    [sym__blank_line_start] = ACTIONS(1760),
    [sym__html_block_1_start] = ACTIONS(1760),
    [sym__html_block_2_start] = ACTIONS(1760),
    [sym__html_block_3_start] = ACTIONS(1760),
    [sym__html_block_4_start] = ACTIONS(1760),
    [sym__html_block_5_start] = ACTIONS(1760),
    [sym__html_block_6_start] = ACTIONS(1760),
    [sym__html_block_7_start] = ACTIONS(1760),
    [sym__pipe_table_start] = ACTIONS(1760),
  },
  [STATE(305)] = {
    [anon_sym_LBRACK] = ACTIONS(1762),
    [anon_sym_RBRACK] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1764),
    [anon_sym_BANG] = ACTIONS(1764),
    [anon_sym_DQUOTE] = ACTIONS(1764),
    [anon_sym_POUND] = ACTIONS(1764),
    [anon_sym_DOLLAR] = ACTIONS(1764),
    [anon_sym_PERCENT] = ACTIONS(1764),
    [anon_sym_AMP] = ACTIONS(1764),
    [anon_sym_SQUOTE] = ACTIONS(1764),
    [anon_sym_STAR] = ACTIONS(1764),
    [anon_sym_PLUS] = ACTIONS(1764),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_DASH] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_SLASH] = ACTIONS(1764),
    [anon_sym_COLON] = ACTIONS(1764),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym_EQ] = ACTIONS(1764),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_AT] = ACTIONS(1764),
    [anon_sym_BSLASH] = ACTIONS(1764),
    [anon_sym_CARET] = ACTIONS(1764),
    [anon_sym__] = ACTIONS(1764),
    [anon_sym_BQUOTE] = ACTIONS(1764),
    [anon_sym_LBRACE] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1764),
    [anon_sym_RBRACE] = ACTIONS(1764),
    [anon_sym_TILDE] = ACTIONS(1764),
    [anon_sym_LPAREN] = ACTIONS(1764),
    [anon_sym_RPAREN] = ACTIONS(1764),
    [aux_sym__word_token1] = ACTIONS(1764),
    [aux_sym__word_token2] = ACTIONS(1764),
    [aux_sym__word_token3] = ACTIONS(1764),
    [sym__whitespace] = ACTIONS(1764),
    [sym__soft_line_ending] = ACTIONS(1764),
    [sym__block_close] = ACTIONS(1764),
    [sym__block_quote_start] = ACTIONS(1764),
    [sym__indented_chunk_start] = ACTIONS(1764),
    [sym_atx_h1_marker] = ACTIONS(1764),
    [sym_atx_h2_marker] = ACTIONS(1764),
    [sym_atx_h3_marker] = ACTIONS(1764),
    [sym_atx_h4_marker] = ACTIONS(1764),
    [sym_atx_h5_marker] = ACTIONS(1764),
    [sym_atx_h6_marker] = ACTIONS(1764),
    [sym__thematic_break] = ACTIONS(1764),
    [sym__list_marker_minus] = ACTIONS(1764),
    [sym__list_marker_plus] = ACTIONS(1764),
    [sym__list_marker_star] = ACTIONS(1764),
    [sym__list_marker_parenthesis] = ACTIONS(1764),
    [sym__list_marker_dot] = ACTIONS(1764),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1764),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1764),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1764),
    [sym__blank_line_start] = ACTIONS(1764),
    [sym__html_block_1_start] = ACTIONS(1764),
    [sym__html_block_2_start] = ACTIONS(1764),
    [sym__html_block_3_start] = ACTIONS(1764),
    [sym__html_block_4_start] = ACTIONS(1764),
    [sym__html_block_5_start] = ACTIONS(1764),
    [sym__html_block_6_start] = ACTIONS(1764),
    [sym__html_block_7_start] = ACTIONS(1764),
    [sym__pipe_table_start] = ACTIONS(1764),
  },
  [STATE(306)] = {
    [anon_sym_LBRACK] = ACTIONS(1766),
    [anon_sym_RBRACK] = ACTIONS(1768),
    [anon_sym_LT] = ACTIONS(1768),
    [anon_sym_GT] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_DQUOTE] = ACTIONS(1768),
    [anon_sym_POUND] = ACTIONS(1768),
    [anon_sym_DOLLAR] = ACTIONS(1768),
    [anon_sym_PERCENT] = ACTIONS(1768),
    [anon_sym_AMP] = ACTIONS(1768),
    [anon_sym_SQUOTE] = ACTIONS(1768),
    [anon_sym_STAR] = ACTIONS(1768),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_COMMA] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(1768),
    [anon_sym_DOT] = ACTIONS(1768),
    [anon_sym_SLASH] = ACTIONS(1768),
    [anon_sym_COLON] = ACTIONS(1768),
    [anon_sym_SEMI] = ACTIONS(1768),
    [anon_sym_EQ] = ACTIONS(1768),
    [anon_sym_QMARK] = ACTIONS(1768),
    [anon_sym_AT] = ACTIONS(1768),
    [anon_sym_BSLASH] = ACTIONS(1768),
    [anon_sym_CARET] = ACTIONS(1768),
    [anon_sym__] = ACTIONS(1768),
    [anon_sym_BQUOTE] = ACTIONS(1768),
    [anon_sym_LBRACE] = ACTIONS(1768),
    [anon_sym_PIPE] = ACTIONS(1768),
    [anon_sym_RBRACE] = ACTIONS(1768),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_LPAREN] = ACTIONS(1768),
    [anon_sym_RPAREN] = ACTIONS(1768),
    [aux_sym__word_token1] = ACTIONS(1768),
    [aux_sym__word_token2] = ACTIONS(1768),
    [aux_sym__word_token3] = ACTIONS(1768),
    [sym__whitespace] = ACTIONS(1768),
    [sym__soft_line_ending] = ACTIONS(1768),
    [sym__block_close] = ACTIONS(1768),
    [sym__block_quote_start] = ACTIONS(1768),
    [sym__indented_chunk_start] = ACTIONS(1768),
    [sym_atx_h1_marker] = ACTIONS(1768),
    [sym_atx_h2_marker] = ACTIONS(1768),
    [sym_atx_h3_marker] = ACTIONS(1768),
    [sym_atx_h4_marker] = ACTIONS(1768),
    [sym_atx_h5_marker] = ACTIONS(1768),
    [sym_atx_h6_marker] = ACTIONS(1768),
    [sym__thematic_break] = ACTIONS(1768),
    [sym__list_marker_minus] = ACTIONS(1768),
    [sym__list_marker_plus] = ACTIONS(1768),
    [sym__list_marker_star] = ACTIONS(1768),
    [sym__list_marker_parenthesis] = ACTIONS(1768),
    [sym__list_marker_dot] = ACTIONS(1768),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1768),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1768),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1768),
    [sym__blank_line_start] = ACTIONS(1768),
    [sym__html_block_1_start] = ACTIONS(1768),
    [sym__html_block_2_start] = ACTIONS(1768),
    [sym__html_block_3_start] = ACTIONS(1768),
    [sym__html_block_4_start] = ACTIONS(1768),
    [sym__html_block_5_start] = ACTIONS(1768),
    [sym__html_block_6_start] = ACTIONS(1768),
    [sym__html_block_7_start] = ACTIONS(1768),
    [sym__pipe_table_start] = ACTIONS(1768),
  },
  [STATE(307)] = {
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_RBRACK] = ACTIONS(1772),
    [anon_sym_LT] = ACTIONS(1772),
    [anon_sym_GT] = ACTIONS(1772),
    [anon_sym_BANG] = ACTIONS(1772),
    [anon_sym_DQUOTE] = ACTIONS(1772),
    [anon_sym_POUND] = ACTIONS(1772),
    [anon_sym_DOLLAR] = ACTIONS(1772),
    [anon_sym_PERCENT] = ACTIONS(1772),
    [anon_sym_AMP] = ACTIONS(1772),
    [anon_sym_SQUOTE] = ACTIONS(1772),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_PLUS] = ACTIONS(1772),
    [anon_sym_COMMA] = ACTIONS(1772),
    [anon_sym_DASH] = ACTIONS(1772),
    [anon_sym_DOT] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1772),
    [anon_sym_COLON] = ACTIONS(1772),
    [anon_sym_SEMI] = ACTIONS(1772),
    [anon_sym_EQ] = ACTIONS(1772),
    [anon_sym_QMARK] = ACTIONS(1772),
    [anon_sym_AT] = ACTIONS(1772),
    [anon_sym_BSLASH] = ACTIONS(1772),
    [anon_sym_CARET] = ACTIONS(1772),
    [anon_sym__] = ACTIONS(1772),
    [anon_sym_BQUOTE] = ACTIONS(1772),
    [anon_sym_LBRACE] = ACTIONS(1772),
    [anon_sym_PIPE] = ACTIONS(1772),
    [anon_sym_RBRACE] = ACTIONS(1772),
    [anon_sym_TILDE] = ACTIONS(1772),
    [anon_sym_LPAREN] = ACTIONS(1772),
    [anon_sym_RPAREN] = ACTIONS(1772),
    [aux_sym__word_token1] = ACTIONS(1772),
    [aux_sym__word_token2] = ACTIONS(1772),
    [aux_sym__word_token3] = ACTIONS(1772),
    [sym__whitespace] = ACTIONS(1772),
    [sym__soft_line_ending] = ACTIONS(1772),
    [sym__block_close] = ACTIONS(1772),
    [sym__block_quote_start] = ACTIONS(1772),
    [sym__indented_chunk_start] = ACTIONS(1772),
    [sym_atx_h1_marker] = ACTIONS(1772),
    [sym_atx_h2_marker] = ACTIONS(1772),
    [sym_atx_h3_marker] = ACTIONS(1772),
    [sym_atx_h4_marker] = ACTIONS(1772),
    [sym_atx_h5_marker] = ACTIONS(1772),
    [sym_atx_h6_marker] = ACTIONS(1772),
    [sym__thematic_break] = ACTIONS(1772),
    [sym__list_marker_minus] = ACTIONS(1772),
    [sym__list_marker_plus] = ACTIONS(1772),
    [sym__list_marker_star] = ACTIONS(1772),
    [sym__list_marker_parenthesis] = ACTIONS(1772),
    [sym__list_marker_dot] = ACTIONS(1772),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1772),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1772),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1772),
    [sym__blank_line_start] = ACTIONS(1772),
    [sym__html_block_1_start] = ACTIONS(1772),
    [sym__html_block_2_start] = ACTIONS(1772),
    [sym__html_block_3_start] = ACTIONS(1772),
    [sym__html_block_4_start] = ACTIONS(1772),
    [sym__html_block_5_start] = ACTIONS(1772),
    [sym__html_block_6_start] = ACTIONS(1772),
    [sym__html_block_7_start] = ACTIONS(1772),
    [sym__pipe_table_start] = ACTIONS(1772),
  },
  [STATE(308)] = {
    [anon_sym_LBRACK] = ACTIONS(1774),
    [anon_sym_RBRACK] = ACTIONS(1776),
    [anon_sym_LT] = ACTIONS(1776),
    [anon_sym_GT] = ACTIONS(1776),
    [anon_sym_BANG] = ACTIONS(1776),
    [anon_sym_DQUOTE] = ACTIONS(1776),
    [anon_sym_POUND] = ACTIONS(1776),
    [anon_sym_DOLLAR] = ACTIONS(1776),
    [anon_sym_PERCENT] = ACTIONS(1776),
    [anon_sym_AMP] = ACTIONS(1776),
    [anon_sym_SQUOTE] = ACTIONS(1776),
    [anon_sym_STAR] = ACTIONS(1776),
    [anon_sym_PLUS] = ACTIONS(1776),
    [anon_sym_COMMA] = ACTIONS(1776),
    [anon_sym_DASH] = ACTIONS(1776),
    [anon_sym_DOT] = ACTIONS(1776),
    [anon_sym_SLASH] = ACTIONS(1776),
    [anon_sym_COLON] = ACTIONS(1776),
    [anon_sym_SEMI] = ACTIONS(1776),
    [anon_sym_EQ] = ACTIONS(1776),
    [anon_sym_QMARK] = ACTIONS(1776),
    [anon_sym_AT] = ACTIONS(1776),
    [anon_sym_BSLASH] = ACTIONS(1776),
    [anon_sym_CARET] = ACTIONS(1776),
    [anon_sym__] = ACTIONS(1776),
    [anon_sym_BQUOTE] = ACTIONS(1776),
    [anon_sym_LBRACE] = ACTIONS(1776),
    [anon_sym_PIPE] = ACTIONS(1776),
    [anon_sym_RBRACE] = ACTIONS(1776),
    [anon_sym_TILDE] = ACTIONS(1776),
    [anon_sym_LPAREN] = ACTIONS(1776),
    [anon_sym_RPAREN] = ACTIONS(1776),
    [aux_sym__word_token1] = ACTIONS(1776),
    [aux_sym__word_token2] = ACTIONS(1776),
    [aux_sym__word_token3] = ACTIONS(1776),
    [sym__whitespace] = ACTIONS(1776),
    [sym__soft_line_ending] = ACTIONS(1776),
    [sym__block_close] = ACTIONS(1776),
    [sym__block_quote_start] = ACTIONS(1776),
    [sym__indented_chunk_start] = ACTIONS(1776),
    [sym_atx_h1_marker] = ACTIONS(1776),
    [sym_atx_h2_marker] = ACTIONS(1776),
    [sym_atx_h3_marker] = ACTIONS(1776),
    [sym_atx_h4_marker] = ACTIONS(1776),
    [sym_atx_h5_marker] = ACTIONS(1776),
    [sym_atx_h6_marker] = ACTIONS(1776),
    [sym__thematic_break] = ACTIONS(1776),
    [sym__list_marker_minus] = ACTIONS(1776),
    [sym__list_marker_plus] = ACTIONS(1776),
    [sym__list_marker_star] = ACTIONS(1776),
    [sym__list_marker_parenthesis] = ACTIONS(1776),
    [sym__list_marker_dot] = ACTIONS(1776),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1776),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1776),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1776),
    [sym__blank_line_start] = ACTIONS(1776),
    [sym__html_block_1_start] = ACTIONS(1776),
    [sym__html_block_2_start] = ACTIONS(1776),
    [sym__html_block_3_start] = ACTIONS(1776),
    [sym__html_block_4_start] = ACTIONS(1776),
    [sym__html_block_5_start] = ACTIONS(1776),
    [sym__html_block_6_start] = ACTIONS(1776),
    [sym__html_block_7_start] = ACTIONS(1776),
    [sym__pipe_table_start] = ACTIONS(1776),
  },
  [STATE(309)] = {
    [anon_sym_LBRACK] = ACTIONS(1186),
    [anon_sym_RBRACK] = ACTIONS(1188),
    [anon_sym_LT] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [anon_sym_POUND] = ACTIONS(1188),
    [anon_sym_DOLLAR] = ACTIONS(1188),
    [anon_sym_PERCENT] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_PLUS] = ACTIONS(1188),
    [anon_sym_COMMA] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_SLASH] = ACTIONS(1188),
    [anon_sym_COLON] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym_EQ] = ACTIONS(1188),
    [anon_sym_QMARK] = ACTIONS(1188),
    [anon_sym_AT] = ACTIONS(1188),
    [anon_sym_BSLASH] = ACTIONS(1188),
    [anon_sym_CARET] = ACTIONS(1188),
    [anon_sym__] = ACTIONS(1188),
    [anon_sym_BQUOTE] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_PIPE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_LPAREN] = ACTIONS(1188),
    [anon_sym_RPAREN] = ACTIONS(1188),
    [aux_sym__word_token1] = ACTIONS(1188),
    [aux_sym__word_token2] = ACTIONS(1188),
    [aux_sym__word_token3] = ACTIONS(1188),
    [sym__whitespace] = ACTIONS(1188),
    [sym__soft_line_ending] = ACTIONS(1188),
    [sym__block_close] = ACTIONS(1188),
    [sym__block_quote_start] = ACTIONS(1188),
    [sym__indented_chunk_start] = ACTIONS(1188),
    [sym_atx_h1_marker] = ACTIONS(1188),
    [sym_atx_h2_marker] = ACTIONS(1188),
    [sym_atx_h3_marker] = ACTIONS(1188),
    [sym_atx_h4_marker] = ACTIONS(1188),
    [sym_atx_h5_marker] = ACTIONS(1188),
    [sym_atx_h6_marker] = ACTIONS(1188),
    [sym__thematic_break] = ACTIONS(1188),
    [sym__list_marker_minus] = ACTIONS(1188),
    [sym__list_marker_plus] = ACTIONS(1188),
    [sym__list_marker_star] = ACTIONS(1188),
    [sym__list_marker_parenthesis] = ACTIONS(1188),
    [sym__list_marker_dot] = ACTIONS(1188),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1188),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1188),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1188),
    [sym__blank_line_start] = ACTIONS(1188),
    [sym__html_block_1_start] = ACTIONS(1188),
    [sym__html_block_2_start] = ACTIONS(1188),
    [sym__html_block_3_start] = ACTIONS(1188),
    [sym__html_block_4_start] = ACTIONS(1188),
    [sym__html_block_5_start] = ACTIONS(1188),
    [sym__html_block_6_start] = ACTIONS(1188),
    [sym__html_block_7_start] = ACTIONS(1188),
    [sym__pipe_table_start] = ACTIONS(1188),
  },
  [STATE(310)] = {
    [ts_builtin_sym_end] = ACTIONS(1656),
    [anon_sym_LBRACK] = ACTIONS(1654),
    [anon_sym_RBRACK] = ACTIONS(1656),
    [anon_sym_LT] = ACTIONS(1656),
    [anon_sym_GT] = ACTIONS(1656),
    [anon_sym_BANG] = ACTIONS(1656),
    [anon_sym_DQUOTE] = ACTIONS(1656),
    [anon_sym_POUND] = ACTIONS(1656),
    [anon_sym_DOLLAR] = ACTIONS(1656),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_AMP] = ACTIONS(1656),
    [anon_sym_SQUOTE] = ACTIONS(1656),
    [anon_sym_STAR] = ACTIONS(1656),
    [anon_sym_PLUS] = ACTIONS(1656),
    [anon_sym_COMMA] = ACTIONS(1656),
    [anon_sym_DASH] = ACTIONS(1656),
    [anon_sym_DOT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1656),
    [anon_sym_COLON] = ACTIONS(1656),
    [anon_sym_SEMI] = ACTIONS(1656),
    [anon_sym_EQ] = ACTIONS(1656),
    [anon_sym_QMARK] = ACTIONS(1656),
    [anon_sym_AT] = ACTIONS(1656),
    [anon_sym_BSLASH] = ACTIONS(1656),
    [anon_sym_CARET] = ACTIONS(1656),
    [anon_sym__] = ACTIONS(1656),
    [anon_sym_BQUOTE] = ACTIONS(1656),
    [anon_sym_LBRACE] = ACTIONS(1656),
    [anon_sym_PIPE] = ACTIONS(1656),
    [anon_sym_RBRACE] = ACTIONS(1656),
    [anon_sym_TILDE] = ACTIONS(1656),
    [anon_sym_LPAREN] = ACTIONS(1656),
    [anon_sym_RPAREN] = ACTIONS(1656),
    [aux_sym__word_token1] = ACTIONS(1656),
    [aux_sym__word_token2] = ACTIONS(1656),
    [aux_sym__word_token3] = ACTIONS(1656),
    [sym__whitespace] = ACTIONS(1656),
    [sym__soft_line_ending] = ACTIONS(1656),
    [sym__block_quote_start] = ACTIONS(1656),
    [sym__indented_chunk_start] = ACTIONS(1656),
    [sym_atx_h1_marker] = ACTIONS(1656),
    [sym_atx_h2_marker] = ACTIONS(1656),
    [sym_atx_h3_marker] = ACTIONS(1656),
    [sym_atx_h4_marker] = ACTIONS(1656),
    [sym_atx_h5_marker] = ACTIONS(1656),
    [sym_atx_h6_marker] = ACTIONS(1656),
    [sym__thematic_break] = ACTIONS(1656),
    [sym__list_marker_minus] = ACTIONS(1656),
    [sym__list_marker_plus] = ACTIONS(1656),
    [sym__list_marker_star] = ACTIONS(1656),
    [sym__list_marker_parenthesis] = ACTIONS(1656),
    [sym__list_marker_dot] = ACTIONS(1656),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1656),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1656),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1656),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1656),
    [sym__blank_line_start] = ACTIONS(1656),
    [sym__html_block_1_start] = ACTIONS(1656),
    [sym__html_block_2_start] = ACTIONS(1656),
    [sym__html_block_3_start] = ACTIONS(1656),
    [sym__html_block_4_start] = ACTIONS(1656),
    [sym__html_block_5_start] = ACTIONS(1656),
    [sym__html_block_6_start] = ACTIONS(1656),
    [sym__html_block_7_start] = ACTIONS(1656),
    [sym__pipe_table_start] = ACTIONS(1656),
  },
  [STATE(311)] = {
    [anon_sym_LBRACK] = ACTIONS(1778),
    [anon_sym_RBRACK] = ACTIONS(1780),
    [anon_sym_LT] = ACTIONS(1780),
    [anon_sym_GT] = ACTIONS(1780),
    [anon_sym_BANG] = ACTIONS(1780),
    [anon_sym_DQUOTE] = ACTIONS(1780),
    [anon_sym_POUND] = ACTIONS(1780),
    [anon_sym_DOLLAR] = ACTIONS(1780),
    [anon_sym_PERCENT] = ACTIONS(1780),
    [anon_sym_AMP] = ACTIONS(1780),
    [anon_sym_SQUOTE] = ACTIONS(1780),
    [anon_sym_STAR] = ACTIONS(1780),
    [anon_sym_PLUS] = ACTIONS(1780),
    [anon_sym_COMMA] = ACTIONS(1780),
    [anon_sym_DASH] = ACTIONS(1780),
    [anon_sym_DOT] = ACTIONS(1780),
    [anon_sym_SLASH] = ACTIONS(1780),
    [anon_sym_COLON] = ACTIONS(1780),
    [anon_sym_SEMI] = ACTIONS(1780),
    [anon_sym_EQ] = ACTIONS(1780),
    [anon_sym_QMARK] = ACTIONS(1780),
    [anon_sym_AT] = ACTIONS(1780),
    [anon_sym_BSLASH] = ACTIONS(1780),
    [anon_sym_CARET] = ACTIONS(1780),
    [anon_sym__] = ACTIONS(1780),
    [anon_sym_BQUOTE] = ACTIONS(1780),
    [anon_sym_LBRACE] = ACTIONS(1780),
    [anon_sym_PIPE] = ACTIONS(1780),
    [anon_sym_RBRACE] = ACTIONS(1780),
    [anon_sym_TILDE] = ACTIONS(1780),
    [anon_sym_LPAREN] = ACTIONS(1780),
    [anon_sym_RPAREN] = ACTIONS(1780),
    [aux_sym__word_token1] = ACTIONS(1780),
    [aux_sym__word_token2] = ACTIONS(1780),
    [aux_sym__word_token3] = ACTIONS(1780),
    [sym__whitespace] = ACTIONS(1780),
    [sym__soft_line_ending] = ACTIONS(1780),
    [sym__block_close] = ACTIONS(1780),
    [sym__block_quote_start] = ACTIONS(1780),
    [sym__indented_chunk_start] = ACTIONS(1780),
    [sym_atx_h1_marker] = ACTIONS(1780),
    [sym_atx_h2_marker] = ACTIONS(1780),
    [sym_atx_h3_marker] = ACTIONS(1780),
    [sym_atx_h4_marker] = ACTIONS(1780),
    [sym_atx_h5_marker] = ACTIONS(1780),
    [sym_atx_h6_marker] = ACTIONS(1780),
    [sym__thematic_break] = ACTIONS(1780),
    [sym__list_marker_minus] = ACTIONS(1780),
    [sym__list_marker_plus] = ACTIONS(1780),
    [sym__list_marker_star] = ACTIONS(1780),
    [sym__list_marker_parenthesis] = ACTIONS(1780),
    [sym__list_marker_dot] = ACTIONS(1780),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1780),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1780),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1780),
    [sym__blank_line_start] = ACTIONS(1780),
    [sym__html_block_1_start] = ACTIONS(1780),
    [sym__html_block_2_start] = ACTIONS(1780),
    [sym__html_block_3_start] = ACTIONS(1780),
    [sym__html_block_4_start] = ACTIONS(1780),
    [sym__html_block_5_start] = ACTIONS(1780),
    [sym__html_block_6_start] = ACTIONS(1780),
    [sym__html_block_7_start] = ACTIONS(1780),
    [sym__pipe_table_start] = ACTIONS(1780),
  },
  [STATE(312)] = {
    [anon_sym_LBRACK] = ACTIONS(1782),
    [anon_sym_RBRACK] = ACTIONS(1784),
    [anon_sym_LT] = ACTIONS(1784),
    [anon_sym_GT] = ACTIONS(1784),
    [anon_sym_BANG] = ACTIONS(1784),
    [anon_sym_DQUOTE] = ACTIONS(1784),
    [anon_sym_POUND] = ACTIONS(1784),
    [anon_sym_DOLLAR] = ACTIONS(1784),
    [anon_sym_PERCENT] = ACTIONS(1784),
    [anon_sym_AMP] = ACTIONS(1784),
    [anon_sym_SQUOTE] = ACTIONS(1784),
    [anon_sym_STAR] = ACTIONS(1784),
    [anon_sym_PLUS] = ACTIONS(1784),
    [anon_sym_COMMA] = ACTIONS(1784),
    [anon_sym_DASH] = ACTIONS(1784),
    [anon_sym_DOT] = ACTIONS(1784),
    [anon_sym_SLASH] = ACTIONS(1784),
    [anon_sym_COLON] = ACTIONS(1784),
    [anon_sym_SEMI] = ACTIONS(1784),
    [anon_sym_EQ] = ACTIONS(1784),
    [anon_sym_QMARK] = ACTIONS(1784),
    [anon_sym_AT] = ACTIONS(1784),
    [anon_sym_BSLASH] = ACTIONS(1784),
    [anon_sym_CARET] = ACTIONS(1784),
    [anon_sym__] = ACTIONS(1784),
    [anon_sym_BQUOTE] = ACTIONS(1784),
    [anon_sym_LBRACE] = ACTIONS(1784),
    [anon_sym_PIPE] = ACTIONS(1784),
    [anon_sym_RBRACE] = ACTIONS(1784),
    [anon_sym_TILDE] = ACTIONS(1784),
    [anon_sym_LPAREN] = ACTIONS(1784),
    [anon_sym_RPAREN] = ACTIONS(1784),
    [aux_sym__word_token1] = ACTIONS(1784),
    [aux_sym__word_token2] = ACTIONS(1784),
    [aux_sym__word_token3] = ACTIONS(1784),
    [sym__whitespace] = ACTIONS(1784),
    [sym__soft_line_ending] = ACTIONS(1784),
    [sym__block_close] = ACTIONS(1784),
    [sym__block_quote_start] = ACTIONS(1784),
    [sym__indented_chunk_start] = ACTIONS(1784),
    [sym_atx_h1_marker] = ACTIONS(1784),
    [sym_atx_h2_marker] = ACTIONS(1784),
    [sym_atx_h3_marker] = ACTIONS(1784),
    [sym_atx_h4_marker] = ACTIONS(1784),
    [sym_atx_h5_marker] = ACTIONS(1784),
    [sym_atx_h6_marker] = ACTIONS(1784),
    [sym__thematic_break] = ACTIONS(1784),
    [sym__list_marker_minus] = ACTIONS(1784),
    [sym__list_marker_plus] = ACTIONS(1784),
    [sym__list_marker_star] = ACTIONS(1784),
    [sym__list_marker_parenthesis] = ACTIONS(1784),
    [sym__list_marker_dot] = ACTIONS(1784),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1784),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1784),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1784),
    [sym__blank_line_start] = ACTIONS(1784),
    [sym__html_block_1_start] = ACTIONS(1784),
    [sym__html_block_2_start] = ACTIONS(1784),
    [sym__html_block_3_start] = ACTIONS(1784),
    [sym__html_block_4_start] = ACTIONS(1784),
    [sym__html_block_5_start] = ACTIONS(1784),
    [sym__html_block_6_start] = ACTIONS(1784),
    [sym__html_block_7_start] = ACTIONS(1784),
    [sym__pipe_table_start] = ACTIONS(1784),
  },
  [STATE(313)] = {
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_RBRACK] = ACTIONS(1206),
    [anon_sym_LT] = ACTIONS(1206),
    [anon_sym_GT] = ACTIONS(1206),
    [anon_sym_BANG] = ACTIONS(1206),
    [anon_sym_DQUOTE] = ACTIONS(1206),
    [anon_sym_POUND] = ACTIONS(1206),
    [anon_sym_DOLLAR] = ACTIONS(1206),
    [anon_sym_PERCENT] = ACTIONS(1206),
    [anon_sym_AMP] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_COMMA] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_SLASH] = ACTIONS(1206),
    [anon_sym_COLON] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1206),
    [anon_sym_EQ] = ACTIONS(1206),
    [anon_sym_QMARK] = ACTIONS(1206),
    [anon_sym_AT] = ACTIONS(1206),
    [anon_sym_BSLASH] = ACTIONS(1206),
    [anon_sym_CARET] = ACTIONS(1206),
    [anon_sym__] = ACTIONS(1206),
    [anon_sym_BQUOTE] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1206),
    [anon_sym_PIPE] = ACTIONS(1206),
    [anon_sym_RBRACE] = ACTIONS(1206),
    [anon_sym_TILDE] = ACTIONS(1206),
    [anon_sym_LPAREN] = ACTIONS(1206),
    [anon_sym_RPAREN] = ACTIONS(1206),
    [aux_sym__word_token1] = ACTIONS(1206),
    [aux_sym__word_token2] = ACTIONS(1206),
    [aux_sym__word_token3] = ACTIONS(1206),
    [sym__whitespace] = ACTIONS(1206),
    [sym__soft_line_ending] = ACTIONS(1206),
    [sym__block_close] = ACTIONS(1206),
    [sym__block_quote_start] = ACTIONS(1206),
    [sym__indented_chunk_start] = ACTIONS(1206),
    [sym_atx_h1_marker] = ACTIONS(1206),
    [sym_atx_h2_marker] = ACTIONS(1206),
    [sym_atx_h3_marker] = ACTIONS(1206),
    [sym_atx_h4_marker] = ACTIONS(1206),
    [sym_atx_h5_marker] = ACTIONS(1206),
    [sym_atx_h6_marker] = ACTIONS(1206),
    [sym__thematic_break] = ACTIONS(1206),
    [sym__list_marker_minus] = ACTIONS(1206),
    [sym__list_marker_plus] = ACTIONS(1206),
    [sym__list_marker_star] = ACTIONS(1206),
    [sym__list_marker_parenthesis] = ACTIONS(1206),
    [sym__list_marker_dot] = ACTIONS(1206),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1206),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1206),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1206),
    [sym__blank_line_start] = ACTIONS(1206),
    [sym__html_block_1_start] = ACTIONS(1206),
    [sym__html_block_2_start] = ACTIONS(1206),
    [sym__html_block_3_start] = ACTIONS(1206),
    [sym__html_block_4_start] = ACTIONS(1206),
    [sym__html_block_5_start] = ACTIONS(1206),
    [sym__html_block_6_start] = ACTIONS(1206),
    [sym__html_block_7_start] = ACTIONS(1206),
    [sym__pipe_table_start] = ACTIONS(1206),
  },
  [STATE(314)] = {
    [ts_builtin_sym_end] = ACTIONS(1660),
    [anon_sym_LBRACK] = ACTIONS(1658),
    [anon_sym_RBRACK] = ACTIONS(1660),
    [anon_sym_LT] = ACTIONS(1660),
    [anon_sym_GT] = ACTIONS(1660),
    [anon_sym_BANG] = ACTIONS(1660),
    [anon_sym_DQUOTE] = ACTIONS(1660),
    [anon_sym_POUND] = ACTIONS(1660),
    [anon_sym_DOLLAR] = ACTIONS(1660),
    [anon_sym_PERCENT] = ACTIONS(1660),
    [anon_sym_AMP] = ACTIONS(1660),
    [anon_sym_SQUOTE] = ACTIONS(1660),
    [anon_sym_STAR] = ACTIONS(1660),
    [anon_sym_PLUS] = ACTIONS(1660),
    [anon_sym_COMMA] = ACTIONS(1660),
    [anon_sym_DASH] = ACTIONS(1660),
    [anon_sym_DOT] = ACTIONS(1660),
    [anon_sym_SLASH] = ACTIONS(1660),
    [anon_sym_COLON] = ACTIONS(1660),
    [anon_sym_SEMI] = ACTIONS(1660),
    [anon_sym_EQ] = ACTIONS(1660),
    [anon_sym_QMARK] = ACTIONS(1660),
    [anon_sym_AT] = ACTIONS(1660),
    [anon_sym_BSLASH] = ACTIONS(1660),
    [anon_sym_CARET] = ACTIONS(1660),
    [anon_sym__] = ACTIONS(1660),
    [anon_sym_BQUOTE] = ACTIONS(1660),
    [anon_sym_LBRACE] = ACTIONS(1660),
    [anon_sym_PIPE] = ACTIONS(1660),
    [anon_sym_RBRACE] = ACTIONS(1660),
    [anon_sym_TILDE] = ACTIONS(1660),
    [anon_sym_LPAREN] = ACTIONS(1660),
    [anon_sym_RPAREN] = ACTIONS(1660),
    [aux_sym__word_token1] = ACTIONS(1660),
    [aux_sym__word_token2] = ACTIONS(1660),
    [aux_sym__word_token3] = ACTIONS(1660),
    [sym__whitespace] = ACTIONS(1660),
    [sym__soft_line_ending] = ACTIONS(1660),
    [sym__block_quote_start] = ACTIONS(1660),
    [sym__indented_chunk_start] = ACTIONS(1660),
    [sym_atx_h1_marker] = ACTIONS(1660),
    [sym_atx_h2_marker] = ACTIONS(1660),
    [sym_atx_h3_marker] = ACTIONS(1660),
    [sym_atx_h4_marker] = ACTIONS(1660),
    [sym_atx_h5_marker] = ACTIONS(1660),
    [sym_atx_h6_marker] = ACTIONS(1660),
    [sym__thematic_break] = ACTIONS(1660),
    [sym__list_marker_minus] = ACTIONS(1660),
    [sym__list_marker_plus] = ACTIONS(1660),
    [sym__list_marker_star] = ACTIONS(1660),
    [sym__list_marker_parenthesis] = ACTIONS(1660),
    [sym__list_marker_dot] = ACTIONS(1660),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1660),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1660),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1660),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1660),
    [sym__blank_line_start] = ACTIONS(1660),
    [sym__html_block_1_start] = ACTIONS(1660),
    [sym__html_block_2_start] = ACTIONS(1660),
    [sym__html_block_3_start] = ACTIONS(1660),
    [sym__html_block_4_start] = ACTIONS(1660),
    [sym__html_block_5_start] = ACTIONS(1660),
    [sym__html_block_6_start] = ACTIONS(1660),
    [sym__html_block_7_start] = ACTIONS(1660),
    [sym__pipe_table_start] = ACTIONS(1660),
  },
  [STATE(315)] = {
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_LT] = ACTIONS(1706),
    [anon_sym_GT] = ACTIONS(1706),
    [anon_sym_BANG] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [anon_sym_POUND] = ACTIONS(1706),
    [anon_sym_DOLLAR] = ACTIONS(1706),
    [anon_sym_PERCENT] = ACTIONS(1706),
    [anon_sym_AMP] = ACTIONS(1706),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_STAR] = ACTIONS(1706),
    [anon_sym_PLUS] = ACTIONS(1706),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_DASH] = ACTIONS(1706),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_SLASH] = ACTIONS(1706),
    [anon_sym_COLON] = ACTIONS(1706),
    [anon_sym_SEMI] = ACTIONS(1706),
    [anon_sym_EQ] = ACTIONS(1706),
    [anon_sym_QMARK] = ACTIONS(1706),
    [anon_sym_AT] = ACTIONS(1706),
    [anon_sym_BSLASH] = ACTIONS(1706),
    [anon_sym_CARET] = ACTIONS(1706),
    [anon_sym__] = ACTIONS(1706),
    [anon_sym_BQUOTE] = ACTIONS(1706),
    [anon_sym_LBRACE] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1706),
    [anon_sym_RBRACE] = ACTIONS(1706),
    [anon_sym_TILDE] = ACTIONS(1706),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_RPAREN] = ACTIONS(1706),
    [aux_sym__word_token1] = ACTIONS(1706),
    [aux_sym__word_token2] = ACTIONS(1706),
    [aux_sym__word_token3] = ACTIONS(1706),
    [sym__whitespace] = ACTIONS(1706),
    [sym__soft_line_ending] = ACTIONS(1706),
    [sym__block_close] = ACTIONS(1706),
    [sym__block_quote_start] = ACTIONS(1706),
    [sym__indented_chunk_start] = ACTIONS(1706),
    [sym_atx_h1_marker] = ACTIONS(1706),
    [sym_atx_h2_marker] = ACTIONS(1706),
    [sym_atx_h3_marker] = ACTIONS(1706),
    [sym_atx_h4_marker] = ACTIONS(1706),
    [sym_atx_h5_marker] = ACTIONS(1706),
    [sym_atx_h6_marker] = ACTIONS(1706),
    [sym__thematic_break] = ACTIONS(1706),
    [sym__list_marker_minus] = ACTIONS(1706),
    [sym__list_marker_plus] = ACTIONS(1706),
    [sym__list_marker_star] = ACTIONS(1706),
    [sym__list_marker_parenthesis] = ACTIONS(1706),
    [sym__list_marker_dot] = ACTIONS(1706),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1706),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1706),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1706),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1706),
    [sym__blank_line_start] = ACTIONS(1706),
    [sym__html_block_1_start] = ACTIONS(1706),
    [sym__html_block_2_start] = ACTIONS(1706),
    [sym__html_block_3_start] = ACTIONS(1706),
    [sym__html_block_4_start] = ACTIONS(1706),
    [sym__html_block_5_start] = ACTIONS(1706),
    [sym__html_block_6_start] = ACTIONS(1706),
    [sym__html_block_7_start] = ACTIONS(1706),
    [sym__pipe_table_start] = ACTIONS(1706),
  },
  [STATE(316)] = {
    [anon_sym_LBRACK] = ACTIONS(1540),
    [anon_sym_RBRACK] = ACTIONS(1538),
    [anon_sym_LT] = ACTIONS(1538),
    [anon_sym_GT] = ACTIONS(1538),
    [anon_sym_BANG] = ACTIONS(1538),
    [anon_sym_DQUOTE] = ACTIONS(1538),
    [anon_sym_POUND] = ACTIONS(1538),
    [anon_sym_DOLLAR] = ACTIONS(1538),
    [anon_sym_PERCENT] = ACTIONS(1538),
    [anon_sym_AMP] = ACTIONS(1538),
    [anon_sym_SQUOTE] = ACTIONS(1538),
    [anon_sym_STAR] = ACTIONS(1538),
    [anon_sym_PLUS] = ACTIONS(1538),
    [anon_sym_COMMA] = ACTIONS(1538),
    [anon_sym_DASH] = ACTIONS(1538),
    [anon_sym_DOT] = ACTIONS(1538),
    [anon_sym_SLASH] = ACTIONS(1538),
    [anon_sym_COLON] = ACTIONS(1538),
    [anon_sym_SEMI] = ACTIONS(1538),
    [anon_sym_EQ] = ACTIONS(1538),
    [anon_sym_QMARK] = ACTIONS(1538),
    [anon_sym_AT] = ACTIONS(1538),
    [anon_sym_BSLASH] = ACTIONS(1538),
    [anon_sym_CARET] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(1538),
    [anon_sym_BQUOTE] = ACTIONS(1538),
    [anon_sym_LBRACE] = ACTIONS(1538),
    [anon_sym_PIPE] = ACTIONS(1538),
    [anon_sym_RBRACE] = ACTIONS(1538),
    [anon_sym_TILDE] = ACTIONS(1538),
    [anon_sym_LPAREN] = ACTIONS(1538),
    [anon_sym_RPAREN] = ACTIONS(1538),
    [aux_sym__word_token1] = ACTIONS(1538),
    [aux_sym__word_token2] = ACTIONS(1538),
    [aux_sym__word_token3] = ACTIONS(1538),
    [sym__whitespace] = ACTIONS(1538),
    [sym__soft_line_ending] = ACTIONS(1538),
    [sym__block_close] = ACTIONS(1538),
    [sym__block_quote_start] = ACTIONS(1538),
    [sym__indented_chunk_start] = ACTIONS(1538),
    [sym_atx_h1_marker] = ACTIONS(1538),
    [sym_atx_h2_marker] = ACTIONS(1538),
    [sym_atx_h3_marker] = ACTIONS(1538),
    [sym_atx_h4_marker] = ACTIONS(1538),
    [sym_atx_h5_marker] = ACTIONS(1538),
    [sym_atx_h6_marker] = ACTIONS(1538),
    [sym__thematic_break] = ACTIONS(1538),
    [sym__list_marker_minus] = ACTIONS(1538),
    [sym__list_marker_plus] = ACTIONS(1538),
    [sym__list_marker_star] = ACTIONS(1538),
    [sym__list_marker_parenthesis] = ACTIONS(1538),
    [sym__list_marker_dot] = ACTIONS(1538),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1538),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1538),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1538),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1538),
    [sym__blank_line_start] = ACTIONS(1538),
    [sym__html_block_1_start] = ACTIONS(1538),
    [sym__html_block_2_start] = ACTIONS(1538),
    [sym__html_block_3_start] = ACTIONS(1538),
    [sym__html_block_4_start] = ACTIONS(1538),
    [sym__html_block_5_start] = ACTIONS(1538),
    [sym__html_block_6_start] = ACTIONS(1538),
    [sym__html_block_7_start] = ACTIONS(1538),
    [sym__pipe_table_start] = ACTIONS(1538),
  },
  [STATE(317)] = {
    [anon_sym_LBRACK] = ACTIONS(1544),
    [anon_sym_RBRACK] = ACTIONS(1542),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_BANG] = ACTIONS(1542),
    [anon_sym_DQUOTE] = ACTIONS(1542),
    [anon_sym_POUND] = ACTIONS(1542),
    [anon_sym_DOLLAR] = ACTIONS(1542),
    [anon_sym_PERCENT] = ACTIONS(1542),
    [anon_sym_AMP] = ACTIONS(1542),
    [anon_sym_SQUOTE] = ACTIONS(1542),
    [anon_sym_STAR] = ACTIONS(1542),
    [anon_sym_PLUS] = ACTIONS(1542),
    [anon_sym_COMMA] = ACTIONS(1542),
    [anon_sym_DASH] = ACTIONS(1542),
    [anon_sym_DOT] = ACTIONS(1542),
    [anon_sym_SLASH] = ACTIONS(1542),
    [anon_sym_COLON] = ACTIONS(1542),
    [anon_sym_SEMI] = ACTIONS(1542),
    [anon_sym_EQ] = ACTIONS(1542),
    [anon_sym_QMARK] = ACTIONS(1542),
    [anon_sym_AT] = ACTIONS(1542),
    [anon_sym_BSLASH] = ACTIONS(1542),
    [anon_sym_CARET] = ACTIONS(1542),
    [anon_sym__] = ACTIONS(1542),
    [anon_sym_BQUOTE] = ACTIONS(1542),
    [anon_sym_LBRACE] = ACTIONS(1542),
    [anon_sym_PIPE] = ACTIONS(1542),
    [anon_sym_RBRACE] = ACTIONS(1542),
    [anon_sym_TILDE] = ACTIONS(1542),
    [anon_sym_LPAREN] = ACTIONS(1542),
    [anon_sym_RPAREN] = ACTIONS(1542),
    [aux_sym__word_token1] = ACTIONS(1542),
    [aux_sym__word_token2] = ACTIONS(1542),
    [aux_sym__word_token3] = ACTIONS(1542),
    [sym__whitespace] = ACTIONS(1542),
    [sym__soft_line_ending] = ACTIONS(1542),
    [sym__block_close] = ACTIONS(1542),
    [sym__block_quote_start] = ACTIONS(1542),
    [sym__indented_chunk_start] = ACTIONS(1542),
    [sym_atx_h1_marker] = ACTIONS(1542),
    [sym_atx_h2_marker] = ACTIONS(1542),
    [sym_atx_h3_marker] = ACTIONS(1542),
    [sym_atx_h4_marker] = ACTIONS(1542),
    [sym_atx_h5_marker] = ACTIONS(1542),
    [sym_atx_h6_marker] = ACTIONS(1542),
    [sym__thematic_break] = ACTIONS(1542),
    [sym__list_marker_minus] = ACTIONS(1542),
    [sym__list_marker_plus] = ACTIONS(1542),
    [sym__list_marker_star] = ACTIONS(1542),
    [sym__list_marker_parenthesis] = ACTIONS(1542),
    [sym__list_marker_dot] = ACTIONS(1542),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1542),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1542),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1542),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1542),
    [sym__blank_line_start] = ACTIONS(1542),
    [sym__html_block_1_start] = ACTIONS(1542),
    [sym__html_block_2_start] = ACTIONS(1542),
    [sym__html_block_3_start] = ACTIONS(1542),
    [sym__html_block_4_start] = ACTIONS(1542),
    [sym__html_block_5_start] = ACTIONS(1542),
    [sym__html_block_6_start] = ACTIONS(1542),
    [sym__html_block_7_start] = ACTIONS(1542),
    [sym__pipe_table_start] = ACTIONS(1542),
  },
  [STATE(318)] = {
    [anon_sym_LBRACK] = ACTIONS(1222),
    [anon_sym_RBRACK] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1224),
    [anon_sym_GT] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [anon_sym_POUND] = ACTIONS(1224),
    [anon_sym_DOLLAR] = ACTIONS(1224),
    [anon_sym_PERCENT] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_PLUS] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_SLASH] = ACTIONS(1224),
    [anon_sym_COLON] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_EQ] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1224),
    [anon_sym_BSLASH] = ACTIONS(1224),
    [anon_sym_CARET] = ACTIONS(1224),
    [anon_sym__] = ACTIONS(1224),
    [anon_sym_BQUOTE] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_PIPE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_RPAREN] = ACTIONS(1224),
    [aux_sym__word_token1] = ACTIONS(1224),
    [aux_sym__word_token2] = ACTIONS(1224),
    [aux_sym__word_token3] = ACTIONS(1224),
    [sym__whitespace] = ACTIONS(1224),
    [sym__soft_line_ending] = ACTIONS(1224),
    [sym__block_close] = ACTIONS(1224),
    [sym__block_quote_start] = ACTIONS(1224),
    [sym__indented_chunk_start] = ACTIONS(1224),
    [sym_atx_h1_marker] = ACTIONS(1224),
    [sym_atx_h2_marker] = ACTIONS(1224),
    [sym_atx_h3_marker] = ACTIONS(1224),
    [sym_atx_h4_marker] = ACTIONS(1224),
    [sym_atx_h5_marker] = ACTIONS(1224),
    [sym_atx_h6_marker] = ACTIONS(1224),
    [sym__thematic_break] = ACTIONS(1224),
    [sym__list_marker_minus] = ACTIONS(1224),
    [sym__list_marker_plus] = ACTIONS(1224),
    [sym__list_marker_star] = ACTIONS(1224),
    [sym__list_marker_parenthesis] = ACTIONS(1224),
    [sym__list_marker_dot] = ACTIONS(1224),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1224),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1224),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1224),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1224),
    [sym__blank_line_start] = ACTIONS(1224),
    [sym__html_block_1_start] = ACTIONS(1224),
    [sym__html_block_2_start] = ACTIONS(1224),
    [sym__html_block_3_start] = ACTIONS(1224),
    [sym__html_block_4_start] = ACTIONS(1224),
    [sym__html_block_5_start] = ACTIONS(1224),
    [sym__html_block_6_start] = ACTIONS(1224),
    [sym__html_block_7_start] = ACTIONS(1224),
    [sym__pipe_table_start] = ACTIONS(1224),
  },
  [STATE(319)] = {
    [ts_builtin_sym_end] = ACTIONS(1294),
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_RBRACK] = ACTIONS(1294),
    [anon_sym_LT] = ACTIONS(1294),
    [anon_sym_GT] = ACTIONS(1294),
    [anon_sym_BANG] = ACTIONS(1294),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [anon_sym_POUND] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1294),
    [anon_sym_PERCENT] = ACTIONS(1294),
    [anon_sym_AMP] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_COMMA] = ACTIONS(1294),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_DOT] = ACTIONS(1294),
    [anon_sym_SLASH] = ACTIONS(1294),
    [anon_sym_COLON] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1294),
    [anon_sym_EQ] = ACTIONS(1294),
    [anon_sym_QMARK] = ACTIONS(1294),
    [anon_sym_AT] = ACTIONS(1294),
    [anon_sym_BSLASH] = ACTIONS(1294),
    [anon_sym_CARET] = ACTIONS(1294),
    [anon_sym__] = ACTIONS(1294),
    [anon_sym_BQUOTE] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1294),
    [anon_sym_PIPE] = ACTIONS(1294),
    [anon_sym_RBRACE] = ACTIONS(1294),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(1294),
    [anon_sym_RPAREN] = ACTIONS(1294),
    [aux_sym__word_token1] = ACTIONS(1294),
    [aux_sym__word_token2] = ACTIONS(1294),
    [aux_sym__word_token3] = ACTIONS(1294),
    [sym__whitespace] = ACTIONS(1294),
    [sym__soft_line_ending] = ACTIONS(1294),
    [sym__block_quote_start] = ACTIONS(1294),
    [sym__indented_chunk_start] = ACTIONS(1294),
    [sym_atx_h1_marker] = ACTIONS(1294),
    [sym_atx_h2_marker] = ACTIONS(1294),
    [sym_atx_h3_marker] = ACTIONS(1294),
    [sym_atx_h4_marker] = ACTIONS(1294),
    [sym_atx_h5_marker] = ACTIONS(1294),
    [sym_atx_h6_marker] = ACTIONS(1294),
    [sym__thematic_break] = ACTIONS(1294),
    [sym__list_marker_minus] = ACTIONS(1294),
    [sym__list_marker_plus] = ACTIONS(1294),
    [sym__list_marker_star] = ACTIONS(1294),
    [sym__list_marker_parenthesis] = ACTIONS(1294),
    [sym__list_marker_dot] = ACTIONS(1294),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1294),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1294),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1294),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1294),
    [sym__blank_line_start] = ACTIONS(1294),
    [sym__html_block_1_start] = ACTIONS(1294),
    [sym__html_block_2_start] = ACTIONS(1294),
    [sym__html_block_3_start] = ACTIONS(1294),
    [sym__html_block_4_start] = ACTIONS(1294),
    [sym__html_block_5_start] = ACTIONS(1294),
    [sym__html_block_6_start] = ACTIONS(1294),
    [sym__html_block_7_start] = ACTIONS(1294),
    [sym__pipe_table_start] = ACTIONS(1294),
  },
  [STATE(320)] = {
    [anon_sym_LBRACK] = ACTIONS(1548),
    [anon_sym_RBRACK] = ACTIONS(1546),
    [anon_sym_LT] = ACTIONS(1546),
    [anon_sym_GT] = ACTIONS(1546),
    [anon_sym_BANG] = ACTIONS(1546),
    [anon_sym_DQUOTE] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_DOLLAR] = ACTIONS(1546),
    [anon_sym_PERCENT] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1546),
    [anon_sym_STAR] = ACTIONS(1546),
    [anon_sym_PLUS] = ACTIONS(1546),
    [anon_sym_COMMA] = ACTIONS(1546),
    [anon_sym_DASH] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1546),
    [anon_sym_SLASH] = ACTIONS(1546),
    [anon_sym_COLON] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1546),
    [anon_sym_EQ] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_AT] = ACTIONS(1546),
    [anon_sym_BSLASH] = ACTIONS(1546),
    [anon_sym_CARET] = ACTIONS(1546),
    [anon_sym__] = ACTIONS(1546),
    [anon_sym_BQUOTE] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_RBRACE] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1546),
    [anon_sym_RPAREN] = ACTIONS(1546),
    [aux_sym__word_token1] = ACTIONS(1546),
    [aux_sym__word_token2] = ACTIONS(1546),
    [aux_sym__word_token3] = ACTIONS(1546),
    [sym__whitespace] = ACTIONS(1546),
    [sym__soft_line_ending] = ACTIONS(1546),
    [sym__block_close] = ACTIONS(1546),
    [sym__block_quote_start] = ACTIONS(1546),
    [sym__indented_chunk_start] = ACTIONS(1546),
    [sym_atx_h1_marker] = ACTIONS(1546),
    [sym_atx_h2_marker] = ACTIONS(1546),
    [sym_atx_h3_marker] = ACTIONS(1546),
    [sym_atx_h4_marker] = ACTIONS(1546),
    [sym_atx_h5_marker] = ACTIONS(1546),
    [sym_atx_h6_marker] = ACTIONS(1546),
    [sym__thematic_break] = ACTIONS(1546),
    [sym__list_marker_minus] = ACTIONS(1546),
    [sym__list_marker_plus] = ACTIONS(1546),
    [sym__list_marker_star] = ACTIONS(1546),
    [sym__list_marker_parenthesis] = ACTIONS(1546),
    [sym__list_marker_dot] = ACTIONS(1546),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1546),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1546),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1546),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1546),
    [sym__blank_line_start] = ACTIONS(1546),
    [sym__html_block_1_start] = ACTIONS(1546),
    [sym__html_block_2_start] = ACTIONS(1546),
    [sym__html_block_3_start] = ACTIONS(1546),
    [sym__html_block_4_start] = ACTIONS(1546),
    [sym__html_block_5_start] = ACTIONS(1546),
    [sym__html_block_6_start] = ACTIONS(1546),
    [sym__html_block_7_start] = ACTIONS(1546),
    [sym__pipe_table_start] = ACTIONS(1546),
  },
  [STATE(321)] = {
    [anon_sym_LBRACK] = ACTIONS(1552),
    [anon_sym_RBRACK] = ACTIONS(1550),
    [anon_sym_LT] = ACTIONS(1550),
    [anon_sym_GT] = ACTIONS(1550),
    [anon_sym_BANG] = ACTIONS(1550),
    [anon_sym_DQUOTE] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_DOLLAR] = ACTIONS(1550),
    [anon_sym_PERCENT] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1550),
    [anon_sym_STAR] = ACTIONS(1550),
    [anon_sym_PLUS] = ACTIONS(1550),
    [anon_sym_COMMA] = ACTIONS(1550),
    [anon_sym_DASH] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1550),
    [anon_sym_SLASH] = ACTIONS(1550),
    [anon_sym_COLON] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_EQ] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_AT] = ACTIONS(1550),
    [anon_sym_BSLASH] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1550),
    [anon_sym__] = ACTIONS(1550),
    [anon_sym_BQUOTE] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_RBRACE] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1550),
    [anon_sym_RPAREN] = ACTIONS(1550),
    [aux_sym__word_token1] = ACTIONS(1550),
    [aux_sym__word_token2] = ACTIONS(1550),
    [aux_sym__word_token3] = ACTIONS(1550),
    [sym__whitespace] = ACTIONS(1550),
    [sym__soft_line_ending] = ACTIONS(1550),
    [sym__block_close] = ACTIONS(1550),
    [sym__block_quote_start] = ACTIONS(1550),
    [sym__indented_chunk_start] = ACTIONS(1550),
    [sym_atx_h1_marker] = ACTIONS(1550),
    [sym_atx_h2_marker] = ACTIONS(1550),
    [sym_atx_h3_marker] = ACTIONS(1550),
    [sym_atx_h4_marker] = ACTIONS(1550),
    [sym_atx_h5_marker] = ACTIONS(1550),
    [sym_atx_h6_marker] = ACTIONS(1550),
    [sym__thematic_break] = ACTIONS(1550),
    [sym__list_marker_minus] = ACTIONS(1550),
    [sym__list_marker_plus] = ACTIONS(1550),
    [sym__list_marker_star] = ACTIONS(1550),
    [sym__list_marker_parenthesis] = ACTIONS(1550),
    [sym__list_marker_dot] = ACTIONS(1550),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1550),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1550),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1550),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1550),
    [sym__blank_line_start] = ACTIONS(1550),
    [sym__html_block_1_start] = ACTIONS(1550),
    [sym__html_block_2_start] = ACTIONS(1550),
    [sym__html_block_3_start] = ACTIONS(1550),
    [sym__html_block_4_start] = ACTIONS(1550),
    [sym__html_block_5_start] = ACTIONS(1550),
    [sym__html_block_6_start] = ACTIONS(1550),
    [sym__html_block_7_start] = ACTIONS(1550),
    [sym__pipe_table_start] = ACTIONS(1550),
  },
  [STATE(322)] = {
    [anon_sym_LBRACK] = ACTIONS(1556),
    [anon_sym_RBRACK] = ACTIONS(1554),
    [anon_sym_LT] = ACTIONS(1554),
    [anon_sym_GT] = ACTIONS(1554),
    [anon_sym_BANG] = ACTIONS(1554),
    [anon_sym_DQUOTE] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_DOLLAR] = ACTIONS(1554),
    [anon_sym_PERCENT] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1554),
    [anon_sym_STAR] = ACTIONS(1554),
    [anon_sym_PLUS] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1554),
    [anon_sym_SLASH] = ACTIONS(1554),
    [anon_sym_COLON] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_EQ] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_AT] = ACTIONS(1554),
    [anon_sym_BSLASH] = ACTIONS(1554),
    [anon_sym_CARET] = ACTIONS(1554),
    [anon_sym__] = ACTIONS(1554),
    [anon_sym_BQUOTE] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_RBRACE] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1554),
    [anon_sym_RPAREN] = ACTIONS(1554),
    [aux_sym__word_token1] = ACTIONS(1554),
    [aux_sym__word_token2] = ACTIONS(1554),
    [aux_sym__word_token3] = ACTIONS(1554),
    [sym__whitespace] = ACTIONS(1554),
    [sym__soft_line_ending] = ACTIONS(1554),
    [sym__block_close] = ACTIONS(1554),
    [sym__block_quote_start] = ACTIONS(1554),
    [sym__indented_chunk_start] = ACTIONS(1554),
    [sym_atx_h1_marker] = ACTIONS(1554),
    [sym_atx_h2_marker] = ACTIONS(1554),
    [sym_atx_h3_marker] = ACTIONS(1554),
    [sym_atx_h4_marker] = ACTIONS(1554),
    [sym_atx_h5_marker] = ACTIONS(1554),
    [sym_atx_h6_marker] = ACTIONS(1554),
    [sym__thematic_break] = ACTIONS(1554),
    [sym__list_marker_minus] = ACTIONS(1554),
    [sym__list_marker_plus] = ACTIONS(1554),
    [sym__list_marker_star] = ACTIONS(1554),
    [sym__list_marker_parenthesis] = ACTIONS(1554),
    [sym__list_marker_dot] = ACTIONS(1554),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1554),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1554),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1554),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1554),
    [sym__blank_line_start] = ACTIONS(1554),
    [sym__html_block_1_start] = ACTIONS(1554),
    [sym__html_block_2_start] = ACTIONS(1554),
    [sym__html_block_3_start] = ACTIONS(1554),
    [sym__html_block_4_start] = ACTIONS(1554),
    [sym__html_block_5_start] = ACTIONS(1554),
    [sym__html_block_6_start] = ACTIONS(1554),
    [sym__html_block_7_start] = ACTIONS(1554),
    [sym__pipe_table_start] = ACTIONS(1554),
  },
  [STATE(323)] = {
    [anon_sym_LBRACK] = ACTIONS(1560),
    [anon_sym_RBRACK] = ACTIONS(1558),
    [anon_sym_LT] = ACTIONS(1558),
    [anon_sym_GT] = ACTIONS(1558),
    [anon_sym_BANG] = ACTIONS(1558),
    [anon_sym_DQUOTE] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [anon_sym_PERCENT] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1558),
    [anon_sym_STAR] = ACTIONS(1558),
    [anon_sym_PLUS] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1558),
    [anon_sym_DASH] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1558),
    [anon_sym_SLASH] = ACTIONS(1558),
    [anon_sym_COLON] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_EQ] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_AT] = ACTIONS(1558),
    [anon_sym_BSLASH] = ACTIONS(1558),
    [anon_sym_CARET] = ACTIONS(1558),
    [anon_sym__] = ACTIONS(1558),
    [anon_sym_BQUOTE] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_RBRACE] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1558),
    [anon_sym_RPAREN] = ACTIONS(1558),
    [aux_sym__word_token1] = ACTIONS(1558),
    [aux_sym__word_token2] = ACTIONS(1558),
    [aux_sym__word_token3] = ACTIONS(1558),
    [sym__whitespace] = ACTIONS(1558),
    [sym__soft_line_ending] = ACTIONS(1558),
    [sym__block_close] = ACTIONS(1558),
    [sym__block_quote_start] = ACTIONS(1558),
    [sym__indented_chunk_start] = ACTIONS(1558),
    [sym_atx_h1_marker] = ACTIONS(1558),
    [sym_atx_h2_marker] = ACTIONS(1558),
    [sym_atx_h3_marker] = ACTIONS(1558),
    [sym_atx_h4_marker] = ACTIONS(1558),
    [sym_atx_h5_marker] = ACTIONS(1558),
    [sym_atx_h6_marker] = ACTIONS(1558),
    [sym__thematic_break] = ACTIONS(1558),
    [sym__list_marker_minus] = ACTIONS(1558),
    [sym__list_marker_plus] = ACTIONS(1558),
    [sym__list_marker_star] = ACTIONS(1558),
    [sym__list_marker_parenthesis] = ACTIONS(1558),
    [sym__list_marker_dot] = ACTIONS(1558),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1558),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1558),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1558),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1558),
    [sym__blank_line_start] = ACTIONS(1558),
    [sym__html_block_1_start] = ACTIONS(1558),
    [sym__html_block_2_start] = ACTIONS(1558),
    [sym__html_block_3_start] = ACTIONS(1558),
    [sym__html_block_4_start] = ACTIONS(1558),
    [sym__html_block_5_start] = ACTIONS(1558),
    [sym__html_block_6_start] = ACTIONS(1558),
    [sym__html_block_7_start] = ACTIONS(1558),
    [sym__pipe_table_start] = ACTIONS(1558),
  },
  [STATE(324)] = {
    [anon_sym_LBRACK] = ACTIONS(1564),
    [anon_sym_RBRACK] = ACTIONS(1562),
    [anon_sym_LT] = ACTIONS(1562),
    [anon_sym_GT] = ACTIONS(1562),
    [anon_sym_BANG] = ACTIONS(1562),
    [anon_sym_DQUOTE] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_DOLLAR] = ACTIONS(1562),
    [anon_sym_PERCENT] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1562),
    [anon_sym_STAR] = ACTIONS(1562),
    [anon_sym_PLUS] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1562),
    [anon_sym_DASH] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1562),
    [anon_sym_SLASH] = ACTIONS(1562),
    [anon_sym_COLON] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_EQ] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_AT] = ACTIONS(1562),
    [anon_sym_BSLASH] = ACTIONS(1562),
    [anon_sym_CARET] = ACTIONS(1562),
    [anon_sym__] = ACTIONS(1562),
    [anon_sym_BQUOTE] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_RBRACE] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1562),
    [anon_sym_RPAREN] = ACTIONS(1562),
    [aux_sym__word_token1] = ACTIONS(1562),
    [aux_sym__word_token2] = ACTIONS(1562),
    [aux_sym__word_token3] = ACTIONS(1562),
    [sym__whitespace] = ACTIONS(1562),
    [sym__soft_line_ending] = ACTIONS(1562),
    [sym__block_close] = ACTIONS(1562),
    [sym__block_quote_start] = ACTIONS(1562),
    [sym__indented_chunk_start] = ACTIONS(1562),
    [sym_atx_h1_marker] = ACTIONS(1562),
    [sym_atx_h2_marker] = ACTIONS(1562),
    [sym_atx_h3_marker] = ACTIONS(1562),
    [sym_atx_h4_marker] = ACTIONS(1562),
    [sym_atx_h5_marker] = ACTIONS(1562),
    [sym_atx_h6_marker] = ACTIONS(1562),
    [sym__thematic_break] = ACTIONS(1562),
    [sym__list_marker_minus] = ACTIONS(1562),
    [sym__list_marker_plus] = ACTIONS(1562),
    [sym__list_marker_star] = ACTIONS(1562),
    [sym__list_marker_parenthesis] = ACTIONS(1562),
    [sym__list_marker_dot] = ACTIONS(1562),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1562),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1562),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1562),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1562),
    [sym__blank_line_start] = ACTIONS(1562),
    [sym__html_block_1_start] = ACTIONS(1562),
    [sym__html_block_2_start] = ACTIONS(1562),
    [sym__html_block_3_start] = ACTIONS(1562),
    [sym__html_block_4_start] = ACTIONS(1562),
    [sym__html_block_5_start] = ACTIONS(1562),
    [sym__html_block_6_start] = ACTIONS(1562),
    [sym__html_block_7_start] = ACTIONS(1562),
    [sym__pipe_table_start] = ACTIONS(1562),
  },
  [STATE(325)] = {
    [anon_sym_LBRACK] = ACTIONS(1568),
    [anon_sym_RBRACK] = ACTIONS(1566),
    [anon_sym_LT] = ACTIONS(1566),
    [anon_sym_GT] = ACTIONS(1566),
    [anon_sym_BANG] = ACTIONS(1566),
    [anon_sym_DQUOTE] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_DOLLAR] = ACTIONS(1566),
    [anon_sym_PERCENT] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1566),
    [anon_sym_PLUS] = ACTIONS(1566),
    [anon_sym_COMMA] = ACTIONS(1566),
    [anon_sym_DASH] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1566),
    [anon_sym_SLASH] = ACTIONS(1566),
    [anon_sym_COLON] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1566),
    [anon_sym_EQ] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_AT] = ACTIONS(1566),
    [anon_sym_BSLASH] = ACTIONS(1566),
    [anon_sym_CARET] = ACTIONS(1566),
    [anon_sym__] = ACTIONS(1566),
    [anon_sym_BQUOTE] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_RBRACE] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1566),
    [anon_sym_RPAREN] = ACTIONS(1566),
    [aux_sym__word_token1] = ACTIONS(1566),
    [aux_sym__word_token2] = ACTIONS(1566),
    [aux_sym__word_token3] = ACTIONS(1566),
    [sym__whitespace] = ACTIONS(1566),
    [sym__soft_line_ending] = ACTIONS(1566),
    [sym__block_close] = ACTIONS(1566),
    [sym__block_quote_start] = ACTIONS(1566),
    [sym__indented_chunk_start] = ACTIONS(1566),
    [sym_atx_h1_marker] = ACTIONS(1566),
    [sym_atx_h2_marker] = ACTIONS(1566),
    [sym_atx_h3_marker] = ACTIONS(1566),
    [sym_atx_h4_marker] = ACTIONS(1566),
    [sym_atx_h5_marker] = ACTIONS(1566),
    [sym_atx_h6_marker] = ACTIONS(1566),
    [sym__thematic_break] = ACTIONS(1566),
    [sym__list_marker_minus] = ACTIONS(1566),
    [sym__list_marker_plus] = ACTIONS(1566),
    [sym__list_marker_star] = ACTIONS(1566),
    [sym__list_marker_parenthesis] = ACTIONS(1566),
    [sym__list_marker_dot] = ACTIONS(1566),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1566),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1566),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1566),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1566),
    [sym__blank_line_start] = ACTIONS(1566),
    [sym__html_block_1_start] = ACTIONS(1566),
    [sym__html_block_2_start] = ACTIONS(1566),
    [sym__html_block_3_start] = ACTIONS(1566),
    [sym__html_block_4_start] = ACTIONS(1566),
    [sym__html_block_5_start] = ACTIONS(1566),
    [sym__html_block_6_start] = ACTIONS(1566),
    [sym__html_block_7_start] = ACTIONS(1566),
    [sym__pipe_table_start] = ACTIONS(1566),
  },
  [STATE(326)] = {
    [anon_sym_LBRACK] = ACTIONS(1257),
    [anon_sym_RBRACK] = ACTIONS(1259),
    [anon_sym_LT] = ACTIONS(1259),
    [anon_sym_GT] = ACTIONS(1259),
    [anon_sym_BANG] = ACTIONS(1259),
    [anon_sym_DQUOTE] = ACTIONS(1259),
    [anon_sym_POUND] = ACTIONS(1259),
    [anon_sym_DOLLAR] = ACTIONS(1259),
    [anon_sym_PERCENT] = ACTIONS(1259),
    [anon_sym_AMP] = ACTIONS(1259),
    [anon_sym_SQUOTE] = ACTIONS(1259),
    [anon_sym_STAR] = ACTIONS(1259),
    [anon_sym_PLUS] = ACTIONS(1259),
    [anon_sym_COMMA] = ACTIONS(1259),
    [anon_sym_DASH] = ACTIONS(1259),
    [anon_sym_DOT] = ACTIONS(1259),
    [anon_sym_SLASH] = ACTIONS(1259),
    [anon_sym_COLON] = ACTIONS(1259),
    [anon_sym_SEMI] = ACTIONS(1259),
    [anon_sym_EQ] = ACTIONS(1259),
    [anon_sym_QMARK] = ACTIONS(1259),
    [anon_sym_AT] = ACTIONS(1259),
    [anon_sym_BSLASH] = ACTIONS(1259),
    [anon_sym_CARET] = ACTIONS(1259),
    [anon_sym__] = ACTIONS(1259),
    [anon_sym_BQUOTE] = ACTIONS(1259),
    [anon_sym_LBRACE] = ACTIONS(1259),
    [anon_sym_PIPE] = ACTIONS(1259),
    [anon_sym_RBRACE] = ACTIONS(1259),
    [anon_sym_TILDE] = ACTIONS(1259),
    [anon_sym_LPAREN] = ACTIONS(1259),
    [anon_sym_RPAREN] = ACTIONS(1259),
    [aux_sym__word_token1] = ACTIONS(1259),
    [aux_sym__word_token2] = ACTIONS(1259),
    [aux_sym__word_token3] = ACTIONS(1259),
    [sym__whitespace] = ACTIONS(1259),
    [sym__soft_line_ending] = ACTIONS(1259),
    [sym__block_close] = ACTIONS(1259),
    [sym__block_quote_start] = ACTIONS(1259),
    [sym__indented_chunk_start] = ACTIONS(1259),
    [sym_atx_h1_marker] = ACTIONS(1259),
    [sym_atx_h2_marker] = ACTIONS(1259),
    [sym_atx_h3_marker] = ACTIONS(1259),
    [sym_atx_h4_marker] = ACTIONS(1259),
    [sym_atx_h5_marker] = ACTIONS(1259),
    [sym_atx_h6_marker] = ACTIONS(1259),
    [sym__thematic_break] = ACTIONS(1259),
    [sym__list_marker_minus] = ACTIONS(1259),
    [sym__list_marker_plus] = ACTIONS(1259),
    [sym__list_marker_star] = ACTIONS(1259),
    [sym__list_marker_parenthesis] = ACTIONS(1259),
    [sym__list_marker_dot] = ACTIONS(1259),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1259),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1259),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1259),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1259),
    [sym__blank_line_start] = ACTIONS(1259),
    [sym__html_block_1_start] = ACTIONS(1259),
    [sym__html_block_2_start] = ACTIONS(1259),
    [sym__html_block_3_start] = ACTIONS(1259),
    [sym__html_block_4_start] = ACTIONS(1259),
    [sym__html_block_5_start] = ACTIONS(1259),
    [sym__html_block_6_start] = ACTIONS(1259),
    [sym__html_block_7_start] = ACTIONS(1259),
    [sym__pipe_table_start] = ACTIONS(1259),
  },
  [STATE(327)] = {
    [ts_builtin_sym_end] = ACTIONS(1300),
    [anon_sym_LBRACK] = ACTIONS(1302),
    [anon_sym_RBRACK] = ACTIONS(1300),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_GT] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [anon_sym_POUND] = ACTIONS(1300),
    [anon_sym_DOLLAR] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_EQ] = ACTIONS(1300),
    [anon_sym_QMARK] = ACTIONS(1300),
    [anon_sym_AT] = ACTIONS(1300),
    [anon_sym_BSLASH] = ACTIONS(1300),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym__] = ACTIONS(1300),
    [anon_sym_BQUOTE] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_RPAREN] = ACTIONS(1300),
    [aux_sym__word_token1] = ACTIONS(1300),
    [aux_sym__word_token2] = ACTIONS(1300),
    [aux_sym__word_token3] = ACTIONS(1300),
    [sym__whitespace] = ACTIONS(1300),
    [sym__soft_line_ending] = ACTIONS(1300),
    [sym__block_quote_start] = ACTIONS(1300),
    [sym__indented_chunk_start] = ACTIONS(1300),
    [sym_atx_h1_marker] = ACTIONS(1300),
    [sym_atx_h2_marker] = ACTIONS(1300),
    [sym_atx_h3_marker] = ACTIONS(1300),
    [sym_atx_h4_marker] = ACTIONS(1300),
    [sym_atx_h5_marker] = ACTIONS(1300),
    [sym_atx_h6_marker] = ACTIONS(1300),
    [sym__thematic_break] = ACTIONS(1300),
    [sym__list_marker_minus] = ACTIONS(1300),
    [sym__list_marker_plus] = ACTIONS(1300),
    [sym__list_marker_star] = ACTIONS(1300),
    [sym__list_marker_parenthesis] = ACTIONS(1300),
    [sym__list_marker_dot] = ACTIONS(1300),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1300),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1300),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1300),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1300),
    [sym__blank_line_start] = ACTIONS(1300),
    [sym__html_block_1_start] = ACTIONS(1300),
    [sym__html_block_2_start] = ACTIONS(1300),
    [sym__html_block_3_start] = ACTIONS(1300),
    [sym__html_block_4_start] = ACTIONS(1300),
    [sym__html_block_5_start] = ACTIONS(1300),
    [sym__html_block_6_start] = ACTIONS(1300),
    [sym__html_block_7_start] = ACTIONS(1300),
    [sym__pipe_table_start] = ACTIONS(1300),
  },
  [STATE(328)] = {
    [ts_builtin_sym_end] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_RBRACK] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1318),
    [anon_sym_GT] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1318),
    [anon_sym_BSLASH] = ACTIONS(1318),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym__] = ACTIONS(1318),
    [anon_sym_BQUOTE] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_RPAREN] = ACTIONS(1318),
    [aux_sym__word_token1] = ACTIONS(1318),
    [aux_sym__word_token2] = ACTIONS(1318),
    [aux_sym__word_token3] = ACTIONS(1318),
    [sym__whitespace] = ACTIONS(1318),
    [sym__soft_line_ending] = ACTIONS(1318),
    [sym__block_quote_start] = ACTIONS(1318),
    [sym__indented_chunk_start] = ACTIONS(1318),
    [sym_atx_h1_marker] = ACTIONS(1318),
    [sym_atx_h2_marker] = ACTIONS(1318),
    [sym_atx_h3_marker] = ACTIONS(1318),
    [sym_atx_h4_marker] = ACTIONS(1318),
    [sym_atx_h5_marker] = ACTIONS(1318),
    [sym_atx_h6_marker] = ACTIONS(1318),
    [sym__thematic_break] = ACTIONS(1318),
    [sym__list_marker_minus] = ACTIONS(1318),
    [sym__list_marker_plus] = ACTIONS(1318),
    [sym__list_marker_star] = ACTIONS(1318),
    [sym__list_marker_parenthesis] = ACTIONS(1318),
    [sym__list_marker_dot] = ACTIONS(1318),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1318),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1318),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1318),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1318),
    [sym__blank_line_start] = ACTIONS(1318),
    [sym__html_block_1_start] = ACTIONS(1318),
    [sym__html_block_2_start] = ACTIONS(1318),
    [sym__html_block_3_start] = ACTIONS(1318),
    [sym__html_block_4_start] = ACTIONS(1318),
    [sym__html_block_5_start] = ACTIONS(1318),
    [sym__html_block_6_start] = ACTIONS(1318),
    [sym__html_block_7_start] = ACTIONS(1318),
    [sym__pipe_table_start] = ACTIONS(1318),
  },
  [STATE(329)] = {
    [ts_builtin_sym_end] = ACTIONS(1470),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_RBRACK] = ACTIONS(1470),
    [anon_sym_LT] = ACTIONS(1470),
    [anon_sym_GT] = ACTIONS(1470),
    [anon_sym_BANG] = ACTIONS(1470),
    [anon_sym_DQUOTE] = ACTIONS(1470),
    [anon_sym_POUND] = ACTIONS(1470),
    [anon_sym_DOLLAR] = ACTIONS(1470),
    [anon_sym_PERCENT] = ACTIONS(1470),
    [anon_sym_AMP] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_STAR] = ACTIONS(1470),
    [anon_sym_PLUS] = ACTIONS(1470),
    [anon_sym_COMMA] = ACTIONS(1470),
    [anon_sym_DASH] = ACTIONS(1470),
    [anon_sym_DOT] = ACTIONS(1470),
    [anon_sym_SLASH] = ACTIONS(1470),
    [anon_sym_COLON] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1470),
    [anon_sym_EQ] = ACTIONS(1470),
    [anon_sym_QMARK] = ACTIONS(1470),
    [anon_sym_AT] = ACTIONS(1470),
    [anon_sym_BSLASH] = ACTIONS(1470),
    [anon_sym_CARET] = ACTIONS(1470),
    [anon_sym__] = ACTIONS(1470),
    [anon_sym_BQUOTE] = ACTIONS(1470),
    [anon_sym_LBRACE] = ACTIONS(1470),
    [anon_sym_PIPE] = ACTIONS(1470),
    [anon_sym_RBRACE] = ACTIONS(1470),
    [anon_sym_TILDE] = ACTIONS(1470),
    [anon_sym_LPAREN] = ACTIONS(1470),
    [anon_sym_RPAREN] = ACTIONS(1470),
    [aux_sym__word_token1] = ACTIONS(1470),
    [aux_sym__word_token2] = ACTIONS(1470),
    [aux_sym__word_token3] = ACTIONS(1470),
    [sym__whitespace] = ACTIONS(1470),
    [sym__soft_line_ending] = ACTIONS(1470),
    [sym__block_quote_start] = ACTIONS(1470),
    [sym__indented_chunk_start] = ACTIONS(1470),
    [sym_atx_h1_marker] = ACTIONS(1470),
    [sym_atx_h2_marker] = ACTIONS(1470),
    [sym_atx_h3_marker] = ACTIONS(1470),
    [sym_atx_h4_marker] = ACTIONS(1470),
    [sym_atx_h5_marker] = ACTIONS(1470),
    [sym_atx_h6_marker] = ACTIONS(1470),
    [sym__thematic_break] = ACTIONS(1470),
    [sym__list_marker_minus] = ACTIONS(1470),
    [sym__list_marker_plus] = ACTIONS(1470),
    [sym__list_marker_star] = ACTIONS(1470),
    [sym__list_marker_parenthesis] = ACTIONS(1470),
    [sym__list_marker_dot] = ACTIONS(1470),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1470),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1470),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1470),
    [sym__blank_line_start] = ACTIONS(1470),
    [sym__html_block_1_start] = ACTIONS(1470),
    [sym__html_block_2_start] = ACTIONS(1470),
    [sym__html_block_3_start] = ACTIONS(1470),
    [sym__html_block_4_start] = ACTIONS(1470),
    [sym__html_block_5_start] = ACTIONS(1470),
    [sym__html_block_6_start] = ACTIONS(1470),
    [sym__html_block_7_start] = ACTIONS(1470),
    [sym__pipe_table_start] = ACTIONS(1470),
  },
  [STATE(330)] = {
    [ts_builtin_sym_end] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1326),
    [anon_sym_RBRACK] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_POUND] = ACTIONS(1324),
    [anon_sym_DOLLAR] = ACTIONS(1324),
    [anon_sym_PERCENT] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_COLON] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_EQ] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_BSLASH] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1324),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_BQUOTE] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_RPAREN] = ACTIONS(1324),
    [aux_sym__word_token1] = ACTIONS(1324),
    [aux_sym__word_token2] = ACTIONS(1324),
    [aux_sym__word_token3] = ACTIONS(1324),
    [sym__whitespace] = ACTIONS(1324),
    [sym__soft_line_ending] = ACTIONS(1324),
    [sym__block_quote_start] = ACTIONS(1324),
    [sym__indented_chunk_start] = ACTIONS(1324),
    [sym_atx_h1_marker] = ACTIONS(1324),
    [sym_atx_h2_marker] = ACTIONS(1324),
    [sym_atx_h3_marker] = ACTIONS(1324),
    [sym_atx_h4_marker] = ACTIONS(1324),
    [sym_atx_h5_marker] = ACTIONS(1324),
    [sym_atx_h6_marker] = ACTIONS(1324),
    [sym__thematic_break] = ACTIONS(1324),
    [sym__list_marker_minus] = ACTIONS(1324),
    [sym__list_marker_plus] = ACTIONS(1324),
    [sym__list_marker_star] = ACTIONS(1324),
    [sym__list_marker_parenthesis] = ACTIONS(1324),
    [sym__list_marker_dot] = ACTIONS(1324),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1324),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1324),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1324),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1324),
    [sym__blank_line_start] = ACTIONS(1324),
    [sym__html_block_1_start] = ACTIONS(1324),
    [sym__html_block_2_start] = ACTIONS(1324),
    [sym__html_block_3_start] = ACTIONS(1324),
    [sym__html_block_4_start] = ACTIONS(1324),
    [sym__html_block_5_start] = ACTIONS(1324),
    [sym__html_block_6_start] = ACTIONS(1324),
    [sym__html_block_7_start] = ACTIONS(1324),
    [sym__pipe_table_start] = ACTIONS(1324),
  },
  [STATE(331)] = {
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_RBRACK] = ACTIONS(1470),
    [anon_sym_LT] = ACTIONS(1470),
    [anon_sym_GT] = ACTIONS(1470),
    [anon_sym_BANG] = ACTIONS(1470),
    [anon_sym_DQUOTE] = ACTIONS(1470),
    [anon_sym_POUND] = ACTIONS(1470),
    [anon_sym_DOLLAR] = ACTIONS(1470),
    [anon_sym_PERCENT] = ACTIONS(1470),
    [anon_sym_AMP] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_STAR] = ACTIONS(1470),
    [anon_sym_PLUS] = ACTIONS(1470),
    [anon_sym_COMMA] = ACTIONS(1470),
    [anon_sym_DASH] = ACTIONS(1470),
    [anon_sym_DOT] = ACTIONS(1470),
    [anon_sym_SLASH] = ACTIONS(1470),
    [anon_sym_COLON] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1470),
    [anon_sym_EQ] = ACTIONS(1470),
    [anon_sym_QMARK] = ACTIONS(1470),
    [anon_sym_AT] = ACTIONS(1470),
    [anon_sym_BSLASH] = ACTIONS(1470),
    [anon_sym_CARET] = ACTIONS(1470),
    [anon_sym__] = ACTIONS(1470),
    [anon_sym_BQUOTE] = ACTIONS(1470),
    [anon_sym_LBRACE] = ACTIONS(1470),
    [anon_sym_PIPE] = ACTIONS(1470),
    [anon_sym_RBRACE] = ACTIONS(1470),
    [anon_sym_TILDE] = ACTIONS(1470),
    [anon_sym_LPAREN] = ACTIONS(1470),
    [anon_sym_RPAREN] = ACTIONS(1470),
    [aux_sym__word_token1] = ACTIONS(1470),
    [aux_sym__word_token2] = ACTIONS(1470),
    [aux_sym__word_token3] = ACTIONS(1470),
    [sym__whitespace] = ACTIONS(1470),
    [sym__soft_line_ending] = ACTIONS(1470),
    [sym__block_close] = ACTIONS(1470),
    [sym__block_quote_start] = ACTIONS(1470),
    [sym__indented_chunk_start] = ACTIONS(1470),
    [sym_atx_h1_marker] = ACTIONS(1470),
    [sym_atx_h2_marker] = ACTIONS(1470),
    [sym_atx_h3_marker] = ACTIONS(1470),
    [sym_atx_h4_marker] = ACTIONS(1470),
    [sym_atx_h5_marker] = ACTIONS(1470),
    [sym_atx_h6_marker] = ACTIONS(1470),
    [sym__thematic_break] = ACTIONS(1470),
    [sym__list_marker_minus] = ACTIONS(1470),
    [sym__list_marker_plus] = ACTIONS(1470),
    [sym__list_marker_star] = ACTIONS(1470),
    [sym__list_marker_parenthesis] = ACTIONS(1470),
    [sym__list_marker_dot] = ACTIONS(1470),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1470),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1470),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1470),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1470),
    [sym__blank_line_start] = ACTIONS(1470),
    [sym__html_block_1_start] = ACTIONS(1470),
    [sym__html_block_2_start] = ACTIONS(1470),
    [sym__html_block_3_start] = ACTIONS(1470),
    [sym__html_block_4_start] = ACTIONS(1470),
    [sym__html_block_5_start] = ACTIONS(1470),
    [sym__html_block_6_start] = ACTIONS(1470),
    [sym__html_block_7_start] = ACTIONS(1470),
    [sym__pipe_table_start] = ACTIONS(1470),
  },
  [STATE(332)] = {
    [ts_builtin_sym_end] = ACTIONS(1740),
    [anon_sym_LBRACK] = ACTIONS(1738),
    [anon_sym_RBRACK] = ACTIONS(1740),
    [anon_sym_LT] = ACTIONS(1740),
    [anon_sym_GT] = ACTIONS(1740),
    [anon_sym_BANG] = ACTIONS(1740),
    [anon_sym_DQUOTE] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_DOLLAR] = ACTIONS(1740),
    [anon_sym_PERCENT] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1740),
    [anon_sym_STAR] = ACTIONS(1740),
    [anon_sym_PLUS] = ACTIONS(1740),
    [anon_sym_COMMA] = ACTIONS(1740),
    [anon_sym_DASH] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1740),
    [anon_sym_SLASH] = ACTIONS(1740),
    [anon_sym_COLON] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1740),
    [anon_sym_EQ] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_AT] = ACTIONS(1740),
    [anon_sym_BSLASH] = ACTIONS(1740),
    [anon_sym_CARET] = ACTIONS(1740),
    [anon_sym__] = ACTIONS(1740),
    [anon_sym_BQUOTE] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_RBRACE] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1740),
    [anon_sym_RPAREN] = ACTIONS(1740),
    [aux_sym__word_token1] = ACTIONS(1740),
    [aux_sym__word_token2] = ACTIONS(1740),
    [aux_sym__word_token3] = ACTIONS(1740),
    [sym__whitespace] = ACTIONS(1740),
    [sym__soft_line_ending] = ACTIONS(1740),
    [sym__block_quote_start] = ACTIONS(1740),
    [sym__indented_chunk_start] = ACTIONS(1740),
    [sym_atx_h1_marker] = ACTIONS(1740),
    [sym_atx_h2_marker] = ACTIONS(1740),
    [sym_atx_h3_marker] = ACTIONS(1740),
    [sym_atx_h4_marker] = ACTIONS(1740),
    [sym_atx_h5_marker] = ACTIONS(1740),
    [sym_atx_h6_marker] = ACTIONS(1740),
    [sym__thematic_break] = ACTIONS(1740),
    [sym__list_marker_minus] = ACTIONS(1740),
    [sym__list_marker_plus] = ACTIONS(1740),
    [sym__list_marker_star] = ACTIONS(1740),
    [sym__list_marker_parenthesis] = ACTIONS(1740),
    [sym__list_marker_dot] = ACTIONS(1740),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1740),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1740),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1740),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1740),
    [sym__blank_line_start] = ACTIONS(1740),
    [sym__html_block_1_start] = ACTIONS(1740),
    [sym__html_block_2_start] = ACTIONS(1740),
    [sym__html_block_3_start] = ACTIONS(1740),
    [sym__html_block_4_start] = ACTIONS(1740),
    [sym__html_block_5_start] = ACTIONS(1740),
    [sym__html_block_6_start] = ACTIONS(1740),
    [sym__html_block_7_start] = ACTIONS(1740),
    [sym__pipe_table_start] = ACTIONS(1740),
  },
  [STATE(333)] = {
    [ts_builtin_sym_end] = ACTIONS(1744),
    [anon_sym_LBRACK] = ACTIONS(1742),
    [anon_sym_RBRACK] = ACTIONS(1744),
    [anon_sym_LT] = ACTIONS(1744),
    [anon_sym_GT] = ACTIONS(1744),
    [anon_sym_BANG] = ACTIONS(1744),
    [anon_sym_DQUOTE] = ACTIONS(1744),
    [anon_sym_POUND] = ACTIONS(1744),
    [anon_sym_DOLLAR] = ACTIONS(1744),
    [anon_sym_PERCENT] = ACTIONS(1744),
    [anon_sym_AMP] = ACTIONS(1744),
    [anon_sym_SQUOTE] = ACTIONS(1744),
    [anon_sym_STAR] = ACTIONS(1744),
    [anon_sym_PLUS] = ACTIONS(1744),
    [anon_sym_COMMA] = ACTIONS(1744),
    [anon_sym_DASH] = ACTIONS(1744),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_SLASH] = ACTIONS(1744),
    [anon_sym_COLON] = ACTIONS(1744),
    [anon_sym_SEMI] = ACTIONS(1744),
    [anon_sym_EQ] = ACTIONS(1744),
    [anon_sym_QMARK] = ACTIONS(1744),
    [anon_sym_AT] = ACTIONS(1744),
    [anon_sym_BSLASH] = ACTIONS(1744),
    [anon_sym_CARET] = ACTIONS(1744),
    [anon_sym__] = ACTIONS(1744),
    [anon_sym_BQUOTE] = ACTIONS(1744),
    [anon_sym_LBRACE] = ACTIONS(1744),
    [anon_sym_PIPE] = ACTIONS(1744),
    [anon_sym_RBRACE] = ACTIONS(1744),
    [anon_sym_TILDE] = ACTIONS(1744),
    [anon_sym_LPAREN] = ACTIONS(1744),
    [anon_sym_RPAREN] = ACTIONS(1744),
    [aux_sym__word_token1] = ACTIONS(1744),
    [aux_sym__word_token2] = ACTIONS(1744),
    [aux_sym__word_token3] = ACTIONS(1744),
    [sym__whitespace] = ACTIONS(1744),
    [sym__soft_line_ending] = ACTIONS(1744),
    [sym__block_quote_start] = ACTIONS(1744),
    [sym__indented_chunk_start] = ACTIONS(1744),
    [sym_atx_h1_marker] = ACTIONS(1744),
    [sym_atx_h2_marker] = ACTIONS(1744),
    [sym_atx_h3_marker] = ACTIONS(1744),
    [sym_atx_h4_marker] = ACTIONS(1744),
    [sym_atx_h5_marker] = ACTIONS(1744),
    [sym_atx_h6_marker] = ACTIONS(1744),
    [sym__thematic_break] = ACTIONS(1744),
    [sym__list_marker_minus] = ACTIONS(1744),
    [sym__list_marker_plus] = ACTIONS(1744),
    [sym__list_marker_star] = ACTIONS(1744),
    [sym__list_marker_parenthesis] = ACTIONS(1744),
    [sym__list_marker_dot] = ACTIONS(1744),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1744),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1744),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1744),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1744),
    [sym__blank_line_start] = ACTIONS(1744),
    [sym__html_block_1_start] = ACTIONS(1744),
    [sym__html_block_2_start] = ACTIONS(1744),
    [sym__html_block_3_start] = ACTIONS(1744),
    [sym__html_block_4_start] = ACTIONS(1744),
    [sym__html_block_5_start] = ACTIONS(1744),
    [sym__html_block_6_start] = ACTIONS(1744),
    [sym__html_block_7_start] = ACTIONS(1744),
    [sym__pipe_table_start] = ACTIONS(1744),
  },
  [STATE(334)] = {
    [ts_builtin_sym_end] = ACTIONS(1748),
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_LT] = ACTIONS(1748),
    [anon_sym_GT] = ACTIONS(1748),
    [anon_sym_BANG] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [anon_sym_POUND] = ACTIONS(1748),
    [anon_sym_DOLLAR] = ACTIONS(1748),
    [anon_sym_PERCENT] = ACTIONS(1748),
    [anon_sym_AMP] = ACTIONS(1748),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_STAR] = ACTIONS(1748),
    [anon_sym_PLUS] = ACTIONS(1748),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_DASH] = ACTIONS(1748),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_SLASH] = ACTIONS(1748),
    [anon_sym_COLON] = ACTIONS(1748),
    [anon_sym_SEMI] = ACTIONS(1748),
    [anon_sym_EQ] = ACTIONS(1748),
    [anon_sym_QMARK] = ACTIONS(1748),
    [anon_sym_AT] = ACTIONS(1748),
    [anon_sym_BSLASH] = ACTIONS(1748),
    [anon_sym_CARET] = ACTIONS(1748),
    [anon_sym__] = ACTIONS(1748),
    [anon_sym_BQUOTE] = ACTIONS(1748),
    [anon_sym_LBRACE] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1748),
    [anon_sym_RBRACE] = ACTIONS(1748),
    [anon_sym_TILDE] = ACTIONS(1748),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_RPAREN] = ACTIONS(1748),
    [aux_sym__word_token1] = ACTIONS(1748),
    [aux_sym__word_token2] = ACTIONS(1748),
    [aux_sym__word_token3] = ACTIONS(1748),
    [sym__whitespace] = ACTIONS(1748),
    [sym__soft_line_ending] = ACTIONS(1748),
    [sym__block_quote_start] = ACTIONS(1748),
    [sym__indented_chunk_start] = ACTIONS(1748),
    [sym_atx_h1_marker] = ACTIONS(1748),
    [sym_atx_h2_marker] = ACTIONS(1748),
    [sym_atx_h3_marker] = ACTIONS(1748),
    [sym_atx_h4_marker] = ACTIONS(1748),
    [sym_atx_h5_marker] = ACTIONS(1748),
    [sym_atx_h6_marker] = ACTIONS(1748),
    [sym__thematic_break] = ACTIONS(1748),
    [sym__list_marker_minus] = ACTIONS(1748),
    [sym__list_marker_plus] = ACTIONS(1748),
    [sym__list_marker_star] = ACTIONS(1748),
    [sym__list_marker_parenthesis] = ACTIONS(1748),
    [sym__list_marker_dot] = ACTIONS(1748),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1748),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1748),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1748),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1748),
    [sym__blank_line_start] = ACTIONS(1748),
    [sym__html_block_1_start] = ACTIONS(1748),
    [sym__html_block_2_start] = ACTIONS(1748),
    [sym__html_block_3_start] = ACTIONS(1748),
    [sym__html_block_4_start] = ACTIONS(1748),
    [sym__html_block_5_start] = ACTIONS(1748),
    [sym__html_block_6_start] = ACTIONS(1748),
    [sym__html_block_7_start] = ACTIONS(1748),
    [sym__pipe_table_start] = ACTIONS(1748),
  },
  [STATE(335)] = {
    [anon_sym_LBRACK] = ACTIONS(1255),
    [anon_sym_RBRACK] = ACTIONS(1253),
    [anon_sym_LT] = ACTIONS(1253),
    [anon_sym_GT] = ACTIONS(1253),
    [anon_sym_BANG] = ACTIONS(1253),
    [anon_sym_DQUOTE] = ACTIONS(1253),
    [anon_sym_POUND] = ACTIONS(1253),
    [anon_sym_DOLLAR] = ACTIONS(1253),
    [anon_sym_PERCENT] = ACTIONS(1253),
    [anon_sym_AMP] = ACTIONS(1253),
    [anon_sym_SQUOTE] = ACTIONS(1253),
    [anon_sym_STAR] = ACTIONS(1253),
    [anon_sym_PLUS] = ACTIONS(1253),
    [anon_sym_COMMA] = ACTIONS(1253),
    [anon_sym_DASH] = ACTIONS(1253),
    [anon_sym_DOT] = ACTIONS(1253),
    [anon_sym_SLASH] = ACTIONS(1253),
    [anon_sym_COLON] = ACTIONS(1253),
    [anon_sym_SEMI] = ACTIONS(1253),
    [anon_sym_EQ] = ACTIONS(1253),
    [anon_sym_QMARK] = ACTIONS(1253),
    [anon_sym_AT] = ACTIONS(1253),
    [anon_sym_BSLASH] = ACTIONS(1253),
    [anon_sym_CARET] = ACTIONS(1253),
    [anon_sym__] = ACTIONS(1253),
    [anon_sym_BQUOTE] = ACTIONS(1253),
    [anon_sym_LBRACE] = ACTIONS(1253),
    [anon_sym_PIPE] = ACTIONS(1253),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_TILDE] = ACTIONS(1253),
    [anon_sym_LPAREN] = ACTIONS(1253),
    [anon_sym_RPAREN] = ACTIONS(1253),
    [aux_sym__word_token1] = ACTIONS(1253),
    [aux_sym__word_token2] = ACTIONS(1253),
    [aux_sym__word_token3] = ACTIONS(1253),
    [sym__whitespace] = ACTIONS(1253),
    [sym__soft_line_ending] = ACTIONS(1253),
    [sym__block_close] = ACTIONS(1253),
    [sym__block_quote_start] = ACTIONS(1253),
    [sym__indented_chunk_start] = ACTIONS(1253),
    [sym_atx_h1_marker] = ACTIONS(1253),
    [sym_atx_h2_marker] = ACTIONS(1253),
    [sym_atx_h3_marker] = ACTIONS(1253),
    [sym_atx_h4_marker] = ACTIONS(1253),
    [sym_atx_h5_marker] = ACTIONS(1253),
    [sym_atx_h6_marker] = ACTIONS(1253),
    [sym__thematic_break] = ACTIONS(1253),
    [sym__list_marker_minus] = ACTIONS(1253),
    [sym__list_marker_plus] = ACTIONS(1253),
    [sym__list_marker_star] = ACTIONS(1253),
    [sym__list_marker_parenthesis] = ACTIONS(1253),
    [sym__list_marker_dot] = ACTIONS(1253),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1253),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1253),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1253),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1253),
    [sym__blank_line_start] = ACTIONS(1253),
    [sym__html_block_1_start] = ACTIONS(1253),
    [sym__html_block_2_start] = ACTIONS(1253),
    [sym__html_block_3_start] = ACTIONS(1253),
    [sym__html_block_4_start] = ACTIONS(1253),
    [sym__html_block_5_start] = ACTIONS(1253),
    [sym__html_block_6_start] = ACTIONS(1253),
    [sym__html_block_7_start] = ACTIONS(1253),
    [sym__pipe_table_start] = ACTIONS(1253),
  },
  [STATE(336)] = {
    [ts_builtin_sym_end] = ACTIONS(1170),
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_RBRACK] = ACTIONS(1170),
    [anon_sym_LT] = ACTIONS(1170),
    [anon_sym_GT] = ACTIONS(1170),
    [anon_sym_BANG] = ACTIONS(1170),
    [anon_sym_DQUOTE] = ACTIONS(1170),
    [anon_sym_POUND] = ACTIONS(1170),
    [anon_sym_DOLLAR] = ACTIONS(1170),
    [anon_sym_PERCENT] = ACTIONS(1170),
    [anon_sym_AMP] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_COMMA] = ACTIONS(1170),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_DOT] = ACTIONS(1170),
    [anon_sym_SLASH] = ACTIONS(1170),
    [anon_sym_COLON] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1170),
    [anon_sym_EQ] = ACTIONS(1170),
    [anon_sym_QMARK] = ACTIONS(1170),
    [anon_sym_AT] = ACTIONS(1170),
    [anon_sym_BSLASH] = ACTIONS(1170),
    [anon_sym_CARET] = ACTIONS(1170),
    [anon_sym__] = ACTIONS(1170),
    [anon_sym_BQUOTE] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1170),
    [anon_sym_PIPE] = ACTIONS(1170),
    [anon_sym_RBRACE] = ACTIONS(1170),
    [anon_sym_TILDE] = ACTIONS(1170),
    [anon_sym_LPAREN] = ACTIONS(1170),
    [anon_sym_RPAREN] = ACTIONS(1170),
    [aux_sym__word_token1] = ACTIONS(1170),
    [aux_sym__word_token2] = ACTIONS(1170),
    [aux_sym__word_token3] = ACTIONS(1170),
    [sym__whitespace] = ACTIONS(1170),
    [sym__soft_line_ending] = ACTIONS(1170),
    [sym__block_quote_start] = ACTIONS(1170),
    [sym__indented_chunk_start] = ACTIONS(1170),
    [sym_atx_h1_marker] = ACTIONS(1170),
    [sym_atx_h2_marker] = ACTIONS(1170),
    [sym_atx_h3_marker] = ACTIONS(1170),
    [sym_atx_h4_marker] = ACTIONS(1170),
    [sym_atx_h5_marker] = ACTIONS(1170),
    [sym_atx_h6_marker] = ACTIONS(1170),
    [sym__thematic_break] = ACTIONS(1170),
    [sym__list_marker_minus] = ACTIONS(1170),
    [sym__list_marker_plus] = ACTIONS(1170),
    [sym__list_marker_star] = ACTIONS(1170),
    [sym__list_marker_parenthesis] = ACTIONS(1170),
    [sym__list_marker_dot] = ACTIONS(1170),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1170),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1170),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1170),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1170),
    [sym__blank_line_start] = ACTIONS(1170),
    [sym__html_block_1_start] = ACTIONS(1170),
    [sym__html_block_2_start] = ACTIONS(1170),
    [sym__html_block_3_start] = ACTIONS(1170),
    [sym__html_block_4_start] = ACTIONS(1170),
    [sym__html_block_5_start] = ACTIONS(1170),
    [sym__html_block_6_start] = ACTIONS(1170),
    [sym__html_block_7_start] = ACTIONS(1170),
    [sym__pipe_table_start] = ACTIONS(1170),
  },
  [STATE(337)] = {
    [ts_builtin_sym_end] = ACTIONS(1752),
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_LT] = ACTIONS(1752),
    [anon_sym_GT] = ACTIONS(1752),
    [anon_sym_BANG] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [anon_sym_POUND] = ACTIONS(1752),
    [anon_sym_DOLLAR] = ACTIONS(1752),
    [anon_sym_PERCENT] = ACTIONS(1752),
    [anon_sym_AMP] = ACTIONS(1752),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_STAR] = ACTIONS(1752),
    [anon_sym_PLUS] = ACTIONS(1752),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_DASH] = ACTIONS(1752),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_SLASH] = ACTIONS(1752),
    [anon_sym_COLON] = ACTIONS(1752),
    [anon_sym_SEMI] = ACTIONS(1752),
    [anon_sym_EQ] = ACTIONS(1752),
    [anon_sym_QMARK] = ACTIONS(1752),
    [anon_sym_AT] = ACTIONS(1752),
    [anon_sym_BSLASH] = ACTIONS(1752),
    [anon_sym_CARET] = ACTIONS(1752),
    [anon_sym__] = ACTIONS(1752),
    [anon_sym_BQUOTE] = ACTIONS(1752),
    [anon_sym_LBRACE] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1752),
    [anon_sym_RBRACE] = ACTIONS(1752),
    [anon_sym_TILDE] = ACTIONS(1752),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_RPAREN] = ACTIONS(1752),
    [aux_sym__word_token1] = ACTIONS(1752),
    [aux_sym__word_token2] = ACTIONS(1752),
    [aux_sym__word_token3] = ACTIONS(1752),
    [sym__whitespace] = ACTIONS(1752),
    [sym__soft_line_ending] = ACTIONS(1752),
    [sym__block_quote_start] = ACTIONS(1752),
    [sym__indented_chunk_start] = ACTIONS(1752),
    [sym_atx_h1_marker] = ACTIONS(1752),
    [sym_atx_h2_marker] = ACTIONS(1752),
    [sym_atx_h3_marker] = ACTIONS(1752),
    [sym_atx_h4_marker] = ACTIONS(1752),
    [sym_atx_h5_marker] = ACTIONS(1752),
    [sym_atx_h6_marker] = ACTIONS(1752),
    [sym__thematic_break] = ACTIONS(1752),
    [sym__list_marker_minus] = ACTIONS(1752),
    [sym__list_marker_plus] = ACTIONS(1752),
    [sym__list_marker_star] = ACTIONS(1752),
    [sym__list_marker_parenthesis] = ACTIONS(1752),
    [sym__list_marker_dot] = ACTIONS(1752),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1752),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1752),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1752),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1752),
    [sym__blank_line_start] = ACTIONS(1752),
    [sym__html_block_1_start] = ACTIONS(1752),
    [sym__html_block_2_start] = ACTIONS(1752),
    [sym__html_block_3_start] = ACTIONS(1752),
    [sym__html_block_4_start] = ACTIONS(1752),
    [sym__html_block_5_start] = ACTIONS(1752),
    [sym__html_block_6_start] = ACTIONS(1752),
    [sym__html_block_7_start] = ACTIONS(1752),
    [sym__pipe_table_start] = ACTIONS(1752),
  },
  [STATE(338)] = {
    [ts_builtin_sym_end] = ACTIONS(1756),
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_LT] = ACTIONS(1756),
    [anon_sym_GT] = ACTIONS(1756),
    [anon_sym_BANG] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [anon_sym_POUND] = ACTIONS(1756),
    [anon_sym_DOLLAR] = ACTIONS(1756),
    [anon_sym_PERCENT] = ACTIONS(1756),
    [anon_sym_AMP] = ACTIONS(1756),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_STAR] = ACTIONS(1756),
    [anon_sym_PLUS] = ACTIONS(1756),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_DASH] = ACTIONS(1756),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_SLASH] = ACTIONS(1756),
    [anon_sym_COLON] = ACTIONS(1756),
    [anon_sym_SEMI] = ACTIONS(1756),
    [anon_sym_EQ] = ACTIONS(1756),
    [anon_sym_QMARK] = ACTIONS(1756),
    [anon_sym_AT] = ACTIONS(1756),
    [anon_sym_BSLASH] = ACTIONS(1756),
    [anon_sym_CARET] = ACTIONS(1756),
    [anon_sym__] = ACTIONS(1756),
    [anon_sym_BQUOTE] = ACTIONS(1756),
    [anon_sym_LBRACE] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1756),
    [anon_sym_RBRACE] = ACTIONS(1756),
    [anon_sym_TILDE] = ACTIONS(1756),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_RPAREN] = ACTIONS(1756),
    [aux_sym__word_token1] = ACTIONS(1756),
    [aux_sym__word_token2] = ACTIONS(1756),
    [aux_sym__word_token3] = ACTIONS(1756),
    [sym__whitespace] = ACTIONS(1756),
    [sym__soft_line_ending] = ACTIONS(1756),
    [sym__block_quote_start] = ACTIONS(1756),
    [sym__indented_chunk_start] = ACTIONS(1756),
    [sym_atx_h1_marker] = ACTIONS(1756),
    [sym_atx_h2_marker] = ACTIONS(1756),
    [sym_atx_h3_marker] = ACTIONS(1756),
    [sym_atx_h4_marker] = ACTIONS(1756),
    [sym_atx_h5_marker] = ACTIONS(1756),
    [sym_atx_h6_marker] = ACTIONS(1756),
    [sym__thematic_break] = ACTIONS(1756),
    [sym__list_marker_minus] = ACTIONS(1756),
    [sym__list_marker_plus] = ACTIONS(1756),
    [sym__list_marker_star] = ACTIONS(1756),
    [sym__list_marker_parenthesis] = ACTIONS(1756),
    [sym__list_marker_dot] = ACTIONS(1756),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1756),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1756),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1756),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1756),
    [sym__blank_line_start] = ACTIONS(1756),
    [sym__html_block_1_start] = ACTIONS(1756),
    [sym__html_block_2_start] = ACTIONS(1756),
    [sym__html_block_3_start] = ACTIONS(1756),
    [sym__html_block_4_start] = ACTIONS(1756),
    [sym__html_block_5_start] = ACTIONS(1756),
    [sym__html_block_6_start] = ACTIONS(1756),
    [sym__html_block_7_start] = ACTIONS(1756),
    [sym__pipe_table_start] = ACTIONS(1756),
  },
  [STATE(339)] = {
    [ts_builtin_sym_end] = ACTIONS(1572),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_LT] = ACTIONS(1572),
    [anon_sym_GT] = ACTIONS(1572),
    [anon_sym_BANG] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [anon_sym_POUND] = ACTIONS(1572),
    [anon_sym_DOLLAR] = ACTIONS(1572),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_AMP] = ACTIONS(1572),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_STAR] = ACTIONS(1572),
    [anon_sym_PLUS] = ACTIONS(1572),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_DASH] = ACTIONS(1572),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1572),
    [anon_sym_COLON] = ACTIONS(1572),
    [anon_sym_SEMI] = ACTIONS(1572),
    [anon_sym_EQ] = ACTIONS(1572),
    [anon_sym_QMARK] = ACTIONS(1572),
    [anon_sym_AT] = ACTIONS(1572),
    [anon_sym_BSLASH] = ACTIONS(1572),
    [anon_sym_CARET] = ACTIONS(1572),
    [anon_sym__] = ACTIONS(1572),
    [anon_sym_BQUOTE] = ACTIONS(1572),
    [anon_sym_LBRACE] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1572),
    [anon_sym_RBRACE] = ACTIONS(1572),
    [anon_sym_TILDE] = ACTIONS(1572),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_RPAREN] = ACTIONS(1572),
    [aux_sym__word_token1] = ACTIONS(1572),
    [aux_sym__word_token2] = ACTIONS(1572),
    [aux_sym__word_token3] = ACTIONS(1572),
    [sym__whitespace] = ACTIONS(1572),
    [sym__soft_line_ending] = ACTIONS(1572),
    [sym__block_quote_start] = ACTIONS(1572),
    [sym__indented_chunk_start] = ACTIONS(1572),
    [sym_atx_h1_marker] = ACTIONS(1572),
    [sym_atx_h2_marker] = ACTIONS(1572),
    [sym_atx_h3_marker] = ACTIONS(1572),
    [sym_atx_h4_marker] = ACTIONS(1572),
    [sym_atx_h5_marker] = ACTIONS(1572),
    [sym_atx_h6_marker] = ACTIONS(1572),
    [sym__thematic_break] = ACTIONS(1572),
    [sym__list_marker_minus] = ACTIONS(1572),
    [sym__list_marker_plus] = ACTIONS(1572),
    [sym__list_marker_star] = ACTIONS(1572),
    [sym__list_marker_parenthesis] = ACTIONS(1572),
    [sym__list_marker_dot] = ACTIONS(1572),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1572),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1572),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1572),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1572),
    [sym__blank_line_start] = ACTIONS(1572),
    [sym__html_block_1_start] = ACTIONS(1572),
    [sym__html_block_2_start] = ACTIONS(1572),
    [sym__html_block_3_start] = ACTIONS(1572),
    [sym__html_block_4_start] = ACTIONS(1572),
    [sym__html_block_5_start] = ACTIONS(1572),
    [sym__html_block_6_start] = ACTIONS(1572),
    [sym__html_block_7_start] = ACTIONS(1572),
    [sym__pipe_table_start] = ACTIONS(1572),
  },
  [STATE(340)] = {
    [ts_builtin_sym_end] = ACTIONS(1760),
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_LT] = ACTIONS(1760),
    [anon_sym_GT] = ACTIONS(1760),
    [anon_sym_BANG] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [anon_sym_POUND] = ACTIONS(1760),
    [anon_sym_DOLLAR] = ACTIONS(1760),
    [anon_sym_PERCENT] = ACTIONS(1760),
    [anon_sym_AMP] = ACTIONS(1760),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_STAR] = ACTIONS(1760),
    [anon_sym_PLUS] = ACTIONS(1760),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_DASH] = ACTIONS(1760),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_SLASH] = ACTIONS(1760),
    [anon_sym_COLON] = ACTIONS(1760),
    [anon_sym_SEMI] = ACTIONS(1760),
    [anon_sym_EQ] = ACTIONS(1760),
    [anon_sym_QMARK] = ACTIONS(1760),
    [anon_sym_AT] = ACTIONS(1760),
    [anon_sym_BSLASH] = ACTIONS(1760),
    [anon_sym_CARET] = ACTIONS(1760),
    [anon_sym__] = ACTIONS(1760),
    [anon_sym_BQUOTE] = ACTIONS(1760),
    [anon_sym_LBRACE] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1760),
    [anon_sym_RBRACE] = ACTIONS(1760),
    [anon_sym_TILDE] = ACTIONS(1760),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_RPAREN] = ACTIONS(1760),
    [aux_sym__word_token1] = ACTIONS(1760),
    [aux_sym__word_token2] = ACTIONS(1760),
    [aux_sym__word_token3] = ACTIONS(1760),
    [sym__whitespace] = ACTIONS(1760),
    [sym__soft_line_ending] = ACTIONS(1760),
    [sym__block_quote_start] = ACTIONS(1760),
    [sym__indented_chunk_start] = ACTIONS(1760),
    [sym_atx_h1_marker] = ACTIONS(1760),
    [sym_atx_h2_marker] = ACTIONS(1760),
    [sym_atx_h3_marker] = ACTIONS(1760),
    [sym_atx_h4_marker] = ACTIONS(1760),
    [sym_atx_h5_marker] = ACTIONS(1760),
    [sym_atx_h6_marker] = ACTIONS(1760),
    [sym__thematic_break] = ACTIONS(1760),
    [sym__list_marker_minus] = ACTIONS(1760),
    [sym__list_marker_plus] = ACTIONS(1760),
    [sym__list_marker_star] = ACTIONS(1760),
    [sym__list_marker_parenthesis] = ACTIONS(1760),
    [sym__list_marker_dot] = ACTIONS(1760),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1760),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1760),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1760),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1760),
    [sym__blank_line_start] = ACTIONS(1760),
    [sym__html_block_1_start] = ACTIONS(1760),
    [sym__html_block_2_start] = ACTIONS(1760),
    [sym__html_block_3_start] = ACTIONS(1760),
    [sym__html_block_4_start] = ACTIONS(1760),
    [sym__html_block_5_start] = ACTIONS(1760),
    [sym__html_block_6_start] = ACTIONS(1760),
    [sym__html_block_7_start] = ACTIONS(1760),
    [sym__pipe_table_start] = ACTIONS(1760),
  },
  [STATE(341)] = {
    [ts_builtin_sym_end] = ACTIONS(1764),
    [anon_sym_LBRACK] = ACTIONS(1762),
    [anon_sym_RBRACK] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1764),
    [anon_sym_BANG] = ACTIONS(1764),
    [anon_sym_DQUOTE] = ACTIONS(1764),
    [anon_sym_POUND] = ACTIONS(1764),
    [anon_sym_DOLLAR] = ACTIONS(1764),
    [anon_sym_PERCENT] = ACTIONS(1764),
    [anon_sym_AMP] = ACTIONS(1764),
    [anon_sym_SQUOTE] = ACTIONS(1764),
    [anon_sym_STAR] = ACTIONS(1764),
    [anon_sym_PLUS] = ACTIONS(1764),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_DASH] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_SLASH] = ACTIONS(1764),
    [anon_sym_COLON] = ACTIONS(1764),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym_EQ] = ACTIONS(1764),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_AT] = ACTIONS(1764),
    [anon_sym_BSLASH] = ACTIONS(1764),
    [anon_sym_CARET] = ACTIONS(1764),
    [anon_sym__] = ACTIONS(1764),
    [anon_sym_BQUOTE] = ACTIONS(1764),
    [anon_sym_LBRACE] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1764),
    [anon_sym_RBRACE] = ACTIONS(1764),
    [anon_sym_TILDE] = ACTIONS(1764),
    [anon_sym_LPAREN] = ACTIONS(1764),
    [anon_sym_RPAREN] = ACTIONS(1764),
    [aux_sym__word_token1] = ACTIONS(1764),
    [aux_sym__word_token2] = ACTIONS(1764),
    [aux_sym__word_token3] = ACTIONS(1764),
    [sym__whitespace] = ACTIONS(1764),
    [sym__soft_line_ending] = ACTIONS(1764),
    [sym__block_quote_start] = ACTIONS(1764),
    [sym__indented_chunk_start] = ACTIONS(1764),
    [sym_atx_h1_marker] = ACTIONS(1764),
    [sym_atx_h2_marker] = ACTIONS(1764),
    [sym_atx_h3_marker] = ACTIONS(1764),
    [sym_atx_h4_marker] = ACTIONS(1764),
    [sym_atx_h5_marker] = ACTIONS(1764),
    [sym_atx_h6_marker] = ACTIONS(1764),
    [sym__thematic_break] = ACTIONS(1764),
    [sym__list_marker_minus] = ACTIONS(1764),
    [sym__list_marker_plus] = ACTIONS(1764),
    [sym__list_marker_star] = ACTIONS(1764),
    [sym__list_marker_parenthesis] = ACTIONS(1764),
    [sym__list_marker_dot] = ACTIONS(1764),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1764),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1764),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1764),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1764),
    [sym__blank_line_start] = ACTIONS(1764),
    [sym__html_block_1_start] = ACTIONS(1764),
    [sym__html_block_2_start] = ACTIONS(1764),
    [sym__html_block_3_start] = ACTIONS(1764),
    [sym__html_block_4_start] = ACTIONS(1764),
    [sym__html_block_5_start] = ACTIONS(1764),
    [sym__html_block_6_start] = ACTIONS(1764),
    [sym__html_block_7_start] = ACTIONS(1764),
    [sym__pipe_table_start] = ACTIONS(1764),
  },
  [STATE(342)] = {
    [ts_builtin_sym_end] = ACTIONS(1768),
    [anon_sym_LBRACK] = ACTIONS(1766),
    [anon_sym_RBRACK] = ACTIONS(1768),
    [anon_sym_LT] = ACTIONS(1768),
    [anon_sym_GT] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_DQUOTE] = ACTIONS(1768),
    [anon_sym_POUND] = ACTIONS(1768),
    [anon_sym_DOLLAR] = ACTIONS(1768),
    [anon_sym_PERCENT] = ACTIONS(1768),
    [anon_sym_AMP] = ACTIONS(1768),
    [anon_sym_SQUOTE] = ACTIONS(1768),
    [anon_sym_STAR] = ACTIONS(1768),
    [anon_sym_PLUS] = ACTIONS(1768),
    [anon_sym_COMMA] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(1768),
    [anon_sym_DOT] = ACTIONS(1768),
    [anon_sym_SLASH] = ACTIONS(1768),
    [anon_sym_COLON] = ACTIONS(1768),
    [anon_sym_SEMI] = ACTIONS(1768),
    [anon_sym_EQ] = ACTIONS(1768),
    [anon_sym_QMARK] = ACTIONS(1768),
    [anon_sym_AT] = ACTIONS(1768),
    [anon_sym_BSLASH] = ACTIONS(1768),
    [anon_sym_CARET] = ACTIONS(1768),
    [anon_sym__] = ACTIONS(1768),
    [anon_sym_BQUOTE] = ACTIONS(1768),
    [anon_sym_LBRACE] = ACTIONS(1768),
    [anon_sym_PIPE] = ACTIONS(1768),
    [anon_sym_RBRACE] = ACTIONS(1768),
    [anon_sym_TILDE] = ACTIONS(1768),
    [anon_sym_LPAREN] = ACTIONS(1768),
    [anon_sym_RPAREN] = ACTIONS(1768),
    [aux_sym__word_token1] = ACTIONS(1768),
    [aux_sym__word_token2] = ACTIONS(1768),
    [aux_sym__word_token3] = ACTIONS(1768),
    [sym__whitespace] = ACTIONS(1768),
    [sym__soft_line_ending] = ACTIONS(1768),
    [sym__block_quote_start] = ACTIONS(1768),
    [sym__indented_chunk_start] = ACTIONS(1768),
    [sym_atx_h1_marker] = ACTIONS(1768),
    [sym_atx_h2_marker] = ACTIONS(1768),
    [sym_atx_h3_marker] = ACTIONS(1768),
    [sym_atx_h4_marker] = ACTIONS(1768),
    [sym_atx_h5_marker] = ACTIONS(1768),
    [sym_atx_h6_marker] = ACTIONS(1768),
    [sym__thematic_break] = ACTIONS(1768),
    [sym__list_marker_minus] = ACTIONS(1768),
    [sym__list_marker_plus] = ACTIONS(1768),
    [sym__list_marker_star] = ACTIONS(1768),
    [sym__list_marker_parenthesis] = ACTIONS(1768),
    [sym__list_marker_dot] = ACTIONS(1768),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1768),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1768),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1768),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1768),
    [sym__blank_line_start] = ACTIONS(1768),
    [sym__html_block_1_start] = ACTIONS(1768),
    [sym__html_block_2_start] = ACTIONS(1768),
    [sym__html_block_3_start] = ACTIONS(1768),
    [sym__html_block_4_start] = ACTIONS(1768),
    [sym__html_block_5_start] = ACTIONS(1768),
    [sym__html_block_6_start] = ACTIONS(1768),
    [sym__html_block_7_start] = ACTIONS(1768),
    [sym__pipe_table_start] = ACTIONS(1768),
  },
  [STATE(343)] = {
    [ts_builtin_sym_end] = ACTIONS(1772),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_RBRACK] = ACTIONS(1772),
    [anon_sym_LT] = ACTIONS(1772),
    [anon_sym_GT] = ACTIONS(1772),
    [anon_sym_BANG] = ACTIONS(1772),
    [anon_sym_DQUOTE] = ACTIONS(1772),
    [anon_sym_POUND] = ACTIONS(1772),
    [anon_sym_DOLLAR] = ACTIONS(1772),
    [anon_sym_PERCENT] = ACTIONS(1772),
    [anon_sym_AMP] = ACTIONS(1772),
    [anon_sym_SQUOTE] = ACTIONS(1772),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_PLUS] = ACTIONS(1772),
    [anon_sym_COMMA] = ACTIONS(1772),
    [anon_sym_DASH] = ACTIONS(1772),
    [anon_sym_DOT] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1772),
    [anon_sym_COLON] = ACTIONS(1772),
    [anon_sym_SEMI] = ACTIONS(1772),
    [anon_sym_EQ] = ACTIONS(1772),
    [anon_sym_QMARK] = ACTIONS(1772),
    [anon_sym_AT] = ACTIONS(1772),
    [anon_sym_BSLASH] = ACTIONS(1772),
    [anon_sym_CARET] = ACTIONS(1772),
    [anon_sym__] = ACTIONS(1772),
    [anon_sym_BQUOTE] = ACTIONS(1772),
    [anon_sym_LBRACE] = ACTIONS(1772),
    [anon_sym_PIPE] = ACTIONS(1772),
    [anon_sym_RBRACE] = ACTIONS(1772),
    [anon_sym_TILDE] = ACTIONS(1772),
    [anon_sym_LPAREN] = ACTIONS(1772),
    [anon_sym_RPAREN] = ACTIONS(1772),
    [aux_sym__word_token1] = ACTIONS(1772),
    [aux_sym__word_token2] = ACTIONS(1772),
    [aux_sym__word_token3] = ACTIONS(1772),
    [sym__whitespace] = ACTIONS(1772),
    [sym__soft_line_ending] = ACTIONS(1772),
    [sym__block_quote_start] = ACTIONS(1772),
    [sym__indented_chunk_start] = ACTIONS(1772),
    [sym_atx_h1_marker] = ACTIONS(1772),
    [sym_atx_h2_marker] = ACTIONS(1772),
    [sym_atx_h3_marker] = ACTIONS(1772),
    [sym_atx_h4_marker] = ACTIONS(1772),
    [sym_atx_h5_marker] = ACTIONS(1772),
    [sym_atx_h6_marker] = ACTIONS(1772),
    [sym__thematic_break] = ACTIONS(1772),
    [sym__list_marker_minus] = ACTIONS(1772),
    [sym__list_marker_plus] = ACTIONS(1772),
    [sym__list_marker_star] = ACTIONS(1772),
    [sym__list_marker_parenthesis] = ACTIONS(1772),
    [sym__list_marker_dot] = ACTIONS(1772),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1772),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1772),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1772),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1772),
    [sym__blank_line_start] = ACTIONS(1772),
    [sym__html_block_1_start] = ACTIONS(1772),
    [sym__html_block_2_start] = ACTIONS(1772),
    [sym__html_block_3_start] = ACTIONS(1772),
    [sym__html_block_4_start] = ACTIONS(1772),
    [sym__html_block_5_start] = ACTIONS(1772),
    [sym__html_block_6_start] = ACTIONS(1772),
    [sym__html_block_7_start] = ACTIONS(1772),
    [sym__pipe_table_start] = ACTIONS(1772),
  },
  [STATE(344)] = {
    [ts_builtin_sym_end] = ACTIONS(1776),
    [anon_sym_LBRACK] = ACTIONS(1774),
    [anon_sym_RBRACK] = ACTIONS(1776),
    [anon_sym_LT] = ACTIONS(1776),
    [anon_sym_GT] = ACTIONS(1776),
    [anon_sym_BANG] = ACTIONS(1776),
    [anon_sym_DQUOTE] = ACTIONS(1776),
    [anon_sym_POUND] = ACTIONS(1776),
    [anon_sym_DOLLAR] = ACTIONS(1776),
    [anon_sym_PERCENT] = ACTIONS(1776),
    [anon_sym_AMP] = ACTIONS(1776),
    [anon_sym_SQUOTE] = ACTIONS(1776),
    [anon_sym_STAR] = ACTIONS(1776),
    [anon_sym_PLUS] = ACTIONS(1776),
    [anon_sym_COMMA] = ACTIONS(1776),
    [anon_sym_DASH] = ACTIONS(1776),
    [anon_sym_DOT] = ACTIONS(1776),
    [anon_sym_SLASH] = ACTIONS(1776),
    [anon_sym_COLON] = ACTIONS(1776),
    [anon_sym_SEMI] = ACTIONS(1776),
    [anon_sym_EQ] = ACTIONS(1776),
    [anon_sym_QMARK] = ACTIONS(1776),
    [anon_sym_AT] = ACTIONS(1776),
    [anon_sym_BSLASH] = ACTIONS(1776),
    [anon_sym_CARET] = ACTIONS(1776),
    [anon_sym__] = ACTIONS(1776),
    [anon_sym_BQUOTE] = ACTIONS(1776),
    [anon_sym_LBRACE] = ACTIONS(1776),
    [anon_sym_PIPE] = ACTIONS(1776),
    [anon_sym_RBRACE] = ACTIONS(1776),
    [anon_sym_TILDE] = ACTIONS(1776),
    [anon_sym_LPAREN] = ACTIONS(1776),
    [anon_sym_RPAREN] = ACTIONS(1776),
    [aux_sym__word_token1] = ACTIONS(1776),
    [aux_sym__word_token2] = ACTIONS(1776),
    [aux_sym__word_token3] = ACTIONS(1776),
    [sym__whitespace] = ACTIONS(1776),
    [sym__soft_line_ending] = ACTIONS(1776),
    [sym__block_quote_start] = ACTIONS(1776),
    [sym__indented_chunk_start] = ACTIONS(1776),
    [sym_atx_h1_marker] = ACTIONS(1776),
    [sym_atx_h2_marker] = ACTIONS(1776),
    [sym_atx_h3_marker] = ACTIONS(1776),
    [sym_atx_h4_marker] = ACTIONS(1776),
    [sym_atx_h5_marker] = ACTIONS(1776),
    [sym_atx_h6_marker] = ACTIONS(1776),
    [sym__thematic_break] = ACTIONS(1776),
    [sym__list_marker_minus] = ACTIONS(1776),
    [sym__list_marker_plus] = ACTIONS(1776),
    [sym__list_marker_star] = ACTIONS(1776),
    [sym__list_marker_parenthesis] = ACTIONS(1776),
    [sym__list_marker_dot] = ACTIONS(1776),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1776),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1776),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1776),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1776),
    [sym__blank_line_start] = ACTIONS(1776),
    [sym__html_block_1_start] = ACTIONS(1776),
    [sym__html_block_2_start] = ACTIONS(1776),
    [sym__html_block_3_start] = ACTIONS(1776),
    [sym__html_block_4_start] = ACTIONS(1776),
    [sym__html_block_5_start] = ACTIONS(1776),
    [sym__html_block_6_start] = ACTIONS(1776),
    [sym__html_block_7_start] = ACTIONS(1776),
    [sym__pipe_table_start] = ACTIONS(1776),
  },
  [STATE(345)] = {
    [ts_builtin_sym_end] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(1186),
    [anon_sym_RBRACK] = ACTIONS(1188),
    [anon_sym_LT] = ACTIONS(1188),
    [anon_sym_GT] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [anon_sym_POUND] = ACTIONS(1188),
    [anon_sym_DOLLAR] = ACTIONS(1188),
    [anon_sym_PERCENT] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_PLUS] = ACTIONS(1188),
    [anon_sym_COMMA] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_SLASH] = ACTIONS(1188),
    [anon_sym_COLON] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym_EQ] = ACTIONS(1188),
    [anon_sym_QMARK] = ACTIONS(1188),
    [anon_sym_AT] = ACTIONS(1188),
    [anon_sym_BSLASH] = ACTIONS(1188),
    [anon_sym_CARET] = ACTIONS(1188),
    [anon_sym__] = ACTIONS(1188),
    [anon_sym_BQUOTE] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_PIPE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_LPAREN] = ACTIONS(1188),
    [anon_sym_RPAREN] = ACTIONS(1188),
    [aux_sym__word_token1] = ACTIONS(1188),
    [aux_sym__word_token2] = ACTIONS(1188),
    [aux_sym__word_token3] = ACTIONS(1188),
    [sym__whitespace] = ACTIONS(1188),
    [sym__soft_line_ending] = ACTIONS(1188),
    [sym__block_quote_start] = ACTIONS(1188),
    [sym__indented_chunk_start] = ACTIONS(1188),
    [sym_atx_h1_marker] = ACTIONS(1188),
    [sym_atx_h2_marker] = ACTIONS(1188),
    [sym_atx_h3_marker] = ACTIONS(1188),
    [sym_atx_h4_marker] = ACTIONS(1188),
    [sym_atx_h5_marker] = ACTIONS(1188),
    [sym_atx_h6_marker] = ACTIONS(1188),
    [sym__thematic_break] = ACTIONS(1188),
    [sym__list_marker_minus] = ACTIONS(1188),
    [sym__list_marker_plus] = ACTIONS(1188),
    [sym__list_marker_star] = ACTIONS(1188),
    [sym__list_marker_parenthesis] = ACTIONS(1188),
    [sym__list_marker_dot] = ACTIONS(1188),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1188),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1188),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1188),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1188),
    [sym__blank_line_start] = ACTIONS(1188),
    [sym__html_block_1_start] = ACTIONS(1188),
    [sym__html_block_2_start] = ACTIONS(1188),
    [sym__html_block_3_start] = ACTIONS(1188),
    [sym__html_block_4_start] = ACTIONS(1188),
    [sym__html_block_5_start] = ACTIONS(1188),
    [sym__html_block_6_start] = ACTIONS(1188),
    [sym__html_block_7_start] = ACTIONS(1188),
    [sym__pipe_table_start] = ACTIONS(1188),
  },
  [STATE(346)] = {
    [ts_builtin_sym_end] = ACTIONS(1664),
    [anon_sym_LBRACK] = ACTIONS(1662),
    [anon_sym_RBRACK] = ACTIONS(1664),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_BANG] = ACTIONS(1664),
    [anon_sym_DQUOTE] = ACTIONS(1664),
    [anon_sym_POUND] = ACTIONS(1664),
    [anon_sym_DOLLAR] = ACTIONS(1664),
    [anon_sym_PERCENT] = ACTIONS(1664),
    [anon_sym_AMP] = ACTIONS(1664),
    [anon_sym_SQUOTE] = ACTIONS(1664),
    [anon_sym_STAR] = ACTIONS(1664),
    [anon_sym_PLUS] = ACTIONS(1664),
    [anon_sym_COMMA] = ACTIONS(1664),
    [anon_sym_DASH] = ACTIONS(1664),
    [anon_sym_DOT] = ACTIONS(1664),
    [anon_sym_SLASH] = ACTIONS(1664),
    [anon_sym_COLON] = ACTIONS(1664),
    [anon_sym_SEMI] = ACTIONS(1664),
    [anon_sym_EQ] = ACTIONS(1664),
    [anon_sym_QMARK] = ACTIONS(1664),
    [anon_sym_AT] = ACTIONS(1664),
    [anon_sym_BSLASH] = ACTIONS(1664),
    [anon_sym_CARET] = ACTIONS(1664),
    [anon_sym__] = ACTIONS(1664),
    [anon_sym_BQUOTE] = ACTIONS(1664),
    [anon_sym_LBRACE] = ACTIONS(1664),
    [anon_sym_PIPE] = ACTIONS(1664),
    [anon_sym_RBRACE] = ACTIONS(1664),
    [anon_sym_TILDE] = ACTIONS(1664),
    [anon_sym_LPAREN] = ACTIONS(1664),
    [anon_sym_RPAREN] = ACTIONS(1664),
    [aux_sym__word_token1] = ACTIONS(1664),
    [aux_sym__word_token2] = ACTIONS(1664),
    [aux_sym__word_token3] = ACTIONS(1664),
    [sym__whitespace] = ACTIONS(1664),
    [sym__soft_line_ending] = ACTIONS(1664),
    [sym__block_quote_start] = ACTIONS(1664),
    [sym__indented_chunk_start] = ACTIONS(1664),
    [sym_atx_h1_marker] = ACTIONS(1664),
    [sym_atx_h2_marker] = ACTIONS(1664),
    [sym_atx_h3_marker] = ACTIONS(1664),
    [sym_atx_h4_marker] = ACTIONS(1664),
    [sym_atx_h5_marker] = ACTIONS(1664),
    [sym_atx_h6_marker] = ACTIONS(1664),
    [sym__thematic_break] = ACTIONS(1664),
    [sym__list_marker_minus] = ACTIONS(1664),
    [sym__list_marker_plus] = ACTIONS(1664),
    [sym__list_marker_star] = ACTIONS(1664),
    [sym__list_marker_parenthesis] = ACTIONS(1664),
    [sym__list_marker_dot] = ACTIONS(1664),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1664),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1664),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1664),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1664),
    [sym__blank_line_start] = ACTIONS(1664),
    [sym__html_block_1_start] = ACTIONS(1664),
    [sym__html_block_2_start] = ACTIONS(1664),
    [sym__html_block_3_start] = ACTIONS(1664),
    [sym__html_block_4_start] = ACTIONS(1664),
    [sym__html_block_5_start] = ACTIONS(1664),
    [sym__html_block_6_start] = ACTIONS(1664),
    [sym__html_block_7_start] = ACTIONS(1664),
    [sym__pipe_table_start] = ACTIONS(1664),
  },
  [STATE(347)] = {
    [ts_builtin_sym_end] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1778),
    [anon_sym_RBRACK] = ACTIONS(1780),
    [anon_sym_LT] = ACTIONS(1780),
    [anon_sym_GT] = ACTIONS(1780),
    [anon_sym_BANG] = ACTIONS(1780),
    [anon_sym_DQUOTE] = ACTIONS(1780),
    [anon_sym_POUND] = ACTIONS(1780),
    [anon_sym_DOLLAR] = ACTIONS(1780),
    [anon_sym_PERCENT] = ACTIONS(1780),
    [anon_sym_AMP] = ACTIONS(1780),
    [anon_sym_SQUOTE] = ACTIONS(1780),
    [anon_sym_STAR] = ACTIONS(1780),
    [anon_sym_PLUS] = ACTIONS(1780),
    [anon_sym_COMMA] = ACTIONS(1780),
    [anon_sym_DASH] = ACTIONS(1780),
    [anon_sym_DOT] = ACTIONS(1780),
    [anon_sym_SLASH] = ACTIONS(1780),
    [anon_sym_COLON] = ACTIONS(1780),
    [anon_sym_SEMI] = ACTIONS(1780),
    [anon_sym_EQ] = ACTIONS(1780),
    [anon_sym_QMARK] = ACTIONS(1780),
    [anon_sym_AT] = ACTIONS(1780),
    [anon_sym_BSLASH] = ACTIONS(1780),
    [anon_sym_CARET] = ACTIONS(1780),
    [anon_sym__] = ACTIONS(1780),
    [anon_sym_BQUOTE] = ACTIONS(1780),
    [anon_sym_LBRACE] = ACTIONS(1780),
    [anon_sym_PIPE] = ACTIONS(1780),
    [anon_sym_RBRACE] = ACTIONS(1780),
    [anon_sym_TILDE] = ACTIONS(1780),
    [anon_sym_LPAREN] = ACTIONS(1780),
    [anon_sym_RPAREN] = ACTIONS(1780),
    [aux_sym__word_token1] = ACTIONS(1780),
    [aux_sym__word_token2] = ACTIONS(1780),
    [aux_sym__word_token3] = ACTIONS(1780),
    [sym__whitespace] = ACTIONS(1780),
    [sym__soft_line_ending] = ACTIONS(1780),
    [sym__block_quote_start] = ACTIONS(1780),
    [sym__indented_chunk_start] = ACTIONS(1780),
    [sym_atx_h1_marker] = ACTIONS(1780),
    [sym_atx_h2_marker] = ACTIONS(1780),
    [sym_atx_h3_marker] = ACTIONS(1780),
    [sym_atx_h4_marker] = ACTIONS(1780),
    [sym_atx_h5_marker] = ACTIONS(1780),
    [sym_atx_h6_marker] = ACTIONS(1780),
    [sym__thematic_break] = ACTIONS(1780),
    [sym__list_marker_minus] = ACTIONS(1780),
    [sym__list_marker_plus] = ACTIONS(1780),
    [sym__list_marker_star] = ACTIONS(1780),
    [sym__list_marker_parenthesis] = ACTIONS(1780),
    [sym__list_marker_dot] = ACTIONS(1780),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1780),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1780),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1780),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1780),
    [sym__blank_line_start] = ACTIONS(1780),
    [sym__html_block_1_start] = ACTIONS(1780),
    [sym__html_block_2_start] = ACTIONS(1780),
    [sym__html_block_3_start] = ACTIONS(1780),
    [sym__html_block_4_start] = ACTIONS(1780),
    [sym__html_block_5_start] = ACTIONS(1780),
    [sym__html_block_6_start] = ACTIONS(1780),
    [sym__html_block_7_start] = ACTIONS(1780),
    [sym__pipe_table_start] = ACTIONS(1780),
  },
  [STATE(348)] = {
    [ts_builtin_sym_end] = ACTIONS(1784),
    [anon_sym_LBRACK] = ACTIONS(1782),
    [anon_sym_RBRACK] = ACTIONS(1784),
    [anon_sym_LT] = ACTIONS(1784),
    [anon_sym_GT] = ACTIONS(1784),
    [anon_sym_BANG] = ACTIONS(1784),
    [anon_sym_DQUOTE] = ACTIONS(1784),
    [anon_sym_POUND] = ACTIONS(1784),
    [anon_sym_DOLLAR] = ACTIONS(1784),
    [anon_sym_PERCENT] = ACTIONS(1784),
    [anon_sym_AMP] = ACTIONS(1784),
    [anon_sym_SQUOTE] = ACTIONS(1784),
    [anon_sym_STAR] = ACTIONS(1784),
    [anon_sym_PLUS] = ACTIONS(1784),
    [anon_sym_COMMA] = ACTIONS(1784),
    [anon_sym_DASH] = ACTIONS(1784),
    [anon_sym_DOT] = ACTIONS(1784),
    [anon_sym_SLASH] = ACTIONS(1784),
    [anon_sym_COLON] = ACTIONS(1784),
    [anon_sym_SEMI] = ACTIONS(1784),
    [anon_sym_EQ] = ACTIONS(1784),
    [anon_sym_QMARK] = ACTIONS(1784),
    [anon_sym_AT] = ACTIONS(1784),
    [anon_sym_BSLASH] = ACTIONS(1784),
    [anon_sym_CARET] = ACTIONS(1784),
    [anon_sym__] = ACTIONS(1784),
    [anon_sym_BQUOTE] = ACTIONS(1784),
    [anon_sym_LBRACE] = ACTIONS(1784),
    [anon_sym_PIPE] = ACTIONS(1784),
    [anon_sym_RBRACE] = ACTIONS(1784),
    [anon_sym_TILDE] = ACTIONS(1784),
    [anon_sym_LPAREN] = ACTIONS(1784),
    [anon_sym_RPAREN] = ACTIONS(1784),
    [aux_sym__word_token1] = ACTIONS(1784),
    [aux_sym__word_token2] = ACTIONS(1784),
    [aux_sym__word_token3] = ACTIONS(1784),
    [sym__whitespace] = ACTIONS(1784),
    [sym__soft_line_ending] = ACTIONS(1784),
    [sym__block_quote_start] = ACTIONS(1784),
    [sym__indented_chunk_start] = ACTIONS(1784),
    [sym_atx_h1_marker] = ACTIONS(1784),
    [sym_atx_h2_marker] = ACTIONS(1784),
    [sym_atx_h3_marker] = ACTIONS(1784),
    [sym_atx_h4_marker] = ACTIONS(1784),
    [sym_atx_h5_marker] = ACTIONS(1784),
    [sym_atx_h6_marker] = ACTIONS(1784),
    [sym__thematic_break] = ACTIONS(1784),
    [sym__list_marker_minus] = ACTIONS(1784),
    [sym__list_marker_plus] = ACTIONS(1784),
    [sym__list_marker_star] = ACTIONS(1784),
    [sym__list_marker_parenthesis] = ACTIONS(1784),
    [sym__list_marker_dot] = ACTIONS(1784),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1784),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1784),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1784),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1784),
    [sym__blank_line_start] = ACTIONS(1784),
    [sym__html_block_1_start] = ACTIONS(1784),
    [sym__html_block_2_start] = ACTIONS(1784),
    [sym__html_block_3_start] = ACTIONS(1784),
    [sym__html_block_4_start] = ACTIONS(1784),
    [sym__html_block_5_start] = ACTIONS(1784),
    [sym__html_block_6_start] = ACTIONS(1784),
    [sym__html_block_7_start] = ACTIONS(1784),
    [sym__pipe_table_start] = ACTIONS(1784),
  },
  [STATE(349)] = {
    [ts_builtin_sym_end] = ACTIONS(1206),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_RBRACK] = ACTIONS(1206),
    [anon_sym_LT] = ACTIONS(1206),
    [anon_sym_GT] = ACTIONS(1206),
    [anon_sym_BANG] = ACTIONS(1206),
    [anon_sym_DQUOTE] = ACTIONS(1206),
    [anon_sym_POUND] = ACTIONS(1206),
    [anon_sym_DOLLAR] = ACTIONS(1206),
    [anon_sym_PERCENT] = ACTIONS(1206),
    [anon_sym_AMP] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_COMMA] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_SLASH] = ACTIONS(1206),
    [anon_sym_COLON] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1206),
    [anon_sym_EQ] = ACTIONS(1206),
    [anon_sym_QMARK] = ACTIONS(1206),
    [anon_sym_AT] = ACTIONS(1206),
    [anon_sym_BSLASH] = ACTIONS(1206),
    [anon_sym_CARET] = ACTIONS(1206),
    [anon_sym__] = ACTIONS(1206),
    [anon_sym_BQUOTE] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1206),
    [anon_sym_PIPE] = ACTIONS(1206),
    [anon_sym_RBRACE] = ACTIONS(1206),
    [anon_sym_TILDE] = ACTIONS(1206),
    [anon_sym_LPAREN] = ACTIONS(1206),
    [anon_sym_RPAREN] = ACTIONS(1206),
    [aux_sym__word_token1] = ACTIONS(1206),
    [aux_sym__word_token2] = ACTIONS(1206),
    [aux_sym__word_token3] = ACTIONS(1206),
    [sym__whitespace] = ACTIONS(1206),
    [sym__soft_line_ending] = ACTIONS(1206),
    [sym__block_quote_start] = ACTIONS(1206),
    [sym__indented_chunk_start] = ACTIONS(1206),
    [sym_atx_h1_marker] = ACTIONS(1206),
    [sym_atx_h2_marker] = ACTIONS(1206),
    [sym_atx_h3_marker] = ACTIONS(1206),
    [sym_atx_h4_marker] = ACTIONS(1206),
    [sym_atx_h5_marker] = ACTIONS(1206),
    [sym_atx_h6_marker] = ACTIONS(1206),
    [sym__thematic_break] = ACTIONS(1206),
    [sym__list_marker_minus] = ACTIONS(1206),
    [sym__list_marker_plus] = ACTIONS(1206),
    [sym__list_marker_star] = ACTIONS(1206),
    [sym__list_marker_parenthesis] = ACTIONS(1206),
    [sym__list_marker_dot] = ACTIONS(1206),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1206),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1206),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1206),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1206),
    [sym__blank_line_start] = ACTIONS(1206),
    [sym__html_block_1_start] = ACTIONS(1206),
    [sym__html_block_2_start] = ACTIONS(1206),
    [sym__html_block_3_start] = ACTIONS(1206),
    [sym__html_block_4_start] = ACTIONS(1206),
    [sym__html_block_5_start] = ACTIONS(1206),
    [sym__html_block_6_start] = ACTIONS(1206),
    [sym__html_block_7_start] = ACTIONS(1206),
    [sym__pipe_table_start] = ACTIONS(1206),
  },
  [STATE(350)] = {
    [ts_builtin_sym_end] = ACTIONS(1732),
    [anon_sym_LBRACK] = ACTIONS(1730),
    [anon_sym_RBRACK] = ACTIONS(1732),
    [anon_sym_LT] = ACTIONS(1732),
    [anon_sym_GT] = ACTIONS(1732),
    [anon_sym_BANG] = ACTIONS(1732),
    [anon_sym_DQUOTE] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_DOLLAR] = ACTIONS(1732),
    [anon_sym_PERCENT] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1732),
    [anon_sym_STAR] = ACTIONS(1732),
    [anon_sym_PLUS] = ACTIONS(1732),
    [anon_sym_COMMA] = ACTIONS(1732),
    [anon_sym_DASH] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1732),
    [anon_sym_SLASH] = ACTIONS(1732),
    [anon_sym_COLON] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym_EQ] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_AT] = ACTIONS(1732),
    [anon_sym_BSLASH] = ACTIONS(1732),
    [anon_sym_CARET] = ACTIONS(1732),
    [anon_sym__] = ACTIONS(1732),
    [anon_sym_BQUOTE] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_RBRACE] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1732),
    [anon_sym_RPAREN] = ACTIONS(1732),
    [aux_sym__word_token1] = ACTIONS(1732),
    [aux_sym__word_token2] = ACTIONS(1732),
    [aux_sym__word_token3] = ACTIONS(1732),
    [sym__whitespace] = ACTIONS(1732),
    [sym__soft_line_ending] = ACTIONS(1732),
    [sym__block_quote_start] = ACTIONS(1732),
    [sym__indented_chunk_start] = ACTIONS(1732),
    [sym_atx_h1_marker] = ACTIONS(1732),
    [sym_atx_h2_marker] = ACTIONS(1732),
    [sym_atx_h3_marker] = ACTIONS(1732),
    [sym_atx_h4_marker] = ACTIONS(1732),
    [sym_atx_h5_marker] = ACTIONS(1732),
    [sym_atx_h6_marker] = ACTIONS(1732),
    [sym__thematic_break] = ACTIONS(1732),
    [sym__list_marker_minus] = ACTIONS(1732),
    [sym__list_marker_plus] = ACTIONS(1732),
    [sym__list_marker_star] = ACTIONS(1732),
    [sym__list_marker_parenthesis] = ACTIONS(1732),
    [sym__list_marker_dot] = ACTIONS(1732),
    [sym__list_marker_minus_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_plus_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_star_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_parenthesis_dont_interrupt] = ACTIONS(1732),
    [sym__list_marker_dot_dont_interrupt] = ACTIONS(1732),
    [sym__fenced_code_block_start_backtick] = ACTIONS(1732),
    [sym__fenced_code_block_start_tilde] = ACTIONS(1732),
    [sym__blank_line_start] = ACTIONS(1732),
    [sym__html_block_1_start] = ACTIONS(1732),
    [sym__html_block_2_start] = ACTIONS(1732),
    [sym__html_block_3_start] = ACTIONS(1732),
    [sym__html_block_4_start] = ACTIONS(1732),
    [sym__html_block_5_start] = ACTIONS(1732),
    [sym__html_block_6_start] = ACTIONS(1732),
    [sym__html_block_7_start] = ACTIONS(1732),
    [sym__pipe_table_start] = ACTIONS(1732),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 11,
    ACTIONS(97), 1,
      sym__html_block_1_start,
    ACTIONS(99), 1,
      sym__html_block_2_start,
    ACTIONS(101), 1,
      sym__html_block_3_start,
    ACTIONS(103), 1,
      sym__html_block_4_start,
    ACTIONS(105), 1,
      sym__html_block_5_start,
    ACTIONS(107), 1,
      sym__html_block_6_start,
    ACTIONS(109), 1,
      sym__html_block_7_start,
    ACTIONS(1786), 1,
      anon_sym_LBRACK,
    STATE(918), 1,
      sym_link_label,
    STATE(184), 7,
      sym__html_block_1,
      sym__html_block_2,
      sym__html_block_3,
      sym__html_block_4,
      sym__html_block_5,
      sym__html_block_6,
      sym__html_block_7,
    ACTIONS(1789), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [77] = 11,
    ACTIONS(47), 1,
      sym__html_block_1_start,
    ACTIONS(49), 1,
      sym__html_block_2_start,
    ACTIONS(51), 1,
      sym__html_block_3_start,
    ACTIONS(53), 1,
      sym__html_block_4_start,
    ACTIONS(55), 1,
      sym__html_block_5_start,
    ACTIONS(57), 1,
      sym__html_block_6_start,
    ACTIONS(59), 1,
      sym__html_block_7_start,
    ACTIONS(1786), 1,
      anon_sym_LBRACK,
    STATE(924), 1,
      sym_link_label,
    STATE(225), 7,
      sym__html_block_1,
      sym__html_block_2,
      sym__html_block_3,
      sym__html_block_4,
      sym__html_block_5,
      sym__html_block_6,
      sym__html_block_7,
    ACTIONS(1789), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [154] = 9,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1795), 1,
      anon_sym_LBRACK,
    ACTIONS(1803), 1,
      sym__soft_line_ending,
    STATE(584), 1,
      sym__last_token_punctuation,
    ACTIONS(1793), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    ACTIONS(1800), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1789), 3,
      sym__line_ending,
      sym__eof,
      anon_sym_RBRACK,
    STATE(404), 5,
      sym_backslash_escape,
      sym__text_inline_no_link,
      sym__soft_line_break,
      sym__word,
      aux_sym_link_label_repeat1,
    ACTIONS(1797), 32,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [221] = 11,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1814), 1,
      sym__whitespace,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(367), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(828), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [290] = 11,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1818), 1,
      sym__whitespace,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(365), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(831), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [359] = 11,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1822), 1,
      sym__whitespace,
    STATE(366), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(832), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [428] = 11,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1824), 1,
      sym__whitespace,
    STATE(374), 1,
      sym_language,
    STATE(376), 1,
      sym__newline,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(853), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [497] = 7,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    STATE(361), 1,
      sym__last_token_punctuation,
    ACTIONS(1830), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1834), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(385), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1828), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [558] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1840), 1,
      sym__block_close,
    ACTIONS(1842), 1,
      sym__fenced_code_block_end_tilde,
    STATE(818), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [620] = 6,
    ACTIONS(1844), 1,
      sym__backslash_escape,
    ACTIONS(1853), 1,
      anon_sym_LPAREN,
    ACTIONS(1850), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1856), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(360), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1847), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [678] = 6,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1860), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1862), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(363), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1858), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [736] = 9,
    ACTIONS(11), 1,
      sym__soft_line_ending,
    ACTIONS(1864), 1,
      anon_sym_LBRACK,
    ACTIONS(1866), 1,
      sym__line_ending,
    ACTIONS(1868), 1,
      sym__eof,
    STATE(111), 1,
      sym__newline,
    STATE(413), 1,
      aux_sym_paragraph_repeat1,
    STATE(506), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(601), 2,
      sym__soft_line_break,
      sym__line,
    ACTIONS(7), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [800] = 6,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1872), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1874), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(360), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1870), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [858] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1876), 1,
      sym__block_close,
    ACTIONS(1878), 1,
      sym__fenced_code_block_end_tilde,
    STATE(854), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [920] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    ACTIONS(1884), 1,
      sym__block_close,
    ACTIONS(1886), 1,
      sym__fenced_code_block_end_backtick,
    STATE(811), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [982] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1884), 1,
      sym__block_close,
    ACTIONS(1886), 1,
      sym__fenced_code_block_end_tilde,
    STATE(812), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1044] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1888), 1,
      sym__block_close,
    ACTIONS(1890), 1,
      sym__fenced_code_block_end_tilde,
    STATE(851), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1106] = 10,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(374), 1,
      sym_language,
    STATE(377), 1,
      sym__newline,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(845), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [1172] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    ACTIONS(1892), 1,
      sym__block_close,
    ACTIONS(1894), 1,
      sym__fenced_code_block_end_backtick,
    STATE(814), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1234] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1892), 1,
      sym__block_close,
    ACTIONS(1894), 1,
      sym__fenced_code_block_end_tilde,
    STATE(815), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1296] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1840), 1,
      sym__block_close,
    ACTIONS(1842), 1,
      sym__fenced_code_block_end_backtick,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    STATE(817), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1358] = 7,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1902), 1,
      sym__line_ending,
    ACTIONS(1896), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    STATE(466), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(1898), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(380), 3,
      sym_backslash_escape,
      sym__line,
      aux_sym_info_string_repeat1,
    ACTIONS(1900), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1418] = 9,
    ACTIONS(11), 1,
      sym__soft_line_ending,
    ACTIONS(1864), 1,
      anon_sym_LBRACK,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(1906), 1,
      sym__eof,
    STATE(326), 1,
      sym__newline,
    STATE(413), 1,
      aux_sym_paragraph_repeat1,
    STATE(506), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(601), 2,
      sym__soft_line_break,
      sym__line,
    ACTIONS(7), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1482] = 7,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1910), 1,
      sym__line_ending,
    ACTIONS(1908), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    STATE(466), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(1898), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(372), 3,
      sym_backslash_escape,
      sym__line,
      aux_sym_info_string_repeat1,
    ACTIONS(1900), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1542] = 9,
    ACTIONS(11), 1,
      sym__soft_line_ending,
    ACTIONS(1864), 1,
      anon_sym_LBRACK,
    ACTIONS(1912), 1,
      sym__line_ending,
    ACTIONS(1914), 1,
      sym__eof,
    STATE(113), 1,
      sym__newline,
    STATE(413), 1,
      aux_sym_paragraph_repeat1,
    STATE(506), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(601), 2,
      sym__soft_line_break,
      sym__line,
    ACTIONS(7), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1606] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    ACTIONS(1888), 1,
      sym__block_close,
    ACTIONS(1890), 1,
      sym__fenced_code_block_end_backtick,
    STATE(850), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1668] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1876), 1,
      sym__block_close,
    ACTIONS(1878), 1,
      sym__fenced_code_block_end_backtick,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    STATE(822), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1730] = 8,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    ACTIONS(1916), 1,
      sym__block_close,
    ACTIONS(1918), 1,
      sym__fenced_code_block_end_backtick,
    STATE(843), 1,
      sym_code_fence_content,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(396), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1792] = 7,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1902), 1,
      sym__line_ending,
    ACTIONS(1920), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    STATE(466), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(1898), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(383), 3,
      sym_backslash_escape,
      sym__line,
      aux_sym_info_string_repeat1,
    ACTIONS(1900), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1852] = 7,
    ACTIONS(1922), 1,
      sym__backslash_escape,
    ACTIONS(1934), 1,
      sym__line_ending,
    ACTIONS(1925), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    STATE(466), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(1928), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(380), 3,
      sym_backslash_escape,
      sym__line,
      aux_sym_info_string_repeat1,
    ACTIONS(1931), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [1912] = 10,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(369), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(835), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [1978] = 10,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(370), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(836), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [2044] = 7,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1936), 1,
      sym__line_ending,
    ACTIONS(1896), 2,
      sym_entity_reference,
      sym_numeric_character_reference,
    STATE(466), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(1898), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(380), 3,
      sym_backslash_escape,
      sym__line,
      aux_sym_info_string_repeat1,
    ACTIONS(1900), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2104] = 8,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(1916), 1,
      sym__block_close,
    ACTIONS(1918), 1,
      sym__fenced_code_block_end_tilde,
    STATE(825), 1,
      sym_code_fence_content,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(407), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2166] = 6,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1872), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1862), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(360), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1870), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [2224] = 6,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1830), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1834), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
    STATE(385), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(1828), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [2282] = 10,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(364), 1,
      sym__newline,
    STATE(374), 1,
      sym_language,
    STATE(414), 1,
      aux_sym_info_string_repeat2,
    STATE(841), 1,
      sym_info_string,
    ACTIONS(1812), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [2348] = 7,
    ACTIONS(1942), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(1944), 1,
      sym__line_ending,
    ACTIONS(1946), 1,
      sym__block_close,
    ACTIONS(1938), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(513), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(399), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_2_repeat1,
    ACTIONS(1940), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2407] = 11,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1956), 1,
      sym__whitespace,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    STATE(493), 1,
      sym__soft_line_break,
    STATE(680), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [2474] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1964), 1,
      anon_sym_DQUOTE,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    STATE(525), 1,
      sym__soft_line_break,
    ACTIONS(1962), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(393), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(1960), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2533] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1964), 1,
      anon_sym_SQUOTE,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    STATE(523), 1,
      sym__soft_line_break,
    ACTIONS(1970), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(394), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(1968), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2592] = 7,
    ACTIONS(1942), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(1944), 1,
      sym__line_ending,
    ACTIONS(1972), 1,
      sym__block_close,
    ACTIONS(1938), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(513), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(399), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_2_repeat1,
    ACTIONS(1940), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2651] = 7,
    ACTIONS(1974), 1,
      sym__backslash_escape,
    ACTIONS(1983), 1,
      anon_sym_DQUOTE,
    ACTIONS(1985), 1,
      sym__soft_line_ending,
    STATE(525), 1,
      sym__soft_line_break,
    ACTIONS(1980), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(393), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(1977), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2710] = 7,
    ACTIONS(1988), 1,
      sym__backslash_escape,
    ACTIONS(1997), 1,
      anon_sym_SQUOTE,
    ACTIONS(1999), 1,
      sym__soft_line_ending,
    STATE(523), 1,
      sym__soft_line_break,
    ACTIONS(1994), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(394), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(1991), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2769] = 7,
    ACTIONS(2006), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2008), 1,
      sym__line_ending,
    ACTIONS(2010), 1,
      sym__block_close,
    ACTIONS(2002), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(518), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(400), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_3_repeat1,
    ACTIONS(2004), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2828] = 6,
    ACTIONS(1820), 1,
      sym__line_ending,
    ACTIONS(1880), 1,
      anon_sym_LBRACK,
    ACTIONS(2012), 2,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(410), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1882), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2885] = 7,
    ACTIONS(2014), 1,
      anon_sym_LBRACK,
    ACTIONS(2020), 1,
      sym__line_ending,
    ACTIONS(2023), 1,
      sym__block_close,
    ACTIONS(2025), 1,
      sym__html_block_1_end,
    STATE(509), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(397), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_1_repeat1,
    ACTIONS(2017), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [2944] = 7,
    ACTIONS(2032), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2034), 1,
      sym__line_ending,
    ACTIONS(2036), 1,
      sym__block_close,
    ACTIONS(2028), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(520), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(401), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_5_repeat1,
    ACTIONS(2030), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3003] = 7,
    ACTIONS(2044), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2047), 1,
      sym__line_ending,
    ACTIONS(2050), 1,
      sym__block_close,
    ACTIONS(2038), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(513), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(399), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_2_repeat1,
    ACTIONS(2041), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3062] = 7,
    ACTIONS(2058), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2061), 1,
      sym__line_ending,
    ACTIONS(2064), 1,
      sym__block_close,
    ACTIONS(2052), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(518), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(400), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_3_repeat1,
    ACTIONS(2055), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3121] = 7,
    ACTIONS(2072), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2075), 1,
      sym__line_ending,
    ACTIONS(2078), 1,
      sym__block_close,
    ACTIONS(2066), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(520), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(401), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_5_repeat1,
    ACTIONS(2069), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3180] = 6,
    ACTIONS(2080), 1,
      sym__backslash_escape,
    ACTIONS(2086), 1,
      anon_sym_RBRACK,
    ACTIONS(2091), 1,
      sym__soft_line_ending,
    ACTIONS(2088), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(402), 5,
      sym_backslash_escape,
      sym__text_inline_no_link,
      sym__soft_line_break,
      sym__word,
      aux_sym_link_label_repeat1,
    ACTIONS(2083), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3237] = 11,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    ACTIONS(2094), 1,
      sym__whitespace,
    STATE(484), 1,
      sym__soft_line_break,
    STATE(667), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [3304] = 6,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    ACTIONS(2098), 1,
      anon_sym_RBRACK,
    ACTIONS(2100), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(402), 5,
      sym_backslash_escape,
      sym__text_inline_no_link,
      sym__soft_line_break,
      sym__word,
      aux_sym_link_label_repeat1,
    ACTIONS(2096), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3361] = 11,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    ACTIONS(2102), 1,
      sym__whitespace,
    STATE(482), 1,
      sym__soft_line_break,
    STATE(664), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [3428] = 7,
    ACTIONS(2104), 1,
      anon_sym_LBRACK,
    ACTIONS(2108), 1,
      sym__line_ending,
    ACTIONS(2110), 1,
      sym__block_close,
    ACTIONS(2112), 1,
      sym__html_block_1_end,
    STATE(509), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(426), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_1_repeat1,
    ACTIONS(2106), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3487] = 6,
    ACTIONS(1816), 1,
      sym__line_ending,
    ACTIONS(1836), 1,
      anon_sym_LBRACK,
    ACTIONS(2012), 2,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(412), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(1838), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3544] = 7,
    ACTIONS(1942), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(1944), 1,
      sym__line_ending,
    ACTIONS(2114), 1,
      sym__block_close,
    ACTIONS(1938), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(513), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(392), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_2_repeat1,
    ACTIONS(1940), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3603] = 7,
    ACTIONS(2006), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2008), 1,
      sym__line_ending,
    ACTIONS(2116), 1,
      sym__block_close,
    ACTIONS(2002), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(518), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(395), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_3_repeat1,
    ACTIONS(2004), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3662] = 6,
    ACTIONS(2118), 1,
      anon_sym_LBRACK,
    ACTIONS(2124), 1,
      sym__line_ending,
    ACTIONS(2127), 2,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
    STATE(499), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(410), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(2121), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3719] = 7,
    ACTIONS(2032), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2034), 1,
      sym__line_ending,
    ACTIONS(2129), 1,
      sym__block_close,
    ACTIONS(2028), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(520), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(398), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_5_repeat1,
    ACTIONS(2030), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3778] = 6,
    ACTIONS(2131), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      sym__line_ending,
    ACTIONS(2127), 2,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
    STATE(501), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(412), 3,
      sym__newline,
      sym__line,
      aux_sym_code_fence_content_repeat1,
    ACTIONS(2134), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3835] = 7,
    ACTIONS(2140), 1,
      anon_sym_LBRACK,
    ACTIONS(2148), 1,
      sym__soft_line_ending,
    STATE(413), 1,
      aux_sym_paragraph_repeat1,
    ACTIONS(2146), 2,
      sym__line_ending,
      sym__eof,
    STATE(506), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(601), 2,
      sym__soft_line_break,
      sym__line,
    ACTIONS(2143), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [3894] = 9,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(1910), 1,
      sym__line_ending,
    ACTIONS(2153), 1,
      sym__whitespace,
    STATE(379), 1,
      sym_language,
    STATE(557), 1,
      aux_sym_info_string_repeat2,
    ACTIONS(2151), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(1810), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(449), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(1808), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [3957] = 10,
    ACTIONS(2161), 1,
      anon_sym_PIPE,
    ACTIONS(2163), 1,
      sym__whitespace,
    STATE(475), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(715), 1,
      sym_pipe_table_cell,
    STATE(794), 1,
      sym_pipe_table_row,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2165), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4022] = 7,
    ACTIONS(2104), 1,
      anon_sym_LBRACK,
    ACTIONS(2108), 1,
      sym__line_ending,
    ACTIONS(2112), 1,
      sym__html_block_1_end,
    ACTIONS(2167), 1,
      sym__block_close,
    STATE(509), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(420), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_1_repeat1,
    ACTIONS(2106), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4081] = 7,
    ACTIONS(1942), 1,
      anon_sym_DASH_DASH_GT,
    ACTIONS(1944), 1,
      sym__line_ending,
    ACTIONS(2169), 1,
      sym__block_close,
    ACTIONS(1938), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(513), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(388), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_2_repeat1,
    ACTIONS(1940), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4140] = 7,
    ACTIONS(2006), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2008), 1,
      sym__line_ending,
    ACTIONS(2171), 1,
      sym__block_close,
    ACTIONS(2002), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(518), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(421), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_3_repeat1,
    ACTIONS(2004), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4199] = 7,
    ACTIONS(2032), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2034), 1,
      sym__line_ending,
    ACTIONS(2173), 1,
      sym__block_close,
    ACTIONS(2028), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(520), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(422), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_5_repeat1,
    ACTIONS(2030), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4258] = 7,
    ACTIONS(2104), 1,
      anon_sym_LBRACK,
    ACTIONS(2108), 1,
      sym__line_ending,
    ACTIONS(2112), 1,
      sym__html_block_1_end,
    ACTIONS(2175), 1,
      sym__block_close,
    STATE(509), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(397), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_1_repeat1,
    ACTIONS(2106), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4317] = 7,
    ACTIONS(2006), 1,
      anon_sym_QMARK_GT,
    ACTIONS(2008), 1,
      sym__line_ending,
    ACTIONS(2177), 1,
      sym__block_close,
    ACTIONS(2002), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(518), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(400), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_3_repeat1,
    ACTIONS(2004), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4376] = 7,
    ACTIONS(2032), 1,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2034), 1,
      sym__line_ending,
    ACTIONS(2179), 1,
      sym__block_close,
    ACTIONS(2028), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(520), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(401), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_5_repeat1,
    ACTIONS(2030), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4435] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    ACTIONS(2185), 1,
      anon_sym_DQUOTE,
    STATE(525), 1,
      sym__soft_line_break,
    ACTIONS(2183), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(390), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat1,
    ACTIONS(2181), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4494] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    ACTIONS(2185), 1,
      anon_sym_SQUOTE,
    STATE(523), 1,
      sym__soft_line_break,
    ACTIONS(2189), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(391), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat2,
    ACTIONS(2187), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4553] = 11,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    ACTIONS(2191), 1,
      sym__whitespace,
    STATE(458), 1,
      sym__soft_line_break,
    STATE(681), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [4620] = 7,
    ACTIONS(2104), 1,
      anon_sym_LBRACK,
    ACTIONS(2108), 1,
      sym__line_ending,
    ACTIONS(2112), 1,
      sym__html_block_1_end,
    ACTIONS(2193), 1,
      sym__block_close,
    STATE(509), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(397), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_1_repeat1,
    ACTIONS(2106), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4679] = 6,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2201), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(437), 3,
      sym__newline,
      sym__line,
      aux_sym__indented_chunk_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4735] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2205), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(430), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4793] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2207), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(436), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4851] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2209), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(444), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4909] = 5,
    STATE(463), 1,
      sym__last_token_punctuation,
    ACTIONS(2213), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(462), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2215), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2211), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [4963] = 5,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    ACTIONS(2217), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(404), 5,
      sym_backslash_escape,
      sym__text_inline_no_link,
      sym__soft_line_break,
      sym__word,
      aux_sym_link_label_repeat1,
    ACTIONS(1793), 34,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5017] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2219), 1,
      anon_sym_GT,
    ACTIONS(2221), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(439), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_4_repeat1,
    ACTIONS(2197), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5075] = 6,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2223), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(442), 3,
      sym__newline,
      sym__line,
      aux_sym__indented_chunk_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5131] = 7,
    ACTIONS(11), 1,
      sym__soft_line_ending,
    ACTIONS(1864), 1,
      anon_sym_LBRACK,
    STATE(26), 1,
      sym_paragraph,
    STATE(373), 1,
      aux_sym_paragraph_repeat1,
    STATE(506), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(601), 2,
      sym__soft_line_break,
      sym__line,
    ACTIONS(7), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5189] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2225), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(444), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5247] = 6,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2227), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(456), 3,
      sym__newline,
      sym__line,
      aux_sym__indented_chunk_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5303] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2219), 1,
      anon_sym_GT,
    ACTIONS(2229), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(445), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_4_repeat1,
    ACTIONS(2197), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5361] = 7,
    ACTIONS(2231), 1,
      anon_sym_LBRACK,
    ACTIONS(2237), 1,
      anon_sym_GT,
    ACTIONS(2240), 1,
      sym__line_ending,
    ACTIONS(2243), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(439), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_4_repeat1,
    ACTIONS(2234), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5419] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2245), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(447), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5477] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2247), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(448), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5535] = 6,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2249), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(456), 3,
      sym__newline,
      sym__line,
      aux_sym__indented_chunk_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5591] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1964), 1,
      anon_sym_RPAREN,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    STATE(530), 1,
      sym__soft_line_break,
    ACTIONS(2253), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(446), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(2251), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5649] = 7,
    ACTIONS(2255), 1,
      anon_sym_LBRACK,
    ACTIONS(2261), 1,
      sym__line_ending,
    ACTIONS(2264), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(444), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2258), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5707] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2219), 1,
      anon_sym_GT,
    ACTIONS(2266), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(439), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_4_repeat1,
    ACTIONS(2197), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5765] = 7,
    ACTIONS(2268), 1,
      sym__backslash_escape,
    ACTIONS(2277), 1,
      anon_sym_RPAREN,
    ACTIONS(2279), 1,
      sym__soft_line_ending,
    STATE(530), 1,
      sym__soft_line_break,
    ACTIONS(2274), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(446), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(2271), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5823] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2282), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(444), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5881] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2203), 1,
      sym__line_ending,
    ACTIONS(2284), 1,
      sym__block_close,
    STATE(561), 1,
      sym__newline,
    STATE(444), 2,
      sym__line,
      aux_sym__html_block_6_repeat1,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2197), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [5939] = 5,
    ACTIONS(1806), 1,
      sym__backslash_escape,
    ACTIONS(2288), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(451), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(2290), 5,
      sym__line_ending,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      sym__whitespace,
    ACTIONS(2286), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [5993] = 7,
    ACTIONS(2195), 1,
      anon_sym_LBRACK,
    ACTIONS(2199), 1,
      sym__line_ending,
    ACTIONS(2219), 1,
      anon_sym_GT,
    ACTIONS(2292), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(433), 3,
      sym__newline,
      sym__line,
      aux_sym__html_block_4_repeat1,
    ACTIONS(2197), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6051] = 5,
    ACTIONS(2294), 1,
      sym__backslash_escape,
    ACTIONS(2300), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(451), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_language_repeat1,
    ACTIONS(2303), 5,
      sym__line_ending,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      sym__whitespace,
    ACTIONS(2297), 31,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [6105] = 7,
    ACTIONS(1791), 1,
      sym__backslash_escape,
    ACTIONS(1966), 1,
      sym__soft_line_ending,
    ACTIONS(2185), 1,
      anon_sym_RPAREN,
    STATE(530), 1,
      sym__soft_line_break,
    ACTIONS(2307), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(443), 3,
      sym_backslash_escape,
      sym__word,
      aux_sym_link_title_repeat3,
    ACTIONS(2305), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6163] = 10,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    STATE(484), 1,
      sym__soft_line_break,
    STATE(667), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [6227] = 10,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    STATE(481), 1,
      sym__soft_line_break,
    STATE(669), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [6291] = 10,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    STATE(493), 1,
      sym__soft_line_break,
    STATE(680), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [6355] = 6,
    ACTIONS(2309), 1,
      anon_sym_LBRACK,
    ACTIONS(2315), 1,
      sym__line_ending,
    ACTIONS(2318), 1,
      sym__block_close,
    STATE(532), 2,
      sym__word,
      aux_sym__line_repeat1,
    STATE(456), 3,
      sym__newline,
      sym__line,
      aux_sym__indented_chunk_repeat1,
    ACTIONS(2312), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6411] = 10,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(1958), 1,
      sym__soft_line_ending,
    STATE(494), 1,
      sym__soft_line_break,
    STATE(685), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [6475] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2320), 1,
      sym__whitespace,
    STATE(680), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [6536] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(260), 1,
      sym__newline,
    STATE(796), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [6595] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(310), 1,
      sym__newline,
    STATE(838), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [6654] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(295), 1,
      sym__newline,
    STATE(810), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [6713] = 4,
    ACTIONS(2332), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(464), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2334), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2330), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6764] = 4,
    ACTIONS(2338), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(465), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2334), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2336), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6815] = 4,
    ACTIONS(2343), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(464), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2346), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2340), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6866] = 4,
    ACTIONS(2332), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(464), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2348), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2330), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6917] = 4,
    STATE(467), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2352), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2350), 4,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
    ACTIONS(2354), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [6968] = 4,
    STATE(467), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2358), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2356), 4,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
    ACTIONS(2361), 33,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [7019] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(266), 1,
      sym__newline,
    STATE(823), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7078] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(213), 1,
      sym__newline,
    STATE(798), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7137] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(214), 1,
      sym__newline,
    STATE(799), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7196] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(215), 1,
      sym__newline,
    STATE(800), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7255] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(216), 1,
      sym__newline,
    STATE(801), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7314] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(217), 1,
      sym__newline,
    STATE(802), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7373] = 8,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    STATE(218), 1,
      sym__newline,
    STATE(803), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7432] = 8,
    ACTIONS(2364), 1,
      sym__whitespace,
    STATE(479), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(700), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2366), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7491] = 5,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2374), 1,
      anon_sym_GT,
    ACTIONS(2372), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(488), 4,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(2370), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [7544] = 8,
    ACTIONS(2376), 1,
      sym__whitespace,
    STATE(479), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(712), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2378), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7603] = 8,
    ACTIONS(2380), 1,
      sym__whitespace,
    STATE(483), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(712), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2382), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7662] = 8,
    ACTIONS(2393), 1,
      sym__whitespace,
    STATE(479), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(579), 1,
      sym__word,
    STATE(855), 1,
      sym_pipe_table_cell,
    ACTIONS(2387), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2396), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2384), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2390), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7721] = 6,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2402), 1,
      anon_sym_LPAREN,
    ACTIONS(2404), 1,
      anon_sym_RPAREN,
    ACTIONS(2400), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(489), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(2398), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [7776] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2406), 1,
      sym__whitespace,
    STATE(682), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [7837] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2408), 1,
      sym__whitespace,
    STATE(667), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [7898] = 8,
    ACTIONS(2410), 1,
      sym__whitespace,
    STATE(479), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(719), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [7957] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2414), 1,
      sym__whitespace,
    STATE(669), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [8018] = 6,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2402), 1,
      anon_sym_LPAREN,
    ACTIONS(2420), 1,
      anon_sym_RPAREN,
    ACTIONS(2418), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(486), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(2416), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8073] = 6,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2402), 1,
      anon_sym_LPAREN,
    ACTIONS(2426), 1,
      anon_sym_RPAREN,
    ACTIONS(2424), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(491), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(2422), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8128] = 5,
    ACTIONS(2428), 1,
      sym__backslash_escape,
    ACTIONS(2437), 1,
      anon_sym_GT,
    ACTIONS(2434), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(487), 4,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(2431), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8181] = 5,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2443), 1,
      anon_sym_GT,
    ACTIONS(2441), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(487), 4,
      sym_backslash_escape,
      sym__text_no_angle,
      sym__word,
      aux_sym_link_destination_repeat1,
    ACTIONS(2439), 33,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8234] = 6,
    ACTIONS(2368), 1,
      sym__backslash_escape,
    ACTIONS(2402), 1,
      anon_sym_LPAREN,
    ACTIONS(2445), 1,
      anon_sym_RPAREN,
    ACTIONS(2424), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(491), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(2422), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8289] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(262), 1,
      sym__newline,
    STATE(813), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8348] = 6,
    ACTIONS(1856), 1,
      anon_sym_RPAREN,
    ACTIONS(2447), 1,
      sym__backslash_escape,
    ACTIONS(2456), 1,
      anon_sym_LPAREN,
    ACTIONS(2453), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(491), 4,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
      aux_sym_link_destination_repeat2,
    ACTIONS(2450), 32,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8403] = 8,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    ACTIONS(2326), 1,
      sym__whitespace,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(293), 1,
      sym__newline,
    STATE(808), 1,
      sym__atx_heading_content,
    STATE(914), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [8462] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2459), 1,
      sym__whitespace,
    STATE(685), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [8523] = 9,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    ACTIONS(2461), 1,
      sym__whitespace,
    STATE(688), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [8584] = 4,
    ACTIONS(2213), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(462), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2215), 4,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
    ACTIONS(2211), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8635] = 2,
    ACTIONS(2463), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1856), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8681] = 2,
    ACTIONS(2467), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2465), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8727] = 7,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(711), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2378), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [8783] = 4,
    ACTIONS(2471), 1,
      anon_sym_LBRACK,
    STATE(500), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
    ACTIONS(2473), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8833] = 4,
    ACTIONS(2475), 1,
      anon_sym_LBRACK,
    STATE(500), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
    ACTIONS(2478), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8883] = 4,
    ACTIONS(2481), 1,
      anon_sym_LBRACK,
    STATE(505), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
    ACTIONS(2483), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8933] = 3,
    ACTIONS(2485), 1,
      sym_block_continuation,
    ACTIONS(1242), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1240), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [8981] = 3,
    ACTIONS(2491), 1,
      sym_block_continuation,
    ACTIONS(2489), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2487), 38,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9029] = 7,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(718), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [9085] = 4,
    ACTIONS(2493), 1,
      anon_sym_LBRACK,
    STATE(505), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
    ACTIONS(2496), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9135] = 4,
    ACTIONS(2499), 1,
      anon_sym_LBRACK,
    STATE(507), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
    ACTIONS(2501), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9185] = 4,
    ACTIONS(2503), 1,
      anon_sym_LBRACK,
    STATE(507), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
    ACTIONS(2506), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9235] = 5,
    STATE(540), 1,
      sym__last_token_punctuation,
    ACTIONS(2215), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2511), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(536), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2509), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9287] = 4,
    ACTIONS(2513), 1,
      anon_sym_LBRACK,
    STATE(529), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
    ACTIONS(2515), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9337] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(682), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [9395] = 9,
    ACTIONS(2523), 1,
      anon_sym_PIPE,
    ACTIONS(2525), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(555), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(789), 1,
      sym_pipe_table_cell,
    STATE(820), 1,
      sym_pipe_table_row,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [9455] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(670), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [9513] = 4,
    ACTIONS(2527), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(515), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2529), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9563] = 2,
    ACTIONS(2533), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2531), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9609] = 4,
    ACTIONS(2535), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    STATE(515), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_DASH_DASH_GT,
    ACTIONS(2538), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9659] = 2,
    ACTIONS(2543), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2541), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9705] = 7,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(734), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2545), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [9761] = 4,
    ACTIONS(2547), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(519), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_QMARK_GT,
    ACTIONS(2549), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9811] = 4,
    ACTIONS(2551), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    STATE(519), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_QMARK_GT,
    ACTIONS(2554), 34,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9861] = 4,
    ACTIONS(2557), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(522), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2350), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2559), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [9911] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(669), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [9969] = 4,
    ACTIONS(2561), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    STATE(522), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK_RBRACK_GT,
    ACTIONS(2564), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10019] = 4,
    ACTIONS(2571), 1,
      sym__soft_line_ending,
    STATE(872), 1,
      sym__soft_line_break,
    ACTIONS(2569), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2567), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10069] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(685), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [10127] = 4,
    ACTIONS(2578), 1,
      sym__soft_line_ending,
    STATE(907), 1,
      sym__soft_line_break,
    ACTIONS(2576), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2574), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10177] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(688), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [10235] = 8,
    ACTIONS(1826), 1,
      sym__backslash_escape,
    ACTIONS(1832), 1,
      anon_sym_LPAREN,
    ACTIONS(1954), 1,
      anon_sym_LT,
    STATE(689), 1,
      sym_link_destination,
    ACTIONS(1950), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    STATE(386), 3,
      sym_backslash_escape,
      sym__link_destination_parenthesis,
      sym__word,
    ACTIONS(1948), 5,
      sym_entity_reference,
      sym_numeric_character_reference,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(1952), 26,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
  [10293] = 9,
    ACTIONS(2523), 1,
      anon_sym_PIPE,
    ACTIONS(2525), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(555), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(789), 1,
      sym_pipe_table_cell,
    STATE(857), 1,
      sym_pipe_table_row,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [10353] = 4,
    ACTIONS(2581), 1,
      anon_sym_LBRACK,
    STATE(529), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2356), 3,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
    ACTIONS(2584), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10403] = 4,
    ACTIONS(2591), 1,
      sym__soft_line_ending,
    STATE(880), 1,
      sym__soft_line_break,
    ACTIONS(2589), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2587), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10452] = 2,
    ACTIONS(2594), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1997), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10497] = 4,
    ACTIONS(2596), 1,
      anon_sym_LBRACK,
    ACTIONS(2350), 2,
      sym__line_ending,
      sym__block_close,
    STATE(558), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2598), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10546] = 2,
    ACTIONS(1472), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1470), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10591] = 4,
    ACTIONS(2215), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2511), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(536), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2509), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10640] = 2,
    ACTIONS(2467), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2465), 37,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10685] = 4,
    ACTIONS(2334), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2602), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(563), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2600), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10734] = 3,
    ACTIONS(1242), 1,
      anon_sym_LBRACK,
    ACTIONS(2604), 1,
      sym_block_continuation,
    ACTIONS(1240), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10781] = 3,
    ACTIONS(1129), 1,
      anon_sym_LBRACK,
    ACTIONS(2606), 1,
      sym_block_continuation,
    ACTIONS(1127), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10828] = 2,
    ACTIONS(2467), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2465), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10873] = 4,
    ACTIONS(2334), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2610), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(564), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2608), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [10922] = 8,
    ACTIONS(2378), 1,
      sym__line_ending,
    ACTIONS(2612), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(549), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(780), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [10979] = 8,
    ACTIONS(2382), 1,
      sym__line_ending,
    ACTIONS(2614), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(565), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(780), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [11036] = 2,
    ACTIONS(2616), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2346), 38,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11081] = 3,
    ACTIONS(1129), 1,
      anon_sym_LBRACK,
    ACTIONS(2618), 1,
      sym_block_continuation,
    ACTIONS(1127), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11128] = 5,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    STATE(593), 1,
      sym__last_token_punctuation,
    ACTIONS(2622), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(592), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2620), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11179] = 3,
    ACTIONS(1129), 1,
      anon_sym_LBRACK,
    ACTIONS(2624), 1,
      sym_block_continuation,
    ACTIONS(1127), 38,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11226] = 3,
    ACTIONS(2626), 1,
      sym_block_continuation,
    ACTIONS(1129), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(1127), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DASH_DASH_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11273] = 3,
    ACTIONS(2628), 1,
      sym_block_continuation,
    ACTIONS(1129), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    ACTIONS(1127), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_QMARK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11320] = 8,
    ACTIONS(2396), 1,
      sym__line_ending,
    ACTIONS(2630), 1,
      sym__whitespace,
    STATE(549), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(579), 1,
      sym__word,
    STATE(806), 1,
      sym_pipe_table_cell,
    ACTIONS(2387), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2384), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2390), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [11377] = 3,
    ACTIONS(2633), 1,
      sym_block_continuation,
    ACTIONS(1129), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(1127), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACK_RBRACK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11424] = 2,
    ACTIONS(2635), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2356), 37,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11469] = 2,
    ACTIONS(2639), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2637), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11514] = 3,
    ACTIONS(1129), 1,
      anon_sym_LBRACK,
    ACTIONS(2641), 1,
      sym_block_continuation,
    ACTIONS(1127), 38,
      sym__line_ending,
      sym__block_close,
      sym__blank_line_start,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11561] = 2,
    ACTIONS(2645), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2643), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11606] = 8,
    ACTIONS(2366), 1,
      sym__line_ending,
    ACTIONS(2647), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(549), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(776), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [11663] = 4,
    ACTIONS(1795), 1,
      anon_sym_LBRACK,
    ACTIONS(2649), 1,
      sym__close_block,
    STATE(619), 1,
      sym__last_token_punctuation,
    ACTIONS(1789), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11712] = 4,
    STATE(557), 1,
      aux_sym_info_string_repeat2,
    ACTIONS(2655), 2,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
    ACTIONS(2653), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2651), 34,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_PIPE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11761] = 4,
    ACTIONS(2658), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 2,
      sym__line_ending,
      sym__block_close,
    STATE(558), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2661), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11810] = 5,
    ACTIONS(1789), 1,
      sym__line_ending,
    ACTIONS(2322), 1,
      anon_sym_LBRACK,
    STATE(913), 1,
      sym__line,
    STATE(583), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2324), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11861] = 2,
    ACTIONS(2664), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2303), 37,
      sym__line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11906] = 4,
    ACTIONS(1502), 1,
      sym__blank_line_start,
    ACTIONS(2666), 1,
      anon_sym_LBRACK,
    STATE(878), 1,
      sym__blank_line,
    ACTIONS(2668), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [11955] = 2,
    ACTIONS(2670), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1983), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12000] = 4,
    ACTIONS(2346), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2675), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(563), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2672), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12049] = 4,
    ACTIONS(2348), 2,
      sym__line_ending,
      anon_sym_PIPE,
    ACTIONS(2602), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(563), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2600), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12098] = 8,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(2678), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(549), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(791), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [12155] = 2,
    ACTIONS(2682), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2680), 38,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12200] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 38,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12244] = 7,
    ACTIONS(2545), 1,
      sym__line_ending,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(787), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [12298] = 2,
    ACTIONS(2635), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2356), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACK_RBRACK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12342] = 4,
    ACTIONS(2356), 1,
      sym__line_ending,
    ACTIONS(2686), 1,
      anon_sym_LBRACK,
    STATE(570), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2689), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12390] = 2,
    ACTIONS(1255), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    ACTIONS(1253), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_QMARK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12434] = 2,
    ACTIONS(2694), 2,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2692), 37,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12478] = 2,
    ACTIONS(1255), 1,
      anon_sym_LBRACK,
    ACTIONS(1253), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12522] = 2,
    ACTIONS(2698), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2696), 37,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12566] = 2,
    ACTIONS(1255), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(1253), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACK_RBRACK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12610] = 2,
    ACTIONS(2700), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2277), 36,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12654] = 3,
    ACTIONS(2702), 1,
      sym_block_continuation,
    ACTIONS(1242), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1240), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12700] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_backtick,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12744] = 4,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2622), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(592), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2620), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12792] = 2,
    ACTIONS(1255), 1,
      anon_sym_LBRACK,
    ACTIONS(1253), 38,
      sym__line_ending,
      sym__block_close,
      sym__blank_line_start,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12836] = 2,
    ACTIONS(2706), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2704), 36,
      sym__soft_line_ending,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12880] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12924] = 4,
    ACTIONS(2350), 1,
      sym__line_ending,
    ACTIONS(2708), 1,
      anon_sym_LBRACK,
    STATE(570), 2,
      sym__word,
      aux_sym__line_repeat1,
    ACTIONS(2710), 35,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [12972] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13016] = 2,
    ACTIONS(2467), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2465), 36,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13060] = 2,
    ACTIONS(2714), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2712), 37,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13104] = 7,
    ACTIONS(2378), 1,
      sym__line_ending,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(783), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [13158] = 3,
    ACTIONS(1129), 1,
      anon_sym_LBRACK,
    ACTIONS(2716), 1,
      sym_block_continuation,
    ACTIONS(1127), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13204] = 2,
    ACTIONS(2718), 1,
      anon_sym_LBRACK,
    ACTIONS(2023), 38,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13248] = 2,
    ACTIONS(1255), 1,
      anon_sym_LBRACK,
    ACTIONS(1253), 38,
      sym__line_ending,
      sym__block_close,
      sym__fenced_code_block_end_tilde,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13292] = 2,
    ACTIONS(2720), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2050), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DASH_DASH_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13336] = 4,
    ACTIONS(2334), 1,
      anon_sym_PIPE,
    ACTIONS(2724), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(606), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2722), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13384] = 4,
    ACTIONS(2334), 1,
      anon_sym_PIPE,
    ACTIONS(2728), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(607), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2726), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13432] = 7,
    ACTIONS(2730), 1,
      sym__whitespace,
    STATE(477), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(495), 1,
      sym__word,
    STATE(700), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [13486] = 7,
    ACTIONS(2732), 1,
      sym__whitespace,
    STATE(534), 1,
      sym__word,
    STATE(541), 1,
      aux_sym_pipe_table_row_repeat1,
    STATE(776), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [13540] = 2,
    ACTIONS(2734), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2396), 37,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13584] = 2,
    ACTIONS(2736), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    ACTIONS(2064), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_QMARK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13628] = 2,
    ACTIONS(2635), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(2356), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DASH_DASH_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13672] = 3,
    ACTIONS(2738), 1,
      anon_sym_LBRACK,
    ACTIONS(2742), 1,
      sym__whitespace,
    ACTIONS(2740), 37,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [13718] = 3,
    ACTIONS(2738), 1,
      anon_sym_LBRACK,
    ACTIONS(2744), 1,
      sym__whitespace,
    ACTIONS(2740), 37,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [13764] = 2,
    ACTIONS(2746), 1,
      anon_sym_LBRACK,
    ACTIONS(2748), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13808] = 2,
    ACTIONS(1255), 1,
      anon_sym_LBRACK,
    ACTIONS(1253), 38,
      sym__line_ending,
      sym__block_close,
      sym__html_block_1_end,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13852] = 2,
    ACTIONS(2750), 2,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
    ACTIONS(2078), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACK_RBRACK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13896] = 7,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(785), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [13950] = 2,
    ACTIONS(2635), 2,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    ACTIONS(2356), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_QMARK_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [13994] = 4,
    ACTIONS(2346), 1,
      anon_sym_PIPE,
    ACTIONS(2755), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(606), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2752), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14042] = 4,
    ACTIONS(2348), 1,
      anon_sym_PIPE,
    ACTIONS(2724), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    STATE(606), 2,
      sym__word,
      aux_sym_pipe_table_cell_repeat1,
    ACTIONS(2722), 34,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14090] = 2,
    ACTIONS(2758), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2382), 37,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14134] = 2,
    ACTIONS(1255), 2,
      anon_sym_LBRACK,
      anon_sym_DASH,
    ACTIONS(1253), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_DASH_DASH_GT,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14178] = 2,
    ACTIONS(1472), 1,
      anon_sym_LBRACK,
    ACTIONS(1470), 38,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14222] = 6,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(718), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14273] = 2,
    ACTIONS(2533), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2531), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [14316] = 6,
    ACTIONS(2760), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(795), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14367] = 6,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(579), 1,
      sym__word,
    STATE(830), 1,
      sym_pipe_table_cell,
    ACTIONS(2764), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2762), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2766), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14418] = 2,
    ACTIONS(2543), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2541), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [14461] = 6,
    ACTIONS(2768), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(701), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14512] = 6,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    STATE(495), 1,
      sym__word,
    STATE(711), 1,
      sym_pipe_table_cell,
    ACTIONS(2157), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2155), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2159), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14563] = 2,
    ACTIONS(1472), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1470), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14606] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14649] = 6,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(579), 1,
      sym__word,
    STATE(824), 1,
      sym_pipe_table_cell,
    ACTIONS(2764), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2762), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2766), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14700] = 2,
    ACTIONS(1255), 1,
      anon_sym_LBRACK,
    ACTIONS(1253), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14743] = 2,
    ACTIONS(2770), 1,
      anon_sym_LBRACK,
    ACTIONS(2772), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14786] = 6,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(783), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [14837] = 2,
    ACTIONS(2776), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(2774), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14880] = 2,
    ACTIONS(2616), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2346), 36,
      sym__line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14923] = 2,
    ACTIONS(2778), 1,
      anon_sym_LBRACK,
    ACTIONS(2243), 37,
      sym__line_ending,
      sym__block_close,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [14966] = 6,
    ACTIONS(2684), 1,
      anon_sym_PIPE,
    STATE(534), 1,
      sym__word,
    STATE(785), 1,
      sym_pipe_table_cell,
    ACTIONS(2519), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2517), 4,
      sym__backslash_escape,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
    ACTIONS(2521), 29,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [15017] = 2,
    ACTIONS(2463), 3,
      anon_sym_LBRACK,
      anon_sym_AMP,
      anon_sym_BSLASH,
    ACTIONS(1856), 35,
      sym__backslash_escape,
      sym_entity_reference,
      sym_numeric_character_reference,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
  [15060] = 2,
    ACTIONS(2758), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2382), 35,
      sym__line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15102] = 2,
    ACTIONS(2734), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2396), 35,
      sym__line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15144] = 2,
    ACTIONS(2635), 1,
      anon_sym_LBRACK,
    ACTIONS(2356), 36,
      sym__line_ending,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_BSLASH,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15186] = 2,
    ACTIONS(2698), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2696), 35,
      sym__line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15228] = 2,
    ACTIONS(2616), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2346), 35,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15270] = 2,
    ACTIONS(2714), 2,
      anon_sym_LBRACK,
      anon_sym_BSLASH,
    ACTIONS(2712), 35,
      sym__line_ending,
      sym__backslash_escape,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      anon_sym_POUND,
      anon_sym_DOLLAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_CARET,
      anon_sym__,
      anon_sym_BQUOTE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      aux_sym__word_token1,
      aux_sym__word_token2,
      aux_sym__word_token3,
      sym__whitespace,
  [15312] = 15,
    ACTIONS(17), 1,
      sym_atx_h1_marker,
    ACTIONS(19), 1,
      sym_atx_h2_marker,
    ACTIONS(21), 1,
      sym_atx_h3_marker,
    ACTIONS(23), 1,
      sym_atx_h4_marker,
    ACTIONS(25), 1,
      sym_atx_h5_marker,
    ACTIONS(27), 1,
      sym_atx_h6_marker,
    ACTIONS(246), 1,
      ts_builtin_sym_end,
    STATE(38), 1,
      sym__atx_heading1,
    STATE(45), 1,
      sym__atx_heading2,
    STATE(54), 1,
      sym__atx_heading3,
    STATE(55), 1,
      sym__atx_heading4,
    STATE(66), 1,
      sym__atx_heading5,
    STATE(72), 1,
      sym__atx_heading6,
    STATE(639), 2,
      sym_section,
      aux_sym_document_repeat2,
    STATE(661), 6,
      sym__section1,
      sym__section2,
      sym__section3,
      sym__section4,
      sym__section5,
      sym__section6,
  [15364] = 15,
    ACTIONS(17), 1,
      sym_atx_h1_marker,
    ACTIONS(19), 1,
      sym_atx_h2_marker,
    ACTIONS(21), 1,
      sym_atx_h3_marker,
    ACTIONS(23), 1,
      sym_atx_h4_marker,
    ACTIONS(25), 1,
      sym_atx_h5_marker,
    ACTIONS(27), 1,
      sym_atx_h6_marker,
    ACTIONS(2780), 1,
      ts_builtin_sym_end,
    STATE(38), 1,
      sym__atx_heading1,
    STATE(45), 1,
      sym__atx_heading2,
    STATE(54), 1,
      sym__atx_heading3,
    STATE(55), 1,
      sym__atx_heading4,
    STATE(66), 1,
      sym__atx_heading5,
    STATE(72), 1,
      sym__atx_heading6,
    STATE(639), 2,
      sym_section,
      aux_sym_document_repeat2,
    STATE(661), 6,
      sym__section1,
      sym__section2,
      sym__section3,
      sym__section4,
      sym__section5,
      sym__section6,
  [15416] = 15,
    ACTIONS(17), 1,
      sym_atx_h1_marker,
    ACTIONS(19), 1,
      sym_atx_h2_marker,
    ACTIONS(21), 1,
      sym_atx_h3_marker,
    ACTIONS(23), 1,
      sym_atx_h4_marker,
    ACTIONS(25), 1,
      sym_atx_h5_marker,
    ACTIONS(27), 1,
      sym_atx_h6_marker,
    ACTIONS(2782), 1,
      ts_builtin_sym_end,
    STATE(38), 1,
      sym__atx_heading1,
    STATE(45), 1,
      sym__atx_heading2,
    STATE(54), 1,
      sym__atx_heading3,
    STATE(55), 1,
      sym__atx_heading4,
    STATE(66), 1,
      sym__atx_heading5,
    STATE(72), 1,
      sym__atx_heading6,
    STATE(639), 2,
      sym_section,
      aux_sym_document_repeat2,
    STATE(661), 6,
      sym__section1,
      sym__section2,
      sym__section3,
      sym__section4,
      sym__section5,
      sym__section6,
  [15468] = 15,
    ACTIONS(17), 1,
      sym_atx_h1_marker,
    ACTIONS(19), 1,
      sym_atx_h2_marker,
    ACTIONS(21), 1,
      sym_atx_h3_marker,
    ACTIONS(23), 1,
      sym_atx_h4_marker,
    ACTIONS(25), 1,
      sym_atx_h5_marker,
    ACTIONS(27), 1,
      sym_atx_h6_marker,
    ACTIONS(2784), 1,
      ts_builtin_sym_end,
    STATE(38), 1,
      sym__atx_heading1,
    STATE(45), 1,
      sym__atx_heading2,
    STATE(54), 1,
      sym__atx_heading3,
    STATE(55), 1,
      sym__atx_heading4,
    STATE(66), 1,
      sym__atx_heading5,
    STATE(72), 1,
      sym__atx_heading6,
    STATE(639), 2,
      sym_section,
      aux_sym_document_repeat2,
    STATE(661), 6,
      sym__section1,
      sym__section2,
      sym__section3,
      sym__section4,
      sym__section5,
      sym__section6,
  [15520] = 15,
    ACTIONS(2786), 1,
      ts_builtin_sym_end,
    ACTIONS(2788), 1,
      sym_atx_h1_marker,
    ACTIONS(2791), 1,
      sym_atx_h2_marker,
    ACTIONS(2794), 1,
      sym_atx_h3_marker,
    ACTIONS(2797), 1,
      sym_atx_h4_marker,
    ACTIONS(2800), 1,
      sym_atx_h5_marker,
    ACTIONS(2803), 1,
      sym_atx_h6_marker,
    STATE(38), 1,
      sym__atx_heading1,
    STATE(45), 1,
      sym__atx_heading2,
    STATE(54), 1,
      sym__atx_heading3,
    STATE(55), 1,
      sym__atx_heading4,
    STATE(66), 1,
      sym__atx_heading5,
    STATE(72), 1,
      sym__atx_heading6,
    STATE(639), 2,
      sym_section,
      aux_sym_document_repeat2,
    STATE(661), 6,
      sym__section1,
      sym__section2,
      sym__section3,
      sym__section4,
      sym__section5,
      sym__section6,
  [15572] = 7,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    ACTIONS(2810), 1,
      sym__whitespace,
    STATE(641), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(713), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2812), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15596] = 7,
    ACTIONS(2814), 1,
      anon_sym_DASH,
    ACTIONS(2817), 1,
      anon_sym_COLON,
    ACTIONS(2820), 1,
      sym__whitespace,
    STATE(641), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(804), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2823), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15620] = 7,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    ACTIONS(2825), 1,
      sym__whitespace,
    STATE(641), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(709), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2827), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15644] = 7,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    ACTIONS(2829), 1,
      sym__whitespace,
    STATE(641), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(695), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2831), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15668] = 3,
    ACTIONS(2833), 1,
      anon_sym_DASH,
    STATE(644), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2836), 6,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_COLON,
      anon_sym_PIPE,
      sym__whitespace,
  [15683] = 4,
    ACTIONS(2838), 1,
      anon_sym_DASH,
    ACTIONS(2840), 1,
      anon_sym_COLON,
    STATE(644), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2842), 5,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
      sym__whitespace,
  [15700] = 8,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    ACTIONS(2848), 1,
      anon_sym_PIPE,
    ACTIONS(2850), 1,
      sym__whitespace,
    STATE(642), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(675), 1,
      sym_pipe_table_delimiter_row,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(804), 1,
      sym_pipe_table_delimiter_cell,
  [15725] = 8,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    ACTIONS(2848), 1,
      anon_sym_PIPE,
    ACTIONS(2850), 1,
      sym__whitespace,
    STATE(642), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(678), 1,
      sym_pipe_table_delimiter_row,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(804), 1,
      sym_pipe_table_delimiter_cell,
  [15750] = 4,
    ACTIONS(2838), 1,
      anon_sym_DASH,
    ACTIONS(2852), 1,
      anon_sym_COLON,
    STATE(644), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2854), 5,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
      sym__whitespace,
  [15767] = 7,
    ACTIONS(1238), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(691), 1,
      sym__soft_line_break,
    STATE(724), 1,
      sym_link_title,
  [15789] = 7,
    ACTIONS(1282), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(663), 1,
      sym__soft_line_break,
    STATE(706), 1,
      sym_link_title,
  [15811] = 5,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(705), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2864), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15829] = 5,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(696), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2812), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [15847] = 7,
    ACTIONS(1277), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(666), 1,
      sym__soft_line_break,
    STATE(697), 1,
      sym_link_title,
  [15869] = 7,
    ACTIONS(1251), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(673), 1,
      sym__soft_line_break,
    STATE(743), 1,
      sym_link_title,
  [15891] = 7,
    ACTIONS(1166), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(683), 1,
      sym__soft_line_break,
    STATE(720), 1,
      sym_link_title,
  [15913] = 7,
    ACTIONS(1161), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(671), 1,
      sym__soft_line_break,
    STATE(731), 1,
      sym_link_title,
  [15935] = 7,
    ACTIONS(1184), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(686), 1,
      sym__soft_line_break,
    STATE(721), 1,
      sym_link_title,
  [15957] = 7,
    ACTIONS(1202), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(687), 1,
      sym__soft_line_break,
    STATE(722), 1,
      sym_link_title,
  [15979] = 7,
    ACTIONS(1272), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(665), 1,
      sym__soft_line_break,
    STATE(703), 1,
      sym_link_title,
  [16001] = 7,
    ACTIONS(1220), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2862), 1,
      sym__soft_line_ending,
    STATE(690), 1,
      sym__soft_line_break,
    STATE(723), 1,
      sym_link_title,
  [16023] = 1,
    ACTIONS(1576), 7,
      sym_atx_h1_marker,
      sym_atx_h2_marker,
      sym_atx_h3_marker,
      sym_atx_h4_marker,
      sym_atx_h5_marker,
      sym_atx_h6_marker,
      ts_builtin_sym_end,
  [16033] = 5,
    ACTIONS(2806), 1,
      anon_sym_DASH,
    ACTIONS(2808), 1,
      anon_sym_COLON,
    STATE(645), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(739), 1,
      sym_pipe_table_delimiter_cell,
    ACTIONS(2831), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16051] = 6,
    ACTIONS(1251), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2866), 1,
      sym__whitespace,
    STATE(743), 1,
      sym_link_title,
  [16070] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2868), 1,
      sym__whitespace,
    ACTIONS(2870), 1,
      sym__soft_line_ending,
    ACTIONS(2872), 1,
      sym__eof,
    STATE(101), 1,
      sym__soft_line_break,
    STATE(291), 1,
      sym__newline,
  [16089] = 6,
    ACTIONS(1277), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2874), 1,
      sym__whitespace,
    STATE(697), 1,
      sym_link_title,
  [16108] = 6,
    ACTIONS(1282), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2876), 1,
      sym__whitespace,
    STATE(706), 1,
      sym_link_title,
  [16127] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2870), 1,
      sym__soft_line_ending,
    ACTIONS(2878), 1,
      sym__whitespace,
    ACTIONS(2880), 1,
      sym__eof,
    STATE(115), 1,
      sym__soft_line_break,
    STATE(336), 1,
      sym__newline,
  [16146] = 2,
    ACTIONS(2882), 1,
      sym_block_continuation,
    ACTIONS(1240), 5,
      sym__no_indented_chunk,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_LPAREN,
      sym__whitespace,
  [16157] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2870), 1,
      sym__soft_line_ending,
    ACTIONS(2884), 1,
      sym__whitespace,
    ACTIONS(2886), 1,
      sym__eof,
    STATE(116), 1,
      sym__soft_line_break,
    STATE(345), 1,
      sym__newline,
  [16176] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2870), 1,
      sym__soft_line_ending,
    ACTIONS(2888), 1,
      sym__whitespace,
    ACTIONS(2890), 1,
      sym__eof,
    STATE(109), 1,
      sym__soft_line_break,
    STATE(187), 1,
      sym__newline,
  [16195] = 6,
    ACTIONS(1272), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2892), 1,
      sym__whitespace,
    STATE(703), 1,
      sym_link_title,
  [16214] = 6,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    ACTIONS(2894), 1,
      sym__whitespace,
    STATE(643), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(804), 1,
      sym_pipe_table_delimiter_cell,
  [16233] = 6,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2896), 1,
      sym__whitespace,
    ACTIONS(2898), 1,
      sym__no_indented_chunk,
    STATE(729), 1,
      sym_link_title,
  [16252] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2900), 1,
      sym__eof,
    ACTIONS(2902), 1,
      sym__pipe_table_line_ending,
    STATE(344), 1,
      sym__newline,
    STATE(415), 1,
      sym__pipe_table_newline,
    STATE(694), 1,
      aux_sym_pipe_table_repeat1,
  [16271] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2902), 1,
      sym__pipe_table_line_ending,
    ACTIONS(2904), 1,
      sym__eof,
    STATE(334), 1,
      sym__newline,
    STATE(415), 1,
      sym__pipe_table_newline,
    STATE(674), 1,
      aux_sym_pipe_table_repeat1,
  [16290] = 1,
    ACTIONS(2823), 6,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_DASH,
      anon_sym_COLON,
      sym__whitespace,
  [16299] = 6,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    ACTIONS(2894), 1,
      sym__whitespace,
    STATE(640), 1,
      aux_sym_pipe_table_delimiter_row_repeat1,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(804), 1,
      sym_pipe_table_delimiter_cell,
  [16318] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2902), 1,
      sym__pipe_table_line_ending,
    ACTIONS(2906), 1,
      sym__eof,
    STATE(299), 1,
      sym__newline,
    STATE(415), 1,
      sym__pipe_table_newline,
    STATE(684), 1,
      aux_sym_pipe_table_repeat1,
  [16337] = 1,
    ACTIONS(2908), 6,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_DASH,
      anon_sym_COLON,
      sym__whitespace,
  [16346] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2910), 1,
      sym__whitespace,
    ACTIONS(2912), 1,
      sym__soft_line_ending,
    ACTIONS(2914), 1,
      sym__eof,
    STATE(103), 1,
      sym__soft_line_break,
    STATE(300), 1,
      sym__newline,
  [16365] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2912), 1,
      sym__soft_line_ending,
    ACTIONS(2916), 1,
      sym__whitespace,
    ACTIONS(2918), 1,
      sym__eof,
    STATE(102), 1,
      sym__soft_line_break,
    STATE(287), 1,
      sym__newline,
  [16384] = 6,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2870), 1,
      sym__soft_line_ending,
    ACTIONS(2920), 1,
      sym__whitespace,
    ACTIONS(2922), 1,
      sym__eof,
    STATE(117), 1,
      sym__soft_line_break,
    STATE(349), 1,
      sym__newline,
  [16403] = 6,
    ACTIONS(1184), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2924), 1,
      sym__whitespace,
    STATE(721), 1,
      sym_link_title,
  [16422] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2902), 1,
      sym__pipe_table_line_ending,
    ACTIONS(2926), 1,
      sym__eof,
    STATE(308), 1,
      sym__newline,
    STATE(415), 1,
      sym__pipe_table_newline,
    STATE(694), 1,
      aux_sym_pipe_table_repeat1,
  [16441] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2912), 1,
      sym__soft_line_ending,
    ACTIONS(2928), 1,
      sym__whitespace,
    ACTIONS(2930), 1,
      sym__eof,
    STATE(104), 1,
      sym__soft_line_break,
    STATE(309), 1,
      sym__newline,
  [16460] = 6,
    ACTIONS(1202), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2932), 1,
      sym__whitespace,
    STATE(722), 1,
      sym_link_title,
  [16479] = 6,
    ACTIONS(1220), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2934), 1,
      sym__whitespace,
    STATE(723), 1,
      sym_link_title,
  [16498] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2912), 1,
      sym__soft_line_ending,
    ACTIONS(2936), 1,
      sym__whitespace,
    ACTIONS(2938), 1,
      sym__eof,
    STATE(105), 1,
      sym__soft_line_break,
    STATE(313), 1,
      sym__newline,
  [16517] = 6,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(2912), 1,
      sym__soft_line_ending,
    ACTIONS(2940), 1,
      sym__whitespace,
    ACTIONS(2942), 1,
      sym__eof,
    STATE(106), 1,
      sym__soft_line_break,
    STATE(318), 1,
      sym__newline,
  [16536] = 6,
    ACTIONS(1238), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2944), 1,
      sym__whitespace,
    STATE(724), 1,
      sym_link_title,
  [16555] = 6,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2946), 1,
      sym__whitespace,
    ACTIONS(2948), 1,
      sym__no_indented_chunk,
    STATE(725), 1,
      sym_link_title,
  [16574] = 1,
    ACTIONS(2950), 6,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_DASH,
      anon_sym_COLON,
      sym__whitespace,
  [16583] = 1,
    ACTIONS(2952), 5,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
      sym__whitespace,
  [16591] = 4,
    ACTIONS(2956), 1,
      sym__pipe_table_line_ending,
    STATE(415), 1,
      sym__pipe_table_newline,
    STATE(694), 1,
      aux_sym_pipe_table_repeat1,
    ACTIONS(2954), 2,
      sym__line_ending,
      sym__eof,
  [16605] = 3,
    ACTIONS(2959), 1,
      anon_sym_PIPE,
    ACTIONS(2961), 1,
      sym__whitespace,
    ACTIONS(2864), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16617] = 3,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(2965), 1,
      sym__whitespace,
    ACTIONS(2831), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16629] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(2969), 1,
      sym__eof,
    STATE(186), 2,
      sym__newline,
      sym__soft_line_break,
  [16643] = 4,
    ACTIONS(2971), 1,
      anon_sym_DASH,
    ACTIONS(2973), 1,
      anon_sym_COLON,
    STATE(714), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2854), 2,
      anon_sym_PIPE,
      sym__whitespace,
  [16657] = 5,
    ACTIONS(1282), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(706), 1,
      sym_link_title,
  [16673] = 3,
    ACTIONS(2975), 1,
      anon_sym_PIPE,
    ACTIONS(2977), 1,
      sym__whitespace,
    ACTIONS(2378), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16685] = 3,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(2981), 1,
      sym__whitespace,
    ACTIONS(2378), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16697] = 1,
    ACTIONS(1470), 5,
      sym__no_indented_chunk,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_LPAREN,
      sym__whitespace,
  [16705] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(2983), 1,
      sym__eof,
    STATE(271), 2,
      sym__newline,
      sym__soft_line_break,
  [16719] = 5,
    ACTIONS(1277), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(697), 1,
      sym_link_title,
  [16735] = 3,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(2985), 1,
      sym__whitespace,
    ACTIONS(2987), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16747] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(2989), 1,
      sym__eof,
    STATE(190), 2,
      sym__newline,
      sym__soft_line_break,
  [16761] = 5,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    ACTIONS(2991), 1,
      anon_sym_PIPE,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(809), 1,
      sym_pipe_table_delimiter_cell,
  [16777] = 2,
    ACTIONS(2993), 1,
      sym_block_continuation,
    ACTIONS(1127), 4,
      anon_sym_DASH,
      anon_sym_COLON,
      anon_sym_PIPE,
      sym__whitespace,
  [16787] = 3,
    ACTIONS(2959), 1,
      anon_sym_PIPE,
    ACTIONS(2995), 1,
      sym__whitespace,
    ACTIONS(2812), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16799] = 5,
    ACTIONS(1251), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(743), 1,
      sym_link_title,
  [16815] = 3,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(2997), 1,
      sym__whitespace,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16827] = 3,
    ACTIONS(2975), 1,
      anon_sym_PIPE,
    ACTIONS(2999), 1,
      sym__whitespace,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16839] = 3,
    ACTIONS(2959), 1,
      anon_sym_PIPE,
    ACTIONS(3001), 1,
      sym__whitespace,
    ACTIONS(2831), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16851] = 3,
    ACTIONS(3003), 1,
      anon_sym_DASH,
    STATE(714), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2836), 3,
      anon_sym_COLON,
      anon_sym_PIPE,
      sym__whitespace,
  [16863] = 3,
    ACTIONS(2975), 1,
      anon_sym_PIPE,
    ACTIONS(3006), 1,
      sym__whitespace,
    ACTIONS(2366), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16875] = 4,
    ACTIONS(2971), 1,
      anon_sym_DASH,
    ACTIONS(3008), 1,
      anon_sym_COLON,
    STATE(714), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    ACTIONS(2842), 2,
      anon_sym_PIPE,
      sym__whitespace,
  [16889] = 5,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(3010), 1,
      sym__no_indented_chunk,
    STATE(726), 1,
      sym_link_title,
  [16905] = 3,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(3012), 1,
      sym__whitespace,
    ACTIONS(2545), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16917] = 3,
    ACTIONS(2975), 1,
      anon_sym_PIPE,
    ACTIONS(3014), 1,
      sym__whitespace,
    ACTIONS(2545), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [16929] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3018), 1,
      sym__eof,
    STATE(311), 2,
      sym__newline,
      sym__soft_line_break,
  [16943] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3020), 1,
      sym__eof,
    STATE(315), 2,
      sym__newline,
      sym__soft_line_break,
  [16957] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3022), 1,
      sym__eof,
    STATE(317), 2,
      sym__newline,
      sym__soft_line_break,
  [16971] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3024), 1,
      sym__eof,
    STATE(321), 2,
      sym__newline,
      sym__soft_line_break,
  [16985] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3026), 1,
      sym__eof,
    STATE(322), 2,
      sym__newline,
      sym__soft_line_break,
  [16999] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3028), 1,
      sym__eof,
    STATE(323), 2,
      sym__newline,
      sym__soft_line_break,
  [17013] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3030), 1,
      sym__eof,
    STATE(324), 2,
      sym__newline,
      sym__soft_line_break,
  [17027] = 4,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3016), 1,
      sym__soft_line_ending,
    ACTIONS(3032), 1,
      sym__eof,
    STATE(325), 2,
      sym__newline,
      sym__soft_line_break,
  [17041] = 5,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2898), 1,
      sym__no_indented_chunk,
    STATE(729), 1,
      sym_link_title,
  [17057] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(3034), 1,
      sym__eof,
    STATE(192), 2,
      sym__newline,
      sym__soft_line_break,
  [17071] = 5,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(3036), 1,
      sym__no_indented_chunk,
    STATE(732), 1,
      sym_link_title,
  [17087] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(3038), 1,
      sym__eof,
    STATE(347), 2,
      sym__newline,
      sym__soft_line_break,
  [17101] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(3040), 1,
      sym__eof,
    STATE(193), 2,
      sym__newline,
      sym__soft_line_break,
  [17115] = 5,
    ACTIONS(1272), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(703), 1,
      sym_link_title,
  [17131] = 3,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(3042), 1,
      sym__whitespace,
    ACTIONS(3044), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17143] = 1,
    ACTIONS(3046), 5,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
      anon_sym_PIPE,
      sym__whitespace,
  [17151] = 5,
    ACTIONS(1184), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(721), 1,
      sym_link_title,
  [17167] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(3048), 1,
      sym__eof,
    STATE(194), 2,
      sym__newline,
      sym__soft_line_break,
  [17181] = 5,
    ACTIONS(1202), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(722), 1,
      sym_link_title,
  [17197] = 3,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(3050), 1,
      sym__whitespace,
    ACTIONS(2864), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17209] = 5,
    ACTIONS(1220), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(723), 1,
      sym_link_title,
  [17225] = 5,
    ACTIONS(1238), 1,
      sym__no_indented_chunk,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(724), 1,
      sym_link_title,
  [17241] = 5,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    ACTIONS(2948), 1,
      sym__no_indented_chunk,
    STATE(725), 1,
      sym_link_title,
  [17257] = 4,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(2967), 1,
      sym__soft_line_ending,
    ACTIONS(3052), 1,
      sym__eof,
    STATE(191), 2,
      sym__newline,
      sym__soft_line_break,
  [17271] = 1,
    ACTIONS(1862), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
  [17278] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(727), 1,
      sym_link_title,
  [17291] = 2,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
    ACTIONS(2545), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17300] = 2,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(2545), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17309] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(729), 1,
      sym_link_title,
  [17322] = 2,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    ACTIONS(2987), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17331] = 2,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(2987), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17340] = 2,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    ACTIONS(2864), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17349] = 4,
    ACTIONS(2844), 1,
      anon_sym_DASH,
    ACTIONS(2846), 1,
      anon_sym_COLON,
    STATE(716), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
    STATE(809), 1,
      sym_pipe_table_delimiter_cell,
  [17362] = 1,
    ACTIONS(1253), 4,
      anon_sym_DASH,
      anon_sym_COLON,
      anon_sym_PIPE,
      sym__whitespace,
  [17369] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(706), 1,
      sym_link_title,
  [17382] = 1,
    ACTIONS(1874), 4,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
      sym__whitespace,
  [17389] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(703), 1,
      sym_link_title,
  [17402] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(732), 1,
      sym_link_title,
  [17415] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(743), 1,
      sym_link_title,
  [17428] = 2,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(2864), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17437] = 2,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17446] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(737), 1,
      sym_link_title,
  [17459] = 2,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
    ACTIONS(3044), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17468] = 2,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(3044), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17477] = 2,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(2378), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17486] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(721), 1,
      sym_link_title,
  [17499] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(722), 1,
      sym_link_title,
  [17512] = 2,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
    ACTIONS(2412), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17521] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(723), 1,
      sym_link_title,
  [17534] = 2,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
    ACTIONS(3058), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17543] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(724), 1,
      sym_link_title,
  [17556] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(697), 1,
      sym_link_title,
  [17569] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(725), 1,
      sym_link_title,
  [17582] = 2,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(2831), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17591] = 4,
    ACTIONS(2856), 1,
      anon_sym_DQUOTE,
    ACTIONS(2858), 1,
      anon_sym_SQUOTE,
    ACTIONS(2860), 1,
      anon_sym_LPAREN,
    STATE(726), 1,
      sym_link_title,
  [17604] = 2,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    ACTIONS(3060), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17613] = 3,
    ACTIONS(2378), 1,
      sym__line_ending,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    ACTIONS(3064), 1,
      sym__whitespace,
  [17623] = 3,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3066), 1,
      sym__eof,
    STATE(269), 1,
      sym__newline,
  [17633] = 3,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(3068), 1,
      sym__eof,
    STATE(314), 1,
      sym__newline,
  [17643] = 3,
    ACTIONS(3070), 1,
      sym__line_ending,
    ACTIONS(3072), 1,
      sym__eof,
    STATE(865), 1,
      sym__newline,
  [17653] = 3,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    ACTIONS(3074), 1,
      sym__whitespace,
  [17663] = 3,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3076), 1,
      sym__eof,
    STATE(219), 1,
      sym__newline,
  [17673] = 3,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3078), 1,
      sym__eof,
    STATE(220), 1,
      sym__newline,
  [17683] = 3,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
    ACTIONS(3082), 1,
      sym__whitespace,
  [17693] = 3,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(3084), 1,
      sym__eof,
    STATE(247), 1,
      sym__newline,
  [17703] = 3,
    ACTIONS(2545), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
    ACTIONS(3086), 1,
      sym__whitespace,
  [17713] = 1,
    ACTIONS(3088), 3,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
  [17719] = 3,
    ACTIONS(3044), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
    ACTIONS(3090), 1,
      sym__whitespace,
  [17729] = 3,
    ACTIONS(1904), 1,
      sym__line_ending,
    ACTIONS(3092), 1,
      sym__eof,
    STATE(270), 1,
      sym__newline,
  [17739] = 3,
    ACTIONS(2366), 1,
      sym__line_ending,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      sym__whitespace,
  [17749] = 3,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(3096), 1,
      sym__eof,
    STATE(346), 1,
      sym__newline,
  [17759] = 3,
    ACTIONS(2545), 1,
      sym__line_ending,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    ACTIONS(3098), 1,
      sym__whitespace,
  [17769] = 3,
    ACTIONS(2328), 1,
      sym__line_ending,
    ACTIONS(3100), 1,
      sym__eof,
    STATE(244), 1,
      sym__newline,
  [17779] = 1,
    ACTIONS(3102), 3,
      sym__line_ending,
      sym__soft_line_ending,
      sym__eof,
  [17785] = 1,
    ACTIONS(2954), 3,
      sym__line_ending,
      sym__eof,
      sym__pipe_table_line_ending,
  [17791] = 3,
    ACTIONS(2378), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
    ACTIONS(3104), 1,
      sym__whitespace,
  [17801] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(202), 1,
      sym__newline,
  [17808] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(890), 1,
      sym__newline,
  [17815] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(248), 1,
      sym__newline,
  [17822] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(249), 1,
      sym__newline,
  [17829] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(250), 1,
      sym__newline,
  [17836] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(251), 1,
      sym__newline,
  [17843] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(252), 1,
      sym__newline,
  [17850] = 2,
    ACTIONS(1904), 1,
      sym__line_ending,
    STATE(253), 1,
      sym__newline,
  [17857] = 2,
    ACTIONS(2959), 1,
      anon_sym_PIPE,
    ACTIONS(3108), 1,
      sym__whitespace,
  [17864] = 2,
    ACTIONS(3044), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
  [17871] = 2,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    ACTIONS(3110), 1,
      sym__whitespace,
  [17878] = 2,
    ACTIONS(2378), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
  [17885] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(209), 1,
      sym__newline,
  [17892] = 2,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
    ACTIONS(3112), 1,
      sym__whitespace,
  [17899] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(210), 1,
      sym__newline,
  [17906] = 2,
    ACTIONS(1892), 1,
      sym__block_close,
    ACTIONS(1894), 1,
      sym__fenced_code_block_end_backtick,
  [17913] = 2,
    ACTIONS(1892), 1,
      sym__block_close,
    ACTIONS(1894), 1,
      sym__fenced_code_block_end_tilde,
  [17920] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(205), 1,
      sym__newline,
  [17927] = 2,
    ACTIONS(1840), 1,
      sym__block_close,
    ACTIONS(1842), 1,
      sym__fenced_code_block_end_backtick,
  [17934] = 2,
    ACTIONS(1840), 1,
      sym__block_close,
    ACTIONS(1842), 1,
      sym__fenced_code_block_end_tilde,
  [17941] = 1,
    ACTIONS(3046), 2,
      anon_sym_PIPE,
      sym__whitespace,
  [17946] = 2,
    ACTIONS(3114), 1,
      sym__block_close,
    ACTIONS(3116), 1,
      sym__fenced_code_block_end_backtick,
  [17953] = 2,
    ACTIONS(3114), 1,
      sym__block_close,
    ACTIONS(3116), 1,
      sym__fenced_code_block_end_tilde,
  [17960] = 2,
    ACTIONS(1240), 1,
      sym__trigger_error,
    ACTIONS(3118), 1,
      sym_block_continuation,
  [17967] = 2,
    ACTIONS(3120), 1,
      sym__line_ending,
    STATE(646), 1,
      sym__newline,
  [17974] = 2,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
  [17981] = 2,
    ACTIONS(1916), 1,
      sym__block_close,
    ACTIONS(1918), 1,
      sym__fenced_code_block_end_backtick,
  [17988] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(207), 1,
      sym__newline,
  [17995] = 2,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
    ACTIONS(3122), 1,
      sym__whitespace,
  [18002] = 2,
    ACTIONS(3124), 1,
      sym__block_close,
    ACTIONS(3126), 1,
      sym__fenced_code_block_end_tilde,
  [18009] = 2,
    ACTIONS(2412), 1,
      sym__line_ending,
    ACTIONS(3128), 1,
      anon_sym_PIPE,
  [18016] = 2,
    ACTIONS(3130), 1,
      anon_sym_DASH,
    STATE(698), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
  [18023] = 2,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(364), 1,
      sym__newline,
  [18030] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(911), 1,
      sym__newline,
  [18037] = 2,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
    ACTIONS(3132), 1,
      sym__whitespace,
  [18044] = 2,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(369), 1,
      sym__newline,
  [18051] = 2,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(370), 1,
      sym__newline,
  [18058] = 2,
    ACTIONS(2545), 1,
      sym__line_ending,
    ACTIONS(3128), 1,
      anon_sym_PIPE,
  [18065] = 2,
    ACTIONS(1127), 1,
      sym__close_block,
    ACTIONS(3134), 1,
      sym_block_continuation,
  [18072] = 2,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(371), 1,
      sym__newline,
  [18079] = 2,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(359), 1,
      sym__newline,
  [18086] = 2,
    ACTIONS(1127), 1,
      sym__block_close,
    ACTIONS(3136), 1,
      sym_block_continuation,
  [18093] = 2,
    ACTIONS(2328), 1,
      sym__line_ending,
    STATE(211), 1,
      sym__newline,
  [18100] = 2,
    ACTIONS(143), 1,
      sym__block_close,
    ACTIONS(3138), 1,
      sym_block_continuation,
  [18107] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(883), 1,
      sym__newline,
  [18114] = 2,
    ACTIONS(1816), 1,
      sym__line_ending,
    STATE(384), 1,
      sym__newline,
  [18121] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(894), 1,
      sym__newline,
  [18128] = 2,
    ACTIONS(3124), 1,
      sym__block_close,
    ACTIONS(3126), 1,
      sym__fenced_code_block_end_backtick,
  [18135] = 2,
    ACTIONS(3044), 1,
      sym__line_ending,
    ACTIONS(3128), 1,
      anon_sym_PIPE,
  [18142] = 2,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(378), 1,
      sym__newline,
  [18149] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(899), 1,
      sym__newline,
  [18156] = 2,
    ACTIONS(2545), 1,
      sym__line_ending,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
  [18163] = 2,
    ACTIONS(3058), 1,
      sym__line_ending,
    ACTIONS(3128), 1,
      anon_sym_PIPE,
  [18170] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(908), 1,
      sym__newline,
  [18177] = 2,
    ACTIONS(1876), 1,
      sym__block_close,
    ACTIONS(1878), 1,
      sym__fenced_code_block_end_backtick,
  [18184] = 2,
    ACTIONS(1876), 1,
      sym__block_close,
    ACTIONS(1878), 1,
      sym__fenced_code_block_end_tilde,
  [18191] = 2,
    ACTIONS(3140), 1,
      anon_sym_DASH,
    STATE(648), 1,
      aux_sym_pipe_table_delimiter_cell_repeat1,
  [18198] = 2,
    ACTIONS(1820), 1,
      sym__line_ending,
    STATE(377), 1,
      sym__newline,
  [18205] = 2,
    ACTIONS(1916), 1,
      sym__block_close,
    ACTIONS(1918), 1,
      sym__fenced_code_block_end_tilde,
  [18212] = 2,
    ACTIONS(2975), 1,
      anon_sym_PIPE,
    ACTIONS(3142), 1,
      sym__whitespace,
  [18219] = 1,
    ACTIONS(2952), 2,
      anon_sym_PIPE,
      sym__whitespace,
  [18224] = 2,
    ACTIONS(3120), 1,
      sym__line_ending,
    STATE(647), 1,
      sym__newline,
  [18231] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(889), 1,
      sym__newline,
  [18238] = 2,
    ACTIONS(3106), 1,
      sym__line_ending,
    STATE(873), 1,
      sym__newline,
  [18245] = 1,
    ACTIONS(3144), 1,
      sym__close_block,
  [18249] = 1,
    ACTIONS(3146), 1,
      sym__block_close,
  [18253] = 1,
    ACTIONS(3148), 1,
      sym__block_close,
  [18257] = 1,
    ACTIONS(3150), 1,
      sym__block_close,
  [18261] = 1,
    ACTIONS(2979), 1,
      anon_sym_PIPE,
  [18265] = 1,
    ACTIONS(1664), 1,
      sym__close_block,
  [18269] = 1,
    ACTIONS(2963), 1,
      anon_sym_PIPE,
  [18273] = 1,
    ACTIONS(3152), 1,
      sym__close_block,
  [18277] = 1,
    ACTIONS(3154), 1,
      sym__block_close,
  [18281] = 1,
    ACTIONS(3156), 1,
      sym__block_close,
  [18285] = 1,
    ACTIONS(3158), 1,
      anon_sym_COLON,
  [18289] = 1,
    ACTIONS(3160), 1,
      sym__close_block,
  [18293] = 1,
    ACTIONS(3162), 1,
      sym__trigger_error,
  [18297] = 1,
    ACTIONS(3124), 1,
      sym__block_close,
  [18301] = 1,
    ACTIONS(1253), 1,
      sym__close_block,
  [18305] = 1,
    ACTIONS(3164), 1,
      ts_builtin_sym_end,
  [18309] = 1,
    ACTIONS(141), 1,
      sym__block_close,
  [18313] = 1,
    ACTIONS(3166), 1,
      sym__block_close,
  [18317] = 1,
    ACTIONS(3168), 1,
      sym__close_block,
  [18321] = 1,
    ACTIONS(3170), 1,
      sym__block_close,
  [18325] = 1,
    ACTIONS(3172), 1,
      sym__trigger_error,
  [18329] = 1,
    ACTIONS(3174), 1,
      sym__block_close,
  [18333] = 1,
    ACTIONS(3176), 1,
      sym__block_close,
  [18337] = 1,
    ACTIONS(3114), 1,
      sym__block_close,
  [18341] = 1,
    ACTIONS(3178), 1,
      sym__block_close,
  [18345] = 1,
    ACTIONS(3180), 1,
      sym__block_close,
  [18349] = 1,
    ACTIONS(1253), 1,
      sym__block_close,
  [18353] = 1,
    ACTIONS(3128), 1,
      anon_sym_PIPE,
  [18357] = 1,
    ACTIONS(3182), 1,
      sym__block_close,
  [18361] = 1,
    ACTIONS(3184), 1,
      sym__block_close,
  [18365] = 1,
    ACTIONS(3186), 1,
      sym__block_close,
  [18369] = 1,
    ACTIONS(3188), 1,
      sym__block_close,
  [18373] = 1,
    ACTIONS(3190), 1,
      sym__block_close,
  [18377] = 1,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
  [18381] = 1,
    ACTIONS(3192), 1,
      sym__block_close,
  [18385] = 1,
    ACTIONS(3194), 1,
      sym__close_block,
  [18389] = 1,
    ACTIONS(3196), 1,
      sym__block_close,
  [18393] = 1,
    ACTIONS(3198), 1,
      sym__block_close,
  [18397] = 1,
    ACTIONS(1470), 1,
      sym__trigger_error,
  [18401] = 1,
    ACTIONS(3200), 1,
      sym__block_close,
  [18405] = 1,
    ACTIONS(3202), 1,
      sym__close_block,
  [18409] = 1,
    ACTIONS(3080), 1,
      anon_sym_PIPE,
  [18413] = 1,
    ACTIONS(3204), 1,
      sym__close_block,
  [18417] = 1,
    ACTIONS(3206), 1,
      sym__close_block,
  [18421] = 1,
    ACTIONS(3208), 1,
      sym__close_block,
  [18425] = 1,
    ACTIONS(3210), 1,
      anon_sym_COLON,
  [18429] = 1,
    ACTIONS(3212), 1,
      sym__close_block,
  [18433] = 1,
    ACTIONS(3214), 1,
      sym__trigger_error,
  [18437] = 1,
    ACTIONS(3216), 1,
      sym__block_close,
  [18441] = 1,
    ACTIONS(3218), 1,
      sym__block_close,
  [18445] = 1,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
  [18449] = 1,
    ACTIONS(3220), 1,
      sym__block_close,
  [18453] = 1,
    ACTIONS(3222), 1,
      sym__block_close,
  [18457] = 1,
    ACTIONS(3224), 1,
      sym__line_ending,
  [18461] = 1,
    ACTIONS(3226), 1,
      sym__line_ending,
  [18465] = 1,
    ACTIONS(3228), 1,
      sym__whitespace,
  [18469] = 1,
    ACTIONS(3230), 1,
      anon_sym_COLON,
  [18473] = 1,
    ACTIONS(3232), 1,
      sym__block_close,
  [18477] = 1,
    ACTIONS(3234), 1,
      anon_sym_COLON,
  [18481] = 1,
    ACTIONS(3236), 1,
      sym__close_block,
  [18485] = 1,
    ACTIONS(3238), 1,
      sym__close_block,
  [18489] = 1,
    ACTIONS(3240), 1,
      sym__close_block,
  [18493] = 1,
    ACTIONS(3242), 1,
      sym__close_block,
  [18497] = 1,
    ACTIONS(3244), 1,
      sym__block_close,
  [18501] = 1,
    ACTIONS(3246), 1,
      anon_sym_COLON,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(351)] = 0,
  [SMALL_STATE(352)] = 77,
  [SMALL_STATE(353)] = 154,
  [SMALL_STATE(354)] = 221,
  [SMALL_STATE(355)] = 290,
  [SMALL_STATE(356)] = 359,
  [SMALL_STATE(357)] = 428,
  [SMALL_STATE(358)] = 497,
  [SMALL_STATE(359)] = 558,
  [SMALL_STATE(360)] = 620,
  [SMALL_STATE(361)] = 678,
  [SMALL_STATE(362)] = 736,
  [SMALL_STATE(363)] = 800,
  [SMALL_STATE(364)] = 858,
  [SMALL_STATE(365)] = 920,
  [SMALL_STATE(366)] = 982,
  [SMALL_STATE(367)] = 1044,
  [SMALL_STATE(368)] = 1106,
  [SMALL_STATE(369)] = 1172,
  [SMALL_STATE(370)] = 1234,
  [SMALL_STATE(371)] = 1296,
  [SMALL_STATE(372)] = 1358,
  [SMALL_STATE(373)] = 1418,
  [SMALL_STATE(374)] = 1482,
  [SMALL_STATE(375)] = 1542,
  [SMALL_STATE(376)] = 1606,
  [SMALL_STATE(377)] = 1668,
  [SMALL_STATE(378)] = 1730,
  [SMALL_STATE(379)] = 1792,
  [SMALL_STATE(380)] = 1852,
  [SMALL_STATE(381)] = 1912,
  [SMALL_STATE(382)] = 1978,
  [SMALL_STATE(383)] = 2044,
  [SMALL_STATE(384)] = 2104,
  [SMALL_STATE(385)] = 2166,
  [SMALL_STATE(386)] = 2224,
  [SMALL_STATE(387)] = 2282,
  [SMALL_STATE(388)] = 2348,
  [SMALL_STATE(389)] = 2407,
  [SMALL_STATE(390)] = 2474,
  [SMALL_STATE(391)] = 2533,
  [SMALL_STATE(392)] = 2592,
  [SMALL_STATE(393)] = 2651,
  [SMALL_STATE(394)] = 2710,
  [SMALL_STATE(395)] = 2769,
  [SMALL_STATE(396)] = 2828,
  [SMALL_STATE(397)] = 2885,
  [SMALL_STATE(398)] = 2944,
  [SMALL_STATE(399)] = 3003,
  [SMALL_STATE(400)] = 3062,
  [SMALL_STATE(401)] = 3121,
  [SMALL_STATE(402)] = 3180,
  [SMALL_STATE(403)] = 3237,
  [SMALL_STATE(404)] = 3304,
  [SMALL_STATE(405)] = 3361,
  [SMALL_STATE(406)] = 3428,
  [SMALL_STATE(407)] = 3487,
  [SMALL_STATE(408)] = 3544,
  [SMALL_STATE(409)] = 3603,
  [SMALL_STATE(410)] = 3662,
  [SMALL_STATE(411)] = 3719,
  [SMALL_STATE(412)] = 3778,
  [SMALL_STATE(413)] = 3835,
  [SMALL_STATE(414)] = 3894,
  [SMALL_STATE(415)] = 3957,
  [SMALL_STATE(416)] = 4022,
  [SMALL_STATE(417)] = 4081,
  [SMALL_STATE(418)] = 4140,
  [SMALL_STATE(419)] = 4199,
  [SMALL_STATE(420)] = 4258,
  [SMALL_STATE(421)] = 4317,
  [SMALL_STATE(422)] = 4376,
  [SMALL_STATE(423)] = 4435,
  [SMALL_STATE(424)] = 4494,
  [SMALL_STATE(425)] = 4553,
  [SMALL_STATE(426)] = 4620,
  [SMALL_STATE(427)] = 4679,
  [SMALL_STATE(428)] = 4735,
  [SMALL_STATE(429)] = 4793,
  [SMALL_STATE(430)] = 4851,
  [SMALL_STATE(431)] = 4909,
  [SMALL_STATE(432)] = 4963,
  [SMALL_STATE(433)] = 5017,
  [SMALL_STATE(434)] = 5075,
  [SMALL_STATE(435)] = 5131,
  [SMALL_STATE(436)] = 5189,
  [SMALL_STATE(437)] = 5247,
  [SMALL_STATE(438)] = 5303,
  [SMALL_STATE(439)] = 5361,
  [SMALL_STATE(440)] = 5419,
  [SMALL_STATE(441)] = 5477,
  [SMALL_STATE(442)] = 5535,
  [SMALL_STATE(443)] = 5591,
  [SMALL_STATE(444)] = 5649,
  [SMALL_STATE(445)] = 5707,
  [SMALL_STATE(446)] = 5765,
  [SMALL_STATE(447)] = 5823,
  [SMALL_STATE(448)] = 5881,
  [SMALL_STATE(449)] = 5939,
  [SMALL_STATE(450)] = 5993,
  [SMALL_STATE(451)] = 6051,
  [SMALL_STATE(452)] = 6105,
  [SMALL_STATE(453)] = 6163,
  [SMALL_STATE(454)] = 6227,
  [SMALL_STATE(455)] = 6291,
  [SMALL_STATE(456)] = 6355,
  [SMALL_STATE(457)] = 6411,
  [SMALL_STATE(458)] = 6475,
  [SMALL_STATE(459)] = 6536,
  [SMALL_STATE(460)] = 6595,
  [SMALL_STATE(461)] = 6654,
  [SMALL_STATE(462)] = 6713,
  [SMALL_STATE(463)] = 6764,
  [SMALL_STATE(464)] = 6815,
  [SMALL_STATE(465)] = 6866,
  [SMALL_STATE(466)] = 6917,
  [SMALL_STATE(467)] = 6968,
  [SMALL_STATE(468)] = 7019,
  [SMALL_STATE(469)] = 7078,
  [SMALL_STATE(470)] = 7137,
  [SMALL_STATE(471)] = 7196,
  [SMALL_STATE(472)] = 7255,
  [SMALL_STATE(473)] = 7314,
  [SMALL_STATE(474)] = 7373,
  [SMALL_STATE(475)] = 7432,
  [SMALL_STATE(476)] = 7491,
  [SMALL_STATE(477)] = 7544,
  [SMALL_STATE(478)] = 7603,
  [SMALL_STATE(479)] = 7662,
  [SMALL_STATE(480)] = 7721,
  [SMALL_STATE(481)] = 7776,
  [SMALL_STATE(482)] = 7837,
  [SMALL_STATE(483)] = 7898,
  [SMALL_STATE(484)] = 7957,
  [SMALL_STATE(485)] = 8018,
  [SMALL_STATE(486)] = 8073,
  [SMALL_STATE(487)] = 8128,
  [SMALL_STATE(488)] = 8181,
  [SMALL_STATE(489)] = 8234,
  [SMALL_STATE(490)] = 8289,
  [SMALL_STATE(491)] = 8348,
  [SMALL_STATE(492)] = 8403,
  [SMALL_STATE(493)] = 8462,
  [SMALL_STATE(494)] = 8523,
  [SMALL_STATE(495)] = 8584,
  [SMALL_STATE(496)] = 8635,
  [SMALL_STATE(497)] = 8681,
  [SMALL_STATE(498)] = 8727,
  [SMALL_STATE(499)] = 8783,
  [SMALL_STATE(500)] = 8833,
  [SMALL_STATE(501)] = 8883,
  [SMALL_STATE(502)] = 8933,
  [SMALL_STATE(503)] = 8981,
  [SMALL_STATE(504)] = 9029,
  [SMALL_STATE(505)] = 9085,
  [SMALL_STATE(506)] = 9135,
  [SMALL_STATE(507)] = 9185,
  [SMALL_STATE(508)] = 9235,
  [SMALL_STATE(509)] = 9287,
  [SMALL_STATE(510)] = 9337,
  [SMALL_STATE(511)] = 9395,
  [SMALL_STATE(512)] = 9455,
  [SMALL_STATE(513)] = 9513,
  [SMALL_STATE(514)] = 9563,
  [SMALL_STATE(515)] = 9609,
  [SMALL_STATE(516)] = 9659,
  [SMALL_STATE(517)] = 9705,
  [SMALL_STATE(518)] = 9761,
  [SMALL_STATE(519)] = 9811,
  [SMALL_STATE(520)] = 9861,
  [SMALL_STATE(521)] = 9911,
  [SMALL_STATE(522)] = 9969,
  [SMALL_STATE(523)] = 10019,
  [SMALL_STATE(524)] = 10069,
  [SMALL_STATE(525)] = 10127,
  [SMALL_STATE(526)] = 10177,
  [SMALL_STATE(527)] = 10235,
  [SMALL_STATE(528)] = 10293,
  [SMALL_STATE(529)] = 10353,
  [SMALL_STATE(530)] = 10403,
  [SMALL_STATE(531)] = 10452,
  [SMALL_STATE(532)] = 10497,
  [SMALL_STATE(533)] = 10546,
  [SMALL_STATE(534)] = 10591,
  [SMALL_STATE(535)] = 10640,
  [SMALL_STATE(536)] = 10685,
  [SMALL_STATE(537)] = 10734,
  [SMALL_STATE(538)] = 10781,
  [SMALL_STATE(539)] = 10828,
  [SMALL_STATE(540)] = 10873,
  [SMALL_STATE(541)] = 10922,
  [SMALL_STATE(542)] = 10979,
  [SMALL_STATE(543)] = 11036,
  [SMALL_STATE(544)] = 11081,
  [SMALL_STATE(545)] = 11128,
  [SMALL_STATE(546)] = 11179,
  [SMALL_STATE(547)] = 11226,
  [SMALL_STATE(548)] = 11273,
  [SMALL_STATE(549)] = 11320,
  [SMALL_STATE(550)] = 11377,
  [SMALL_STATE(551)] = 11424,
  [SMALL_STATE(552)] = 11469,
  [SMALL_STATE(553)] = 11514,
  [SMALL_STATE(554)] = 11561,
  [SMALL_STATE(555)] = 11606,
  [SMALL_STATE(556)] = 11663,
  [SMALL_STATE(557)] = 11712,
  [SMALL_STATE(558)] = 11761,
  [SMALL_STATE(559)] = 11810,
  [SMALL_STATE(560)] = 11861,
  [SMALL_STATE(561)] = 11906,
  [SMALL_STATE(562)] = 11955,
  [SMALL_STATE(563)] = 12000,
  [SMALL_STATE(564)] = 12049,
  [SMALL_STATE(565)] = 12098,
  [SMALL_STATE(566)] = 12155,
  [SMALL_STATE(567)] = 12200,
  [SMALL_STATE(568)] = 12244,
  [SMALL_STATE(569)] = 12298,
  [SMALL_STATE(570)] = 12342,
  [SMALL_STATE(571)] = 12390,
  [SMALL_STATE(572)] = 12434,
  [SMALL_STATE(573)] = 12478,
  [SMALL_STATE(574)] = 12522,
  [SMALL_STATE(575)] = 12566,
  [SMALL_STATE(576)] = 12610,
  [SMALL_STATE(577)] = 12654,
  [SMALL_STATE(578)] = 12700,
  [SMALL_STATE(579)] = 12744,
  [SMALL_STATE(580)] = 12792,
  [SMALL_STATE(581)] = 12836,
  [SMALL_STATE(582)] = 12880,
  [SMALL_STATE(583)] = 12924,
  [SMALL_STATE(584)] = 12972,
  [SMALL_STATE(585)] = 13016,
  [SMALL_STATE(586)] = 13060,
  [SMALL_STATE(587)] = 13104,
  [SMALL_STATE(588)] = 13158,
  [SMALL_STATE(589)] = 13204,
  [SMALL_STATE(590)] = 13248,
  [SMALL_STATE(591)] = 13292,
  [SMALL_STATE(592)] = 13336,
  [SMALL_STATE(593)] = 13384,
  [SMALL_STATE(594)] = 13432,
  [SMALL_STATE(595)] = 13486,
  [SMALL_STATE(596)] = 13540,
  [SMALL_STATE(597)] = 13584,
  [SMALL_STATE(598)] = 13628,
  [SMALL_STATE(599)] = 13672,
  [SMALL_STATE(600)] = 13718,
  [SMALL_STATE(601)] = 13764,
  [SMALL_STATE(602)] = 13808,
  [SMALL_STATE(603)] = 13852,
  [SMALL_STATE(604)] = 13896,
  [SMALL_STATE(605)] = 13950,
  [SMALL_STATE(606)] = 13994,
  [SMALL_STATE(607)] = 14042,
  [SMALL_STATE(608)] = 14090,
  [SMALL_STATE(609)] = 14134,
  [SMALL_STATE(610)] = 14178,
  [SMALL_STATE(611)] = 14222,
  [SMALL_STATE(612)] = 14273,
  [SMALL_STATE(613)] = 14316,
  [SMALL_STATE(614)] = 14367,
  [SMALL_STATE(615)] = 14418,
  [SMALL_STATE(616)] = 14461,
  [SMALL_STATE(617)] = 14512,
  [SMALL_STATE(618)] = 14563,
  [SMALL_STATE(619)] = 14606,
  [SMALL_STATE(620)] = 14649,
  [SMALL_STATE(621)] = 14700,
  [SMALL_STATE(622)] = 14743,
  [SMALL_STATE(623)] = 14786,
  [SMALL_STATE(624)] = 14837,
  [SMALL_STATE(625)] = 14880,
  [SMALL_STATE(626)] = 14923,
  [SMALL_STATE(627)] = 14966,
  [SMALL_STATE(628)] = 15017,
  [SMALL_STATE(629)] = 15060,
  [SMALL_STATE(630)] = 15102,
  [SMALL_STATE(631)] = 15144,
  [SMALL_STATE(632)] = 15186,
  [SMALL_STATE(633)] = 15228,
  [SMALL_STATE(634)] = 15270,
  [SMALL_STATE(635)] = 15312,
  [SMALL_STATE(636)] = 15364,
  [SMALL_STATE(637)] = 15416,
  [SMALL_STATE(638)] = 15468,
  [SMALL_STATE(639)] = 15520,
  [SMALL_STATE(640)] = 15572,
  [SMALL_STATE(641)] = 15596,
  [SMALL_STATE(642)] = 15620,
  [SMALL_STATE(643)] = 15644,
  [SMALL_STATE(644)] = 15668,
  [SMALL_STATE(645)] = 15683,
  [SMALL_STATE(646)] = 15700,
  [SMALL_STATE(647)] = 15725,
  [SMALL_STATE(648)] = 15750,
  [SMALL_STATE(649)] = 15767,
  [SMALL_STATE(650)] = 15789,
  [SMALL_STATE(651)] = 15811,
  [SMALL_STATE(652)] = 15829,
  [SMALL_STATE(653)] = 15847,
  [SMALL_STATE(654)] = 15869,
  [SMALL_STATE(655)] = 15891,
  [SMALL_STATE(656)] = 15913,
  [SMALL_STATE(657)] = 15935,
  [SMALL_STATE(658)] = 15957,
  [SMALL_STATE(659)] = 15979,
  [SMALL_STATE(660)] = 16001,
  [SMALL_STATE(661)] = 16023,
  [SMALL_STATE(662)] = 16033,
  [SMALL_STATE(663)] = 16051,
  [SMALL_STATE(664)] = 16070,
  [SMALL_STATE(665)] = 16089,
  [SMALL_STATE(666)] = 16108,
  [SMALL_STATE(667)] = 16127,
  [SMALL_STATE(668)] = 16146,
  [SMALL_STATE(669)] = 16157,
  [SMALL_STATE(670)] = 16176,
  [SMALL_STATE(671)] = 16195,
  [SMALL_STATE(672)] = 16214,
  [SMALL_STATE(673)] = 16233,
  [SMALL_STATE(674)] = 16252,
  [SMALL_STATE(675)] = 16271,
  [SMALL_STATE(676)] = 16290,
  [SMALL_STATE(677)] = 16299,
  [SMALL_STATE(678)] = 16318,
  [SMALL_STATE(679)] = 16337,
  [SMALL_STATE(680)] = 16346,
  [SMALL_STATE(681)] = 16365,
  [SMALL_STATE(682)] = 16384,
  [SMALL_STATE(683)] = 16403,
  [SMALL_STATE(684)] = 16422,
  [SMALL_STATE(685)] = 16441,
  [SMALL_STATE(686)] = 16460,
  [SMALL_STATE(687)] = 16479,
  [SMALL_STATE(688)] = 16498,
  [SMALL_STATE(689)] = 16517,
  [SMALL_STATE(690)] = 16536,
  [SMALL_STATE(691)] = 16555,
  [SMALL_STATE(692)] = 16574,
  [SMALL_STATE(693)] = 16583,
  [SMALL_STATE(694)] = 16591,
  [SMALL_STATE(695)] = 16605,
  [SMALL_STATE(696)] = 16617,
  [SMALL_STATE(697)] = 16629,
  [SMALL_STATE(698)] = 16643,
  [SMALL_STATE(699)] = 16657,
  [SMALL_STATE(700)] = 16673,
  [SMALL_STATE(701)] = 16685,
  [SMALL_STATE(702)] = 16697,
  [SMALL_STATE(703)] = 16705,
  [SMALL_STATE(704)] = 16719,
  [SMALL_STATE(705)] = 16735,
  [SMALL_STATE(706)] = 16747,
  [SMALL_STATE(707)] = 16761,
  [SMALL_STATE(708)] = 16777,
  [SMALL_STATE(709)] = 16787,
  [SMALL_STATE(710)] = 16799,
  [SMALL_STATE(711)] = 16815,
  [SMALL_STATE(712)] = 16827,
  [SMALL_STATE(713)] = 16839,
  [SMALL_STATE(714)] = 16851,
  [SMALL_STATE(715)] = 16863,
  [SMALL_STATE(716)] = 16875,
  [SMALL_STATE(717)] = 16889,
  [SMALL_STATE(718)] = 16905,
  [SMALL_STATE(719)] = 16917,
  [SMALL_STATE(720)] = 16929,
  [SMALL_STATE(721)] = 16943,
  [SMALL_STATE(722)] = 16957,
  [SMALL_STATE(723)] = 16971,
  [SMALL_STATE(724)] = 16985,
  [SMALL_STATE(725)] = 16999,
  [SMALL_STATE(726)] = 17013,
  [SMALL_STATE(727)] = 17027,
  [SMALL_STATE(728)] = 17041,
  [SMALL_STATE(729)] = 17057,
  [SMALL_STATE(730)] = 17071,
  [SMALL_STATE(731)] = 17087,
  [SMALL_STATE(732)] = 17101,
  [SMALL_STATE(733)] = 17115,
  [SMALL_STATE(734)] = 17131,
  [SMALL_STATE(735)] = 17143,
  [SMALL_STATE(736)] = 17151,
  [SMALL_STATE(737)] = 17167,
  [SMALL_STATE(738)] = 17181,
  [SMALL_STATE(739)] = 17197,
  [SMALL_STATE(740)] = 17209,
  [SMALL_STATE(741)] = 17225,
  [SMALL_STATE(742)] = 17241,
  [SMALL_STATE(743)] = 17257,
  [SMALL_STATE(744)] = 17271,
  [SMALL_STATE(745)] = 17278,
  [SMALL_STATE(746)] = 17291,
  [SMALL_STATE(747)] = 17300,
  [SMALL_STATE(748)] = 17309,
  [SMALL_STATE(749)] = 17322,
  [SMALL_STATE(750)] = 17331,
  [SMALL_STATE(751)] = 17340,
  [SMALL_STATE(752)] = 17349,
  [SMALL_STATE(753)] = 17362,
  [SMALL_STATE(754)] = 17369,
  [SMALL_STATE(755)] = 17382,
  [SMALL_STATE(756)] = 17389,
  [SMALL_STATE(757)] = 17402,
  [SMALL_STATE(758)] = 17415,
  [SMALL_STATE(759)] = 17428,
  [SMALL_STATE(760)] = 17437,
  [SMALL_STATE(761)] = 17446,
  [SMALL_STATE(762)] = 17459,
  [SMALL_STATE(763)] = 17468,
  [SMALL_STATE(764)] = 17477,
  [SMALL_STATE(765)] = 17486,
  [SMALL_STATE(766)] = 17499,
  [SMALL_STATE(767)] = 17512,
  [SMALL_STATE(768)] = 17521,
  [SMALL_STATE(769)] = 17534,
  [SMALL_STATE(770)] = 17543,
  [SMALL_STATE(771)] = 17556,
  [SMALL_STATE(772)] = 17569,
  [SMALL_STATE(773)] = 17582,
  [SMALL_STATE(774)] = 17591,
  [SMALL_STATE(775)] = 17604,
  [SMALL_STATE(776)] = 17613,
  [SMALL_STATE(777)] = 17623,
  [SMALL_STATE(778)] = 17633,
  [SMALL_STATE(779)] = 17643,
  [SMALL_STATE(780)] = 17653,
  [SMALL_STATE(781)] = 17663,
  [SMALL_STATE(782)] = 17673,
  [SMALL_STATE(783)] = 17683,
  [SMALL_STATE(784)] = 17693,
  [SMALL_STATE(785)] = 17703,
  [SMALL_STATE(786)] = 17713,
  [SMALL_STATE(787)] = 17719,
  [SMALL_STATE(788)] = 17729,
  [SMALL_STATE(789)] = 17739,
  [SMALL_STATE(790)] = 17749,
  [SMALL_STATE(791)] = 17759,
  [SMALL_STATE(792)] = 17769,
  [SMALL_STATE(793)] = 17779,
  [SMALL_STATE(794)] = 17785,
  [SMALL_STATE(795)] = 17791,
  [SMALL_STATE(796)] = 17801,
  [SMALL_STATE(797)] = 17808,
  [SMALL_STATE(798)] = 17815,
  [SMALL_STATE(799)] = 17822,
  [SMALL_STATE(800)] = 17829,
  [SMALL_STATE(801)] = 17836,
  [SMALL_STATE(802)] = 17843,
  [SMALL_STATE(803)] = 17850,
  [SMALL_STATE(804)] = 17857,
  [SMALL_STATE(805)] = 17864,
  [SMALL_STATE(806)] = 17871,
  [SMALL_STATE(807)] = 17878,
  [SMALL_STATE(808)] = 17885,
  [SMALL_STATE(809)] = 17892,
  [SMALL_STATE(810)] = 17899,
  [SMALL_STATE(811)] = 17906,
  [SMALL_STATE(812)] = 17913,
  [SMALL_STATE(813)] = 17920,
  [SMALL_STATE(814)] = 17927,
  [SMALL_STATE(815)] = 17934,
  [SMALL_STATE(816)] = 17941,
  [SMALL_STATE(817)] = 17946,
  [SMALL_STATE(818)] = 17953,
  [SMALL_STATE(819)] = 17960,
  [SMALL_STATE(820)] = 17967,
  [SMALL_STATE(821)] = 17974,
  [SMALL_STATE(822)] = 17981,
  [SMALL_STATE(823)] = 17988,
  [SMALL_STATE(824)] = 17995,
  [SMALL_STATE(825)] = 18002,
  [SMALL_STATE(826)] = 18009,
  [SMALL_STATE(827)] = 18016,
  [SMALL_STATE(828)] = 18023,
  [SMALL_STATE(829)] = 18030,
  [SMALL_STATE(830)] = 18037,
  [SMALL_STATE(831)] = 18044,
  [SMALL_STATE(832)] = 18051,
  [SMALL_STATE(833)] = 18058,
  [SMALL_STATE(834)] = 18065,
  [SMALL_STATE(835)] = 18072,
  [SMALL_STATE(836)] = 18079,
  [SMALL_STATE(837)] = 18086,
  [SMALL_STATE(838)] = 18093,
  [SMALL_STATE(839)] = 18100,
  [SMALL_STATE(840)] = 18107,
  [SMALL_STATE(841)] = 18114,
  [SMALL_STATE(842)] = 18121,
  [SMALL_STATE(843)] = 18128,
  [SMALL_STATE(844)] = 18135,
  [SMALL_STATE(845)] = 18142,
  [SMALL_STATE(846)] = 18149,
  [SMALL_STATE(847)] = 18156,
  [SMALL_STATE(848)] = 18163,
  [SMALL_STATE(849)] = 18170,
  [SMALL_STATE(850)] = 18177,
  [SMALL_STATE(851)] = 18184,
  [SMALL_STATE(852)] = 18191,
  [SMALL_STATE(853)] = 18198,
  [SMALL_STATE(854)] = 18205,
  [SMALL_STATE(855)] = 18212,
  [SMALL_STATE(856)] = 18219,
  [SMALL_STATE(857)] = 18224,
  [SMALL_STATE(858)] = 18231,
  [SMALL_STATE(859)] = 18238,
  [SMALL_STATE(860)] = 18245,
  [SMALL_STATE(861)] = 18249,
  [SMALL_STATE(862)] = 18253,
  [SMALL_STATE(863)] = 18257,
  [SMALL_STATE(864)] = 18261,
  [SMALL_STATE(865)] = 18265,
  [SMALL_STATE(866)] = 18269,
  [SMALL_STATE(867)] = 18273,
  [SMALL_STATE(868)] = 18277,
  [SMALL_STATE(869)] = 18281,
  [SMALL_STATE(870)] = 18285,
  [SMALL_STATE(871)] = 18289,
  [SMALL_STATE(872)] = 18293,
  [SMALL_STATE(873)] = 18297,
  [SMALL_STATE(874)] = 18301,
  [SMALL_STATE(875)] = 18305,
  [SMALL_STATE(876)] = 18309,
  [SMALL_STATE(877)] = 18313,
  [SMALL_STATE(878)] = 18317,
  [SMALL_STATE(879)] = 18321,
  [SMALL_STATE(880)] = 18325,
  [SMALL_STATE(881)] = 18329,
  [SMALL_STATE(882)] = 18333,
  [SMALL_STATE(883)] = 18337,
  [SMALL_STATE(884)] = 18341,
  [SMALL_STATE(885)] = 18345,
  [SMALL_STATE(886)] = 18349,
  [SMALL_STATE(887)] = 18353,
  [SMALL_STATE(888)] = 18357,
  [SMALL_STATE(889)] = 18361,
  [SMALL_STATE(890)] = 18365,
  [SMALL_STATE(891)] = 18369,
  [SMALL_STATE(892)] = 18373,
  [SMALL_STATE(893)] = 18377,
  [SMALL_STATE(894)] = 18381,
  [SMALL_STATE(895)] = 18385,
  [SMALL_STATE(896)] = 18389,
  [SMALL_STATE(897)] = 18393,
  [SMALL_STATE(898)] = 18397,
  [SMALL_STATE(899)] = 18401,
  [SMALL_STATE(900)] = 18405,
  [SMALL_STATE(901)] = 18409,
  [SMALL_STATE(902)] = 18413,
  [SMALL_STATE(903)] = 18417,
  [SMALL_STATE(904)] = 18421,
  [SMALL_STATE(905)] = 18425,
  [SMALL_STATE(906)] = 18429,
  [SMALL_STATE(907)] = 18433,
  [SMALL_STATE(908)] = 18437,
  [SMALL_STATE(909)] = 18441,
  [SMALL_STATE(910)] = 18445,
  [SMALL_STATE(911)] = 18449,
  [SMALL_STATE(912)] = 18453,
  [SMALL_STATE(913)] = 18457,
  [SMALL_STATE(914)] = 18461,
  [SMALL_STATE(915)] = 18465,
  [SMALL_STATE(916)] = 18469,
  [SMALL_STATE(917)] = 18473,
  [SMALL_STATE(918)] = 18477,
  [SMALL_STATE(919)] = 18481,
  [SMALL_STATE(920)] = 18485,
  [SMALL_STATE(921)] = 18489,
  [SMALL_STATE(922)] = 18493,
  [SMALL_STATE(923)] = 18497,
  [SMALL_STATE(924)] = 18501,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 0, 0, 0),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT(353),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [89] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_content, 4, 0, 0),
  [143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_content, 3, 0, 0),
  [145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_content, 1, 0, 0),
  [147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0),
  [161] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [164] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [167] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(469),
  [170] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(470),
  [173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(471),
  [176] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(472),
  [179] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(473),
  [182] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [188] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [197] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [203] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [209] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [212] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [218] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [221] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [224] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [227] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [230] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_quote_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 1, 0, 2),
  [240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 2, 0, 4),
  [242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 1, 0, 0),
  [248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section1, 1, 0, 1),
  [250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0),
  [252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [270] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(490),
  [273] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(468),
  [276] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(492),
  [279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [282] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [285] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [288] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [291] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [303] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [306] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [318] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [339] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(470),
  [348] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(471),
  [351] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(472),
  [354] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(473),
  [357] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [363] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [366] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [369] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [372] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [375] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [381] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [387] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [393] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section1, 2, 0, 1),
  [398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0),
  [400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [409] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(468),
  [421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(492),
  [424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [433] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [436] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [439] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [454] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [457] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [463] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [466] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [469] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [472] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [475] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [478] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section2, 2, 0, 1),
  [483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section2, 1, 0, 1),
  [485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [488] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(471),
  [497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(472),
  [500] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(473),
  [503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [509] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [512] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [515] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [518] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [521] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [527] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [530] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [533] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [536] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [539] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0),
  [544] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [550] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [553] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(492),
  [565] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [571] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [574] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [577] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [583] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [586] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [589] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [592] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [595] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [598] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [601] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [604] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [610] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [619] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section3, 2, 0, 1),
  [624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section3, 1, 0, 1),
  [626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [629] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [632] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [635] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(472),
  [638] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(473),
  [641] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [644] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [653] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [668] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [671] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [674] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [677] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section4, 1, 0, 1),
  [682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0),
  [684] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [687] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [690] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [693] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [696] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [699] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [708] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [711] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [717] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [720] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [723] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [744] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [747] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [753] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(473),
  [771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [774] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [777] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [780] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [783] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [786] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [789] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [792] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [798] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [801] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [804] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section4, 2, 0, 1),
  [812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section5, 2, 0, 1),
  [814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section5, 1, 0, 1),
  [816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0),
  [818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [821] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [845] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [848] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [860] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [863] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [866] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [869] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [875] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [878] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [881] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [884] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(474),
  [902] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [905] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [911] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section6, 2, 0, 1),
  [940] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(353),
  [943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0),
  [954] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [957] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [963] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [966] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [969] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [972] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [975] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [978] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(355),
  [981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(356),
  [984] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [987] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(416),
  [990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(417),
  [993] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(438),
  [999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(419),
  [1002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(440),
  [1005] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(441),
  [1008] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [1011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__section6, 1, 0, 1),
  [1013] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(352),
  [1016] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [1019] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [1022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(778),
  [1025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(357),
  [1028] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(354),
  [1031] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [1034] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [1037] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [1040] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(409),
  [1043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(450),
  [1046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(411),
  [1049] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(428),
  [1052] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [1055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(511),
  [1058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_minus, 1, 0, 0),
  [1060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_minus, 1, 0, 0),
  [1062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_dot, 1, 0, 0),
  [1064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_dot, 1, 0, 0),
  [1066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_plus_repeat1, 2, 0, 0),
  [1068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__list_plus_repeat1, 2, 0, 0),
  [1070] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_plus_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [1073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_minus_repeat1, 2, 0, 0),
  [1075] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__list_minus_repeat1, 2, 0, 0),
  [1077] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_minus_repeat1, 2, 0, 0), SHIFT_REPEAT(239),
  [1080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_star_repeat1, 2, 0, 0),
  [1082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__list_star_repeat1, 2, 0, 0),
  [1084] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_star_repeat1, 2, 0, 0), SHIFT_REPEAT(198),
  [1087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indented_code_block, 1, 0, 0),
  [1089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indented_code_block, 1, 0, 0),
  [1091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_plus, 1, 0, 0),
  [1093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_plus, 1, 0, 0),
  [1095] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_star, 1, 0, 0),
  [1097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_star, 1, 0, 0),
  [1099] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_parenthesis, 1, 0, 0),
  [1101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_parenthesis, 1, 0, 0),
  [1103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_dot_repeat1, 2, 0, 0),
  [1105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__list_dot_repeat1, 2, 0, 0),
  [1107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_dot_repeat1, 2, 0, 0), SHIFT_REPEAT(201),
  [1110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_parenthesis_repeat1, 2, 0, 0),
  [1112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__list_parenthesis_repeat1, 2, 0, 0),
  [1114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_parenthesis_repeat1, 2, 0, 0), SHIFT_REPEAT(199),
  [1117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0),
  [1119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0),
  [1121] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [1124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [1127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__newline, 1, 0, 0),
  [1129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__newline, 1, 0, 0),
  [1131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [1133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [1135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indented_code_block, 2, 0, 0),
  [1137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indented_code_block, 2, 0, 0),
  [1139] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0), SHIFT_REPEAT(427),
  [1142] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_indented_code_block_repeat1, 2, 0, 0), SHIFT_REPEAT(790),
  [1145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0),
  [1147] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 4, 10, 0),
  [1149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0), SHIFT(423),
  [1152] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0), SHIFT(424),
  [1155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0), SHIFT(452),
  [1158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0), SHIFT(733),
  [1161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [1163] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 4, 10, 0), SHIFT(736),
  [1166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [1168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 5, 10, 0),
  [1170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0),
  [1172] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0), SHIFT(423),
  [1175] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0), SHIFT(424),
  [1178] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0), SHIFT(452),
  [1181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0), SHIFT(738),
  [1184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [1186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 6, 10, 0),
  [1188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0),
  [1190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0), SHIFT(423),
  [1193] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0), SHIFT(424),
  [1196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0), SHIFT(452),
  [1199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0), SHIFT(740),
  [1202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [1204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 7, 10, 0),
  [1206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0),
  [1208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0), SHIFT(423),
  [1211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0), SHIFT(424),
  [1214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0), SHIFT(452),
  [1217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0), SHIFT(741),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [1222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 8, 10, 0),
  [1224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0),
  [1226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0), SHIFT(423),
  [1229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0), SHIFT(424),
  [1232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0), SHIFT(452),
  [1235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0), SHIFT(742),
  [1238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [1240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__soft_line_break, 1, 0, 0),
  [1242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__soft_line_break, 1, 0, 0),
  [1244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [1246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [1248] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 10, 0), SHIFT(728),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [1253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__newline, 2, 0, 0),
  [1255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__newline, 2, 0, 0),
  [1257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_paragraph, 2, 0, 5),
  [1259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_paragraph, 2, 0, 5),
  [1261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_not_section, 1, 0, 0),
  [1263] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_not_section, 1, 0, 0),
  [1265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [1269] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 5, 10, 0), SHIFT(704),
  [1272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [1274] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 10, 0), SHIFT(699),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [1279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 10, 0), SHIFT(710),
  [1282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [1284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [1286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [1288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_5, 3, 0, 0),
  [1290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_5, 3, 0, 0),
  [1292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [1294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_minus, 4, 0, 0),
  [1296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_minus, 4, 0, 0),
  [1298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [1300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_star, 4, 0, 0),
  [1302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_star, 4, 0, 0),
  [1304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [1306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_4, 2, 0, 0),
  [1308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_4, 2, 0, 0),
  [1310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [1312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_5, 2, 0, 0),
  [1314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_5, 2, 0, 0),
  [1316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_dot, 4, 0, 0),
  [1320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_dot, 4, 0, 0),
  [1322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [1324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_parenthesis, 4, 0, 0),
  [1326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_parenthesis, 4, 0, 0),
  [1328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [1330] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_quote, 3, 0, 0),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_quote, 3, 0, 0),
  [1334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [1336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indented_chunk, 3, 0, 0),
  [1338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indented_chunk, 3, 0, 0),
  [1340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [1342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_1, 3, 0, 0),
  [1344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_1, 3, 0, 0),
  [1346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_2, 3, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_2, 3, 0, 0),
  [1352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [1354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_3, 3, 0, 0),
  [1356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_3, 3, 0, 0),
  [1358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [1360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_4, 3, 0, 0),
  [1362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_4, 3, 0, 0),
  [1364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [1366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indented_chunk, 2, 0, 0),
  [1368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indented_chunk, 2, 0, 0),
  [1370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_6, 3, 0, 0),
  [1374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_6, 3, 0, 0),
  [1376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [1378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_7, 3, 0, 0),
  [1380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_7, 3, 0, 0),
  [1382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [1384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [1386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [1388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [1390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [1392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [1394] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_quote, 2, 0, 0),
  [1396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_quote, 2, 0, 0),
  [1398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [1400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_plus, 3, 0, 0),
  [1402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_plus, 3, 0, 0),
  [1404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [1406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_minus, 3, 0, 0),
  [1408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_minus, 3, 0, 0),
  [1410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [1412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_star, 3, 0, 0),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_star, 3, 0, 0),
  [1416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [1418] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_dot, 3, 0, 0),
  [1420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_dot, 3, 0, 0),
  [1422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [1424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_parenthesis, 3, 0, 0),
  [1426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_parenthesis, 3, 0, 0),
  [1428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [1430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_quote, 4, 0, 0),
  [1432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_quote, 4, 0, 0),
  [1434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [1436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [1438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_plus, 4, 0, 0),
  [1440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_plus, 4, 0, 0),
  [1442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [1444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [1446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [1448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [1450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [1452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [1454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_1, 2, 0, 0),
  [1456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_1, 2, 0, 0),
  [1458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [1460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [1462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_3, 2, 0, 0),
  [1464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_3, 2, 0, 0),
  [1466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [1468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [1470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__soft_line_break, 2, 0, 0),
  [1472] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__soft_line_break, 2, 0, 0),
  [1474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [1478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [1480] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_6, 2, 0, 0),
  [1482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_6, 2, 0, 0),
  [1484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [1486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_7, 2, 0, 0),
  [1488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_7, 2, 0, 0),
  [1490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [1492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [1494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [1496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [1498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [1500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [1502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [1504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_2, 2, 0, 0),
  [1506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_2, 2, 0, 0),
  [1508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [1510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [1512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [1514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [1516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [1518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [1520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [1522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [1524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [1532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [1534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_html_block, 2, 0, 0),
  [1536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_html_block, 2, 0, 0),
  [1538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 8, 0, 0),
  [1540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 8, 0, 0),
  [1542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 8, 20, 0),
  [1544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 8, 20, 0),
  [1546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 9, 0, 0),
  [1548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 9, 0, 0),
  [1550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 9, 20, 0),
  [1552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 9, 20, 0),
  [1554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 10, 20, 0),
  [1556] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 10, 20, 0),
  [1558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 11, 20, 0),
  [1560] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 11, 20, 0),
  [1562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 12, 20, 0),
  [1564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 12, 20, 0),
  [1566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 13, 20, 0),
  [1568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 13, 20, 0),
  [1570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_document_repeat1, 1, 0, 0),
  [1572] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 1, 0, 0),
  [1574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_section, 1, 0, 0),
  [1576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_section, 1, 0, 0),
  [1578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_marker_plus, 1, 0, 0),
  [1580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_marker_plus, 1, 0, 0),
  [1582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_marker_star, 1, 0, 0),
  [1584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_marker_star, 1, 0, 0),
  [1586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_marker_parenthesis, 1, 0, 0),
  [1588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_marker_parenthesis, 1, 0, 0),
  [1590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_marker_dot, 1, 0, 0),
  [1592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_marker_dot, 1, 0, 0),
  [1594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading1, 3, 0, 7),
  [1596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading1, 3, 0, 7),
  [1598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_not_section, 1, 0, 1),
  [1600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_not_section, 1, 0, 1),
  [1602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading2, 3, 0, 7),
  [1604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading2, 3, 0, 7),
  [1606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_html_block, 1, 0, 0),
  [1608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_html_block, 1, 0, 0),
  [1610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading3, 3, 0, 7),
  [1612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading3, 3, 0, 7),
  [1614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 1, 0, 0),
  [1616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 1, 0, 0),
  [1618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading4, 3, 0, 7),
  [1620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading4, 3, 0, 7),
  [1622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading5, 3, 0, 7),
  [1624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading5, 3, 0, 7),
  [1626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading6, 3, 0, 7),
  [1628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading6, 3, 0, 7),
  [1630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 3, 0, 0),
  [1632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 3, 0, 0),
  [1634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading1, 2, 0, 0),
  [1636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading1, 2, 0, 0),
  [1638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading2, 2, 0, 0),
  [1640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading2, 2, 0, 0),
  [1642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading3, 2, 0, 0),
  [1644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading3, 2, 0, 0),
  [1646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading4, 2, 0, 0),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading4, 2, 0, 0),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading5, 2, 0, 0),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading5, 2, 0, 0),
  [1654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__atx_heading6, 2, 0, 0),
  [1656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading6, 2, 0, 0),
  [1658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_thematic_break, 2, 0, 0),
  [1660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_thematic_break, 2, 0, 0),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__blank_line, 2, 0, 0),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__blank_line, 2, 0, 0),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__section1_repeat1, 1, 0, 2),
  [1668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section1_repeat1, 1, 0, 2),
  [1670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__section2_repeat1, 1, 0, 2),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section2_repeat1, 1, 0, 2),
  [1674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__section3_repeat1, 1, 0, 2),
  [1676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section3_repeat1, 1, 0, 2),
  [1678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__section4_repeat1, 1, 0, 2),
  [1680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section4_repeat1, 1, 0, 2),
  [1682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__section5_repeat1, 1, 0, 2),
  [1684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__section5_repeat1, 1, 0, 2),
  [1686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_marker_minus, 1, 0, 0),
  [1688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_marker_minus, 1, 0, 0),
  [1690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__setext_heading1, 3, 0, 9),
  [1692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__setext_heading1, 3, 0, 9),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__setext_heading2, 3, 0, 9),
  [1696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__setext_heading2, 3, 0, 9),
  [1698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indented_chunk, 4, 0, 0),
  [1700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indented_chunk, 4, 0, 0),
  [1702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 4, 0, 0),
  [1704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 4, 0, 0),
  [1706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 7, 20, 0),
  [1708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 7, 20, 0),
  [1710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_2, 4, 0, 0),
  [1712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_2, 4, 0, 0),
  [1714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_3, 4, 0, 0),
  [1716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_3, 4, 0, 0),
  [1718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_4, 4, 0, 0),
  [1720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_4, 4, 0, 0),
  [1722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_5, 4, 0, 0),
  [1724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_5, 4, 0, 0),
  [1726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_6, 4, 0, 0),
  [1728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_6, 4, 0, 0),
  [1730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_1, 4, 0, 0),
  [1732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_1, 4, 0, 0),
  [1734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__html_block_7, 4, 0, 0),
  [1736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__html_block_7, 4, 0, 0),
  [1738] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_quote, 5, 0, 0),
  [1740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_quote, 5, 0, 0),
  [1742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 5, 0, 0),
  [1744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 5, 0, 0),
  [1746] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pipe_table, 5, 0, 11),
  [1748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table, 5, 0, 11),
  [1750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_plus, 5, 0, 0),
  [1752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_plus, 5, 0, 0),
  [1754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_minus, 5, 0, 0),
  [1756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_minus, 5, 0, 0),
  [1758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_star, 5, 0, 0),
  [1760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_star, 5, 0, 0),
  [1762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_dot, 5, 0, 0),
  [1764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_dot, 5, 0, 0),
  [1766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__list_item_parenthesis, 5, 0, 0),
  [1768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_item_parenthesis, 5, 0, 0),
  [1770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 6, 0, 0),
  [1772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 6, 0, 0),
  [1774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pipe_table, 6, 0, 11),
  [1776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table, 6, 0, 11),
  [1778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_link_reference_definition, 6, 20, 0),
  [1780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_reference_definition, 6, 20, 0),
  [1782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fenced_code_block, 7, 0, 0),
  [1784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fenced_code_block, 7, 0, 0),
  [1786] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0), SHIFT(432),
  [1789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0),
  [1791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [1793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [1795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0),
  [1797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0), SHIFT(404),
  [1800] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0), SHIFT(404),
  [1803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 1, 0, 0), SHIFT(502),
  [1806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [1808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [1810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(449),
  [1812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [1814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [1816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [1818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [1820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [1822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [1824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [1826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [1828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [1830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [1832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [1834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 1, 10, 0),
  [1836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [1838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [1840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [1842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [1844] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(497),
  [1847] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(360),
  [1850] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(360),
  [1853] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(480),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0),
  [1858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [1860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [1862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 2, 10, 0),
  [1864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [1866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [1868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [1870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [1872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [1874] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_destination, 3, 10, 0),
  [1876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [1878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [1880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [1882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [1884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [1886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [1888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [1890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [1892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [1894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [1896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [1898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(466),
  [1900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [1902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_info_string, 2, 0, 0),
  [1904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [1906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [1908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [1910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_info_string, 1, 0, 0),
  [1912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [1914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [1916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [1918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [1920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [1922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_info_string_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [1925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_info_string_repeat1, 2, 0, 0), SHIFT_REPEAT(380),
  [1928] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_info_string_repeat1, 2, 0, 0), SHIFT_REPEAT(466),
  [1931] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_info_string_repeat1, 2, 0, 0), SHIFT_REPEAT(466),
  [1934] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_info_string_repeat1, 2, 0, 0),
  [1936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_info_string, 3, 0, 0),
  [1938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [1940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [1942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [1944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [1946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [1948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [1950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [1952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [1954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [1956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [1958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [1960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [1962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [1964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [1966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [1968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [1970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [1972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [1974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(539),
  [1977] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [1980] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [1983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [1985] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0), SHIFT_REPEAT(502),
  [1988] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(539),
  [1991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(394),
  [1994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(394),
  [1997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [1999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0), SHIFT_REPEAT(502),
  [2002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(518),
  [2004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [2006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [2008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [2010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [2012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_code_fence_content, 1, 0, 0),
  [2014] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0), SHIFT_REPEAT(509),
  [2017] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0), SHIFT_REPEAT(509),
  [2020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0), SHIFT_REPEAT(546),
  [2023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0),
  [2025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0), SHIFT_REPEAT(895),
  [2028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [2030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [2032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [2034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [2036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [2038] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0), SHIFT_REPEAT(513),
  [2041] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0), SHIFT_REPEAT(513),
  [2044] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0), SHIFT_REPEAT(903),
  [2047] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0), SHIFT_REPEAT(547),
  [2050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0),
  [2052] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0), SHIFT_REPEAT(518),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0), SHIFT_REPEAT(518),
  [2058] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0), SHIFT_REPEAT(900),
  [2061] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0), SHIFT_REPEAT(548),
  [2064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0),
  [2066] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0), SHIFT_REPEAT(520),
  [2069] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0), SHIFT_REPEAT(520),
  [2072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0), SHIFT_REPEAT(867),
  [2075] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0), SHIFT_REPEAT(550),
  [2078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0),
  [2080] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(539),
  [2083] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(402),
  [2086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0),
  [2088] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(402),
  [2091] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_label_repeat1, 2, 0, 0), SHIFT_REPEAT(502),
  [2094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [2096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [2100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [2102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [2104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(509),
  [2106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [2108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [2110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [2112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [2114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [2116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [2118] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(499),
  [2121] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(499),
  [2124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(538),
  [2127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0),
  [2129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [2131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(501),
  [2134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(501),
  [2137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_code_fence_content_repeat1, 2, 0, 0), SHIFT_REPEAT(544),
  [2140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_paragraph_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [2143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_paragraph_repeat1, 2, 0, 0), SHIFT_REPEAT(506),
  [2146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_paragraph_repeat1, 2, 0, 0),
  [2148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_paragraph_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [2151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [2153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [2155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [2157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(431),
  [2159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [2161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [2163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [2165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_repeat1, 1, 0, 0),
  [2167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [2169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [2171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [2173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [2177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [2181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [2183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [2185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [2189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [2191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [2193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [2195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [2197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [2199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [2201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [2203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [2207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [2209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [2211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [2215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_cell, 1, 0, 0),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [2231] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2234] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2237] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0), SHIFT_REPEAT(556),
  [2240] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0), SHIFT_REPEAT(588),
  [2243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0),
  [2245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [2249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [2251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [2253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(446),
  [2255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__html_block_6_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_6_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__html_block_6_repeat1, 2, 0, 0), SHIFT_REPEAT(553),
  [2264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_6_repeat1, 2, 0, 0),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [2268] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(539),
  [2271] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(446),
  [2274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(446),
  [2277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [2279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0), SHIFT_REPEAT(502),
  [2282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [2284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [2286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [2288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(451),
  [2290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_language, 1, 0, 0),
  [2292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [2294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_language_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [2297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_language_repeat1, 2, 0, 0), SHIFT_REPEAT(451),
  [2300] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_language_repeat1, 2, 0, 0), SHIFT_REPEAT(451),
  [2303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_language_repeat1, 2, 0, 0),
  [2305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [2307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(443),
  [2309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__indented_chunk_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__indented_chunk_repeat1, 2, 0, 0), SHIFT_REPEAT(532),
  [2315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__indented_chunk_repeat1, 2, 0, 0), SHIFT_REPEAT(588),
  [2318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__indented_chunk_repeat1, 2, 0, 0),
  [2320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(583),
  [2324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [2326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [2328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [2330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [2332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(464),
  [2334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_cell, 2, 0, 0),
  [2336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [2338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(465),
  [2340] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(464),
  [2343] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(464),
  [2346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0),
  [2348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_cell, 3, 0, 0),
  [2350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__line, 1, 0, 0),
  [2352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(467),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [2356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0),
  [2358] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(467),
  [2361] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(467),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [2366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 1, 0, 0),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [2370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [2372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [2374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [2378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 2, 0, 0),
  [2380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [2382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 8),
  [2384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0), SHIFT_REPEAT(579),
  [2387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0), SHIFT_REPEAT(545),
  [2390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0), SHIFT_REPEAT(545),
  [2393] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0), SHIFT_REPEAT(614),
  [2396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0),
  [2398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [2400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [2402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [2404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [2406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [2408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [2410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [2412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 3, 0, 0),
  [2414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [2418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(486),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [2424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(491),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [2428] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(585),
  [2431] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(487),
  [2434] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0), SHIFT_REPEAT(487),
  [2437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat1, 2, 0, 0),
  [2439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [2441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [2443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [2445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [2447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(585),
  [2450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(491),
  [2453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(491),
  [2456] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0), SHIFT_REPEAT(485),
  [2459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [2461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [2463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_destination_repeat2, 2, 0, 0),
  [2465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_backslash_escape, 1, 0, 0),
  [2467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_backslash_escape, 1, 0, 0),
  [2469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [2471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [2473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [2475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [2478] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [2481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [2483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [2485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [2487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pipe_table_newline, 1, 0, 0),
  [2489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pipe_table_newline, 1, 0, 0),
  [2491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [2493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(505),
  [2496] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(505),
  [2499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(507),
  [2501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [2503] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(507),
  [2506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(507),
  [2509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(536),
  [2511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [2513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [2515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [2517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(508),
  [2521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [2523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [2525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [2527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(515),
  [2529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [2531] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_destination_parenthesis, 3, 0, 0),
  [2533] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__link_destination_parenthesis, 3, 0, 0),
  [2535] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(515),
  [2538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(515),
  [2541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0),
  [2543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__link_destination_parenthesis, 2, 0, 0),
  [2545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 4, 0, 0),
  [2547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(519),
  [2549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2551] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(519),
  [2554] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(519),
  [2557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [2559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [2561] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(522),
  [2564] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(522),
  [2567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [2569] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0),
  [2571] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 1, 0, 0), SHIFT_REPEAT(819),
  [2574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [2576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0),
  [2578] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 1, 0, 0), SHIFT_REPEAT(819),
  [2581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(529),
  [2584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(529),
  [2587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [2589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0),
  [2591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 1, 0, 0), SHIFT_REPEAT(819),
  [2594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 2, 0, 0),
  [2596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(558),
  [2598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [2600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [2602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [2604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [2610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(564),
  [2612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [2616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [2622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [2630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0), SHIFT_REPEAT(620),
  [2633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [2635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0),
  [2637] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat1, 3, 0, 0),
  [2639] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 3, 0, 0),
  [2641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [2643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat2, 3, 0, 0),
  [2645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat2, 3, 0, 0),
  [2647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [2649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [2651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_info_string_repeat2, 2, 0, 0),
  [2653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_info_string_repeat2, 2, 0, 0),
  [2655] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_info_string_repeat2, 2, 0, 0), SHIFT_REPEAT(557),
  [2658] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(558),
  [2661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(558),
  [2664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_language_repeat1, 2, 0, 0),
  [2666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_6_repeat1, 1, 0, 0),
  [2668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_6_repeat1, 1, 0, 0),
  [2670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat1, 2, 0, 0),
  [2672] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(563),
  [2675] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(563),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [2680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pipe_table_newline, 2, 0, 0),
  [2682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pipe_table_newline, 2, 0, 0),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [2686] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(570),
  [2689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__line_repeat1, 2, 0, 0), SHIFT_REPEAT(570),
  [2692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_inline_no_link, 2, 0, 0),
  [2694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_inline_no_link, 2, 0, 0),
  [2696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 4, 0, 0),
  [2698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipe_table_row_repeat1, 4, 0, 0),
  [2700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 2, 0, 0),
  [2702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [2704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_link_title_repeat3, 3, 0, 0),
  [2706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_link_title_repeat3, 3, 0, 0),
  [2708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [2710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [2712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_row_repeat1, 3, 0, 0),
  [2714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipe_table_row_repeat1, 3, 0, 0),
  [2716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [2718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_1_repeat1, 2, 0, 0),
  [2720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_2_repeat1, 2, 0, 0),
  [2722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [2724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(606),
  [2726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [2728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [2730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [2732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [2734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 0),
  [2736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_3_repeat1, 2, 0, 0),
  [2738] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__word, 1, 0, 0),
  [2740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__word, 1, 0, 0),
  [2742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_task_list_marker_checked, 1, 0, 0),
  [2744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_task_list_marker_unchecked, 1, 0, 0),
  [2746] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_paragraph_repeat1, 1, 0, 0),
  [2748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_paragraph_repeat1, 1, 0, 0),
  [2750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_5_repeat1, 2, 0, 0),
  [2752] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(606),
  [2755] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pipe_table_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(606),
  [2758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pipe_table_row_repeat1, 2, 0, 8),
  [2760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [2762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [2764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(545),
  [2766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [2768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [2770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_6_repeat1, 3, 0, 0),
  [2772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__html_block_6_repeat1, 3, 0, 0),
  [2774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__text_no_angle, 2, 0, 0),
  [2776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__text_no_angle, 2, 0, 0),
  [2778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__html_block_4_repeat1, 2, 0, 0),
  [2780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 2, 0, 0),
  [2782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 3, 0, 4),
  [2784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 2, 0, 2),
  [2786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0),
  [2788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(459),
  [2791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(490),
  [2794] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(468),
  [2797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(492),
  [2800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(461),
  [2803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(460),
  [2806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [2808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [2810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [2812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 2, 0, 0),
  [2814] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 2, 0, 0), SHIFT_REPEAT(716),
  [2817] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 2, 0, 0), SHIFT_REPEAT(827),
  [2820] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 2, 0, 0), SHIFT_REPEAT(752),
  [2823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 2, 0, 0),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [2827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 1, 0, 0),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [2831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 3, 0, 0),
  [2833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(644),
  [2836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_cell_repeat1, 2, 0, 0),
  [2838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [2840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [2842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_cell, 1, 0, 0),
  [2844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [2846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [2848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [2850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [2852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [2854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_cell, 2, 0, 10),
  [2856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [2858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [2860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [2862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [2864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 4, 0, 0),
  [2866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [2868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [2870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [2874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [2876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [2878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [2880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [2882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [2884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [2886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [2888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [2890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [2892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [2894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [2898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [2900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [2904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [2906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [2908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 3, 0, 0),
  [2910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [2912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [2914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [2918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [2920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [2922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [2926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [2928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [2930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [2932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [2934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [2938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [2944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [2948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [2950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_row_repeat1, 4, 0, 0),
  [2952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_cell, 2, 0, 12),
  [2954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pipe_table_repeat1, 2, 0, 0),
  [2956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_repeat1, 2, 0, 0), SHIFT_REPEAT(503),
  [2959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [2963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [2967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [2969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [2971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [2977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [2979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [2981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [2983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [2985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [2987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 5, 0, 0),
  [2989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [2991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [2995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [2997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [2999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [3001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [3003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pipe_table_delimiter_cell_repeat1, 2, 0, 0), SHIFT_REPEAT(714),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [3012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [3016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [3020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [3022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [3024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [3028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [3034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [3036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [3038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [3040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [3042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [3044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 5, 0, 0),
  [3046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_cell, 3, 0, 13),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [3058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_row, 6, 0, 0),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipe_table_delimiter_row, 6, 0, 0),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [3088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_title, 3, 0, 0),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [3102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_title, 2, 0, 0),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [3110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [3112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [3124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [3126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [3128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [3130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [3136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [3138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [3142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [3144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [3146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [3148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [3150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [3152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [3154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [3156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [3158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_link_label, 3, 0, 0),
  [3160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [3164] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [3166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [3172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [3174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [3176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [3178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [3180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [3182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [3186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [3188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [3192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [3196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [3198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [3200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [3202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [3204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [3206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [3208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [3210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [3212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [3214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [3216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [3218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [3220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [3222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [3224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading_content, 2, 0, 6),
  [3226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__atx_heading_content, 1, 0, 3),
  [3228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [3230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [3232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [3234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [3236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [3238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [3240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [3242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [3244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [3246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__line_ending = 0,
  ts_external_token__soft_line_ending = 1,
  ts_external_token__block_close = 2,
  ts_external_token_block_continuation = 3,
  ts_external_token__block_quote_start = 4,
  ts_external_token__indented_chunk_start = 5,
  ts_external_token_atx_h1_marker = 6,
  ts_external_token_atx_h2_marker = 7,
  ts_external_token_atx_h3_marker = 8,
  ts_external_token_atx_h4_marker = 9,
  ts_external_token_atx_h5_marker = 10,
  ts_external_token_atx_h6_marker = 11,
  ts_external_token_setext_h1_underline = 12,
  ts_external_token_setext_h2_underline = 13,
  ts_external_token__thematic_break = 14,
  ts_external_token__list_marker_minus = 15,
  ts_external_token__list_marker_plus = 16,
  ts_external_token__list_marker_star = 17,
  ts_external_token__list_marker_parenthesis = 18,
  ts_external_token__list_marker_dot = 19,
  ts_external_token__list_marker_minus_dont_interrupt = 20,
  ts_external_token__list_marker_plus_dont_interrupt = 21,
  ts_external_token__list_marker_star_dont_interrupt = 22,
  ts_external_token__list_marker_parenthesis_dont_interrupt = 23,
  ts_external_token__list_marker_dot_dont_interrupt = 24,
  ts_external_token__fenced_code_block_start_backtick = 25,
  ts_external_token__fenced_code_block_start_tilde = 26,
  ts_external_token__blank_line_start = 27,
  ts_external_token__fenced_code_block_end_backtick = 28,
  ts_external_token__fenced_code_block_end_tilde = 29,
  ts_external_token__html_block_1_start = 30,
  ts_external_token__html_block_1_end = 31,
  ts_external_token__html_block_2_start = 32,
  ts_external_token__html_block_3_start = 33,
  ts_external_token__html_block_4_start = 34,
  ts_external_token__html_block_5_start = 35,
  ts_external_token__html_block_6_start = 36,
  ts_external_token__html_block_7_start = 37,
  ts_external_token__close_block = 38,
  ts_external_token__no_indented_chunk = 39,
  ts_external_token__error = 40,
  ts_external_token__trigger_error = 41,
  ts_external_token__eof = 42,
  ts_external_token_minus_metadata = 43,
  ts_external_token_plus_metadata = 44,
  ts_external_token__pipe_table_start = 45,
  ts_external_token__pipe_table_line_ending = 46,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__line_ending] = sym__line_ending,
  [ts_external_token__soft_line_ending] = sym__soft_line_ending,
  [ts_external_token__block_close] = sym__block_close,
  [ts_external_token_block_continuation] = sym_block_continuation,
  [ts_external_token__block_quote_start] = sym__block_quote_start,
  [ts_external_token__indented_chunk_start] = sym__indented_chunk_start,
  [ts_external_token_atx_h1_marker] = sym_atx_h1_marker,
  [ts_external_token_atx_h2_marker] = sym_atx_h2_marker,
  [ts_external_token_atx_h3_marker] = sym_atx_h3_marker,
  [ts_external_token_atx_h4_marker] = sym_atx_h4_marker,
  [ts_external_token_atx_h5_marker] = sym_atx_h5_marker,
  [ts_external_token_atx_h6_marker] = sym_atx_h6_marker,
  [ts_external_token_setext_h1_underline] = sym_setext_h1_underline,
  [ts_external_token_setext_h2_underline] = sym_setext_h2_underline,
  [ts_external_token__thematic_break] = sym__thematic_break,
  [ts_external_token__list_marker_minus] = sym__list_marker_minus,
  [ts_external_token__list_marker_plus] = sym__list_marker_plus,
  [ts_external_token__list_marker_star] = sym__list_marker_star,
  [ts_external_token__list_marker_parenthesis] = sym__list_marker_parenthesis,
  [ts_external_token__list_marker_dot] = sym__list_marker_dot,
  [ts_external_token__list_marker_minus_dont_interrupt] = sym__list_marker_minus_dont_interrupt,
  [ts_external_token__list_marker_plus_dont_interrupt] = sym__list_marker_plus_dont_interrupt,
  [ts_external_token__list_marker_star_dont_interrupt] = sym__list_marker_star_dont_interrupt,
  [ts_external_token__list_marker_parenthesis_dont_interrupt] = sym__list_marker_parenthesis_dont_interrupt,
  [ts_external_token__list_marker_dot_dont_interrupt] = sym__list_marker_dot_dont_interrupt,
  [ts_external_token__fenced_code_block_start_backtick] = sym__fenced_code_block_start_backtick,
  [ts_external_token__fenced_code_block_start_tilde] = sym__fenced_code_block_start_tilde,
  [ts_external_token__blank_line_start] = sym__blank_line_start,
  [ts_external_token__fenced_code_block_end_backtick] = sym__fenced_code_block_end_backtick,
  [ts_external_token__fenced_code_block_end_tilde] = sym__fenced_code_block_end_tilde,
  [ts_external_token__html_block_1_start] = sym__html_block_1_start,
  [ts_external_token__html_block_1_end] = sym__html_block_1_end,
  [ts_external_token__html_block_2_start] = sym__html_block_2_start,
  [ts_external_token__html_block_3_start] = sym__html_block_3_start,
  [ts_external_token__html_block_4_start] = sym__html_block_4_start,
  [ts_external_token__html_block_5_start] = sym__html_block_5_start,
  [ts_external_token__html_block_6_start] = sym__html_block_6_start,
  [ts_external_token__html_block_7_start] = sym__html_block_7_start,
  [ts_external_token__close_block] = sym__close_block,
  [ts_external_token__no_indented_chunk] = sym__no_indented_chunk,
  [ts_external_token__error] = sym__error,
  [ts_external_token__trigger_error] = sym__trigger_error,
  [ts_external_token__eof] = sym__eof,
  [ts_external_token_minus_metadata] = sym_minus_metadata,
  [ts_external_token_plus_metadata] = sym_plus_metadata,
  [ts_external_token__pipe_table_start] = sym__pipe_table_start,
  [ts_external_token__pipe_table_line_ending] = sym__pipe_table_line_ending,
};

static const bool ts_external_scanner_states[48][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token_setext_h1_underline] = true,
    [ts_external_token_setext_h2_underline] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__fenced_code_block_end_backtick] = true,
    [ts_external_token__fenced_code_block_end_tilde] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_1_end] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__close_block] = true,
    [ts_external_token__no_indented_chunk] = true,
    [ts_external_token__error] = true,
    [ts_external_token__trigger_error] = true,
    [ts_external_token__eof] = true,
    [ts_external_token_minus_metadata] = true,
    [ts_external_token_plus_metadata] = true,
    [ts_external_token__pipe_table_start] = true,
    [ts_external_token__pipe_table_line_ending] = true,
  },
  [2] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token_minus_metadata] = true,
    [ts_external_token_plus_metadata] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [3] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [4] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [5] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [6] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [7] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token_setext_h1_underline] = true,
    [ts_external_token_setext_h2_underline] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [8] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token_setext_h1_underline] = true,
    [ts_external_token_setext_h2_underline] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [9] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__no_indented_chunk] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [10] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__no_indented_chunk] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [11] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__no_indented_chunk] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [12] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__no_indented_chunk] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [13] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token_setext_h1_underline] = true,
    [ts_external_token_setext_h2_underline] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [14] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__block_quote_start] = true,
    [ts_external_token__indented_chunk_start] = true,
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
    [ts_external_token_setext_h1_underline] = true,
    [ts_external_token_setext_h2_underline] = true,
    [ts_external_token__thematic_break] = true,
    [ts_external_token__list_marker_minus] = true,
    [ts_external_token__list_marker_plus] = true,
    [ts_external_token__list_marker_star] = true,
    [ts_external_token__list_marker_parenthesis] = true,
    [ts_external_token__list_marker_dot] = true,
    [ts_external_token__list_marker_minus_dont_interrupt] = true,
    [ts_external_token__list_marker_plus_dont_interrupt] = true,
    [ts_external_token__list_marker_star_dont_interrupt] = true,
    [ts_external_token__list_marker_parenthesis_dont_interrupt] = true,
    [ts_external_token__list_marker_dot_dont_interrupt] = true,
    [ts_external_token__fenced_code_block_start_backtick] = true,
    [ts_external_token__fenced_code_block_start_tilde] = true,
    [ts_external_token__blank_line_start] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__pipe_table_start] = true,
  },
  [15] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__html_block_1_start] = true,
    [ts_external_token__html_block_2_start] = true,
    [ts_external_token__html_block_3_start] = true,
    [ts_external_token__html_block_4_start] = true,
    [ts_external_token__html_block_5_start] = true,
    [ts_external_token__html_block_6_start] = true,
    [ts_external_token__html_block_7_start] = true,
    [ts_external_token__eof] = true,
  },
  [16] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__eof] = true,
  },
  [17] = {
    [ts_external_token__line_ending] = true,
  },
  [18] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__fenced_code_block_end_tilde] = true,
  },
  [19] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__fenced_code_block_end_backtick] = true,
  },
  [20] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
  },
  [21] = {
    [ts_external_token__soft_line_ending] = true,
  },
  [22] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__html_block_1_end] = true,
  },
  [23] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__eof] = true,
    [ts_external_token__pipe_table_line_ending] = true,
  },
  [24] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token_block_continuation] = true,
  },
  [25] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__eof] = true,
    [ts_external_token__pipe_table_line_ending] = true,
  },
  [26] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__eof] = true,
  },
  [27] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__fenced_code_block_end_backtick] = true,
  },
  [28] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__fenced_code_block_end_tilde] = true,
  },
  [29] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__html_block_1_end] = true,
  },
  [30] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
  },
  [31] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
    [ts_external_token__blank_line_start] = true,
  },
  [32] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__close_block] = true,
  },
  [33] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__block_close] = true,
    [ts_external_token__blank_line_start] = true,
  },
  [34] = {
    [ts_external_token_block_continuation] = true,
  },
  [35] = {
    [ts_external_token_atx_h1_marker] = true,
    [ts_external_token_atx_h2_marker] = true,
    [ts_external_token_atx_h3_marker] = true,
    [ts_external_token_atx_h4_marker] = true,
    [ts_external_token_atx_h5_marker] = true,
    [ts_external_token_atx_h6_marker] = true,
  },
  [36] = {
    [ts_external_token__soft_line_ending] = true,
    [ts_external_token__no_indented_chunk] = true,
  },
  [37] = {
    [ts_external_token__no_indented_chunk] = true,
  },
  [38] = {
    [ts_external_token_block_continuation] = true,
    [ts_external_token__no_indented_chunk] = true,
  },
  [39] = {
    [ts_external_token__line_ending] = true,
    [ts_external_token__eof] = true,
  },
  [40] = {
    [ts_external_token__block_close] = true,
    [ts_external_token__fenced_code_block_end_backtick] = true,
  },
  [41] = {
    [ts_external_token__block_close] = true,
    [ts_external_token__fenced_code_block_end_tilde] = true,
  },
  [42] = {
    [ts_external_token_block_continuation] = true,
    [ts_external_token__trigger_error] = true,
  },
  [43] = {
    [ts_external_token_block_continuation] = true,
    [ts_external_token__close_block] = true,
  },
  [44] = {
    [ts_external_token__block_close] = true,
    [ts_external_token_block_continuation] = true,
  },
  [45] = {
    [ts_external_token__close_block] = true,
  },
  [46] = {
    [ts_external_token__block_close] = true,
  },
  [47] = {
    [ts_external_token__trigger_error] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_markdown_external_scanner_create(void);
void tree_sitter_markdown_external_scanner_destroy(void *);
bool tree_sitter_markdown_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_markdown_external_scanner_serialize(void *, char *);
void tree_sitter_markdown_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_markdown(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_markdown_external_scanner_create,
      tree_sitter_markdown_external_scanner_destroy,
      tree_sitter_markdown_external_scanner_scan,
      tree_sitter_markdown_external_scanner_serialize,
      tree_sitter_markdown_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "markdown",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 5,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
